/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: base_utils.cc,v 1.8 2009/01/01 07:46:43 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/base_utils.h"

#include "libochusha/misc_utils.h"

#include <cassert>
#include <cstdlib>
#include <cstring>


using namespace Ochusha;

namespace Ochusha
{


const char *Base64Encoder::base64 =
  "ABCDEFGHIJKLMNOP"
  "QRSTUVWXYZabcdef"
  "ghijklmnopqrstuv"
  "wxyz0123456789+/";


const char *Base64Encoder::base64url =
  "ABCDEFGHIJKLMNOP"
  "QRSTUVWXYZabcdef"
  "ghijklmnopqrstuv"
  "wxyz0123456789-_";


Base64Encoder::Base64Encoder(size_t buffer_len, const char *alphabet)
  : Buffer(buffer_len),
    alphabet(alphabet),
    result_buffer(NULL),
    pending_length(0)
{
  assert(alphabet != NULL);
}


Base64Encoder::Base64Encoder(int fd, bool gzip, const char *alphabet)
  : Buffer(fd, gzip),
    alphabet(alphabet),
    result_buffer(NULL),
    pending_length(0)
{
  assert(alphabet != NULL);
}


Base64Encoder::Base64Encoder(Buffer *buffer, const char *alphabet)
  : Buffer(static_cast<size_t>(0)),
    alphabet(alphabet),
    result_buffer(buffer),
    pending_length(0)
{
  assert(alphabet != NULL);
  assert(buffer != NULL);
}


Base64Encoder::~Base64Encoder()
{
}


char *
Base64Encoder::encode_text_to_base64(const char *text)
{
  static Base64Encoder encoder(static_cast<size_t>(4096), base64);
  static Lock lock("Base64Encoder");

  LockBlock lock_obj(lock);
  encoder.clear(0);
  if (encoder.append_data(text, strlen(text), 0) && encoder.fix(0))
    return strndup(encoder.get_buffer(0), encoder.get_length(0));
  return NULL;
}


char *
Base64Encoder::encode_text_to_base64url(const char *text)
{
  static Base64Encoder encoder(static_cast<size_t>(4096), base64url);
  static Lock lock("Base64URLEncoder");

  LockBlock lock_obj(lock);
  encoder.clear(0);
  if (encoder.append_data(text, strlen(text), 0) && encoder.fix(0))
    return strndup(encoder.get_buffer(0), encoder.get_length(0));
  return NULL;
}


bool
Base64Encoder::append_result(const char *data, size_t length, int lock_id)
{
  if (result_buffer != NULL)
    return result_buffer->append_data(data, length, 0);

  return Buffer::append_data(data, length, lock_id);
}


#define BASE64_PAD '='


bool
Base64Encoder::fix(int lock_id)
{
  assert(pending_length >= 0 && pending_length < 3);

  // Ⱦüʬν
  bool result;
  char trailing[4];
  if (pending_length == 1)
    {
      trailing[0] = alphabet[pending_buffer[0] >> 2];
      trailing[1] = alphabet[(pending_buffer[0] << 4) & 0x3f];
      trailing[2] = BASE64_PAD;
      trailing[3] = BASE64_PAD;
      result = append_result(trailing, 4, lock_id);
    }
  else if (pending_length == 2)
    {
      trailing[0] = alphabet[pending_buffer[0] >> 2];
      trailing[1] = alphabet[(pending_buffer[0] << 4 | pending_buffer[1] >> 4)
			     & 0x3f];
      trailing[2] = alphabet[(pending_buffer[1] << 2) & 0x3f];
      trailing[3] = BASE64_PAD;
      result = append_result(trailing, 4, lock_id);
    }
  else
    result = true;

  Buffer::fix();

  return result;
}


bool
Base64Encoder::clear(int lock_id)
{
  pending_length = 0;
  bool result = true;
  if (result_buffer != NULL)
    result = result_buffer->clear(0);

  result &= Buffer::clear(lock_id);

  return result;
}


bool
Base64Encoder::append_data(const char *data, size_t length, int lock_id)
{
  assert(data != NULL && length >= 0);

  if (length == 0)
    return true;

  if (pending_length > 0)
    {
      while (length > 0 && pending_length < 3)
	{
	  pending_buffer[pending_length++] = *data++;
	  --length;
	}

      if (pending_length == 3)
	{
	  pending_length = 0;
	  if (!append_data(pending_buffer, 3, lock_id))
	    return false;
	}
    }

  if (pending_length > 0)
    {
      assert(length == 0);
      return true;
    }

  size_t full_groups = length / 3;
  size_t partial_length = full_groups * 4;
  const char *tail = data + full_groups * 3;

  char *tmp_buf = static_cast<char *>(malloc(partial_length));
  if (tmp_buf == NULL)
    return false;	// out of memory

  char *dst = tmp_buf;
  while (data < tail)
    {
      *dst++ = alphabet[data[0] >> 2];
      *dst++ = alphabet[(data[0] << 4 | data[1] >> 4) & 0x3f];
      *dst++ = alphabet[(data[1] << 2 | data[2] >> 6) & 0x3f];
      *dst++ = alphabet[data[2] & 0x3f];
      data += 3;
    }
  assert(data == tail);

  bool result = append_result(tmp_buf, partial_length, lock_id);
  free(tmp_buf);

  if (!result)
    return false;

  size_t trailing = length % 3;
  assert(pending_length == 0);
  while (trailing-- > 0)
    pending_buffer[pending_length++] = *data++;

  assert(pending_length < 3);

  return true;
}


#define IV 0x80
#define PD 0x40


const unsigned char Base64Decoder::base64[] = {
  IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV,
  IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV,
  IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, 62, IV, IV, IV, 63,
  52, 53, 54, 55, 56, 57, 58, 59, 60, 61, IV, IV, IV, PD, IV, IV,
  IV,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
  15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, IV, IV, IV, IV, IV,
  IV, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
  41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, IV, IV, IV, IV, IV,
};

const unsigned char Base64Decoder::base64url[] = {
  IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV,
  IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV,
  IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, IV, 62, IV, IV,
  52, 53, 54, 55, 56, 57, 58, 59, 60, 61, IV, IV, IV, PD, IV, IV,
  IV,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
  15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, IV, IV, IV, IV, 63,
  IV, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
  41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, IV, IV, IV, IV, IV,
};


Base64Decoder::Base64Decoder(size_t buffer_len, const unsigned char *dictionary)
  : Buffer(buffer_len),
    dictionary(dictionary),
    result_buffer(NULL),
    pending_length(0),
    ignore_garbage(false)
{
  assert(dictionary != NULL);
}


Base64Decoder::Base64Decoder(int fd, bool gzip, const unsigned char *dictionary)
  : Buffer(fd, gzip),
    dictionary(dictionary),
    result_buffer(NULL),
    pending_length(0),
    ignore_garbage(false)
{
  assert(dictionary != NULL);
}


Base64Decoder::Base64Decoder(Buffer *buffer, const unsigned char *dictionary)
  : Buffer(static_cast<size_t>(0)),
    dictionary(dictionary),
    result_buffer(buffer),
    pending_length(0),
    ignore_garbage(false)
{
  assert(dictionary != NULL);
  assert(buffer != NULL);
}


Base64Decoder::~Base64Decoder()
{
}


char *
Base64Decoder::decode_text_from_base64(const char *text)
{
  static Base64Decoder decoder(static_cast<size_t>(4096), base64);
  static Lock lock("Base64Decoder");

  LockBlock lock_obj(lock);
  decoder.clear(0);
  if (decoder.append_data(text, strlen(text), 0) && decoder.fix(0))
    return strndup(decoder.get_buffer(0), decoder.get_length(0));
  return NULL;
}


char *
Base64Decoder::decode_text_from_base64url(const char *text)
{
  static Base64Decoder decoder(static_cast<size_t>(4096), base64url);
  static Lock lock("Base64URLDecoder");

  LockBlock lock_obj(lock);
  decoder.clear(0);
  if (decoder.append_data(text, strlen(text), 0) && decoder.fix(0))
    return strndup(decoder.get_buffer(0), decoder.get_length(0));
  return NULL;
}


bool
Base64Decoder::append_result(const char *data, size_t length, int lock_id)
{
  if (result_buffer != NULL)
    return result_buffer->append_data(data, length, 0);

  return Buffer::append_data(data, length, lock_id);
}


bool
Base64Decoder::fix(int lock_id)
{
  Buffer::fix();
  if (result_buffer != NULL)
    result_buffer->fix();

  return pending_length == 0;  // Ⱦüʬä饨顼
}


bool
Base64Decoder::clear(int lock_id)
{
  pending_length = 0;
  bool result = true;
  if (result_buffer != NULL)
    result = result_buffer->clear(0);

  result &= Buffer::clear(lock_id);

  return result;
}


int
Base64Decoder::collect_word(const unsigned char *src, const unsigned char *tail,
			    unsigned char dst[])
{
  const unsigned char *cur_src = src;
  unsigned char *cur_dst = dst;

  while (cur_src < tail)
    {
      unsigned char c = *cur_src++;

      if (c == '\n' || c == '\r')
	continue;

      if ((c & 0x80) != 0 || (dictionary[c] == IV))
	{
	  if (!ignore_garbage)
	    return -1;
	  else
	    continue;
	}

      *cur_dst++ = c;

      if (cur_dst == dst + 4)
	break;
    }

  while (cur_dst < dst + 4)
    *cur_dst++ = '\0';

  return cur_src - src;
}


bool
Base64Decoder::append_data(const char *data, size_t length, int lock_id)
{
  assert(data != NULL && length >= 0);

  if (length == 0)
    return true;

  if (pending_length > 0)
    {
      while (length > 0 && pending_length < 4)
	{
	  unsigned char c = *data++;
	  if ((c & 0x80) == 0 && (dictionary[c] & IV) == 0)
	    pending_buffer[pending_length++] = c;
	  else if (c != '\n' && c != '\r' && !ignore_garbage)
	    return false;
	  --length;
	}

      if (pending_length == 4)
	{
	  pending_length = 0;
	  if (!append_data(pending_buffer, 4, lock_id))
	    return false;
	}
    }

  if (pending_length > 0)
    {
      assert(length == 0);
      return true;
    }

  const unsigned char *src = reinterpret_cast<const unsigned char *>(data);
  char *tmp_buf = static_cast<char *>(malloc(length / 4 * 3));
  const unsigned char *tail = src + length;
  char *dst = tmp_buf;
  size_t decoded_length = 0;

  while (src < tail)
    {
      unsigned char quad[4];
      int bytes_ate = collect_word(src, tail, quad);
      if (bytes_ate == -1)
	break;	// invalid source
      src += bytes_ate;

      if (quad[3] != '\0')
	{
	  int num_pad = 0;
	  unsigned int word
	    = (dictionary[quad[0]] << 18) | (dictionary[quad[1]] << 12);
	  if (quad[2] != BASE64_PAD)
	    {
	      word |= (dictionary[quad[2]] << 6);

	      if (quad[3] != BASE64_PAD)
		{
		  word |= dictionary[quad[3]];
		  num_pad = 0;
		  decoded_length += 3;
		}
	      else
		{
		  num_pad = 1;
		  decoded_length += 2;
		}
	    }
	  else
	    {
	      num_pad = 2;
	      decoded_length++;
	    }

	  *dst++ = word >> 16;
	  if (num_pad < 2)
	    {
	      *dst++ = (word >> 8) & 0xff;
	      if (num_pad == 0)
		*dst++ = word & 0xff;
	    }
	}
      else
	{
	  assert(pending_length == 0);
	  int i;
	  for (i = 0; i < 3; i++)
	    {
	      if (quad[i] != '\0')
		pending_buffer[i] = quad[i];
	      else
		break;
	    }
	  assert(i < 3);
	  pending_length = i;
	}
    }

  bool result;
  if (src == tail)
    result = append_result(tmp_buf, dst - tmp_buf, lock_id);
  else
    result = false;
  free(tmp_buf);

  if (!result)
    return false;

  assert(pending_length < 4);

  return true;
}


}
