/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: bbs_table.cc,v 1.47 2009/01/01 07:46:43 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/bbs_table.h"

#include "libochusha/bbs_2ch_utils.h"
#include "libochusha/bbs_jbbs_utils.h"
#include "libochusha/bulletin_board.h"
#include "libochusha/configuration.h"
#include "libochusha/constants.h"
#include "libochusha/cp932_utils.h"
#include "libochusha/http_headers.h"
#include "libochusha/lock.h"
#include "libochusha/network_agent.h"
#include "libochusha/ref_count.h"
#include "libochusha/regexp.h"
#include "libochusha/repository.h"
#include "libochusha/sax_context.h"
#include "libochusha/text_utils.h"
#include "libochusha/thread_pool.h"
#include "libochusha/tree.h"
#include "libochusha/uri_utils.h"
#include "libochusha/misc_utils.h"

#include <algorithm>
#include <cassert>
#include <cstring>
#include <cstdlib>
#include <iostream>
#include <string>
#include <vector>

#include <zlib.h>


namespace Ochusha
{

BBSTable::BBSTable()
  : RootFolder(N_("Boardlist"))
{
}


BBSTable::~BBSTable()
{
}


bool
BBSTable::read_boardlist_xml(Repository &repository)
{
  return read_tree_xml(repository, OCHUSHA_BOARDLIST_XML);
}


bool
BBSTable::write_boardlist_xml(Repository &repository)
{
  return write_tree_xml(repository, OCHUSHA_BOARDLIST_XML);
}


TreeElement *
BBSTable::copy(const TreeElement *element)
{
  assert(element != NULL);
  assert(!element->is_folder());

  const BulletinBoard *board = dynamic_cast<const BulletinBoard *>(element);
  assert(board != NULL);

  // ƱĤФBulletinBoard֥Ȥ϶ͭߤΤǥԡʤ
  return const_cast<BulletinBoard *>(board);
}


const char *
BBSTable::get_root_element_name() const
{
  return "boardlist";
}


const char *
BBSTable::get_leaf_element_name() const
{
  return "board";
}


Folder *
BBSTable::folder_new(time_t tc, time_t tlm, unsigned int sn)
{
  return new Folder(FOLDER_TYPE_UNKNOWN, NULL, tc, tlm, sn);
}


Folder *
BBSTable::folder_new(const Folder *folder, unsigned int sn)
{
  return new Folder(folder, sn);
}


TreeElement *
BBSTable::leaf_element_new(AttributeTable &attributes, time_t tc, time_t tlm)
{
  const char *base_url = attributes.get_string("base_url");
  if (base_url == NULL)
    return NULL;

  BulletinBoard *board = lookup_board_by_url(base_url);
  if (board == NULL)
    {
      board = new BulletinBoard(attributes, tc, tlm);
      register_board(board, base_url);
    }

  return board;
}


Folder *
BBSTable::lookup_category(const char *name) const
{
  return dynamic_cast<Folder *>(category_table.lookup(name));
}


void
BBSTable::register_board(BulletinBoard *board, const char *base_url)
{
  board_table.insert(base_url, board);

  if (!board->is_in_boardlist())
    {
      board->set_in_boardlist();
      all_boards.push_back(board);
    }
}


void
BBSTable::unregister_board(BulletinBoard *board, const char *base_url)
{
  void *entry = board_table.lookup(base_url);
  if (entry == board)
    board_table.remove(base_url);
}


BulletinBoard *
BBSTable::lookup_board_by_url(const char *url) const
{
  return dynamic_cast<BulletinBoard *>(board_table.lookup(url));
}


class ParseBBSMenuJob : public WorkerJob, public has_slots
{
private:
  BBSTable *table;
  Configuration &config;
  NetworkAgentHandle agent;
  BufferHandle buffer;

  IconvBuffer iconv_buffer;

  Regexp category_pattern;
  Regexp category_end_pattern;
  Regexp board_pattern;

  volatile bool worker_is_running;

  static Monitor monitor;


public:
  ParseBBSMenuJob(BBSTable *table, Configuration &config, NetworkAgent *agent)
    : table(table), config(config), agent(agent), buffer(NULL),
      iconv_buffer(IconvBuffer::UTF8_ENCODING, IconvBuffer::CP932_ENCODING),
      category_pattern("(?i:<b>)([^<]*)(?i:</b>)", "CP932"),
      category_end_pattern("(?i:<b>)|(?i:</body>)", "CP932"),
      board_pattern("(?i:<a[[:space:]]+href=)([^ \">]+|\"[^\">]+\")[^>]*>([^<]+)(?i:</a>)", "CP932"),
      worker_is_running(false)
  {
    agent->prepare_request.connect(this, &ParseBBSMenuJob::prepare);
    agent->access_finished.connect(this, &ParseBBSMenuJob::finished);
    agent->access_failed.connect(this, &ParseBBSMenuJob::failed);
    agent->access_terminated.connect(this, &ParseBBSMenuJob::terminated);
  }
  ~ParseBBSMenuJob()
  {
#ifdef VERBOSE_DESTRUCTOR
    cerr << "ParseBBSMenuJob destructed.\n";
#endif
  }

  virtual void work(const WorkerThread *employee);
  virtual const char *description() const
  { return "parsing of bbsmenu.html"; }

  void prepare(NetworkAgent *agent, RefCount *user_data);
  void finished(NetworkAgent *agent, RefCount *user_data);
  void failed(NetworkAgent *agent, const char *reason, RefCount *user_data);
  void terminated(NetworkAgent *agent, RefCount *user_data);

  void parse();
  void extract_boards(Folder *folder,
		      const char *board_part_start, size_t board_part_len);
};


Monitor ParseBBSMenuJob::monitor("ParseBBSMenuJob");
static Lock update_boardlist_lock("Boardlist Updater");
static bool boardlist_updating = false;


void
ParseBBSMenuJob::work(const WorkerThread *employee)
{
  monitor.enter();

  worker_is_running = true;
  monitor.notify();	// NetworkAgentΥåɤƱ롣

  monitor.wait();	// ͥåȥδλԤ
  monitor.exit();

  buffer = agent->get_buffer();
  if (buffer != NULL)
    parse();		// ͥåȥｪλ

  agent = NULL;
  buffer = NULL;

  LockBlock lock_obj(update_boardlist_lock);
  boardlist_updating = false;
}


void
ParseBBSMenuJob::prepare(NetworkAgent *agent, RefCount *user_data)
{
  LockBlock lock_obj(monitor);

  agent->set_header(HTTPHeader::Accept_Encoding, "deflate, gzip");
  agent->set_header(HTTPHeader::User_Agent, OCHUSHA_USER_AGENT);

  if (config.get_enable_proxy_to_read())
    agent->set_proxy_uri(config.get_proxy_uri_to_read());

  if (!worker_is_running)
    monitor.wait();	// ParseBBSMenuJobѤΥåɤεưԤ
  // ParseBBSMenuJobѤΥåɤư顢ϡ
}


void
ParseBBSMenuJob::finished(NetworkAgent *agent, RefCount *user_data)
{
  LockBlock lock_obj(monitor);
  if (agent->get_status_code() != 200)
    buffer = NULL;
  monitor.notify();
}


void
ParseBBSMenuJob::failed(NetworkAgent *agent, const char *reason,
			RefCount *user_data)
{
  LockBlock lock_obj(monitor);
  buffer = NULL;
  monitor.notify();
  cerr << "Updating of boardlist failed: " << reason << endl;
}


void
ParseBBSMenuJob::terminated(NetworkAgent *agent, RefCount *user_data)
{
  LockBlock lock_obj(monitor);
  buffer = NULL;
  monitor.notify();
  cerr << "Updating of boardlist terminated.\n";
}


void
ParseBBSMenuJob::extract_boards(Folder *folder, const char *board_part_start,
				size_t board_part_len)
{
  const char *cur_pos = board_part_start;
  while (board_part_len > 0 && board_pattern.match(cur_pos, board_part_len))
    {
      char *url;
      if (board_pattern.match_begin(1)[0] != '"')
	url = strndup(board_pattern.match_begin(1),
		      board_pattern.match_len(1));
      else
	url = strndup(board_pattern.match_begin(1) + 1,
		      board_pattern.match_len(1) - 2);
      iconv_buffer.clear();
      iconv_buffer.append(board_pattern.match_begin(2),
			  board_pattern.match_len(2));
      const char *board_name = iconv_buffer.get_string();
#if 0
      cerr << "Board candidate: " << board_name << ", " << url << endl;
#endif
      BulletinBoard *board = table->lookup_board_by_url(url);
      if (board == NULL)
	{
	  board = table->guess_board(url);
	  if (board != NULL)
	    {
	      // İž
	      // base_urlurlѹƤ륹Ƚ̤ξǤ
	      // datեΰưʤɤsignalϥɥˤǤ
	      table->board_moved(table, board, url);
	      table->register_board(board, url);
	    }
	}

      if (board == NULL)
	{
	  // 
	  int type = BBSTable::guess_board_type(url);
	  char *base_url = BBSTable::guess_board_url(url);
#if 0
	  if (base_url != NULL)
	    cerr << "type=" << type << ", base_url=" << base_url << endl;
	  else
	    cerr << "type=" << type << ", url=" << url << endl;
#endif
	  if (type != OCHUSHA_BBS_TYPE_UNKNOWN
	      && type != OCHUSHA_BBS_TYPE_2CH_COMPATIBLE
	      && base_url != NULL)
	    {
	      board = new BulletinBoard();
	      board->set_bbs_type(type);
	      board->set_base_url(base_url);
	      board->set_name(board_name);

	      table->register_board(board, base_url);
	    }
	  if (base_url != NULL)
	    free(base_url);
	}
      else
	board->set_name(board_name);	// ̾ι

      if (board != NULL)
	{
	  BulletinBoardHandle b = board;
	  folder->remove(board);
	  folder->append(board);
	}

      free(url);

      cur_pos = board_pattern.match_end() + 1;
      board_part_len = category_end_pattern.match_begin() - cur_pos;
    }
}


void
ParseBBSMenuJob::parse()
{
  RWLock *rwlock = buffer->get_lock();
  int lock_id = rwlock->rdlock();

  const char *html_buf = buffer->get_buffer(lock_id);
  size_t html_len = buffer->get_length(lock_id);

  const char *cur_pos = html_buf;
  size_t rest_len = html_len;
  Folder *prev_folder = NULL;
  while (rest_len > 0 && category_pattern.match(cur_pos, rest_len))
    {
      iconv_buffer.clear();
      iconv_buffer.append(category_pattern.match_begin(1),
			  category_pattern.match_len(1));
      const char *category_name = iconv_buffer.get_string();
					   
      Folder *folder = table->lookup_category(category_name);
      bool category_is_new;
      if (folder == NULL)
	{
	  folder = table->RootFolder::folder_new();	// θʤġ
	  folder->set_folder_type(FOLDER_TYPE_BBSMENU_CATEGORY);
	  folder->set_name(category_name);
	  category_is_new = true;
	}
      else
	category_is_new = false;

#if 0
      cerr << "Parsing for category: " << category_name << endl;
#endif

      cur_pos = category_pattern.match_end() + 1;
      rest_len = html_buf + html_len - cur_pos;

      if (category_end_pattern.match(cur_pos, rest_len))
	{
	  extract_boards(folder, cur_pos,
			 category_end_pattern.match_begin() - cur_pos);
	  cur_pos = category_end_pattern.match_begin();
	  rest_len = html_buf + html_len - cur_pos;
	}
      else
	rest_len = 0;	// λ

      if (folder->is_empty())
	delete folder;
      else
	{
	  if (category_is_new)
	    table->append(folder, prev_folder);
	  prev_folder = folder;
	}
    }

  rwlock->unlock(lock_id);
  table->tree_updated(table);
}


void
BBSTable::update_boardlist(Configuration &config, ThreadPool &pool,
			   NetworkAgentBroker &broker)
{
  LockBlock lock_obj(update_boardlist_lock);
  if (boardlist_updating)
    return;

  NetworkAgent *agent = broker.employ_agent();
  ParseBBSMenuJob *job = new ParseBBSMenuJob(this, config, agent);
  pool.commit(job);

  Buffer *buffer = new Buffer();
  const URI &uri = config.get_bbsmenu_uri();
  agent->get(uri.get_uri(), buffer, job);
}


void
BBSTable::prepend(TreeElement *element, TreeElement *next_element)
{
  assert(element != NULL);
  if (element->is_folder())
    {
      Folder *folder = dynamic_cast<Folder *>(element);
      if (folder->get_folder_type() == FOLDER_TYPE_BBSMENU_CATEGORY)
	{
	  Folder *old = lookup_category(folder->get_name());
	  if (old != NULL)
	    {
	      if (folder == old)
		return;
	      else
		remove(old);
	    }
	  category_table.insert(folder->get_name(), folder);
	}
    }
  Folder::prepend(element, next_element);
}


void
BBSTable::append(TreeElement *element, TreeElement *prev_element)
{
  assert(element != NULL);
  if (element->is_folder())
    {
      Folder *folder = dynamic_cast<Folder *>(element);
      if (folder->get_folder_type() == FOLDER_TYPE_BBSMENU_CATEGORY)
	{
	  Folder *old = lookup_category(folder->get_name());
	  if (old != NULL)
	    {
	      if (folder == old)
		return;
	      else
		remove(old);
	    }
	  category_table.insert(folder->get_name(), folder);
	}
    }
  Folder::append(element, prev_element);
}


void
BBSTable::remove(TreeElement *element)
{
  assert(element != NULL);
  if (element->is_folder())
    {
      Folder *folder = dynamic_cast<Folder *>(element);

      if (folder->get_folder_type() == FOLDER_TYPE_BBSMENU_CATEGORY)
	category_table.remove(folder->get_name());
    }
  Folder::remove(element);
}


BulletinBoard *
BBSTable::guess_board(const char *url)
{
  if (url == NULL)
    return NULL;

  BulletinBoard *board = lookup_board_by_url(url);
  if (board != NULL)
    return board;

  char *board_url = guess_board_url(url);

  if (board_url == NULL)
    return NULL;	// 롣

  board = lookup_board_by_url(board_url);
  if (board != NULL)
    {
      free(board_url);
      return board;
    }

  URI uri(board_url);
  free(board_url);

  if (!uri.is_valid())
    return NULL;

  board = BBS2ch::guess_board(uri);
  if (board != NULL)
    return board;

  board = BBSMachiBBS::guess_board(uri);
  if (board != NULL)
    return board;

  board = BBSJbbsLivedoor::guess_board(uri);
  return board;
}


char *
BBSTable::guess_board_url(const char *url)
{
  if (url == NULL)
    return NULL;

  URI uri(url);
  if (!uri.is_valid())
    return NULL;

  char *board_url = BBS2ch::guess_board_url(uri);
  if (board_url != NULL)
    return board_url;

  board_url = BBSMachiBBS::guess_board_url(uri);
  if (board_url != NULL)
    return board_url;

  board_url = BBSJbbsLivedoor::guess_board_url(uri);
  if (board_url != NULL)
    return board_url;

  return BBS2chCompatible::guess_board_url(uri);
}


int
BBSTable::guess_board_type(const char *url)
{
  if (url == NULL)
    return OCHUSHA_BBS_TYPE_UNKNOWN;

  URI uri(url);
  if (!uri.is_valid())
    return OCHUSHA_BBS_TYPE_UNKNOWN;

  int board_type = BBS2ch::guess_board_type(uri);
  if (board_type != OCHUSHA_BBS_TYPE_UNKNOWN)
    return board_type;

  board_type = BBSMachiBBS::guess_board_type(uri);
  if (board_type != OCHUSHA_BBS_TYPE_UNKNOWN)
    return board_type;

  board_type = BBSJbbsLivedoor::guess_board_type(uri);
  if (board_type != OCHUSHA_BBS_TYPE_UNKNOWN)
    return board_type;

  return OCHUSHA_BBS_TYPE_2CH_COMPATIBLE;
}


}
