/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: bbs_thread.cc,v 1.55 2009/01/01 07:46:43 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/bulletin_board.h"
#include "libochusha/bbs_thread.h"

#include "libochusha/bbs_2ch_utils.h"
#include "libochusha/bbs_jbbs_utils.h"
#include "libochusha/configuration.h"
#include "libochusha/text_utils.h"
#include "libochusha/misc_utils.h"

#include <climits>
#include <cstdlib>
#include <cstring>

#include <iostream>

#include <zlib.h>


#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif


namespace Ochusha
{


#undef VERBOSE_DESTRUCTOR


Response::Response()
  : name(NULL), mailto(NULL), date(NULL), id(NULL), message(NULL), title(NULL),
    flags(0)
{
}


Response::~Response()
{
  if (name != NULL)
    free(name);

  if (mailto != NULL)
    free(mailto);

  if (date != NULL)
    free(date);

  if (id != NULL)
    free(id);

  if (message != NULL)
    free(message);

  if (title != NULL)
    free(title);
}


void
Response::set_name(const char *value)
{
  if (name != NULL)
    free(name);
  name = value != NULL ? strdup(value) : NULL;
}


void
Response::set_mailto(const char *value)
{
  if (mailto != NULL)
    free(mailto);
  mailto = value != NULL ? strdup(value) : NULL;
}


void
Response::set_date(const char *value)
{
  if (date != NULL)
    free(date);
  date = value != NULL ? strdup(value) : NULL;
}


void
Response::set_id(const char *value)
{
  if (id != NULL)
    free(id);
  id = value != NULL ? strdup(value) : NULL;
}


void
Response::set_message(const char *value)
{
  if (message != NULL)
    free(message);
  message = value != NULL ? strdup(value) : NULL;
}


void
Response::set_title(const char *value)
{
  if (title != NULL)
    free(title);
  title = value != NULL ? strdup(value) : NULL;
}


ResponseCursor::ResponseCursor(BBSThread *thread,
			       Configuration &config, Repository &repository,
			       Buffer *buffer, NetworkAgent *agent,
			       const char *bbs_enc)
  : thread(thread),
    config(config),
    repository(repository),
    iconv_buffer(IconvBuffer::UTF8_ENCODING, bbs_enc),
    buffer(buffer),
    agent(agent)
{
  // ͤξ硢㳰ŪüʾȳƥϹ⡹1000쥹
  // ʤΤǡ2ɤʤhints1000餤ǽ˼äƤޤ
  hints.reserve(1024);

  if (agent == NULL)
    return;

  agent->access_started.connect(this, &ResponseCursor::access_started_cb);
  agent->access_progressed.connect(this, &ResponseCursor::access_progressed_cb);
  agent->access_finished.connect(this, &ResponseCursor::access_finished_cb);
  agent->cache_is_dirty.connect(this, &ResponseCursor::cache_is_dirty_cb);
  agent->access_failed.connect(this, &ResponseCursor::access_failed_cb);
  agent->access_terminated.connect(this, &ResponseCursor::access_terminated_cb);
}


ResponseCursor::~ResponseCursor()
{
#ifdef VERBOSE_DESTRUCTOR
  fprintf(stderr, "ResponseCursor(@%p) destructed\n", this);
#endif
}


void
ResponseCursor::clear_hints()
{
  thread->set_number_of_responses_got(0);
  thread->set_dat_filesize(0);


  int lock_id = buffer->get_lock()->wrlock();

  while (!hints.empty())
    hints.pop_back();

  buffer->get_lock()->unlock(lock_id);
}


void
ResponseCursor::set_hints(unsigned int res_num, int offset, size_t length)
{
  assert(res_num > 0);
  assert(hints.size() >= res_num - 1);
  assert((offset & RESPONSE_HINT_MASK) == 0);

  thread->set_number_of_responses_got(res_num);
  thread->set_dat_filesize(offset + length);

  if (hints.capacity() >= res_num)
    {
      if (hints.size() == res_num - 1)
	hints.push_back(ResponseHint(offset, length));
      else
	hints[res_num - 1] = ResponseHint(offset, length);
      return;
    }

  assert(hints.size() == res_num - 1);


  int lock_id = buffer->get_lock()->wrlock();

  hints.push_back(ResponseHint(offset, length));

  buffer->get_lock()->unlock(lock_id);
}


void
ResponseCursor::set_broken(unsigned int res_num, int offset, size_t length)
{
  assert(res_num > 0);
  assert(hints.size() >= res_num - 1);
  assert((offset & RESPONSE_HINT_MASK) == 0);

  if (thread->get_number_of_responses_got() < static_cast<int>(res_num))
    {
      thread->set_number_of_responses_got(res_num);
      thread->set_dat_filesize(offset + length);
    }

  if (hints.capacity() >= res_num)
    {
      if (hints.size() == res_num - 1)
	hints.push_back(ResponseHint(RESPONSE_HINT_BROKEN | offset, length));
      else
	hints[res_num - 1]
	  = ResponseHint(RESPONSE_HINT_BROKEN | offset, length);
      return;
    }

  assert(hints.size() == res_num - 1);


  int lock_id = buffer->get_lock()->wrlock();

  hints.push_back(ResponseHint(RESPONSE_HINT_BROKEN | offset, length));

  buffer->get_lock()->unlock(lock_id);
}


void
ResponseCursor::set_deleted(unsigned int res_num, int offset)
{
  assert(res_num > 0);
  assert(hints.size() >= res_num - 1);
  assert((offset & RESPONSE_HINT_MASK) == 0);

  thread->set_number_of_responses_got(res_num);

  if (hints.capacity() >= res_num)
    {
      if (hints.size() == res_num - 1)
	hints.push_back(ResponseHint(RESPONSE_HINT_DELETED | offset, 0));
      else
	hints[res_num - 1] = ResponseHint(RESPONSE_HINT_DELETED | offset, 0);
      return;
    }

  assert(hints.size() == res_num - 1);


  int lock_id = buffer->get_lock()->wrlock();

  hints.push_back(ResponseHint(RESPONSE_HINT_DELETED | offset, 0));

  buffer->get_lock()->unlock(lock_id);
}


int
ResponseCursor::get_next_offset()
{
  unsigned int res_num = hints.size();
  if (res_num > 0)
    {
      unsigned int prev = res_num - 1;
      return (hints[prev].offset & ~RESPONSE_HINT_MASK) + hints[prev].length;
    }

  return 0;
}


bool
ResponseCursor::has_response(unsigned int res_num, int lock_id)
{
  assert(res_num > 0);
#ifndef NDEBUG
  if (hints.size() >= res_num)
    {
      assert((hints[res_num - 1].offset & RESPONSE_HINT_DELETED) != 0
	      || hints[res_num - 1].length > 0);

      return true;
    }
  return false;
#else
  return hints.size() >= res_num;
#endif
}


Response &
ResponseCursor::get_response(unsigned int res_num, int lock_id)
{
  assert(res_num > 0);
  assert(has_response(res_num, lock_id));

  unsigned int index = res_num - 1;

  int offset = hints[index].offset;
  if ((offset & RESPONSE_HINT_MASK) == 0)
    {
      response.clear_flags();
      if (parse_response(res_num, buffer->get_buffer(lock_id) + offset,
			 hints[index].length))
	return response;

      set_broken(res_num, offset, hints[res_num].length);

      response.set_broken();
      return response;
    }

  switch (offset & RESPONSE_HINT_MASK)
    {
    case RESPONSE_HINT_BROKEN:
      response.set_broken();
      return response;

    case RESPONSE_HINT_DELETED:
      response.set_deleted();
      return response;
    }

  // never reached here
#ifndef NDEBUG
  assert(0);
#else
  response.set_not_ready();
  return response;
#endif
}


bool
ResponseCursor::finished(int lock_id)
{
  return buffer->is_fixed();
}


NetworkAgent::NetworkAgentState
ResponseCursor::get_agent_state(int lock_id)
{
  if (finished(lock_id) || static_cast<NetworkAgent *>(agent) == NULL)
    return NetworkAgent::STATE_INITIAL;

  return agent->get_state();
}


void
ResponseCursor::access_started_cb(NetworkAgent *agent, RefCount *data)
{
  access_started();
}


void
ResponseCursor::access_progressed_cb(NetworkAgent *agent, size_t data_read,
				     uintmax_t total_len, RefCount *data)
{
  access_progressed();
}


void
ResponseCursor::cache_is_dirty_cb(NetworkAgent *agent, RefCount *data)
{
  cache_is_dirty();
}


void
ResponseCursor::access_finished_cb(NetworkAgent *agent, RefCount *data)
{
  access_finished();
}


void
ResponseCursor::access_failed_cb(NetworkAgent *agent, const char *reason,
				 RefCount *data)
{
  access_failed();
}


void
ResponseCursor::access_terminated_cb(NetworkAgent *agent, RefCount *data)
{
  access_terminated();
}


BBSThread::BBSThread(BulletinBoard *board, const char *id, const char *title)
  : board(board),
    id(NULL),
    title(NULL),
    last_modified(NULL),
    date(NULL),
    kako_html(NULL),
    grave(NULL),
    trigger_url(NULL),
    rank(0),
    previous_rank(0),
    number_of_responses_on_server(0),
    number_of_responses_got(0),
    number_of_responses_read(0),
    number_of_new_responses(0),
    bookmark_position(0),
    last_view_response(0),
    shown_count(0),
    flags(0),
    ui_flags(0),
    last_name(NULL),
    last_mail(NULL),
    dat_filesize(0),
    alive(false)
{
  set_id(id);
  set_title(title);
}


BBSThread::~BBSThread()
{
  if (last_mail != NULL)
    free(last_mail);

  if (last_name != NULL)
    free(last_name);

  if (trigger_url != NULL)
    free(trigger_url);

  if (grave != NULL)
    free(grave);

  if (kako_html != NULL)
    free(kako_html);

  if (date != NULL)
    free(date);

  if (last_modified != NULL)
    free(last_modified);

  if (title != NULL)
    free(title);

  if (id != NULL)
    free(id);
}


int
BBSThread::get_bbs_type() const
{
  BulletinBoard *board = get_board();

  if (grave == NULL && kako_html == NULL)
    return board->get_bbs_type();

  URI uri;
  if (kako_html != NULL)
    uri.parse_uri(kako_html);

  if (grave != NULL && !uri.is_valid())
    uri.parse_uri(grave);

  int result;
  switch (board->get_bbs_type())
    {
    case OCHUSHA_BBS_TYPE_2CH:
    case OCHUSHA_BBS_TYPE_2CHLIKE_EUCJP:
    case OCHUSHA_BBS_TYPE_2CH_COMPATIBLE:
    case OCHUSHA_BBS_TYPE_2CH_HEADLINE:
    case OCHUSHA_BBS_TYPE_2CH_BE:
    case OCHUSHA_BBS_TYPE_BBSPINK:
    case OCHUSHA_BBS_TYPE_BBSPINK_HEADLINE:
    case OCHUSHA_BBS_TYPE_2CH_COMPATIBLE_RO:
      result = BBS2ch::guess_board_type(uri);
      break;

    case OCHUSHA_BBS_TYPE_MACHIBBS:
      result = BBSMachiBBS::guess_board_type(uri);
      break;


    case OCHUSHA_BBS_TYPE_JBBS_LIVEDOOR:
      result = BBSJbbsLivedoor::guess_board_type(uri);
      break;

    default:
      result = OCHUSHA_BBS_TYPE_UNKNOWN;
      break;
    }

  if (result != OCHUSHA_BBS_TYPE_UNKNOWN)
    return result;

  return board->get_bbs_type();
}


ResponseCursor *
BBSThread::get_responses(Configuration &config,
			 Repository &repository,
			 NetworkAgentBroker &broker)
{
  assert(board != NULL);
  switch (get_board()->get_bbs_type())
    {
    case OCHUSHA_BBS_TYPE_2CH:
    case OCHUSHA_BBS_TYPE_2CHLIKE_EUCJP:
    case OCHUSHA_BBS_TYPE_2CH_COMPATIBLE:
    case OCHUSHA_BBS_TYPE_2CH_HEADLINE:
    case OCHUSHA_BBS_TYPE_2CH_BE:
    case OCHUSHA_BBS_TYPE_BBSPINK:
    case OCHUSHA_BBS_TYPE_BBSPINK_HEADLINE:
    case OCHUSHA_BBS_TYPE_2CH_COMPATIBLE_RO:
      return new ResponseCursor2chDAT(this, config, repository,
				      broker.employ_agent());

    case OCHUSHA_BBS_TYPE_MACHIBBS:
      return new ResponseCursorMachiBBS(this, config, repository,
					broker.employ_agent());

    case OCHUSHA_BBS_TYPE_JBBS_LIVEDOOR:
      return new ResponseCursorJbbsLivedoor(this, config, repository,
					    broker.employ_agent());

#if 0
    case OCHUSHA_BBS_TYPE_JBBS:
    case OCHUSHA_BBS_TYPE_MITINOKU:
      return new ResponseCursorJBBS(this, config, repository,
				    broker.employ_agent());
#endif
    default:
      break;
    }
  return NULL;
}


PostResponseAgent *
BBSThread::get_post_agent(Configuration &config, NetworkAgentBroker &broker)
{
  if (config.get_offline())
    return NULL;

  assert(board != NULL);
  switch (get_board()->get_bbs_type())
    {
    case OCHUSHA_BBS_TYPE_2CH:
    case OCHUSHA_BBS_TYPE_2CHLIKE_EUCJP:
    case OCHUSHA_BBS_TYPE_2CH_COMPATIBLE:
    case OCHUSHA_BBS_TYPE_2CH_HEADLINE:
    case OCHUSHA_BBS_TYPE_2CH_BE:
    case OCHUSHA_BBS_TYPE_BBSPINK:
    case OCHUSHA_BBS_TYPE_BBSPINK_HEADLINE:
      return new PostResponseAgent2ch(this, config, broker.employ_agent());

    case OCHUSHA_BBS_TYPE_MACHIBBS:
    case OCHUSHA_BBS_TYPE_JBBS_LIVEDOOR:
      return new PostResponseAgentJbbs(this, config, broker.employ_agent());

#if 0
    case OCHUSHA_BBS_TYPE_JBBS:
    case OCHUSHA_BBS_TYPE_MITINOKU:
#endif
    case OCHUSHA_BBS_TYPE_2CH_COMPATIBLE_RO:
    default:
      break;
    }
  return NULL;
}


int
BBSThread::get_flags() const
{
  return flags;
}


void
BBSThread::set_flags(int value)
{
  if (flags == value)
    return;

  flags = value;
  get_board()->notify_minor_modification();
}


void
BBSThread::set_number_of_responses_got(int value)
{
  if (number_of_responses_got == value)
    return;

  number_of_responses_got = value;
  get_board()->notify_minor_modification();
}


void
BBSThread::set_number_of_responses_read(int value)
{
  if (number_of_responses_read == value)
    return;

  number_of_responses_read = value;
  get_board()->notify_minor_modification();
}


void
BBSThread::set_bookmark_position(int value)
{
  if (bookmark_position == value)
    return;
  bookmark_position = value;
  get_board()->notify_major_modification();
}


void
BBSThread::set_last_view_response(int value)
{
  if (value < 0 || value > number_of_responses_got)
    value = 0;

  if (last_view_response == value)
    return;

  last_view_response = value;
  get_board()->notify_major_modification();
}


void
BBSThread::set_shown_count(int value)
{
  if (value < 0)
    value = 0;

  if (shown_count == value)
    return;

  shown_count = value;
  get_board()->notify_minor_modification();
}


void
BBSThread::set_dat_dropped(bool value)
{
  int old_val = get_flags();
  set_flags((old_val & ~DAT_DROPPED) | (value ? DAT_DROPPED : 0));
}


void
BBSThread::set_stopped(bool value)
{
  int old_val = get_flags();
  set_flags((old_val & ~STOPPED) | (value ? STOPPED : 0));
}


void
BBSThread::set_kako_thread(bool value)
{
  int old_val = get_flags();
  set_flags((old_val & ~(KAKO_THREAD | DAT_DROPPED))
	    | (value ? (KAKO_THREAD | DAT_DROPPED) : 0));
}


void
BBSThread::set_viewer_used(bool value)
{
  int old_val = get_flags();
  set_flags((old_val & ~(DAT_VIEWER | DAT_DROPPED))
	    | (value ? (DAT_VIEWER | DAT_DROPPED) : 0));
}


void
BBSThread::set_fake_dat(bool value)
{
  int old_val = get_flags();
  set_flags((old_val & ~FAKE_DAT) | (value ? FAKE_DAT : 0));
}


void
BBSThread::set_tmp_thread(bool value)
{
  int old_val = get_flags();
  set_flags((old_val & ~TMP_THREAD) | (value ? TMP_THREAD : 0));
}


void
BBSThread::set_id(const char *value)
{
  if (id != NULL && value != NULL && strcmp(id, value) == 0)
    return;

  if (id != NULL)
    free(id);

  id = value != NULL ? strdup(value) : NULL;
}


void
BBSThread::set_title(const char *value)
{
  char *new_title;
  if (value != NULL)
    new_title = wipe_string(value);
  else
    new_title = NULL;

  if (new_title != NULL && title != NULL && strcmp(new_title, title) == 0)
    {
      free(new_title);
      return;
    }

  if (title != NULL)
    free(title);

  title = new_title;
  set_tmp_thread(title == NULL);
  get_board()->notify_major_modification();
}


void
BBSThread::set_last_modified(const char *value)
{
  if (value != NULL && last_modified != NULL
      && strcmp(value, last_modified) == 0)
    return;

  if (last_modified != NULL)
    free(last_modified);
  last_modified = value != NULL ? strdup(value) : NULL;
  get_board()->notify_major_modification();
}


void
BBSThread::set_date(const char *value)
{
  if (date != NULL)
    free(date);
  date = value != NULL ? strdup(value) : NULL;
}


void
BBSThread::set_kako_html(const char *value)
{
  if (value != NULL && kako_html != NULL && strcmp(value, kako_html) == 0)
    return;

  if (kako_html != NULL)
    free(kako_html);
  kako_html = value != NULL ? strdup(value) : NULL;
  get_board()->notify_major_modification();
}


void
BBSThread::set_grave(const char *value)
{
  if (value != NULL && grave != NULL && strcmp(value, grave) == 0)
    return;

  if (grave != NULL)
    free(grave);

  grave = value != NULL ? strdup(value) : NULL;
  get_board()->notify_major_modification();
}


void
BBSThread::set_trigger_url(const char *value)
{
  if (value != NULL && trigger_url != NULL && strcmp(value, trigger_url) == 0)
    return;

  if (trigger_url != NULL)
    free(trigger_url);

  trigger_url = value != NULL ? strdup(value) : NULL;
}


void
BBSThread::set_marked(bool value)
{
  ui_flags = (ui_flags & ~MARKED) | (value ? MARKED : 0);
  get_board()->notify_major_modification();
}


void
BBSThread::set_hidden(bool value)
{
  ui_flags = (ui_flags & ~HIDDEN) | (value ? HIDDEN : 0);
  get_board()->notify_major_modification();
}


void
BBSThread::set_last_post_sage(bool value)
{
  ui_flags = (ui_flags & ~LAST_POST_SAGE) | (value ? LAST_POST_SAGE : 0);
  get_board()->notify_major_modification();
}


void
BBSThread::set_last_post_be(bool value)
{
  ui_flags = (ui_flags & ~LAST_POST_BE) | (value ? LAST_POST_BE : 0);
  get_board()->notify_major_modification();
}


void
BBSThread::set_last_post_viewer(bool value)
{
  ui_flags = (ui_flags & ~LAST_POST_VIEWER) | (value ? LAST_POST_VIEWER : 0);
  get_board()->notify_major_modification();
}


void
BBSThread::set_last_name(const char *value)
{
  if (last_name != NULL)
    free(last_name);
  last_name = value != NULL ? strdup(value) : NULL;
  get_board()->notify_major_modification();
}


void
BBSThread::set_last_mail(const char *value)
{
  if (last_mail != NULL)
    free(last_mail);
  last_mail = value != NULL ? strdup(value) : NULL;
  get_board()->notify_major_modification();
}


void
BBSThread::set_number_of_responses_on_server(int value)
{
  if (number_of_responses_on_server < value)
    number_of_new_responses = value - number_of_responses_on_server;
  else
    number_of_new_responses = 0;
  number_of_responses_on_server = value;
}


void
BBSThread::recover_information(Configuration &config, Repository &repository)
{
  assert(board != NULL);

  char dat_file[PATH_MAX];
  if (excavate_dat_file(repository, dat_file, PATH_MAX) == NULL)
    return;

  ResponseCursorHandle cursor;
  switch (get_board()->get_bbs_type())
    {
    case OCHUSHA_BBS_TYPE_2CH:
    case OCHUSHA_BBS_TYPE_2CHLIKE_EUCJP:
    case OCHUSHA_BBS_TYPE_2CH_COMPATIBLE:
    case OCHUSHA_BBS_TYPE_2CH_HEADLINE:
    case OCHUSHA_BBS_TYPE_2CH_BE:
    case OCHUSHA_BBS_TYPE_BBSPINK:
    case OCHUSHA_BBS_TYPE_BBSPINK_HEADLINE:
    case OCHUSHA_BBS_TYPE_2CH_COMPATIBLE_RO:
      cursor = new ResponseCursor2chDAT(this, config, repository, NULL);
      break;

    case OCHUSHA_BBS_TYPE_MACHIBBS:
      cursor = new ResponseCursorMachiBBS(this, config, repository, NULL);
      break;

    case OCHUSHA_BBS_TYPE_JBBS_LIVEDOOR:
      cursor = new ResponseCursorJbbsLivedoor(this, config, repository, NULL);
      break;

#if 0
    case OCHUSHA_BBS_TYPE_JBBS:
    case OCHUSHA_BBS_TYPE_MITINOKU:
      cursor = new ResponseCursorJBBS(this, config, repository, NULL);
#endif
    default:
      cursor = NULL;
    }

  set_number_of_responses_read(number_of_responses_got);

  if (cursor != NULL)
    {
      int lock_id = cursor->lock();
      if (cursor->has_response(1, lock_id))
	{
	  Response &response = cursor->get_response(1, lock_id);
	  const char *val = response.get_title();
	  if (title == NULL || val != NULL)
	    set_title(val);
	}
      cursor->unlock(lock_id);
    }
}


char *
BBSThread::get_dat_url(char *buf, size_t buf_len) const
{
  if (buf == NULL || id == NULL)
    return NULL;

  BulletinBoard *board = get_board();
  assert(board != NULL);

  URI base_uri;
  string server;
  if (kako_html != NULL)
    {
#warning "Revisit here!"
      if (!base_uri.parse_uri(kako_html) || base_uri.get_server() == NULL)
	return NULL;
      server = base_uri.get_server();

      if (!base_uri.parse_uri(board->get_base_url())
	  || base_uri.get_scheme() == NULL)
	return NULL;
    }
  else if (grave != NULL)
    {
      if (!base_uri.parse_uri(grave) || base_uri.get_server() == NULL)
	return NULL;
      server = base_uri.get_server();

      if (!base_uri.parse_uri(board->get_base_url())
	  || base_uri.get_scheme() == NULL)
	return NULL;
    }
  else
    {
      if (!base_uri.parse_uri(board->get_base_url())
	  || base_uri.get_server() == NULL || base_uri.get_scheme() == NULL)
	return NULL;
      server = base_uri.get_server();
    }

  if (snprintf(buf, buf_len, "%s://%s%sdat/%s.dat",
	       base_uri.get_scheme(), server.c_str(), base_uri.get_path(), id)
      < static_cast<int>(buf_len))
    return buf;

  return NULL;
}


char *
BBSThread::excavate_dat_file(Repository &repository, char *buf, size_t buf_len)
{
  BulletinBoard *board = get_board();
  assert(board != NULL);

  URI base_uri(board->get_base_url());
  if (!base_uri.is_valid()
      || base_uri.get_server() == NULL || base_uri.get_scheme() == NULL)
    return NULL;

  off_t st_size = 0;
  time_t mtime = 0;
  string grave_dat;
  string grave_server;
  string grave_file;
  char url[PATH_MAX];
  BBSDataManager &bbs_data = repository.get_bbs_data();

  if (snprintf(url, PATH_MAX, "%sdat/%s.dat",
	       board->get_base_url(), id) < PATH_MAX
      && bbs_data.find_file(url, buf, buf_len) != NULL)
    {
      struct stat sb;
      if (stat(buf, &sb) == 0)
	{
	  grave_dat = url;
	  grave_server = base_uri.get_server();
	  grave_file = buf;
	  st_size = sb.st_size;
	  mtime = sb.st_mtime;
	}
    }

  list<char *>::const_iterator iter = board->movement_history_begin();
  list<char *>::const_iterator end_iter = board->movement_history_end();
  while (iter != end_iter)
    {
      if (snprintf(url, PATH_MAX, "%s://%s%sdat/%s.dat", base_uri.get_scheme(),
		   *iter, base_uri.get_path(), id) < PATH_MAX
	  && bbs_data.find_file(url, buf, buf_len) != NULL)
	{
	  struct stat sb;
	  if (stat(buf, &sb) == 0 && sb.st_size > st_size)
	    {
	      if (sb.st_mtime > mtime)
		bbs_data.remove_file(grave_dat.c_str());
	      else
		{
		  fprintf(stderr,
			  LIB_("\"%s\" is smaller but newer than \"%s\"\n"),
			  grave_file.c_str(), buf);
		}
	      grave_dat = url;
	      grave_server = *iter;
	      grave_file = buf;
	      st_size = sb.st_size;
	      mtime = sb.st_mtime;
	    }
	}
      ++iter;
    }

  if (!grave_file.empty())
    {
      assert(grave_file.length() < buf_len);

      if (last_modified == NULL)
	{
	  char lm[64];
	  lm[0] = '\0';
	  set_last_modified(get_gmt_date_from_utc(mtime, lm, 64));
	}

      if (!is_alive())
	{
	  // DATƤ륹ξgrave
	  char *dummy_url = get_url();
	  if (dummy_url != NULL)
	    {
	      URI dummy_uri(dummy_url);
	      if (dummy_uri.is_valid() && dummy_uri.get_scheme() != NULL)
		{
		  string grave_url = dummy_uri.get_scheme();
		  grave_url.append("://").append(grave_server)
		    .append(dummy_uri.get_path());
		  const char *query = dummy_uri.get_query();
		  if (query != NULL && *query != '\0')
		    {
		      if (*query != '?')
			grave_url.append("?");
		      grave_url.append(dummy_uri.get_query());
		    }
		  set_grave(grave_url.c_str());
		}

	      free(dummy_url);
	    }
	}

      return strcpy(buf, grave_file.c_str());
    }

  return NULL;
}


size_t
BBSThread::get_dat_filesize(Repository &repository)
{
  if (dat_filesize != 0)
    return dat_filesize;

  char url[PATH_MAX];
  if (get_dat_url(url, PATH_MAX) == NULL)
    return 0;

  char dat_path[PATH_MAX];
  if (repository.get_bbs_data().find_file(url, dat_path, PATH_MAX) == NULL
      && excavate_dat_file(repository, dat_path, PATH_MAX) == NULL)
    return 0;

  struct stat sb;
  if (stat(dat_path, &sb) != 0)
    sb.st_size = 0;

  set_dat_filesize(sb.st_size);
  return get_dat_filesize();
}


char *
BBSThread::find_dat_file(Repository &repository, char *buf, size_t buf_len,
			 bool to_create)
{
  char url[PATH_MAX];
  if (get_dat_url(url, PATH_MAX) == NULL)
    return NULL;

  if (repository.get_bbs_data().find_file(url, buf, buf_len) != NULL)
    return buf;

  if (excavate_dat_file(repository, buf, buf_len) != NULL)
    {
      if (is_alive())
	{
	  repository.get_bbs_data().move_file(buf, url);
	  return repository.get_bbs_data().find_file(url, buf, buf_len);
	}
      return buf;
    }

  if (to_create)
    return repository.get_bbs_data().get_pathname(url, buf, buf_len);

  return NULL;
}


int
BBSThread::open_dat_file(Repository &repository, int flags)
{
  char url[PATH_MAX];
  if (get_dat_url(url, PATH_MAX) == NULL)
    return -1;

  int fd = repository.get_bbs_data().open_file(url, flags);
  if (fd >= 0 || (flags & O_CREAT) != 0)
    return fd;

  char dat_path[PATH_MAX];
  if (excavate_dat_file(repository, dat_path, PATH_MAX) == NULL)
    return -1;

  fd = open(dat_path, flags);

  if (is_alive())
    repository.get_bbs_data().move_file(dat_path, url);

  return fd;
}


void
BBSThread::remove_dat_file(Repository &repository)
{
  char url[PATH_MAX];
  if (get_dat_url(url, PATH_MAX) == NULL)
    return;

  char dat_path[PATH_MAX];
  if (excavate_dat_file(repository, dat_path, PATH_MAX) != NULL)
    repository.get_bbs_data().move_file(dat_path, url);

  repository.get_bbs_data().remove_file(url);
}


void
BBSThread::rename_dat_file_for_backup(Repository &repository)
{
  char url[PATH_MAX];
  if (get_dat_url(url, PATH_MAX) == NULL)
    return;

  char dat_path[PATH_MAX];
  if (repository.get_bbs_data().find_file(url, dat_path, PATH_MAX) == NULL)
    {
      if (excavate_dat_file(repository, dat_path, PATH_MAX) == NULL)
	return;

      if (is_alive())
	{
	  string prev_path(dat_path);
	  repository.get_bbs_data().move_file(dat_path, url);
	  if (repository.get_bbs_data().find_file(url, dat_path, PATH_MAX) == NULL)
	    {
	      // ư˼Ԥ硣
	      snprintf(dat_path, PATH_MAX, "%s", prev_path.c_str());
	    }
	}
    }

  string dat_bak_url(url);
  dat_bak_url.append(".bak");
  repository.get_bbs_data().move_file(dat_path, dat_bak_url.c_str());
}


bool
BBSThread::parse_url(const char *url, int *from_p, int *to_p)
{
  assert(get_board() != NULL);

  int from = 0;
  int to = 0;
  BBSThread *target_thread
    = get_board()->lookup_bbs_thread_by_url(url, &from, &to);
  if (target_thread == this)
    {
      if (from_p != NULL)
	*from_p = from;
      if (to_p != NULL)
	*to_p = to;

      return true;
    }

  return false;
}


char *
BBSThread::get_url(int from, int to)
{
  return get_board()->get_url(this, from, to);
}


bool
BBSThread::is_trackback_supported()
{
  switch (get_board()->get_bbs_type())
    {
    case OCHUSHA_BBS_TYPE_2CH:
    case OCHUSHA_BBS_TYPE_2CH_BE:
    case OCHUSHA_BBS_TYPE_BBSPINK:
    case OCHUSHA_BBS_TYPE_JBBS_LIVEDOOR:
      return true;

    default:
      break;
    }

  return false;
}


char *
BBSThread::get_trackback_url(int res_num)
{
  switch (get_board()->get_bbs_type())
    {
    case OCHUSHA_BBS_TYPE_2CH:
    case OCHUSHA_BBS_TYPE_2CH_BE:
    case OCHUSHA_BBS_TYPE_BBSPINK:
      return BBS2ch::get_trackback_url(this, res_num);

    case OCHUSHA_BBS_TYPE_JBBS_LIVEDOOR:
      return BBSJbbsLivedoor::get_trackback_url(this, res_num);

    default:
      break;
    }

  return NULL;
}


bool
BBSThread::is_a_boned(unsigned int res_num)
{
  list<unsigned int>::const_iterator iter = a_bone_list.begin();
  list<unsigned int>::const_iterator end_iter = a_bone_list.end();
  while (iter != end_iter)
    if (*iter++ == res_num)
      return true;

  return false;
}


void
BBSThread::register_a_bone(unsigned int res_num)
{
  list<unsigned int>::iterator iter = a_bone_list.begin();
  list<unsigned int>::iterator end_iter = a_bone_list.end();
  while (iter != end_iter)
    {
      if (*iter == res_num)
	return;

      if (*iter > res_num)
	break;
      ++iter;
    }
  if (iter != end_iter)
    a_bone_list.insert(iter, res_num);
  else
    a_bone_list.push_back(res_num);

  get_board()->notify_major_modification();
}


void
BBSThread::unregister_a_bone(unsigned int res_num)
{
  a_bone_list.remove(res_num);

  get_board()->notify_major_modification();
}


time_t
BBSThread::get_last_modified_utc()
{
  if (last_modified != NULL)
    return get_utc_from_date(last_modified);

  return -1;
}


time_t
BBSThread::get_birthtime()
{
  // ߤΤȤIDbirthtimeˤʤäƤ褦ʷǼĤˤбƤʤ
  if (id != NULL)
    {
      long value;	// time_t32bitɡġ
      if (sscanf(id, "%ld", &value) == 1)
	return static_cast<time_t>(value);
    }

  return -1;
}


void
BBSThread::reset(Repository &repository)
{
  set_last_modified(NULL);
  set_date(NULL);
  set_kako_html(NULL);
  set_grave(NULL);
  set_number_of_responses_got(0);
  set_number_of_responses_read(0);
  set_dat_filesize(0);
  flags &= ~(DAT_VIEWER | KAKO_THREAD | STOPPED);
  ui_flags = 0;
  set_last_name(NULL);
  set_last_mail(NULL);
  remove_dat_file(repository);
}


void
BBSThread::read_threadlist_element(AttributeTable &attributes)
{
  set_last_modified(attributes.get_string("last_modified"));
  set_kako_html(attributes.get_string("kako_html"));
  set_grave(attributes.get_string("grave"));

  rank = attributes.get_int("rank");
  previous_rank = rank;
  number_of_responses_on_server
    = attributes.get_int("number_of_responses_on_server");

  if (attributes.lookup("number_of_responses_read") == NULL)
    {
      number_of_responses_got
	= attributes.get_int("number_of_responses_got");
      number_of_responses_read = number_of_responses_got;
    }
  else
    {
      number_of_responses_got
	= attributes.get_int("number_of_responses_read");
      number_of_responses_read = number_of_responses_got;
    }

  bookmark_position = attributes.get_int("bookmark_position");
  last_view_response = attributes.get_int("last_view_response");
  shown_count = attributes.get_int("shown_count");

  flags = attributes.get_int("flags");
  ui_flags = attributes.get_int("ui_flags");

  set_last_name(attributes.get_string("last_name"));
  set_last_mail(attributes.get_string("last_mail"));

  const char *a_bone_list = attributes.get_string("a_bone_list");
  if (a_bone_list != NULL)
    {
      const char *cur_pos = a_bone_list;
      while (*cur_pos != '\0')
	{
	  unsigned int res_num;
	  if (sscanf(cur_pos, "%u", &res_num) == 1)
	    register_a_bone(res_num);
	  cur_pos = strchr(cur_pos, ',');
	  if (cur_pos == NULL)
	    break;
	  ++cur_pos;
	}
    }

  a_bone.read_xml_elements(attributes);
  read_extended_threadlist_element(attributes);
}


#define OUTPUT_ATTRIBUTE_BOOLEAN(gzfile, attribute)			\
  do									\
    {									\
      if (attribute)							\
	gzprintf(gzfile,						\
		 "      <attribute name=\"" #attribute "\">\n"		\
		 "        <boolean val=\"%s\"/>\n"			\
		 "      </attribute>\n",				\
		 (attribute) ? "true" : "false");			\
    }									\
  while (0)

#define OUTPUT_ATTRIBUTE_INT(gzfile, attribute)				\
  do									\
    {									\
      if (attribute)							\
	gzprintf(gzfile,						\
		 "      <attribute name=\"" #attribute "\">\n"		\
		 "        <int val=\"%d\"/>\n"				\
		 "      </attribute>\n", attribute);			\
    }									\
  while (0)

#define OUTPUT_ATTRIBUTE_STRING(gzfile, attribute)			\
  do									\
    {									\
      if (attribute != NULL)						\
	{								\
	  string text;							\
	  gzprintf(gzfile,						\
		   "      <attribute name=\"" #attribute "\">\n"	\
		   "        <string>%s</string>\n"			\
		   "      </attribute>\n",				\
		   append_markup_escaped_text(text, attribute).c_str());\
	}								\
    }									\
  while (0)


void
BBSThread::write_threadlist_element(gzFile threadlist_xml)
{
  if (number_of_responses_got == 0 && (flags & DAT_DROPPED))
    return;	// DATեäƤʤDATϤ줤äѤ˺롣

  gzprintf(threadlist_xml, "    <thread>\n");

  OUTPUT_ATTRIBUTE_STRING(threadlist_xml, id);
  OUTPUT_ATTRIBUTE_STRING(threadlist_xml, title);
  OUTPUT_ATTRIBUTE_STRING(threadlist_xml, last_modified);

  OUTPUT_ATTRIBUTE_STRING(threadlist_xml, kako_html);
  OUTPUT_ATTRIBUTE_STRING(threadlist_xml, grave);

  OUTPUT_ATTRIBUTE_INT(threadlist_xml, rank);

  OUTPUT_ATTRIBUTE_INT(threadlist_xml, number_of_responses_on_server);
  OUTPUT_ATTRIBUTE_INT(threadlist_xml, number_of_responses_got);

  OUTPUT_ATTRIBUTE_INT(threadlist_xml, bookmark_position);
  OUTPUT_ATTRIBUTE_INT(threadlist_xml, last_view_response);
  OUTPUT_ATTRIBUTE_INT(threadlist_xml, shown_count);

  OUTPUT_ATTRIBUTE_INT(threadlist_xml, flags);
  OUTPUT_ATTRIBUTE_INT(threadlist_xml, ui_flags);

  OUTPUT_ATTRIBUTE_STRING(threadlist_xml, last_name);
  OUTPUT_ATTRIBUTE_STRING(threadlist_xml, last_mail);

  if (!a_bone_list.empty())
    {
      list<unsigned int>::const_iterator iter = a_bone_list.begin();
      list<unsigned int>::const_iterator end_iter = a_bone_list.end();

      string tmp_string;
      char tmp_buf[64];
      while (iter != end_iter)
	{
	  snprintf(tmp_buf, 64, "%u", *iter);
	  if (iter == a_bone_list.begin())
	    tmp_string.append(tmp_buf);
	  else
	    tmp_string.append(",").append(tmp_buf);
	  ++iter;
	}
      const char *a_bone_list = tmp_string.c_str();
      OUTPUT_ATTRIBUTE_STRING(threadlist_xml, a_bone_list);
    }

  a_bone.write_xml_elements(threadlist_xml, "      ");
  write_extended_threadlist_element(threadlist_xml);

  gzprintf(threadlist_xml, "    </thread>\n");
} 


void
BBSThread::read_extended_threadlist_element(AttributeTable &attributes)
{
}


void
BBSThread::write_extended_threadlist_element(gzFile threadlist_xml) const
{
}


}
