/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: bbs_utils.cc,v 1.3 2009/01/01 07:46:43 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/bbs_utils.h"

#include "libochusha/bulletin_board.h"
#include "libochusha/configuration.h"
#include "libochusha/html_utils.h"
#include "libochusha/misc_utils.h"

#include <string>

#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <cstring>


namespace Ochusha
{

PostResponseAgent::PostResponseAgent(BulletinBoard *board,
				     Configuration &config,
				     NetworkAgent *agent)
  : board(board),
    thread(NULL),
    agent(agent),
    config(config)
{
}


PostResponseAgent::PostResponseAgent(BBSThread *thread, Configuration &config,
				     NetworkAgent *agent)
  : board(NULL),
    thread(thread),
    agent(agent),
    config(config)
{
  if (thread != NULL)
    board = thread->get_board();
}


PostResponseAgent::~PostResponseAgent()
{
#ifdef VERBOSE_DESTRUCTOR
  fprintf(stderr, "PostResponseAgent(@%p) destructed\n", this);
#endif
}


BulletinBoard *
PostResponseAgent::get_board()
{
  return dynamic_cast<BulletinBoard *>(static_cast<RefCount *>(board));
}


BBSThread *
PostResponseAgent::get_thread()
{
  return dynamic_cast<BBSThread *>(static_cast<RefCount *>(thread));
}


NetworkAgent *
PostResponseAgent::get_agent()
{
  return agent;
}


void
PostResponseAgent::set_2ch_be_session_id(const char *id)
{
}


void
PostResponseAgent::set_2ch_viewer_session_id(const char *id)
{
}


BoardURLSeeker::BoardURLSeeker(BulletinBoard *board, Configuration &config,
			       NetworkAgent *agent)
  : board(board),
    agent(agent),
    board_url(NULL),
    config(config)
{
  agent->prepare_request.connect(this, &BoardURLSeeker::prepare_request_cb);
  agent->response_header_received.connect(this, &BoardURLSeeker::response_header_received_cb);
  agent->response_body_received.connect(this, &BoardURLSeeker::response_body_received_cb);
  agent->access_finished.connect(this, &BoardURLSeeker::access_finished_cb);
  agent->access_failed.connect(this, &BoardURLSeeker::access_failed_cb);
  agent->access_terminated.connect(this, &BoardURLSeeker::access_terminated_cb);
}


BoardURLSeeker::~BoardURLSeeker()
{
  if (board_url != NULL)
    free(board_url);

#ifdef VERBOSE_DESTRUCTOR
  fprintf(stderr, "BoardURLSeeker destructed.\n");
#endif
}


void
BoardURLSeeker::prepare_request_cb(NetworkAgent *agent, RefCount *data)
{
  agent->set_header(HTTPHeader::Accept_Encoding, "deflate, gzip");
  agent->set_header(HTTPHeader::User_Agent, OCHUSHA_USER_AGENT);

  if (config.get_enable_proxy_to_read())
    agent->set_proxy_uri(config.get_proxy_uri_to_read());
}


void
BoardURLSeeker::response_header_received_cb(NetworkAgent *agent, RefCount *data)
{
  // LocationإåФ򻲾
  const char *value = agent->get_header(HTTPHeader::Location);
  if (value == NULL)
    return;

  set_board_url(value);
}


class LocationExtractor : public HTMLSAXParser
{
private:
  string script_text;
  bool in_script_element;


protected:
  virtual void start_element(const char *name, const char *const attrs[]);
  virtual void end_element(const char *name);
  virtual void characters(const char *ch, size_t len);


public:
  LocationExtractor() : in_script_element(false) {}
  virtual ~LocationExtractor() {}

  inline const char *get_script_text() { return script_text.c_str(); }
};


void
LocationExtractor::start_element(const char *name, const char *const attrs[])
{
  if (strcasecmp(name, "script") == 0)
    {
      in_script_element = true;
      return;
    }

  return;
}


void
LocationExtractor::end_element(const char *name)
{
  if (strcasecmp(name, "script") == 0)
    in_script_element = false;
}


void
LocationExtractor::characters(const char *ch, size_t len)
{
  if (in_script_element)
    script_text.append(ch, len);
}


void
BoardURLSeeker::response_body_received_cb(NetworkAgent *agent, RefCount *data)
{
  const char *value = agent->get_header(HTTPHeader::Content_Type);
  if (value == NULL || strstr(value, "text/html") == NULL)
    return;

  LocationExtractor extractor;

  Buffer *buffer = agent->get_buffer();

  int lock_id = buffer->get_lock()->rdlock();

  extractor.parse_text(buffer->get_buffer(lock_id),
		       buffer->get_length(lock_id));

  buffer->get_lock()->unlock(lock_id);

  const char *script = extractor.get_script_text();

  // ʡad-hocjavascriptȲ᤹ΤݤʤΤǡġ
  const char *tmp_p = strstr(script, "location.href=\"");
  if (tmp_p == NULL)
    return;

  const char *url_p = tmp_p + 15;	// strlen("location.href=\"") == 15
  const char *dq_p = strchr(url_p, '"');
  if (dq_p == NULL || url_p == dq_p)
    return;

  string url(url_p, dq_p - url_p);
  set_board_url(url.c_str());
}


void
BoardURLSeeker::access_finished_cb(NetworkAgent *agent, RefCount *data)
{
  if (board_url != NULL)
    url_found(this, board_url);
  else
    seek_failed(this);
}


void
BoardURLSeeker::access_failed_cb(NetworkAgent *agent, const char *reason,
				 RefCount *data)
{
  seek_failed(this);
}


void
BoardURLSeeker::access_terminated_cb(NetworkAgent *agent, RefCount *data)
{
  seek_failed(this);
}


BulletinBoard *
BoardURLSeeker::get_board()
{
  return dynamic_cast<BulletinBoard *>(static_cast<RefCount *>(board));
}


NetworkAgent *
BoardURLSeeker::get_agent()
{
  return agent;
}


void
BoardURLSeeker::set_board_url(const char *url)
{
  if (board_url != NULL && url != NULL && strcmp(board_url, url) == 0)
    return;

  if (board_url != NULL)
    free(board_url);

  board_url = url != NULL ? strdup(url) : NULL;
}


void
BoardURLSeeker::seek()
{
  agent->get(get_board()->get_base_url(), new Buffer(), this);
}


}
