/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: bookmarks.cc,v 1.16 2009/01/01 07:46:43 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/bookmarks.h"

#include "libochusha/constants.h"
#include "libochusha/misc_utils.h"

#include <cstring>
#include <zlib.h>


namespace Ochusha
{

#define OUTPUT_ATTRIBUTE_BOOLEAN(gzfile, attribute, ind)		\
  do									\
    {									\
      if (attribute)							\
	gzprintf(gzfile,						\
		 "%s<attribute name=\"" #attribute "\">\n"		\
		 "%s  <boolean val=\"true\"/>\n"			\
		 "%s</attribute>\n", ind, ind, ind);			\
    }									\
  while (0)

#define OUTPUT_ATTRIBUTE_INT(gzfile, attribute, ind)			\
  do									\
    {									\
      if (attribute)							\
	gzprintf(gzfile,						\
		 "%s<attribute name=\"" #attribute "\">\n"		\
		 "%s  <int val=\"%d\"/>\n"				\
		 "%s</attribute>\n", ind, ind, attribute, ind);		\
    }									\
  while (0)

#define OUTPUT_ATTRIBUTE_STRING(gzfile, attribute, ind)			\
  do									\
    {									\
      if (attribute != NULL)						\
	{								\
	  string text;							\
	  gzprintf(gzfile,						\
		   "%s<attribute name=\"" #attribute "\">\n"		\
		   "%s  <string>%s</string>\n"				\
		   "%s</attribute>\n", ind, ind,			\
		   append_markup_escaped_text(text, attribute).c_str(),	\
		   ind);						\
	}								\
    }									\
  while (0)


BookmarkEntry::BookmarkEntry(const char *name, const char *url,
			     time_t tc, time_t tlm)
  : TreeElement(name, tc, tlm),
    url(NULL)
{
  set_url(url);
}


BookmarkEntry::BookmarkEntry(AttributeTable &attributes,
			     time_t tc, time_t tlm)
  : TreeElement(attributes, tc, tlm),
    url(NULL)
{
  set_url(attributes.get_string("url"));
}


BookmarkEntry::~BookmarkEntry()
{
  if (url != NULL)
    free(url);
}


void
BookmarkEntry::set_url(const char *url)
{
  if (this->url != NULL)
    free(this->url);

  this->url = url != NULL ? strdup(url) : NULL;
}


bool
BookmarkEntry::is_folder() const
{
  return false;
}


const char *
BookmarkEntry::get_element_name() const
{
  return "entry";
}


void
BookmarkEntry::read_extended_tree_element(AttributeTable &attributes)
{
  set_url(attributes.get_string("url"));
}


void
BookmarkEntry::write_extended_tree_element(WriteTree &args)
{
  const char *indent = args.indent->c_str();
  OUTPUT_ATTRIBUTE_STRING(args.tree_xml, url, indent);
}


Bookmarks::Bookmarks()
  : RootFolder(N_("Bookmarks"))
{
}


TreeElement *
Bookmarks::copy(const TreeElement *element)
{
  assert(!element->is_folder());
  assert(element->get_root_folder() == this);

  const BookmarkEntry *entry = dynamic_cast<const BookmarkEntry *>(element);
  assert(entry != NULL);

  BookmarkEntry *result = new BookmarkEntry(entry->get_name(), entry->get_url(),
					    0, 0);
  result->set_description(entry->get_description());

  return result;
}


bool
Bookmarks::read_bookmarks_xml(Repository &repository)
{
  return read_tree_xml(repository, OCHUSHA_BOOKMARKS_XML);
}


bool
Bookmarks::write_bookmarks_xml(Repository &repository)
{
  return write_tree_xml(repository, OCHUSHA_BOOKMARKS_XML);
}


const char *
Bookmarks::get_root_element_name() const
{
  return "bookmarks";
}


const char *
Bookmarks::get_leaf_element_name() const
{
  return "entry";
}


Folder *
Bookmarks::folder_new(time_t tc, time_t tlm, unsigned int sn)
{
  return new Folder(FOLDER_TYPE_BOOKMARKS_FOLDER, NULL, tc, tlm, sn);
}


Folder *
Bookmarks::folder_new(const Folder *folder, unsigned int sn)
{
  return new Folder(folder, sn);
}


TreeElement *
Bookmarks::leaf_element_new(AttributeTable &attributes,
			    time_t tc, time_t tlm)
{
  return new BookmarkEntry(attributes, tc, tlm);
}

}
