/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: buffer.cc,v 1.23 2009/01/01 07:46:43 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/buffer.h"

#include "libochusha/misc_utils.h"


#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#ifdef HAVE_MMAP
# ifdef HAVE_SYS_MMAN_H
#  include <sys/mman.h>
# endif

# ifndef MAP_NOCORE
#  define MAP_NOCORE 0
# endif
# ifndef MAP_NOSYNC
#  define MAP_NOSYNC 0
# endif
#endif	/* HAVE_MMAP */

#ifdef HAVE_SYS_UIO_H
# include <sys/uio.h>
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif

#include <zlib.h>


namespace Ochusha
{

#define INITIAL_CHUNK_SIZE	0x1000		// 4K
#define MAXIMUM_CHUNK_SIZE	0x10000		// 64K


Buffer::Buffer(size_t buffer_len)
  : fd(-1),
    gzfile(NULL),
    lock(LockManager::rwlock_new())
{
  // եʤ
  if (buffer_len > 0)
    buffer = static_cast<char *>(malloc(buffer_len));
  else
    buffer = NULL;

  if (buffer != NULL)
    buffer_length = buffer_len;
  else
    buffer_length = 0;

  data_length = 0;

  fixed = false;
  mmapped = false;
}


// 񤭹ѥХåե롣
Buffer::Buffer(int fd, bool gzip)
  : fd(fd),
    gzfile(NULL),
    lock(LockManager::rwlock_new())
{
  if (fd >= 0)
    {
      if (gzip)
	{
	  gzfile = gzdopen(fd, "wb");
	  if (gzfile != NULL)
	    this->fd = -1;
	}
    }
  buffer = NULL;
  buffer_length = 0;

  data_length = 0;

  fixed = false;
  mmapped = false;
}


// gzipʽ񤭹ѥХåե롣
Buffer::Buffer(gzFile gzfile)
  : fd(-1),
    gzfile(gzfile),
    lock(LockManager::rwlock_new())
{
  assert(gzfile != NULL);
  buffer = NULL;
  buffer_length = 0;

  data_length = 0;

  fixed = false;
  mmapped = false;
}


Buffer::~Buffer()
{
  if (fd >= 0)
    close(fd);
  if (gzfile != NULL)
    gzclose(gzfile);

#ifdef HAVE_MMAP
  if (mmapped)
    munmap(buffer, buffer_length);
  else
#endif
  if (buffer != NULL)
    free(buffer);
}


// lock_idreader/writeråɬפǤ뤳ȤŤ˼ĥƤ
size_t
Buffer::get_length(int lock_id)
{
  return data_length;
}


// lock_idreader/writeråɬפǤ뤳ȤŤ˼ĥƤ
const char *
Buffer::get_buffer(int lock_id)
{
  if (fd >= 0 || gzfile != NULL)
    return NULL;	// 񤭹ѥХåե

  return buffer;
}


void
Buffer::fix()
{
  fixed = true;
  if (fd >= 0)
    {
      close(fd);
      fd = -1;
    }
  else if (gzfile != NULL)
    {
      gzclose(gzfile);
      gzfile = NULL;
    }
}


bool
Buffer::read_file(int fd)
{
  if (fd < 0)
    return false;

  assert(!fixed);

  gzFile gzfile = gzdopen(fd, "rb");
  if (gzfile == NULL)
    {
      close(fd);
      return false;
    }

  size_t chunk_size = INITIAL_CHUNK_SIZE;
  int result = 0;

  while (!gzeof(gzfile))
    {
      size_t available_space = ensure_free_space(chunk_size, 0);
      if (available_space < chunk_size)
	{
	  result = -1;
	  break;
	}

      chunk_size = MIN(available_space, MAXIMUM_CHUNK_SIZE);
      result = gzread(gzfile, buffer + data_length, chunk_size);
      if (result == -1)
	break;

      int lock_id = lock->wrlock();
      data_length += result;
      lock->unlock(lock_id);
    }

  gzclose(gzfile);

  return result >= 0;
}


#ifdef HAVE_MMAP

#define GZIP_HEADER_RESERVED	0xe0	/* bits 5..7: reserved */

static bool
is_gzip_file(int fd)
{
  off_t len = lseek(fd, 0, SEEK_END);

  if (lseek(fd, 0, SEEK_SET) == 0)
    {
      if (len < 10)
	return false;	// ե뾮
      else
	{
	  unsigned char header[4];
	  if (pread(fd, &header, 4, 0) == 4)
	    {
	      if (header[0] == 0x1f && header[1] == 0x8b
		  && header[2] == Z_DEFLATED
		  && (header[3] & GZIP_HEADER_RESERVED) == 0)
		return true;
	    }
	}
    }

  return false;
}
#endif


bool
Buffer::map_file(int fd)
{
  if (fixed)
    return false;

  if (this->fd >= 0 || gzfile != NULL)
    return false;

  int lock_id = lock->wrlock();

  data_length = 0;

  bool result;
#ifdef HAVE_MMAP
  if (is_gzip_file(fd))
    {
      lock->unlock(lock_id);
      result = read_file(fd);
    }
  else
    {
      if (buffer != NULL)
	free(buffer);
      buffer_length = static_cast<size_t>(lseek(fd, 0, SEEK_END));
      buffer = static_cast<char *>(mmap(NULL, buffer_length, PROT_READ,
					MAP_PRIVATE | MAP_NOCORE | MAP_NOSYNC,
					fd, 0));
      if (buffer != MAP_FAILED)
	{
	  data_length = buffer_length;
	  mmapped = true;
	  result = true;
	}
      else
	{
	  buffer_length = 0;
	  data_length = 0;
	  buffer = NULL;
	  result = false;
	}
      lock->unlock(lock_id);
      close(fd);
    }
#else
  lock->unlock(lock_id);
  result = read_file(fd);
#endif
  fixed = true;

  return result;
}


bool
Buffer::write_file(int fd, bool gzip)
{
  if (!fixed)
    return false;

  if (this->fd >= 0 || this->gzfile != NULL)
    return false;	// 񤭹ѥХåե

  if (fd < 0)
    return false;

  bool result;

  int lock_id = lock->rdlock();

  if (gzip)
    {
      int len = 0;
      gzFile gzfile = gzdopen(fd, "wb");
      if (gzfile != NULL)
	{
	  len = gzwrite(gzfile,
			const_cast<char *>(get_buffer(lock_id)),
			get_length(lock_id));
	  gzclose(gzfile);
	}
      else
	close(fd);

      result = len != 0;
    }
  else
    {
      size_t buf_len = get_length(lock_id);
      ssize_t len = write(fd,
			  const_cast<char *>(get_buffer(lock_id)), buf_len);
      close(fd);
      result = (static_cast<size_t>(len) == buf_len);
    }

  lock->unlock(lock_id);

  return result;
}

bool
Buffer::write_file(gzFile gzfile)
{
  if (!fixed)
    return false;

  if (this->fd >= 0 || this->gzfile != NULL)
    return false;	// 񤭹ѥХåե

  if (gzfile == NULL)
    return false;

  bool result;

  int lock_id = lock->rdlock();

  int len = gzwrite(gzfile,
		    const_cast<char *>(get_buffer(lock_id)),
		    get_length(lock_id));
  gzclose(gzfile);
  result = len != 0;

  lock->unlock(lock_id);

  return result;
}


/*
 * writer API
 */
// Хåեˤ롣
bool
Buffer::clear(int lock_id)
{
  if (fd >= 0 || gzfile != NULL)
    {
      cerr << "Clearing write only buffer isn't supported.\n";
      return false;
    }

  if (data_length == 0)
    return true;	// 

  if (lock_id != 0)
    {
      data_length = 0;
      return true;
    }

  lock_id = lock->wrlock();
  data_length = 0;
  lock->unlock(lock_id);

  return true;
}


// Хåե˥ǡɵ롣
bool
Buffer::append_data(const char *data, size_t length, int lock_id)
{
  if (data == NULL)
    return false;

  if (length == 0)
    return true;

  if (fd >= 0)
    {
      // 񤭹ѤʤΤǥå̵롣
      bool result;
      result = write(fd, data, length) != -1;
      if (result)
	data_length += length;
      return result;
    }

  if (gzfile != NULL)
    {
      // 񤭹ѤʤΤǥå̵롣
      bool result = static_cast<size_t>(gzwrite(gzfile,
						const_cast<char *>(data),
						length)) != length;
      if (result)
	data_length += length;
      return result;
    }

  if (ensure_free_space(length, lock_id) < length)
    return false;

  memcpy(buffer + data_length, data, length);

  if (lock_id != 0)
    {
      data_length += length;
      return true;
    }

  lock_id = lock->wrlock();
  data_length += length;
  lock->unlock(lock_id);

  return true;
}


// Хåեΰ̣ͭǡĹ򥻥åȤ롣
bool
Buffer::set_length(size_t length, int lock_id)
{
  if (fd >= 0 || gzfile != NULL)
    {
      cerr << "Setting length of write only buffer isn't supported.\n";
      return false;
    }

  if (buffer_length < length)
    return false;

  if (lock_id != 0)
    {
      data_length = length;
      return true;
    }

  lock_id = lock->wrlock();
  data_length = length;
  lock->unlock(lock_id);

  return true;
}


size_t
Buffer::ensure_free_space(size_t length, int lock_id)
{
  size_t avail_len = buffer_length - data_length;
  if (avail_len >= length)
    return avail_len;

  size_t new_buf_len
    = buffer_length > 0 ? buffer_length * 2 : DEFAULT_BUFFER_LENGTH;

  avail_len = new_buf_len - data_length;
  while (avail_len < length)
    {
      new_buf_len *= 2;
      avail_len = new_buf_len - data_length;
    }

  if (lock_id != 0)
    {
      char *old_buf = static_cast<char *>(buffer);
      buffer = static_cast<char *>(realloc(buffer, new_buf_len));
      if (buffer != NULL)
	buffer_length = new_buf_len;
      else
	buffer = old_buf;	// Out of Memory

      return buffer_length - data_length;
    }

  lock_id = lock->wrlock();
  char *old_buf = static_cast<char *>(buffer);
  buffer = static_cast<char *>(realloc(buffer, new_buf_len));
  if (buffer != NULL)
    buffer_length = new_buf_len;
  else
    buffer = old_buf;	// Out of Memory
  lock->unlock(lock_id);

  return buffer_length - data_length;
}


}

