/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: bulletin_board.cc,v 1.60 2009/01/01 07:46:43 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/bulletin_board.h"

#include "libochusha/bbs_2ch_utils.h"
#include "libochusha/bbs_jbbs_utils.h"
#include "libochusha/bbs_table.h"
#include "libochusha/bookmarks.h"
#include "libochusha/configuration.h"
#include "libochusha/constants.h"
#include "libochusha/misc_utils.h"
#include "libochusha/regexp.h"
#include "libochusha/repository.h"
#include "libochusha/sax_context.h"
#include "libochusha/uri_utils.h"

#include <climits>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <vector>
#include <string>

#include <zlib.h>


namespace Ochusha
{


// #define VERBOSE_DESTRUCTOR	1
// #define DEBUG_SAX_HANDLER	1
static Lock bulletin_board_giant("BulletinBoard");


ThreadlistUpdater::ThreadlistUpdater(BulletinBoard *board,
				     Configuration &config,
				     Repository &repository,
				     NetworkAgent *agent)
  : board(board),
    base_uri(board->get_base_url()),
    config(config),
    repository(repository),
    agent(agent),
    lock(LockManager::rwlock_new())
{
  assert(libochusha_giant.is_locked_by_self());
  assert(bulletin_board_giant.is_locked_by_self());

  board->updater = this;
}


ThreadlistUpdater::~ThreadlistUpdater()
{
#ifdef VERBOSE_DESTRUCTOR
  cerr << "ThreadlistUpdater destructed\n";
#endif
}


void
ThreadlistUpdater::recover_threadlist()
{
  int lock_id = lock->wrlock();
  board->recover_threadlist(this);
  lock->unlock(lock_id);
}


void
ThreadlistUpdater::update_threadlist()
{
  int lock_id = lock->wrlock();
  board->update_threadlist(this);
  lock->unlock(lock_id);
}


void
ThreadlistUpdater::finalize()
{
  assert(!bulletin_board_giant.is_locked_by_self());


  bulletin_board_giant.lock();

  board->updater = NULL;

  bulletin_board_giant.unlock();


  agent = NULL;
  disconnect_all();
}


// ʤ(⤷ϤǤʤ)
class DummyUpdater : public ThreadlistUpdater
{
public:
  DummyUpdater(BulletinBoard *board, Configuration &config,
	       Repository &repository, NetworkAgent *agent);
  virtual ~DummyUpdater();
};


DummyUpdater::DummyUpdater(BulletinBoard *board, Configuration &config,
			   Repository &repository, NetworkAgent *agent)
  : ThreadlistUpdater(board, config, repository, agent)
{
  recover_threadlist();
}


DummyUpdater::~DummyUpdater()
{
#ifdef VERBOSE_DESTRUCTOR
  cerr << "DummyUpdater destructed\n";
#endif
}


class ThreadEnumeratorJob : public WorkerJob, public has_slots
{
private:
  smart_ptr<InterboardThreadlistUpdater> updater;

  Monitor monitor;

  bool failed;
  bool terminated;

  void state_changed_cb();

  // ԷϤΤߴƻ
  void access_failed_cb();
  void access_terminated_cb();


public:
  ThreadEnumeratorJob(InterboardThreadlistUpdater *updater);
  virtual void work(const WorkerThread *employee);
  virtual const char *description() const
  { return "enemerating BBS threads"; }
};


InterboardThreadlistUpdater::InterboardThreadlistUpdater(BulletinBoard *board,
							 Configuration &config,
							 Repository &repository,
							 ThreadPool &pool,
							 NetworkAgent *agent,
							 bool update)
  : ThreadlistUpdater(board, config, repository, agent),
    pool(pool),
    update(update)
{
  assert(board->get_bbs_type() < OCHUSHA_BBS_TYPE_UNKNOWN);
}


InterboardThreadlistUpdater::~InterboardThreadlistUpdater()
{
#ifdef VERBOSE_DESTRUCTOR
  cerr << "InterboardThreadlistUpdater destructed\n";
#endif
}


ThreadlistIterator *
InterboardThreadlistUpdater::get_threadlist(BulletinBoard *board,
					    bool no_update)
{
  time_t previous_time = board->previous_access_time;
  time_t current_time = time(NULL);
  if (previous_time + MINIMUM_BOARD_UPDATE_INTERVAL < current_time)
    board->previous_access_time = current_time;
  else
    no_update = true;


  LockBlock global_lock(libochusha_giant);
  LockBlock local_lock(bulletin_board_giant);

  return board->get_threadlist(config, repository, pool, agent,
			       update && !no_update);
}


void
InterboardThreadlistUpdater::notify_started(BulletinBoard *board)
{
  if (agent != NULL && !update)
    agent->access_started(agent, agent->get_user_data());
}


void
InterboardThreadlistUpdater::notify_finished(bool succeeded)
{
  bulletin_board_giant.lock();

  update_threadlist();

  bulletin_board_giant.unlock();


  if (agent != NULL && !update)
    {
      if (succeeded)
	agent->access_finished(agent, agent->get_user_data());
      else
	agent->access_failed(agent, LIB_("Unknown reason"),
			     agent->get_user_data());
    }


  finalize();
}


void
InterboardThreadlistUpdater::add_board_to_be_updated(BulletinBoard *board_to_be_updated)
{
  assert(static_cast<WorkerJob *>(job) == NULL);

  if (board_to_be_updated->get_bbs_type() <= OCHUSHA_BBS_TYPE_UNKNOWN)
    return;

  if (listed_boards.find(board_to_be_updated) == listed_boards.end())
    {
      boards_to_be_updated.push_back(board_to_be_updated);
      listed_boards.insert(board_to_be_updated);
    }
}


void
InterboardThreadlistUpdater::start()
{
  assert(static_cast<WorkerJob *>(job) == NULL);

  job = new ThreadEnumeratorJob(this);
  pool.commit(job, true);
}


bool
InterboardThreadlistUpdater::enumerate_threadlist(ThreadlistIterator *iterator)
{
  smart_ptr<ThreadlistIterator> iter = iterator;

  int num_iterated = 0;

  int rdlock_id = iter->lock();
  int wrlock_id = lock->wrlock();

  while (iter->has_more_threads(rdlock_id))
    {
      BBSThread *thread = iter->next_thread(rdlock_id);
      assert(thread != NULL);

      if (enumerated_threads.find(thread) == enumerated_threads.end()
	  && should_be_enumerated(thread))
	{
	  updated_list.push_back(thread);
	  enumerated_threads.insert(thread);
	  ++num_iterated;
	}
    }
  assert(iter->finished(rdlock_id));

  lock->unlock(wrlock_id);
  iter->unlock(rdlock_id);

  return num_iterated > 0;
}


ThreadEnumeratorJob::ThreadEnumeratorJob(InterboardThreadlistUpdater *updater)
  : updater(updater),
    monitor("ThreadEnumeratorJob"),
    failed(false),
    terminated(false)
{
#ifdef CHECK_LOCK_ORDER
  assert(libochusha_giant.is_locked_by_self());
  // "libochusha""ThreadEnumeratorJob"ȤåϿƤᡣ
  monitor.enter();
  monitor.exit();
#endif
}


void
ThreadEnumeratorJob::work(const WorkerThread *employee)
{
  BulletinBoardList::iterator iter = updater->boards_to_be_updated.begin();
  BulletinBoardList::iterator end_iter = updater->boards_to_be_updated.end();
  while (iter != end_iter)
    {
      if (employee->get_state() == WorkerThread::DIE_NOW)
	break;
      BulletinBoard *board = *iter++;

      updater->notify_started(board);

      smart_ptr<ThreadlistIterator> threadlist
	= updater->get_threadlist(board, terminated);

      if (threadlist != NULL)
	{
	  libochusha_giant.lock();

	  monitor.enter();

	  threadlist->state_changed
	    .connect(this, &ThreadEnumeratorJob::state_changed_cb);
	  threadlist->access_failed
	    .connect(this, &ThreadEnumeratorJob::access_failed_cb);
	  threadlist->access_terminated
	    .connect(this, &ThreadEnumeratorJob::access_terminated_cb);


	  libochusha_giant.unlock();


	  while (!threadlist->finished(0))	// lock_idϥ
	    monitor.wait();

	  bool changed = updater->enumerate_threadlist(threadlist);
	  monitor.exit();

	  if (changed)
	    updater->state_changed();
	}
    }

  updater->notify_finished(!failed);
}


void
ThreadEnumeratorJob::state_changed_cb()
{
  monitor.enter();
  monitor.notify();
  monitor.exit();
}


void
ThreadEnumeratorJob::access_failed_cb()
{
  failed = true;
}


void
ThreadEnumeratorJob::access_terminated_cb()
{
  terminated = true;
}


class AllThreadEnumerator : public InterboardThreadlistUpdater
{
protected:
  virtual bool should_be_enumerated(BBSThread *thread);


public:
  AllThreadEnumerator(BulletinBoard *board, Configuration &config,
		      Repository &repository, ThreadPool &pool,
		      NetworkAgent *agent);
};


AllThreadEnumerator::AllThreadEnumerator(BulletinBoard *board,
					 Configuration &config,
					 Repository &repository,
					 ThreadPool &pool,
					 NetworkAgent *agent)
  : InterboardThreadlistUpdater(board, config, repository, pool, agent, false)
{
  assert(board->get_bbs_type() == OCHUSHA_BBS_TYPE_PSEUDO_FOLDER);

#ifndef NDEBUG
  char *algorithm
    = BulletinBoard::get_algorithm_from_pseudo_folder_url(board->get_base_url());
  assert(algorithm != NULL);
  assert(strcmp(algorithm, "all-threads") == 0);
  free(algorithm);
#endif

  RootFolder *root = board->get_root_folder();
  assert(root != NULL);
  BBSTable *table = dynamic_cast<BBSTable *>(root);
  assert(table != NULL);

  vector<TreeElementHandle>::iterator iter = table->all_boards_begin();
  vector<TreeElementHandle>::iterator end_iter = table->all_boards_end();
  while (iter != end_iter)
    {
      TreeElement *element = *iter++;
      BulletinBoard *brd = dynamic_cast<BulletinBoard *>(element);
      if (brd != NULL && brd->get_bbs_type() >= OCHUSHA_BBS_TYPE_2CH)
	add_board_to_be_updated(brd);
    }

  start();
}


bool
AllThreadEnumerator::should_be_enumerated(BBSThread *thread)
{
  return thread->get_number_of_responses_got() > 0;
}


class BookmarkFolderUpdater : public InterboardThreadlistUpdater
{
private:
  set<BBSThread *> all_threads;
  set<BulletinBoard *> all_boards;

  // should_be_enumeratedtrue֤ˤϿ֤ɤ
  set<BBSThread *> listed_threads;

  void setup_for_update(BBSTable *table, BookmarkEntry *entry);
  void setup_for_update(BBSTable *table, Folder *folder);


protected:
  virtual bool should_be_enumerated(BBSThread *thread);


public:
  BookmarkFolderUpdater(BulletinBoard *board, Configuration &config,
			Repository &repository, ThreadPool &pool,
			NetworkAgent *agent, bool update);
};


BookmarkFolderUpdater::BookmarkFolderUpdater(BulletinBoard *board,
					     Configuration &config,
					     Repository &repository,
					     ThreadPool &pool,
					     NetworkAgent *agent, bool update)
  : InterboardThreadlistUpdater(board, config, repository, pool, agent,
				update && (board->get_bbs_type()
					   != OCHUSHA_BBS_TYPE_PASSIVE_FOLDER))
{
  /*
   * MEMO: 졢PSEUDO_FOLDERĤalgorithmȤ
   *       "threads-recently-read""threads-recently-posted"Τ褦
   *       Τ硢Ĥ򤽤ΤޤήѤ褦ʴ
   */
  assert(board->get_bbs_type() == OCHUSHA_BBS_TYPE_ACTIVE_FOLDER
	 || board->get_bbs_type() == OCHUSHA_BBS_TYPE_PASSIVE_FOLDER);

  RootFolder *root = board->get_root_folder();
  assert(root != NULL);
  BBSTable *table = dynamic_cast<BBSTable *>(root);
  assert(table != NULL);

  Folder *folder = board->get_correspond_folder();
  if (folder != NULL)
    setup_for_update(table, folder);

  start();
}


void
BookmarkFolderUpdater::setup_for_update(BBSTable *table, BookmarkEntry *entry)
{
  if (entry == NULL)
    return;

  const char *url = entry->get_url();
  if (url == NULL)
    return;

  BulletinBoard *board = table->guess_board(url);
  if (board == NULL || board->get_bbs_type() <= OCHUSHA_BBS_TYPE_UNKNOWN)
    return;

  board->read_threadlist_xml(repository);
  add_board_to_be_updated(board);

  // entryURLؤΥ֥åޡʤall_boardsˤĤBulletinBoard *
  // URLؤΥ֥åޡʤall_threadsˤΥBBSThread *
  // 줾insert롣
  BBSURLInfo info(url,
		  board->get_bbs_type() != OCHUSHA_BBS_TYPE_JBBS_LIVEDOOR
		  ? 1 : 2);
  if (info.get_thread_id() == NULL)
    {
      all_boards.insert(board);
      if (strcmp(board->get_base_url(), url) != 0)
	entry->set_url(board->get_base_url());
      return;
    }

  BBSThread *thread = board->lookup_bbs_thread_by_url(url);
  if (thread != NULL)
    {
      all_threads.insert(thread);
      return;
    }
}


void
BookmarkFolderUpdater::setup_for_update(BBSTable *table, Folder *folder)
{
  if (folder == NULL)
    return;

  TreeElementList::const_iterator iter = folder->elements_begin();
  TreeElementList::const_iterator end_iter = folder->elements_end();
  while (iter != end_iter)
    {
      TreeElement *element = *iter++;
      if (element->is_folder())
	setup_for_update(table, dynamic_cast<Folder *>(element));
      else
	setup_for_update(table, dynamic_cast<BookmarkEntry *>(element));
    }
}


bool
BookmarkFolderUpdater::should_be_enumerated(BBSThread *thread)
{
  if (listed_threads.find(thread) != listed_threads.end())
    return false;

  if (all_boards.find(thread->get_board()) != all_boards.end()
      || all_threads.find(thread) != all_threads.end())
    {
      listed_threads.insert(thread);

      return true;
    }

  return false;
}


class UpdaterActivityWatcher : public NetworkActivityNotifier
{
private:
  ThreadlistIterator *iterator;
  bool failed;
  int magic;	// ǥХå


protected:
  virtual void access_started(NetworkAgent *agent);
  virtual void access_progressed(NetworkAgent *agent, size_t data_read,
				 uintmax_t total_len);
  virtual void access_finished(NetworkAgent *agent);
  virtual void access_failed(NetworkAgent *agent);
  virtual void access_terminated(NetworkAgent *agent);


public:
  UpdaterActivityWatcher(NetworkAgent *agent, ThreadlistIterator *iterator);
  virtual ~UpdaterActivityWatcher();
};


UpdaterActivityWatcher::UpdaterActivityWatcher(NetworkAgent *agent,
					       ThreadlistIterator *iterator)
  : NetworkActivityNotifier(agent),
    iterator(iterator),
    failed(false),
    magic(0x14142136)
{
}


UpdaterActivityWatcher::~UpdaterActivityWatcher()
{
  magic = 0;
}


void
UpdaterActivityWatcher::access_started(NetworkAgent *agent)
{
  assert(magic == 0x14142136);

  iterator->access_started();
}


void
UpdaterActivityWatcher::access_progressed(NetworkAgent *agent, size_t data_read,
					  uintmax_t total_len)
{
  assert(magic == 0x14142136);

  iterator->access_progressed();
}


void
UpdaterActivityWatcher::access_finished(NetworkAgent *agent)
{
  assert(magic == 0x14142136);

  if (!failed)
    iterator->access_finished();
  else
    iterator->access_failed();
}


void
UpdaterActivityWatcher::access_failed(NetworkAgent *agent)
{
  assert(magic == 0x14142136);

  failed = true;
  iterator->access_failed();
}


void
UpdaterActivityWatcher::access_terminated(NetworkAgent *agent)
{
  assert(magic == 0x14142136);

  failed = true;
  iterator->access_terminated();
}


ThreadlistIterator::ThreadlistIterator(ThreadlistUpdater *updater)
  : updater(updater),
    notifier(NULL),
    updated_list(updater->updated_list),
    index(0),
    error_occurred(false)
{
  assert(bulletin_board_giant.is_locked_by_self());

  NetworkAgent *agent = updater->agent;
  if (agent != NULL)
    notifier = new UpdaterActivityWatcher(agent, this);

  updater->state_changed.connect(this, &ThreadlistIterator::state_changed_cb);
  updater->board_moved.connect(this, &ThreadlistIterator::board_moved_cb);
  updater->error_occurred.connect(this, &ThreadlistIterator::error_occurred_cb);
}


ThreadlistIterator::~ThreadlistIterator()
{
#ifdef VERBOSE_DESTRUCTOR
  cerr << "ThreadlistIterator destructed\n";
#endif
  if (notifier != NULL)
    delete notifier;
}


bool
ThreadlistIterator::has_more_threads(int lock_id)
{
  BBSThreadVector::iterator iter = updated_list.begin();
  return (iter + index) < updated_list.end();
}


BBSThread *
ThreadlistIterator::next_thread(int lock_id)
{
  BBSThreadVector::iterator iter = updated_list.begin();
  if ((iter + index) < updated_list.end())
    return iter[index++];

  return NULL;
}


bool
ThreadlistIterator::finished(int lock_id)
{
  return updater->board->updater != updater;
}


NetworkAgent::NetworkAgentState
ThreadlistIterator::get_agent_state(int lock_id)
{
  if (notifier == NULL || finished(lock_id))
    return NetworkAgent::STATE_INITIAL;

  return notifier->get_agent_state();
}


void
ThreadlistIterator::state_changed_cb()
{
  state_changed();
}


void
ThreadlistIterator::board_moved_cb()
{
  board_moved();
}


void
ThreadlistIterator::error_occurred_cb()
{
  error_occurred = true;
}


DATFileExplorer::DATFileExplorer(BulletinBoard *board, Configuration &config,
				 Repository &repository)
  : ThreadlistUpdater(board, config, repository, NULL)
{
}


DATFileExplorer::~DATFileExplorer()
{
}


class DATFileExplorerImpl : public DATFileExplorer
{
  friend class DATFileExplorerJob;

private:
  smart_ptr<WorkerJob> job;


public:
  DATFileExplorerImpl(BulletinBoard *board, Configuration &config,
		      Repository &repository);
  virtual ~DATFileExplorerImpl();

  virtual bool explore(ThreadPool &thread_pool);

  void notify_dat_file_found(const char *id, const char *fullpath,
			     struct stat *sb);
  void notify_finished();
};


class DATFileExplorerJob : public WorkerJob
{
private:
  smart_ptr<DATFileExplorerImpl> impl;
  BulletinBoardHandle board;
  Repository &repository;

  URI base_uri;


  static void foreach_file(const char *fullpath, const char *filename,
			   struct stat *sb, void *user_data);
  void explore_for_server(const char *server);


public:
  DATFileExplorerJob(DATFileExplorerImpl *impl, BulletinBoard *board,
		     Repository &repository);
  virtual ~DATFileExplorerJob();

  virtual void work(const WorkerThread *employee);
  virtual const char *description() const
  { return "Exploring DAT cache directory to lookup DAT files stay alone."; }
};


DATFileExplorerImpl::DATFileExplorerImpl(BulletinBoard *board,
					 Configuration &config,
					 Repository &repository)
  : DATFileExplorer(board, config, repository)
{
  // ǰžꥹȤ򹹿Ƥ
  // ԹʻΤη̡boardlist.xmlľʤɡmovement_history
  // ƤƤ⡢$OCHUSHA_HOME/cache/ʲΥǥ쥯ȥ깽¤
  // СиΤ뻪ˤĤƤǤ롣
  URI base_uri(board->get_base_url());
  if (!base_uri.is_valid()
      || base_uri.get_scheme() == NULL || base_uri.get_path() == NULL)
    return;

  list<string> servers;
  BBSDataManager &bbs_data = repository.get_bbs_data();
  bbs_data.enumerate_servers(servers);

  list<string> matched_servers;
  char url[PATH_MAX];
  char dirname[PATH_MAX];
  while (!servers.empty())
    {
      string &server = servers.front();
      if (snprintf(url, PATH_MAX, "%s://%s%s",
		   base_uri.get_scheme(), server.c_str(), base_uri.get_path())
	  < PATH_MAX)
	{
	  if (bbs_data.find_directory(url, dirname, PATH_MAX) == dirname)
	    {
	      URI uri(url);
	      if (uri.is_valid())
		{
		  BulletinBoard *gb;
		  switch (board->get_bbs_type())
		    {
		    case OCHUSHA_BBS_TYPE_2CH:
		    case OCHUSHA_BBS_TYPE_2CH_HEADLINE:
		    case OCHUSHA_BBS_TYPE_2CH_BE:
		    case OCHUSHA_BBS_TYPE_BBSPINK:
		    case OCHUSHA_BBS_TYPE_BBSPINK_HEADLINE:
		      gb = BBS2ch::guess_board(uri);
		      break;

		    case OCHUSHA_BBS_TYPE_MACHIBBS:
		      gb = BBSMachiBBS::guess_board(uri);
		      break;

		    case OCHUSHA_BBS_TYPE_JBBS_LIVEDOOR:
		      gb = BBSJbbsLivedoor::guess_board(uri);
		      break;

		    default:
		      gb = NULL;
		      break;
		    }

		  if (gb == board)
		    matched_servers.push_front(server);
		}
	    }
	}
      servers.pop_front();
    }

  if (matched_servers.empty())
    return;


  list<char *> &history = board->get_movement_history();
  while (!matched_servers.empty())
    {
      string &server = matched_servers.front();

      if (strcmp(base_uri.get_server(), server.c_str()) != 0)
	{
	  // movement_historyΥ֤å
	  list<char *>::iterator end_iter = history.end();
	  list<char *>::iterator where = find_if(history.begin(), end_iter,
						 CStringEqual(server.c_str()));
	  if (where == end_iter)
	    history.push_back(strdup(server.c_str()));
	}

      matched_servers.pop_front();
    }
}


DATFileExplorerImpl::~DATFileExplorerImpl()
{
}


bool
DATFileExplorerImpl::explore(ThreadPool &thread_pool)
{
  LockBlock lock_obj(bulletin_board_giant);

  if (job != NULL)
    return false;

  job = new DATFileExplorerJob(this, board, repository);
  thread_pool.commit(job);
  return true;
}


void
DATFileExplorerImpl::notify_dat_file_found(const char *id, const char *fullpath,
					   struct stat *sb)
{
  // ޤidˤĤơboard->lookup_bbs_thread_by_id(id)롣
  // BBSThread֥ȤĤʤä⤷ϡɾ¸
  // ʤСDATե»ҤˤʤäDATեʤΤǡ
  // 롣
  BBSThread *thread = board->lookup_bbs_thread_by_id(id);
  if (thread != NULL
      && thread->get_title() != NULL && thread->get_title()[0] != '\0'
      && thread->get_number_of_responses_got() > 0
      && thread->get_last_modified() != NULL
      && (thread->is_alive()
	  || thread->get_grave() != NULL || thread->get_kako_html() != NULL))
    return;

  if (thread == NULL)
    thread = board->thread_new(id, NULL);

  bool recovering;
  if (thread->is_tmp_thread())
    {
      thread->set_dat_dropped(true);
      recovering = false;
    }
  else
    recovering = true;

  bool is_alive = thread->is_alive();

  thread->set_dat_dropped(true);
  thread->recover_information(config, repository);
  thread->set_dat_dropped(!is_alive);

  make_alive(thread);

  int lock_id = lock->wrlock();
  updated_list.push_back(thread);
  lock->unlock(lock_id);

  if (recovering)
    information_recovered(board, thread);
  else
    dat_file_found(board, thread);
}


void
DATFileExplorerImpl::notify_finished()
{
  bulletin_board_giant.lock();

  recover_threadlist();

  bulletin_board_giant.unlock();


  finalize();
  state_changed();

  job = NULL;

  finished(board);
}


DATFileExplorerJob::DATFileExplorerJob(DATFileExplorerImpl *impl,
				       BulletinBoard *board,
				       Repository &repository)
  : impl(impl),
    board(board),
    repository(repository),
    base_uri(board->get_base_url())
{
}


DATFileExplorerJob::~DATFileExplorerJob()
{
}


void
DATFileExplorerJob::foreach_file(const char *fullpath, const char *filename,
				 struct stat *sb, void *user_data)
{
  const char *suffix = strstr(filename, ".dat");
  if (suffix == NULL || suffix == filename
      || (strcmp(suffix, ".dat") != 0 && strcmp(suffix, ".dat.gz") != 0))
    return;

  DATFileExplorerJob *job = static_cast<DATFileExplorerJob *>(user_data);
  char *id = strndup(filename, suffix - filename);
  job->impl->notify_dat_file_found(id, fullpath, sb);
  free(id);
}


void
DATFileExplorerJob::explore_for_server(const char *server)
{
  if (server == NULL || *server == '\0')
    return;

  char url[PATH_MAX];
  if (snprintf(url, PATH_MAX, "%s://%s%sdat/",
	       base_uri.get_scheme(), server, base_uri.get_path()) < PATH_MAX)
    {
      char dirname[PATH_MAX];
      if (repository.get_bbs_data().find_directory(url, dirname, PATH_MAX)
	  == NULL)
	return;

      repository.foreach_file(dirname, foreach_file, this);
    }
}


void
DATFileExplorerJob::work(const WorkerThread *employee)
{
  vector<string> servers;

  list<char *>::const_iterator iter = board->movement_history_begin();
  list<char *>::const_iterator end_iter = board->movement_history_end();
  while (iter != end_iter)
    servers.push_back(string(*iter++));

  if (base_uri.is_valid() && base_uri.get_path() != NULL)
    {
      // boardθԻȲбDATե륭åΤ꤫ꡢ
      // ƤdatեˤĤimpl->notify_dat_file_found(id, fullpath, sb);
      vector<string>::const_iterator iter = servers.begin();
      vector<string>::const_iterator end_iter = servers.end();
      while (iter != end_iter)
	{
	  const string &server = *iter++;
	  explore_for_server(server.c_str());
	}

      explore_for_server(base_uri.get_server());
    }

  impl->notify_finished();
}


BulletinBoard::BulletinBoard(time_t tc, time_t tlm)
  : TreeElement(NULL, tc, tlm),
    previous_access_time(0),
    bbs_type(OCHUSHA_BBS_TYPE_UNKNOWN),
    base_url(NULL),
    last_modified(NULL),
    date(NULL),
    short_name(NULL),
    rule(NULL),
    ui_flags(0),
    last_name(NULL),
    last_mail(NULL),
    previous_server(NULL),
    dirtyness(NOT_DIRTY),
    threadlist_read(false),
    in_boardlist(false)
{
}


BulletinBoard::BulletinBoard(AttributeTable &attributes, time_t tc, time_t tlm)
  : TreeElement(attributes, tc, tlm),
    previous_access_time(0),
    bbs_type(OCHUSHA_BBS_TYPE_UNKNOWN),
    base_url(NULL),
    last_modified(NULL),
    date(NULL),
    short_name(NULL),
    rule(NULL),
    ui_flags(0),
    last_name(NULL),
    last_mail(NULL),
    previous_server(NULL),
    dirtyness(NOT_DIRTY),
    threadlist_read(false),
    in_boardlist(false)
{
  BulletinBoard::read_extended_tree_element(attributes);
}


BulletinBoard::~BulletinBoard()
{
  if (base_url != NULL)
    free(base_url);

  if (last_modified != NULL)
    free(last_modified);

  if (date != NULL)
    free(date);

  if (short_name != NULL)
    free(short_name);

  if (rule != NULL)
    free(rule);

  if (last_name != NULL)
    free(last_name);

  if (last_mail != NULL)
    free(last_mail);

  if (previous_server != NULL)
    free(previous_server);

  while (!movement_history.empty())
    {
      char *server = movement_history.front();
      free(server);
      movement_history.pop_front();
    }
}


PostResponseAgent *
BulletinBoard::get_new_thread_agent(Configuration &config,
				    NetworkAgentBroker &broker)
{
  if (config.get_offline())
    return NULL;

  switch (get_bbs_type())
    {
    case OCHUSHA_BBS_TYPE_2CH:
    case OCHUSHA_BBS_TYPE_2CHLIKE_EUCJP:
    case OCHUSHA_BBS_TYPE_2CH_COMPATIBLE:
    case OCHUSHA_BBS_TYPE_2CH_BE:
    case OCHUSHA_BBS_TYPE_BBSPINK:
      return new PostResponseAgent2ch(this, config, broker.employ_agent());

    case OCHUSHA_BBS_TYPE_MACHIBBS:
    case OCHUSHA_BBS_TYPE_JBBS_LIVEDOOR:
      return new PostResponseAgentJbbs(this, config, broker.employ_agent());

    default:
#if 0
    case OCHUSHA_BBS_TYPE_JBBS:
    case OCHUSHA_BBS_TYPE_MITINOKU:
#endif
    case OCHUSHA_BBS_TYPE_2CH_HEADLINE:
    case OCHUSHA_BBS_TYPE_BBSPINK_HEADLINE:
    case OCHUSHA_BBS_TYPE_2CH_COMPATIBLE_RO:
      break;
    }
  return NULL;
}


BBSThread *
BulletinBoard::lookup_bbs_thread_by_id(const char *id)
{
  BBSThread *thread = dynamic_cast<BBSThread *>(thread_table.lookup(id));
  if (thread != NULL)
    return thread;

  // idбBBSThread֥Ȥʤäȶäݤ֤ۤ
  return thread_new(id, NULL);
}


BBSThread *
BulletinBoard::lookup_bbs_thread_by_url(const char *url,
					int *from_p, int *to_p)
{
  BBSTable *table = dynamic_cast<BBSTable *>(get_root_folder());
  assert(table != NULL);

  BulletinBoard *target_board = table->guess_board(url);
  if (target_board != NULL && target_board != this)
    return NULL;	// İ㤤URL

  char *id;
  switch (get_bbs_type())
    {
    case OCHUSHA_BBS_TYPE_2CH:
    case OCHUSHA_BBS_TYPE_2CHLIKE_EUCJP:
    case OCHUSHA_BBS_TYPE_2CH_COMPATIBLE:
    case OCHUSHA_BBS_TYPE_2CH_HEADLINE:
    case OCHUSHA_BBS_TYPE_2CH_BE:
    case OCHUSHA_BBS_TYPE_BBSPINK:
    case OCHUSHA_BBS_TYPE_BBSPINK_HEADLINE:
    case OCHUSHA_BBS_TYPE_2CH_COMPATIBLE_RO:
      id = BBS2ch::parse_thread_url(url, from_p, to_p);
      break;

    case OCHUSHA_BBS_TYPE_MACHIBBS:
      id = BBSMachiBBS::parse_thread_url(url, from_p, to_p);
      break;

    case OCHUSHA_BBS_TYPE_JBBS_LIVEDOOR:
      id = BBSJbbsLivedoor::parse_thread_url(url, from_p, to_p);
      break;

    default:
      id = NULL;
    }

  if (id == NULL)
    return NULL;

  BBSThread *thread = lookup_bbs_thread_by_id(id);
  free(id);

  if (thread->is_tmp_thread())
    thread->set_trigger_url(url);

  return thread;
}


int
BulletinBoard::how_dirty() const
{
  return dirtyness == NOT_DIRTY
    ? NOT_DIRTY
    : (dirtyness < URGENT_SAVE_RECOMMENDED
       ? MINOR_MODIFICATION : URGENT_SAVE_RECOMMENDED);
}


ThreadlistIterator *
BulletinBoard::get_threadlist(Configuration &config, Repository &repository,
			      ThreadPool &pool, NetworkAgent *agent,
			      bool update)
{
  assert(libochusha_giant.is_locked_by_self());
  assert(bulletin_board_giant.is_locked_by_self()
	 && bulletin_board_giant.get_count() == 1);

  smart_ptr<ThreadlistUpdater> updater = this->updater;

  if (static_cast<ThreadlistUpdater *>(updater) == NULL)
    { /* ̸ɤ߹ߡǤʤ */
      // ͥåȥν
      BBSThreadVector::iterator iter = threadlist.begin();
      BBSThreadVector::iterator end_iter = threadlist.end();
      while (iter != end_iter)
	{
	  BBSThread *thread = *iter;
	  thread->alive = false;
	  iter++;
	}

      if (!update || config.get_offline() || agent == NULL)
	{
	  updater = new DummyUpdater(this, config, repository, agent);

	  bulletin_board_giant.unlock();
	  updater->finalize();
	  bulletin_board_giant.lock();
	}
      else switch (bbs_type)
	{
	case OCHUSHA_BBS_TYPE_2CH:
	case OCHUSHA_BBS_TYPE_2CH_COMPATIBLE:
	case OCHUSHA_BBS_TYPE_2CH_HEADLINE:
	case OCHUSHA_BBS_TYPE_BBSPINK:
	case OCHUSHA_BBS_TYPE_BBSPINK_HEADLINE:
	case OCHUSHA_BBS_TYPE_2CH_COMPATIBLE_RO:
	  updater = new SubjectTxtParser(this, config, repository, agent,
					 IconvBuffer::CP932_ENCODING,
					 "^([[:digit:]]+)\\.dat<>(.*)\\(([[:digit:]]+)\\)$");
	  break;

	case OCHUSHA_BBS_TYPE_2CHLIKE_EUCJP:
	case OCHUSHA_BBS_TYPE_2CH_BE:
	  updater = new SubjectTxtParser(this, config, repository, agent,
					 IconvBuffer::EUCJP_ENCODING,
					 "^([[:digit:]]+)\\.dat<>(.*)\\(([[:digit:]]+)\\)$");
	  break;

	case OCHUSHA_BBS_TYPE_MACHIBBS:
	  updater = new SubjectTxtParser(this, config, repository, agent,
					 IconvBuffer::CP932_ENCODING,
					 "^([[:digit:]]+)\\.cgi,(.*)\\(([[:digit:]]+)\\)$");
	  break;

	case OCHUSHA_BBS_TYPE_JBBS:
	case OCHUSHA_BBS_TYPE_JBBS_LIVEDOOR:
	  updater = new SubjectTxtParser(this, config, repository, agent,
					 IconvBuffer::EUCJP_ENCODING,
					 "^([[:digit:]]+)\\.cgi,(.*)\\(([[:digit:]]+)\\)$");
	  break;

	case OCHUSHA_BBS_TYPE_PSEUDO_FOLDER:
	case OCHUSHA_BBS_TYPE_PASSIVE_FOLDER:	// ׼
	case OCHUSHA_BBS_TYPE_ACTIVE_FOLDER:	// ׼
	case OCHUSHA_BBS_TYPE_VIRTUAL_BOARD:	// ѻ
	case OCHUSHA_BBS_TYPE_MITINOKU:		// ׼
	default:
	  updater = new DummyUpdater(this, config, repository, agent);

	  bulletin_board_giant.unlock();
	  updater->finalize();
	  bulletin_board_giant.lock();
	  break;
	}
    }

  return new ThreadlistIterator(updater);
}


ThreadlistIterator *
BulletinBoard::get_threadlist(Configuration &config, Repository &repository,
			      ThreadPool &pool, NetworkAgentBroker &broker,
			      bool update)
{
  assert(!libochusha_giant.is_locked_by_self());
  assert(!bulletin_board_giant.is_locked_by_self());

  LockBlock global_lock(libochusha_giant);
  LockBlock local_lock(bulletin_board_giant);

  smart_ptr<ThreadlistUpdater> updater = this->updater;

  if (static_cast<ThreadlistUpdater *>(updater) == NULL)
    { /* ̸ɤ߹ߡǤʤ */
      // ͥåȥν
      NetworkAgent *agent = broker.employ_agent();
      if (agent == NULL)
	return NULL;	// Out of Memory?

      switch (bbs_type)
	{
	case OCHUSHA_BBS_TYPE_2CH:
	case OCHUSHA_BBS_TYPE_2CH_COMPATIBLE:
	case OCHUSHA_BBS_TYPE_2CH_HEADLINE:
	case OCHUSHA_BBS_TYPE_BBSPINK:
	case OCHUSHA_BBS_TYPE_BBSPINK_HEADLINE:
	case OCHUSHA_BBS_TYPE_2CH_COMPATIBLE_RO:
	case OCHUSHA_BBS_TYPE_2CHLIKE_EUCJP:
	case OCHUSHA_BBS_TYPE_2CH_BE:
	case OCHUSHA_BBS_TYPE_MACHIBBS:
	case OCHUSHA_BBS_TYPE_JBBS:
	case OCHUSHA_BBS_TYPE_JBBS_LIVEDOOR:
	  return get_threadlist(config, repository, pool, agent, update);

	case OCHUSHA_BBS_TYPE_PASSIVE_FOLDER:
	  update = false;
	  /* FALLTHROUGH */
	case OCHUSHA_BBS_TYPE_ACTIVE_FOLDER:
	  updater = new BookmarkFolderUpdater(this, config, repository, pool,
					      agent, update);
	  break;

	case OCHUSHA_BBS_TYPE_PSEUDO_FOLDER:
	  {
	    char *algorithm
	      = get_algorithm_from_pseudo_folder_url(get_base_url());
	    if (algorithm != NULL)
	      {
		if (strcmp(algorithm, "all-threads") == 0)
		  {
		    updater = new AllThreadEnumerator(this, config, repository,
						      pool, agent);
		    free(algorithm);
		    break;
		  }
		free(algorithm);
	      }
	  }
	  /* FALLTHROUGH */
	case OCHUSHA_BBS_TYPE_VIRTUAL_BOARD:	// ѻ
	case OCHUSHA_BBS_TYPE_MITINOKU:		// ׼
	  updater = new DummyUpdater(this, config, repository, agent);

	  bulletin_board_giant.unlock();
	  updater->finalize();
	  bulletin_board_giant.lock();
	  break;
	}
    }

  return new ThreadlistIterator(updater);
}


DATFileExplorer *
BulletinBoard::get_dat_file_explorer(Configuration &config,
				     Repository &repository)
{
  LockBlock global_lock(libochusha_giant);
  LockBlock local_lock(bulletin_board_giant);

  smart_ptr<ThreadlistUpdater> updater = this->updater;

  if (updater != NULL)
    return NULL;	// ̸DATեȯ⤷ϥι

  return new DATFileExplorerImpl(this, config, repository);
}


void
BulletinBoard::update_threadlist(ThreadlistUpdater *updater)
{
  LockBlock lock_obj(bulletin_board_giant);

  // updaterΥꥹȤupdater򻲾ȤƤThreadlistIteratorƤΤ
  // ޤŬڤʾ֤˹θ幹줿ꥹȤthreadlist
  // ˥ԡ롣
  BBSThreadVector::iterator iter = threadlist.begin();
  BBSThreadVector::iterator end_iter = threadlist.end();
  while (iter != end_iter)
    {
      BBSThread *thread = *iter;
      if (!thread->alive)
	{
	  // thread->aliveʥϴupdaterΥꥹȤ˴ޤޤƤΤǡ
	  // Ǥ̵롣
	  if (thread->get_number_of_responses_got() > 0)
	    {
	      // DATե뤬¸ߤϤΥϥꥹȤ˻Ĥ
	      thread->set_number_of_responses_on_server(0);
	      thread->alive = true;
	      thread->set_dat_dropped(true);
	      if (thread->get_grave() == NULL)
		{
		  char *url = thread->get_url();
		  thread->set_grave(url);
		  free(url);
		}
	      updater->updated_list.push_back(thread);
	    }
	  else
	    thread_table.remove(thread->get_id());
	}

      iter++;
    }

  threadlist = updater->updated_list;
  dirtyness += URGENT_SAVE_RECOMMENDED;
}


void
BulletinBoard::recover_threadlist(ThreadlistUpdater *updater)
{
  LockBlock lock_obj(bulletin_board_giant);

  // ThreadlistIteratorƤupdaterΥꥹȤľΥꥹȤ
  // Ƥޤ褦ˤĤĤޤ碌롣
  BBSThreadVector::iterator iter = threadlist.begin();
  BBSThreadVector::iterator end_iter = threadlist.end();
  while (iter != end_iter)
    {
      BBSThread *thread = *iter;
      if (!thread->alive)
	{
	  // thread->aliveʥupdated_list˴ޤޤƤΤ̵롣
	  thread->alive = true;
	  updater->updated_list.push_back(thread);
	}
      iter++;
    }
}


const char *
BulletinBoard::get_element_name() const
{
  return "board";
}


BBSThread *
BulletinBoard::thread_new(const char *id, const char *title)
{
  BBSThread *thread = new BBSThread(this, id, title);
  threadlist.push_back(thread);
  thread_table.insert(id, thread);

  notify_modified();
  return thread;
}


void
BulletinBoard::set_base_url(const char *new_base_url, bool internal)
{
  if (new_base_url != NULL && base_url != NULL
      && strcmp(new_base_url, base_url) == 0)
    return;	// Ѳ̵

  set_last_modified(NULL, internal);
  set_date(NULL, internal);

  if (base_url != NULL)
    {
      URI old_uri(base_url);
      const char *server = old_uri.get_server();
      if (server != NULL)
	{
	  // movement_historyΥ֤å
	  list<char *>::iterator where;
	  list<char *>::iterator end_iter = movement_history.end();
	  where = find_if(movement_history.begin(), end_iter,
			  CStringEqual(server));
	  if (where == end_iter)
	    movement_history.push_front(strdup(server));
	}
      free(base_url);
      base_url = NULL;
    }

  if (new_base_url != NULL && *new_base_url != '\0')
    {
      size_t len = strlen(new_base_url);
      if (new_base_url[len - 1] == '/')
	base_url = strdup(new_base_url);
      else
	{
	  char *new_url = static_cast<char *>(malloc(len + 2));
	  if (new_url != NULL)
	    {
	      memcpy(new_url, new_base_url, len);
	      new_url[len] = '/';
	      new_url[len + 1] = '\0';
	    }
	  base_url = new_url;
	}
    }
  else
    base_url = NULL;

  if (!internal)
    notify_modified();

  URI new_uri(base_url);
  if (!new_uri.is_valid())
    return;	// ̿Ūġ

  const char *new_server = new_uri.get_server();
  if (new_server == NULL)
    return;	// ̿Ūġ

  // movement_historyΥ֤å
  list<char *>::iterator where;
  list<char *>::iterator end_iter = movement_history.end();
  where = find_if(movement_history.begin(), end_iter,
		  CStringEqual(new_server));
  if (where != end_iter)
    {
      char *tmp_string = *where;
      movement_history.erase(where);
      free(tmp_string);
    }

  switch (bbs_type)
    {
    case OCHUSHA_BBS_TYPE_2CH:
    case OCHUSHA_BBS_TYPE_2CH_HEADLINE:
    case OCHUSHA_BBS_TYPE_2CH_BE:
    case OCHUSHA_BBS_TYPE_BBSPINK:
    case OCHUSHA_BBS_TYPE_BBSPINK_HEADLINE:
    case OCHUSHA_BBS_TYPE_2CH_COMPATIBLE:
      BBS2ch::register_board(this);
      return;

    case OCHUSHA_BBS_TYPE_MACHIBBS:
      BBSMachiBBS::register_board(this);
      return;

    case OCHUSHA_BBS_TYPE_JBBS_LIVEDOOR:
      BBSJbbsLivedoor::register_board(this);
      return;
    }
}


void
BulletinBoard::set_bbs_type(int type, bool internal)
{
  if (bbs_type == type)
    return;

  bbs_type = type;

  if (!internal)
    notify_modified();

  switch (type)
    {
    case OCHUSHA_BBS_TYPE_2CH:
    case OCHUSHA_BBS_TYPE_2CH_HEADLINE:
    case OCHUSHA_BBS_TYPE_2CH_BE:
    case OCHUSHA_BBS_TYPE_BBSPINK:
    case OCHUSHA_BBS_TYPE_BBSPINK_HEADLINE:
      BBS2ch::register_board(this);
      return;

    case OCHUSHA_BBS_TYPE_MACHIBBS:
      BBSMachiBBS::register_board(this);
      return;

    case OCHUSHA_BBS_TYPE_JBBS_LIVEDOOR:
      BBSJbbsLivedoor::register_board(this);
      return;
    }
}


char *
BulletinBoard::get_id() const
{
  switch (bbs_type)
    {
    case OCHUSHA_BBS_TYPE_2CH:
    case OCHUSHA_BBS_TYPE_2CH_COMPATIBLE:
    case OCHUSHA_BBS_TYPE_2CH_HEADLINE:
    case OCHUSHA_BBS_TYPE_2CH_BE:
    case OCHUSHA_BBS_TYPE_BBSPINK:
    case OCHUSHA_BBS_TYPE_BBSPINK_HEADLINE:
    case OCHUSHA_BBS_TYPE_2CH_COMPATIBLE_RO:
      return BBS2ch::get_board_id(this);

    case OCHUSHA_BBS_TYPE_MACHIBBS:
      return BBSMachiBBS::get_board_id(this);

    case OCHUSHA_BBS_TYPE_JBBS_LIVEDOOR:
      return BBSJbbsLivedoor::get_board_id(this);
    }

  return NULL;
}


char *
BulletinBoard::get_bbs_root() const
{
  switch (bbs_type)
    {
    case OCHUSHA_BBS_TYPE_2CH:
    case OCHUSHA_BBS_TYPE_2CH_COMPATIBLE:
    case OCHUSHA_BBS_TYPE_2CH_HEADLINE:
    case OCHUSHA_BBS_TYPE_2CH_BE:
    case OCHUSHA_BBS_TYPE_BBSPINK:
    case OCHUSHA_BBS_TYPE_BBSPINK_HEADLINE:
    case OCHUSHA_BBS_TYPE_2CH_COMPATIBLE_RO:
      return BBS2ch::get_bbs_root(this);

    case OCHUSHA_BBS_TYPE_MACHIBBS:
      return BBSMachiBBS::get_bbs_root(this);

    case OCHUSHA_BBS_TYPE_JBBS_LIVEDOOR:
      return BBSJbbsLivedoor::get_bbs_root(this);
    }

  return NULL;
}


char *
BulletinBoard::get_url(BBSThread *thread, int from, int to) const
{
  assert(thread->get_board() == this);
  switch (bbs_type)
    {
    case OCHUSHA_BBS_TYPE_2CH:
    case OCHUSHA_BBS_TYPE_2CHLIKE_EUCJP:
    case OCHUSHA_BBS_TYPE_2CH_COMPATIBLE:
    case OCHUSHA_BBS_TYPE_2CH_HEADLINE:
    case OCHUSHA_BBS_TYPE_2CH_BE:
    case OCHUSHA_BBS_TYPE_BBSPINK:
    case OCHUSHA_BBS_TYPE_BBSPINK_HEADLINE:
    case OCHUSHA_BBS_TYPE_2CH_COMPATIBLE_RO:
      return BBS2ch::get_url(thread, from, to);

    case OCHUSHA_BBS_TYPE_MACHIBBS:
      return BBSMachiBBS::get_url(thread, from, to);

    case OCHUSHA_BBS_TYPE_JBBS:
    case OCHUSHA_BBS_TYPE_JBBS_LIVEDOOR:
      return BBSJbbsLivedoor::get_url(thread, from, to);

    case OCHUSHA_BBS_TYPE_MITINOKU:
      break;
    }
  return NULL;
}


void
BulletinBoard::set_last_modified(const char *value, bool internal)
{
  if (last_modified != NULL && value != NULL
      && strcmp(last_modified, value) == 0)
    return;

  if (last_modified != NULL)
    free(last_modified);
  last_modified = value != NULL ? strdup(value) : NULL;

  if (!internal)
    notify_modified();
}


void
BulletinBoard::set_date(const char *value, bool internal)
{
  if (date != NULL && value != NULL && strcmp(date, value) == 0)
    return;

  if (date != NULL)
    free(date);
  date = value != NULL ? strdup(value) : NULL;

  if (!internal)
    notify_modified();
}


void
BulletinBoard::set_short_name(const char *value, bool internal)
{
  if (short_name != NULL && value != NULL && strcmp(short_name, value) == 0)
    return;

  if (short_name != NULL)
    free(short_name);

  if (value != NULL
      && (*value == '\0'
	  || (get_name() != NULL && strcmp(value, get_name()) == 0)))
    value = NULL;

  short_name = value != NULL ? strdup(value) : NULL;

  if (!internal)
    notify_modified();
}


void
BulletinBoard::set_rule(const char *value, bool internal)
{
  if (rule != NULL && value != NULL && strcmp(rule, value) == 0)
    return;

  if (rule != NULL)
    free(rule);
  rule = value != NULL ? strdup(value) : NULL;

  if (!internal)
    notify_modified();
}


void
BulletinBoard::set_previous_server(const char *value)
{
  if (previous_server != NULL)
    free(previous_server);
  previous_server = value != NULL ? strdup(value) : NULL;
}


void
BulletinBoard::set_mailto_mode(MailtoMode mode, bool internal)
{
  int prev_flags = ui_flags;

  int effective_mode = mode >= MAILTO_MODE_DEFAULT && mode <= MAILTO_MODE_HIDE
    ? mode : MAILTO_MODE_DEFAULT;
  ui_flags = (ui_flags & ~MAILTO_MODE_MASK) | effective_mode;

  if (ui_flags != prev_flags && !internal)
    notify_modified();
}


void
BulletinBoard::set_last_post_sage(bool value, bool internal)
{
  int prev_flags = ui_flags;

  ui_flags = (ui_flags & ~LAST_POST_SAGE) | (value ? LAST_POST_SAGE : 0);

  if (ui_flags != prev_flags && !internal)
    notify_modified();
}


void
BulletinBoard::set_last_post_be(bool value, bool internal)
{
  int prev_flags = ui_flags;

  ui_flags = (ui_flags & ~LAST_POST_BE) | (value ? LAST_POST_BE : 0);

  if (ui_flags != prev_flags && !internal)
    notify_modified();
}


void
BulletinBoard::set_last_post_viewer(bool value, bool internal)
{
  int prev_flags = ui_flags;

  ui_flags = (ui_flags & ~LAST_POST_VIEWER) | (value ? LAST_POST_VIEWER : 0);

  if (ui_flags != prev_flags && !internal)
    notify_modified();
}


void
BulletinBoard::set_last_name(const char *value, bool internal)
{
  if (last_name != NULL && value != NULL && strcmp(last_name, value) == 0)
    return;

  if (last_name != NULL)
    free(last_name);
  last_name = value != NULL ? strdup(value) : NULL;

  if (!internal)
    notify_modified();
}


void
BulletinBoard::set_last_mail(const char *value, bool internal)
{
  if (last_mail != NULL && value != NULL && strcmp(last_mail, value) == 0)
    return;

  if (last_mail != NULL)
    free(last_mail);
  last_mail = value != NULL ? strdup(value) : NULL;

  if (!internal)
    notify_modified();
}


void
BulletinBoard::get_all_bbs_type_list(vector<int> &bbs_type_list)
{
  bbs_type_list.push_back(OCHUSHA_BBS_TYPE_2CH_COMPATIBLE_RO);
  bbs_type_list.push_back(OCHUSHA_BBS_TYPE_2CH_COMPATIBLE);
  bbs_type_list.push_back(OCHUSHA_BBS_TYPE_JBBS_LIVEDOOR);
  bbs_type_list.push_back(OCHUSHA_BBS_TYPE_2CHLIKE_EUCJP);
  bbs_type_list.push_back(OCHUSHA_BBS_TYPE_2CH);
  bbs_type_list.push_back(OCHUSHA_BBS_TYPE_BBSPINK);
  bbs_type_list.push_back(OCHUSHA_BBS_TYPE_MACHIBBS);
  bbs_type_list.push_back(OCHUSHA_BBS_TYPE_2CH_HEADLINE);
  bbs_type_list.push_back(OCHUSHA_BBS_TYPE_BBSPINK_HEADLINE);
  bbs_type_list.push_back(OCHUSHA_BBS_TYPE_2CH_BE);
  bbs_type_list.push_back(OCHUSHA_BBS_TYPE_ACTIVE_FOLDER);
  bbs_type_list.push_back(OCHUSHA_BBS_TYPE_PASSIVE_FOLDER);
  bbs_type_list.push_back(OCHUSHA_BBS_TYPE_PSEUDO_FOLDER);
}


static const char * const bbs_type_description[] =
{
  N_("Pseudo Folder"),
  N_("Passive Folder"),
  N_("Active Folder"),
  NULL,	// 첾ġѻ
  N_("Unknown Type BBS"),
  N_("2ch BBS"),
  N_("2ch-like(EUC-JP)"),
  N_("JBBS"),
  N_("Machi BBS"),
  N_("JBBS Livedoor"),
  NULL,	//N_("Mitinoku BBS"),	// ̤ѻߡ
  N_("2ch Compatible"),
  N_("2ch Headline"),
  N_("2ch Be BBS"),
  NULL,	//N_("2ch - Reserved."),
  N_("BBSPink"),
  N_("2ch Compatible(R/O)"),
  N_("BBSPink Headline")
};


const char *
BulletinBoard::get_bbs_type_description(int type)
{
  const char *desc;
  if (type >= OCHUSHA_BBS_TYPE_PSEUDO_FOLDER
      && type <= OCHUSHA_BBS_TYPE_BBSPINK_HEADLINE)
    desc = bbs_type_description[type + 5];
  else
    desc = NULL;

  return desc != NULL ? dgettext(PACKAGE, desc) : NULL;
}


char *
BulletinBoard::synthesize_active_folder_url(unsigned int sn)
{
  char buf[PATH_MAX];
  if (snprintf(buf, PATH_MAX, "oabf:///sn-%u/", sn) < PATH_MAX)
    return strdup(buf);
  return NULL;
}


unsigned int
BulletinBoard::get_serial_number_from_active_folder_url(const char *url)
{
  if (url == NULL)
    return 0;

  unsigned int sn;
  if (sscanf(url, "oabf:///sn-%u/", &sn) == 1)
    return sn;

  return 0;
}


bool
BulletinBoard::is_active_folder_url(const char *url)
{
  return strncmp("oabf:///sn-", url, 11) == 0
    && get_serial_number_from_active_folder_url(url) != 0;
}


char *
BulletinBoard::synthesize_pseudo_folder_url(const char *algorithm)
{
  char buf[PATH_MAX];
  if (snprintf(buf, PATH_MAX, "opbf:///%s/", algorithm) < PATH_MAX)
    return strdup(buf);
  return NULL;
}


char *
BulletinBoard::get_algorithm_from_pseudo_folder_url(const char *url)
{
  if (url == NULL)
    return NULL;

  Regexp url_pattern("opbf:///([^/]+)/");

  if (url_pattern.match(url, strlen(url)))
    return strndup(url_pattern.match_begin(1), url_pattern.match_len(1));

  return NULL;
}


const char *
BulletinBoard::get_description_about_pseudo_folder_url(const char *url)
{
  if (!is_pseudo_folder_url(url))
    return NULL;

  char *algorithm = get_algorithm_from_pseudo_folder_url(url);
  if (algorithm == NULL)
    return NULL;

  const char *result;
  if (strcmp(algorithm, "all-threads") == 0)
    result = LIB_("All Threads Got");
  else
    result = NULL;

  free(algorithm);
  return result;
}


bool
BulletinBoard::is_pseudo_folder_url(const char *url)
{
  char *algorithm = get_algorithm_from_pseudo_folder_url(url);
  if (algorithm == NULL)
    return false;

  // ΤȤall-threadsʤΤǡġġ
  bool result = strcmp(algorithm, "all-threads") == 0;

  free(algorithm);
  return result;
}




#define OUTPUT_ATTRIBUTE_BOOLEAN(gzfile, attribute, ind)		\
  do									\
    {									\
      if (attribute)							\
	gzprintf(gzfile,						\
		 "%s<attribute name=\"" #attribute "\">\n"		\
		 "%s  <boolean val=\"true\"/>\n"			\
		 "%s</attribute>\n", ind, ind, ind);			\
    }									\
  while (0)

#define OUTPUT_ATTRIBUTE_INT(gzfile, attribute, ind)			\
  do									\
    {									\
      if (attribute)							\
	gzprintf(gzfile,						\
		 "%s<attribute name=\"" #attribute "\">\n"		\
		 "%s  <int val=\"%d\"/>\n"				\
		 "%s</attribute>\n", ind, ind, attribute, ind);		\
    }									\
  while (0)

#define OUTPUT_ATTRIBUTE_STRING(gzfile, attribute, ind)			\
  do									\
    {									\
      if (attribute != NULL)						\
	{								\
	  string text;							\
	  gzprintf(gzfile,						\
		   "%s<attribute name=\"" #attribute "\">\n"		\
		   "%s  <string>%s</string>\n"				\
		   "%s</attribute>\n", ind, ind,			\
		   append_markup_escaped_text(text, attribute).c_str(),	\
		   ind);						\
	}								\
    }									\
  while (0)


void
BulletinBoard::write_extended_tree_element(WriteTree &args)
{
  gzFile boardlist_xml = args.tree_xml;
  const char *indent = args.indent->c_str();

  OUTPUT_ATTRIBUTE_INT(boardlist_xml, bbs_type, indent);
  OUTPUT_ATTRIBUTE_STRING(boardlist_xml, base_url, indent);
  OUTPUT_ATTRIBUTE_STRING(boardlist_xml, previous_server, indent);
  OUTPUT_ATTRIBUTE_STRING(boardlist_xml, last_modified, indent);

  OUTPUT_ATTRIBUTE_STRING(boardlist_xml, short_name, indent);
  OUTPUT_ATTRIBUTE_STRING(boardlist_xml, rule, indent);
  OUTPUT_ATTRIBUTE_INT(boardlist_xml, ui_flags, indent);
  OUTPUT_ATTRIBUTE_STRING(boardlist_xml, last_name, indent);
  OUTPUT_ATTRIBUTE_STRING(boardlist_xml, last_mail, indent);

  if (!movement_history.empty())
    {
      list<char *>::const_iterator iter = movement_history.begin();
      list<char *>::const_iterator end_iter = movement_history.end();

      string tmp_string;
      while (iter != end_iter)
	{
	  const char *server = *iter++;
	  if (tmp_string.empty())
	    tmp_string.append(server);
	  else
	    tmp_string.append(",").append(server);
	}
      const char *server_history = tmp_string.c_str();
      OUTPUT_ATTRIBUTE_STRING(boardlist_xml, server_history, indent);
    }

  a_bone.write_xml_elements(boardlist_xml, indent);
}


void
BulletinBoard::read_extended_tree_element(AttributeTable &attributes)
{
  set_bbs_type(attributes.get_int("bbs_type"), true);
  set_base_url(attributes.get_string("base_url"), true);
  set_last_modified(attributes.get_string("last_modified"), true);

  set_short_name(attributes.get_string("short_name"), true);
  set_rule(attributes.get_string("rule"), true);
  ui_flags = attributes.get_int("ui_flags");
  set_last_name(attributes.get_string("last_name"), true);
  set_last_mail(attributes.get_string("last_mail"), true);

  const char *tmp = attributes.get_string("previous_server");
  if (tmp != NULL)
    set_previous_server(tmp);

  const char *server_history = attributes.get_string("server_history");
  const char *cur_pos = server_history;

  URI current_uri(get_base_url());
  if (current_uri.is_valid())
    {
      const char *current_server = current_uri.get_server();
      while (cur_pos != NULL && *cur_pos != '\0')
	{
	  const char *tmp_pos = strchr(cur_pos, ',');
	  if (tmp_pos == NULL)
	    tmp_pos = cur_pos + strlen(cur_pos);

	  char *server = strndup(cur_pos, tmp_pos - cur_pos);
	  if (server != NULL)
	    {
	      list<char *>::iterator where;
	      list<char *>::iterator end_iter = movement_history.end();
	      where = find_if(movement_history.begin(), end_iter,
			      CStringEqual(server));
	      if (where == end_iter
		  && (current_server == NULL
		      || strcmp(current_server, server) != 0))
		movement_history.push_back(server);
	      else
		free(server);
	    }
	  if (*tmp_pos == '\0')
	    break;

	  cur_pos = tmp_pos + 1;
	}
    }

  a_bone.read_xml_elements(attributes);
}


bool
BulletinBoard::write_threadlist_xml(Repository &repository)
{
  LockBlock lock_obj(bulletin_board_giant);

  if (dirtyness == NOT_DIRTY)
    return true;	// ǿξ󤬵ϿѤʤΤǤʤˤ⤷ʤ

  int prev_dirtyness = dirtyness;
  dirtyness = NOT_DIRTY;

  if (threadlist.empty())
    return true;	// äݤʤΤϵϿʤ

  switch (bbs_type)
    {
    case OCHUSHA_BBS_TYPE_PSEUDO_FOLDER:
    case OCHUSHA_BBS_TYPE_PASSIVE_FOLDER:
    case OCHUSHA_BBS_TYPE_ACTIVE_FOLDER:
    case OCHUSHA_BBS_TYPE_VIRTUAL_BOARD:	// ѻ
    case OCHUSHA_BBS_TYPE_MITINOKU:		// ̤ݡ
      return true;

    default:
      break;
    }

  URI base_uri(get_base_url());
  if (!base_uri.is_valid())
    return false;	// äȤ롣

  char fake_url[PATH_MAX];
  if (snprintf(fake_url, PATH_MAX, "%s" OCHUSHA_THREADLIST_XML, get_base_url())
      >= PATH_MAX)
    return false;	// Ĺ

  int fd = repository.get_bbs_data().open_file(fake_url,
					       O_WRONLY | O_TRUNC | O_CREAT);
  if (fd < 0)
    {
      cerr << "Couldn't open " OCHUSHA_THREADLIST_XML " to write.\n";
      return false;
    }

  gzFile threadlist_xml = gzdopen(fd, "w");
  if (threadlist_xml == NULL)
    {
      cerr << "Couldn't gzdopen for " OCHUSHA_THREADLIST_XML "\n";
      close(fd);
      return false;
    }

  gzprintf(threadlist_xml, "<?xml version=\"1.0\"?>\n");
  gzprintf(threadlist_xml, "<ochusha>\n");
  gzprintf(threadlist_xml, "  <threadlist>\n");

  BBSThreadVector::iterator iter = threadlist.begin();
  BBSThreadVector::iterator end_iter = threadlist.end();
  while (iter != end_iter)
    {
      (*iter)->write_threadlist_element(threadlist_xml);
      iter++;
    }

  gzprintf(threadlist_xml, "  </threadlist>\n");
  gzprintf(threadlist_xml, "</ochusha>\n");

  if (gzclose(threadlist_xml) == 0)
    return true;

  // ¸ԤʤΤǡ괺礤ᤷƤ
  dirtyness = prev_dirtyness;
  return false;
}


class ThreadlistXMLContext : public SAXContext
{
private:
  enum SAXState
  {
    SAX_INITIAL,
    SAX_OCHUSHA,
    SAX_THREADLIST,
    SAX_THREAD,
    SAX_THREAD_ATTRIBUTE,
    SAX_THREAD_ATTRIBUTE_BOOLEAN,
    SAX_THREAD_ATTRIBUTE_INT,
    SAX_THREAD_ATTRIBUTE_STRING,
    SAX_ACCEPTED,
    SAX_ERROR
  } state;

  BulletinBoard *board;
  AttributeTable thread_attrs;


protected:
  void start_element_handler(const char *name, const char **attrs);
  void end_element_handler(const char *name);
  void characters_handler(const char *ch, size_t len);


public:
  ThreadlistXMLContext(BulletinBoard *board) : state(SAX_INITIAL), board(board)
  {}
  inline bool is_accepted() { return state == SAX_ACCEPTED; }
};


void
ThreadlistXMLContext::start_element_handler(const char *name,
					    const char **attrs)
{
  switch (state)
    {
    case SAX_INITIAL:
      if (strcmp(name, "ochusha") == 0) { state = SAX_OCHUSHA; return; }
      break;	// 顼

    case SAX_OCHUSHA:
      if (strcmp(name, "threadlist") == 0) { state = SAX_THREADLIST; return; }
      break;	// 顼

    case SAX_THREADLIST:
      if (strcmp(name, "thread") == 0) { state = SAX_THREAD; return; }
      break;	// 顼

    case SAX_THREAD:
      if (strcmp(name, "attribute") == 0
	  && attrs != NULL && strcmp(attrs[0], "name") == 0)
	{
	  state = SAX_THREAD_ATTRIBUTE;
	  thread_attrs.set_attr_name(attrs[1]);
	  return;
	}
      break;	// 顼

    case SAX_THREAD_ATTRIBUTE:
      if (attrs != NULL && strcmp(attrs[0], "val") == 0)
	{
	  // int/booleanβǽͭ
	  if (strcmp(name, "int") == 0)
	    {
	      state = SAX_THREAD_ATTRIBUTE_INT;
	      thread_attrs.set_attr_value(attrs[1]);
	      return;
	    }
	  if (strcmp(name, "boolean") == 0)
	    {
	      state = SAX_THREAD_ATTRIBUTE_BOOLEAN;
	      thread_attrs.set_attr_value(attrs[1]);
	      return;
	    }
#ifdef DEBUG_SAX_HANDLER
	  cerr << "element unexpected in state(" << state << ") found: "
	       << name << endl;
#endif
	  break;	// 顼
	}
      else if (strcmp(name, "string") == 0)
	{
	  state = SAX_THREAD_ATTRIBUTE_STRING;
	  return;
	}
      break;		// 顼

    case SAX_THREAD_ATTRIBUTE_INT:
    case SAX_THREAD_ATTRIBUTE_BOOLEAN:
    case SAX_THREAD_ATTRIBUTE_STRING:
    case SAX_ACCEPTED:
    case SAX_ERROR:
      break;	// 顼

    default:
      cerr << "start_element_handler is called in unknown state: "
	   << state << endl;
    }
  state = SAX_ERROR;
}


void
ThreadlistXMLContext::end_element_handler(const char *name)
{
  switch (state)
    {
    case SAX_OCHUSHA:
      if (strcmp(name, "ochusha") == 0) { state = SAX_ACCEPTED; return; }
      break;	// 顼

    case SAX_THREADLIST:
      if (strcmp(name, "threadlist") == 0) { state = SAX_OCHUSHA; return; }
      break;	// 顼

    case SAX_THREAD:
      if (strcmp(name, "thread") == 0)
	{
	  const char *thread_id = thread_attrs.get_string("id");
	  const char *thread_title = thread_attrs.get_string("title");
	  BBSThread *thread = board->lookup_bbs_thread_by_id(thread_id);
#if 0
	  if (thread == NULL)
	    {
	      if (thread_title == NULL)
		thread_title = "";
	      thread = board->thread_new(thread_id, thread_title);
	    }
#else
	  thread->set_title(thread_title != NULL ? thread_title : "");
#endif

	  if (thread == NULL)	// Out of Memory
	    {
	      thread_attrs.clear();
	      break;	// 顼
	    }

	  thread->read_threadlist_element(thread_attrs);
	  thread_attrs.clear();
	  state = SAX_THREADLIST;
	  return;
	}
      break;	// 顼

    case SAX_THREAD_ATTRIBUTE:
      if (strcmp(name, "attribute") == 0)
	{
	  thread_attrs.register_current_attribute();
	  state = SAX_THREAD;
	  return;
	}
      break;	// 顼

    case SAX_THREAD_ATTRIBUTE_BOOLEAN:
      if (strcmp(name, "boolean") == 0)
	{ state = SAX_THREAD_ATTRIBUTE; return; }
      break;	// 顼

    case SAX_THREAD_ATTRIBUTE_INT:
      if (strcmp(name, "int") == 0) { state = SAX_THREAD_ATTRIBUTE; return; }
      break;	// 顼

    case SAX_THREAD_ATTRIBUTE_STRING:
      if (strcmp(name, "string") == 0)
	{ state = SAX_THREAD_ATTRIBUTE; return; }
      break;	// 顼

    case SAX_INITIAL:
    case SAX_ACCEPTED:
    case SAX_ERROR:
#ifdef DEBUG_SAX_HANDLER
      cerr << "end_element_handler called in invalid state: " << state << "\n";
#endif
      break;	// 顼
    }
#ifdef DEBUG_SAX_HANDLER
  cerr << "Invalid document: </" << name << "> appeared in state=" << state
       << "\n";
#endif
  state = SAX_ERROR;
}


void
ThreadlistXMLContext::characters_handler(const char *ch, size_t len)
{
  if (state == SAX_THREAD_ATTRIBUTE_STRING)
    thread_attrs.append_attr_value(ch, len);
}


bool
BulletinBoard::read_threadlist_xml(Repository &repository)
{
  if (threadlist_read)
    return true;

  LockBlock lock_obj(bulletin_board_giant);

  if (threadlist_read)
    return true;

  int previous_dirtyness = dirtyness;

  switch (bbs_type)
    {
    case OCHUSHA_BBS_TYPE_PSEUDO_FOLDER:
    case OCHUSHA_BBS_TYPE_PASSIVE_FOLDER:
    case OCHUSHA_BBS_TYPE_ACTIVE_FOLDER:
    case OCHUSHA_BBS_TYPE_VIRTUAL_BOARD:	// ѻ
    case OCHUSHA_BBS_TYPE_MITINOKU:		// ̤ݡ
      threadlist_read = true;
      return true;

    default:
      break;
    }

  URI base_uri(get_base_url());
  if (!base_uri.is_valid() && bbs_type != OCHUSHA_BBS_TYPE_VIRTUAL_BOARD)
    {
      set_last_modified(NULL);
      threadlist_read = true;
      return false;	// äȤ롣
    }

  char fake_url[PATH_MAX];
  if (snprintf(fake_url, PATH_MAX, "%s" OCHUSHA_THREADLIST_XML, get_base_url())
      >= PATH_MAX)
    {
      set_last_modified(NULL);
      threadlist_read = true;
      return false;	// Ĺ
    }

  char *fullpath = repository.get_bbs_data().find_file(fake_url);
  if (fullpath == NULL)
    {
      set_last_modified(NULL);
      threadlist_read = true;
      return false;	// ե뤬¸ߤʤ
    }

  ThreadlistXMLContext context(this);
  context.parse_xml_file(fullpath);
  free(fullpath);

  // threadlist.xmlɤ߹ʬʤ顢ƤʤΤ
  // ľͤᤷƤ
  dirtyness = previous_dirtyness;
  threadlist_read = true;

  if (context.is_accepted())
    return true;

  set_last_modified(NULL);
  return false;
}

}
