/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: configuration.cc,v 1.23 2009/01/01 07:46:43 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/configuration.h"

#include "libochusha/base_utils.h"
#include "libochusha/constants.h"
#include "libochusha/hashtable.h"
#include "libochusha/misc_utils.h"
#include "libochusha/repository.h"
#include "libochusha/sax_context.h"
#include "libochusha/uri_utils.h"

#include <cstddef>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <string>

#include <zlib.h>


#define OUTPUT_ATTRIBUTE_INT(gzfile, attribute)				\
  gzprintf(gzfile,							\
	   "    <attribute name=\"" #attribute "\">\n"			\
	   "      <int val=\"%d\"/>\n"					\
	   "    </attribute>\n", attribute)

#define OUTPUT_ATTRIBUTE_BOOLEAN(gzfile, attribute)			\
  gzprintf(gzfile,							\
	  "    <attribute name=\"" #attribute "\">\n"			\
	  "      <boolean val=\"%s\"/>\n"				\
	  "    </attribute>\n", attribute ? "true" : "false" )

#define OUTPUT_ATTRIBUTE_STRING(gzfile, attribute)			\
  do									\
    {									\
      if (attribute != NULL)						\
	{								\
	  string text;							\
	  gzprintf(gzfile,						\
		   "    <attribute name=\"" #attribute "\">\n"		\
		   "      <string>%s</string>\n"			\
		   "    </attribute>\n",				\
		   append_markup_escaped_text(text, attribute).c_str());\
	}								\
    }									\
  while (0)

#define OUTPUT_ATTRIBUTE_URI(gzfile, attribute)				\
  do									\
    {									\
      const char *uri = (attribute).get_uri();				\
      if (uri != NULL)							\
	{								\
	  char *enc_uri = Base64Encoder::encode_text_to_base64url(uri);	\
	  gzprintf(gzfile,						\
		   "    <attribute name=\"" #attribute "\">\n"		\
		   "      <string>%s</string>\n"			\
		   "    </attribute>\n", enc_uri);			\
	  free(enc_uri);						\
	}								\
    }									\
  while (0)


namespace Ochusha
{

Configuration::Configuration()
  : config_version(OCHUSHA_CURRENT_CONFIG_VERSION),
    bbsmenu_uri(OCHUSHA_DEFAULT_BBSMENU_URL),
    login_be_uri(OCHUSHA_2CH_BE_LOGIN_URL),
    login_viewer_uri(OCHUSHA_2CH_VIEWER_LOGIN_URL),
    offline(false),
    enable_proxy_to_read(false),
    enable_proxy_to_post(false),
    attributes(103)
{
}


Configuration::~Configuration()
{
}


void
Configuration::set_bbsmenu_uri(const char *url)
{
  if (url != NULL && strcmp(url, OCHUSHA_DEFAULT_BBSMENU_URL_OLD) == 0)
    return;

  bbsmenu_uri.parse_uri(url);
  if (bbsmenu_uri.is_valid())
    return;

  bbsmenu_uri.parse_uri(OCHUSHA_DEFAULT_BBSMENU_URL);
}


void
Configuration::set_login_be_uri(const char *url)
{
  login_be_uri.parse_uri(url);
  if (login_be_uri.is_valid())
    return;

  login_be_uri.parse_uri(OCHUSHA_2CH_BE_LOGIN_URL);
}


void
Configuration::set_login_viewer_uri(const char *url)
{
  login_viewer_uri.parse_uri(url);
  if (login_viewer_uri.is_valid())
    return;

  login_viewer_uri.parse_uri(OCHUSHA_2CH_VIEWER_LOGIN_URL);
}


bool
Configuration::write_config_xml(Repository &repository)
{
  int fd = repository.open_file(OCHUSHA_CONFIG_XML, O_WRONLY|O_TRUNC|O_CREAT);
  if (fd < 0)
    {
      cerr << "Couldn't open " OCHUSHA_CONFIG_XML " to write.\n";
      return false;
    }

  gzFile config_xml = gzdopen(fd, "w");
  if (config_xml == NULL)
    {
      cerr << "Couldn't gzdopen for config.xml.\n";
      return false;
    }

  gzprintf(config_xml, "<?xml version=\"1.0\"?>\n");
  gzprintf(config_xml, "<ochusha>\n");
  gzprintf(config_xml, "  <configuration>\n");

  config_version = OCHUSHA_CURRENT_CONFIG_VERSION;
  OUTPUT_ATTRIBUTE_INT(config_xml, config_version);

  OUTPUT_ATTRIBUTE_URI(config_xml, bbsmenu_uri);
  OUTPUT_ATTRIBUTE_URI(config_xml, login_be_uri);
  OUTPUT_ATTRIBUTE_URI(config_xml, login_viewer_uri);

  OUTPUT_ATTRIBUTE_BOOLEAN(config_xml, offline);

  OUTPUT_ATTRIBUTE_BOOLEAN(config_xml, enable_proxy_to_read);
  OUTPUT_ATTRIBUTE_URI(config_xml, proxy_uri_to_read);

  OUTPUT_ATTRIBUTE_BOOLEAN(config_xml, enable_proxy_to_post);
  OUTPUT_ATTRIBUTE_URI(config_xml, proxy_uri_to_post);

  a_bone.write_xml_elements(config_xml, "    ");

  write_extended_elements(config_xml);

  gzprintf(config_xml, "  </configuration>\n");
  gzprintf(config_xml, "</ochusha>\n");

  return gzclose(config_xml) == 0;
}


class ConfigXMLContext : public SAXContext
{
private:
  friend class Configuration;

  enum SAXState
  {
    SAX_INITIAL,
    SAX_OCHUSHA,
    SAX_CONFIGURATION,
    SAX_CONFIGURATION_ATTRIBUTE,
    SAX_CONFIGURATION_ATTRIBUTE_BOOLEAN,
    SAX_CONFIGURATION_ATTRIBUTE_INT,
    SAX_CONFIGURATION_ATTRIBUTE_STRING,
    SAX_ACCEPTED,
    SAX_ERROR
  } state;

  AttributeTable &attributes;


protected:
  void start_element_handler(const char *name, const char **attrs);
  void end_element_handler(const char *name);
  void characters_handler(const char *ch, size_t len);

public:
  ConfigXMLContext(AttributeTable &attributes) : attributes(attributes)
  {
    state = SAX_INITIAL;
    attributes.clear();
  }

  inline bool is_accepted()
  {
    return state == SAX_ACCEPTED;
  }
};


void
ConfigXMLContext::start_element_handler(const char *name, const char **attrs)
{
  switch (state)
    {
    case SAX_INITIAL:
      if (strcmp(name, "ochusha") == 0)
	{ state = SAX_OCHUSHA; return; }
      break;	/* 顼 */

    case SAX_OCHUSHA:
#if 1	/* COMPAT: ߴΤ */
      if (strcmp(name, "configuration") == 0
	  || strcmp(name, "preference") == 0)
#else
      if (strcmp(name, "configuration") == 0)
#endif
	{ state = SAX_CONFIGURATION; return; }
      break;	/* 顼 */

    case SAX_CONFIGURATION:
      if (strcmp(name, "attribute") == 0
	  && attrs != NULL && strcmp(attrs[0], "name") == 0)
	{
	  state = SAX_CONFIGURATION_ATTRIBUTE;
	  attributes.set_attr_name(attrs[1]);
	  return;
	}
      break;	/* 顼 */

    case SAX_CONFIGURATION_ATTRIBUTE:
      if (attrs != NULL && strcmp(attrs[0], "val") == 0)
	{
	  /* int/booleanβǽ */
	  if (strcmp(name, "int") == 0)
	    {
	      state = SAX_CONFIGURATION_ATTRIBUTE_INT;
	    }
	  else if (strcmp(name, "boolean") == 0)
	    {
	      state = SAX_CONFIGURATION_ATTRIBUTE_BOOLEAN;
	    }
	  else
	    {
	      break;	/* 顼 */
	    }

	  attributes.set_attr_value(attrs[1]);
	  return;
	}
      else if (strcmp(name, "string") == 0)
	{
	  state = SAX_CONFIGURATION_ATTRIBUTE_STRING;
	  return;
	}
      break;	/* 顼 */

    case SAX_CONFIGURATION_ATTRIBUTE_STRING:
    case SAX_CONFIGURATION_ATTRIBUTE_BOOLEAN:
    case SAX_CONFIGURATION_ATTRIBUTE_INT:
    case SAX_ACCEPTED:
    case SAX_ERROR:
      break;	/* 顼 */

    default:
      cerr << "start_handler is called in unknown state: " << state << "\n";
    }
  state = SAX_ERROR;
}


void
ConfigXMLContext::end_element_handler(const char *name)
{
  switch (state)
    {
    case SAX_OCHUSHA:
      if (strcmp(name, "ochusha") == 0)
	{ state = SAX_ACCEPTED; return; }
      break;	/* 顼 */

    case SAX_CONFIGURATION:
#if 1	/* COMPAT: ߴΤ */
      if (strcmp(name, "configuration") == 0
	  || strcmp(name, "preference") == 0)
#else
      if (strcmp(name, "configuration") == 0)
#endif
	{ state = SAX_OCHUSHA; return; }
      break;	/* 顼 */

    case SAX_CONFIGURATION_ATTRIBUTE:
      if (strcmp(name, "attribute") == 0)
	{
	  state = SAX_CONFIGURATION;
	  attributes.register_current_attribute();
	  return;
	}
      break;	/* 顼 */

    case SAX_CONFIGURATION_ATTRIBUTE_STRING:
      if (strcmp(name, "string") == 0)
	{
	  state = SAX_CONFIGURATION_ATTRIBUTE;
	  return;
	}
      break;	/* 顼 */

    case SAX_CONFIGURATION_ATTRIBUTE_BOOLEAN:
      if (strcmp(name, "boolean") == 0)
	{ state = SAX_CONFIGURATION_ATTRIBUTE; return; }
      break;	/* 顼 */

    case SAX_CONFIGURATION_ATTRIBUTE_INT:
      if (strcmp(name, "int") == 0)
	{ state = SAX_CONFIGURATION_ATTRIBUTE; return; }
      break;	/* 顼 */

    case SAX_INITIAL:
    case SAX_ACCEPTED:
    case SAX_ERROR:
      break;	/* 顼 */

    default:
#ifdef DEBUG_SAX_HANDLER
      cerr << "endHandler called in unknown state: " << state << "\n";
#endif
      break;	/* 顼 */
    }
#ifdef DEBUG_SAX_HANDLER
  cerr << "Invalid document: </" << name << "> appeared in state="
       << state << "\n";
#endif
  state = SAX_ERROR;
}


void
ConfigXMLContext::characters_handler(const char *ch, size_t len)
{
  if (state == SAX_CONFIGURATION_ATTRIBUTE_STRING)
    attributes.append_attr_value(ch, len);
}


void
Configuration::read_config_xml(Repository &repository)
{
  char *pathname = repository.find_file(OCHUSHA_CONFIG_XML);
  if (pathname == NULL)
    {
      config_version = OCHUSHA_CURRENT_CONFIG_VERSION;
      return;
    }

  ConfigXMLContext context(attributes);
  context.parse_xml_file(pathname);

  if (context.is_accepted())
    {
      config_version = attributes.get_int("config_version");

      const char *value = attributes.get_string("bbsmenu_url");
      if (value != NULL)
	set_bbsmenu_uri(value);

      value = attributes.get_string("bbsmenu_uri");
      if (value != NULL && *value != '\0')
	{
	  char *uri = Base64Decoder::decode_text_from_base64url(value);
	  set_bbsmenu_uri(uri);
	  free(uri);
	}

      value = attributes.get_string("login_be_uri");
      if (value != NULL && *value != '\0')
	{
	  char *uri = Base64Decoder::decode_text_from_base64url(value);
	  set_login_be_uri(uri);
	  free(uri);
	}

      value = attributes.get_string("login_viewer_uri");
      if (value != NULL && *value != '\0')
	{
	  char *uri = Base64Decoder::decode_text_from_base64url(value);
	  set_login_viewer_uri(uri);
	  free(uri);
	}

      set_offline(attributes.get_bool("offline"));


      set_enable_proxy_to_read(attributes.get_bool("enable_proxy_to_read"));
      value = attributes.get_string("proxy_uri_to_read");
      if (value != NULL && *value != '\0')
	{
	  char *uri = Base64Decoder::decode_text_from_base64url(value);
	  set_proxy_uri_to_read(uri);
	  free(uri);
	}


      set_enable_proxy_to_post(attributes.get_bool("enable_proxy_to_post"));
      value = attributes.get_string("proxy_uri_to_post");
      if (value != NULL && *value != '\0')
	{
	  char *uri = Base64Decoder::decode_text_from_base64url(value);
	  set_proxy_uri_to_post(uri);
	  free(uri);
	}


      a_bone.read_xml_elements(attributes);
      read_extended_elements(attributes);
    }

  free(pathname);
}

}
