/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: cookie_manager.cc,v 1.7 2009/01/01 07:46:43 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/cookie_manager.h"

#include "libochusha/lock.h"
#include "libochusha/misc_utils.h"
#include "libochusha/sqlite_utils.h"

#include <string>
#include <vector>

#include <sqlite3.h>

#include <cassert>
#include <cctype>
#include <cstdio>
#include <cstdlib>
#include <cstring>


// #define DEBUG_COOKIES		1


namespace Ochusha
{


#define OCHUSHA_COOKIES_TABLE_DEFINITION "CREATE TABLE IF NOT EXISTS ochusha_cookies (id INTEGER PRIMARY KEY, name TEXT, value TEXT, domain TEXT, path TEXT, expiry INTEGER, comment TEXT, commentURL TEXT, version INTEGER, port TEXT, lastAccessed INTEGER, isSecure INTEGER);"


#define COLUMN_ID		0
#define COLUMN_NAME		1
#define COLUMN_VALUE		2
#define COLUMN_DOMAIN		3
#define COLUMN_PATH		4
#define COLUMN_EXPIRY		5
#define COLUMN_COMMENT		6
#define COLUMN_COMMENT_URL	7
#define COLUMN_VERSION		8
#define COLUMN_PORT		9
#define COLUMN_LAST_ACCESSED	10
#define COLUMN_IS_SECURE	11


#define OCHUSHA_COOKIES_DELETE_ALL_COOKIES "DELETE FROM ochusha_cookies;"


#define INDEX_ID		1
#define INDEX_NAME		2
#define INDEX_VALUE		3
#define INDEX_DOMAIN		4
#define INDEX_PATH		5
#define INDEX_EXPIRY		6
#define INDEX_COMMENT		7
#define INDEX_COMMENT_URL	8
#define INDEX_VERSION		9
#define INDEX_PORT		10
#define INDEX_LAST_ACCESSED	11
#define INDEX_IS_SECURE		12

#define OCHUSHA_COOKIES_SELECT_ALL_COOKIE "SELECT * FROM ochusha_cookies;"
#define OCHUSHA_COOKIES_SELECT_MATCHED_DOMAIN "SELECT * FROM ochusha_cookies WHERE domain = ?4;"
#define OCHUSHA_COOKIES_SELECT_MATCHED_COOKIE "SELECT * FROM ochusha_cookies WHERE name = ?2 AND domain = ?4 AND path = ?5;"
#define OCHUSHA_COOKIES_INSERT_NEW_ROW "INSERT INTO ochusha_cookies (name, value, domain, path, expiry, comment, commentURL, version, port, lastAccessed, isSecure) VALUES(?2, ?3, ?4, ?5, ?6, ?7, ?8, ?9, ?10, ?11, ?12);"
#define OCHUSHA_COOKIES_UPDATE_MATCHED_ID "UPDATE ochusha_cookies SET value = ?3, expiry = ?6, comment = ?7, commentURL = ?8, version = ?9, port = ?10, lastAccessed = ?11, isSecure = ?12 WHERE id = ?1;"
#define OCHUSHA_COOKIES_DELETE_MATCHED_ID "DELETE FROM ochusha_cookies WHERE id = ?1;"


CookieManager::CookieManager(const char *db_filename)
  : lock("CookieManager"),
    db_handle(NULL),
    select_all_cookie(NULL),
    select_matched_domain(NULL), select_matched_cookie(NULL),
    insert_new_row(NULL), update_matched_id(NULL), delete_matched_id(NULL)
{
  assert(db_filename != NULL);

  db_handle = new SQLite3(db_filename);

  // tableν
  db_handle->get_table(OCHUSHA_COOKIES_TABLE_DEFINITION);


  // prepared statementν
  select_all_cookie
    = db_handle->statement_new(OCHUSHA_COOKIES_SELECT_ALL_COOKIE);

  select_matched_domain
    = db_handle->statement_new(OCHUSHA_COOKIES_SELECT_MATCHED_DOMAIN);

  select_matched_cookie
    = db_handle->statement_new(OCHUSHA_COOKIES_SELECT_MATCHED_COOKIE);

  insert_new_row
    = db_handle->statement_new(OCHUSHA_COOKIES_INSERT_NEW_ROW);

  update_matched_id
    = db_handle->statement_new(OCHUSHA_COOKIES_UPDATE_MATCHED_ID);

  delete_matched_id
    = db_handle->statement_new(OCHUSHA_COOKIES_DELETE_MATCHED_ID);
}


CookieManager::~CookieManager()
{
  drop_expired_cookies(true);

  delete select_all_cookie;
  delete select_matched_domain;
  delete select_matched_cookie;
  delete insert_new_row;
  delete update_matched_id;
  delete delete_matched_id;
  delete db_handle;
}


void
CookieManager::drop_expired_cookies(bool is_at_exit)
{
  assert(db_handle != NULL);

  LockBlock block(lock);

  time_t now = time(NULL);

  vector<sqlite3_int64> expired_ids;
  select_all_cookie->reset();

  int result;
  while ((result = select_all_cookie->step()) == SQLITE_ROW)
    {
      time_t expiry
	= static_cast<time_t>(select_all_cookie->get_column_int64(COLUMN_EXPIRY));
      if (expiry != -1
	  && !(!is_at_exit && expiry == 0)
	  && expiry < now)
	expired_ids.push_back(select_all_cookie->get_column_int64(COLUMN_ID));
    }

  select_all_cookie->reset();

  if (result != SQLITE_DONE)
    {
      fprintf(stderr, "CookieManager::drop_expired_cookies() failed.\n");
    }

  vector<sqlite3_int64>::iterator iter = expired_ids.begin();
  vector<sqlite3_int64>::iterator end_iter = expired_ids.end();
  while (iter != end_iter)
    {
      sqlite3_int64 id = *iter++;

      delete_matched_id->clear_bindings();
      delete_matched_id->reset();
      if (delete_matched_id->bind(INDEX_ID, id))
	{
	  int del_result;
	  do
	    {
	      del_result = delete_matched_id->step();
	    }
	  while (del_result == SQLITE_ROW);
	}
    }

  delete_matched_id->clear_bindings();
  delete_matched_id->reset();
}


bool
CookieManager::update_cookies_db(HTTPCookies &cookies)
{
  assert(lock.is_locked_by_self());
  assert(db_handle != NULL);

  time_t now = time(NULL);

  vector<HTTPCookie>::const_iterator iter = cookies.begin();
  vector<HTTPCookie>::const_iterator end_iter = cookies.end();

  while (iter != end_iter)
    {
      const HTTPCookie &cookie = *iter++;

      HTTPCookie::HTTPCookieSpec spec = cookie.get_spec();
      const char *name = cookie.get_name();
      const char *value = cookie.get_value();
      const char *domain = cookie.get_domain();
      const char *path = cookie.get_path();
      time_t expiry = cookie.get_expires();
      const char *comment = cookie.get_comment();
      const char *comment_url = cookie.get_comment_url();
      const char *port = cookie.get_port();
      bool secure = cookie.get_secure();

      sqlite3_int64 row_id = 0;
      vector<sqlite3_int64> duplicated_entries;

      assert(name != NULL);
      assert(domain != NULL);
      assert(path != NULL);
#ifdef DEBUG_COOKIES
#define REGULARIZE(text) (text != NULL ? text : "NULL")
      fprintf(stderr,
	      "Cookie {\n"
	      "  spec = %d, name = %s,  value = %s,  domain = %s,  path = %s,\n"
	      "  expiry = %lld,  comment = %s,  comment_url = %s,\n"
	      "  port = %s,  secure = %d\n",
	      spec, name, REGULARIZE(value), domain, path, static_cast<long long>(expiry),
	      REGULARIZE(comment), REGULARIZE(comment_url), REGULARIZE(port),
	      secure);
#endif

      // ǰΤ
      select_matched_cookie->reset();
      select_matched_cookie->clear_bindings();

      int result;
      if (select_matched_cookie->bind_static(INDEX_NAME, name)
	  && select_matched_cookie->bind_static(INDEX_DOMAIN, domain)
	  && select_matched_cookie->bind_static(INDEX_PATH, path))
	{
	  while ((result = select_matched_cookie->step()) == SQLITE_ROW)
	    {
	      if (row_id != 0)
		duplicated_entries.push_back(row_id);
	      row_id = select_matched_cookie->get_column_int64(COLUMN_ID);
	    }
	}
      else
	{
	  // Ϥ褿ƵǽȤ
	  select_matched_cookie->clear_bindings();
	  select_matched_cookie->reset();
	  return false;
	}
      select_matched_cookie->clear_bindings();
      select_matched_cookie->reset();


      if (!duplicated_entries.empty())
	{
	  // Ʊname, domain, pathбrowʣȤʾ
	  // ˤʤä顢äƤޤʤäȤˤ롣

	  duplicated_entries.push_back(row_id);
	  row_id = 0;

	  while (!duplicated_entries.empty())
	    {
	      delete_matched_id->clear_bindings();
	      delete_matched_id->reset();

	      if (delete_matched_id->bind(INDEX_ID, duplicated_entries.back()))
		{
		  int del_result;
		  do
		    {
		      del_result = delete_matched_id->step();
		    }
		  while (del_result == SQLITE_ROW);
		}

	      duplicated_entries.pop_back();
	    }

	  delete_matched_id->clear_bindings();
	  delete_matched_id->reset();
	}


      if (result == SQLITE_DONE)
	{
	  SQLite3Statement *stmt;
	  bool bindings_ok = true;
	  if (row_id == 0)
	    {
	      // INSERT
	      stmt = insert_new_row;
	      stmt->clear_bindings();
	      stmt->reset();

	      bindings_ok = stmt->bind_static(INDEX_NAME, name)
		&& stmt->bind_static(INDEX_DOMAIN, domain)
		&& stmt->bind_static(INDEX_PATH, path);
	    }
	  else
	    {
	      // UPDATE
	      stmt = update_matched_id;
	      stmt->clear_bindings();
	      stmt->reset();

	      bindings_ok = stmt->bind(INDEX_ID, row_id);
	    }

	  if (bindings_ok
	      && stmt->bind_static(INDEX_VALUE, value)
	      && stmt->bind(INDEX_EXPIRY, static_cast<sqlite3_int64>(expiry))
	      && stmt->bind_static(INDEX_COMMENT, comment)
	      && stmt->bind_static(INDEX_COMMENT_URL, comment_url)
	      && stmt->bind(INDEX_VERSION, static_cast<int>(spec))
	      && stmt->bind_static(INDEX_PORT, port)
	      && stmt->bind(INDEX_LAST_ACCESSED, static_cast<sqlite3_int64>(now))
	      && stmt->bind(INDEX_IS_SECURE, secure))
	    stmt->step();

	  stmt->clear_bindings();
	  stmt->reset();
	}
    }

  return true;
}


bool
CookieManager::set_cookie(const URI &uri, const char *value)
{
  if (db_handle->get_db_handle() == NULL)
    return false;

  HTTPCookies cookies;
  bool result = cookies.parse_set_cookie_value(uri, value);

  if (result)
    {
      LockBlock block(lock);
      return update_cookies_db(cookies);
    }
  return false;
}


bool
CookieManager::set_cookie2(const URI &uri, const char *value)
{
  if (db_handle->get_db_handle() == NULL)
    return false;

  HTTPCookies cookies;
  bool result = cookies.parse_set_cookie2_value(uri, value);

  if (result)
    {
      LockBlock block(lock);
      return update_cookies_db(cookies);
    }
  return false;
}


char *
CookieManager::prepare_cookie(const URI &uri)
{
  if (db_handle->get_db_handle() == NULL)
    return NULL;

  time_t now = time(NULL);
  if (now == -1)
    return NULL;

  assert(uri.get_scheme() != NULL);
  assert(uri.get_server() != NULL);

  vector<HTTPCookie> cookies;
  {
    LockBlock block(lock);

    select_matched_domain->clear_bindings();
    select_matched_domain->reset();

    // ʤ顢CookieΥɥᥤϥ֥ɥᥤˤޥåʤФʤʤ
    // ݤʤΤǼ괺ܤ롣
    if (select_matched_domain->bind(INDEX_DOMAIN, uri.get_server()))
      {
	const char *req_path = uri.get_path();
	if (req_path == NULL || req_path[0] == '\0')
	  req_path = "/";
	size_t req_path_len = strlen(req_path);
	bool is_secure_req = strcmp(uri.get_scheme(), "https") == 0;

	while (select_matched_domain->step() == SQLITE_ROW)
	  {
	    const char *name
	      = select_matched_domain->get_column_text(COLUMN_NAME);
	    const char *value
	      = select_matched_domain->get_column_text(COLUMN_VALUE);
	    const char *domain
	      = select_matched_domain->get_column_text(COLUMN_DOMAIN);
	    const char *path
	      = select_matched_domain->get_column_text(COLUMN_PATH);
	    time_t expiry
	      = static_cast<time_t>(select_matched_domain->get_column_int64(COLUMN_EXPIRY));
	    bool secure
	      = select_matched_domain->get_column_int(COLUMN_IS_SECURE);
	    HTTPCookie::HTTPCookieSpec spec
	      = static_cast<HTTPCookie::HTTPCookieSpec>(select_matched_domain->get_column_int(COLUMN_VERSION));

	    if (path == NULL || path[0] == '\0')
	      path = "/";
	    size_t path_len = strlen(path);

	    if ((expiry == -1 || expiry == 0 || expiry >= now)
		&& (!secure || is_secure_req)
		&& path_len <= req_path_len
		&& strncmp(path, req_path, path_len) == 0)
	      {
		// name֤äƤн褬ɬס
		cookies.push_back(HTTPCookie(spec, name, value, domain, path));
	      }
	  }
      }

    select_matched_domain->clear_bindings();
    select_matched_domain->reset();
  }

  vector<HTTPCookie>::iterator iter = cookies.begin();
  vector<HTTPCookie>::iterator end_iter = cookies.end();
  string cookie_value_buf;

  while (iter != end_iter)
    {
      HTTPCookie &cookie = *iter++;
      if (cookie_value_buf.c_str()[0] != '\0')
	cookie_value_buf.append("; ");

      cookie_value_buf.append(cookie.get_name());
      cookie_value_buf.append("=");
      cookie_value_buf.append(cookie.get_value());
    }

  return cookie_value_buf.c_str()[0] != '\0'
    ? strdup(cookie_value_buf.c_str()) : NULL;
}


const HTTPCookie *
CookieManager::lookup_cookie(const URI &uri) const
{
  // APIκƹˤĤ̤
  return NULL;
}


void
CookieManager::delete_cookie(const HTTPCookie *cookie)
{
  // APIκƹˤĤ̤
}


void
CookieManager::delete_all_cookies()
{
  if (db_handle->get_db_handle() == NULL)
    return;

  LockBlock block(lock);
  db_handle->get_table(OCHUSHA_COOKIES_DELETE_ALL_COOKIES);
}


}
