/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: hashtable.cc,v 1.13 2009/01/01 07:46:43 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/hashtable.h"


namespace Ochusha
{


AttributeTable::Entry::Entry(unsigned int hash, const char *key,
			     const char *value, HashtableEntry *next)
  : HashtableEntry(hash, strdup(key), strdup(value), next)
{
}


AttributeTable::Entry::~Entry()
{
  free(const_cast<char *>(key));
  free(const_cast<char *>(value));
}


const char *
AttributeTable::Entry::set_value(const char *value)
{
  return HashtableEntry::set_value(strdup(value));
}


unsigned int
AttributeTable::calculate_hash(const char *key) const
{
  return cstring_case_sensitive_hash(key);
}


bool
AttributeTable::equals(const char *key1, const char *key2) const
{
  return strcmp(key1, key2) == 0;
}


AttributeTable::HashtableEntry *
AttributeTable::entry_new(unsigned int hash, const char *key, const char *value,
			  AttributeTable::HashtableEntry *next)
{
  return new Entry(hash, key, value, next);
}


void
AttributeTable::insert(const char *name, const char *value)
{
  const char *val = lookup(name);
  hashtable<const char *, const char *>::insert(name, value);
  if (val != NULL)
    free(const_cast<char *>(val));
  return;
}


void
AttributeTable::remove(const char *name)
{
  hashtable<const char *, const char *>::remove(name);
  return;
}


RefCountHashtable::Entry::Entry(unsigned int hash, const char *key,
				RefCount *value, HashtableEntry *next)
  : HashtableEntry(hash, strdup(key), value, next)
{
  if (value != NULL)
    value->ref();
}


RefCountHashtable::Entry::~Entry()
{
  if (value != NULL)
    value->unref();
  free(const_cast<char *>(key));
}


RefCount *
RefCountHashtable::Entry::set_value(RefCount *value)
{
  if (value != NULL)
    value->ref();

  RefCount *prev = HashtableEntry::set_value(value);
  if (prev != NULL)
    prev->unref();

  return prev;
}


unsigned int
RefCountHashtable::calculate_hash(const char *key) const
{
  return cstring_case_sensitive_hash(key);
}


bool
RefCountHashtable::equals(const char *key1, const char *key2) const
{
  return strcmp(key1, key2) == 0;
}


RefCountHashtable::HashtableEntry *
RefCountHashtable::entry_new(unsigned int hash, const char *key,
			     RefCount *value,
			     RefCountHashtable::HashtableEntry *next)
{
  return new Entry(hash, key, value, next);
}


CStringToPointerHashtable::Entry::Entry(unsigned int hash, const char *key,
					void *value, HashtableEntry *next)
  : HashtableEntry(hash, strdup(key), value, next)
{
}


CStringToPointerHashtable::Entry::~Entry()
{
  free(const_cast<char *>(key));
}


unsigned int
CStringToPointerHashtable::calculate_hash(const char *key) const
{
  return cstring_case_sensitive_hash(key);
}


bool
CStringToPointerHashtable::equals(const char *key1, const char *key2) const
{
  return strcmp(key1, key2) == 0;
}


CStringToPointerHashtable::HashtableEntry *
CStringToPointerHashtable::entry_new(unsigned int hash, const char *key,
				void *value,
				CStringToPointerHashtable::HashtableEntry *next)
{
  return new Entry(hash, key, value, next);
}

}
