/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: hashtable.h,v 1.19 2010/08/15 15:36:11 fuyu Exp $
 */

#ifndef _LIBOCHUSHA_HASHTABLE_H_
#define _LIBOCHUSHA_HASHTABLE_H_

#ifdef __cplusplus

#if defined(__GNUG__) && !defined(__ICC)
#pragma interface
#endif

#include "libochusha/ref_count.h"

#include <string>

#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <cstdio>
#include <cstring>


using namespace std;
namespace Ochusha
{

#define LOAD_FACTOR	0.75

  template <typename KeyType, typename ValueType>
  class hashtable
  {
  protected:
    struct HashtableEntry
    {
      unsigned int hash;
      KeyType key;
      ValueType value;
      HashtableEntry *next;

      HashtableEntry(unsigned int hash, KeyType key, ValueType value,
		     HashtableEntry *next = NULL)
	: hash(hash), key(key), value(value), next(next) {}
      virtual ~HashtableEntry() { if (next != NULL) delete next; }

      inline void set_next(HashtableEntry *next) { this->next = next; }
      inline HashtableEntry *get_next() { return next; }
      virtual ValueType set_value(ValueType value)
      { ValueType prev = this->value; this->value = value; return prev; }
      inline ValueType get_value() { return value; }
      inline KeyType get_key() { return key; }
    };


  private:
    int count;
    int threshold;
    HashtableEntry **table;
    size_t table_length;

    void rehash()
    {
      int old_capacity = table_length;
      HashtableEntry **old_table = table;

      int new_capacity = old_capacity * 2 + 1;
      HashtableEntry **new_table
	= static_cast<HashtableEntry **>(calloc(new_capacity,
						sizeof(HashtableEntry *)));
      assert(new_table != NULL);
      table = new_table;
      table_length = new_capacity;
      threshold = static_cast<int>(new_capacity * LOAD_FACTOR);

      for (int i = old_capacity; i-- > 0;)
	for (HashtableEntry *old_entry = old_table[i]; old_entry != NULL;)
	  {
	    HashtableEntry *entry = old_entry;
	    old_entry = entry->next;

	    size_t index = entry->hash % new_capacity;
	    entry->next = new_table[index];
	    new_table[index] = entry;
	  }

      free(old_table);
    }


  protected:
    virtual unsigned int calculate_hash(KeyType key) const = 0;
    virtual bool equals(KeyType key1, KeyType key2) const = 0;

    virtual HashtableEntry *entry_new(unsigned int hash, KeyType key,
				      ValueType value,
				      HashtableEntry *next = NULL)
    {
      return new HashtableEntry(hash, key, value, next);
    }


  public:
    hashtable(int initial_capacity = 101)
      : count(0)
    {
      if (initial_capacity <= 0)
	initial_capacity = 101;

      table_length = initial_capacity;
      table = static_cast<HashtableEntry **>(calloc(initial_capacity,
						    sizeof(HashtableEntry *)));
      assert(table != NULL);
      threshold = static_cast<int>(initial_capacity * LOAD_FACTOR);
    }

    virtual ~hashtable()
    {
      for (int i = table_length; i-- > 0;)
	if (table[i] != NULL)
	  delete table[i];

      free(table);
    }

    inline int get_count() const { return count; }
    inline bool is_empty() const { return count == 0; }

    virtual void insert(KeyType key, ValueType value)
    {
      unsigned int hash = calculate_hash(key);
      size_t index = hash % table_length;
      for (HashtableEntry *e = table[index]; e != NULL; e = e->next)
	{
	  if (e->hash == hash && equals(e->key, key))
	    {
	      e->set_value(value);
	      return;
	    }
	}

      if (count >= threshold)
	{
	  rehash();
	  index = hash % table_length;
	}

      HashtableEntry *e = entry_new(hash, key, value, table[index]);
      table[index] = e;
      ++count;
    }

    virtual ValueType lookup(KeyType key) const
    {
      unsigned int hash = calculate_hash(key);
      size_t index = hash % table_length;
      for (HashtableEntry *e = table[index]; e != NULL; e = e->next)
	{
	  if (e->hash == hash && equals(e->key, key))
	    return e->value;
	}
      return NULL;
    }

    virtual void remove(KeyType key)
    {
      unsigned int hash = calculate_hash(key);
      size_t index = hash % table_length;
      for (HashtableEntry *e = table[index], *prev = NULL;
	   e != NULL;
	   prev = e, e = e->next)
	{
	  if (e->hash == hash && equals(e->key, key))
	    {
	      if (prev != NULL)
		prev->next = e->next;
	      else
		table[index] = e->next;
	      --count;
	      e->next= NULL;
	      delete e;
	      return;
	    }
	}
    }

    typedef void (*Scanner)(ValueType value, void *scanner_data, KeyType key);

    virtual void for_each(Scanner scanner, void *scanner_data)
    {
      assert(scanner != NULL);
      for (int i = table_length; i-- > 0;)
	for (HashtableEntry *e = table[i]; e != NULL; e = e->next)
	  (*scanner)(e->value, scanner_data, e->key);
    }

    virtual void clear()
    {
      for (int i = table_length; i-- > 0;)
	if (table[i] != NULL)
	  delete table[i];

      memset(table, 0, sizeof(HashtableEntry *) * table_length);
      count = 0;
    }
  };


  static inline unsigned int cstring_case_sensitive_hash(const char *str)
  {
    const char *tail = str + strlen(str);
    unsigned int h = 0;
    while (str < tail)
      h = 31 * h + *str++;
    return h;
  }


  static inline unsigned int cstring_case_insensitive_hash(const char *str)
  {
    const char *tail = str + strlen(str);
    unsigned int h = 0;
    while (str < tail)
      {
	unsigned int c = *str++;
	h = 31 * h + ((c >= 'a' && c <= 'z') ? c - ('a' - 'A') : c);
      }
    return h;
  }


  class AttributeTable : public hashtable<const char *, const char *>
  {
  protected:
    struct Entry : public HashtableEntry
    {
      Entry(unsigned int hash, const char *key, const char *value,
	    HashtableEntry *next);
      virtual ~Entry();
      virtual const char *set_value(const char *value);
    };


    virtual unsigned int calculate_hash(const char *key) const;
    virtual bool equals(const char *key1, const char *key2) const;
    virtual HashtableEntry *entry_new(unsigned int hash, const char *key,
				      const char *value, HashtableEntry *next);


  private:
    string name_string;
    string value_string;


  public:
    AttributeTable(int initial_capacity = 13)
      : hashtable<const char *, const char *>(initial_capacity) {}

    virtual void insert(const char *name, const char *value);
    virtual void remove(const char *name);

    inline bool get_bool(const char *attr_name) const
    {
      const char *value = lookup(attr_name);
      if (value == NULL)
	return false;

      return strcmp(value, "true") == 0 ? true : false;
    }

    inline int get_int(const char *attr_name) const
    {
      const char *value = lookup(attr_name);
      if (value == NULL)
	return 0;

      int num;
      if (sscanf(value, "%d", &num) == 0)
	num = 0;
      return num;
    }

    inline long get_long(const char *attr_name) const
    {
      const char *value = lookup(attr_name);
      if (value == NULL)
	return 0;

      long num;
      if (sscanf(value, "%ld", &num) == 0)
	num = 0;
      return num;
    }

    inline const char *get_string(const char *attr_name) const
    { return lookup(attr_name); }

    inline char *get_string_copy(const char *attr_name) const
    {
      const char *value = lookup(attr_name);
      return value != NULL ? strdup(value) : NULL;
    }

    inline void set_attr_name(const char *name) { name_string = name; }
    inline void set_attr_value(const char *value) { value_string = value; }
    inline void append_attr_value(const char *value, size_t len)
    { value_string.append(value, len); }
    inline void register_current_attribute()
    {
      if (name_string.length() > 0 && value_string.length() > 0)
	insert(name_string.c_str(), value_string.c_str());
      name_string.erase();
      value_string.erase();
    }
  };


  class RefCountHashtable : public hashtable<const char *, RefCount *>
  {
  protected:
    struct Entry : public HashtableEntry
    {
      Entry(unsigned int hash, const char *key, RefCount *value,
	    HashtableEntry *next);
      virtual ~Entry();
      virtual RefCount *set_value(RefCount *value);
    };


    virtual unsigned int calculate_hash(const char *key) const;
    virtual bool equals(const char *key1, const char *key2) const;
    virtual HashtableEntry *entry_new(unsigned int hash, const char *key,
				      RefCount *value, HashtableEntry *next);


  public:
    RefCountHashtable(int initial_capacity = 101)
      : hashtable<const char *, RefCount *>(initial_capacity) {}
  };


  class CStringToPointerHashtable : public hashtable<const char *, void *>
  {
  protected:
    struct Entry : public HashtableEntry
    {
      Entry(unsigned int hash, const char *key, void *value,
	    HashtableEntry *next);
      virtual ~Entry();
    };

    virtual unsigned int calculate_hash(const char *key) const;
    virtual bool equals(const char *key1, const char *key2) const;
    virtual HashtableEntry *entry_new(unsigned int hash, const char *key,
				      void *value, HashtableEntry *next);


  public:
    CStringToPointerHashtable(int initial_capacity = 101)
      : hashtable<const char *, void *>(initial_capacity) {}
  };

}


#endif	/* __cplusplus */

#endif	/* _LIBOCHUSHA_HASHTABLE_H_ */
