/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: history_manager.cc,v 1.4 2009/01/01 07:46:43 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/history_manager.h"

#include "libochusha/misc_utils.h"
#include "libochusha/sqlite_utils.h"
#include "libochusha/uri_utils.h"

#include <cassert>
#include <cerrno>
#include <cstdio>
#include <cstdlib>
#include <cstring>

#ifdef HAVE_TIME_H
#include <time.h>
#endif


namespace Ochusha
{


History::History(sqlite3_int64 id, const char *url, const char *comment,
		 time_t first_accessed, time_t last_accessed,
		 int number_of_accesses)
  : id(id),
    first_accessed(first_accessed),
    last_accessed(last_accessed),
    number_of_accesses(number_of_accesses)
{
  assert(url != NULL);
  this->url = strdup(url);
  this->comment = comment != NULL ? strdup(comment) : NULL;
}


History::History(const char *url, time_t access_time, const char *comment)
  : id(0),
    first_accessed(0),
    last_accessed(access_time),
    number_of_accesses(0)
{
  assert(url != NULL);
  this->url = strdup(url);
  this->comment = comment != NULL ? strdup(comment) : NULL;
}


History::History(const History &history)
  : id(history.id),
    url(strdup(history.url)),
    first_accessed(history.first_accessed),
    last_accessed(history.last_accessed),
    number_of_accesses(history.number_of_accesses)
{
  if (history.comment != NULL)
    comment = strdup(history.comment);
}


History::~History()
{
  free(url);

  if (comment != NULL)
    free(comment);
}


void
History::set_comment(const char *comment)
{
  if (comment == NULL)
    return;

  if (this->comment != NULL)
    free(this->comment);

  this->comment = comment[0] != '\0' ? strdup(comment) : NULL;
}


#define OCHUSHA_HISTORIES_TABLE_DEFINITION "CREATE TABLE IF NOT EXISTS ochusha_histories (id INTEGER PRIMARY KEY, uri TEXT, comment TEXT, firstAccessed INTEGER, lastAccessed INTEGER, numberOfAccesses INTEGER);"


#define COLUMN_ID			0
#define COLUMN_URI			1
#define COLUMN_COMMENT			2
#define COLUMN_FIRST_ACCESSED		3
#define COLUMN_LAST_ACCESSED		4
#define COLUMN_NUMBER_OF_ACCESSES	5

#define OCHUSHA_HISTORIES_SELECT_ALL_HISTORIES	"SELECT * FROM ochusha_histories;"
#define OCHUSHA_HISTORIES_SELECT_MATCHED_URL	"SELECT * FROM ochusha_histories WHERE uri = ?2;"
#define OCHUSHA_HISTORIES_INSERT_NEW_ROW	"INSERT INTO ochusha_histories (uri, comment, firstAccessed, lastAccessed, numberOfAccesses) VALUES(?2, ?3, ?4, ?5, ?6);"
#define OCHUSHA_HISTORIES_UPDATE_MATCHED_ID	"UPDATE ochusha_histories SET comment = ?3, lastAccessed = ?5, numberOfAccesses = ?6 WHERE id = ?1;"
#define OCHUSHA_HISTORIES_DELETE_MATCHED_ID	"DELETE FROM ochusha_histories WHERE id = ?1;"

#define OCHUSHA_HISTORIES_DELETE_ALL_HISTORIES	"DELETE FROM ochusha_histories;"


#define INDEX_ID			1
#define INDEX_URI			2
#define INDEX_COMMENT			3
#define INDEX_FIRST_ACCESSED		4
#define INDEX_LAST_ACCESSED		5
#define INDEX_NUMBER_OF_ACCESSES	6


HistoryManager::HistoryManager(const char *db_filename,
			       bool allow_duplicated_records)
  : lock("HistoryManager"),
    db_handle(NULL),
    allow_duplicated_records(allow_duplicated_records)
{
  assert(db_filename != NULL);

  db_handle = new SQLite3(db_filename);
  db_handle->get_table(OCHUSHA_HISTORIES_TABLE_DEFINITION);

  select_all_histories
    = db_handle->statement_new(OCHUSHA_HISTORIES_SELECT_ALL_HISTORIES);

  select_matched_url
    = db_handle->statement_new(OCHUSHA_HISTORIES_SELECT_MATCHED_URL);
  
  insert_new_row
    = db_handle->statement_new(OCHUSHA_HISTORIES_INSERT_NEW_ROW);

  update_matched_id
    = db_handle->statement_new(OCHUSHA_HISTORIES_UPDATE_MATCHED_ID);

  delete_matched_id
    = db_handle->statement_new(OCHUSHA_HISTORIES_DELETE_MATCHED_ID);
}


HistoryManager::~HistoryManager()
{
  delete select_all_histories;
  delete select_matched_url;
  delete insert_new_row;
  delete update_matched_id;
  delete delete_matched_id;
  delete db_handle;
}


bool
HistoryManager::has_history(const char *url)
{
  assert(url != NULL);
  assert(url[0] != '\0');

  LockBlock block(lock);

  select_matched_url->clear_bindings();
  select_matched_url->reset();

  bool result;
  if (select_matched_url->bind_static(INDEX_URI, url))
    result = (select_matched_url->step() == SQLITE_ROW);
  else
    result = false;

  select_matched_url->clear_bindings();
  select_matched_url->reset();

  return result;
}


History *
HistoryManager::get_history(const char *url)
{
  if (allow_duplicated_records)
    return NULL;	// ̤APIȤäƤȤǡġ

  assert(url != NULL);
  assert(url[0] != '\0');

  LockBlock block(lock);

  select_matched_url->clear_bindings();
  select_matched_url->reset();

  History *history;
  if (select_matched_url->bind_static(INDEX_URI, url)
      && (select_matched_url->step() == SQLITE_ROW))
    history
      = new History(select_matched_url->get_column_int64(COLUMN_ID),
		    select_matched_url->get_column_text(COLUMN_URI),
		    select_matched_url->get_column_text(COLUMN_COMMENT),
		    select_matched_url->get_column_int64(COLUMN_FIRST_ACCESSED),
		    select_matched_url->get_column_int64(COLUMN_LAST_ACCESSED),
		    select_matched_url->get_column_int(COLUMN_NUMBER_OF_ACCESSES));
  else
    history = NULL;

  select_matched_url->clear_bindings();
  select_matched_url->reset();

  return history;
}


bool
HistoryManager::record(const char *url, const char *comment, time_t access_time)
{
  assert(url != NULL);
  assert(url[0] != '\0');

  if (access_time == static_cast<time_t>(-1))
    {
      access_time = time(NULL);

      if (access_time == static_cast<time_t>(-1))
	{
	  ErrorMessage error_message(errno);
	  fprintf(stderr,
		  "HistoryManager::record(...): time(2) failed:  %s\n",
		  error_message.get_message());
	  return false;
	}
    }

  History *history;
  if (!allow_duplicated_records)
    history = get_history(url);
  else
    history = NULL;

  if (history != NULL)
    {
      if (history->get_last_accessed() < access_time)
	history->last_accessed = access_time;
      ++history->number_of_accesses;
      history->set_comment(comment);
      bool result = update(*history);
      delete history;
      return result;
    }

  LockBlock block(lock);

  insert_new_row->clear_bindings();
  insert_new_row->reset();

  bool result;
  if (insert_new_row->bind_static(INDEX_URI, url)
      && insert_new_row->bind_static(INDEX_COMMENT, comment)
      && insert_new_row->bind(INDEX_FIRST_ACCESSED,
			      static_cast<sqlite3_int64>(access_time))
      && insert_new_row->bind(INDEX_LAST_ACCESSED,
			      static_cast<sqlite3_int64>(access_time))
      && insert_new_row->bind(INDEX_NUMBER_OF_ACCESSES, 1))
    result = (insert_new_row->step() == SQLITE_DONE);
  else
    result = false;

  insert_new_row->clear_bindings();
  insert_new_row->reset();

  return result;
}


bool
HistoryManager::update(const History &history)
{
  assert(history.get_id() != 0);

  LockBlock block(lock);

  update_matched_id->clear_bindings();
  update_matched_id->reset();

  bool result;
  if (update_matched_id->bind(INDEX_ID, history.get_id())
      && update_matched_id->bind_static(INDEX_COMMENT, history.get_comment())
      && update_matched_id->bind(INDEX_LAST_ACCESSED,
				 static_cast<sqlite3_int64>(history.get_last_accessed()))
      && update_matched_id->bind(INDEX_NUMBER_OF_ACCESSES,
				 history.get_number_of_accesses()))
    result = (update_matched_id->step() == SQLITE_DONE);
  else
    result = false;

  update_matched_id->clear_bindings();
  update_matched_id->reset();

  return false;
}


void
HistoryManager::remove(const History &history)
{
  assert(history.get_id() != 0);

  LockBlock block(lock);

  delete_matched_id->clear_bindings();
  delete_matched_id->reset();

  if (delete_matched_id->bind(INDEX_ID, history.get_id()))
    while (delete_matched_id->step() == SQLITE_ROW)
      ;	// IDuniqueʤΤǣǽλϤġġ

  return;
}


void
HistoryManager::remove_all()
{
  if (db_handle->get_db_handle() == NULL)
    return;

  LockBlock block(lock);
  db_handle->get_table(OCHUSHA_HISTORIES_DELETE_ALL_HISTORIES);
}


Histories *
HistoryManager::get_histories()
{
  Histories *histories = new Histories();

  // DB˥ƳǤбHistoryhistoriesɲä롣

  return histories;
}


}
