/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: http_cookies.cc,v 1.9 2009/01/01 07:46:43 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/http_cookies.h"

#include "libochusha/misc_utils.h"
#include "libochusha/uri_utils.h"

#include <cassert>
#include <cctype>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <ctime>
#include <string>


namespace Ochusha
{


static const char *
skip_whitespace(const char *text, const char *tail_pos)
{
  assert(text != NULL);
  while (text < tail_pos
	 && (*text == ' ' || *text == '\t' || *text == '\r' || *text == '\n'))
    ++text;

  return text;
}


class HTTPSyntaxToken
{
public:
  enum HTTPSyntaxTokenType {
    UNKNOWN = 0,
    TOKEN,
    QUOTED_STRING,
    DIGIT,
    HEX,
    COMMENT,
    SEPARATOR,
    CONTROL,
    LINEAR_WHITE_SPACE,
    CRLF
  };


private:
  const char *token_start;
  size_t token_length;
  HTTPSyntaxTokenType token_type;


public:
  HTTPSyntaxToken();

  bool tokenize(const char *start_pos, const char *tail_pos);
  // This is a special version of tokenizer that accept ROGUE cookies.
  bool tokenize_value(const char *start_pos, const char *tail_pos);
  inline const char *get_token_start() const { return token_start; }
  inline size_t get_token_length() const { return token_length; }
  inline HTTPSyntaxTokenType get_token_type() const { return token_type; }

  inline bool is_http_token() const
  { return token_type == TOKEN || token_type == DIGIT || token_type == HEX; }
};


HTTPSyntaxToken::HTTPSyntaxToken()
  : token_start(NULL),
    token_length(0),
    token_type(UNKNOWN)
{
}


bool
HTTPSyntaxToken::tokenize(const char *start_pos, const char *tail_pos)
{
  token_start = start_pos;
  token_length = 0;
  token_type = UNKNOWN;

  if (start_pos == NULL || start_pos >= tail_pos)
    return false;

  int comment_nesting_level = 0;
  const char *cur_pos = start_pos;
  while (cur_pos < tail_pos)
    {
      int c = *cur_pos++;
      switch (token_type)
	{
	case UNKNOWN:
	  if (c == ' ' || c == '\t')
	    {
	      token_type = LINEAR_WHITE_SPACE;
	      continue;
	    }
	  if (c == '"')
	    {
	      token_type = QUOTED_STRING;
	      continue;
	    }
	  if (c == '(')
	    {
	      token_type = COMMENT;
	      comment_nesting_level = 1;
	      continue;
	    }
	  if (isdigit(c))
	    {
	      token_type = DIGIT;
	      continue;
	    }
	  if (isxdigit(c))
	    {
	      token_type = HEX;
	      continue;
	    }
#if 1
	  if (strchr("),;=", c) != NULL)
#else
	  if (strchr(")<>@,;:\\/[]?={}", c) != NULL)
#endif
	    {
	      token_type = SEPARATOR;
	      goto done;
	    }
	  if (c < 0x20 || c == 0x7f)
	    {
	      if (c == '\r')
		{
		  if (cur_pos < tail_pos && *cur_pos == '\n')
		    {
		      token_type = CRLF;
		      ++cur_pos;
		      continue;
		    }
		}
	      token_type = CONTROL;
	      goto done;
	    }
	  token_type = TOKEN;
	  continue;

	case TOKEN:
	  if (strchr("()<>@,;:\\\"/[]?={} \t", c) == NULL
	      && c >= 0x20 && c != 0x7f)
	    continue;
	  --cur_pos;
	  goto done;

	case QUOTED_STRING:
	    if (c == '\\')
	      {
		if (cur_pos < tail_pos)
		  {
		    ++cur_pos;
		    continue;
		  }
		token_type = UNKNOWN;
		goto done;
	      }
	    if (c == '"')
	      goto done;
	    if (cur_pos == tail_pos)
	      {
		token_type = UNKNOWN;	// unterminated quoted-string
		goto done;
	      }
	    continue;

	case DIGIT:
	  if (isdigit(c))
	    continue;
	  if (isxdigit(c))
	    {
	      token_type = HEX;
	      continue;
	    }
	  if (strchr("()<>@,;:\\\"/[]?={} ", c) == NULL
	      && c >= 0x20 && c != 0x7f)
	    {
	      token_type = TOKEN;
	      continue;
	    }
	  --cur_pos;
	  goto done;
	  
	case HEX:
	  if (isxdigit(c))
	    continue;
	  if (strchr("(,;\"= \t\r\n", c) == NULL)
	    {
	      token_type = TOKEN;
	      if (c == '\\')
		{
		  if (cur_pos < tail_pos)
		    ++cur_pos;
		  else
		    {
		      token_type = UNKNOWN;	// unterminated token
		      goto done;
		    }
		}
	      continue;
	    }
	  --cur_pos;
	  goto done;

	case COMMENT:
	  if (c == ')')
	    {
	      if (--comment_nesting_level == 0)
		goto done;
	      continue;
	    }
	  if (c == '\\')
	    {
	      if (cur_pos < tail_pos)
		{
		  ++cur_pos;
		  continue;
		}
	      token_type = UNKNOWN;	// unterminated comment
	      goto done;
	    }
	  if (c == '(')
	    ++comment_nesting_level;
	  if (c > 0x20 && c != 0x7f)
	    continue;
	  token_type = UNKNOWN;		// unterminated comment
	  goto done;

	case LINEAR_WHITE_SPACE:
	  if (c == ' ' || c == '\t')
	    continue;
	  --cur_pos;
	  goto done;

	case CRLF:
	  if (c == ' ' || c == '\t')
	    {
	      token_type = LINEAR_WHITE_SPACE;
	      continue;
	    }
	  --cur_pos;
	  goto done;

	case SEPARATOR:
	case CONTROL:
	  abort();	// never reached.
	}
    }

  assert(token_type != UNKNOWN && token_type != SEPARATOR
	 && token_type != CONTROL);

  if (token_type != TOKEN && token_type != DIGIT && token_type != HEX)
    {
      // odd token
      token_type = UNKNOWN;
    }

 done:
  token_length = cur_pos - start_pos;

#ifdef DEBUG_TOKENIZER
  string token(token_start, token_length);
  fprintf(stderr, "HTTPSyntaxToken::tokenize(): token_type=%d, %s\n",
	  token_type, token.c_str());
#endif

  return token_type != UNKNOWN;
}


/*
 * This is a special version of tokenize() that accepts *ROGUE* Cookies.
 * Although specifications about HTTP Cookies, e.g. Netscape's, RFC2109 and
 * RFC2965, deny appearance of white space at where VALUE placed, there's
 * some instances, i.e. HTTP servers, that generate such Cookies.
 * Ignoring them may be better choice for such case, but It's not clear
 * at this time.
 *
 * `tokenize_value()' accepts a sequence of characters excluding comma and
 * semicolon as the VALUE part of Cookie.  White spaces are treated as part
 * of VALUE.
 */
bool
HTTPSyntaxToken::tokenize_value(const char *start_pos, const char *tail_pos)
{
  token_start = start_pos;
  token_length = 0;
  token_type = UNKNOWN;

  if (start_pos == NULL || start_pos >= tail_pos)
    return false;

  const char *cur_pos = start_pos;
  while (cur_pos < tail_pos)
    {
      int c = *cur_pos++;
      switch (token_type)
	{
	case UNKNOWN:
	  if (c == '"')
	    {
	      token_type = QUOTED_STRING;
	      continue;
	    }
	  if (c == ',' || c == ';')
	    {
	      token_type = SEPARATOR;
	      goto done;
	    }
	  token_type = TOKEN;
	  continue;

	case TOKEN:
	  if (c != ',' && c != ';' && c >= 0x20 && c != 0x7f)
	    continue;
	  --cur_pos;
	  goto done;

	case QUOTED_STRING:
	  if (c == '\\')
	    {
	      if (cur_pos < tail_pos)
		{
		  ++cur_pos;
		  continue;
		}
	      token_type = UNKNOWN;
	      goto done;
	    }
	  if (c == '"')
	    goto done;
	  if (cur_pos == tail_pos)
	    {
	      token_type = UNKNOWN;	// unterminated quoted-string
	      goto done;
	    }
	  continue;

	case SEPARATOR:
	case CONTROL:
	case CRLF:
	case LINEAR_WHITE_SPACE:
	case COMMENT:
	case DIGIT:
	case HEX:
	  abort();	// never reached.
	}
    }

 done:
  token_length = cur_pos - start_pos;

  return token_type != UNKNOWN;
}


class CookieParser
{
private:
  const char *name_pos;
  size_t name_length;
  const char *value_pos;
  size_t value_length;
  const char *expires_pos;
  size_t expires_length;
  const char *path_pos;
  size_t path_length;
  const char *domain_pos;
  size_t domain_length;
  bool secure;
  const char *comment_pos;
  size_t comment_length;
  const char *max_age_pos;
  size_t max_age_length;
  const char *version_pos;
  size_t version_length;
  const char *comment_url_pos;
  size_t comment_url_length;
  bool discard;
  const char *port_pos;
  size_t port_length;


public:
  CookieParser();
  const char *parse_set_cookie_value(const char *start_pos,
				     const char *tail_pos);

  inline const char *get_name_pos() const { return name_pos; }
  inline size_t get_name_length() const { return name_length; }
  inline const char *get_value_pos() const { return value_pos; }
  inline size_t get_value_length() const { return value_length; }
  inline const char *get_expires_pos() const { return expires_pos; }
  inline size_t get_expires_length() const { return expires_length; }
  inline const char *get_path_pos() const { return path_pos; }
  inline size_t get_path_length() const { return path_length; }
  inline const char *get_domain_pos() const { return domain_pos; }
  inline size_t get_domain_length() const { return domain_length; }
  inline bool get_secure() const { return secure; }
  inline const char *get_comment_pos() const { return comment_pos; }
  inline size_t get_comment_length() const { return comment_length; }
  inline const char *get_max_age_pos() const { return max_age_pos; }
  inline size_t get_max_age_length() const { return max_age_length; }
  inline const char *get_version_pos() const { return version_pos; }
  inline size_t get_version_length() const { return version_length; }
  inline const char *get_comment_url_pos() const { return comment_url_pos; }
  inline size_t get_comment_url_length() const { return comment_url_length; }
  inline bool get_discard() const { return discard; }
  inline const char *get_port_pos() const { return port_pos; }
  inline size_t get_port_length() const { return port_length; }
};


CookieParser::CookieParser()
  : name_pos(NULL), name_length(0), value_pos(NULL), value_length(0),
    expires_pos(NULL), expires_length(0), path_pos(NULL), path_length(0),
    domain_pos(NULL), domain_length(0), secure(false),
    comment_pos(NULL), comment_length(0), max_age_pos(NULL), max_age_length(0),
    version_pos(NULL), version_length(0),
    comment_url_pos(NULL), comment_url_length(0), discard(false),
    port_pos(NULL), port_length(0)
{
}


const char *
CookieParser::parse_set_cookie_value(const char *start_pos,
				     const char *tail_pos)
{
  if (name_pos != NULL)
    {
      name_pos = NULL;
      name_length = 0;
      value_pos = NULL;
      value_length = 0;
      expires_pos = NULL;
      expires_length = 0;
      path_pos = NULL;
      path_length = 0;
      domain_pos = NULL;
      domain_length = 0;
      secure = false;
      comment_pos = NULL;
      comment_length = 0;
      max_age_pos = NULL;
      max_age_length = 0;
      version_pos = NULL;
      version_length = 0;
      comment_url_pos = NULL;
      comment_url_length = 0;
      discard = false;
      port_pos = NULL;
      port_length = 0;
    }
#ifndef NDEBUG
  else
    {
      assert(name_length == 0);
      assert(value_pos == NULL && value_length == 0);
      assert(expires_pos == NULL && expires_length == 0);
      assert(path_pos == NULL && path_length == 0);
      assert(domain_pos == NULL && domain_length == 0);
      assert(!secure);
      assert(comment_pos == NULL && comment_length == 0);
      assert(max_age_pos == NULL && max_age_length == 0);
      assert(version_pos == NULL && version_length == 0);
      assert(comment_url_pos == NULL && comment_url_length == 0);
      assert(!discard);
      assert(port_pos == NULL && port_length == 0);
    }
#endif

  enum {
    STATE_INITIAL = 0,			// expect NAME
    STATE_NAME_FOUND,			// expect `='
    STATE_NAME_EQUAL_FOUND,		// expect value
    STATE_VALUE_FOUND,			// expect semi-colon or comma
    STATE_ATTR_INITIAL,			// expect name of an attribute
    STATE_ATTR_NAME_FOUND,		// expect `='
    STATE_ATTR_NAME_EQUAL_FOUND		// expect value
  } state = STATE_INITIAL;

  const char *cur_pos = start_pos;

  const char *attr_name_pos = NULL;
  size_t attr_name_length = 0;
  while (cur_pos < tail_pos)
    {
      HTTPSyntaxToken token;
      if (token.tokenize(cur_pos, tail_pos))
	{
	  HTTPSyntaxToken::HTTPSyntaxTokenType token_type
	    = token.get_token_type();
	  switch (state)
	    {
	    case STATE_INITIAL:
	      if (token.is_http_token())
		{
		  state = STATE_NAME_FOUND;
		  name_pos = token.get_token_start();
		  name_length = token.get_token_length();
		  break;
		}
	      if (token_type == HTTPSyntaxToken::LINEAR_WHITE_SPACE)
		break;
	      if (token_type == HTTPSyntaxToken::SEPARATOR)
		{
		  int c = token.get_token_start()[0];
		  if (c == ',')
		    break;	// We should ignore empty cookie element
		}
	      return NULL;	// error

	    case STATE_NAME_FOUND:
	      if (token_type == HTTPSyntaxToken::SEPARATOR
		  && token.get_token_start()[0] == '=')
		{
#if 1	// temporal hack
		  // we accepts ROGUE Cookies for the time being.
		  cur_pos += token.get_token_length();
		  if (token.tokenize_value(cur_pos, tail_pos))
		    {
		      state = STATE_VALUE_FOUND;

		      token_type = token.get_token_type();
		      if (token_type == HTTPSyntaxToken::SEPARATOR)
			{
			  value_pos = "";
			  value_length = 0;
			  continue;
			}

		      assert(token_type == HTTPSyntaxToken::QUOTED_STRING
			     || token_type == HTTPSyntaxToken::TOKEN);

		      value_pos = token.get_token_start();
		      value_length = token.get_token_length();
		      break;
		    }
#else
		  state = STATE_NAME_EQUAL_FOUND;
#endif
		  break;
		}
	      else
		return NULL;
	      if (token_type != HTTPSyntaxToken::LINEAR_WHITE_SPACE)
		return NULL;
	      break;

	    case STATE_NAME_EQUAL_FOUND:
	      if (token.is_http_token()
		  || token_type == HTTPSyntaxToken::QUOTED_STRING)
		{
		  state = STATE_VALUE_FOUND;
		  value_pos = token.get_token_start();
		  value_length = token.get_token_length();
		  break;
		}
	      if (token_type == HTTPSyntaxToken::LINEAR_WHITE_SPACE)
		break;
	      if (token_type == HTTPSyntaxToken::SEPARATOR)
		{
		  int c = token.get_token_start()[0];
		  if (c == ';' || c == ',')
		    {
		      state = STATE_VALUE_FOUND;
		      value_pos = "";
		      value_length = 0;
		      continue;
		    }
		}
	      return NULL;

	    case STATE_VALUE_FOUND:
	      if (token_type == HTTPSyntaxToken::LINEAR_WHITE_SPACE)
		break;
	      if (token_type == HTTPSyntaxToken::SEPARATOR)
		{
		  int c = token.get_token_start()[0];
		  if (c == ';')
		    {
		      state = STATE_ATTR_INITIAL;
		      break;
		    }
		  if (c == ',')
		    {
		      cur_pos += token.get_token_length();
		      goto done;
		    }
		}
	      goto done;	// invalid separator found -- should be error?

	    case STATE_ATTR_INITIAL:
	      if (token.is_http_token())
		{
		  state = STATE_ATTR_NAME_FOUND;
		  attr_name_pos = token.get_token_start();
		  attr_name_length = token.get_token_length();
		  break;
		}
	      if (token_type == HTTPSyntaxToken::LINEAR_WHITE_SPACE)
		break;
	      if (token_type == HTTPSyntaxToken::SEPARATOR)
		{
		  int c = token.get_token_start()[0];
		  if (c == ';')
		    break;	// just ignore empty attr
		  if (c == ',')
		    {
		      cur_pos += token.get_token_length();
		      goto done;
		    }
		}
	      goto done;

	    case STATE_ATTR_NAME_FOUND:
	      if (token_type == HTTPSyntaxToken::SEPARATOR)
		{
		  int c = token.get_token_start()[0];
		  if (c == '=')
		    {
		      state = STATE_ATTR_NAME_EQUAL_FOUND;
		      break;
		    }
		  if (c == ',' || c == ';')
		    {
		      if (attr_name_length == 6)
			{
			  if (strncasecmp(attr_name_pos, "secure", 6) == 0)
			    secure = true;
			}
		      else if (attr_name_length == 7)
			{
			  if (strncasecmp(attr_name_pos, "discard", 7) == 0)
			    discard = true;
			}

		      if (c == ',')
			{
			  cur_pos += token.get_token_length();
			  goto done;
			}
		      state = STATE_ATTR_INITIAL;
		      attr_name_pos = NULL;
		      attr_name_length = 0;
		      break;
		    }
		  goto done;	// invalid separator found -- should be error?
		}
	      if (token_type == HTTPSyntaxToken::LINEAR_WHITE_SPACE)
		break;
	      goto done;	// unexpected entity found -- should be error?

	    case STATE_ATTR_NAME_EQUAL_FOUND:
	      if (token_type == HTTPSyntaxToken::LINEAR_WHITE_SPACE)
		break;
	      else
		{
		  const char *attr_value_pos;
		  size_t attr_value_length;
		  if (token.is_http_token()
		      || token_type == HTTPSyntaxToken::QUOTED_STRING)
		    {
		      attr_value_pos = token.get_token_start();
		      attr_value_length = token.get_token_length();

		      if (attr_name_length == 7
			  && strncasecmp(attr_name_pos, "expires", 7) == 0
			  && attr_value_length >= 3
			  && (strncasecmp(attr_value_pos, "Sun", 3) == 0
			      || strncasecmp(attr_value_pos, "Mon", 3) == 0
			      || strncasecmp(attr_value_pos, "Tue", 3) == 0
			      || strncasecmp(attr_value_pos, "Wed", 3) == 0
			      || strncasecmp(attr_value_pos, "Thu", 3) == 0
			      || strncasecmp(attr_value_pos, "Fri", 3) == 0
			      || strncasecmp(attr_value_pos, "Sat", 3) == 0))
			{
			  // Should be treated too much specially!

			  // well formed, as specified by Netscape,
			  // DATE is like;
			  // Wdy, DD-Mon-YYYY HH:MM:SS GMT
			  // However, rogue DATE values like;
			  // Weekday, DD-Mon-YYYY HH:MM:SS GMT
			  // are frequently received.

			  cur_pos += attr_value_length;
			  cur_pos = skip_whitespace(cur_pos, tail_pos);
			  if (cur_pos < tail_pos && *cur_pos == ',')
			    ++cur_pos;
			  while (cur_pos < tail_pos
				 && *cur_pos != ';' && *cur_pos != ',')
			    ++cur_pos;

			  expires_pos = attr_value_pos;
			  expires_length = cur_pos - expires_pos;
			  state = STATE_ATTR_INITIAL;
			  attr_name_pos = NULL;
			  continue;
			}
		    }
		  else
		    {
		      attr_value_pos = "";
		      attr_value_length = 0;
		    }
		  if (attr_name_length == 4)
		    {
		      if (strncasecmp(attr_name_pos, "path", 4) == 0)
			{
			  path_pos = attr_value_pos;
			  path_length = attr_value_length;
			}
		      else if (strncasecmp(attr_name_pos, "port", 4) == 0)
			{
			  port_pos = attr_value_pos;
			  port_length = attr_value_length;
			}
		    }
		  else if (attr_name_length == 6)
		    {
		      if (strncasecmp(attr_name_pos, "domain", 6) == 0)
			{
			  domain_pos = attr_value_pos;
			  domain_length = attr_value_length;
			}
		      else if (strncasecmp(attr_name_pos, "secure", 6) == 0)
			{
			  // should this be an error?
			  secure = true;
			}
		    }
		  else if (attr_name_length == 7)
		    {
		      if (strncasecmp(attr_name_pos, "comment", 7) == 0)
			{
			  comment_pos = attr_value_pos;
			  comment_length = attr_value_length;
			}
		      else if (strncasecmp(attr_name_pos, "max-age", 7) == 0)
			{
			  max_age_pos = attr_value_pos;
			  max_age_length = attr_value_length;
			}
		      else if (strncasecmp(attr_name_pos, "version", 7) == 0)
			{
			  version_pos = attr_value_pos;
			  version_length = attr_value_length;
			}
		      else if (strncasecmp(attr_name_pos, "discard", 7) == 0)
			{
			  // should this be an error?
			  discard = true;
			}
		      else if (strncasecmp(attr_name_pos, "expires", 7) == 0)
			{
			  // Only reached here for not-well formed DATE
			  expires_pos = attr_value_pos;
			  expires_length = attr_value_length;
			}
		    }
		  else if (attr_name_length == 10)
		    {
		      if (strncasecmp(attr_name_pos, "commenturl", 10) == 0)
			{
			  comment_url_pos = attr_value_pos;
			  comment_url_length = attr_value_length;
			}
		    }
		  if (token_type == HTTPSyntaxToken::SEPARATOR
		      && token.get_token_start()[0] == ',')
		    {
		      cur_pos += token.get_token_length();
		      goto done;
		    }
		  state = STATE_ATTR_INITIAL;
		  attr_name_pos = 0;
		  break;
		}
	    }
	}
      else
	break;
      cur_pos += token.get_token_length();
    }

 done:
  if (name_pos != NULL)
    return cur_pos;
  return NULL;
}


HTTPCookie::HTTPCookie(HTTPCookieSpec spec, const char *name, const char *value,
		       const char *domain, const char *path, time_t expires,
		       const char *comment, const char *comment_url,
		       const char *port, bool secure)
  : spec(spec),
    name(name != NULL ? strdup(name) : NULL),
    value(value != NULL ? strdup(value) : NULL),
    domain(domain != NULL ? strdup(domain) : NULL),
    path(path != NULL ? strdup(path) : NULL),
    expires(expires),
    comment(comment != NULL ? strdup(comment) : NULL),
    comment_url(comment_url != NULL ? strdup(comment_url) : NULL),
    port(port != NULL ? strdup(port) : NULL),
    secure(secure)
{
}


HTTPCookie::HTTPCookie(HTTPCookieSpec spec, const string *name,
		       const string *value, const string *domain,
		       const string *path, time_t expires,
		       const string *comment, const string *comment_url,
		       const string *port, bool secure)
  : spec(spec),
    name(name != NULL && name->c_str()[0] != '\0'
	 ? strdup(name->c_str()) : NULL),
    value(value != NULL && value->c_str()[0] != '\0'
	  ? strdup(value->c_str()) : NULL),
    domain(domain != NULL && domain->c_str()[0] != '\0'
	   ? strdup(domain->c_str()) : NULL),
    path(path != NULL && path->c_str()[0] != '\0'
	 ? strdup(path->c_str()) : NULL),
    expires(expires),
    comment(comment != NULL && comment->c_str()[0] != '\0'
	    ? strdup(comment->c_str()) : NULL),
    comment_url(comment_url != NULL && comment_url->c_str()[0] != '\0'
		? strdup(comment_url->c_str()) : NULL),
    port(port != NULL && port->c_str()[0] != '\0'
	 ? strdup(port->c_str()) : NULL),
    secure(secure)
{
}


HTTPCookie::HTTPCookie(const HTTPCookie &cookie)
  : spec(cookie.spec),
    name(cookie.name != NULL ? strdup(cookie.name) : NULL),
    value(cookie.value != NULL ? strdup(cookie.value) : NULL),
    domain(cookie.domain != NULL ? strdup(cookie.domain) : NULL),
    path(cookie.path != NULL ? strdup(cookie.path) : NULL),
    expires(cookie.expires),
    comment(cookie.comment != NULL ? strdup(cookie.comment) : NULL),
    comment_url(cookie.comment_url != NULL ? strdup(cookie.comment_url) : NULL),
    port(cookie.port != NULL ? strdup(cookie.port) : NULL),
    secure(cookie.secure)
{
}


HTTPCookie::~HTTPCookie()
{
  if (name != NULL)
    free(name);

  if (value != NULL)
    free(value);

  if (domain != NULL)
    free(domain);

  if (path != NULL)
    free(path);

  if (comment != NULL)
    free(comment);

  if (comment_url != NULL)
    free(comment_url);

  if (port != NULL)
    free(port);
}


void
HTTPCookie::set_name(const char *name_start, size_t length)
{
  if (name != NULL)
    free(name);
  name = name_start != NULL ? strndup(name_start, length) : NULL;
}


void
HTTPCookie::set_value(const char *value_start, size_t length)
{
  if (value != NULL)
    free(value);
  value = value_start != NULL ? strndup(value_start, length) : NULL;
}


void
HTTPCookie::set_domain(const char *domain_start, size_t length)
{
  if (domain != NULL)
    free(domain);
  domain = domain_start != NULL ? strndup(domain_start, length) : NULL;
}


void
HTTPCookie::set_path(const char *path_start, size_t length)
{
  if (path != NULL)
    free(path);
  path = path_start != NULL ? strndup(path_start, length) : NULL;
}


void
HTTPCookie::set_comment(const char *comment_start, size_t length)
{
  if (comment != NULL)
    free(comment);
  comment = comment_start != NULL ? strndup(comment_start, length) : NULL;
}


void
HTTPCookie::set_comment_url(const char *comment_url_start, size_t length)
{
  if (comment_url != NULL)
    free(comment_url);
  comment_url = comment_url_start != NULL
    ? strndup(comment_url_start, length) : NULL;
}


void
HTTPCookie::set_port(const char *port_start, size_t length)
{
  if (port != NULL)
    free(port);
  port = port_start != NULL ? strndup(port_start, length) : NULL;
}


bool
HTTPCookies::parse_set_cookie_value(const URI &uri, const char *value,
				    bool is_set_cookie2)
{
  if (value == NULL)
    return false;

  const char *tail_pos = value + strlen(value);
  const char *cur_pos = skip_whitespace(value, tail_pos);

  while (cur_pos < tail_pos)
    {
      CookieParser cookie;

      cur_pos = cookie.parse_set_cookie_value(cur_pos, tail_pos);
      if (cur_pos != NULL)
	{
	  string name;
	  assert(cookie.get_name_pos() != NULL);
	  assert(cookie.get_name_length() != 0);
	  name.append(cookie.get_name_pos(), cookie.get_name_length());

	  string val;
	  assert(cookie.get_value_pos() != NULL);
	  assert(cookie.get_value_length() != 0);
	  val.append(cookie.get_value_pos(), cookie.get_value_length());

	  string path;
	  if (cookie.get_path_pos() != NULL && cookie.get_path_length() != 0)
	    path.append(cookie.get_path_pos(), cookie.get_path_length());
	  else if (uri.get_path() != NULL && uri.get_path()[0] != '\0')
	    path.append(uri.get_path());
	  else
	    path.append("/");

	  time_t expiry = static_cast<time_t>(-1);
	  if (cookie.get_expires_pos() != NULL
	      && cookie.get_expires_length() != 0)
	    {
	      string datetime(cookie.get_expires_pos(),
			      cookie.get_expires_length());
	      expiry = get_utc_from_date(datetime.c_str());
	    }

	  string domain;
	  if (cookie.get_domain_pos() != NULL
	      && cookie.get_domain_length() != 0)
	    domain.append(cookie.get_domain_pos(), cookie.get_domain_length());
	  else
	    domain.append(uri.get_server());

	  HTTPCookie::HTTPCookieSpec spec;
	  if (is_set_cookie2)
	    spec = HTTPCookie::SPEC_RFC2965;
	  else if (cookie.get_version_pos() != NULL
	      && cookie.get_version_length() != 0)
	    {
	      string version(cookie.get_version_pos(),
			     cookie.get_version_length());
	      int v = -1;
	      if (sscanf(version.c_str(), "%d", &v) == 1)
		{
		  if (v == 1)
		    spec = HTTPCookie::SPEC_RFC2109;
		  else if (v == 2)
		    spec = HTTPCookie::SPEC_RFC2965;
		  else
		    spec = HTTPCookie::SPEC_UNKNOWN;
		}
	      else
		spec = HTTPCookie::SPEC_UNKNOWN;
	    }
	  else
	    spec = HTTPCookie::SPEC_NETSCAPE;

	  string comment;
	  if ((spec == HTTPCookie::SPEC_RFC2109
	       || spec == HTTPCookie::SPEC_RFC2965)
	      && cookie.get_comment_pos() != NULL
	      && cookie.get_comment_length() != 0)
	    {
	      if (cookie.get_max_age_pos() != NULL
		  && cookie.get_max_age_length() != 0)
		{
		  string delta(cookie.get_max_age_pos(),
			       cookie.get_max_age_length());
		  int d = -1;
		  if (sscanf(delta.c_str(), "%d", &d) == 1)
		    {
		      if (d > 0)
			{
			  time_t now = time(NULL);
			  expiry = now + d;
			}
		    }
		}

	      comment.append(cookie.get_comment_pos(),
			     cookie.get_comment_length());
	    }

	  string comment_url;
	  string port;
	  if (spec == HTTPCookie::SPEC_RFC2965)
	    {
	      if (cookie.get_discard())
		expiry = 0;

	      if (cookie.get_comment_url_pos() != NULL
		  && cookie.get_comment_url_length() != 0)
		comment_url.append(cookie.get_comment_url_pos(),
				   cookie.get_comment_url_length());

	      if (cookie.get_port_pos() != NULL
		  && cookie.get_port_length() != 0)
		port.append(cookie.get_port_pos(), cookie.get_port_length());
	    }	      

	  cookies.push_back(HTTPCookie(spec, &name, &val, &domain, &path,
				       expiry, &comment, &comment_url,
				       &port, cookie.get_secure()));
	}
      break;
    }

  return !cookies.empty();
}


bool
HTTPCookies::parse_set_cookie_value(const URI &uri, const char *value)
{
  return parse_set_cookie_value(uri, value, false);
}


bool
HTTPCookies::parse_set_cookie2_value(const URI &uri, const char *value)
{
  return parse_set_cookie_value(uri, value, true);
}


}
