/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: http_handle.cc,v 1.15 2009/01/01 07:46:43 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/http_handle.h"

#include "libochusha/base_utils.h"


#include <cassert>
#include <cstdlib>
#include <cstring>

#include <string>

#include <poll.h>


#define POLLING_INTERVAL_OUTPUT_MILLISEC		0	// 0ms

#define POLLING_INTERVAL_LOW_LATENCY_UNIT		10	// 10ms
#define POLLING_INTERVAL_HIGH_THROUGHPUT_UNIT		50	// 100ms

//#define ENABLE_DEBUG_OUTPUT 1


namespace Ochusha
{

HTTPHandle::HTTPHandle()
  : uri(NULL),
    uri_user(NULL),
    uri_passwd(NULL),
    proxy(NULL),
    proxy_user(NULL),
    proxy_passwd(NULL),
    last_interval(10),
    policy(HTTP_POLICY_LOW_LATENCY),
    state(STATE_INITIAL),
    asynchronous(false),
    nonblock(false)
{
}


HTTPHandle::~HTTPHandle()
{
  if (uri != NULL)
    delete uri;

  if (uri_user != NULL)
    free(uri_user);

  if (uri_passwd != NULL)
    free(uri_passwd);

  if (proxy != NULL)
    delete proxy;

  if (proxy_user != NULL)
    free(proxy_user);

  if (proxy_passwd != NULL)
    free(proxy_passwd);
}


bool
HTTPHandle::set_uri(const char *uri)
{
  if (this->uri != NULL)
    {
      // Should we check the equality between given URI and past URI?
      delete this->uri;
    }

  this->uri = new URI(uri);
  if (!this->uri->is_valid())
    {
      delete this->uri;
      this->uri = NULL;
    }

  connection.set_uri(this->uri);
  return this->uri != NULL;
}


bool
HTTPHandle::set_uri(const URI &uri)
{
  if (this->uri != NULL)
    {
      // Should we check the equality between given URI and the past URI?
      delete this->uri;
    }

  this->uri = new URI(uri);
  if (!this->uri->is_valid())
    {
      delete this->uri;
      this->uri = NULL;
    }

  connection.set_uri(this->uri);
  return this->uri != NULL;
}


bool
HTTPHandle::set_authinfo(const char *user, const char *passwd)
{
  if (uri_user != NULL)
    free(uri_user);

  if (uri_passwd != NULL)
    free(uri_passwd);

  uri_user = user != NULL ? strdup(user) : NULL;
  uri_passwd = passwd != NULL ? strdup(passwd) : NULL;

  return true;
}


bool
HTTPHandle::set_proxy(const char *uri)
{
  if (proxy != NULL)
    {
      // Should we check the equality between given URI and past URI?
      delete proxy;
    }

  proxy = new URI(uri);
  if (!proxy->is_valid())
    {
      delete proxy;
      proxy = NULL;
    }

  connection.set_proxy_uri(proxy);
  return proxy != NULL;
}


bool
HTTPHandle::set_proxy(const URI &uri)
{
  if (proxy != NULL)
    {
      // Should we check the equality between given URI and past URI?
      delete proxy;
    }

  proxy = new URI(uri);
  if (!proxy->is_valid())
    {
      delete proxy;
      proxy = NULL;
    }

  connection.set_proxy_uri(proxy);
  return proxy != NULL;
}


bool
HTTPHandle::set_proxy_authinfo(const char *user, const char *passwd)
{
  if (proxy_user != NULL)
    free(proxy_user);

  if (proxy_passwd != NULL)
    free(proxy_passwd);

  proxy_user = user != NULL ? strdup(user) : NULL;
  proxy_passwd = passwd != NULL ? strdup(passwd) : NULL;

  return true;
}


void
HTTPHandle::set_asynchronous(bool value)
{
  if (asynchronous == value)
    return;

  asynchronous = value;
  connection.set_asynchronous(value);
}


void
HTTPHandle::set_nonblock(bool value)
{
  if (nonblock == value)
    return;

  nonblock = value;
  connection.set_nonblock(value);
}


void
HTTPHandle::set_policy(HTTPPolicy value)
{
  policy = value;
}


static bool
compose_basic_authtoken(string &result_buf, const URI *uri,
			const char *user, const char *pass)
{
  string credentials_src;
  if (user != NULL && *user != '\0')
    {
      credentials_src.append(user).append(":");
      if (pass != NULL)
	credentials_src.append(pass);
    }
  else if (uri != NULL)
    {
      const char *user_part = uri->get_user();
      if (user_part != NULL && *user_part != '\0')
	credentials_src.append(user_part);
    }

  if (credentials_src.length() == 0)
    return false;

  char *credentials
    = Base64Encoder::encode_text_to_base64(credentials_src.c_str());
  if (credentials == NULL)
    return false;

  result_buf.clear();
  result_buf.append("Basic ").append(credentials);

  free(credentials);

  return true;
}


bool
HTTPHandle::prepare()
{
  if (uri == NULL)
    return false;

  if (uri->get_server() == NULL || uri->get_scheme() == NULL)
    return false;

  if (proxy == NULL)
    {
#ifdef ENABLE_OPENSSL
      if (strcmp(uri->get_scheme(), "http") != 0
	  && strcmp(uri->get_scheme(), "https") != 0)
	return false;
#else
      if (strcmp(uri->get_scheme(), "http") != 0)
	return false;
#endif
    }
  else
    {
      if (proxy->get_server() == NULL || proxy->get_scheme() == NULL)
	return false;
#ifdef ENABLE_OPENSSL
      if (strcmp(proxy->get_scheme(), "http") != 0
	  && strcmp(proxy->get_scheme(), "https") != 0)
	return false;
#else
      if (strcmp(proxy->get_scheme(), "http") != 0)
	return false;
#endif
    }

  {
    string authtoken;
    if (compose_basic_authtoken(authtoken, uri, uri_user, uri_passwd))
      {
	request.clear_header(HTTPHeader::Authorization);
	request.clear_header(HTTPHeader::WWW_Authenticate);
	request.set_header(HTTPHeader::Authorization, authtoken.c_str());
      }
    else
      {
	request.clear_header(HTTPHeader::Authorization);
	request.clear_header(HTTPHeader::WWW_Authenticate);
	request.set_header(HTTPHeader::WWW_Authenticate, "");
      }
  }

  {
    string authtoken;
    if (compose_basic_authtoken(authtoken, proxy, proxy_user, proxy_passwd))
      {
	request.set_header(HTTPHeader::Proxy_Authorization, authtoken.c_str());
      }
  }

  connection.set_uri(uri);
  connection.set_proxy_uri(proxy);

  request.set_uri(uri);
  request.set_proxy_uri(proxy);
  return request.prepare();
}


HTTPStatus
HTTPHandle::process()
{
  if (state == STATE_INITIAL)
    {
#ifdef ENABLE_DEBUG_OUTPUT
      fprintf(stderr, "HTTPHandle@%p::process(): <1>\n", this);
#endif
      HTTPStatus result;
      do
	{
	  result = connection.connect();
	  if (result == HTTP_NOT_DONE && asynchronous)
	    return HTTP_NOT_DONE;
	}
      while (result == HTTP_NOT_DONE);

      if (result == HTTP_DONE)
	state = STATE_CONNECTION_ESTABLISHED;
      else
	{
	  assert(result == HTTP_ERROR);
	  state = STATE_CONNECTION_ERROR;
	  return HTTP_ERROR;
	}
    }

  if (state == STATE_CONNECTION_ESTABLISHED)
    {
#ifdef ENABLE_DEBUG_OUTPUT
      fprintf(stderr, "HTTPHandle@%p::process(): <2>\n", this);
#endif
      HTTPStatus result;
      do
	{
	  result = request.send(connection);
	  if (result == HTTP_NOT_DONE && asynchronous)
	    return HTTP_NOT_DONE;
	}
      while (result == HTTP_NOT_DONE);

      if (result == HTTP_DONE)
	state = STATE_REQUEST_DONE;
      else
	{
	  assert(result == HTTP_ERROR);
	  state = STATE_REQUEST_ERROR;
	  return HTTP_ERROR;
	}
    }

  if (state == STATE_REQUEST_DONE)
    {
#ifdef ENABLE_DEBUG_OUTPUT
      fprintf(stderr, "HTTPHandle@%p::process(): <3>\n", this);
#endif
      HTTPStatus result;
      do
	{
	  result = response.receive_status(connection);
	  if (result == HTTP_NOT_DONE && asynchronous)
	    return HTTP_NOT_DONE;
	}
      while (result == HTTP_NOT_DONE);

      if (result == HTTP_DONE)
	state = STATE_RESPONSE_STATUS_RECEIVED;
      else
	{
	  assert(result == HTTP_ERROR);
	  state = STATE_RESPONSE_STATUS_ERROR;
	  return HTTP_ERROR;
	}
    }

  if (state == STATE_RESPONSE_STATUS_RECEIVED)
    {
#ifdef ENABLE_DEBUG_OUTPUT
      fprintf(stderr, "HTTPHandle@%p::process(): <4>\n", this);
#endif
      HTTPStatus result;
      do
	{
	  result = response.receive_headers(connection);
	  if (result == HTTP_NOT_DONE && asynchronous)
	    return HTTP_NOT_DONE;
	}
      while (result == HTTP_NOT_DONE);

      if (result == HTTP_DONE)
	state = STATE_RESPONSE_HEADERS_RECEIVED;
      else
	{
	  assert(result == HTTP_ERROR);
	  state = STATE_RESPONSE_HEADERS_ERROR;
	  return HTTP_ERROR;
	}
    }

  if (state == STATE_RESPONSE_HEADERS_RECEIVED)
    {
#ifdef ENABLE_DEBUG_OUTPUT
      fprintf(stderr, "HTTPHandle@%p::process(): <5>\n", this);
#endif
      if (request.get_request_method() == HTTP_REQUEST_METHOD_HEAD)
	{
	  state = STATE_RESPONSE_BODY_RECEIVED;
	  return HTTP_DONE;
	}

      HTTPStatus result;
      do
	{
	  result = response.receive_body(connection);
	  if (result == HTTP_NOT_DONE && asynchronous)
	    return HTTP_NOT_DONE;
	}
      while (result == HTTP_NOT_DONE);

      if (result == HTTP_DONE)
	state = STATE_RESPONSE_BODY_RECEIVED;
      else
	{
	  assert(result == HTTP_ERROR);
	  state = STATE_RESPONSE_BODY_ERROR;
	  return HTTP_ERROR;
	}
    }

  if (state == STATE_RESPONSE_BODY_RECEIVED)
    {
#ifdef ENABLE_DEBUG_OUTPUT
      fprintf(stderr, "HTTPHandle@%p::process(): <6>\n", this);
#endif
      // Should we treat trailing headers for chunked transfer encoding here?
      return HTTP_DONE;
    }

  return HTTP_ERROR;
}


void
HTTPHandle::flush_response_body_buffer()
{
  response.flush_buffer();
}


void
HTTPHandle::clean()
{
  connection.clean();
  request.clean();
  response.clean();

  state = STATE_INITIAL;

  if (uri != NULL)
    {
      delete uri;
      uri = NULL;
    }

  if (uri_user != NULL)
    {
      free(uri_user);
      uri_user = NULL;
    }

  if (uri_passwd != NULL)
    {
      free(uri_passwd);
      uri_passwd = NULL;
    }

  if (proxy != NULL)
    {
      delete proxy;
      proxy = NULL;
    }

  if (proxy_user != NULL)
    {
      free(proxy_user);
      proxy_user = NULL;
    }

  if (proxy_passwd != NULL)
    {
      free(proxy_passwd);
      proxy_passwd = NULL;
    }

  asynchronous = false;
  nonblock = false;
}


const char *
HTTPHandle::get_error_string()
{
  switch (state)
    {
    case STATE_CONNECTION_ERROR:
      return connection.get_error_string();

    case STATE_REQUEST_ERROR:
      return "cannot send a request";

    case STATE_RESPONSE_STATUS_ERROR:
    case STATE_RESPONSE_HEADERS_ERROR:
    case STATE_RESPONSE_BODY_ERROR:
      return response.get_error_string();

    default:
      return NULL;
    }
}


int
HTTPHandle::get_poll_event_mask()
{
  return connection.get_poll_event_mask();
}


int
HTTPHandle::get_prefered_polling_interval_millisec()
{
  int event_mask = connection.get_poll_event_mask();
  if ((event_mask & POLLOUT) != 0)
    return POLLING_INTERVAL_OUTPUT_MILLISEC;

  if (event_mask == 0)
    return 0;

  int interval = last_interval;
  switch (connection.get_feeling())
    {
    case HTTPTransport::FEELING_HUNGRY:
      interval += 2;
      break;

    case HTTPTransport::FEELING_FULL:
      interval /= 2;	// Should this be interval=0 instead?
      break;

    case HTTPTransport::FEELING_GOOD:
      break;
    }

  if (interval > 20)
    last_interval = 20;	// Too long interval should never good.
  else
    {
      assert(interval >= 0);
      last_interval = interval;
    }

  return policy == HTTP_POLICY_LOW_LATENCY
    ? (last_interval * POLLING_INTERVAL_LOW_LATENCY_UNIT)
    : (last_interval * POLLING_INTERVAL_HIGH_THROUGHPUT_UNIT);
}


bool
HTTPHandle::is_connected() const
{
  return state >= STATE_CONNECTION_ESTABLISHED
    && state != STATE_CONNECTION_ERROR;
}


bool
HTTPHandle::has_request_done() const
{
  return state >= STATE_REQUEST_DONE
    && state != STATE_CONNECTION_ERROR
    && state != STATE_REQUEST_ERROR;
}


bool
HTTPHandle::is_status_available() const
{
  return state >= STATE_RESPONSE_STATUS_RECEIVED
    && state != STATE_CONNECTION_ERROR
    && state != STATE_REQUEST_ERROR
    && state != STATE_RESPONSE_STATUS_ERROR;
}


bool
HTTPHandle::is_headers_available() const
{
  return state >= STATE_RESPONSE_HEADERS_RECEIVED
    && state != STATE_CONNECTION_ERROR
    && state != STATE_REQUEST_ERROR
    && state != STATE_RESPONSE_STATUS_ERROR
    && state != STATE_RESPONSE_HEADERS_ERROR;
}


}
