/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: http_handle.h,v 1.11 2009/01/01 07:46:43 fuyu Exp $
 */

#ifndef _LIBOCHUSHA_HTTP_HANDLE_H_
#define _LIBOCHUSHA_HTTP_HANDLE_H_

#ifdef __cplusplus

#if defined(__GNUG__) && !defined(__ICC)
#pragma interface
#endif

#include "libochusha/http_enums.h"
#include "libochusha/http_headers.h"
#include "libochusha/http_request.h"
#include "libochusha/http_response.h"
#include "libochusha/http_transport.h"

#include <stdint.h>


using namespace std;
namespace Ochusha
{

  class HTTPHandle
  {
  private:
    HTTPTransport connection;
    HTTPRequest request;
    HTTPResponse response;

    URI *uri;
    char *uri_user;
    char *uri_passwd;

    URI *proxy;
    char *proxy_user;
    char *proxy_passwd;

    int last_interval;
    HTTPPolicy policy;

    enum {
      STATE_INITIAL = 0,
      STATE_CONNECTION_ESTABLISHED,
      STATE_REQUEST_DONE,
      STATE_RESPONSE_STATUS_RECEIVED,
      STATE_RESPONSE_HEADERS_RECEIVED,
      STATE_RESPONSE_BODY_RECEIVED,
      STATE_CONNECTION_ERROR,
      STATE_REQUEST_ERROR,
      STATE_RESPONSE_STATUS_ERROR,
      STATE_RESPONSE_HEADERS_ERROR,
      STATE_RESPONSE_BODY_ERROR
    } state;

    bool asynchronous;
    bool nonblock;


  public:
    HTTPHandle();
    virtual ~HTTPHandle();


    /*
     * Sets the URI for a request.
     *
     * This will return false if the URI is invalid.
     */
    bool set_uri(const char *uri);
    bool set_uri(const URI &uri);

    inline const URI *get_uri() const { return uri; }


    /*
     * Sets your username/password pair.
     */
    bool set_authinfo(const char *user, const char *passwd);


    /*
     * Sets the proxy URI for a request.
     *
     * This will return false if the URI is invalid.
     */
    bool set_proxy(const char *uri);
    bool set_proxy(const URI &uri);

    inline const URI *get_proxy() const { return proxy; }


    /*
     * Sets your username/password pair for proxy.
     */
    bool set_proxy_authinfo(const char *user, const char *passwd);


    /*
     * Sets the request method.
     *
     * This will return false if the request type is invalid or unsupported.
     */
    inline bool set_method(HTTPRequestMethod method)
    { return request.set_request_method(method); }


    /*
     * Sets the body.
     *
     * This will return false if the request type doesn't support it.
     */
    inline bool set_body(const char *body, size_t length)
    { return request.set_body(body, length); }


    /*
     * Sets the synchronous/asynchronous mode.
     */
    void set_asynchronous(bool value);


    /*
     * Sets the blocking/nonblocking mode.
     */
    void set_nonblock(bool value);


    /*
     * Sets the policy to be achieved for this connection.
     */
    void set_policy(HTTPPolicy value);


    /*
     * Prepares a request.
     * Call this before trying to process a request or if you change the
     * URI.
     *
     * This will return false if preparation failed.
     */
    bool prepare();


    /*
     * Sets locations for trusted CA certificates.
     * Call this after preparation of a request and before trying to
     * process a request for https request.
     *
     * This is a simple wrapper of SSL_CTX_load_verify_locations() function
     * of OpenSSL.  If ca_file == NULL and ca_path == NULL, verifications
     * of server certificates will be omitted, i.e. blindly trusted.
     */
    inline void load_verify_locations(const char *ca_file, const char *ca_path)
    { connection.set_verify_locations(ca_file, ca_path); }


    /*
     * Sets a random request header.
     */
    inline void set_request_header(const char *name, const char *value)
    { request.set_header(name, value); }


    /*
     * Gets the value of a random request header.
     */
    inline const char *get_request_header(const char *name) const
    { return request.get_header(name); }


    /*
     * Clears the value of a random request header.
     */
    inline void clear_request_header(const char *name)
    { request.clear_header(name); }


    /*
     * Processes a request.
     *
     * This returns the current status of the request being processed.
     */
    HTTPStatus process();


    /*
     * Flushes the buffered received data (so far) to reduce memory usage.
     */
    void flush_response_body_buffer();


    /*
     * Gets the value of a random response header.
     */
    inline const char *get_response_header(const char *name) const
    { return response.get_header(name); }


    /*
     * Gets all response headers.
     */
    inline const HTTPHeaders &get_headers() const
    { return response.get_headers(); }


    /*
     * Aborts a currently running request.
     *
     * This will return false if failed.
     */
    inline bool close() { return connection.close(); }


    /*
     * Cleans a request.
     */
    void clean();


    /*
     * Gets the socket file descriptor associated with a particular
     * connection.
     */
    inline int get_socket() const { return connection.get_socket(); }


    /*
     * Gets the buffered entity body.
     */
    inline const char *get_body() const { return response.get_buffer(); }


    /*
     * Get the length of buffered entity body.
     */
    inline size_t get_body_length() const { return response.get_length(); }


    /*
     * Gets the total length of entity body.
     *
     * This returns the value given by Content-Length response header,
     * otherwise 0 will be returned.
     */
    inline uintmax_t get_content_length() const
    { return response.get_content_length(); }


    /*
     * Gets the total length of entity body received so far.
     */
    inline uintmax_t get_total_length() const
    { return response.get_total_length(); }


    /*
     * Gets the error message for a request that has failed.
     */
    const char *get_error_string();


    /*
     * Returns the status code.
     */
    inline int get_status_code() const { return response.get_status_code(); }


    /*
     * Returns the reason phrase.
     */
    inline const char *get_reason_phrase() const
    { return response.get_reason_phrase(); }


    /*
     * Returns the poll event mask being waited by this particular
     * connection.
     */
    int get_poll_event_mask();


    /*
     * Returns the prefered polling interval in milliseconds for the
     * expected next event.
     */
    int get_prefered_polling_interval_millisec();


    /*
     * Tests whether the connection has established or not.
     */
    bool is_connected() const;


    /*
     * Tests whether the request has been sended or not.
     */
    bool has_request_done() const;


    /*
     * Tests whether the HTTP status has been received or not.
     */
    bool is_status_available() const;


    /*
     * Tests whether the HTTP response headers have been received or not.
     */
    bool is_headers_available() const;


  };

}


#endif	/* __cplusplus */

#endif	/* _LIBOCHUSHA_HTTP_HANDLE_H_ */
