/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: http_headers.cc,v 1.7 2009/01/01 07:46:43 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/http_headers.h"

#include "libochusha/misc_utils.h"

#include <cassert>
#include <cstdlib>
#include <cstring>
#include <string>

// #define DEBUG_COOKIE	1


namespace Ochusha
{

#ifndef HAVE_STRNDUP
static char *
strndup(const char *str, size_t length)
{
  char *result = static_cast<char *>(malloc(length + 1));
  if (result == NULL)
    return result;

  memcpy(result, str, length);
  result[length] = '\0';

  return result;
}
#endif


// general headers (from RFC2616 Section 4.5)
const char * const HTTPHeader::Cache_Control = "Cache-Control";
const char * const HTTPHeader::Connection = "Connection";
const char * const HTTPHeader::Date = "Date";
const char * const HTTPHeader::Pragma = "Pragma";
const char * const HTTPHeader::Trailer = "Trailer";
const char * const HTTPHeader::Transfer_Encoding = "Transfer-Encoding";
const char * const HTTPHeader::Update = "Update";
const char * const HTTPHeader::Via = "Via";
const char * const HTTPHeader::Warning = "Warning";

// request headers (from RFC2616 Section 5.3)
const char * const HTTPHeader::Accept = "Accept";
const char * const HTTPHeader::Accept_Charset = "Accept-Charset";
const char * const HTTPHeader::Accept_Encoding = "Accept-Encoding";
const char * const HTTPHeader::Accept_Language = "Accept-Language";
const char * const HTTPHeader::Authorization = "Authorization";
const char * const HTTPHeader::Expect = "Expect";
const char * const HTTPHeader::From = "From";
const char * const HTTPHeader::Host = "Host";
const char * const HTTPHeader::If_Match = "If-Match";
const char * const HTTPHeader::If_Modified_Since = "If-Modified-Since";
const char * const HTTPHeader::If_None_Match = "If-None-Match";
const char * const HTTPHeader::If_Range = "If-Range";
const char * const HTTPHeader::If_Unmodified_Since = "If-Unmodified-Since";
const char * const HTTPHeader::Max_Forwards = "Max-Forwards";
const char * const HTTPHeader::Proxy_Authorization = "Proxy-Authorization";
const char * const HTTPHeader::Range = "Range";
const char * const HTTPHeader::Referer = "Referer";
const char * const HTTPHeader::TE = "TE";
const char * const HTTPHeader::User_Agent = "User-Agent";

// response headers (from RFC2616 Section 6.2)
const char * const HTTPHeader::Accept_Ranges = "Accept-Ranges";
const char * const HTTPHeader::Age = "Age";
const char * const HTTPHeader::ETag = "ETag";
const char * const HTTPHeader::Location = "Location";
const char * const HTTPHeader::Proxy_Authenticate = "Proxy-Authenticate";
const char * const HTTPHeader::Retry_After = "Retry-After";
const char * const HTTPHeader::Server = "Server";
const char * const HTTPHeader::Vary = "Vary";
const char * const HTTPHeader::WWW_Authenticate = "WWW-Authenticate";

// entity headers (from RFC2616 Section 7.1)
const char * const HTTPHeader::Allow = "Allow";
const char * const HTTPHeader::Content_Encoding = "Content-Encoding";
const char * const HTTPHeader::Content_Language = "Content-Language";
const char * const HTTPHeader::Content_Length = "Content-Length";
const char * const HTTPHeader::Content_Location = "Content-Location";
const char * const HTTPHeader::Content_MD5 = "Content-MD5";
const char * const HTTPHeader::Content_Range = "Content-Range";
const char * const HTTPHeader::Content_Type = "Content-Type";
const char * const HTTPHeader::Expires = "Expires";
const char * const HTTPHeader::Last_Modified = "Last-Modified";

// other headers (from RFC2109 and RFC2965)
const char * const HTTPHeader::Cookie = "Cookie";
const char * const HTTPHeader::Set_Cookie = "Set-Cookie";
const char * const HTTPHeader::Cookie2 = "Cookie2";
const char * const HTTPHeader::Set_Cookie2 = "Set-Cookie2";



HTTPHeader::HTTPHeader(const char *name, const char *value)
  : name(strdup(name)),
    value(strdup(value))
{
}


HTTPHeader::HTTPHeader(const char *name_start, size_t name_length,
		       const char *value_start, size_t value_length)
  : name(strndup(name_start, name_length)),
    value(strndup(value_start, value_length))
{
}


HTTPHeader::HTTPHeader(const HTTPHeader &header)
  : name(strdup(header.name)),
    value(strdup(header.value))
{
}


HTTPHeader::~HTTPHeader()
{
  free(name);
  free(value);
}


void
HTTPHeader::set_value(const char *value)
{
  assert(value != NULL);

  set_value(value, strlen(value));
}


void
HTTPHeader::set_value(const char *value, size_t value_length)
{
  assert(value != NULL);

  if (value_length == 0)
    return;

  string value_buf;
  if (this->value != NULL)
    {
      if (this->value[0] != '\0')
	value_buf.append(this->value).append(", ");
      free(this->value);
    }

  value_buf.append(value, value_length);
  this->value = strdup(value_buf.c_str());
}


HTTPHeader *
HTTPHeaders::lookup_header(const char *name)
{
  vector<HTTPHeader>::iterator iter = headers.begin();
  vector<HTTPHeader>::iterator end_iter = headers.end();
  while (iter != end_iter)
    {
      HTTPHeader &header = *iter++;
      if (strcasecmp(header.get_name(), name) == 0)
	return &header;
    }

  return NULL;
}


HTTPHeader *
HTTPHeaders::lookup_header(const char *name_start, size_t name_length)
{
  vector<HTTPHeader>::iterator iter = headers.begin();
  vector<HTTPHeader>::iterator end_iter = headers.end();
  while (iter != end_iter)
    {
      HTTPHeader &header = *iter++;
      if (strncasecmp(header.get_name(), name_start, name_length) == 0
	  && strlen(header.get_name()) == name_length)
	return &header;
    }

  return NULL;
}


void
HTTPHeaders::clear_value(const char *name)
{
  vector<HTTPHeader>::iterator iter = headers.begin();
  vector<HTTPHeader>::iterator end_iter = headers.end();
  while (iter != end_iter)
    {
      HTTPHeader &header = *iter;
      if (strcasecmp(header.get_name(), name) == 0)
	{
	  headers.erase(iter);
	  return;
	}
      ++iter;
    }
}


bool
HTTPHeaders::set_value(const char *name, const char *value)
{
#ifdef DEBUG_COOKIE
  if (strcasecmp(name, HTTPHeader::Set_Cookie) == 0
      || strcasecmp(name, HTTPHeader::Set_Cookie2) == 0)
    fprintf(stderr, "%s: %s\n", name, value);
#endif

  headers.push_back(HTTPHeader(name, value));

  return true;
}


bool
HTTPHeaders::set_value(const char *name_start, size_t name_length,
		       const char *value_start, size_t value_length)
{
#ifdef DEBUG_COOKIE
  if ((name_length == 10
       && strncasecmp(name_start, HTTPHeader::Set_Cookie, 10) == 0)
      || (name_length == 11
	  && strncasecmp(name_start, HTTPHeader::Set_Cookie2, 11) == 0))
    {
      string name_buf;
      name_buf.append(name_start, name_length);
      string value_buf;
      value_buf.append(value_start, value_length);
      fprintf(stderr, "%s: %s\n", name_buf.c_str(), value_buf.c_str());
    }
#endif

  headers.push_back(HTTPHeader(name_start, name_length,
			       value_start, value_length));

  return true;
}


const char *
HTTPHeaders::get_value(const char *name) const
{
  HTTPHeader *header = const_cast<HTTPHeaders *>(this)->lookup_header(name);
  if (header != NULL)
    return header->value;

  return NULL;
}


const HTTPHeader *
HTTPHeaders::get_header(const char *name) const
{
  return const_cast<HTTPHeaders *>(this)->lookup_header(name);
}


void
HTTPHeaders::clean()
{
  while (!headers.empty())
    headers.pop_back();
}

}
