/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: http_request.cc,v 1.10 2009/01/01 07:46:43 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/http_request.h"

#include "libochusha/http_transport.h"

#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <string>


// #define ENABLE_DEBUG_OUTPUT 1

namespace Ochusha
{

const char * const
HTTPRequest::http_request_method_strings[] = {
  "OPTIONS",
  "GET",
  "HEAD",
  "POST",
  "PUT",
  "DELETE",
  "TRACE",
  "CONNECT"
};


HTTPRequest::HTTPRequest(const URI *uri, const URI *proxy_uri)
  : uri((uri != NULL && uri->is_valid()) ? uri : NULL),
    proxy_uri((proxy_uri != NULL && proxy_uri->is_valid()) ? proxy_uri : NULL),
    method(HTTP_REQUEST_METHOD_OPTIONS),
    major_version(1),
    minor_version(1),
    body(NULL),
    body_length(0),
    state(STATE_INITIAL)
{
}


HTTPRequest::~HTTPRequest()
{
  if (body != NULL)
    free(body);
}


bool
HTTPRequest::set_uri(const URI *uri)
{
  if (state != STATE_INITIAL)
    return false;

  if (uri != NULL && !uri->is_valid())
    {
      uri = NULL;
      return false;
    }

  this->uri = uri;
  return true;
}


bool
HTTPRequest::set_proxy_uri(const URI *proxy_uri)
{
  if (state != STATE_INITIAL)
    return false;

  if (proxy_uri != NULL && !proxy_uri->is_valid())
    {
      this->proxy_uri = NULL;
      return false;
    }

  this->proxy_uri = proxy_uri;
  return true;
}


bool
HTTPRequest::set_request_method(HTTPRequestMethod method)
{
  if (state != STATE_INITIAL)
    return false;

  if (!is_method_supported(method))
    return false;

  this->method = method;
  return true;
}


bool
HTTPRequest::set_body(const char *body, size_t body_length)
{
  if (body == NULL && body_length > 0)
    return false;

  // TRACE MUST NOT include an entity (RFC2616 Section 9.8)
  if (method == HTTP_REQUEST_METHOD_TRACE)
    return false;

  if (this->body != NULL)
    {
      free(this->body);
      this->body = NULL;
      this->body_length = 0;
    }

  if (body != NULL)
    {
      this->body = static_cast<char *>(calloc(body_length, sizeof(char)));
      if (this->body == NULL)
	return false;	// out of memory
      memcpy(this->body, body, body_length);
      this->body_length = body_length;
    }

  return true;
}


bool
HTTPRequest::prepare()
{
  if (state != STATE_INITIAL)
    return false;

  if (uri == NULL || !uri->is_valid() || uri->get_server() == NULL)
    return false;

  if (headers.get_value(HTTPHeader::Host) == NULL)
    headers.set_value(HTTPHeader::Host, uri->get_server());

  if (method != HTTP_REQUEST_METHOD_TRACE && body_length > 0)
    {
      char buf[30];
      snprintf(buf, 30, "%lld", static_cast<long long>(body_length));
      headers.set_value(HTTPHeader::Content_Length, buf);
    }

  if (headers.get_value(HTTPHeader::User_Agent) == NULL)
    headers.set_value(HTTPHeader::User_Agent, PACKAGE "/" VERSION);

  return true;
}


HTTPStatus
HTTPRequest::prepare_request(HTTPTransport &connection)
{
  if (uri == NULL)
    {
      state = STATE_ERROR;
      return HTTP_ERROR;
    }
  assert(uri->is_valid());

  // Compose the Request-Line (RFC2616 Section 5.1)
  string request_line;

  request_line.append(get_request_method_string()).append(" ");
  if (proxy_uri == NULL)
    {
      if (uri->get_path() != NULL)
	{
	  request_line.append(uri->get_path());
	  if (uri->get_query() != NULL)
	    request_line.append("?").append(uri->get_query());
	}
      else
	request_line.append("/");
    }
  else
    {
      assert(proxy_uri->is_valid());
      request_line.append(uri->get_uri());
    }

  char buf[64];
  snprintf(buf, 64, " HTTP/%u.%u", get_major_version(), get_minor_version());
  request_line.append(buf).append(HTTP_CRLF);

  connection.append_data_to_buffer(request_line.c_str(),
				   strlen(request_line.c_str()));
  return HTTP_DONE;
}


HTTPStatus
HTTPRequest::prepare_headers(HTTPTransport &connection)
{
  vector<HTTPHeader>::const_iterator iter = headers.begin();
  vector<HTTPHeader>::const_iterator end_iter = headers.end();

  while (iter != end_iter)
    {
      const HTTPHeader &header = *iter++;
      connection.append_data_to_buffer(header.get_name(),
				       strlen(header.get_name()));
      connection.append_data_to_buffer(": ", 2);
      if (header.get_value() != NULL)
	connection.append_data_to_buffer(header.get_value(),
					 strlen(header.get_value()));
      connection.append_data_to_buffer(HTTP_CRLF, 2);
    }

  connection.append_data_to_buffer(HTTP_CRLF, 2);
  return HTTP_DONE;
}


HTTPStatus
HTTPRequest::prepare_body(HTTPTransport &connection)
{
  if (body_length == 0 || method == HTTP_REQUEST_METHOD_TRACE)
    {
      state = STATE_DONE;
      return HTTP_DONE;
    }

  assert(body != NULL);

  if (headers.get_value(HTTPHeader::Content_Length) == NULL)
    {
      state = STATE_ERROR;
      return HTTP_ERROR;
    }

  connection.append_data_to_buffer(body, body_length);

#ifdef ENABLE_DEBUG_OUTPUT
  fprintf(stderr, "HTTPRequest::prepare_body(): body=\"%s\"\n", body);
#endif

  return HTTP_DONE;
}


HTTPStatus
HTTPRequest::send(HTTPTransport &connection)
{
  if (state == STATE_INITIAL)
    {
#ifdef ENABLE_DEBUG_OUTPUT
      fprintf(stderr, "HTTPRequest::send() <1>\n");
#endif
      HTTPStatus result = prepare_request(connection);

      if (result != HTTP_DONE)
	return result;

      state = STATE_SENDING_REQUEST;
    }

  if (state == STATE_SENDING_REQUEST)
    {
#ifdef ENABLE_DEBUG_OUTPUT
      fprintf(stderr, "HTTPRequest::send() <2>: request=\"%s\"\n",
	      connection.get_incoming_buffer());
#endif

      HTTPStatus result = connection.write_buffer();
      if (result != HTTP_DONE)
	return result;

      connection.reset_buffer();
      result = prepare_headers(connection);
      if (result != HTTP_DONE)
	return result;

      state = STATE_SENDING_HEADERS;
    }

  if (state == STATE_SENDING_HEADERS)
    {
#ifdef ENABLE_DEBUG_OUTPUT
      fprintf(stderr, "HTTPRequest::send() <3>: request=\"%s\"\n",
	      connection.get_incoming_buffer());
#endif
      HTTPStatus result = connection.write_buffer();
      if (result != HTTP_DONE)
	return result;

      connection.reset_buffer();
      result = prepare_body(connection);
      if (result != HTTP_DONE)
	return result;

      if (state == STATE_SENDING_HEADERS)
	state = STATE_SENDING_BODY;
    }

  if (state == STATE_SENDING_BODY)
    {
#ifdef ENABLE_DEBUG_OUTPUT
      fprintf(stderr, "HTTPRequest::send() <4>: request=\"%s\"\n",
	      connection.get_incoming_buffer());
#endif
      HTTPStatus result = connection.write_buffer();
      if (result != HTTP_DONE)
	return result;

      connection.reset_buffer();

      state = STATE_DONE;
      return HTTP_DONE;
    }

#ifdef ENABLE_DEBUG_OUTPUT
  fprintf(stderr, "HTTPRequest::send() <5>: request=\"%s\"\n",
	  connection.get_incoming_buffer());
#endif

  if (state == STATE_DONE)
    return HTTP_DONE;

#ifdef ENABLE_DEBUG_OUTPUT
  fprintf(stderr, "HTTPRequest::send() <6>: request=\"%s\"\n",
	  connection.get_incoming_buffer());
#endif

  assert(state == STATE_ERROR);
  return HTTP_ERROR;
}


void
HTTPRequest::clean()
{
  uri = NULL;
  proxy_uri = NULL;
  method = HTTP_REQUEST_METHOD_OPTIONS;
  major_version = 1;
  minor_version = 1;
  if (body != NULL)
    {
      free(body);
      body = NULL;
    }
  body_length = 0;
  state = STATE_INITIAL;
}


}
