/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: http_response.cc,v 1.20 2009/01/01 07:46:43 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/http_response.h"

#include "libochusha/http_transport.h"
#include "libochusha/misc_utils.h"

#include <cassert>
#include <cctype>
#include <cstdio>
#include <cstdlib>
#include <cstring>

#include <zlib.h>


//#define ENABLE_DEBUG_OUTPUT		1
//#define ENABLE_DEBUG_OUTPUT_MOST	1


namespace Ochusha
{

#define INITIAL_BUFFER_SIZE	4096
#define INFLATION_BUFFER_SIZE	8192


HTTPResponseBodyBuffer::HTTPResponseBodyBuffer()
  : buffer(NULL),
    data_length(0),
    buffer_size(0),
    total_length(0),
    error_description(NULL)
{
}


HTTPResponseBodyBuffer::~HTTPResponseBodyBuffer()
{
  if (buffer != NULL)
    free(buffer);
}


size_t
HTTPResponseBodyBuffer::ensure_free_space(size_t length)
{
  if (buffer_size - data_length >= length)
    return buffer_size - data_length;

  size_t new_size = buffer_size != 0 ? buffer_size * 2 : INITIAL_BUFFER_SIZE;
  while (new_size - data_length < length)
    new_size *= 2;

  char *old_buf = buffer;
  buffer = static_cast<char *>(realloc(buffer, new_size));

  if (buffer != NULL)
    buffer_size = new_size;
  else
    buffer = old_buf;	// out of memory

  return buffer_size - data_length;
}


ssize_t
HTTPResponseBodyBuffer::eat(const char *data, size_t length)
{
  if (length == 0)
    return 0;

  if (ensure_free_space(length + 1) < length + 1)
    {
      error_description = LIB_("Out of memory");
      return -1;
    }

  memcpy(buffer + data_length, data, length);
  data_length += length;
  buffer[data_length] = '\0';

  total_length += length;

  return length;
}


class HTTPResponseBodyBufferZLib : public HTTPResponseBodyBuffer
{
private:
  z_stream zstream;
  char *inflation_buffer;
  enum {
    STATE_INITIAL = 0,
    STATE_INFLATING,
    STATE_DONE,
    STATE_ERROR
  } state;


public:
  HTTPResponseBodyBufferZLib(const char *encoding);
  virtual ~HTTPResponseBodyBufferZLib();
  virtual ssize_t eat(const char *data, size_t length);
};


#define HEAD_CRC     0x02 /* bit 1 set: header CRC present */
#define EXTRA_FIELD  0x04 /* bit 2 set: extra field present */
#define ORIG_NAME    0x08 /* bit 3 set: original file name present */
#define COMMENT      0x10 /* bit 4 set: file comment present */
#define RESERVED     0xE0 /* bits 5..7: reserved */


static ssize_t
check_gzip_header(const unsigned char *buffer, size_t buffer_length)
{
  unsigned int method;	// method byte
  unsigned int flags;	// flags byte
  size_t len;

  static const unsigned char gz_magic[] = { 0x1f, 0x8b, '\0' };
  const unsigned char *cur_pos = buffer;
  const unsigned char *tail_pos = buffer + buffer_length;

  if (buffer_length < 10)
    return 0;		// gzip header has at least 10bytes.

  // find the gzip magic header
  if (buffer[0] != gz_magic[0]
      || buffer[1] != gz_magic[1])
    return -1;

  cur_pos += 2;	// skip magic

  method = *cur_pos++;
  flags = *cur_pos++;

  if (method != Z_DEFLATED || (flags & RESERVED) != 0)
    return -1;

  // discard time, xflags and the OS code
  cur_pos += 6;

  if ((flags & EXTRA_FIELD) != 0)
    {
      if (cur_pos + 2 >= tail_pos)
	return 0;

      // skip the extra field
      len = static_cast<unsigned int>(*cur_pos++);
      len += (static_cast<unsigned int>(*cur_pos++)) << 8;
      cur_pos += len;

      if (cur_pos > tail_pos)
	return 0;
    }

  if ((flags & ORIG_NAME) != 0)
    {
      // skip the original file name
      do
	{
	  if (cur_pos >= tail_pos)
	    return 0;
	}
      while (*cur_pos++ != '\0');
    }

  if ((flags & COMMENT) != 0)
    {
      // skip the .gz file comment
      do
	{
	  if (cur_pos >= tail_pos)
	    return 0;
	}
      while (*cur_pos++ != '\0');
    }

  if ((flags & HEAD_CRC) != 0)
    {
      // skip the header crc
      cur_pos += 2;
      if (cur_pos > tail_pos)
	return 0;
    }

  return cur_pos - buffer;
}


HTTPResponseBodyBufferZLib::HTTPResponseBodyBufferZLib(const char *encoding)
  : inflation_buffer(NULL),
    state(STATE_INITIAL)
{
  zstream.next_in = Z_NULL;
  zstream.avail_in = 0;

  zstream.next_out = Z_NULL;
  zstream.avail_out = 0;

  zstream.msg = Z_NULL;
  zstream.zalloc = Z_NULL;
  zstream.zfree = Z_NULL;
  zstream.opaque = Z_NULL;

  int result;
  if (strcmp(encoding, "gzip") == 0)
    {
      result = inflateInit2(&zstream, -MAX_WBITS);
    }
  else
    {
      result = inflateInit(&zstream);
      state = STATE_INFLATING;
    }

  if (result == Z_OK)
    {
      inflation_buffer = static_cast<char *>(malloc(INFLATION_BUFFER_SIZE));
      if (inflation_buffer != NULL)
	return;
    }

  error_description
    = LIB_("Unexpected initialization error, may be out of memory");
  state = STATE_ERROR;
}


HTTPResponseBodyBufferZLib::~HTTPResponseBodyBufferZLib()
{
  if (inflation_buffer != NULL)
    free(inflation_buffer);

  inflateEnd(&zstream);
}


ssize_t
HTTPResponseBodyBufferZLib::eat(const char *data, size_t length)
{
  if (state == STATE_ERROR)
    return -1;

  if (state == STATE_DONE)
    {
      // Should this be treated as an error?
      // Just drop away trailing garbages at this time.
      return length;
    }

  if (length == 0)
    return 0;

#ifdef ENABLE_DEBUG_OUTPUT_MOST
  fprintf(stderr, "HTTPResponseBodyBufferZLib::eat(%p, %u)\n", data, length);
#endif
  assert(data != NULL);

  if (state == STATE_INITIAL)
    {
      ssize_t gzip_header_len
	= check_gzip_header(reinterpret_cast<const unsigned char *>(data),
			    length);
      if (gzip_header_len < 0)
	{
	  error_description = LIB_("Invalid gzip header");
	  state = STATE_ERROR;
	  return -1;
	}

#ifdef ENABLE_DEBUG_OUTPUT
	  fprintf(stderr, "HTTPResponseBodyBufferZLib:: gzip_header_len=%d\n", gzip_header_len);
#endif

      if (gzip_header_len == 0)
	return 0;		// no enough input so far

      state = STATE_INFLATING;
      zstream.next_in
	= reinterpret_cast<Bytef *>(const_cast<char *>(data + gzip_header_len));
      zstream.avail_in = length - gzip_header_len;
#ifdef ENABLE_DEBUG_OUTPUT
      fprintf(stderr, "HTTPResponseBodyBufferZLib:: gzip_header_len=%d\n", gzip_header_len);
#endif
      if (zstream.avail_in == 0)
	return gzip_header_len;
    }
  else
    {
      assert(state == STATE_INFLATING);
      zstream.next_in = reinterpret_cast<Bytef *>(const_cast<char *>(data));
      zstream.avail_in = length;
    }

  while (true)
    {
      zstream.next_out = reinterpret_cast<Bytef *>(inflation_buffer);
      zstream.avail_out = INFLATION_BUFFER_SIZE;

      int result = inflate(&zstream, Z_SYNC_FLUSH);
#ifdef ENABLE_DEBUG_OUTPUT_MOST
      if (zstream.msg != NULL)
	fprintf(stderr, "HTTPResponseBodyBufferZLib::eat(): result=%d, msg=\"%s\"\n", result, zstream.msg);
      else
	fprintf(stderr, "HTTPResponseBodyBufferZLib::eat(): result=%d\n", result);
#endif
      switch (result)
	{
	case Z_STREAM_END:
	  state = STATE_DONE;
	  /* FALLTHROUGH */
	case Z_OK:
	  {
	    ssize_t inflated_len = INFLATION_BUFFER_SIZE - zstream.avail_out;
	    if (inflated_len > 0)
	      {
		ssize_t len = HTTPResponseBodyBuffer::eat(inflation_buffer,
							  inflated_len);
		if (len != inflated_len)
		  {
		    state = STATE_ERROR;	// out of memory?
		    return -1;
		  }
		if (state != STATE_DONE)
		  continue;
	      }
#ifdef ENABLE_DEBUG_OUTPUT_MOST
	    fprintf(stderr, "HTTPResponseBodyBufferZLib::eat(): inflated_len=%d state=%d, zstream.avail_in=%d\n", inflated_len, state, zstream.avail_in);
#endif
	  }
	  /* FALLTHROUGH */
	case Z_BUF_ERROR:
	  return length - zstream.avail_in;

	case Z_DATA_ERROR:
	  result = inflateSync(&zstream);
	  if (result == Z_OK)
	    continue;
	  if (result == Z_BUF_ERROR)
	    return length - zstream.avail_in;

#ifdef ENABLE_DEBUG_OUTPUT
	  if (zstream.msg != NULL)
	    fprintf(stderr, "HTTPResponseBodyBufferZLib::eat(): <2> result=%d, msg=\"%s\"\n", result, zstream.msg);
	  else
	    fprintf(stderr, "HTTPResponseBodyBufferZLib::eat(): <2> result=%d\n", result);
#endif

	default:
	  if (zstream.msg != NULL)
	    error_description = zstream.msg;
	  else
	    error_description = LIB_("Unrecoverable zlib error");
	  state = STATE_ERROR;
	  return -1;
	}
    }
}


HTTPResponseBodyReader::HTTPResponseBodyReader(HTTPTransport &connection,
					       HTTPResponseBodyBuffer &buffer)
  : connection(connection),
    buffer(buffer),
    error_description(NULL)
{
}


HTTPResponseBodyReader::~HTTPResponseBodyReader()
{
}


static size_t
calculate_chunksize(uintmax_t length)
{
  if (length <= 0x7fffffff)
    return static_cast<size_t>(length);
  return 0x7fffffff;
}


class HTTPResponseBodyReaderContentLength : public HTTPResponseBodyReader
{
private:
  uintmax_t content_length;
  uintmax_t bytes_read;

  enum {
    STATE_INITIAL = 0,
    STATE_READING,
    STATE_DONE,
    STATE_ERROR
  } state;


public:
  HTTPResponseBodyReaderContentLength(HTTPTransport &connection,
				      HTTPResponseBodyBuffer &buffer,
				      uintmax_t length);
  virtual ~HTTPResponseBodyReaderContentLength();

  virtual uintmax_t get_content_length() const;
  virtual HTTPStatus read();
};


HTTPResponseBodyReaderContentLength::HTTPResponseBodyReaderContentLength(
						HTTPTransport &connection,
						HTTPResponseBodyBuffer &buffer,
						uintmax_t length)
  : HTTPResponseBodyReader(connection, buffer),
    content_length(length),
    bytes_read(0),
    state(length > 0 ? STATE_INITIAL : STATE_DONE)
{
}


HTTPResponseBodyReaderContentLength::~HTTPResponseBodyReaderContentLength()
{
}


uintmax_t
HTTPResponseBodyReaderContentLength::get_content_length() const
{
  return content_length;
}


HTTPStatus
HTTPResponseBodyReaderContentLength::read()
{
  if (state == STATE_ERROR)
    return HTTP_ERROR;

  if (state == STATE_DONE)
    return HTTP_DONE;

  if (state == STATE_INITIAL)
    state = STATE_READING;

  uintmax_t bytes_to_read = content_length - bytes_read;
  assert(bytes_to_read > 0);

  while (bytes_to_read > 0)
    {
      size_t unit = calculate_chunksize(bytes_to_read);
      HTTPStatus result = connection.read_into_buffer(unit);
      if (result != HTTP_ERROR)
	{
	  const char *incoming_buffer = connection.get_incoming_buffer();
	  ssize_t len = buffer.eat(incoming_buffer,
				   connection.get_incoming_data_length());
#ifdef ENABLE_DEBUG_OUTPUT_MOST
	  fprintf(stderr, "HTTPResponesBodyReaderContentLength::read(): len=%d, result=%d, bytes_to_read=%lld, bytes_read=%lld\n", len, result, bytes_to_read, bytes_read);
#endif
	  if (len > 0)
	    {
	      connection.clip_buffer(incoming_buffer + len);
	      bytes_to_read -= len;
	      bytes_read += len;
	    }
	  else if (len < 0)
	    {
	      error_description = buffer.get_error_string();
	      state = STATE_ERROR;
	      return HTTP_ERROR;	// out of memory?
	    }

	  if (bytes_read == content_length)
	    {
	      state = STATE_DONE;
	      return HTTP_DONE;
	    }

	  if (result == HTTP_NOT_DONE)
	    return result;

	  if (bytes_to_read > 0 && connection.get_incoming_data_length() > 0)
	    {
	      incoming_buffer = connection.get_incoming_buffer();
	      len = buffer.eat(incoming_buffer,
			       connection.get_incoming_data_length());
	      if (len > 0)
		{
		  connection.clip_buffer(incoming_buffer + len);
		  bytes_to_read -= len;
		  bytes_read += len;

		  if (bytes_read == content_length)
		    {
		      state = STATE_DONE;
		      return HTTP_DONE;
		    }
		}
	    }

	  if (connection.get_last_io_result() == 0 && bytes_to_read > 0)
	    {
	      error_description = LIB_("Unexpected EOF");
	      state = STATE_ERROR;	// unexpected EOF
	      return HTTP_ERROR;
	    }

	  continue;			// read more data
	}
      // error
      error_description = connection.get_error_string();
      state = STATE_ERROR;
      return HTTP_ERROR;
    }

  assert(bytes_read == content_length);
  state = STATE_DONE;
  return HTTP_DONE;
}


class HTTPResponseBodyReaderLegacy : public HTTPResponseBodyReader
{
private:
  enum {
    STATE_INITIAL = 0,
    STATE_READING,
    STATE_DONE,
    STATE_ERROR
  } state;


public:
  HTTPResponseBodyReaderLegacy(HTTPTransport &connection,
			       HTTPResponseBodyBuffer &buffer);
  virtual ~HTTPResponseBodyReaderLegacy();

  virtual uintmax_t get_content_length() const;
  virtual HTTPStatus read();
};


HTTPResponseBodyReaderLegacy::HTTPResponseBodyReaderLegacy(
						HTTPTransport &connection,
						HTTPResponseBodyBuffer &buffer)
  : HTTPResponseBodyReader(connection, buffer)
{
}


HTTPResponseBodyReaderLegacy::~HTTPResponseBodyReaderLegacy()
{
}


uintmax_t
HTTPResponseBodyReaderLegacy::get_content_length() const
{
  return 0;	// unknown
}


HTTPStatus
HTTPResponseBodyReaderLegacy::read()
{
  if (state == STATE_ERROR)
    return HTTP_ERROR;

  if (state == STATE_DONE)
    return HTTP_DONE;

  if (state == STATE_INITIAL)
    state = STATE_READING;

  while (true)
    {
      HTTPStatus result = connection.read_into_buffer(0);
#ifdef ENABLE_DEBUG_OUTPUT_MOST
      fprintf(stderr, "HTTPResponseBodyReaderLegacy::read(): result=%d\n",
	      result);
      fprintf(stderr, "  len=%d, last_io_result=%d\n",
	      connection.get_incoming_data_length(),
	      connection.get_last_io_result());
#endif
      if (result != HTTP_ERROR)
	{
	  const char *incoming_buffer = connection.get_incoming_buffer();
	  ssize_t len = buffer.eat(incoming_buffer,
				   connection.get_incoming_data_length());
	  if (len < 0)
	    {
	      error_description = buffer.get_error_string();
	      state = STATE_ERROR;
	      return HTTP_ERROR;	// out of memory?
	    }

	  if (result == HTTP_NOT_DONE)
	    return result;

	  if (connection.get_last_io_result() == 0)
	    {
	      state = STATE_DONE;	// EOF reached
	      return HTTP_DONE;
	    }

	  continue;			// read more data
	}

      error_description = connection.get_error_string();
      state = STATE_ERROR;
      return HTTP_ERROR;
    }
}


class HTTPResponseBodyReaderChunked : public HTTPResponseBodyReader
{
private:
  size_t current_chunk_size;
  size_t bytes_read;
  enum {
    STATE_INITIAL,
    STATE_READING_CHUNK,
    STATE_DONE,
    STATE_ERROR
  } state;


public:
  HTTPResponseBodyReaderChunked(HTTPTransport &connection,
				HTTPResponseBodyBuffer &buffer);
  virtual ~HTTPResponseBodyReaderChunked();

  virtual uintmax_t get_content_length() const;
  virtual HTTPStatus read();
};


HTTPResponseBodyReaderChunked::HTTPResponseBodyReaderChunked(
						HTTPTransport &connection,
						HTTPResponseBodyBuffer &buffer)
  : HTTPResponseBodyReader(connection, buffer),
    current_chunk_size(0),
    bytes_read(0),
    state(STATE_INITIAL)
{
}


HTTPResponseBodyReaderChunked::~HTTPResponseBodyReaderChunked()
{
}


uintmax_t
HTTPResponseBodyReaderChunked::get_content_length() const
{
  return 0;	// unknown
}


HTTPStatus
HTTPResponseBodyReaderChunked::read()
{
  if (state == STATE_ERROR)
    return HTTP_ERROR;

  if (state == STATE_DONE)
    return HTTP_DONE;

  bool network_accessed = false;
  const char *incoming_buffer = connection.get_incoming_buffer();
  const char *cur_pos = incoming_buffer;

#ifdef ENABLE_DEBUG_OUTPUT_MOST
  fprintf(stderr, "HTTPResponseBodyReaderChunked::read(): incoming_buffer=%p, length=%u\n",
	  incoming_buffer, connection.get_incoming_data_length());
  char *tmp_string = strndup(incoming_buffer,
			     connection.get_incoming_data_length());
  const unsigned char *tmp_pos = (unsigned char *)tmp_string;
  while (*tmp_pos != '\0')
    {
      fprintf(stderr, "%02x \'%c\'\n", *tmp_pos, *tmp_pos);
      ++tmp_pos;
    }
  free(tmp_string);
#endif

  while (true)
    {
      if (state == STATE_INITIAL)
	{
	  // determine the next chunk-size
	  const char *crlf_pos = connection.find_pattern_from(cur_pos,
							      HTTP_CRLF, 2);
	  if (crlf_pos == NULL)
	    {
	      if (network_accessed)
		{
		  connection.clip_buffer(cur_pos);
		  return HTTP_NOT_DONE;
		}

	      // We need more data here.
	      connection.clip_buffer(cur_pos);
	      HTTPStatus result = connection.read_into_buffer(0);
	      if (result != HTTP_ERROR)
		{
		  incoming_buffer = connection.get_incoming_buffer();
		  cur_pos = incoming_buffer;

		  crlf_pos = connection.find_pattern_from(cur_pos,
							  HTTP_CRLF, 2);
		  if (crlf_pos == NULL)
		    {
		      if (result == HTTP_NOT_DONE)
			return HTTP_NOT_DONE;

		      error_description
			= LIB_("Badly formed response: chunk-size not found");
		      state = STATE_ERROR;
		      return HTTP_ERROR;
		    }
		  network_accessed = true;
		}
	      else
		{
		  error_description = connection.get_error_string();
		  state = STATE_ERROR;
		  return HTTP_ERROR;
		}
	    }

	  assert(crlf_pos != NULL);
	  assert(cur_pos < crlf_pos);
	  assert(incoming_buffer <= cur_pos);

	  current_chunk_size = 0;
	  bytes_read = 0;
	  int i = 0;
	  while (cur_pos < crlf_pos)
	    {
	      unsigned int c = *reinterpret_cast<unsigned char *>(const_cast<char *>(cur_pos));
	      if (c >= '0' && c <= '9')
		{
		  current_chunk_size <<= 4;
		  current_chunk_size += c - '0';
		}
	      else if (c >= 'A' && c <= 'F')
		{
		  current_chunk_size <<= 4;
		  current_chunk_size += c - 'A' + 10;
		}
	      else if (c >= 'a' && c <= 'f')
		{
		  current_chunk_size <<= 4;
		  current_chunk_size += c - 'a' + 10;
		}
	      else
		break;

	      ++cur_pos;
	      ++i;
	    }

	  if (i > SIZEOF_SIZE_T)
	    {
	      error_description = LIB_("Too large chunk-size specified");
	      state = STATE_ERROR;	// too large chunk-size
	      return HTTP_ERROR;
	    }

	  // skip trailing white spaces
	  while (cur_pos < crlf_pos && (*cur_pos == ' ' || *cur_pos == '\t'))
	    ++cur_pos;

	  if (cur_pos != crlf_pos && *cur_pos != ';')
	    {
	      error_description
		= LIB_("Badly formed response: chunk-size spec is wrong");
	      state = STATE_ERROR;
	      return HTTP_ERROR;	// invalid chunk-size spec
	    }

	  state = STATE_READING_CHUNK;
	  cur_pos = crlf_pos + 2;
#ifdef ENABLE_DEBUG_OUTPUT
	  fprintf(stderr, "chunk-size: %lu\n",
		  static_cast<unsigned long>(current_chunk_size));
#endif
	}

      assert(state == STATE_READING_CHUNK);

      while (bytes_read < current_chunk_size)
	{
	  size_t left_in_incoming_buffer
	    = connection.get_incoming_data_length() - (cur_pos - incoming_buffer);
	  assert(left_in_incoming_buffer >= 0);
	  size_t bytes_to_read = current_chunk_size - bytes_read;

	  left_in_incoming_buffer = left_in_incoming_buffer < bytes_to_read
	    ? left_in_incoming_buffer : bytes_to_read;

	  if (left_in_incoming_buffer > 0)
	    {
#ifdef ENABLE_DEBUG_OUTPUT_MOST
	      fprintf(stderr, "chunked: bytes_read=%u, bytes_to_read=%u, left_in_incoming_buffer=%u\n", bytes_read, bytes_to_read, left_in_incoming_buffer);
#endif
	      ssize_t len = buffer.eat(cur_pos, left_in_incoming_buffer);
	      if (len < 0)
		{
		  error_description = buffer.get_error_string();
		  state = STATE_ERROR;
		  return HTTP_ERROR;	// out of memory?
		}
	      bytes_to_read -= len;
	      bytes_read += len;
	      cur_pos += len;
	    }

	  if (!network_accessed)
	    {
	      connection.clip_buffer(cur_pos);
	      HTTPStatus result = connection.read_into_buffer(0);
	      if (result != HTTP_ERROR)
		{
		  incoming_buffer = connection.get_incoming_buffer();
		  cur_pos = incoming_buffer;
		  network_accessed = true;
		  continue;
		}
	      error_description = connection.get_error_string();
	      state = STATE_ERROR;
	      return HTTP_ERROR;
	    }

	  break;
	}

      assert(bytes_read <= current_chunk_size);

      if (bytes_read == current_chunk_size)
	{
	  if (current_chunk_size == 0)
	    {
	      state = STATE_DONE;
	      return HTTP_DONE;
	    }

	  const char *crlf_pos
	    = connection.find_pattern_from(cur_pos, HTTP_CRLF, 2);
	  if (cur_pos == crlf_pos)
	    {
	      cur_pos += 2;
	      state = STATE_INITIAL;
	      continue;
	    }

	  if (crlf_pos == NULL)
	    {
	      if (!network_accessed)
		{
		  connection.clip_buffer(cur_pos);
		  HTTPStatus result = connection.read_into_buffer(0);
		  if (result != HTTP_ERROR)
		    {
		      incoming_buffer = connection.get_incoming_buffer();
		      cur_pos = incoming_buffer;
		      network_accessed = true;
		      continue;
		    }

		  error_description = connection.get_error_string();
		  state = STATE_ERROR;
		  return HTTP_ERROR;
		}
	    }
	  else
	    {
	      error_description
		= LIB_("Badly formed response: unterminated chunk");
	      state = STATE_ERROR;
	      return HTTP_ERROR;
	    }
	}

      connection.clip_buffer(cur_pos);
      return HTTP_NOT_DONE;
    }
}


HTTPResponse::HTTPResponse()
  : major_version(1),
    minor_version(1),
    status_code(0),
    reason_phrase(NULL),
    buffer(NULL),
    reader(NULL),
    state(STATE_INITIAL)
{
}


HTTPResponse::~HTTPResponse()
{
  if (reason_phrase != NULL)
    free(reason_phrase);

  if (reader != NULL)
    delete reader;

  if (buffer != NULL)
    delete buffer;
}


HTTPStatus
HTTPResponse::receive_status(HTTPTransport &connection)
{
  char buf[32];

  if (state == STATE_ERROR)
    return HTTP_ERROR;

  if (state >= STATE_RECEIVING_HEADERS)
    return HTTP_DONE;

  HTTPStatus result = connection.read_into_buffer();
  if (result == HTTP_ERROR)
    {
      state = STATE_ERROR;
      error_description.clear();
      error_description.append(LIB_("Cannot receive any responses: "));
      error_description.append(connection.get_error_string());
      return HTTP_ERROR;
    }

  const char *crlf_pos = connection.find_pattern(HTTP_CRLF, 2);
  if (crlf_pos == NULL)
    {
      if (result == HTTP_DONE && connection.get_last_io_result() == 0)
	{
	  state = STATE_ERROR;
	  error_description.clear();
	  error_description.append(LIB_("Too short response"));
	  return HTTP_ERROR;
	}
	    
      return HTTP_NOT_DONE;
    }

  // The Status-Line of HTTP response is from tmp_pos to crlf_pos here.
  const char *tmp_pos = connection.get_incoming_buffer();

  // "HTTP/1.1 200 \r\n" is an example of the valid shortest Status-Line.
  // This has 13 characters before CRLF.
#ifdef ENABLE_DEBUG_OUTPUT
  string tmp_buf;
  tmp_buf.append(tmp_pos, crlf_pos - tmp_pos).append("\n");
  fprintf(stderr, "%s", tmp_buf.c_str());
#endif
  if (crlf_pos - tmp_pos < 14)
    {
      state = STATE_ERROR;
      error_description.clear();
      error_description.append(LIB_("Status-Line too short: "));
      error_description.append("\"");
      error_description.append(tmp_pos, crlf_pos - tmp_pos);
      error_description.append("\"");
      return HTTP_ERROR;
    }


  // Check the comformance of HTTP response.
  if (strncmp(tmp_pos, "HTTP/", 5) != 0)
    {
      state = STATE_ERROR;
      error_description.clear();
      error_description.append(LIB_("Not an HTTP response: "));
      error_description.append("\"");
      error_description.append(tmp_pos, crlf_pos - tmp_pos);
      error_description.append("\"");
      return HTTP_ERROR;
    }

  tmp_pos += 5;

  // let's dig the major version
  const char *delim_pos = tmp_pos;
  while (isdigit(*delim_pos))
    ++delim_pos;

  if (*delim_pos != '.')
    {
      state = STATE_ERROR;
      error_description.clear();
      error_description.append(LIB_("Not a valid HTTP Status-Line: "));
      error_description.append("\"");
      tmp_pos -= 5;
      error_description.append(tmp_pos, crlf_pos - tmp_pos);
      error_description.append("\"");
      return HTTP_ERROR;
    }

  // digits between tmp_pos (included) to delim_pos (excluded) is the
  // major version of HTTP response.
  if (delim_pos - tmp_pos > 31)
    {
      // This should be treated as NOT an error, because very large
      // version number can appear in the far future!
      // But it can be effectively avoided by specifying modest version
      // number in our request.
      state = STATE_ERROR;
      error_description.clear();
      error_description.append(LIB_("Too large major version of response HTTP protocol: "));
      error_description.append("\"");
      tmp_pos -= 5;
      error_description.append(tmp_pos, crlf_pos - tmp_pos);
      error_description.append("\"");
      return HTTP_ERROR;
    }

  memcpy(buf, tmp_pos, delim_pos - tmp_pos);
  buf[delim_pos - tmp_pos] = '\0';
  if (sscanf(buf, "%u", &major_version) != 1)
    {
      state = STATE_ERROR;
      error_description.clear();
      error_description.append(LIB_("IMPLEMENTATION ERROR OCCURED"));
      return HTTP_ERROR;
    }

  tmp_pos = ++delim_pos;

  // let's dig the minor version
  delim_pos = tmp_pos;
  while (isdigit(*delim_pos))
    ++delim_pos;

  if (*delim_pos != ' ')
    {
      state = STATE_ERROR;
      error_description.clear();
      error_description.append(LIB_("Not a valid HTTP Status-Line: "));
      tmp_pos = connection.get_incoming_buffer();
      error_description.append("\"");
      error_description.append(tmp_pos, crlf_pos - tmp_pos);
      error_description.append("\"");
      return HTTP_ERROR;
    }

  // digits between tmp_pos (included) to delim_pos (excluded) is the
  // mainor version of HTTP response.
  if (delim_pos - tmp_pos > 31)
    {
      // This should be treated as NOT an error, because very large
      // version number can appear in the far future!
      state = STATE_ERROR;
      error_description.clear();
      error_description.append(LIB_("Too large minor version of response HTTP protocol: "));
      tmp_pos = connection.get_incoming_buffer();
      error_description.append("\"");
      error_description.append(tmp_pos, crlf_pos - tmp_pos);
      error_description.append("\"");
      return HTTP_ERROR;
    }

  memcpy(buf, tmp_pos, delim_pos - tmp_pos);
  buf[delim_pos - tmp_pos] = '\0';
  if (sscanf(buf, "%u", &minor_version) != 1)
    {
      state = STATE_ERROR;
      error_description.clear();
      error_description.append(LIB_("IMPLEMENTATION ERROR OCCURED"));
      return HTTP_ERROR;
    }

  tmp_pos = ++delim_pos;

  // let's dig the Status-Code
  delim_pos = tmp_pos;
  while (isdigit(*delim_pos))
    ++delim_pos;

  // The Status-Code element is a 3-digit integer result code of the attempt to
  // understand and satisfy the request.  (RFC2616 Section 6.1.1)
  if (delim_pos - tmp_pos != 3 || *delim_pos != ' ')
    {
      state = STATE_ERROR;
      error_description.clear();
      error_description.append(LIB_("Not a valid HTTP Status-Line: "));
      tmp_pos = connection.get_incoming_buffer();
      error_description.append("\"");
      error_description.append(tmp_pos, crlf_pos - tmp_pos);
      error_description.append("\"");
      return HTTP_ERROR;
    }
  status_code
    = (tmp_pos[0] - '0') * 100 + (tmp_pos[1] - '0') * 10 + (tmp_pos[2] - '0');

  tmp_pos = ++delim_pos;
  // The rest of line, i.e. between tmp_pos (included) and crlf_pos (excluded),
  // is the Reason-Phrase.
  assert(reason_phrase == NULL);
  reason_phrase = strndup(tmp_pos, crlf_pos - tmp_pos);

#ifdef ENABLE_DEBUG_OUTPUT
  if (status_code == 500)
    {
      tmp_pos = connection.get_incoming_buffer();
      const char *tail_pos = tmp_pos + connection.get_incoming_data_length();
      int i = 0;
      while (tmp_pos < tail_pos)
	{
	  if (i > 0)
	    fprintf(stderr, " ");

	  int c = *reinterpret_cast<const unsigned char *>(tmp_pos);
	  fprintf(stderr, "%02x(%c)", c, c > 0x20 ? c : ' ');

	  if (++i == 16)
	    {
	      fprintf(stderr, "\n");
	      i = 0;
	    }

	  ++tmp_pos;
	}
      fprintf(stderr, "\n");
    }
#endif

  state = STATE_RECEIVING_HEADERS;

  tmp_pos = connection.find_pattern_from(crlf_pos, HTTP_CRLF HTTP_CRLF, 4);
  size_t term_len = 4;
  if (tmp_pos == NULL)
    {
      // There seems to be broken servers ;-<
      tmp_pos = connection.find_pattern_from(crlf_pos, HTTP_CRLF "\n", 3);
      if (tmp_pos != NULL)
	term_len = 3;
    }

  if (tmp_pos != NULL)
    {
      // incoming buffer seems to have all response headers already.
      if (tmp_pos != crlf_pos)	// otherwise, there's no header!
	{
	  result = parse_headers(crlf_pos + 2, tmp_pos);
	  if (result != HTTP_DONE)
	    return result;
	}
      state = STATE_RECEIVING_BODY;

      connection.clip_buffer(tmp_pos + term_len);
      return HTTP_DONE;
    }

  connection.clip_buffer(crlf_pos + 2);
  return HTTP_DONE;
}


HTTPStatus
HTTPResponse::parse_headers(const char *buf_top, const char *tail_pos)
{
  assert(state == STATE_RECEIVING_HEADERS);
  assert(buf_top != NULL);
  assert(tail_pos != NULL);

  const char *tmp_pos = buf_top;
  const char *delim_pos = tmp_pos;
  while (delim_pos <= tail_pos)
    {
      if (delim_pos[0] == '\r' && delim_pos[1] == '\n'
	  && delim_pos[2] != ' ' && delim_pos[2] != '\t')
	{
	  // tmp_pos (included) to delim_pos (excluded) is a
	  // message header here.
	  const char *colon_pos = tmp_pos;
	  while (*colon_pos != ':' && colon_pos < delim_pos)
	    ++colon_pos;

	  const char *name_start = tmp_pos;
	  size_t name_length = colon_pos - tmp_pos;
	  if (*colon_pos != ':' || name_length == 0)
	    {
	      state = STATE_ERROR;
	      error_description.clear();
	      error_description.append(LIB_("Invalid response-header: "));
	      error_description.append("\"");
	      error_description.append(tmp_pos, delim_pos - tmp_pos);
	      error_description.append("\"");
	      return HTTP_ERROR;
	    }

	  tmp_pos = delim_pos + 2;

	  const char *value_start = colon_pos + 1;
	  while (value_start < delim_pos
		 && (*value_start == ' ' || *value_start == '\t'
		     || *value_start == '\r' || *value_start == '\n'))
	    ++value_start;	// skip leading Linear White Spaces

	  while (value_start < delim_pos
		 && (delim_pos[-1] == ' ' || delim_pos[-1] == '\t'
		     || delim_pos[-1] == '\r' || delim_pos[-1] == '\n'))
	    --delim_pos;	// skip trailing LWS

#if 0
	  if (value_start == delim_pos)
	    {
	      // Should we just ignore this? XXX
	      state = STATE_ERROR;
	      error_description.clear();
	      error_description.append(LIB_("Empty response-header: "));
	      error_description.append("\"");
	      error_description.append(name_start, delim_pos - name_start);
	      error_description.append("\"");
	      return HTTP_ERROR;
	    }
#endif

	  headers.set_value(name_start, name_length,
			    value_start, delim_pos - value_start);
#ifdef ENABLE_DEBUG_OUTPUT
	  if (value_start == delim_pos)
	    {
	      char *tmp_name = strndup(name_start, name_length);
	      char *tmp_val = strndup(value_start, delim_pos - value_start);
	      fprintf(stderr, "response-header: %s: %s\n", tmp_name, tmp_val);
	      free(tmp_name);
	      free(tmp_val);
	    }
#endif

	  delim_pos = tmp_pos;
	}
      else
	++delim_pos;
    }

  return HTTP_DONE;
}


HTTPStatus
HTTPResponse::receive_headers(HTTPTransport &connection)
{
  if (state == STATE_ERROR)
    return HTTP_ERROR;

  if (state >= STATE_RECEIVING_BODY)
    return HTTP_DONE;

  HTTPStatus result = connection.read_into_buffer();
  if (result == HTTP_ERROR)
    {
      state = STATE_ERROR;
      error_description.clear();
      error_description.append(LIB_("Cannot receive any headers: "));
      error_description.append(connection.get_error_string());
      return HTTP_ERROR;
    }

  const char *crlf_pos = connection.find_pattern(HTTP_CRLF, 2);
  const char *crlfcrlf_pos = connection.find_pattern(HTTP_CRLF HTTP_CRLF, 4);
  size_t term_len = 4;
  if (crlf_pos != NULL && crlfcrlf_pos == NULL)
    {
      // There seems to be broken servers ;-<
      crlfcrlf_pos = connection.find_pattern(HTTP_CRLF "\n", 3);
      if (crlfcrlf_pos != NULL)
	term_len = 3;
    }

  if (crlf_pos == NULL || crlfcrlf_pos == NULL)
    {
      if (crlf_pos != NULL && crlf_pos == connection.get_incoming_buffer())
	{
	  // There's no header!
	  state = STATE_RECEIVING_BODY;

	  connection.clip_buffer(crlf_pos + 2);
	  return HTTP_DONE;
	}

      if (result == HTTP_DONE && connection.get_last_io_result() == 0)
	{
	  state = STATE_ERROR;
	  error_description.clear();
	  error_description.append(LIB_("Too short response"));
	  return HTTP_ERROR;
	}

      return HTTP_NOT_DONE;
    }

#ifdef ENABLE_DEBUG_OUTPUT_MOST
  fprintf(stderr, "HTTPResponse::receive_headers(): <1>\n");
#endif

  result = parse_headers(connection.get_incoming_buffer(), crlfcrlf_pos);

#ifdef ENABLE_DEBUG_OUTPUT_MOST
  fprintf(stderr, "HTTPResponse::receive_headers(): <2> result=%d\n", result);
#endif

  if (result != HTTP_DONE)
    return result;

#ifdef ENABLE_DEBUG_OUTPUT_MOST
  fprintf(stderr, "HTTPResponse::receive_headers(): <3>\n");
#endif

  state = STATE_RECEIVING_BODY;
  connection.clip_buffer(crlfcrlf_pos + term_len);
  return HTTP_DONE;
}


HTTPStatus
HTTPResponse::receive_body(HTTPTransport &connection)
{
  if (state == STATE_ERROR)
    return HTTP_ERROR;

  if (state == STATE_DONE)
    return HTTP_DONE;

  if (reader == NULL)
    {
      const char *content_encoding_value
	= headers.get_value(HTTPHeader::Content_Encoding);

      if (content_encoding_value != NULL
	  && (strcasecmp(content_encoding_value, "gzip") == 0
	      || strcasecmp(content_encoding_value, "deflate") == 0
	      || strcasecmp(content_encoding_value, "zlib") == 0))
	buffer = new HTTPResponseBodyBufferZLib(content_encoding_value);
      else
	buffer = new HTTPResponseBodyBuffer();
	  

      const char *transfer_encoding_value
	= headers.get_value(HTTPHeader::Transfer_Encoding);

      if (transfer_encoding_value != NULL)
	{
	  if (strcasecmp(transfer_encoding_value, "chunked") == 0)
	    reader = new HTTPResponseBodyReaderChunked(connection, *buffer);
	  else
	    {
	      state = STATE_ERROR;
	      error_description.clear();
	      error_description.append(LIB_("Unknown token for Transfer-Encoding: "));
	      error_description.append(transfer_encoding_value);
	      return HTTP_ERROR;
	    }
	}
      else
	{
	  const char *content_length_value
	    = headers.get_value(HTTPHeader::Content_Length);
	  if (content_length_value != NULL)
	    {
	      uintmax_t content_length = 0;
	      const unsigned char *digit
		= reinterpret_cast<const unsigned char *>(content_length_value);
	      while (isdigit(*digit))
		{
		  content_length *= 10;
		  content_length += *digit - '0';
		  ++digit;
		}

	      if (*digit != '\0' || *content_length_value == '\0')
		{
		  state = STATE_ERROR;
		  error_description.clear();
		  error_description.append(LIB_("Invalid Content-Length: "));
		  error_description.append(content_length_value);
		  return HTTP_ERROR;
		}

	      reader = new HTTPResponseBodyReaderContentLength(connection,
							       *buffer,
							       content_length);
	    }
	  else if (major_version == 1 && minor_version == 1)
	    {
	      // empty message-body
	      state = STATE_DONE;
	      return HTTP_DONE;
	    }
	  else
	    reader = new HTTPResponseBodyReaderLegacy(connection, *buffer);
	}

      if (reader == NULL)
	{
	  state = STATE_ERROR;
	  error_description.clear();
	  error_description.append(LIB_("Out of memory"));
	  return HTTP_ERROR;
	}
    }

  HTTPStatus result = reader->read();
  if (result != HTTP_DONE)
    return result;

  const char *connection_value
    = headers.get_value(HTTPHeader::Connection);
  if (connection_value != NULL && strcasecmp(connection_value, "close") == 0)
    connection.close();

  return HTTP_DONE;
}


void
HTTPResponse::clean()
{
  major_version = 1;
  minor_version = 1;
  status_code = 0;
  if (reason_phrase != NULL)
    {
      free(reason_phrase);
      reason_phrase = NULL;
    }

  headers.clean();

  if (buffer != NULL)
    {
      delete buffer;
      buffer = NULL;
    }

  if (reader != NULL)
    {
      delete reader;
      reader = NULL;
    }

  error_description.clear();
}

}
