/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: http_response.h,v 1.10 2010/08/15 15:36:11 fuyu Exp $
 */

#ifndef _LIBOCHUSHA_HTTP_RESPONSE_H_
#define _LIBOCHUSHA_HTTP_RESPONSE_H_

#ifdef __cplusplus

#if defined(__GNUG__) && !defined(__ICC)
#pragma interface
#endif

#include "libochusha/http_enums.h"
#include "libochusha/http_headers.h"

#if defined(HAVE_STDINT_H) || !defined(HAVE_CONFIG_H)
#include <stdint.h>
#else
#ifdef HAVE_INTTYPES_H
#include <inttypes.h>
#endif
#endif

#include <string>
#include <stdint.h>


using namespace std;
namespace Ochusha
{

  class HTTPResponseBodyBuffer
  {
  private:
    char *buffer;
    size_t data_length;
    size_t buffer_size;
    uintmax_t total_length;

    size_t ensure_free_space(size_t length);


  protected:
    const char *error_description;


  public:
    HTTPResponseBodyBuffer();
    virtual ~HTTPResponseBodyBuffer();
    virtual ssize_t eat(const char *data, size_t length);

    inline uintmax_t get_total_length() const { return total_length; }
    inline void flush() { data_length = 0; }
    inline const char *get_buffer() const { return buffer; }
    inline size_t get_length() const { return data_length; }
    inline const char *get_error_string() const { return error_description; }
  };


  class HTTPTransport;

  class HTTPResponseBodyReader
  {
  protected:
    HTTPTransport &connection;
    HTTPResponseBodyBuffer &buffer;

    const char *error_description;


  public:
    HTTPResponseBodyReader(HTTPTransport &connection,
			   HTTPResponseBodyBuffer &buffer);
    virtual ~HTTPResponseBodyReader();

    virtual uintmax_t get_content_length() const = 0;
    virtual HTTPStatus read() = 0;

    inline const char *get_error_string() const { return error_description; }
  };


  class HTTPResponse
  {
  private:
    unsigned int major_version;
    unsigned int minor_version;
    unsigned int status_code;
    char *reason_phrase;

    HTTPHeaders headers;

    HTTPResponseBodyBuffer *buffer;
    HTTPResponseBodyReader *reader;

    string error_description;


    enum {
      STATE_INITIAL = 0,
      STATE_RECEIVING_HEADERS,
      STATE_RECEIVING_BODY,
      STATE_DONE,
      STATE_ERROR
    } state;


    HTTPStatus parse_headers(const char *buf_top, const char *tail_pos);


  public:
    HTTPResponse();
    ~HTTPResponse();

    inline unsigned int get_major_version() const { return major_version; }
    inline unsigned int get_minor_version() const { return minor_version; }
    inline unsigned int get_status_code() const { return status_code; }
    inline const char *get_reason_phrase() const { return reason_phrase; }

    inline const HTTPHeaders &get_headers() const { return headers; }
    inline const char *get_header(const char *name) const
    { return headers.get_value(name); }

    HTTPStatus receive_status(HTTPTransport &connection);
    HTTPStatus receive_headers(HTTPTransport &connection);
    HTTPStatus receive_body(HTTPTransport &connection);

    inline uintmax_t get_content_length() const
    { return reader != NULL ? reader->get_content_length() : 0; }
    inline uintmax_t get_total_length() const
    { return buffer != NULL ? buffer->get_total_length() : 0; }
    inline const char *get_buffer() const
    { return buffer != NULL ? buffer->get_buffer() : NULL; }
    inline size_t get_length() const
    { return buffer != NULL ? buffer->get_length() : 0; }
    inline void flush_buffer()
    { if (buffer != NULL) buffer->flush(); }

    inline const char *get_error_string() const
    { return error_description.c_str(); }
    void clean();
  };

}


#endif	/* __cplusplus */

#endif	/* _LIBOCHUSHA_HTTP_RESPONSE_H_ */
