/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: http_transport.cc,v 1.16 2009/01/01 07:46:43 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/http_transport.h"

#include <cassert>
#include <cerrno>
#include <cstdio>
#include <cstdlib>
#include <cstring>

#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif

#include <netdb.h>
#include <netinet/in.h>

#include <poll.h>

#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif

#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef ENABLE_OPENSSL
#include <openssl/err.h>
#include <openssl/ssl.h>
#endif

#if defined(__FreeBSD_version) && (__FreeBSD_version < 501107)
#define NEED_RESOLVER_LOCK 1
#endif

#if !defined(HAVE_GETADDRINFO)
#define NEED_RESOLVER_LOCK 1
#endif

#ifdef NEED_RESOLVER_LOCK
# ifdef HAVE_PTHREAD_H
#include <pthread.h>
# else
#undef NEED_RESOLVER_LOCK
#warning "We need lock mechanism for thread safeness!"
# endif
#endif


#define IO_BUF_LENGTH_INITIAL 8192
#define IO_BUF_LENGTH_MAXIMUM 262144

// #define ENABLE_DEBUG_OUTPUT 1

namespace Ochusha
{

#ifdef ENABLE_OPENSSL
class SSLInitializer
{
public:
  SSLInitializer()
  {
    SSL_library_init();
    SSL_load_error_strings();
  }
};


static SSLInitializer ssl_initializer;
#endif


class LockPrimitive
{
#ifdef NEED_RESOLVER_LOCK
private:
  pthread_mutex_t mutex;
#endif

public:
  LockPrimitive()
  {
#ifdef NEED_RESOLVER_LOCK
#ifdef NDEBUG
    pthread_mutex_init(&mutex, NULL);
#else
    int result = pthread_mutex_init(&mutex, NULL);
    assert(result == 0);
#endif
#endif
  }

  ~LockPrimitive()
  {
#ifdef NEED_RESOLVER_LOCK
    pthread_mutex_destroy(&mutex);
#endif
  }

  inline void lock()
  {
#ifdef NEED_RESOLVER_LOCK
#ifdef NDEBUG
    pthread_mutex_lock(&mutex);
#else
    int result = pthread_mutex_lock(&mutex);
    assert(result == 0);
#endif
#endif
  }

  inline void unlock()
  {
#ifdef NEED_RESOLVER_LOCK
#ifdef NDEBUG
    pthread_mutex_unlock(&mutex);
#else
    int result = pthread_mutex_unlock(&mutex);
    assert(result == 0);
#endif
#endif
  }
};


class ResolverLock
{
private:
  static LockPrimitive lock;

public:
  inline ResolverLock() { lock.lock(); }
  inline ~ResolverLock() { lock.unlock(); }
};


LockPrimitive ResolverLock::lock;


class Addrinfo
{
private:
#ifdef HAVE_GETADDRINFO
  struct addrinfo *addrinfo;
#else
  struct hostent *hostinfo;
  struct sockaddr_in saddr;
#endif
  int last_errno;
#ifndef HAVE_GAI_STRERROR
  ErrorMessage error_buf;
#endif


public:
  Addrinfo(const char *host, unsigned short port);
  ~Addrinfo();

  struct sockaddr *get_addr();
  size_t get_addrlen();
  const char *get_error_string();
};


#ifdef HAVE_GETADDRINFO
Addrinfo::Addrinfo(const char *host, unsigned short port)
  : addrinfo(NULL),
    last_errno(0)
{
  char serv[16];
  struct addrinfo hints;

  snprintf(serv, 16, "%u", port);
  memset(&hints, 0, sizeof(hints));
  hints.ai_family = AF_INET;
  hints.ai_socktype = SOCK_STREAM;

  ResolverLock lock;
  int result = getaddrinfo(host, serv, &hints, &addrinfo);
  if (result == 0)
    return;

  last_errno = result;
  if (addrinfo != NULL)
    {
      freeaddrinfo(addrinfo);
      addrinfo = NULL;
    }
}


Addrinfo::~Addrinfo()
{
  if (addrinfo != NULL)
    freeaddrinfo(addrinfo);
}


inline struct sockaddr *
Addrinfo::get_addr()
{
  return addrinfo != NULL ? addrinfo->ai_addr : NULL;
}


inline size_t
Addrinfo::get_addrlen()
{
  return addrinfo != NULL ? addrinfo->ai_addrlen : 0;
}


const char *
Addrinfo::get_error_string()
{
#ifdef HAVE_GAI_STRERROR
  return gai_strerror(last_errno);
#else
  switch (last_errno)
    {
#ifdef EAI_ADDRFAMILY
    case EAI_ADDRFAMILY:
      return LIB_("Network addresses in the requested address family not available on specified network host");
#endif


#ifdef EAI_AGAIN
    case EAI_AGAIN:
      return LIB_("Temporary failure in name resolution");
#endif


#ifdef EAI_BADFLAGS
    case EAI_BADFLAGS:
      return LIB_("Invalid value for ai_flags");
#endif


#ifdef EAI_BADHINTS
    case EAI_BADHINTS:
      return LIB_("Invalid value for hints");
#endif


#ifdef EAI_FAIL
    case EAI_FAIL:
      return LIB_("Non-recoverable failure in name resolution");
#endif


#ifdef EAI_FAMILY
    case EAI_FAMILY:
      return LIB_("Requested ai_family not supported");
#endif


#ifdef EAI_MEMORY
    case EAI_MEMORY:
      return LIB_("Out of memory");
#endif


#ifdef EAI_NODATA
    case EAI_NODATA:
      return LIB_("Requested name not have IP address");
#endif


#ifdef EAI_NONAME
    case EAI_NONAME:
      return LIB_("Requested hostname or servname not provided, or not known");
#endif


#ifdef EAI_OVERFLOW
    case EAI_OVERFLOW:
      return LIB_("Argument buffer overflow");
#endif


#ifdef EAI_PROTOCOL
    case EAI_PROTOCOL:
      return LIB_("Resolved protocol is unknown");
#endif


#ifdef EAI_SERVICE
    case EAI_SERVICE:
      return LIB_("Requested servname not supported for ai_socktype");
#endif


#ifdef EAI_SOCKTYPE
    case EAI_SOCKTYPE:
      return LIB_("Requested ai_socktype not supported");
#endif


#ifdef EAI_SYSTEM
    case EAI_SYSTEM:
      error_buf.set_errno(last_errno);
      return error_buf.get_message();
#endif
    }


  error_buf.set_unknown_errno(last_errno);
  return error_buf.get_message();
#endif
}


#else
Addrinfo::Addrinfo(const char *host, unsigned short port)
 : last_errno(0)
{
  memset(saddr, 0, sizeof(struct sockaddr_in));

  ResolverLock lock;
  hostinfo = gethostbyname(host);
  if (hostinfo != NULL)
    {
      memcpy(&saddr.sin_addr.s_addr, hostinfo->h_addr_list[0], sizeof(char *));
      saddr.sin_family = AF_INET;
      saddr.sin_port = htons(port);
    }
  else
    {
      last_errno = h_errno;
    }
}


Addrinfo::~Addrinfo()
{
}


inline struct sockaddr *
Addrinfo::get_addr()
{
  return hostinfo != NULL ? &saddr : NULL;
}


inline size_t
Addrinfo::get_addrlen()
{
  return hostinfo != NULL ? sizeof(struct sockaddr_in) : 0;
}


const char *
Addrinfo::get_error_string()
{
#ifdef HAVE_HSTRERROR
  return hstrerror(last_errno);
#else
  switch (last_errno)
    {
#ifdef HOST_NOT_FOUND
    case HOST_NOT_FOUND:
      return LIB_("Host not found");
#endif


#ifdef TRY_AGAIN
    case TRY_AGAIN:
      return LIB_("Try again");
#endif


#ifdef NO_RECOVERY
    case NO_RECOVERY:
      return LIB_("Non-recoverable error");
#endif


#ifdef NO_DATA
    case NO_DATA:
      return LIB_("No data available for requested name");
#endif


#ifdef NO_ADDRESS
    case NO_ADDRESS:
      return LIB_("Requested name not have IP address");
#endif
    }
#endif

  error_buf.set_unknown_errno(last_errno);
  return error_buf.get_message();
}
#endif


HTTPTransportConnection::HTTPTransportConnection(const char *host,
						 unsigned short port)
  : addrinfo(new Addrinfo(host, port)),
    state(STATE_INITIAL),
    fd(-1),
    poll_event_mask(0),
    error_type(ERROR_TYPE_UNINITIALIZED),
    last_errno(0),
    asynchronous(false),
    nonblock(false)
{
}


HTTPTransportConnection::~HTTPTransportConnection()
{
  if (addrinfo != NULL)
    delete addrinfo;
}


bool
HTTPTransportConnection::load_verify_locations(const char *file,
					       const char *path)
{
  return true;
}


static void
update_flags(int fd, bool async, bool nonblock)
{
  int flags = fcntl(fd, F_GETFL);
  if (async && nonblock)
    flags |= O_NONBLOCK;
  else
    flags &= ~O_NONBLOCK;

  fcntl(fd, F_SETFL, flags);
}


void
HTTPTransportConnection::set_asynchronous(bool value)
{
  if (asynchronous == value)
    return;

  asynchronous = value;
  update_flags(fd, asynchronous, nonblock);
}


void
HTTPTransportConnection::set_nonblock(bool value)
{
  if (nonblock == value)
    return;

  nonblock = value;
  update_flags(fd, asynchronous, nonblock);
}


HTTPStatus
HTTPTransportConnection::connect()
{
  if (addrinfo == NULL)
    {
      set_error_type(ERROR_TYPE_ERRNO);
      set_last_errno(ENOMEM);
      return HTTP_ERROR;
    }

  if (addrinfo->get_addr() == NULL)
    {
      set_error_type(ERROR_TYPE_HOST);
      return HTTP_ERROR;
    }

  if (state == STATE_INITIAL)
    {
      if ((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
	  set_error_type(ERROR_TYPE_ERRNO);
	  set_last_errno(errno);
	  return HTTP_ERROR;
	}

      update_flags(fd, asynchronous, nonblock);
      state = STATE_CONNECT;
    }

  if (state == STATE_CONNECT)
    {
      assert(addrinfo->get_addrlen() > 0);

      int result = ::connect(fd, addrinfo->get_addr(), addrinfo->get_addrlen());
      if (result < 0)
	{
	  if (errno == EINPROGRESS)
	    {
	      set_poll_event_mask(POLLOUT);
	      state = STATE_INPROGRESS;
	      set_last_errno(errno);
	      return HTTP_NOT_DONE;
	    }

	  state = STATE_ERROR;
	  set_error_type(ERROR_TYPE_ERRNO);
	  set_last_errno(errno);
	  return HTTP_ERROR;
	}
    }

  if (state == STATE_INPROGRESS)
    {
      int value = 0;
      socklen_t value_len = sizeof(int);
      if (getsockopt(fd, SOL_SOCKET, SO_ERROR, &value, &value_len) == 0)
	{
	  if (value == 0)
	    {
	      state = STATE_ESTABLISHED;
	      set_last_errno(0);
	      return HTTP_DONE;
	    }

	  set_last_errno(value);
	}
      else
	set_last_errno(errno);

      state = STATE_ERROR;
      set_error_type(ERROR_TYPE_ERRNO);
      return HTTP_ERROR;
    }

  state = STATE_ESTABLISHED;
  return HTTP_DONE;
}


ssize_t
HTTPTransportConnection::read(void *buf, size_t nbytes)
{
  ssize_t result = ::read(fd, buf, nbytes);
  if (result >= 0)
    {
      set_poll_event_mask(0);
      return result;
    }

  set_last_errno(errno);
  if (get_last_errno() == EAGAIN)
    set_poll_event_mask(POLLIN);
  else
    set_error_type(ERROR_TYPE_ERRNO);

  return result;
}


ssize_t
HTTPTransportConnection::write(void *buf, size_t nbytes)
{
  ssize_t result = ::write(fd, buf, nbytes);
  if (static_cast<size_t>(result) == nbytes || result == 0)
    {
      set_poll_event_mask(0);
      return result;
    }

  if (result > 0)
    {
      set_poll_event_mask(POLLOUT);
      return result;
    }

  set_last_errno(errno);
  if (get_last_errno() == EAGAIN)
    set_poll_event_mask(POLLOUT);
  else
    set_error_type(ERROR_TYPE_ERRNO);

  return result;
}


int
HTTPTransportConnection::close()
{
  if (fd == -1)
    return 0;

  int old_fd = fd;
  fd = -1;
  return ::close(old_fd);
}


const char *
HTTPTransportConnection::get_error_string()
{
  if (addrinfo == NULL)
    return LIB_("Out of memory");

  if (addrinfo->get_addr() == NULL)
    return addrinfo->get_error_string();

  if (error_type == ERROR_TYPE_ERRNO)
    {
      error_buf.set_errno(get_last_errno());
      return error_buf.get_message();
    }

  if (error_type == ERROR_TYPE_UNINITIALIZED)
    return LIB_("Error type is not initialized, not an error?");

  return LIB_("Unknown error");
}


#ifdef ENABLE_OPENSSL
class HTTPTransportConnectionSSL : public HTTPTransportConnection
{
private:
  SSL_CTX *ssl_ctx;
  SSL *ssl;
  bool ssl_no_verify;
  long ssl_verify_result;


public:
  HTTPTransportConnectionSSL(const char *host, unsigned short port);
  virtual ~HTTPTransportConnectionSSL();
  virtual bool load_verify_locations(const char *file, const char *path);
  virtual HTTPStatus connect();
  virtual ssize_t read(void *buf, size_t nbytes);
  virtual ssize_t write(void *buf, size_t nbytes);
  virtual int close();
  virtual const char *get_error_string();
};


HTTPTransportConnectionSSL::HTTPTransportConnectionSSL(const char *host,
						       unsigned short port)
  : HTTPTransportConnection(host, port),
    ssl_ctx(SSL_CTX_new(SSLv23_client_method())),
    ssl(NULL),
    ssl_no_verify(true),
    ssl_verify_result(X509_V_OK)
{
}


HTTPTransportConnectionSSL::~HTTPTransportConnectionSSL()
{
  if (ssl != NULL)
    {
      SSL_shutdown(ssl);
      SSL_free(ssl);
    }

  if (ssl_ctx != NULL)
    SSL_CTX_free(ssl_ctx);
}


bool
HTTPTransportConnectionSSL::load_verify_locations(const char *file,
						  const char *path)
{
  if (ssl_ctx != NULL)
    {
      SSL_CTX_load_verify_locations(ssl_ctx, file, path);
      return true;
    }

  return false;
}


HTTPStatus
HTTPTransportConnectionSSL::connect()
{
  assert(ssl_ctx != NULL);

  if (get_state() != STATE_ESTABLISHED)
    {
      HTTPStatus result = HTTPTransportConnection::connect();
      if (result != HTTP_DONE)
	return result;
    }

  if (get_state() == STATE_ESTABLISHED)
    {
      assert(ssl == NULL);

      ssl = SSL_new(ssl_ctx);
      if (ssl == NULL)
	{
	  set_state(STATE_ERROR);
	  set_error_type(ERROR_TYPE_ERRNO);
	  set_last_errno(errno);
	  return HTTP_ERROR;
	}

      if (!SSL_set_fd(ssl, get_socket()))
	{
	  set_state(STATE_ERROR);
	  set_error_type(ERROR_TYPE_ERRNO);
	  set_last_errno(errno);
	  return HTTP_ERROR;
	}

      int err;
      if (!(err = SSL_connect(ssl)))
	{
	  set_state(STATE_ERROR);
	  int ssl_err = SSL_get_error(ssl, err);
	  if (ssl_err == SSL_ERROR_SYSCALL)
	    {
	      set_error_type(ERROR_TYPE_ERRNO);
	      set_last_errno(errno);
	    }
	  else
	    {
	      set_error_type(ERROR_TYPE_SSL);
	      set_last_errno(ssl_err);
	    }

	  return HTTP_ERROR;
	}

      if (!ssl_no_verify)
	{
	  ssl_verify_result = SSL_get_verify_result(ssl);
	  if (ssl_verify_result != X509_V_OK)
	    {
	      set_state(STATE_ERROR);
	      set_error_type(ERROR_TYPE_SSL_VERIFICATION);
	      return HTTP_ERROR;
	    }
	}

      return HTTP_DONE;
    }

  set_state(STATE_ERROR);
  set_error_type(ERROR_TYPE_ERRNO);
  set_last_errno(errno);
  return HTTP_ERROR;
}


ssize_t
HTTPTransportConnectionSSL::read(void *buf, size_t nbytes)
{
  assert(ssl != NULL);
  int result = SSL_read(ssl, buf, nbytes);
  if (result >= 0)
    {
      set_poll_event_mask(0);
      return result;
    }

  int ssl_err = SSL_get_error(ssl, result);
  if (ssl_err == SSL_ERROR_WANT_READ || ssl_err == SSL_ERROR_WANT_WRITE)
    {
      if (ssl_err == SSL_ERROR_WANT_READ)
	set_poll_event_mask(POLLIN);
      else
	set_poll_event_mask(POLLOUT);

      set_last_errno(EAGAIN);
      set_error_type(ERROR_TYPE_ERRNO);

      return result;
    }

  int new_errno;
  if (ssl_err == SSL_ERROR_SYSCALL)
    {
      new_errno = errno;
      set_error_type(ERROR_TYPE_ERRNO);
    }
  else
    {
      new_errno = ssl_err;
      set_error_type(ERROR_TYPE_SSL);
    }

  set_poll_event_mask(0);
  set_last_errno(new_errno);

  return result;
}


ssize_t
HTTPTransportConnectionSSL::write(void *buf, size_t nbytes)
{
  assert(ssl != NULL);

  int result = SSL_write(ssl, buf, nbytes);
  if (static_cast<size_t>(result) == nbytes || result == 0)
    {
      set_poll_event_mask(0);
      return result;
    }

  if (result > 0)
    {
      set_poll_event_mask(POLLOUT);
      return result;
    }

  int ssl_err = SSL_get_error(ssl, result);
  if (ssl_err == SSL_ERROR_WANT_READ || ssl_err == SSL_ERROR_WANT_WRITE)
    {
      if (ssl_err == SSL_ERROR_WANT_READ)
	set_poll_event_mask(POLLIN);
      else
	set_poll_event_mask(POLLOUT);

      set_last_errno(EAGAIN);
      set_error_type(ERROR_TYPE_ERRNO);
      return result;
    }

  int new_errno;
  if (ssl_err == SSL_ERROR_SYSCALL)
    {
      new_errno = errno;
      set_error_type(ERROR_TYPE_ERRNO);
    }
  else
    {
      new_errno = ssl_err;
      set_error_type(ERROR_TYPE_SSL);
    }

  set_poll_event_mask(0);
  set_last_errno(new_errno);
  return result;
}


int
HTTPTransportConnectionSSL::close()
{
  if (ssl != NULL)
    {
      SSL_shutdown(ssl);
      SSL_free(ssl);
      ssl = NULL;
    }

  if (ssl_ctx != NULL)
    {
      SSL_CTX_free(ssl_ctx);
      ssl_ctx = NULL;
    }

  return HTTPTransportConnection::close();
}


const char *
HTTPTransportConnectionSSL::get_error_string()
{
  if (get_error_type() == ERROR_TYPE_SSL)
    {
      char *buf = error_buf.get_buf();
      ERR_error_string_n(get_last_errno(), buf, ERROR_MESSAGE_BUFLEN);
      return buf;
    }

  if (get_error_type() == ERROR_TYPE_SSL_VERIFICATION)
    {
      static const char * const x509_error_strings[] = {
	LIB_("OK"),
	LIB_("Unknown SSL verification error"),
	LIB_("Unable to get issuer certificate"),
	LIB_("Unable to get certificate CRL"),
	LIB_("Unable to decrypt certificate's signature"),

	LIB_("Unable to decrypt CRL's signature"),
	LIB_("Unable to decode issuer public key"),
	LIB_("Certificate signature failure"),
	LIB_("CRL signature failure"),
	LIB_("Certificate is not yet valid"),

	LIB_("Certificate has expired"),
	LIB_("CRL is not yet valid"),
	LIB_("CRL has expired"),
	LIB_("Format error in certificate's notBefore field"),
	LIB_("Format error in certificate's notAfter field"),

	LIB_("Format error in CRL's lastUpdate field"),
	LIB_("Format error in CRL's nextUpdate field"),
	LIB_("Out of memory"),
	LIB_("Self signed certificate"),
	LIB_("Self signed certificate in certificate chain"),

	LIB_("Unable to get local issuer certificate"),
	LIB_("Unable to verify the first certificate"),
	LIB_("Certificate chain too long"),
	LIB_("Certificate revoked"),
	LIB_("Invalid CA certificate"),

	LIB_("Path length constraint exceeded"),
	LIB_("Unsupported certificate purpose"),
	LIB_("Certificate not trusted"),
	LIB_("Certificate rejected"),
	LIB_("Subject issuer mismatch"),

	LIB_("Authority and subject key identifier mismatch"),
	LIB_("Authority and issuer serial number mismatch"),
	LIB_("Key usage does not include certificate signing")
      };

      if (ssl_verify_result >= X509_V_OK
	  && ssl_verify_result <= X509_V_ERR_KEYUSAGE_NO_CERTSIGN)
	return x509_error_strings[ssl_verify_result];

      return LIB_("Unknown SSL verification error");
    }

  return HTTPTransportConnection::get_error_string();
}
#endif


HTTPTransport::HTTPTransport(const URI *uri, const URI *proxy_uri)
  : uri(uri),
    proxy_uri(uri),
    ca_file(NULL),
    ca_path(NULL),
    connection(NULL),
    io_buf_length(IO_BUF_LENGTH_INITIAL),
    io_buf(static_cast<char *>(calloc(io_buf_length, sizeof(char)))),
    io_buf_alloc(0),
    io_buf_io_done(0),
    last_io_result(0),
    last_errno(0),
    feeling(FEELING_GOOD),
    asynchronous(false),
    nonblock(false)
{
  assert(io_buf != NULL);
}


HTTPTransport::~HTTPTransport()
{
  if (ca_file != NULL)
    free(ca_file);

  if (ca_path != NULL)
    free(ca_path);

  if (connection != NULL)
    delete connection;

  if (io_buf != NULL)
    free(io_buf);
}


void
HTTPTransport::set_uri(const URI *uri)
{
  if (this->uri == uri)
    return;

  if (connection != NULL)
    {
      delete connection;
      connection = NULL;
    }

  this->uri = uri;
}


void
HTTPTransport::set_proxy_uri(const URI *uri)
{
  if (this->proxy_uri == uri)
    return;

  if (connection != NULL)
    {
      delete connection;
      connection = NULL;
    }

  this->proxy_uri = uri;
}


void
HTTPTransport::set_verify_locations(const char *file, const char *path)
{
  if (ca_file != NULL)
    {
      free(ca_file);
      ca_file = NULL;
    }

  if (ca_path != NULL)
    {
      free(ca_path);
      ca_path = NULL;
    }

  if (file != NULL)
    ca_file = strdup(file);

  if (path != NULL)
    ca_path = strdup(path);

  if (connection != NULL)
    connection->load_verify_locations(file, path);
}


void
HTTPTransport::reset_buffer()
{
  io_buf_alloc = 0;
  io_buf_io_done = 0;
}


void
HTTPTransport::clip_buffer(const char *clip_to)
{
  assert(clip_to >= io_buf && clip_to <= io_buf + io_buf_alloc);

  io_buf_alloc -= clip_to - io_buf;
  if (io_buf_alloc > 0)
    memmove(io_buf, clip_to, io_buf_alloc);

  io_buf_io_done = 0;
}


const char *
HTTPTransport::find_pattern(const char *pattern, size_t pattern_length)
{
  const char *tail = io_buf + io_buf_alloc - pattern_length;
  for (const char *cur_pos = io_buf; cur_pos <= tail; ++cur_pos)
    if (*cur_pos == *pattern && memcmp(cur_pos, pattern, pattern_length) == 0)
      return cur_pos;

  return NULL;
}


const char *
HTTPTransport::find_pattern_from(const char *from,
				 const char *pattern, size_t pattern_length)
{
  if (from < io_buf)
    return NULL;

  const char *tail = io_buf + io_buf_alloc - pattern_length;
  for (const char *cur_pos = from; cur_pos <= tail; ++cur_pos)
    if (*cur_pos == *pattern && memcmp(cur_pos, pattern, pattern_length) == 0)
      return cur_pos;

  return NULL;
}


void
HTTPTransport::set_asynchronous(bool value)
{
  if (asynchronous == value)
    return;

  asynchronous = value;
  if (connection != NULL)
    connection->set_asynchronous(value);
}


void
HTTPTransport::set_nonblock(bool value)
{
  if (nonblock == value)
    return;

  nonblock = value;
  if (connection != NULL)
    connection->set_nonblock(value);
}


HTTPStatus
HTTPTransport::connect()
{
  if (connection != NULL)
    return connection->connect();

#ifndef NDEBUG
#ifdef ENABLE_DEBUG_OUTPUT
  if (uri == NULL)
    fprintf(stderr, "HTTPTransport::connect(): uri == NULL\n");
#endif

  assert(uri != NULL);

#else
  if (uri == NULL)
    return HTTP_ERROR;
#endif

  const URI *peer = proxy_uri == NULL ? uri : proxy_uri;

  assert(peer->get_scheme() != NULL);
#ifdef ENABLE_OPENSSL
  assert(strcmp(peer->get_scheme(), "http") == 0
	 || strcmp(peer->get_scheme(), "https") == 0);
#else
  assert(strcmp(peer->get_scheme(), "http") == 0);
#endif
  assert(peer->get_server() != NULL);

  unsigned short port;
  if (strcmp(peer->get_scheme(), "http") == 0)
    {
      port = peer->get_port();
      if (port == 0)
	port = 80;	// standard port number for HTTP.

      connection = new HTTPTransportConnection(peer->get_server(), port);
    }
  else
    {
#ifdef ENABLE_OPENSSL
      // https
      port = peer->get_port();
      if (port == 0)
	port = 443;	// standard port number for HTTPS.

      connection = new HTTPTransportConnectionSSL(peer->get_server(), port);
#else
      return HTTP_ERROR;
#endif
    }

  if (ca_file != NULL || ca_path != NULL)
    connection->load_verify_locations(ca_file, ca_path);

  connection->set_asynchronous(asynchronous);
  connection->set_nonblock(nonblock);

  return connection->connect();
}


bool
HTTPTransport::close()
{
  if (connection != NULL)
    return connection->close() == 0;

  return false;
}


void
HTTPTransport::clean()
{
  if (connection != NULL)
    {
      delete connection;
      connection = NULL;
    }

  uri = NULL;
  proxy_uri = NULL;
  if (ca_file != NULL)
    {
      free(ca_file);
      ca_file = NULL;
    }

  if (ca_path != NULL)
    {
      free(ca_path);
      ca_path = NULL;
    }

  io_buf_length = IO_BUF_LENGTH_INITIAL;
  char *old_buf = io_buf;
  io_buf = static_cast<char *>(realloc(io_buf, io_buf_length));
  if (io_buf == NULL)
    io_buf = old_buf;

  io_buf_alloc = 0;
  io_buf_io_done = 0;
  last_io_result = 0;
  asynchronous = false;
  nonblock = false;
}


bool
HTTPTransport::ensure_enough_space(size_t length)
{
  if (calculate_space_available() < length)
    {
      size_t old_buf_length = io_buf_length;
      do
	{
	  io_buf_length *= 2;
	}
      while (calculate_space_available() < length);

      char *old_buf = io_buf;
      io_buf = static_cast<char *>(realloc(io_buf, io_buf_length));
      if (io_buf == NULL)
	{
	  io_buf = old_buf;
	  io_buf_length = old_buf_length;
	  last_errno = ENOMEM;
	  return false;		// out of memory
	}
    }

  return true;
}


bool
HTTPTransport::append_data_to_buffer(const char *data, size_t length)
{
  if (ensure_enough_space(length))
    {
      memcpy(io_buf + io_buf_alloc, data, length);
      io_buf_alloc += length;
      return true;
    }

  return false;		// out of memory
}


HTTPStatus
HTTPTransport::write_buffer(size_t length)
{
  size_t io_left = io_buf_alloc - io_buf_io_done;
  if (io_left == 0)
    {
      last_io_result = 0;
      last_errno = 0;
      return HTTP_DONE;	// buffer is empty
    }

  size_t bytes_to_write = (length > 0 && length < io_left) ? length : io_left;

  assert(bytes_to_write + io_buf_io_done <= io_buf_alloc);
  ssize_t result = last_io_result
    = connection->write(io_buf + io_buf_io_done, bytes_to_write);

  if (result < 0)
    {
      last_errno = connection->get_last_errno();
      if (last_errno == EAGAIN)
	return HTTP_NOT_DONE;

      if (last_errno == EINTR)
	result = 0;
      else
	{
#ifdef ENABLE_DEBUG_OUTPUT
	  fprintf(stderr, "HTTPTransport::write_buffer(): last_errno=%d\n",
		  last_errno);
#endif
	  return HTTP_ERROR;
	}
    }
  else if (result == 0)
    return HTTP_DONE;

  io_buf_io_done += result;
  return bytes_to_write == static_cast<size_t>(result)
    ? HTTP_DONE : HTTP_NOT_DONE;
}


HTTPStatus
HTTPTransport::read_into_buffer(size_t length)
{
#ifdef ENABLE_DEBUG_OUTPUT
  fprintf(stderr, "HTTPTransport::read_into_buffer(%u)\n", length);
#endif
  size_t io_left = io_buf_length - io_buf_alloc;
  if (io_left == 0)
    {
      size_t new_len = io_buf_length * 2;
      new_len
	= (new_len < IO_BUF_LENGTH_MAXIMUM) ? new_len : IO_BUF_LENGTH_MAXIMUM;
      if (ensure_enough_space(new_len - io_buf_length))
	io_left = io_buf_length - io_buf_alloc;
      else
	{
	  // buffer overflow
	  last_errno = ENOMEM;	// should treat this in more appropriate way.
	  feeling = FEELING_FULL;
	  return HTTP_ERROR;
	}
    }

  size_t bytes_to_read = (length > 0 && length < io_left) ? length : io_left;

  assert(bytes_to_read + io_buf_alloc <= io_buf_length);
  ssize_t result = last_io_result
    = connection->read(io_buf + io_buf_alloc, bytes_to_read);

#if 0
  fprintf(stderr,
	  "HTTPTransport::read_into_buffer(%lu): result=%ld bytes_to_read=%lu\n",
	  static_cast<unsigned long>(length),
	  static_cast<long>(result), static_cast<unsigned long>(bytes_to_read));
#endif

  if (result < 0)
    {
      last_errno = connection->get_last_errno();
      feeling = FEELING_HUNGRY;

#ifdef ENABLE_DEBUG_OUTPUT
      fprintf(stderr, "HTTPTransport::read_into_buffer(): last_errno=%d\n",
	      last_errno);
#endif
      if (last_errno == EAGAIN || last_errno == EINTR)
	return HTTP_NOT_DONE;

      return HTTP_ERROR;
    }

  if (result == 0)
    {
      feeling = FEELING_GOOD;
      return HTTP_DONE;	// EOF
    }

  io_buf_alloc += result;

  if ((length == 0 || length > io_left)
      && io_buf_alloc == io_buf_length)
    {
      if (io_buf_length < IO_BUF_LENGTH_MAXIMUM)
	{
	  // enlarge io_buf for the next read().
	  size_t new_len = io_buf_length * 2;
	  new_len = (new_len < IO_BUF_LENGTH_MAXIMUM)
	    ? new_len : IO_BUF_LENGTH_MAXIMUM;
	  feeling = ensure_enough_space(new_len - io_buf_length)
	    ? FEELING_GOOD : FEELING_FULL; // we can ignore the failure here.
	}
      else
	feeling = FEELING_FULL;
    }
  else
    feeling = (length == io_buf_alloc) ? FEELING_GOOD : FEELING_HUNGRY;

  return bytes_to_read == static_cast<size_t>(result)
    ? HTTP_DONE : HTTP_NOT_DONE;
}


const char *
HTTPTransport::get_error_string()
{
  if (last_errno == ENOMEM)
    return LIB_("Out of memory");

  if (connection != NULL)
    return connection->get_error_string();

  return LIB_("Unknown error");
}

}
