/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: http_transport.h,v 1.9 2009/01/01 07:46:43 fuyu Exp $
 */

#ifndef _LIBOCHUSHA_HTTP_TRANSPORT_H_
#define _LIBOCHUSHA_HTTP_TRANSPORT_H_

#ifdef __cplusplus

#if defined(__GNUG__) && !defined(__ICC)
#pragma interface
#endif

#include "libochusha/http_enums.h"
#include "libochusha/misc_utils.h"
#include "libochusha/uri_utils.h"

#include <cstdlib>
#include <cstring>


using namespace std;
namespace Ochusha
{

#define HTTP_ERROR_BUFLEN		(256)

  class Addrinfo;
  class HTTPTransport;

  class HTTPTransportConnection
  {
    friend class HTTPTransport;

  protected:
    enum HTTPTransportState {
      STATE_INITIAL = 0,
      STATE_CONNECT,
      STATE_INPROGRESS,
      STATE_ESTABLISHED,
      STATE_ERROR
    };


    enum HTTPTransportErrorType {
      ERROR_TYPE_UNINITIALIZED = 0,
      ERROR_TYPE_HOST,
      ERROR_TYPE_ERRNO,
      ERROR_TYPE_SSL,
      ERROR_TYPE_SSL_VERIFICATION
    };


  private:
    Addrinfo *addrinfo;
    HTTPTransportState state;
    int fd;
    int poll_event_mask;
    HTTPTransportErrorType error_type;
    int last_errno;
    bool asynchronous;
    bool nonblock;


  protected:
    ErrorMessage error_buf;

    inline void set_poll_event_mask(int value) { poll_event_mask = value; }
    inline HTTPTransportState get_state() const { return state; }
    inline void set_state(HTTPTransportState new_state) { state = new_state; }
    inline void set_error_type(HTTPTransportErrorType new_error_type)
    { error_type = new_error_type; }
    inline void set_last_errno(int new_errno) { last_errno = new_errno; }


  public:
    HTTPTransportConnection(const char *host, unsigned short port);
    virtual ~HTTPTransportConnection();
    virtual bool load_verify_locations(const char *file, const char *path);
    virtual void set_asynchronous(bool value);
    virtual void set_nonblock(bool value);
    virtual HTTPStatus connect();
    virtual ssize_t read(void *buf, size_t nbytes);
    virtual ssize_t write(void *buf, size_t nbytes);
    virtual int close();

    inline int get_socket() const { return fd; }
    inline int get_poll_event_mask() const { return poll_event_mask; }
    inline HTTPTransportErrorType get_error_type() const { return error_type; }
    inline int get_last_errno() const { return last_errno; }
    virtual const char *get_error_string();
  };


  class HTTPTransport
  {
  public:
    enum ReadFeeling {
      FEELING_GOOD,
      FEELING_HUNGRY,
      FEELING_FULL
    };


  private:
    const URI *uri;
    const URI *proxy_uri;
    char *ca_file;
    char *ca_path;

    HTTPTransportConnection *connection;

    size_t io_buf_length;
    char *io_buf;
    size_t io_buf_alloc;
    size_t io_buf_io_done;

    ssize_t last_io_result;
    int last_errno;

    ReadFeeling feeling;

    bool asynchronous;
    bool nonblock;


    bool ensure_enough_space(size_t length);
    inline size_t calculate_space_available() const
    { return io_buf_length - io_buf_alloc; }


  public:
    HTTPTransport(const URI *uri = NULL, const URI *proxy_uri = NULL);
    ~HTTPTransport();

    void set_uri(const URI *uri);
    void set_proxy_uri(const URI *uri);
    void set_verify_locations(const char *file, const char *path);

    void reset_buffer();
    void clip_buffer(const char *clip_to);
    const char *find_pattern(const char *pattern, size_t pattern_length);
    inline const char *find_pattern(const char *pattern)
    { return find_pattern(pattern, strlen(pattern)); }
    const char *find_pattern_from(const char *from,
				  const char *pattern, size_t pattern_length);
    inline const char *find_pattern_from(const char *from, const char *pattern)
    { return find_pattern_from(from, pattern, strlen(pattern)); }

    void set_asynchronous(bool value);
    void set_nonblock(bool value);
    HTTPStatus connect();
    inline int get_socket() const
    { return connection != NULL ? connection->get_socket() : -1; }
    inline int get_poll_event_mask() const
    { return connection != NULL ? connection->get_poll_event_mask() : 0; }
    inline ReadFeeling get_feeling() const { return feeling; }
    bool close();
    void clean();

    bool append_data_to_buffer(const char *data, size_t length);
    inline bool append_data_to_buffer(const char *data)
    { return append_data_to_buffer(data, strlen(data)); }

    /*
     * Send buffered data.
     *
     * When 0 `length' is given, this try to send all buffered data,
     * otherwise, only MIN(length, buffered-data-length) data will be sended.
     *
     * This returns:
     *   HTTP_DONE - if all or `length' buffered data had sended.
     *   HTTP_NOT_DONE - if there's data still in buffer.
     *   HTTP_ERROR - something wrong occured.
     */
    HTTPStatus write_buffer(size_t length = 0);

    /*
     * Receive data into buffer.
     *
     * When 0 `length' is given, this try to receive data that fully fill
     * internal incoming buffer, otherwise,
     * only MIN(length, internal-buffer-size) data will be received.
     *
     * This returns:
     *   HTTP_DONE - if internal incoming buffer is fully filled,
     *               non 0 `length' data is available in the incoming buffer,
     *               or EOF received.
     *   HTTP_ERROR - something wrong occured.
     *   HTTP_NOT_DONE - otherwise.
     */
    HTTPStatus read_into_buffer(size_t length = 0);

    inline const char *get_incoming_buffer() const { return io_buf; }
    inline size_t get_incoming_data_length() const { return io_buf_alloc; }
    inline ssize_t get_last_io_result() const { return last_io_result; }
    inline int get_last_errno() const { return last_errno; }
    const char *get_error_string();
  };

}


#endif	/* __cplusplus */

#endif	/* _LIBOCHUSHA_HTTP_TRANSPORT_H_ */
