/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: lock.cc,v 1.26 2009/01/01 07:46:43 fuyu Exp $
 */

#include "config.h"

#define _LIBOCHUSHA_LOCK_CC_	1

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/lock.h"

#include "libochusha/misc_utils.h"

#include <cassert>
#include <cerrno>
#include <cstddef>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <iostream>

#include <pthread.h>

#if defined(HAVE_CLOCK_GETTIME) || defined(HAVE_GETTIMEOFDAY)
# ifdef HAVE_SYS_TIME_H
#  include <sys/time.h>
# endif
#endif


namespace Ochusha
{

// #define DUMP_THREAD_INFO_FOR_DYING_THREAD	1
// #define VERBOSE_DESTRUCTOR	1


Monitor libochusha_giant("libochusha *GIANT*");


#ifdef CHECK_LOCK_ORDER

static pthread_once_t lock_set_giant_once = PTHREAD_ONCE_INIT;
static pthread_mutex_t lock_set_giant;


static void
lock_set_giant_init()
{
  if (pthread_mutex_init(&lock_set_giant, NULL) != 0)
    {
      cerr << "Couldn't initialize a mutex: lock_set_giant\n";
      abort();
    }
}


pthread_once_t ThreadInfo::thread_info_key_once = PTHREAD_ONCE_INIT;
pthread_key_t ThreadInfo::thread_info_key;


void
ThreadInfo::thread_info_key_destroy(void *data)
{
  ThreadInfo *info = static_cast<ThreadInfo *>(data);
  if (info != NULL)
    {
#ifdef DUMP_THREAD_INFO_FOR_DYING_THREAD
      int result = pthread_mutex_lock(&lock_set_giant);
      if (result != 0)
	{
	  ErrorMessage error_message(result);
	  fprintf(stderr, "Couldn't lock a mutex: lock_set_giant: %s\n",
		  error_message.get_message());
	  abort();
	}


      info->dump_info();


      result = pthread_mutex_unlock(&lock_set_giant);
      if (result != 0)
	{
	  ErrorMessage error_message(result);
	  fprintf(stderr, "Couldn't unlock a mutex: lock_set_giant: %s\n",
		  error_message.get_message());
	  abort();
	}
#endif
      delete info;
    }
}


void
ThreadInfo::thread_info_key_init()
{
  if (pthread_key_create(&thread_info_key, thread_info_key_destroy) != 0)
    {
      cerr << "ThreadInfo::thread_info_key_init(): pthread_key_create() failed.\n";
      abort();
    }
}


void
ThreadInfo::set_current(ThreadInfo *info)
{
  assert(info != NULL);

  pthread_once(&lock_set_giant_once, lock_set_giant_init);
  pthread_once(&thread_info_key_once, thread_info_key_init);
  if (pthread_setspecific(thread_info_key, info) != 0)
    {
      cerr << "ThreadInfo::set_current(): pthread_setspecific() failed.\n";
      abort();
    }
}


ThreadInfo *
ThreadInfo::get_current()
{
  void *data = pthread_getspecific(thread_info_key);
  if (data != NULL)
    return static_cast<ThreadInfo *>(data);

  return new ThreadInfo("<unknown thread>");
}


ThreadInfo::ThreadInfo(const char *description)
  : description(description),
    waiting_monitor(NULL),
    maximum_number_of_simultaneous_locks(0)
{
  set_current(this);
}


ThreadInfo::~ThreadInfo()
{
  assert(holding_locks.empty());
  assert(waiting_monitor == NULL);
}


void
ThreadInfo::notify_locked(Lock *lock)
{
  assert(lock != NULL);
  assert(waiting_monitor != lock);
  assert(holding_locks.find(lock) == holding_locks.end());

  holding_locks.insert(lock);
  int size = holding_locks.size();
  if (size > maximum_number_of_simultaneous_locks)
    maximum_number_of_simultaneous_locks = size;
}


void
ThreadInfo::notify_unlocked(Lock *lock)
{
  assert(lock != NULL);
  assert(waiting_monitor != lock);
  assert(holding_locks.find(lock) != holding_locks.end());

  holding_locks.erase(lock);
}


void
ThreadInfo::notify_waited(Monitor *monitor)
{
  assert(monitor != NULL);
  assert(waiting_monitor == NULL);
  assert(holding_locks.find(monitor) != holding_locks.end());

  holding_locks.erase(monitor);
  waiting_monitor = monitor;
}


void
ThreadInfo::notify_signaled(Monitor *monitor)
{
  assert(monitor != NULL);
  assert(waiting_monitor == monitor);
  assert(holding_locks.find(monitor) == holding_locks.end());

  holding_locks.insert(monitor);
  waiting_monitor = NULL;
}


const char *
ThreadInfo::get_description()
{
  return description.c_str();
}


void
ThreadInfo::dump_info_body()
{
  LockSet::const_iterator iter = holding_locks.begin();
  LockSet::const_iterator end_iter = holding_locks.end();
  string buf("  holding_locks: ");
  while (iter != end_iter)
    {
      Lock *lock = *iter++;
      lock->dump_brief_description(buf);
      buf.append(", ");
    }
  fprintf(stderr, "%s\n", buf.c_str());

  buf = "  waiting_monitor: ";
  if (waiting_monitor != NULL)
    waiting_monitor->dump_brief_description(buf);
  fprintf(stderr, "%s\n", buf.c_str());

  fprintf(stderr, "  maximum_number_of_simultaneous_locks: %d\n",
	  maximum_number_of_simultaneous_locks);
}


void
ThreadInfo::dump_info()
{
  fprintf(stderr, "ThreadInfo(\"%s\") {\n", get_description());

  dump_info_body();

  fprintf(stderr, "}\n");
}
#endif


static bool lock_manager_destructed = false;

#ifdef CHECK_LOCK_ORDER
#endif


LockManager LockManager::lock_manager;


LockManager::LockManager()
  : lock("LockManager")
{
}


RWLock *
LockManager::rwlock_new()
{
  assert(!lock_manager_destructed);

  return lock_manager.get_rwlock();
}


LockManager::~LockManager()
{
  assert(this == &lock_manager);
  lock_manager_destructed = true;

  while (!free_rwlock_list.empty())
    {
      RWLock *rwlock = free_rwlock_list.front();
      free_rwlock_list.pop_front();
      if (rwlock->ref_count != 1)
	cerr << "Something wrong with RWLock(@" << rwlock << ")\n";
      delete rwlock;
    }
}


Lock::Lock(const char *description)
  : description(description),
#ifdef CHECK_LOCK_ORDER
    locker_info(NULL),
#endif
    count(0)
{
#ifdef CHECK_LOCK_ORDER
  pthread_once(&lock_set_giant_once, lock_set_giant_init);
#endif

  if (pthread_mutex_init(&mutex, NULL) != 0)
    {
      cerr << "Couldn't initialize a mutex: " << description << endl;
      abort();
    }

  memset(&locker, 0, sizeof(pthread_t));
}


Lock::~Lock()
{
#ifdef CHECK_LOCK_ORDER
  assert(locker_info == NULL);

  int result = pthread_mutex_lock(&lock_set_giant);
  if (result != 0)
    {
      ErrorMessage error_message(result);
      fprintf(stderr, "Couldn't lock a mutex: lock_set_giant: %s\n",
	      error_message.get_message());
      abort();
    }


  LockSet::iterator iter = predecessor_locks.begin();
  LockSet::iterator end_iter = predecessor_locks.end();
  while (iter != end_iter)
    {
      Lock *pred_lock = *iter++;
      assert(pred_lock->successor_locks.find(this)
	     != pred_lock->successor_locks.end());
      pred_lock->successor_locks.erase(this);
    }

  iter = successor_locks.begin();
  end_iter = successor_locks.end();
  while (iter != end_iter)
    {
      Lock *succ_lock = *iter++;
      assert(succ_lock->predecessor_locks.find(this)
	     != succ_lock->predecessor_locks.end());
      succ_lock->predecessor_locks.erase(this);
    }


  result = pthread_mutex_unlock(&lock_set_giant);
  if (result != 0)
    {
      ErrorMessage error_message(result);
      fprintf(stderr, "Couldn't unlock a mutex: lock_set_giant: %s\n",
	      error_message.get_message());
      abort();
    }
#endif


  if (pthread_mutex_destroy(&mutex) != 0)
    {
      cerr << "Couldn't destroy a mutex: " << description << endl;
      abort();
    }
}


#ifdef CHECK_LOCK_ORDER
void
Lock::check_lock_order(ThreadInfo *info)
{
  assert(count > 0 || !pthread_equal(locker, pthread_self()));

  LockList reversed_locks;


  int result = pthread_mutex_lock(&lock_set_giant);
  if (result != 0)
    {
      ErrorMessage error_message(result);
      fprintf(stderr, "Couldn't lock a mutex: lock_set_giant: %s\n",
	      error_message.get_message());
      abort();
    }


  LockSet::const_iterator iter = info->holding_locks_begin();
  LockSet::const_iterator end_iter = info->holding_locks_end();
  while (iter != end_iter)
    {
      const Lock *pred_lock = *iter++;
      if (successor_locks.find(const_cast<Lock *>(pred_lock))
	  != successor_locks.end())
	{
	  // Lock Order Reversal!
	  reversed_locks.push_front(const_cast<Lock *>(pred_lock));
	}
    }

  if (reversed_locks.empty())
    {
      result = pthread_mutex_unlock(&lock_set_giant);
      if (result != 0)
	{
	  ErrorMessage error_message(result);
	  fprintf(stderr, "Couldn't unlock a mutex: lock_set_giant: %s\n",
		  error_message.get_message());
	  abort();
	}

      return;
    }

  fprintf(stderr, "*** Lock Order Reversal *** found in\n");
  info->dump_info();
  dump_description();
  fprintf(stderr,
	  "Attempting to hold the above lock, \"%s\", in reverse order with locks below:\n",
	  get_description());
  while (!reversed_locks.empty())
    {
      Lock *succ_lock = reversed_locks.front();
      succ_lock->dump_description();

      reversed_locks.pop_front();
    }


  result = pthread_mutex_unlock(&lock_set_giant);
  if (result != 0)
    {
      ErrorMessage error_message(result);
      fprintf(stderr, "Couldn't unlock a mutex: lock_set_giant: %s\n",
	      error_message.get_message());
    }

  abort();
}


void
Lock::record_lock_order(ThreadInfo *info)
{
  int result = pthread_mutex_lock(&lock_set_giant);
  if (result != 0)
    {
      ErrorMessage error_message(result);
      fprintf(stderr, "Couldn't lock a mutex: lock_set_giant: %s\n",
	      error_message.get_message());
      abort();
    }


  LockSet::iterator iter = info->holding_locks.begin();
  LockSet::iterator end_iter = info->holding_locks.end();
  while (iter != end_iter)
    {
      Lock *pred_lock = *iter++;
      if (pred_lock == this)
	continue;

      pred_lock->successor_locks.insert(this);
      assert(pred_lock->predecessor_locks.find(this)
	     == pred_lock->predecessor_locks.end());

      predecessor_locks.insert(pred_lock);
      assert(successor_locks.find(pred_lock) == successor_locks.end());
    }


  result = pthread_mutex_unlock(&lock_set_giant);
  if (result != 0)
    {
      ErrorMessage error_message(result);
      fprintf(stderr, "Couldn't unlock a mutex: lock_set_giant: %s\n",
	      error_message.get_message());
      abort();
    }
}
#endif


void
Lock::lock()
{
  if (count > 0 && pthread_equal(locker, pthread_self()))
    {
      // recursive2ܰʹߤlock()ξ礳롣
      // pthread_equal(locker, pthread_self())trueξˤΤߡ
      // countΰ̣ݾڤƤ뤳Ȥա
      ++count;
      return;
    }

#ifdef CHECK_LOCK_ORDER
  ThreadInfo *info = ThreadInfo::get_current();
  check_lock_order(info);
#endif

  int result = pthread_mutex_lock(&mutex);
  if (result == 0)
    {
      assert(count == 0);

      locker = pthread_self();

#ifdef CHECK_LOCK_ORDER
      locker_info = info;
      locker_info->notify_locked(this);

      record_lock_order(info);
#endif

      count = 1;
      return;
    }

  ErrorMessage error_message(result);
  fprintf(stderr, "Couldn't lock a mutex: %s: %s\n",
	  description.c_str(), error_message.get_message());
  abort();
}


bool
Lock::trylock()
{
  if (count > 0 && pthread_equal(locker, pthread_self()))
    {
      ++count;
      return true;
    }

  int result;
  if ((result = pthread_mutex_trylock(&mutex)) == 0)
    {
      assert(count == 0);
      locker = pthread_self();

#ifdef CHECK_LOCK_ORDER
      locker_info = ThreadInfo::get_current();
      locker_info->notify_locked(this);
#endif

      count = 1;
      return true;
    }

  assert(result == EBUSY);
  return false;
}


void
Lock::unlock()
{
  if (count < 1 || !pthread_equal(locker, pthread_self()))
    {
      cerr << "Couldn't unlock a Lock held by another thread!: ";
      dump_description();
      abort();
    }

  if (--count > 0)
    return;

  assert(count == 0);
  memset(&locker, 0, sizeof(pthread_t));

#ifdef CHECK_LOCK_ORDER
  assert(locker_info == ThreadInfo::get_current());

  locker_info->notify_unlocked(this);
  locker_info = NULL;
#endif

  int result = pthread_mutex_unlock(&mutex);
  if (result == 0)
    return;

  ErrorMessage error_message(result);
  fprintf(stderr, "Couldn't unlock a mutex: %s: %s\n",
	  description.c_str(), error_message.get_message());
  abort();
}


bool
Lock::is_locked_by_self()
{
  return count > 0 && pthread_equal(locker, pthread_self()) != 0;
}


const char *
Lock::get_typename() const
{
  return "Lock";
}


void
Lock::dump_brief_description(string &buf) const
{
  buf.append(get_typename()).append("(\"").append(get_description());
  char trailing[64];
  snprintf(trailing, 64, "\")@%p", this);
  buf.append(trailing);
}


void
Lock::dump_description() const
{
#ifdef CHECK_LOCK_ORDER
  fprintf(stderr, "%s(\"%s\")@%p {\n",
	  get_typename(), get_description(), this);

  fprintf(stderr, "  predecessor_locks: ");
  LockSet::const_iterator iter = predecessor_locks.begin();
  LockSet::const_iterator end_iter = predecessor_locks.end();
  while (iter != end_iter)
    {
      Lock *pred = *iter++;
      fprintf(stderr, "%s(\"%s\")@%p, ",
	      pred->get_typename(), pred->get_description(), pred);
    }
  fprintf(stderr, "\n");


  fprintf(stderr, "  successor_locks: ");
  iter = successor_locks.begin();
  end_iter = successor_locks.end();
  while (iter != end_iter)
    {
      Lock *succ = *iter++;
      fprintf(stderr, "%s(\"%s\")@%p, ",
	      succ->get_typename(), succ->get_description(), succ);
    }
  fprintf(stderr, "\n");


  fprintf(stderr, "}\n");
#else
  fprintf(stderr, "%s(\"%s\")@%p\n",
	  get_typename(), get_description(), this);
#endif
}


Monitor::Monitor(const char *description)
  : Lock(description)
{
  if (pthread_cond_init(&cond, NULL) != 0)
    {
      cerr << "Couldn't initialize a condition: " << description << endl;
      abort();
    }
}


Monitor::~Monitor()
{
  if (pthread_cond_destroy(&cond) != 0)
    {
      cerr << "Couldn't destroy a condition: " << get_description() << endl;
      abort();
    }
}


const char *
Monitor::get_typename() const
{
  return "Monitor";
}


void
Monitor::wait()
{
  assert(is_locked_by_self());
  assert(count > 0);

  int prev_count = count;
  count = 0;
  memset(&locker, 0, sizeof(pthread_t));
#ifdef CHECK_LOCK_ORDER
  assert(locker_info == ThreadInfo::get_current());

  locker_info->notify_waited(this);
  locker_info = NULL;
#endif

  if (pthread_cond_wait(&cond, &mutex) != 0)
    {
      cerr << "Couldn't wait a condition: " << get_description() << endl;
      abort();
    }

  assert(count == 0);

  locker = pthread_self();

#ifdef CHECK_LOCK_ORDER
  locker_info = ThreadInfo::get_current();
  locker_info->notify_signaled(this);
#endif

  count = prev_count;
}


bool
Monitor::wait(long ms)
{
  assert(is_locked_by_self());
  assert(count > 0);

#if defined(HAVE_CLOCK_GETTIME) || defined(HAVE_GETTIMEOFDAY)
  struct timespec abstime;
  if (ms > 0)
    {
# ifdef HAVE_CLOCKGETTIME
      if (clock_gettime(CLOCK_REALTIME, &abstime) == 0)
	{
	  long nsec = abstime.tv_nsec + (ms % 1000) * 1000000;
	  abstime.tv_sec += (ms / 1000) + (nsec / 1000000000);
	  abstime.tv_nsec = nsec % 1000000000;
	}
      else
	ms = 0;
# else
      struct timeval abstime_now;
      if (gettimeofday(&abstime_now, NULL) == 0)
	{
	  long nsec;
	  abstime.tv_sec = abstime_now.tv_sec;
	  abstime.tv_nsec = abstime_now.tv_usec * 1000;

	  nsec = abstime.tv_nsec + (ms % 1000) * 1000000;
	  abstime.tv_sec += (ms / 1000) + (nsec / 1000000000);
	  abstime.tv_nsec = nsec % 1000000000;
	}
      else
	ms = 0;
# endif
    }

  if (ms == 0)
    {
      wait();
      return true;
    }

  int prev_count = count;
  count = 0;
  memset(&locker, 0, sizeof(pthread_t));
# ifdef CHECK_LOCK_ORDER
  assert(locker_info == ThreadInfo::get_current());

  locker_info->notify_waited(this);
  locker_info = NULL;
# endif

  int result = pthread_cond_timedwait(&cond, &mutex, &abstime);
  if (result != 0 && result != ETIMEDOUT && result != EINTR)
    {
      ErrorMessage error_message(result);
      fprintf(stderr, "Couldn't wait a condition: %s : %s\n",
	      get_description(), error_message.get_message());
      abort();
    }

  assert(count == 0);

  locker = pthread_self();

# ifdef CHECK_LOCK_ORDER
  locker_info = ThreadInfo::get_current();
  locker_info->notify_signaled(this);
# endif

  count = prev_count;

  return result == 0;
#else
  wait();
  return true;
#endif
}


void
Monitor::notify()
{
  if (pthread_cond_signal(&cond) != 0)
    {
      cerr << "Couldn't signal a condition: " << get_description() << endl;
      abort();
    }
}


void
Monitor::notify_all()
{
  if (pthread_cond_broadcast(&cond) != 0)
    {
      cerr << "Couldn't broadcast a condition: " << get_description() << endl;
      abort();
    }
}


RWLock::RWLock() : ref_count(1)
{
#ifdef USE_PTHREAD_RWLOCK
  if (pthread_rwlock_init(&rwlock, NULL) != 0)
    {
      cerr << "Couldn't initialize an rwlock.\n";
      abort();
    }
#else
  if (pthread_mutex_init(&mutex, NULL) != 0)
    {
      cerr << "Couldn't initialize a mutex.\n";
      abort();
    }

  if (pthread_cond_init(&reader_cond, NULL) != 0)
    {
      cerr << "Couldn't initialize a condition.\n";
      abort();
    }

  if (pthread_cond_init(&writer_cond, NULL) != 0)
    {
      cerr << "Couldn't initialize a condition.\n";
      abort();
    }

  number_of_readers = 0;
  number_of_writers = 0;
#endif
}


RWLock::~RWLock()
{
#ifdef VERBOSE_DESTRUCTOR
  cerr << "RWLock(@" << this << ") destructed.\n";
#endif
#ifdef USE_PTHREAD_RWLOCK
  if (pthread_rwlock_destroy(&rwlock) != 0)
    {
      cerr << "Couldn't destroy a rwlock.\n";
      abort();
    }
#else
  if (number_of_readers != 0 || number_of_writers != 0)
    {
      cerr << "RWLock is busy.\n";
      abort();
    }

  if (pthread_cond_destroy(&writer_cond) != 0)
    {
      cerr << "Couldn't destroy a condition.\n";
      abort();
    }

  if (pthread_cond_destroy(&reader_cond) != 0)
    {
      cerr << "Couldn't destroy a condition.\n";
      abort();
    }

  if (pthread_mutex_destroy(&mutex) != 0)
    {
      cerr << "Couldn't destroy a mutex.\n";
      abort();
    }
#endif
}


void
RWLock::release_rwlock()
{
#ifdef VERBOSE_DESTRUCTOR
  cerr << "RWLock(@" << this << ") released.\n"; 
#endif
#ifndef USE_PTHREAD_RWLOCK
  assert(number_of_readers == 0);
  assert(number_of_writers == 0);
#endif
  if (ref_count == 1)
    {
      if (lock_manager_destructed)
	delete this;
      else
	LockManager::lock_manager.rwlock_free(this);
      return;
    }

  assert(ref_count == 0);

  delete this;
  return;
}


int
RWLock::rdlock()
{
#ifdef USE_PTHREAD_RWLOCK
  if (pthread_rwlock_rdlock(&rwlock) != 0)
    {
      cerr << "Couldn't rdlock a rwlock.\n";
      abort();
    }
  return 2;				// Ȥƻʤʡ
#else
  mutex_lock();

  while (number_of_writers > 0)
    reader_cond_wait();

  ++number_of_readers;
  int lock_id = 2;

  // readeråʤΤ¾readerwriterϵ̵̡
  reader_cond_broadcast();
  mutex_unlock();

  return lock_id;
#endif
}


int
RWLock::tryrdlock()
{
#ifdef USE_PTHREAD_RWLOCK
  int result = pthread_rwlock_tryrdlock(&rwlock);

  if (result == 0)
    return 2;			// Ȥƻʤʡ

  if (result == EBUSY)
    return 0;			// åƤwriter롣

  cerr << "Couldn't tryrdlock a rwlock.\n";
  abort();
#else
  mutex_lock();

  int lock_id;
  if (number_of_writers == 0)
    {
      ++number_of_readers;
      lock_id = 2;
      // readeråʤΤ¾readerwriterϵ̵̡
      reader_cond_broadcast();
    }
  else
    lock_id = 0;

  mutex_unlock();

  return lock_id;
#endif
}


int
RWLock::wrlock()
{
#ifdef USE_PTHREAD_RWLOCK
  if (pthread_rwlock_wrlock(&rwlock) != 0)
    {
      cerr << "Couldn't wrlock a rwlock.\n";
      abort();
    }
  return 3;				// Ȥƻʤʡ
#else
  mutex_lock();

  ++number_of_writers;			// waitϿ

  while (number_of_readers > 0)
    writer_cond_wait();			// ƤreaderΥåԤ

  int lock_id = 3;			// writerlock_idϴȤ롣

  mutex_unlock();

  return lock_id;
#endif
}


int
RWLock::trywrlock()
{
#ifdef USE_PTHREAD_RWLOCK
  int result = pthread_rwlock_trywrlock(&rwlock);

  if (result == 0)
    return 3;			// Ȥƻʤʡ

  if (result == EBUSY)
    return 0;			// åƤwriter롣

  cerr << "Couldn't tryrdlock a rwlock.\n";
  abort();
#else
  mutex_lock();

  int lock_id;
  if (number_of_readers == 0)
    {
      ++number_of_writers;
      lock_id = 3;			// writerlock_idϴȤ롣
    }
  else
    lock_id = 0;

  mutex_unlock();

  return lock_id;
#endif
}


int
RWLock::unlock(int lock_id)
{
#ifdef USE_PTHREAD_RWLOCK
  if (pthread_rwlock_unlock(&rwlock) != 0)
    {
      cerr << "Couldn't unlock an rwlock.\n";
      abort();
    }
  return lock_id;
#else
  if (lock_id & 1)
    return release_writer_lock(lock_id);
  return release_reader_lock(lock_id);
#endif
}


#ifndef USE_PTHREAD_RWLOCK
int
RWLock::release_reader_lock(int lock_id)
{
  assert((lock_id & 1) == 0);
  mutex_lock();

  if (--number_of_readers < 0)
    {
      cerr << "Something wrong.  It seems not to be locked.\n";
      abort();	// 
    }

  if (number_of_readers == 0)	// ʬǸreaderäΤǡ
    writer_cond_broadcast();	// writer򵯤ͤꡣ

  mutex_unlock();

  return lock_id;
}


int
RWLock::release_writer_lock(int lock_id)
{
  assert(lock_id & 1);
  mutex_lock();

  if (--number_of_writers < 0)
    {
      cerr << "Something wrong.  It seems not to be locked.\n";
      abort();	// 
    }

  if (number_of_writers == 0)	// ʬǸwriteräΤǡ
    reader_cond_broadcast();	// Եreader򵯤ͤꡣ

  mutex_unlock();

  return lock_id;
}
#endif


}
