/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: lock.h,v 1.22 2009/01/01 07:46:43 fuyu Exp $
 */

#ifndef _LIBOCHUSHA_LOCK_H_
#define _LIBOCHUSHA_LOCK_H_

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#ifdef __cplusplus

#if defined(__GNUG__) && !defined(__ICC)
#pragma interface
#endif

#ifdef HAVE_PTHREAD_H
# include "libochusha/ref_count.h"

# include <cassert>
# include <cstdlib>

# include <iostream>
# include <list>
# include <set>
# include <string>

# include <pthread.h>

# ifdef HAVE_PTHREAD_RWLOCK_INIT
#  define USE_PTHREAD_RWLOCK	1
# endif

#endif


#define CHECK_LOCK_ORDER	1


using namespace std;
namespace Ochusha
{

#ifdef HAVE_PTHREAD_H
  class Lock;
  typedef list<Lock *> LockList;
  typedef set<Lock *> LockSet;

  class Monitor;

# ifdef CHECK_LOCK_ORDER
  class ThreadInfo
  {
    friend class Lock;
    friend class LockManager;
    friend class Monitor;

  private:
    string description;
    LockSet holding_locks;
    Monitor *waiting_monitor;
    int maximum_number_of_simultaneous_locks;

    inline LockSet::iterator holding_locks_begin()
    { return holding_locks.begin(); }
    inline LockSet::iterator holding_locks_end()
    { return holding_locks.end(); }

    void notify_locked(Lock *lock);
    void notify_unlocked(Lock *lock);
    void notify_waited(Monitor *monitor);
    void notify_signaled(Monitor *monitor);

    static pthread_once_t thread_info_key_once;
    static pthread_key_t thread_info_key;

    static void thread_info_key_init();
    static void thread_info_key_destroy(void *data);
    static void set_current(ThreadInfo *info);


  protected:
    virtual ~ThreadInfo();
    void dump_info_body();


  public:
    // newǥȥåɤޤ롣delete⤪Ǥɤ
    ThreadInfo(const char *description);
    virtual const char *get_description();
    virtual void dump_info();

    static ThreadInfo *get_current();
  };
# endif


  // Recursive Lock
  class Lock
  {
  private:
    string description;

#ifdef CHECK_LOCK_ORDER
    // ()Ĥ¾this->mutexϻȤʤ
    LockSet predecessor_locks;
    LockSet successor_locks;
    // ()Ĥ¾this->mutexϻȤʤ
#endif


  protected:
    pthread_mutex_t mutex;
    pthread_t locker;

#ifdef CHECK_LOCK_ORDER
    ThreadInfo *locker_info;
#endif

    int count;


#ifdef CHECK_LOCK_ORDER
    void check_lock_order(ThreadInfo *info);
    void record_lock_order(ThreadInfo *info);
#endif


  public:
    Lock(const char *description);
    virtual ~Lock();

    void lock();
    bool trylock();
    void unlock();

    bool is_locked_by_self();

    virtual const char *get_typename() const;
    inline const char *get_description() const { return description.c_str(); }
    virtual void dump_brief_description(string &buf) const;
    virtual void dump_description() const;

    inline int get_count() const { return count; }
  };


  class LockBlock
  {
  private:
    Lock *lock;


  public:
    inline LockBlock(Lock *lock) : lock(lock) { lock->lock(); }
    inline LockBlock(Lock &lock) : lock(&lock) { this->lock->lock(); }
    inline ~LockBlock() { lock->unlock(); }
  };


  class Monitor : public Lock
  {
  private:
    pthread_cond_t cond;


  public:
    Monitor(const char *description);
    virtual ~Monitor();

    virtual const char *get_typename() const;

    inline void enter() { lock(); }	// ñʤ̾
    inline void exit() { unlock(); }	// ñʤ̾
    inline void leave() { unlock(); }	// ñʤ̾

    void wait();
    bool wait(long ms);
    void notify();
    void notify_all();
  };


  class MonitorBlock
  {
  private:
    Monitor *monitor;


  public:
    inline MonitorBlock(Monitor *monitor) : monitor(monitor)
    { monitor->enter(); }
    inline MonitorBlock(Monitor &monitor) : monitor(&monitor)
    { monitor.enter(); }
    inline ~MonitorBlock()
    { monitor->notify_all(); monitor->exit(); }
    inline void wait() { monitor->wait(); }
    inline bool wait(long ms) { return monitor->wait(ms); }
    inline void notify() { monitor->notify(); }
    inline void notify_all() { monitor->notify_all(); }
  };


  class RWLock
  {
    friend class LockManager;

  private:
    int ref_count;
#ifdef USE_PTHREAD_RWLOCK
    pthread_rwlock_t rwlock;
#else
    pthread_mutex_t mutex;
    pthread_cond_t reader_cond;	// reader lockwriterβԤ
    pthread_cond_t writer_cond;	// writer lockreaderβԤ

    int number_of_readers;
    int number_of_writers;
#endif
    RWLock();

    void release_rwlock();


  public:
    ~RWLock();
    inline void ref() { ++ref_count; }
    inline void unref() { if (--ref_count > 1) return; release_rwlock(); }

    int rdlock();
    int tryrdlock();

    int wrlock();
    int trywrlock();

    int unlock(int lock_id);


#ifndef USE_PTHREAD_RWLOCK
  private:
    inline void mutex_lock()
    {
      if (pthread_mutex_lock(&mutex) != 0)
	{
	  cerr << "Couldn't lock a mutex.\n";
	  abort();
	}
    }

    inline void mutex_unlock()
    {
      if (pthread_mutex_unlock(&mutex) != 0)
	{
	  cerr << "Couldn't unlock a mutex.\n";
	  abort();
	}
    }

    inline void reader_cond_wait()
    {
      if (pthread_cond_wait(&reader_cond, &mutex) != 0)
	{
	  cerr << "Couldn't wait a reader condition.\n";
	  abort();
	}
    }

    inline void reader_cond_signal()
    {
      if (pthread_cond_signal(&reader_cond) != 0)
	{
	  cerr << "Couldn't signal a reader condition.\n";
	  abort();
	}
    }

    inline void reader_cond_broadcast()
    {
      if (pthread_cond_broadcast(&reader_cond) != 0)
	{
	  cerr << "Couldn't broadcast a reader condition.\n";
	  abort();
	}
    }

    inline void writer_cond_wait()
    {
      if (pthread_cond_wait(&writer_cond, &mutex) != 0)
	{
	  cerr << "Couldn't wait a writer condition.\n";
	  abort();
	}
    }

    inline void writer_cond_signal()
    {
      if (pthread_cond_signal(&writer_cond) != 0)
	{
	  cerr << "Couldn't signal a writer condition.\n";
	  abort();
	}
    }

    inline void writer_cond_broadcast()
    {
      if (pthread_cond_broadcast(&writer_cond) != 0)
	{
	  cerr << "Couldn't broadcast a writer condition.\n";
	  abort();
	}
    }

    int release_reader_lock(int lock_id);
    int release_writer_lock(int lock_id);
#endif
  };


  class LockManager
  {
    friend class Lock;
    friend class Monitor;
    friend class RWLock;

  private:
    Lock lock;

    list<RWLock *> free_rwlock_list;

    inline RWLock *get_rwlock()
    {
      LockBlock lock_obj(lock);
      if (free_rwlock_list.empty())
	return new RWLock();
      RWLock *rwlock = free_rwlock_list.front();
      free_rwlock_list.pop_front();
      return rwlock;
    }

    inline void rwlock_free(RWLock *rwlock)
    {
      LockBlock lock_obj(lock);
      assert(rwlock->ref_count == 1);
      free_rwlock_list.push_front(rwlock);
    }


    LockManager();

    static LockManager lock_manager;


  public:
    ~LockManager();

    static RWLock *rwlock_new();
  };


#ifndef _LIBOCHUSHA_LOCK_CC_
extern Monitor libochusha_giant;
#endif


#endif	/* HAVE_PTHREAD_H */

}


#endif	/* __cplusplus */

#endif	/* _LIBOCHUSHA_LOCK_H_ */
