/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: misc_utils.cc,v 1.17 2009/01/01 07:46:43 fuyu Exp $
 */

#ifdef __linux__
#define _XOPEN_SOURCE
#endif
#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/misc_utils.h"

#include "libochusha/lock.h"
#include "libochusha/regexp.h"

#include <string>

#include <cassert>
#include <cerrno>
#include <climits>
#include <cstdio>
#include <cstdlib>
#include <cstring>

#ifdef HAVE_LOCALE_H
#include <locale.h>
#endif

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

#ifdef HAVE_TIME_H
#include <time.h>
#endif


namespace Ochusha
{

ErrorMessage::ErrorMessage()
  : error_number(0)
{
  buf[0] = '\0';
}


ErrorMessage::ErrorMessage(int errnum)
  : error_number(errnum)
{
#ifdef HAVE_STRERROR_R
  char message[512];
  strerror_r(errnum, message, 512);
  const char *err = message;
#else
  const char *err = strerror(errnum);
#endif
  snprintf(buf, 512, "%s (%d)", err, errnum);
}


void
ErrorMessage::set_errno(int errnum)
{
  error_number = errnum;
#ifdef HAVE_STRERROR_R
  char message[512];
  strerror_r(errnum, message, 512);
  const char *err = message;
#else
  const char *err = strerror(errnum);
#endif
  snprintf(buf, 512, "%s (%d)", err, errnum);
}


void
ErrorMessage::set_unknown_errno(int errnum)
{
  error_number = errnum;
  snprintf(buf, 512, LIB_("Unknown error: %d"), errnum);
}


string &
append_markup_escaped_text(string &str, const char *text)
{
  const char *cur_pos = text;

  while (true)
    {
      const char *tmp_pos = strpbrk(cur_pos, "&<>'\"");

      if (tmp_pos == NULL)
	{
	  str.append(cur_pos);
	  return str;
	}

      if (tmp_pos - cur_pos > 0)
	str.append(cur_pos, tmp_pos - cur_pos);

      switch (*tmp_pos)
	{
	case '&':
	  str.append("&amp;"); break;
	case '<':
	  str.append("&lt;"); break;
	case '>':
	  str.append("&gt;"); break;
	case '\'':
	  str.append("&apos;"); break;
	case '"':
	  str.append("&quot;"); break;
	}
      cur_pos = tmp_pos + 1;
    }
}


}


/*
 * mkdir_p
 *
 * mkdir -pޥɤΤ褦ˡpathǻꤵ줿ǥ쥯ȥŪ
 * 롣
 * ǥ쥯ȥ꤬¸ߤ뤫ǥ쥯ȥκˤ0
 * ʳξ-1֤
 */
int
mkdir_p(const char *path)
{
  char buffer[PATH_MAX];
  char *cur_pos;

  assert(path != NULL);

  strncpy(buffer, path, PATH_MAX);
  buffer[PATH_MAX - 1] = '\0';
  size_t len = strlen(buffer);
  if (buffer[len - 1] == '/')
    buffer[len - 1] = '\0';	// Ǹ'/'̵롣

  cur_pos = buffer;
  while (1)
    {
      int result;
      struct stat sb;
      char tmp_char;

      cur_pos = strchr(cur_pos, '/');
      if (cur_pos != NULL)
	{
	  tmp_char = *(++cur_pos);
	  *cur_pos = '\0';
	}
      else
	tmp_char = '\0';

      result = stat(buffer, &sb);

      if (result == -1)
	{
	  if (errno != ENOENT)
	    return -1;
	
	  result = mkdir(buffer, 0x1c0);
	  if (result != 0)
	    return -1;
	}
      else
	if (!S_ISDIR(sb.st_mode))
	  return -1;

      if (tmp_char == '\0')
	return 0;

      *cur_pos = tmp_char;
    }
}


char *
markup_escape_text(const char *text)
{
  string tmp;
  Ochusha::append_markup_escaped_text(tmp, text);
  return strdup(tmp.c_str());
}


/*
 * Ϥޥɥ饤ȤƻȤ˥פɬפʸ
 * ƥפʸΥԡ֤
 * Ǥϡ#\#֤
 */
char *
shell_escape_text(const char *text)
{
  string command_line;
  const char *cur_pos = text;
  while (*cur_pos != '\0')
    {
      if (*cur_pos != '#')
	command_line.append(cur_pos, 1);
      else
	command_line.append("\\#");
      ++cur_pos;
    }

  return strdup(command_line.c_str());
}


class DateParserHelper
{
private:
#ifndef HAVE_TIMEGM
  time_t delta;
#endif

public:
  DateParserHelper();
#ifndef HAVE_TIMEGM
  inline time_t get_delta() { return delta; }
#endif
};


DateParserHelper::DateParserHelper()
{
#ifdef HAVE_LOCALE_H
  static char *lc_time = NULL;

  if (lc_time == NULL)
    {
      lc_time = setlocale(LC_TIME, "C");
      tzset();
    }
#endif

#ifndef HAVE_TIMEGM
  time_t cur_time = time(NULL);
  struct tm *cur_time_gm_p;
#ifdef HAVE_GMTIME_R
  struct tm cur_time_gm;
  memset(&cur_time_gm, 0, sizeof(struct tm));
  cur_time_gm_p = gmtime_r(&cur_time, &cur_time_gm);
#else
  cur_time_gm_p = gmtime(&cur_time);
#endif

  delta = cur_time - mktime(cur_time_gm_p);
#endif
}


static DateParserHelper date_parser_helper;


#ifndef HAVE_TIMEGM
time_t
timegm(struct tm *tm)
{
  return mktime(tm) + date_parser_helper.get_delta();
}
#endif


static const char *
gmt_parse_date(const char *date_string, struct tm *tm)
{
  if (date_string == NULL)
    return NULL;

  static Ochusha::Lock lock("MiscUtils::gmt_parse_date");

  Ochusha::LockBlock lock_obj(lock);
  static Ochusha::Regexp date_pattern("([FMTSW][aehoru][deintu]\\w*),\\s+([0-3][0-9])(?:\\s+|-)([ADFJMNOS][aceopu][bcglnprtvy])(?:\\s+|-)([0-9]+\\s+[0-2][0-9]:[0-5][0-9]:[0-6][0-9]\\s+GMT)");

  if (!date_pattern.match(date_string, strlen(date_string)))
    return NULL;

  const char *rest_string = date_pattern.match_end(0);

  // ƬʸʬӤʤʹߤ˥ڥ󥰥ߥȤäƤ̵롣
  int week;
  const char *w = date_pattern.match_begin(1);
  if (strncmp(w, "Sun", 3) == 0)
    week = 0;
  else if (strncmp(w, "Mon", 3) == 0)
    week = 1;
  else if (strncmp(w, "Tue", 3) == 0)
    week = 2;
  else if (strncmp(w, "Wed", 3) == 0)
    week = 3;
  else if (strncmp(w, "Thu", 3) == 0)
    week = 4;
  else if (strncmp(w, "Fri", 3) == 0)
    week = 5;
  else if (strncmp(w, "Sat", 3) == 0)
    week = 6;
  else
    return NULL;
  
  int day = 0;
  if (sscanf(date_pattern.match_begin(2), "%d", &day) != 1)
    return NULL;	// date_patternޥåƤ뤫ͭʤϤ

  int month;
  const char *m = date_pattern.match_begin(3);
  if (strncmp(m, "Jan", 3) == 0)
    month = 0;
  else if (strncmp(m, "Feb", 3) == 0)
    month = 1;
  else if (strncmp(m, "Mar", 3) == 0)
    month = 2;
  else if (strncmp(m, "Apr", 3) == 0)
    month = 3;
  else if (strncmp(m, "May", 3) == 0)
    month = 4;
  else if (strncmp(m, "Jun", 3) == 0)
    month = 5;
  else if (strncmp(m, "Jul", 3) == 0)
    month = 6;
  else if (strncmp(m, "Aug", 3) == 0)
    month = 7;
  else if (strncmp(m, "Sep", 3) == 0)
    month = 8;
  else if (strncmp(m, "Oct", 3) == 0)
    month = 9;
  else if (strncmp(m, "Nov", 3) == 0)
    month = 10;
  else if (strncmp(m, "Dec", 3) == 0)
    month = 11;
  else
    return NULL;

  int year = 0;
  int hour = 0;
  int minute = 0;
  int second = 0;
  if (sscanf(date_pattern.match_begin(4), "%d %d:%d:%d GMT",
	     &year, &hour, &minute, &second) != 4)
    return NULL;	// date_patternޥåƤ뤫ͭʤϤ

  if (hour > 23)
    return NULL;	// error

  if (minute > 59)
    return NULL;	// error

  if (second > 60)	// 뤦äΤ60äޤǤk
    return NULL;	// error

  if (tm != NULL)
    {
      tm->tm_sec = second;
      tm->tm_min = minute;
      tm->tm_hour = hour;
      tm->tm_mday = day;
      tm->tm_mon = month;
      tm->tm_year = year - 1900;
      tm->tm_wday = week;
    }

  return rest_string;
}


/*
 * get_utc_from_date
 *
 * Ϳ줿ʸɽʸȤƲᤷUTC֤
 */
time_t
get_utc_from_date(const char *date_string)
{
  if (date_string == NULL)
    return static_cast<time_t>(-1);

  struct tm tm;
  memset(&tm, 0, sizeof(struct tm));

  if (gmt_parse_date(date_string, &tm) == NULL
      && strptime(date_string, "%a, %d %b %Y %T", &tm) == NULL
      && strptime(date_string, "%a, %d-%b-%Y %T", &tm) == NULL
      && strptime(date_string, "%A, %d-%b-%Y %T", &tm) == NULL)
    return static_cast<time_t>(-1);

  /*
   * ʡꤤtimezoneξäƤΤݤʤΤǡ
   * GMTʳξϼ¹ԴĶƱtimezoneǤȷĤ롣
   */
  return strstr(date_string, "GMT") != NULL ? timegm(&tm) : mktime(&tm);
}


/*
 * get_gmt_date_from_utc
 *
 * Ϳ줿GMTɽʸ֤Last-Modifiedѡ
 * ̤bufǻꤵ줿buf_lenĹΰ˳Ǽ롣
 * ˤbuf֤餫Υ顼⤷buf_lenǤ϶֤­
 * äˤNULL֤
 */
char *
get_gmt_date_from_utc(time_t utc_time, char *buf, size_t buf_len)
{
#if defined(HAVE_GMTIME) || defined(HAVE_GMTIME_R)

  struct tm *gmt_p;
#ifdef HAVE_GMTIME_R
  struct tm gmt;
  memset(&gmt, 0, sizeof(struct tm));
  gmt_p = gmtime_r(&utc_time, &gmt);
#else
  gmt_p = gmtime(&utc_time);
#endif

  if (gmt_p == NULL)
    return NULL;	// 

  static const char *const week[] = {
    "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"
  };

  static const char *const month[] = {
    "Jan", "Feb", "Mar", "Apr", "May", "Jun",
    "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
  };

  if (snprintf(buf, buf_len, "%s, %02d %s %04d %02d:%02d:%02d GMT",
	       week[gmt_p->tm_wday], gmt_p->tm_mday,
	       month[gmt_p->tm_mon], gmt_p->tm_year + 1900,
	       gmt_p->tm_hour, gmt_p->tm_min, gmt_p->tm_sec)
      < static_cast<int>(buf_len))
    return buf;

  return NULL;
#else
  return NULL;		// 
#endif
}


/*
 * Ĺǽstrpbrkؿ
 * stringĹlenʸcharset˴ޤޤǽʸؤΥݥ󥿤֤
 * ʤNULL֤
 */
char *
mempbrk(const char *str, const char *charset, size_t len)
{
  while (len > 0)
    {
      const char *c = charset;
      char s = *str;
      while (*c)
	{
	  if (*c == s)
	    return const_cast<char *>(str);	// strpbrk˹碌
	  c++;
	}
      str++;
      len--;
    }
  return NULL;
}


char *
reverse_strpbrk(const char *str, const char *charset)
{
  const char *s = str;
  while (*s != '\0')
    {
      if (strchr(charset, *s) == NULL)
	return const_cast<char *>(s);
      s++;
    }

  return NULL;
}


#ifndef HAVE_STRNSTR
/*-
 * Copyright (c) 2001 Mike Barcroft <mike@FreeBSD.org>
 * Copyright (c) 1990, 1993
 *      The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Chris Torek.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

char *
strnstr(const char *s, const char *find, size_t slen)
{
  char c, sc;
  size_t len;

  if ((c = *find++) != '\0')
    {
      len = strlen(find);
      do
	{
	  do
	    {
	      if ((sc = *s++) == '\0' || slen-- < 1)
		return NULL;
	    }
	  while (sc != c);

	  if (len > slen)
	    return NULL;
	}
      while (strncmp(s, find, len) != 0);

      s--;
    }
  return const_cast<char *>(s);
}
#endif


#ifndef HAVE_STRDUP
char *
strdup(const char *src)
{
  char *result = static_cast<char *>(malloc(strlen(src) + 1));
  if (result == NULL)
    return NULL;	// Out of Memory
  strcpy(result, src);
  return result;
}
#endif


#ifndef HAVE_STRNDUP
char *
strndup(const char *src, size_t len)
{
  char *result = static_cast<char *>(malloc(len + 1));
  if (result == NULL)
    return NULL;	// Out of Memory

  strncpy(result, src, len);
  result[len] = '\0';

  return result;
}
#endif


#ifndef HAVE_STRCASECMP
/*
 * Copyright (c) 1987, 1993
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
int
strcasecmp(const char *s1, const char *s2)
{
  register const unsigned char *us1 = (const unsigned char *)s1;
  register const unsigned char *us2 = (const unsigned char *)s2;

  while (tolower(*us1) == tolower(*us2++))
    if (*us1++ == '\0')
      return 0;
  return tolower(*us1) - tolower(*--us2);
}
#endif


#ifndef HAVE_STRNCASECMP
int
strncasecmp(const char *s1, const char *s2, size_t n)
{
  if (n != 0)
    {
      register const unsigned char *us1 = (const unsigned char *)s1;
      register const unsigned char *us2 = (const unsigned char *)s2;
      do
	{
	  if (tolower(*us1) != tolower(*us2++))
	    return tolower(*us1) - tolower(*--us2);
	  if (*us1++ == '\0')
	    break;
	}
      while (--n != 0);
    }
  return 0;
}
#endif


# ifndef HAVE_STRCASESTR
char *
strcasestr(const char *big, const char *little)
{
  int c;
  size_t len;
  if ((c = *little++) != '\0')
    {
      c = tolower(c);
      len = strlen(little);
      do
	{
	  char bc;
	  do
	    {
	      if ((bc = *big++) == '\0')
		return NULL;
	    }
	  while (tolower(bc) != c);
	}
      while (strncasecmp(big, little, len) != 0);

      big--;
    }
  return const_cast<char *>(big);
}
# endif


# ifndef HAVE_STRNCASESTR
char *
strncasestr(const char *big, const char *little, size_t blen)
{
  int c;
  size_t len;
  if ((c = *little++) != '\0')
    {
      c = tolower(c);
      len = strlen(little);
      do
	{
	  char bc;
	  do
	    {
	      if ((bc = *big++) == '\0' || --blen < 0)
		return NULL;
	    }
	  while (tolower(bc) != c);

	  if (len > blen)
	    return NULL;
	}
      while (strncasecmp(big, little, len) != 0);

      big--;
    }
  return const_cast<char *>(big);
}
# endif


char *
strnrstr(const char *big, const char *little, size_t blen)
{
  size_t len = strlen(little);
  const char *max_pos = big + blen - len;
  const char *cur_pos = max_pos;

  int c = *little;
  while (big <= cur_pos)
    { /* ޤΨŪǤʤΤϵˤʤ*/
      if (*cur_pos == c && strncmp(cur_pos, little, len) == 0)
	return const_cast<char *>(cur_pos);
      cur_pos--;
    }

  return NULL;
}


char *
strncaserstr(const char *big, const char *little, size_t blen)
{
  size_t len = strlen(little);
  const char *max_pos = big + blen - len;
  const char *cur_pos = max_pos;

  int c = tolower(*little);
  while (big <= cur_pos)
    { // ޤΨŪǤʤΤϵˤʤ
      if (tolower(*cur_pos) == c && strncasecmp(cur_pos, little, len) == 0)
	return const_cast<char *>(cur_pos);
      cur_pos--;
    }

  return NULL;
}
