/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: network_connection_pool.cc,v 1.24 2009/01/01 07:46:43 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/network_connection_pool.h"

#include "libochusha/misc_utils.h"
#include "libochusha/uri_utils.h"


#include <cassert>
#include <list>

#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif


namespace Ochusha
{

#define MAXIMUM_NUMBER_OF_CONNECTIONS_LIMIT	4
#define MINIMUM_PENALTY_SEC			60	// 1ʬ
#define MAXIMUM_PENALTY_SEC			720	// 2
#undef CHASE_REQUEST


typedef enum {
  REQUEST_OK,
  SERVER_BUSY,
  SERVER_DEAD
} ServerState;


class ServerInfo : public RefCount
{
private:
  int maximum_number_of_connections;
  int current_number_of_connections;
  struct timeval penalty_expiration_time;
  long penalty_sec;

  list<ConnectionRequest *> request_queue;

  ServerState check_state();


public:
  ServerInfo();
  ~ServerInfo();

  void append_request(ConnectionRequest *request);

  void notify_access_done();
  void notify_connection_failure();
  void notify_busy();
};


ServerInfo::ServerInfo()
  : maximum_number_of_connections(1),
    current_number_of_connections(0),
    penalty_sec(0)
{
  penalty_expiration_time.tv_sec = 0;
  penalty_expiration_time.tv_usec = 0;
}


ServerInfo::~ServerInfo()
{
  while (!request_queue.empty())
    {
      ConnectionRequest *request = request_queue.front();
      request->notify_denied();
      request_queue.pop_front();
    }
}


ServerState
ServerInfo::check_state()
{
  if (penalty_sec > 0 && maximum_number_of_connections == 0)
    {
      // ³絬桽Ȥ¿ʬ˴
      struct timeval current_time;
      gettimeofday(&current_time, NULL);
      if (current_time.tv_sec < penalty_expiration_time.tv_sec)
	return SERVER_DEAD;	// 絬³
      // 絬
      maximum_number_of_connections = 1;
    }

  if (current_number_of_connections >= maximum_number_of_connections)
    return SERVER_BUSY;		// ͥ

  return REQUEST_OK;
}


void
ServerInfo::append_request(ConnectionRequest *request)
{
  if (request_queue.empty())
    {
      switch (check_state())
	{
	case REQUEST_OK:
#ifdef CHASE_REQUEST
	  HTTPHandle *hdl = new HTTPHandle();
	  fprintf(stderr, "append_request: url=\"%s\", hdl=%p\n",
		  request->get_url(), hdl);
	  request->notify_allowed(hdl);
#else
	  request->notify_allowed(new HTTPHandle());
#endif
	  return;

	case SERVER_DEAD:
	  request->notify_denied();
	  return;

	case SERVER_BUSY:
	  if (!request->notify_busy())
	    return;
	}
    }
  request_queue.push_back(request);
}


void
ServerInfo::notify_access_done()
{
  // ̵λ
  current_number_of_connections--;
  if (penalty_sec == 0)
    maximum_number_of_connections = MIN(maximum_number_of_connections + 1,
					MAXIMUM_NUMBER_OF_CONNECTIONS_LIMIT);
  else
    penalty_sec = 0;	// 絬

  while (!request_queue.empty()
	 && current_number_of_connections < maximum_number_of_connections)
    {
      ConnectionRequest *request = request_queue.front();
      request_queue.pop_front();
#ifdef CHASE_REQUEST
      HTTPHandle *hdl = new HTTPHandle();
      fprintf(stderr, "notify_access_done: url=\"%s\", hdl=%p\n",
	      request->get_url(), hdl);
      request->notify_allowed(hdl);
#else
      request->notify_allowed(new HTTPHandle());
#endif
      ++current_number_of_connections;
    }
}


void
ServerInfo::notify_connection_failure()
{
  // ¿ʬ˴ʤΤǻä
  current_number_of_connections--;
  if (maximum_number_of_connections > 0)
    {
      maximum_number_of_connections = 0;
      if (penalty_sec == 0)
	penalty_sec = MINIMUM_PENALTY_SEC;
      else
	penalty_sec = MIN(penalty_sec * 5, MAXIMUM_PENALTY_SEC);

      gettimeofday(&penalty_expiration_time, NULL);
      penalty_expiration_time.tv_sec += penalty_sec;
    }

  // 塼äƤꥯȤԤ롣
  while (!request_queue.empty())
    {
      ConnectionRequest *request = request_queue.front();
      request_queue.pop_front();
      request->notify_denied();
    }
}


void
ServerInfo::notify_busy()
{
  current_number_of_connections--;
  maximum_number_of_connections = MAX(1, maximum_number_of_connections / 2);
  if (penalty_sec > 0)
    penalty_sec = 0;

  while (!request_queue.empty()
	 && current_number_of_connections < maximum_number_of_connections)
    {
      ConnectionRequest *request = request_queue.front();
      request_queue.pop_front();
#ifdef CHASE_REQUEST
      HTTPHandle *hdl = new HTTPHandle();
      fprintf(stderr, "notify_busy: url=\"%s\", hdl=%p\n",
	      request->get_url(), hdl);
      request->notify_allowed(hdl);
#else
      request->notify_allowed(new HTTPHandle());
#endif
      ++current_number_of_connections;
    }
}


ConnectionRequest::ConnectionRequest()
{
}


ConnectionRequest::~ConnectionRequest()
{
}


NetworkConnectionPool::NetworkConnectionPool(int max_connections)
  : server_info_table(101),
    lock("NetworkConnectionPool")
{
}


NetworkConnectionPool::~NetworkConnectionPool()
{
}


void
NetworkConnectionPool::request_connection(ConnectionRequest *request)
{
  assert(request != NULL);
  URI uri(request->get_url());

  if (!uri.is_valid())
    {
      request->notify_denied();
      return;
    }

  const char *scheme = uri.get_scheme();
  if (scheme == NULL
      || (strcmp(scheme, "http") != 0 && strcmp(scheme, "https") != 0))
    {
      request->notify_denied();
      return;
    }

  const char *server = uri.get_server();
  if (server == NULL)
    {
      request->notify_denied();
      return;
    }

  LockBlock lock_obj(lock);
  ServerInfo *info
    = dynamic_cast<ServerInfo *>(server_info_table.lookup(server));

  if (info == NULL)
    {
      info = new ServerInfo();
      if (info == NULL)
	{
	  request->notify_denied();
	  return;	// Out of memory.
	}
      server_info_table.insert(server, info);
    }

  info->append_request(request);
}


void
NetworkConnectionPool::release_http_handle(const char *url,
					   HTTPHandle *handle,
					   ConnectionState state)
{
  if (handle == NULL)
    return;

#ifdef CHASE_REQUEST
  fprintf(stderr, "release_http_request: url=\"%s\", handle=%p\n",
	  url, handle);
#endif
  URI uri(url);
  assert(uri.is_valid());

#ifndef NDEBUG
  const char *scheme = uri.get_scheme();
  assert(scheme != NULL);
  assert(strcmp(scheme, "http") == 0 || strcmp(scheme, "https") == 0);
#endif

  const char *server = uri.get_server();
  assert(server != NULL);

  // requestȤ󤹤ȡHTTP/1.1persistent connectionѤǤ뤬
  // ݤʤΤǸ󤷡
  delete handle;

  LockBlock lock_obj(lock);
  ServerInfo *info
    = dynamic_cast<ServerInfo *>(server_info_table.lookup(server));
  assert(info != NULL);

  switch (state)
    {
    case ACCESS_SUCCEEDED:
    case ACCESS_TERMINATED:
      info->notify_access_done();
      break;

    case ACCESS_BUSY:
      info->notify_busy();
      break;

    case ACCESS_FAILED:
      info->notify_connection_failure();
      break;
    }
}

}
