/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: poll_event_manager.cc,v 1.28 2009/01/01 07:46:43 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/poll_event_manager.h"

#include <cassert>
#include <cerrno>
#include <cstdlib>
#include <iostream>


#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#include <poll.h>
#include <pthread.h>

#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif


//#define ENABLE_OUTPUT_STATISTICS	1
//#define DEBUG_EVENT_LOOP		1


#define DEFAULT_CONNECTION_TIMEOUT_MILLISEC	(15 * 1000)	/* 15 */
#define DEFAULT_IO_TIMEOUT_MILLISEC		(30 * 1000)	/* 30 */
#define MINIMUM_TIMEOUT_MILLISEC		10		/* 10ms */


namespace Ochusha
{


static void
canonicalize_timeval(struct timeval &tv)
{
  tv.tv_sec += tv.tv_usec / 1000000;
  tv.tv_usec = tv.tv_usec % 1000000;
}


class PollEventManager::Event : public RefCount
{
private:
  smart_ptr<PollEventListener> listener;
  int fd;		// listenereventԤäƤfd

  PollEventManager::Priority priority;
  struct pollfd *result;
  int timeout;
  int effective_timeout;
  struct timeval timeout_time;

  short events;		// ԤäƤeventpollͿeventsͤΤΡ
  short pending_events;
  bool polled;
  bool previous_timeout_left;


public:
  inline PollEventListener *get_listener() { return listener; }
  inline int get_fd() const { return fd; }
  inline PollEventManager::Priority get_priority() const { return priority; }
  inline bool get_polled() const { return polled; }

  inline short get_events() const { return result->events; }
  inline short get_revents() const { return result->revents; }


  void debug_print() const
  {
    fprintf(stderr,
	    "Event(%p): fd=%d, events=0x%x, effective_timeout=%d, polled=%d",
	    this, fd, events, effective_timeout, polled);
    if (result != NULL)
      fprintf(stderr, ", result(%d, %x, %x)\n",
	      result->fd, result->events, result->revents);
    else
      fprintf(stderr, "\n");
  }


  inline void update_events(short new_events, struct timeval &current_time,
			    int interval)
  {
    if (interval == 0)
      {
	events = new_events;
	pending_events = 0;
	return;
      }

    events = 0;
    pending_events = new_events;

    previous_timeout_left = false;

    timeout_time.tv_sec = current_time.tv_sec;
    timeout_time.tv_usec = current_time.tv_usec + (interval * 1000);
    canonicalize_timeval(timeout_time);

#ifdef DEBUG_EVENT_LOOP
    fprintf(stderr,
	    "update_events: current_time is: { %lu, %lu }\n",
	    static_cast<unsigned long>(current_time.tv_sec),
	    static_cast<unsigned long>(current_time.tv_usec));
    fprintf(stderr,
	    "update_events: this(%p) will be wait for: { %lu, %lu }\n",
	    this,
	    static_cast<unsigned long>(timeout_time.tv_sec),
	    static_cast<unsigned long>(timeout_time.tv_usec));
#endif
  }


  inline void set_timeout(int new_timeout)
  {
    if (timeout <= 0)
      {
	if (timeout == PollEventManager::TIMEOUT_CONNECTION
	    || timeout == PollEventManager::TIMEOUT_IO)
	  effective_timeout = new_timeout;
      }
  }


  inline void update_timeout_time(struct timeval &current_time)
  {
    if (pending_events != 0)
      {
	assert(events == 0);
	if (timeout_time.tv_sec < current_time.tv_sec
	    || (timeout_time.tv_sec == current_time.tv_sec
		&& timeout_time.tv_usec <= current_time.tv_usec))
	  {
	    events = pending_events;
	    pending_events = 0;
#ifdef DEBUG_EVENT_LOOP
	    fprintf(stderr,
		    "update_timeout_time: waiting done for this(%p) at: { %lu, %lu }\n",
		    this,
		    static_cast<unsigned long>(current_time.tv_sec),
		    static_cast<unsigned long>(current_time.tv_usec));
#endif
	  }
	else
	  return;
      }

    if (effective_timeout <= 0 || previous_timeout_left)
      return;

    timeout_time.tv_sec = current_time.tv_sec;
    timeout_time.tv_usec = current_time.tv_usec + (effective_timeout * 1000);
    canonicalize_timeval(timeout_time);

#ifdef DEBUG_EVENT_LOOP
    fprintf(stderr,
	    "update_timeout_time: current_time is: { %lu, %lu }\n",
	    static_cast<unsigned long>(current_time.tv_sec),
	    static_cast<unsigned long>(current_time.tv_usec));
    fprintf(stderr,
	    "update_timeout_time: this(%p) will be wait for: { %lu, %lu }\n",
	    this,
	    static_cast<unsigned long>(timeout_time.tv_sec),
	    static_cast<unsigned long>(timeout_time.tv_usec));
#endif
  }


  void setup_pollfd(struct pollfd *fds,
		    struct timeval &closest_timeout,
		    struct timeval &current_time)
  {
    update_timeout_time(current_time);

    polled = true;

    fds->fd = fd;
    fds->events = events;
    fds->revents = 0;
    result = fds;

    if (effective_timeout > 0)
      {
	if (closest_timeout.tv_sec == 0)
	  {
	    closest_timeout.tv_sec = timeout_time.tv_sec;
	    closest_timeout.tv_usec = timeout_time.tv_usec;
	  }
	else if (timeout_time.tv_sec < closest_timeout.tv_sec
		 || (timeout_time.tv_sec == closest_timeout.tv_sec
		     && timeout_time.tv_usec < closest_timeout.tv_usec))
	  {
	    closest_timeout.tv_sec = timeout_time.tv_sec;
	    closest_timeout.tv_usec = timeout_time.tv_usec;
	  }
      }
#ifdef DEBUG_EVENT_LOOP
    fprintf(stderr,
	    "setup_pollfd(%p...): closest_timeout so far is at: { %lu, %lu }\n",
	    this,
	    static_cast<unsigned long>(closest_timeout.tv_sec),
	    static_cast<unsigned long>(closest_timeout.tv_usec));
#endif
  }


  inline bool is_event_occured() const
  { return result != NULL
      && (result->revents & (events | POLLERR | POLLHUP | POLLNVAL)) != 0; }


  inline bool is_timeout_expired(struct timeval &current_time)
  {
    if (effective_timeout <= 0 || !polled || pending_events != 0)
      return false;

    previous_timeout_left =
      !(timeout_time.tv_sec < current_time.tv_sec
	|| (timeout_time.tv_sec == current_time.tv_sec
	    && timeout_time.tv_usec <= current_time.tv_usec));

    return !previous_timeout_left;
  }


  inline Event(PollEventListener *listener, int fd, short events,
	       PollEventManager::Priority priority,
	       int timeout, int effective_timeout, int interval)
    : listener(listener), fd(fd), priority(priority),
      result(NULL), timeout(timeout), effective_timeout(effective_timeout),
      polled(false), previous_timeout_left(false)
  {
    if (effective_timeout <= 0)
      {
	timeout_time.tv_sec = 0;
	timeout_time.tv_usec = 0;
      }

    struct timeval current_time;
    gettimeofday(&current_time, NULL);

    set_timeout(effective_timeout);
    update_events(events, current_time, interval);
  }


  ~Event()
  {
#ifdef DEBUG_EVENT_LOOP
    cerr << "PollEventManager::Event destructed.\n";
#endif
  }
};


PollEventManager::PollEventManager()
  : lock("PollEventManager"),
    number_of_events(0),
    total_number_of_events(0),
    total_number_of_contention(0),
    timeout_connection(DEFAULT_CONNECTION_TIMEOUT_MILLISEC),
    timeout_io(DEFAULT_IO_TIMEOUT_MILLISEC)
{
  if (pipe(command_pipe) != 0)
    {
      cerr << "Couldn't create a pipe.\n";
      exit(EXIT_FAILURE);
    }

  fcntl(command_pipe[0], F_SETFL,
	fcntl(command_pipe[0], F_GETFL) | O_NONBLOCK);

  if (pthread_create(&event_loop_thread_id, NULL,
		     event_loop_thread_main, this) != 0)
    {
      cerr << "Couldn't create event loop thread.\n";
      exit(EXIT_FAILURE);
    }
}


PollEventManager::~PollEventManager()
{
  write(command_pipe[1], "DIENOW", 7);
  close(command_pipe[1]);

  if (pthread_join(event_loop_thread_id, NULL) != 0)
    cerr << "Couldn't join the event loop thread.\n";
#ifdef ENABLE_OUTPUT_STATISTICS
  fprintf(stderr, "PollEventManager: total-events=%lu, total-contention=%lu\n",
	  total_number_of_events, total_number_of_contention);
#endif
}


void
PollEventManager::event_loop()
{
  struct pollfd *fds = NULL;
  bool quit_event_loop = false;

  lock.lock();
  while (true)
    {
      if (fds != NULL)
	free(fds);

      unsigned int nfds = number_of_events + 1;
      fds = static_cast<struct pollfd *>(calloc(nfds, sizeof(struct pollfd)));
      if (fds == NULL)
	{
	  cerr << "PollEventManager: Out of Memory.\n";
	  exit(EXIT_FAILURE);	// Out of memoryǺƵǽ
	}

      fds[0].fd = command_pipe[0];
      fds[0].events = POLLIN | POLLPRI;

      struct timeval current_time;
      gettimeofday(&current_time, NULL);

#ifdef DEBUG_EVENT_LOOP
      fprintf(stderr,
	      "event_loop: current_time is: { %lu, %lu }\n",
	      static_cast<unsigned long>(current_time.tv_sec),
	      static_cast<unsigned long>(current_time.tv_usec));
#endif

      struct timeval closest_timeout = { 0, 0 };

      EventList::iterator iter = event_list.begin();
      EventList::iterator end_iter = event_list.end();
      struct pollfd *pollfd_element = fds;

      while (iter != end_iter)
	{
	  (*iter)->setup_pollfd(++pollfd_element,
				closest_timeout, current_time);
	  ++iter;
	}
      lock.unlock();

      int timeout;

      if (closest_timeout.tv_sec < current_time.tv_sec
	  || (closest_timeout.tv_sec == current_time.tv_sec
	      && closest_timeout.tv_usec < current_time.tv_usec))
	timeout = -1;
      else
	{
	  timeout = (closest_timeout.tv_sec - current_time.tv_sec) * 1000
	    + ((closest_timeout.tv_usec - current_time.tv_usec) / 1000);

	  assert(timeout >= 0);
	  if (timeout < MINIMUM_TIMEOUT_MILLISEC)
	    timeout = MINIMUM_TIMEOUT_MILLISEC;		// 
	}

#ifdef DEBUG_EVENT_LOOP
      fprintf(stderr,
	      "event_loop: closest_timeout is at: { %lu, %lu }, timeout=%d\n",
	      static_cast<unsigned long>(closest_timeout.tv_sec),
	      static_cast<unsigned long>(closest_timeout.tv_usec),
	      timeout);
#endif

      int n_events;
      do
	{
	  n_events = poll(fds, nfds, timeout);
	}
      while (n_events == -1 && errno == EINTR);

      // command_pipe[0]å
      bool terminate_all = false;
      if (fds[0].revents & fds[0].events)
	{
	  if (fds[0].revents & (POLLIN | POLLPRI))
	    {
	      char buf[7];
	      while (read(command_pipe[0], buf, 7) == 7)
		if (strcmp(buf, "DIENOW") == 0)
		  {
		    if (command_pipe[0] != -1)
		      {
			close(command_pipe[0]);
			command_pipe[0] = -1;
		      }
		    quit_event_loop = true;
		  }
		else if (strcmp(buf, "GIVEUP") == 0)
		  terminate_all = true;
	    }
	  else if (fds[0].revents & (POLLERR | POLLHUP | POLLNVAL))
	    {
	      quit_event_loop = true;
	    }
	}

      Priority priority = PRIORITY_LOW;
      gettimeofday(&current_time, NULL);

      lock.lock();
      EventList::iterator it = event_list.begin();
      while (it != event_list.end())
	{
	  Event *event = *it;

#ifdef DEBUG_EVENT_LOOP
	  event->debug_print();
#endif

	  if (quit_event_loop || terminate_all)
	    {
	      event->get_listener()->terminate(event->get_fd());
	      event_list.erase(it++);
	      number_of_events--;
	    }
	  else if (event->get_polled() && event->get_priority() <= priority
		   && (event->is_event_occured()
		       || event->is_timeout_expired(current_time)))
	    {
	      priority = event->get_priority();
	      int result
		= event->get_listener()->handle_event(event->get_fd(),
						      event->get_events(),
						      event->get_revents());
	      short next_req = GET_NEXT_EVENT_REQUEST(result);
	      int interval = GET_NEXT_EVENT_POLL_INTERVAL(result);

	      ++total_number_of_events;
	      if (result == EVENT_RESULT_CONTENTION)
		++total_number_of_contention;

	      if (next_req != 0)
		{
		  event->update_events(next_req, current_time, interval);
		  // timeout֤򾡼˽񤭴Τϥʥ
		  // EVENT_RESULT_CONTINUEʤꤹ٤
		  event->set_timeout(timeout_io);
		  it++;
		}
	      else
		{
		  event_list.erase(it++);
		  number_of_events--;
		}
	    }
	  else
	    {
#ifdef DEBUG_EVENT_LOOP
	      fprintf(stderr,
		      "polled=%d priority={%d, %d} is_event_occured=%d is_timeout_expired=%d\n",
		      event->get_polled(), event->get_priority(), priority,
		      event->is_event_occured(),
		      event->is_timeout_expired(current_time));
#endif
	      it++;
	    }
	}

      if (quit_event_loop)
	{
	  free(fds);
	  lock.unlock();
	  return;
	}
    }
}


void *
PollEventManager::event_loop_thread_main(void *arg)
{
#ifdef CHECK_LOCK_ORDER
  new ThreadInfo("<<PollEventManager>>");
#endif

  PollEventManager *manager = static_cast<PollEventManager *>(arg);
  manager->event_loop();
  return NULL;
}


void *
PollEventManager::request_poll(PollEventListener *listener,
			       int fd, short events,
			       Priority priority, int timeout)
{
  int effective_timeout;
  if (timeout == TIMEOUT_CONNECTION)
    effective_timeout = timeout_connection;
  else if (timeout == TIMEOUT_IO)
    effective_timeout = timeout_io;
  else if (timeout == TIMEOUT_NEVER)
    effective_timeout = -1;
  else
    effective_timeout = 0;

  EventHandle event = new Event(listener, fd, events, priority,
				timeout, effective_timeout, 0);

  lock.lock();
  number_of_events++;
  EventList::iterator it = event_list.begin();
  while (it != event_list.end())
    {
      if (priority < (*it)->get_priority())
	break;

      if ((*it)->get_fd() == fd)
	event_list.erase(it++);
      else
	it++;
    }

  event_list.insert(it, event);

  if (it != event_list.end())
    {
      it++;
      while (it != event_list.end())
	{
	  if ((*it)->get_fd() == fd)
	    event_list.erase(it++);
	  else
	    it++;
	}
    }

  write(command_pipe[1], "WAKEUP", 7);
  lock.unlock();
  return event;
}


void
PollEventManager::terminate_all()
{
  write(command_pipe[1], "GIVEUP", 7);
}


}
