/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: regexp.cc,v 1.20 2009/01/01 07:46:43 fuyu Exp $
 */

#include "config.h"

#include "oniguruma.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/regexp.h"

#include <iostream>
#include <cstddef>
#include <cstdlib>
#include <cstring>


namespace Ochusha
{

Regexp::Regexp(const char *pattern, const char *encoding)
  : regexp(NULL),
    region(NULL),
    last_match_text(NULL),
    regexp_available(false)
{
  if (pattern != NULL)
    this->compile(pattern, encoding);
}


Regexp::~Regexp()
{
  if (regexp != NULL)
    onig_free(regexp);

  if (region != NULL)
    onig_region_free(region, 1);
}


bool
Regexp::compile(const char *pattern, const char *encoding)
{
  OnigEncoding enc;

  regexp_available = false;
  last_match_text = NULL;

  if (pattern == NULL || *pattern == '\0')
    return regexp_available;

  if (encoding != NULL)
    {
      // ĥ뵤̵
      if (strcmp(encoding, "CP932") == 0)
	enc = ONIG_ENCODING_SJIS;
      else if (strcmp(encoding, "EUC-JP") == 0)
	enc = ONIG_ENCODING_EUC_JP;
      else if (strcmp(encoding, "Shift_JIS") == 0)
	enc = ONIG_ENCODING_SJIS;
      else
	enc = ONIG_ENCODING_UTF8;
    }
  else
    enc = ONIG_ENCODING_UTF8;

  if (regexp != NULL)
    {
      onig_free(regexp);
      regexp = NULL;
    }

  regexp_available
    = onig_new(&regexp,
	       reinterpret_cast<UChar *>(const_cast<char *>(pattern)),
	       reinterpret_cast<UChar *>(const_cast<char *>(pattern)) + strlen(pattern),
	       ONIG_OPTION_DEFAULT,
	       enc, ONIG_SYNTAX_RUBY, NULL) == ONIG_NORMAL;

  if (regexp_available)
    {
      if (region == NULL)
	region = onig_region_new();
      else
	onig_region_clear(region);
    }
#if 0
  else
    cerr << "Regexp compilation error: " << pattern << endl;
#endif

  return regexp_available;
}

  
bool
Regexp::match(const char *text, int len)
{
  if (text == NULL || !regexp_available)
    return false;

  size_t text_len = len >= 0 ? len : strlen(text);
  UChar *text_tail
    = reinterpret_cast<UChar *>(const_cast<char *>(text)) + text_len;
  int offset = onig_search(regexp,
			   reinterpret_cast<UChar *>(const_cast<char *>(text)),
			   text_tail,
			   reinterpret_cast<UChar *>(const_cast<char *>(text)),
			   text_tail, region, 0);
  if (offset == ONIG_MISMATCH)
    last_match_text = NULL;
  else
    last_match_text = text;

  return offset != ONIG_MISMATCH;
}


const char *
Regexp::match_begin(int group)
{
  if (last_match_text == NULL || region->num_regs <= group
      || region->beg[group] == ONIG_REGION_NOTPOS)
    return NULL;

  return last_match_text + region->beg[group];
}


const char *
Regexp::match_end(int group)
{
  if (last_match_text == NULL || region->num_regs <= group
      || region->end[group] == ONIG_REGION_NOTPOS)
    return NULL;

  return last_match_text + region->end[group];
}


size_t
Regexp::match_len(int group)
{
  if (last_match_text == NULL || region->num_regs <= group
      || region->beg[group] == ONIG_REGION_NOTPOS
      || region->end[group] == ONIG_REGION_NOTPOS)
    return 0;

  return region->end[group] - region->beg[group];
}

}


#ifdef REGEX_TEST
using namespace Ochusha;

int
main(int argc, char *argv[])
{
  Regexp board_pattern("<[Aa][[:space:]]+(?i:href)=([^ \">]+)[^>]*>([^<]+)</[Aa]>",
		       "CP932");
  if (board_pattern.match("<A HREF=http://news8.2ch.net/eq/>Rinji Jishin</A><br>\n"
			  "<A HREF=http://news8.2ch.net/eqplus/>Rinji Jishin+</A>\n", -1))
    {
      cerr << "matched.\n";
    }
  else
    {
      cerr << "not matched.\n";
    }

  Regexp subject_pattern("^([[:digit:]]+)\\.dat<>.*\\(([[:digit:]]+)\\)$",
			 "CP932");
  cerr << "pattern.is_valid(): " << subject_pattern.is_valid() << endl;

  const char *text =
    "1041737196.dat<>= Linux Games = General Thread Part2 (733)\n"
    "1082297069.dat<>Debian GNU/Linux Thread Ver. 23 (721)\n";
  if (subject_pattern.match(text, strchr(text, '\n') - text))
    {
      cerr << "matched: len=" << subject_pattern.match_len() << endl;
    }
  else
    {
      cerr << "not matched.\n";
    }
}
#endif
