/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: sqlite_utils.cc,v 1.2 2009/01/01 07:46:43 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/sqlite_utils.h"

#include <cassert>
#include <cstdlib>
#include <cstdio>


#define DEBUG_SQLITE3_USAGE	1


namespace Ochusha
{


SQLite3::SQLite3(const char *db_filename)
  : db_handle(NULL),
    query_result(NULL), number_of_rows(0), number_of_columns(0),
    error_message(NULL)
{
  assert(db_filename != NULL);

  int result = sqlite3_open(db_filename, &db_handle);
  if (result != SQLITE_OK)
    {
      if (db_handle != NULL)
	{
#ifdef DEBUG_SQLITE3_USAGE
	  fprintf(stderr, "sqlite3_open(\"%s\") failed: %d: %s\n",
		  db_filename, result, sqlite3_errmsg(db_handle));

#endif
	  sqlite3_close(db_handle);
	  db_handle = NULL;
	}
#ifdef DEBUG_SQLITE3_USAGE
      else
	fprintf(stderr, "sqlite3_open(\"%s\") failed: %d\n",
		db_filename, result);
#endif
      return;
    }
}


SQLite3::~SQLite3()
{
  if (db_handle != NULL)
    sqlite3_close(db_handle);
  db_handle = NULL;
}


void
SQLite3::reset_table()
{
  if (query_result != NULL)
    {
      sqlite3_free_table(query_result);
      query_result = NULL;
    }

  number_of_rows = 0;
  number_of_columns = 0;

  if (error_message != NULL)
    {
      sqlite3_free(error_message);
      error_message = NULL;
    }
}


bool
SQLite3:: get_table(const char *sql)
{
  if (db_handle == NULL)
    return false;

  assert(sql != NULL);

  reset_table();
  int result = sqlite3_get_table(db_handle, sql, &query_result, &number_of_rows,
				 &number_of_columns, &error_message);
  if (result == SQLITE_OK)
    return true;

#ifdef DEBUG_SQLITE3_USAGE
  if (error_message != NULL)
    fprintf(stderr, "sqlite3_get_table() failed for \"%s\": %d: %s\n",
	    sql, result, error_message);
  else
    fprintf(stderr, "sqlite3_get_table() failed for \"%s\": %d\n",
	    sql, result);
#endif

  return false;
}


SQLite3Statement::SQLite3Statement(SQLite3 &db_handle, const char *sql_templ)
  : db_handle(db_handle.get_db_handle()), statement(NULL)
{
  if (this->db_handle == NULL)
    return;

  int result = sqlite3_prepare_v2(this->db_handle, sql_templ, -1,
				  &statement, NULL);
  if (result != SQLITE_OK)
    {
#ifdef DEBUG_SQLITE3_USAGE
      fprintf(stderr,
	      "sqlite3_prepare_v2() for \"%s\" failed: %d: %s\n",
	      sql_templ, result, sqlite3_errmsg(this->db_handle));
#endif
    }
}


SQLite3Statement::~SQLite3Statement()
{
  if (statement != NULL)
    {
      sqlite3_finalize(statement);
      statement = NULL;
    }
}


bool
SQLite3Statement::bind(int index, int value)
{
  if (statement == NULL)
    return false;

  int result = sqlite3_bind_int(statement, index, value);
  if (result == SQLITE_OK)
    return true;

#ifdef DEBUG_SQLITE3_USAGE
  fprintf(stderr,
	  "SQLite3Statement::bind(%d, %d): sqlite3_bind_int() failed: %d: %s\n",
	  index, value, result, sqlite3_errmsg(db_handle));
#endif

  return false;
}


bool
SQLite3Statement::bind(int index, sqlite3_int64 value)
{
  if (statement == NULL)
    return false;

  int result = sqlite3_bind_int64(statement, index, value);
  if (result == SQLITE_OK)
    return true;

#ifdef DEBUG_SQLITE3_USAGE
  fprintf(stderr,
	  "SQLite3Statement::bind(%d, %lld): sqlite3_bind_int64() failed: %d: %s\n",
	  index, value, result, sqlite3_errmsg(db_handle));
#endif

  return false;
}


bool
SQLite3Statement::bind(int index, const char *value)
{
  if (statement == NULL)
    return false;

  if (value != NULL)
    {
      int result = sqlite3_bind_text(statement, index, value, -1,
				     SQLITE_TRANSIENT);
      if (result == SQLITE_OK)
	return true;
#ifdef DEBUG_SQLITE3_USAGE
      fprintf(stderr,
	      "SQLite3Statement::bind(%d, \"%s\"): sqlite3_bind_text() failed: %d: %s\n",
	      index, value, result, sqlite3_errmsg(db_handle));
#endif
      return false;
    }

  int result = sqlite3_bind_null(statement, index);
  if (result == SQLITE_OK)
    return true;

#ifdef DEBUG_SQLITE3_USAGE
      fprintf(stderr,
	      "SQLite3Statement::bind(%d, NULL): sqlite3_bind_null() failed: %d: %s\n",
	      index, result, sqlite3_errmsg(db_handle));
#endif

  return false;
}


bool
SQLite3Statement::bind_static(int index, const char *value)
{
  if (statement == NULL)
    return false;


  if (value != NULL)
    {
      int result = sqlite3_bind_text(statement, index, value, -1,
				     SQLITE_STATIC);
      if (result == SQLITE_OK)
	return true;
#ifdef DEBUG_SQLITE3_USAGE
      fprintf(stderr,
	      "SQLite3Statement::bind_static(%d, \"%s\"): sqlite3_bind_text() failed: %d: %s\n",
	      index, value, result, sqlite3_errmsg(db_handle));
#endif
      return false;
    }

  int result = sqlite3_bind_null(statement, index);
  if (result == SQLITE_OK)
    return true;

#ifdef DEBUG_SQLITE3_USAGE
      fprintf(stderr,
	      "SQLite3Statement::bind_static(%d, NULL): sqlite3_bind_null() failed: %d: %s\n",
	      index, result, sqlite3_errmsg(db_handle));
#endif

  return false;
}


int
SQLite3Statement::step()
{
  assert(statement != NULL);

  int result = sqlite3_step(statement);

#ifdef DEBUG_SQLITE3_USAGE
  if (result != SQLITE_ROW && result != SQLITE_DONE)
    {
      fprintf(stderr, "sqlite3_step() failed: %d: %s\n",
	      result, sqlite3_errmsg(db_handle));
    }
#endif

  return result;
}


}
