/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: text_utils.cc,v 1.24 2009/01/01 07:46:43 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/text_utils.h"

#include "libochusha/cp932_utils.h"
#include "libochusha/eucjp_utils.h"
#include "libochusha/misc_utils.h"

#include <cerrno>
#include <climits>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <string>

#include <iconv.h>

#include <libxml/xmlstring.h>


using namespace Ochusha;

namespace Ochusha
{


size_t
CStringBuffer::expand()
{
  size_t str_len = tail - buffer;
  size_t old_len = buffer_length;
  char *old_buf = buffer;
  buffer_length *= 2;
  if (buffer != default_buffer)
    {
      buffer = static_cast<char *>(realloc(buffer, buffer_length));
      if (buffer != NULL)
	{
	  tail = buffer + str_len;
	  return buffer_length - old_len;
	}
    }
  else
    {
      buffer = static_cast<char *>(malloc(buffer_length));
      if (buffer != NULL)
	{
	  memcpy(buffer, old_buf, old_len);
	  tail = buffer + str_len;
	  *tail = '\0';
	  return buffer_length - old_len;
	}
    }
  /* out of memory */
  buffer = old_buf;
  buffer_length = old_len;
  return buffer_length - old_len;
}


bool
CStringBuffer::ensure_capacity(size_t required_len)
{
  size_t old_len = buffer_length;
  size_t str_len = tail - buffer;

  if (old_len >= required_len)
    return true;

  while (buffer_length < required_len)
    buffer_length *= 2;

  char *old_buf = buffer;
  if (buffer != default_buffer)
    {
      buffer = static_cast<char *>(realloc(buffer, buffer_length));
      if (buffer != NULL)
	{
	  tail = buffer + str_len;
	  return true;
	}
    }
  else
    {
      buffer = static_cast<char *>(malloc(buffer_length));
      if (buffer != NULL)
	{
	  memcpy(buffer, old_buf, old_len);
	  tail = buffer + str_len;
	  *tail = '\0';
	  return true;
	}
    }
  /* out of memory */
  buffer = old_buf;
  buffer_length = old_len;
  return false;
}


void
CStringBuffer::append(const char *text, size_t length)
{
  if (length == 0)
    return;

  if (get_string_length() + length + 1 > get_buffer_length()
      && !ensure_capacity(get_string_length() + length + 1))
    {
      fprintf(stderr, "CStringBuffer::append(): out of memory\n");
      return;	// ȤꤢΤơ
    }
  memcpy(tail, text, length);
  tail += length;
  *tail = '\0';
}


const char * const IconvBuffer::CP932_ENCODING = "CP932";
const char * const IconvBuffer::EUCJP_ENCODING = "EUC-JP";
const char * const IconvBuffer::UTF8_ENCODING = "UTF-8";
const char * const IconvBuffer::UTF8_ENCODING_IGNORE = "UTF-8//IGNORE";


IconvBuffer::IconvBuffer(const char *to_encoding, const char *from_encoding)
{
  to_encoding = get_canonical_encoding_name(to_encoding);
  from_encoding = get_canonical_encoding_name(from_encoding);
  if (to_encoding == UTF8_ENCODING)
    {
      // helper뤫⤷ʤ
      if (from_encoding == CP932_ENCODING)
	{
	  converter = iconv_open(UTF8_ENCODING, CP932_ENCODING);
	  helper = cp932_to_utf8_helper;
	}
      else if (from_encoding == EUCJP_ENCODING)
	{
	  converter = iconv_open(UTF8_ENCODING, EUCJP_ENCODING);
	  helper = eucjp_to_utf8_helper;
	}
      else
	{
	  // helperʤ켺ԤʸƦѤhelperѰդ
	  // ⤷ʤŪˤCP932EUC-JPɬ׽ʬ
	  converter = iconv_open(UTF8_ENCODING_IGNORE, from_encoding);
	  helper = NULL;
	}
    }
  else if (to_encoding == from_encoding
	   || strcasecmp(to_encoding, from_encoding) == 0)
    {
      converter = NULL;
      helper = NULL;
    }
  else
    {
      // UTF-8CP932UTF-8EUC-JPˤϤhelperѰդĤ
      // ΤȤϤʤΤǤĤIGNORE
      // ͤǼĤؤν񤭹ߤΤѴξ硢ʸλ
      // ˤƤޤȤ⤢ꡢhelperȤѰդĤ
      // 򤳤Ȥ߹ΤϾ㤤äݤΤǡΤ
      // helperŪͿ륳󥹥ȥ饯Ѱդ롣
      char to[256];
      if (strstr(to_encoding, "//IGNORE") == NULL)
	{
	  snprintf(to, sizeof(to), "%s//IGNORE", to_encoding);
	  to_encoding = to;
	}
      converter = iconv_open(to_encoding, from_encoding);
      helper = NULL;
    }
  if (converter == reinterpret_cast<iconv_t>(-1))
    {
      converter = NULL;	// Ѵʤ⡼ɤˤƤ
      fprintf(stderr, "iconv_open(\"%s\", \"%s\") failed\n",
	      to_encoding, from_encoding);
    }
}


IconvBuffer::IconvBuffer(const char *to_encoding, const char *from_encoding,
			 IconvHelperFunc helper)
{
  converter = iconv_open(to_encoding, from_encoding);
  helper = helper;
  if (converter == reinterpret_cast<iconv_t>(-1))
    {
      converter = NULL;	// Ѵʤ⡼ɤˤƤ
      fprintf(stderr, "iconv_open(\"%s\", \"%s\") failed\n",
	      to_encoding, from_encoding);
    }
}


IconvBuffer::~IconvBuffer()
{
  if (converter != NULL)
    iconv_close(converter);
}


void
IconvBuffer::append(const char *text, size_t length)
{
  if (length == 0)
    return;
  if (converter == NULL)
    {
      CStringBuffer::append(text, length);
      return;
    }

  ICONV_CONST char *inbuf = const_cast<ICONV_CONST char *>(text);
  size_t outbytesleft = buffer_length - (tail - buffer) - 1;
  size_t inbytesleft = length;

  while (true)
    {
      size_t size;
      do
	{
	  size = iconv(converter,
		       &inbuf, &inbytesleft, &tail, &outbytesleft);
	  *tail = '\0';	// outbytesleftǽ餫1餷Ƥ뤫
	  if (size != static_cast<size_t>(-1))
	    return;	// ｪλ

	  if (errno == EILSEQ && helper != NULL)
	    size = (*helper)(const_cast<const char **>(&inbuf), &inbytesleft,
			     &tail, &outbytesleft);
	  *tail = '\0';	// outbytesleftǽ餫1餷Ƥ뤫
	}
      while (size != static_cast<size_t>(-1));

      if (errno == E2BIG)
	{
	  size_t extra = expand();
	  if (extra == 0)
	    return ;	// out of Memory
	  outbytesleft += extra;
	}
      else
	{
#if DEBUG_ICONV
	  fprintf(stderr, "iconv failed(errno = %d)\n", errno);
	  if (inbytesleft >= 3)
	    fprintf(stderr, "character code = 0x%02x 0x%02x 0x%02x ...\n",
		    static_cast<int>(inbuf[0]),
		    static_cast<int>(inbuf[1]),
		    static_cast<int>(inbuf[2]));
	  else if (inbytesleft == 2) 
	    fprintf(stderr, "character code = 0x%02x 0x%02x\n",
		    static_cast<int>(inbuf[0]), static_cast<int>(inbuf[1]));
	  else
	    fprintf(stderr, "character code = 0x%02x\n",
		    static_cast<int>(inbuf[0]));
#endif
	  return;	// Ѵ
	}
    }
}

}


/*
 * get_canonical_encoding_name:
 *   Ϳ줿encodingʸencoding̾ȸʤƲϤΤäƤ
 *   encoding̾ǤˤlibochushaŪintern줿encodingʸ
 *   ؤΥݥ󥿤֤
 *   ΤʤΤä顢Ϳ줿encoding򤽤Τޤ֤
 */
const char *
get_canonical_encoding_name(const char *encoding)
{
  if (encoding == IconvBuffer::UTF8_ENCODING)
    return encoding;
  if (encoding == IconvBuffer::CP932_ENCODING)
    return encoding;
  if (encoding == IconvBuffer::EUCJP_ENCODING)
    return encoding;

  if (strcasecmp(encoding, "UTF-8") == 0)
    return IconvBuffer::UTF8_ENCODING;
  if (strcasecmp(encoding, IconvBuffer::CP932_ENCODING) == 0
      || strcasecmp(encoding, "Shift_JIS") == 0
      || strcasecmp(encoding, "Shift-JIS") == 0)
    return IconvBuffer::CP932_ENCODING;
  if (strcasecmp(encoding, IconvBuffer::EUCJP_ENCODING) == 0)
    return IconvBuffer::EUCJP_ENCODING;
  return encoding;	// Τʤencoding
}


/*
 * get_encoding_from_content_type:
 *   Ϳ줿content_typeHTTPContent-Type쥹ݥ󥹥إå
 *   ȸʤƲϤΤäƤcharsetޤޤƤ餽λ
 *   libochushaŪencoding֤̾
 *   charsetꤵƤʤΤʤΤǤˤNULL
 *   ֤
 */
const char *
get_encoding_from_content_type(const char *content_type)
{
  if (content_type == NULL)
    return NULL;	// ɤƤcall siteǥåϤ
  			// ʤΤNULL

  const char *encoding = NULL;
  const char *charset = strcasestr(content_type, "charset=");
  if (charset != NULL)
    {
      charset += 8;
      if (*charset == '"' || *charset == '\'')
	++charset;
      const char *tail = strpbrk(charset, " ;\"'\n\r\t");
      size_t len;
      if (tail != NULL)
	len = tail - charset;
      else
	len = strlen(charset);

      if (strncasecmp(charset, "Shift_JIS", len) == 0
	  || strncasecmp(charset, "Shift-JIS", len) == 0
	  || strncasecmp(charset, "X-Shift_JIS", len) == 0
	  || strncasecmp(charset, "X-Shift-JIS", len) == 0)
	encoding = IconvBuffer::CP932_ENCODING;
      else if (strncasecmp(charset, "EUC-JP", len) == 0
	       || strncasecmp(charset, "X-EUC-JP", len) == 0)
	encoding = IconvBuffer::EUCJP_ENCODING;
      else if (strncasecmp(charset, "UTF-8", len) == 0
	       || strncasecmp(charset, "X-UTF-8", len) == 0)
	encoding = IconvBuffer::UTF8_ENCODING;
    }
  return encoding;
}


/*
 * wipe_string:
 *   UTF-8ʸstring˴ޤޤnon-printableʸƬ
 *   ʸΥԡ֤
 */
char *
wipe_string(const char *text)
{
  size_t length = strlen(text);

  const unsigned char *head = reinterpret_cast<const unsigned char *>(text);
  const unsigned char *tail = head + length;

  // ޥХȤʶɤФ褦ˤ٤ɤ̯ʤȤ
  // ǤפäݤΤ1ХȤʶ' 'Τߤθ롣

  // Ƭnon-printableʸӶɤФ
  while (head < tail && *head <= 0x20)
    ++head;

  // non-printableʸӶɤФ
  while (head < tail && tail[-1] <= 0x20)
    --tail;

  // non-printableʸȤʬ̵̸뤬
  // ʤΤǵˤʤ
  char *result = static_cast<char *>(malloc(tail - head + 1));
  if (result == NULL)
    return NULL;	// Out of Memory

  int i = 0;
  while (head < tail)
    {
      unsigned char c = *head++;
      if (c >= 0x20)	// printable
	result[i++] = c;
    }
  result[i] = '\0';

  return result;
}


// TODO: ̾ɤʤ
// *text_pUTF-8ʸȸʤ1ʸɤ߼ꡢUCS-2()֤
static int
eat_unicode_char(const unsigned char **text_p, size_t *len_p) {
  if (text_p == NULL || *text_p == NULL)
    return -1;
  if (len_p == NULL)
    return -1;
  if (*len_p < 1)
    return -1;

  const unsigned char *text = *text_p;
  size_t len = *len_p;

  unsigned int c = text[0];
  if (c & 0x80)
    {
      if (len < 2)
	return -1;
      if ((text[1] & 0xc0) != 0x80)
	return -1;
      if ((c & 0xe0) == 0xe0)
	{
	  if (len < 3)
	    return -1;
	  if ((text[2] & 0xc0) != 0x80)
	    return -1;
	  if ((c & 0xf0) == 0xf0)
	    {
	      if (len < 4)
		return -1;
	      if ((c & 0xf8) != 0xf0 || (text[3] & 0xc0) != 0x80)
		return -1;
	      /* 4Хʸ */
	      c = (text[0] & 0x7) << 18;
	      c |= (text[1] & 0x3f) << 12;
	      c |= (text[2] & 0x3f) << 6;
	      c |= text[3] & 0x3f;
	      *len_p = len - 4;
	      *text_p = text + 4;
	    }
	  else
	    {
	      /* 3Хʸ */
	      c = (text[0] & 0xf) << 12;
	      c |= (text[1] & 0x3f) << 6;
	      c |= text[2] & 0x3f;
	      *len_p = len - 3;
	      *text_p = text + 3;
	    }
	}
      else
	{
	  /* 2Хʸ */
	  c = (text[0] & 0x1f) << 6;
	  c |= text[1] & 0x3f;
	  *len_p = len - 2;
	  *text_p = text + 2;
	}
    }
  else
    {
      *len_p = len - 1;
      *text_p = text + 1;
    }

  return c;
}


int
parse_int(const char *src, size_t len)
{
  int result = 0;
  size_t tmp_len = len;
  const unsigned char *tmp_pos = reinterpret_cast<const unsigned char *>(src);

  do
    {
      int c = eat_unicode_char(&tmp_pos, &tmp_len);
      if (c >= '0' && c <= '9')
	result = result * 10 + c - '0';
      else if (c >= 65296 && c <= 65305)	// ''unicode65296
	result = result * 10 + c - 65296;	// ''unicode65305
      else
	break;
    }
  while (tmp_len > 0);

  return result;
}
