/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: thread_pool.cc,v 1.19 2009/01/01 07:46:43 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/thread_pool.h"

#include "libochusha/misc_utils.h"

#include <algorithm>
#include <cstdlib>
#include <pthread.h>


// #define DEBUG_THREAD	1
// #define DEBUG_THREAD_MORE 1


namespace Ochusha
{

#ifdef CHECK_LOCK_ORDER
class WorkerThreadInfo : public ThreadInfo
{
private:
  WorkerThread *self;


public:
  WorkerThreadInfo(WorkerThread *self)
    : ThreadInfo("WorkerThread"), self(self) {}
  virtual void dump_info();
};


void
WorkerThreadInfo::dump_info()
{
  fprintf(stderr, "WorkerThreadInfo() {\n");

  const WorkerJob *job = self->get_current_job();
  if (job != NULL)
    fprintf(stderr, "  current_job: %s\n", job->description());
  else
    fprintf(stderr, "  current_job:\n");

  dump_info_body();

  fprintf(stderr, "}\n");
}
#endif


WorkerThread::WorkerThread(ThreadPool &pool, WorkerJob *job)
  : pool(pool),
    state(job == NULL ? RUNNABLE : DETACHED),
    current_job(job)
{
  int result = pthread_create(&id, NULL, worker_thread_main, this);
  if (result == 0)
    {
      if (job != NULL)
	{
	  result = pthread_detach(id);
	  if (result != 0)
	    goto error_exit;
	}

      return;
    }

 error_exit:
  ErrorMessage error_message(result);
  fprintf(stderr, LIB_("Couldn't create a WorkerThread: %s\n"),
	  error_message.get_message());

  state = DEAD;
}


void *
WorkerThread::worker_thread_main(void *arg)
{
  WorkerThread *self = static_cast<WorkerThread *>(arg);

#ifdef CHECK_LOCK_ORDER
  new WorkerThreadInfo(self);
#endif

  self->pool.monitor.enter();

  if (self->state == DETACHED)
    {
      assert(self->current_job != NULL);

      self->state = BUSY;
      self->pool.monitor.exit();
#if DEBUG_THREAD_MORE
      cerr << "Starting one-shot Job: "
	   << self->current_job->description() << endl;
#endif
      self->current_job->work(self);

#if DEBUG_THREAD_MORE
      cerr << "Finished one-shot Job: "
	   << self->current_job->description() << endl;
#endif

      delete self;

      return NULL;
    }

  while (true)
    {
      if (self->state == DIE_NOW)
	{
	  self->state = DEAD;
	  self->pool.number_of_threads--;
	  self->pool.monitor.exit();

	  return NULL;
	}

      if (self->state == RUNNABLE && !self->pool.pending_jobs.empty())
	{
	  self->current_job = self->pool.pending_jobs.front();
	  self->pool.pending_jobs.pop_front();
	  self->state = BUSY;
	  self->pool.monitor.exit();

#if DEBUG_THREAD_MORE
	  cerr << "Starting Job: " << self->current_job->description() << endl;
#endif
	  self->current_job->work(self);

#if DEBUG_THREAD_MORE
	  cerr << "Finished Job: " << self->current_job->description() << endl;
#endif
	  /*
	   * jobreferenceΥ֥åsmart_pträ硢
	   * Υ֥åΥפ򳰤줿destructorư
	   * Ȥˤʤ롣destructor(pool.monitorʳ)餫Υå
	   * 褦ʤΤȡåʤʤΤǤ
	   * ŪjobreferenceõƤξ硢destructor
	   * ¹ԻˤThreadPool/WorkerThread˴طåϲ
	   * ƤʤΤǡ¿
	   */
	  self->current_job = NULL;
	  self->pool.monitor.enter();
	  if (self->state == BUSY)
	    self->state = RUNNABLE;
	  continue;
	}

      if (self->state == RUNNABLE)
	{
	  self->pool.number_of_idle_threads++;
	  self->pool.monitor.wait();
	  self->pool.number_of_idle_threads--;
	}
    }
}


bool
WorkerThread::join()
{
  int result = pthread_join(id, NULL);
  if (result == 0)
    return true;

  ErrorMessage error_message(result);
  fprintf(stderr, LIB_("Couldn't Join a WorkerThread(%p): %s\n"),
	  this, error_message.get_message());
  return false;
}


ThreadPool::ThreadPool(unsigned int initial_num_threads,
		       unsigned int maximum_num_threads)
  : monitor("ThreadPool"),
    maximum_number_of_threads(maximum_num_threads)
{
  number_of_threads = 0;
  number_of_idle_threads = 0;

  LockBlock lock_obj(monitor);
  for (unsigned int i = 0; i < initial_num_threads; i++)
    {
      if (!employ_new_worker())
	break;
    }
}


struct JoinThreadFunc
{
  inline void operator()(WorkerThread *worker)
  { if (worker->join()) delete worker; }
};


ThreadPool::~ThreadPool()
{
  monitor.enter();
  maximum_number_of_threads = 0;
  number_of_threads = 0;
  cancel_all();
  terminate_workers();
  monitor.exit();

  for_each(worker_threads.begin(), worker_threads.end(), JoinThreadFunc());
}


bool
ThreadPool::employ_new_worker()
{
#ifdef DEBUG_THREAD
  fprintf(stderr, "employ_new_worker(): #threads=%d, max#threads=%d\n",
	  number_of_threads, maximum_number_of_threads);
#endif
  if (number_of_threads < maximum_number_of_threads)
    {
      WorkerThread *worker = new WorkerThread(*this);
      if (worker->get_state() != WorkerThread::DEAD)
	{
	  number_of_threads++;
	  worker_threads.push_back(worker);
	  return true;
	}

      delete worker;
    }

  return false;
}


bool
ThreadPool::commit(WorkerJob *job, bool force)
{
  smart_ptr<WorkerJob> new_job = job;

  LockBlock lock_obj(monitor);
  if (force)
    {
      WorkerThread * worker = new WorkerThread(*this, job);
      if (worker->get_state() == WorkerThread::DETACHED)
	return true;

      delete worker;
      return false;
    }

  pending_jobs.push_back(job);

  bool result;
  if (number_of_idle_threads > 0)
    result = true;
  else
    result = employ_new_worker();
  monitor.notify();

  return result;
}


void
ThreadPool::cancel_all()
{
  monitor.enter();
  while (!pending_jobs.empty())
    {
      smart_ptr<WorkerJob> job = pending_jobs.front();
      pending_jobs.pop_front();
      monitor.exit();
      job = NULL;
      monitor.enter();
    }
  monitor.exit();
}


struct SetDieNowFunc
{
  inline void operator()(WorkerThread *worker)
  { worker->state = WorkerThread::DIE_NOW; }
};


void
ThreadPool::terminate_workers()
{
  LockBlock lock_obj(monitor);
  for_each(worker_threads.begin(), worker_threads.end(), SetDieNowFunc());
  monitor.notify_all();
}

}


#ifdef TP_TEST

using namespace Ochusha;

class TestJob : public WorkerJob
{
private:
  int id;

public:
  TestJob(int id) : id(id) {}
  ~TestJob()
  {
    cerr << "TestJob(id=" << id << ", @" << this << ") deleted.\n";
  }

  void work(const WorkerThread *employee)
  {
    cerr << "Hello, I'm thread " << employee << endl;
    monitor.enter();
    count++;
    monitor.notify();
    monitor.exit();
    while (employee->get_state() != WorkerThread::DIE_NOW);
  }

  static Monitor monitor;
  static int count;
};


Monitor TestJob::monitor;
int TestJob::count = 0;


int
main(int argc, char *argv[])
{
  ThreadPool pool(1, 3);

  TestJob::monitor.enter();
  WorkerJob *job = new TestJob(1);
  smart_ptr<WorkerJob> first_job = job;
  cerr << "commit: " << pool.commit(job) << endl;
  job = new TestJob(2);
  cerr << "commit: " << pool.commit(job) << endl;
  job = new TestJob(3);
  cerr << "commit: " << pool.commit(job) << endl;

  while (TestJob::count < 3)
    TestJob::monitor.wait();
  TestJob::monitor.exit();
  cerr << "Now all jobs should be done.\n";
}
#endif
