/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: tree.cc,v 1.23 2009/01/01 07:46:43 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/tree.h"

#include "libochusha/misc_utils.h"
#include "libochusha/sax_context.h"

#include <algorithm>
#include <iostream>
#include <string>
#include <vector>

#include <cstring>
#include <zlib.h>


namespace Ochusha
{

#define OUTPUT_ATTRIBUTE_BOOLEAN(gzfile, attribute, ind)		\
  do									\
    {									\
      if (attribute)							\
	gzprintf(gzfile,						\
		 "%s<attribute name=\"" #attribute "\">\n"		\
		 "%s  <boolean val=\"true\"/>\n"			\
		 "%s</attribute>\n", ind, ind, ind);			\
    }									\
  while (0)

#define OUTPUT_ATTRIBUTE_INT(gzfile, attribute, ind)			\
  do									\
    {									\
      if (attribute)							\
	gzprintf(gzfile,						\
		 "%s<attribute name=\"" #attribute "\">\n"		\
		 "%s  <int val=\"%d\"/>\n"				\
		 "%s</attribute>\n", ind, ind, attribute, ind);		\
    }									\
  while (0)

#define OUTPUT_ATTRIBUTE_LONG(gzfile, attribute, ind)			\
  do									\
    {									\
      if (attribute)							\
	gzprintf(gzfile,						\
		 "%s<attribute name=\"" #attribute "\">\n"		\
		 "%s  <long val=\"%ld\"/>\n"				\
		 "%s</attribute>\n", ind, ind,				\
		 static_cast<long>(attribute), ind);			\
    }									\
  while (0)

#define OUTPUT_ATTRIBUTE_STRING(gzfile, attribute, ind)			\
  do									\
    {									\
      if (attribute != NULL)						\
	{								\
	  string text;							\
	  gzprintf(gzfile,						\
		   "%s<attribute name=\"" #attribute "\">\n"		\
		   "%s  <string>%s</string>\n"				\
		   "%s</attribute>\n", ind, ind,			\
		   append_markup_escaped_text(text, attribute).c_str(),	\
		   ind);						\
	}								\
    }									\
  while (0)


TreeElement::TreeElement(const char *name, time_t tc, time_t last_mod)
  : name(NULL),
    description(NULL),
    hidden(false),
    alone(true),
    root_folder(NULL)
{
  set_name(name, true);
  if (last_mod != 0)
    time_of_last_modification = last_mod;
  else
    time_of_last_modification = time(NULL);

  if (tc != 0)
    time_of_creation = tc;
  else
    time_of_creation = time_of_last_modification;
}


TreeElement::TreeElement(AttributeTable &attributes, time_t tc, time_t last_mod)
  : name(NULL),
    description(NULL),
    hidden(false),
    alone(true),
    root_folder(NULL)
{
  const char *str = attributes.get_string("name");
  set_name(str != NULL ? str : "", true);
  set_description(attributes.get_string("description"), true);
  set_hidden(attributes.get_bool("hidden"));

  if (last_mod != 0)
    time_of_last_modification = last_mod;
  else
    time_of_last_modification = time(NULL);

  if (tc != 0)
    time_of_creation = tc;
  else
    time_of_creation = time_of_last_modification;
}


TreeElement::~TreeElement()
{
  if (name != NULL)
    free(name);

  if (description != NULL)
    free(description);
}


bool
TreeElement::is_separator() const
{
  return false;
}


void
TreeElement::notify_modified()
{
  time_of_last_modification = time(NULL);
  if (root_folder != this && root_folder != NULL)
    root_folder->notify_modified();
}


void
TreeElement::set_alone(bool value)
{
  alone = value;
  if (!is_folder())
    return;

  Folder *folder = dynamic_cast<Folder *>(this);
  assert(folder != NULL);

  TreeElementList::iterator iter = folder->elements.begin();
  TreeElementList::iterator end_iter = folder->elements.end();
  while (iter != end_iter)
    {
      TreeElement *element = *iter++;
      element->set_alone(value);
    }
}


RootFolder *
TreeElement::get_root_folder() const
{
  return dynamic_cast<RootFolder *>(root_folder);
}


void
TreeElement::set_root_folder(RootFolder *folder)
{
  if (root_folder == folder || folder == NULL)
    return;

  assert(root_folder == NULL);
  root_folder = folder;
}


void
TreeElement::set_name(const char *value, bool internal)
{
  if (value == name
      || (value != NULL && name != NULL && strcmp(value, name) == 0))
    return;

  if (name != NULL)
    free(name);

  name = (value != NULL && *value != '\0') ? strdup(value) : NULL;
  if (!internal)
    notify_modified();
}


void
TreeElement::set_description(const char *value, bool internal)
{
  if (value == description
      || (value != NULL && description != NULL
	  && strcmp(value, description) == 0))
    return;

  if (description != NULL)
    free(description);

  description = (value != NULL && *value != '\0') ? strdup(value) : NULL;
  if (!internal)
    notify_modified();
}


void
TreeElement::read_tree_element(AttributeTable &attributes)
{
  set_name(attributes.get_string("name"), true);
  set_description(attributes.get_string("description"), true);
  hidden = attributes.get_bool("hidden");

  read_extended_tree_element(attributes);
}


void
TreeElement::write_tree_element(WriteTree &args)
{
  const char *name = get_name();
  if (name == NULL)
    return;

  string *current_indent = args.indent;
  gzFile tree_xml = args.tree_xml;

  if (!is_folder())
    gzprintf(tree_xml,
	     "%s<%s tc=\"%lu\" tlm=\"%lu\">\n",
	     current_indent->c_str(), get_element_name(),
	     static_cast<unsigned long>(time_of_creation),
	     static_cast<unsigned long>(time_of_last_modification));
  else
    {
      Folder *folder = dynamic_cast<Folder *>(this);
      gzprintf(tree_xml,
	       "%s<%s tc=\"%lu\" tlm=\"%lu\" sn=\"%u\">\n",
	       current_indent->c_str(), get_element_name(),
	       static_cast<unsigned long>(time_of_creation),
	       static_cast<unsigned long>(time_of_last_modification),
	       folder->get_serial_number());
    }

  string new_indent = *current_indent + "  ";
  const char *indent = new_indent.c_str();

  OUTPUT_ATTRIBUTE_STRING(tree_xml, name, indent);
  OUTPUT_ATTRIBUTE_STRING(tree_xml, description, indent);
  OUTPUT_ATTRIBUTE_BOOLEAN(tree_xml, hidden, indent);

  args.indent = &new_indent;
  write_extended_tree_element(args);
  args.indent = current_indent;

  gzprintf(tree_xml, "%s</%s>\n", current_indent->c_str(), get_element_name());
  args.indent = current_indent;
}


Folder::Folder(int type, const char *name,
	       time_t tc, time_t last_mod, unsigned int sn)
  : TreeElement(name, tc, last_mod),
    type(type),
    expanded(false),
    serial_number(sn)
{
}


Folder::Folder(const Folder *folder, unsigned int sn)
  : TreeElement(folder->get_name(), 0, 0),
    type(folder->get_folder_type()),
    expanded(folder->is_expanded()),
    serial_number(sn)
{
  set_root_folder(folder->get_root_folder());
  set_description(folder->get_description());
  set_hidden(folder->is_hidden());
}


void
Folder::set_folder_type(int value)
{
  if (type == value)
    return;

  type = value;
  notify_modified();
}


void
Folder::set_expanded(bool value)
{
  if (expanded == value)
    return;

  expanded = value;
  notify_modified();
}


void
Folder::prepend(TreeElement *element, TreeElement *next)
{
  if (element == NULL)
    return;

  assert(element->is_alone() || !element->is_folder());

  element->set_root_folder(get_root_folder());

  if (next == NULL)
    elements.insert(elements.begin(), element);
  else
    {
      TreeElementList::iterator where
	= find(elements.begin(), elements.end(), next);
      if (where != elements.end())
	elements.insert(where, element);
      else
	elements.insert(elements.begin(), element);
    }

  element->set_alone(false);

  notify_modified();
}


void
Folder::append(TreeElement *element, TreeElement *prev)
{
  if (element == NULL)
    return;

  assert(element->is_alone() || !element->is_folder());

  element->set_root_folder(get_root_folder());

  if (prev == NULL)
    {
#if 0
      cerr << "Folder::append(" << element << ", NULL)\n";
#endif
      elements.push_back(element);
    }
  else
    {
      TreeElementList::iterator where
	= find(elements.begin(), elements.end(), prev);
      if (where != elements.end())
	elements.insert(++where, element);
      else
	elements.push_back(element);
    }

  element->set_alone(false);

  notify_modified();
}


void
Folder::remove(TreeElement *element)
{
  if (element == NULL)
    return;

  TreeElementList::iterator where
    = find(elements.begin(), elements.end(), element);
  if (where != elements.end())
    {
      TreeElementHandle elm = element;
      elements.erase(where);

      element->set_alone(true);

      notify_modified();
    }
}


bool
Folder::is_folder() const
{
  return true;
}


const char *
Folder::get_element_name() const
{
  return "folder";
}


void
Folder::read_extended_tree_element(AttributeTable &attributes)
{
  type = attributes.get_int("type");
#if 1	/* COMPAT: ߴΤ */
  if (type == 0)
    type = attributes.get_int("category_type");
#endif
  expanded = attributes.get_bool("expanded");
}


void
Folder::write_extended_tree_element(WriteTree &args)
{
  string *current_indent = args.indent;
  gzFile tree_xml = args.tree_xml;

  const char *indent = current_indent->c_str();
  OUTPUT_ATTRIBUTE_INT(tree_xml, type, indent);
  OUTPUT_ATTRIBUTE_BOOLEAN(tree_xml, expanded, indent);

  string new_indent = *current_indent + "  ";
  args.indent = &new_indent;
  for_each(elements.begin(), elements.end(), args);

  args.indent = current_indent;
}


RootFolder::RootFolder(const char *name, time_t tc, time_t last_mod)
  : Folder(FOLDER_TYPE_ROOT_FOLDER, name, tc, last_mod, 0),
    last_serial_number(0),
    save_time(0)
{
  set_root_folder(this);
}


Folder *
RootFolder::folder_new_internal(time_t tc, time_t last_mod, unsigned int sn)
{
  if (sn > last_serial_number)
    last_serial_number = sn;

  Folder *folder
    = folder_new(tc, last_mod, sn != 0 ? sn : ++last_serial_number);
  folder->set_root_folder(this);
  return folder;
}


Folder *
RootFolder::folder_new()
{
  Folder *folder = folder_new(0, 0, ++last_serial_number);
  folder->set_root_folder(this);
  return folder;
}


Folder *
RootFolder::copy(const Folder *folder)
{
  assert(folder != NULL);
  assert(folder != this);
  assert(folder->get_root_folder() == this);

  Folder *cloned_folder = folder_new(folder, ++last_serial_number);
  TreeElementList::const_iterator iter = folder->elements_begin();
  TreeElementList::const_iterator end_iter = folder->elements_end();
  while (iter != end_iter)
    {
      TreeElement *element = *iter++;
      TreeElement *cloned_element;
      if (element->is_folder())
	cloned_element = copy(dynamic_cast<Folder *>(element));
      else
	cloned_element = copy(element);

      cloned_folder->append(cloned_element);
    }

  return cloned_folder;
}


class TreeXMLContext : public SAXContext
{
private:
  enum SAXState
  {
    SAX_INITIAL,
    SAX_OCHUSHA,
    SAX_ROOT_FOLDER,
    SAX_FOLDER,
    SAX_FOLDER_ATTRIBUTE,
    SAX_FOLDER_ATTRIBUTE_BOOLEAN,
    SAX_FOLDER_ATTRIBUTE_INT,
    SAX_FOLDER_ATTRIBUTE_STRING,
    SAX_FOLDER_LEAF,
    SAX_FOLDER_LEAF_ATTRIBUTE,
    SAX_FOLDER_LEAF_ATTRIBUTE_BOOLEAN,
    SAX_FOLDER_LEAF_ATTRIBUTE_INT,
    SAX_FOLDER_LEAF_ATTRIBUTE_STRING,
    SAX_ACCEPTED,
    SAX_ERROR
  } state;

  time_t time_of_read_start;
  RootFolder *root_folder;
  const char *root_element_name;
  const char *leaf_element_name;
  time_t leaf_time_of_creation;
  time_t leaf_time_of_last_modification;

  struct FolderContext
  {
    Folder *folder;
    AttributeTable attrs;

    FolderContext(Folder *folder)
    {
      this->folder = folder;
    }
  };

  vector<FolderContext *> folder_stack;
  FolderContext *current_folder_context;

  AttributeTable leaf_attrs;


protected:
  void start_element_handler(const char *name, const char **attrs);
  void end_element_handler(const char *name);
  void characters_handler(const char *ch, size_t len);


public:
  TreeXMLContext(RootFolder *root_folder)
    : state(SAX_INITIAL), root_folder(root_folder)
  {
    time_of_read_start = time(NULL);
    current_folder_context = new FolderContext(root_folder);
    root_element_name = root_folder->get_root_element_name();
    leaf_element_name = root_folder->get_leaf_element_name();
  }
  ~TreeXMLContext()
  {
    while (!folder_stack.empty())
      pop_folder();
    delete current_folder_context;
  }

  inline void push_folder(Folder *folder)
  {
    folder_stack.push_back(current_folder_context);
    current_folder_context = new FolderContext(folder);
  }

  inline void pop_folder()
  {
    if (folder_stack.empty())
      return;	// rootեϺǸޤǻĤƤ
    delete current_folder_context;
    current_folder_context = folder_stack.back();
    folder_stack.pop_back();
  }

  inline Folder *get_current_folder()
  { return current_folder_context->folder; }

  inline AttributeTable &get_current_folder_attrs()
  { return current_folder_context->attrs; }

  inline bool is_accepted() { return state == SAX_ACCEPTED; }
};


void
TreeXMLContext::start_element_handler(const char *name, const char **attrs)
{
  switch (state)
    {
    case SAX_INITIAL:
      if (strcmp(name, "ochusha") == 0) { state = SAX_OCHUSHA; return; }
      break;	// 顼

    case SAX_OCHUSHA:
      if (strcmp(name, root_element_name) == 0)
	{
	  unsigned long tc = 0;
	  unsigned long tlm = 0;
	  unsigned int lsn = 0;
	  if (attrs != NULL)
	    {
	      const char **cur_attrs = attrs;
	      while (*cur_attrs != NULL)
		{
		  if (cur_attrs[1] == NULL)
		    break;
		  if (strcmp(cur_attrs[0], "tc") == 0)
		    {
		      if (sscanf(cur_attrs[1], "%lu", &tc) != 1)
			tc = 0;
		    }
		  if (strcmp(cur_attrs[0], "tlm") == 0
		      || strcmp(cur_attrs[0],
				"time_of_last_modification") == 0)
		    {
		      if (sscanf(cur_attrs[1], "%lu", &tlm) != 1)
			tlm = 0;
		    }
		  else if (strcmp(cur_attrs[0], "lsn") == 0
			   || strcmp(cur_attrs[0], "last_serial_number") == 0)
		    {
		      if (sscanf(cur_attrs[1], "%u", &lsn) != 1)
			lsn = 0;
		    }
		  cur_attrs += 2;
		}
	    }

	  time_t tmp_time;
	  if (tlm != 0)
	    tmp_time = static_cast<time_t>(tlm);
	  else
	    tmp_time = time_of_read_start;

	  root_folder->time_of_last_modification = tmp_time;

	  if (tc != 0)
	    tmp_time = static_cast<time_t>(tc);
	  root_folder->time_of_creation = tc;

	  root_folder->last_serial_number = lsn;

	  state = SAX_ROOT_FOLDER;
	  return;
	}
      break;	// 顼

    case SAX_ROOT_FOLDER:
    case SAX_FOLDER:
    case SAX_FOLDER_LEAF:
      if (strcmp(name, "attribute") == 0
	  && attrs != NULL && strcmp(attrs[0], "name") == 0)
	{
	  switch (state)
	    {
	    case SAX_ROOT_FOLDER:
	    case SAX_FOLDER:
	      state = SAX_FOLDER_ATTRIBUTE;
	      get_current_folder_attrs().set_attr_name(attrs[1]);
	      return;

	    case SAX_FOLDER_LEAF:
	      state = SAX_FOLDER_LEAF_ATTRIBUTE;
	      leaf_attrs.set_attr_name(attrs[1]);
	      return;

	    default:
	      fprintf(stderr, "Wrong implementation found.\n");
	      abort();
	    }
	}
      else if (state == SAX_FOLDER || state == SAX_ROOT_FOLDER)
	{
	  if (strcmp(name, leaf_element_name) == 0)
	    {
	      unsigned long tc = 0;
	      unsigned long tlm = 0;
	      if (attrs != NULL)
		{
		  const char **cur_attrs = attrs;
		  while (*cur_attrs != NULL)
		    {
		      if (cur_attrs[1] == NULL)
			break;
		      if (strcmp(cur_attrs[0], "tc") == 0)
			{
			  if (sscanf(cur_attrs[1], "%lu", &tc) != 1)
			    tc = 0;
			}
		      else if (strcmp(cur_attrs[0], "tlm") == 0
			       || strcmp(cur_attrs[0],
					 "time_of_last_modification") == 0)
			{
			  if (sscanf(cur_attrs[1], "%lu", &tlm) != 1)
			    tlm = 0;
			}
		      cur_attrs += 2;
		    }
		}

	      if (tlm != 0)
		leaf_time_of_last_modification = static_cast<time_t>(tlm);
	      else
		leaf_time_of_last_modification = time_of_read_start;

	      if (tc != 0)
		leaf_time_of_creation = static_cast<time_t>(tc);
	      else
		leaf_time_of_creation = leaf_time_of_last_modification;

	      state = SAX_FOLDER_LEAF; return;
	    }
#if 1	// COMPAT: ߴΤ
	  else if (strcmp(name, "folder") == 0
		   || strcmp(name, "category") == 0)
#else
	  else if (strcmp(name, "folder") == 0)
#endif
	    {
	      unsigned long tc = 0;
	      unsigned long tlm = 0;
	      unsigned int sn = 0;
	      if (attrs != NULL)
		{
		  const char **cur_attrs = attrs;
		  while (*cur_attrs != NULL)
		    {
		      if (cur_attrs[1] == NULL)
			break;
		      if (strcmp(cur_attrs[0], "tc") == 0)
			{
			  if (sscanf(cur_attrs[1], "%lu", &tc) != 1)
			    tc = 0;
			}
		      else if (strcmp(cur_attrs[0], "tlm") == 0
			       || strcmp(cur_attrs[0],
					 "time_of_last_modification") == 0)
			{
			  if (sscanf(cur_attrs[1], "%lu", &tlm) != 1)
			    tlm = 0;
			}
		      else if (strcmp(cur_attrs[0], "sn") == 0
			       || strcmp(cur_attrs[0], "serial_number") == 0)
			{
			  if (sscanf(cur_attrs[1], "%u", &sn) != 1)
			    sn = 0;
			}
		      cur_attrs += 2;
		    }
		}

	      time_t last_mod;
	      if (tlm != 0)
		last_mod = static_cast<time_t>(tlm);
	      else
		last_mod = time_of_read_start;

	      time_t time_creation;
	      if (tc != 0)
		time_creation = static_cast<time_t>(tc);
	      else
		time_creation = last_mod;

	      state = SAX_FOLDER;
	      push_folder(root_folder->folder_new_internal(time_creation,
							   last_mod, sn));
	      return;
	    }
	}
      break;	// 顼


    case SAX_FOLDER_ATTRIBUTE:
    case SAX_FOLDER_LEAF_ATTRIBUTE:
      if (attrs != NULL && strcmp(attrs[0], "val") == 0)
	{
	  // int/booleanβǽ
	  if (strcmp(name, "int") == 0)
	    {
	      switch (state)
		{
		case SAX_FOLDER_ATTRIBUTE:
		  get_current_folder_attrs().set_attr_value(attrs[1]);
		  state = SAX_FOLDER_ATTRIBUTE_INT;
		  return;

		case SAX_FOLDER_LEAF_ATTRIBUTE:
		  state = SAX_FOLDER_LEAF_ATTRIBUTE_INT;
		  leaf_attrs.set_attr_value(attrs[1]);
		  return;

		default:
		  cerr << "Wrong implementation found.\n";
		  abort();
		}
	    }
	  if (strcmp(name, "boolean") == 0)
	    {
	      switch (state)
		{
		case SAX_FOLDER_ATTRIBUTE:
		  state = SAX_FOLDER_ATTRIBUTE_BOOLEAN;
		  get_current_folder_attrs().set_attr_value(attrs[1]);
		  return;

		case SAX_FOLDER_LEAF_ATTRIBUTE:
		  state = SAX_FOLDER_LEAF_ATTRIBUTE_BOOLEAN;
		  leaf_attrs.set_attr_value(attrs[1]);
		  return;

		default:
		  cerr << "Wrong implementation found.\n";
		  abort();
		}
	    }

#if DEBUG_SAX_HANDLER
	  cerr << "element unexpected in state(" << state << ") found: "
	       << name << endl;
#endif
	  break;	// 顼
	}
      else if (strcmp(name, "string") == 0)
	{
	  switch (state)
	    {
	    case SAX_FOLDER_ATTRIBUTE:
	      state = SAX_FOLDER_ATTRIBUTE_STRING;
	      return;

	    case SAX_FOLDER_LEAF_ATTRIBUTE:
	      state = SAX_FOLDER_LEAF_ATTRIBUTE_STRING;
	      return;

	    default:
	      break;	// 顼
	    }
#if DEBUG_SAX_HANDLER
	  cerr << "string element is unexpected in state(" << state << ")\n";
#endif
	}
      break;	// 顼

    case SAX_FOLDER_ATTRIBUTE_STRING:
    case SAX_FOLDER_ATTRIBUTE_BOOLEAN:
    case SAX_FOLDER_LEAF_ATTRIBUTE_STRING:
    case SAX_FOLDER_LEAF_ATTRIBUTE_INT:
    case SAX_ACCEPTED:
    case SAX_ERROR:
      break;	// 顼

    default:
      cerr << "start_element_handler is called in unknown state: "
	   << state << endl;
    }
  state = SAX_ERROR;
}


void
TreeXMLContext::end_element_handler(const char *name)
{
  switch (state)
    {
    case SAX_OCHUSHA:
      if (strcmp(name, "ochusha") == 0) { state = SAX_ACCEPTED; return; }
      break;	// 顼

    case SAX_ROOT_FOLDER:
      if (strcmp(name, root_element_name) == 0)
	{ state = SAX_OCHUSHA; return; }
      break;	// 顼

    case SAX_FOLDER:
#if 1	// COMPAT: ߴΤ
      if (strcmp(name, "folder") == 0 || strcmp(name, "category") == 0)
#else
      if (strcmp(name, "folder") == 0)
#endif
	{
	  Folder *folder = get_current_folder();
	  folder->read_tree_element(get_current_folder_attrs());
	  pop_folder();

	  // XMLե뤫ɤ߹ˤϥॹפѤäߤ
	  // ʤΤǥ
	  time_t last_mod = get_current_folder()->time_of_last_modification;
	  get_current_folder()->append(folder);
	  get_current_folder()->time_of_last_modification = last_mod;

	  if (folder_stack.empty())
	    state = SAX_ROOT_FOLDER;
	  return;
	}
      break;	// 顼

    case SAX_FOLDER_ATTRIBUTE:
      if (strcmp(name, "attribute") == 0)
	{
	  get_current_folder_attrs().register_current_attribute();
	  if (folder_stack.empty())
	    state = SAX_ROOT_FOLDER;
	  else
	    state = SAX_FOLDER;
	  return;
	}
      break;	// 顼

    case SAX_FOLDER_LEAF_ATTRIBUTE:
      if (strcmp(name, "attribute") == 0)
	{
	  leaf_attrs.register_current_attribute();
	  state = SAX_FOLDER_LEAF;
	  return;
	}
      break;	// 顼

    case SAX_FOLDER_ATTRIBUTE_STRING:
      if (strcmp(name, "string") == 0)
	{ state = SAX_FOLDER_ATTRIBUTE; return; }
      break;	// 顼

    case SAX_FOLDER_LEAF_ATTRIBUTE_STRING:
      if (strcmp(name, "string") == 0)
	{ state = SAX_FOLDER_LEAF_ATTRIBUTE; return; }
      break;	// 顼

    case SAX_FOLDER_ATTRIBUTE_BOOLEAN:
      if (strcmp(name, "boolean") == 0)
	{ state = SAX_FOLDER_ATTRIBUTE; return; }
      break;	// 顼

    case SAX_FOLDER_ATTRIBUTE_INT:
      if (strcmp(name, "int") == 0) { state = SAX_FOLDER_ATTRIBUTE; return; }
      break;	// 顼

    case SAX_FOLDER_LEAF:
      if (strcmp(name, leaf_element_name) == 0)
	{
	  TreeElement *leaf
	    = root_folder->leaf_element_new(leaf_attrs,
					    leaf_time_of_creation,
					    leaf_time_of_last_modification);
	  if (leaf != NULL)
	    {
	      // XMLե뤫ɤ߹ˤϥॹפѤä
	      // ߤʤΤǥ
	      time_t last_mod = get_current_folder()->time_of_last_modification;
	      TreeElementHandle elm = leaf;
	      get_current_folder()->remove(leaf);
	      get_current_folder()->append(leaf);
	      get_current_folder()->time_of_last_modification = last_mod;
	    }

	  leaf_attrs.clear();
	  if (folder_stack.empty())
	    state = SAX_ROOT_FOLDER;
	  else
	    state = SAX_FOLDER;
	  return;
	}
      break;	// 顼

    case SAX_FOLDER_LEAF_ATTRIBUTE_BOOLEAN:
      if (strcmp(name, "boolean") == 0)
	{ state = SAX_FOLDER_LEAF_ATTRIBUTE; return; }
      break;	// 顼

    case SAX_FOLDER_LEAF_ATTRIBUTE_INT:
      if (strcmp(name, "int") == 0)
	{ state = SAX_FOLDER_LEAF_ATTRIBUTE; return; }
      break;	// 顼

    case SAX_INITIAL:
    case SAX_ACCEPTED:
    case SAX_ERROR:
#if DEBUG_SAX_HANDLER
      cerr << "end_element_handler called in invalid state: " << state << "\n";
#endif
      break;	/* 顼 */
    }
#if DEBUG_SAX_HANDLER
  cerr << "Invalid document: </" << name << "> appeared in state=" << state
       << "\n";
#endif
  state = SAX_ERROR;
}


void
TreeXMLContext::characters_handler(const char *ch, size_t len)
{
  if (state == SAX_FOLDER_ATTRIBUTE_STRING)
    get_current_folder_attrs().append_attr_value(ch, len);
  else if (state == SAX_FOLDER_LEAF_ATTRIBUTE_STRING)
    leaf_attrs.append_attr_value(ch, len);
}


bool
RootFolder::read_tree_xml(Repository &repository, const char *filename)
{
  char *pathname = repository.find_file(filename);
  if (pathname == NULL)
    return false;	// ե뤬¸ߤʤ

  TreeXMLContext context(this);
  context.parse_xml_file(pathname);
  free(pathname);

  if (context.is_accepted())
    {
      tree_updated(this);

      save_time = get_time_of_last_modification();

      return true;
    }

  return false;
}


bool
RootFolder::write_tree_xml(Repository &repository, const char *filename)
{
  int fd = repository.open_file(filename, O_WRONLY | O_TRUNC | O_CREAT);
  if (fd < 0)
    {
      cerr << "Couldn't open " << filename << " to write.\n";
      return false;
    }

  gzFile tree_xml = gzdopen(fd, "w");
  if (tree_xml == NULL)
    {
      cerr << "Couldn't gzdopen for " << filename << endl;
      close(fd);
      return false;
    }

  gzprintf(tree_xml, "<?xml version=\"1.0\"?>\n");
  gzprintf(tree_xml, "<ochusha>\n");
  gzprintf(tree_xml,
	   "  <%s tc=\"%lu\" tlm=\"%lu\" lsn=\"%u\">\n",
	   get_root_element_name(),
	   static_cast<unsigned long>(get_time_of_creation()),
	   static_cast<unsigned long>(get_time_of_last_modification()),
	   last_serial_number);

  string indent = "    ";
  for_each(elements.begin(), elements.end(), WriteTree(tree_xml, &indent));

  gzprintf(tree_xml, "  </%s>\n", get_root_element_name());
  gzprintf(tree_xml, "</ochusha>\n");

  gzclose(tree_xml);

  save_time = time(NULL);

  return true;
}


}
