/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: tree.h,v 1.23 2009/01/01 07:46:43 fuyu Exp $
 */

#ifndef _LIBOCHUSHA_TREE_H_
#define _LIBOCHUSHA_TREE_H_

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef __cplusplus

#if defined(__GNUG__) && !defined(__ICC)
#pragma interface
#endif

#include "libochusha/hashtable.h"
#include "libochusha/ref_count.h"
#include "libochusha/repository.h"
#include "libochusha/sigslot.h"

#include <list>
#include <string>

#if defined(HAVE_TIME_H) || !defined(HAVE_CONFIG_H)
#include <time.h>
#endif

#include <zlib.h>


using namespace std;
namespace Ochusha
{

  class Folder;
  class RootFolder;
  class TreeElement;
  class TreeXMLContext;

  struct WriteTree
  {
    gzFile tree_xml;
    string *indent;

    WriteTree(gzFile file, string *indent)
      : tree_xml(file), indent(indent) {}

    inline void operator()(TreeElement *element);
  };


  class TreeElement : public RefCount
  {
    friend class Folder;
    friend class TreeXMLContext;

  private:
    char *name;
    char *description;
    bool hidden;	/* UIѤġ */
    bool alone;

    time_t time_of_creation;
    time_t time_of_last_modification;

    Folder *root_folder;


    void set_name(const char *value, bool internal);
    void set_description(const char *value, bool internal);


  protected:
    void notify_modified();
    void set_alone(bool value);


  public:
    TreeElement(const char *name, time_t tc, time_t last_mod);
    TreeElement(AttributeTable &attributes, time_t tc, time_t last_mod);
    virtual ~TreeElement();

    inline const char *get_name() const { return name; }
    inline void set_name(const char *value) { set_name(value, false); }

    inline const char *get_description() const { return description; }
    inline void set_description(const char *value)
    { set_description(value, false); }

    inline time_t get_time_of_creation() const { return time_of_creation; }
    inline time_t get_time_of_last_modification() const
    { return time_of_last_modification; }

    RootFolder *get_root_folder() const;
    void set_root_folder(RootFolder *folder);

    virtual bool is_folder() const = 0;
    virtual bool is_separator() const;	// for future extension.
    inline void show() { hidden = false; }
    inline bool is_shown() const { return !hidden; }
    inline void hide() { hidden = true; }
    inline bool is_hidden() const { return hidden; }
    inline void set_hidden(bool value) { hidden = value; }
    inline bool is_alone() const { return alone; }

    virtual const char *get_element_name() const = 0;
    virtual void read_tree_element(AttributeTable &attributes);
    virtual void write_tree_element(WriteTree &args);
    virtual void read_extended_tree_element(AttributeTable &attributes) = 0;
    virtual void write_extended_tree_element(WriteTree &args) = 0;
  };


  typedef smart_ptr<TreeElement> TreeElementHandle;
  typedef list<TreeElementHandle> TreeElementList;


  inline void WriteTree::operator()(TreeElement *element)
  { element->write_tree_element(*this); }


  enum {
    FOLDER_TYPE_UNKNOWN = -2,
    FOLDER_TYPE_ROOT_FOLDER = -1,
    FOLDER_TYPE_BBSMENU_CATEGORY = 0,
    FOLDER_TYPE_OCHUSHA_MY_FOLDER,
    FOLDER_TYPE_USER_FOLDER,
    FOLDER_TYPE_BOOKMARKS_FOLDER,
    FOLDER_TYPE_OCHUSHA_MY_BOOKMARK,
    FOLDER_TYPE_PSEUDO_FOLDER,
    FOLDER_TYPE_OABF_FOLDER
  };


  class Folder : public TreeElement
  {
    friend class TreeElement;

  private:
    int type;		/* ե */
    bool expanded;	/* UIѤġ */

    unsigned int serial_number;


  protected:
    TreeElementList elements;


  public:
    Folder(int type, const char *name,
	   time_t tc, time_t last_mod, unsigned int sn);
    Folder(const Folder *folder, unsigned int sn);

    void set_folder_type(int value);
    inline int get_folder_type() const { return type; }

    void set_expanded(bool value);
    inline bool is_expanded() const { return expanded; }

    inline unsigned int get_serial_number() const { return serial_number; }

    inline TreeElementList &get_elements()
    { return elements; }

    inline TreeElementList::const_iterator elements_begin() const
    { return elements.begin(); }

    inline TreeElementList::const_iterator elements_end() const
    { return elements.end(); }

    virtual void prepend(TreeElement *element,
			 TreeElement *next = NULL);
    virtual void append(TreeElement *element,
			TreeElement *prev = NULL);
    virtual void remove(TreeElement *element);

    virtual bool is_folder() const;
    bool is_empty() const { return elements.empty(); }
    
    virtual const char *get_element_name() const;
    virtual void read_extended_tree_element(AttributeTable &attributes);
    virtual void write_extended_tree_element(WriteTree &args);
  };


  typedef smart_ptr<Folder> FolderHandle;


  class RootFolder : public Folder
  {
    friend class TreeXMLContext;
  private:
    unsigned int last_serial_number;
    time_t save_time;


    Folder *folder_new_internal(time_t tc, time_t last_mod, unsigned int sn);


  protected:
    virtual Folder *folder_new(time_t tc, time_t last_mod, unsigned int sn) = 0;
    virtual Folder *folder_new(const Folder *folder, unsigned int sn) = 0;
    virtual TreeElement *leaf_element_new(AttributeTable &attributes,
					  time_t tc, time_t last_mod) = 0;


  public:
    RootFolder(const char *name = NULL, time_t tc = 0, time_t last_mod = 0);

    Folder *folder_new();
    virtual Folder *copy(const Folder *folder);
    virtual TreeElement *copy(const TreeElement *element) = 0;

    virtual const char *get_root_element_name() const = 0;
    virtual const char *get_leaf_element_name() const = 0;
    virtual bool write_tree_xml(Repository &repository, const char *filename);
    virtual bool read_tree_xml(Repository &repository, const char *filename);

    inline bool is_dirty() const
    { return save_time < get_time_of_last_modification(); }

    // void tree_updated(this);
    signal1<RootFolder *> tree_updated;
  };


  typedef smart_ptr<RootFolder> RootFolderHandle;


}


#endif	/* __cplusplus */

#endif	/* _LIBOCHUSHA_TREE_H_ */
