/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: uri_utils.cc,v 1.19 2009/01/01 07:46:43 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/uri_utils.h"

#include "libochusha/text_utils.h"
#include "libochusha/misc_utils.h"

#include <string>

#include <cassert>
#include <cstring>
#include <ctype.h>


namespace Ochusha
{

URI::URI(const char *uri)
  : full_uri(NULL),
    xml_uri(NULL)
{
  if (uri != NULL)
    parse_uri(uri);
}


URI::URI(const URI &uri)
  : full_uri(NULL),
    xml_uri(NULL)
{
  if (uri.is_valid())
    parse_uri(uri.get_uri());
}


URI::~URI()
{
  if (xml_uri != NULL)
    xmlFreeURI(xml_uri);
  if (full_uri != NULL)
    free(full_uri);
}


bool
URI::parse_uri(const char *uri)
{
  if (uri != NULL && full_uri != NULL && strcmp(uri, full_uri) == 0)
    return true;

  if (xml_uri != NULL)
    xmlFreeURI(xml_uri);

  if (full_uri != NULL)
    free(full_uri);

  if (uri != NULL && *uri != '\0')
    xml_uri = xmlParseURI(uri);
  else
    xml_uri = NULL;

  if (xml_uri != NULL)
    full_uri = strdup(uri);
  else
    full_uri = NULL;

  return full_uri != NULL;
}


char *
QueryKeyTuple::get_name() const
{
  return strndup(name, name_len);
}


char *
QueryKeyTuple::get_value() const
{
  return strndup(value, value_len);
}


QueryParser::QueryParser(const char *query)
{
  query_pos = query;
  query_key.name = NULL;
  query_key.name_len = 0;
  query_key.value = NULL;
  query_key.value_len = 0;
}


QueryParser::QueryParser(const URI &uri)
{
  query_pos = uri.get_query();
  query_key.name = NULL;
  query_key.name_len = 0;
  query_key.value = NULL;
  query_key.value_len = 0;
}


QueryParser::QueryParser(const URI *uri)
{
  query_pos = uri->get_query();
  query_key.name = NULL;
  query_key.name_len = 0;
  query_key.value = NULL;
  query_key.value_len = 0;
}


const QueryKeyTuple *
QueryParser::next_key()
{
  const char *name_pos = query_pos;
  if (name_pos == NULL)
    return NULL;

  const char *tmp_pos;
  while ((tmp_pos = strpbrk(name_pos, "&=")) != NULL)
    {
      if (*tmp_pos == '&')
	{
	  name_pos = tmp_pos + 1;
	  continue;
	}

      query_key.name = name_pos;
      query_key.name_len = tmp_pos - name_pos;

      const char *term = strchr(tmp_pos + 1, '&');
      query_pos = term;

      if (term == NULL)
	term = tmp_pos + strlen(tmp_pos);
      /* λǡ*term'&''\0' */

      query_key.value = tmp_pos + 1;
      query_key.value_len = term - tmp_pos - 1;

      return &query_key;
    }

  query_pos = NULL;
  return NULL;
}


char *
URI::encode_string(const char *src)
{
  if (src == NULL)
    return NULL;

  CStringBuffer buffer;

  const unsigned char *src_pos = reinterpret_cast<const unsigned char *>(src);
  size_t result_len = 0;
  unsigned char c;

  while ((c = *src_pos) != '\0')
    {
      if (isalnum(c)
	  || c == '*' || c == '-' || c == '.' || c == '@' || c == '_')
	{
	  if (buffer.ensure_capacity(result_len + 1))
	    buffer.get_buffer()[result_len++] = c;
	  else
	    return NULL;	// Out of Memory
	}
      else if (c == ' ')
	{
	  if (buffer.ensure_capacity(result_len + 1))
	    buffer.get_buffer()[result_len++] = '+';
	  else
	    return NULL;	// Out of Memory
	}
      else
	{
	  if (buffer.ensure_capacity((result_len + 3) + 1))
	    result_len += snprintf(buffer.get_buffer() + result_len, 4,
				   "%%%02X", c);
	  else
	    return NULL;	// Out of Memory
	}
      src_pos++;
    }

  return strndup(buffer.get_buffer(), result_len);
}


char *
URI::decode_string(const char *src)
{
  if (src == NULL)
    return NULL;

  CStringBuffer buffer;

  const char *cur_src = src;
  size_t dst_len = 0;

  while (*cur_src != '\0')
    {
      if (!buffer.ensure_capacity(cur_src - src + 1))
	{
	  return NULL;
	}
      char *cur_buf = buffer.get_buffer();
      if (*cur_src != '%')
	{
	  if (*cur_src == '+')
	    cur_buf[dst_len++] = ' ';
	  else
	    cur_buf[dst_len++] = *cur_src;
	  cur_src++;
	}
      else
	{
	  unsigned int code;
	  int result = sscanf(cur_src, "%%%02x", &code);
	  if (result == 1)
	    {
	      cur_buf[dst_len++] = static_cast<char>(code);
	      cur_src += 3;
	    }
	  else
	    cur_buf[dst_len++] = *cur_src++;
	}
    }

  return strndup(buffer.get_buffer(), dst_len);
}


BBSURLInfo::BBSURLInfo(const char *url, int board_id_depth)
  : server(NULL),
    bbs_root(NULL),
    bbs_dir(NULL),
    board_id(NULL),
    thread_id(NULL),
    from(0),
    to(0)
{
  URI uri(url);
  if (uri.is_valid())
    parse(uri, board_id_depth);
}


BBSURLInfo::BBSURLInfo(const URI &uri, int board_id_depth)
  : server(NULL),
    bbs_root(NULL),
    bbs_dir(NULL),
    board_id(NULL),
    thread_id(NULL),
    from(0),
    to(0)
{
  if (uri.is_valid())
    parse(uri, board_id_depth);
}


BBSURLInfo::~BBSURLInfo()
{
  if (server != NULL)
    free(server);

  if (bbs_root != NULL)
    free(bbs_root);

  if (bbs_dir != NULL)
    free(bbs_dir);

  if (board_id != NULL)
    free(board_id);

  if (thread_id != NULL)
    free(thread_id);
}


void
BBSURLInfo::set_server(const char *value, size_t len)
{
  if (server != NULL)
    free(server);

  server = (value != NULL && *value != '\0') ? strndup(value, len) : NULL;
}


// '/'⤪kˤСencode_string()Ūʤ
static char *
encode_string_hack(const char *src, size_t len)
{
  if (src == NULL)
    return NULL;

  CStringBuffer buffer;

  const unsigned char *src_pos = reinterpret_cast<const unsigned char *>(src);
  const unsigned char *tail_pos = src_pos + len;
  size_t result_len = 0;

  while (src_pos < tail_pos)
    {
      unsigned char c = *src_pos++;
      if (isalnum(c)
	  || c == '*' || c == '-' || c == '.' || c == '@' || c == '_' || c == '/')
	{
	  if (buffer.ensure_capacity(result_len + 1))
	    buffer.get_buffer()[result_len++] = c;
	  else
	    return NULL;	// Out of Memory
	}
      else if (c == ' ')
	{
	  if (buffer.ensure_capacity(result_len + 1))
	    buffer.get_buffer()[result_len++] = '+';
	  else
	    return NULL;	// Out of Memory
	}
      else
	{
	  if (buffer.ensure_capacity((result_len + 3) + 1))
	    result_len += snprintf(buffer.get_buffer() + result_len, 4,
				   "%%%02X", c);
	  else
	    return NULL;	// Out of Memory
	}
    }

  return strndup(buffer.get_buffer(), result_len);
}


void
BBSURLInfo::set_bbs_root(const char *value, size_t len)
{
  if (bbs_root != NULL)
    free(bbs_root);

  bbs_root = (value != NULL && *value != '\0')
    ? encode_string_hack(value, len) : NULL;
}


void
BBSURLInfo::set_bbs_dir(const char *value, size_t len)
{
  if (bbs_dir != NULL)
    free(bbs_dir);

  bbs_dir = (value != NULL && *value != '\0')
    ? encode_string_hack(value, len) : NULL;
}


void
BBSURLInfo::set_board_id(const char *value, size_t len)
{
  if (board_id != NULL)
    free(board_id);

  board_id = (value != NULL && *value != '\0')
    ? encode_string_hack(value, len) : NULL;
}


void
BBSURLInfo::set_thread_id(const char *value, size_t len)
{
  if (thread_id != NULL)
    free(thread_id);

  thread_id = (value != NULL && *value != '\0')
    ? encode_string_hack(value, len) : NULL;
}


void
BBSURLInfo::parse(const URI &uri, int board_id_depth)
{
  assert(uri.is_valid());
  assert(board_id_depth > 0);

  if (uri.get_server() != NULL)
    set_server(uri.get_server(), strlen(uri.get_server()));
  else
    set_server(NULL, 0);

  const char *path = uri.get_path();
  const char *cgi = strstr(path, "/read.cgi");
  size_t cgi_len = 0;
  if (cgi != NULL)
    cgi_len = 9;
  else
    {
      cgi = strstr(path, "/read.pl");
      if (cgi != NULL)
	cgi_len = 8;
      else
	{
	  cgi = strstr(path, "/read.html");
	  if (cgi != NULL)
	    cgi_len = 10;
	  else
	    {
	      cgi = strstr(path, "/read.so");
	      if (cgi != NULL)
		cgi_len = 8;
	    }
	}
    }

  if (cgi != NULL)
    {
      // URLäݤ
      if (*path != '/')
	{
	  // ХѥäݤΤǺǽslashޤɤФ
	  path = strchr(path, '/');
	  assert(path != NULL);
	}

      const char *slash_pos = cgi;
      size_t len = 0;
      if (slash_pos != path)
	while ((--slash_pos)[0] != '/')
	  ++len;

      set_bbs_dir(cgi - len, len + 1);
      if (path == uri.get_path())
	{
	  // Хѥǡpathcgi - lenμޤǤbbs_root
	  assert(cgi - len - path > 0);
	  set_bbs_root(path, cgi - len - path);
	}

      if (uri.get_query() == NULL)
	{
	  // PathInfoȸʤƲ
	  path = cgi + cgi_len;
	  const char *slash_pos = path;
	  if (*path != '/')
	    {
	      if (*path == '\0')
		{
		  const char *frag = uri.get_fragment();
		  if (frag != NULL)
		    {
		      // javascriptѡ
		      // http://dso.2ch.net/test/read.cgi#myanmar/1155409133/
		      // ߤۡ
		      slash_pos = strchr(frag, '/');
		      if (slash_pos != NULL && slash_pos != frag)
			{
			  path = frag - 1;	// 
			  goto parse_board_id;
			}
		    }
		}
	      return;	// 
	    }

	  for (int i = 0; i < board_id_depth; i++)
	    {
	      slash_pos = strchr(slash_pos + 1, '/');
	      if (slash_pos == NULL)
		break;
	    }

	parse_board_id:
	  const char *id_pos = path + 1;
	  if (slash_pos == NULL || slash_pos == id_pos)
	    return;	// 

	  set_board_id(id_pos, slash_pos - id_pos);

	  id_pos = slash_pos + 1;
	  slash_pos = strchr(id_pos, '/');
	  if (slash_pos != NULL && id_pos != slash_pos)
	    {
	      // IDȴԤƤʬߤǤǽ̵
	      set_thread_id(id_pos, slash_pos - id_pos);

	      if (slash_pos[1] != '\0')
		{
		  if (slash_pos[1] != 'l')
		    {
		      unsigned int f;
		      unsigned int t;
		      int count = sscanf(slash_pos + 1, "%u-%u", &f, &t);
		      if (count >= 1)
			from = f;
		      if (count == 2)
			to = t;
		    }
		  else
		    {
		      int f;
		      int count = sscanf(slash_pos + 2, "%d", &f);
		      if (count == 1)
			from = -f;
		    }
		}
	    }
	}
      else
	{
	  // CGI queryȤߤʤƲ
	  QueryParser qp(uri);
	  char *bbs = NULL;
	  char *dir = NULL;
	  while (qp.has_more_keys())
	    {
	      const QueryKeyTuple *tuple = qp.next_key();
	      if (tuple == NULL)
		break;

	      if (tuple->get_name_len() == 2)
		{
		  if (strncasecmp(tuple->get_name_raw(), "ST", 2) == 0)
		    {
		      char *value = tuple->get_value();
		      if (value != NULL)
			{
			  unsigned int f;
			  if (sscanf(value, "%u", &f) == 1)
			    from = f;
			  free(value);
			}
		    }
		  else if (strncasecmp(tuple->get_name_raw(), "TO", 2) == 0)
		    {
		      char *value = tuple->get_value();
		      if (value != NULL)
			{
			  unsigned int t;
			  if (sscanf(value, "%u", &t) == 1)
			    to = t;
			  free(value);
			}
		    }
		}
	      else if (tuple->get_name_len() == 3)
		{
		  if (strncasecmp(tuple->get_name_raw(), "BBS", 3) == 0)
		    {
		      if (bbs != NULL)
			free(bbs);
		      bbs = tuple->get_value();
		    }
		  else if (strncasecmp(tuple->get_name_raw(), "DIR", 3) == 0)
		    {
		      if (dir != NULL)
			free(dir);
		      dir = tuple->get_value();
		    }
		  else if (strncasecmp(tuple->get_name_raw(), "KEY", 3) == 0)
		    {
		      set_thread_id(tuple->get_value_raw(),
				    tuple->get_value_len());
		    }
		  else if (strncasecmp(tuple->get_name_raw(), "END", 3) == 0)
		    {
		      char *value = tuple->get_value();
		      if (value != NULL)
			{
			  unsigned int t;
			  if (sscanf(value, "%u", &t) == 1)
			    to = t;
			  free(value);
			}
		    }
		}
	      else if (tuple->get_name_len() == 4
		  && strncasecmp(tuple->get_name_raw(), "LAST", 4) == 0)
		{
		  char *value = tuple->get_value();
		  if (value != NULL)
		    {
		      int f;
		      if (sscanf(value, "%d", &f) == 1)
			from = -f;
		      free(value);
		    }
		}
	      else if (tuple->get_name_len() == 5
		       && strncasecmp(tuple->get_name_raw(), "START", 5) == 0)
		{
		  char *value = tuple->get_value();
		  if (value != NULL)
		    {
		      unsigned int f;
		      if (sscanf(value, "%u", &f) == 1)
			from = f;
		      free(value);
		    }
		}
	    }

	  if (dir != NULL && bbs != NULL)
	    {
	      string id_buf;
	      id_buf.append(dir).append("/").append(bbs);
	      set_board_id(id_buf.c_str(), id_buf.length());
	    }
	  else if (bbs != NULL)
	    {
	      set_board_id(bbs, strlen(bbs));
	    }
	  if (dir != NULL)
	    free(dir);
	  if (bbs != NULL)
	    free(bbs);
	}
      return;
    }

  // URLäݤ
  // URLξ硢bbs_dirϿ¬Ǥʤ
  if (*path != '/')
    return;	// ХѥäݤΤǤ

  size_t len = strlen(path);
  if (path[len - 1] != '/')
    return;

  const char *tmp_pos = path + len - 2;
  int i = board_id_depth;
  while (tmp_pos >= path)
    {
      if (*tmp_pos == '/' && --i == 0)
	break;
      --tmp_pos;
    }

  if (i != 0)
    return;	// 

  assert(tmp_pos >= path);

  set_board_id(tmp_pos + 1, strlen(tmp_pos + 1) - 1);
  set_bbs_root(path, tmp_pos - path + 1);
}


}


#ifdef URI_TEST
#include <iostream>
#include <string>


#define NONNULL_STRING(val) ((val) ? val : "<null>")


int
main(int argc, char *argv[])
{
  Ochusha::URI uri("http://fuyu:pass@pc3.2ch.net/test/read.cgi/linux/1079912788/");
  cout << "uri1: " << NONNULL_STRING(uri.get_uri()) << endl;
  cout << "scheme: " << NONNULL_STRING(uri.get_scheme()) << endl;
  cout << "opaque: " << NONNULL_STRING(uri.get_opaque()) << endl;
  cout << "authority: " << NONNULL_STRING(uri.get_authority()) << endl;
  cout << "server: " << NONNULL_STRING(uri.get_server()) << endl;
  cout << "user: " << NONNULL_STRING(uri.get_user()) << endl;
  cout << "port: " << uri.get_port() << endl;
  cout << "path: " << NONNULL_STRING(uri.get_path()) << endl;
  cout << "query: " << NONNULL_STRING(uri.get_query()) << endl;
  cout << "fragment: " << NONNULL_STRING(uri.get_fragment()) << endl << endl;

  Ochusha::URI uri2("http://kanto.machi.to/bbs/read.pl?BBS=tokyo&KEY=1081721841&START=10&END=10&NOFIRST=");
  cout << "uri: " << NONNULL_STRING(uri2.get_uri()) << endl;
  cout << "scheme: " << NONNULL_STRING(uri2.get_scheme()) << endl;
  cout << "opaque: " << NONNULL_STRING(uri2.get_opaque()) << endl;
  cout << "authority: " << NONNULL_STRING(uri2.get_authority()) << endl;
  cout << "server: " << NONNULL_STRING(uri2.get_server()) << endl;
  cout << "user: " << NONNULL_STRING(uri2.get_user()) << endl;
  cout << "port: " << uri2.get_port() << endl;
  cout << "path: " << NONNULL_STRING(uri2.get_path()) << endl;
  cout << "query: " << NONNULL_STRING(uri2.get_query()) << endl;
  cout << "fragment: " << NONNULL_STRING(uri2.get_fragment()) << endl << endl;

  Ochusha::QueryParser qp(uri2);
  while (qp.has_more_keys())
    {
      const Ochusha::QueryKeyTuple *tuple = qp.next_key();
      string name(tuple->get_name_raw(), tuple->get_name_len());
      string val(tuple->get_value_raw(), tuple->get_value_len());
      cout << "KEY: " << name << " = " << val << endl;
    }

  const char *url3 = "http://2ch.senshin.org/offevent/cgi-bin/img-box/img20040126200341.jpg";
  cout << "url3: " << url3 << endl;
  char *str = Ochusha::URI::encode_string(url3);
  cout << "encoded url3: " << NONNULL_STRING(str) << endl;
  char *str2 = Ochusha::URI::decode_string(str);
  cout << "decoded(encoded url3): " << NONNULL_STRING(str2) << endl;
  free(str);
  free(str2);

  return 0;
}

#endif
