/*
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: combo_action.c,v 1.7 2009/01/01 07:46:51 fuyu Exp $
 */

#include "config.h"

#include "libochushagtk/combo_action.h"
#include "libochushagtk/marshal.h"

#include <gtk/gtk.h>


static void ochusha_combo_action_class_init(OchushaComboActionClass *klass);
static void ochusha_combo_action_init(OchushaComboAction *action);
static void ochusha_combo_action_finalize(GObject *object);

static GtkWidget *create_menu_item(GtkAction *action);
static GtkWidget *create_tool_item(GtkAction *action);

static void ochusha_combo_action_set_active_internal(OchushaComboAction *action,
						     GtkTreeIter *iter,
						     GtkTreePath *path);
#if GTK_CHECK_VERSION(2,12,0)
static gboolean combo_box_query_tooltip_cb(GtkWidget *widget, gint x, gint y,
					   gboolean keyboard_tooltip,
					   GtkTooltip *tooltip,
					   gpointer user_data);
#endif


GType
ochusha_combo_action_get_type(void)
{
  static GType combo_action_type = 0;

  if (combo_action_type == 0)
    {
      static const GTypeInfo combo_action_info =
	{
	  sizeof(OchushaComboActionClass),
	  NULL,	/* base_init */
	  NULL, /* base_finalize */
	  (GClassInitFunc)ochusha_combo_action_class_init,
	  NULL,	/* class_finalize */
	  NULL,	/* class_data */
	  sizeof(OchushaComboAction),
	  0,	/* n_preallocs */
	  (GInstanceInitFunc)ochusha_combo_action_init,
	};

      combo_action_type = g_type_register_static(GTK_TYPE_ACTION,
						 "OchushaComboAction",
						 &combo_action_info, 0);
    }

  return combo_action_type;
}


enum {
  CHANGED_SIGNAL = 0,
  LAST_SIGNAL
};


static GObjectClass *parent_class = NULL;
static gint combo_action_signals[LAST_SIGNAL] = { 0 };


static void
ochusha_combo_action_class_init(OchushaComboActionClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
  GtkActionClass *action_class = GTK_ACTION_CLASS(klass);

  parent_class = g_type_class_peek_parent(klass);

  /* as a GObject */
  gobject_class->finalize = ochusha_combo_action_finalize;

  /* as a GtkAction */
  action_class->menu_item_type = GTK_TYPE_IMAGE_MENU_ITEM;
  action_class->toolbar_item_type = GTK_TYPE_TOOL_ITEM;

  action_class->create_menu_item = create_menu_item;
  action_class->create_tool_item = create_tool_item;

  combo_action_signals[CHANGED_SIGNAL]
    = g_signal_new("changed",
		   G_TYPE_FROM_CLASS(klass),
		   G_SIGNAL_RUN_LAST,
		   G_STRUCT_OFFSET(OchushaComboActionClass, changed),
		   NULL, NULL,
		   libochushagtk_marshal_VOID__OBJECT_BOXED_BOXED,
		   G_TYPE_NONE, 3,
		   GTK_TYPE_TREE_MODEL,
		   GTK_TYPE_TREE_ITER | G_SIGNAL_TYPE_STATIC_SCOPE,
		   GTK_TYPE_TREE_PATH | G_SIGNAL_TYPE_STATIC_SCOPE);

  klass->changed = NULL;
}


static GtkMenuItem *
menu_item_new_cb(OchushaTreeMenuHelper *helper, GtkMenuShell *shell,
		 GtkTreeModel *model, GtkTreeIter *iter,
		 OchushaComboAction *action)
{
  GtkMenuItem *leader = ochusha_tree_menu_helper_get_nth_menu_item(helper,
								   shell, 0);
  if (leader != NULL && GTK_IS_RADIO_MENU_ITEM(leader))
    {
      GtkRadioMenuItem *group = GTK_RADIO_MENU_ITEM(leader);
      return GTK_MENU_ITEM(gtk_radio_menu_item_new_from_widget(group));
    }

  return GTK_MENU_ITEM(gtk_radio_menu_item_new(NULL));
}


static void
activate_cb(OchushaTreeMenuHelper *helper, GtkTreeModel *model,
	    GtkTreeIter *iter, GtkTreePath *path, guint button,
	    OchushaComboAction *action)
{
  ochusha_combo_action_set_active_internal(action, iter, path);
}


static void
ochusha_combo_action_init(OchushaComboAction *action)
{
  action->helper = ochusha_tree_menu_helper_new();
  g_signal_connect(action->helper, "menu_item_new",
		   G_CALLBACK(menu_item_new_cb), action);
  g_signal_connect(action->helper, "activate",
		   G_CALLBACK(activate_cb), action);

  action->model = NULL;
  action->active_row = NULL;

  action->layout_data_func = NULL;
  action->layout_data_func_data = NULL;
  action->layout_data_func_data_destroy = NULL;

  action->tooltip_data_func = NULL;
  action->tooltip_data_func_data = NULL;
  action->tooltip_data_func_data_destroy = NULL;

  action->combo_list = NULL;
}


static void
ochusha_combo_action_finalize(GObject *object)
{
  OchushaComboAction *action;
  g_return_if_fail(OCHUSHA_IS_COMBO_ACTION(object));

  action = OCHUSHA_COMBO_ACTION(object);

  g_return_if_fail(action->combo_list == NULL);

  if (action->model != NULL)
    {
      g_object_unref(action->model);
      action->model = NULL;
    }

  if (action->layout_data_func_data_destroy != NULL)
    {
      (*action->layout_data_func_data_destroy)(action->layout_data_func_data);
      action->layout_data_func_data = NULL;
      action->layout_data_func_data_destroy = NULL;
    }

  if (action->tooltip_data_func_data_destroy != NULL)
    {
      (*action->tooltip_data_func_data_destroy)(action->tooltip_data_func_data);
      action->tooltip_data_func_data = NULL;
      action->tooltip_data_func_data_destroy = NULL;
    }

  if (G_OBJECT_CLASS(parent_class)->finalize != NULL)
    (*G_OBJECT_CLASS(parent_class)->finalize)(object);
}


OchushaComboAction *
ochusha_combo_action_new(const gchar *name, const gchar *label,
			 const gchar *tooltip, const gchar *stock_id)
{
  OchushaComboAction *action = g_object_new(OCHUSHA_TYPE_COMBO_ACTION,
					    "name", name,
					    "label", label,
					    "tooltip", tooltip,
					    "stock_id", stock_id,
					    NULL);
  return action;
}


OchushaComboAction *
ochusha_combo_action_new_with_model(const gchar *name,
				    const gchar *label,
				    const gchar *tooltip,
				    const gchar *stock_id,
				    GtkTreeModel *model)
{
  g_return_val_if_fail(model != NULL, NULL);
  g_return_val_if_fail(GTK_IS_TREE_MODEL(model), NULL);

  OchushaComboAction *action = g_object_new(OCHUSHA_TYPE_COMBO_ACTION,
					    "name", name,
					    "label", label,
					    "tooltip", tooltip,
					    "stock_id", stock_id,
					    NULL);

  ochusha_combo_action_set_model(action, model);

  return action;
}


static GtkWidget *
create_menu_item(GtkAction *action)
{
  OchushaComboAction *combo_action = OCHUSHA_COMBO_ACTION(action);

  GtkWidget *menu_item = g_object_new(GTK_TYPE_IMAGE_MENU_ITEM, NULL);
  GtkWidget *submenu = gtk_menu_new();

  ochusha_tree_menu_helper_attach_tree_menu(combo_action->helper,
					    GTK_MENU_SHELL(submenu));
  gtk_widget_show(submenu);
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(menu_item), submenu);

  return menu_item;
}


static void
combo_box_destroy_cb(GtkWidget *widget, OchushaComboAction *action)
{
  action->combo_list = g_list_remove(action->combo_list, widget);
}


static void
combo_changed_cb(GtkComboBox *combo, OchushaComboAction *action)
{
  GtkTreeIter iter;
  if (gtk_combo_box_get_active_iter(combo, &iter))
    {
      GtkTreePath *path = gtk_tree_model_get_path(action->model, &iter);
      ochusha_combo_action_set_active_internal(action, &iter, path);
      gtk_tree_path_free(path);
    }
}


#if GTK_CHECK_VERSION(2,12,0)
static gboolean
combo_box_query_tooltip_cb(GtkWidget *widget, gint x, gint y,
			   gboolean keyboard_tooltip,
			   GtkTooltip *tooltip, gpointer user_data)
{
  OchushaComboAction *action;
  GtkTreeIter iter;
  g_return_val_if_fail(GTK_IS_COMBO_BOX(widget), FALSE);
  g_return_val_if_fail(OCHUSHA_IS_COMBO_ACTION(user_data), FALSE);

  action = OCHUSHA_COMBO_ACTION(user_data);

  if (action->tooltip_data_func == NULL)
    return FALSE;

  if (gtk_combo_box_get_active_iter(GTK_COMBO_BOX(widget), &iter))
    {
      (*action->tooltip_data_func)(tooltip, action->model, &iter,
				   action->tooltip_data_func_data);
      return TRUE;
    }

  return FALSE;
}
#endif


static GtkWidget *
create_combo_box(OchushaComboAction *action)
{
  GtkWidget *combo;
  GtkCellRenderer *renderer;

  if (action->model != NULL)
    combo = gtk_combo_box_new_with_model(action->model);
  else
    combo = gtk_combo_box_new();

  g_signal_connect(combo, "destroy",
		   G_CALLBACK(combo_box_destroy_cb), action);

#if GTK_CHECK_VERSION(2,12,0)
  g_signal_connect(combo, "query-tooltip",
		   G_CALLBACK(combo_box_query_tooltip_cb), action);
  g_object_set(G_OBJECT(combo), "has-tooltip", TRUE, NULL);
#endif

  renderer = gtk_cell_renderer_text_new();
  gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(combo), renderer, TRUE);
  if (action->layout_data_func != NULL)
    gtk_cell_layout_set_cell_data_func(GTK_CELL_LAYOUT(combo), renderer,
				       action->layout_data_func,
				       action->layout_data_func_data,
				       NULL);

  action->combo_list = g_list_prepend(action->combo_list, combo);
  g_signal_connect(combo, "changed", G_CALLBACK(combo_changed_cb), action);
  if (gtk_tree_row_reference_valid(action->active_row))
    {
      GtkTreePath *path
	= gtk_tree_row_reference_get_path(action->active_row);
      GtkTreeModel *model = action->model;
      GtkTreeIter iter;

      gtk_tree_model_get_iter(model, &iter, path);
      gtk_combo_box_set_active_iter(GTK_COMBO_BOX(combo), &iter);

      gtk_tree_path_free(path);
    }

  return combo;
}


static void
ochusha_combo_action_set_active_internal(OchushaComboAction *action,
					 GtkTreeIter *iter,
					 GtkTreePath *path)
{
  GList *menu_item_list;
  gint depth;

  g_return_if_fail(path != NULL);

  depth = gtk_tree_path_get_depth(path);
  g_return_if_fail(depth == 1);

  if (gtk_tree_row_reference_valid(action->active_row))
    {
      GtkTreePath *active_path
	= gtk_tree_row_reference_get_path(action->active_row);
      gint path_cmp = gtk_tree_path_compare(path, active_path);
      gtk_tree_path_free(active_path);
      if (path_cmp == 0)
	return;
    }
  else if (action->active_row != NULL)
    gtk_tree_row_reference_free(action->active_row);

  action->active_row = gtk_tree_row_reference_new(action->model, path);

  menu_item_list
    = ochusha_tree_menu_helper_collect_menu_items(action->helper, path);
  if (menu_item_list != NULL)
    {
      GList *list = menu_item_list;
      while (list != NULL)
	{
	  gtk_check_menu_item_set_active(list->data, TRUE);
	  list = g_list_next(list);
	}

      g_list_free(menu_item_list);
    }

  if (action->combo_list != NULL)
    {
      GList *list = action->combo_list;
      while (list != NULL)
	{
	  gtk_combo_box_set_active_iter(list->data, iter);
	  list = g_list_next(list);
	}
    }

  g_signal_emit(action, combo_action_signals[CHANGED_SIGNAL], 0,
		action->model, iter, path);
}


static GtkWidget *
create_tool_item(GtkAction *action)
{
  OchushaComboAction *combo_action = OCHUSHA_COMBO_ACTION(action);

  GtkWidget *tool_item = g_object_new(GTK_TYPE_TOOL_ITEM, NULL);
  GtkWidget *combo = create_combo_box(combo_action);
  gtk_combo_box_set_focus_on_click(GTK_COMBO_BOX(combo), FALSE);
  gtk_widget_show(combo);
  gtk_container_add(GTK_CONTAINER(tool_item), combo);

  return tool_item;
}


GtkWidget *
ochusha_combo_action_create_combo_box(OchushaComboAction *action)
{
  g_return_val_if_fail(OCHUSHA_IS_COMBO_ACTION(action), NULL);

  return create_combo_box(action);
}


void
ochusha_combo_action_set_model(OchushaComboAction *action,
			       GtkTreeModel *model)
{
  GList *list;
  g_return_if_fail(OCHUSHA_IS_COMBO_ACTION(action));
  g_return_if_fail(GTK_IS_TREE_MODEL(model));

  list = action->combo_list;
  while (list != NULL)
    {
      gtk_combo_box_set_model(list->data, model);
      list = g_list_next(list);
    }

  g_object_ref(model);

  if (action->model != NULL)
    g_object_unref(action->model);

  action->model = model;

  ochusha_tree_menu_helper_set_model(action->helper, model);
}


GtkTreeModel *
ochusha_combo_action_get_model(OchushaComboAction *action)
{
  g_return_val_if_fail(OCHUSHA_IS_COMBO_ACTION(action), NULL);

  return action->model;
}


void
ochusha_combo_action_set_menu_data_func(OchushaComboAction *action,
					OchushaTreeMenuDataFunc func,
					gpointer func_data,
					GDestroyNotify destroy)
{
  g_return_if_fail(OCHUSHA_IS_COMBO_ACTION(action));

  ochusha_tree_menu_helper_set_menu_data_func(action->helper,
					      func, func_data, destroy);
}


void
ochusha_combo_action_set_cell_data_func(OchushaComboAction *action,
					GtkCellLayoutDataFunc func,
					gpointer func_data,
					GDestroyNotify destroy)
{
  g_return_if_fail(OCHUSHA_IS_COMBO_ACTION(action));

  if (action->layout_data_func_data_destroy != NULL)
    (*action->layout_data_func_data_destroy)(action->layout_data_func_data);

  action->layout_data_func = func;
  action->layout_data_func_data = func_data;
  action->layout_data_func_data_destroy = destroy;
}


void
ochusha_combo_action_set_tooltip_data_func(OchushaComboAction *action,
					   OchushaTooltipDataFunc func,
					   gpointer func_data,
					   GDestroyNotify destroy)
{
  g_return_if_fail(OCHUSHA_IS_COMBO_ACTION(action));

  if (action->tooltip_data_func_data_destroy != NULL)
    (*action->tooltip_data_func_data_destroy)(action->tooltip_data_func_data);

  action->tooltip_data_func = func;
  action->tooltip_data_func_data = func_data;
  action->tooltip_data_func_data_destroy = destroy;
}


gboolean
ochusha_combo_action_get_active_iter(OchushaComboAction *action,
				     GtkTreeIter *iter)
{
  g_return_val_if_fail(OCHUSHA_IS_COMBO_ACTION(action), FALSE);

  if (gtk_tree_row_reference_valid(action->active_row))
    {
      GtkTreePath *path = gtk_tree_row_reference_get_path(action->active_row);
      gboolean result = gtk_tree_model_get_iter(action->model, iter, path);
      gtk_tree_path_free(path);

      return result;
    }

  return FALSE;
}


void
ochusha_combo_action_set_active_iter(OchushaComboAction *action,
				     GtkTreeIter *iter)
{
  GtkTreePath *path;

  g_return_if_fail(OCHUSHA_IS_COMBO_ACTION(action));

  path = gtk_tree_model_get_path(action->model, iter);
  ochusha_combo_action_set_active_internal(action, iter, path);
  gtk_tree_path_free(path);
}
