/*
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: entry_action.c,v 1.7 2009/01/01 07:46:51 fuyu Exp $
 */

#include "config.h"

#include "libochushagtk/entry_action.h"
#include "libochushagtk/marshal.h"

#include <gtk/gtk.h>

#include <string.h>


static void ochusha_entry_action_class_init(OchushaEntryActionClass *klass);
static void ochusha_entry_action_init(OchushaEntryAction *action);
static void ochusha_entry_action_finalize(GObject *object);

static GtkWidget *create_tool_item(GtkAction *action);


GType
ochusha_entry_action_get_type(void)
{
  static GType entry_action_type = 0;

  if (entry_action_type == 0)
    {
      static const GTypeInfo entry_action_info =
	{
	  sizeof(OchushaEntryActionClass),
	  NULL,	/* base_init */
	  NULL, /* base_finalize */
	  (GClassInitFunc)ochusha_entry_action_class_init,
	  NULL,	/* class_finalize */
	  NULL,	/* class_data */
	  sizeof(OchushaEntryAction),
	  0,	/* n_preallocs */
	  (GInstanceInitFunc)ochusha_entry_action_init,
	};

      entry_action_type = g_type_register_static(GTK_TYPE_ACTION,
						 "OchushaEntryAction",
						 &entry_action_info, 0);
    }

  return entry_action_type;
}


enum {
  ENTRY_ACTIVATE_SIGNAL = 0,
  LAST_SIGNAL
};


static GObjectClass *parent_class = NULL;
static gint entry_action_signals[LAST_SIGNAL] = { 0 };


static void
ochusha_entry_action_class_init(OchushaEntryActionClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
  GtkActionClass *action_class = GTK_ACTION_CLASS(klass);

  parent_class = g_type_class_peek_parent(klass);

  /* as a GObject */
  gobject_class->finalize = ochusha_entry_action_finalize;

  /* as a GtkAction */
  action_class->menu_item_type = GTK_TYPE_IMAGE_MENU_ITEM;
  action_class->toolbar_item_type = GTK_TYPE_TOOL_ITEM;

  action_class->create_tool_item = create_tool_item;

  entry_action_signals[ENTRY_ACTIVATE_SIGNAL]
    = g_signal_new("entry_activate",
		   G_TYPE_FROM_CLASS(klass),
		   G_SIGNAL_RUN_LAST,
		   G_STRUCT_OFFSET(OchushaEntryActionClass, entry_activate),
		   NULL, NULL,
		   libochushagtk_marshal_VOID__STRING,
		   G_TYPE_NONE, 1,
		   G_TYPE_STRING | G_SIGNAL_TYPE_STATIC_SCOPE);

  klass->entry_activate = NULL;
}


static void
ochusha_entry_action_init(OchushaEntryAction *action)
{
  action->width_chars = 0;
  action->model = GTK_TREE_MODEL(gtk_list_store_new(1, G_TYPE_STRING));
  action->widget_list = NULL;
  action->enable_completion = FALSE;
}


static void
ochusha_entry_action_finalize(GObject *object)
{
  OchushaEntryAction *action = OCHUSHA_ENTRY_ACTION(object);

  g_return_if_fail(action->widget_list == NULL);

  if (action->model != NULL)
    {
      g_object_unref(action->model);
      action->model = NULL;
    }

  if (G_OBJECT_CLASS(parent_class)->finalize != NULL)
    (*G_OBJECT_CLASS(parent_class)->finalize)(object);
}


static gboolean
completion_match(GtkEntryCompletion *completion, const gchar *key,
		 GtkTreeIter *iter, gpointer data)
{
  OchushaEntryAction *action = data;
  gchar *text = NULL;
  gboolean result;

  gtk_tree_model_get(action->model, iter, 0, &text, -1);
  g_return_val_if_fail(text != NULL, FALSE);

  result = strstr(text, key) != NULL;

  g_free(text);
  return result;
}


static void
setup_completion(OchushaEntryAction *action, GtkComboBoxEntry *combo_entry)
{
  GtkEntry *entry = GTK_ENTRY(GTK_BIN(combo_entry)->child);
  GtkEntryCompletion *completion = gtk_entry_completion_new();

  gtk_entry_completion_set_model(completion, action->model);
  gtk_entry_completion_set_match_func(completion, completion_match,
				      action, NULL);

  gtk_entry_completion_set_text_column(completion, 0);

  gtk_entry_set_completion(entry, completion);

  g_object_unref(completion);
}


static void
activate_cb(GtkEntry *entry, OchushaEntryAction *action)
{
  g_signal_emit(action, entry_action_signals[ENTRY_ACTIVATE_SIGNAL],
		0,
		gtk_entry_get_text(entry));
}


static void
destroy_cb(GtkWidget *entry, OchushaEntryAction *action)
{
  action->widget_list = g_list_remove(action->widget_list, entry);
}


static GtkWidget *
create_entry_widget(OchushaEntryAction *action)
{
  GtkWidget *widget = gtk_combo_box_entry_new_with_model(action->model, 0);
  GtkComboBoxEntry *combo_entry = GTK_COMBO_BOX_ENTRY(widget);
  GtkEntry *entry = GTK_ENTRY(GTK_BIN(combo_entry)->child);

  if (action->enable_completion)
    setup_completion(action, combo_entry);

  if (action->width_chars != 0)
    gtk_entry_set_width_chars(entry, action->width_chars);

  g_signal_connect(entry, "activate",
		   G_CALLBACK(activate_cb), action);
  g_signal_connect(widget, "destroy",
		   G_CALLBACK(destroy_cb), action);

  return widget;
}


static GtkWidget *
create_tool_item(GtkAction *action)
{
  OchushaEntryAction *entry_action = OCHUSHA_ENTRY_ACTION(action);
  GtkWidget *item = g_object_new(GTK_TYPE_TOOL_ITEM, NULL);
  GtkWidget *entry = create_entry_widget(entry_action);

  gtk_combo_box_set_focus_on_click(GTK_COMBO_BOX(entry), FALSE);
  gtk_widget_show(entry);
  gtk_container_add(GTK_CONTAINER(item), entry);
  gtk_container_set_border_width(GTK_CONTAINER(item), 3);

  gtk_tool_item_set_expand(GTK_TOOL_ITEM(item), TRUE);

  return item;
}


OchushaEntryAction *
ochusha_entry_action_new(const gchar *name, const gchar *label,
			 const gchar *tooltip, const gchar *stock_id)
{
  OchushaEntryAction *action = g_object_new(OCHUSHA_TYPE_ENTRY_ACTION,
					    "name", name,
					    "label", label,
					    "tooltip", tooltip,
					    "stock_id", stock_id,
					    NULL);
  return action;
}


gint
ochusha_entry_action_get_default_width_chars(OchushaEntryAction *action)
{
  g_return_val_if_fail(OCHUSHA_IS_ENTRY_ACTION(action), 0);
  return action->width_chars;
}


void
ochusha_entry_action_set_default_width_chars(OchushaEntryAction *action,
					     gint n_chars)
{
  g_return_if_fail(OCHUSHA_IS_ENTRY_ACTION(action));
  action->width_chars = n_chars;
}


gboolean
ochusha_entry_action_get_enable_completion(OchushaEntryAction *action)
{
  g_return_val_if_fail(OCHUSHA_IS_ENTRY_ACTION(action), FALSE);

  return action->enable_completion;
}


void
ochusha_entry_action_set_enable_completion(OchushaEntryAction *action,
					   gboolean value)
{
  g_return_if_fail(OCHUSHA_IS_ENTRY_ACTION(action));

  if (value == action->enable_completion)
    return;

  action->enable_completion = value;

  if (value)
    {
      GList *list;
      list = action->widget_list;
      while (list != NULL)
	{
	  setup_completion(action, list->data);
	  list = g_list_next(list);
	}
    }
  else
    {
      GList *list;
      list = action->widget_list;
      while (list != NULL)
	{
	  GtkEntry *entry = GTK_ENTRY(GTK_BIN(list->data)->child);
	  gtk_entry_set_completion(entry, NULL);
	  list = g_list_next(list);
	}
    }
}


void
ochusha_entry_action_insert_completion_text(OchushaEntryAction *action,
					    gint index, const gchar *text)
{
  g_return_if_fail(OCHUSHA_IS_ENTRY_ACTION(action));
  g_return_if_fail(text != NULL);

  if (*text == '\0')
    return;

  GtkTreeIter iter;
  if (gtk_tree_model_get_iter_first(action->model, &iter))
    {
      do
	{
	  gchar *completion_text = NULL;
	  gtk_tree_model_get(action->model, &iter, 0, &completion_text, -1);
	  g_return_if_fail(completion_text != NULL);

	  if (strcmp(text, completion_text) == 0)
	    {
	      gtk_list_store_move_after(GTK_LIST_STORE(action->model),
					&iter, NULL);
	      g_free(completion_text);
	      return;
	    }
	  g_free(completion_text);
	}
      while (gtk_tree_model_iter_next(action->model, &iter));
    }

  gtk_list_store_insert_with_values(GTK_LIST_STORE(action->model), NULL, index,
				    0, text,
				    -1);
}


GtkWidget *
ochusha_entry_action_create_entry_widget(OchushaEntryAction *action)
{
  g_return_val_if_fail(OCHUSHA_IS_ENTRY_ACTION(action), NULL);

  return create_entry_widget(action);
}

