/*
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: hypertext_buffer.c,v 1.25 2009/01/01 07:46:51 fuyu Exp $
 */

#include "config.h"

#include "libochushagtk/hypertext_buffer.h"

#ifdef HAVE_STDINT_H
#include <stdint.h>
#else
# ifdef HAVE_INTTYPES_H
# include <inttpyes.h>
# else
typedef long intptr_t;
# endif
#endif

#include <stdio.h>
#include <string.h>


static void ochusha_hypertext_buffer_class_init(OchushaHypertextBufferClass *klass);
static void ochusha_hypertext_buffer_init(OchushaHypertextBuffer *buffer);
static void ochusha_hypertext_buffer_finalize(GObject *object);


typedef struct _LinkInfo LinkInfo;

static LinkInfo *ochusha_hypertext_buffer_insert_link_info(
						OchushaHypertextBuffer *buffer,
						gint offset, size_t length,
						const gchar *link);
static LinkInfo *ochusha_hypertext_buffer_lookup_link_info(
						OchushaHypertextBuffer *buffer,
						gint offset);
static void ochusha_hypertext_buffer_remove_link_info(
						OchushaHypertextBuffer *buffer,
						LinkInfo *info);
static void ochusha_hypertext_buffer_adjust_link_info(
						OchushaHypertextBuffer *buffer,
						gint offset, gint value);
static void ochusha_hypertext_buffer_clear_link_info(
						OchushaHypertextBuffer *buffer);


typedef struct _ChildInfo ChildInfo;

static ChildInfo *ochusha_hypertext_buffer_insert_child_info(
						OchushaHypertextBuffer *buffer,
						GtkWidget *widget, gint offset);
static ChildInfo *ochusha_hypertext_buffer_lookup_child_info(
						OchushaHypertextBuffer *buffer,
						GtkWidget *widget);
static void ochusha_hypertext_buffer_remove_child_info(
						OchushaHypertextBuffer *buffer,
						ChildInfo *info);
static void ochusha_hypertext_buffer_adjust_child_info(
						OchushaHypertextBuffer *buffer,
						gint offset, gint value);
static void ochusha_hypertext_buffer_clear_child_info(
						OchushaHypertextBuffer *buffer);


GType
ochusha_hypertext_buffer_get_type(void)
{
  static GType hypertext_buffer_type = 0;

  if (hypertext_buffer_type == 0)
    {
      static const GTypeInfo hypertext_buffer_info =
	{
	  sizeof(OchushaHypertextBufferClass),
	  NULL, /* base_init */
	  NULL, /* base_finalize */
	  (GClassInitFunc)ochusha_hypertext_buffer_class_init,
	  NULL, /* class_finalize */
	  NULL, /* class_data */
	  sizeof(OchushaHypertextBuffer),
	  0,	/* n_preallocs */
	  (GInstanceInitFunc)ochusha_hypertext_buffer_init,
	};

      hypertext_buffer_type = g_type_register_static(GTK_TYPE_TEXT_BUFFER,
						     "OchushaHypertextBuffer",
						     &hypertext_buffer_info,
						     0);
    }
  
  return hypertext_buffer_type;
}


static GtkTextBufferClass *parent_class = NULL;
static OchushaHypertextTagTable *default_tag_table = NULL;


OchushaHypertextTagTable *
ochusha_hypertext_buffer_get_default_tag_table(void)
{
  if (default_tag_table != NULL)
    return default_tag_table;

  default_tag_table = ochusha_hypertext_tag_table_new();
  return default_tag_table;
}


static void
ochusha_hypertext_buffer_class_init(OchushaHypertextBufferClass *klass)
{
  GObjectClass *o_class = (GObjectClass *)klass;

  parent_class = g_type_class_peek_parent(klass);

  /* GObject signals */
  o_class->finalize = ochusha_hypertext_buffer_finalize;

  ochusha_hypertext_buffer_get_default_tag_table();
}


struct _LinkInfo
{
  gint offset;
  size_t length;
  char *link;
};


static LinkInfo *
ochusha_hypertext_buffer_insert_link_info(OchushaHypertextBuffer *buffer,
					  gint offset, size_t length,
					  const gchar *link)
{
  LinkInfo *info;
  g_return_val_if_fail(OCHUSHA_IS_HYPERTEXT_BUFFER(buffer), NULL);
  g_return_val_if_fail(link != NULL, NULL);
  g_return_val_if_fail(length > 0, NULL);
  g_return_val_if_fail(ochusha_hypertext_buffer_lookup_link_info(
								buffer,
								offset) == NULL,
		       NULL);

#if GLIB_CHECK_VERSION(2,10,0)
  info = g_slice_new(LinkInfo);
#else
  info = g_mem_chunk_alloc(buffer->link_info_chunk);
#endif
  info->offset = offset;
  info->length = length;
  info->link = g_strdup(link);
  g_tree_insert(buffer->link_info, info, info);

  return info;
}


static LinkInfo *
ochusha_hypertext_buffer_lookup_link_info(OchushaHypertextBuffer *buffer,
					  gint offset)
{
  LinkInfo key = { offset, 0, NULL };
  g_return_val_if_fail(OCHUSHA_IS_HYPERTEXT_BUFFER(buffer), NULL);

  return g_tree_lookup(buffer->link_info, &key);
}


static void
ochusha_hypertext_buffer_remove_link_info(OchushaHypertextBuffer *buffer,
					  LinkInfo *info)
{
  g_return_if_fail(OCHUSHA_IS_HYPERTEXT_BUFFER(buffer));
  g_return_if_fail(info != NULL);
  g_return_if_fail(ochusha_hypertext_buffer_lookup_link_info(
							buffer,
							info->offset) != NULL);

  g_tree_remove(buffer->link_info, info);
#if GLIB_CHECK_VERSION(2,10,0)
  g_slice_free(LinkInfo, info);
#else
  g_mem_chunk_free(buffer->link_info_chunk, info);
#endif
}


typedef struct _AdjustInfoArgs
{
  gint offset;
  gint value;
} AdjustInfoArgs;


static gboolean
adjust_link_info(LinkInfo *key, LinkInfo *value, AdjustInfoArgs *args)
{
  g_return_val_if_fail(key == value, TRUE);

  if (value->offset >= args->offset)
    value->offset += args->value;

  return FALSE;
}


/*
 * buffer->link_info˴ޤޤƤLinkInfo *infoˤĤơ
 * info->offset >= offsetʤinfo->offset += value롣
 */
static void
ochusha_hypertext_buffer_adjust_link_info(OchushaHypertextBuffer *buffer,
					  gint offset, gint value)
{
  AdjustInfoArgs args = { offset, value };

  g_return_if_fail(OCHUSHA_IS_HYPERTEXT_BUFFER(buffer));

  if (value == 0)
    return;

  g_tree_foreach(buffer->link_info, (GTraverseFunc)adjust_link_info, &args);
}


static gint
link_info_compare_func(LinkInfo *key0, LinkInfo *key1, gpointer unused)
{
  gint offset0 = key0->offset;
  gint offset1 = key1->offset;

  if (key0->length == 0)
    {
      g_return_val_if_fail(key1->length > 0, 0);

       if (offset0 >= offset1 && offset0 < (offset1 + key1->length))
	 return 0;
       return offset0 - offset1;
    }

  if (offset1 >= offset0 && (offset1 < offset0 + key0->length))
    return 0;
  return offset0 - offset1;

}


static void
link_info_destroy(LinkInfo *info)
{
  g_return_if_fail(info->link != NULL);
  g_free(info->link);
  info->link = NULL;
}


static void
ochusha_hypertext_buffer_clear_link_info(OchushaHypertextBuffer *buffer)
{
  g_return_if_fail(OCHUSHA_IS_HYPERTEXT_BUFFER(buffer));

  if (buffer->link_info != NULL)
    g_tree_destroy(buffer->link_info);

#if !GLIB_CHECK_VERSION(2,10,0)
  if (buffer->link_info_chunk != NULL)
    g_mem_chunk_destroy(buffer->link_info_chunk);

  /*
   * LinkInfoILP32ILP64ʤ3ɡLP642.5()
   * IA32Υڡ4096ʤΤǡlcm6144061440
   * IA3215120ĤΥ󥯤ޤǤchunkɲóƤʤ
   * ºݤˤGLibǴΰȤƥ֥åȤ˿(6.57)ɤ
   * ;ʬ˳ƤΤǤʬ61425ХȤˤƤ
   *
   * θߤΥɽǤϡƥ쥹˥쥹֡Ǥʤᥤ
   * IDȥ쥹ʸΥ󥯤󥯤ȤưΤǡʿŪˤϥ쥹
   * 45ĤΥ󥯤ĤޤꡢͤΥʤ顢
   * Ͻʬǽʬ­롣ޤΥƥ64k٤Υ
   * ȤΤ˵ȤɬפʤƤʤġġ
   */
  buffer->link_info_chunk = g_mem_chunk_new("OchushaHypertextBuffer::LinkInfo",
					    sizeof(LinkInfo), 61425,
					    G_ALLOC_AND_FREE);
#endif
  buffer->link_info = g_tree_new_full((GCompareDataFunc)link_info_compare_func,
				      NULL, NULL,
				      (GDestroyNotify)link_info_destroy);
}


struct _ChildInfo
{
  GtkWidget *widget;
  gint offset;
};


static ChildInfo *
ochusha_hypertext_buffer_insert_child_info(OchushaHypertextBuffer *buffer,
					   GtkWidget *widget, gint offset)
{
  ChildInfo *info;
  g_return_val_if_fail(OCHUSHA_IS_HYPERTEXT_BUFFER(buffer), NULL);
  g_return_val_if_fail(widget != NULL, NULL);
  g_return_val_if_fail(GTK_IS_WIDGET(widget), NULL);
  g_return_val_if_fail(ochusha_hypertext_buffer_lookup_child_info(
								buffer,
								widget) == NULL,
		       NULL);

#if GLIB_CHECK_VERSION(2,10,0)
  info = g_slice_new(ChildInfo);
#else
  info = g_mem_chunk_alloc(buffer->child_info_chunk);
#endif
  info->widget = widget;
  info->offset = offset;
  g_tree_insert(buffer->child_info, widget, info);

  return info;
}


static ChildInfo *
ochusha_hypertext_buffer_lookup_child_info(OchushaHypertextBuffer *buffer,
					   GtkWidget *widget)
{
  g_return_val_if_fail(OCHUSHA_IS_HYPERTEXT_BUFFER(buffer), NULL);

  return g_tree_lookup(buffer->child_info, widget);
}


static void
ochusha_hypertext_buffer_remove_child_info(OchushaHypertextBuffer *buffer,
					   ChildInfo *info)
{
  g_return_if_fail(OCHUSHA_IS_HYPERTEXT_BUFFER(buffer));
  g_return_if_fail(ochusha_hypertext_buffer_lookup_child_info(
							buffer,
							info->widget) != NULL);

  g_tree_remove(buffer->child_info, info->widget);
#if GLIB_CHECK_VERSION(2,10,0)
  g_slice_free(ChildInfo, info);
#else
  g_mem_chunk_free(buffer->child_info_chunk, info);
#endif
}


static gboolean
adjust_child_info(GtkWidget *key, ChildInfo *value, AdjustInfoArgs *args)
{
  g_return_val_if_fail(key == value->widget, TRUE);

  if (value->offset >= args->offset)
    value->offset += args->value;

  return FALSE;
}


/*
 * buffer->child_info˴ޤޤƤChildInfo *infoˤĤơ
 * info->offset >= offsetʤinfo->offset += value롣
 */
static void
ochusha_hypertext_buffer_adjust_child_info(OchushaHypertextBuffer *buffer,
					   gint offset, gint value)
{
  AdjustInfoArgs args = { offset, value };

  g_return_if_fail(OCHUSHA_IS_HYPERTEXT_BUFFER(buffer));

  if (value == 0)
    return;

  g_tree_foreach(buffer->child_info, (GTraverseFunc)adjust_child_info, &args);
}


static gint
child_info_compare_func(GtkWidget *key0, GtkWidget *key1)
{
  return key0 - key1;
}


static void
ochusha_hypertext_buffer_clear_child_info(OchushaHypertextBuffer *buffer)
{
  g_return_if_fail(OCHUSHA_IS_HYPERTEXT_BUFFER(buffer));

  if (buffer->child_info != NULL)
    g_tree_destroy(buffer->child_info);
#if !GLIB_CHECK_VERSION(2,10,0)
  if (buffer->child_info_chunk != NULL)
    g_mem_chunk_destroy(buffer->link_info_chunk);

  /*
   * ChildInfoILP32ILP64ʤ2ɡLP641.5()
   * IA32Υڡ4096ʤΤǡlcm1228812288
   * IA3211536ĤΥåȤޤǤchunkɲóƤ
   * ʤºݤˤGLibǴΰȤƥ֥åȤ˿(6.57)
   * ;ʬ˳ƤΤǤʬ12276ХȤˤƤ
   *
   * ġĤȻפäåȤϲȤǤʤΤǡ
   * ̵̤¿롣4000ХȤ餤Ǥ䡣
   */
  buffer->child_info_chunk
    = g_mem_chunk_new("OchushaHypertextBuffer::ChildInfo",
		      sizeof(ChildInfo), 4000,
		      G_ALLOC_AND_FREE);
#endif
  buffer->child_info = g_tree_new((GCompareFunc)child_info_compare_func);
}


static void
ochusha_hypertext_buffer_init(OchushaHypertextBuffer *buffer)
{
  buffer->tags = NULL;

  buffer->link_tag = NULL;
  buffer->link_start_offset = 0;

#if !GLIB_CHECK_VERSION(2,10,0)
  buffer->link_info_chunk = NULL;
#endif
  buffer->link_info = NULL;
  ochusha_hypertext_buffer_clear_link_info(buffer);

#if !GLIB_CHECK_VERSION(2,10,0)
  buffer->child_info_chunk = NULL;
#endif
  buffer->child_info = NULL;
  ochusha_hypertext_buffer_clear_child_info(buffer);
}


static void
ochusha_hypertext_buffer_finalize(GObject *object)
{
  OchushaHypertextBuffer *buffer = OCHUSHA_HYPERTEXT_BUFFER(object);

#if 0
  fprintf(stderr, "ochusha_hypertext_buffer_finalize(%p)\n", object);
#endif
  if (buffer->tags != NULL)
    {
      g_slist_free(buffer->tags);
      buffer->tags = NULL;
    }

  if (buffer->link_info != NULL)
    {
      g_tree_destroy(buffer->link_info);
      buffer->link_info = NULL;
    }

#if !GLIB_CHECK_VERSION(2,10,0)
  if (buffer->link_info_chunk != NULL)
    {
      g_mem_chunk_destroy(buffer->link_info_chunk);
      buffer->link_info_chunk = NULL;
    }
#endif

  if (buffer->child_info != NULL)
    {
      g_tree_destroy(buffer->child_info);
      buffer->child_info = NULL;
    }

#if !GLIB_CHECK_VERSION(2,10,0)
  if (buffer->child_info_chunk != NULL)
    {
      g_mem_chunk_destroy(buffer->child_info_chunk);
      buffer->child_info_chunk = NULL;
    }
#endif

  if (G_OBJECT_CLASS(parent_class)->finalize != NULL)
    (*G_OBJECT_CLASS(parent_class)->finalize)(object);
}


OchushaHypertextBuffer *
ochusha_hypertext_buffer_new(OchushaHypertextTagTable *table)
{
  OchushaHypertextBuffer *buffer
    = g_object_new(OCHUSHA_TYPE_HYPERTEXT_BUFFER,
		   "tag_table", table != NULL ? table : default_tag_table,
		   NULL);
  gtk_text_buffer_get_start_iter(&buffer->text_buffer, &buffer->iter);

  return buffer;
}


void
ochusha_hypertext_buffer_clear(OchushaHypertextBuffer *buffer)
{
  GtkTextIter end_iter;

  g_return_if_fail(OCHUSHA_IS_HYPERTEXT_BUFFER(buffer));

  gtk_text_buffer_get_start_iter(&buffer->text_buffer, &buffer->iter);
  gtk_text_buffer_get_end_iter(&buffer->text_buffer, &end_iter);

  gtk_text_buffer_delete(&buffer->text_buffer, &buffer->iter, &end_iter);

  buffer->link_tag = NULL;
  buffer->link_start_offset = 0;

  ochusha_hypertext_buffer_clear_link_info(buffer);
  ochusha_hypertext_buffer_clear_child_info(buffer);
}


gint
ochusha_hypertext_buffer_get_current_offset(OchushaHypertextBuffer *buffer)
{
  g_return_val_if_fail(OCHUSHA_IS_HYPERTEXT_BUFFER(buffer), 0);

  return gtk_text_iter_get_offset(&buffer->iter);
}


gint
ochusha_hypertext_buffer_get_link_offset(OchushaHypertextBuffer *buffer,
					 GtkTextIter *iter)
{
  LinkInfo *info;
  gint offset;

  g_return_val_if_fail(OCHUSHA_IS_HYPERTEXT_BUFFER(buffer), -1);
  g_return_val_if_fail(iter != NULL, -1);

  offset = gtk_text_iter_get_offset(iter);
  g_return_val_if_fail(offset >= 0, -1);

  info = ochusha_hypertext_buffer_lookup_link_info(buffer, offset);
  return info != NULL ? info->offset : -1;
}


const char *
ochusha_hypertext_buffer_get_link_at_offset(OchushaHypertextBuffer *buffer,
					    gint offset)
{
  LinkInfo *info;

  g_return_val_if_fail(OCHUSHA_IS_HYPERTEXT_BUFFER(buffer), NULL);
  g_return_val_if_fail(offset >= 0, NULL);

  info = ochusha_hypertext_buffer_lookup_link_info(buffer, offset);
  return info != NULL ? info->link : NULL;
}


void
ochusha_hypertext_buffer_place_cursor_at_offset(OchushaHypertextBuffer *buffer,
						gint offset)
{
  GtkTextIter iter;

  g_return_if_fail(OCHUSHA_IS_HYPERTEXT_BUFFER(buffer));

  gtk_text_buffer_get_iter_at_offset(&buffer->text_buffer, &iter, offset);
  gtk_text_buffer_place_cursor(&buffer->text_buffer, &iter);
}


GtkTextMark *
ochusha_hypertext_buffer_create_mark(OchushaHypertextBuffer *buffer,
				     const gchar *mark_name)
{
  g_return_val_if_fail(OCHUSHA_IS_HYPERTEXT_BUFFER(buffer), NULL);

  return gtk_text_buffer_create_mark(&buffer->text_buffer, mark_name,
				     &buffer->iter, TRUE); 
}


GtkTextMark *
ochusha_hypertext_buffer_create_mark_at_offset(OchushaHypertextBuffer *buffer,
					       const gchar *mark_name,
					       gint offset)
{
  GtkTextIter iter;

  g_return_val_if_fail(OCHUSHA_IS_HYPERTEXT_BUFFER(buffer), NULL);

  gtk_text_buffer_get_iter_at_offset(&buffer->text_buffer, &iter, offset);
  return gtk_text_buffer_create_mark(&buffer->text_buffer, mark_name,
				     &iter, TRUE); 
}


void
ochusha_hypertext_buffer_delete_mark(OchushaHypertextBuffer *buffer,
				     GtkTextMark *mark)
{
  g_return_if_fail(OCHUSHA_IS_HYPERTEXT_BUFFER(buffer));
  gtk_text_buffer_delete_mark(&buffer->text_buffer, mark);
}


static void
apply_tags(GtkTextBuffer *buffer, GSList *tags,
	   GtkTextIter *start_iter, GtkTextIter *end_iter)
{
  while (tags != NULL)
    {
      gtk_text_buffer_apply_tag(buffer, tags->data, start_iter, end_iter);
      tags = tags->next;
    }
}


void
ochusha_hypertext_buffer_append_text(OchushaHypertextBuffer *buffer,
				     const gchar *text, gint len)
{
  g_return_if_fail(OCHUSHA_IS_HYPERTEXT_BUFFER(buffer));
  g_return_if_fail(text != NULL);

  if (len < 0)
    len = strlen(text);

  if (len > 0)
    {
      GtkTextIter start_iter;
      gint start_offset;

      start_offset = gtk_text_iter_get_offset(&buffer->iter);

      gtk_text_buffer_insert(&buffer->text_buffer, &buffer->iter, text, len);

      gtk_text_buffer_get_iter_at_offset(&buffer->text_buffer,
					 &start_iter, start_offset);
      apply_tags(&buffer->text_buffer, buffer->tags,
		 &start_iter, &buffer->iter);
    }
}


void
ochusha_hypertext_buffer_append_text_with_tag(OchushaHypertextBuffer *buffer,
					      const gchar *text, gint len,
					      GtkTextTag *tag)
{
  g_return_if_fail(OCHUSHA_IS_HYPERTEXT_BUFFER(buffer));
  g_return_if_fail(text != NULL);

  if (len < 0)
    len = strlen(text);

  if (len > 0)
    {
      GtkTextIter start_iter;
      gint start_offset;

      start_offset = gtk_text_iter_get_offset(&buffer->iter);

      gtk_text_buffer_insert_with_tags(&buffer->text_buffer, &buffer->iter,
				       text, len,
				       tag, buffer->link_tag, NULL);

      gtk_text_buffer_get_iter_at_offset(&buffer->text_buffer,
					 &start_iter, start_offset);
    }
}


#if 0
gint
ochusha_hypertext_buffer_insert_text_at_offset(OchushaHypertextBuffer *buffer,
					       const gchar *text, gint len)
{
  return 0;
}


gint
ochusha_hypertext_buffer_remove_range(OchushaHypertextBuffer *buffer,
				      gint start_offset, gint end_offset)
{
  return 0;
}
#endif


gboolean
ochusha_hypertext_buffer_is_at_start_of_line(OchushaHypertextBuffer *buffer)
{
  g_return_val_if_fail(OCHUSHA_IS_HYPERTEXT_BUFFER(buffer), FALSE);
  return gtk_text_iter_starts_line(&buffer->iter);
}


GtkTextChildAnchor *
ochusha_hypertext_buffer_append_child(OchushaHypertextBuffer *buffer,
				      gpointer key)
{
  gint offset;
  GtkTextChildAnchor *anchor;
  GtkTextIter start_iter;

  g_return_val_if_fail(OCHUSHA_IS_HYPERTEXT_BUFFER(buffer), NULL);

  offset = gtk_text_iter_get_offset(&buffer->iter);
  ochusha_hypertext_buffer_insert_child_info(buffer, key, offset);

  anchor = gtk_text_buffer_create_child_anchor(&buffer->text_buffer,
					       &buffer->iter);

  gtk_text_buffer_get_iter_at_offset(&buffer->text_buffer, &start_iter, offset);
  apply_tags(&buffer->text_buffer, buffer->tags, &start_iter, &buffer->iter);

  return anchor;
}


GtkTextChildAnchor *
ochusha_hypertext_buffer_insert_child_at_offset(OchushaHypertextBuffer *buffer,
						gpointer key, gint offset)
{
  GtkTextIter start_iter;
  GtkTextIter iter;
  GtkTextChildAnchor *anchor;
  gint adjust_value;

  g_return_val_if_fail(OCHUSHA_IS_HYPERTEXT_BUFFER(buffer), NULL);

  gtk_text_buffer_get_iter_at_offset(&buffer->text_buffer, &iter, offset);

  anchor = gtk_text_buffer_create_child_anchor(&buffer->text_buffer, &iter);

  gtk_text_buffer_get_iter_at_offset(&buffer->text_buffer, &start_iter, offset);
  apply_tags(&buffer->text_buffer, buffer->tags, &start_iter, &iter);

  adjust_value = gtk_text_iter_get_offset(&iter) - offset;
  g_assert(adjust_value == 1);

  ochusha_hypertext_buffer_adjust_link_info(buffer, offset, adjust_value);
  ochusha_hypertext_buffer_adjust_child_info(buffer, offset, adjust_value);
  ochusha_hypertext_buffer_insert_child_info(buffer, key, offset);

  gtk_text_buffer_get_end_iter(&buffer->text_buffer, &buffer->iter);

  return anchor;
}


gint
ochusha_hypertext_buffer_remove_child(OchushaHypertextBuffer *buffer,
				      gpointer key)
{
  GtkTextIter start;
  GtkTextIter end;
  gint offset;
  ChildInfo *info;

  g_return_val_if_fail(OCHUSHA_IS_HYPERTEXT_BUFFER(buffer), -1);
  g_return_val_if_fail(key != NULL, -1);

  info = ochusha_hypertext_buffer_lookup_child_info(buffer, key);
  g_return_val_if_fail(info != NULL, -1);

  gtk_text_buffer_get_iter_at_offset(&buffer->text_buffer,
				     &start, info->offset);
  gtk_text_buffer_get_iter_at_offset(&buffer->text_buffer,
				     &end, info->offset + 1);
  gtk_text_buffer_delete(&buffer->text_buffer, &start, &end);
  gtk_text_buffer_get_end_iter(&buffer->text_buffer, &buffer->iter);

  offset = info->offset;
  ochusha_hypertext_buffer_remove_child_info(buffer, info);

  ochusha_hypertext_buffer_adjust_link_info(buffer, offset, -1);
  ochusha_hypertext_buffer_adjust_child_info(buffer, offset, -1);

  return offset;
}


GtkTextTag *
ochusha_hypertext_buffer_get_tag_by_name(OchushaHypertextBuffer *buffer,
					 const gchar *name)
{
  OchushaHypertextTagTable *table;

  g_return_val_if_fail(OCHUSHA_IS_HYPERTEXT_BUFFER(buffer), NULL);

  table = OCHUSHA_HYPERTEXT_TAG_TABLE(buffer->text_buffer.tag_table);
  return ochusha_hypertext_tag_table_lookup_tag(table, name);
}


void
ochusha_hypertext_buffer_push_tag_by_name(OchushaHypertextBuffer *buffer,
					  const gchar *name)
{
  OchushaHypertextTagTable *table;
  GtkTextTag *tag;

  g_return_if_fail(OCHUSHA_IS_HYPERTEXT_BUFFER(buffer));
  
  table = OCHUSHA_HYPERTEXT_TAG_TABLE(buffer->text_buffer.tag_table);
  tag = ochusha_hypertext_tag_table_lookup_tag(table, name);

  if (tag != NULL)
    ochusha_hypertext_buffer_push_tag(buffer, tag);
}


void
ochusha_hypertext_buffer_push_tag(OchushaHypertextBuffer *buffer,
				  GtkTextTag *tag)
{
  g_return_if_fail(OCHUSHA_IS_HYPERTEXT_BUFFER(buffer));
  g_return_if_fail(GTK_IS_TEXT_TAG(tag));

  buffer->tags = g_slist_prepend(buffer->tags, tag);
}


GtkTextTag *
ochusha_hypertext_buffer_pop_tag(OchushaHypertextBuffer *buffer)
{
  GtkTextTag *tag;
  GSList *head_entry;

  g_return_val_if_fail(OCHUSHA_IS_HYPERTEXT_BUFFER(buffer), NULL);

  head_entry = buffer->tags;
  if (head_entry == NULL)
    return NULL;

  tag = GTK_TEXT_TAG(head_entry->data);
  buffer->tags = g_slist_remove_link(buffer->tags, head_entry);
  g_slist_free(head_entry);

  g_return_val_if_fail(buffer->link_tag != tag, NULL);

  return tag;
}


void
ochusha_hypertext_buffer_pop_tags(OchushaHypertextBuffer *buffer,
				  GtkTextTag *tag)
{
  GtkTextTag *popped_tag;
  GSList *entry;

  g_return_if_fail(OCHUSHA_IS_HYPERTEXT_BUFFER(buffer));
  g_return_if_fail(GTK_IS_TEXT_TAG(tag));

  entry = buffer->tags;
  while (entry != NULL)
    {
      if (entry->data == tag)
	break;
      entry = entry->next;
    }

  g_return_if_fail(entry != NULL);

  do
    {
      popped_tag = ochusha_hypertext_buffer_pop_tag(buffer);
      g_return_if_fail(popped_tag != NULL);
    }
  while (popped_tag != tag);
}


void
ochusha_hypertext_buffer_push_link_tag(OchushaHypertextBuffer *buffer,
				       GtkTextTag *tag, const gchar *link)
{
  g_return_if_fail(OCHUSHA_IS_HYPERTEXT_BUFFER(buffer));
  g_return_if_fail(GTK_IS_TEXT_TAG(tag));

  if (buffer->link_tag != NULL)
    {
#ifndef NDEBUG
      fprintf(stderr, "Unmatched link tag found!\n");
#endif
      ochusha_hypertext_buffer_pop_link_tag(buffer, buffer->link_tag);
    }

  buffer->link_start_offset = gtk_text_iter_get_offset(&buffer->iter);
  buffer->link_tag = tag;

  /* 󥯤Ĺνͤ1ˤƤpop˿ĹĴ */
  ochusha_hypertext_buffer_insert_link_info(buffer, buffer->link_start_offset,
					    1, link);
  ochusha_hypertext_buffer_push_tag(buffer, tag);
}


void
ochusha_hypertext_buffer_pop_link_tag(OchushaHypertextBuffer *buffer,
				      GtkTextTag *tag)
{
  gint offset;
  LinkInfo *info;

  g_return_if_fail(OCHUSHA_IS_HYPERTEXT_BUFFER(buffer));
  g_return_if_fail(GTK_IS_TEXT_TAG(tag));
  g_return_if_fail(buffer->link_tag != NULL);
  g_return_if_fail(buffer->link_tag == tag);

  offset = gtk_text_iter_get_offset(&buffer->iter);
  g_return_if_fail(offset >= buffer->link_start_offset);

  info = ochusha_hypertext_buffer_lookup_link_info(buffer,
						   buffer->link_start_offset);
  g_return_if_fail(info != NULL);
  if (offset != buffer->link_start_offset)
    info->length = offset - buffer->link_start_offset;
  else
    ochusha_hypertext_buffer_remove_link_info(buffer, info);

  buffer->link_tag = NULL;
  ochusha_hypertext_buffer_pop_tags(buffer, tag);
}
