/*
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: hypertext_view.c,v 1.23 2009/01/01 07:46:51 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#include "libochushagtk/hypertext_view.h"
#include "libochushagtk/hypertext_buffer.h"
#include "libochushagtk/marshal.h"
#include "libochushagtk/searchable_private.h"

#include <gdk/gdkkeysyms.h>
#include <glib.h>
#include <gtk/gtk.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


static void ochusha_hypertext_view_class_init(OchushaHypertextViewClass *klass);
static void ochusha_hypertext_view_searchable_init(OchushaSearchableIface *iface);
static void ochusha_hypertext_view_init(OchushaHypertextView *view);
static void ochusha_hypertext_view_finalize(GObject *object);
static void ochusha_hypertext_view_destroy(GtkObject *object);
static void ochusha_hypertext_view_realize(GtkWidget *widget);

static void emit_link_mouse_over(OchushaHypertextView *view,
				 GdkEventMotion *event, const gchar *link);
static void emit_link_mouse_out(OchushaHypertextView *view,
				GdkEventAny *event);

static gboolean ochusha_hypertext_view_motion_notify(GtkWidget *widget,
						     GdkEventMotion *event);
static gboolean ochusha_hypertext_view_leave_notify(GtkWidget *widget,
						    GdkEventCrossing *event);
static gboolean ochusha_hypertext_view_button_press(GtkWidget *widget,
						    GdkEventButton *event);
static gboolean ochusha_hypertext_view_button_release(GtkWidget *widget,
						      GdkEventButton *event);

static gboolean ochusha_hypertext_view_find_next(OchushaSearchable *searchable,
						 OchushaSearchKey *key);
static void ochusha_hypertext_view_terminate_search(OchushaSearchable *searchable,
						    OchushaSearchKey *key);


GType
ochusha_hypertext_view_get_type(void)
{
  static GType hypertext_view_type = 0;

  if (hypertext_view_type == 0)
    {
      static const GTypeInfo hypertext_view_info =
	{
	  sizeof(OchushaHypertextViewClass),
	  NULL, /* base_init */
	  NULL, /* base_finalize */
	  (GClassInitFunc)ochusha_hypertext_view_class_init,
	  NULL, /* class_finalize */
	  NULL, /* class_data */
	  sizeof(OchushaHypertextView),
	  0,	/* n_preallocs */
	  (GInstanceInitFunc)ochusha_hypertext_view_init
	};

      static const GInterfaceInfo searchable_info =
	{
	  (GInterfaceInitFunc)ochusha_hypertext_view_searchable_init,
	  NULL,
	  NULL
	};

      hypertext_view_type = g_type_register_static(GTK_TYPE_TEXT_VIEW,
						   "OchushaHypertextView",
						   &hypertext_view_info, 0);

      g_type_add_interface_static(hypertext_view_type, OCHUSHA_TYPE_SEARCHABLE,
				  &searchable_info);
    }
  
  return hypertext_view_type;
}


enum {
  LINK_MOUSE_OVER_SIGNAL = 0,
  LINK_MOUSE_OUT_SIGNAL,
  LINK_MOUSE_PRESS_SIGNAL,
  LINK_MOUSE_RELEASE_SIGNAL,
  START_INTERACTIVE_SEARCH_SIGNAL,
  LAST_SIGNAL
};


static GtkTextViewClass *parent_class = NULL;
static gint hypertext_view_signals[LAST_SIGNAL] = { 0, 0, 0, 0, 0 };


static void
ochusha_hypertext_view_class_init(OchushaHypertextViewClass *klass)
{
  GObjectClass *o_class = (GObjectClass *)klass;
  GtkObjectClass *object_class = (GtkObjectClass *)klass;
  GtkWidgetClass *widget_class = (GtkWidgetClass *)klass;

  parent_class = g_type_class_peek_parent(klass);

  /* GObject signals */
  o_class->finalize = ochusha_hypertext_view_finalize;

  /* GtkObject signals */
  object_class->destroy = ochusha_hypertext_view_destroy;

  hypertext_view_signals[LINK_MOUSE_OVER_SIGNAL] =
    g_signal_new("link_mouse_over",
		 G_TYPE_FROM_CLASS(klass),
		 G_SIGNAL_RUN_LAST,
		 G_STRUCT_OFFSET(OchushaHypertextViewClass, link_mouse_over),
		 NULL, NULL,
		 libochushagtk_marshal_VOID__BOXED_STRING,
		 G_TYPE_NONE, 2,
		 GDK_TYPE_EVENT | G_SIGNAL_TYPE_STATIC_SCOPE,
		 G_TYPE_STRING);
  hypertext_view_signals[LINK_MOUSE_OUT_SIGNAL] =
    g_signal_new("link_mouse_out",
		 G_TYPE_FROM_CLASS(klass),
		 G_SIGNAL_RUN_LAST,
		 G_STRUCT_OFFSET(OchushaHypertextViewClass, link_mouse_out),
		 NULL, NULL,
		 libochushagtk_marshal_VOID__BOXED,
		 G_TYPE_NONE, 1,
		 GDK_TYPE_EVENT | G_SIGNAL_TYPE_STATIC_SCOPE);
  hypertext_view_signals[LINK_MOUSE_PRESS_SIGNAL] =
    g_signal_new("link_mouse_press",
		 G_TYPE_FROM_CLASS(klass),
		 G_SIGNAL_RUN_LAST,
		 G_STRUCT_OFFSET(OchushaHypertextViewClass, link_mouse_press),
		 NULL, NULL,
		 libochushagtk_marshal_BOOLEAN__BOXED_STRING,
		 G_TYPE_BOOLEAN, 2,
		 GDK_TYPE_EVENT | G_SIGNAL_TYPE_STATIC_SCOPE,
		 G_TYPE_STRING);
  hypertext_view_signals[LINK_MOUSE_RELEASE_SIGNAL] =
    g_signal_new("link_mouse_release",
		 G_TYPE_FROM_CLASS(klass),
		 G_SIGNAL_RUN_LAST,
		 G_STRUCT_OFFSET(OchushaHypertextViewClass,
				 link_mouse_release),
		 NULL, NULL,
		 libochushagtk_marshal_BOOLEAN__BOXED_STRING,
		 G_TYPE_BOOLEAN, 2,
		 GDK_TYPE_EVENT | G_SIGNAL_TYPE_STATIC_SCOPE,
		 G_TYPE_STRING);
  hypertext_view_signals[START_INTERACTIVE_SEARCH_SIGNAL] =
    g_signal_new("start_interactive_search",
		 G_TYPE_FROM_CLASS(klass),
		 G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
		 G_STRUCT_OFFSET(OchushaHypertextViewClass,
				 start_interactive_search),
		 NULL, NULL,
		 libochushagtk_marshal_BOOLEAN__INT,
		 G_TYPE_BOOLEAN, 1,
		 G_TYPE_INT);

  widget_class->realize = ochusha_hypertext_view_realize;
  widget_class->motion_notify_event = ochusha_hypertext_view_motion_notify;
  widget_class->leave_notify_event = ochusha_hypertext_view_leave_notify;
  widget_class->button_press_event = ochusha_hypertext_view_button_press;
  widget_class->button_release_event = ochusha_hypertext_view_button_release;

  klass->link_mouse_over = NULL;
  klass->link_mouse_out = NULL;
  klass->link_mouse_press = NULL;
  klass->link_mouse_release = NULL;

  klass->start_interactive_search = NULL;
}


static void
ochusha_hypertext_view_searchable_init(OchushaSearchableIface *iface)
{
  iface->find_next = ochusha_hypertext_view_find_next;
  iface->terminate_search = ochusha_hypertext_view_terminate_search;
}


static void
ochusha_hypertext_view_init(OchushaHypertextView *view)
{
  view->ibeam_cursor = NULL;
  view->hand_cursor = NULL;

  gtk_text_view_set_wrap_mode(&view->text_view, GTK_WRAP_CHAR);
  gtk_text_view_set_editable(&view->text_view, FALSE);
  gtk_text_view_set_cursor_visible(&view->text_view, FALSE);

  view->last_event_offset = -1;

  view->last_match_offset = 0;
  view->search_wrapped = FALSE;
  view->last_key = NULL;
  view->last_key_id = 0;
}


static void
ochusha_hypertext_view_finalize(GObject *object)
{

#if 0
  OchushaHypertextView *view = OCHUSHA_HYPERTEXT_VIEW(object);
#endif
#if 0
  fprintf(stderr, "ochusha_hypertext_view_finalize(%p)\n", object);
#endif

  if (G_OBJECT_CLASS(parent_class)->finalize != NULL)
    (*G_OBJECT_CLASS(parent_class)->finalize)(object);
}


static void
ochusha_hypertext_view_destroy(GtkObject *object)
{
  OchushaHypertextView *view = OCHUSHA_HYPERTEXT_VIEW(object);

#if 0
  fprintf(stderr, "ochusha_hypertext_view_destroy(%p)\n", object);
#endif

  if (view->ibeam_cursor != NULL)
    {
      gdk_cursor_unref(view->ibeam_cursor);
      view->ibeam_cursor = NULL;
    }

  if (view->hand_cursor != NULL)
    {
      gdk_cursor_unref(view->hand_cursor);
      view->hand_cursor = NULL;
    }

  if (GTK_OBJECT_CLASS(parent_class)->destroy != NULL)
    (*GTK_OBJECT_CLASS(parent_class)->destroy)(object);

}


static void
ochusha_hypertext_view_realize(GtkWidget *widget)
{
  GdkDisplay *display;
  GdkWindow *window;
  OchushaHypertextView *view;

  if (GTK_WIDGET_CLASS(parent_class)->realize)
    (*GTK_WIDGET_CLASS(parent_class)->realize)(widget);

  window = gtk_text_view_get_window(GTK_TEXT_VIEW(widget),
				    GTK_TEXT_WINDOW_TEXT);
  gdk_window_set_events(window,
			gdk_window_get_events(window)
			| GDK_POINTER_MOTION_MASK
			| GDK_ENTER_NOTIFY_MASK
			| GDK_LEAVE_NOTIFY_MASK);

  display = gtk_widget_get_display(widget);
  view = OCHUSHA_HYPERTEXT_VIEW(widget);

  if (view->ibeam_cursor != NULL)
    gdk_cursor_unref(view->ibeam_cursor);
  view->ibeam_cursor = gdk_cursor_new_for_display(display, GDK_XTERM);

  if (view->hand_cursor != NULL)
    gdk_cursor_unref(view->hand_cursor);
  view->hand_cursor = gdk_cursor_new_for_display(display, GDK_HAND2);
}


static gboolean
ochusha_hypertext_view_get_iter_from_event(GtkTextView *text_view,
					   GtkTextIter *iter,
					   GdkEventAny *event)
{
  gint buf_x;
  gint buf_y;

  switch (event->type)
    {
    case GDK_MOTION_NOTIFY:
      {
	GdkEventMotion *me = (GdkEventMotion *)event;
	gtk_text_view_window_to_buffer_coords(text_view, GTK_TEXT_WINDOW_TEXT,
					      me->x, me->y, &buf_x, &buf_y);
	break;
      }

    case GDK_BUTTON_PRESS:
    case GDK_2BUTTON_PRESS:
    case GDK_3BUTTON_PRESS:
    case GDK_BUTTON_RELEASE:
      {
	GdkEventButton *be = (GdkEventButton *)event;
	gtk_text_view_window_to_buffer_coords(text_view, GTK_TEXT_WINDOW_TEXT,
					      be->x, be->y, &buf_x, &buf_y);
	break;
      }

    default:
      return FALSE;
    }

  gtk_text_view_get_iter_at_location(text_view, iter, buf_x, buf_y);
  return TRUE;
}


static void
emit_link_mouse_over(OchushaHypertextView *view, GdkEventMotion *event,
		     const gchar *link)
{
  gdk_window_set_cursor(gtk_text_view_get_window(&view->text_view,
						 GTK_TEXT_WINDOW_TEXT),
			view->hand_cursor);

  g_signal_emit(view,
		hypertext_view_signals[LINK_MOUSE_OVER_SIGNAL],
		0,
		event,
		link);
}


static void
emit_link_mouse_out(OchushaHypertextView *view, GdkEventAny *event)
{
  gdk_window_set_cursor(gtk_text_view_get_window(&view->text_view,
						 GTK_TEXT_WINDOW_TEXT),
			view->ibeam_cursor);

  g_signal_emit(view,
		hypertext_view_signals[LINK_MOUSE_OUT_SIGNAL],
		0,
		event);
}


static gboolean
ochusha_hypertext_view_motion_notify(GtkWidget *widget, GdkEventMotion *event)
{
  GtkTextView *text_view;
  OchushaHypertextView *view;
  GtkTextIter iter;

  g_return_val_if_fail(OCHUSHA_IS_HYPERTEXT_VIEW(widget)
		       && event->type == GDK_MOTION_NOTIFY, FALSE);

  text_view = GTK_TEXT_VIEW(widget);
  view = OCHUSHA_HYPERTEXT_VIEW(widget);

  if (ochusha_hypertext_view_get_iter_from_event(text_view, &iter,
						 (GdkEventAny *)event))
    {
      GtkTextBuffer *text_buffer = gtk_text_view_get_buffer(text_view);
      GtkTextIter start;
      GtkTextIter end;
      gchar *text;

      if (gtk_text_buffer_get_selection_bounds(text_buffer, &start, &end)
	  && gtk_text_iter_in_range(&iter, &start, &end)
	  && (text = gtk_text_iter_get_visible_text(&start, &end)) != NULL)
	{
	  gint offset = gtk_text_iter_get_offset(&start);
	  if (offset != view->last_event_offset)
	    {
	      if (view->last_event_offset != -1)
		emit_link_mouse_out(view, (GdkEventAny *)event);

	      view->last_event_offset = offset;
	      emit_link_mouse_over(view, event, text);
	    }

	  g_free(text);
	}
      else
	{
	  OchushaHypertextBuffer *buffer
	    = OCHUSHA_HYPERTEXT_BUFFER(text_buffer);
	  gint offset = ochusha_hypertext_buffer_get_link_offset(buffer, &iter);
	  if (offset == -1)
	    {
	      if (view->last_event_offset != -1)
		{
		  emit_link_mouse_out(view, (GdkEventAny *)event);
		  view->last_event_offset = -1;
		}
	    }
	  else if (offset != view->last_event_offset)
	    {
	      const gchar *link;

	      if (view->last_event_offset != -1)
		emit_link_mouse_out(view, (GdkEventAny *)event);

	      link = ochusha_hypertext_buffer_get_link_at_offset(buffer,
								 offset);
	      if (link != NULL)
		{
		  view->last_event_offset = offset;
		  emit_link_mouse_over(view, event, link);
		}
	      else
		view->last_event_offset = -1;
	    }
	}
    }
  else if (view->last_event_offset != -1)
    {
      emit_link_mouse_out(view, (GdkEventAny *)event);
      view->last_event_offset = -1;
    }

  if (GTK_WIDGET_CLASS(parent_class)->motion_notify_event != NULL)
    return (*GTK_WIDGET_CLASS(parent_class)->motion_notify_event)(widget,
								  event);
  return FALSE;
}


static gboolean
ochusha_hypertext_view_leave_notify(GtkWidget *widget, GdkEventCrossing *event)
{
  OchushaHypertextView *view = OCHUSHA_HYPERTEXT_VIEW(widget);

  if (view->last_event_offset != -1)
    {
      emit_link_mouse_out(view, (GdkEventAny *)event);
      view->last_event_offset = -1;
    }

  if (GTK_WIDGET_CLASS(parent_class)->leave_notify_event != NULL)
    return (*GTK_WIDGET_CLASS(parent_class)->leave_notify_event)(widget,
								 event);
  return FALSE;
}


static gboolean
ochusha_hypertext_view_button_press(GtkWidget *widget, GdkEventButton *event)
{
  gboolean result = FALSE;
  GtkTextView *text_view;
  GtkTextIter iter;

  g_return_val_if_fail(OCHUSHA_IS_HYPERTEXT_VIEW(widget)
		       && (event->type == GDK_BUTTON_PRESS
			   || event->type == GDK_2BUTTON_PRESS
			   || event->type == GDK_3BUTTON_PRESS), FALSE);
  text_view = GTK_TEXT_VIEW(widget);

  if (ochusha_hypertext_view_get_iter_from_event(text_view, &iter,
						 (GdkEventAny *)event))
    {
      OchushaHypertextBuffer *buffer
	= OCHUSHA_HYPERTEXT_BUFFER(gtk_text_view_get_buffer(text_view));
      gint offset = ochusha_hypertext_buffer_get_link_offset(buffer, &iter);
      OchushaHypertextView *view = OCHUSHA_HYPERTEXT_VIEW(widget);

      if (view->last_event_offset != -1 && view->last_event_offset != offset)
	{
	  emit_link_mouse_out(view, (GdkEventAny *)event);
	  view->last_event_offset = -1;
	}

      if (offset != -1)
	{
	  const gchar *link
	    = ochusha_hypertext_buffer_get_link_at_offset(buffer, offset);

	  if (link != NULL)
	    g_signal_emit(text_view,
			  hypertext_view_signals[LINK_MOUSE_PRESS_SIGNAL],
			  0,
			  event,
			  link,
			  &result);
	}
    }

  if (!result && GTK_WIDGET_CLASS(parent_class)->button_press_event != NULL)
    return (*GTK_WIDGET_CLASS(parent_class)->button_press_event)(widget,
								 event);
  return result;
}


static gboolean
ochusha_hypertext_view_button_release(GtkWidget *widget,
				      GdkEventButton *event)
{
  gboolean result = FALSE;
  GtkTextView *text_view;
  GtkTextIter iter;

  g_return_val_if_fail(GTK_IS_TEXT_VIEW(widget)
		       && event->type == GDK_BUTTON_RELEASE, FALSE);
  text_view = GTK_TEXT_VIEW(widget);

  if (ochusha_hypertext_view_get_iter_from_event(text_view, &iter,
						 (GdkEventAny *)event))
    {
      OchushaHypertextBuffer *buffer
	= OCHUSHA_HYPERTEXT_BUFFER(gtk_text_view_get_buffer(text_view));
      gint offset = ochusha_hypertext_buffer_get_link_offset(buffer, &iter);
      OchushaHypertextView *view = OCHUSHA_HYPERTEXT_VIEW(widget);

      if (view->last_event_offset != -1 && view->last_event_offset != offset)
	{
	  emit_link_mouse_out(view, (GdkEventAny *)event);
	  view->last_event_offset = -1;
	}

      if (offset != -1)
	{
	  const gchar *link
	    = ochusha_hypertext_buffer_get_link_at_offset(buffer, offset);

	  if (link != NULL)
	    g_signal_emit(text_view,
			  hypertext_view_signals[LINK_MOUSE_RELEASE_SIGNAL],
			  0,
			  event,
			  link,
			  &result);
	}
    }

  if (GTK_WIDGET_CLASS(parent_class)->button_release_event != NULL)
    return (*GTK_WIDGET_CLASS(parent_class)->button_release_event)(widget,
								   event);
  return result;
}


GtkWidget *
ochusha_hypertext_view_new(OchushaHypertextBuffer *buffer)
{
  GtkTextView *text_view = g_object_new(OCHUSHA_TYPE_HYPERTEXT_VIEW, NULL);
  gtk_text_view_set_buffer(text_view, GTK_TEXT_BUFFER(buffer));

  return GTK_WIDGET(text_view);
}


gint
ochusha_hypertext_view_append_widget(OchushaHypertextView *view,
				     GtkWidget *widget,
				     gint hborder, gint vborder)
{
  GtkTextBuffer *text_buffer;
  OchushaHypertextBuffer *buffer;
  gint offset;
  GtkTextChildAnchor *anchor;
  GtkWidget *parent_widget;

  g_return_val_if_fail(OCHUSHA_IS_HYPERTEXT_VIEW(view), -1);
  g_return_val_if_fail(GTK_IS_WIDGET(widget), -1);

  text_buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(view));
  g_return_val_if_fail(text_buffer != NULL, -1);
  g_return_val_if_fail(OCHUSHA_IS_HYPERTEXT_BUFFER(text_buffer), -1);

  buffer = OCHUSHA_HYPERTEXT_BUFFER(text_buffer);

  offset = ochusha_hypertext_buffer_get_current_offset(buffer);
  anchor = ochusha_hypertext_buffer_append_child(buffer, widget);

  /* parent_widgetν */
  if (hborder == 0 && vborder == 0)
    parent_widget = widget;	// borderʤϤΤޤ޻Ȥ
  else if (hborder < vborder)
    {
      guint padding = vborder - hborder;
      parent_widget = gtk_vbox_new(FALSE, 0);
      gtk_container_set_border_width(GTK_CONTAINER(parent_widget), hborder);
      gtk_widget_show(widget);
      gtk_box_pack_start(GTK_BOX(parent_widget), widget, TRUE, TRUE, padding);
    }
  else
    {
      guint padding = hborder - vborder;
      parent_widget = gtk_hbox_new(FALSE, 0);
      gtk_container_set_border_width(GTK_CONTAINER(parent_widget), vborder);
      gtk_widget_show(widget);
      gtk_box_pack_start(GTK_BOX(parent_widget), widget, TRUE, TRUE, padding);
    }

  gtk_widget_show(parent_widget);
  gtk_text_view_add_child_at_anchor(GTK_TEXT_VIEW(view), parent_widget, anchor);

  return offset;
}


gint
ochusha_hypertext_view_insert_widget_at_offset(OchushaHypertextView *view,
					       GtkWidget *widget,
					       gint hborder, gint vborder,
					       gint offset)
{
  GtkWidget *parent_widget;
  GtkTextBuffer *text_buffer;
  OchushaHypertextBuffer *buffer;
  GtkTextChildAnchor *anchor;

  g_return_val_if_fail(OCHUSHA_IS_HYPERTEXT_VIEW(view), -1);
  g_return_val_if_fail(GTK_IS_WIDGET(widget), -1);

  text_buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(view));
  g_return_val_if_fail(text_buffer != NULL, -1);
  g_return_val_if_fail(OCHUSHA_IS_HYPERTEXT_BUFFER(text_buffer), -1);

  buffer = OCHUSHA_HYPERTEXT_BUFFER(text_buffer);

  anchor = ochusha_hypertext_buffer_insert_child_at_offset(buffer,
							   widget, offset);

  /* parent_widgetν */
  if (hborder == 0 && vborder == 0)
    parent_widget = widget;	// borderʤϤΤޤ޻Ȥ
  else if (hborder < vborder)
    {
      guint padding = vborder - hborder;
      parent_widget = gtk_vbox_new(FALSE, 0);
      gtk_container_set_border_width(GTK_CONTAINER(parent_widget), hborder);
      gtk_widget_show(widget);
      gtk_box_pack_start(GTK_BOX(parent_widget), widget, TRUE, TRUE, padding);
    }
  else
    {
      guint padding = hborder - vborder;
      parent_widget = gtk_hbox_new(FALSE, 0);
      gtk_container_set_border_width(GTK_CONTAINER(parent_widget), vborder);
      gtk_widget_show(widget);
      gtk_box_pack_start(GTK_BOX(parent_widget), widget, TRUE, TRUE, padding);
    }

  gtk_widget_show(parent_widget);
  gtk_text_view_add_child_at_anchor(GTK_TEXT_VIEW(view), parent_widget, anchor);

  return offset;
}


gint
ochusha_hypertext_view_remove_widget(OchushaHypertextView *view,
				     GtkWidget *widget)
{
  GtkTextBuffer *text_buffer;
  OchushaHypertextBuffer *buffer;

  g_return_val_if_fail(OCHUSHA_IS_HYPERTEXT_VIEW(view), -1);

  text_buffer = gtk_text_view_get_buffer(&view->text_view);
  buffer = OCHUSHA_HYPERTEXT_BUFFER(text_buffer);

  return ochusha_hypertext_buffer_remove_child(buffer, widget);
}


void
ochusha_hypertext_view_scroll_to_start(OchushaHypertextView *view)
{
  GtkTextIter iter;
  GtkTextBuffer *buffer;

  g_return_if_fail(OCHUSHA_IS_HYPERTEXT_VIEW(view));

  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(view));
  g_return_if_fail(buffer != NULL);

  gtk_text_buffer_get_start_iter(buffer, &iter);
  gtk_text_buffer_place_cursor(buffer, &iter);
  gtk_text_view_scroll_to_iter(GTK_TEXT_VIEW(view),
			       &iter, 0.0, FALSE, 0.0, 0.0);
}


void
ochusha_hypertext_view_scroll_to_end(OchushaHypertextView *view)
{
  GtkTextIter iter;
  GtkTextBuffer *buffer;

  g_return_if_fail(OCHUSHA_IS_HYPERTEXT_VIEW(view));

  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(view));
  g_return_if_fail(buffer != NULL);

  gtk_text_buffer_get_end_iter(buffer, &iter);
  gtk_text_buffer_place_cursor(buffer, &iter);
  gtk_text_view_scroll_to_iter(GTK_TEXT_VIEW(view),
			       &iter, 0.0, FALSE, 0.0, 0.0);
}


gint
ochusha_hypertext_view_scroll_to_offset(OchushaHypertextView *view,
					gint offset)
{
  GtkTextIter iter;
  GtkTextBuffer *text_buffer;
  gint previous_offset;

  g_return_val_if_fail(OCHUSHA_IS_HYPERTEXT_VIEW(view), 0);

  text_buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(view));

  gtk_text_view_place_cursor_onscreen(GTK_TEXT_VIEW(view));
  gtk_text_buffer_get_iter_at_mark(text_buffer, &iter,
				   gtk_text_buffer_get_insert(text_buffer));
  previous_offset = gtk_text_iter_get_offset(&iter);

  gtk_text_buffer_get_iter_at_offset(text_buffer, &iter, offset);
  gtk_text_buffer_place_cursor(text_buffer, &iter);
  gtk_text_view_scroll_to_iter(GTK_TEXT_VIEW(view),
			       &iter, 0.0, FALSE, 0.0, 0.0);

  return previous_offset;
}


void
ochusha_hypertext_view_scroll_to_mark(OchushaHypertextView *view,
				      GtkTextMark *mark)
{
  GtkTextIter iter;
  GtkTextBuffer *buffer;

  g_return_if_fail(OCHUSHA_IS_HYPERTEXT_VIEW(view) && mark != NULL);

  buffer = gtk_text_mark_get_buffer(mark);
  g_return_if_fail(buffer != NULL);

  gtk_text_buffer_get_iter_at_mark(buffer, &iter, mark);
  gtk_text_buffer_place_cursor(buffer, &iter);

  gtk_text_view_scroll_to_mark(GTK_TEXT_VIEW(view), mark, 0.1, TRUE, 0.0, 0.1);
}


gint
ochusha_hypertext_view_get_visible_offset(OchushaHypertextView *view)
{
  GdkRectangle visible_rect;
  GtkTextIter iter;

  g_return_val_if_fail(OCHUSHA_IS_HYPERTEXT_VIEW(view), 0);

  gtk_text_view_get_visible_rect(GTK_TEXT_VIEW(view), &visible_rect);
  gtk_text_view_get_iter_at_location(GTK_TEXT_VIEW(view), &iter,
				     visible_rect.x, visible_rect.y);
  return gtk_text_iter_get_offset(&iter);
}


void
ochusha_hypertext_view_make_offset_visible(OchushaHypertextView *view,
					   gint offset)
{
  GtkTextIter iter;
  GtkTextBuffer *text_buffer;

  g_return_if_fail(OCHUSHA_IS_HYPERTEXT_VIEW(view));

  text_buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(view));

  gtk_text_buffer_get_iter_at_offset(text_buffer, &iter, offset);
  gtk_text_buffer_place_cursor(text_buffer, &iter);
  gtk_text_view_scroll_to_iter(GTK_TEXT_VIEW(view),
			       &iter, 0.0, TRUE, 0.0, 0.0);
}


static void
ochusha_hypertext_view_scroll_to_offset_range_visible(OchushaHypertextView *view,
						      GtkTextBuffer *buffer,
						      gint start_offset,
						      gint end_offset)
{
  /* ޥåʬ򤷡ʬɽ褦˥ */
  GtkTextIter iter;
  GtkTextMark *insert_mark = gtk_text_buffer_get_mark(buffer, "insert");
  GtkTextMark *selection_bound_mark
    = gtk_text_buffer_get_mark(buffer, "selection_bound");

  gtk_text_buffer_get_iter_at_offset(buffer, &iter, start_offset);
  gtk_text_buffer_move_mark(buffer, insert_mark, &iter);
  gtk_text_buffer_get_iter_at_offset(buffer, &iter, end_offset);
  gtk_text_buffer_move_mark(buffer, selection_bound_mark, &iter);
  gtk_text_view_scroll_to_mark(GTK_TEXT_VIEW(view), insert_mark,
			       0.1, TRUE, 0.2, 0.2);
}


static gboolean
ochusha_hypertext_view_find_forward_first_match(OchushaHypertextView *view,
						GtkTextIter *start_iter,
						GtkTextIter *end_iter,
						gint *match_start_offset_p,
						gint *match_end_offset_p)
{
  OchushaSearchKey *key = view->last_key;
  gboolean result = FALSE;
  gchar *text;
#if 0
  fprintf(stderr, "find_forward_first_match: key=\"%s\"\n", key->key_string);
#endif

  /*
   * MEMO: gtk_text_iter_get_text()䤽֤ϥԡ֤Ƥ뤳Ȥ
   *       աХåեΤоݤ˸ȤʤȡХåեݤȤΥԡ
   *       롣ٹʤΤǤʺ٤ʤȤ̵Ǥ
   */
  text = gtk_text_iter_get_slice(start_iter, end_iter);
  g_return_val_if_fail(text != NULL, FALSE);

  result = ochusha_search_key_match_text(key, text, -1);

  if (result)
    {
      gint offset = gtk_text_iter_get_offset(start_iter);
      const gchar *match_begin
	= ochusha_search_key_last_match_begin(key);

      *match_start_offset_p = g_utf8_strlen(text, match_begin - text) + offset;
      *match_end_offset_p
	= g_utf8_strlen(match_begin,
			ochusha_search_key_last_match_len(key))
	+ (*match_start_offset_p);

#if 0
      fprintf(stderr, "===FORWARD==-=====================================\n");
      fprintf(stderr, "offset=%d, text=%p, match_begin=%p, match_len=%d\n",
	      offset, text, match_begin,
	      ochusha_search_key_last_match_len(key));
      fprintf(stderr, "match_start_offset=%d, match_end_offset=%d\n",
	      *match_start_offset_p, *match_end_offset_p);
#endif
    }

  g_free(text);

  return result;
}


static gboolean
ochusha_hypertext_view_find_backward_first_match(OchushaHypertextView *view,
						 GtkTextIter *start_iter,
						 GtkTextIter *end_iter,
						 gint *match_start_offset_p,
						 gint *match_end_offset_p)
{
  OchushaSearchKey *key = view->last_key;
  gboolean result = FALSE;
  gchar *text;
  /*
   * MEMO: gtk_text_iter_get_text()䤽֤ϥԡ֤Ƥ뤳Ȥ
   *       աХåեΤоݤ˸ȤʤȡХåեݤȤΥԡ
   *       롣ٹʤΤǤʺ٤ʤȤ̵Ǥ
   */
  text = gtk_text_iter_get_slice(start_iter, end_iter);
  g_return_val_if_fail(text != NULL, FALSE);

  result = ochusha_search_key_match_text(key, text, -1);

  if (result)
    {
      gint offset = gtk_text_iter_get_offset(start_iter);
      const gchar *match_begin
	= ochusha_search_key_last_match_begin(key);

      *match_start_offset_p = g_utf8_strlen(text, match_begin - text) + offset;
      *match_end_offset_p
	= g_utf8_strlen(match_begin,
			ochusha_search_key_last_match_len(key))
	+ (*match_start_offset_p);

#if 0
      fprintf(stderr, "===BACKWARD=======================================\n");
      fprintf(stderr, "offset=%d, text=%p, match_begin=%p, match_len=%d\n",
	      offset, text, match_begin,
	      ochusha_search_key_last_match_len(key));
      fprintf(stderr, "match_start_offset=%d, match_end_offset=%d\n",
	      *match_start_offset_p, *match_end_offset_p);
#endif
    }

  g_free(text);

  return result;
}


static gboolean
ochusha_hypertext_view_do_search_forward(OchushaHypertextView *view,
					 gboolean include_current_pos)
{
  OchushaSearchKey *key = view->last_key;
  GtkTextBuffer *text_buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(view));
  GtkTextIter start;
  GtkTextIter end;
  gint start_offset = 0;
  gint end_offset = 0;

  if (key->key_string == NULL)
    return FALSE;

  if (key->use_regexp && !key->regexp_available)
    return FALSE;

  gtk_text_buffer_get_iter_at_offset(text_buffer, &start,
				     view->last_match_offset);
  if (!include_current_pos)
    {
      if (!gtk_text_iter_forward_char(&start))
	{
	  view->last_match_offset = 0;
	  view->search_wrapped = TRUE;
	  return FALSE;
	}
    }
  gtk_text_buffer_get_end_iter(text_buffer, &end);

  if (ochusha_hypertext_view_find_forward_first_match(view, &start, &end,
						       &start_offset,
						       &end_offset))
    {
      ochusha_hypertext_view_scroll_to_offset_range_visible(view, text_buffer,
							    start_offset,
							    end_offset);
      view->last_match_offset = start_offset;
      return TRUE;
    }
  else
    {
      view->last_match_offset = 0;
      view->search_wrapped = TRUE;
    }

  return FALSE;
}


static gboolean
ochusha_hypertext_view_do_search_backward(OchushaHypertextView *view,
					  gboolean include_current_pos)
{
  OchushaSearchKey *key = view->last_key;
  GtkTextBuffer *text_buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(view));
  GtkTextIter start;
  GtkTextIter end;
  gint start_offset = 0;
  gint end_offset = 0;

  if (key->key_string == NULL)
    return FALSE;

  if (key->use_regexp && !key->regexp_available)
    return FALSE;

  gtk_text_buffer_get_start_iter(text_buffer, &start);
  gtk_text_buffer_get_iter_at_offset(text_buffer, &end,
				     view->last_match_offset);

  if (include_current_pos)
    {
      if (key->use_regexp)
	{
	  /* ʤ */
	  if (!gtk_text_iter_forward_line(&end))
	    gtk_text_buffer_get_end_iter(text_buffer, &end);
	}
      else
	gtk_text_iter_forward_chars(&end,
				    g_utf8_strlen(key->key_string, -1));
    }
  else
    {
      if (!gtk_text_iter_backward_char(&end))
	{
	  gtk_text_buffer_get_end_iter(text_buffer, &end);
	  view->last_match_offset = gtk_text_iter_get_offset(&end);
	  view->search_wrapped = TRUE;
	  return FALSE;
	}
    }

  if (ochusha_hypertext_view_find_backward_first_match(view, &start, &end,
							&start_offset,
							&end_offset))
    {
      ochusha_hypertext_view_scroll_to_offset_range_visible(view, text_buffer,
							    start_offset,
							    end_offset);
      view->last_match_offset = start_offset;
      return TRUE;
    }
  else
    {
      gtk_text_buffer_get_end_iter(text_buffer, &end);
      view->last_match_offset = gtk_text_iter_get_offset(&end);
      view->search_wrapped = TRUE;
    }

  return FALSE;
}


static void
ochusha_hypertext_view_unselect_text(OchushaHypertextView *view)
{
  GtkTextBuffer *text_buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(view));
  GtkTextMark *insert_mark = gtk_text_buffer_get_mark(text_buffer, "insert");
  GtkTextMark *selection_bound_mark
    = gtk_text_buffer_get_mark(text_buffer, "selection_bound");
  GtkTextIter iter;

  gtk_text_buffer_get_iter_at_mark(text_buffer, &iter, insert_mark);
  gtk_text_buffer_move_mark(text_buffer, selection_bound_mark, &iter);
}


static gboolean
ochusha_hypertext_view_find_next(OchushaSearchable *searchable,
				 OchushaSearchKey *key)
{
  gboolean result;
  OchushaHypertextView *view = OCHUSHA_HYPERTEXT_VIEW(searchable);
  gboolean include_current_pos
    = (view->last_key != key || key->key_id != view->last_key_id
       || view->search_wrapped);

  g_return_val_if_fail(view->last_key == NULL || view->last_key == key, FALSE);

  view->last_key = key;
  view->last_key_id = key->key_id;
  view->search_wrapped = FALSE;

  if (key->key_string == NULL || *key->key_string == '\0')
    {
      ochusha_hypertext_view_unselect_text(view);
      return TRUE;
    }

  if (key->direction == OCHUSHA_SEARCH_DIRECTION_FORWARD)
    result = ochusha_hypertext_view_do_search_forward(view,
						      include_current_pos);
  else
    result = ochusha_hypertext_view_do_search_backward(view,
						       include_current_pos);

  if (!result)
    ochusha_hypertext_view_unselect_text(view);

  return result;
}


static void
ochusha_hypertext_view_terminate_search(OchushaSearchable *searchable,
					OchushaSearchKey *key)
{
  OchushaHypertextView *view = OCHUSHA_HYPERTEXT_VIEW(searchable);
  view->last_match_offset = 0;
  view->last_key = NULL;
  view->last_key_id = 0;
  view->search_wrapped = FALSE;
  ochusha_hypertext_view_unselect_text(view);
}


gchar *
ochusha_hypertext_view_get_selected_text(OchushaHypertextView *view)
{
  GtkTextBuffer *buffer;
  GtkTextIter start;
  GtkTextIter end;

  g_return_val_if_fail(OCHUSHA_IS_HYPERTEXT_VIEW(view), NULL);
  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(view));

  g_return_val_if_fail(buffer != NULL, NULL);

  if (!gtk_text_buffer_get_selection_bounds(buffer, &start, &end))
    return NULL;	/* 򤵤Ƥʤ */

  return gtk_text_iter_get_visible_text(&start, &end);
}

