/*
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: menu_action.c,v 1.5 2009/01/01 07:46:51 fuyu Exp $
 */

#include "config.h"

#include "libochushagtk/marshal.h"
#include "libochushagtk/menu_action.h"

#include <gtk/gtk.h>


static void ochusha_menu_action_class_init(OchushaMenuActionClass *klass);
static void ochusha_menu_action_init(OchushaMenuAction *action);
static void ochusha_menu_action_finalize(GObject *object);

static void activate_cb(OchushaTreeMenuHelper *helper, GtkTreeModel *model,
			GtkTreeIter *iter, GtkTreePath *path, guint button,
			OchushaMenuAction *action);

static GtkWidget *create_menu_item(GtkAction *action);


GType
ochusha_menu_action_get_type(void)
{
  static GType menu_action_type = 0;

  if (menu_action_type == 0)
    {
      static const GTypeInfo menu_action_info =
	{
	  sizeof(OchushaMenuActionClass),
	  NULL,	/* base_init */
	  NULL, /* base_finalize */
	  (GClassInitFunc)ochusha_menu_action_class_init,
	  NULL,	/* class_finalize */
	  NULL,	/* class_data */
	  sizeof(OchushaMenuAction),
	  0,	/* n_preallocs */
	  (GInstanceInitFunc)ochusha_menu_action_init,
	};

      menu_action_type = g_type_register_static(GTK_TYPE_ACTION,
						"OchushaMenuAction",
						&menu_action_info, 0);
    }

  return menu_action_type;
}


enum {
  MENU_ACTIVATE_SIGNAL = 0,
  LAST_SIGNAL
};


static GObjectClass *parent_class = NULL;
static gint menu_action_signals[LAST_SIGNAL] = { 0 };


static void
ochusha_menu_action_class_init(OchushaMenuActionClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
  GtkActionClass *action_class = GTK_ACTION_CLASS(klass);

  parent_class = g_type_class_peek_parent(klass);

  /* as a GObject */
  gobject_class->finalize = ochusha_menu_action_finalize;

  /* as a GtkAction */
  action_class->menu_item_type = GTK_TYPE_IMAGE_MENU_ITEM;
  action_class->toolbar_item_type = GTK_TYPE_MENU_TOOL_BUTTON;

  action_class->create_menu_item = create_menu_item;


  menu_action_signals[MENU_ACTIVATE_SIGNAL]
    = g_signal_new("menu_activate",
		   G_TYPE_FROM_CLASS(klass),
		   G_SIGNAL_RUN_LAST,
		   G_STRUCT_OFFSET(OchushaMenuActionClass, menu_activate),
		   NULL, NULL,
		   libochushagtk_marshal_VOID__OBJECT_BOXED_BOXED,
		   G_TYPE_NONE, 3,
		   GTK_TYPE_TREE_MODEL,
		   GTK_TYPE_TREE_ITER | G_SIGNAL_TYPE_STATIC_SCOPE,
		   GTK_TYPE_TREE_PATH | G_SIGNAL_TYPE_STATIC_SCOPE);

  klass->menu_activate = NULL;
}


static void
ochusha_menu_action_init(OchushaMenuAction *action)
{
  action->helper = ochusha_tree_menu_helper_new();
  g_signal_connect(action->helper, "activate",
		   G_CALLBACK(activate_cb), action);
}


static void
ochusha_menu_action_finalize(GObject *object)
{
  OchushaMenuAction *action;
  g_return_if_fail(OCHUSHA_IS_MENU_ACTION(object));

  action = OCHUSHA_MENU_ACTION(object);

  if (action->helper != NULL)
    {
      g_object_unref(action->helper);
      action->helper = NULL;
    }

  if (G_OBJECT_CLASS(parent_class)->finalize != NULL)
    (*G_OBJECT_CLASS(parent_class)->finalize)(object);
}


static void
activate_cb(OchushaTreeMenuHelper *helper, GtkTreeModel *model,
	    GtkTreeIter *iter, GtkTreePath *path, guint button,
	    OchushaMenuAction *action)
{
  g_signal_emit(action, menu_action_signals[MENU_ACTIVATE_SIGNAL],
		0,
		model, iter, path);
}


static GtkWidget *
create_menu_item(GtkAction *action)
{
  OchushaMenuAction *menu_action = OCHUSHA_MENU_ACTION(action);

  GtkWidget *menu_item = g_object_new(GTK_TYPE_IMAGE_MENU_ITEM, NULL);
  GtkWidget *submenu = gtk_menu_new();

  ochusha_tree_menu_helper_attach_tree_menu(menu_action->helper,
					    GTK_MENU_SHELL(submenu));
  gtk_widget_show(submenu);
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(menu_item), submenu);

  return menu_item;
}


OchushaMenuAction *
ochusha_menu_action_new(const gchar *name, const gchar *label,
			const gchar *tooltip, const gchar *stock_id)
{
  OchushaMenuAction *action = g_object_new(OCHUSHA_TYPE_MENU_ACTION,
					   "name", name,
					   "label", label,
					   "tooltip", tooltip,
					   "stock_id", stock_id,
					   NULL);
  return action;
}


OchushaMenuAction *
ochusha_menu_action_new_with_model(const gchar *name,
				   const gchar *label,
				   const gchar *tooltip,
				   const gchar *stock_id,
				   GtkTreeModel *model)
{
  g_return_val_if_fail(model != NULL, NULL);
  g_return_val_if_fail(GTK_IS_TREE_MODEL(model), NULL);

  OchushaMenuAction *action = g_object_new(OCHUSHA_TYPE_MENU_ACTION,
					   "name", name,
					   "label", label,
					   "tooltip", tooltip,
					   "stock_id", stock_id,
					   NULL);

  ochusha_menu_action_set_model(action, model);

  return action;
}


void
ochusha_menu_action_set_model(OchushaMenuAction *action, GtkTreeModel *model)
{
  g_return_if_fail(OCHUSHA_IS_MENU_ACTION(action));
  g_return_if_fail(GTK_IS_TREE_MODEL(model));

  ochusha_tree_menu_helper_set_model(action->helper, model);
}


GtkTreeModel *
ochusha_menu_action_get_model(OchushaMenuAction *action)
{
  g_return_val_if_fail(OCHUSHA_IS_MENU_ACTION(action), NULL);

  return ochusha_tree_menu_helper_get_model(action->helper);
}


void
ochusha_menu_action_set_menu_data_func(OchushaMenuAction *action,
				       OchushaTreeMenuDataFunc func,
				       gpointer func_data,
				       GDestroyNotify destroy)
{
  g_return_if_fail(OCHUSHA_IS_MENU_ACTION(action));

  ochusha_tree_menu_helper_set_menu_data_func(action->helper, func, func_data,
					      destroy);
}


void
ochusha_menu_action_set_row_separator_func(OchushaMenuAction *action,
					   GtkTreeViewRowSeparatorFunc func,
					   gpointer func_data,
					   GDestroyNotify destroy)
{
  g_return_if_fail(OCHUSHA_IS_MENU_ACTION(action));

  ochusha_tree_menu_helper_set_row_separator_func(action->helper, func,
						  func_data, destroy);
}


void
ochusha_menu_action_set_row_sensitive_func(OchushaMenuAction *action,
					   OchushaTreeMenuRowSensitiveFunc func,
					   gpointer func_data,
					   GDestroyNotify destroy)
{
  g_return_if_fail(OCHUSHA_IS_MENU_ACTION(action));

  ochusha_tree_menu_helper_set_row_sensitive_func(action->helper, func,
						  func_data, destroy);
}
