/*
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: regex_editor.c,v 1.15 2009/01/01 07:46:51 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#include "libochushagtk/regex_editor.h"

#include <gtk/gtk.h>

#include <oniguruma.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


static void ochusha_regex_editor_class_init(OchushaRegexEditorClass *klass);
static void ochusha_regex_editor_init(OchushaRegexEditor *editor);
static void ochusha_regex_editor_finalize(GObject *object);
static void ochusha_regex_editor_destroy(GtkObject *object);
static void ochusha_regex_editor_text_buffer_changed_cb(GtkTextBuffer *buffer,
							OchushaRegexEditor *editor);


GType
ochusha_regex_editor_get_type(void)
{
  static GType regex_editor_type = 0;

  if (regex_editor_type == 0)
    {
      static const GTypeInfo regex_editor_info =
	{
	  sizeof(OchushaRegexEditorClass),
	  NULL,	/* base_init */
	  NULL,	/* base_finalize */
	  (GClassInitFunc)ochusha_regex_editor_class_init,
	  NULL,	/* class_finalize */
	  NULL,	/* class_data */
	  sizeof(OchushaRegexEditor),
	  0,	/* n_preallocs */
	  (GInstanceInitFunc)ochusha_regex_editor_init,
	};

      regex_editor_type = g_type_register_static(GTK_TYPE_VBOX,
						 "OchushaRegexEditor",
						 &regex_editor_info, 0);
    }

  return regex_editor_type;
}


static GtkVBoxClass *parent_class;


static void
ochusha_regex_editor_class_init(OchushaRegexEditorClass *klass)
{
  GObjectClass *o_class = (GObjectClass *)klass;
  GtkObjectClass *object_class = (GtkObjectClass *)klass;

  parent_class = g_type_class_peek_parent(klass);

  o_class->finalize = ochusha_regex_editor_finalize;
  object_class->destroy = ochusha_regex_editor_destroy;
}


static void
ochusha_regex_editor_init(OchushaRegexEditor *editor)
{
  GtkTextBuffer *regex_buffer = gtk_text_buffer_new(NULL);
  GtkTextTag *error_tag
    = gtk_text_buffer_create_tag(regex_buffer, "error-line",
				 "foreground", "red",
				 "foreground-set", TRUE,
				 NULL);
  GtkWidget *regex_view = gtk_text_view_new_with_buffer(regex_buffer);
  GtkWidget *message = gtk_label_new("");
  GtkWidget *scrolled_window = gtk_scrolled_window_new(NULL, NULL);
  GtkWidget *hbox = gtk_hbox_new(FALSE, 0);

  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_window),
				 GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_widget_show(regex_view);
  gtk_container_add(GTK_CONTAINER(scrolled_window), regex_view);
  gtk_widget_show(scrolled_window);
  gtk_box_pack_start(GTK_BOX(editor), scrolled_window, TRUE, TRUE, 0);

  gtk_widget_show(message);
  gtk_box_pack_start(GTK_BOX(hbox), message, TRUE, TRUE, 0);
  gtk_widget_show(hbox);
  gtk_box_pack_start(GTK_BOX(editor), hbox, FALSE, FALSE, 0);

  editor->regex_view = GTK_TEXT_VIEW(regex_view);
  editor->regex_buffer = regex_buffer;
  editor->error_tag = error_tag;
  editor->message = GTK_LABEL(message);

  g_signal_connect(regex_buffer, "changed",
		   G_CALLBACK(ochusha_regex_editor_text_buffer_changed_cb),
		   editor);
}


static void
ochusha_regex_editor_finalize(GObject *object)
{
  OchushaRegexEditor *editor = OCHUSHA_REGEX_EDITOR(object);

  if (editor->regex_buffer != NULL)
    {
      g_object_unref(editor->regex_buffer);
      editor->regex_buffer = NULL;
    }

  if (G_OBJECT_CLASS(parent_class)->finalize != NULL)
    (*G_OBJECT_CLASS(parent_class)->finalize)(object);
}


static void
ochusha_regex_editor_destroy(GtkObject *object)
{
  if (GTK_OBJECT_CLASS(parent_class)->destroy != NULL)
    (*GTK_OBJECT_CLASS(parent_class)->destroy)(object);
}


static gboolean
ochusha_regex_editor_is_valid_regular_expression(const gchar *exp,
						 size_t exp_len)
{
  regex_t *regex;
  gboolean result = onig_new(&regex, (UChar *)exp, (UChar *)exp + exp_len,
			     ONIG_OPTION_NONE, ONIG_ENCODING_UTF8,
			     ONIG_SYNTAX_RUBY, NULL) == ONIG_NORMAL;
  onig_free(regex);
  return result;
}


static gboolean
ochusha_regex_editor_check_regular_expressions(OchushaRegexEditor *editor)
{
  GtkTextBuffer *buffer;
  gint num_lines;
  gint line;
  gboolean result = TRUE;

  g_return_val_if_fail(OCHUSHA_IS_REGEX_EDITOR(editor), FALSE);

  buffer = editor->regex_buffer;
  num_lines = gtk_text_buffer_get_line_count(buffer);
  for (line = 0; line < num_lines; line++)
    {
      GtkTextIter start_iter;
      GtkTextIter end_iter;
      gchar *line_text;
      gchar *eol_pos;

      gtk_text_buffer_get_iter_at_line(buffer, &start_iter, line);
      if (line < num_lines - 1)
	gtk_text_buffer_get_iter_at_line(buffer, &end_iter, line + 1);
      else
	gtk_text_buffer_get_end_iter(buffer, &end_iter);
      line_text = gtk_text_buffer_get_text(buffer,
					   &start_iter, &end_iter, FALSE);
      g_return_val_if_fail(line_text != NULL, FALSE);

      eol_pos = strchr(line_text, '\n');
      if (eol_pos != NULL)
	*eol_pos = '\0';

      if (ochusha_regex_editor_is_valid_regular_expression(line_text,
							   strlen(line_text)))
	{
	  gtk_text_buffer_remove_tag(buffer, editor->error_tag,
				     &start_iter, &end_iter);
	}
      else
	{
	  gtk_text_buffer_apply_tag(buffer, editor->error_tag,
				    &start_iter, &end_iter);
	  result = FALSE;
	}

      g_free(line_text);
    }

  return result;
}


static void
ochusha_regex_editor_text_buffer_changed_cb(GtkTextBuffer *buffer,
					    OchushaRegexEditor *editor)
{
  g_return_if_fail(GTK_IS_TEXT_BUFFER(buffer));
  g_return_if_fail(OCHUSHA_IS_REGEX_EDITOR(editor));

  if (ochusha_regex_editor_check_regular_expressions(editor))
    gtk_label_set_text(editor->message, "");
  else
    gtk_label_set_text(editor->message, LIB_("Invalid Regular Expression"));
}


/*
 * Ϳ줿textγƹԤbranchȤƷҤǰĤɽˤޤȤ᤿
 * ʸ֤
 *
 * ()ɽȤʤԤϼΤƤ롣
 */
static gchar *
ochusha_regex_editor_synthesize_regular_expression(const gchar *text)
{
  gchar *result = NULL;
  gchar *tmp_pos;

  g_return_val_if_fail(text != NULL, g_strdup(""));
  if (*text == '\0')
    return g_strdup("");

  tmp_pos = (gchar *)text;	/* ޤä㤤*/
  while (*tmp_pos != '\0')
    {
      gchar *end_pos = strchr(tmp_pos, '\n');
      if (end_pos == NULL)
	end_pos = tmp_pos + strlen(tmp_pos) - 1;
      else
	end_pos--;	/* '\n'μޤ */

      if (tmp_pos == end_pos)
	goto next_line;

      if (ochusha_regex_editor_is_valid_regular_expression(tmp_pos,
							   end_pos - tmp_pos))
	{
	  gchar *sub_exp = g_strndup(tmp_pos, end_pos - tmp_pos + 1);
	  if (result == NULL)
	    result = sub_exp;
	  else
	    {
	      /* ٹ롣*/
	      gchar *old_result = result;
	      result = g_strconcat(old_result, "|", sub_exp, NULL);
	      g_free(old_result);
	      g_free(sub_exp);
	    }
	}

    next_line:
      tmp_pos = end_pos + 1;
      if (*tmp_pos == '\n')
	tmp_pos++;	/* '\n'ɤФ */
    }

  if (result == NULL)
    return g_strdup("");

  return result;
}


GtkWidget *
ochusha_regex_editor_new(const gchar *regex_text)
{
  OchushaRegexEditor *editor = g_object_new(OCHUSHA_TYPE_REGEX_EDITOR,
					    NULL);

  if (regex_text != NULL)
    ochusha_regex_editor_set_regex_text(editor, regex_text);

  return GTK_WIDGET(editor);
}


void
ochusha_regex_editor_set_regex_text(OchushaRegexEditor *editor,
				    const gchar *text)
{
  GtkTextIter current_iter;
  GtkTextIter end_iter;

  g_return_if_fail(OCHUSHA_IS_REGEX_EDITOR(editor));

  gtk_text_buffer_get_start_iter(editor->regex_buffer, &current_iter);
  gtk_text_buffer_get_end_iter(editor->regex_buffer, &end_iter);

  gtk_text_buffer_delete(editor->regex_buffer, &current_iter, &end_iter);

  if (text != NULL && *text != '\0')
    {
      /* branchԤˤФ餷ƥХåե */
      int nest = 0;
      gchar *branch_start = (gchar *)text;	/* ޤä㤤 */
      gchar *tmp_pos = branch_start;
      while (*tmp_pos != '\0')
	{
	  if (*tmp_pos == '\\')
	    tmp_pos++;
	  else if (*tmp_pos == '(')
	    nest++;
	  else if (*tmp_pos == ')')
	    nest--;
	  else if (*tmp_pos == '|' && nest == 0)
	    {
	      int len = tmp_pos - branch_start;
	      if (len > 0)
		{
		  gtk_text_buffer_insert(editor->regex_buffer, &current_iter,
					 branch_start, len);
		  gtk_text_buffer_insert(editor->regex_buffer, &current_iter,
					 "\n", -1);
		}
	      branch_start = tmp_pos + 1;
	    }
	  tmp_pos++;
	}

      if (*branch_start != '\0' && branch_start < tmp_pos)
	{
	  int len = tmp_pos - branch_start;
	  if (len > 0)
	    gtk_text_buffer_insert(editor->regex_buffer, &current_iter,
				   branch_start, len);
	}
    }
}


void
ochusha_regex_editor_append_pattern(OchushaRegexEditor *editor,
				    const gchar *pattern)
{
  GtkTextIter current_iter;

  g_return_if_fail(OCHUSHA_IS_REGEX_EDITOR(editor));
  g_return_if_fail(pattern != NULL);

  if (*pattern == '\0')
    return;

  gtk_text_buffer_get_end_iter(editor->regex_buffer, &current_iter);

  if (gtk_text_iter_backward_char(&current_iter))
    {
      gunichar c = gtk_text_iter_get_char(&current_iter);
      gtk_text_iter_forward_char(&current_iter);
      if (c != '\n')
	gtk_text_buffer_insert(editor->regex_buffer, &current_iter, "\n", 1);
    }

  gtk_text_buffer_insert(editor->regex_buffer, &current_iter, pattern, -1);
}


void
ochusha_regex_editor_append_pattern_escaped(OchushaRegexEditor *editor,
					    const gchar *pattern)
{
  size_t pattern_len;
  gchar *escaped_pattern;
  size_t escaped_buf_len;
  const gchar *cur_pos;
  const gchar *tail_pos;
  gchar *buf_pos;

  g_return_if_fail(pattern != NULL);
  if (*pattern == '\0')
    return;

  pattern_len = strlen(pattern);
  escaped_buf_len = pattern_len * 2;	/* ٹ뤫 */
  escaped_pattern = malloc(escaped_buf_len);
  g_return_if_fail(escaped_pattern != NULL);

  cur_pos = pattern;
  tail_pos = cur_pos + pattern_len;
  buf_pos = escaped_pattern;

  while (cur_pos < tail_pos)
    {
      int c = *cur_pos++;
      switch (c)
	{
	case '\\':
	case '|':
	case '(':
	case ')':
	case '[':
	case ']':
	case '.':
	case '*':
	case '+':
	case '?':
	case '{':
	case '}':
	case '^':
	case '$':
	  *buf_pos++ = '\\';
	  break;
	}

      *buf_pos++ = c;
    }
  *buf_pos = '\0';

  ochusha_regex_editor_append_pattern(editor, escaped_pattern);

  free(escaped_pattern);
}


gchar *
ochusha_regex_editor_get_regex_text(OchushaRegexEditor *editor)
{
  GtkTextIter start_iter;
  GtkTextIter end_iter;
  gchar *buffer_text;
  gchar *result;

  g_return_val_if_fail(OCHUSHA_IS_REGEX_EDITOR(editor), NULL);

  gtk_text_buffer_get_start_iter(editor->regex_buffer, &start_iter);
  gtk_text_buffer_get_end_iter(editor->regex_buffer, &end_iter);
  buffer_text = gtk_text_buffer_get_text(editor->regex_buffer, &start_iter,
					 &end_iter, FALSE);
  result = ochusha_regex_editor_synthesize_regular_expression(buffer_text);
  g_free(buffer_text);

  return result;
}
