/*
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: search_window.c,v 1.10 2009/01/01 07:46:51 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#include "libochushagtk/search_window.h"
#include "libochushagtk/searchable_private.h"
#include "libochushagtk/marshal.h"

#include <glib.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


static void ochusha_search_window_class_init(OchushaSearchWindowClass *klass);

static void ochusha_search_window_init(OchushaSearchWindow *window);
static void ochusha_search_window_finalize(GObject *object);
static void ochusha_search_window_set_property(GObject *object, guint prop_id,
					       const GValue *value,
					       GParamSpec *pspec);
static void ochusha_search_window_get_property(GObject *object, guint prop_id,
					       GValue *value,
					       GParamSpec *pspec);

static void ochusha_search_window_destroy(GtkObject *object);
static gboolean ochusha_search_window_delete_event(GtkWidget *widget,
						   GdkEventAny *event);
static void ochusha_search_window_real_response(OchushaSearchWindow *window,
						gint response_id);
static void ochusha_search_window_find_prev(OchushaSearchWindow *window);
static void ochusha_search_window_find_next(OchushaSearchWindow *window);
static gboolean ochusha_search_window_key_press_event(GtkWidget *widget,
						      GdkEventKey *event);


GType
ochusha_search_window_get_type(void)
{
  static GType search_window_type = 0;

  if (search_window_type == 0)
    {
      static const GTypeInfo search_window_info =
	{
	  sizeof(OchushaSearchWindowClass),
	  NULL,	/* base_init */
	  NULL, /* base_finalize */
	  (GClassInitFunc)ochusha_search_window_class_init,
	  NULL,	/* class_finalize */
	  NULL, /* class_data */
	  sizeof(OchushaSearchWindow),
	  0,	/* n_preallocs */
	  (GInstanceInitFunc)ochusha_search_window_init,
	};

      search_window_type = g_type_register_static(GTK_TYPE_WINDOW,
						  "OchushaSearchWindow",
						  &search_window_info, 0);
    }

  return search_window_type;
}


enum {
  RESPONSE_SIGNAL,
  SEARCH_REQUEST_SIGNAL,
  FIND_PREVIOUS_SIGNAL,
  FIND_NEXT_SIGNAL,
  LAST_SIGNAL
};


enum {
  PROP_0,
  PROP_KEY
};


static GtkWindowClass *parent_class = NULL;
static int search_window_signals[LAST_SIGNAL] = { 0, 0, 0, 0 };


static void
ochusha_search_window_class_init(OchushaSearchWindowClass *klass)
{
  GObjectClass *o_class = (GObjectClass *)klass;
  GtkObjectClass *object_class = (GtkObjectClass *)klass;
  GtkWidgetClass *widget_class = (GtkWidgetClass *)klass;
  parent_class = g_type_class_peek_parent(klass);

  /* GObject signals */
  o_class->finalize = ochusha_search_window_finalize;
  o_class->set_property = ochusha_search_window_set_property;
  o_class->get_property = ochusha_search_window_get_property;

  g_object_class_install_property(o_class, PROP_KEY,
	g_param_spec_object("key",
			    PROP_("SearchKey"),
			    PROP_("Search Key object represents the query"),
			    OCHUSHA_TYPE_SEARCH_KEY,
			    G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));


  /* GtkObject signals */
  object_class->destroy = ochusha_search_window_destroy;

  search_window_signals[RESPONSE_SIGNAL] =
    g_signal_new("response",
		 G_TYPE_FROM_CLASS(klass),
		 G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
		 G_STRUCT_OFFSET(OchushaSearchWindowClass, response),
		 NULL, NULL,
		 libochushagtk_marshal_VOID__INT,
		 G_TYPE_NONE, 1,
		 G_TYPE_INT);
  search_window_signals[SEARCH_REQUEST_SIGNAL] =
    g_signal_new("search_request",
		 G_TYPE_FROM_CLASS(klass),
		 G_SIGNAL_RUN_LAST,
		 G_STRUCT_OFFSET(OchushaSearchWindowClass, search_request),
		 NULL, NULL,
		 libochushagtk_marshal_OBJECT__VOID,
		 OCHUSHA_TYPE_SEARCHABLE, 0);
  search_window_signals[FIND_PREVIOUS_SIGNAL] =
    g_signal_new("find_previous",
		 G_TYPE_FROM_CLASS(klass),
		 G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
		 G_STRUCT_OFFSET(OchushaSearchWindowClass, find_previous),
		 NULL, NULL,
		 libochushagtk_marshal_VOID__VOID,
		 G_TYPE_NONE, 0);
  search_window_signals[FIND_NEXT_SIGNAL] =
    g_signal_new("find_next",
		 G_TYPE_FROM_CLASS(klass),
		 G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
		 G_STRUCT_OFFSET(OchushaSearchWindowClass, find_next),
		 NULL, NULL,
		 libochushagtk_marshal_VOID__VOID,
		 G_TYPE_NONE, 0);

  widget_class->delete_event = ochusha_search_window_delete_event;
  widget_class->key_press_event = ochusha_search_window_key_press_event;

  klass->response = ochusha_search_window_real_response;
  klass->search_request = NULL;
  klass->find_previous = ochusha_search_window_find_prev;
  klass->find_next = ochusha_search_window_find_next;
}


static void
ochusha_search_window_entry_changed_cb(GtkEntry *entry,
				       OchushaSearchWindow *window)
{
  ochusha_search_key_set_key(window->key, gtk_entry_get_text(entry));
}


static void
ochusha_search_window_match_case_toggled_cb(GtkToggleButton *button,
					    OchushaSearchWindow *window)
{
  ochusha_search_key_set_match_case(window->key,
				    gtk_toggle_button_get_active(button));
}


static void
ochusha_search_window_use_regexp_toggled_cb(GtkToggleButton *button,
					    OchushaSearchWindow *window)
{
  ochusha_search_key_set_use_regexp(window->key,
				    gtk_toggle_button_get_active(button));
}


static void
ochusha_search_window_direction_down_toggled_cb(GtkToggleButton *button,
						OchushaSearchWindow *window)
{
  ochusha_search_key_set_direction(window->key,
				   gtk_toggle_button_get_active(button)
				   ? OCHUSHA_SEARCH_DIRECTION_FORWARD
				   : OCHUSHA_SEARCH_DIRECTION_BACKWARD);
}


static void
ochusha_search_window_show_no_match_text(OchushaSearchWindow *window)
{
#if !GTK_CHECK_VERSION(2,12,0)
  GdkDisplay *display;
#endif

  if (window->statusbar_message_id != 0)
    gtk_statusbar_remove(window->statusbar, window->statusbar_context_id,
			 window->statusbar_message_id);
  window->statusbar_message_id
    = gtk_statusbar_push(window->statusbar, window->statusbar_context_id,
			 LIB_("No Matching Text"));

#if GTK_CHECK_VERSION(2,12,0)
  gtk_widget_error_bell(GTK_WIDGET(window));
#else
  display = gtk_widget_get_display(GTK_WIDGET(window));
  if (display != NULL)
    gdk_display_beep(display);
#endif
}


static void
ochusha_search_window_real_find_next(OchushaSearchWindow *window)
{
  OchushaSearchable *searchable = NULL;

  g_signal_emit(window, search_window_signals[SEARCH_REQUEST_SIGNAL],
		0,
		&searchable);

  if (OCHUSHA_IS_SEARCHABLE(searchable))
    {
      if (window->last_search_target != searchable)
	{
	  if (window->last_search_target != NULL)
	    {
	      ochusha_searchable_terminate_search(window->last_search_target,
						  window->key);
	      g_signal_handler_disconnect(window->last_search_target,
					  window->last_search_target_destroy_id);
	    }

	  window->last_search_target = searchable;
	  window->last_search_target_destroy_id
	    = g_signal_connect(searchable, "destroy",
			       G_CALLBACK(gtk_widget_destroyed),
			       &window->last_search_target);
	}
    }

  if (searchable == NULL
      || ochusha_searchable_find_next(searchable, window->key))
    {
      if (window->statusbar_message_id != 0)
	gtk_statusbar_remove(window->statusbar, window->statusbar_context_id,
			     window->statusbar_message_id);
      window->statusbar_message_id = 0;
    }
  else
    ochusha_search_window_show_no_match_text(window);
}


static void
ochusha_search_window_terminate_search(OchushaSearchWindow *window)
{
  g_return_if_fail(OCHUSHA_IS_SEARCH_WINDOW(window));

  if (window->last_search_target != NULL)
    {
      ochusha_searchable_terminate_search(window->last_search_target,
					  window->key);
      g_signal_handler_disconnect(window->last_search_target,
				  window->last_search_target_destroy_id);
    }
  window->last_search_target = NULL;

  ochusha_search_window_response(OCHUSHA_SEARCH_WINDOW(window),
				 GTK_RESPONSE_CANCEL);
}


static void
ochusha_search_window_init(OchushaSearchWindow *window)
{
  /* ɥΤΥƥ */
  GtkWidget *container_vbox;
  GtkWidget *container_hbox;

  /* container */
  GtkWidget *vbox;
  GtkWidget *button_box;

  /* vbox */
  GtkWidget *main_box;
  GtkWidget *option_box;


  /* option_box */
  GtkWidget *search_option_box;
  GtkWidget *search_direction_frame;

  /* search_direction_frame */
  GtkWidget *direction_box;

  GtkWidget *tmp_widget;

  /* button_box */
  GtkWidget *next_button;
  GtkWidget *cancel_button;

  container_vbox = gtk_vbox_new(FALSE, 5);
  gtk_container_set_border_width(GTK_CONTAINER(&window->parent_object), 5);
  gtk_container_add(GTK_CONTAINER(&window->parent_object), container_vbox);

  container_hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(container_vbox), container_hbox, TRUE, TRUE, 0);

  window->statusbar = GTK_STATUSBAR(gtk_statusbar_new());
  gtk_box_pack_start(GTK_BOX(container_vbox),
		     GTK_WIDGET(window->statusbar), FALSE, FALSE, 0);
  window->statusbar_context_id
    = gtk_statusbar_get_context_id(window->statusbar, "search-result");
  window->statusbar_message_id = 0;


  vbox = gtk_vbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(container_hbox), vbox, TRUE, TRUE, 0);

  button_box = gtk_vbutton_box_new();
  gtk_button_box_set_layout(GTK_BUTTON_BOX(button_box), GTK_BUTTONBOX_START);
  gtk_box_set_spacing(GTK_BOX(button_box), 5);
  gtk_box_pack_end(GTK_BOX(container_hbox), button_box, FALSE, FALSE, 5);


  /* vbox */
  main_box = gtk_hbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(vbox), main_box, TRUE, TRUE, 0);

  option_box = gtk_hbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(vbox), option_box, FALSE, FALSE, 0);


  /* main_box */
  gtk_box_pack_start(GTK_BOX(main_box),
		     gtk_label_new(LIB_("Find What: ")), FALSE, FALSE, 0);

  tmp_widget = gtk_entry_new();
  g_signal_connect(tmp_widget, "changed",
		   G_CALLBACK(ochusha_search_window_entry_changed_cb),
		   window);
  gtk_box_pack_start(GTK_BOX(main_box), tmp_widget, TRUE, TRUE, 0);
  window->entry = GTK_ENTRY(tmp_widget);


  /* option_box */
  search_option_box = gtk_vbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(option_box), search_option_box, FALSE, FALSE, 0);

  search_direction_frame = gtk_frame_new(LIB_("Direction"));
  gtk_box_pack_start(GTK_BOX(option_box),
		     search_direction_frame, FALSE, FALSE, 0);


  /* search_option_box */
  tmp_widget = gtk_check_button_new_with_mnemonic(LIB_("Mat_ch case"));
  g_signal_connect(tmp_widget, "toggled",
		   G_CALLBACK(ochusha_search_window_match_case_toggled_cb),
		   window);
  gtk_box_pack_start(GTK_BOX(search_option_box), tmp_widget, FALSE, FALSE, 0);
  window->match_case_button = GTK_TOGGLE_BUTTON(tmp_widget);

  tmp_widget = gtk_check_button_new_with_mnemonic(LIB_("Regular _expression"));
  g_signal_connect(tmp_widget, "toggled",
		   G_CALLBACK(ochusha_search_window_use_regexp_toggled_cb),
		   window);
  gtk_box_pack_start(GTK_BOX(search_option_box), tmp_widget, FALSE, FALSE, 0);
  window->use_regexp_button = GTK_TOGGLE_BUTTON(tmp_widget);


  /* search_direction_frame */
  direction_box = gtk_hbox_new(TRUE, 5);
  gtk_container_add(GTK_CONTAINER(search_direction_frame), direction_box);

  /* direction_box */
  tmp_widget = gtk_radio_button_new_with_label(NULL, LIB_("Up"));
  gtk_box_pack_start(GTK_BOX(direction_box), tmp_widget, FALSE, FALSE, 0);
  window->direction_up_button = GTK_TOGGLE_BUTTON(tmp_widget);

  tmp_widget
    = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(tmp_widget),
						  LIB_("Down"));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_widget), TRUE);
  g_signal_connect(tmp_widget, "toggled",
		   G_CALLBACK(ochusha_search_window_direction_down_toggled_cb),
		   window);
  gtk_box_pack_start(GTK_BOX(direction_box), tmp_widget, FALSE, FALSE, 0);
  window->direction_down_button = GTK_TOGGLE_BUTTON(tmp_widget);


  /* button_box */
  next_button = gtk_button_new_with_mnemonic(LIB_("_Find Next"));
  g_signal_connect_swapped(next_button, "clicked",
			   G_CALLBACK(ochusha_search_window_real_find_next),
			   window);
  gtk_box_pack_start(GTK_BOX(button_box), next_button, FALSE, FALSE, 0);

  cancel_button = gtk_button_new_with_label(LIB_("Cancel"));
  g_signal_connect_swapped(cancel_button, "clicked",
			   G_CALLBACK(ochusha_search_window_terminate_search),
			   window);
  gtk_box_pack_start(GTK_BOX(button_box), cancel_button, FALSE, FALSE, 0);

  gtk_widget_show_all(container_vbox);

  window->enable_incremental_search = FALSE;

  window->key = NULL;
  window->key_changed_id = 0;
  window->last_search_target = NULL;
  window->last_search_target_destroy_id = 0;

  gtk_window_set_type_hint(GTK_WINDOW(window), GDK_WINDOW_TYPE_HINT_DIALOG);
  gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER_ON_PARENT);
}


static void
ochusha_search_window_finalize(GObject *object)
{

  OchushaSearchWindow *window = OCHUSHA_SEARCH_WINDOW(object);

  if (window->key != NULL)
    {
      g_signal_handler_disconnect(window->key, window->key_changed_id);
      g_object_unref(window->key);
      window->key = NULL;
    }

#if DEBUG_WIDGET_MOST
  fprintf(stderr, "ochusha_search_window_finalize(%p)\n", object);
#endif

  if (G_OBJECT_CLASS(parent_class)->finalize != NULL)
    (*G_OBJECT_CLASS(parent_class)->finalize)(object);
}


static void
ochusha_search_window_set_property(GObject *object, guint prop_id,
				   const GValue *value, GParamSpec *pspec)
{
  OchushaSearchWindow *window = OCHUSHA_SEARCH_WINDOW(object);

  switch (prop_id)
    {
    case PROP_KEY:
      ochusha_search_window_set_key(window, g_value_get_object(value));
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
      break;
    }
}


static void
ochusha_search_window_get_property(GObject *object, guint prop_id,
				   GValue *value, GParamSpec *pspec)
{
  OchushaSearchWindow *window;

  g_return_if_fail(OCHUSHA_IS_SEARCH_WINDOW(object));

  window = OCHUSHA_SEARCH_WINDOW(object);

  switch (prop_id)
    {
    case PROP_KEY:
      g_value_set_object(value, window->key);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
      break;
    }
}


static void
ochusha_search_window_destroy(GtkObject *object)
{
  OchushaSearchWindow *window = OCHUSHA_SEARCH_WINDOW(object);

#if DEBUG_WIDGET_MOST
  fprintf(stderr, "ochusha_search_window_destroy(%p)\n", object);
#endif

  if (window->last_search_target != NULL)
    {
      g_signal_handler_disconnect(window->last_search_target,
				  window->last_search_target_destroy_id);
      window->last_search_target = NULL;
    }

  if (GTK_OBJECT_CLASS(parent_class)->destroy)
    (*GTK_OBJECT_CLASS(parent_class)->destroy)(object);
}


static gboolean
ochusha_search_window_delete_event(GtkWidget *widget, GdkEventAny *event)
{
  /* emit response signal */
  ochusha_search_window_response(OCHUSHA_SEARCH_WINDOW(widget),
				 GTK_RESPONSE_DELETE_EVENT);

  if (GTK_WIDGET_CLASS(parent_class)->delete_event)
    return (*GTK_WIDGET_CLASS(parent_class)->delete_event)(widget, event);

  return FALSE;	/* ͭʤϤ*/
}


static void
ochusha_search_window_real_response(OchushaSearchWindow *window,
				    gint response_id)
{
  if (window->statusbar_message_id != 0)
    gtk_statusbar_remove(window->statusbar, window->statusbar_context_id,
			 window->statusbar_message_id);
  window->statusbar_message_id = 0;
}


static void
ochusha_search_window_find_prev(OchushaSearchWindow *window)
{
  gtk_toggle_button_set_active(window->direction_up_button, TRUE);
  ochusha_search_window_real_find_next(window);
}


static void
ochusha_search_window_find_next(OchushaSearchWindow *window)
{
  gtk_toggle_button_set_active(window->direction_down_button, TRUE);
  ochusha_search_window_real_find_next(window);
}


static gboolean
ochusha_search_window_key_press_event(GtkWidget *widget, GdkEventKey *event)
{
  OchushaSearchWindow *window = OCHUSHA_SEARCH_WINDOW(widget);

  if (event->keyval == GDK_Escape)
    {
      ochusha_search_window_response(OCHUSHA_SEARCH_WINDOW(window),
				     GTK_RESPONSE_CANCEL);
      return TRUE;
    }

  if (event->keyval == GDK_Up)
    {
      ochusha_search_window_find_prev(window);
      return TRUE;
    }

  if (event->keyval == GDK_Down)
    {
      ochusha_search_window_find_next(window);
      return TRUE;
    }

  if (event->keyval == GDK_Return)
    {
      ochusha_search_window_real_find_next(window);
      return TRUE;
    }

  if (GTK_WIDGET_CLASS(parent_class)->key_press_event)
    return (*GTK_WIDGET_CLASS(parent_class)->key_press_event)(widget, event);

  return FALSE;	/* ͭʤ*/
}


static void
ochusha_search_window_key_changed_cb(OchushaSearchKey *key,
				     guint key_id,
				     OchushaSearchWindow *window)
{
  g_return_if_fail(OCHUSHA_IS_SEARCH_WINDOW(window));
  g_return_if_fail(window->key == key);

  if (key->key_string != NULL)
    gtk_entry_set_text(window->entry, key->key_string);
  gtk_toggle_button_set_active(window->match_case_button, key->match_case);
  gtk_toggle_button_set_active(window->use_regexp_button, key->use_regexp);
  if (key->direction == OCHUSHA_SEARCH_DIRECTION_FORWARD)
    gtk_toggle_button_set_active(window->direction_down_button, TRUE);
  else
    gtk_toggle_button_set_active(window->direction_up_button, TRUE);

  if (window->enable_incremental_search)
    ochusha_search_window_real_find_next(window);
}


GtkWidget *
ochusha_search_window_new(void)
{
  return g_object_new(OCHUSHA_TYPE_SEARCH_WINDOW,
		      "key", ochusha_search_key_new(), NULL);
}


GtkWidget *
ochusha_search_window_new_with_key(OchushaSearchKey *key)
{
  g_return_val_if_fail(OCHUSHA_IS_SEARCH_KEY(key),
		       ochusha_search_window_new());

  g_object_ref(key);
  return g_object_new(OCHUSHA_TYPE_SEARCH_WINDOW,
		      "key", key, NULL);
}


void
ochusha_search_window_set_key(OchushaSearchWindow *window,
			      OchushaSearchKey *key)
{
  /* Constructor Only */
  g_return_if_fail(OCHUSHA_IS_SEARCH_WINDOW(window));
  g_return_if_fail(OCHUSHA_IS_SEARCH_KEY(key));
  g_return_if_fail(window->key == NULL);

  window->key = key;
  window->key_changed_id
    = g_signal_connect(key, "key_changed",
		       G_CALLBACK(ochusha_search_window_key_changed_cb),
		       window);

  if (key->key_string != NULL)
    gtk_entry_set_text(window->entry, key->key_string);

  gtk_toggle_button_set_active(window->match_case_button, key->match_case);
  gtk_toggle_button_set_active(window->use_regexp_button, key->use_regexp);

  if (key->direction == OCHUSHA_SEARCH_DIRECTION_FORWARD)
    gtk_toggle_button_set_active(window->direction_down_button, TRUE);
  else
    gtk_toggle_button_set_active(window->direction_up_button, TRUE);
}


OchushaSearchKey *
ochusha_search_window_get_key(OchushaSearchWindow *window)
{
  g_return_val_if_fail(OCHUSHA_IS_SEARCH_WINDOW(window), NULL);

  return window->key;
}


void
ochusha_search_window_set_enable_incremental_search(OchushaSearchWindow *window,
						 gboolean enable)
{
  g_return_if_fail(OCHUSHA_IS_SEARCH_WINDOW(window));

  window->enable_incremental_search = enable;
}


gboolean
ochusha_search_window_get_enable_incremental_search(OchushaSearchWindow *window)
{
  g_return_val_if_fail(OCHUSHA_IS_SEARCH_WINDOW(window), FALSE);

  return window->enable_incremental_search;
}


void
ochusha_search_window_response(OchushaSearchWindow *window, gint response_id)
{
  g_return_if_fail(OCHUSHA_IS_SEARCH_WINDOW(window));

  g_signal_emit(window, search_window_signals[RESPONSE_SIGNAL],
		0,
		response_id);
}


void
ochusha_search_window_set_key_selected(OchushaSearchWindow *window)
{
  g_return_if_fail(OCHUSHA_IS_SEARCH_WINDOW(window));

  if (window->statusbar_message_id != 0)
    gtk_statusbar_remove(window->statusbar, window->statusbar_context_id,
			 window->statusbar_message_id);
  window->statusbar_message_id = 0;

  gtk_editable_select_region(GTK_EDITABLE(window->entry), 0, -1);
  gtk_widget_grab_focus(GTK_WIDGET(window->entry));
}
