/*
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: searchable.c,v 1.11 2009/01/01 07:46:51 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"
#include "libochusha/misc_utils.h"
#include "libochushagtk/searchable.h"
#include "libochushagtk/searchable_private.h"

#include "libochushagtk/marshal.h"

#include <glib-object.h>

#include <oniguruma.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


static void ochusha_searchable_base_init(gpointer klass);


GType
ochusha_search_direction_get_type(void)
{
  static GType search_direction_type = 0;

  if (search_direction_type == 0)
    {
      static const GEnumValue values[] =
	{
	  {
	    OCHUSHA_SEARCH_DIRECTION_FORWARD,
	    (gchar *)"OCHUSHA_SEARCH_DIRECTION_FORWARD",
	    (gchar *)"search_forward"
	  },
	  {
	    OCHUSHA_SEARCH_DIRECTION_BACKWARD,
	    (gchar *)"OCHUSHA_SEARCH_DIRECTION_BACKWARD",
	    (gchar *)"search_backward"
	  }
	};

      search_direction_type = g_enum_register_static("OchushaSearchDirection",
						     values);
    }

  return search_direction_type;
}


GType
ochusha_search_action_get_type(void)
{
  static GType search_action_type = 0;

  if (search_action_type == 0)
    {
      static const GEnumValue values[] =
	{
	  {
	    OCHUSHA_SEARCH_ACTION_FORWARD,
	    (gchar *)"OCHUSHA_SEARCH_ACTION_FORWARD",
	    (gchar *)"search_forward"
	  },
	  {
	    OCHUSHA_SEARCH_ACTION_BACKWARD,
	    (gchar *)"OCHUSHA_SEARCH_ACTION_BACKWARD",
	    (gchar *)"search_backward"
	  },
	  {
	    OCHUSHA_SEARCH_ACTION_REGEXP_FORWARD,
	    (gchar *)"OCHUSHA_SEARCH_ACTION_REGEXP_FORWARD",
	    (gchar *)"search_regexp_forward"
	  },
	  {
	    OCHUSHA_SEARCH_ACTION_REGEXP_BACKWARD,
	    (gchar *)"OCHUSHA_SEARCH_ACTION_REGEXP_BACKWARD",
	    (gchar *)"search_regexp_backward"
	  },
	  {
	    OCHUSHA_SEARCH_ACTION_TERMINATE,
	    (gchar *)"OCHUSHA_SEARCH_ACTION_TERMINATE",
	    (gchar *)"terminate_search"
	  },
	};

      search_action_type = g_enum_register_static("OchushaSearchAction",
						  values);
    }

  return search_action_type;
}


static void ochusha_search_key_class_init(OchushaSearchKeyClass *klass);
static void ochusha_search_key_init(OchushaSearchKey *key);
static void ochusha_search_key_finalize(GObject *object);
static void ochusha_search_key_set_property(GObject *object, guint prop_id,
					    const GValue *value,
					    GParamSpec *pspec);
static void ochusha_search_key_get_property(GObject *object, guint prop_id,
					    GValue *value, GParamSpec *pspec);


GType
ochusha_search_key_get_type(void)
{
  static GType search_key_type = 0;

  if (search_key_type == 0)
    {
      static const GTypeInfo search_key_info =
	{
	  sizeof(OchushaSearchKeyClass),
	  NULL, /* base_init */
	  NULL, /* base_finalize */
	  (GClassInitFunc)ochusha_search_key_class_init,
	  NULL, /* class_finalize */
	  NULL, /* class_data */
	  sizeof(OchushaSearchKey),
	  0,	/* n_preallocs */
	  (GInstanceInitFunc)ochusha_search_key_init,
	};

      search_key_type = g_type_register_static(G_TYPE_OBJECT,
					       "OchushaSearchKey",
					       &search_key_info, 0);
    }
  
  return search_key_type;
}


enum {
  SEARCH_KEY_KEY_CHANGED_SIGNAL,
  LAST_SIGNAL
};


enum {
  PROP_0,
  PROP_KEY_STRING,
  PROP_DIRECTION,
  PROP_MATCH_CASE,
  PROP_USE_REGEXP
};


static GObjectClass *search_key_parent_class;
static int search_key_signals[LAST_SIGNAL] = { 0 };


static void
ochusha_search_key_class_init(OchushaSearchKeyClass *klass)
{
  GObjectClass *o_class = (GObjectClass *)klass;

  search_key_parent_class = g_type_class_peek_parent(klass);


  /* GObject signals */
  o_class->finalize = ochusha_search_key_finalize;
  o_class->set_property = ochusha_search_key_set_property;
  o_class->get_property = ochusha_search_key_get_property;

  g_object_class_install_property(o_class,
				  PROP_KEY_STRING,
				  g_param_spec_string("key_string",
						      PROP_("KeyString"),
						      PROP_("String with which Search Performed"),
						      NULL,
						      G_PARAM_READWRITE));
  g_object_class_install_property(o_class,
				  PROP_DIRECTION,
				  g_param_spec_enum("direction",
						    PROP_("Direction"),
						    PROP_("Search Direction"),
						    OCHUSHA_TYPE_SEARCH_DIRECTION,
						    OCHUSHA_SEARCH_DIRECTION_FORWARD,
						    G_PARAM_READWRITE));
  g_object_class_install_property(o_class,
				  PROP_MATCH_CASE,
				  g_param_spec_boolean("match_case",
						       PROP_("MatchCase"),
						       PROP_("Match Case"),
						       FALSE,
						       G_PARAM_READWRITE));
  g_object_class_install_property(o_class,
				  PROP_USE_REGEXP,
				  g_param_spec_boolean("use_regexp",
						       PROP_("UseRegexp"),
						       PROP_("Use Regular Expression"),
						       FALSE,
						       G_PARAM_READWRITE));

  search_key_signals[SEARCH_KEY_KEY_CHANGED_SIGNAL]
    = g_signal_new("key_changed",
		   G_TYPE_FROM_CLASS(klass),
		   G_SIGNAL_RUN_LAST,
		   G_STRUCT_OFFSET(OchushaSearchKeyClass, key_changed),
		   NULL, NULL,
		   libochushagtk_marshal_VOID__INT,
		   G_TYPE_NONE, 1,
		   G_TYPE_UINT);

  klass->key_changed = NULL;
}


static void
ochusha_search_key_init(OchushaSearchKey *key)
{
  key->key_string = NULL;
  key->key_len = 0;
  key->regexp = NULL;
  key->region = NULL;
  key->last_match_pos = NULL;

  key->key_id = 1;

  key->regexp_available = FALSE;
  key->direction = OCHUSHA_SEARCH_DIRECTION_FORWARD;
  key->match_case = FALSE;
  key->use_regexp = FALSE;
}


static void
ochusha_search_key_finalize(GObject *object)
{
  OchushaSearchKey *key = (OchushaSearchKey *)object;

  if (key->key_string != NULL)
    {
      g_free(key->key_string);
      key->key_string = NULL;
    }

  if (key->regexp != NULL)
    {
      onig_free(key->regexp);
      key->regexp = NULL;
      key->regexp_available = FALSE;
    }

  if (key->region != NULL)
    {
      onig_region_free(key->region, 1);
      key->region = NULL;
    }

  if (G_OBJECT_CLASS(search_key_parent_class)->finalize != NULL)
    (*G_OBJECT_CLASS(search_key_parent_class)->finalize)(object);
}


static void
ochusha_search_key_set_property(GObject *object, guint prop_id,
				const GValue *value, GParamSpec *pspec)
{
  OchushaSearchKey *key = (OchushaSearchKey *)object;

  switch (prop_id)
    {
    case PROP_KEY_STRING:
      ochusha_search_key_set_key(key, g_value_get_string(value));
      break;

    case PROP_DIRECTION:
      ochusha_search_key_set_direction(key, g_value_get_enum(value));
      break;

    case PROP_MATCH_CASE:
      ochusha_search_key_set_match_case(key, g_value_get_boolean(value));
      break;

    case PROP_USE_REGEXP:
      ochusha_search_key_set_use_regexp(key, g_value_get_boolean(value));
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
      break;
    }
}


static void
ochusha_search_key_get_property(GObject *object, guint prop_id,
				GValue *value, GParamSpec *pspec)
{
  OchushaSearchKey *key;

  g_return_if_fail(OCHUSHA_IS_SEARCH_KEY(object));
  key = OCHUSHA_SEARCH_KEY(object);

  switch (prop_id)
    {
    case PROP_KEY_STRING:
      g_value_set_string(value, key->key_string);
      break;

    case PROP_DIRECTION:
      g_value_set_enum(value, key->direction);
      break;

    case PROP_MATCH_CASE:
      g_value_set_boolean(value, key->match_case);
      break;

    case PROP_USE_REGEXP:
      g_value_set_boolean(value, key->use_regexp);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
      break;
    }
}


OchushaSearchKey *
ochusha_search_key_new(void)
{
  return g_object_new(OCHUSHA_TYPE_SEARCH_KEY, NULL);
}


OchushaSearchKey *
ochusha_search_key_new_with_initial_key(const gchar *key_string,
					OchushaSearchDirectionType direction,
					gboolean match_case,
					gboolean use_regexp)
{
  OchushaSearchKey *key = g_object_new(OCHUSHA_TYPE_SEARCH_KEY, NULL);
  ochusha_search_key_set_full(key, key_string, direction,
			      match_case, use_regexp);
  return key;
}


static gboolean
ochusha_search_key_set_key_internal(OchushaSearchKey *key,
				    const gchar *key_string)
{
  if (key->key_string == key_string)
    return FALSE;	/* ξNULL */

  if (key->key_string != NULL)
    {
      if (key_string != NULL && strcmp(key->key_string, key_string) == 0)
	return FALSE;	/* ѹʤ */
      g_free(key->key_string);
    }

  if (key_string == NULL)
    {
      key->key_string = NULL;
      key->key_len = 0;
    }
  else
    {
      key->key_string = strdup(key_string);
      key->key_len = strlen(key_string);
    }

  g_object_notify(G_OBJECT(key), "key_string");
  return TRUE;
}


static gboolean
ochusha_search_key_set_direction_internal(OchushaSearchKey *key,
					  OchushaSearchDirectionType direction)
{
  if (key->direction == direction)
    return FALSE;

  key->direction = direction;
  g_object_notify(G_OBJECT(key), "direction");
  return TRUE;
}


static gboolean
ochusha_search_key_set_match_case_internal(OchushaSearchKey *key,
					   gboolean match_case)
{
  if (key->match_case == match_case)
    return FALSE;

  key->match_case = match_case;
  g_object_notify(G_OBJECT(key), "match_case");
  return TRUE;
}


static gboolean
ochusha_search_key_set_use_regexp_internal(OchushaSearchKey *key,
					   gboolean use_regexp)
{
  if (key->use_regexp == use_regexp)
    return FALSE;

  key->use_regexp = use_regexp;
  g_object_notify(G_OBJECT(key), "use_regexp");
  return TRUE;
}


static void
ochusha_search_key_update_key_id(OchushaSearchKey *key)
{
  key->last_match_pos = NULL;

  if (key->use_regexp)
    {
      if (key->key_string != NULL)
	{
	  if (key->regexp != NULL)
	    {
	      onig_free(key->regexp);
	      key->regexp = NULL;
	    }

	  key->regexp_available
	    = onig_new(&key->regexp,
		       (UChar *)key->key_string,
		       (UChar *)key->key_string + key->key_len,
		       ONIG_OPTION_EXTEND | ONIG_OPTION_NEGATE_SINGLELINE
		       | key->match_case ? 0 : ONIG_OPTION_IGNORECASE,
		       ONIG_ENCODING_UTF8,
		       ONIG_SYNTAX_POSIX_EXTENDED, NULL) == ONIG_NORMAL;

	  if (key->region != NULL)
	    onig_region_clear(key->region);
	  else
	    key->region = onig_region_new();
	}
      else
	key->regexp_available = FALSE;

      if (!key->regexp_available)
	return;
    }

  g_signal_emit(key, search_key_signals[SEARCH_KEY_KEY_CHANGED_SIGNAL],
		0,
		++key->key_id);
}


void
ochusha_search_key_set_full(OchushaSearchKey *key, const gchar *key_string,
			    OchushaSearchDirectionType direction,
			    gboolean match_case, gboolean use_regexp)
{
  gboolean changed = FALSE;
  g_return_if_fail(OCHUSHA_IS_SEARCH_KEY(key));

  ochusha_search_key_set_direction_internal(key, direction);

  changed = ochusha_search_key_set_match_case_internal(key, match_case)
    | ochusha_search_key_set_use_regexp_internal(key, use_regexp)
    | ochusha_search_key_set_key_internal(key, key_string);

  if (changed)
    ochusha_search_key_update_key_id(key);
}


void
ochusha_search_key_set_key(OchushaSearchKey *key, const gchar *key_string)
{
  g_return_if_fail(OCHUSHA_IS_SEARCH_KEY(key));
  if (ochusha_search_key_set_key_internal(key, key_string))
    ochusha_search_key_update_key_id(key);
}


const gchar *
ochusha_search_key_get_key(OchushaSearchKey *key)
{
  g_return_val_if_fail(OCHUSHA_IS_SEARCH_KEY(key), NULL);
  return key->key_string;
}


void
ochusha_search_key_set_direction(OchushaSearchKey *key,
				 OchushaSearchDirectionType direction)
{
  g_return_if_fail(OCHUSHA_IS_SEARCH_KEY(key));
  ochusha_search_key_set_direction_internal(key, direction);
}


OchushaSearchDirectionType
ochusha_search_key_get_direction(OchushaSearchKey *key)
{
  g_return_val_if_fail(OCHUSHA_IS_SEARCH_KEY(key), 0);
  return key->direction;
}


void
ochusha_search_key_set_match_case(OchushaSearchKey *key, gboolean match_case)
{
  g_return_if_fail(OCHUSHA_IS_SEARCH_KEY(key));
  if (ochusha_search_key_set_match_case_internal(key, match_case))
    ochusha_search_key_update_key_id(key);
}


gboolean
ochusha_search_key_get_match_case(OchushaSearchKey *key)
{
  g_return_val_if_fail(OCHUSHA_IS_SEARCH_KEY(key), FALSE);
  return key->match_case;
}


void
ochusha_search_key_set_use_regexp(OchushaSearchKey *key, gboolean use_regexp)
{
  g_return_if_fail(OCHUSHA_IS_SEARCH_KEY(key));
  if (ochusha_search_key_set_use_regexp_internal(key, use_regexp))
    ochusha_search_key_update_key_id(key);
}


gboolean
ochusha_search_key_get_use_regexp(OchushaSearchKey *key)
{
  g_return_val_if_fail(OCHUSHA_IS_SEARCH_KEY(key), FALSE);
  return key->use_regexp;
}


gboolean
ochusha_search_key_match_text(OchushaSearchKey *key,
			      const gchar *text, int len)
{
  g_return_val_if_fail(OCHUSHA_IS_SEARCH_KEY(key), FALSE);

  if (key->key_len == 0)
    return FALSE;

  if (key->use_regexp)
    {
      UChar *text_tail;
      int offset;
      if (!key->regexp_available)
	return FALSE;

      if (len < 0)
	text_tail = (UChar *)text + strlen(text);
      else
	text_tail = (UChar *)text;

      offset = onig_search(key->regexp, (UChar *)text, text_tail,
			   (UChar *)text, text_tail, key->region, 0);

      if (offset == ONIG_MISMATCH)
	{
	  key->last_match_pos = NULL;
	  key->last_match_len = 0;
	  return FALSE;
	}

      if (key->direction == OCHUSHA_SEARCH_DIRECTION_BACKWARD)
	{
	  /*
	   * ĤϤɤͤƤ٤ΤǻȤʤ˱ۤȤϤʤġġ
	   */
	  UChar *tmp_text = (UChar *)g_utf8_next_char(text + offset);
	  while (tmp_text < text_tail)
	    {
	      int tmp_offset = offset;
	      offset = onig_search(key->regexp, (UChar *)text, text_tail,
				   tmp_text, text_tail, key->region, 0);
	      if (offset == ONIG_MISMATCH)
		{
		  offset = tmp_offset;
		  tmp_offset = onig_search(key->regexp,
					   (UChar *)text, text_tail,
					   (UChar *)text + offset, text_tail,
					   key->region, 0);
		  break;
		}
	      tmp_text = (UChar *)g_utf8_next_char(text + tmp_offset);
	    }
	}

      g_assert(offset != ONIG_MISMATCH);

      key->last_match_pos = text + offset;
      key->last_match_len = key->region->end[0] - key->region->beg[0];
      return TRUE;
    }
  else
    {
      /* match_caseƤͤ(w */
      size_t text_len = len < 0 ? strlen(text) : len;
      if (key->direction == OCHUSHA_SEARCH_DIRECTION_FORWARD)
	{
	  if (key->match_case)
	    key->last_match_pos = strnstr(text, key->key_string, text_len);
	  else
	    key->last_match_pos = strncasestr(text, key->key_string, text_len);
	}
      else
	{
	  if (key->match_case)
	    key->last_match_pos = strnrstr(text, key->key_string, text_len);
	  else
	    key->last_match_pos = strncaserstr(text, key->key_string, text_len);
	}
      key->last_match_len = key->key_len;

#if 0
      fprintf(stderr, "find %s in %s(len=%d): %p\n",
	      key->key_string, text, text_len, key->last_match_pos);
#endif

      return key->last_match_pos != NULL;
    }
}


const gchar *
ochusha_search_key_last_match_begin(OchushaSearchKey *key)
{
  g_return_val_if_fail(OCHUSHA_IS_SEARCH_KEY(key), NULL);
  g_return_val_if_fail(key->last_match_pos != NULL, NULL);

  return key->last_match_pos;
}


const gchar *
ochusha_search_key_last_match_end(OchushaSearchKey *key)
{
  g_return_val_if_fail(OCHUSHA_IS_SEARCH_KEY(key), NULL);
  g_return_val_if_fail(key->last_match_pos != NULL, NULL);

  return key->last_match_pos + key->last_match_len;
}


size_t
ochusha_search_key_last_match_len(OchushaSearchKey *key)
{
  g_return_val_if_fail(OCHUSHA_IS_SEARCH_KEY(key), 0);

  return key->last_match_len;
}


GType
ochusha_searchable_get_type(void)
{
  static GType searchable_type = 0;

  if (searchable_type == 0)
    {
      static const GTypeInfo searchable_info =
	{
	  sizeof(OchushaSearchableIface),
	  ochusha_searchable_base_init,
	  NULL,		/* base finalize */
	  NULL,
	  NULL,		/* class_finalize */
	  NULL,		/* class_data */
	  0,
	  0,		/* n_preallocs */
	  NULL
	};

      searchable_type
	= g_type_register_static(G_TYPE_INTERFACE, "OchushaSearchable",
				 &searchable_info, 0);
      g_type_interface_add_prerequisite(searchable_type, G_TYPE_OBJECT);
    }

  return searchable_type;
}


static void
ochusha_searchable_base_init(gpointer klass)
{
  static gboolean initialized = FALSE;

  if (!initialized)
    {
      /* ʥȤä餳 */

      initialized = TRUE;
    }
}


gboolean
ochusha_searchable_find_next(OchushaSearchable *searchable,
			     OchushaSearchKey *key)
{
  g_return_val_if_fail(OCHUSHA_IS_SEARCHABLE(searchable), FALSE);
  g_return_val_if_fail(OCHUSHA_IS_SEARCH_KEY(key), FALSE);

  if (OCHUSHA_SEARCHABLE_GET_IFACE(searchable)->find_next)
    return (*OCHUSHA_SEARCHABLE_GET_IFACE(searchable)->find_next)(searchable,
								  key);

  return FALSE;
}


void
ochusha_searchable_terminate_search(OchushaSearchable *searchable,
				    OchushaSearchKey *key)
{
  g_return_if_fail(OCHUSHA_IS_SEARCHABLE(searchable));

  if (OCHUSHA_SEARCHABLE_GET_IFACE(searchable)->terminate_search)
    (*OCHUSHA_SEARCHABLE_GET_IFACE(searchable)->terminate_search)(searchable,
								  key);
}
