/*
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: searchbar.c,v 1.10 2009/01/01 07:46:51 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#include "libochushagtk/icon_label.h"
#include "libochushagtk/searchbar.h"
#include "libochushagtk/searchable_private.h"
#include "libochushagtk/marshal.h"

#include <glib.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


static void ochusha_searchbar_class_init(OchushaSearchbarClass *klass);

static void ochusha_searchbar_init(OchushaSearchbar *bar);
static void ochusha_searchbar_finalize(GObject *object);
static void ochusha_searchbar_set_property(GObject *object, guint prop_id,
					   const GValue *value,
					   GParamSpec *pspec);
static void ochusha_searchbar_get_property(GObject *object, guint prop_id,
					   GValue *value,
					   GParamSpec *pspec);

static void ochusha_searchbar_destroy(GtkObject *object);
static gboolean ochusha_searchbar_key_press_event(GtkWidget *widget,
						  GdkEventKey *event);
static void ochusha_searchbar_find_next(OchushaSearchbar *bar);
static void ochusha_searchbar_find_prev(OchushaSearchbar *bar);
static void ochusha_searchbar_real_response(OchushaSearchbar *bar,
					    int response_id);


GType
ochusha_searchbar_get_type(void)
{
  static GType searchbar_type = 0;

  if (searchbar_type == 0)
    {
      static const GTypeInfo searchbar_info =
	{
	  sizeof(OchushaSearchbarClass),
	  NULL,	/* base_init */
	  NULL,	/* base_finalize */
	  (GClassInitFunc)ochusha_searchbar_class_init,
	  NULL,	/* class_finalize */
	  NULL,	/* class_data */
	  sizeof(OchushaSearchbar),
	  0,	/* n_preallocs */
	(GInstanceInitFunc)ochusha_searchbar_init
	};

      searchbar_type = g_type_register_static(GTK_TYPE_HBOX,
					      "OchushaSearchbar",
					      &searchbar_info, 0);
    }

  return searchbar_type;
}


enum {
  RESPONSE_SIGNAL,
  SEARCH_REQUEST_SIGNAL,
  FIND_PREVIOUS_SIGNAL,
  FIND_NEXT_SIGNAL,
  LAST_SIGNAL
};


enum {
  PROP_0,
  PROP_KEY
};


static GtkHBoxClass *parent_class = NULL;
static int searchbar_signals[LAST_SIGNAL] = { 0, 0, 0, 0 };


static void
ochusha_searchbar_class_init(OchushaSearchbarClass *klass)
{
  GObjectClass *o_class = (GObjectClass *)klass;
  GtkObjectClass *object_class = (GtkObjectClass *)klass;
  GtkWidgetClass *widget_class = (GtkWidgetClass *)klass;
  parent_class = g_type_class_peek_parent(klass);

  /* GObject signals */
  o_class->finalize = ochusha_searchbar_finalize;
  o_class->set_property = ochusha_searchbar_set_property;
  o_class->get_property = ochusha_searchbar_get_property;

  g_object_class_install_property(o_class, PROP_KEY,
	g_param_spec_object("key",
			    PROP_("SearchKey"),
			    PROP_("Search Key object represents the query"),
			    OCHUSHA_TYPE_SEARCH_KEY,
			    G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));

  /* GtkObject signals */
  object_class->destroy = ochusha_searchbar_destroy;

  searchbar_signals[RESPONSE_SIGNAL] =
    g_signal_new("response",
		 G_TYPE_FROM_CLASS(klass),
		 G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
		 G_STRUCT_OFFSET(OchushaSearchbarClass, response),
		 NULL, NULL,
		 libochushagtk_marshal_VOID__INT,
		 G_TYPE_NONE, 1,
		 G_TYPE_INT);
  searchbar_signals[SEARCH_REQUEST_SIGNAL] =
    g_signal_new("search_request",
		 G_TYPE_FROM_CLASS(klass),
		 G_SIGNAL_RUN_LAST,
		 G_STRUCT_OFFSET(OchushaSearchbarClass, search_request),
		 NULL, NULL,
		 libochushagtk_marshal_OBJECT__VOID,
		 OCHUSHA_TYPE_SEARCHABLE, 0);
  searchbar_signals[FIND_PREVIOUS_SIGNAL] =
    g_signal_new("find_previous",
		 G_TYPE_FROM_CLASS(klass),
		 G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
		 G_STRUCT_OFFSET(OchushaSearchbarClass, find_previous),
		 NULL, NULL,
		 libochushagtk_marshal_VOID__VOID,
		 G_TYPE_NONE, 0);
  searchbar_signals[FIND_NEXT_SIGNAL] =
    g_signal_new("find_next",
		 G_TYPE_FROM_CLASS(klass),
		 G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
		 G_STRUCT_OFFSET(OchushaSearchbarClass, find_next),
		 NULL, NULL,
		 libochushagtk_marshal_VOID__VOID,
		 G_TYPE_NONE, 0);

  widget_class->key_press_event = ochusha_searchbar_key_press_event;

  klass->response = ochusha_searchbar_real_response;
  klass->search_request = NULL;
  klass->find_previous = ochusha_searchbar_find_prev;
  klass->find_next = ochusha_searchbar_find_next;
}


static void
ochusha_searchbar_entry_changed_cb(GtkEntry *entry, OchushaSearchbar *bar)
{
  ochusha_search_key_set_key(bar->key, gtk_entry_get_text(entry));
}


static void
ochusha_searchbar_match_case_toggled_cb(GtkToggleButton *button,
					OchushaSearchbar *bar)
{
  ochusha_search_key_set_match_case(bar->key,
				    gtk_toggle_button_get_active(button));
}


static void
ochusha_searchbar_use_regexp_toggled_cb(GtkToggleButton *button,
					OchushaSearchbar *bar)
{
  ochusha_search_key_set_use_regexp(bar->key,
				    gtk_toggle_button_get_active(button));
}


static void
ochusha_searchbar_show_no_match_text(OchushaSearchbar *bar)
{
#if !GTK_CHECK_VERSION(2,12,0)
  GdkDisplay *display;
#endif

  /*
   * ʥǤȯԤ뤫ġĤȻפäfirefoxΤϸСαü
   * åФƤ
   */
  if (bar->statusbar_message_id != 0)
    gtk_statusbar_remove(bar->statusbar, bar->statusbar_context_id,
			 bar->statusbar_message_id);
  bar->statusbar_message_id
    = gtk_statusbar_push(bar->statusbar, bar->statusbar_context_id,
			 LIB_("No Matching Text"));

#if GTK_CHECK_VERSION(2,12,0)
  gtk_widget_error_bell(GTK_WIDGET(bar));
#else
  display = gtk_widget_get_display(GTK_WIDGET(bar));
  if (display != NULL)
    gdk_display_beep(display);
#endif
}


static void
ochusha_searchbar_advance_search(OchushaSearchbar *bar)
{
  OchushaSearchable *searchable = NULL;
  g_signal_emit(bar, searchbar_signals[SEARCH_REQUEST_SIGNAL],
		0,
		&searchable);

  if (OCHUSHA_IS_SEARCHABLE(searchable))
    {
      if (bar->last_search_target != searchable)
	{
	  if (bar->last_search_target != NULL)
	    {
	      ochusha_searchable_terminate_search(bar->last_search_target,
						  bar->key);
	      g_signal_handler_disconnect(bar->last_search_target,
					  bar->last_search_target_destroy_id);
	    }
	  bar->last_search_target = searchable;
	  bar->last_search_target_destroy_id
	    = g_signal_connect(searchable, "destroy",
			       G_CALLBACK(gtk_widget_destroyed),
			       &bar->last_search_target);
	}
    }

  if (searchable == NULL
      || ochusha_searchable_find_next(searchable, bar->key))
    {
      if (bar->statusbar_message_id != 0)
	gtk_statusbar_remove(bar->statusbar, bar->statusbar_context_id,
			     bar->statusbar_message_id);
      bar->statusbar_message_id = 0;
    }
  else
    ochusha_searchbar_show_no_match_text(bar);
}


static void
ochusha_searchbar_find_next(OchushaSearchbar *bar)
{
  ochusha_search_key_set_direction(bar->key, OCHUSHA_SEARCH_DIRECTION_FORWARD);
  ochusha_searchbar_advance_search(bar);
}


static void
ochusha_searchbar_find_prev(OchushaSearchbar *bar)
{
  ochusha_search_key_set_direction(bar->key, OCHUSHA_SEARCH_DIRECTION_BACKWARD);
  ochusha_searchbar_advance_search(bar);
}


static void
ochusha_searchbar_terminate_search(OchushaSearchbar *bar)
{
  g_return_if_fail(OCHUSHA_IS_SEARCHBAR(bar));

  if (bar->last_search_target != NULL)
    {
      ochusha_searchable_terminate_search(bar->last_search_target,
					  bar->key);
      g_signal_handler_disconnect(bar->last_search_target,
				  bar->last_search_target_destroy_id);
    }
  bar->last_search_target = NULL;

  ochusha_searchbar_response(OCHUSHA_SEARCHBAR(bar), GTK_RESPONSE_CANCEL);
}


static void
ochusha_searchbar_init(OchushaSearchbar *bar)
{
  GtkBox *box = GTK_BOX(bar);
  GtkWidget *widget = gtk_button_new();
  GtkWidget *tmp_widget;
  gtk_container_add(GTK_CONTAINER(widget),
		    gtk_image_new_from_stock(GTK_STOCK_CLOSE,
					     GTK_ICON_SIZE_MENU));
  gtk_button_set_focus_on_click(GTK_BUTTON(widget), FALSE);
  gtk_button_set_relief(GTK_BUTTON(widget), GTK_RELIEF_NONE);
  GTK_WIDGET_UNSET_FLAGS(widget, GTK_CAN_FOCUS);
  gtk_widget_show_all(widget);
  gtk_box_pack_start(box, widget, FALSE, FALSE, 0);
  g_signal_connect_swapped(widget, "clicked",
			   G_CALLBACK(ochusha_searchbar_terminate_search), bar);

  widget = gtk_label_new(LIB_("Find:"));
  gtk_widget_show(widget);
  gtk_box_pack_start(box, widget, FALSE, FALSE, 4);

  widget = gtk_entry_new();
  g_signal_connect(widget, "changed",
		   G_CALLBACK(ochusha_searchbar_entry_changed_cb), bar);
  gtk_widget_grab_focus(widget);
  gtk_widget_show(widget);
  gtk_box_pack_start(box, widget, FALSE, FALSE, 0);
  bar->entry = GTK_ENTRY(widget);

  widget = gtk_button_new();
  tmp_widget
    = ochusha_icon_label_new_with_mnemonic_from_stock(LIB_("Find _Next"),
						      GTK_STOCK_GO_DOWN);
  gtk_container_add(GTK_CONTAINER(widget), tmp_widget);
  gtk_button_set_focus_on_click(GTK_BUTTON(widget), FALSE);
  gtk_button_set_relief(GTK_BUTTON(widget), GTK_RELIEF_NONE);
  g_signal_connect_swapped(widget, "clicked",
			   G_CALLBACK(ochusha_searchbar_find_next), bar);
  GTK_WIDGET_UNSET_FLAGS(widget, GTK_CAN_FOCUS);
  gtk_widget_show_all(widget);
  gtk_box_pack_start(box, widget, FALSE, FALSE, 0);
  bar->find_next_button = widget;

  widget = gtk_button_new();
  tmp_widget
    = ochusha_icon_label_new_with_mnemonic_from_stock(LIB_("Find _Previous"),
						      GTK_STOCK_GO_UP);
  gtk_container_add(GTK_CONTAINER(widget), tmp_widget);
  gtk_button_set_focus_on_click(GTK_BUTTON(widget), FALSE);
  gtk_button_set_relief(GTK_BUTTON(widget), GTK_RELIEF_NONE);
  g_signal_connect_swapped(widget, "clicked",
			   G_CALLBACK(ochusha_searchbar_find_prev), bar);
  GTK_WIDGET_UNSET_FLAGS(widget, GTK_CAN_FOCUS);
  gtk_widget_show_all(widget);
  gtk_box_pack_start(box, widget, FALSE, FALSE, 0);
  bar->find_prev_button = widget;

  widget = gtk_check_button_new_with_mnemonic(LIB_("Mat_ch case"));
  gtk_button_set_focus_on_click(GTK_BUTTON(widget), FALSE);
  g_signal_connect(widget, "toggled",
		   G_CALLBACK(ochusha_searchbar_match_case_toggled_cb), bar);
  GTK_WIDGET_UNSET_FLAGS(widget, GTK_CAN_FOCUS);
  gtk_widget_show(widget);
  gtk_box_pack_start(box, widget, FALSE, FALSE, 4);
  bar->match_case_button = GTK_TOGGLE_BUTTON(widget);

  widget = gtk_check_button_new_with_mnemonic(LIB_("Regular _expression"));
  gtk_button_set_focus_on_click(GTK_BUTTON(widget), FALSE);
  g_signal_connect(widget, "toggled",
		   G_CALLBACK(ochusha_searchbar_use_regexp_toggled_cb), bar);
  GTK_WIDGET_UNSET_FLAGS(widget, GTK_CAN_FOCUS);
  gtk_widget_show(widget);
  gtk_box_pack_start(box, widget, FALSE, FALSE, 0);
  bar->use_regexp_button = GTK_TOGGLE_BUTTON(widget);

  widget = gtk_statusbar_new();
  gtk_widget_show(widget);
  gtk_box_pack_start(box, widget, TRUE, TRUE, 4);
  bar->statusbar = GTK_STATUSBAR(widget);
  gtk_statusbar_set_has_resize_grip(bar->statusbar, FALSE);
  bar->statusbar_context_id
    = gtk_statusbar_get_context_id(bar->statusbar, "search-result");
  bar->statusbar_message_id = 0;

  bar->enable_incremental_search = FALSE;

  bar->key = NULL;
  bar->key_changed_id = 0;
  bar->last_search_target = NULL;
  bar->last_search_target_destroy_id = 0;
}


static void
ochusha_searchbar_finalize(GObject *object)
{
  OchushaSearchbar *bar = OCHUSHA_SEARCHBAR(object);

  if (bar->key != NULL)
    {
      g_signal_handler_disconnect(bar->key, bar->key_changed_id);
      g_object_unref(bar->key);
      bar->key = NULL;
    }

#if DEBUG_WIDGET_MOST
  fprintf(stderr, "ochusha_searchbar_finalize(%p)\n", object);
#endif

  if (G_OBJECT_CLASS(parent_class)->finalize != NULL)
    (*G_OBJECT_CLASS(parent_class)->finalize)(object);
}


static void
ochusha_searchbar_set_property(GObject *object, guint prop_id,
			       const GValue *value, GParamSpec *pspec)
{
  OchushaSearchbar *bar = OCHUSHA_SEARCHBAR(object);

  switch (prop_id)
    {
    case PROP_KEY:
      ochusha_searchbar_set_key(bar, g_value_get_object(value));
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
      break;
    }
}


static void
ochusha_searchbar_get_property(GObject *object, guint prop_id,
			       GValue *value, GParamSpec *pspec)
{
  OchushaSearchbar *bar;
  g_return_if_fail(OCHUSHA_IS_SEARCHBAR(object));

  bar = OCHUSHA_SEARCHBAR(object);

  switch (prop_id)
    {
    case PROP_KEY:
      g_value_set_object(value, bar->key);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
      break;
    }
}


static void
ochusha_searchbar_destroy(GtkObject *object)
{
  OchushaSearchbar *bar = OCHUSHA_SEARCHBAR(object);

#if DEBUG_WIDGET_MOST
  fprintf(stderr, "ochusha_searchbar_destroy(%p)\n", object);
#endif

  if (bar->last_search_target != NULL)
    {
      g_signal_handler_disconnect(bar->last_search_target,
				  bar->last_search_target_destroy_id);
      bar->last_search_target = NULL;
    }

  if (GTK_OBJECT_CLASS(parent_class)->destroy)
    (*GTK_OBJECT_CLASS(parent_class)->destroy)(object);
}


static gboolean
ochusha_searchbar_key_press_event(GtkWidget *widget, GdkEventKey *event)
{
  OchushaSearchbar *bar = OCHUSHA_SEARCHBAR(widget);

  if (event->keyval == GDK_Escape)
    {
      ochusha_searchbar_response(OCHUSHA_SEARCHBAR(bar), GTK_RESPONSE_CANCEL);
      return TRUE;
    }

  if (event->keyval == GDK_Up)
    {
      ochusha_searchbar_find_prev(bar);
      return TRUE;
    }

  if (event->keyval == GDK_Down || event->keyval == GDK_Return)
    {
      ochusha_searchbar_find_next(bar);
      return TRUE;
    }

  if (GTK_WIDGET_CLASS(parent_class)->key_press_event)
    return (*GTK_WIDGET_CLASS(parent_class)->key_press_event)(widget, event);

  return FALSE;	/* ͭʤ*/
}


static void
ochusha_searchbar_key_changed_cb(OchushaSearchKey *key, guint key_id,
				 OchushaSearchbar *bar)
{
  g_return_if_fail(OCHUSHA_IS_SEARCHBAR(bar));
  g_return_if_fail(bar->key == key);

  if (key->key_string != NULL)
    gtk_entry_set_text(bar->entry, key->key_string);
  gtk_toggle_button_set_active(bar->match_case_button, key->match_case);
  gtk_toggle_button_set_active(bar->use_regexp_button, key->use_regexp);

  if (key->key_string != NULL && *key->key_string != '\0')
    {
      gtk_widget_set_sensitive(bar->find_next_button, TRUE);
      gtk_widget_set_sensitive(bar->find_prev_button, TRUE);
      if (bar->enable_incremental_search)
	ochusha_searchbar_find_next(bar);
    }
  else
    {
      gtk_widget_set_sensitive(bar->find_next_button, FALSE);
      gtk_widget_set_sensitive(bar->find_prev_button, FALSE);
    }
}


static void
ochusha_searchbar_real_response(OchushaSearchbar *bar, int response_id)
{
  if (response_id != GTK_RESPONSE_CANCEL)
    return;

  if (bar->statusbar_message_id != 0)
    gtk_statusbar_remove(bar->statusbar, bar->statusbar_context_id,
			 bar->statusbar_message_id);
  bar->statusbar_message_id = 0;
}


GtkWidget *
ochusha_searchbar_new(void)
{
  return g_object_new(OCHUSHA_TYPE_SEARCHBAR,
		      "key", ochusha_search_key_new(), NULL);
}


GtkWidget *
ochusha_searchbar_new_with_key(OchushaSearchKey *key)
{
  g_return_val_if_fail(OCHUSHA_IS_SEARCH_KEY(key),
		       ochusha_searchbar_new());

  g_object_ref(key);
  return g_object_new(OCHUSHA_TYPE_SEARCHBAR,
		      "key", key, NULL);
}


void
ochusha_searchbar_set_key(OchushaSearchbar *bar, OchushaSearchKey *key)
{
  gboolean sensitive;
  /* Constructor Only */
  g_return_if_fail(OCHUSHA_IS_SEARCHBAR(bar));
  g_return_if_fail(OCHUSHA_IS_SEARCH_KEY(key));
  g_return_if_fail(bar->key == NULL);

  bar->key = key;
  bar->key_changed_id
    = g_signal_connect(key, "key_changed",
		       G_CALLBACK(ochusha_searchbar_key_changed_cb), bar);

  if (key->key_string != NULL)
    gtk_entry_set_text(bar->entry, key->key_string);

  gtk_toggle_button_set_active(bar->match_case_button, key->match_case);
  gtk_toggle_button_set_active(bar->use_regexp_button, key->use_regexp);

  sensitive = key->key_string != NULL && *key->key_string != '\0';
  gtk_widget_set_sensitive(bar->find_next_button, sensitive);
  gtk_widget_set_sensitive(bar->find_prev_button, sensitive);
}


OchushaSearchKey *
ochusha_searchbar_get_key(OchushaSearchbar *bar)
{
  g_return_val_if_fail(OCHUSHA_IS_SEARCHBAR(bar), NULL);

  return bar->key;
}


void
ochusha_searchbar_set_enable_incremental_search(OchushaSearchbar *bar,
						gboolean enable)
{
  g_return_if_fail(OCHUSHA_IS_SEARCHBAR(bar));
  bar->enable_incremental_search = enable;
}


gboolean
ochusha_searchbar_get_enable_incremental_search(OchushaSearchbar *bar)
{
  g_return_val_if_fail(OCHUSHA_IS_SEARCHBAR(bar), FALSE);
  return bar->enable_incremental_search;
}


void
ochusha_searchbar_response(OchushaSearchbar *bar, int response_id)
{
  g_return_if_fail(OCHUSHA_IS_SEARCHBAR(bar));
  g_signal_emit(bar, searchbar_signals[RESPONSE_SIGNAL],
		0,
		response_id);
}


void
ochusha_searchbar_set_key_selected(OchushaSearchbar *bar)
{
  g_return_if_fail(OCHUSHA_IS_SEARCHBAR(bar));

  if (bar->statusbar_message_id != 0)
    gtk_statusbar_remove(bar->statusbar, bar->statusbar_context_id,
			 bar->statusbar_message_id);
  bar->statusbar_message_id = 0;

  gtk_editable_select_region(GTK_EDITABLE(bar->entry), 0, -1);
  gtk_widget_grab_focus(GTK_WIDGET(bar->entry));
}


void
ochusha_searchbar_grab_focus(OchushaSearchbar *bar)
{
  g_return_if_fail(OCHUSHA_IS_SEARCHBAR(bar));

  gtk_widget_grab_focus(GTK_WIDGET(bar->entry));
}
