/*
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: tree_selector_view.c,v 1.20 2009/01/01 07:46:51 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"
#include "libochushagtk/searchable.h"
#include "libochushagtk/searchable_private.h"
#include "libochushagtk/tree_selector_view.h"
#include "libochushagtk/marshal.h"

#include <gdk/gdkkeysyms.h>
#include <glib.h>
#include <gtk/gtk.h>

#include <stdio.h>


static void ochusha_tree_selector_view_class_init(OchushaTreeSelectorViewClass *klass);
static void ochusha_tree_selector_view_searchable_init(OchushaSearchableIface *iface);
static void ochusha_tree_selector_view_init(OchushaTreeSelectorView *view);
static void ochusha_tree_selector_view_finalize(GObject *object);
static void ochusha_tree_selector_view_destroy(GtkObject *object);
static gboolean ochusha_tree_selector_view_motion_notify(GtkWidget *widget,
							 GdkEventMotion *event);
static gboolean ochusha_tree_selector_view_leave_notify(GtkWidget *widget,
							GdkEventCrossing *event);
static gboolean ochusha_tree_selector_view_button_press(GtkWidget *widget,
							GdkEventButton *event);
static gboolean ochusha_tree_selector_view_button_release(GtkWidget *widget,
							  GdkEventButton *event);
static gint ochusha_tree_selector_view_key_press_event(GtkWidget *widget,
						       GdkEventKey *event);

static void ochusha_tree_selector_view_center_cursor_on_screen(GtkTreeView *view);
static gboolean ochusha_tree_selector_view_find_next(OchushaSearchable *searchable,
						     OchushaSearchKey *key);
static void ochusha_tree_selector_view_terminate_search(OchushaSearchable *searchable,
							OchushaSearchKey *key);


GType
ochusha_tree_selector_view_get_type(void)
{
  static GType tree_selector_view_type = 0;

  if (tree_selector_view_type == 0)
    {
      static const GTypeInfo tree_selector_view_info =
        {
          sizeof(OchushaTreeSelectorViewClass),
          NULL, /* base_init */
          NULL, /* base_finalize */
          (GClassInitFunc)ochusha_tree_selector_view_class_init,
          NULL, /* class_finalize */
          NULL, /* class_data */
          sizeof(OchushaTreeSelectorView),
          0,    /* n_preallocs */
          (GInstanceInitFunc)ochusha_tree_selector_view_init,
        };

      static const GInterfaceInfo searchable_info =
	{
	  (GInterfaceInitFunc)ochusha_tree_selector_view_searchable_init,
	  NULL,
	  NULL
	};

      tree_selector_view_type
	= g_type_register_static(GTK_TYPE_TREE_VIEW, "OchushaTreeSelectorView",
				 &tree_selector_view_info, 0);

      g_type_add_interface_static(tree_selector_view_type,
				  OCHUSHA_TYPE_SEARCHABLE, &searchable_info);
    }

  return tree_selector_view_type;
}


enum {
  ROW_MOUSE_OVER_SIGNAL,
  ROW_MOUSE_OUT_SIGNAL,
  ROW_BUTTON_PRESS_SIGNAL,
  ROW_BUTTON_RELEASE_SIGNAL,

  COLUMN_RESIZED_SIGNAL,

  CENTER_CURSOR_ON_SCREEN_SIGNAL,

  LAST_SIGNAL
};


static GtkTreeViewClass *parent_class = NULL;
static int tree_selector_view_signals[LAST_SIGNAL] = { 0, 0, 0, 0, 0, 0 };


static void
ochusha_tree_selector_view_class_init(OchushaTreeSelectorViewClass *klass)
{
  GObjectClass *o_class = (GObjectClass *)klass;
  GtkObjectClass *object_class = (GtkObjectClass *)klass;
  GtkWidgetClass *widget_class = (GtkWidgetClass *)klass;

  parent_class = g_type_class_peek_parent(klass);

  /* GObject signals */
  o_class->finalize = ochusha_tree_selector_view_finalize;

  /* GtkObject signals */
  object_class->destroy = ochusha_tree_selector_view_destroy;

  /* GtkWidget signals */
  widget_class->motion_notify_event = ochusha_tree_selector_view_motion_notify;
  widget_class->leave_notify_event = ochusha_tree_selector_view_leave_notify;
  widget_class->button_press_event = ochusha_tree_selector_view_button_press;
  widget_class->button_release_event
    = ochusha_tree_selector_view_button_release;
  widget_class->key_press_event = ochusha_tree_selector_view_key_press_event;

  tree_selector_view_signals[ROW_MOUSE_OVER_SIGNAL]
    = g_signal_new("row_mouse_over",
		   G_TYPE_FROM_CLASS(klass),
		   G_SIGNAL_RUN_LAST,
		   G_STRUCT_OFFSET(OchushaTreeSelectorViewClass,
				   row_mouse_over),
		   NULL, NULL,
		   libochushagtk_marshal_BOOLEAN__BOXED_OBJECT_BOXED,
		   G_TYPE_BOOLEAN, 3,
		   GTK_TYPE_TREE_PATH | G_SIGNAL_TYPE_STATIC_SCOPE,
		   GTK_TYPE_TREE_VIEW_COLUMN,
		   GDK_TYPE_EVENT | G_SIGNAL_TYPE_STATIC_SCOPE);
  tree_selector_view_signals[ROW_MOUSE_OUT_SIGNAL]
    = g_signal_new("row_mouse_out",
		   G_TYPE_FROM_CLASS(klass),
		   G_SIGNAL_RUN_LAST,
		   G_STRUCT_OFFSET(OchushaTreeSelectorViewClass,
				   row_mouse_out),
		   NULL, NULL,
		   libochushagtk_marshal_BOOLEAN__BOXED,
		   G_TYPE_BOOLEAN, 1,
		   GTK_TYPE_TREE_PATH | G_SIGNAL_TYPE_STATIC_SCOPE);
  tree_selector_view_signals[ROW_BUTTON_PRESS_SIGNAL]
    = g_signal_new("row_button_press",
		   G_TYPE_FROM_CLASS(klass),
		   G_SIGNAL_RUN_LAST,
		   G_STRUCT_OFFSET(OchushaTreeSelectorViewClass,
				   row_button_press),
		   NULL, NULL,
		   libochushagtk_marshal_BOOLEAN__BOXED_OBJECT_BOXED,
		   G_TYPE_BOOLEAN, 3,
		   GTK_TYPE_TREE_PATH | G_SIGNAL_TYPE_STATIC_SCOPE,
		   GTK_TYPE_TREE_VIEW_COLUMN,
		   GDK_TYPE_EVENT | G_SIGNAL_TYPE_STATIC_SCOPE);
  tree_selector_view_signals[ROW_BUTTON_RELEASE_SIGNAL]
    = g_signal_new("row_button_release",
		   G_TYPE_FROM_CLASS(klass),
		   G_SIGNAL_RUN_LAST,
		   G_STRUCT_OFFSET(OchushaTreeSelectorViewClass,
				   row_button_release),
		   NULL, NULL,
		   libochushagtk_marshal_BOOLEAN__BOXED_OBJECT_BOXED,
		   G_TYPE_BOOLEAN, 3,
		   GTK_TYPE_TREE_PATH | G_SIGNAL_TYPE_STATIC_SCOPE,
		   GTK_TYPE_TREE_VIEW_COLUMN,
		   GDK_TYPE_EVENT | G_SIGNAL_TYPE_STATIC_SCOPE);
  tree_selector_view_signals[COLUMN_RESIZED_SIGNAL]
    = g_signal_new("column_resized",
		   G_TYPE_FROM_CLASS(klass),
		   G_SIGNAL_RUN_LAST,
		   G_STRUCT_OFFSET(OchushaTreeSelectorViewClass,
				   column_resized),
		   NULL, NULL,
		   libochushagtk_marshal_VOID__OBJECT_BOOLEAN_INT,
		   G_TYPE_NONE, 3,
		   GTK_TYPE_TREE_VIEW_COLUMN,
		   G_TYPE_BOOLEAN,
		   G_TYPE_INT);

  tree_selector_view_signals[CENTER_CURSOR_ON_SCREEN_SIGNAL]
    = g_signal_new("center_cursor_on_screen",
		   G_TYPE_FROM_CLASS(klass),
		   G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
		   G_STRUCT_OFFSET(OchushaTreeSelectorViewClass,
				   center_cursor_on_screen),
		   NULL, NULL,
		   libochushagtk_marshal_VOID__VOID,
		   G_TYPE_NONE, 0);

  klass->row_mouse_over = NULL;
  klass->row_mouse_out = NULL;
  klass->row_button_press = NULL;
  klass->row_button_release = NULL;
  klass->column_resized = NULL;
  klass->center_cursor_on_screen
    = ochusha_tree_selector_view_center_cursor_on_screen;
}


static void
ochusha_tree_selector_view_searchable_init(OchushaSearchableIface *iface)
{
  iface->find_next = ochusha_tree_selector_view_find_next;
  iface->terminate_search = ochusha_tree_selector_view_terminate_search;
}


static void
ochusha_tree_selector_view_init(OchushaTreeSelectorView *view)
{
  view->last_event_path = NULL;

  view->last_search_pos = NULL;
  view->last_key = NULL;
  view->last_key_id = 0;
  view->search_wrapped = FALSE;

  view->search_func = NULL;
  view->search_destroy = NULL;
  view->search_data = NULL;
}


static void
ochusha_tree_selector_view_finalize(GObject *object)
{
#if 0
  OchushaTreeSelectorView *view = (OchushaTreeSelectorView *)object;
#endif

  if (G_OBJECT_CLASS(parent_class)->finalize != NULL)
    (*G_OBJECT_CLASS(parent_class)->finalize)(object);
}


static void
ochusha_tree_selector_view_destroy(GtkObject *object)
{
  OchushaTreeSelectorView *view = (OchushaTreeSelectorView *)object;

  if (view->last_event_path != NULL)
    {
      gtk_tree_path_free(view->last_event_path);
      view->last_event_path = NULL;
    }

  if (view->search_destroy != NULL)
    {
      (*view->search_destroy)(view->search_data);
      view->search_destroy = NULL;
      view->search_data = NULL;
    }

  if (view->last_search_pos != NULL)
    {
      gtk_tree_row_reference_free(view->last_search_pos);
      view->last_search_pos = NULL;
    }

  if (GTK_OBJECT_CLASS(parent_class)->destroy != NULL)
    (*GTK_OBJECT_CLASS(parent_class)->destroy)(object);
}


static gboolean
ochusha_tree_selector_view_motion_notify(GtkWidget *widget,
					 GdkEventMotion *event)
{
  OchushaTreeSelectorView *view = OCHUSHA_TREE_SELECTOR_VIEW(widget);
  GtkTreePath *path = NULL;
  GtkTreeViewColumn *column = NULL;
  gboolean result = FALSE;
  GList *column_list;
  GList *list;
  gboolean use_resized_width = FALSE;
  gint resized_width = 0;

  if (!gtk_tree_view_get_path_at_pos(&view->tree_view, event->x, event->y,
				     &path, &column, NULL, NULL))
    goto done;

  if (view->last_event_path != NULL)
    {
      if (gtk_tree_path_compare(view->last_event_path, path) != 0)
	{
	  g_signal_emit(view, tree_selector_view_signals[ROW_MOUSE_OUT_SIGNAL],
			0,
			view->last_event_path,
			&result);
	  gtk_tree_path_free(view->last_event_path);
	  view->last_event_path = NULL;
	}
      else
	goto done;
    }

  result = FALSE;

  if (event->window != gtk_tree_view_get_bin_window(&view->tree_view))
    goto done;

  view->last_event_path = path;
  g_signal_emit(view, tree_selector_view_signals[ROW_MOUSE_OVER_SIGNAL],
		0,
		path, column, event,
		&result);

  path = NULL;

 done:
  if (path != NULL)
    gtk_tree_path_free(path);

  column_list = gtk_tree_view_get_columns(GTK_TREE_VIEW(widget));
  list = column_list;
  while (list != NULL)
    {
      column = list->data;
      if (column->window == event->window)
	break;
      column = NULL;
      list = list->next;
    }
  g_list_free(column_list);

  if (column != NULL)
    {
      use_resized_width = column->use_resized_width;
      resized_width = column->resized_width;
    }

  if (!result && GTK_WIDGET_CLASS(parent_class)->motion_notify_event != NULL)
    result = (*GTK_WIDGET_CLASS(parent_class)->motion_notify_event)(widget,
								    event);

  if (column != NULL
      && (use_resized_width != column->use_resized_width
	  || resized_width != column->resized_width))
    g_signal_emit(view, tree_selector_view_signals[COLUMN_RESIZED_SIGNAL],
		  0,
		  column, column->use_resized_width, column->resized_width);

  return result;
}


static gboolean
ochusha_tree_selector_view_leave_notify(GtkWidget *widget,
					GdkEventCrossing *event)
{
  OchushaTreeSelectorView *view = OCHUSHA_TREE_SELECTOR_VIEW(widget);
  GtkTreePath *path = NULL;
  gboolean result = FALSE;

  if (view->last_event_path != NULL)
    {
      g_signal_emit(view, tree_selector_view_signals[ROW_MOUSE_OUT_SIGNAL],
		    0,
		    view->last_event_path,
		    &result);
      gtk_tree_path_free(view->last_event_path);
      view->last_event_path = NULL;
    }

  if (path != NULL)
    gtk_tree_path_free(path);

  if (!result && GTK_WIDGET_CLASS(parent_class)->leave_notify_event != NULL)
    result = (*GTK_WIDGET_CLASS(parent_class)->leave_notify_event)(widget,
								   event);
  return result;
}


static gboolean
ochusha_tree_selector_view_button_press(GtkWidget *widget,
					GdkEventButton *event)
{
  OchushaTreeSelectorView *view = OCHUSHA_TREE_SELECTOR_VIEW(widget);
  GtkTreePath *path = NULL;
  GtkTreeViewColumn *column = NULL;
  gboolean result = FALSE;
  GList *column_list;
  GList *list;
  gboolean use_resized_width = FALSE;
  gint resized_width = 0;

  if (event->window != gtk_tree_view_get_bin_window(&view->tree_view))
    goto done;

  if (!gtk_tree_view_get_path_at_pos(&view->tree_view, event->x, event->y,
				     &path, &column, NULL, NULL))
    goto done;

  g_signal_emit(view, tree_selector_view_signals[ROW_BUTTON_PRESS_SIGNAL],
		0,
		path, column, event,
		&result);

 done:
  if (path != NULL)
    gtk_tree_path_free(path);

  column_list = gtk_tree_view_get_columns(GTK_TREE_VIEW(widget));
  list = column_list;
  while (list != NULL)
    {
      column = list->data;
      if (column->window == event->window)
	break;
      column = NULL;
      list = list->next;
    }
  g_list_free(column_list);

  if (column != NULL)
    {
      use_resized_width = column->use_resized_width;
      resized_width = column->resized_width;
    }

  if (!result && GTK_WIDGET_CLASS(parent_class)->button_press_event != NULL)
    result = (*GTK_WIDGET_CLASS(parent_class)->button_press_event)(widget,
								   event);

  if (column != NULL
      && (use_resized_width != column->use_resized_width
	  || resized_width != column->resized_width))
    g_signal_emit(view, tree_selector_view_signals[COLUMN_RESIZED_SIGNAL],
		  0,
		  column, column->use_resized_width, column->resized_width);

  return result;
}


static gboolean
ochusha_tree_selector_view_button_release(GtkWidget *widget,
					  GdkEventButton *event)
{
  OchushaTreeSelectorView *view = OCHUSHA_TREE_SELECTOR_VIEW(widget);
  GtkTreePath *path = NULL;
  GtkTreeViewColumn *column = NULL;
  gboolean result = FALSE;

  if (event->window != gtk_tree_view_get_bin_window(&view->tree_view))
    goto done;

  if (!gtk_tree_view_get_path_at_pos(&view->tree_view, event->x, event->y,
				     &path, &column, NULL, NULL))
    goto done;

  g_signal_emit(view, tree_selector_view_signals[ROW_BUTTON_RELEASE_SIGNAL],
		0,
		path, column, event,
		&result);

 done:
  if (path != NULL)
    gtk_tree_path_free(path);

  if (!result && GTK_WIDGET_CLASS(parent_class)->button_release_event != NULL)
    result = (*GTK_WIDGET_CLASS(parent_class)->button_release_event)(widget,
								     event);
  return result;
}


static gint
ochusha_tree_selector_view_key_press_event(GtkWidget *widget,
					   GdkEventKey *event)
{
  /*
   * ŪˤϡμťΤϥϡɥǥ󥰤ΤǤϤʤ
   * scheme٥˲widgetȤƤϤ˾٤ǤϤʤ
   * Ū˰ڤ̵뤹褦ˤƤŪˤ
   * key_press_event˥ϥɥꤹɤ
   */
  if (event->keyval > GDK_space && event->keyval <= GDK_ydiaeresis
      && (event->state
	  & (GDK_CONTROL_MASK | GDK_MOD1_MASK | GDK_MOD2_MASK | GDK_MOD3_MASK
	     | GDK_MOD4_MASK | GDK_MOD5_MASK)) == 0)
    {
      return TRUE;
    }

  if (GTK_WIDGET_CLASS(parent_class)->key_press_event != NULL)
    return GTK_WIDGET_CLASS(parent_class)->key_press_event(widget, event);

  return FALSE;
}


static void
ochusha_tree_selector_view_center_cursor_on_screen(GtkTreeView *view)
{
  GtkTreePath *path;
  g_return_if_fail(GTK_IS_TREE_VIEW(view));

  gtk_tree_view_get_cursor(view, &path, NULL);
  if (path == NULL)
    return;

  gtk_tree_view_scroll_to_cell(view, path, NULL, TRUE, 0.5, 0.0);
  gtk_tree_path_free(path);
}


GtkWidget *
ochusha_tree_selector_view_new(void)
{
  return g_object_new(OCHUSHA_TYPE_TREE_SELECTOR_VIEW, NULL);
}


GtkWidget *
ochusha_tree_selector_view_new_with_model(GtkTreeModel *model)
{
  return g_object_new(OCHUSHA_TYPE_TREE_SELECTOR_VIEW,
		      "model", model, NULL);
}


void
ochusha_tree_selector_view_set_search_equal_func(OchushaTreeSelectorView *view,
			OchushaTreeSelectorViewSearchEqualFunc search_func,
			gpointer search_data,
			GDestroyNotify search_destroy)
{
  g_return_if_fail(OCHUSHA_IS_TREE_SELECTOR_VIEW(view));

  if (view->search_destroy != NULL)
    (*view->search_destroy)(view->search_data);

  view->search_func = search_func;
  view->search_data = search_data;
  view->search_destroy = search_destroy;
}


/*
 * iterpre orderdepth firstˤ뼡ΰ֤ذư롣
 * ư(¸ߤ)ˤTRUE֤ʳξˤ
 * FALSE֤
 */
static gboolean
ochusha_tree_model_next_iter(GtkTreeModel *model, GtkTreeIter *iter)
{
  GtkTreeIter *tmp_iter;
  gboolean result = FALSE;

  g_return_val_if_fail(GTK_IS_TREE_MODEL(model), FALSE);
  g_return_val_if_fail(iter != NULL, FALSE);

  tmp_iter = gtk_tree_iter_copy(iter);

  if (gtk_tree_model_iter_has_child(model, iter))
    result = gtk_tree_model_iter_children(model, iter, tmp_iter);
  else
    {
      if (gtk_tree_model_iter_next(model, tmp_iter))
	result = gtk_tree_model_iter_next(model, iter);
      else
	{
	  GtkTreeIter tmp_parent;
	  gtk_tree_iter_free(tmp_iter);
	  tmp_iter = gtk_tree_iter_copy(iter);
	  while (gtk_tree_model_iter_parent(model, &tmp_parent, tmp_iter))
	    {
	      /* λǡtmp_iteriterοƤؤƤ */
	      gtk_tree_iter_free(tmp_iter);
	      tmp_iter = gtk_tree_iter_copy(&tmp_parent);
	      if (gtk_tree_model_iter_next(model, &tmp_parent))
		{
		  GtkTreePath *path = gtk_tree_model_get_path(model,
							      &tmp_parent);
		  result = gtk_tree_model_get_iter(model, iter, path);
		  gtk_tree_path_free(path);
		  break;
		}
	    }
	}
    }

  gtk_tree_iter_free(tmp_iter);

  return result;
}


/*
 * pathiterpost orderdepth firstˤΰ֤ذư롣
 * ư(¸ߤ)ˤTRUE֤ʳξˤ
 * FALSE֤
 */
static gboolean
ochusha_tree_model_prev_iter(GtkTreeModel *model, GtkTreeIter *iter)
{
  GtkTreePath *path;

  g_return_val_if_fail(GTK_IS_TREE_MODEL(model), FALSE);
  g_return_val_if_fail(iter != NULL, FALSE);

  path = gtk_tree_model_get_path(model, iter);

  if (!gtk_tree_path_prev(path))
    {
      gboolean result;
      if (gtk_tree_path_get_depth(path) > 1 && gtk_tree_path_up(path))
	{
	  result = TRUE;
	  gtk_tree_model_get_iter(model, iter, path);
	}
      else
	result = FALSE;

      gtk_tree_path_free(path);
      return result;
    }

  gtk_tree_model_get_iter(model, iter, path);
  while (gtk_tree_model_iter_has_child(model, iter))
    {
      int n_children = gtk_tree_model_iter_n_children(model, iter);
      GtkTreeIter *tmp_iter = gtk_tree_iter_copy(iter);
      gtk_tree_model_iter_nth_child(model, iter, tmp_iter, n_children - 1);
      gtk_tree_iter_free(tmp_iter);
    }
  gtk_tree_path_free(path);

  return TRUE;
}


#define ochusha_tree_model_get_iter_first(model, iter)	\
  gtk_tree_model_get_iter_first((model), (iter))


static gboolean
ochusha_tree_model_get_iter_last(GtkTreeModel *model, GtkTreeIter *iter)
{
  GtkTreeIter *tmp_iter = NULL;

  g_return_val_if_fail(GTK_IS_TREE_MODEL(model), FALSE);
  g_return_val_if_fail(iter != NULL, FALSE);

  while (tmp_iter == NULL || gtk_tree_model_iter_has_child(model, tmp_iter))
    {
      int n_children = gtk_tree_model_iter_n_children(model, tmp_iter);
      gtk_tree_model_iter_nth_child(model, iter, tmp_iter, n_children - 1);
      if (tmp_iter != NULL)
	gtk_tree_iter_free(tmp_iter);
      tmp_iter = gtk_tree_iter_copy(iter);
    }
  if (tmp_iter == NULL)
    return FALSE;	/* ä */

  gtk_tree_iter_free(tmp_iter);
  return TRUE;
}


static gboolean
ochusha_tree_selector_view_find_next(OchushaSearchable *searchable,
				     OchushaSearchKey *key)
{
  OchushaTreeSelectorView *view = OCHUSHA_TREE_SELECTOR_VIEW(searchable);
  gboolean include_current_pos
    = (view->last_key != key || key->key_id != view->last_key_id
       || view->search_wrapped);
  GtkTreeModel *model = gtk_tree_view_get_model(&view->tree_view);
  GtkTreeSelection *selection = gtk_tree_view_get_selection(&view->tree_view);
  GtkTreePath *path;
  GtkTreeIter iter;

  g_return_val_if_fail(view->last_key == NULL || view->last_key == key, FALSE);
  g_return_val_if_fail(view->search_func != NULL, FALSE);

  view->last_key = key;
  view->last_key_id = key->key_id;
  view->search_wrapped = FALSE;

  if (key->key_string == NULL || *key->key_string == '\0')
    return TRUE;

  gtk_tree_selection_unselect_all(selection);

  if (view->last_search_pos != NULL
      && !gtk_tree_row_reference_valid(view->last_search_pos))
    {
      gtk_tree_row_reference_free(view->last_search_pos);
      view->last_search_pos = NULL;
    }

  if (view->last_search_pos != NULL)
    {
      path = gtk_tree_row_reference_get_path(view->last_search_pos);
      gtk_tree_model_get_iter(model, &iter, path);
      gtk_tree_path_free(path);
    }
  else
    {
      if (key->direction == OCHUSHA_SEARCH_DIRECTION_FORWARD)
	{
	  if (!ochusha_tree_model_get_iter_first(model, &iter))
	    return TRUE;	/* ä */
	}
      else
	{
	  if (!ochusha_tree_model_get_iter_last(model, &iter))
	    return TRUE;	/* ä */
	}
    }

  do
    {
      if (!include_current_pos)
	{
	  if (key->direction == OCHUSHA_SEARCH_DIRECTION_FORWARD)
	    {
	      if (!ochusha_tree_model_next_iter(model, &iter))
		{
		  view->search_wrapped = TRUE;
		  if (view->last_search_pos != NULL)
		    {
		      gtk_tree_row_reference_free(view->last_search_pos);
		      view->last_search_pos = NULL;
		    }
		  return FALSE;
		}
	    }
	  else
	    {
	      if (!ochusha_tree_model_prev_iter(model, &iter))
		{
		  view->search_wrapped = TRUE;
		  if (view->last_search_pos != NULL)
		    {
		      gtk_tree_row_reference_free(view->last_search_pos);
		      view->last_search_pos = NULL;
		    }
		  return FALSE;
		}
	    }
	}
      include_current_pos = FALSE;
    }
  while (!(*view->search_func)(view, model, &iter, key, view->search_data));

  /* iterΰ֤˥ޥåˤΤߤ롣 */
  path = gtk_tree_model_get_path(model, &iter);
  gtk_tree_view_expand_to_path(&view->tree_view, path);
  gtk_tree_view_scroll_to_cell(&view->tree_view, path, NULL, TRUE, 0.5f, 0.0f);
  gtk_tree_selection_select_iter(selection, &iter);
  gtk_tree_view_set_cursor(&view->tree_view, path, NULL, FALSE);
  view->last_search_pos = gtk_tree_row_reference_new(model, path);
  gtk_tree_path_free(path);

  return TRUE;
}


static void
ochusha_tree_selector_view_terminate_search(OchushaSearchable *searchable,
					    OchushaSearchKey *key)
{
  OchushaTreeSelectorView *view = OCHUSHA_TREE_SELECTOR_VIEW(searchable);
  if (view->last_search_pos != NULL)
    {
      gtk_tree_row_reference_free(view->last_search_pos);
      view->last_search_pos = NULL;
    }
  view->last_key = NULL;
  view->last_key_id = 0;
  view->search_wrapped = FALSE;
}
