/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: bytecode_buffer.h,v 1.11 2009/01/02 03:07:32 fuyu Exp $
 */

#ifndef _LIBTSCORE_BYTECODE_BUFFER_H_
#define _LIBTSCORE_BYTECODE_BUFFER_H_

#ifdef __cplusplus

#if defined(__GNUG__) && !defined(__ICC)
#pragma interface
#endif

#include "config.h"
#include "libtscore/virtual_machine.h"
#include "libtscore/vmopcode.h"

#include <cassert>
#include <cstddef>


using namespace std;
namespace LibTSCore
{

  /*
   * BytecodeBuffer manages a sequence of bytecodes for each Scheme procedure.
   */
  class BytecodeBuffer
  {
  private:
    char *buffer;
    size_t buffer_length;
    size_t offset;
    bool include_forward_branch;

    static const size_t INITIAL_BYTECODE_BUFFER_CAPACITY = 16;


    void enlarge_buffer(size_t capacity);
    inline void ensure_capacity(size_t capacity)
    {
      if (buffer_length >= capacity)
	return;

      enlarge_buffer(capacity);
    }


  public:
    BytecodeBuffer();
    ~BytecodeBuffer();

    void optimize_bytecode();

    inline void set_include_forward_branch() { include_forward_branch = true; }
    inline const char *get_bytecode() const { return buffer; }
    inline size_t get_current_offset() const { return offset; }
    inline void set_opcode(size_t off, TSVMOpcode bytecode)
    {
      assert(off >= 0 && off < buffer_length);

      buffer[off] = static_cast<char>(bytecode);
    }

    inline void set_unsigned_8(size_t off, unsigned long value)
    {
      assert(value >= 0 && value <= 0xff);

      buffer[off] = static_cast<char>(value & 0xff);
    }

    inline void set_unsigned_16(size_t off, unsigned long value)
    {
      assert(value >= 0 && value <= 0xffff);

      buffer[off++] = static_cast<char>(value & 0xff);
      buffer[off] = static_cast<char>((value >> 8) & 0xff);
    }

    inline void set_unsigned_32(size_t off, unsigned long value)
    {
#if SIZEOF_LONG > 4
      assert(value >= 0 && value <= 0xffffffff);
#endif
      buffer[off++] = static_cast<char>(value & 0xff);
      buffer[off++] = static_cast<char>((value >> 8) & 0xff);
      buffer[off++] = static_cast<char>((value >> 16) & 0xff);
      buffer[off] = static_cast<char>((value >> 24) & 0xff);
    }

#if SIZEOF_LONG > 4
    inline void set_unsigned_64(size_t off, unsigned long value)
    {
      buffer[off++] = static_cast<char>(value & 0xff);
      buffer[off++] = static_cast<char>((value >> 8) & 0xff);
      buffer[off++] = static_cast<char>((value >> 16) & 0xff);
      buffer[off++] = static_cast<char>((value >> 24) & 0xff);
      buffer[off++] = static_cast<char>((value >> 32) & 0xff);
      buffer[off++] = static_cast<char>((value >> 40) & 0xff);
      buffer[off++] = static_cast<char>((value >> 48) & 0xff);
      buffer[off] = static_cast<char>((value >> 56) & 0xff);
      return;
    }
#endif

    inline void set_signed_8(size_t off, long value)
    {
      assert(value >= -128 && value <= 127);

      buffer[off] = static_cast<char>(value & 0xff);
    }

    inline void set_signed_16(size_t off, long value)
    {
      assert(value >= -32768 && value <= 32767);

      buffer[off++] = static_cast<char>(value & 0xff);
      buffer[off] = static_cast<char>((value >> 8) & 0xff);
    }

    inline void set_signed_32(size_t off, long value)
    {
#if SIZEOF_LONG > 4
      assert(value >= -2147483648 && value <= 2147483647);
#endif
      buffer[off++] = static_cast<char>(value & 0xff);
      buffer[off++] = static_cast<char>((value >> 8) & 0xff);
      buffer[off++] = static_cast<char>((value >> 16) & 0xff);
      buffer[off] = static_cast<char>((value >> 24) & 0xff);
    }

#if SIZEOF_LONG > 4
    inline void set_signed_64(size_t off, long value)
    {
      buffer[off++] = static_cast<char>(value & 0xff);
      buffer[off++] = static_cast<char>((value >> 8) & 0xff);
      buffer[off++] = static_cast<char>((value >> 16) & 0xff);
      buffer[off++] = static_cast<char>((value >> 24) & 0xff);
      buffer[off++] = static_cast<char>((value >> 32) & 0xff);
      buffer[off++] = static_cast<char>((value >> 40) & 0xff);
      buffer[off++] = static_cast<char>((value >> 48) & 0xff);
      buffer[off] = static_cast<char>((value >> 56) & 0xff);
    }
#endif

    inline size_t append(TSVMOpcode bytecode)
    {
      ensure_capacity(offset + 1);

      buffer[offset++] = static_cast<char>(bytecode);
      return 1;
    }

    inline size_t append_unsigned_8(unsigned long value)
    {
      assert(value >= 0 && value <= 0xff);

      ensure_capacity(offset + 1);

      buffer[offset++] = static_cast<char>(value & 0xff);
      return 1;
    }
    inline size_t append_unsigned_16(unsigned long value)
    {
      assert(value >= 0 && value <= 0xffff);

      ensure_capacity(offset + 2);

      buffer[offset++] = static_cast<char>(value & 0xff);
      buffer[offset++] = static_cast<char>((value >> 8) & 0xff);
      return 2;
    }
    inline size_t append_unsigned_32(unsigned long value)
    {
#if SIZEOF_LONG > 4
      assert(value >= 0 && value <= 0xffffffff);
#endif
      ensure_capacity(offset + 4);

      buffer[offset++] = static_cast<char>(value & 0xff);
      buffer[offset++] = static_cast<char>((value >> 8) & 0xff);
      buffer[offset++] = static_cast<char>((value >> 16) & 0xff);
      buffer[offset++] = static_cast<char>((value >> 24) & 0xff);
      return 4;
    }

#if SIZEOF_LONG > 4
    inline size_t append_unsigned_64(unsigned long value)
    {
      ensure_capacity(offset + 8);

      buffer[offset++] = static_cast<char>(value & 0xff);
      buffer[offset++] = static_cast<char>((value >> 8) & 0xff);
      buffer[offset++] = static_cast<char>((value >> 16) & 0xff);
      buffer[offset++] = static_cast<char>((value >> 24) & 0xff);
      buffer[offset++] = static_cast<char>((value >> 32) & 0xff);
      buffer[offset++] = static_cast<char>((value >> 40) & 0xff);
      buffer[offset++] = static_cast<char>((value >> 48) & 0xff);
      buffer[offset++] = static_cast<char>((value >> 56) & 0xff);
      return 8;
    }
#endif

    inline size_t append_signed_8(long value)
    {
      assert(value >= -128 && value <= 127);

      ensure_capacity(offset + 1);

      buffer[offset++] = static_cast<char>(value & 0xff);
      return 1;
    }
    inline size_t append_signed_16(long value)
    {
      assert(value >= -32768 && value <= 32767);

      ensure_capacity(offset + 2);

      buffer[offset++] = static_cast<char>(value & 0xff);
      buffer[offset++] = static_cast<char>((value >> 8) & 0xff);
      return 2;
    }
    inline size_t append_signed_32(long value)
    {
#if SIZEOF_LONG > 4
      assert(value >= -2147483648 && value <= 2147483647);
#endif
      ensure_capacity(offset + 4);

      buffer[offset++] = static_cast<char>(value & 0xff);
      buffer[offset++] = static_cast<char>((value >> 8) & 0xff);
      buffer[offset++] = static_cast<char>((value >> 16) & 0xff);
      buffer[offset++] = static_cast<char>((value >> 24) & 0xff);
      return 4;
    }

#if SIZEOF_LONG > 4
    inline size_t append_signed_64(long value)
    {
      ensure_capacity(offset + 8);

      buffer[offset++] = static_cast<char>(value & 0xff);
      buffer[offset++] = static_cast<char>((value >> 8) & 0xff);
      buffer[offset++] = static_cast<char>((value >> 16) & 0xff);
      buffer[offset++] = static_cast<char>((value >> 24) & 0xff);
      buffer[offset++] = static_cast<char>((value >> 32) & 0xff);
      buffer[offset++] = static_cast<char>((value >> 40) & 0xff);
      buffer[offset++] = static_cast<char>((value >> 48) & 0xff);
      buffer[offset++] = static_cast<char>((value >> 56) & 0xff);
      return 8;
    }
#endif

    size_t append_unsigned(unsigned long value);
    size_t append_signed(long value);
  };
}

#endif

#endif	/* _LIBTSCORE_BYTECODE_BUFFER_H_ */
