/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: cell.cc,v 1.46 2010/08/15 15:39:47 fuyu Exp $
 */

#include "config.h"

#include <oniguruma.h>

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libtscore/cell.h"

#include "libtscore/memory_system.h"
#include "libtscore/procedure.h"
#include "libtscore/r5rs_misc_procedure.h"
#include "libtscore/syntax.h"


#include <cctype>
#include <cstdio>


namespace LibTSCore
{

#ifdef __GNUC__
struct SpecialCellInitializer
{
  static Cell nil;
  static Cell t;
  static Cell f;
  static Cell eof;
  static Cell unspecified;
  static Cell sink;
  static Cell trampoline;

  SpecialCellInitializer();
};


Cell __attribute__((aligned(sizeof(Cell)))) SpecialCellInitializer::nil;
Cell __attribute__((aligned(sizeof(Cell)))) SpecialCellInitializer::t;
Cell __attribute__((aligned(sizeof(Cell)))) SpecialCellInitializer::f;
Cell __attribute__((aligned(sizeof(Cell)))) SpecialCellInitializer::eof;
Cell __attribute__((aligned(sizeof(Cell)))) SpecialCellInitializer::unspecified;
Cell __attribute__((aligned(sizeof(Cell)))) SpecialCellInitializer::sink;
Cell __attribute__((aligned(sizeof(Cell)))) SpecialCellInitializer::trampoline;


Cell *Cell::nil_cell = &SpecialCellInitializer::nil;
Cell *Cell::t_cell = &SpecialCellInitializer::t;
Cell *Cell::f_cell = &SpecialCellInitializer::f;
Cell *Cell::eof_cell = &SpecialCellInitializer::eof;
Cell *Cell::unspecified_cell = &SpecialCellInitializer::unspecified;
Cell *Cell::sink_cell = &SpecialCellInitializer::sink;
Cell *Cell::trampoline_cell = &SpecialCellInitializer::trampoline;


SpecialCellInitializer::SpecialCellInitializer()
{
  nil.mk_constant();
  t.mk_constant();
  f.mk_constant();
  eof.mk_constant();
  unspecified.mk_constant();
  sink.mk_constant();
  trampoline.mk_constant();
}

static SpecialCellInitializer dummy;
#else
Cell *Cell::nil_cell;
Cell *Cell::t_cell;
Cell *Cell::f_cell;
Cell *Cell::eof_cell;
Cell *Cell::unspecified_cell;
Cell *Cell::sink_cell;
Cell *Cell::trampoline_cell;


struct SpecialCellInitializer
{
  static Cell constants[8];
  SpecialCellInitializer()
  {
    char *pointer = reinterpret_cast<char *>(&constants);
    size_t mod = reinterpret_cast<intptr_t>(pointer) % sizeof(Cell);
    Cell *aligned
      = reinterpret_cast<Cell *>(mod > 0
				 ? pointer + sizeof(Cell) - mod : pointer);

    Cell::nil_cell = (&aligned[0])->mk_constant();
    Cell::t_cell = (&aligned[1])->mk_constant();
    Cell::f_cell = (&aligned[2])->mk_constant();
    Cell::eof_cell = (&aligned[3])->mk_constant();
    Cell::unspecified_cell = (&aligned[4])->mk_constant();
    Cell::sink_cell = (&aligned[5])->mk_constant();
    Cell::trampoline_cell = (&aligned[6])->mk_constant();
  }
};
Cell SpecialCellInitializer::constants[8];
static SpecialCellInitializer dummy;
#endif


static const char * const cell_type_names[] =
{
  "free",
  "number",
  "character",
  "string",
  "foreign object",
  "native procedure",
  NULL,
  "port",
  "time",
  "date",
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  "symbol",
  "pair",
  "vector",
  "constant pool",
  "substring",
  "environment",
  "binding",
  "regexp",
  "syntax",
  "macro transformer",
  "closure template",
  "closure",
  "continuation",
  "promise",
  "forced promise",
  "toplevel frame"
};


static inline bool
is_valid_cell_type(long type)
{
  if (type >= 0 && type <= Cell::TOPLEVEL_FRAME_CELL
      && cell_type_names[type] != NULL)
    return true;

  return false;
}


const char *
Cell::get_cell_type_string() const
{
  long type = cell_type();
  if (is_valid_cell_type(type))
    return cell_type_names[type];

  return "!!! INVALID TYPE CELL !!!";
}


char *
Cell::to_string() const
{
  OutputStringPort string_buf;
  to_string(&string_buf);
  return strdup(string_buf.get_output_string());
}


Port *
Cell::to_string(Port *port, bool toplevel) const
{
  char buf[256];
  long type = cell_type();
  switch (type)
    {
    case NUMBER_CELL:
      if (is_integer())
	{
	  snprintf(buf, sizeof(buf), "%ld", get_integer0());
	  port->append(buf);
	  break;
	}
      else
	{
	  snprintf(buf, sizeof(buf), "%lf", get_real());
	  port->append(buf);
	  break;
	}

    case CHARACTER_CELL:
      {
	long c = get_character();
	switch (c)
	  {
	  case ' ':
	    port->append("#\\space");
	    break;

	  case '\n':
	    port->append("#\\newline");
	    break;

	  case '\r':
	    port->append("#\\return");
	    break;

	  case '\t':
	    port->append("#\\tab");
	    break;

	  default:
	    if (isprint(static_cast<int>(c)))
	      port->append("#\\")->append(c);
	    else
	      {
		snprintf(buf, sizeof(buf), "#\\x%lx", c & 0xff);
		port->append(buf);
	      }
	  }
	break;
      }

    case STRING_CELL:
    case SUBSTRING_CELL:
      {
	port->append('"');
	const char *s = get_string();
	size_t len = get_strlen();
	for (size_t i = 0; i < len; i++)
	  {
	    switch (*s)
	      {
	      case '\n':
		port->append("\\n");
		break;

	      case '\r':
		port->append("\\r");
		break;

	      case '\t':
		port->append("\\t");
		break;

	      case '"':
		port->append("\\\"");
		break;

	      case '\\':
		port->append("\\\\");
		break;

	      default:
		port->append(*s);
	      }
	    s++;
	  }

	port->append('"');
	break;
      }

    case FOREIGN_OBJECT_CELL:
      if (get_foreign_destructor() != NULL)
	snprintf(buf, sizeof(buf), "#<foreign: data=%p, destructor=%p, option_data=%p>",
		 get_foreign_pointer(),
		 get_foreign_destructor(), get_foreign_option_data());
      else
	snprintf(buf, sizeof(buf), "#<foreign: data=%p>",
		 get_foreign_pointer());

      port->append(buf);
      break;

    case NATIVE_PROCEDURE_CELL:
      snprintf(buf, sizeof(buf), "#<native procedure: %s>",
	       get_native_procedure()->get_procedure_name(this));
      port->append(buf);
      break;

    case PORT_CELL:
      snprintf(buf, sizeof(buf), "#<port: %p>", get_port());
      port->append(buf);
      break;

    case TIME_CELL:
      {
	snprintf(buf, sizeof(buf),
		 "#<time: type=%ld, nanosecond=%ld, second=%ld>",
		 get_time_type(),
		 get_time_nanosecond(),
		 get_time_second());
	port->append(buf);
	break;
      }

#ifdef HAVE_TIME_H
    case DATE_CELL:
      {
	const struct tm *val = get_date();
	snprintf(buf, sizeof(buf), "#<date: %p={ tm_sec=%d, tm_min=%d, tm_hour=%d, tm_mday=%d, tm_mon=%d, tm_year=%d, tm_wday=%d, tm_yday=%d, tm_isdst=%d }, nanosecond=%ld, zone_offset=%ld>",
		 val, val->tm_sec, val->tm_min, val->tm_hour, val->tm_mday,
		 val->tm_mon, val->tm_year, val->tm_wday, val->tm_yday,
		 val->tm_isdst,
		 get_date_nanosecond(), get_date_zone_offset());
	port->append(buf);
	break;
      }
#endif

    case SYMBOL_CELL:
      port->append(get_symname());
      break;

    case PAIR_CELL:
      if (toplevel)
	{
	  port->append('(');
	  get_car()->to_string(port, false);
	  port->append(" . ");
	  get_cdr()->to_string(port, false);
	  port->append(')');
	}
      else
	{
	  snprintf(buf, sizeof(buf),
		   "#<pair: car=%p cdr=%p>", get_car(), get_cdr());
	  port->append(buf);
	}
      break;

    case VECTOR_CELL:
      snprintf(buf, sizeof(buf), "#<vector: length=%ld>", get_vector_length());
      port->append(buf);
      break;

    case BINDING_CELL:
      port->append(get_car()->get_symname());
      break;

    case REGEXP_CELL:
      {
	port->append("#/");
	Cell *rx_string = get_regexp_string();
	const char *s = rx_string->get_string();
	size_t len = rx_string->get_strlen();
	for (size_t i = 0; i < len; i++)
	  {
	    if (s[i] == '/')
	      port->append('\\');

	    port->append(s[i]);
	  }

	port->append('/');
	break;
      }

    case SYNTAX_CELL:
      port->append("#<syntax: ")->append(get_car()->get_symname());
      if (!is_macro_syntax())
	port->append('>');
      else
	{
	  char buf[256];
	  snprintf(buf, sizeof(buf), ": transformer=%p>",
		   get_macro_transformer());
	  port->append(buf);
	}

      break;

    case CONTINUATION_CELL:
      {
	char buf[256];
	if (is_full_continuation())
	  snprintf(buf, sizeof(buf),
		   "#<continuation: %p, proc=%p, frame=%p, pc=%p, inport=%p, outport=%p, depth=%ld>",
		   this, get_continuation_procedure(),
		   get_continuation_frame(), get_continuation_pc(),
		   get_continuation_input_port(),
		   get_continuation_output_port(), get_continuation_depth());
	else
	  snprintf(buf, sizeof(buf),
		   "#<continuation: %p, proc=%p, frame=%p, pc=%p>",
		   this, get_continuation_procedure(),
		   get_continuation_frame(), get_continuation_pc());

	port->append(buf);
	break;
      }

    case PROMISE_CELL:
      if (is_forced_promise())
	{
	  port->append("#<promise(forced): ");
	  get_memoized_value()->to_string(port, false);
	  port->append(">");
	}
      else
	{
	  port->append("#<promise: ");
	  get_promise_procedure()->to_string(port, false);
	  port->append(">");
	}
      break;

    case MACRO_TRANSFORMER_CELL:
    case RESERVED_ATOM0_CELL:
    case RESERVED_ATOM1_CELL:
    case RESERVED_ATOM2_CELL:
    case RESERVED_ATOM3_CELL:
    case RESERVED_ATOM4_CELL:
    case FREE_CELL:
    case RESERVED_NON_ATOM_CELL:
    default:
      if (this == Cell::nil())
	port->append("()");
      else if (this == Cell::t())
	port->append("#t");
      else if (this == Cell::f())
	port->append("#f");
      else if (this == Cell::eof())
	port->append("#eof");
      else if (this == Cell::unspecified())
	port->append("#<unspecified>");
      else
	{
	  snprintf(buf, sizeof(buf), "#<%s: %p>",
		   get_cell_type_string(), this);
	  port->append(buf);
	}

      break;
    }

  return port;
}


Cell::Cell()
{
  set_flags(GC_MARK_BIT);
  contents.cell.object.cons.car = this;
  contents.cell.object.cons.cdr = this;
}


Cell *
Cell::finalize()
{
  assert(!is_pseudo_cell());

  long type = cell_type();
  assert(is_valid_cell_type(type));

  switch (type)
    {
    case STRING_CELL:
      if ((get_flags() & STATIC_STRING_FLAG) == 0)
	free(const_cast<char *>(get_string()));

      break;

    case PORT_CELL:
      delete get_port();
      break;

    case DATE_CELL:
      free(const_cast<struct tm *>(get_date()));
      break;

    case FOREIGN_OBJECT_CELL:
      if (get_foreign_destructor() != NULL)
	(*get_foreign_destructor())(get_foreign_pointer(),
				    get_foreign_option_data());

      break;

    case REGEXP_CELL:
      if (get_regexp() != NULL)
	onig_free(get_regexp());

      break;

    case ENVIRONMENT_CELL:
      delete get_environment();
      break;

    case CONSTANT_POOL_CELL:
      delete get_constant_pool();
      break;

    case SYNTAX_CELL:
      if ((get_flags() & STATIC_SYNTAX_FLAG) == 0)
	delete get_syntax();

      break;

    case CLOSURE_TEMPLATE_CELL:
      delete get_closure_bytecode();
      break;

    case VECTOR_CELL:
      {
	long size = get_vector_size();
	mk_free_cell();
	Cell *elm = this;
	for (long i = 0; i < size; i++)
	  {
	    elm->mk_free_cell(this + i + 1);
	    elm = this + i + 1;
	    elm->set_flags(FREE_CELL | GC_MARK_BIT);
	  }

	return elm->mk_free_cell();
      }

    default:
      break;
    }

  set_flags(FREE_CELL | GC_MARK_BIT);
  return this;
}


Cell *
Cell::mk_vector(long len, Cell *fill)
{
  set_flags(VECTOR_CELL);
  contents.cell.object.cons.car = Cell::nil();
  contents.cell.object.cons.cdr = Cell::nil();
  contents.cell.object.cons.option.length = len;

  long size = get_vector_size(len);
  for (long i = 0; i < size; i++)
    {
      Cell *slot = this + 1 + i;
      slot->contents.vector.elements[0] = fill;
      slot->contents.vector.elements[1] = fill;
      slot->contents.vector.elements[2] = fill;
      slot->contents.vector.elements[3] = fill;
    }

  return this;
}


Cell *
Cell::mk_substring_copy(Cell *str, Cell *offset, Cell *len)
{
  size_t copy_len = len->get_integer();
  char *copy_str = static_cast<char *>(malloc(copy_len + 1));
  if (copy_str == NULL)
    {
      // emergency!
      // Here we blindly share the storage for substring of mutable string.
      set_flags(SUBSTRING_CELL | IMMUTABLE_FLAG);
      contents.cell.object.cons.car = str;
      contents.cell.object.cons.cdr = len;
      contents.cell.object.cons.option.substring
	= str->get_string() + offset->get_integer();
      return this;
    }
  memcpy(copy_str, str->get_string() + offset->get_integer(), copy_len);
  copy_str[copy_len] = '\0';
  return mk_string(copy_str, copy_len)->set_immutable();
}


long
Cell::string_hash()
{
  bool cachable;
  const char *str;
  const char *tail;
  long h;
  if (cell_type() == STRING_CELL)
    {
      if (is_immutable())
	{
	  h = contents.cell.object.string.hash;
	  if (h != 0)
	    return h;

	  cachable = true;
	}
      else
	{
	  cachable = false;
	  h = 0;
	}

      str = contents.cell.object.string.pointer;
      tail = str + contents.cell.object.string.length;
    }
  else
    {
      cachable = false;
      str = contents.cell.object.cons.option.substring;
      tail = str + contents.cell.object.cons.cdr->get_integer0();
      h = 0;
    }

  while (str < tail)
    h = 31 * h + *str++;

  if (cachable)
    contents.cell.object.string.hash = h;

  return h;
}


bool
Cell::string_equals(Cell *str)
{
  long this_type = cell_type();
  size_t this_len;
  const char *this_str;
  if (this_type == STRING_CELL)
    {
      this_str = contents.cell.object.string.pointer;
      this_len = contents.cell.object.string.length;
    }
  else if (this_type == SUBSTRING_CELL)
    {
      this_str = contents.cell.object.cons.option.substring;
      this_len = contents.cell.object.cons.cdr->get_integer0();
    }
  else
    return false;

  long str_type = str->cell_type();
  size_t str_len;
  const char *str_str;
  if (str_type == STRING_CELL)
    {
      str_str = str->contents.cell.object.string.pointer;
      str_len = str->contents.cell.object.string.length;
    }
  else if (str_type == SUBSTRING_CELL)
    {
      str_str = str->contents.cell.object.cons.option.substring;
      str_len = str->contents.cell.object.cons.cdr->get_integer0();
    }
  else
    return false;

  if (this_len != str_len)
    return false;

  return strncmp(this_str, str_str, this_len) == 0;
}


bool
Cell::string_equals(const char *str)
{
  long this_type = cell_type();
  size_t this_len;
  const char *this_str;
  if (this_type == STRING_CELL)
    {
      this_str = contents.cell.object.string.pointer;
      this_len = contents.cell.object.string.length;
    }
  else if (this_type == SUBSTRING_CELL)
    {
      this_str = contents.cell.object.cons.option.substring;
      this_len = contents.cell.object.cons.cdr->get_integer0();
    }
  else
    return false;

  size_t str_len = strlen(str);
  if (this_len != str_len)
    return false;

  return strncmp(this_str, str, this_len) == 0;
}


long
Cell::string_hash(const char *str, size_t len)
{
  const char *tail = str + len;
  long h = 0;
  while (str < tail)
    h = 31 * h + *str++;

  return h;
}


long
Cell::symbol_hash(const char *str, size_t len)
{
  const char *tail = str + len;
  long h = 0;
  while (str < tail)
    h = 31 * h + tolower(*str++);

  return h;
}


Cell *
Cell::mk_regexp(Cell *rx_string, bool ignore_case)
{
  assert(!is_pseudo_cell());

  contents.cell.object.cons.option.rx = NULL;
  OnigErrorInfo err_info;
  UChar *regexp
    = reinterpret_cast<UChar *>(const_cast<char *>(rx_string->get_string()));
  int err_code = onig_new(&contents.cell.object.cons.option.rx,
			  regexp, regexp + rx_string->get_strlen(),
			  (ignore_case ? ONIG_OPTION_IGNORECASE : 0),
			  ONIG_ENCODING_UTF8,
			  ONIG_SYNTAX_RUBY, &err_info);
  if (err_code == ONIG_NORMAL)
    {
      set_flags(REGEXP_CELL);
      contents.cell.object.cons.car = rx_string;
      contents.cell.object.cons.cdr = this;
      return this;
    }

  UChar err_buf[ONIG_MAX_ERROR_MESSAGE_LEN];
  int err_len = onig_error_code_to_str(err_buf, err_code, &err_info);
  char *err_str = static_cast<char *>(calloc(err_len + 1, sizeof(char)));
  return mk_string(strncpy(err_str, reinterpret_cast<const char *>(err_buf),
			   err_len), err_len);
}


Cell *
Cell::toplevel_environment_new(MemorySystem &ms)
{
  Environment *hash_env = new HashEnvironment(ms);
  Cell *bindings = ms.get_vector(113);
  Cell *env = ms.get_cell(bindings);
  if (env != Cell::sink())
    return env->mk_environment(Cell::nil(), hash_env, bindings);

  return Cell::nil();
}


Cell *
Cell::closed_environment_new(MemorySystem &ms, const Cell *parent)
{
  Environment *list_env = new ListEnvironment(ms);
  Cell *par = const_cast<Cell *>(parent);
  Cell *env = ms.get_cell(par);
  if (env != Cell::sink())
    return env->mk_environment(par, list_env, Cell::nil());

  return Cell::nil();
}


long
Cell::calculate_environment_distance(const Cell *environment) const
{
  assert(environment->is_environment());

  long distance = 0;
  const Cell *env = this;
  while (env != Cell::nil())
    {
      assert(env->is_environment());

      if (environment == env)
	return distance;

      env = env->get_parent_environment();
      ++distance;
    }

  return -1;	// should never reached here!
}


Cell *
Cell::toplevel_frame_new(MemorySystem &ms, long initial_frame_size)
{
  Cell *storage = ms.get_vector(round_vector_size(initial_frame_size),
				Cell::unspecified0());
  if (storage == Cell::sink())
    return Cell::nil();

  assert(storage->is_vector());

  Cell *frame = ms.get_cell(storage);
  return frame->mk_toplevel_frame(storage);
}


Cell *
Cell::scheme_frame_new(MemorySystem &ms, Cell *parent, long frame_size)
{
  assert(frame_size > 0);
  assert(round_vector_size(frame_size) > 0);

  StackRoot par(ms, parent);
  Cell *frame = ms.get_vector_from_young_space(round_vector_size(frame_size),
					       Cell::unspecified0());
  if (frame != Cell::sink())
    return frame->mk_scheme_frame(par);

  return Cell::nil();
}


void
ConstantPool::reset(Cell *constant_pool)
{
  Cell *table = constant_pool->get_constant_table();
  if (table != Cell::nil())
    {
      long n_items = get_number_of_items();
      for (long i = 0; i < n_items; i++)
	memory_system.set_vector_element(table, i, Cell::unspecified());
    }

  number_of_items = 0;
}


long
ConstantPool::register_constant(Cell *constant_pool, const Cell *value)
{
  Cell *table = constant_pool->get_constant_table();
  long n_items = get_number_of_items();
  for (long i = 0; i < n_items; i++)
    if (R5RSMiscProcedure::equal(table->get_vector_element(i), value))
      return i;

  long slot = new_item_slot();
#if 0
  fprintf(stderr, "register_constant(%p, %p) => %ld\n",
	  constant_pool, value, slot);
#endif
  long len;
  if (table != Cell::nil())
    len = table->get_vector_length();
  else
    len = 0;

  if (slot < len)
    {
      memory_system.set_vector_element(table, slot, const_cast<Cell *>(value));
      return slot;
    }

  StackRoot cpool(memory_system, constant_pool);
  StackRoot val(memory_system, const_cast<Cell *>(value));
  if (table != Cell::nil())
    {
      // Should we enlarge tables more than 4 at once?
      table = memory_system.copy_vector(len + 4, table, Cell::unspecified0());
    }
  else
    {
      table = memory_system.get_vector(4, Cell::unspecified0());
    }

  assert(slot < table->get_vector_length());

  memory_system.set_car(cpool, table);
  memory_system.set_vector_element(table, slot, val);
  return slot;
}


Cell *
Cell::constant_pool_new(MemorySystem &ms)
{
  return ms.get_cell()->mk_constant_pool(new ConstantPool(ms));
}


Cell *
Cell::continuation_new(MemorySystem &ms, Cell *procedure, Cell *frame,
		       const char *pc, long depth,
		       Cell *input_port, Cell *output_port)
{
  assert(procedure->is_closure() || procedure == Cell::nil());
  assert(frame->is_scheme_frame() || frame == Cell::nil());

  frame->set_frame_closed();
  StackRoot proc(ms, procedure);
  StackRoot frm(ms, frame);
  StackRoot inport(ms, input_port);
  Cell *env = ms.get_cons(output_port);
  env = ms.get_cons(inport, env);
  env = ms.get_cons(frm, env);
  Cell *continuation = ms.get_cell(env);
  return continuation->mk_continuation(proc, env, new Continuation(pc, depth));
}


Cell *
Cell::memoize_promise_value(MemorySystem &ms, Cell *value)
{
  assert(is_promise());

  if (is_forced_promise())
    {
#if 0
      fprintf(stderr, "A value already memoized for this promise: %p\n", this);
#endif
      return this;
    }

  set_flags(PROMISE_CELL | FORCED_PROMISE_FLAG);
  // ms.set_car(this, Cell::nil());	// <= we can do this
  ms.set_cdr(this, value);
  return this;
}


#ifdef HAVE_TIME_H
Cell *
Cell::mk_date(const struct tm *value, long nanosecond, long zone_offset)
{
  set_flags(DATE_CELL);
  void *val = malloc(sizeof(struct tm));
  if (val != NULL)
    memcpy(val, value, sizeof(struct tm));

  contents.cell.object.date.value = static_cast<struct tm *>(val);
  contents.cell.object.date.nanosecond = nanosecond;
  contents.cell.object.date.zone_offset = zone_offset;
  return this;
}
#endif

}
