/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: cell.h,v 1.54 2009/01/01 07:47:00 fuyu Exp $
 */

#ifndef _LIBTSCORE_CELL_H_
#define _LIBTSCORE_CELL_H_

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef __cplusplus

#if defined(__GNUG__) && !defined(__ICC)
#pragma interface
#endif

#include "libtscore/environment.h"
#define _DEFINE_CELL_	1
#include "libtscore/number.h"
#undef _DEFINE_CELL_
#include "libtscore/port.h"
#include "libtscore/procedure.h"

#include <cassert>
#include <cstdlib>

#ifdef HAVE_TIME_H
#include <time.h>
#endif

#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#ifdef HAVE_STDINT_H
#include <stdint.h>
#else
# ifdef HAVE_INTTYPES_H
# include <inttypes.h>
# else
typedef long intptr_t;
# endif
#endif

#ifndef ONIGURUMA_H
typedef struct re_pattern_buffer regex_t;
typedef struct re_registers OnigRegion;
#endif


using namespace std;
namespace LibTSCore
{
  class Cell;
  class Environment;
  class MemorySystem;
  class TSCore;

  /*
   * ConstantPool represents a table that manages constant values.
   *
   * This actually doesn't have any storage to store constant values.
   * Instead, all accessors take a pointer to Cell that holds actual
   * pointers.  This design allows each instance of ConstantPool
   * to be collected by usual GC activity.
   */
  class ConstantPool
  {
  private:
    MemorySystem &memory_system;
    long number_of_items;

    inline long new_item_slot() { return number_of_items++; }


  public:
    ConstantPool(MemorySystem &ms) : memory_system(ms), number_of_items(0) {}


    /*
     * Makes constant pool empty for recycling.
     */
    void reset(Cell *constant_pool);


    /*
     * Returns the number of items managed by this ConstantPool.
     */
    inline long get_number_of_items() const { return number_of_items; }


    /*
     * Returns the pointer to Cell of constant value.
     */
    inline Cell *load_constant(const Cell *constant_pool, long index) const;


    /*
     * Registers specified pointer to Cell as a constant value.
     *
     * This returns the index to the constant value within the pool.
     * Before registering, this lookups the table and if there already
     * exists `equal' value within the pool, the index returned is the
     * index to pre-existing value and the constant value will be shared.
     */
    long register_constant(Cell *constant_pool, const Cell *value);
  };


#define CAR(p)			((p)->get_car())
#define CDR(p)			((p)->get_cdr())
#define CAAR(p)			CAR(CAR(p))
#define CADR(p)			CAR(CDR(p))
#define CDAR(p)			CDR(CAR(p))
#define CDDR(p)			CDR(CDR(p))
#define CAAAR(p)		CAR(CAR(CAR(p)))
#define CAADR(p)		CAR(CAR(CDR(p)))
#define CADAR(p)		CAR(CDR(CAR(p)))
#define CDADR(p)		CDR(CAR(CDR(p)))
#define CDDAR(p)		CDR(CDR(CAR(p)))
#define CADDR(p)		CAR(CDR(CDR(p)))
#define CDDDR(p)		CDR(CDR(CDR(p)))
#define CADAAR(p)		CAR(CDR(CAR(CAR(p))))
#define CADDDR(p)		CAR(CDR(CDR(CDR(p))))
#define CDDDDR(p)		CDR(CDR(CDR(CDR(p))))


  typedef void (*ForeignObjectDestructFunc)(void *, void *);


  class Syntax;

  // The Cell class INTENTIONALLY uses no virtual functions!
  // We don't want to give extra spaces for C++ vtbl within each Cell.
  // Current 4 pointers/words sized representation might be the maximum
  // luxury for Cell representation.
  class Cell
  {
    friend class ConstantPool;
    friend class Define;
    friend class DefineSyntax;
    friend class Environment;
    friend class HashEnvironment;
    friend class LetSyntax;
    friend class LetrecSyntax;
    friend class ListEnvironment;
    friend class MacroSyntax;
    friend class MemorySystem;
    friend class NativeProcedure;
    friend class Quasiquote;
    friend class R5RSListProcedure;
    friend class R5RSMiscProcedure;
    friend class Register;
    friend class SpecialCellInitializer;
    friend class Syntax;
    friend class TSCore;
    friend class TSCoreProcedure;
    friend class Unquote;
    friend class UnquoteSplicing;
    friend class VirtualMachine;


  private:
    static Cell *nil_cell;
    static Cell *t_cell;
    static Cell *f_cell;
    static Cell *eof_cell;
    static Cell *unspecified_cell;
    static Cell *sink_cell;
    static Cell *trampoline_cell;

    static inline Cell *&nil0() { return nil_cell; }
    static inline Cell *&t0() { return t_cell; }
    static inline Cell *&f0() { return f_cell; }
    static inline Cell *&eof0() { return eof_cell; }
    static inline Cell *&unspecified0() { return unspecified_cell; }
    static inline Cell *&sink0() { return sink_cell; }
    static inline Cell *&trampoline0() { return trampoline_cell; }


  public:
    // These returns (Cell *&) instead of (const Cell *&).  This is
    // intentional to allow simpler implementation of TSCore and
    // MemorySystem.  You should not alter Cells accessed via
    // the pointer given by these.  These are treated as constant
    // value within Scheme world, the base level.  But in the meta-level
    // (implementation level of our Scheme system), the responsibility to
    // treat these as constants lies with implementer.
    static inline Cell *const &nil() { return nil_cell; }
    static inline Cell *const &t() { return t_cell; }
    static inline Cell *const &f() { return f_cell; }
    static inline Cell *const &eof() { return eof_cell; }
    static inline Cell *const &unspecified() { return unspecified_cell; }
    static inline Cell *const &sink() { return sink_cell; }
    static inline Cell *const &trampoline() { return trampoline_cell; }

    static inline Cell *const ts_bool(bool val) { return val ? t() : f(); }


    /*
     * CellType will be rearranged to allow more efficient type checks
     * for some frequently checked type.  At this time, these are
     * not assigned in too clever way.
     *
     * We use the word `atom' to show the Cell doesn't have any references
     * to other Cells.
     *
     * `vector' Cells are special in our system for effieient heap usage.
     * The `vector' Cell is just a beginning mark of a vector and it
     * only contains length of the `vector'.  All vector elements are
     * stored within Cells following just after the `vector' Cell in
     * quite condense manner.  Because of such representation of vectors,
     * we need special treatments of vectors for memory managements, i.e.
     * GC.
     */
    enum CellType
    {
      // atom
      FREE_CELL = 0,				// 0 0000
      NUMBER_CELL = 1,				// 0 0001 same as pseudo type
      CHARACTER_CELL = 2,			// 0 0010 same as pseudo type
      STRING_CELL = 3,				// 0 0011
      FOREIGN_OBJECT_CELL = 4,			// 0 0100 same as pseudo type
      NATIVE_PROCEDURE_CELL = 5,		// 0 0101
      RESERVED_ATOM0_CELL = 6,			// 0 0110 for another pseudo
      PORT_CELL = 7,				// 0 0111
      TIME_CELL = 8,				// 0 1000
      DATE_CELL = 9,				// 0 1001
      RESERVED_ATOM1_CELL = 10,			// 0 1010
      RESERVED_ATOM2_CELL = 11,			// 0 1011
      RESERVED_ATOM3_CELL = 12,			// 0 1100
      RESERVED_ATOM4_CELL = 13,			// 0 1101
      RESERVED_ATOM5_CELL = 14,			// 0 1110
      RESERVED_ATOM6_CELL = 15,			// 0 1111

      // non-atom
      SYMBOL_CELL = 16,				// 1 0000
      PAIR_CELL = 17,				// 1 0001
      VECTOR_CELL = 18,				// 1 0010
      CONSTANT_POOL_CELL = 19,			// 1 0011
      SUBSTRING_CELL = 20,			// 1 0100
      ENVIRONMENT_CELL = 21,			// 1 0101
      BINDING_CELL = 22,			// 1 0110
      REGEXP_CELL = 23,				// 1 0111
      SYNTAX_CELL = 24,				// 1 1000
      MACRO_TRANSFORMER_CELL = 25,		// 1 1001
      CLOSURE_TEMPLATE_CELL = 26,		// 1 1010
      CLOSURE_CELL = 27,			// 1 1011
      CONTINUATION_CELL = 28,			// 1 1100
      PROMISE_CELL = 29,			// 1 1101
      TOPLEVEL_FRAME_CELL = 30,			// 1 1110
      RESERVED_NON_ATOM_CELL = 31		// 1 1111 reserved
    };


#define NON_ATOM_FLAG		0x10L
#define CELL_TYPE_MASK		0x1fL		// enough to store CellType

#define FIXNUM_FLAG			0x1000000L
#define SCHEME_FRAME_FLAG		0x1000000L
#define FORCED_PROMISE_FLAG		0x1000000L
#define STATIC_STRING_FLAG		0x2000000L
#define STATIC_SYNTAX_FLAG		0x2000000L
#define INTERACTIVE_PORT_FLAG		0x2000000L
#define CLOSED_FRAME_FLAG		0x2000000L
#define FULL_CONTINUATION_FLAG		0x2000000L
#define IMMUTABLE_FLAG			0x4000000L


// for GC
#define STICK_YOUNG_FLAG	0x8000000L
#define FORWARD_POINTER_FLAG	0x10000000L
#define SURVIVED_CELL_FLAG	0x20000000L
#define GC_MARK_BIT		0x40000000L
#define GC_FLAG_BIT		0x80000000L


  private:
    union
    { // union of cell and vector storage
      struct
      {
	union
	{ // union of each type of cell
	  struct
	  {
	    void *pad0;
	    void *pad1;
	    void *pad2;
	    long flags;
	  } generic;

	  struct
	  {
	    union
	    {
	      long ivalue;
	      double rvalue;
	    } value;
	    void *pad;
#if SIZEOF_VOID_P == SIZE_OF_DOUBLE
	    void *pad1;
#endif
	    long flags;
	  } number;

	  struct
	  {
	    char *pointer;
	    size_t length;
	    long hash;
	    long flags;
	  } string;

	  struct
	  {
	    Cell *car;
	    Cell *cdr;
	    union
	    {
	      long length;
	      regex_t *rx;
	      const char *substring;
	      long slot;
	      Environment *environment;
	      ConstantPool *constant_pool;
	      Syntax *syntax;
	      BytecodeProcedure *bytecode;
	      const char *pc;
	      Continuation *continuation;
	    } option;
	    long flags;
	  } cons;

	  struct
	  {
	    Port *port;
	    void *pad0;
	    void *pad1;
	    long flags;
	  } port;

	  struct
	  {
	    long type;
	    long nanosecond;
	    long second;
	    long flags;
	  } time;

#ifdef HAVE_TIME_H
	  struct
	  {
	    struct tm *value;
	    long nanosecond;
	    long zone_offset;
	    long flags;
	  } date;
#endif

	  struct
	  {
	    void *data;
	    const NativeProcedure::Info *info;
	    NativeProcedure *proc;
	    long flags;
	  } procedure;

	  struct
	  {
	    void *pointer;
	    ForeignObjectDestructFunc destructor;
	    void *option_data;
	    long flags;
	  } fo;
	} object;
      } cell;

      // The current representation of vector is like
      // |Vector Cell|e0, e1, e2, e3|e4, e5, e6, e7|...|
      struct
      {
	Cell *elements[4];
      } vector;
    } contents;



    /*
     * PSEUDO objects
     *   We use special object representation that encodes its contents
     *   within each pointer's self for integers, characters and foreign
     *   objects.  Since all allocations of Cells we do ensure
     *   that Cells to be aligned at sizeof(Cell) boundary and sizeof(Cell)
     *   in our current representation is 16 bytes on ILP32 and 32 bytes
     *   on (I)LP64 architecture, LSB parts of valid pointer to Cell
     *   always have bit pattern 0000(ILP32) and 00000(LP64) respectively.
     *
     *   We treat (Cell *)s whose least significant 3 bits are non-zero
     *   as pseudo object and the 3 bits shows the type of object as
     *   follows:
     */
  private:
#define PSEUDO_CELL_INTEGER_TAG		1	// 001
#define PSEUDO_CELL_CHARACTER_TAG	2	// 010
#define PSEUDO_CELL_POINTER_TAG		4	// 100
#define PSEUDO_CELL_RESERVED_TAG	6	// 110

#define PSEUDO_CELL_TAG_MASK		7	// 111
#define PSEUDO_CELL_TAG_BITS		3

    inline bool is_pseudo_cell() const
    { return (reinterpret_cast<intptr_t>(this) & PSEUDO_CELL_TAG_MASK) != 0; }

    // This returns 0 for non-pseudo (Cell *).
    inline long pseudo_cell_type() const
    {
      long flags = reinterpret_cast<intptr_t>(this);
      return (flags & PSEUDO_CELL_INTEGER_TAG) != 0
	? NUMBER_CELL : (flags & PSEUDO_CELL_TAG_MASK);
    }


    /*
     * PSEUDO integer:  If the long `value' can be represented in
     * { BIT_SIZE_OF(Cell *) - 1} bits, we represent the value as
     * a pseudo (Cell *).  Since `integer' is one of the most frequently
     * used type of object, we want to make the range of integers
     * represented as pseudo Cell as larger range as possible.
     */
  public:
    static inline Cell *mk_small_integer(long value)
    {
      // Here we assume sizeof(Cell *) == sizeof(long)
      long tmp_val = value << 1;
      if ((tmp_val >> 1) == value)
	return reinterpret_cast<Cell *>(tmp_val | PSEUDO_CELL_INTEGER_TAG);

      return NULL;
    }


  private:
    inline bool is_pseudo_integer_cell() const
    {
      return (reinterpret_cast<intptr_t>(this) & PSEUDO_CELL_INTEGER_TAG) != 0;
    }

    inline long get_pseudo_integer() const
    {
      assert(is_pseudo_integer_cell());

      return reinterpret_cast<intptr_t>(this) >> 1;
    }



    /*
     * Character, we always represent each Scheme character object
     * as pseudo Cell, because we use UTF-8 encoding for our string
     * representation and treat them as is, 8 bits octet streams.
     * Therefore, each characters have only 8 bits information.
     *
     * There's another design choice such as treat Unicode characters
     * like Java instead of raw stream of UTF-8.  But we know using
     * Unicode isn't very useful for real world application.  There's
     * very few applications that requires treatment of indivisual
     * characters instead of strings.  If you don't need indivisual
     * character treatment, using raw UTF-8 streams are much easier
     * and more efficient than wide char Unicode.
     *
     * NOTE: Since R6RS defines that `character' is Unicode character,
     *       we may change our definition of `character' in the future.
     */
  public:
    static inline Cell *mk_character(long c)
    {
      return reinterpret_cast<Cell *>((c << PSEUDO_CELL_TAG_BITS)
				      | PSEUDO_CELL_CHARACTER_TAG);
    }

    inline long get_character() const
    {
      return reinterpret_cast<intptr_t>(this) >> PSEUDO_CELL_TAG_BITS;
    }

    inline bool is_character() const { return cell_type() == CHARACTER_CELL; }


    /*
     * PSEUDO foreign object:  Generic native pointers are frequently
     * aligned at some good boundary.  Currently, we only use the
     * least significant 3 bits to represent type of pseudo object.
     * Therefore, there's chance to make foreign pointers as pseudo
     * objects if condition met.  If given foreign pointer is aligned
     * at 3 bits boundary, i.e. 8 bytes boundary, we represent the foreign
     * object as pseudo foreign object Cell by simply `or'ing with
     * PSEUDO_CELL_POINTER_TAG to show it is pseudo foreign object.
     *
     * This returns NULL if given pointer doesn't met the condition
     * to be represented as pseudo object.
     */
  public:
    static inline Cell *mk_foreign_object(void *data)
    {
      if (reinterpret_cast<intptr_t>(data) & PSEUDO_CELL_TAG_MASK)
	return NULL;

      return reinterpret_cast<Cell *>(reinterpret_cast<intptr_t>(data)
				      | PSEUDO_CELL_POINTER_TAG);
    }


  private:
    inline void *get_pseudo_foreign_pointer() const
    {
      return reinterpret_cast<void *>(reinterpret_cast<intptr_t>(this)
				      & ~PSEUDO_CELL_TAG_MASK);
    }



    /*
     * INTERNAL APIs for managements of a Cell.
     */
    inline void set_flags(long flags)
    {
      assert(!is_pseudo_cell());

      contents.cell.object.generic.flags = flags;
    }

    inline long get_flags() const
    {
      assert(!is_pseudo_cell());

      return contents.cell.object.generic.flags;
    }


  public:
    inline bool is_immutable() const
    {
      return (get_flags() & IMMUTABLE_FLAG) != 0;
    }

    inline Cell *set_immutable()
    {
      set_flags(get_flags() | IMMUTABLE_FLAG);
      return this;
    }


    // for inspection
    inline long cell_type() const
    {
      long pseudo_type = pseudo_cell_type();
      if (pseudo_type != 0)
	return pseudo_type;

      return get_flags() & CELL_TYPE_MASK;
    }

    const char *get_cell_type_string() const;


    /*
     * Returns C string representation of this Cell.
     * The respoinsibility to free() returned pointer lies with caller.
     */
    char *to_string() const;


    /*
     * Appends string representation of this Cell into given Port.
     * This will be used by Port representation to eat (Cell *).
     */
    inline Port *to_string(Port *port) const { return to_string(port, true); }


  private:
    Port *to_string(Port *port, bool toplevel) const;


  private:
    // for GC
    inline bool is_atom() const
    {
      return is_pseudo_cell() || (get_flags() & NON_ATOM_FLAG) == 0;
    }

    inline bool is_forward_pointer() const
    {
      return !is_pseudo_cell() && (get_flags() & FORWARD_POINTER_FLAG) != 0;
    }

    inline Cell *mk_forward_pointer(Cell *p)
    {
      set_flags(FORWARD_POINTER_FLAG);
      contents.cell.object.cons.car = p;
      p->set_flags(p->get_flags() | SURVIVED_CELL_FLAG);
      return this;
    }

    inline Cell *get_forward_pointer() const
    {
      assert(is_forward_pointer());

      return contents.cell.object.cons.car;
    }

    inline bool is_survived_cell() const
    {
      return (get_flags() & (SURVIVED_CELL_FLAG | STICK_YOUNG_FLAG))
	== SURVIVED_CELL_FLAG;
    }

    inline bool is_gc_marked() const
    {
      return is_pseudo_cell() || (get_flags() & GC_MARK_BIT) != 0;
    }

    inline Cell *set_gc_mark()
    {
      set_flags(get_flags() | GC_MARK_BIT);
      return this;
    }

    inline Cell *unset_gc_mark()
    {
      set_flags(get_flags() & ~GC_MARK_BIT);
      return this;
    }

    inline bool is_gc_flag_set() const
    {
      return (get_flags() & GC_FLAG_BIT) != 0;
    }

    inline Cell *set_gc_flag()
    {
      set_flags(get_flags() | GC_FLAG_BIT);
      return this;
    }

    inline Cell *unset_gc_flag()
    {
      set_flags(get_flags() & ~GC_FLAG_BIT);
      return this;
    }



    Cell *finalize();

    inline Cell *mk_free_cell(Cell *next = Cell::nil())
    {
      set_flags(FREE_CELL | GC_MARK_BIT);
      contents.cell.object.cons.car = Cell::nil();
      contents.cell.object.cons.cdr = next;
      contents.cell.object.cons.option.length = 0;
      return this;
    }


    Cell();



  public:
    // Only for initialization of global constant atoms, e.g. nil, #t and #f.
    inline Cell *mk_constant()
    {
      set_flags(GC_MARK_BIT);
      contents.cell.object.cons.car = this;
      contents.cell.object.cons.cdr = this;
      return this;
    }


    /*
     * Pair, the most essential type of Scheme object
     */
    inline Cell *mk_pair(Cell *car, Cell *cdr, bool immutable = false)
    {
      if (immutable)
	set_flags(PAIR_CELL | IMMUTABLE_FLAG);
      else
	set_flags(PAIR_CELL);

      contents.cell.object.cons.car = car;
      contents.cell.object.cons.cdr = cdr;
      return this;
    }

    inline bool is_pair() const { return cell_type() == PAIR_CELL; }

    inline Cell *get_car() const
    {
      return contents.cell.object.cons.car;
    }

    inline Cell *get_cdr() const
    {
      return contents.cell.object.cons.cdr;
    }


  private:
    // These are private here since our memory system requires
    // `write barrier' and there's no space to do it by Cell's self.
    // Use methods provided by MemorySystem to alter car and cdr!
    inline Cell *set_car(Cell *car)
    {
      contents.cell.object.cons.car = car;
      return this;
    }

    inline Cell *set_cdr(Cell *cdr)
    {
      contents.cell.object.cons.cdr = cdr;
      return this;
    }

    inline Cell *&car() { return contents.cell.object.cons.car; }
    inline Cell *&cdr() { return contents.cell.object.cons.cdr; }


  public:
    /*
     * String, we have three kinds of string representation.
     * - ordinal strings have own storage represents the text
     *   as C string, i.e. array of char terminated by '\0'.  In this
     *   case, strlen(cell->get_string()) == cell->get_strlen().
     * - static strings don't have own storage.  It must be
     *   IMMUTABLE and there's no action performed during finalization
     *   such as free() of native storage.  This is useful to make
     *   Scheme string from native C++ code.
     * - substrings don't have own storage.  It refers other
     *   IMMUTABLE string's storage for the native representation.
     */
    inline bool is_string() const
    {
      long type = cell_type();
      return type == STRING_CELL || type == SUBSTRING_CELL;
    }

    inline Cell *mk_string(char *str, size_t len)
    {
      set_flags(STRING_CELL);
      contents.cell.object.string.pointer = str;
      contents.cell.object.string.length = len;
      contents.cell.object.string.hash = 0;
      return this;
    }

    inline Cell *mk_static_string(const char *str, size_t len)
    {
      set_flags(STRING_CELL | STATIC_STRING_FLAG | IMMUTABLE_FLAG);
      contents.cell.object.string.pointer = const_cast<char *>(str);
      contents.cell.object.string.length = len;
      contents.cell.object.string.hash = 0;
      return this;
    }

    // This makes ordinal string by copying native representation.
    Cell *mk_substring_copy(Cell *str, Cell *offset, Cell *len);

    // This makes shallow copy of substring if given source string is
    // immutable.
    inline Cell *mk_substring_reference(Cell *str, Cell *offset, Cell *len)
    {
      assert(str->is_string());

      if (str->is_immutable())
	{
	  set_flags(SUBSTRING_CELL | IMMUTABLE_FLAG);
	  contents.cell.object.cons.car = str;
	  contents.cell.object.cons.cdr = len;
	  contents.cell.object.cons.option.substring
	    = str->get_string() + offset->get_integer();
	}

      return mk_substring_copy(str, offset, len);
    }

    // This may returns the C string longer than actual Scheme string
    // if this is actually a substring of other.
    // You must take care of the length of Scheme string.
    inline const char *get_string() const
    {
      if (cell_type() == STRING_CELL)
	return contents.cell.object.string.pointer;

      assert(cell_type() == SUBSTRING_CELL);

      return contents.cell.object.cons.option.substring;
    }

    inline size_t get_strlen() const
    {
      if (cell_type() == STRING_CELL)
	return contents.cell.object.string.length;

      assert(cell_type() == SUBSTRING_CELL);

      return contents.cell.object.cons.cdr->get_integer0();
    }

    inline char *get_string0() const
    {
      assert(cell_type() == STRING_CELL);

      return contents.cell.object.string.pointer;
    }

    long string_hash();
    bool string_equals(Cell *str);
    bool string_equals(const char *str);
    static long string_hash(const char *str, size_t len);
    static inline long string_hash(const char *str)
    {
      return string_hash(str, strlen(str));
    }



    /*
     * Regular Expression, we use native regular expression
     * library oniguruma for this.
     *
     * This Scheme object is not an useful C++ object but just a
     * placeholder of oniguruma's regular expression object
     *
     * We decided not to encapsulate the fact that we use oniguruma.
     * Especially, we don't supply any useful methods of Cell to perform
     * regular expression matching because our dedicated memory system
     * don't allow Cell to allocate another Cell by itself, and thus,
     * such utility methods can produce only in native representation.
     * It is usually not very useful for user Scheme program, and the
     * core parts of our Scheme system should provide Scheme level
     * utility procedures for users.  Implementations of such procedures
     * often require native access to oniguruma's object for efficiency.
     * Therefore, odds encapsulation helps none of us and users!
     */

    // This may return string cell that shows the error during regexp
    // compilation, otherwise, returns a constructed regexp cell.
    Cell *mk_regexp(Cell *rx_string, bool ignore_case = false);

    inline bool is_regexp() const { return cell_type() == REGEXP_CELL; }

    // This returns the pointer to internal(compiled) representation of the
    // regular expression, it actually is oniguruma's regex_t.
    inline regex_t *get_regexp() const
    {
      assert(is_regexp());

      return contents.cell.object.cons.option.rx;
    }

    // Returns the string Cell from which this regular expression object
    // is compiled.
    inline Cell *get_regexp_string() const
    {
      assert(is_regexp());

      return contents.cell.object.cons.car;
    }


    /*
     * Number, we only support fixed integer as exact number and double
     * precision floating point value as inexact number at this time.
     * Integers may or may not have Cell allocated for their representation.
     */
    inline Cell *mk_integer(long value)
    {
      set_flags(NUMBER_CELL | FIXNUM_FLAG);
      contents.cell.object.number.value.ivalue = value;
      return this;
    }

    inline Cell *mk_real(double value)
    {
      set_flags(NUMBER_CELL);
      contents.cell.object.number.value.rvalue = value;
      return this;
    }

    inline Cell *mk_number(Number &number)
    {
      if (number.is_fixnum())
	{
	  set_flags(NUMBER_CELL | FIXNUM_FLAG);
	  contents.cell.object.number.value.ivalue = number.get_ivalue0();
	}
      else
	{
	  set_flags(NUMBER_CELL);
	  contents.cell.object.number.value.rvalue = number.get_rvalue0();
	}

      return this;
    }


    inline bool is_number() const
    {
      return is_pseudo_integer_cell() || cell_type() == NUMBER_CELL;
    }

    inline bool is_complex() const { return false; }
    inline bool is_real() const { return is_number() && !is_complex(); }

    inline bool is_integer() const
    {
      return is_pseudo_integer_cell()
	|| (cell_type() == NUMBER_CELL && (get_flags() & FIXNUM_FLAG));
    }

    inline Cell *set_integer(long value)
    {
      assert(!is_pseudo_cell());

      contents.cell.object.number.value.ivalue = value;
      return this;
    }

    inline long get_integer() const
    {
      if (is_pseudo_integer_cell())
	return get_pseudo_integer();

      assert(cell_type() == NUMBER_CELL);

      return (get_flags() & FIXNUM_FLAG)
	? contents.cell.object.number.value.ivalue
	: static_cast<long>(contents.cell.object.number.value.rvalue);
    }

    // get `long' value with *NO* type-conversion even for FLONUM.
    inline long get_integer0() const
    {
      if (is_pseudo_integer_cell())
	return get_pseudo_integer();

      assert(cell_type() == NUMBER_CELL);

      return contents.cell.object.number.value.ivalue;
    }

    inline Cell *set_real(double value)
    {
      contents.cell.object.number.value.rvalue = value;
      set_flags(get_flags() & ~FIXNUM_FLAG);
      return this;
    }

    inline double get_real() const
    {
      if (is_pseudo_integer_cell())
	return static_cast<double>(get_pseudo_integer());

      assert(cell_type() == NUMBER_CELL);

      return (get_flags() & FIXNUM_FLAG)
	? static_cast<double>(contents.cell.object.number.value.ivalue)
	: contents.cell.object.number.value.rvalue;
    }

    // get `double' value with *NO* type-conversion even for FIXNUM.
    inline double get_real0() const
    {
      assert(!is_pseudo_cell());

      return contents.cell.object.number.value.rvalue;
    }



    /*
     * Vector
     */
    Cell *mk_vector(long len, Cell *fill);

    inline bool is_vector() const { return cell_type() == VECTOR_CELL; }

    inline long get_vector_length() const
    {
      assert(is_vector());

      return contents.cell.object.cons.option.length;
    }

    inline Cell *get_vector_element(long index) const
    {
      assert(is_vector());

      long subindex = index % 4;
      index /= 4;
      return (this + 1 + index)->contents.vector.elements[subindex];
    }


  private:
    // This is private because of the same reason as Pairs.
    inline Cell *set_vector_element(long index, Cell *element)
    {
      assert(is_vector());

      long subindex = index % 4;
      index /= 4;
      (this + 1 + index)->contents.vector.elements[subindex] = element;
      return this;
    }

    // Only useful for our memory system
    static inline long get_vector_size(size_t len)
    {
      return (len / 4) + ((len % 4) > 0 ? 1 : 0);
    }

    static inline long round_vector_size(size_t len)
    {
      return get_vector_size(len) * 4;
    }

    inline long get_vector_size() const
    {
      assert(is_vector());

      return get_vector_size(contents.cell.object.cons.option.length);
    }



  public:
    /*
     * Foreign Object, useful especially to exchange something between
     * foreign functions and scheme programs.
     *
     * You can specify the ForeignObjectDestructFunc that will be invoked
     * when the Cell for foreign object is collected as garbage.
     * The non-NULL ForeignObjectDestructFunc will be invoked like:
     *   (*func)(data, option_data);
     *
     * You can store optional data that may corresponds to `data', e.g.
     * type-information.
     */
    inline Cell *mk_foreign_object(void *data, ForeignObjectDestructFunc func,
				   void *option_data)
    {
      set_flags(FOREIGN_OBJECT_CELL);
      contents.cell.object.fo.pointer = data;
      contents.cell.object.fo.destructor = func;
      contents.cell.object.fo.option_data = option_data;
      return this;
    }

    inline bool is_foreign_object() const
    {
      return cell_type() == FOREIGN_OBJECT_CELL;
    }

    inline void *get_foreign_pointer() const
    {
      if (is_pseudo_cell())
	return get_pseudo_foreign_pointer();

      assert(is_foreign_object());

      return contents.cell.object.fo.pointer;
    }

    inline ForeignObjectDestructFunc get_foreign_destructor() const
    {
      if (is_pseudo_cell())
	return NULL;

      assert(is_foreign_object());

      return contents.cell.object.fo.destructor;
    }

    inline void *get_foreign_option_data() const
    {
      if (is_pseudo_cell())
	return NULL;

      assert(is_foreign_object());

      return contents.cell.object.fo.option_data;
    }



    /*
     * Ad-hoc types used by our Scheme system's self.
     */
    inline Cell *mk_symbol(Cell *name, Cell *next)
    {
      assert(name->is_string());
      assert(name->is_immutable());

      set_flags(SYMBOL_CELL | IMMUTABLE_FLAG);
      contents.cell.object.cons.car = name;
      contents.cell.object.cons.cdr = next;
      return this;
    }

    inline bool is_symbol() const { return cell_type() == SYMBOL_CELL; }

    inline Cell *get_symbol_string() const
    {
      assert(is_symbol());

      return get_car();
    }

    inline const char *get_symname() const
    {
      assert(is_symbol());

      return contents.cell.object.cons.car->get_string();
    }

    inline long symbol_hash() const
    {
      assert(is_symbol());

      return contents.cell.object.cons.car->string_hash();
    }
    static long symbol_hash(const char *str, size_t len);
    static inline long symbol_hash(const char *str)
    {
      return symbol_hash(str, strlen(str));
    }



    /*
     * The `environment' manages only bindings and has no storage
     * for each `variable's.
     * The `scheme frame' manages only storage for each `variable
     * binding's and delegates management of `binding's to its corresponding
     * `environment'.
     *
     * We have two concrete implementations of `environment' to perform
     * efficient management of bindings.  One uses hashtable and
     * the other uses singly linked list.
     *
     * hashtable representation is only used for `toplevel' environments.
     * list representation is used for non-toplevel environments.
     */
    static Cell *toplevel_environment_new(MemorySystem &ms);
    static Cell *closed_environment_new(MemorySystem &ms, const Cell *parent);

    inline bool is_environment() const
    {
      return cell_type() == ENVIRONMENT_CELL;
    }

    inline bool is_toplevel_environment() const
    {
      return is_environment() && get_parent_environment() == Cell::nil();
    }


    inline Environment *get_environment() const
    {
      assert(is_environment());

      return contents.cell.object.cons.option.environment;
    }

    inline Cell *get_parent_environment() const
    {
      assert(is_environment());

      return contents.cell.object.cons.car;
    }

    inline Cell *get_bindings() const
    {
      assert(is_environment());

      return contents.cell.object.cons.cdr;
    }

    inline const Cell *lookup(const Cell *symbol, bool all) const
    {
      assert(is_environment());

      return get_environment()->lookup(this, symbol, all);
    }

    inline bool is_bound(const Cell *symbol, bool all) const
    {
      assert(is_environment());

      return get_environment()->is_bound(this, symbol, all);
    }


  private:
    inline Cell *mk_environment(Cell *parent, Environment *env,
				Cell *bindings)
    {
      assert(parent == Cell::nil() || parent->is_environment());

      set_flags(ENVIRONMENT_CELL);
      contents.cell.object.cons.car = parent;
      contents.cell.object.cons.cdr = bindings;
      contents.cell.object.cons.option.environment = env;
      return this;
    }



    /*
     * Syntax related Cell objects.
     *
     * Almost only used by the compiler and virtual machine.
     */
  private:
    inline const Cell *bind_syntax(const Cell *syntax)
    {
      assert(is_environment());

      return get_environment()->bind_syntax(this, syntax);
    }

    inline void unbind_syntax(const Cell *syntax)
    {
      assert(is_environment());

      get_environment()->unbind_syntax(this, syntax);
    }

    inline const Cell *bind_variable(const Cell *symbol)
    {
      assert(is_environment());

      return get_environment()->bind_variable(this, symbol);
    }

    inline const Cell *make_unbound_variable(const Cell *symbol)
    {
      assert(is_environment());

      return get_environment()->make_unbound_variable(this, symbol);
    }

    inline void unbind_variable(const Cell *symbol)
    {
      assert(is_environment());

      get_environment()->unbind_variable(this, symbol);
    }

    inline void rebind_binding(const Cell *binding)
    {
      assert(is_environment());

      get_environment()->rebind_binding(this, binding);
    }


  public:
    inline long allocate_anonymous_slot()
    {
      assert(is_environment());

      return get_environment()->allocate_anonymous_slot(this);
    }

    inline long allocate_stack_slot()
    {
      assert(is_environment());

      return get_environment()->allocate_stack_slot();
    }

    inline long get_current_stack_pointer() const
    {
      assert(is_environment());

      return get_environment()->get_current_stack_pointer();
    }

    inline void dispose_stack_slots(long depth)
    {
      assert(is_environment());

      return get_environment()->dispose_stack_slots(depth);
    }

    long calculate_environment_distance(const Cell *environment) const;

    inline long get_required_frame_size() const
    {
      assert(is_environment());

      return get_environment()->estimate_storage_size();
    }


    // For variable bindings
    inline Cell *mk_binding(const Cell *symbol, const Cell *env, long slot)
    {
      set_flags(BINDING_CELL | IMMUTABLE_FLAG);
      contents.cell.object.cons.car = const_cast<Cell *>(symbol);
      contents.cell.object.cons.cdr = const_cast<Cell *>(env);
      contents.cell.object.cons.option.slot = slot;
      return this;
    }

    inline bool is_binding() const { return cell_type() == BINDING_CELL; }

    inline const Cell *get_bound_symbol() const
    {
      assert(is_binding());

      return contents.cell.object.cons.car;
    }

    inline const Cell *get_bound_environment() const
    {
      assert(is_binding());

      return contents.cell.object.cons.cdr;
    }

    inline long get_slot_location() const
    {
      assert(is_binding());

      return contents.cell.object.cons.option.slot;
    }

    inline Cell *set_slot_location(long slot)
    {
      assert(is_binding());

      contents.cell.object.cons.option.slot = slot;
      return this;
    }



    /*
     * Frames, almost only used by the virtual machine.
     */
  private:
    inline Cell *mk_toplevel_frame(Cell *slots)
    {
      set_flags(TOPLEVEL_FRAME_CELL | CLOSED_FRAME_FLAG);
      contents.cell.object.cons.car = Cell::nil();
      contents.cell.object.cons.cdr = slots;	// must be a vector
      return this;
    }

    inline Cell *mk_scheme_frame(Cell *parent)
    {
      set_flags(VECTOR_CELL | SCHEME_FRAME_FLAG);
      contents.cell.object.cons.car = parent;
      contents.cell.object.cons.cdr = Cell::nil();
      return this;
    }


  public:
    static Cell *toplevel_frame_new(MemorySystem &ms, long initial_frame_size);
    static Cell *scheme_frame_new(MemorySystem &ms, Cell *parent,
				  long frame_size);


    inline bool is_toplevel_frame() const
    {
      return cell_type() == TOPLEVEL_FRAME_CELL;
    }

    inline Cell *get_variable_slots() const
    {
      assert(is_toplevel_frame());

      return contents.cell.object.cons.cdr;
    }

    inline bool is_scheme_frame() const
    {
      return is_vector() && (get_flags() & SCHEME_FRAME_FLAG) != 0;
    }

    inline Cell *set_frame_closed()
    {
      set_flags(get_flags() | CLOSED_FRAME_FLAG);
      return this;
    }

    inline bool is_frame_closed() const
    {
      assert(is_scheme_frame());

      return (get_flags() & CLOSED_FRAME_FLAG) != 0;
    }

    inline Cell *get_parent_frame() const
    {
      assert(is_scheme_frame() || is_toplevel_frame());

      return contents.cell.object.cons.car;
    }

#if 0
    inline Cell *set_parent_frame(Cell *parent)
    {
      assert(is_scheme_frame());
      assert(parent->is_scheme_frame() || parent->is_toplevel_frame());

      contents.cell.object.cons.car = parent;
      return this;
    }
#endif

    inline long get_frame_size() const
    {
      assert(is_scheme_frame());

      return get_vector_length();
    }

    /*
     * Loads a value within scheme frame from specified slot.
     */
    inline Cell *load_variable(long slot) const
    {
      assert(is_scheme_frame());

      if (slot < get_vector_length())
	return get_vector_element(slot);

      return Cell::unspecified();	// unbound variable
    }

    /*
     * Stores a value into scheme frame at specified slot.
     * Because of memory barrier, we need a reference to the MemorySystem.
     */
    inline Cell *store_variable(MemorySystem &ms, long slot, Cell *value);

    /*
     * Loads a value at specified slot of toplevel scheme frame.
     */
    inline Cell *load_toplevel_variable(long slot) const
    {
      assert(is_toplevel_frame());

      Cell *storage = get_variable_slots();
      if (slot < storage->get_vector_length())
	return storage->get_vector_element(slot);

      return Cell::unspecified();	// unbound variable
    }

    /*
     * Stores a value into toplevel scheme frame at specified slot.
     * Because the toplevel environment/frame can grow incrementally at
     * runtime, we need special treatment for this.
     */
    inline Cell *store_toplevel_variable(Cell *environment,
					 long slot, Cell *value)
    {
      return environment->get_environment()->store_variable(this, slot, value);
    }


    /*
     * Syntaxes, almost only used by the bytecompiler during compilation.
     */
    inline Cell *mk_syntax(const Cell *symbol, Syntax *syntax,
			   const Cell *transformer = Cell::nil())
    {
      assert(transformer == Cell::nil()
	     || transformer->is_macro_transformer());

      set_flags(SYNTAX_CELL | IMMUTABLE_FLAG);
      contents.cell.object.cons.car = const_cast<Cell *>(symbol);
      contents.cell.object.cons.cdr = const_cast<Cell *>(transformer);
      contents.cell.object.cons.option.syntax = syntax;
      return this;
    }

    inline Cell *mk_static_syntax(const Cell *symbol, Syntax *syntax,
				  const Cell *transformer = Cell::nil())
    {
      assert(transformer == Cell::nil()
	     || transformer->is_macro_transformer());

      set_flags(SYNTAX_CELL | STATIC_SYNTAX_FLAG | IMMUTABLE_FLAG);
      contents.cell.object.cons.car = const_cast<Cell *>(symbol);
      contents.cell.object.cons.cdr = const_cast<Cell *>(transformer);
      contents.cell.object.cons.option.syntax = syntax;
      return this;
    }

    inline bool is_syntax() const { return cell_type() == SYNTAX_CELL; }

    inline bool is_macro_syntax() const
    {
      return is_syntax() && get_macro_transformer() != Cell::nil();
    }

    inline Cell *get_syntax_symbol() const
    {
      assert(is_syntax());

      return contents.cell.object.cons.car;
    }

    inline const char *get_syntax_name() const
    {
      return get_syntax_symbol()->get_symname();
    }

    inline Syntax *get_syntax() const
    {
      assert(is_syntax());

      return contents.cell.object.cons.option.syntax;
    }

    inline const Cell *get_macro_transformer() const
    {
      assert(is_syntax());

      return contents.cell.object.cons.cdr;
    }



    inline Cell *mk_macro_transformer(const Cell *environment_literals_pair,
				      const Cell *rules)
    {
      assert(environment_literals_pair->is_pair());
      assert(CAR(environment_literals_pair)->is_environment());

      set_flags(MACRO_TRANSFORMER_CELL | IMMUTABLE_FLAG);
      contents.cell.object.cons.car
	= const_cast<Cell *>(environment_literals_pair);
      contents.cell.object.cons.cdr = const_cast<Cell *>(rules);
      return this;
    }

    inline bool is_macro_transformer() const
    { return cell_type() == MACRO_TRANSFORMER_CELL; }

    inline Cell *get_macro_transformer_environment() const
    {
      assert(is_macro_transformer());

      return CAR(contents.cell.object.cons.car);
    }

    inline const Cell *get_macro_transformer_literals() const
    {
      assert(is_macro_transformer());

      return CDR(contents.cell.object.cons.car);
    }

    inline const Cell *get_macro_transformer_syntax_rules() const
    {
      assert(is_macro_transformer());

      return contents.cell.object.cons.cdr;
    }


    static Cell *constant_pool_new(MemorySystem &ms);

    inline Cell *mk_constant_pool(ConstantPool *constant_pool)
    {
      set_flags(CONSTANT_POOL_CELL);
      contents.cell.object.cons.car = Cell::nil();
      contents.cell.object.cons.cdr = Cell::nil();
      contents.cell.object.cons.option.constant_pool = constant_pool;
      return this;
    }

    inline bool is_constant_pool() const
    {
      return cell_type() == CONSTANT_POOL_CELL;
    }

    inline ConstantPool *get_constant_pool() const
    {
      assert(is_constant_pool());

      return contents.cell.object.cons.option.constant_pool;
    }

    inline Cell *get_constant_table() const
    {
      assert(is_constant_pool());

      return contents.cell.object.cons.car;
    }

    inline Cell *load_constant(long index) const
    {
      assert(is_constant_pool());

      return get_constant_pool()->load_constant(this, index);
    }

    inline long register_constant(const Cell *value)
    {
      assert(is_constant_pool());

      return get_constant_pool()->register_constant(this, value);
    }

    inline void reset_constant_pool()
    {
      assert(is_constant_pool());

      get_constant_pool()->reset(this);
    }



    inline Cell *mk_native_procedure(NativeProcedure *proc,
				     const NativeProcedure::Info *info,
				     void *data)
    {
      set_flags(NATIVE_PROCEDURE_CELL);
      contents.cell.object.procedure.proc = proc;
      contents.cell.object.procedure.info = info;
      contents.cell.object.procedure.data = data;
      return this;
    }

    inline bool is_native_procedure() const
    {
      return cell_type() == NATIVE_PROCEDURE_CELL;
    }

    inline NativeProcedure *get_native_procedure() const
    {
      assert(is_native_procedure());

      return contents.cell.object.procedure.proc;
    }

    inline const NativeProcedure::Info *get_native_procedure_info() const
    {
      assert(is_native_procedure());

      return contents.cell.object.procedure.info;
    }

    inline void *get_native_procedure_data() const
    {
      assert(is_native_procedure());

      return contents.cell.object.procedure.data;
    }



    inline Cell *mk_closure_template(Cell *constant_pool, Cell *code,
				     BytecodeProcedure *bytecode)
    {
      set_flags(CLOSURE_TEMPLATE_CELL);
      contents.cell.object.cons.car = constant_pool;
      contents.cell.object.cons.cdr = code;
      contents.cell.object.cons.option.bytecode = bytecode;
      return this;
    }

    inline bool is_closure_template() const
    {
      return cell_type() == CLOSURE_TEMPLATE_CELL;
    }

    inline Cell *get_closure_constant_pool0() const
    {
      assert(is_closure_template());

      return contents.cell.object.cons.car;
    }

    inline Cell *get_closure_code0() const
    {
      assert(is_closure_template());

      return contents.cell.object.cons.cdr;
    }

    inline BytecodeProcedure *get_closure_bytecode0() const
    {
      assert(is_closure_template());

      return contents.cell.object.cons.option.bytecode;
    }

    inline Cell *mk_closure(Cell *closure_template, Cell *frame)
    {
      assert(closure_template->is_closure_template());
      assert(frame->is_scheme_frame() || frame == Cell::nil());

      if (frame != Cell::nil())
	frame->set_frame_closed();

      set_flags(CLOSURE_CELL);
      contents.cell.object.cons.car = closure_template;
      contents.cell.object.cons.cdr = frame;
      return this;
    }

    inline bool is_closure() const { return cell_type() == CLOSURE_CELL; }

    inline Cell *get_closure_template() const
    {
      assert(is_closure());

      return contents.cell.object.cons.car;
    }

    inline Cell *get_closure_frame() const
    {
      assert(is_closure());

      return contents.cell.object.cons.cdr;
    }

    inline long get_closure_frame_size() const
    {
      assert(is_closure());

      return get_closure_bytecode()->get_frame_size();
    }

    inline Cell *get_closure_constant_pool() const
    {
      if (is_closure())
	return get_closure_template()->get_closure_constant_pool0();

      assert(is_closure_template());

      return contents.cell.object.cons.car;
    }

    inline Cell *get_closure_code() const
    {
      if (is_closure())
	return get_closure_template()->get_closure_code0();

      assert(is_closure_template());

      return contents.cell.object.cons.cdr;
    }

    inline BytecodeProcedure *get_closure_bytecode() const
    {
      if (is_closure())
	return get_closure_template()->get_closure_bytecode0();

      assert(is_closure_template());

      return contents.cell.object.cons.option.bytecode;
    }



    // This is only used for creation of meta_return special continuation.
    inline Cell *mk_constant_continuation(const char *pc)
    {
      // Makes cheaper continuation only used VM's internal.
      set_flags(CONTINUATION_CELL);
      contents.cell.object.cons.car = Cell::nil();
      contents.cell.object.cons.cdr = Cell::nil();
      contents.cell.object.cons.option.pc = pc;
      return this;
    }

    inline Cell *mk_continuation(Cell *procedure, Cell *frame, const char *pc)
    {
      assert(procedure->is_closure());
      assert(frame->is_scheme_frame());

      // Makes cheaper continuation only used VM's internal.
      set_flags(CONTINUATION_CELL);
      contents.cell.object.cons.car = procedure;
      contents.cell.object.cons.cdr = frame;
      contents.cell.object.cons.option.pc = pc;
      return this;
    }

    static Cell *continuation_new(MemorySystem &ms,
				  Cell *procedure, Cell *frame,
				  const char *pc, long depth,
				  Cell *input_port, Cell *output_port);

    inline Cell *mk_continuation(Cell *procedure,
				 Cell *continuation_environment,
				 Continuation *continuation)
    {
      // Makes FULL continuation.
      set_flags(CONTINUATION_CELL | FULL_CONTINUATION_FLAG);
      contents.cell.object.cons.car = procedure;
      contents.cell.object.cons.cdr = continuation_environment;
      contents.cell.object.cons.option.continuation = continuation;
      return this;
    }

    inline bool is_continuation() const
    {
      return cell_type() == CONTINUATION_CELL;
    }

    inline bool is_full_continuation() const
    {
      return cell_type() == CONTINUATION_CELL
	&& (get_flags() & FULL_CONTINUATION_FLAG) != 0;
    }

    inline Cell *get_continuation_procedure() const
    {
      assert(is_continuation());

      return contents.cell.object.cons.car;
    }

    inline const char *get_continuation_pc() const
    {
      assert(is_continuation());

      if ((get_flags() & FULL_CONTINUATION_FLAG) == 0)
	return contents.cell.object.cons.option.pc;

      return contents.cell.object.cons.option.continuation->get_pc();
    }

    inline Cell *get_continuation_frame() const
    {
      assert(is_continuation());

      if ((get_flags() & FULL_CONTINUATION_FLAG) == 0)
	return contents.cell.object.cons.cdr;

      return CAR(contents.cell.object.cons.cdr);
    }

    inline Cell *get_continuation_input_port() const
    {
      assert(is_continuation() && (get_flags() & FULL_CONTINUATION_FLAG));

      return CADR(contents.cell.object.cons.cdr);
    }

    inline Cell *get_continuation_output_port() const
    {
      assert(is_continuation() && (get_flags() & FULL_CONTINUATION_FLAG));

      return CADDR(contents.cell.object.cons.cdr);
    }

    inline long get_continuation_depth() const
    {
      assert(is_continuation() && (get_flags() & FULL_CONTINUATION_FLAG));

      return contents.cell.object.cons.option.continuation->get_depth();
    }

    inline Continuation *get_continuation() const
    {
      assert(is_continuation());

      if ((get_flags() & FULL_CONTINUATION_FLAG) == 0)
	return Continuation::get_singleton();

      return contents.cell.object.cons.option.continuation;
    }

    inline bool is_procedure() const
    {
      long type = cell_type();
      return type == CLOSURE_CELL
	|| type == NATIVE_PROCEDURE_CELL
	|| type == CONTINUATION_CELL;
    }

    inline Procedure *get_procedure_invoker() const
    {
      assert(is_procedure());

      long type = cell_type();
      if (type == CLOSURE_CELL)
	return get_closure_bytecode();

      if (type == NATIVE_PROCEDURE_CELL)
	return get_native_procedure();

      return get_continuation();
    }


    inline bool is_promise() const { return cell_type() == PROMISE_CELL; }

    inline bool is_forced_promise() const
    {
      return is_promise() && (get_flags() & FORCED_PROMISE_FLAG) != 0;
    }

    inline Cell *mk_promise(Cell *procedure)
    {
      assert(procedure->is_procedure());

      set_flags(PROMISE_CELL);
      contents.cell.object.cons.car = procedure;
      contents.cell.object.cons.cdr = Cell::nil();
      return this;
    }

    inline Cell *get_promise_procedure() const
    {
      assert(is_promise());
      assert(!is_forced_promise());

      return contents.cell.object.cons.car;
    }

    Cell *memoize_promise_value(MemorySystem &ms, Cell *value);

    inline Cell *get_memoized_value() const
    {
      assert(is_forced_promise());

      return contents.cell.object.cons.cdr;
    }



    inline Cell *mk_port(Port *port)
    {
      set_flags(PORT_CELL);
      contents.cell.object.port.port = port;
      return this;
    }

    inline bool is_port() const { return cell_type() == PORT_CELL; }

    inline bool is_input_port() const
    {
      return is_port() && contents.cell.object.port.port->is_input_port();
    }

    inline bool is_output_port() const
    {
      return is_port() && contents.cell.object.port.port->is_output_port();
    }

    inline Port *get_port() const
    {
      assert(is_port());

      return contents.cell.object.port.port;
    }

    inline void close_port()
    {
      assert(is_port());

      get_port()->close();
    }

    inline Cell *set_interactive(bool interactive = true)
    {
      if (interactive)
	set_flags(get_flags() | INTERACTIVE_PORT_FLAG);
      else
	set_flags(get_flags() & ~INTERACTIVE_PORT_FLAG);

      return this;
    }

    inline bool is_interactive_port() const
    {
      return cell_type() == PORT_CELL
	&& (get_flags() & INTERACTIVE_PORT_FLAG) != 0;
    }



    inline Cell *mk_time(long time_type, long nanosecond, long second)
    {
      set_flags(TIME_CELL);
      contents.cell.object.time.type = time_type;
      contents.cell.object.time.nanosecond = nanosecond;
      contents.cell.object.time.second = second;
      return this;
    }

    inline bool is_time() const { return cell_type() == TIME_CELL; }

    inline long get_time_type() const
    {
      assert(is_time());

      return contents.cell.object.time.type;
    }

    inline Cell *set_time_type(long time_type)
    {
      assert(is_time());

      contents.cell.object.time.type = time_type;
      return this;
    }

    inline long get_time_nanosecond() const
    {
      assert(is_time());

      return contents.cell.object.time.nanosecond;
    }

    inline Cell *set_time_nanosecond(long nanosecond)
    {
      assert(is_time());

      contents.cell.object.time.nanosecond = nanosecond;
      return this;
    }

    inline long get_time_second() const
    {
      assert(is_time());

      return contents.cell.object.time.second;
    }

    inline Cell *set_time_second(long second)
    {
      assert(is_time());

      contents.cell.object.time.second = second;
      return this;
    }


#ifdef HAVE_TIME_H
    Cell *mk_date(const struct tm *value, long nanosecond, long zone_offset);

    const struct tm *get_date() const
    {
      assert(is_date());

      return contents.cell.object.date.value;
    }

    inline long get_date_nanosecond() const
    {
      assert(is_date());

      return contents.cell.object.date.nanosecond;
    }

    inline long get_date_second() const
    {
      assert(is_date());

      return contents.cell.object.date.value->tm_sec;
    }

    inline long get_date_minute() const
    {
      assert(is_date());

      return contents.cell.object.date.value->tm_min;
    }

    inline long get_date_hour() const
    {
      assert(is_date());

      return contents.cell.object.date.value->tm_hour;
    }

    inline long get_date_day() const
    {
      assert(is_date());

      return contents.cell.object.date.value->tm_mday;
    }

    inline long get_date_month() const
    {
      assert(is_date());

      return contents.cell.object.date.value->tm_mon + 1;
    }

    inline long get_date_year() const
    {
      assert(is_date());

      return contents.cell.object.date.value->tm_year + 1900;
    }

    inline long get_date_week_day() const
    {
      assert(is_date());

      return contents.cell.object.date.value->tm_wday;
    }

    inline long get_date_year_day() const
    {
      assert(is_date());

      return contents.cell.object.date.value->tm_yday;
    }

    inline long get_date_zone_offset() const
    {
      assert(is_date());

      return contents.cell.object.date.zone_offset;
    }
#endif
    inline bool is_date() const { return cell_type() == DATE_CELL; }

  };


  inline Cell *ConstantPool::load_constant(const Cell *constant_pool,
					   long index) const
  {
    assert(constant_pool->cell_type() == Cell::CONSTANT_POOL_CELL);
    assert(index < get_number_of_items());

    return constant_pool->get_constant_table()->get_vector_element(index);
  }


}


#endif	/* __cplusplus */

#endif	/* _LIBTSCORE_CELL_H_ */

#if defined(_MEMORY_SYSTEM_DEFINED_) && !defined(_CELL_INLINE_FUNC_DEFINED_)

namespace LibTSCore
{
  /*
   * Stores a value into scheme frame at specified slot.
   * Because of memory barrier, we need a reference to the MemorySystem.
   */
  inline Cell *Cell::store_variable(MemorySystem &ms, long slot,
				    Cell *value)
  {
    assert(is_scheme_frame());
    assert(slot < get_vector_length());

    ms.set_vector_element(this, slot, value);
    return value;
  }
}

#define _CELL_INLINE_FUNC_DEFINED_	1
#endif
