/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: environment.cc,v 1.9 2009/01/01 07:47:00 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libtscore/environment.h"

#include "libtscore/cell.h"
#include "libtscore/memory_system.h"

#include <cassert>
#include <cstdio>
#include <cstdlib>


namespace LibTSCore
{

Environment::Environment(MemorySystem &ms)
  : number_of_variables_bound(0L),
    stack_bottom(0L),
    stack_top(0L),
    maximum_stack_depth(0L),
    memory_system(ms)
{}


const Cell *
Environment::bind_syntax(Cell *env, const Cell *syntax)
{
  assert(env->is_environment());

  return bind(env, syntax->get_syntax_symbol(), syntax);
}


void
Environment::unbind_syntax(Cell *env, const Cell *syntax)
{
  assert(env->is_environment());

  const Cell *symbol = syntax->get_syntax_symbol();
  const Cell *binding = lookup0(env->get_bindings(), symbol);
  if (binding == syntax)
    unbind(env, symbol);
}


void
Environment::unbind_variable(Cell *env, const Cell *symbol)
{
  assert(env->is_environment());

  const Cell *binding = lookup0(env->get_bindings(), symbol);
  if (binding->is_binding())
    unbind(env, symbol);
}


const Cell *
Environment::bind_variable(Cell *env, const Cell *symbol)
{
  assert(env->is_environment());

  const Cell *binding = lookup0(env->get_bindings(), symbol);
  if (binding->is_syntax() || binding == Cell::nil())
    {
      Cell *sym = const_cast<Cell *>(symbol);
      Cell *new_binding = memory_system.get_cell(sym, env);
      new_binding->mk_binding(sym, env, new_variable_slot());
      return bind(env, sym, new_binding);
    }

  if (binding->get_slot_location() == -1)
    const_cast<Cell *>(binding)->set_slot_location(new_variable_slot());

  return binding;
}


const Cell *
Environment::make_unbound_variable(Cell *env, const Cell *symbol)
{
  assert(env->is_environment());

  const Cell *binding = lookup0(env->get_bindings(), symbol);
  if (binding->is_syntax() || binding == Cell::nil())
    {
      Cell *sym = const_cast<Cell *>(symbol);
      Cell *new_binding = memory_system.get_cell(sym, env);
      new_binding->mk_binding(sym, env, -1);
      return bind(env, sym, new_binding);
    }

  return binding;
}


void
Environment::rebind_binding(Cell *env, const Cell *binding)
{
  if (binding->is_syntax())
    bind_syntax(env, binding);
  else if (binding->is_binding())
    bind(env, binding->get_bound_symbol(), binding);
  else
    unbind(env, CAR(binding));
}


const Cell *
Environment::lookup(const Cell *env, const Cell *symbol, bool all) const
{
  assert(env->is_environment());

  const Cell *binding = lookup0(env->get_bindings(), symbol);
  if (binding != Cell::nil() || !all)
    return binding;

  Cell *parent = env->get_parent_environment();
  if (parent != Cell::nil())
    return parent->lookup(symbol, all);

  return Cell::nil();
}


bool
Environment::is_bound(const Cell *env, const Cell *symbol, bool all) const
{
  return lookup(env, symbol, all) != Cell::nil();
}


Cell *
Environment::store_variable(Cell *frame, long slot, Cell *value) const
{
  assert(frame->is_toplevel_frame());

  Cell *storage = frame->get_variable_slots();
  if (slot < storage->get_vector_length())
    {
      memory_system.set_vector_element(storage, slot, value);
      return Cell::unspecified();
    }

  // we must enlarge storage here.
  StackRoot tmp_frame(memory_system, frame);
  StackRoot val(memory_system, value);
  storage = memory_system.copy_vector(estimate_storage_size(), storage,
				      Cell::unspecified0());
  if (storage == Cell::sink())
    {
      fprintf(stderr, "Out of memory\n");
      return NULL;
    }

  assert(slot < storage->get_vector_length());

  memory_system.set_cdr(tmp_frame, storage);
  memory_system.set_vector_element(storage, slot, val);
  return Cell::unspecified();
}


const Cell *
HashEnvironment::bind(Cell *env, const Cell *symbol, const Cell *binding)
{
  StackRoot bindings(memory_system, env->get_bindings());
  long location
    = static_cast<unsigned long>(symbol->symbol_hash())
    % static_cast<unsigned long>(bindings->get_vector_length());
  Cell *candidate = bindings->get_vector_element(location);

  // candidate is a list of `binding object' and `syntax object'
  // each of these types are implemented as cons Cell whose car has
  // symbol.
  for (Cell *pred = Cell::nil(), *entry = candidate;
       entry != Cell::nil();
       pred = entry, entry = CDR(entry))
    {
      Cell *cs = CAAR(entry);
      if (cs == symbol)
	{
	  assert(CAR(entry)->is_syntax());

	  if (pred != Cell::nil())
	    memory_system.set_cdr(pred, CDR(entry));
	  else
	    candidate = CDR(entry);

	  break;
	}
    }

  // At here, candidate should be top of the list within the same
  // hash slot and there's no bindings for `symbol' in the list.
  Cell *bin = const_cast<Cell *>(binding);
  candidate = memory_system.get_cons(bin, candidate);
  memory_system.set_vector_element(bindings, location, candidate);
  return bin;
}


void
HashEnvironment::unbind(Cell *env, const Cell *symbol)
{
  Cell *bindings = env->get_bindings();
  long location
    = static_cast<unsigned long>(symbol->symbol_hash())
    % static_cast<unsigned long>(bindings->get_vector_length());
  Cell *candidate = bindings->get_vector_element(location);
  for (Cell *pred = Cell::nil(), *entry = candidate;
       entry != Cell::nil();
       pred = entry, entry = CDR(entry))
    {
      Cell *cs = CAAR(entry);
      if (cs == symbol)
	{
	  if (pred != Cell::nil())
	    memory_system.set_cdr(pred, CDR(entry));
	  else
	    candidate = CDR(entry);

	  memory_system.set_vector_element(bindings, location, candidate);
	  return;
	}
    }
}


const Cell *
HashEnvironment::lookup0(const Cell *bindings, const Cell *symbol) const
{
  long location
    = static_cast<unsigned long>(symbol->symbol_hash())
    % static_cast<unsigned long>(bindings->get_vector_length());
  for (Cell *candidate = bindings->get_vector_element(location);
       candidate != Cell::nil();
       candidate = CDR(candidate))
    {
      Cell *cs = CAAR(candidate);
      if (cs == symbol)
	return CAR(candidate);
    }

  return Cell::nil();
}


long
HashEnvironment::estimate_storage_size() const
{
  // Parameters below should be adjusted for each application domain.
  // Since HashEnvironment usually corresponds to top level environment,
  // number of variables bound may grow up in an unlimited manner.

  // Note: Although we use HashEnvironment only for top level environment,
  //       HashEnvironment itself isn't limited to be used for top
  //       level environment.  Therefore we should check the number of
  //       variables currently bound here.  However, top level environment
  //       is always instantiated before any variables bound, and thus,
  //       this method usually returns 128 -- the magic number should
  //       be changed in the future.
  long n_var = get_number_of_variables_bound() + get_maximum_stack_depth();
  return n_var > 128 ? n_var + 32 : 128;
}


HashEnvironment::HashEnvironment(MemorySystem &ms)
  : Environment(ms)
{
}


const Cell *
ListEnvironment::bind(Cell *env, const Cell *symbol, const Cell *binding)
{
  StackRoot tmp_env(memory_system, env);
  StackRoot bindings(memory_system, env->get_bindings());
  Cell *candidate = bindings;

  // candidate is a list of `binding object' and `syntax object'
  // each of these types are implemented as cons Cell whose car has
  // symbol.
  for (Cell *pred = Cell::nil(), *entry = candidate;
       entry != Cell::nil();
       pred = entry, entry = CDR(entry))
    {
      Cell *cs = CAAR(entry);
      if (cs == symbol)
	{
	  assert(CAR(entry)->is_syntax());
	  if (pred != Cell::nil())
	    memory_system.set_cdr(pred, CDR(entry));
	  else
	    candidate = CDR(entry);

	  break;
	}
    }

  // At here, candidate should be top of the list within the same
  // hash slot and there's no bindings for `symbol' in the list.
  Cell *bin = const_cast<Cell *>(binding);
  candidate = memory_system.get_cons(bin, candidate);
  memory_system.set_cdr(tmp_env, candidate);
  return bin;
}


void
ListEnvironment::unbind(Cell *env, const Cell *symbol)
{
  Cell *bindings = env->get_bindings();
  Cell *candidate = bindings;
  for (Cell *pred = Cell::nil(), *entry = candidate;
       entry != Cell::nil();
       pred = entry, entry = CDR(entry))
    {
      Cell *cs = CAAR(entry);
      if (cs == symbol)
	{
	  if (pred != Cell::nil())
	    memory_system.set_cdr(pred, CDR(entry));
	  else
	    candidate = CDR(entry);

	  memory_system.set_cdr(env, candidate);
	  return;
	}
    }
}


const Cell *
ListEnvironment::lookup0(const Cell *bindings, const Cell *symbol) const
{
#if 1
  for (const Cell *candidate = bindings;
       candidate != Cell::nil();
       candidate = CDR(candidate))
    {
      Cell *cs = CAAR(candidate);
      if (cs == symbol)
	return CAR(candidate);
    }
#else
  const Cell *candidate = bindings;
  if (candidate != Cell::nil())
    {
      for (; candidate != Cell::nil(); candidate = CDR(candidate))
	{
	  Cell *cs = CAAR(candidate);
	  if (cs == symbol)
	    return CAR(candidate);
	}
    }
#endif

  return Cell::nil();
}


long
ListEnvironment::estimate_storage_size() const
{
  // Parameters below should be adjusted for each application domain.
  // Since we don't any kind of compilation, the maximum number of
  // variables will be bound to a given environment cannot be known
  // at all.
  return get_number_of_variables_bound() + get_maximum_stack_depth();
}


ListEnvironment::ListEnvironment(MemorySystem &ms)
  : Environment(ms)
{
}


}
