/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: environment.h,v 1.8 2009/01/01 07:47:00 fuyu Exp $
 */

#ifndef _LIBTSCORE_ENVIRONMENT_H_
#define _LIBTSCORE_ENVIRONMENT_H_

#ifdef __cplusplus

#if defined(__GNUG__) && !defined(__ICC)
#pragma interface
#endif

#include <cassert>


using namespace std;
namespace LibTSCore
{

  class Cell;
  class MemorySystem;

  /*
   * An identifier may name a type of syntax, or it may name a location
   * where a value can be stored.  An identifier that names a type of
   * syntax is called a `syntactic keyword' and is said to be `bound' to
   * that syntax.  An identifier that names a location is called a
   * `variable' and is said to be `bound' to that location.  The set of
   * all visible bindings in effect at some point in a program is known
   * as the `ENVIRONMENT' in effect at that point.
   *             -- Revised^5 Report on the Algorithmic Language Scheme
   *
   * Each instance of this class represents an `ENVIRONMENT' in a program
   * point within our Scheme system.
   */
  class Environment
  {
    friend class Cell;

  private:
    long number_of_variables_bound;
    long stack_bottom;
    long stack_top;
    long maximum_stack_depth;

    inline long new_variable_slot()
    {
      if (stack_top > number_of_variables_bound)
	number_of_variables_bound = stack_top;

      return number_of_variables_bound++;
    }


  protected:
    MemorySystem &memory_system;

    virtual const Cell *bind(Cell *env, const Cell *symbol,
			     const Cell *binding) = 0;
    virtual void unbind(Cell *env, const Cell *symbol) = 0;
    virtual const Cell *lookup0(const Cell *bindings,
				const Cell *symbol) const = 0;
    virtual long estimate_storage_size() const = 0;


  public:
    Environment(MemorySystem &ms);
    virtual ~Environment() {}


    /*
     * Returns the number of variables bound within this Environment.
     *
     * See also descriptions about `environment/scheme frame'
     * placed around definition of environment Cell below.
     */
    inline long get_number_of_variables_bound() const
    {
      return number_of_variables_bound;
    }


    /*
     * Returns the maximum depth of stack slots simultaneouly used.
     */
    inline long get_maximum_stack_depth() const
    {
      return maximum_stack_depth;
    }


    /*
     * Returns the `binding object' that is either a syntax Cell
     * or a pair of (env . slot#).  C++ level Envrionment totally ignore
     * what it is, instead it only provides a repository that manages
     * relations between symbol Cells and binding objects.
     *
     * The `env' must be an `environment' corresponds to this.
     * When `all' is true, parent environments are also looked up
     * recursively until a binding is found or top level environment
     * if there's no binding.
     */
    const Cell *lookup(const Cell *env, const Cell *symbol, bool all) const;


    /*
     * Tests whether the `symbol' has been bound within this Environment.
     *
     * The `env' must be an `environment' corresponds to this.
     * When `all' is true, parent environments are also tested
     * recursively until a binding is found or top level environment
     * if there's no binding.
     */
    bool is_bound(const Cell *env, const Cell *symbol, bool all) const;


    /*
     * Binds a syntax within this Environment.  The symbol to which the
     * syntax bound is retrieved from syntax Cell.
     */
    const Cell *bind_syntax(Cell *env, const Cell *syntax);


    /*
     * Unbinds a syntax within this Environment.
     */
    void unbind_syntax(Cell *env, const Cell *syntax);


    /*
     * Unbinds a variable within this Environment.
     */
    void unbind_variable(Cell *env, const Cell *symbol);


    /*
     * Binds a symbol to a variable to, conceptually, newly allocated
     * slot.  The actual storage for the slot will never be allocated
     * by this because the storage belongs to `scheme frame'
     * and C++ Environment only belongs to `environment'.
     */
    const Cell *bind_variable(Cell *env, const Cell *symbol);


    /*
     * Makes a binding object that will be used as a placeholder for
     * an unbound variable during hygienic rewriting of macro transformation.
     */
    const Cell *make_unbound_variable(Cell *env, const Cell *symbol);


    /*
     * Rebinds previously bound syntax or variable.
     */
    void rebind_binding(Cell *env, const Cell *binding);


    /*
     * Creates an anonymous variable slot.  This is only used for the
     * implicit argument, i.e. continuation.
     */
    inline long allocate_anonymous_slot(Cell *env)
    {
      return new_variable_slot();
    }


    /*
     * Stores a value in the location specified by the pair of scheme
     * frame and slot number.
     *
     * `frame' must be a scheme frame that represents the variable
     * storage for given `environment' that is associated with this
     * Environment.
     *
     * This method is only for so-called toplevel environment/frame
     * that is expected to grow incrementally.  We need not to worry
     * about other environments/frames because their size and bound
     * locations are byte-compilation time constant.
     */
    Cell *store_variable(Cell *frame, long slot, Cell *value) const;


    /*
     * Allocates a slot that will be used like a stack variable.
     */
    inline long allocate_stack_slot()
    {
      if (stack_bottom >= number_of_variables_bound)
	{
	  if (maximum_stack_depth < (stack_top - stack_bottom + 1))
	    maximum_stack_depth = stack_top - stack_bottom + 1;

	  return stack_top++;
	}

      stack_top = stack_bottom = number_of_variables_bound;
      return stack_top++;
    }


    /*
     * Returns the current stack top.
     */
    inline long get_current_stack_pointer() const { return stack_top; }


    /*
     * Disposes all stack slots placed at shallower than specified `depth'.
     */
    inline void dispose_stack_slots(long depth)
    {
      assert(depth >= stack_bottom);

      stack_top = depth;
      if (stack_top < number_of_variables_bound)
	number_of_variables_bound = stack_top;
    }
  };


  /*
   * HashEnvironment is an implementation of Environment intended
   * to be used for representations of Environment in which many
   * bindings will be recorded, e.g. top level environment.
   */
  class HashEnvironment : public Environment
  {
  protected:
    virtual const Cell *bind(Cell *env, const Cell *symbol,
			     const Cell *binding);

    virtual void unbind(Cell *env, const Cell *symbol);

    virtual const Cell *lookup0(const Cell *bindings, const Cell *symbol) const;

    virtual long estimate_storage_size() const;


  public:
    HashEnvironment(MemorySystem &ms);
  };


  /*
   * ListEnvironment
   */
  class ListEnvironment : public Environment
  {
  protected:
    virtual const Cell *bind(Cell *env, const Cell *symbol,
			     const Cell *binding);

    virtual void unbind(Cell *env, const Cell *symbol);

    virtual const Cell *lookup0(const Cell *bindings, const Cell *symbol) const;

    virtual long estimate_storage_size() const;


  public:
    ListEnvironment(MemorySystem &ms);
  };


}


#endif	/* __cplusplus */

#endif	/* _LIBTSCORE_ENVIRONMENT_H_ */
