/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: gettext_procedure.cc,v 1.1 2009/01/05 12:16:19 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libtscore/gettext_procedure.h"

#include "libtscore/tscore.h"

#ifdef ENABLE_NLS
#include <libintl.h>
#endif


#include <cstring>


namespace LibTSCore
{

NativeProcedure::Info GettextProcedure::procedures[] = {
  { &GettextProcedure::gettext,
    "_", TST_STRING, 1, 1 },
  { &GettextProcedure::gettext,
    "gettext", TST_STRING, 1, 1 },
  { &GettextProcedure::ngettext,
    "ngettext", TST_STRING TST_STRING TST_NONNEG, 3, 3 },
  { &GettextProcedure::textdomain,
    "textdomain", TST_STRING, 0, 1 },
  { &GettextProcedure::bindtextdomain,
    "bindtextdomain", TST_STRING TST_STRING, 1, 2 }
};


Cell *
GettextProcedure::assign_procedures(Cell *environment, Cell *frame)
{
  TSStackRoot env(core, environment);
  TSStackRoot frm(core, frame);
  size_t n_procs = sizeof(procedures) / sizeof(Info);
  for (size_t i = 0; i < n_procs; i++)
    assign_procedure(env, frm, &procedures[i], this);

  return env;
}


#define ARG(n)	context.frame->load_variable(argument_pointer + (n))


Cell *
GettextProcedure::gettext(VirtualMachine &vm, Register &context,
			  unsigned long argument_pointer,
			  unsigned long nargs, void *data)
{
#ifdef ENABLE_NLS
  Cell *str = ARG(0);
  char *tmp_msgid;
  const char *cstr = str->get_string();
  const char *msgid;
  if (strlen(cstr) == str->get_strlen())
    {
      tmp_msgid = NULL;
      msgid = cstr;
    }
  else
    {
      size_t len = str->get_strlen();
      tmp_msgid = static_cast<char *>(malloc(len + 1));
      if (tmp_msgid == NULL)
	return signal_error(vm, "gettext: Out of Memory: ", str);

      memcpy(tmp_msgid, cstr, len);
      tmp_msgid[len] = '\0';

      msgid = tmp_msgid;
    }

  const char *message = ::gettext(msgid);
  Cell *result = vm.mk_static_string(message);

  if (tmp_msgid != NULL)
    free(tmp_msgid);

  return result;
#else
  return ARG(0);
#endif
}


Cell *
GettextProcedure::ngettext(VirtualMachine &vm, Register &context,
			   unsigned long argument_pointer,
			   unsigned long nargs, void *data)
{
#ifdef ENABLE_NLS
  Cell *str = ARG(0);
  const char *cstr = str->get_string();
  char *tmp_msgid;
  const char *msgid;
  if (strlen(cstr) == str->get_strlen())
    {
      tmp_msgid = NULL;
      msgid = cstr;
    }
  else
    {
      size_t len = str->get_strlen();
      tmp_msgid = static_cast<char *>(malloc(len + 1));
      if (tmp_msgid == NULL)
	return signal_error(vm, "ngettext: Out of Memory: ", str);

      memcpy(tmp_msgid, cstr, len);
      tmp_msgid[len] = '\0';

      msgid = tmp_msgid;
    }

  str = ARG(1);
  cstr = str->get_string();
  char *tmp_plural;
  const char *msgid_plural;
  if (strlen(cstr) == str->get_strlen())
    {
      tmp_plural = NULL;
      msgid_plural = cstr;
    }
  else
    {
      size_t len = str->get_strlen();
      tmp_plural = static_cast<char *>(malloc(len + 1));
      if (tmp_plural == NULL)
	return signal_error(vm, "ngettext: Out of Memory: ", str);

      memcpy(tmp_plural, cstr, len);
      tmp_plural[len] = '\0';

      msgid_plural = tmp_plural;
    }

  const char *message = ::ngettext(msgid, msgid_plural, ARG(2)->get_integer0());
  Cell *result = vm.mk_static_string(message);

  if (tmp_plural != NULL)
    free(tmp_plural);

  if (tmp_msgid != NULL)
    free(tmp_msgid);

  return result;
#else
  return ARG(2)->get_integer0() == 1 ? ARG(0) : ARG(1);
#endif
}


Cell *
GettextProcedure::textdomain(VirtualMachine &vm, Register &context,
			     unsigned long argument_pointer,
			     unsigned long nargs, void *data)
{
#ifdef ENABLE_NLS
  char *tmp_domain;
  const char *domain;

  if (nargs == 0)
    {
      tmp_domain = NULL;
      domain = NULL;
    }
  else
    {
      Cell *str = ARG(0);
      if (str->get_strlen() == 0)
	return signal_error(vm,
			    "textdomain: domainname must be a non-empty string: ",
			    str);

      const char *cstr = str->get_string();
      if (strlen(cstr) == str->get_strlen())
	{
	  tmp_domain = NULL;
	  domain = cstr;
	}
      else
	{
	  size_t len = str->get_strlen();
	  tmp_domain = static_cast<char *>(malloc(len + 1));
	  if (tmp_domain == NULL)
	    return signal_error(vm, "textdomain: Out of Memory: ", str);

	  memcpy(tmp_domain, cstr, len);
	  tmp_domain[len] = '\0';

	  domain = tmp_domain;
	}
    }

  const char *current_domain = ::textdomain(domain);
  Cell *result = vm.mk_string(current_domain);

  if (tmp_domain != NULL)
    free(tmp_domain);

  return result;
#else
  return Cell::f();
#endif
}


Cell *
GettextProcedure::bindtextdomain(VirtualMachine &vm, Register &context,
				 unsigned long argument_pointer,
				 unsigned long nargs, void *data)
{
#ifdef ENABLE_NLS
  Cell *str = ARG(0);
  if (str->get_strlen() == 0)
    return signal_error(vm,
			"bindtextdomain: domainname must be a non-empty string: ",
			str);

  const char *cstr = str->get_string();
  char *tmp_domain;
  const char *domain;
  if (strlen(cstr) == str->get_strlen())
    {
      tmp_domain = NULL;
      domain = cstr;
    }
  else
    {
      size_t len = str->get_strlen();
      tmp_domain = static_cast<char *>(malloc(len + 1));
      if (tmp_domain == NULL)
	return signal_error(vm, "bindtextdomain: Out of Memory: ", str);

      memcpy(tmp_domain, cstr, len);
      tmp_domain[len] = '\0';

      domain = tmp_domain;
    }

  char *tmp_dirname;
  const char *dirname;
  if (nargs == 1)
    {
      tmp_dirname = NULL;
      dirname = NULL;
    }
  else
    {
      str = ARG(1);
      cstr = str->get_string();
      if (str->get_strlen() == 0)
	{
	  tmp_dirname = NULL;
	  dirname = NULL;
	}
      else if (strlen(cstr) == str->get_strlen())
	{
	  tmp_dirname = NULL;
	  dirname = cstr;
	}
      else
	{
	  size_t len = str->get_strlen();
	  tmp_dirname = static_cast<char *>(malloc(len + 1));
	  if (tmp_dirname == NULL)
	    return signal_error(vm, "bindtextdomain: Out of Memory: ", str);

	  memcpy(tmp_dirname, cstr, len);
	  tmp_dirname[len] = '\0';

	  dirname = tmp_dirname;
	}
    }

  const char *current_dirname = ::bindtextdomain(domain, dirname);
  Cell *result = vm.mk_string(current_dirname);

  if (tmp_dirname != NULL)
    free(tmp_dirname);

  if (tmp_domain != NULL)
    free(tmp_domain);

  return result;
#else
  return Cell::f();
#endif
}


}
