/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: macro_syntax.cc,v 1.14 2009/01/01 07:47:00 fuyu Exp $
 */

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libtscore/macro_syntax.h"

#include "libtscore/bytecode_buffer.h"
#include "libtscore/virtual_machine.h"
#include "libtscore/r5rs_misc_procedure.h"
#include "libtscore/syntax.h"

#include <algorithm>
#include <vector>

#include <cassert>


namespace LibTSCore
{

size_t
MacroSyntax::emit_bytecode(VirtualMachine &vm, Register &context,
			   BytecodeBuffer *buffer,
			   Cell *expression, Cell *environment,
			   Cell *constant_pool, Cell *hint,
			   bool is_tail_context)
{
  char buf[1024];
  const Cell *macro;
  if (expression->is_symbol())
    macro = environment->lookup(expression, true);
  else if (expression->is_syntax())
    macro = expression;
  else if (CAR(expression)->is_symbol())
    macro = environment->lookup(CAR(expression), true);
  else
    macro = CAR(expression);

  assert(macro->is_macro_syntax());

  const Cell *transformer = macro->get_macro_transformer();
  Cell *rule = match_syntax_rules(vm, expression, environment, transformer);

  if (rule == Cell::nil())
    {
      snprintf(buf, sizeof(buf), "%s: no pattern matched: ",
	       macro->get_syntax_name());
      return signal_error(vm, buf, expression);
    }

  if (vm.get_bytecompile_tracing(context))
    {
      Port *port = context.output_port->get_port();
      port->append("MacroSyntax::expand with (syntax-rule):\n(pattern): ");
      Writer::write_datum(CAR(rule), port, true, true);
      port->append("\n(template): ");
      Writer::write_datum(CADR(rule), port, true, true);
      port->append("\n(form): ");
      Writer::write_datum(expression, port, true, true);
      port->append('\n');
    }

  VMStackRoot form(vm, expression);
  VMStackRoot env(vm, environment);
  VMStackRoot macro_env(vm, transformer->get_macro_transformer_environment());
  VMStackRoot cpool(vm, constant_pool);
  VMStackRoot hnt(vm, hint);
  VMStackRoot literals(vm, transformer->get_macro_transformer_literals());
  Cell *expanded_form = expand_macro(vm, form, literals, rule, env, macro_env);

  if (vm.get_bytecompile_tracing(context))
    {
      Port *port = context.output_port->get_port();
      port->append("(expanded): ");
      Writer::write_datum(expanded_form, port, true, true);
      port->append('\n');
    }

  assert(expanded_form != Cell::nil());

  return bytecompile(vm, context, buffer,
		     expanded_form, env, cpool, hnt, is_tail_context);
}


bool
MacroSyntax::match_list_pattern(VirtualMachine &vm,
				Cell *expression, Cell *pattern,
				vector<const Cell *> &valid_literals,
				vector<const Cell *> &invalid_literals)
{
  if (!expression->is_pair())
    return false;

  Cell *flist = expression;
  Cell *plist = pattern;
  Cell *pat = CAR(pattern);
  while (flist->is_pair())
    {
      if (!match_pattern(vm, CAR(flist), pat,
			 valid_literals, invalid_literals))
	return false;

      if (plist == Cell::nil())
	return true;	// improper list P's Pn+1 matched nth cdr F

      flist = CDR(flist);
      if (!CDR(plist)->is_pair())
	{
	  if (CDR(plist) == Cell::nil())
	    {
	      plist = Cell::nil();
	      break;
	    }

	  // P is an improper list
	  pat = CDR(plist);
	  plist = Cell::nil();
	  continue;
	}
      else
	{
	  if  (CADR(plist) != vm.core.ellipsis_symbol)
	    {
	      plist = CDR(plist);
	      pat = CAR(plist);
	    }
	}
    }

  if (flist == Cell::nil())
    {
      if (plist == Cell::nil())
	return true;	// proper list P matched proper list F

      if (plist->is_pair() && CDR(plist)->is_pair()
	  && CADR(plist) == vm.core.ellipsis_symbol)
	return true;	// (P1 ... Pn Pn+1 <ellipsis>) matched
    }

  // flist != Cell::nil() only match improper list P but control
  // never reaches here for improper list P
  return false;
}


bool
MacroSyntax::match_vector_pattern(VirtualMachine &vm,
				  Cell *expression, Cell *pattern,
				  vector<const Cell *> &valid_literals,
				  vector<const Cell *> &invalid_literals)
{
  if (!expression->is_vector())
    return false;

  long plen = pattern->get_vector_length();
  long flen = expression->get_vector_length();
  if (flen < plen)
    {
      if (pattern->get_vector_element(plen - 1) != vm.core.ellipsis_symbol
	  || flen < plen - 2)
	return false;

      plen = flen;	// we only need to match expression's length here
    }
  else if (flen > plen)
    {
      if (plen < 2
	  || pattern->get_vector_element(plen - 1) != vm.core.ellipsis_symbol)
	return false;

      Cell *pat = pattern->get_vector_element(plen - 2);
      for (long i = plen - 2; i < flen; i++)
	if (!match_pattern(vm, expression->get_vector_element(i), pat,
			   valid_literals, invalid_literals))
	  return false;

      plen -= 2;
      flen = plen;	// remaining elements of expression matched Pn+1
    }
  else
    {
      if (plen >= 2
	  && pattern->get_vector_element(plen - 1) == vm.core.ellipsis_symbol)
	{
	  Cell *pat = pattern->get_vector_element(plen - 2);
	  for (long i = plen - 2; i < plen; i++)
	    if (!match_pattern(vm, expression->get_vector_element(i), pat,
			       valid_literals, invalid_literals))
	      return false;

	  plen -= 2;
	  flen = plen;
	}
    }

  for (long i = 0; i < plen; i++)
    if (!match_pattern(vm, expression->get_vector_element(i),
		       pattern->get_vector_element(i),
		       valid_literals, invalid_literals))
      return false;

  return true;
}


bool
MacroSyntax::match_pattern(VirtualMachine &vm,
			   Cell *expression, Cell *pattern,
			   vector<const Cell *> &valid_literals,
			   vector<const Cell *> &invalid_literals)
{
  if (pattern->is_symbol())
    {
      if (find(invalid_literals.begin(), invalid_literals.end(), pattern)
	  != invalid_literals.end())
	return false;	// differently bound literal identifier

      if (find(valid_literals.begin(), valid_literals.end(), pattern)
	  == valid_literals.end())
	return true;	// non-literal identifier

      return pattern == expression
	|| (expression->is_syntax()
	    && expression->get_syntax_symbol() == pattern);
    }

  if (pattern->is_pair())
    return match_list_pattern(vm, expression, pattern,
			      valid_literals, invalid_literals);

  if (pattern->is_vector())
    return match_vector_pattern(vm, expression, pattern,
				valid_literals, invalid_literals);

  return R5RSMiscProcedure::equal(pattern, expression);
}


Cell *
MacroSyntax::match_syntax_rules(VirtualMachine &vm,
				Cell *expression, Cell *environment,
				const Cell *transformer)
{
  vector<const Cell *> valid_literals;
  vector<const Cell *> invalid_literals;

  const Cell *literals = transformer->get_macro_transformer_literals();
  Cell *macro_env = transformer->get_macro_transformer_environment();
  while (literals != Cell::nil())
    {
      const Cell *symbol = CAR(literals);
      if (environment->lookup(symbol, true) == macro_env->lookup(symbol, true))
	valid_literals.push_back(symbol);
      else
	invalid_literals.push_back(symbol);

      literals = CDR(literals);
    }

  const Cell *syntax_rules = transformer->get_macro_transformer_syntax_rules();
  while (syntax_rules != Cell::nil())
    {
      Cell *rule = CAR(syntax_rules);
      if (match_pattern(vm, expression, CAR(rule),
			valid_literals, invalid_literals))
	return rule;

      syntax_rules = CDR(syntax_rules);
    }

  return Cell::nil();
}


const Cell *
MacroSyntax::match_ellipsis_form(VirtualMachine &vm,
				 const Cell *expression, const Cell *pattern,
				 const Cell *sub_pattern)
{
  if (!pattern->is_pair() && !pattern->is_vector())
    return NULL;

  if (pattern == sub_pattern)
    return expression;

  assert(pattern != Cell::nil());
  assert(sub_pattern != Cell::nil());
  assert(pattern != vm.core.ellipsis_symbol);
  assert(sub_pattern != vm.core.ellipsis_symbol);

  if (pattern->is_pair())
    {
      const Cell *pattern_list = pattern;
      const Cell *form_list = expression;

      // Look for sub_pattern within current pattern depth first.
      do
	{
	  if (CAR(pattern_list) == sub_pattern)
	    {
	      if (CDR(pattern_list)->is_pair()
		  && CADR(pattern_list) == vm.core.ellipsis_symbol)
		{
		  assert(CDDR(pattern_list) == Cell::nil());
		  return form_list;
		}

	      return CAR(form_list);
	    }

	  pattern_list = CDR(pattern_list);
	  if (form_list->is_pair())
	    form_list = CDR(form_list);
	  else
	    form_list = Cell::nil();	// result can be null
	}
      while (pattern_list->is_pair());

      // `sub_pattern' should be included somewhere deeper in the `pattern'.
      pattern_list = pattern;
      form_list = expression;
      const Cell *sub_form;
      if (form_list->is_pair())
	sub_form = CAR(form_list);
      else
	sub_form = form_list;

      do
	{
	  const Cell *result = match_ellipsis_form(vm, sub_form,
						   CAR(pattern_list),
						   sub_pattern);
	  if (result != NULL)
	    return result;

	  pattern_list = CDR(pattern_list);
	  if (form_list->is_pair())
	    {
	      form_list = CDR(form_list);
	      if (form_list->is_pair())
		sub_form = CAR(form_list);
	      else
		sub_form = form_list;
	    }
	  else
	    sub_form = Cell::nil();	// result can be null
	}
      while (pattern_list->is_pair());

      if (pattern_list != Cell::nil())
	{
	  assert(pattern_list->is_vector());

	  return match_ellipsis_form(vm, sub_form, pattern_list, sub_pattern);
	}

      // should never reached because `expression' matches `pattern'
      // and `sub_pattern' should be included in the `pattern'.
#ifndef NDEBUG
      assert(0);
#else
      return NULL;
#endif
    }

  // pattern is vector
  long plen = pattern->get_vector_length();
  assert(plen > 0);

  if (pattern->get_vector_element(plen - 1) == vm.core.ellipsis_symbol)
    {
      // Look for sub_pattern within current pattern depth first.
      assert(plen > 1);

      if (pattern->get_vector_element(plen - 2) == sub_pattern)
	{
	  // found!
	  if (expression == Cell::nil())
	    return Cell::nil();	// match omitted pattern

	  VMStackRoot form(vm, expression);
	  VMStackRoot result(vm, Cell::nil());

	  assert(expression->is_vector());

	  long flen = expression->get_vector_length();
	  plen -= 2;
	  while (--flen >= plen)
	    {
	      Cell *elm = expression->get_vector_element(flen);
	      result = vm.cons(elm, result);
	    }

	  return result;
	}
    }

  long i = 0;
  if (expression->is_vector())
    {
      for (long flen = expression->get_vector_length();
	   i < plen && i < flen;
	   i++)
	{
	  const Cell *sub_form
	    = match_ellipsis_form(vm, expression->get_vector_element(i),
				  pattern->get_vector_element(i),
				  sub_pattern);

	  if (sub_form != NULL)
	    return sub_form;
	}
    }

  while (i < plen)
    {
      // sub pattern may be included within omitted part of another ellipsis
      // pattern.
      const Cell *sub_form
	= match_ellipsis_form(vm, Cell::nil(),
			      pattern->get_vector_element(i++),
			      sub_pattern);
      if (sub_form != NULL)
	return sub_form;
    }

  // should never reached because `expression' match `pattern'
  // and `sub_pattern' should be included in the `pattern'.
#ifndef NDEBUG
  assert(0);
#else
  return NULL;
#endif
}


static const Cell *
find_ellipsis_predecessor(const Cell *pattern, const Cell *ellipsis_symbol)
{
  const Cell *pred = Cell::nil();
  if (pattern->is_pair())
    {
      const Cell *list = pattern;
      while (true)
	{
	  pred = CAR(list);
	  list = CDR(list);
	  if (list->is_pair())
	    {
	      if (CAR(list) == ellipsis_symbol)
		break;

	      continue;
	    }

	  break;
	}

      assert(list->is_pair());
      assert(CAR(list) == ellipsis_symbol);
    }
  else
    {
      assert(pattern->is_vector());

      long len = pattern->get_vector_length() - 1;
      assert(len > 0);
      assert(pattern->get_vector_element(len) == ellipsis_symbol);

      pred = pattern->get_vector_element(len - 1);
    }

  return pred;
}


Cell *
MacroSyntax::transform_ellipsis(VirtualMachine &vm,
				const Cell *expression,
				Cell *&literals, Cell *&template_alist,
				const Cell *pattern, const Cell *templ,
				Cell *&environment, Cell *&macro_environment)
{
  VMStackRoot form(vm, expression);
  VMStackRoot pat(vm, const_cast<Cell *>(pattern));
  VMStackRoot tpl(vm, const_cast<Cell *>(templ));

  VMStackRoot sub_pattern(vm, Cell::nil());
  const Cell *alist = template_alist;
  while (alist != Cell::nil())
    {
      if (CAAR(alist) == templ)
	{
	  sub_pattern = CDAR(alist);
	  break;
	}

      alist = CDR(alist);
    }
  assert(sub_pattern != Cell::nil());

  const Cell *pred = find_ellipsis_predecessor(pat, vm.core.ellipsis_symbol);
  assert(pred != Cell::nil());

  if (pred != sub_pattern)
    {
      const Cell *sub_form = match_ellipsis_form(vm, form, pat, sub_pattern);
      assert(sub_form != NULL);

      const Cell *new_pred
	= find_ellipsis_predecessor(sub_pattern, vm.core.ellipsis_symbol);

      if (pred != new_pred)
	return transform_ellipsis(vm, sub_form, literals, template_alist,
				  sub_pattern, tpl,
				  environment, macro_environment);
    }

  sub_pattern = const_cast<Cell *>(pred);

  const Cell *match_form = match_ellipsis_form(vm, form, pat, sub_pattern);
  assert(match_form != NULL);

  VMStackRoot sub_form(vm, match_form);

  // sub_form here is a null or a list of sub forms each match `sub_pattern'.
  VMStackRoot result(vm, Cell::nil());
  while (sub_form->is_pair())
    {
      // CAR(sub_form) should exactly match `sub_pattern'.
      Cell *sub_result = transform_expression(vm, CAR(sub_form),
					      literals, template_alist,
					      sub_pattern, tpl,
					      environment, macro_environment);

      if (sub_result != Cell::nil())
	result = vm.cons(sub_result, result);

      sub_form = CDR(sub_form);
    }
  assert(sub_form == Cell::nil());

  result = R5RSListProcedure::list_reverse_in_place(vm.get_memory_system(),
						    Cell::nil(), result, true);

  return result;
}


const Cell *
MacroSyntax::match_sub_form(VirtualMachine &vm,
			    const Cell *expression, const Cell *pattern,
			    const Cell *sub_pattern)
{
  if (sub_pattern == NULL)
    sub_pattern = Cell::nil();

  if (pattern == sub_pattern)
    return expression;

  if (sub_pattern == Cell::nil())
    return Cell::nil();

  if (pattern->is_pair())
    {
      assert(expression->is_pair());

      const Cell *form_list = expression;
      const Cell *pattern_list = pattern;
      assert(pattern_list != Cell::nil());

      do
	{
	  assert(pattern != Cell::nil());
	  assert(pattern_list->is_pair());
	  assert(pattern_list != Cell::nil());

	  const Cell *sub_form = match_sub_form(vm, CAR(form_list),
						CAR(pattern_list),
						sub_pattern);
	  if (sub_form != NULL)
	    return sub_form;

	  pattern_list = CDR(pattern_list);
	  if (form_list->is_pair())
	    form_list = CDR(form_list);
	  else
	    return NULL;
	}
      while (pattern_list->is_pair());

      if (pattern_list != Cell::nil())
	return match_sub_form(vm, form_list, pattern_list, sub_pattern);

      return NULL;
    }

  if (pattern->is_vector())
    {
      assert(expression->is_vector());

      long len = pattern->get_vector_length();
      for (long i = 0; i < len; i++)
	{
	  const Cell *sub_form
	    = match_sub_form(vm, expression->get_vector_element(i),
			     pattern->get_vector_element(i),
			     sub_pattern);
	  if (sub_form != NULL)
	    return sub_form;
	}
    }

  return NULL;
}


Cell *
MacroSyntax::transform_list(VirtualMachine &vm,
			    const Cell *expression,
			    Cell *&literals, Cell *&template_alist,
			    const Cell *pattern, const Cell *templ,
			    Cell *&environment, Cell *&macro_environment)
{
  assert(templ->is_pair());
  assert(expression != NULL);

  VMStackRoot form(vm, expression);
  VMStackRoot pat(vm, pattern);
  VMStackRoot tpl(vm, templ);
  VMStackRoot result(vm, Cell::nil());
  do
    {
      const Cell *sub_templ = CAR(tpl);
      tpl = CDR(tpl);
      if (tpl->is_pair() && CAR(tpl) == vm.core.ellipsis_symbol)
	{
	  tpl = CDR(tpl);
	  Cell *sub_result = transform_ellipsis(vm, form,
						literals, template_alist,
						pat, sub_templ,
						environment,
						macro_environment);

	  while (sub_result != Cell::nil())
	    {
	      Cell *sub_sub_form = CAR(sub_result);
	      result = vm.cons(sub_sub_form, result);
	      sub_result = CDR(sub_result);
	    }
	}
      else
	{
	  Cell *sub_form = transform_expression(vm, form,
						literals, template_alist,
						pat, sub_templ,
						environment,
						macro_environment);
	  result = vm.cons(sub_form, result);
	}
    }
  while (tpl->is_pair());

  if (tpl == Cell::nil())
    return R5RSListProcedure::list_reverse_in_place(vm.get_memory_system(),
						    Cell::nil(), result,
						    true);
  Cell *rest = transform_expression(vm, form,
				    literals, template_alist,
				    pat, tpl,
				    environment, macro_environment);
  return R5RSListProcedure::list_reverse_in_place(vm.get_memory_system(),
						  rest, result, true);
}


Cell *
MacroSyntax::transform_vector(VirtualMachine &vm,
			      const Cell *expression,
			      Cell *&literals, Cell *&template_alist,
			      const Cell *pattern, const Cell *templ,
			      Cell *&environment, Cell *&macro_environment)
{
  assert(templ->is_vector());
  assert(expression != NULL);

  VMStackRoot form(vm, expression);
  VMStackRoot pat(vm, pattern);
  VMStackRoot tpl(vm, templ);

  return Cell::nil();
}


static bool
is_literal(const Cell *literals, const Cell *template_alist,
	   const Cell *symbol)
{
  const Cell *list = literals;
  while (list != Cell::nil())
    {
      if (CAR(list) == symbol)
	return true;
      list = CDR(list);
    }

  list = template_alist;
  while (list != Cell::nil())
    {
      if (CAAR(list) == symbol)
	return CDAR(list) == Cell::nil();

      list = CDR(list);
    }

  return false;
}


Cell *
MacroSyntax::transform_expression(VirtualMachine &vm,
				  const Cell *expression,
				  Cell *&literals, Cell *&template_alist,
				  const Cell *pattern, const Cell *templ,
				  Cell *&environment,
				  Cell *&macro_environment)
{
  assert(expression != NULL);
  const Cell *sub_pattern;
  if (templ->is_symbol())
    {
      sub_pattern = templ;
    }
  else
    {
      sub_pattern = NULL;
      for (const Cell *alist = template_alist;
	   alist != Cell::nil();
	   alist = CDR(alist))
	{
	  if (CAAR(alist) == templ)
	    {
	      sub_pattern = CDAR(alist);
	      break;
	    }
	}
    }

  if (templ->is_symbol())
    {
      assert(templ != vm.core.ellipsis_symbol);

      if (!is_literal(literals, template_alist, templ))
	{
	  const Cell *sub_form = match_sub_form(vm, expression,
						pattern, sub_pattern);
	  assert(sub_form != NULL);

	  if (sub_pattern == templ)
	    return const_cast<Cell *>(sub_form);

	  if (sub_pattern->is_pair())
	    {
	      do
		{
		  if (CAR(sub_pattern) == templ)
		    return const_cast<Cell *>(sub_form);

		  sub_pattern = CDR(sub_pattern);
		  if (sub_form->is_pair())
		    sub_form = CDR(sub_form);
		}
	      while (sub_pattern->is_pair());
	    }
	  else if (sub_pattern->is_vector())
	    {
	      assert(sub_form->is_vector());

	      long len = sub_pattern->get_vector_length();
	      for (long i = 0; i < len; i++)
		{
		  if (sub_pattern->get_vector_element(i) == templ)
		    {
		      VMStackRoot vec(vm, sub_form);
		      len = vec->get_vector_length();
		      VMStackRoot result(vm, Cell::nil());
		      while (i < len)
			{
			  Cell *elm = vec->get_vector_element(i++);
			  result = vm.cons(elm, result);
			}

		      return R5RSListProcedure::list_reverse_in_place(
							vm.get_memory_system(),
							Cell::nil(), result,
							true);
		    }
		}
	    }

	  assert(sub_form != NULL);

	  return Cell::nil();
	}

      // templ isn't a pattern variable.
      // In this case, `templ' is inserted into output as a literal identifier.
      const Cell *binding = macro_environment->lookup(templ, true);
      if (binding != Cell::nil())
	return const_cast<Cell *>(binding);

      binding = macro_environment->make_unbound_variable(templ);
      return const_cast<Cell *>(binding);
    }

  const Cell *sub_form = match_sub_form(vm, expression, pattern, sub_pattern);
  if (sub_pattern == NULL)
    sub_pattern = Cell::nil();

  if (templ->is_pair())
    return transform_list(vm, sub_form, literals, template_alist,
			  sub_pattern, templ, environment, macro_environment);
  if (templ->is_vector())
    return transform_vector(vm, sub_form, literals, template_alist,
			    sub_pattern, templ,
			    environment, macro_environment);

  return const_cast<Cell *>(templ);	// templ must be a constant value!
}


Cell *
MacroSyntax::expand_macro(VirtualMachine &vm,
			  Cell *expression,
			  Cell *&literals, const Cell *syntax_rule,
			  Cell *&environment, Cell *&macro_environment)
{
  VMStackRoot template_alist(vm, CDDR(syntax_rule));
  return transform_expression(vm, expression, literals, template_alist,
			      CAR(syntax_rule), CADR(syntax_rule),
			      environment, macro_environment);
}

}
