/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: memory_system.cc,v 1.46 2009/01/01 07:47:00 fuyu Exp $
 */

#ifdef __linux__
#define _XOPEN_SOURCE 600
#endif
#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libtscore/memory_system.h"

#include "libtscore/tscore.h"

#include <algorithm>

#include <cstdio>


#if defined(HAVE_MMAP) && !defined(HAVE_POSIX_MEMALIGN)
# ifdef HAVE_SYS_MMAN_H
#  include <sys/mman.h>
# endif
# if !defined(MAP_ANON) && defined(MAP_ANONYMOUS)
#  define MAP_ANON MAP_ANONYMOUS
# endif
# if !defined(MAP_ANON) || !defined(HAVE_MUNMAP)
#  undef HAVE_MMAP
# endif
#endif


#ifdef HAVE_GETTIMEOFDAY
#include <sys/time.h>
#undef ACCUMULATE_GC_TIME
#endif


namespace LibTSCore
{

static inline Cell *
align_cell_boundary(void *pointer)
{
  if ((reinterpret_cast<intptr_t>(pointer) % sizeof(Cell)) == 0)
    return static_cast<Cell *>(pointer);	// already aligned

  return reinterpret_cast<Cell *>(sizeof(Cell)
				  * ((reinterpret_cast<intptr_t>(pointer)
				      / sizeof(Cell)) + 1));
}


MemorySystem::MemorySystem(TSCore &core,
			   size_t initial_n_cell_segments,
			   size_t maximum_n_cell_segments,
			   size_t cell_segment_size)
  : free_cells(Cell::nil()),
    n_free_cells(0),
    cell_segment_size(cell_segment_size),
    maximum_n_cell_segments(maximum_n_cell_segments),
    core(core),
    require_full_gc(false),
    verbose_gc(false),
    no_memory(false),
    total_gc_time(0.0)
{
  pagesize = 4096;
  number_of_cells_in_page = pagesize / sizeof(Cell);
  Cell::sink()->mk_free_cell(Cell::nil());

  long i;
  long test_bit = 1;
  for (i = 0; test_bit < pagesize; i++)
    if ((test_bit <<= 1) == pagesize)
      break;

  card_mark_index_shift = i + 1;

  for (i = 0; i < CARD_MARK_SIZE; i++)
    card_mark[i] = 0;

#ifdef HAVE_POSIX_MEMALIGN
  raw_young_segment = NULL;
  if (posix_memalign(&raw_young_segment, sizeof(Cell),
		     raw_segment_size() * 2) != 0)
    {
      no_memory = true;
      return;
    }

  memset(raw_young_segment, 0, raw_segment_size() * 2);
#else
# ifdef HAVE_MMAP
  raw_young_segment = mmap(0, raw_segment_size() * 2,
			   PROT_READ | PROT_WRITE,
			   MAP_PRIVATE | MAP_ANON, -1, 0);
  if (raw_young_segment == MAP_FAILED)
    {
      no_memory = true;
      return;
    }
# else
  raw_young_segment = calloc(cell_segment_size * 2, sizeof(Cell));
  if (raw_young_segment == NULL)
    {
      no_memory = true;
      return;
    }
# endif
#endif

  Cell *aligned_seg = align_cell_boundary(raw_young_segment);
  young_segments[0] = aligned_seg;
  next = aligned_seg;
  young_space = next;
  if (aligned_seg == raw_young_segment)
    {
      limit = aligned_seg + cell_segment_size;
      young_segment_limit = limit + cell_segment_size;
    }
  else
    {
      limit = aligned_seg + cell_segment_size - 1;
      young_segment_limit = limit + cell_segment_size - 1;
    }

  young_segments[1] = limit;

  allocate_cell_segments(initial_n_cell_segments);


  // This makes no_memory state only if no segment allocated.
  // Should this treat the case that allocated segments are smaller
  // than initial_n_segment as no_memory situation?
  if (free_cells == Cell::nil())
    no_memory = true;
}


MemorySystem::~MemorySystem()
{
  vector<void *>::iterator iter = raw_segments.begin();
  vector<void *>::iterator end_iter = raw_segments.end();
  while (iter != end_iter)
    {
#if defined(HAVE_POSIX_MEMALIGN) || !defined(HAVE_MMAP)
      free(*iter++);
#else
      munmap(*iter++, raw_segment_size());
#endif
    }

#if defined(HAVE_POSIX_MEMALIGN) || !defined(HAVE_MMAP)
  free(raw_young_segment);
#else
  munmap(raw_young_segment, raw_segment_size() * 2);
#endif

#ifdef ACCUMULATE_GC_TIME
  fprintf(stderr, "total gc time: %lf msec.\n", total_gc_time);
#endif
}


bool
MemorySystem::is_tenured_cell(Cell *cell) const
{
  size_t n_segs = cell_segments.size();
  for (size_t i = 0; i < n_segs; i++)
    {
      void *raw_seg = raw_segments[i];
      Cell *cell_seg = cell_segments[i];
      size_t seg_size = cell_segment_size - (raw_seg == cell_seg ? 0 : 1);
      Cell *seg_tail = cell_seg + seg_size;
      if (cell >= cell_seg && cell < seg_tail)
	return true;
    }

  return false;
}


void
MemorySystem::dump_segments_table() const
{
  vector<void *>::const_iterator iter = raw_segments.begin();
  vector<void *>::const_iterator end_iter = raw_segments.end();
  fprintf(stderr, "raw_segments: ");
  while (iter != end_iter)
    fprintf(stderr, "%p, ", *iter++);

  fprintf(stderr, "\n");

  vector<Cell *>::const_iterator iter1 = cell_segments.begin();
  vector<Cell *>::const_iterator end_iter1 = cell_segments.end();
  fprintf(stderr, "cell_segments: ");
  while (iter1 != end_iter1)
    fprintf(stderr, "%p, ", *iter1++);

  fprintf(stderr, "\n");
}


void
MemorySystem::register_static_root(Cell **root)
{
  static_roots.push_back(root);
}


void
MemorySystem::unregister_static_root(Cell **root)
{
  vector<Cell **>::iterator where
    = find(static_roots.begin(), static_roots.end(), root);
  if (where != static_roots.end())
    static_roots.erase(where);
  else
    fprintf(stderr, "Cannot unregister unknown root: %p\n", root);
}


void
MemorySystem::register_external_root(Cell **root)
{
  assert(external_roots.find(root) == external_roots.end());

  external_roots.insert(root);
}


void
MemorySystem::unregister_external_root(Cell **root)
{
  set<Cell **>::iterator where = external_roots.find(root);
  if (where != external_roots.end())
    external_roots.erase(where);
  else
    fprintf(stderr, "Unknown external root: %p\n", root);
}


void
MemorySystem::register_root_enumerator(Enumerator enumerator, void *data)
{
  enumerators.push_back(EnumeratorRecord(enumerator, data));
}


struct EnumeratorRecordEquals
{
  MemorySystem::Enumerator enumerator;
  void *data;

  EnumeratorRecordEquals(MemorySystem::Enumerator enumerator, void *data)
    : enumerator(enumerator), data(data) {}
  inline bool operator()(MemorySystem::EnumeratorRecord &record) const
  { return enumerator == record.enumerator && data == record.data; }
};


void
MemorySystem::unregister_root_enumerator(Enumerator enumerator, void *data)
{
  list<EnumeratorRecord>::iterator where;
  where = find_if(enumerators.begin(), enumerators.end(),
		  EnumeratorRecordEquals(enumerator, data));
  if (where != enumerators.end())
    enumerators.erase(where);
  else
    fprintf(stderr, "Unknown root enumerator: { %p, %p }\n", enumerator, data);
}


bool
MemorySystem::forward_vector_elements(Cell *vector,
				      const Cell *from, const Cell *from_limit)
{
  // We can treat vector in young space as it has old->young reference
  // because the result will not be used at all.
  bool exist_old_to_young_reference = is_young_cell(vector);

  long i = 0;
  long size = vector->get_vector_size();
  for (; i < size && !exist_old_to_young_reference; i++)
    {
      Cell *slot = vector + 1 + i;
      exist_old_to_young_reference
	|= is_young_cell(forward(&slot->contents.vector.elements[0],
				 from, from_limit));
      exist_old_to_young_reference
	|= is_young_cell(forward(&slot->contents.vector.elements[1],
				 from, from_limit));
      exist_old_to_young_reference
	|= is_young_cell(forward(&slot->contents.vector.elements[2],
				 from, from_limit));
      exist_old_to_young_reference
	|= is_young_cell(forward(&slot->contents.vector.elements[3],
				 from, from_limit));
    }

  for (; i < size; i++)
    {
      Cell *slot = vector + 1 + i;
      forward(&slot->contents.vector.elements[0], from, from_limit);
      forward(&slot->contents.vector.elements[1], from, from_limit);
      forward(&slot->contents.vector.elements[2], from, from_limit);
      forward(&slot->contents.vector.elements[3], from, from_limit);
    }

  return exist_old_to_young_reference;
}


void
MemorySystem::tenure(Cell *p, const Cell *from, const Cell *from_limit)
{
  if (p->is_vector())
    {
      size_t size = p->get_vector_size() + 1;
      Cell *q = get_tenured_consecutive_cells(size, p, false);
      // p passed as an argument of get_tenured_consecutive_cells should
      // have non-sense because we don't allow GC be occured here!.
      if (q != Cell::sink())
	{
	  // do copy elements.
	  memcpy(q, p, sizeof(Cell) * size);
	  p->finalize();
	  p->mk_forward_pointer(q);
	  forward(&q->car(), from, from_limit);
	  forward(&q->cdr(), from, from_limit);
	  forward_vector_elements(q, from, from_limit);
	}
    }
  else
    {
      Cell *q = get_tenured_cell0();
      if (q != Cell::sink())
	{
	  *q = *p;
	  p->mk_forward_pointer(q);

	  if (!q->is_atom())
	    {
	      forward(&q->car(), from, from_limit);
	      forward(&q->cdr(), from, from_limit);
	    }
	}
    }
  if (p->is_forward_pointer())
    return;

  require_full_gc = true;
#if 0
  fprintf(stderr, "Cannot tenure %p at this time.\n", p);
#endif
  chase(p, from, from_limit);		// Let p be alive on young space again.
}


void
MemorySystem::chase(Cell *p, const Cell *from, const Cell *from_limit)
{
  do
    {
      Cell *q = next++;
      q->finalize();
      *q = *p;		// copy
      p->mk_forward_pointer(q);
      Cell *r = NULL;
      if (q->is_vector())
	{
	  // copy elements
	  long len = q->get_vector_size();
	  for (long i = 0; i < len; i++)
	    {
	      Cell *slot = next++;
	      slot->finalize();
	      *slot = *(p + 1 + i);
	      (p + 1 + i)->mk_free_cell();
	    }
	}

      if (!q->is_atom())
	{
	  Cell *val = CAR(q);
	  if (!val->is_pseudo_cell()
	      && val >= from && val < from_limit
	      && !val->is_forward_pointer())
	    {
	      r = val;
	    }
	  else
	    {
	      val = CDR(q);
	      if (!val->is_pseudo_cell()
		  && val >= from && val < from_limit
		  && !val->is_forward_pointer())
		{
		  r = val;
		}
	    }
	}

      p = r;
    }
  while (p != NULL);
}


void
MemorySystem::forward_by_remembered_set(const Cell *from,
					const Cell *from_limit)
{
  size_t n_segs = cell_segments.size();
  for (size_t i = 0; i < n_segs; i++)
    {
      void *raw_seg = raw_segments[i];
      Cell *cell_seg = cell_segments[i];
      size_t seg_size = cell_segment_size - (raw_seg == cell_seg ? 0 : 1);
      Cell *seg_tail = cell_seg + seg_size;

      Cell *p = cell_seg;
      while (p < seg_tail)
	{
	  long idx = card_index(p); 
	  Cell *page_tail = get_tail_boundary(p);
	  page_tail = (page_tail < seg_tail) ? page_tail : seg_tail;
	  if (card_mark[idx])
	    {
	      long new_mark = 2;
	      while (p < page_tail)
		{
		  if (!p->is_atom())
		    {
		      forward(&p->car(), from, from_limit);
		      forward(&p->cdr(), from, from_limit);
		      if (is_young_cell(CAR(p)) || is_young_cell(CDR(p)))
			new_mark |= 1;
		    }

		  if (p->is_vector())
		    {
		      if (forward_vector_elements(p, from, from_limit))
			new_mark |= 1;
		      p += p->get_vector_size();
		    }

		  p++;
		}

	      card_mark[idx] = new_mark;
	    }
	  else
	    p = page_tail;
	}
    }
}


void
MemorySystem::alter_young_space()
{
  if (limit <= young_segments[1])
    next = young_segments[1];
  else
    next = young_segments[0];

  young_space = next;
  limit = next + (cell_segment_size
		  - (raw_young_segment == young_segments[0] ? 0 : 1));
}


void
MemorySystem::minor_gc(Cell *&a, Cell *&d, vector<Cell **> *dyn_roots,
		       bool finalize_now)
{
#ifdef HAVE_GETTIMEOFDAY
  struct timeval gc_start;
#endif
  if (require_full_gc)
    {
      gc(a, d, true);
      return;
    }

  if (verbose_gc)
    {
#ifdef HAVE_GETTIMEOFDAY
      gettimeofday(&gc_start, NULL);
#endif
      core.output_message("minor gc...");
      number_of_tenured_cells = 0;
    }

#if defined(HAVE_GETTIMEOFDAY) && defined(ACCUMULATE_GC_TIME)
  struct timeval gc_start_ac;
  gettimeofday(&gc_start_ac, NULL);
#endif

  Cell *from_space = young_space;
  Cell *from_limit = next;

  // We need to finalize all Cells left in current young space.
  // minor_gc() can be invoked even if there's more space left.
  while (next < limit)
    {
      next = next->finalize();
      next++;
    }

  alter_young_space();
  Cell *to_space = next;

  set<Cell **>::iterator ext_iter = external_roots.begin();
  set<Cell **>::iterator ext_end_iter = external_roots.end();
  while (ext_iter != ext_end_iter)
    forward(*ext_iter++, from_space, from_limit);

  forward_by_remembered_set(from_space, from_limit);

  vector<Cell **> dynamic_roots_local;
  if (dyn_roots == NULL)
    dyn_roots = &dynamic_roots_local;

  vector<Cell **> &dynamic_roots = *dyn_roots;
  list<EnumeratorRecord>::iterator enum_iter = enumerators.begin();
  list<EnumeratorRecord>::iterator enum_end_iter = enumerators.end();
  while (enum_iter != enum_end_iter)
    {
      EnumeratorRecord &record = *enum_iter++;
      (*record.enumerator)(dynamic_roots, record.data);
    }
    
  vector<Cell **>::iterator dyn_iter = dynamic_roots.begin();
  vector<Cell **>::iterator dyn_end_iter = dynamic_roots.end();
  while (dyn_iter != dyn_end_iter)
    forward(*dyn_iter++, from_space, from_limit);

  vector<Cell **>::iterator stack_iter = stack_roots.begin();
  vector<Cell **>::iterator stack_end_iter = stack_roots.end();
  while (stack_iter != stack_end_iter)
    forward(*stack_iter++, from_space, from_limit);

  vector<Cell **>::iterator st_iter = static_roots.begin();
  vector<Cell **>::iterator st_end_iter = static_roots.end();
  while (st_iter != st_end_iter)
    forward(*st_iter++, from_space, from_limit);

  forward(&a, from_space, from_limit);
  forward(&d, from_space, from_limit);

  Cell *scan = to_space;
  while (scan < next)
    {
      if (!scan->is_atom())
	{
	  forward(&scan->car(), from_space, from_limit);
	  forward(&scan->cdr(), from_space, from_limit);
	}

      if (scan->is_vector())
	{
	  forward_vector_elements(scan, from_space, from_limit);
	  scan += scan->get_vector_size();
	}

      ++scan;
    }

  for (long i = 0; i < CARD_MARK_SIZE; i++)
    card_mark[i] &= 1;

  if (finalize_now)
    {
      while (from_space < from_limit)
	{
	  if (!from_space->is_forward_pointer())
	    from_space = from_space->finalize();

	  from_space++;
	}
    }

#if defined(HAVE_GETTIMEOFDAY) && defined(ACCUMULATE_GC_TIME)
  struct timeval gc_end_ac;
  gettimeofday(&gc_end_ac, NULL);
  double start_msec = gc_start_ac.tv_sec * 1000.0 + gc_start_ac.tv_usec / 1000.0;
  double end_msec = gc_end_ac.tv_sec * 1000.0 + gc_end_ac.tv_usec / 1000.0;
  total_gc_time += end_msec - start_msec;
#endif

  if (verbose_gc)
    {
#ifdef HAVE_GETTIMEOFDAY
      struct timeval gc_end;
      gettimeofday(&gc_end, NULL);

      double start_msec = gc_start.tv_sec * 1000.0 + gc_start.tv_usec / 1000.0;
      double end_msec = gc_end.tv_sec * 1000.0 + gc_end.tv_usec / 1000.0;
      char msg[1024];
      snprintf(msg, 1024,
	       "done: %lu cells were reclaimed (%lu tenured) in %lf msec.\n",
	       static_cast<unsigned long>(limit - next),
	       static_cast<unsigned long>(number_of_tenured_cells),
	       end_msec - start_msec);
      core.output_message(msg);
#else
      char msg[1024];
      snprintf(msg, 1024,
	       "done: %lu cells were reclaimed (%lu tenured).\n",
	       static_cast<unsigned long>(limit - next),
	       static_cast<unsigned long>(number_of_tenured_cells));
      core.output_message(msg);
#endif
    }

  Cell::nil()->unset_gc_mark();	// needless?
  if (require_full_gc)
    {
      gc(a, d, false);
      return;
    }
}


Cell *
MemorySystem::get_cell_slow(Cell *&a, Cell *&d)
{
  if (next >= limit)
    minor_gc(a, d);

  if (next < limit)
    {
      next->finalize();
      return next++;
    }

  return get_tenured_cell(a, d);
}


Cell *
MemorySystem::get_tenured_cell_slow(Cell *&a, Cell *&d)
{
  if (no_memory)
    return Cell::sink();

  gc(a, d, false);	// minor GC should have been done before.

  if (free_cells == Cell::nil())
    {
      no_memory = true;
      return Cell::sink();
    }

  --n_free_cells;
  Cell *result = free_cells;
  free_cells = free_cells->get_cdr();
  card_mark[card_index(result)] = 3;
  return result;
}


long
MemorySystem::count_consecutive_cells(Cell *x, long length)
{
  long n = 1;
  while (CDR(x) == x + 1)
    {
      x = CDR(x);
      n++;
      if (n >= length)
	return n;
    }

  return n;
}


Cell *
MemorySystem::find_consecutive_cells(long length)
{
  Cell *cur = free_cells;
  Cell *tail = Cell::nil();
  while (cur != Cell::nil())
    {
      long count = count_consecutive_cells(cur, length);
      if (count >= length)
	{
	  // Theres at least count consecutive cells headed by cur
	  if (tail == Cell::nil())
	    free_cells = (cur + length - 1)->get_cdr();
	  else
	    tail->set_cdr((cur + length - 1)->get_cdr());

	  n_free_cells -= length;
	  return cur;
	}

      tail = cur + count - 1;
      cur = tail->get_cdr();
    }

  return Cell::nil();
}


Cell *
MemorySystem::young_space_get_consecutive_cells(long n, Cell *&fill)
{
  if (next + n >= limit)
    minor_gc(fill, fill);

  if (next + n < limit)
    {
      Cell *result = next;
      next += n;
      for (Cell *p = result; p < next; p++)
	p = p->finalize();

      return result;
    }

  return Cell::sink();		// no enough memory in young space!
}


Cell *
MemorySystem::get_tenured_consecutive_cells(long length, Cell *&fill,
					    bool allow_gc)
{
  if (no_memory)
    return Cell::sink();

  Cell *nil;
  Cell *result = find_consecutive_cells(length);
  if (result != Cell::nil())
    goto allocated;

  if (!allow_gc)
    return Cell::sink();	// out of memory, it's O.K. during young GC.

  nil = Cell::nil();
  gc(fill, nil, false);		// young GC will make old space smaller
  result = find_consecutive_cells(length);
  if (result != Cell::nil())
    goto allocated;

  if (n_free_cells >= cell_segments.size() * 32
      && free_cells != Cell::nil()
      && allocate_cell_segments() == 0)
    return Cell::sink();

  result = find_consecutive_cells(length);
  if (result != Cell::nil())
    goto allocated;

  fprintf(stderr, "Cannot allocate enough memory even though new cell segment allocated!\n");
  fprintf(stderr, "n_free_cells=%lu, requested size=%lu\n",
	  static_cast<unsigned long>(n_free_cells),
	  static_cast<unsigned long>(length));

  //no_memory = true;
  return Cell::sink();


 allocated:
  number_of_tenured_cells += length;
  Cell *tail = get_tail_boundary(result + length);
  Cell *p = result;
  do
    {
      card_mark[card_index(p)] = 3;
      p += number_of_cells_in_page;
    }
  while (p < tail);

  return result;
}


long
MemorySystem::allocate_cell_segments(long n_cell_segments)
{
  long n;
  for (n = 0; n < n_cell_segments; n++)
    {
      if (cell_segments.size() >= maximum_n_cell_segments)
	break;

#ifdef HAVE_POSIX_MEMALIGN
      void *raw_seg = NULL;
      if (posix_memalign(&raw_seg, sizeof(Cell), raw_segment_size()) != 0)
	break;

      memset(raw_seg, 0, raw_segment_size());
#else
# ifdef HAVE_MMAP
      void *raw_seg = mmap(0, raw_segment_size(), PROT_READ | PROT_WRITE,
			   MAP_PRIVATE | MAP_ANON, -1, 0);
      if (raw_seg == MAP_FAILED)
	break;
# else
      void *raw_seg = calloc(cell_segment_size, sizeof(Cell));
      if (raw_seg == NULL)
	break;
# endif
#endif

      raw_segments.push_back(raw_seg);

      Cell *cell_seg = align_cell_boundary(raw_seg);
      cell_segments.push_back(cell_seg);

      Cell *tail;
      if (raw_seg == cell_seg)
	{
	  n_free_cells += cell_segment_size;
	  tail = cell_seg + cell_segment_size;
	}
      else
	{
	  n_free_cells += cell_segment_size - 1;
	  tail = cell_seg + cell_segment_size - 1;
	}

      Cell *p;
      for (p = cell_seg; p < tail; p++)
	p->mk_free_cell(p + 1);

      tail = p - 1;
      if (free_cells == Cell::nil() || p < free_cells)
	{
	  tail->mk_free_cell(free_cells);
	  free_cells = cell_seg;
	}
      else
	{
	  p = free_cells;
	  while (CDR(p) != Cell::nil() && CDR(p) < cell_seg)
	    p = CDR(p);

	  tail->mk_free_cell(CDR(p));
	  p->mk_free_cell(tail);
	}
    }

  if (n > 0)
    {
      sort(raw_segments.begin(), raw_segments.end());
      sort(cell_segments.begin(), cell_segments.end());
    }

  return n;
}


/*
 * This mark algorithm is a straight forward implementation of
 * the Deutsch-Schorr-Waite pointer reversal algorithm except for
 * treatment of 'Vector'.
 *
 * The original TinyScheme said that they also use D-S-W algorithm
 * based on Knuth's book.  We reimplement this algorithm based
 * on `Garbage Collection' by Richard Jones and Rafael Lins.
 *
 * See Algorithm 4.3, page 84.
 */
void
MemorySystem::mark(Cell *&r)
{
  if (r->is_pseudo_cell())
    return;	// pseudo cell never affected by any GC actions.

  Cell *current = r;
  Cell *previous = Cell::nil();

  while (true)
    {
      // -- follow left pointers
      while (current != Cell::nil() && !current->is_gc_marked())
	{
	  current->set_gc_mark();
	  if (current->is_vector())
	    {
	      // For vector, we don't care about depth of recursion.
	      // The possible depth of recursion is bound by the length
	      // of vector->vector chain, i.e. if a vector is stored
	      // within a vector as an element, it increments the recursion
	      // depth by 1.  We believe it isn't a problem for real world
	      // application at all because use of vector in vector in
	      // vector in... vector is unusual.  Please note,
	      // cycles of vector in vector relation will never cause
	      // unlimited recursion since we set gc mark at first.
	      for (long i = 0, len = current->get_vector_size(); i < len; i++)
		{
		  Cell *slot = current + i + 1;
		  mark(slot->contents.vector.elements[0]);
		  mark(slot->contents.vector.elements[1]);
		  mark(slot->contents.vector.elements[2]);
		  mark(slot->contents.vector.elements[3]);
		}
	    }
	  if (!current->is_atom())
	    {
	      Cell *next = CAR(current);
	      current->set_car(previous);
	      previous = current;
	      current = next;

	      assert(current != NULL);
	    }
	}

      // -- retreat
      while (previous != Cell::nil() && previous->is_gc_flag_set())
	{
	  previous->unset_gc_flag();
	  Cell *next = CDR(previous);
	  previous->set_cdr(current);
	  current = previous;
	  previous = next;

	  assert(current != NULL);
	}

      if (previous != Cell::nil())
	{
	  // switch to right subgraph
	  previous->set_gc_flag();
	  Cell *next = CAR(previous);
	  previous->set_car(current);
	  current = CDR(previous);
	  previous->set_cdr(next);

	  assert(current != NULL);
	}
      else
	return;	// done;
    }
}


// This is a Mark and Sweep garbage collector.
void
MemorySystem::gc(Cell *&a, Cell *&d, bool with_minor_gc)
{
#ifdef HAVE_GETTIMEOFDAY
  struct timeval gc_start;
#endif
  require_full_gc = false;
  vector<Cell **> dynamic_roots;
  if (with_minor_gc)
    minor_gc(a, d, &dynamic_roots, true);

  require_full_gc = false;

  if (verbose_gc)
    {
#ifdef HAVE_GETTIMEOFDAY
      gettimeofday(&gc_start, NULL);
#endif
      core.output_message("gc...");
    }

#if defined(HAVE_GETTIMEOFDAY) && defined(ACCUMULATE_GC_TIME)
  struct timeval gc_start_ac;
  gettimeofday(&gc_start_ac, NULL);
#endif

  // mark from static roots
  vector<Cell **>::iterator iter = static_roots.begin();
  vector<Cell **>::iterator end_iter = static_roots.end();
  while (iter != end_iter)
    mark(**iter++);

  // mark from stack roots
  vector<Cell **>::iterator stack_iter = stack_roots.begin();
  vector<Cell **>::iterator stack_end_iter = stack_roots.end();
  while (stack_iter != stack_end_iter)
    mark(**stack_iter++);

  // mark from external roots
  set<Cell **>::iterator ext_iter = external_roots.begin();
  set<Cell **>::iterator ext_end_iter = external_roots.end();
  while (ext_iter != ext_end_iter)
    mark(**ext_iter++);

  // mark from young space
  Cell *cur = young_space;
  while (cur < next)
    {
      mark(cur);
      if (cur->is_vector())
	cur += cur->get_vector_size();

      ++cur;
    }

  // When minor GC performed just before major GC, dynamic_roots
  // already enumerated at minor GC, and thus, we don't need to
  // enumerate them again here.
  if (!with_minor_gc)
    {
      list<EnumeratorRecord>::iterator enum_iter = enumerators.begin();
      list<EnumeratorRecord>::iterator enum_end_iter = enumerators.end();
      while (enum_iter != enum_end_iter)
	{
	  EnumeratorRecord &record = *enum_iter++;
	  (*record.enumerator)(dynamic_roots, record.data);
	}
    }

  // mark from interpreter's roots
  vector<Cell **>::iterator dyn_iter = dynamic_roots.begin();
  vector<Cell **>::iterator dyn_end_iter = dynamic_roots.end();
  while (dyn_iter != dyn_end_iter)
    mark(**dyn_iter++);

  mark(a);
  mark(d);


  // collect garbages
  Cell::nil()->unset_gc_mark();	// needless?
  n_free_cells = 0;
  free_cells = Cell::nil();

  /*
   * free_cells is a list of free Cells kept sorted by address order
   * so as to maintain consecutive cells in an easy manner for vectors.
   * Here we scan cells, which are also kept sorted by address order,
   * downwards to build the fresh free_cells.
   */
  Cell *free_cell_tail = Cell::nil();
  for (size_t i = 0, size = cell_segments.size(); i < size; i++)
    {
      void *raw_seg = raw_segments[i];
      Cell *cell_seg = cell_segments[i];
      Cell *cell_seg_limit
	= cell_seg + cell_segment_size - (raw_seg == cell_seg ? 0 : 1);
      Cell *p = cell_seg;
      if (free_cell_tail != Cell::nil())
	while (p <= free_cell_tail)
	  p++;

      while (p < cell_seg_limit)
	{
	  if (p->is_gc_marked())
	    {
	      p->unset_gc_mark();		// live cell
	      if (p->is_vector())
		p += p->get_vector_size();

	      ++p;
	    }
	  else
	    {
	      Cell *p_tail = p->finalize();
	      do
		{
		  if (free_cells != Cell::nil())
		    free_cell_tail->set_cdr(p);
		  else
		    free_cells = p;

		  free_cell_tail = p++;
		  ++n_free_cells;
		}
	      while (p < p_tail);
	    }
	}
    }

  free_cell_tail->mk_free_cell(Cell::nil());

  // unmark Cells in young space
  cur = young_space;
  while (cur < next)
    {
      cur->unset_gc_mark();
      cur++;
    }

  if (with_minor_gc)
    {
      // finalize Cells left in young space to be allocated soon.
      while (cur < limit)
	{
	  cur = cur->finalize();
	  cur++;
	}
    }

  if (n_free_cells < cell_segments.size() * 32 || free_cells == Cell::nil())
    {
      if (allocate_cell_segments() > 0)
	{
	  if (verbose_gc)
	    core.output_message("old space expanded...");
	}
      else
	{
	  if (verbose_gc)
	    fprintf(stderr, "maximum number of cell segments exceeded.\n");
	}
    }

#if defined(HAVE_GETTIMEOFDAY) && defined(ACCUMULATE_GC_TIME)
  struct timeval gc_end_ac;
  gettimeofday(&gc_end_ac, NULL);
  double start_msec
    = gc_start_ac.tv_sec * 1000.0 + gc_start_ac.tv_usec / 1000.0;
  double end_msec = gc_end_ac.tv_sec * 1000.0 + gc_end_ac.tv_usec / 1000.0;
  total_gc_time += end_msec - start_msec;
#endif

  if (verbose_gc)
    {
#ifdef HAVE_GETTIMEOFDAY
      struct timeval gc_end;
      gettimeofday(&gc_end, NULL);

      double start_msec = gc_start.tv_sec * 1000.0 + gc_start.tv_usec / 1000.0;
      double end_msec = gc_end.tv_sec * 1000.0 + gc_end.tv_usec / 1000.0;
      char msg[1024];
      snprintf(msg, 1024,
	       "done: %lu cells were recovered in %lf msec.\n",
	       static_cast<unsigned long>(n_free_cells), end_msec - start_msec);
      core.output_message(msg);
#else
      char msg[1024];
      snprintf(msg, 1024,
	       "done: %lu cells were recovered.\n",
	       static_cast<unsigned long>(n_free_cells));
      core.output_message(msg);
#endif
    }

  if (n_free_cells > 0)
    no_memory = false;
}


Cell *
MemorySystem::get_vector_slow(size_t len, Cell *&fill)
{
  if (no_memory)
    {
      fprintf(stderr, "Cannot allocate vector: len=%lu\n",
	      static_cast<unsigned long>(len));
      return Cell::sink();
    }

  size_t size = Cell::get_vector_size(len) + 1;
  Cell *t = get_tenured_consecutive_cells(size, fill);
  if (t != Cell::sink())
    return t->mk_vector(len, fill);

  if (size >= (cell_segment_size / 10))
    {
      // try to allocate from young space.
      t = young_space_get_consecutive_cells(size, fill);
    }

  if (t != Cell::sink())
    return t->mk_vector(len, fill);

  fprintf(stderr, "Cannot allocate vector: len=%lu\n",
	  static_cast<unsigned long>(len));
  return Cell::sink();
}


Cell *
MemorySystem::copy_vector(size_t len, Cell *&src_vector, Cell *&fill)
{
  StackRoot src(*this, src_vector);
  Cell *vector = get_vector(len, fill);
  size_t src_len = src->get_vector_length();
  for (size_t i = 0, copy_len = src_len < len ? src_len : len;
       i < copy_len; i++)
    vector->set_vector_element(i, src->get_vector_element(i));

  // We don't need to worry about write barrier for newly allocated
  // vector.  Even though it is pre-tenured, the card_mark[] already marked.
  return vector;
}


static char *
allocate_string(size_t len, const char *str, char fill)
{
  char *storage = static_cast<char *>(malloc(len + 1));
  if (storage == NULL)
    return NULL;

  if (str != NULL)
    strncpy(storage, str, len);
  else
    memset(storage, fill, len);

  storage[len] = '\0';
  return storage;
}


Cell *
MemorySystem::get_counted_string(const char *str, size_t len)
{
  Cell *x = get_cell();
  char *storage = allocate_string(len, str, '\0');
  return storage != NULL ? x->mk_string(storage, len) : Cell::sink();
}


Cell *
MemorySystem::get_empty_string(size_t len, char fill)
{
  Cell *x = get_cell();
  char *storage = allocate_string(len, NULL, fill);
  return storage != NULL ? x->mk_string(storage, len) : Cell::sink();
}


Cell *
MemorySystem::get_substring_copy(Cell *&src, long start, long end)
{
  assert(is_valid_pointer(src));
  assert(src->is_string());

  Cell *offset = Cell::mk_small_integer(start);
  if (offset == NULL)
    offset = get_cell(src)->mk_integer(start);

  long len_val = end - start;
  Cell *len = Cell::mk_small_integer(len_val);
  if (len == NULL)
    len = get_cell(src, offset)->mk_integer(len_val);

  StackRoot str(*this, src);
  StackRoot result(*this, get_cell(offset, len));
  return result->mk_substring_copy(str, offset, len);
}


Cell *
MemorySystem::get_substring_reference(Cell *&src, long start, long end)
{
  assert(is_valid_pointer(src));
  assert(src->is_string());

  Cell *offset = Cell::mk_small_integer(start);
  if (offset == NULL)
    offset = get_cell(src)->mk_integer(start);

  long len_val = end - start;
  Cell *len = Cell::mk_small_integer(len_val);
  if (len == NULL)
    len = get_cell(src, offset)->mk_integer(len_val);

  StackRoot str(*this, src);
  StackRoot result(*this, get_cell(offset, len));
  return result->mk_substring_reference(str, offset, len);
}

}
