/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: memory_system.h,v 1.38 2009/01/01 07:47:00 fuyu Exp $
 */

#ifndef _LIBTSCORE_MEMORY_SYSTEM_H_
#define _LIBTSCORE_MEMORY_SYSTEM_H_

#ifdef __cplusplus

#if defined(__GNUG__) && !defined(__ICC)
#pragma interface
#endif

#include "libtscore/cell.h"

#include <algorithm>
#include <list>
#include <set>
#include <vector>

#include <cstring>


using namespace std;
namespace LibTSCore
{

  class TSCore;

  /*
   * MemorySystem represents garbage collected heaps.
   * This currently employs generational GC.  The young generation GC
   * is a copying GC using mark card table by simple write barriers.
   * The tenured generation GC is ordinal mark and sweep GC.
   *
   * Since we use copying GC for young generation, using this
   * memory system requires quite a lot of care.  Please do not store
   * pointers to Cell within the heap managed by this class without
   * special concern.  Otherwise, your pointer to Cell will soon
   * be invalidated due to the GC!
   *
   * Don't worry too much!  We try to provide some helper class to
   * let pointers be up to date.  And the most important thing is,
   * users shouldn't need to touch this directly.  Depend on TSCore
   * instead of its building blocks such as this class.
   */
  class MemorySystem
  {
    friend class EnumeratorRecordEquals;
    friend class TSCore;
    friend class StackRoot;
    friend class Reader;
    friend class Register;

  private:
    // young generation heap
    Cell *young_space;			// pointer to start of current space
    Cell *next;				// pointer to next allocation
    Cell *limit;			// pointer to current space boundary

    Cell *young_segments[2];		// pointers to aligned young space
    Cell *young_segment_limit;		// pointer to young space boundary


    // old generation heap
    Cell *free_cells;
    size_t n_free_cells;

    vector<Cell *> cell_segments;	// pointers to aligned old space


    /*
     * The `static_roots' remember the root-sets that may be considered
     * long-living, i.e. registered in early stage of applications using
     * TSCore, usually at initialization time, and unregistered when the
     * application preparing to quit.
    */
    vector<Cell **> static_roots;

    /*
     * The `stack_roots' remember the root-sets that may be considered
     * short-living and/or at least the pattern of
     * registering/unregistering having good shape, i.e. in FILO order
     * like stack allocated objects in C++.
     * This field is mutated only by `StackRoot' defined below.
     */
    vector<Cell **> stack_roots;

    /*
     * The `external_roots' remember the root-sets that may be
     * registered/unregistered in an random manner.
     */
    set<Cell **> external_roots;


    typedef void (*Enumerator)(vector<Cell **> &dynamic_roots,
			       void *data);
    struct EnumeratorRecord
    {
      Enumerator enumerator;
      void *data;
      EnumeratorRecord(Enumerator enumerator, void *data)
	: enumerator(enumerator), data(data) {}
    };
    list<EnumeratorRecord> enumerators;

    void *raw_young_segment;		// raw pointers to young space
    vector<void *> raw_segments;	// raw pointers to old space

    static const long CARD_MARK_INDEX_MASK = 0xff;
    static const long CARD_MARK_SIZE = 0x100;

    long card_mark_index_shift;
    char card_mark[CARD_MARK_SIZE];
    long pagesize;
    long number_of_cells_in_page;

    size_t cell_segment_size;		// # of cells for each cell segment
    size_t maximum_n_cell_segments;	// maximum # of cell segments

    TSCore &core;

    bool require_full_gc;		// shows next GC should be full GC
    bool verbose_gc;
    bool no_memory;

    size_t number_of_tenured_cells;	// during last minor GC
    double total_gc_time;

    inline size_t raw_segment_size() const
    {
      return sizeof(Cell) * cell_segment_size;
    }

    void dump_segments_table() const;	// only for debugging

    void alter_young_space();

    inline Cell *forward(Cell **p, const Cell *from, const Cell *from_limit)
    {
      Cell *cell = *p;
      if (cell->is_pseudo_cell() || cell < from || cell >= from_limit)
	return cell;

      if (!cell->is_forward_pointer())
	{
	  if (cell->is_survived_cell())
	    tenure(cell, from, from_limit);
	  else
	    chase(cell, from, from_limit);
	}
      *p = cell->get_forward_pointer();

      assert(*p);

      return *p;
    }

    bool forward_vector_elements(Cell *vector,
				 const Cell *from, const Cell *from_limit);
    void tenure(Cell *p, const Cell *from, const Cell *from_limit);
    void chase(Cell *p, const Cell *from, const Cell *from_limit);
    void forward_by_remembered_set(const Cell *from, const Cell *from_limit);

    bool is_tenured_cell(Cell *cell) const;
    inline bool is_young_cell(Cell *cell) const
    {
      return cell >= young_space && cell < limit;
    }

    inline long is_constant_cell(Cell *cell) const
    {
      return cell == Cell::nil() || cell == Cell::t() || cell == Cell::f()
	|| cell == Cell::eof() || cell == Cell::unspecified()
	|| cell == Cell::sink();
    }

    inline bool is_valid_pointer(Cell *cell) const
    {
      if (!cell->is_pseudo_cell() && !is_constant_cell(cell)
	  && !is_young_cell(cell) && !is_tenured_cell(cell))
	{
	  fprintf(stderr, "cell@%p points to invalid young space.\n", cell);
	  return false;
	}

      return true;
    }

    inline long card_index(Cell *p)
    {
      return (reinterpret_cast<long>(p) >> card_mark_index_shift)
	& CARD_MARK_INDEX_MASK;
    }

    void minor_gc(Cell *&a, Cell *&d, vector<Cell **> *dyn_roots = NULL,
		  bool finalize_now = false);

    Cell *get_cell_slow(Cell *&a, Cell *&d);

    inline Cell *get_tenured_cell0()
    {
      if (free_cells != Cell::nil())
	{
	  ++number_of_tenured_cells;
	  --n_free_cells;
	  Cell *cell = free_cells;
	  free_cells = free_cells->get_cdr();
	  card_mark[card_index(cell)] = 3;
	  return cell;
	}

      return Cell::sink();
    }

    Cell *get_tenured_cell_slow(Cell *&a, Cell *&d);

    long count_consecutive_cells(Cell *x, long required);
    Cell *find_consecutive_cells(long required);
    Cell *young_space_get_consecutive_cells(long n, Cell *&fill);


    inline Cell *get_head_boundary(Cell *cell)
    {
      return reinterpret_cast<Cell *>(reinterpret_cast<intptr_t>(cell)
				      & ~(pagesize - 1));
    }

    inline Cell *get_tail_boundary(Cell *cell)
    {
      return reinterpret_cast<Cell *>(reinterpret_cast<intptr_t>(cell)
				      & ~(pagesize - 1))
	+ number_of_cells_in_page;
    }

    // Returns a Cell * that precedes at least n consecutive cells.
    Cell *get_tenured_consecutive_cells(long n, Cell *&fill,
					bool allow_gc = true);


    // Mark all Cells reachable from given Cell as live for garbage collection.
    void mark(Cell *&a);

    // Do garbage collection.  `a' and `d' will be treated as live cells.
    void gc(Cell *&a, Cell *&d, bool with_minor_gc);

    // Try to allocate another n_cell_segments cell_segments.
    long allocate_cell_segments(long n_cell_segments = 1);


  public:
    static const long INITIAL_N_CELL_SEGMENTS = 2;
    static const long MAXIMUM_N_CELL_SEGMENTS = 8;

    // This makes each cell segment size approximately 32KB on 32bit systems.
    // static const size_t DEFAULT_CELL_SEGMENT_SIZE = 2048;

    // This makes each cell segment size approximately 64KB on 32bit systems.
    // static const size_t DEFAULT_CELL_SEGMENT_SIZE = 4096;

    // This makes each cell segment size approximately 128KB on 32bit systems.
    static const size_t DEFAULT_CELL_SEGMENT_SIZE = 8192;


    /*
     * The constructor allocates 4 segments each has cell_segment_size
     * at first.  2 segments are for younger generation heap, and another
     * 2 segments are for tenured generation.
     */
    MemorySystem(TSCore &core,
		 size_t initial_n_cell_segments = INITIAL_N_CELL_SEGMENTS,
		 size_t maximum_n_cell_segments = MAXIMUM_N_CELL_SEGMENTS,
		 size_t cell_segment_size = DEFAULT_CELL_SEGMENT_SIZE);
    ~MemorySystem();

    void register_static_root(Cell **root);
    void unregister_static_root(Cell **root);
    void register_external_root(Cell **root);
    void unregister_external_root(Cell **root);

    void register_root_enumerator(Enumerator enumerator, void *data);
    void unregister_root_enumerator(Enumerator enumerator, void *data);

    inline bool out_of_memory() const { return no_memory; }
    inline bool get_verbose_gc() const { return verbose_gc; }
    inline void set_verbose_gc(bool value) { verbose_gc = value; }


    /*
     * Marks card table for given cell as it has a pointer to young
     * generation heap.
     *
     * We don't check neither `cell' that will have a pointer stored
     * is actually in old space nor a pointer to be stored in `cell'
     * is actually in young space.  Although such checks can help to
     * reduce false marking that make minor GC a little expensive,
     * they also take another cost for each writing.
     *
     * To pay costs for each writing seems more expensive than
     * to accept a little more cost for each minor GC with our current
     * GC algorithm.
     */
    inline void write_barrier(Cell *cell)
    {
      assert(is_valid_pointer(cell));

      card_mark[card_index(cell)] = 1;
      return;
    }

    inline Cell *set_car(Cell *cell, Cell *car)
    {
      write_barrier(cell);
      cell->set_car(car);
      return cell;
    }

    inline Cell *set_cdr(Cell *cell, Cell *cdr)
    {
      write_barrier(cell);
      cell->set_cdr(cdr);
      return cell;
    }

    inline Cell *get_cons(Cell *&car = Cell::nil0(),
			  Cell *&cdr = Cell::nil0(), bool immutable = false)
    {
      Cell *cell = get_cell(car, cdr);
      return cell->mk_pair(car, cdr, immutable);
    }

    inline Cell *set_vector_element(Cell *vec, long index, Cell *elm)
    {
      write_barrier(vec);
      vec->set_vector_element(index, elm);
      return vec;
    }

    // Return pointer to a free Cell.
    inline Cell *get_cell(Cell *&a = Cell::nil0(), Cell *&d = Cell::nil0())
    {
      assert(is_valid_pointer(a));
      assert(is_valid_pointer(d));

      if (next < limit)
	{
	  next->finalize();
	  return next++;
	}

      return get_cell_slow(a, d);
    }

    inline Cell *get_vector(size_t len, Cell *&fill = Cell::nil0())
    {
      size_t size = Cell::get_vector_size(len) + 1;
      if (size < (cell_segment_size / 10))
	{
	  Cell *y = young_space_get_consecutive_cells(size, fill);
	  if (y != Cell::sink())
	    {
	      assert(is_valid_pointer(y));
	      assert(is_valid_pointer(fill));

	      return y->mk_vector(len, fill);
	    }
	}

      return get_vector_slow(len, fill);
    }

    Cell *get_vector_slow(size_t len, Cell *&fill = Cell::nil0());
    Cell *copy_vector(size_t len, Cell *&src, Cell *&fill = Cell::nil0());

    inline Cell *get_vector_from_young_space(size_t len,
					     Cell *&fill = Cell::nil0())
    {
      size_t size = Cell::get_vector_size(len) + 1;
      Cell *result = young_space_get_consecutive_cells(size, fill);
      if (result != Cell::sink())
	{
	  assert(is_valid_pointer(result));
	  assert(is_valid_pointer(fill));

	  result->mk_vector(len, fill);
	}

      return result;
    }

    // Do full garbage collection.  `a' and `d' will be treated as live cells.
    inline void gc(Cell *&a = Cell::nil0(), Cell *&d = Cell::nil0())
    {
      gc(a, d, true);
    }

    // Return a free Cell from old space.
    // `a' and `d' will be treated as live cells when GC occurs.
    // inlining fast case may be good.
    inline Cell *get_tenured_cell(Cell *&a = Cell::nil0(),
				  Cell *&d = Cell::nil0())
    {
      Cell *cell = get_tenured_cell0();
      if (cell != Cell::nil())
	return cell;

      return get_tenured_cell_slow(a, d);	// Find free cell after GC
    }

    Cell *get_counted_string(const char *str, size_t len);
    Cell *get_empty_string(size_t len, char fill);
    Cell *get_substring_copy(Cell *&src, long start, long end);
    Cell *get_substring_reference(Cell *&src, long start, long end);
    inline Cell *get_static_string(const char *str)
    {
      return get_cell()->mk_static_string(str, strlen(str));
    }

  };


  class StackRoot
  {
  private:
    MemorySystem &memory_system;
    Cell *pointer;


  public:
    /*
     * Do NOT instantiate via new StackRoot(...)!
     *
     * StackRoot, associated MemorySystem, is designed to be constructed
     * and destructed in the Last In First Out order, StackRoots allocated
     * on the stack, i.e. as local objects, will automatically behave in
     * such manner.
     */
    inline StackRoot(MemorySystem &ms, Cell *cell = Cell::nil())
      : memory_system(ms),
	pointer(cell)
    {
      memory_system.stack_roots.push_back(&pointer);
    }

    inline ~StackRoot()
    {
      /*
       * According to Programming Language C++ 3rd edition,
       * objects allocated on stack at the same nesting level
       * must be destructed in the reverse order of their construction.
       * Therefore, we actually don't need to worry about
       * where's the pointer placed on stack_roots vector,
       * if StackRoot is used as designed.
       */
      assert(memory_system.stack_roots.back() == &pointer);

      memory_system.stack_roots.pop_back();	// should be OK.
    }

    inline operator Cell *&() { return pointer; }
    inline Cell *&operator->() { return pointer; }
    inline StackRoot &operator=(Cell *cell) { pointer = cell; return *this; }
    inline StackRoot &operator=(const Cell *cell)
    {
      pointer = const_cast<Cell *>(cell);
      return *this;
    }

    inline StackRoot &operator=(const StackRoot &root)
    {
      pointer = root.pointer;
      return *this;
    }
  };

}


#define _MEMORY_SYSTEM_DEFINED_	1

#endif	/* __cplusplus */

#endif	/* _LIBTSCORE_MEMORY_SYSTEM_H_ */
