/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: number.h,v 1.19 2009/01/01 07:47:00 fuyu Exp $
 */

#ifndef _LIBTSCORE_NUMBER_H_
#define _LIBTSCORE_NUMBER_H_

#ifdef __cplusplus

#if defined(__GNUG__) && !defined(__ICC)
#pragma interface
#endif

#include <stdlib.h>


using namespace std;
namespace LibTSCore
{

  class Cell;

  /*
   * Number represents number tower of Scheme.
   *
   * This implementation is still far away to conform to the complete set
   * of number tower defined in R5RS.  We also have a plan to provide
   * more complete set of nubmer tower, but the motivation and priority
   * is quite low for our problem domain.
   */
  class Number
  {
  private:
    union
    {
      long ivalue;
      double rvalue;
    } value;

    bool fixnum;


  public:
    inline Number() {}
    inline Number(long val) { value.ivalue = val; fixnum = true; }
    inline Number(double val) { value.rvalue = val; fixnum = false; }
    inline Number(Number &val)
    {
      if (val.fixnum)
	{
	  value.ivalue = val.value.ivalue;
	  fixnum = true;
	}
      else
	{
	  value.rvalue = val.value.rvalue;
	  fixnum = false;
	}
    }

    Number(Cell *cell);

    inline void set_value(Number val)
    {
      if (val.fixnum)
	{
	  value.ivalue = val.value.ivalue;
	  fixnum = true;
	}
      else
	{
	  value.rvalue = val.value.rvalue;
	  fixnum = false;
	}
    }

    inline bool is_fixnum() const { return fixnum; }

    inline void set_ivalue(long val)
    {
      value.ivalue = val;
      fixnum = true;
    }

    inline void set_ivalue0(long val) { value.ivalue = val; }

    inline void set_rvalue(double val)
    {
      value.rvalue = val;
      fixnum = false;
    }

    inline void set_rvalue0(double val) { value.rvalue = val; }

    inline long get_ivalue() const
    {
      return fixnum ? value.ivalue : static_cast<long>(value.rvalue);
    }

    inline double get_rvalue() const
    {
      return fixnum ? static_cast<double>(value.ivalue) : value.rvalue;
    }

    inline long get_ivalue0() const { return value.ivalue; }
    inline double get_rvalue0() const { return value.rvalue; }

    inline void add(Number &val)
    {
      bool new_fixnum = fixnum & val.fixnum;
      if (new_fixnum)
	value.ivalue += val.value.ivalue;
      else
	value.rvalue = get_rvalue() + val.get_rvalue();

      fixnum = new_fixnum;
    }

    inline void mul(Number &val)
    {
      bool new_fixnum = fixnum & val.fixnum;
      if (new_fixnum)
	value.ivalue *= val.value.ivalue;
      else
	value.rvalue = get_rvalue() * val.get_rvalue();

      fixnum = new_fixnum;
    }

    inline void div(Number &val)
    {
      bool new_fixnum = fixnum && val.fixnum && val.value.ivalue != 0;
      if (new_fixnum)
	value.ivalue /= val.value.ivalue;
      else
	value.rvalue = get_rvalue() / val.get_rvalue();

      fixnum = new_fixnum;
    }

    inline void intdiv(Number &val)
    {
      bool new_fixnum = fixnum & val.fixnum;
      if (new_fixnum)
	value.ivalue /= val.value.ivalue;
      else
	value.rvalue = get_rvalue() / val.get_rvalue();

      fixnum = new_fixnum;
    }

    inline void sub(Number &val)
    {
      bool new_fixnum = fixnum & val.fixnum;
      if (new_fixnum)
	value.ivalue -= val.value.ivalue;
      else
	value.rvalue = get_rvalue() - val.get_rvalue();

      fixnum = new_fixnum;
    }

    inline void rem(Number &val)
    {
      bool new_fixnum = fixnum & val.fixnum;
      long e1 = get_ivalue();
      long e2 = val.get_ivalue();
      long res = e1 % e2;
      if (res * e1 < 0)
	{
	  e2 = labs(e2);
	  if (res > 0)
	    res -= e2;
	  else
	    res += e2;
	}

      if (new_fixnum)
	value.ivalue = res;
      else
	value.rvalue = static_cast<double>(res);

      fixnum = new_fixnum;
    }

    inline void mod(Number &val)
    {
      bool new_fixnum = fixnum & val.fixnum;
      long e1 = get_ivalue();
      long e2 = val.get_ivalue();
      long res = e1 % e2;
      if (res * e2 < 0)
	{
	  e2 = labs(e2);
	  if (res > 0)
	    res -= e2;
	  else
	    res += e2;
	}

      if (new_fixnum)
	value.ivalue = res;
      else
	value.rvalue = static_cast<double>(res);

      fixnum = new_fixnum;
    }

    inline bool eq(Number &val)
    {
      if (fixnum && val.fixnum)
	return value.ivalue == val.value.ivalue;

      return get_rvalue() == val.get_rvalue();
    }

    inline bool gt(Number &val)
    {
      if (fixnum && val.fixnum)
	return value.ivalue > val.value.ivalue;

      return get_rvalue() > val.get_rvalue();
    }

    inline bool lt(Number &val)
    {
      if (fixnum && val.fixnum)
	return value.ivalue < val.value.ivalue;

      return get_rvalue() < val.get_rvalue();
    }

    inline bool ge(Number &val) { return !lt(val); }
    inline bool le(Number &val) { return !gt(val); }

    Number &operator=(Cell *cell);
    void add(Cell *cell);
    void mul(Cell *cell);
    void div(Cell *cell);
    void intdiv(Cell *cell);
    void sub(Cell *cell);
    void rem(Cell *cell);
    void mod(Cell *cell);
    bool eq(Cell *cell);
    bool gt(Cell *cell);
    bool lt(Cell *cell);
    inline bool ge(Cell *cell) { return !lt(cell); }
    inline bool le(Cell *cell) { return !gt(cell); }
  };


}


#endif	/* __cplusplus */

#endif	/* _LIBTSCORE_NUMBER_H_ */
