/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: port.cc,v 1.22 2009/01/01 07:47:00 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libtscore/port.h"

#include "libtscore/cell.h"

#include <cstdio>
#include <cstdlib>
#include <iostream>

#include <sys/poll.h>


using namespace std;
namespace LibTSCore
{


bool
Port::char_ready()
{
  return false;
}


long
Port::peek_char()
{
  return EOF;
}


long
Port::get_char()
{
  return EOF;
}


Port *
Port::unget_char(long c)
{
  return this;
}


const char *
Port::get_remaining_string() const
{
  return "";
}


size_t
Port::get_remaining_strlen() const
{
  return 0;
}


void
Port::flush_remaining_input()
{
  return;
}


Port *
Port::append(long c)
{
  return this;
}


Port *
Port::append(const char *str)
{
  return this;
}


Port *
Port::append(const Cell *cell)
{
  return cell->to_string(this);
}


Port *
Port::append(const char *str, size_t len)
{
  return this;
}


Port *
Port::flush()
{
  return this;
}


const char *
Port::get_output_string() const
{
  return "";
}


size_t
Port::get_output_strlen() const
{
  return 0;
}


void
Port::set_transcript_port(TranscriptPort *port)
{
}


InputFilePort::InputFilePort(FILE *fin, bool closeit)
  : Port(PORT_FILE | PORT_INPUT),
    file(fin), transcript_port(NULL), closeit(closeit)
{
  fd = fileno(fin);
}


InputFilePort::~InputFilePort()
{
  if (file != NULL && closeit)
    fclose(file);
}


void
InputFilePort::close()
{
  if (file != NULL && closeit)
    fclose(file);

  file = NULL;
  port_type = PORT_FREE;
}


bool
InputFilePort::char_ready()
{
  /*
   * Note: This actually doesn't work as expected because of buffering
   *       performed by stdio!
   */
  if (file == NULL || feof(file))
    return true;	// return true on the EOF!

  struct pollfd ufd = { fd, POLLIN, 0 };
  return poll(&ufd, 1, 0) > 0;
}


void
InputFilePort::flush_remaining_input()
{
  if (file == NULL || feof(file))
    return;

  fseek(file, 0, SEEK_END);	// why doesn't this work as expected?
  return;
}


long
InputFilePort::peek_char()
{
  if (file == NULL)
    return EOF;

  long c = fgetc(file);
  if (c != EOF)
    ungetc(c, file);

  return c;
}


long
InputFilePort::get_char()
{
  if (file == NULL)
    return EOF;

  long c = static_cast<long>(fgetc(file));
  if (transcript_port == NULL)
    return c;

  transcript_port->append(c);
  return c;
}


Port *
InputFilePort::unget_char(long c)
{
  if (c != EOF && file != NULL)
    ungetc(c, file);

  if (transcript_port != NULL)
    transcript_port->unput_char(c);

 return this;
}


void
InputFilePort::set_transcript_port(TranscriptPort *port)
{
  transcript_port = port;
}


OutputFilePort::OutputFilePort(FILE *fout, bool closeit)
  : Port(PORT_FILE | PORT_OUTPUT),
    file(fout), transcript_port(NULL), closeit(closeit)
{
}


OutputFilePort::~OutputFilePort()
{
  if (file != NULL && closeit)
    fclose(file);
}


void
OutputFilePort::close()
{
  if (file != NULL && closeit)
    fclose(file);

  file = NULL;
  port_type = PORT_FREE;
}


Port *
OutputFilePort::append(long c)
{
  if (c != EOF && file != NULL)
    {
      fputc(c, file);
      if (c == '\n')
	fflush(file);
    }

  if (transcript_port != NULL)
    transcript_port->append(c);

  return this;
}


Port *
OutputFilePort::append(const char *str)
{
  if (file != NULL)
    {
      fputs(str, file);
      if (str[strlen(str) - 1] == '\n')
	fflush(file);
    }

  if (transcript_port != NULL)
    transcript_port->append(str);

  return this;
}


Port *
OutputFilePort::append(const char *str, size_t len)
{
  if (len == 0)
    return this;

  if (file != NULL)
    {
      fwrite(str, 1, len, file);
      if (str[len - 1] == '\n')
	fflush(file);
    }

  if (transcript_port != NULL)
    transcript_port->append(str, len);

  return this;
}


Port *
OutputFilePort::flush()
{
  if (file != NULL)
    fflush(file);

  if (transcript_port != NULL)
    transcript_port->flush();

  return this;
}


void
OutputFilePort::set_transcript_port(TranscriptPort *port)
{
  transcript_port = port;
}


InputStringPort::InputStringPort(const char *str, size_t len, bool copy)
  : Port(PORT_STRING | PORT_INPUT | (copy ? PORT_OWN_STRING : 0))
{
  if (len == 0 && str != NULL)
    len = strlen(str);

  if (copy)
    {
      char *buf = static_cast<char *>(malloc(len + 1));
      if (buf == NULL)
	{
	  cerr << "Cannot allocate enough room for string Port\n";
	  exit(EXIT_FAILURE);	// give up to continue.
	}

      memcpy(buf, str, len);
      buf[len] = '\0';
      storage = buf;
      cur = buf;
      tail = buf + len;
      // port_type |= PORT_OWN_STRING;
    }
  else
    {
      storage = const_cast<char *>(str);
      cur = const_cast<char *>(str);
      tail = const_cast<char *>(str) + len;
    }
}


InputStringPort::~InputStringPort()
{
  if (storage != NULL && (port_type & PORT_OWN_STRING) != 0)
    free(storage);
}


void
InputStringPort::close()
{
  if (storage != NULL && (port_type & PORT_OWN_STRING) != 0)
    free(storage);

  storage = NULL;
  cur = NULL;
  tail = NULL;
  port_type = PORT_FREE;
}


bool
InputStringPort::char_ready()
{
  return true;	// always true!
}


long
InputStringPort::peek_char()
{
  return cur < tail ? *cur : EOF;
}


long
InputStringPort::get_char()
{
  return cur < tail ? *cur++ : EOF;
}


Port *
InputStringPort::unget_char(long c)
{
  if (cur > storage)
    --cur;

  return this;
}


const char *
InputStringPort::get_remaining_string() const
{
  return cur;
}


size_t
InputStringPort::get_remaining_strlen() const
{
  return tail - cur;
}


void
InputStringPort::flush_remaining_input()
{
  tail = cur;
}


void
OutputStringPort::enlarge_capacity(size_t required_space)
{
  // redundant check
  if (get_capacity_left() >= required_space)
    return;

  size_t new_capacity;
  if (storage == NULL)
    new_capacity = INITIAL_STRING_PORT_CAPACITY;
  else
    new_capacity = capacity;

  size_t capacity_used = tail - storage;
  while (new_capacity - capacity_used < required_space)
    new_capacity *= 2;

  char *old_storage = storage;
  storage = static_cast<char *>(realloc(old_storage, new_capacity));
  if (storage != NULL)
    {
      cur = storage + (cur - old_storage);
      tail = storage + (tail - old_storage);
      capacity = new_capacity;
    }
  else
    {
      fprintf(stderr, "Cannot allocate enough room for string Port\n");
      exit(EXIT_FAILURE);	// give up to continue.
    }
}


OutputStringPort::OutputStringPort()
  : Port(PORT_STRING | PORT_OUTPUT),
    storage(NULL), cur(NULL), tail(NULL), capacity(0)
{
}


OutputStringPort::~OutputStringPort()
{
  if (storage != NULL)
    free(storage);
}


void
OutputStringPort::close()
{
  if (storage != NULL)
    free(storage);

  storage = NULL;
  cur = NULL;
  tail = NULL;
  capacity = 0;
  port_type = PORT_FREE;
}


Port *
OutputStringPort::append(long c)
{
  if (port_type != PORT_FREE)
    {
      ensure_capacity(2);
      *tail++ = c;
      *tail = '\0';
    }

  return this;
}


Port *
OutputStringPort::append(const char *str)
{
  if (port_type != PORT_FREE)
    {
      size_t len = strlen(str);
      ensure_capacity(len + 1);
      strcpy(tail, str);
      tail += len;
    }

  return this;
}


Port *
OutputStringPort::append(const char *str, size_t len)
{
  if (port_type != PORT_FREE)
    {
      ensure_capacity(len + 1);
      memcpy(tail, str, len);
      tail += len;
      *tail = '\0';
    }

  return this;
}


const char *
OutputStringPort::get_output_string() const
{
  return storage != NULL ? storage : "";
}


size_t
OutputStringPort::get_output_strlen() const
{
  return tail - storage;
}


TranscriptPort::TranscriptPort(FILE *fout)
  : Port(PORT_FILE | PORT_OUTPUT), file(fout), unput_count(0)
{
}


TranscriptPort::~TranscriptPort()
{
  if (file != NULL)
    fclose(file);
}


void
TranscriptPort::close()
{
  if (file != NULL)
    fclose(file);

  file = NULL;
  port_type = PORT_FREE;
}


Port *
TranscriptPort::append(long c)
{
  if (unput_count == 0)
    {
      if (c != EOF && file != NULL)
	{
	  fputc(c, file);
	  if (c == '\n')
	    fflush(file);
	}
    }
  else
    --unput_count;

  return this;
}


Port *
TranscriptPort::append(const char *str)
{
  if (unput_count > 0)
    while (*str++ != '\0' && --unput_count > 0)
      ;

  if (unput_count == 0 && file != NULL)
    {
      fputs(str, file);
      if (str[strlen(str) - 1] == '\n')
	fflush(file);
    }

  return this;
}


Port *
TranscriptPort::append(const char *str, size_t len)
{
  if (unput_count > len)
    {
      unput_count -= len;
      return this;
    }

  len -= unput_count;
  str += unput_count;
  unput_count = 0;

  if (file != NULL)
    {
      fwrite(str, 1, len, file);
      if (str[len - 1] == '\n')
	fflush(file);
    }

  return this;
}


Port *
TranscriptPort::flush()
{
  if (file != NULL)
    fflush(file);

  return this;
}


Port *
TranscriptPort::unput_char(long c)
{
  ++unput_count;
  return this;
}

}
