/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: port.h,v 1.24 2009/01/01 07:47:00 fuyu Exp $
 */

#ifndef _LIBTSCORE_PORT_H_
#define _LIBTSCORE_PORT_H_

#ifdef __cplusplus

#if defined(__GNUG__) && !defined(__ICC)
#pragma interface
#endif

#include <cstdio>
#include <cstring>


using namespace std;
namespace LibTSCore
{

#define PORT_FREE	0
#define PORT_FILE	1
#define PORT_STRING	2
#define PORT_INPUT	16
#define PORT_OUTPUT	32
#define PORT_OWN_STRING	64

  class Cell;
  class TranscriptPort;

  /*
   * Port provides functionalities to implement I/O port defined in
   * R5RS.
   * This is essentially just a building block of TSCore, and not
   * too useful for general purpose.
   */
  class Port
  {
  protected:
    unsigned long port_type;


  public:
    Port(long flags) : port_type(flags) {}
    virtual ~Port() {}

    inline bool is_free() const { return port_type == PORT_FREE; }
    inline bool is_file_port() const { return (port_type & PORT_FILE) != 0; }

    inline bool is_string_port() const
    {
      return (port_type & PORT_STRING) != 0;
    }

    inline bool is_input_port() const { return (port_type & PORT_INPUT) != 0; }

    inline bool is_output_port() const
    {
      return (port_type & PORT_OUTPUT) != 0;
    }

    virtual void close() = 0;

    virtual bool char_ready();

    virtual long peek_char();
    virtual long get_char();
    virtual Port *unget_char(long c);

    virtual const char *get_remaining_string() const;
    virtual size_t get_remaining_strlen() const;

    virtual void flush_remaining_input();

    virtual Port *append(long c);
    virtual Port *append(const char *str);
    virtual Port *append(const char *str, size_t len);
    virtual Port *append(const Cell *cell);
    virtual Port *flush();

    virtual const char *get_output_string() const;
    virtual size_t get_output_strlen() const;

    virtual void set_transcript_port(TranscriptPort *port);
  };


  class InputFilePort : public Port
  {
  private:
    FILE *file;
    TranscriptPort *transcript_port;
    long closeit;
    int fd;


  public:
    InputFilePort(FILE *fin, bool closeit = false);
    virtual ~InputFilePort();

    virtual void close();

    virtual bool char_ready();

    virtual long peek_char();
    virtual long get_char();
    virtual Port *unget_char(long c);

    virtual void flush_remaining_input();

    virtual void set_transcript_port(TranscriptPort *port);
  };


  class OutputFilePort : public Port
  {
  private:
    FILE *file;
    TranscriptPort *transcript_port;
    long closeit;


  public:
    OutputFilePort(FILE *fout, bool closeit = false);
    virtual ~OutputFilePort();

    virtual void close();

    virtual Port *append(long c);
    virtual Port *append(const char *str);
    virtual Port *append(const char *str, size_t len);
    virtual Port *flush();

    virtual void set_transcript_port(TranscriptPort *port);
  };


  class InputStringPort : public Port
  {
  private:
    char *storage;		// pointer to the entire storage
    char *cur;			// current position for input
    char *tail;			// current position for output


  public:
    InputStringPort(const char *str, size_t len = 0, bool copy = true);
    virtual ~InputStringPort();

    virtual void close();

    virtual bool char_ready();

    virtual long peek_char();
    virtual long get_char();
    virtual Port *unget_char(long c);

    virtual const char *get_remaining_string() const;
    virtual size_t get_remaining_strlen() const;

    virtual void flush_remaining_input();
  };


  class OutputStringPort : public Port
  {
  private:
    static const size_t INITIAL_STRING_PORT_CAPACITY = 256;

    char *storage;		// pointer to the entire storage
    char *cur;			// current position for input
    char *tail;			// current position for output
    size_t capacity;		// the capacity of the entire storage

    inline size_t get_capacity_left() const
    {
      return storage + capacity - tail;
    }


    /*
     * Ensures string Port to have enough room to output `required_space'
     * bytes.
     */
    inline void ensure_capacity(size_t required_space)
    {
      if (get_capacity_left() >= required_space)
	return;

      enlarge_capacity(required_space);
    }

    void enlarge_capacity(size_t required_space);


  public:
    OutputStringPort();
    virtual ~OutputStringPort();

    virtual void close();

    virtual Port *append(long c);
    virtual Port *append(const char *str);
    virtual Port *append(const char *str, size_t len);

    inline Port *append(OutputStringPort &port)
    {
      return append(port.get_output_string(), port.get_output_strlen());
    }

    inline Port *append(OutputStringPort *port)
    {
      return append(port->get_output_string(), port->get_output_strlen());
    }

    virtual const char *get_output_string() const;
    virtual size_t get_output_strlen() const;

    inline Port *clear()
    {
      tail = storage;
      if (tail != NULL)
	*tail = '\0';

      return this;
    }
  };


  class TranscriptPort : public Port
  {
  private:
    FILE *file;
    size_t unput_count;


  public:
    TranscriptPort(FILE *fout);
    virtual ~TranscriptPort();

    virtual void close();

    virtual Port *append(long c);
    virtual Port *append(const char *str);
    virtual Port *append(const char *str, size_t len);
    virtual Port *flush();

    virtual Port *unput_char(long c);
  };


}


#endif	/* __cplusplus */

#endif	/* _LIBTSCORE_PORT_H_ */
