/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: procedure.cc,v 1.20 2009/01/02 03:07:32 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libtscore/procedure.h"

#include "libtscore/bytecode_buffer.h"
#include "libtscore/cell.h"
#include "libtscore/port.h"
#include "libtscore/r5rs_list_procedure.h"
#include "libtscore/tscore.h"
#include "libtscore/writer.h"

#include <cstdio>


namespace LibTSCore
{

/* For arguments test */
bool
Procedure::is_any(Cell *cell)
{
  return true;
}


bool
Procedure::is_string(Cell *cell)
{
  return cell->is_string();
}


bool
Procedure::is_symbol(Cell *cell)
{
  return cell->is_symbol();
}


bool
Procedure::is_port(Cell *cell)
{
  return cell->is_port();
}


bool
Procedure::is_input_port(Cell *cell)
{
  return cell->is_input_port();
}


bool
Procedure::is_output_port(Cell *cell)
{
  return cell->is_output_port();
}


bool
Procedure::is_environment(Cell *cell)
{
  return cell->is_environment();
}


bool
Procedure::is_pair(Cell *cell)
{
  return cell->is_pair();
}


bool
Procedure::is_list(Cell *cell)
{
  return R5RSListProcedure::is_list(cell);
}


bool
Procedure::is_character(Cell *cell)
{
  return cell->is_character();
}


bool
Procedure::is_vector(Cell *cell)
{
  return cell->is_vector();
}


bool
Procedure::is_number(Cell *cell)
{
  return cell->is_number();
}


bool
Procedure::is_integer(Cell *cell)
{
  return cell->is_integer();
}


bool
Procedure::is_non_negative_integer(Cell *cell)
{
  return cell->is_integer() && cell->get_integer0() >= 0;
}


bool
Procedure::is_regexp(Cell *cell)
{
  return cell->is_regexp();
}


bool
Procedure::is_output_string_port(Cell *cell)
{
  if (!cell->is_output_port())
    return false;

  return cell->get_port()->is_string_port();
}


bool
Procedure::is_input_string_port(Cell *cell)
{
  if (!cell->is_input_port())
    return false;

  return cell->get_port()->is_string_port();
}


bool
Procedure::is_closure(Cell *cell)
{
  return cell->is_closure();
}


bool
Procedure::is_mutable_string(Cell *cell)
{
  return cell->is_string() && !cell->is_immutable();
}


bool
Procedure::is_mutable_pair(Cell *cell)
{
  return cell->is_pair() && !cell->is_immutable();
}


bool
Procedure::is_mutable_vector(Cell *cell)
{
  return cell->is_vector() && !cell->is_immutable();
}


bool
Procedure::is_promise(Cell *cell)
{
  //return cell->is_forced_promise() || cell->is_promise();
  return false;
}


bool
Procedure::is_foreign_object(Cell *cell)
{
  return cell->is_foreign_object();
}


bool
Procedure::is_procedure(Cell *cell)
{
  return cell->is_procedure();
}


bool
Procedure::is_time(Cell *cell)
{
  return cell->is_time();
}


bool
Procedure::is_date(Cell *cell)
{
  return cell->is_date();
}


bool
Procedure::is_real(Cell *cell)
{
  return cell->is_real();
}


/* Correspond carefully with defined TST_XXX! */
Procedure::ArgumentTest Procedure::argument_tests[] = {
  { NULL,				NULL },			// no test
  { Procedure::is_any,			NULL },
  { Procedure::is_string,		"string" },
  { Procedure::is_symbol,		"symbol" },
  { Procedure::is_port,			"port" },
  { Procedure::is_input_port,		"input port" },
  { Procedure::is_output_port,		"output_port" },
  { Procedure::is_environment,		"environment" },
  { Procedure::is_pair,			"pair" },
  { Procedure::is_list,			"list" },
  { Procedure::is_character,		"character" },
  { Procedure::is_vector,		"vector" },
  { Procedure::is_number,		"number" },
  { Procedure::is_integer,		"integer" },
  { Procedure::is_non_negative_integer,	"non negative integer" },
  { Procedure::is_regexp,		"regular expression" },
  { Procedure::is_output_string_port,	"output string port" },
  { Procedure::is_input_string_port,	"input string port" },
  { Procedure::is_closure,		"closure" },
  { Procedure::is_mutable_string,	"mutable string" },
  { Procedure::is_mutable_pair,		"mutable pair" },
  { Procedure::is_mutable_vector,	"mutable vector" },
  { Procedure::is_promise,		"promise" },
  { Procedure::is_foreign_object,	"foreign object" },
  { Procedure::is_procedure,		"procedure" },
  { Procedure::is_time,			"time" },
  { Procedure::is_date,			"date" },
  { Procedure::is_real,			"real" }
};


Cell *
Procedure::signal_error(VirtualMachine &vm, const char *message,
			const Cell *value)
{
  OutputStringPort *error_buffer = vm.get_error_buffer();
  error_buffer->clear();
  error_buffer->append(message);
  Writer::write_datum(value, error_buffer, false);
  return NULL;
}


BytecodeProcedure::~BytecodeProcedure()
{
  delete buffer;
}


Cell *
BytecodeProcedure::call(VirtualMachine &vm, Cell *procedure,
			Register &caller_context, Cell *&caller_frame,
			long args_pointer, long nargs)
{
  if (nargs == minimum_arity ||
      (minimum_arity < nargs && minimum_arity < maximum_arity))
    {
#if 0
      fprintf(stderr, "setting cp for procedure %p\n", procedure);
#endif
      caller_context.constant_pool = procedure->get_closure_constant_pool();

      Cell *frame;
      if (caller_frame == caller_context.trampoline_frame)
	{
	  assert(args_pointer == 0);

	  VMStackRoot proc(vm, procedure);
	  Cell *continuation = caller_frame->load_variable(0);
	  if (continuation == Cell::nil())
	    {
	      continuation = vm.mk_current_continuation(caller_context);
	      caller_frame->store_variable(vm.memory_system, 0, continuation);
	    }

	  caller_context.procedure = proc;

	  // In this case, we can use trampoline_frame as the callee frame
	  // because it is always freshly allocated and having enough room.
	  assert(caller_frame->get_frame_size() >= get_frame_size());

	  frame = caller_frame;
	  // `(car frame)' hold the parent frame.
	  // Should this be written like below?
	  // frame->set_parent_frame(vm.memory_system,
	  //                         proc->get_closure_frame());
	  vm.memory_system.set_car(frame, proc->get_closure_frame());
	  goto prepare_arguments;
	}
      else
	{
	  VMStackRoot continuation(vm,
		caller_context.frame->load_variable(args_pointer));
	  if (continuation == Cell::nil())
	    {
	      // The call isn't a tail-call.
	      // In this case, we need to create a new internal continuation.
	      // This may take a little expensive cost, and we will find
	      // some optimization here or VM's internal to recycle
	      // intra-procedural continuation object, since such continuation
	      // object is needed only one for each executed procedure.
	      VMStackRoot proc(vm, procedure);
	      continuation = vm.mk_current_continuation(caller_context);
	      caller_context.procedure = proc;
	      frame = vm.frame_new(proc->get_closure_frame(),
				   get_frame_size());
	    }
	  else
	    {
	      assert(continuation->is_continuation());

	      // The call is a tail-call.
	      // In this case, we try to recycle the current frame if it has
	      // enough room for callee frame.
	      frame = caller_context.frame;
	      if (frame != Cell::nil() && !frame->is_frame_closed())
		{
		  if (frame->get_frame_size() < get_frame_size())
		    {
		      // caller frame cannot be recycled because of the size
		      if (caller_context.procedure->is_closure())
			{
			  BytecodeProcedure *caller
			    = caller_context.procedure->get_closure_bytecode();
			  // Enlarge caller's frame size requisition for
			  // future tail-calls.
			  caller->frame_size = get_frame_size();
			}
		      caller_context.procedure = procedure;
		      frame = vm.frame_new(procedure->get_closure_frame(),
					   get_frame_size());
		    }
		  else
		    {
		      // `(car frame)' hold the parent frame.
		      // Should this be written like below?
		      // frame->set_parent_frame(vm.memory_system,
		      //                         proc->get_closure_frame());
		      vm.memory_system.set_car(frame,
					       procedure->get_closure_frame());
		      caller_context.procedure = procedure;
		    }
		}
	      else
		{
		  caller_context.procedure = procedure;
		  frame = vm.frame_new(procedure->get_closure_frame(),
				       get_frame_size());
		}
	    }

	  if (frame == Cell::nil())
	    return signal_error(vm, "Out of Memory: ", procedure);

	  frame->store_variable(vm.memory_system, 0, continuation);
	}

    prepare_arguments:
      if (caller_frame != caller_context.trampoline_frame)
	for (long i = 1; i < minimum_arity; i++)
	  frame->store_variable(vm.memory_system, i,
				caller_frame->load_variable(args_pointer + i));

      if (nargs == minimum_arity)
	{
	  if (maximum_arity > minimum_arity)
	    frame->store_variable(vm.memory_system,
				  minimum_arity, Cell::nil());

	  if (caller_context.frame == frame && args_pointer > 0)
	    {
	      long frame_size = frame->get_frame_size();
	      for (long i = minimum_arity + 1; i < frame_size; i++)
		frame->store_variable(vm.memory_system, i, Cell::unspecified());
	    }

	  caller_context.frame = frame;
	}
      else
	{
	  Cell *rest_args = Cell::nil();
	  VMStackRoot frm(vm, frame);
	  while (nargs-- > minimum_arity)
	    {
	      Cell *arg = caller_frame->load_variable(args_pointer + nargs);
	      rest_args = vm.mk_cons(arg, rest_args, false);
	    }

	  frm->store_variable(vm.memory_system, minimum_arity, rest_args);
	  if (caller_context.frame == frm)
	    {
	      long frame_size = frm->get_frame_size();
	      for (long i = minimum_arity + 1; i < frame_size; i++)
		frm->store_variable(vm.memory_system, i, Cell::unspecified());
	    }

	  caller_context.frame = frm;
	}

      if (vm.get_tracing(caller_context))
	{
	  Port *port = caller_context.output_port->get_port();
	  port->append("apply: ");
	  Writer::write_datum(caller_context.procedure, port, false);
	  port->append("\n   ");
	  long i;
	  for (i = 1; i < minimum_arity; i++)
	    {
	      Writer::write_datum(caller_context.frame->load_variable(i),
				  port, true, true);
	      port->append(' ');
	    }

	  if (minimum_arity < maximum_arity)
	    Writer::write_datum(caller_context.frame->load_variable(i),
				port, true, true);

	  port->append("\n");
	}

#ifndef NDEBUG
      if (caller_context.trampoline_frame != Cell::nil())
	{
	  // just for implementation checks
	  caller_context.trampoline_frame = Cell::nil();
	  caller_context.nargs = 0;
	}
#endif

      caller_context.pc = buffer->get_bytecode();
      return Cell::nil();	// return to interpreter loop.
    }

#if 0
  fprintf(stderr, "min_arity=%ld, max_arity=%ld, nargs=%ld\n",
	  minimum_arity, maximum_arity, nargs);
#endif
  return signal_error(vm, "Wrong number of arguments: ", procedure);
}


Continuation Continuation::singleton(NULL, 0);


Cell *
Continuation::call(VirtualMachine &vm, Cell *procedure,
		   Register &caller_context, Cell *&caller_frame,
		   long args_pointer, long nargs)
{
  Cell *value = caller_frame->load_variable(args_pointer + 1);
  if (caller_context.frame == caller_frame)
    {
      caller_context.trampoline_frame = Cell::nil();
      caller_context.nargs = 0;
    }

  return apply(vm, procedure, caller_context, value);
}


Cell *
Continuation::apply(VirtualMachine &vm, Cell *continuation,
		    Register &caller_context, Cell *value)
{
  caller_context.value = value;
  caller_context.pc = continuation->get_continuation_pc();
  caller_context.procedure = continuation->get_continuation_procedure();

  if (continuation->is_full_continuation())
    {
      caller_context.input_port = continuation->get_continuation_input_port();
      caller_context.output_port
	= continuation->get_continuation_output_port();
      caller_context.constant_pool
	= caller_context.procedure->get_closure_constant_pool();
    }
  else
    {
      if (caller_context.procedure != Cell::nil())
	{
#if 0
	  fprintf(stderr, "setting cp for continuation procedure %p\n",
		  caller_context.procedure);
#endif
	  caller_context.constant_pool
	    = caller_context.procedure->get_closure_constant_pool();
	  vm.recycle_continuation(continuation);
	}
#if 0
      else
	{
	  fprintf(stderr, "continuation procedure is nil!\n");
	}
#endif
    }

  if (vm.get_tracing(caller_context))
    {
      Port *port = caller_context.output_port->get_port();
      port->append("Continue with: ");
      Writer::write_datum(caller_context.value, port, true, true);
      port->append("\n");
    }

  caller_context.frame = continuation->get_continuation_frame();
#if 0
  fprintf(stderr, "continuation frame=%p\n", caller_context.frame);
#endif
  return Cell::nil();
}


void
NativeProcedure::assign_procedure(Cell *environment, Cell *frame,
				  const Info *info, void *data)
{
  TSStackRoot env(core, environment);
  TSStackRoot frm(core, frame);
  TSStackRoot proc(core, core.mk_native_procedure(this, info, data));
  Cell *symbol = core.mk_static_symbol(info->name);
  const Cell *binding = env->bind_variable(symbol);
  frm->store_toplevel_variable(env, binding->get_slot_location(), proc);
}


const char *
NativeProcedure::get_procedure_name(const Cell *procedure) const
{
  return procedure->get_native_procedure_info()->name;
}


const char *
NativeProcedure::test_arguments(OutputStringPort &error_buf, const Info *info,
				Cell *frame, long fp, long nargs) const
{
  char message[512];

  /* Check number of arguments first. */
  if (nargs < info->minimum_arity)
    {
      snprintf(message, sizeof(message), "%s: needs%s %d argument%s",
	       info->name,
	       info->minimum_arity == info->maximum_arity ? "" : " at least",
	       info->minimum_arity,
	       info->minimum_arity > 1 ? "s" : "");
      goto error;
    }

  if (nargs > info->maximum_arity)
    {
      snprintf(message, sizeof(message), "%s: needs%s %d argument%s",
	       info->name,
	       info->minimum_arity == info->maximum_arity ? "" : " at most",
	       info->maximum_arity,
	       info->maximum_arity > 1 ? "s" : "");
      goto error;
    }

  if (info->encoded_argument_tests != NULL)
    {
      const char *encoded_tests = info->encoded_argument_tests;
      for (long i = 0; i < nargs; i++)
	{
	  Cell *arg = frame->load_variable(fp + i);
	  long requirement = *encoded_tests;
	  if (!argument_tests[requirement].func(arg))
	    {
	      snprintf(message, sizeof(message),
		       "%s: argument #%ld must be of %s : ",
		       info->name, i + 1,
		       argument_tests[requirement].type);
	      error_buf.clear();
	      error_buf.append(message);
	      Writer::write_datum(arg, &error_buf, false);
	      return error_buf.get_output_string();
	    }

	  // The last type of encoded_tests are continuously used for
	  // arguments having variable length.
	  if (encoded_tests[1] != '\0')
	    ++encoded_tests;
	}
    }

  return NULL;

 error:
  error_buf.clear();
  error_buf.append(message);
  return error_buf.get_output_string();
}


Cell *
NativeProcedure::call(VirtualMachine &vm, Cell *procedure,
		      Register &caller_context, Cell *&caller_frame,
		      long args_pointer, long nargs)
{
  assert(procedure->is_native_procedure());
  assert(procedure->get_native_procedure() == this);

  // ignore the first argument, continuation, given by VirtualMachine
  // temporariry here.
  ++args_pointer;
  --nargs;

  VMStackRoot frame(vm, caller_context.frame);
  if (caller_context.trampoline_frame == caller_frame)
    {
      caller_context.frame = caller_frame;
      caller_context.trampoline_frame = Cell::nil();
      caller_context.nargs = 0;
    }

  const Info *info = procedure->get_native_procedure_info();
  void *data = procedure->get_native_procedure_data();
  const char *err = test_arguments(*vm.get_error_buffer(), info,
				   caller_context.frame,
				   args_pointer, nargs);
  if (err == NULL)
    {
      if (vm.get_tracing(caller_context))
	{
	  Port *port = caller_context.output_port->get_port();
	  port->append("apply: ");
	  Writer::write_datum(procedure, port, false);
	  port->append("\n   ");
	  for (long i = 0; i < nargs; i++)
	    {
	      Writer::write_datum(caller_context.frame->load_variable(args_pointer + i),
				  port, true, true);
	      port->append(' ');
	    }

	  port->append("\n");
	}

      Cell *result = (*info->procedure)(vm, caller_context, args_pointer,
					nargs, data);

      if (result != NULL)
	{
	  // procedure call is succeeded.
	  if (result != Cell::trampoline())
	    {
	      // usual case.
	      Cell *continuation
		= caller_context.frame->load_variable(args_pointer - 1);
	      if (continuation == Cell::nil())
		{
		  // This procedure call wasn't a tail-call.
		  // Since all of our native procedure implementations
		  // that return non Cell::unspecified() never change
		  // interpreter's state, we can simply return here
		  // to continue with interpreter's state.
		  caller_context.value = result;
		  caller_context.frame = frame;
		  if (vm.get_tracing(caller_context))
		    {
		      Port *port = caller_context.output_port->get_port();
		      port->append("Continue with: ");
		      Writer::write_datum(result, port, true, true);
		      port->append('\n');
		    }

		  return Cell::nil();
		}

	      Continuation *cont = continuation->get_continuation();
	      return cont->apply(vm, continuation, caller_context, result);
	    }

	  // special case that executed procedure performs something that
	  // changes the interpreter's state, e.g. call/cc.
	  return Cell::unspecified();
	}

      return result;	// result == NULL means error occured.
    }

  return NULL;
}

}
