/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: procedure.h,v 1.15 2009/01/01 07:47:00 fuyu Exp $
 */

#ifndef _LIBTSCORE_PROCEDURE_H_
#define _LIBTSCORE_PROCEDURE_H_

#ifdef __cplusplus

#if defined(__GNUG__) && !defined(__ICC)
#pragma interface
#endif

#include "libtscore/port.h"


using namespace std;
namespace LibTSCore
{
  class Cell;
  class OutputStringPort;
  class Register;
  class TSCore;
  class VirtualMachine;
  class Procedure
  {
    friend class VirtualMachine;

  protected:
    typedef bool (*TestPredicate)(Cell *cell);
    struct ArgumentTest
    {
      TestPredicate func;
      const char *type;
    };

    static ArgumentTest argument_tests[];

    static bool is_any(Cell *cell);
    static bool is_string(Cell *cell);
    static bool is_symbol(Cell *cell);
    static bool is_port(Cell *cell);
    static bool is_input_port(Cell *cell);
    static bool is_output_port(Cell *cell);
    static bool is_environment(Cell *cell);
    static bool is_pair(Cell *cell);
    static bool is_list(Cell *list);
    static bool is_character(Cell *cell);
    static bool is_vector(Cell *cell);
    static bool is_number(Cell *cell);
    static bool is_integer(Cell *cell);
    static bool is_non_negative_integer(Cell *cell);
    static bool is_regexp(Cell *cell);
    static bool is_output_string_port(Cell *cell);
    static bool is_input_string_port(Cell *cell);
    static bool is_closure(Cell *cell);
    static bool is_mutable_string(Cell *cell);
    static bool is_mutable_pair(Cell *cell);
    static bool is_mutable_vector(Cell *cell);
    static bool is_promise(Cell *cell);
    static bool is_foreign_object(Cell *cell);
    static bool is_procedure(Cell *cell);
    static bool is_time(Cell *cell);
    static bool is_date(Cell *cell);
    static bool is_real(Cell *cell);


    // call is always invoked from VirtualMachine's interpreter loop.
    // The first argument is always the continuation in our convention.
    // If this returns NULL, the VirtualMachine will treat it as
    // error situation.
    virtual Cell *call(VirtualMachine &vm, Cell *procedure,
		       Register &caller_context, Cell *&caller_frame,
		       long args_pointer, long nargs) = 0;

    static Cell *signal_error(VirtualMachine &vm,
			      const char *message, const Cell *value = NULL);


  public:
    virtual ~Procedure() {}
  };


  class BytecodeBuffer;
  class BytecodeProcedure : public Procedure
  {
    friend class Writer;
  private:
    BytecodeBuffer *buffer;
    long frame_size;
    long minimum_arity;
    long maximum_arity;

    inline BytecodeBuffer *get_buffer() const { return buffer; }


  protected:
    Cell *call(VirtualMachine &vm, Cell *procedure,
	       Register &caller_context, Cell *&caller_frame,
	       long args_pointer, long nargs);
    inline void set_frame_size(long size) { frame_size = size; }


  public:
    BytecodeProcedure(BytecodeBuffer *buffer, long size,
		      long min_arity, long max_arity)
      : buffer(buffer), frame_size(size),
	minimum_arity(min_arity), maximum_arity(max_arity) {}
    virtual ~BytecodeProcedure();

    inline long get_frame_size() const { return frame_size; }
    inline long get_minimum_arity() const { return minimum_arity; }
    inline long get_maximum_arity() const { return maximum_arity; }
  };


  class Continuation : public Procedure
  {
    friend class NativeProcedure;
    friend class VirtualMachine;
  private:
    static Continuation singleton;
    const char *pc;
    long depth;


  protected:
    Cell *call(VirtualMachine &vm, Cell *procedure,
	       Register &caller_context, Cell *&caller_frame,
	       long args_pointer, long nargs);

    Cell *apply(VirtualMachine &vm, Cell *continuation,
		Register &context, Cell *value);


  public:
    Continuation(const char *pc, long depth) : pc(pc), depth(depth) {}
    ~Continuation() {}
    inline const char *get_pc() const { return pc; }
    inline long get_depth() const { return depth; }

    static Continuation *get_singleton() { return &singleton; }
  };


  /*
   * NativeProcedure is the base class of native procedure implementations
   * of our Scheme interpreter.
   *
   * This defines a set of methods that are used by our interpreter and
   * concrete class of this that implements native procedures.
   */
  class NativeProcedure : public Procedure
  {
  public:
    typedef Cell *(*ProcedureImplementation)(VirtualMachine &vm, Register &context, unsigned long argument_pointer, unsigned long nargs, void *data);


    struct Info
    {
      ProcedureImplementation procedure;
      const char *name;
      const char *encoded_argument_tests;
      unsigned short minimum_arity;
      unsigned short maximum_arity;
    };


  private:
    /*
     * This returns NULL if given `args' are confortable for the `procedure',
     * return the error message string otherwise.
     */
    const char *test_arguments(OutputStringPort &error_buf, const Info *info,
			       Cell *frame, long fp, long nargs) const;


  protected:
    TSCore &core;

    void assign_procedure(Cell *environment, Cell *frame,
			  const Info *info, void *data);
    Cell *call(VirtualMachine &vm, Cell *procedure,
	       Register &caller_context, Cell *&caller_frame,
	       long args_pointer, long nargs);


  public:
    NativeProcedure(TSCore &tscore) : core(tscore) {}
    virtual ~NativeProcedure() {}

    const char *get_procedure_name(const Cell *procedure) const;

    // This is invoked by TSCore
    virtual Cell *assign_procedures(Cell *environment, Cell *frame) = 0;
  };

#define TST_NONE		0
#define TST_ANY			"\001"
#define TST_STRING		"\002"
#define TST_SYMBOL		"\003"
#define TST_PORT		"\004"
#define TST_INPORT		"\005"
#define TST_OUTPORT		"\006"
#define TST_ENVIRONMENT		"\007"
#define TST_PAIR		"\010"
#define TST_LIST		"\011"
#define TST_CHAR		"\012"
#define TST_VECTOR		"\013"
#define TST_NUMBER		"\014"
#define TST_INTEGER		"\015"
#define TST_NONNEG		"\016"
#define TST_REGEXP		"\017"
#define TST_STR_OUTPORT		"\020"
#define TST_STR_INPORT		"\021"
#define TST_CLOSURE		"\022"
#define TST_MUTABLE_STRING	"\023"
#define TST_MUTABLE_PAIR	"\024"
#define TST_MUTABLE_VECTOR	"\025"
#define TST_PROMISE		"\026"
#define TST_FOREIGN_OBJECT	"\027"
#define TST_PROCEDURE		"\030"
#define TST_TIME		"\031"
#define TST_DATE		"\032"
#define TST_REAL		"\033"

#define INF_ARG			0xffff
}


#endif

#endif	/* _LIBTSCORE_PROCEDURE_H_ */
