/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: r5rs_derived_syntax.cc,v 1.18 2009/01/01 07:47:00 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libtscore/r5rs_derived_syntax.h"

#include "libtscore/bytecode_buffer.h"
#include "libtscore/virtual_machine.h"
#include "libtscore/syntax.h"


#include <cassert>
#include <vector>


namespace LibTSCore
{

size_t
And::emit_bytecode(VirtualMachine &vm, Register &context,
		   BytecodeBuffer *buffer,
		   Cell *expression, Cell *environment,
		   Cell *constant_pool, Cell *hint,
		   bool is_tail_context)
{
  assert(expression->is_pair());

  if (!is_tail_context)
    buffer->set_include_forward_branch();

  VMStackRoot form(vm, expression);
  VMStackRoot env(vm, environment);
  VMStackRoot cpool(vm, constant_pool);
  VMStackRoot hnt(vm, extend_hint(vm, CAR(form), hint));

  size_t total_len = 0;
  vector<size_t> branch_offsets;
  VMStackRoot args(vm, CDR(form));
  while (args != Cell::nil())
    {
      Cell *arg = CAR(args);
      args = CDR(args);

      size_t len = bytecompile(vm, context, buffer, arg, env, cpool, hnt,
			       false);
      if (len == 0)
	return signal_error(vm, "and: syntax error: ", form);

      total_len += len;

      if (args == Cell::nil())
	break;

      if (is_tail_context)
	{
	  total_len += buffer->append(OP_IF_T1);
	  total_len += buffer->append_signed_8(2);
	  total_len += buffer->append(OP_APPLY_CONTINUATION0);
	}
      else
	{
	  total_len += buffer->append(OP_IF_F2);	// a little conservative
	  branch_offsets.push_back(buffer->get_current_offset());
	  total_len += buffer->append_signed_16(2);	// dummy
	}
    }

  size_t current_offset = buffer->get_current_offset();

  vector<size_t>::const_iterator iter = branch_offsets.begin();
  vector<size_t>::const_iterator end_iter = branch_offsets.end();
  while (iter != end_iter)
    {
      size_t branch_off = *iter++;
      size_t offset = current_offset - branch_off;
      if (offset > 32767)
	return signal_error(vm, "unsupported too long <test> sequences: ",
			    form);

      buffer->set_signed_16(branch_off, offset);
    }

  if (is_tail_context)
    return total_len + buffer->append(OP_APPLY_CONTINUATION0);

  return total_len;
}


size_t
Or::emit_bytecode(VirtualMachine &vm, Register &context,
		  BytecodeBuffer *buffer,
		  Cell *expression, Cell *environment,
		  Cell *constant_pool, Cell *hint,
		  bool is_tail_context)
{
  assert(expression->is_pair());

  if (!is_tail_context)
    buffer->set_include_forward_branch();

  VMStackRoot form(vm, expression);
  VMStackRoot env(vm, environment);
  VMStackRoot cpool(vm, constant_pool);
  VMStackRoot hnt(vm, extend_hint(vm, CAR(form), hint));

  size_t total_len = 0;
  vector<size_t> branch_offsets;
  VMStackRoot args(vm, CDR(form));
  while (args != Cell::nil())
    {
      Cell *arg = CAR(args);
      args = CDR(args);

      size_t len = bytecompile(vm, context, buffer, arg, env, cpool, hnt,
			       false);
      if (len == 0)
	return signal_error(vm, "and: syntax error: ", form);

      total_len += len;

      if (args == Cell::nil())
	break;

      if (is_tail_context)
	{
	  total_len += buffer->append(OP_IF_F1);
	  total_len += buffer->append_signed_8(2);
	  total_len += buffer->append(OP_APPLY_CONTINUATION0);
	}
      else
	{
	  total_len += buffer->append(OP_IF_T2);	// a little conservative
	  branch_offsets.push_back(buffer->get_current_offset());
	  total_len += buffer->append_signed_16(2);	// dummy
	}
    }

  size_t current_offset = buffer->get_current_offset();

  vector<size_t>::const_iterator iter = branch_offsets.begin();
  vector<size_t>::const_iterator end_iter = branch_offsets.end();
  while (iter != end_iter)
    {
      size_t branch_off = *iter++;
      size_t offset = current_offset - branch_off;
      if (offset > 32767)
	return signal_error(vm, "unsupported too long <test> sequences: ",
			    form);

      buffer->set_signed_16(branch_off, offset);
    }

  if (is_tail_context)
    return total_len + buffer->append(OP_APPLY_CONTINUATION0);

  return total_len;
}


size_t
LetAsterisk::emit_bytecode(VirtualMachine &vm, Register &context,
			   BytecodeBuffer *buffer,
			   Cell *expression, Cell *environment,
			   Cell *constant_pool, Cell *hint,
			   bool is_tail_context)
{
  assert(expression->is_pair());

  VMStackRoot form(vm, expression);
  VMStackRoot env(vm, environment);
  VMStackRoot cpool(vm, constant_pool);
  VMStackRoot hnt(vm, extend_hint(vm, CAR(form), hint));

  if (!CDR(form)->is_pair() || !Letrec::check_bindings(vm, CADR(form), hnt))
    return signal_error(vm, "let*: syntax error: ", form);

  size_t len = 0;
  VMStackRoot letast_bindings(vm, CADR(form));
  VMStackRoot body(vm, CDDR(form));
  while (letast_bindings->is_pair())
    {
      Cell *letast_binding = CAR(letast_bindings);
      assert(letast_binding->is_pair());
      assert(CDR(letast_binding)->is_pair());
      assert(CDDR(letast_binding) == Cell::nil());

      Cell *init = CADR(letast_binding);
      size_t init_len = bytecompile(vm, context, buffer, init,
				    env, cpool, hnt, false);
      if (init_len == 0)
	{
	  signal_error(vm, "wrong binding specification: ",
		       CAR(letast_bindings));
	  return signal_error(vm, "let*: syntax error: ", form);
	}

      len += init_len;

      const Cell *variable = CAAR(letast_bindings); 
      variable = get_unique_symbol(vm, env, variable);
      Cell *rename_from = CAAR(letast_bindings);
      body = rename_variable(vm, body, rename_from, variable);
      Cell *tmp = rename_variable(vm, CDR(letast_bindings),
				  rename_from, variable);
      vm.set_cdr(letast_bindings, tmp);
      vm.set_car(CAR(letast_bindings), const_cast<Cell *>(variable));

      const Cell *binding = bind_variable(vm, env, variable, hnt);
      assert(binding != Cell::nil());

      variable = binding->get_bound_symbol();
      len += emit_variable_assignment(vm, context, buffer, variable,
				      env, cpool);
      letast_bindings = CDR(letast_bindings);
    }

  assert(letast_bindings == Cell::nil());

  // emit body
  if (!CDDR(form)->is_pair())
    return len;	// empty body should be treated as error?

  size_t body_len = bytecompile_sequence(vm, context, buffer, body,
					 env, cpool, hnt, is_tail_context);
  if (body_len > 0)
    return len + body_len;

  return signal_error(vm, "let*: syntax error: ", form);
}


size_t
Letrec::emit_letrec(VirtualMachine &vm, Register &context,
		    BytecodeBuffer *buffer,
		    Cell *expression, Cell *environment,
		    Cell *constant_pool, Cell *hint,
		    Cell *letrec_bindings, Cell *letrec_body,
		    bool is_tail_context)
{
  assert(expression->is_pair());

  VMStackRoot form(vm, expression);
  VMStackRoot env(vm, environment);
  VMStackRoot cpool(vm, constant_pool);
  VMStackRoot hnt(vm, hint);
  VMStackRoot bindings(vm, letrec_bindings);
  VMStackRoot body(vm, letrec_body);

  VMStackRoot tmp_bindings(vm, bindings);
  VMStackRoot variable(vm, Cell::nil());

  while (tmp_bindings->is_pair())
    {
      Cell *letrec_binding = CAR(tmp_bindings);
      assert(letrec_binding->is_pair());
      assert(CDR(letrec_binding)->is_pair());
      assert(CDDR(letrec_binding) == Cell::nil());

      variable = CAR(letrec_binding);
      const Cell *binding;
      if (variable->is_symbol())
	binding = env->lookup(variable, false);
      else
	{
	  assert(variable->is_binding());

	  binding = variable;
	}

      VMStackRoot rename_from(vm, variable);
      variable = const_cast<Cell *>(get_unique_symbol(vm, env, rename_from));
      bindings = rename_variable(vm, bindings, rename_from, variable);
      body = rename_variable(vm, body, rename_from, variable);

      if (bind_variable(vm, env, variable, hnt) == Cell::nil())
	return signal_error(vm, "wrong binding specification: ",
			    CAR(tmp_bindings));

      tmp_bindings = CDR(tmp_bindings);
    }

  assert(tmp_bindings == Cell::nil());

  // emit local variable initializations
  tmp_bindings = static_cast<Cell *>(bindings);
  size_t len = 0;
  while (tmp_bindings->is_pair())
    {
      Cell *letrec_binding = CAR(tmp_bindings);
      Cell *init = CADR(letrec_binding);
      size_t init_len = bytecompile(vm, context, buffer, init,
				    env, cpool, hnt, false);
      if (init_len == 0)
	return signal_error(vm, "wrong binding specification: ",
			    CAR(tmp_bindings));

      len += init_len;
      len += emit_variable_assignment(vm, context, buffer, CAAR(tmp_bindings),
				      env, cpool);
      tmp_bindings = CDR(tmp_bindings);
    }

  if (body == Cell::nil())
    {
      if (is_tail_context)
	return len + buffer->append(OP_APPLY_CONTINUATION0);

      return len;
    }

  size_t body_len = bytecompile_sequence(vm, context, buffer, body,
					 env, cpool, hnt, is_tail_context);
  if (body_len > 0)
    return len + body_len;

  return 0;
}


bool
Letrec::check_bindings(VirtualMachine &vm, Cell *bindings, Cell *hint)
{
  VMStackRoot list(vm, bindings);
  VMStackRoot hnt(vm, hint);
  VMStackRoot dummy_env(vm, Cell::closed_environment_new(vm.get_memory_system(),
							 Cell::nil()));
  while (list->is_pair())
    {
      Cell *binding = CAR(list);
      if (!binding->is_pair()
	  || !CDR(binding)->is_pair()
	  || CDDR(binding) != Cell::nil())
	{
	  signal_error(vm, "wrong binding specification: ", binding);
	  return false;
	}

      const Cell *variable = CAR(binding);
      if (variable->is_binding())
	variable = variable->get_bound_symbol();
      else if (!variable->is_symbol())
	{
	  signal_error(vm, "not a symbol: ", variable);
	  signal_error(vm, "wrong binding specification: ", binding);
	  return false;
	}

      const Cell *name = dummy_env->lookup(variable, false);
      if (name != Cell::nil())
	{
	  signal_error(vm, "duplicated variable: ", variable);
	  signal_error(vm, "wrong binding specification: ", binding);
	  return false;
	}

      if (bind_variable(vm, dummy_env, variable, hnt) == Cell::nil())
	{
	  signal_error(vm, "wrong binding specification: ", CAR(list));
	  return false;
	}

      list = CDR(list);
    }

  if (list == Cell::nil())
    return true;

  signal_error(vm, "unexpected form: ", list);
  return false;
}


size_t
Letrec::emit_bytecode(VirtualMachine &vm, Register &context,
		      BytecodeBuffer *buffer,
		      Cell *expression, Cell *environment,
		      Cell *constant_pool, Cell *hint,
		      bool is_tail_context)
{
  assert(expression->is_pair());

  VMStackRoot form(vm, expression);
  VMStackRoot env(vm, environment);
  VMStackRoot cpool(vm, constant_pool);
  VMStackRoot hnt(vm, extend_hint(vm, CAR(form), hint));

  if (CDR(form)->is_pair() && check_bindings(vm, CADR(form), hnt))
    {
      size_t len = emit_letrec(vm, context, buffer, form, env, cpool, hnt,
			       CADR(form), CDDR(form), is_tail_context);
      if (len > 0)
	return len;
    }

  return signal_error(vm, "letrec: syntax error: ", form);
}


size_t
Begin::emit_bytecode(VirtualMachine &vm, Register &context,
		     BytecodeBuffer *buffer,
		     Cell *expression, Cell *environment,
		     Cell *constant_pool, Cell *hint,
		     bool is_tail_context)
{
  assert(expression->is_pair());

  VMStackRoot form(vm, expression);
  VMStackRoot env(vm, environment);
  VMStackRoot cpool(vm, constant_pool);
  VMStackRoot hnt(vm, extend_hint(vm, CAR(form), hint));

  size_t len = bytecompile_sequence(vm, context, buffer, CDR(form),
				    env, cpool, hnt, is_tail_context);
  if (len > 0)
    return len;

  return signal_error(vm, "begin: syntax error: ", form);
}


size_t
Delay::emit_bytecode(VirtualMachine &vm, Register &context,
		     BytecodeBuffer *buffer,
		     Cell *expression, Cell *environment,
		     Cell *constant_pool, Cell *hint,
		     bool is_tail_context)
{
  assert(expression->is_pair());

  if (!CDR(expression)->is_pair())
    return signal_error(vm, "delay: syntax error, empty expression: ",
			expression);

  VMStackRoot form(vm, expression);
  VMStackRoot env(vm, environment);
  VMStackRoot cpool(vm, constant_pool);
  VMStackRoot hnt(vm, extend_hint(vm, CAR(form), hint));

  size_t len = Lambda::emit_lambda(vm, context, buffer, Cell::nil(),
				   env, cpool, hnt, Cell::nil(), CDR(form));
  if (len > 0)
    {
      len += buffer->append(OP_PROMISE);

      if (is_tail_context)
	len += buffer->append(OP_APPLY_CONTINUATION0);

      return len;
    }

  return signal_error(vm, "delay: syntax error: ", form);
}


bool
Quasiquote::find_unquote(VirtualMachine &vm, Register &context,
			 Cell *environment, Cell *expression, long nesting)
{
  if (vm.get_bytecompile_tracing(context))
    {
      Port *port = context.output_port->get_port();
      char buf[256];
      snprintf(buf, sizeof(buf), "Quasiquote::find_unquote<nesting=%ld>: ",
	       nesting);
      port->append(buf);
      Writer::write_datum(expression, port, false);
      port->append('\n');
    }

  Cell *unquote;
  if (environment->lookup(vm.core.unquote_symbol, true)
      == vm.core.internal_unquote_syntax)
    unquote = vm.core.unquote_symbol;
  else
    unquote = Cell::nil();

  Cell *unquote_splicing;
  if (environment->lookup(vm.core.unquote_splicing_symbol, true)
      == vm.core.internal_unquote_splicing_syntax)
    unquote_splicing = vm.core.unquote_splicing_symbol;
  else
    unquote_splicing = Cell::nil();

  if ((!expression->is_pair() && !expression->is_vector())
      || (unquote == Cell::nil() && unquote_splicing == Cell::nil()))
    return false;

  Syntax *syntax = lookup_syntax(expression, environment);
  if (syntax == &vm.core.syntax_quasiquote)
    {
      if (CDR(expression)->is_pair() && CDDR(expression) == Cell::nil())
	return find_unquote(vm, context, environment,
			    CADR(expression), nesting + 1);

      return false;	// syntax error
    }

  if (expression->is_pair())
    {
      // list quasiquote template
      Cell *op = CAR(expression);
      if ((unquote != Cell::nil () && op == unquote)
	  || (unquote_splicing != Cell::nil() && op == unquote_splicing))
	{
	  if (nesting == 1)
	    return true;

	  if (CDR(expression)->is_pair()
	      && CDDR(expression) == Cell::nil())
	    return find_unquote(vm, context, environment,
				CADR(expression), nesting - 1);

	  return false;	// syntax error
	}

      while (expression->is_pair())
	{
	  if (find_unquote(vm, context, environment, CAR(expression), nesting))
	    return true;

	  expression = CDR(expression);
	}

      return find_unquote(vm, context, environment, expression, nesting);
    }

  // vector quasiquote template
  long len = expression->get_vector_length();
  for (long i = 0; i < len; i++)
    {
      if (find_unquote(vm, context, environment,
		       expression->get_vector_element(i), nesting))
	return true;
    }

  return false;
}


size_t
Quasiquote::emit_qq_template(VirtualMachine &vm, Register &context,
			     BytecodeBuffer *buffer,
			     Cell *qq_template, long nesting,
			     Cell *environment, Cell *constant_pool,
			     Cell *hint, bool list_qq)
{
  if (vm.get_bytecompile_tracing(context))
    {
      Port *port = context.output_port->get_port();
      char buf[256];
      snprintf(buf, sizeof(buf), "Quasiquote::emit_qq_template<nesting=%ld>: ",
	       nesting);
      port->append(buf);
      Writer::write_datum(qq_template, port, false);
      port->append('\n');
    }

  assert(lookup_syntax(qq_template, environment) == this);
  assert(nesting > 0);

  if (!CDR(qq_template)->is_pair())
    return signal_error(vm, "quasiquote: syntax error: ", qq_template);

  if (CDDR(qq_template) != Cell::nil())
    {
      signal_error(vm, "unexpected form: ", CDDR(qq_template));
      return signal_error(vm, "quasiquote: syntax error: ", qq_template);
    }

  assert(find_unquote(vm, context, environment,
		      CADR(qq_template), nesting + 1));

  VMStackRoot form(vm, qq_template);

  size_t len = emit_list_qq_template(vm, context, buffer, form, nesting + 1,
				     environment, constant_pool, hint,
				     list_qq);

  if (len > 0)
    return len;

  return signal_error(vm, "quasiquote: syntax error: ", form);
}


size_t
Quasiquote::emit_bytecode(VirtualMachine &vm, Register &context,
			  BytecodeBuffer *buffer,
			  Cell *expression, Cell *environment,
			  Cell *constant_pool, Cell *hint,
			  bool is_tail_context)
{
  assert(expression->is_pair());

  if (!CDR(expression)->is_pair())
    return signal_error(vm, "quasiquote: syntax error: ", expression);

  if (CDDR(expression) != Cell::nil())
    {
      signal_error(vm, "unexpected form: ", CDDR(expression));
      return signal_error(vm, "quasiquote: syntax error: ", expression);
    }

  if (vm.get_bytecompile_tracing(context))
    {
      Port *port = context.output_port->get_port();
      port->append("Quasiquote::emit_bytecode: ");
      Writer::write_datum(expression, port, false);
      port->append('\n');
    }

  VMStackRoot form(vm, expression);
  VMStackRoot env(vm, environment);
  VMStackRoot cpool(vm, constant_pool);
  VMStackRoot hnt(vm, extend_hint(vm, CAR(form), hint));

  /*
   * `unquote' and `unquote-splicing' should not be globally available
   * syntaxes, therefore, we define them before entering
   * <quasiquotation 1> context.
   *
   * Note: here we check the existence of binding for `unquote' and
   * `unquote-splicing' not to overwrite pre-bound user defined syntaxes
   * or variables in the environment.
   */
  if (environment->lookup(vm.core.unquote_symbol, true) == Cell::nil())
    environment->bind_syntax(vm.core.internal_unquote_syntax);

  if (environment->lookup(vm.core.unquote_splicing_symbol, true) == Cell::nil())
    environment->bind_syntax(vm.core.internal_unquote_splicing_syntax);

  /*
   * We implement Quasiquote::emit_bytecode() only invoked in
   * <quasiquotation 0> context, the nesting level should always
   * be started with 1 here.
   */
  size_t len = Syntax::emit_qq_template(vm, context, buffer, CADR(form), 1,
					env, cpool, hnt, false);

  /*
   * `unquote' and `unquote-splicing' should not be globally available
   * syntaxes, therefore, we undefine them before leaving the
   * <quasiquotation 1> context.
   */
  environment->unbind_syntax(vm.core.internal_unquote_syntax);
  environment->unbind_syntax(vm.core.internal_unquote_splicing_syntax);

  if (len > 0)
    {
      if (!is_tail_context)
	return len;

      return len + buffer->append(OP_APPLY_CONTINUATION0);
    }

  return signal_error(vm, "quasiquote: syntax error: ", form);
}


bool
Quasiquote::is_quasiquote_syntax()
{
  return true;
}


size_t
Unquote::emit_bytecode(VirtualMachine &vm, Register &context,
		       BytecodeBuffer *buffer,
		       Cell *expression, Cell *environment,
		       Cell *constant_pool, Cell *hint,
		       bool is_tail_context)
{
  // should never be invoked
  return 0;
}


bool
Unquote::is_quasiquote_syntax()
{
  return true;
}


size_t
Unquote::emit_qq_template(VirtualMachine &vm, Register &context,
			  BytecodeBuffer *buffer,
			  Cell *qq_template, long nesting,
			  Cell *environment, Cell *constant_pool,
			  Cell *hint, bool list_qq)
{
  if (vm.get_bytecompile_tracing(context))
    {
      Port *port = context.output_port->get_port();
      char buf[256];
      snprintf(buf, sizeof(buf), "Unquote::emit_qq_template<nesting=%ld>: ",
	       nesting);
      port->append(buf);
      Writer::write_datum(qq_template, port, false);
      port->append('\n');
    }

  assert(lookup_syntax(qq_template, environment) == this);
  assert(nesting > 0);

  if (!CDR(qq_template)->is_pair())
    return signal_error(vm, "unquote: syntax error: ", qq_template);

  if (CDDR(qq_template) != Cell::nil())
    {
      signal_error(vm, "unexpected form: ", CDDR(qq_template));
      return signal_error(vm, "unquote: syntax error: ", qq_template);
    }

  /*
   * `unquote' and `unquote-splicing' should not be globally available
   * syntaxes, therefore, we undefine them before leaving the
   * <quasiquotation 1> context.
   */
  if (nesting == 1)
    {
      environment->unbind_syntax(vm.core.internal_unquote_syntax);
      environment->unbind_syntax(vm.core.internal_unquote_splicing_syntax);
    }
#ifndef NDEBUG
  else
    assert(Quasiquote::find_unquote(vm, context, environment,
				    CADR(qq_template), nesting - 1));
#endif

  VMStackRoot form(vm, qq_template);
  VMStackRoot env(vm, environment);
  VMStackRoot cpool(vm, constant_pool);

  size_t len;
  if (nesting == 1)
    len = bytecompile(vm, context, buffer, CADR(form),
		      env, cpool, hint, false);
  else
    len = emit_list_qq_template(vm, context, buffer, form, nesting - 1,
				env, cpool, hint, false);

  if (env->lookup(vm.core.unquote_symbol, true) == Cell::nil())
    env->bind_syntax(vm.core.internal_unquote_syntax);

  if (env->lookup(vm.core.unquote_splicing_symbol, true) == Cell::nil())
    env->bind_syntax(vm.core.internal_unquote_splicing_syntax);

  if (!list_qq)
    return len;

  if (len > 0)
    {
      unsigned long fp = env->get_current_stack_pointer() - 3;
      assert(fp > 0);

      len += emit_local_variable_assignment(buffer, fp + 1);
      return len + emit_known_procedure_call(vm, context, buffer,
					     vm.core.internal_cons_procedure,
					     env, cpool, fp, 3);
    }

  return signal_error(vm, "unquote: syntax error: ", form);
}


size_t
UnquoteSplicing::emit_bytecode(VirtualMachine &vm, Register &context,
			       BytecodeBuffer *buffer,
			       Cell *expression, Cell *environment,
			       Cell *constant_pool, Cell *hint,
			       bool is_tail_context)
{
  // should never be invoked
  return 0;
}


bool
UnquoteSplicing::is_quasiquote_syntax()
{
  return true;
}


size_t
UnquoteSplicing::emit_qq_template(VirtualMachine &vm, Register &context,
				  BytecodeBuffer *buffer,
				  Cell *qq_template, long nesting,
				  Cell *environment, Cell *constant_pool,
				  Cell *hint, bool list_qq)
{
  if (vm.get_bytecompile_tracing(context))
    {
      Port *port = context.output_port->get_port();
      char buf[256];
      snprintf(buf, sizeof(buf),
	       "UnquoteSplicing::emit_qq_template<nesting=%ld>: ",
	       nesting);
      port->append(buf);
      Writer::write_datum(qq_template, port, false);
      port->append('\n');
    }

  assert(lookup_syntax(qq_template, environment) == this);
  assert(nesting > 0);

  if (!CDR(qq_template)->is_pair())
    return signal_error(vm, "unquote-splicing: syntax error: ", qq_template);

  if (CDDR(qq_template) != Cell::nil())
    {
      signal_error(vm, "unexpected form: ", CDDR(qq_template));
      return signal_error(vm, "unquote-splicing: syntax error: ", qq_template);
    }

  /*
   * `unquote' and `unquote-splicing' should not be globally available
   * syntaxes, therefore, we undefine them before leaving the
   * <quasiquotation 1> context.
   */
  if (nesting == 1)
    {
      environment->unbind_syntax(vm.core.internal_unquote_syntax);
      environment->unbind_syntax(vm.core.internal_unquote_splicing_syntax);
    }
#ifndef NDEBUG
  else
    assert(Quasiquote::find_unquote(vm, context, environment,
				    CADR(qq_template), nesting - 1));
#endif

  VMStackRoot form(vm, qq_template);
  VMStackRoot env(vm, environment);
  VMStackRoot cpool(vm, constant_pool);

  size_t len;
  if (nesting == 1)
    len = bytecompile(vm, context, buffer, CADR(form),
		      env, cpool, hint, false);
  else
    len = emit_list_qq_template(vm, context, buffer, form, nesting - 1,
				env, cpool, hint, false);

  if (env->lookup(vm.core.unquote_symbol, true) == Cell::nil())
    env->bind_syntax(vm.core.internal_unquote_syntax);

  if (env->lookup(vm.core.unquote_splicing_symbol, true) == Cell::nil())
    env->bind_syntax(vm.core.internal_unquote_splicing_syntax);

  if (!list_qq)
    return len;

  if (len > 0)
    {
      unsigned long fp = env->get_current_stack_pointer() - 3;
      assert(fp > 0);

      len += emit_local_variable_assignment(buffer, fp + 1);
      return len + emit_known_procedure_call(vm, context, buffer,
					     vm.core.internal_append_procedure,
					     env, cpool, fp, 3);
    }

  return signal_error(vm, "unquote-splicing: syntax error: ", form);
}

}
