/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: r5rs_list_procedure.cc,v 1.21 2010/08/15 15:39:47 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libtscore/r5rs_list_procedure.h"

#include "libtscore/bytecode_buffer.h"
#include "libtscore/tscore.h"
#include "libtscore/virtual_machine.h"


namespace LibTSCore
{

NativeProcedure::Info R5RSListProcedure::procedures[] = {
  { &R5RSListProcedure::is_pair,
    "pair?", TST_NONE, 1, 1 },
  { &R5RSListProcedure::is_null,
    "null?", TST_NONE, 1, 1 },
  { &R5RSListProcedure::is_list,
    "list?", TST_NONE, 1, 1 },
  { &R5RSListProcedure::car,
    "car", TST_PAIR, 1, 1 },
  { &R5RSListProcedure::cdr,
    "cdr", TST_PAIR, 1, 1 },
  { &R5RSListProcedure::cons,
    "cons", TST_ANY, 2, 2 },
  { &R5RSListProcedure::set_car,
    "set-car!", TST_MUTABLE_PAIR TST_ANY, 2, 2 },
  { &R5RSListProcedure::set_cdr,
    "set-cdr!", TST_MUTABLE_PAIR TST_ANY, 2, 2 },
  { &R5RSListProcedure::list,
    "list", TST_NONE, 0, INF_ARG },
  { &R5RSListProcedure::length,
    "length", TST_PAIR, 1, 1 },
  { &R5RSListProcedure::append,
    "append", TST_ANY, 1, INF_ARG },
  { &R5RSListProcedure::reverse,
    "reverse", TST_PAIR, 1, 1 },
  { &R5RSListProcedure::list_tail,
    "list-tail", TST_PAIR TST_NONNEG, 2, 2 },
  { &R5RSListProcedure::list_ref,
    "list-ref", TST_PAIR TST_NONNEG, 2, 2 },

  { &R5RSListProcedure::memq,
    "memq", TST_ANY TST_ANY, 2, 2 },
  { &R5RSListProcedure::memv,
    "memv", TST_ANY TST_ANY, 2, 2 },
  { &R5RSListProcedure::member,
    "member", TST_ANY TST_ANY, 2, 2 },

  { &R5RSListProcedure::assq,
    "assq", TST_ANY TST_ANY, 2, 2 },
  { &R5RSListProcedure::assv,
    "assv", TST_ANY TST_ANY, 2, 2 },
  { &R5RSListProcedure::assoc,
    "assoc", TST_ANY TST_ANY, 2, 2 },

  { &R5RSListProcedure::is_vector,
    "vector?", TST_NONE, 1, 1 },
  { &R5RSListProcedure::make_vector,
    "make-vector", TST_NONNEG TST_ANY, 1, 2 },
  { &R5RSListProcedure::vector,
    "vector", TST_ANY, 0, INF_ARG },
  { &R5RSListProcedure::vector_length,
    "vector-length", TST_VECTOR, 1, 1 },
  { &R5RSListProcedure::vector_ref,
    "vector-ref", TST_VECTOR TST_NONNEG, 2, 2 },
  { &R5RSListProcedure::vector_set,
    "vector-set!", TST_MUTABLE_VECTOR TST_NONNEG TST_ANY, 3, 3 },
  { &R5RSListProcedure::vector2list,
    "vector->list", TST_VECTOR, 1, 1 },
  { &R5RSListProcedure::list2vector,
    "list->vector", TST_PAIR, 1, 1 },
  { &R5RSListProcedure::vector_fill,
    "vector-fill!", TST_MUTABLE_VECTOR TST_ANY, 2, 2 }
};


struct BytecodeProcedureInfo
{
  const char *name;
  long frame_size;
  long min_arity;
  long max_arity;
  size_t bytecode_length;
  const char *bytecode;
};


static BytecodeProcedureInfo bytecode_procedures[] = {
  { "caar", 4, 2, 2, 5,
    OPS_LOAD_L1 "\001" OPS_CAR OPS_CAR OPS_APPLY_CONTINUATION0 },
  { "cdar", 4, 2, 2, 5,
    OPS_LOAD_L1 "\001" OPS_CAR OPS_CDR OPS_APPLY_CONTINUATION0 },
  { "cadr", 4, 2, 2, 5,
    OPS_LOAD_L1 "\001" OPS_CDR OPS_CAR OPS_APPLY_CONTINUATION0 },
  { "cddr", 4, 2, 2, 5,
    OPS_LOAD_L1 "\001" OPS_CDR OPS_CDR OPS_APPLY_CONTINUATION0 },
  { "caaar", 4, 2, 2, 6,
    OPS_LOAD_L1 "\001" OPS_CAR OPS_CAR OPS_CAR OPS_APPLY_CONTINUATION0 },
  { "cadar", 4, 2, 2, 6,
    OPS_LOAD_L1 "\001" OPS_CAR OPS_CDR OPS_CAR OPS_APPLY_CONTINUATION0 },
  { "caadr", 4, 2, 2, 6,
    OPS_LOAD_L1 "\001" OPS_CDR OPS_CAR OPS_CAR OPS_APPLY_CONTINUATION0 },
  { "caddr", 4, 2, 2, 6,
    OPS_LOAD_L1 "\001" OPS_CDR OPS_CDR OPS_CAR OPS_APPLY_CONTINUATION0 },
  { "cdaar", 4, 2, 2, 6,
    OPS_LOAD_L1 "\001" OPS_CAR OPS_CAR OPS_CDR OPS_APPLY_CONTINUATION0 },
  { "cddar", 4, 2, 2, 6,
    OPS_LOAD_L1 "\001" OPS_CAR OPS_CDR OPS_CDR OPS_APPLY_CONTINUATION0 },
  { "cdadr", 4, 2, 2, 6,
    OPS_LOAD_L1 "\001" OPS_CDR OPS_CAR OPS_CDR OPS_APPLY_CONTINUATION0 },
  { "cdddr", 4, 2, 2, 6,
    OPS_LOAD_L1 "\001" OPS_CDR OPS_CDR OPS_CDR OPS_APPLY_CONTINUATION0 },
  { "caaaar", 4, 2, 2, 7,
    OPS_LOAD_L1 "\001" OPS_CAR OPS_CAR OPS_CAR OPS_CAR OPS_APPLY_CONTINUATION0 },
  { "caadar", 4, 2, 2, 7,
    OPS_LOAD_L1 "\001" OPS_CAR OPS_CDR OPS_CAR OPS_CAR OPS_APPLY_CONTINUATION0 },
  { "caaadr", 4, 2, 2, 7,
    OPS_LOAD_L1 "\001" OPS_CDR OPS_CAR OPS_CAR OPS_CAR OPS_APPLY_CONTINUATION0 },
  { "caaddr", 4, 2, 2, 7,
    OPS_LOAD_L1 "\001" OPS_CDR OPS_CDR OPS_CAR OPS_CAR OPS_APPLY_CONTINUATION0 },
  { "cadaar", 4, 2, 2, 7,
    OPS_LOAD_L1 "\001" OPS_CAR OPS_CAR OPS_CDR OPS_CAR OPS_APPLY_CONTINUATION0 },
  { "caddar", 4, 2, 2, 7,
    OPS_LOAD_L1 "\001" OPS_CAR OPS_CDR OPS_CDR OPS_CAR OPS_APPLY_CONTINUATION0 },
  { "cadadr", 4, 2, 2, 7,
    OPS_LOAD_L1 "\001" OPS_CDR OPS_CAR OPS_CDR OPS_CAR OPS_APPLY_CONTINUATION0 },
  { "cadddr", 4, 2, 2, 7,
    OPS_LOAD_L1 "\001" OPS_CDR OPS_CDR OPS_CDR OPS_CAR OPS_APPLY_CONTINUATION0 },
  { "cdaaar", 4, 2, 2, 7,
    OPS_LOAD_L1 "\001" OPS_CAR OPS_CAR OPS_CAR OPS_CDR OPS_APPLY_CONTINUATION0 },
  { "cdadar", 4, 2, 2, 7,
    OPS_LOAD_L1 "\001" OPS_CAR OPS_CDR OPS_CAR OPS_CDR OPS_APPLY_CONTINUATION0 },
  { "cdaadr", 4, 2, 2, 7,
    OPS_LOAD_L1 "\001" OPS_CDR OPS_CAR OPS_CAR OPS_CDR OPS_APPLY_CONTINUATION0 },
  { "cdaddr", 4, 2, 2, 7,
    OPS_LOAD_L1 "\001" OPS_CDR OPS_CDR OPS_CAR OPS_CDR OPS_APPLY_CONTINUATION0 },
  { "cddaar", 4, 2, 2, 7,
    OPS_LOAD_L1 "\001" OPS_CAR OPS_CAR OPS_CDR OPS_CDR OPS_APPLY_CONTINUATION0 },
  { "cdddar", 4, 2, 2, 7,
    OPS_LOAD_L1 "\001" OPS_CAR OPS_CDR OPS_CDR OPS_CDR OPS_APPLY_CONTINUATION0 },
  { "cddadr", 4, 2, 2, 7,
    OPS_LOAD_L1 "\001" OPS_CDR OPS_CAR OPS_CDR OPS_CDR OPS_APPLY_CONTINUATION0 },
  { "cddddr", 4, 2, 2, 7,
    OPS_LOAD_L1 "\001" OPS_CDR OPS_CDR OPS_CDR OPS_CDR OPS_APPLY_CONTINUATION0 }
};


Cell *
R5RSListProcedure::assign_procedures(Cell *environment, Cell *frame)
{
  TSStackRoot env(core, environment);
  TSStackRoot frm(core, frame);
  size_t n_procs = sizeof(procedures) / sizeof(Info);
  for (size_t i = 0; i < n_procs; i++)
    assign_procedure(env, frm, &procedures[i], this);

  n_procs = sizeof(bytecode_procedures) / sizeof(BytecodeProcedureInfo);
  for (size_t i = 0; i < n_procs; i++)
    assign_bytecode_procedure(env, frm, &bytecode_procedures[i]);

  return env;
}


void
R5RSListProcedure::assign_bytecode_procedure(Cell *environment, Cell *frame,
					     const BytecodeProcedureInfo *info)
{
  TSStackRoot env(core, environment);
  TSStackRoot frm(core, frame);

  BytecodeBuffer *bytecode = new BytecodeBuffer();
  assert(strlen(info->bytecode) == info->bytecode_length);

  for (size_t i = 0; i < info->bytecode_length; i++)
    bytecode->append(static_cast<TSVMOpcode>(info->bytecode[i]));

  BytecodeProcedure *proc
    = new BytecodeProcedure(bytecode, info->frame_size,
			    info->min_arity, info->max_arity);
  TSStackRoot tmpl(core, core.mk_closure_template(Cell::nil0(), Cell::nil0(),
						  proc));
  TSStackRoot closure(core, core.mk_closure(tmpl, Cell::nil0()));
  Cell *symbol = core.mk_static_symbol(info->name);
  const Cell *binding = env->bind_variable(symbol);
  frm->store_toplevel_variable(env, binding->get_slot_location(), closure);
}


bool
R5RSListProcedure::is_list(Cell *cell)
{
  Cell *fast = cell;
  Cell *slow = cell;
  while (true)
    {
      if (!fast->is_pair())
	return fast == Cell::nil();

      fast = CDR(fast);
      if (!fast->is_pair())
	return fast == Cell::nil();

      fast = CDR(fast);
      slow = CDR(slow);
      if (fast == slow)
	{
	  /* the fast (Cell *) has looped back around and caught up
	     with the slow (Cell *), hence the structure is circular,
	     not of finite length, and therefore not a list */
	  return false;
	}
    }
}


long
R5RSListProcedure::list_length(Cell *list)
{
  long len = 0;
  for (len = 0; list->is_pair(); list = CDR(list))
    len++;

  if (list == Cell::nil())
    return len;

  return -1;
}


Cell *
R5RSListProcedure::list_append(MemorySystem &memory_system, Cell *a, Cell *b)
{
  Cell *p = b;
  Cell *q;

  if (a != Cell::nil())
    {
      a = list_reverse(memory_system, a);
      if (a == NULL)
	return NULL;	// error

      while (a != Cell::nil())
        {
          q = CDR(a);
          memory_system.set_cdr(a, p);
          p = a;
          a = q;
        }
    }

  return p;
}


#define CAR0(p)			((p)->car())
#define CDR0(p)			((p)->cdr())


Cell *
R5RSListProcedure::list_reverse(MemorySystem &memory_system, Cell *&list)
{
  Cell *p = Cell::nil();
  StackRoot cur(memory_system, list);
  for (; cur->is_pair(); cur = CDR(cur))
    p = memory_system.get_cons(CAR0(cur), p);

  if (cur == Cell::nil())
    return p;
  return NULL;
}


Cell *
R5RSListProcedure::list_reverse_in_place(MemorySystem &memory_system,
					 Cell *term, Cell *list,
					 bool set_immutable)
{
  Cell *p = list;
  Cell *result = term;
  Cell *q;

  if (set_immutable)
    {
      while (p != Cell::nil())
	{
	  if (!p->is_pair())
	    return NULL;	// error

	  q = CDR(p);
	  memory_system.set_cdr(p, result)->set_immutable();
	  result = p;
	  p = q;
	}
    }
  else
    {
      while (p != Cell::nil())
	{
	  if (!p->is_pair())
	    return NULL;	// error

	  q = CDR(p);
	  memory_system.set_cdr(p, result);
	  result = p;
	  p = q;
	}
    }

  return result;
}


#define ARG(n)	context.frame->load_variable(argument_pointer + (n))


Cell *
R5RSListProcedure::is_pair(VirtualMachine &vm, Register &context,
			   unsigned long argument_pointer,
			   unsigned long nargs, void *data)
{
  return Cell::ts_bool(ARG(0)->is_pair());
}


Cell *
R5RSListProcedure::is_null(VirtualMachine &vm, Register &context,
			   unsigned long argument_pointer,
			   unsigned long nargs, void *data)
{
  return Cell::ts_bool(ARG(0) == Cell::nil());
}


Cell *
R5RSListProcedure::is_list(VirtualMachine &vm, Register &context,
			   unsigned long argument_pointer,
			   unsigned long nargs, void *data)
{
  return Cell::ts_bool(is_list(ARG(0)));
}


Cell *
R5RSListProcedure::car(VirtualMachine &vm, Register &context,
		       unsigned long argument_pointer,
		       unsigned long nargs, void *data)
{
  return CAR(ARG(0));
}


Cell *
R5RSListProcedure::cdr(VirtualMachine &vm, Register &context,
		       unsigned long argument_pointer,
		       unsigned long nargs, void *data)
{
  return CDR(ARG(0));
}


Cell *
R5RSListProcedure::cons(VirtualMachine &vm, Register &context,
			unsigned long argument_pointer,
			unsigned long nargs, void *data)
{
  Cell *a = ARG(0);
  Cell *d = ARG(1);
  return vm.cons(a, d);
}


Cell *
R5RSListProcedure::set_car(VirtualMachine &vm, Register &context,
			   unsigned long argument_pointer,
			   unsigned long nargs, void *data)
{
  vm.set_car(ARG(0), ARG(1));
  return Cell::unspecified();
}


Cell *
R5RSListProcedure::set_cdr(VirtualMachine &vm, Register &context,
			   unsigned long argument_pointer,
			   unsigned long nargs, void *data)
{
  vm.set_cdr(ARG(0), ARG(1));
  return Cell::unspecified();
}


Cell *
R5RSListProcedure::list(VirtualMachine &vm, Register &context,
			unsigned long argument_pointer,
			unsigned long nargs, void *data)
{
  VMStackRoot result(vm, Cell::nil());
  while (nargs > 0)
    {
      Cell *elm = ARG(--nargs);
      result = vm.cons(elm, result);
    }

  return result;
}


Cell *
R5RSListProcedure::length(VirtualMachine &vm, Register &context,
			  unsigned long argument_pointer,
			  unsigned long nargs, void *data)
{
  long len = list_length(ARG(0));
  if (len >= 0)
    return vm.mk_integer(len);

  return signal_error(vm, "length: not a list:", ARG(0));
}


Cell *
R5RSListProcedure::append(VirtualMachine &vm, Register &context,
			  unsigned long argument_pointer,
			  unsigned long nargs, void *data)
{
  Cell *result = ARG(0);
  for (unsigned long i = 1; i < nargs; i++)
    {
      result = list_append(vm.get_memory_system(), result, ARG(i));
      if (result == NULL)
	{
	  char buf[256];
	  snprintf(buf, sizeof(buf),
		   "append: argument #%ld must be of list: ", i);
	  return signal_error(vm, buf, ARG(i - 1));
	}
    }

  return result;
}


Cell *
R5RSListProcedure::reverse(VirtualMachine &vm, Register &context,
			   unsigned long argument_pointer,
			   unsigned long nargs, void *data)
{
  Cell *list = ARG(0);
  return list_reverse(vm.get_memory_system(), list);
}


Cell *
R5RSListProcedure::list_tail(VirtualMachine &vm, Register &context,
			     unsigned long argument_pointer,
			     unsigned long nargs, void *data)
{
  Cell *list = ARG(0);
  long k = ARG(1)->get_integer0();
  while (CDR(list) != Cell::nil() && k > 0)
    {
      list = CDR(list);
      --k;
    }

  if (k == 0)
    return list;

  return signal_error(vm, "list-tail: list has elements fewer than:", ARG(1));
}


Cell *
R5RSListProcedure::list_ref(VirtualMachine &vm, Register &context,
			    unsigned long argument_pointer,
			    unsigned long nargs, void *data)
{
  Cell *list = ARG(0);
  long k = ARG(1)->get_integer0();
  while (CDR(list) != Cell::nil() && k > 0)
    {
      list = CDR(list);
      --k;
    }

  if (k == 0 && list->is_pair())
    return CAR(list);

  return signal_error(vm, "list-ref: list has elements fewer than:", ARG(1));
}


Cell *
R5RSListProcedure::memq(VirtualMachine &vm, Register &context,
			unsigned long argument_pointer,
			unsigned long nargs, void *data)
{
  Cell *key = ARG(0);
  Cell *fast;
  Cell *slow;
  slow = fast = ARG(1);
  while (true)
    {
      if (!fast->is_pair())
	{
	  if (fast == Cell::nil())
	    return Cell::f();

	  return signal_error(vm, "memq: argument #2 must be of: list", ARG(1));
	}

      if (CAR(fast) == key)
	return Cell::t();

      fast = CDR(fast);
      if (!fast->is_pair())
	{
	  if (fast == Cell::nil())
	    return Cell::f();

	  return signal_error(vm, "memq: argument #2 must be of: list", ARG(1));
	}

      if (CAR(fast) == key)
	return Cell::t();

      fast = CDR(fast);
      slow = CDR(slow);
      if (fast == slow)
	return signal_error(vm, "memq: argument #2 must be of: list", ARG(1));
    }
}


Cell *
R5RSListProcedure::memv(VirtualMachine &vm, Register &context,
			unsigned long argument_pointer,
			unsigned long nargs, void *data)
{
  Cell *key = ARG(0);
  Cell *fast;
  Cell *slow;
  slow = fast = ARG(1);
  while (true)
    {
      if (!fast->is_pair())
	{
	  if (fast == Cell::nil())
	    return Cell::f();

	  return signal_error(vm, "memv: argument #2 must be of: list", ARG(1));
	}

      if (R5RSMiscProcedure::eqv(CAR(fast), key))
	return Cell::t();

      fast = CDR(fast);
      if (!fast->is_pair())
	{
	  if (fast == Cell::nil())
	    return Cell::f();

	  return signal_error(vm, "memv: argument #2 must be of: list", ARG(1));
	}

      if (R5RSMiscProcedure::eqv(CAR(fast), key))
	return Cell::t();

      fast = CDR(fast);
      slow = CDR(slow);
      if (fast == slow)
	return signal_error(vm, "memv: argument #2 must be of: list", ARG(1));
    }
}


Cell *
R5RSListProcedure::member(VirtualMachine &vm, Register &context,
			  unsigned long argument_pointer,
			  unsigned long nargs, void *data)
{
  Cell *key = ARG(0);
  Cell *fast;
  Cell *slow;
  slow = fast = ARG(1);
  while (true)
    {
      if (!fast->is_pair())
	{
	  if (fast == Cell::nil())
	    return Cell::f();

	  return signal_error(vm,
			      "member: argument #2 must be of: list", ARG(1));
	}

      if (R5RSMiscProcedure::equal(CAR(fast), key))
	return Cell::t();

      fast = CDR(fast);
      if (!fast->is_pair())
	{
	  if (fast == Cell::nil())
	    return Cell::f();

	  return signal_error(vm,
			      "member: argument #2 must be of: list", ARG(1));
	}

      if (R5RSMiscProcedure::equal(CAR(fast), key))
	return Cell::t();

      fast = CDR(fast);
      slow = CDR(slow);
      if (fast == slow)
	return signal_error(vm, "member: argument #2 must be of: list", ARG(1));
    }
}



Cell *
R5RSListProcedure::assq(VirtualMachine &vm, Register &context,
			unsigned long argument_pointer,
			unsigned long nargs, void *data)
{
  Cell *key = ARG(0);
  Cell *fast;
  Cell *slow;
  slow = fast = ARG(1);
  while (true)
    {
      if (!fast->is_pair())
	{
	  if (fast == Cell::nil())
	    return Cell::f();

	  return signal_error(vm,
			      "assq: argument #2 must be of alist: ", ARG(1));
	}

      if (CAR(fast)->is_pair())
	{
	  if (CAAR(fast) == key)
	    return CAR(fast);
	}
      else
	return signal_error(vm, "assq: argument #2 must be of alist: ", ARG(1));

      fast = CDR(fast);
      if (!fast->is_pair())
	{
	  if (fast == Cell::nil())
	    return Cell::f();

	  return signal_error(vm,
			      "assq: argument #2 must be of alist: ", ARG(1));
	}

      if (CAR(fast)->is_pair())
	{
	  if (CAAR(fast) == key)
	    return CAR(fast);
	}
      else
	return signal_error(vm, "assq: argument #2 must be of alist: ", ARG(1));

      fast = CDR(fast);
      slow = CDR(slow);
      if (fast == slow)
	return signal_error(vm, "assq: argument #2 must be of alist: ", ARG(1));
    }
}


Cell *
R5RSListProcedure::assv(VirtualMachine &vm, Register &context,
			unsigned long argument_pointer,
			unsigned long nargs, void *data)
{
  Cell *key = ARG(0);
  Cell *fast;
  Cell *slow;
  slow = fast = ARG(1);
  while (true)
    {
      if (!fast->is_pair())
	{
	  if (fast == Cell::nil())
	    return Cell::f();

	  return signal_error(vm,
			      "assv: argument #2 must be of alist: ", ARG(1));
	}

      if (CAR(fast)->is_pair())
	{
	  if (R5RSMiscProcedure::eqv(CAAR(fast), key))
	    return CAR(fast);
	}
      else
	return signal_error(vm, "assv: argument #2 must be of alist: ", ARG(1));

      fast = CDR(fast);
      if (!fast->is_pair())
	{
	  if (fast == Cell::nil())
	    return Cell::f();

	  return signal_error(vm,
			      "assv: argument #2 must be of alist: ", ARG(1));
	}

      if (CAR(fast)->is_pair())
	{
	  if (R5RSMiscProcedure::eqv(CAAR(fast), key))
	    return CAR(fast);
	}
      else
	return signal_error(vm, "assv: argument #2 must be of alist: ", ARG(1));

      fast = CDR(fast);
      slow = CDR(slow);
      if (fast == slow)
	return signal_error(vm, "assv: argument #2 must be of alist: ", ARG(1));
    }
}


Cell *
R5RSListProcedure::assoc(VirtualMachine &vm, Register &context,
			 unsigned long argument_pointer,
			 unsigned long nargs, void *data)
{
  Cell *key = ARG(0);
  Cell *fast;
  Cell *slow;
  slow = fast = ARG(1);
  while (true)
    {
      if (!fast->is_pair())
	{
	  if (fast == Cell::nil())
	    return Cell::f();

	  return signal_error(vm,
			      "assoc: argument #2 must be of alist: ", ARG(1));
	}

      if (CAR(fast)->is_pair())
	{
	  if (R5RSMiscProcedure::equal(CAAR(fast), key))
	    return CAR(fast);
	}
      else
	return signal_error(vm,
			    "assoc: argument #2 must be of alist: ", ARG(1));

      fast = CDR(fast);
      if (!fast->is_pair())
	{
	  if (fast == Cell::nil())
	    return Cell::f();

	  return signal_error(vm,
			      "assoc: argument #2 must be of alist: ", ARG(1));
	}

      if (CAR(fast)->is_pair())
	{
	  if (R5RSMiscProcedure::equal(CAAR(fast), key))
	    return CAR(fast);
	}
      else
	return signal_error(vm,
			    "assoc: argument #2 must be of alist: ", ARG(1));

      fast = CDR(fast);
      slow = CDR(slow);
      if (fast == slow)
	return signal_error(vm,
			    "assoc: argument #2 must be of alist: ", ARG(1));
    }
}


Cell *
R5RSListProcedure::is_vector(VirtualMachine &vm, Register &context,
			     unsigned long argument_pointer,
			     unsigned long nargs, void *data)
{
  return Cell::ts_bool(ARG(0)->is_vector());
}


Cell *
R5RSListProcedure::make_vector(VirtualMachine &vm, Register &context,
			       unsigned long argument_pointer,
			       unsigned long nargs, void *data)
{
  long len = ARG(0)->get_integer0();

  if (nargs == 1)
    return vm.mk_vector(len);

  Cell *fill = ARG(1);
  return vm.mk_vector(len, fill);
}


Cell *
R5RSListProcedure::vector(VirtualMachine &vm, Register &context,
			  unsigned long argument_pointer,
			  unsigned long nargs, void *data)
{
  Cell *result = vm.mk_vector(nargs);
  for (unsigned long i = 0; i < nargs; i++)
    vm.set_vector_element(result, i, ARG(i));

  return result;
}


Cell *
R5RSListProcedure::vector_length(VirtualMachine &vm, Register &context,
				 unsigned long argument_pointer,
				 unsigned long nargs, void *data)
{
  return vm.mk_integer(ARG(0)->get_vector_length());
}


Cell *
R5RSListProcedure::vector_ref(VirtualMachine &vm, Register &context,
			      unsigned long argument_pointer,
			      unsigned long nargs, void *data)
{
  long index = ARG(1)->get_integer0();
  if (index < ARG(0)->get_vector_length())
    return ARG(0)->get_vector_element(index);

  return signal_error(vm, "vector-ref: index out of bounds:", ARG(1));
}


Cell *
R5RSListProcedure::vector_set(VirtualMachine &vm, Register &context,
			      unsigned long argument_pointer,
			      unsigned long nargs, void *data)
{
  long index = ARG(1)->get_integer();
  if (index < ARG(0)->get_vector_length())
    {
      vm.set_vector_element(ARG(0), index, ARG(2));
      return ARG(2);
    }

  return signal_error(vm, "vector-set!: index out of bounds:", ARG(1));
}


Cell *
R5RSListProcedure::vector2list(VirtualMachine &vm, Register &context,
			       unsigned long argument_pointer,
			       unsigned long nargs, void *data)
{
  VMStackRoot vector(vm, ARG(0));
  long len = vector->get_vector_length();
  Cell *list = Cell::nil();
  for (long i = 0; i < len; i++)
    {
      Cell *elm = vector->get_vector_element(i);
      list = vm.cons(elm, list);
    }

  return list_reverse_in_place(vm.get_memory_system(),
			       Cell::nil(), list, false);
}


Cell *
R5RSListProcedure::list2vector(VirtualMachine &vm, Register &context,
			       unsigned long argument_pointer,
			       unsigned long nargs, void *data)
{
  VMStackRoot list(vm, ARG(0));
  long len = list_length(list);
  if (len < 0)
    return signal_error(vm, "length->vector: not a list:", list);

  Cell *vector = vm.mk_vector(len);
  for (long i = 0; i < len; i++)
    {
      vm.set_vector_element(vector, i, CAR(list));
      list = CDR(list);
    }

  return vector;
}


Cell *
R5RSListProcedure::vector_fill(VirtualMachine &vm, Register &context,
			       unsigned long argument_pointer,
			       unsigned long nargs, void *data)
{
  Cell *vector = ARG(0);
  Cell *fill = ARG(1);
  long len = vector->get_vector_length();
  for (long i = 0; i < len; i++)
    vm.set_vector_element(vector, i, fill);

  return vector;
}

}
