/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: r5rs_macro_syntax.cc,v 1.13 2009/01/01 07:47:00 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libtscore/r5rs_macro_syntax.h"

#include "libtscore/bytecode_buffer.h"
#include "libtscore/virtual_machine.h"
#include "libtscore/syntax.h"

#include <algorithm>
#include <vector>

#include <cassert>


#define ENABLE_MACRO_KEYWORD_CHECK	1


namespace LibTSCore
{

size_t
DefineSyntax::emit_bytecode(VirtualMachine &vm, Register &context,
			    BytecodeBuffer *buffer,
			    Cell *expression, Cell *environment,
			    Cell *constant_pool, Cell *hint,
			    bool is_tail_context)
{
  assert(expression->is_pair());

  if (!is_toplevel(vm, environment, hint))
    return signal_error(vm, "define-syntax: only allowed at toplevel: ",
			expression);

  Cell *tmp_env = environment;
  while (true)
    {
      Cell *parent = tmp_env->get_parent_environment();
      if (parent == Cell::nil())
	{
	  if (tmp_env == vm.core.get_interaction_environment())
	    break;

	  return signal_error(vm, "define-syntax: in immutable environment: ",
			      expression);
	}

      tmp_env = parent;
    }

  VMStackRoot toplevel_environment(vm, tmp_env);

  if (!CDR(expression)->is_pair()
      || !CDDR(expression)->is_pair())
    return signal_error(vm, "define-syntax: syntax error: ", expression);

  if (CDDDR(expression) != Cell::nil())
    {
      signal_error(vm, "unexpected form: ", CDDDR(expression));
      return signal_error(vm, "define-syntax: syntax error: ", expression);
    }

  VMStackRoot keyword(vm, CADR(expression));
  if (!keyword->is_symbol())
    {
      signal_error(vm, "not a symbol: ", keyword);
      return signal_error(vm, "define-syntax: syntax error: ", expression);
    }

  VMStackRoot form(vm, expression);
  VMStackRoot env(vm, environment);
  VMStackRoot cpool(vm, constant_pool);
  const Cell *transformer = SyntaxRules::parse(vm, keyword, CADDR(form),
					       toplevel_environment);
  if (transformer == NULL)
    return signal_error(vm, "define-syntax: syntax error: ", form);

  VMStackRoot syntax(vm, vm.core.mk_macro_syntax(keyword, transformer));
  toplevel_environment->bind_syntax(syntax);

  size_t len = emit_literal_expression(vm, context, buffer, syntax, cpool);
  assert(len > 0);

  if (!is_tail_context)
    return len;

  return len + buffer->append(OP_APPLY_CONTINUATION0);
}


size_t
LetSyntax::emit_bytecode(VirtualMachine &vm, Register &context,
			 BytecodeBuffer *buffer,
			 Cell *expression, Cell *environment,
			 Cell *constant_pool, Cell *hint,
			 bool is_tail_context)
{
  return signal_error(vm, "let-syntax: not implemented yet");
}


size_t
LetrecSyntax::emit_bytecode(VirtualMachine &vm, Register &context,
			    BytecodeBuffer *buffer,
			    Cell *expression, Cell *environment,
			    Cell *constant_pool, Cell *hint,
			    bool is_tail_context)
{
  VMStackRoot form(vm, expression);
  VMStackRoot env(vm, environment);
  VMStackRoot cpool(vm, constant_pool);
  VMStackRoot hnt(vm, extend_hint(vm, CAR(form), hint));

  if (!CDR(form)->is_pair() || !Letrec::check_bindings(vm, CADR(form), hnt))
    return signal_error(vm, "letrec-syntax: syntax error: ", form);

  Cell *nil = Cell::nil();
  VMStackRoot tmp_bindings(vm, CADR(form));
  VMStackRoot new_bindings(vm, nil);
  VMStackRoot keyword(vm, nil);
  while (tmp_bindings->is_pair())
    {
      Cell *binding = CAR(tmp_bindings);
      keyword = CAR(binding);
      if (keyword->is_binding())
	keyword = const_cast<Cell *>(keyword->get_bound_symbol());
      else if (keyword->is_syntax())
	keyword = const_cast<Cell *>(keyword->get_syntax_symbol());
#ifndef NDEBUG
      else
	{
	  if (!keyword->is_symbol())
	    signal_error(vm, "letrec-syntax: what is this?: ", keyword);

	  assert(keyword->is_symbol());
	}
#endif
      const Cell *transformer = SyntaxRules::parse(vm,
						   keyword, CADR(binding),
						   env);
      if (transformer == NULL)
	return signal_error(vm, "letrec-syntax: syntax error: ", form);

      binding = vm.core.mk_macro_syntax(keyword, transformer);
      new_bindings = vm.cons(binding, new_bindings);
      tmp_bindings = CDR(tmp_bindings);
    }

  VMStackRoot old_bindings(vm, nil);
  tmp_bindings = new_bindings;
  while (tmp_bindings->is_pair())
    {
      Cell *syntax = CAR(tmp_bindings);
      keyword = const_cast<Cell *>(syntax->get_syntax_symbol());
      Cell *binding = const_cast<Cell *>(env->lookup(keyword, false));
      if (binding == nil)
	binding = vm.cons(keyword, nil);

      old_bindings = vm.cons(binding, old_bindings);
      environment->bind_syntax(CAR(tmp_bindings));
      tmp_bindings = CDR(tmp_bindings);
    }

  size_t len = bytecompile_sequence(vm, context, buffer, CDDR(form),
				    env, cpool, hnt, is_tail_context);

  tmp_bindings = old_bindings;
  while (tmp_bindings->is_pair())
    {
      Cell *binding = CAR(tmp_bindings);
      env->rebind_binding(binding);
      tmp_bindings = CDR(tmp_bindings);
    }

  if (len > 0)
    return len;

  return 0;	// error
}


const Cell *
SyntaxRules::parse(VirtualMachine &vm, Cell *keyword, Cell *expression,
		   Cell *environment)
{
  if (!expression->is_pair() || CAR(expression) != vm.core.syntax_rules_symbol)
    {
      Syntax::signal_error(vm, "not a syntax-rules: ", CAR(expression));
      return NULL;
    }

  if (environment->lookup(vm.core.syntax_rules_symbol, true) != Cell::nil())
    return NULL;	// syntax-rules have bound to something meaningful!

  if (CDR(expression) == Cell::nil())
    {
      Syntax::signal_error(vm, "empty syntax-rules: ", expression);
      return NULL;
    }

  if (!CADR(expression)->is_pair() && CADR(expression) != Cell::nil())
    {
      Syntax::signal_error(vm, "literals must be of list: ", CADR(expression));
      Syntax::signal_error(vm, "syntax-rules: syntax error: ", expression);
      return NULL;
    }

  VMStackRoot env(vm, environment);
  VMStackRoot form(vm, expression);
  VMStackRoot non_pattern_variables(vm, Cell::nil());
  non_pattern_variables = vm.cons(keyword, non_pattern_variables);
  non_pattern_variables = vm.cons(vm.core.ellipsis_symbol,
				  non_pattern_variables);

  // Checks for <literals>
  VMStackRoot list(vm, CADR(form));
  while (list->is_pair())
    {
      Cell *id = CAR(list);
      if (!id->is_symbol())
	{
	  Syntax::signal_error(vm, "not a symbol: ", CAR(list));
	  Syntax::signal_error(vm, "syntax-rules: syntax error: ", form);
	  return NULL;
	}

      if (id == vm.core.ellipsis_symbol)
	{
	  Syntax::signal_error(vm, "wrong literal identifier: ", CAR(list));
	  Syntax::signal_error(vm, "syntax-rules: syntax error: ", form);
	  return NULL;
	}

      non_pattern_variables = vm.cons(id, non_pattern_variables);
      list = CDR(list);
    }

  if (list != Cell::nil())
    {
      Syntax::signal_error(vm, "literals must be of proper list: ",
			   CADR(form));
      Syntax::signal_error(vm, "syntax-rules: syntax error: ", form);
      return NULL;
    }

  if (CDDR(form) == Cell::nil())
    {
      Syntax::signal_error(vm, "no syntax rule: ", form);
      return NULL;
    }

  VMStackRoot rules(vm, CDDR(form));
  VMStackRoot internal_rules(vm, Cell::nil());
  while (rules->is_pair())
    {
      Cell *result = parse_rule(vm, CAR(rules),
				keyword, non_pattern_variables);
      if (result == NULL)
	{
	  Syntax::signal_error(vm, "wrong syntax rule: ", CAR(rules));
	  Syntax::signal_error(vm, "syntax-rules: syntax error: ", form);
	  return NULL;
	}

      internal_rules = vm.cons(result, internal_rules);
      rules = CDR(rules);
    }

  if (rules != Cell::nil())
    {
      Syntax::signal_error(vm, "not a proper list: ", CDDR(form));
      Syntax::signal_error(vm, "syntax-rules: syntax error: ", form);
      return NULL;
    }

  internal_rules
    = R5RSListProcedure::list_reverse_in_place(vm.get_memory_system(),
					       Cell::nil(), internal_rules,
					       true);

  Cell *literals = CADR(form);
  VMStackRoot env_literals(vm, vm.core.mk_cons(env, literals, true));
  return vm.core.memory_system.get_cell()
    ->mk_macro_transformer(env_literals, internal_rules);
}


Cell *
SyntaxRules::parse_rule(VirtualMachine &vm, Cell *syntax_rule, Cell *keyword,
			Cell *&non_pattern_variables)
{
  VMStackRoot rule(vm, syntax_rule);
  if (!rule->is_pair() || !CDR(rule)->is_pair() || CDDR(rule) != Cell::nil())
    {
      Syntax::signal_error(vm, "wrong syntax rule: ", rule);
      return NULL;
    }

  VMStackRoot pattern_variables(vm, Cell::nil());
  Cell *pattern = parse_pattern(vm, CAR(rule), keyword,
				pattern_variables, non_pattern_variables);
  if (pattern == NULL)
    return NULL;

  VMStackRoot template_alist(vm, Cell::nil());
  Cell *templ = parse_template(vm, CADR(rule), template_alist,
			       CAR(rule), pattern_variables);
  if (templ == NULL)
    return NULL;

  VMStackRoot internal_rule(vm, template_alist);
  templ = CADR(rule);
  internal_rule = vm.cons(templ, internal_rule);
  pattern = CAR(rule);
  internal_rule = vm.cons(pattern, internal_rule);

  return internal_rule;
}


Cell *
SyntaxRules::parse_pattern0(VirtualMachine &vm, Cell *pattern,
			    Cell *&pattern_variables,
			    Cell *&non_pattern_variables)
{
  if (pattern->is_string()
      || pattern->is_character()
      || pattern == Cell::t() || pattern == Cell::f() || pattern == Cell::nil()
      || pattern->is_number()
      || pattern->is_regexp())
    return pattern;

  if (pattern->is_symbol() && pattern != vm.core.ellipsis_symbol)
    {
      Cell *vars = non_pattern_variables;
      while (vars != Cell::nil())
	{
	  if (CAR(vars) == pattern)
	    return pattern;
	  vars = CDR(vars);
	}

      vars = pattern_variables;
      while (vars != Cell::nil())
	{
	  if (CAR(vars) == pattern)
	    {
	      Syntax::signal_error(vm, "duplicated variable: ", pattern);
	      return NULL;
	    }

	  vars = CDR(vars);
	}

      pattern_variables = vm.cons(pattern, pattern_variables);
      return pattern;
    }

  if (pattern->is_pair())
    {
      Cell *sub_pattern = pattern;
      while (sub_pattern->is_pair())
	{
	  Cell *sub = CAR(sub_pattern);
	  if ((sub != vm.core.ellipsis_symbol
	       || CAR(pattern) == sub
	       || CDR(sub_pattern) != Cell::nil())
	      && parse_pattern0(vm, sub,
				pattern_variables,
				non_pattern_variables) == NULL)
	    return NULL;

	  sub_pattern = CDR(sub_pattern);
	}

      if (sub_pattern != Cell::nil()
	  && parse_pattern0(vm, sub_pattern,
			    pattern_variables,
			    non_pattern_variables) == NULL)
	return NULL;

      return pattern;
    }

  if (pattern->is_vector())
    {
      long len = pattern->get_vector_length();
      for (long i = 0; i < len; i++)
	{
	  Cell *elm = pattern->get_vector_element(i);
	  if ((elm != vm.core.ellipsis_symbol || i == 0 || i + 1 < len)
	      && parse_pattern0(vm, elm,
				pattern_variables,
				non_pattern_variables) == NULL)
	    return NULL;
	}

      return pattern;
    }

  if (pattern == vm.core.ellipsis_symbol)
    {
      Syntax::signal_error(vm, "unexpected identifier: ...");
      return NULL;
    }

  return NULL;
}


Cell *
SyntaxRules::parse_pattern(VirtualMachine &vm, Cell *pattern, Cell *keyword,
			   Cell *&pattern_variables,
			   Cell *&non_pattern_variables)
{
  if (pattern->is_pair())
    {
#ifdef ENABLE_MACRO_KEYWORD_CHECK
      // This doesn't actually conform to R5RS, but I don't want to
      // see annoying macros!
      if (CAR(pattern) != keyword)
	{
	  Syntax::signal_error(vm, "should be the syntactic keyword: ",
			       CAR(pattern));
	  return NULL;
	}
#endif
      if (parse_pattern0(vm, pattern,
			 pattern_variables, non_pattern_variables) == NULL)
	{
	  Syntax::signal_error(vm, "wrong pattern: ", pattern);
	  return NULL;
	}

      return pattern;
    }

  if (pattern->is_vector())
    {
#ifdef ENABLE_MACRO_KEYWORD_CHECK
      // This doesn't actually conform to R5RS, but I don't want to
      // see annoying macros!
      if (pattern->get_vector_length() == 0
	  || pattern->get_vector_element(0) != keyword)
	{
	  Syntax::signal_error(vm, "should be the syntactic keyword: ",
			       pattern->get_vector_element(0));
	  return NULL;
	}
#endif
      if (parse_pattern0(vm, pattern,
			 pattern_variables, non_pattern_variables) == NULL)
	{
	  Syntax::signal_error(vm, "wrong pattern: ", pattern);
	  return NULL;
	}

      return pattern;
    }
#ifdef ENABLE_MACRO_KEYWORD_CHECK
  if (pattern != keyword)
    {
      Syntax::signal_error(vm, "should be the syntactic keyword: ", pattern);
      return NULL;
    }
#endif
  return pattern;
}


Cell *
SyntaxRules::parse_template_datum(VirtualMachine &vm, Cell *templ)
{
  if (templ->is_string()
      || templ->is_character()
      || templ == Cell::t() || templ == Cell::f() || templ == Cell::nil()
      || templ->is_number()
      || templ->is_regexp()
      || templ->is_syntax()
      || templ->is_binding())
    return Cell::nil();

  return NULL;
}


Cell *
SyntaxRules::alist_cons(VirtualMachine &vm,
			Cell *key, Cell *datum, Cell *alist)
{
  VMStackRoot alist_prev(vm, alist);
  Cell *element = vm.cons(key, datum);
  return vm.cons(element, alist_prev);
}


Cell *
SyntaxRules::find_inner_most_sub_pattern(VirtualMachine &vm,
					 Cell *src_template,
					 Cell *match_pattern,
					 bool with_ellipsis,
					 Cell *&template_alist)
{
  if (match_pattern == src_template)
    return with_ellipsis ? NULL : match_pattern;

  VMStackRoot templ(vm, src_template);
  VMStackRoot pattern(vm, match_pattern);

  if (with_ellipsis)
    {
      Cell *alist = template_alist;
      while (alist != Cell::nil())
	{
	  Cell *elm = CAR(alist);
	  if (CAR(elm) == templ)
	    return CDR(elm);

	  alist = CDR(alist);
	}
    }

  if (pattern->is_pair())
    {
      VMStackRoot pattern_list(vm, pattern);
      VMStackRoot sub_pattern(vm, Cell::nil());
      while (true)
	{
	  sub_pattern = CAR(pattern_list);
	  Cell *sub = find_inner_most_sub_pattern(vm, templ,
						  sub_pattern,
						  with_ellipsis,
						  template_alist);
	  if (sub != NULL)
	    return sub;

	  pattern_list = CDR(pattern_list);
	  if (pattern_list->is_pair())
	    {
	      if (CAR(pattern_list) == vm.core.ellipsis_symbol)
		{
		  Cell *inner
		    = find_inner_most_sub_pattern(vm, templ, sub_pattern,
						  false,
						  template_alist);
		  if (inner == NULL)
		    return NULL;

		  template_alist = alist_cons(vm, templ, pattern,
					      template_alist);
		  return pattern;
		}

	      continue;
	    }

	  if (pattern_list != Cell::nil())
	    return find_inner_most_sub_pattern(vm, templ,
					       pattern_list, with_ellipsis,
					       template_alist);
	  return NULL;
	}
    }

  if (pattern->is_vector())
    {
      long len = pattern->get_vector_length() - 1;
      if (len < 0)
	return NULL;

      long i = 0;
      VMStackRoot sub_pattern(vm, Cell::nil());
      for (; i < len; i++)
	{
	  sub_pattern = pattern->get_vector_element(i);
	  Cell *sub = find_inner_most_sub_pattern(vm, templ, sub_pattern,
						  with_ellipsis,
						  template_alist);
	  if (sub != NULL)
	    return sub;
	}

      Cell *tail = pattern->get_vector_element(i);
      if (tail != vm.core.ellipsis_symbol)
	return find_inner_most_sub_pattern(vm, templ, tail, with_ellipsis,
					   template_alist);

      sub_pattern = find_inner_most_sub_pattern(vm, templ, sub_pattern,
						false,
						template_alist);

      if (sub_pattern != NULL)
	{
	  template_alist = alist_cons(vm, templ, pattern, template_alist);
	  return pattern;
	}
    }

  return NULL;
}


static Cell *
find_sub_pattern(Cell *pattern, Cell *sub_pattern)
{
  if (pattern == sub_pattern)
    return pattern;

  if (pattern->is_pair())
    {
      Cell *list = pattern;
      while (list->is_pair())
	{
	  if (find_sub_pattern(CAR(list), sub_pattern) != NULL)
	    return pattern;

	  list = CDR(list);
	}

      if (list == sub_pattern)
	return pattern;

      return NULL;
    }

  if (pattern->is_vector())
    {
      long len = pattern->get_vector_length();
      for (long i = 0; i < len; i++)
	{
	  if (find_sub_pattern(pattern->get_vector_element(i),
			       sub_pattern) != NULL)
	    return pattern;
	}
    }

  return NULL;
}


Cell *
SyntaxRules::find_common_sub_pattern(VirtualMachine &vm,
				     Cell *sub_pattern0, Cell *sub_pattern1,
				     Cell *pattern)
{
  if (sub_pattern0 == sub_pattern1)
    return sub_pattern0;

  if (sub_pattern0 == Cell::nil())
    return sub_pattern1;

  if (sub_pattern1 == Cell::nil())
    return sub_pattern0;

  if (find_sub_pattern(sub_pattern0, sub_pattern1) != NULL)
    return sub_pattern0;

  if (find_sub_pattern(sub_pattern1, sub_pattern0) != NULL)
    return sub_pattern1;

  if (pattern->is_pair())
    {
      Cell *list = pattern;
      while (list->is_pair())
	{
	  Cell *sub_pattern = CAR(list);
	  Cell *sub = find_common_sub_pattern(vm, sub_pattern0, sub_pattern1,
					      sub_pattern);
	  if (sub != NULL)
	    return sub;

	  list = CDR(list);
	  if (find_sub_pattern(sub_pattern, sub_pattern0) != NULL)
	    {
	      while (list->is_pair())
		{
		  if (find_sub_pattern(CAR(list), sub_pattern1) != NULL)
		    return pattern;

		  list = CDR(list);
		}

	      if (list != Cell::nil()
		  && find_sub_pattern(list, sub_pattern1) != NULL)
		return pattern;

	      return NULL;
	    }

	  if (find_sub_pattern(sub_pattern, sub_pattern1) != NULL)
	    {
	      while (list->is_pair())
		{
		  if (find_sub_pattern(CAR(list), sub_pattern0) != NULL)
		    return pattern;

		  list = CDR(list);
		}

	      if (list != Cell::nil()
		  && find_sub_pattern(list, sub_pattern0) != NULL)
		return pattern;

	      return NULL;
	    }
	}

      return find_common_sub_pattern(vm, sub_pattern0, sub_pattern1, list);
    }

  if (pattern->is_vector())
    {
      long len = pattern->get_vector_length();
      for (long i = 0; i < len; i++)
	{
	  Cell *sub_pattern = pattern->get_vector_element(i);
	  Cell *sub = find_common_sub_pattern(vm, sub_pattern0, sub_pattern1,
					      sub_pattern);
	  if (sub != NULL)
	    return sub;

	  if (find_sub_pattern(sub_pattern, sub_pattern0) != NULL)
	    {
	      while (++i < len)
		{
		  if (find_sub_pattern(pattern->get_vector_element(i),
				       sub_pattern1) != NULL)
		    return pattern;
		}

	      return NULL;
	    }

	  if (find_sub_pattern(sub_pattern, sub_pattern1) != NULL)
	    {
	      while (++i < len)
		{
		  if (find_sub_pattern(pattern->get_vector_element(i),
				       sub_pattern0) != NULL)
		    return pattern;
		}

	      return NULL;
	    }
	}
    }

  return NULL;
}


Cell *
SyntaxRules::parse_template(VirtualMachine &vm, Cell *src_template,
			    Cell *&template_alist,
			    Cell *match_pattern, Cell *&pattern_variables)
{
  VMStackRoot templ(vm, src_template);
  VMStackRoot pattern(vm, match_pattern);
  if (!templ->is_pair() && !templ->is_vector())
    {
      if (templ->is_symbol())
	{
	  Cell *variables = pattern_variables;
	  while (variables != Cell::nil())
	    {
	      if (CAR(variables) == templ)
		return templ;

	      variables = CDR(variables);
	    }

	  // templ will be treated as a literal identifier
	  variables = template_alist;
	  while (variables != Cell::nil())
	    {
	      if (CAR(variables) == templ)
		{
		  assert(CDR(variables) == Cell::nil());
		  return Cell::nil();
		}

	      variables = CDR(variables);
	    }

	  template_alist = alist_cons(vm, templ, Cell::nil(), template_alist);
	  return Cell::nil();	// treated as a literal identifier
	}

      return parse_template_datum(vm, templ);
    }

  if (templ->is_pair())
    {
      VMStackRoot common_pattern(vm, Cell::nil());
      VMStackRoot list(vm, templ);
      VMStackRoot sub_templ(vm, CAR(list));
      VMStackRoot sub_pattern(vm, Cell::nil());
      while (true)
	{
	  Cell *sub = parse_template(vm, sub_templ, template_alist,
				     pattern, pattern_variables);
	  if (sub == NULL)
	    {
	      Syntax::signal_error(vm, "wrong template: ", templ);
	      return NULL;
	    }

	  sub_pattern = sub;
	  common_pattern = find_common_sub_pattern(vm,
						   common_pattern, sub_pattern,
						   pattern);
	  list = CDR(list);
	  if (list->is_pair())
	    {
	      sub_templ = CAR(list);
	      if (sub_templ == vm.core.ellipsis_symbol)
		{
		  sub = find_inner_most_sub_pattern(vm, sub_pattern,
						    pattern, true,
						    template_alist);
		  if (sub != NULL)
		    {
		      sub_pattern = sub;
		      common_pattern = find_common_sub_pattern(vm,
							       common_pattern,
							       sub_pattern,
							       pattern);
		      template_alist = alist_cons(vm, templ, common_pattern,
						  template_alist);
		      list = CDR(list);
		      if (list->is_pair())
			{
			  sub_templ = CAR(list);
			  continue;
			}

		      return common_pattern;
		    }

		  Syntax::signal_error(vm, "wrong template: ", templ);
		  return NULL;
		}

	      continue;
	    }

	  sub = parse_template(vm, list, template_alist,
			       pattern, pattern_variables);
	  if (sub == NULL)
	    {
	      Syntax::signal_error(vm, "wrong template: ", templ);
	      return NULL;
	    }

	  sub_pattern = sub;
	  common_pattern
	    = find_common_sub_pattern(vm, common_pattern, sub_pattern,
				      pattern);
	  template_alist = alist_cons(vm, templ, common_pattern,
				      template_alist);
	  return common_pattern;
	}
    }

  long len = templ->get_vector_length() - 1;
  if (len < 0)
    return Cell::nil();

  VMStackRoot common_pattern(vm, Cell::nil());
  long i = 0;
  VMStackRoot sub_templ(vm, templ->get_vector_element(i));
  VMStackRoot sub_pattern(vm, Cell::nil());
  while (++i < len)
    {
      Cell *sub = parse_template(vm, sub_templ, template_alist, 
				 pattern, pattern_variables);
      if (sub == NULL)
	{
	  Syntax::signal_error(vm, "wrong template: ", templ);
	  return NULL;
	}

      sub_pattern = sub;
      common_pattern = find_common_sub_pattern(vm, common_pattern, sub_pattern,
					       pattern);
      sub_templ = templ->get_vector_element(i);
      if (sub_templ == vm.core.ellipsis_symbol)
	{
	  sub = find_inner_most_sub_pattern(vm, sub_pattern, pattern, true,
					    template_alist);
	  if (sub != NULL)
	    {
	      sub_pattern = sub;
	      common_pattern = find_common_sub_pattern(vm, common_pattern,
						       sub_pattern, pattern);
	      template_alist = alist_cons(vm, templ, common_pattern,
					  template_alist);
	      if (++i < len)
		{
		  sub_templ = templ->get_vector_element(i);
		  continue;
		}

	      return common_pattern;
	    }

	  Syntax::signal_error(vm, "wrong template: ", templ);
	  return NULL;
	}
    }
  Cell *sub = parse_template(vm, sub_templ, template_alist,
			     pattern, pattern_variables);
  if (sub != NULL)
    {
      sub_pattern = sub;
      common_pattern = find_common_sub_pattern(vm, common_pattern,
					       sub_pattern, pattern);
      template_alist = alist_cons(vm, templ, common_pattern, template_alist);
      return common_pattern;
    }
  Syntax::signal_error(vm, "wrong template: ", templ);
  return NULL;
}

}
