/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: r5rs_math_procedure.cc,v 1.14 2009/01/01 07:47:00 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libtscore/r5rs_math_procedure.h"

#include "libtscore/tscore.h"
#include "libtscore/virtual_machine.h"
#include "libtscore/writer.h"

#include <cassert>
#include <cfloat>
#include <climits>
#include <cmath>
#include <cstdio>


namespace LibTSCore
{

NativeProcedure::Info R5RSMathProcedure::procedures[] = {
  { &R5RSMathProcedure::is_number,
    "number?", TST_ANY, 1, 1 },
  { &R5RSMathProcedure::is_complex,
    "complex?", TST_ANY, 1, 1 },
  { &R5RSMathProcedure::is_real,
    "real?", TST_ANY, 1, 1 },
  { &R5RSMathProcedure::is_rational,
    "rational?", TST_ANY, 1, 1 },
  { &R5RSMathProcedure::is_integer,
    "integer?", TST_ANY, 1, 1 },
  { &R5RSMathProcedure::is_exact,
    "exact?", TST_ANY, 1, 1 },
  { &R5RSMathProcedure::is_inexact,
    "inexact?", TST_ANY, 1, 1 },

  { &R5RSMathProcedure::number_eq,
    "=", TST_NUMBER, 2, INF_ARG },
  { &R5RSMathProcedure::number_lt,
    "<", TST_NUMBER, 2, INF_ARG },
  { &R5RSMathProcedure::number_gt,
    ">", TST_NUMBER, 2, INF_ARG },
  { &R5RSMathProcedure::number_le,
    "<=", TST_NUMBER, 2, INF_ARG },
  { &R5RSMathProcedure::number_ge,
    ">=", TST_NUMBER, 2, INF_ARG },

  { &R5RSMathProcedure::is_zero,
    "zero?", TST_NUMBER, 1, 1 },
  { &R5RSMathProcedure::is_positive,
    "positive?", TST_NUMBER, 1, 1 },
  { &R5RSMathProcedure::is_negative,
    "negative?", TST_NUMBER, 1, 1 },
  { &R5RSMathProcedure::is_odd,
    "odd?", TST_NUMBER, 1, 1 },
  { &R5RSMathProcedure::is_even,
    "even?", TST_NUMBER, 1, 1 },

  { &R5RSMathProcedure::number_max,
    "max", TST_NUMBER, 2, INF_ARG },
  { &R5RSMathProcedure::number_min,
    "min", TST_NUMBER, 2, INF_ARG },

  { &R5RSMathProcedure::number_add,
    "+", TST_NUMBER, 0, INF_ARG },
  { &R5RSMathProcedure::number_mul,
    "*", TST_NUMBER, 0, INF_ARG },
  { &R5RSMathProcedure::number_sub,
    "-", TST_NUMBER, 1, INF_ARG },
  { &R5RSMathProcedure::number_div,
    "/", TST_NUMBER, 1, INF_ARG },

  { &R5RSMathProcedure::number_abs,
    "abs", TST_NUMBER, 1, 1 },

  { &R5RSMathProcedure::number_quotient,
    "quotient", TST_NUMBER, 2, 2 },
  { &R5RSMathProcedure::number_remainder,
    "remainder", TST_NUMBER, 2, 2 },
  { &R5RSMathProcedure::number_modulo,
    "modulo", TST_NUMBER, 2, 2 },

  { &R5RSMathProcedure::number_gcd,
    "gcd", TST_NUMBER, 0, INF_ARG },
  { &R5RSMathProcedure::number_lcm,
    "lcm", TST_NUMBER, 0, INF_ARG },
#if 0
  { &R5RSMathProcedure::number_numerator,
    "numerator", TST_NUMBER, 1, 1 },
  { &R5RSMathProcedure::number_denominator,
    "denominator", TST_NUMBER, 1, 1 },
#endif

  { &R5RSMathProcedure::number_floor,
    "floor", TST_NUMBER, 1, 1 },
  { &R5RSMathProcedure::number_ceiling,
    "ceiling", TST_NUMBER, 1, 1 },
  { &R5RSMathProcedure::number_truncate,
    "truncate", TST_NUMBER, 1, 1 },
  { &R5RSMathProcedure::number_round,
    "round", TST_NUMBER, 1, 1 },

#if 0
  { &R5RSMathProcedure::number_rationalize,
    "rationalize", TST_NUMBER, 2, 2 },
#endif

  { &R5RSMathProcedure::number_exp,
    "exp", TST_NUMBER, 1, 1 },
  { &R5RSMathProcedure::number_log,
    "log", TST_NUMBER, 1, 1 },
  { &R5RSMathProcedure::number_sin,
    "sin", TST_NUMBER, 1, 1 },
  { &R5RSMathProcedure::number_cos,
    "cos", TST_NUMBER, 1, 1 },
  { &R5RSMathProcedure::number_tan,
    "tan", TST_NUMBER, 1, 1 },
  { &R5RSMathProcedure::number_asin,
    "asin", TST_NUMBER, 1, 1 },
  { &R5RSMathProcedure::number_acos,
    "acos", TST_NUMBER, 1, 1 },
  { &R5RSMathProcedure::number_atan,
    "atan", TST_NUMBER, 1, 2 },

  { &R5RSMathProcedure::number_sqrt,
    "sqrt", TST_NUMBER, 1, 1 },
  { &R5RSMathProcedure::number_expt,
    "expt", TST_NUMBER, 2, 2 },

#if 0
  { &R5RSMathProcedure::make_rectangular,
    "make-rectangular", TST_NUMBER, 2, 2 },
  { &R5RSMathProcedure::make_polar,
    "make-polar", TST_NUMBER, 2, 2 },
  { &R5RSMathProcedure::real_part,
    "real-part", TST_NUMBER, 1, 1 },
  { &R5RSMathProcedure::imag_part,
    "imag-part", TST_NUMBER, 1, 1 },
  { &R5RSMathProcedure::magnitude,
    "magnitude", TST_NUMBER, 1, 1 },
  { &R5RSMathProcedure::angle,
    "angle", TST_NUMBER, 1, 1 },
#endif

  { &R5RSMathProcedure::exact2inexact,
    "exact->inexact", TST_NUMBER, 1, 1 },
  { &R5RSMathProcedure::inexact2exact,
    "inexact->exact", TST_NUMBER, 1, 1 },

  { &R5RSMathProcedure::number2string,
    "number->string", TST_NUMBER TST_INTEGER, 1, 2 },
  { &R5RSMathProcedure::string2number,
    "string->number", TST_STRING TST_INTEGER, 1, 2 },
};


Cell *
R5RSMathProcedure::assign_procedures(Cell *environment, Cell *frame)
{
  TSStackRoot env(core, environment);
  TSStackRoot frm(core, frame);
  size_t n_procs = sizeof(procedures) / sizeof(Info);
  for (size_t i = 0; i < n_procs; i++)
    assign_procedure(env, frm, &procedures[i], this);

  return env;
}


#define ARG(n)	context.frame->load_variable(argument_pointer + (n))


Cell *
R5RSMathProcedure::is_number(VirtualMachine &vm, Register &context,
			     unsigned long argument_pointer,
			     unsigned long nargs, void *data)
{
  return Cell::ts_bool(ARG(0)->is_number());
}


Cell *
R5RSMathProcedure::is_complex(VirtualMachine &vm, Register &context,
			      unsigned long argument_pointer,
			      unsigned long nargs, void *data)
{
  return Cell::ts_bool(ARG(0)->is_number());
}


Cell *
R5RSMathProcedure::is_real(VirtualMachine &vm, Register &context,
			   unsigned long argument_pointer,
			   unsigned long nargs, void *data)
{
  return Cell::ts_bool(ARG(0)->is_number());
}


Cell *
R5RSMathProcedure::is_rational(VirtualMachine &vm, Register &context,
			       unsigned long argument_pointer,
			       unsigned long nargs, void *data)
{
  return Cell::ts_bool(ARG(0)->is_number());
}


Cell *
R5RSMathProcedure::is_integer(VirtualMachine &vm, Register &context,
			      unsigned long argument_pointer,
			      unsigned long nargs, void *data)
{
  return Cell::ts_bool(ARG(0)->is_integer());
}


Cell *
R5RSMathProcedure::is_exact(VirtualMachine &vm, Register &context,
			    unsigned long argument_pointer,
			    unsigned long nargs, void *data)
{
  return Cell::ts_bool(ARG(0)->is_integer());
}


Cell *
R5RSMathProcedure::is_inexact(VirtualMachine &vm, Register &context,
			      unsigned long argument_pointer,
			      unsigned long nargs, void *data)
{
  return Cell::ts_bool(!ARG(0)->is_integer() && ARG(0)->is_number());
}


Cell *
R5RSMathProcedure::number_eq(VirtualMachine &vm, Register &context,
			     unsigned long argument_pointer,
			     unsigned long nargs, void *data)
{
  Number value(ARG(0));
  if (!value.eq(ARG(1)))
    return Cell::f();

  for (unsigned long i = 2; i < nargs; i++)
    if (!value.eq(ARG(i)))
      return Cell::f();

  return Cell::t();
}


Cell *
R5RSMathProcedure::number_lt(VirtualMachine &vm, Register &context,
			     unsigned long argument_pointer,
			     unsigned long nargs, void *data)
{
  Number value(ARG(0));
  if (!value.lt(ARG(1)))
    return Cell::f();

  value = ARG(1);
  for (unsigned long i = 2; i < nargs; i++)
    {
      Cell *elm = ARG(i);
      if (!value.lt(elm))
	return Cell::f();

      value = elm;
    }

  return Cell::t();
}


Cell *
R5RSMathProcedure::number_gt(VirtualMachine &vm, Register &context,
			     unsigned long argument_pointer,
			     unsigned long nargs, void *data)
{
  Number value(ARG(0));
  if (!value.gt(ARG(1)))
    return Cell::f();

  value = ARG(1);
  for (unsigned long i = 2; i < nargs; i++)
    {
      Cell *elm = ARG(i);
      if (!value.gt(elm))
	return Cell::f();

      value = elm;
    }

  return Cell::t();
}


Cell *
R5RSMathProcedure::number_le(VirtualMachine &vm, Register &context,
			     unsigned long argument_pointer,
			     unsigned long nargs, void *data)
{
  Number value(ARG(0));
  if (!value.le(ARG(1)))
    return Cell::f();

  value = ARG(1);
  for (unsigned long i = 2; i < nargs; i++)
    {
      Cell *elm = ARG(i);
      if (!value.le(elm))
	return Cell::f();

      value = elm;
    }

  return Cell::t();
}


Cell *
R5RSMathProcedure::number_ge(VirtualMachine &vm, Register &context,
			     unsigned long argument_pointer,
			     unsigned long nargs, void *data)
{
  Number value(ARG(0));
  if (!value.ge(ARG(1)))
    return Cell::f();

  value = ARG(1);
  for (unsigned long i = 2; i < nargs; i++)
    {
      Cell *elm = ARG(i);
      if (!value.ge(elm))
	return Cell::f();

      value = elm;
    }

  return Cell::t();
}



Cell *
R5RSMathProcedure::is_zero(VirtualMachine &vm, Register &context,
			   unsigned long argument_pointer,
			   unsigned long nargs, void *data)
{
  return Cell::ts_bool(ARG(0)->get_integer() == 0);
}


Cell *
R5RSMathProcedure::is_positive(VirtualMachine &vm, Register &context,
			       unsigned long argument_pointer,
			       unsigned long nargs, void *data)
{
  return Cell::ts_bool(ARG(0)->get_real() > 0.0);
}


Cell *
R5RSMathProcedure::is_negative(VirtualMachine &vm, Register &context,
			       unsigned long argument_pointer,
			       unsigned long nargs, void *data)
{
  return Cell::ts_bool(ARG(0)->get_real() < 0.0);
}


Cell *
R5RSMathProcedure::is_odd(VirtualMachine &vm, Register &context,
			  unsigned long argument_pointer,
			  unsigned long nargs, void *data)
{
  return Cell::ts_bool((ARG(0)->get_integer() & 1) != 0);
}


Cell *
R5RSMathProcedure::is_even(VirtualMachine &vm, Register &context,
			   unsigned long argument_pointer,
			   unsigned long nargs, void *data)
{
  return Cell::ts_bool((ARG(0)->get_integer() & 1) == 0);
}


Cell *
R5RSMathProcedure::number_max(VirtualMachine &vm, Register &context,
			      unsigned long argument_pointer,
			      unsigned long nargs, void *data)
{
  Number value(ARG(0));
  Cell *elm = ARG(1);
  if (value.lt(elm))
    value = elm;

  for (unsigned long i = 2; i < nargs; i++)
    {
      elm = ARG(i);
      if (value.lt(elm))
	value = elm;
    }

  return vm.mk_number(value);
}


Cell *
R5RSMathProcedure::number_min(VirtualMachine &vm, Register &context,
			      unsigned long argument_pointer,
			      unsigned long nargs, void *data)
{
  Number value(ARG(0));
  Cell *elm = ARG(1);
  if (value.gt(elm))
    value = elm;

  for (unsigned long i = 2; i < nargs; i++)
    {
      elm = ARG(i);
      if (value.gt(elm))
	value = elm;
    }

  return vm.mk_number(value);
}


static Number num_zero(0L);
static Number num_one(1L);


Cell *
R5RSMathProcedure::number_add(VirtualMachine &vm, Register &context,
			      unsigned long argument_pointer,
			      unsigned long nargs, void *data)
{
  Number value = num_zero;
  for (unsigned long i = 0; i < nargs; i++)
    value.add(ARG(i));

  return vm.mk_number(value);
}


Cell *
R5RSMathProcedure::number_mul(VirtualMachine &vm, Register &context,
			      unsigned long argument_pointer,
			      unsigned long nargs, void *data)
{
  Number value = num_one;
  for (unsigned long i = 0; i < nargs; i++)
    value.mul(ARG(i));

  return vm.mk_number(value);
}


Cell *
R5RSMathProcedure::number_sub(VirtualMachine &vm, Register &context,
			      unsigned long argument_pointer,
			      unsigned long nargs, void *data)
{
  Number value;
  if (nargs == 1)
    {
      value = num_zero;
      value.sub(ARG(0));
      return vm.mk_number(value);
    }

  value = ARG(0);
  for (unsigned long i = 1; i < nargs; i++)
    value.sub(ARG(i));

  return vm.mk_number(value);
}


static inline bool
is_zero_double(double x)
{
  return x < DBL_MIN && x > -DBL_MIN;
}


Cell *
R5RSMathProcedure::number_div(VirtualMachine &vm, Register &context,
			      unsigned long argument_pointer,
			      unsigned long nargs, void *data)
{
  Number value;
  if (nargs == 1)
    {
      value = num_one;
      value.div(ARG(0));
      return vm.mk_number(value);
    }

  value = ARG(0);
  for (unsigned long i = 1; i < nargs; i++)
    {
      Cell *elm = ARG(i);
      if (!is_zero_double(elm->get_real()))
	value.div(elm);
      else
	return signal_error(vm, "/: division by zero");
    }

  return vm.mk_number(value);
}



Cell *
R5RSMathProcedure::number_abs(VirtualMachine &vm, Register &context,
			      unsigned long argument_pointer,
			      unsigned long nargs, void *data)
{
  if (ARG(0)->get_real() >= 0.0)
    return ARG(0);

  Number value = num_zero;
  value.sub(ARG(0));
  return vm.mk_number(value);
}


Cell *
R5RSMathProcedure::number_quotient(VirtualMachine &vm, Register &context,
				   unsigned long argument_pointer,
				   unsigned long nargs, void *data)
{
  Number value(ARG(0));
  Cell *divisor = ARG(1);
  if (divisor->get_integer() != 0)
    {
      value.intdiv(divisor);
      return vm.mk_number(value);
    }

  return signal_error(vm, "quotient: division by zero");
}


Cell *
R5RSMathProcedure::number_remainder(VirtualMachine &vm, Register &context,
				    unsigned long argument_pointer,
				    unsigned long nargs, void *data)
{
  Number value(ARG(0));
  Cell *divisor = ARG(1);
  if (divisor->get_integer() != 0)
    {
      value.rem(divisor);
      return vm.mk_number(value);
    }

  return signal_error(vm, "remainder: division by zero");
}


Cell *
R5RSMathProcedure::number_modulo(VirtualMachine &vm, Register &context,
				 unsigned long argument_pointer,
				 unsigned long nargs, void *data)
{
  Number value(ARG(0));
  Cell *divisor = ARG(1);
  if (divisor->get_integer() != 0)
    {
      value.mod(divisor);
      return vm.mk_number(value);
    }

  return signal_error(vm, "modulo: division by zero");
}


static inline void
gcd2(Number &a, Number &b)
{
  while (true)
    {
      assert(b.get_ivalue() >= 0);
      if (b.get_ivalue() == 0)
	return;

      a.rem(b);
      assert(a.get_ivalue() >= 0);

      if (a.get_ivalue() == 0)
	{
	  a = b;
	  return;
	}

      b.rem(a);
    }
}


Cell *
R5RSMathProcedure::number_gcd(VirtualMachine &vm, Register &context,
			      unsigned long argument_pointer,
			      unsigned long nargs, void *data)
{
  if (nargs == 0)
    return vm.mk_number(num_zero);

  Cell *a = ARG(0);

  Number c;
  if (a->get_real() >= 0.0)
    c = a;
  else
    {
      c = num_zero;
      c.sub(a);
    }

  for (unsigned long i = 1; i < nargs; i++)
    {
      Number b;
      Cell *elm = ARG(i);
      if (elm->get_real() >= 0.0)
	b = elm;
      else
	{
	  b = num_zero;
	  b.sub(elm);
	}
      gcd2(c, b);
    }

  return vm.mk_number(c);
}


Cell *
R5RSMathProcedure::number_lcm(VirtualMachine &vm, Register &context,
			      unsigned long argument_pointer,
			      unsigned long nargs, void *data)
{
  if (nargs == 0)
    return vm.mk_number(num_one);

  Cell *a = ARG(0);

  Number c;
  if (a->get_real() >= 0.0)
    c = a;
  else
    {
      c = num_zero;
      c.sub(a);
    }

  for (unsigned long i = 1; i < nargs; i++)
    {
      Number a(c);
      Number b;
      Cell *elm = ARG(i);
      if (elm->get_real() >= 0.0)
	b = elm;
      else
	{
	  b = num_zero;
	  b.sub(elm);
	}

      if (c.get_rvalue() == 0.0)
	break;

      if (b.get_rvalue() == 0.0)
	{
	  c = b;
	  break;
	}

      Number d(b);
      gcd2(a, b);
      c.intdiv(a);
      c.mul(d);
    }
  return vm.mk_number(c);
}


#if 0
Cell *
R5RSMathProcedure::number_numerator(VirtualMachine &vm, Register &context,
				    unsigned long argument_pointer,
				    unsigned long nargs, void *data)
{
}


Cell *
R5RSMathProcedure::number_denominator(VirtualMachine &vm, Register &context,
				      unsigned long argument_pointer,
				      unsigned long nargs, void *data)
{
}
#endif


Cell *
R5RSMathProcedure::number_floor(VirtualMachine &vm, Register &context,
				unsigned long argument_pointer,
				unsigned long nargs, void *data)
{
  return vm.mk_real(floor(ARG(0)->get_real()));
}


Cell *
R5RSMathProcedure::number_ceiling(VirtualMachine &vm, Register &context,
				  unsigned long argument_pointer,
				  unsigned long nargs, void *data)
{
  return vm.mk_real(ceil(ARG(0)->get_real()));
}


Cell *
R5RSMathProcedure::number_truncate(VirtualMachine &vm, Register &context,
				   unsigned long argument_pointer,
				   unsigned long nargs, void *data)
{
  double x = ARG(0)->get_real();
  if (x > 0.0)
    return vm.mk_real(floor(x));

  return vm.mk_real(ceil(x));
}


Cell *
R5RSMathProcedure::number_round(VirtualMachine &vm, Register &context,
				unsigned long argument_pointer,
				unsigned long nargs, void *data)
{
  double x = ARG(0)->get_real();
  double fl = floor(x);
  double ce = ceil(x);
  double dfl = x - fl;
  double dce = ce -x;

  if (dfl > dce)
    return vm.mk_real(ce);

  if (dfl < dce)
    return vm.mk_real(fl);

  if (fmod(fl, 2.0) == 0.0)
    return vm.mk_real(fl);

  return vm.mk_real(ce);
}


#if 0
Cell *
R5RSMathProcedure::number_rationalize(VirtualMachine &vm, Register &context,
				      unsigned long argument_pointer,
				      unsigned long nargs, void *data)
{
}
#endif


Cell *
R5RSMathProcedure::number_exp(VirtualMachine &vm, Register &context,
			      unsigned long argument_pointer,
			      unsigned long nargs, void *data)
{
  return vm.mk_real(exp(ARG(0)->get_real()));
}


Cell *
R5RSMathProcedure::number_log(VirtualMachine &vm, Register &context,
			      unsigned long argument_pointer,
			      unsigned long nargs, void *data)
{
  return vm.mk_real(log(ARG(0)->get_real()));
}


Cell *
R5RSMathProcedure::number_sin(VirtualMachine &vm, Register &context,
			      unsigned long argument_pointer,
			      unsigned long nargs, void *data)
{
  return vm.mk_real(sin(ARG(0)->get_real()));
}


Cell *
R5RSMathProcedure::number_cos(VirtualMachine &vm, Register &context,
			      unsigned long argument_pointer,
			      unsigned long nargs, void *data)
{
  return vm.mk_real(cos(ARG(0)->get_real()));
}


Cell *
R5RSMathProcedure::number_tan(VirtualMachine &vm, Register &context,
			      unsigned long argument_pointer,
			      unsigned long nargs, void *data)
{
  return vm.mk_real(tan(ARG(0)->get_real()));
}


Cell *
R5RSMathProcedure::number_asin(VirtualMachine &vm, Register &context,
			       unsigned long argument_pointer,
			       unsigned long nargs, void *data)
{
  return vm.mk_real(asin(ARG(0)->get_real()));
}


Cell *
R5RSMathProcedure::number_acos(VirtualMachine &vm, Register &context,
			       unsigned long argument_pointer,
			       unsigned long nargs, void *data)
{
  return vm.mk_real(acos(ARG(0)->get_real()));
}


Cell *
R5RSMathProcedure::number_atan(VirtualMachine &vm, Register &context,
			       unsigned long argument_pointer,
			       unsigned long nargs, void *data)
{
  if (nargs == 1)
    return vm.mk_real(atan(ARG(0)->get_real()));

  return vm.mk_real(atan2(ARG(0)->get_real(), ARG(1)->get_real()));
}


Cell *
R5RSMathProcedure::number_sqrt(VirtualMachine &vm, Register &context,
			       unsigned long argument_pointer,
			       unsigned long nargs, void *data)
{
  return vm.mk_real(sqrt(ARG(0)->get_real()));
}


Cell *
R5RSMathProcedure::number_expt(VirtualMachine &vm, Register &context,
			       unsigned long argument_pointer,
			       unsigned long nargs, void *data)
{
  return vm.mk_real(pow(ARG(0)->get_real(), ARG(1)->get_real()));
}


#if 0
Cell *
R5RSMathProcedure::make_rectangular(VirtualMachine &vm, Register &context,
				    unsigned long argument_pointer,
				    unsigned long nargs, void *data)
{
}


Cell *
R5RSMathProcedure::make_polar(VirtualMachine &vm, Register &context,
			      unsigned long argument_pointer,
			      unsigned long nargs, void *data)
{
}


Cell *
R5RSMathProcedure::real_part(VirtualMachine &vm, Register &context,
			     unsigned long argument_pointer,
			     unsigned long nargs, void *data)
{
}


Cell *
R5RSMathProcedure::imag_part(VirtualMachine &vm, Register &context,
			     unsigned long argument_pointer,
			     unsigned long nargs, void *data)
{
}


Cell *
R5RSMathProcedure::magnitude(VirtualMachine &vm, Register &context,
			     unsigned long argument_pointer,
			     unsigned long nargs, void *data)
{
}


Cell *
R5RSMathProcedure::angle(VirtualMachine &vm, Register &context,
			 unsigned long argument_pointer,
			 unsigned long nargs, void *data)
{
}
#endif


Cell *
R5RSMathProcedure::exact2inexact(VirtualMachine &vm, Register &context,
				 unsigned long argument_pointer,
				 unsigned long nargs, void *data)
{
  if (!ARG(0)->is_integer())
    return ARG(0);

  return vm.mk_real(ARG(0)->get_real());
}


Cell *
R5RSMathProcedure::inexact2exact(VirtualMachine &vm, Register &context,
				 unsigned long argument_pointer,
				 unsigned long nargs, void *data)
{
  if (ARG(0)->is_integer())
    return ARG(0);

  return vm.mk_integer(ARG(0)->get_integer());
}


Cell *
R5RSMathProcedure::number2string(VirtualMachine &vm, Register &context,
				 unsigned long argument_pointer,
				 unsigned long nargs, void *data)
{
  OutputStringPort string_buf;
  const Cell *err;
  long radix;
  if (nargs == 1)
    radix = 10;
  else
    {
      radix = ARG(1)->get_integer0();
      if (radix != 10 && radix != 16 && radix != 2 && radix != 8)
	return signal_error(vm,
			    "number->string: invalid radix specified:",
			    ARG(1));
    }

  err = Writer::write_number(ARG(0), &string_buf, radix);
  assert(err == Cell::nil());

  return vm.mk_counted_string(string_buf.get_output_string(),
			      string_buf.get_output_strlen());
}


Cell *
R5RSMathProcedure::string2number(VirtualMachine &vm, Register &context,
				 unsigned long argument_pointer,
				 unsigned long nargs, void *data)
{
  if (nargs == 2)
    {
      OutputStringPort string_buf;
      switch (ARG(1)->get_integer0())
	{
	case 2:
	  string_buf.append("#b");
	  break;

	case 8:
	  string_buf.append("#o");
	  break;

	case 10:
	  break;

	case 16:
	  string_buf.append("#x");
	  break;

	default:
	  return signal_error(vm,
			      "string->number: invalid radix specified:",
			      ARG(1));
	}

      string_buf.append(ARG(0)->get_string(), ARG(0)->get_strlen());
      InputStringPort input_port(string_buf.get_output_string(),
				 string_buf.get_output_strlen(), false);
      Cell *datum = vm.get_reader()->read_datum(&input_port);
      if (datum != NULL)
	{
	  if (datum->is_number())
	    return datum;

	  return Cell::f();
	}
      // Retry without specified default radix.
    }

  InputStringPort input_port(ARG(0)->get_string(),
			     ARG(0)->get_strlen(), false);
  Cell *datum = vm.get_reader()->read_datum(&input_port);
  if (datum != NULL && datum->is_number())
    return datum;

  return Cell::f();
}

}
