/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: r5rs_misc_procedure.cc,v 1.17 2010/08/15 15:39:47 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libtscore/r5rs_misc_procedure.h"

#include "libtscore/bytecode_buffer.h"
#include "libtscore/tscore.h"
#include "libtscore/virtual_machine.h"

#include <vector>

#include <cassert>


namespace LibTSCore
{

NativeProcedure::Info R5RSMiscProcedure::procedures[] = {
  { &R5RSMiscProcedure::eq,
    "eq?", TST_ANY, 2, 2 },
  { &R5RSMiscProcedure::eqv,
    "eqv?", TST_ANY, 2, 2 },
  { &R5RSMiscProcedure::equal,
    "equal?", TST_ANY, 2, 2 },
  { &R5RSMiscProcedure::is_not,
    "not", TST_ANY, 1, 1 },
  { &R5RSMiscProcedure::is_boolean,
    "boolean?", TST_NONE, 1, 1 },
  { &R5RSMiscProcedure::is_symbol,
    "symbol?", TST_NONE, 1, 1 },
  { &R5RSMiscProcedure::symbol2string,
    "symbol->string", TST_SYMBOL, 1, 1 },
  { &R5RSMiscProcedure::string2symbol,
    "string->symbol", TST_STRING, 1, 1 },
  { &R5RSMiscProcedure::is_procedure,
    "procedure?", TST_NONE, 1, 1 },
  { &R5RSMiscProcedure::apply,
    "apply", TST_PROCEDURE TST_ANY, 2, INF_ARG },
  { &R5RSMiscProcedure::map,
    "map", TST_PROCEDURE TST_LIST, 2, INF_ARG },
  { &R5RSMiscProcedure::for_each,
    "for-each", TST_PROCEDURE TST_LIST, 2, INF_ARG },
#if 0
  { &R5RSMiscProcedure::force,
    "force", TST_PROMISE, 1, 1 },
#endif
  { &R5RSMiscProcedure::call_cc,
    "call-with-current-continuation", TST_PROCEDURE, 1, 1 },
  { &R5RSMiscProcedure::call_cc,
    "call/cc", TST_PROCEDURE, 1, 1 },
  { &R5RSMiscProcedure::values,
    "values", TST_NONE, 0, INF_ARG },
  { &R5RSMiscProcedure::call_with_values,
    "call-with-values", TST_PROCEDURE, 2, 2 },
  { &R5RSMiscProcedure::dynamic_wind,
    "dynamic-wind", TST_PROCEDURE, 3, 3 },
  { &R5RSMiscProcedure::eval,
    "eval", TST_ANY TST_ENVIRONMENT, 2, 2 },
  { &R5RSMiscProcedure::scheme_report_environment,
    "scheme-report-environment", TST_NONNEG, 1, 1 },
  { &R5RSMiscProcedure::null_environment,
    "null-environment", TST_NONNEG, 1, 1 },
  { &R5RSMiscProcedure::interaction_environment,
    "interaction-environment", TST_NONE, 0, 0 }
};


struct BytecodeProcedureInfo
{
  const char *name;
  long frame_size;
  long min_arity;
  long max_arity;
  size_t bytecode_length;
  const char *bytecode;
};


static BytecodeProcedureInfo bytecode_procedures[] = {
  { "force", 4, 2, 2, 10,
    OPS_LOAD_L1 "\001" OPS_IF_FORCED1 "\006"
    OPS_CALL_CC1 "\002" "\001" OPS_FORCE_L1 "\001"
    OPS_APPLY_CONTINUATION0
  }
};


Cell *
R5RSMiscProcedure::assign_procedures(Cell *environment, Cell *frame)
{
  TSStackRoot env(core, environment);
  TSStackRoot frm(core, frame);
  size_t n_procs = sizeof(procedures) / sizeof(Info);
  for (size_t i = 0; i < n_procs; i++)
    assign_procedure(env, frm, &procedures[i], this);

  n_procs = sizeof(bytecode_procedures) / sizeof(BytecodeProcedureInfo);
  for (size_t i = 0; i < n_procs; i++)
    assign_bytecode_procedure(env, frm, &bytecode_procedures[i]);

  return env;
}


void
R5RSMiscProcedure::assign_bytecode_procedure(Cell *environment, Cell *frame,
					     const BytecodeProcedureInfo *info)
{
  TSStackRoot env(core, environment);
  TSStackRoot frm(core, frame);

  BytecodeBuffer *bytecode = new BytecodeBuffer();
  assert(strlen(info->bytecode) == info->bytecode_length);

  for (size_t i = 0; i < info->bytecode_length; i++)
    bytecode->append(static_cast<TSVMOpcode>(info->bytecode[i]));

  BytecodeProcedure *proc
    = new BytecodeProcedure(bytecode, info->frame_size,
			    info->min_arity, info->max_arity);
  TSStackRoot tmpl(core, core.mk_closure_template(Cell::nil0(), Cell::nil0(),
						  proc));
  TSStackRoot closure(core, core.mk_closure(tmpl, Cell::nil0()));
  Cell *symbol = core.mk_static_symbol(info->name);
  const Cell *binding = env->bind_variable(symbol);
  frm->store_toplevel_variable(env, binding->get_slot_location(), closure);
}


bool
R5RSMiscProcedure::eqv(const Cell *x, const Cell *y)
{
  if (x->is_forced_promise())
    x = x->get_memoized_value();
  if (y->is_forced_promise())
    y = y->get_memoized_value();

  if (x == y)
    return true;

  if (x->is_string())
    {
      if (y->is_string())
	return x->get_strlen() == y->get_strlen()
	  && strncmp(x->get_string(), y->get_string(), x->get_strlen()) == 0;

      return false;
    }

  if (x->is_number())
    {
      if (y->is_number())
        {
          Number n;
          n = const_cast<Cell *>(x);
          return n.eq(const_cast<Cell *>(y));
        }

      return false;
    }

  if (x->is_native_procedure())
    {
      if (y->is_native_procedure())
	return x->get_native_procedure_info() == y->get_native_procedure_info();

      return false;
    }

  return false;
}


enum EqualState {
  EQUAL_STATE_INITIAL,
  EQUAL_STATE_PAIR,
  EQUAL_STATE_VECTOR
};


struct EqualContext
{
  EqualState state;
  long vector_index;
  const Cell *x;
  const Cell *y;
};


bool
R5RSMiscProcedure::equal(const Cell *x, const Cell *y)
{
  EqualContext current_context = { EQUAL_STATE_INITIAL, 0, x, y };
  vector<EqualContext> equal_stack;

  while (true)
    {
      x = current_context.x;
      y = current_context.y;

      if (x->is_pair())
	{
	  if (!y->is_pair())
	    return false;

	  if (current_context.state == EQUAL_STATE_INITIAL)
	    {
	      current_context.state = EQUAL_STATE_PAIR;
	      equal_stack.push_back(current_context);
	      current_context.state = EQUAL_STATE_INITIAL;
	      current_context.x = CAR(x);
	      current_context.y = CAR(y);
	      continue;
	    }

	  assert(current_context.state == EQUAL_STATE_PAIR);

	  x = CDR(x);
	  y = CDR(y);
	  if (x == Cell::nil())
	    {
	      if (y != Cell::nil())
		return false;

	      current_context.state = EQUAL_STATE_INITIAL;
	      goto done;
	    }

	  if (x->is_pair())
	    {
	      if (!y->is_pair())
		return false;

	      current_context.x = x;
	      current_context.y = y;
	      equal_stack.push_back(current_context);
	      current_context.state = EQUAL_STATE_INITIAL;
	      current_context.x = CAR(x);
	      current_context.y = CAR(y);
	      continue;
	    }
	  // The last elements for improper lists.
	  current_context.state = EQUAL_STATE_INITIAL;
	}

      if (x->is_vector())
	{
	  if (current_context.vector_index == 0)
	    {
	      if (!y->is_vector())
		return false;
	      if (x->get_vector_length() != y->get_vector_length())
		return false;

	      current_context.state = EQUAL_STATE_VECTOR;
	    }

	  assert(current_context.state == EQUAL_STATE_VECTOR);

	  if (x->get_vector_length() > current_context.vector_index)
	    {
	      assert(current_context.x == x);
	      assert(current_context.y == y);

	      equal_stack.push_back(current_context);
	      current_context.state = EQUAL_STATE_INITIAL;
	      current_context.x
		= x->get_vector_element(current_context.vector_index);
	      current_context.y
		= y->get_vector_element(current_context.vector_index);
	      current_context.vector_index = 0;
	      continue;
	    }
	  else
	    {
	      current_context.state = EQUAL_STATE_INITIAL;
	      goto done;
	    }
	}

      if (!eqv(x, y))
	return false;

    done:
      assert(current_context.state == EQUAL_STATE_INITIAL);

      if (current_context.state == EQUAL_STATE_INITIAL)
	{
	  if (equal_stack.empty())
	    return true;

	  current_context = equal_stack.back();
	  equal_stack.pop_back();
	  if (current_context.state == EQUAL_STATE_VECTOR)
	    ++current_context.vector_index;
	}
    }
}


#define ARG(n)	context.frame->load_variable(argument_pointer + (n))


Cell *
R5RSMiscProcedure::eq(VirtualMachine &vm, Register &context,
		      unsigned long argument_pointer,
		      unsigned long nargs, void *data)
{
  return Cell::ts_bool(ARG(0) == ARG(1));
}


Cell *
R5RSMiscProcedure::eqv(VirtualMachine &vm, Register &context,
		       unsigned long argument_pointer,
		       unsigned long nargs, void *data)
{
  return Cell::ts_bool(eqv(ARG(0), ARG(1)));
}


Cell *
R5RSMiscProcedure::equal(VirtualMachine &vm, Register &context,
			 unsigned long argument_pointer,
			 unsigned long nargs, void *data)
{
  return Cell::ts_bool(equal(ARG(0), ARG(1)));
}


Cell *
R5RSMiscProcedure::is_not(VirtualMachine &vm, Register &context,
			  unsigned long argument_pointer,
			  unsigned long nargs, void *data)
{
  return Cell::ts_bool(ARG(0) == Cell::f());
}


Cell *
R5RSMiscProcedure::is_boolean(VirtualMachine &vm, Register &context,
			      unsigned long argument_pointer,
			      unsigned long nargs, void *data)
{
  Cell *arg = ARG(0);
  return Cell::ts_bool(arg == Cell::f() || arg == Cell::t());
}


Cell *
R5RSMiscProcedure::is_symbol(VirtualMachine &vm, Register &context,
			     unsigned long argument_pointer,
			     unsigned long nargs, void *data)
{
  return Cell::ts_bool(ARG(0)->is_symbol());
}


Cell *
R5RSMiscProcedure::symbol2string(VirtualMachine &vm, Register &context,
				 unsigned long argument_pointer,
				 unsigned long nargs, void *data)
{
  return ARG(0)->get_symbol_string();
}


Cell *
R5RSMiscProcedure::string2symbol(VirtualMachine &vm, Register &context,
				 unsigned long argument_pointer,
				 unsigned long nargs, void *data)
{
  Cell *arg = ARG(0);
  if (!arg->is_immutable())
    arg = vm.mk_counted_string(arg->get_string(), arg->get_strlen());

  return vm.core.mk_symbol(arg);
}


Cell *
R5RSMiscProcedure::is_procedure(VirtualMachine &vm, Register &context,
				unsigned long argument_pointer,
				unsigned long nargs, void *data)
{
  return Cell::ts_bool(ARG(0)->is_procedure());
}


static const char trampoline_bytecode[] = OPS_TRAMPOLINE;


Cell *
R5RSMiscProcedure::apply(VirtualMachine &vm, Register &context,
			 unsigned long argument_pointer,
			 unsigned long nargs, void *data)
{
  assert(context.trampoline_frame == Cell::nil());
  assert(context.nargs == 0);
  assert(context.saved_pc == NULL);

  unsigned long rest_args_index = nargs - 1;
  VMStackRoot rest_args(vm, ARG(rest_args_index));
  if (!R5RSListProcedure::is_list(rest_args))
    return signal_error(vm, "apply: rest-args must be of list:", rest_args);

  Cell *proc = ARG(0);
  context.value = proc;

  long len = nargs - 2 + R5RSListProcedure::list_length(rest_args);
  long frame_size;
  if (proc->is_closure())
    {
      frame_size = proc->get_closure_frame_size();
      if (len >= frame_size)
	frame_size = len + 1;
    }
  else
    frame_size = len + 1;

  Cell *frame = vm.frame_new(Cell::nil(), frame_size);
  // ARG(-1) contains the continuation of `apply' procedure.
  frame->store_variable(vm.memory_system, 0, ARG(-1));

  long i = 1;
  for (unsigned long n = 1; n < rest_args_index; n++)
    frame->store_variable(vm.memory_system, i++, ARG(n));

  while (i <= len)
    {
      frame->store_variable(vm.memory_system, i++, CAR(rest_args));
      rest_args = CDR(rest_args);
    }

  context.trampoline_frame = frame;
  context.nargs = len + 1;
  context.saved_pc = context.pc;
  context.pc = trampoline_bytecode;

  return Cell::trampoline();
}


Cell *
R5RSMiscProcedure::map(VirtualMachine &vm, Register &context,
		       unsigned long argument_pointer,
		       unsigned long nargs, void *data)
{
#warning "Implement this!: map"
  return signal_error(vm, "map: not implemented yet");
}


Cell *
R5RSMiscProcedure::for_each(VirtualMachine &vm, Register &context,
			    unsigned long argument_pointer,
			    unsigned long nargs, void *data)
{
#warning "Implement this!: for-each"
  return signal_error(vm, "for-each: not implemented yet");
}


Cell *
R5RSMiscProcedure::call_cc(VirtualMachine &vm, Register &context,
			   unsigned long argument_pointer,
			   unsigned long nargs, void *data)
{
  assert(context.trampoline_frame == Cell::nil());
  assert(context.nargs == 0);
  assert(context.saved_pc == NULL);

  Cell *proc = ARG(0);
  context.value = proc;

  long frame_size;
  if (proc->is_closure())
    {
      frame_size = proc->get_closure_frame_size();
      if (frame_size < 2)
	frame_size = 2;
    }
  else
    frame_size = 2;

  // ARG(-1) contains the continuation of `call-with-current-continuation'
  // procedure.
  VMStackRoot cont0(vm, ARG(-1));
  VMStackRoot continuation(vm, Cell::nil());
  if (cont0 != Cell::nil())
    {
      // call/cc is used for a tail-call.
      context.procedure = cont0->get_continuation_procedure();
      assert(context.procedure->is_closure() || cont0 == vm.core.meta_return);

      if (context.procedure->is_closure())
	{
	  context.pc = cont0->get_continuation_pc();
	  context.constant_pool
	    = context.procedure->get_closure_constant_pool();
	  context.frame = cont0->get_continuation_frame();
	  continuation = vm.mk_current_continuation_full(context);
	}
      else
	continuation = cont0;
    }
  else
    continuation = vm.mk_current_continuation_full(context);


  Cell *frame = vm.frame_new(Cell::nil(), frame_size);
  frame->store_variable(vm.memory_system, 0, cont0);
  frame->store_variable(vm.memory_system, 1, continuation);

  context.trampoline_frame = frame;
  context.nargs = 2;
  context.saved_pc = context.pc;
  context.pc = trampoline_bytecode;

  return Cell::trampoline();
}


Cell *
R5RSMiscProcedure::values(VirtualMachine &vm, Register &context,
			  unsigned long argument_pointer,
			  unsigned long nargs, void *data)
{
#warning "Implement this!: values"
  return signal_error(vm, "values: not implemented yet");
}


Cell *
R5RSMiscProcedure::call_with_values(VirtualMachine &vm, Register &context,
				    unsigned long argument_pointer,
				    unsigned long nargs, void *data)
{
#warning "Implement this!: call-with-values"
  return signal_error(vm, "call-with-values: not implemented yet");
}


Cell *
R5RSMiscProcedure::dynamic_wind(VirtualMachine &vm, Register &context,
				unsigned long argument_pointer,
				unsigned long nargs, void *data)
{
#warning "Implement this!: dynamic-wind"
  return signal_error(vm, "dynamic-wind: not implemented yet");
}


Cell *
R5RSMiscProcedure::eval(VirtualMachine &vm, Register &context,
			unsigned long argument_pointer,
			unsigned long nargs, void *data)
{
  /*
   * XXX: This implementation isn't properly tail-recursive.
   */
  Register eval_context(vm, ARG(1), Cell::nil(), Cell::nil(), Cell::nil(),
			context.input_port, context.output_port);

  return vm.eval(eval_context, ARG(0), ARG(1));
}


Cell *
R5RSMiscProcedure::scheme_report_environment(VirtualMachine &vm,
					     Register &context,
					     unsigned long argument_pointer,
					     unsigned long nargs, void *data)
{
  if (ARG(0)->get_integer0() == 5)
    return vm.core.get_r5rs_environment();

  return signal_error(vm,
		      "scheme-report-environment: unsupported version: ",
		      ARG(0));
}


Cell *
R5RSMiscProcedure::null_environment(VirtualMachine &vm, Register &context,
				    unsigned long argument_pointer,
				    unsigned long nargs, void *data)
{
  if (ARG(0)->get_integer0() == 5)
    return vm.core.get_null_environment();

  return signal_error(vm, "null-environment: unsupported version:", ARG(0));
}


Cell *
R5RSMiscProcedure::interaction_environment(VirtualMachine &vm,
					   Register &context,
					   unsigned long argument_pointer,
					   unsigned long nargs, void *data)
{
  return vm.core.get_interaction_environment();
}

}
