/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: r5rs_primitive_syntax.cc,v 1.16 2010/08/15 15:39:47 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libtscore/r5rs_primitive_syntax.h"

#include "libtscore/bytecode_buffer.h"
#include "libtscore/syntax.h"
#include "libtscore/virtual_machine.h"
#include "libtscore/writer.h"


#include <cassert>


namespace LibTSCore
{

size_t
Define::emit_bytecode(VirtualMachine &vm, Register &context,
		      BytecodeBuffer *buffer,
		      Cell *expression, Cell *environment,
		      Cell *constant_pool, Cell *hint,
		      bool is_tail_context)
{
  assert(expression->is_pair());
  assert(is_toplevel(vm, environment, hint));

  Cell *tmp_env = environment;
  while (true)
    {
      Cell *parent = tmp_env->get_parent_environment();
      if (parent == Cell::nil())
	{
	  if (tmp_env == vm.core.get_interaction_environment())
	    break;

	  return signal_error(vm, "define: in immutable environment: ",
			      expression);
	}

      tmp_env = parent;
    }

  VMStackRoot toplevel_environment(vm, tmp_env);

  if (!CDR(expression)->is_pair()
      || !CDDR(expression)->is_pair())
    return signal_error(vm, "define: syntax error: ", expression);

  VMStackRoot form(vm, expression);
  Cell *first = CADR(expression);
  Cell *second = CDDR(expression);

  VMStackRoot variable(vm, Cell::nil());
  VMStackRoot env(vm, environment);
  VMStackRoot cpool(vm, constant_pool);
  VMStackRoot hnt(vm, extend_hint(vm, CAR(form), hint));

  size_t len;
  VMStackRoot name(vm, Cell::nil());
  if (first->is_pair())
    {
      // procedure definition
      if (CAR(first)->is_symbol())
	variable = CAR(first);
      else if (CAR(first)->is_binding())
	variable = const_cast<Cell *>(CAR(first)->get_bound_symbol());
      else
	return signal_error(vm, "define: syntax error: ", form);

      if (toplevel_environment->lookup(variable, false)
	  == Cell::nil())
	name = static_cast<Cell *>(variable);

      if (bind_variable(vm, toplevel_environment, variable, hnt) == Cell::nil())
	return signal_error(vm, "define: syntax error: ", form);

      len = Lambda::emit_lambda(vm, context, buffer, Cell::nil(), env,
				cpool, hnt, CDADR(form), CDDR(form));
    }
  else
    {
      if (first->is_symbol())
	variable = first;
      else if (first->is_binding())
	variable = const_cast<Cell *>(first->get_bound_symbol());
      else 
	return signal_error(vm, "define: syntax error: ", form);

      if (CDR(second) != Cell::nil())
	return signal_error(vm, "define: syntax error: ", form);

      if (toplevel_environment->lookup(variable, false) == Cell::nil())
	name = static_cast<Cell *>(variable);

      if (bind_variable(vm, toplevel_environment, variable, hnt) == Cell::nil())
	return signal_error(vm, "define: syntax error: ", form);

      len = bytecompile(vm, context, buffer,
			CADDR(form), env, cpool, hnt, false);
    }

  if (len == 0)
    {
      if (name != Cell::nil())
	toplevel_environment->unbind_variable(name);

      return signal_error(vm, "define: syntax error: ", form);
    }

  len += emit_variable_assignment(vm, context, buffer, variable,
				  toplevel_environment, cpool);
  if (is_tail_context)
    return len + buffer->append(OP_APPLY_CONTINUATION0);

  return len;
}


size_t
Quote::emit_bytecode(VirtualMachine &vm, Register &context,
		     BytecodeBuffer *buffer,
		     Cell *expression, Cell *environment,
		     Cell *constant_pool, Cell *hint,
		     bool is_tail_context)
{
  assert(expression->is_pair());

  size_t len = emit_literal_expression(vm, context, buffer, CADR(expression),
				       constant_pool);
  assert(len > 0);

  if (!is_tail_context)
    return len;

  return len + buffer->append(OP_APPLY_CONTINUATION0);
}


Cell *
Lambda::closure_template_new(VirtualMachine &vm, Register &context,
			     Cell *expression, Cell *environment,
			     Cell *constant_pool, Cell *hint,
			     Cell *formal_args, Cell *procedure_body)
{
  VMStackRoot form(vm, expression);
  VMStackRoot env(vm, environment);
  VMStackRoot cpool(vm, constant_pool);
  VMStackRoot hnt(vm, hint);
  VMStackRoot formal(vm, formal_args);
  VMStackRoot body(vm, procedure_body);

  env = vm.environment_new(env);
  if (env == Cell::nil())
    return 0;	// out of memory

  long min_arity;
  long max_arity;
#ifdef NDEBUG
  env->allocate_anonymous_slot();
#else
  long cont_slot = env->allocate_anonymous_slot();
  assert(cont_slot == 0);	// for implicit continuation
#endif

  VMStackRoot sym(vm, Cell::nil());
  min_arity = 1;
  while (true)
    {
      if (formal == Cell::nil())
	{
	  max_arity = min_arity;
	  break;
	}

      if (formal->is_pair())
	{
	  sym = CAR(formal);
	  const Cell *binding;
	  if (sym->is_binding())
	    sym = const_cast<Cell *>(sym->get_bound_symbol());
	  else if (!sym->is_symbol())
	    {
	      signal_error(vm, "not a symbol: ", sym);
	      return Cell::nil();
	    }

	  binding = env->lookup(sym, false);
	  if (binding != Cell::nil())
	    {
	      signal_error(vm, "duplicated formal argument: ",
			   sym->get_symbol_string());
	      return Cell::nil();
	    }
	  
	  binding = bind_variable(vm, env, sym, hnt);
	  if (binding == Cell::nil())
	    return Cell::nil();

	  if (CAR(formal)->is_binding())
	    body = rename_variable(vm, body, CAR(formal), sym);

	  min_arity++;
	}
      else
	{
	  const Cell *binding;
	  sym = formal;
	  if (sym->is_binding())
	    sym = const_cast<Cell *>(sym->get_bound_symbol());
	  else if (!sym->is_symbol())
	    {
	      signal_error(vm, "not a symbol: ", formal);
	      return Cell::nil();
	    }

	  // (lambda (var1 ... varN . varN+1) <body>) syntax
	  binding = env->lookup(sym, false);
	  if (binding != Cell::nil())
	    {
	      signal_error(vm, "duplicated formal argument: ",
			   sym->get_symbol_string());
	      return Cell::nil();
	    }

	  binding = bind_variable(vm, env, sym, hnt);
	  if (binding == Cell::nil())
	    return Cell::nil();

	  if (formal->is_binding())
	    body = rename_variable(vm, body, formal, sym);

	  max_arity = min_arity + 1;
	  break;
	}

      formal = CDR(formal);
    }

  // emit codes for body.
  BytecodeBuffer *bytecode = new BytecodeBuffer();
  size_t len = bytecompile_sequence(vm, context, bytecode,
				    body, env, cpool, hint, true);
  if (len > 0)
    {
      BytecodeProcedure *proc
	= new BytecodeProcedure(bytecode,
				env->get_required_frame_size(),
				min_arity, max_arity);
      Cell *tmpl = vm.mk_closure_template(cpool, form, proc);

      if (vm.get_bytecompile_tracing(context))
	{
	  Port *port = context.output_port->get_port();
	  Writer::write_datum(tmpl, port, false, true);
	  port->append('\n');
	}

      return tmpl;
    }

  // bytecompilation error
  delete bytecode;
  return Cell::nil();
}


size_t
Lambda::emit_lambda(VirtualMachine &vm, Register &context,
		    BytecodeBuffer *buffer,
		    Cell *expression, Cell *environment,
		    Cell *constant_pool, Cell *hint,
		    Cell *formal_args, Cell *procedure_body)
{
  VMStackRoot form(vm, expression);
  VMStackRoot env(vm, environment);
  VMStackRoot cpool(vm, constant_pool);
  VMStackRoot hnt(vm, hint);
  VMStackRoot formal(vm, formal_args);
  VMStackRoot body(vm, procedure_body);

  if (form == Cell::nil())
    {
      form = vm.mk_cons(formal, body, true);
      form = vm.mk_cons(vm.core.internal_lambda_symbol, form, true);
    }

  Cell *proc_cpool = Cell::constant_pool_new(vm.get_memory_system());
  Cell *closure_template = closure_template_new(vm, context, form, env,
						proc_cpool, hnt, formal, body);
  size_t len;
  if (closure_template != Cell::nil())
    {
      long index = cpool->register_constant(closure_template);
      size_t off = buffer->get_current_offset();
      len = buffer->append(OP_NOP);		// dummy;
      size_t imm_len = buffer->append_unsigned(index);
      switch (imm_len)
	{
	case 1:
	  buffer->set_opcode(off, OP_CONST_P1);
	  break;

	case 2:
	  buffer->set_opcode(off, OP_CONST_P2);
	  break;

	case 4:
	  buffer->set_opcode(off, OP_CONST_P4);
	  break;

	default:
#ifndef NDEBUG
	  assert(0);	// should never reached
#else
	  abort();		// should never reached
#endif
	}

      len += imm_len;
      len += buffer->append(OP_LAMBDA);
    }
  else
    len = 0;	// error

  return len;
}


size_t
Lambda::emit_bytecode(VirtualMachine &vm, Register &context,
		      BytecodeBuffer *buffer,
		      Cell *expression, Cell *environment,
		      Cell *constant_pool, Cell *hint,
		      bool is_tail_context)
{
  assert(expression->is_pair());

  VMStackRoot form(vm, expression);
  VMStackRoot env(vm, environment);
  VMStackRoot cpool(vm, constant_pool);
  VMStackRoot hnt(vm, extend_hint(vm, CAR(form), hint));

  Cell *tmp_exp = CDR(form);	// skip the keyword for `lambda' syntax
  if (tmp_exp->is_pair())
    {
      Cell *formal = CAR(tmp_exp);
      tmp_exp = CDR(tmp_exp);
      if (tmp_exp->is_pair())
	{
	  size_t len = emit_lambda(vm, context, buffer, form, env, cpool, hnt,
				   formal, tmp_exp);
	  if (len > 0)
	    {
	      if (is_tail_context)
		len += buffer->append(OP_APPLY_CONTINUATION0);

	      return len;
	    }
	}
    }

  return signal_error(vm, "lambda: syntax error: ", form);
}


size_t
If::emit_if(VirtualMachine &vm, Register &context, BytecodeBuffer *buffer,
	    Cell *expression,
	    Cell *environment, Cell *constant_pool, Cell *hint,
	    Cell *test, Cell *consequent, bool is_tail_context)
{
  VMStackRoot form(vm, expression);
  VMStackRoot env(vm, environment);
  VMStackRoot cpool(vm, constant_pool);
  VMStackRoot conseq(vm, consequent);
  VMStackRoot hnt(vm, extend_hint(vm, CAR(form), hint));

  size_t len = bytecompile(vm, context, buffer, test, env, cpool, hnt, false);
  if (len == 0)
    return signal_error(vm, "if: syntax error: ", form);

  if (is_tail_context)
    {
      len += buffer->append(OP_IF_T1);
      len += buffer->append_signed_8(2);
      len += buffer->append(OP_APPLY_CONTINUATION0);
      size_t conseq_len = bytecompile(vm, context, buffer, conseq,
				      env, cpool, hnt, true);
      if (conseq_len > 0)
	return len + conseq_len;

      return signal_error(vm, "if: syntax error: ", form);
    }

  buffer->set_include_forward_branch();

  len += buffer->append(OP_IF_F2);	// We are a little conservative here.
  size_t branch_off = buffer->get_current_offset();
  len += buffer->append_signed_16(2);	// dummy

  size_t conseq_len = bytecompile(vm, context, buffer, conseq,
				  env, cpool, hnt, false);
  if (conseq_len == 0)
    return signal_error(vm, "if: syntax error: ", form);

  size_t offset = buffer->get_current_offset() - branch_off;
  if (offset > 32767)
    return signal_error(vm, "unsupported too long consequent expression: ",
			form);

  buffer->set_signed_16(branch_off, offset);

  return len + conseq_len;
}


size_t
If::emit_if(VirtualMachine &vm, Register &context, BytecodeBuffer *buffer,
	    Cell *expression,
	    Cell *environment, Cell *constant_pool, Cell *hint,
	    Cell *test, Cell *consequent, Cell *alternate,
	    bool is_tail_context)
{
  VMStackRoot form(vm, expression);
  VMStackRoot env(vm, environment);
  VMStackRoot cpool(vm, constant_pool);
  VMStackRoot conseq(vm, consequent);
  VMStackRoot alt(vm, alternate);
  VMStackRoot hnt(vm, extend_hint(vm, CAR(form), hint));

  size_t len = bytecompile(vm, context, buffer, test, env, cpool, hnt, false);
  if (len == 0)
    return signal_error(vm, "if: syntax error: ", form);

  len += buffer->append(OP_IF_F2);	// We are a little conservative.
  size_t branch_off = buffer->get_current_offset();
  len += buffer->append_signed_16(2);	// dummy

  size_t conseq_len = bytecompile(vm, context, buffer, conseq, env, cpool, hnt,
				  is_tail_context);
  if (conseq_len == 0)
    return signal_error(vm, "if: syntax error: ", form);

  len += conseq_len;

  if (is_tail_context)
    {
      size_t offset = buffer->get_current_offset() - branch_off;
      if (offset > 32767)
	return signal_error(vm, "unsupported too long consequent expression: ",
			    form);

      buffer->set_signed_16(branch_off, offset);

      size_t alt_len = bytecompile(vm, context, buffer, alt, env, cpool, hnt,
				   is_tail_context);
      if (alt_len == 0)
	return signal_error(vm, "if: syntax error: ", form);

      return len + alt_len;
    }

  buffer->set_include_forward_branch();
  conseq_len += buffer->append(OP_GOTO2);	// a little conservative.
  size_t goto_off = buffer->get_current_offset();
  conseq_len += buffer->append_signed_16(2);	// dummy;

  size_t offset = buffer->get_current_offset() - branch_off;
  if (offset > 32767)
    return signal_error(vm, "unsupported too long consequent expression: ",
			form);

  buffer->set_signed_16(branch_off, offset);

  size_t alt_len = bytecompile(vm, context, buffer, alt, env, cpool, hnt,
			       is_tail_context);
  if (alt_len == 0)
    return signal_error(vm, "if: syntax error: ", form);

  offset = buffer->get_current_offset() - goto_off;
  if (offset > 32767)
    return signal_error(vm, "unsupported too long alternate expression: ",
			form);

  buffer->set_signed_16(goto_off, offset);

  return len + alt_len;
}


size_t
If::emit_bytecode(VirtualMachine &vm, Register &context,
		  BytecodeBuffer *buffer,
		  Cell *expression, Cell *environment,
		  Cell *constant_pool, Cell *hint,
		  bool is_tail_context)
{
  assert(expression->is_pair());

  if (!CDR(expression)->is_pair()
      || !CDDR(expression)->is_pair())
    return signal_error(vm, "if: syntax error: ", expression);

  Cell *test = CADR(expression);
  Cell *consequent = CADDR(expression);
  if (CDDDR(expression) == Cell::nil())
    return emit_if(vm, context, buffer, expression, environment, constant_pool,
		   hint,
		   test, consequent, is_tail_context);

  Cell *alternate = CADDDR(expression);
  if (CDDDDR(expression) != Cell::nil())
    return signal_error(vm, "if: syntax error: ", expression);

  return emit_if(vm, context, buffer, expression, environment, constant_pool,
		 hint,
		 test, consequent, alternate, is_tail_context);
}


size_t
Assignment::emit_bytecode(VirtualMachine &vm, Register &context,
			  BytecodeBuffer *buffer,
			  Cell *expression, Cell *environment,
			  Cell *constant_pool, Cell *hint,
			  bool is_tail_context)
{
  assert(expression->is_pair());

  if (!CDR(expression)->is_pair()
      || !CDDR(expression)->is_pair())
    return signal_error(vm, "set!: syntax error: ", expression);

  VMStackRoot form(vm, expression);
  VMStackRoot variable(vm, Cell::nil());
  VMStackRoot env(vm, environment);
  VMStackRoot cpool(vm, constant_pool);
  VMStackRoot hnt(vm, extend_hint(vm, CAR(form), hint));

  Cell *first = CADR(form);
  if (first->is_pair())
    return signal_error(vm, "set!: syntax error: ", expression);

  if (first->is_symbol())
    variable = first;
  else if (first->is_binding())
    variable = const_cast<Cell *>(first->get_bound_symbol());
  else
    return signal_error(vm, "set!: syntax error: ", expression);

  Cell *second = CDDR(form);
  if (CDR(second) != Cell::nil())
    return signal_error(vm, "set!: syntax error: ", expression);

  size_t len = bytecompile(vm, context, buffer, CAR(second),
			   environment, constant_pool, hnt, false);

  if (len == 0)
    return signal_error(vm, "set!: syntax error: ", form);

  size_t set_len = emit_variable_assignment(vm, context, buffer, variable,
					    env, cpool);
  if (set_len == 0)
    return signal_error(vm, "set!: syntax error: ", form);

  if (is_tail_context)
    return len + set_len + buffer->append(OP_APPLY_CONTINUATION0);

  return len + set_len;
}

}
