/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: r5rs_string_procedure.cc,v 1.13 2010/08/15 15:39:47 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libtscore/r5rs_string_procedure.h"

#include "libtscore/port.h"
#include "libtscore/tscore.h"
#include "libtscore/virtual_machine.h"

#include <cstring>


namespace LibTSCore
{

NativeProcedure::Info R5RSStringProcedure::procedures[] = {
  { &R5RSStringProcedure::is_string,
    "string?", TST_NONE, 1, 1 },
  { &R5RSStringProcedure::make_string,
    "make-string", TST_NONNEG TST_CHAR, 1, 2 },
  { &R5RSStringProcedure::string_from_chars,
    "string", TST_CHAR, 0, INF_ARG },
  { &R5RSStringProcedure::string_length,
    "string-length", TST_STRING, 1, 1 },
  { &R5RSStringProcedure::string_ref,
    "string-ref", TST_STRING TST_NONNEG, 2, 2 },
  { &R5RSStringProcedure::string_set,
    "string-set!", TST_MUTABLE_STRING TST_NONNEG TST_CHAR, 3, 3 },
  { &R5RSStringProcedure::string_eq,
    "string=?", TST_STRING, 2, 2 },
  { &R5RSStringProcedure::string_ci_eq,
    "string-ci=?", TST_STRING, 2, 2 },
  { &R5RSStringProcedure::string_lt,
    "string<?", TST_STRING, 2, 2 },
  { &R5RSStringProcedure::string_gt,
    "string>?", TST_STRING, 2, 2 },
  { &R5RSStringProcedure::string_le,
    "string<=?", TST_STRING, 2, 2 },
  { &R5RSStringProcedure::string_ge,
    "string>=?", TST_STRING, 2, 2 },
  { &R5RSStringProcedure::string_ci_lt,
    "string-ci<?", TST_STRING, 2, 2 },
  { &R5RSStringProcedure::string_ci_gt,
    "string-ci>?", TST_STRING, 2, 2 },
  { &R5RSStringProcedure::string_ci_le,
    "string-ci<=?", TST_STRING, 2, 2 },
  { &R5RSStringProcedure::string_ci_ge,
    "string-ci>=?", TST_STRING, 2, 2 },
  { &R5RSStringProcedure::substring,
    "substring", TST_STRING TST_NONNEG, 2, 3 },
  { &R5RSStringProcedure::string_append,
    "string-append", TST_STRING, 0, INF_ARG },
  { &R5RSStringProcedure::string2list,
    "string->list", TST_STRING, 1, 1 },
  { &R5RSStringProcedure::list2string,
    "list->string", TST_LIST, 1, 1 },
  { &R5RSStringProcedure::string_copy,
    "string-copy", TST_STRING, 1, 1 },
  { &R5RSStringProcedure::string_fill,
    "string-fill!", TST_MUTABLE_STRING TST_CHAR, 2, 2 }
};


Cell *
R5RSStringProcedure::assign_procedures(Cell *environment, Cell *frame)
{
  TSStackRoot env(core, environment);
  TSStackRoot frm(core, frame);
  size_t n_procs = sizeof(procedures) / sizeof(Info);
  for (size_t i = 0; i < n_procs; i++)
    assign_procedure(env, frm, &procedures[i], this);

  return env;
}


#define ARG(n)	context.frame->load_variable(argument_pointer + (n))


Cell *
R5RSStringProcedure::is_string(VirtualMachine &vm, Register &context,
			       unsigned long argument_pointer,
			       unsigned long nargs, void *data)
{
  return Cell::ts_bool(ARG(0)->is_string());
}


Cell *
R5RSStringProcedure::make_string(VirtualMachine &vm, Register &context,
				 unsigned long argument_pointer,
				 unsigned long nargs, void *data)
{
  char fill;
  size_t len = ARG(0)->get_integer0();
  if (nargs == 2)
    fill = static_cast<char>(ARG(1)->get_character());
  else
    fill = ' ';
  return vm.mk_empty_string(len, fill);
}


Cell *
R5RSStringProcedure::string_from_chars(VirtualMachine &vm, Register &context,
				       unsigned long argument_pointer,
				       unsigned long nargs, void *data)
{
  OutputStringPort string_buf;
  for (unsigned long i = 0; i < nargs; i++)
    {
      Cell *c = ARG(i);
      if (c->is_character())
	string_buf.append(c->get_character());
      else
	return signal_error(vm, "string: not a character: ", c);
    }
  return vm.mk_counted_string(string_buf.get_output_string(),
			      string_buf.get_output_strlen());
}


Cell *
R5RSStringProcedure::string_length(VirtualMachine &vm, Register &context,
				   unsigned long argument_pointer,
				   unsigned long nargs, void *data)
{
  return vm.mk_integer(static_cast<long>(ARG(0)->get_strlen()));
}


Cell *
R5RSStringProcedure::string_ref(VirtualMachine &vm, Register &context,
				unsigned long argument_pointer,
				unsigned long nargs, void *data)
{
  Cell *arg0 = ARG(0);
  const unsigned char *str
    = reinterpret_cast<const unsigned char *>(arg0->get_string());
  size_t index = ARG(1)->get_integer0();

  if (index >= arg0->get_strlen())
    return signal_error(vm, "string-ref: index out of bounds: ", ARG(1));
  return vm.mk_character(str[index]);
}


Cell *
R5RSStringProcedure::string_set(VirtualMachine &vm, Register &context,
				unsigned long argument_pointer,
				unsigned long nargs, void *data)
{
  Cell *arg0 = ARG(0);
  char *str = arg0->get_string0();
  size_t index = ARG(1)->get_integer();

  if (index >= arg0->get_strlen())
    return signal_error(vm, "string-set!: index out of bounds: ", ARG(1));

  str[index] = static_cast<char>(ARG(2)->get_character());
  return arg0;
}


Cell *
R5RSStringProcedure::string_eq(VirtualMachine &vm, Register &context,
			       unsigned long argument_pointer,
			       unsigned long nargs, void *data)
{
  Cell *arg0 = ARG(0);
  Cell *arg1 = ARG(1);
  size_t len = arg0->get_strlen();
  if (len != arg1->get_strlen())
    return Cell::f();

  const char *lstr = arg0->get_string();
  const char *rstr = arg1->get_string();
  return Cell::ts_bool(strncmp(lstr, rstr, len) == 0);
}


Cell *
R5RSStringProcedure::string_ci_eq(VirtualMachine &vm, Register &context,
				  unsigned long argument_pointer,
				  unsigned long nargs, void *data)
{
  Cell *arg0 = ARG(0);
  Cell *arg1 = ARG(1);
  size_t len = arg0->get_strlen();
  if (len != arg1->get_strlen())
    return Cell::f();

  const char *lstr = arg0->get_string();
  const char *rstr = arg1->get_string();
  return Cell::ts_bool(strncasecmp(lstr, rstr, len) == 0);
}


Cell *
R5RSStringProcedure::string_lt(VirtualMachine &vm, Register &context,
			       unsigned long argument_pointer,
			       unsigned long nargs, void *data)
{
  Cell *arg0 = ARG(0);
  Cell *arg1 = ARG(1);
  size_t len = arg0->get_strlen();
  if (len > arg1->get_strlen())
    len = arg1->get_strlen();

  const char *lstr = arg0->get_string();
  const char *rstr = arg1->get_string();
  return Cell::ts_bool(strncmp(lstr, rstr, len) < 0);
}


Cell *
R5RSStringProcedure::string_gt(VirtualMachine &vm, Register &context,
			       unsigned long argument_pointer,
			       unsigned long nargs, void *data)
{
  Cell *arg0 = ARG(0);
  Cell *arg1 = ARG(1);
  size_t len = arg0->get_strlen();
  if (len > arg1->get_strlen())
    len = arg1->get_strlen();

  const char *lstr = arg0->get_string();
  const char *rstr = arg1->get_string();
  return Cell::ts_bool(strncmp(lstr, rstr, len) > 0);
}


Cell *
R5RSStringProcedure::string_le(VirtualMachine &vm, Register &context,
			       unsigned long argument_pointer,
			       unsigned long nargs, void *data)
{
  Cell *arg0 = ARG(0);
  Cell *arg1 = ARG(1);
  size_t len = arg0->get_strlen();
  if (len > arg1->get_strlen())
    len = arg1->get_strlen();

  const char *lstr = arg0->get_string();
  const char *rstr = arg1->get_string();
  return Cell::ts_bool(strncmp(lstr, rstr, len) <= 0);
}


Cell *
R5RSStringProcedure::string_ge(VirtualMachine &vm, Register &context,
			       unsigned long argument_pointer,
			       unsigned long nargs, void *data)
{
  Cell *arg0 = ARG(0);
  Cell *arg1 = ARG(1);
  size_t len = arg0->get_strlen();
  if (len > arg1->get_strlen())
    len = arg1->get_strlen();

  const char *lstr = arg0->get_string();
  const char *rstr = arg1->get_string();
  return Cell::ts_bool(strncmp(lstr, rstr, len) >= 0);
}


Cell *
R5RSStringProcedure::string_ci_lt(VirtualMachine &vm, Register &context,
				  unsigned long argument_pointer,
				  unsigned long nargs, void *data)
{
  Cell *arg0 = ARG(0);
  Cell *arg1 = ARG(1);
  size_t len = arg0->get_strlen();
  if (len > arg1->get_strlen())
    len = arg1->get_strlen();

  const char *lstr = arg0->get_string();
  const char *rstr = arg1->get_string();
  return Cell::ts_bool(strncasecmp(lstr, rstr, len) < 0);
}


Cell *
R5RSStringProcedure::string_ci_gt(VirtualMachine &vm, Register &context,
				  unsigned long argument_pointer,
				  unsigned long nargs, void *data)
{
  Cell *arg0 = ARG(0);
  Cell *arg1 = ARG(1);
  size_t len = arg0->get_strlen();
  if (len > arg1->get_strlen())
    len = arg1->get_strlen();

  const char *lstr = arg0->get_string();
  const char *rstr = arg1->get_string();
  return Cell::ts_bool(strncasecmp(lstr, rstr, len) > 0);
}


Cell *
R5RSStringProcedure::string_ci_le(VirtualMachine &vm, Register &context,
				  unsigned long argument_pointer,
				  unsigned long nargs, void *data)
{
  Cell *arg0 = ARG(0);
  Cell *arg1 = ARG(1);
  size_t len = arg0->get_strlen();
  if (len > arg1->get_strlen())
    len = arg1->get_strlen();

  const char *lstr = arg0->get_string();
  const char *rstr = arg1->get_string();
  return Cell::ts_bool(strncasecmp(lstr, rstr, len) <= 0);
}


Cell *
R5RSStringProcedure::string_ci_ge(VirtualMachine &vm, Register &context,
				  unsigned long argument_pointer,
				  unsigned long nargs, void *data)
{
  Cell *arg0 = ARG(0);
  Cell *arg1 = ARG(1);
  size_t len = arg0->get_strlen();
  if (len > arg1->get_strlen())
    len = arg1->get_strlen();

  const char *lstr = arg0->get_string();
  const char *rstr = arg1->get_string();
  return Cell::ts_bool(strncasecmp(lstr, rstr, len) >= 0);
}


Cell *
R5RSStringProcedure::substring(VirtualMachine &vm, Register &context,
			       unsigned long argument_pointer,
			       unsigned long nargs, void *data)
{
  Cell *arg0 = ARG(0);
  Cell *arg1 = ARG(1);
  size_t index0 = arg1->get_integer0();
  if (index0 > arg0->get_strlen())
    return signal_error(vm, "substring: start out of bounds: ", ARG(1));

  size_t index1;
  if (nargs == 3)
    {
      index1 = ARG(2)->get_integer0();
      if (index1 > arg0->get_strlen() || index1 < index0)
	return signal_error(vm, "substring: end out of bounds: ", ARG(2));
    }
  else
    index1 = arg0->get_strlen();

  return vm.mk_substring_copy(arg0, index0, index1);
}


Cell *
R5RSStringProcedure::string_append(VirtualMachine &vm, Register &context,
				   unsigned long argument_pointer,
				   unsigned long nargs, void *data)
{
  OutputStringPort string_buf;
  for (unsigned long i = 0; i < nargs; i++)
    {
      Cell *str = ARG(i);
      string_buf.append(str->get_string(), str->get_strlen());
    }
  return vm.mk_counted_string(string_buf.get_output_string(),
			      string_buf.get_output_strlen());
}


Cell *
R5RSStringProcedure::string2list(VirtualMachine &vm, Register &context,
				 unsigned long argument_pointer,
				 unsigned long nargs, void *data)
{
  Cell *arg0 = ARG(0);
  const unsigned char *cstr
    = reinterpret_cast<const unsigned char *>(arg0->get_string());
  size_t len = arg0->get_strlen();
  VMStackRoot list(vm, Cell::nil());
  while (len-- > 0)
    {
      Cell *c = vm.mk_character(cstr[len]);
      list = vm.cons(c, list);
    }
  return list;
}


Cell *
R5RSStringProcedure::list2string(VirtualMachine &vm, Register &context,
				 unsigned long argument_pointer,
				 unsigned long nargs, void *data)
{
  OutputStringPort string_buf;
  Cell *list = ARG(0);
  while (list->is_pair())
    {
      if (CAR(list)->is_character())
	string_buf.append(CAR(list)->get_character());
      else
	return signal_error(vm, "list->string: not a character: ", CAR(list));
      list = CDR(list);
    }
  if (list != Cell::nil())
    return signal_error(vm, "list->string: argument #1 must be of list: ",
			ARG(0));

  return vm.mk_counted_string(string_buf.get_output_string(),
			      string_buf.get_output_strlen());
}


Cell *
R5RSStringProcedure::string_copy(VirtualMachine &vm, Register &context,
				 unsigned long argument_pointer,
				 unsigned long nargs, void *data)
{
  Cell *arg0 = ARG(0);
  return vm.mk_counted_string(arg0->get_string(), arg0->get_strlen());
}


Cell *
R5RSStringProcedure::string_fill(VirtualMachine &vm, Register &context,
				 unsigned long argument_pointer,
				 unsigned long nargs, void *data)
{
  Cell *arg0 = ARG(0);
  Cell *arg1 = ARG(1);
  char *cstr = const_cast<char *>(arg0->get_string());
  size_t len = arg0->get_strlen();
  char fill = static_cast<char>(arg1->get_character());
  for (size_t i = 0; i < len; i++)
    cstr[i] = fill;
  return arg0;
}

}
