/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: reader.cc,v 1.23 2009/01/01 07:47:01 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libtscore/reader.h"

#include "libtscore/memory_system.h"
#include "libtscore/tscore.h"
#include "libtscore/virtual_machine.h"

#include <cctype>
#include <cmath>
#include <cstdio>
#include <cstring>


namespace LibTSCore
{

static int
is_digit_2(int c)
{
  return c == '0' || c == '1';
}


static int
is_digit_8(int c)
{
  return c >= '0' && c <= '7';
}


static inline bool
is_exponent_marker(long c)
{
  return strchr("esfdlESFDL", c) != NULL;
}


static inline bool
is_real_terminater(long c)
{
  return c == EOF || strchr(" \t\n\r()\";@+-", c) != NULL;
}


static inline bool
is_one_of(const char *group, long c)
{
  return c == EOF || strchr(group, c) != NULL;
}


static inline bool
is_delimiter(long c)
{
  return c == EOF || strchr(" \t\n\r()\";", c) != NULL;
}


static inline bool
is_special_initial(long c)
{
  return strchr("!$%&*/:<=>?^_~", c) != NULL;
}


static inline bool
is_initial(long c)
{
  return isalpha(c) != 0 || is_special_initial(c);
}


static inline bool
is_special_subsequent(long c)
{
  return strchr("+-.@", c) != NULL;
}


static inline bool
is_subsequent(long c)
{
  return is_initial(c) || isdigit(c) != 0 || is_special_subsequent(c);
}


typedef int (*DigitPredicate)(int);
static DigitPredicate digit_predicate_funcs[] =
{
  is_digit_2,
  is_digit_8,
  static_cast<DigitPredicate>(isdigit),
  static_cast<DigitPredicate>(isxdigit)
};


static long
sscanf_2(const char *src)
{
  long value = 0;
  while (*src != '\0' && (*src == '0' || *src == '1'))
    {
      value <<= 1;
      value += *src - '0';
      ++src;
    }

  return value;
}


static long
sscanf_8(const char *src)
{
  long value;
  if (sscanf(src, "%lo", &value) != 1)
    value = 0;

  return value;
}


static long
sscanf_10(const char *src)
{
  long value;
  if (sscanf(src, "%ld", &value) != 1)
    value = 0;

  return value;
}


static long
sscanf_16(const char *src)
{
  long value;
  if (sscanf(src, "%lx", &value) != 1)
    value = 0;

  return value;
}


typedef long (*DigitSscanf)(const char *src);
static DigitSscanf digit_sscanf_funcs[] =
{
  sscanf_2,
  sscanf_8,
  sscanf_10,
  sscanf_16,
};


long
Reader::check_prefix_r(Port *input_port, Radix &radix, Exactness &exactness)
{
  long c;
  while (true)
    {
      c = input_port->get_char();
      if (c == EOF)
	return c;

      if (c != '#')
	break;

      buffer.append(c);
      c = input_port->get_char();
      switch (c)
	{
	case 'B':
	case 'b':
	  if (radix == RADIX_UNKNOWN)
	    {
	      buffer.append(c);
	      radix = RADIX_2;
	      continue;
	    }
	  break;

	case 'D':
	case 'd':
	  if (radix == RADIX_UNKNOWN)
	    {
	      buffer.append(c);
	      radix = RADIX_10;
	      continue;
	    }
	  break;

	case 'O':
	case 'o':
	  if (radix == RADIX_UNKNOWN)
	    {
	      buffer.append(c);
	      radix = RADIX_8;
	      continue;
	    }
	  break;

	case 'X':
	case 'x':
	  if (radix == RADIX_UNKNOWN)
	    {
	      buffer.append(c);
	      radix = RADIX_16;
	      continue;
	    }
	  break;

	case 'E':
	case 'e':
	  if (exactness == EXACTNESS_UNKNOWN)
	    {
	      buffer.append(c);
	      exactness = EXACTNESS_EXACT;
	      continue;
	    }
	  break;

	case 'I':
	case 'i':
	  if (exactness == EXACTNESS_UNKNOWN)
	    {
	      buffer.append(c);
	      exactness = EXACTNESS_INEXACT;
	      continue;
	    }
	  break;
	}

      // error!
      input_port->unget_char(c);
      return EOF;
    }

  if (radix == RADIX_UNKNOWN)
    radix = RADIX_10;

  return c;
}


static char *
replace_sharp(const char *str)
{
  size_t len = strlen(str);
  char *result = static_cast<char *>(malloc(len + 1));
  if (result == NULL)
    {
      fprintf(stderr, "Out of Memory!\n");
      exit(EXIT_FAILURE);
    }

  const char *p = str;
  char *q = result;
  char c;
  do
    {
      c = *p++;
      if (c != '#')
	*q = c;
      else
	*q = '0';

      ++q;
    }
  while (c != '\0');

  return result;
}


static char *
replace_exponent_marker_in_place(char *str)
{
  char *p = str;
  while (*p != '\0')
    {
      if (is_exponent_marker(*p))
	*p = 'e';

      ++p;
    }
  return str;
}


// TODO: Should this be revised to construct Number with respect to the
// `exactness'.  We completely ignore it at this time.
long
Reader::read_real(Port *input_port, Radix radix, Exactness exactness,
		  Number &real)
{
  enum {
    INITIAL, SIGN0, INTEGER, RATIONAL0, RATIONAL,
    DECIMAL0, DECIMAL, DECIMAL_SUFFIX0, DECIMAL_SUFFIX1, DECIMAL_SUFFIX
  } state = INITIAL;

  DigitPredicate is_digit_r = digit_predicate_funcs[radix];
  DigitSscanf digit_sscanf = digit_sscanf_funcs[radix];
  size_t offset = buffer.get_output_strlen();

  int c = static_cast<int>(input_port->get_char());
  bool sharp = false;
  long numerator = 0;
  while (true)
    {
      switch (state)
	{
	case INITIAL:
	  if ((*is_digit_r)(c))
	    {
	      buffer.append(c);
	      state = INTEGER;
	      break;
	    }

	  if (c == '+' || c == '-')
	    {
	      buffer.append(c);
	      state = SIGN0;
	      break;
	    }

	  if (c == '.' && radix == RADIX_10)
	    {
	      buffer.append(c);
	      state = DECIMAL0;
	      break;
	    }

	  goto error;


	case SIGN0:
	  if ((*is_digit_r)(c))
	    {
	      buffer.append(c);
	      state = INTEGER;
	      break;
	    }

	  if (c == 'i' || c == 'I')
	    {
	      // Case of "+i" or "-i", we supply omitted `1' here.
	      if (buffer.get_output_string()[offset] == '+')
		real.set_ivalue(1);
	      else
		real.set_ivalue(-1);

	      return c;
	    }

	  goto error;


	case INTEGER:
	  if (!sharp && (*is_digit_r)(c))
	    {
	      buffer.append(c);
	      break;
	    }

	  if (c == '#')
	    {
	      buffer.append(c);
	      sharp = true;
	      break;
	    }

	  if (c == '.' && radix == RADIX_10)
	    {
	      buffer.append(c);
	      state = DECIMAL;
	      break;
	    }

	  if (c == '/')
	    {
	      if (!sharp)
		numerator = (*digit_sscanf)(buffer.get_output_string()
					    + offset);
	      else
		{
		  char *tmp_buf = replace_sharp(buffer.get_output_string()
						+ offset);
		  numerator = (*digit_sscanf)(tmp_buf);
		  free(tmp_buf);
		}

	      buffer.append(c);
	      state = RATIONAL0;
	      offset = buffer.get_output_strlen();
	      sharp = false;	// allow digits again
	      break;
	    }

	  if (is_real_terminater(c))
	    {
	      if (!sharp)
		real.set_ivalue((*digit_sscanf)(buffer.get_output_string()
						+ offset));
	      else
		{
		  char *tmp_buf = replace_sharp(buffer.get_output_string()
						+ offset);
		  real.set_ivalue((*digit_sscanf)(tmp_buf));
		  free(tmp_buf);
		}

	      return c;
	    }

	  goto error;


	case RATIONAL0:
	  if ((*is_digit_r)(c))
	    {
	      buffer.append(c);
	      state = RATIONAL;
	      break;
	    }

	  goto error;


	case RATIONAL:
	  if (!sharp && (*is_digit_r)(c))
	    {
	      buffer.append(c);
	      break;
	    }

	  if (c == '#')
	    {
	      buffer.append(c);
	      sharp = true;
	      break;
	    }

	  if (is_real_terminater(c))
	    {
	      long denominator;
	      if (!sharp)
		denominator = (*digit_sscanf)(buffer.get_output_string()
					      + offset);
	      else
		{
		  char *tmp_buf = replace_sharp(buffer.get_output_string()
						+ offset);
		  denominator = (*digit_sscanf)(tmp_buf);
		  free(tmp_buf);
		}

	      if (denominator != 0 && (numerator % denominator) == 0)
		{
		  real.set_ivalue(numerator / denominator);
		  return c;
		}

	      real.set_rvalue(static_cast<double>(numerator)
			      / static_cast<double>(denominator));
	      return c;
	    }

	  goto error;

	  
	case DECIMAL0:
	  if (isdigit(c))
	    {
	      buffer.append(c);
	      state = DECIMAL;
	      break;
	    }

	  goto error;


	case DECIMAL:
	  if (!sharp && isdigit(c))
	    {
	      buffer.append(c);
	      break;
	    }

	  if (c == '#')
	    {
	      buffer.append(c);
	      sharp = true;
	      break;
	    }

	  if (is_real_terminater(c))
	    {
	      double value;
	      if (!sharp)
		{
		  if (sscanf(buffer.get_output_string() + offset, "%lf",
			     &value) != 1)
		    value = 0.0;
		}
	      else
		{
		  char *tmp_buf = replace_sharp(buffer.get_output_string()
						+ offset);
		  if (sscanf(tmp_buf, "%lf", &value) != 1)
		    value = 0.0;

		  free(tmp_buf);
		}

	      real.set_rvalue(value);
	      return c;
	    }

	  if (is_exponent_marker(c))
	    {
	      buffer.append('e');
	      state = DECIMAL_SUFFIX0;
	      break;
	    }

	  goto error;


	case DECIMAL_SUFFIX0:
	  if (c == '+' || c == '-')
	    {
	      buffer.append(c);
	      state = DECIMAL_SUFFIX1;
	      break;
	    }

	  if (isdigit(c))
	    {
	      buffer.append(c);
	      state = DECIMAL_SUFFIX;
	      break;
	    }

	  goto error;


	case DECIMAL_SUFFIX1:
	  if (isdigit(c))
	    {
	      buffer.append(c);
	      state = DECIMAL_SUFFIX;
	      break;
	    }

	  goto error;


	case DECIMAL_SUFFIX:
	  if (isdigit(c))
	    {
	      buffer.append(c);
	      break;
	    }

	  if (is_real_terminater(c))
	    {
	      char *tmp_buf = replace_sharp(buffer.get_output_string()
					    + offset);
	      replace_exponent_marker_in_place(tmp_buf);

	      double value;
	      if (sscanf(tmp_buf, "%lf", &value) != 1)
		value = 0.0;
	      free(tmp_buf);
	      real.set_rvalue(value);
	      return c;
	    }

	  goto error;
	}

      c = input_port->get_char();
    }

 error:
  // input_port->unget_char(c);
  return 0;
}


Token
Reader::read_number(Port *input_port, Radix radix, Exactness exactness)
{
  long c;
  if ((c = check_prefix_r(input_port, radix, exactness)) == EOF)
    return TOKEN_INCOMPLETE_NUMBER;

  input_port->unget_char(c);

  c = read_real(input_port, radix, exactness, number);
  if (c != 0)
    {
      if (is_delimiter(c))
	{
	  input_port->unget_char(c);
	  return TOKEN_NUMBER;
	}

      if (c == 'i' || c == 'I')
	{
	  // number contains imaginary part of complex here.
	  number.set_rvalue(NAN);
	  buffer.append(c);
	  return TOKEN_UNSUPPORTED_NUMBER;
	}

      if (c == '@')
	{
	  buffer.append(c);
	  c = read_real(input_port, radix, exactness, number);
	  number.set_rvalue(NAN);
	  if (c != 0 && is_delimiter(c))
	    return TOKEN_UNSUPPORTED_NUMBER;
	}

      else if (c == '+' || c == '-')
	{
	  input_port->unget_char(c);
	  c = read_real(input_port, radix, exactness, number);
	  number.set_rvalue(NAN);
	  if (c != 0 && is_delimiter(c))
	    return TOKEN_UNSUPPORTED_NUMBER;
	}

      if (c != 0)
	input_port->unget_char(c);
    }

  return TOKEN_BAD_FORMED_EXPRESSION;
}


Token
Reader::read_token(Port *input_port)
{
  long c;
  // skip all white spaces at first.
  while (is_one_of(" \n\r\t", (c = input_port->get_char())))
    if (c == EOF)
      return TOKEN_EOF;

  switch (c)
    {
    case '(':
      return TOKEN_L_PAREN;


    case ')':
      return TOKEN_R_PAREN;


    case '\'':
      return TOKEN_QUOTE;


    case '`':
      return TOKEN_BACKQUOTE;


    case ',':
      c = input_port->get_char();
      if (c == '@')
	return TOKEN_COMMA_ATSIGN;

      input_port->unget_char(c);
      return TOKEN_COMMA;


    case '.':
      c = input_port->get_char();
      if (c != '.')
	{
	  input_port->unget_char(c);
	  return TOKEN_PERIOD;
	}

      c = input_port->get_char();
      if (c == '.')
	{
	  buffer.clear()->append("...", 3);
	  return TOKEN_IDENTIFIER;
	}

      input_port->unget_char('.');
      input_port->unget_char(c);
      return TOKEN_PERIOD;


    case '+':
    case '-':
      if (is_delimiter(input_port->peek_char()))
	{
	  buffer.clear()->append(c);
	  return TOKEN_IDENTIFIER;
	}

      input_port->unget_char(c);
      return read_number(input_port);


    case '"':
      buffer.clear();
      do
	{
	  c = input_port->get_char();
	  if (c != EOF && c != '"')
	    {
	      if (c != '\\')
		buffer.append(c);
	      else
		{
		  long c1 = input_port->get_char();
		  // Here we append one character follows '\' blindly.
		  // Should we treat escape sequences here?
		  switch (c1)
		    {
		    case 'n':
		      buffer.append('\n');
		      break;


		    case 't':
		      buffer.append('\t');
		      break;


		    case 'r':
		      buffer.append('\r');
		      break;


		    default:
		      if (c1 != EOF)
			buffer.append(c1);
		    }
		}
	    }

	  if (c == EOF)
	    return TOKEN_UNTERMINATED_STRING;
	}
      while (c != '"');

      return TOKEN_STRING;


    case ';':
      buffer.clear();
    one_line_comment:
      do
	{
	  buffer.append(c);
	  c = input_port->get_char();
	}
      while (!is_one_of("\n\r", c));

      return TOKEN_COMMENT;


    case '#':
      c = input_port->get_char();
      switch (c)
	{
	case '!':
	  buffer.clear()->append('#');
	  goto one_line_comment;


	case '(':
	  return TOKEN_L_VECTOR;


	case 'f':
	case 'F':
	  return TOKEN_BOOLEAN_FALSE;


	case 't':
	case 'T':
	  return TOKEN_BOOLEAN_TRUE;


	case '|':
	  {
	    buffer.clear()->append("#|", 2);
	    long n = 1;
	    c = input_port->get_char();
	    while (true)
	      {
		buffer.append(c);
		long c1 = input_port->get_char();
		if (c == '|' && c1 == '#')
		  {
		    buffer.append(c1);
		    if (--n == 0)
		      break;

		    c = input_port->get_char();
		  }
		else if (c == '#' && c1 == '|')
		  {
		    buffer.append(c1);
		    ++n;
		    c = input_port->get_char();
		  }
		else if (c == EOF)
		  return TOKEN_UNTERMINATED_BLOCK_COMMENT;
		else
		  c = c1;
	      }

	    return TOKEN_COMMENT;
	  }


	case '/':
	  {
	    buffer.clear();
	    do
	      {
		c = input_port->get_char();
		buffer.append(c);
		if (c != '\\')
		  {
		    if (c == EOF)
		      return TOKEN_UNTERMINATED_REGEXP;
		  }
		else
		  {
		    long c1 = input_port->get_char();
		    if (c1 == EOF)
		      return TOKEN_UNTERMINATED_REGEXP;

		    buffer.append(c1);
		  }
	      }
	    while (c != '/');

	    while (true)
	      {
		/*
		 * We have at least one option for each literal regular
		 * expression, e.g. #/regexp/i
		 * The optional `i' specifies case insensitiveness for
		 * regular expression.  The number of options may be
		 * extended in the future, therefore, here we read
		 * characters beyond the regular expression delimiter
		 * '/' to the Scheme delimiter such as spaces and
		 * parentheses.
		 */
		c = input_port->get_char();
		if (is_delimiter(c))
		  {
		    input_port->unget_char(c);
		    break;
		  }

		buffer.append(c);
	      }

	    return TOKEN_REGEXP;
	  }


	case '\\':
	  buffer.clear();
	  while (!is_delimiter(c = input_port->get_char()))
	    buffer.append(c);

	  if (buffer.get_output_strlen() == 0)
	    buffer.append(c);		// the delimiter is the character.
	  else
	    input_port->unget_char(c);	// push back the delimiter.

	  if (buffer.get_output_strlen() > 0)
	    return TOKEN_CHARACTER;

	  return TOKEN_INCOMPLETE_CHARACTER;


	case 'B':
	case 'b':
	  return read_number(input_port, RADIX_2);


	case 'D':
	case 'd':
	  return read_number(input_port, RADIX_10);


	case 'O':
	case 'o':
	  return read_number(input_port, RADIX_8);


	case 'X':
	case 'x':
	  return read_number(input_port, RADIX_16);


	case 'E':
	case 'e':
	  return read_number(input_port, RADIX_UNKNOWN, EXACTNESS_EXACT);


	case 'I':
	case 'i':
	  return read_number(input_port, RADIX_UNKNOWN, EXACTNESS_INEXACT);


	default:
	  // unknown sharp constant
	  buffer.clear();
	  while (!is_delimiter(c))
	    {
	      buffer.append(c);
	      c = input_port->get_char();
	    }

	  input_port->unget_char(c);	// push back the delimiter
	  return TOKEN_UNKNOWN_SHARP;
	}


    default:	// may be identifier or number
      if (isdigit(c))
	{
	  input_port->unget_char(c);
	  return read_number(input_port);
	}

#if 0	// relaxed identifier reader
      buffer.clear()->append(c);
      while (true)
	{
	  c = input_port->get_char();
	  
	  if (!is_delimiter(c))
	    buffer.append(c);
	  else
	    {
	      input_port->unget_char(c);
	      break;
	    }
	}
      return TOKEN_IDENTIFIER;
#else	// identifier reader strictly conform to R5RS spec.
      buffer.clear()->append(c);
      if (is_initial(c))
	{
	  while (true)
	    {
	      c = input_port->get_char();

	      if (!is_delimiter(c) && is_subsequent(c))
		{
		  buffer.append(c);
		  continue;
		}
	      else if (is_delimiter(c))
		{
		  input_port->unget_char(c);
		  break;
		}

	      return TOKEN_INVALID_IDENTIFIER;
	    }

	  return TOKEN_IDENTIFIER;
	}

      return TOKEN_INVALID_IDENTIFIER;
#endif
    }
}


Cell *
Reader::parse_character_spec()
{
  // According to R5RS, there's only two charnames, newline and space.
  // We allow more than that, especially, these are almost complete
  // clone of gauche's specification.  The only exception is we don't
  // support #\uXXXX encoding and multibyte character as a literal
  // character because we don't support wide character at all.
  size_t len = buffer.get_output_strlen();
  const char *spec = buffer.get_output_string();
  switch (len)
    {
    case 1:
      return vm.mk_character(reinterpret_cast<const unsigned char *>(spec)[0]);


    case 2:
      if (strcasecmp(spec, "lf") == 0 || strcasecmp(spec, "nl") == 0)
	return vm.mk_character('\n');

      if (strcasecmp(spec, "cr") == 0)
	return vm.mk_character('\r');

      if (strcasecmp(spec, "ht") == 0)
	return vm.mk_character('\t');

      break;


    case 3:
      if (strcasecmp(spec, "tab") == 0)
	return vm.mk_character('\t');

      if (strcasecmp(spec, "esc") == 0)
	return vm.mk_character(0x1b);

      if (strcasecmp(spec, "del") == 0)
	return vm.mk_character(0x7f);

      break;


    case 4:
      if (strcasecmp(spec, "null") == 0)
	return vm.mk_character(0);

      if (strcasecmp(spec, "page") == 0)
	return vm.mk_character(0x0c);

      break;


    case 5:
      if (strcasecmp(spec, "space") == 0)
	return vm.mk_character(' ');

      break;


    case 6:
      if (strcasecmp(spec, "return") == 0)
	return vm.mk_character('\r');

      if (strcasecmp(spec, "escape") == 0)
	return vm.mk_character(0x1b);

      if (strcasecmp(spec, "delete") == 0)
	return vm.mk_character(0x7f);

      break;


    case 7:
      if (strcasecmp(spec, "newline") == 0)
	return vm.mk_character('\n');

      break;


    default:
      break;
    }

  if (*spec == 'x' || *spec == 'X')
    {
      const char *p = spec + 1;
      while (isxdigit(*p++))
	if (*p == '\0')
	  {
	    long c;
	    if (sscanf(spec + 1, "%lx", &c) != 1)
	      c = 0;

	    return vm.mk_character(c & 0xff);
	  }
    }

  error_message.clear();
  error_message.append("Unsupported character literal: #\\");
  error_message.append(buffer);
  return Cell::nil();
}


Cell *
Reader::parse_regexp_spec()
{
  OutputStringPort regexp;
  // At first, regular expression body will be determined and cleaned up.
  const char *p = buffer.get_output_string();
  while (*p != '/')	// We know there's terminater '/' within the buffer.
    {
      if (*p != '\\')
	regexp.append(*p++);
      else
	{
	  // We know there's at least one non-NUL char exist after `\' ensured.
	  if (*++p != '/')
	    regexp.append('\\')->append(*p++);
	  else
	    regexp.append(*p++);	// remove leading \ that escapes `/'.
	}
    }

  if (p[1] == '\0' || (p[1] == 'i' && p[2] == '\0'))
    {
      Cell *rx_str = vm.mk_counted_string(regexp.get_output_string(),
					  regexp.get_output_strlen());
      Cell *rx = vm.mk_regexp(rx_str, p[1] != '\0');

      if (rx->is_regexp())
	return rx;

      error_message.clear()
	->append("Regular expression compilation error: ")
	->append(rx->get_string(), rx->get_strlen())->append(": #/");
    }
  else
    {
      error_message.clear()
	->append("Invalid regular expression literal: #/");
    }

  error_message.append(buffer);
  return Cell::nil();	// invalid literal regular expression option.
}


static const char *const error_message_string[] = {
  "Invalid sharp constant: #",
  "Unterminated string: \"",
  "Unterminated block comment: ",
  "Unterminated regexp: #/",
  "Incomplete number: ",
  "Incomplete character: #\\",
  "Unsupported number: ",
  "Bad formed expression: ",
  "Invalid Identifier: "
};


Cell *
Reader::read_datum(Port *input_port)
{
  Cell *nil = Cell::nil();
  Cell *datum = nil;
  do
    {
      Token token;
      while ((token = read_token(input_port)) == TOKEN_COMMENT)
	; // skip all coments here

#if 0
      fprintf(stderr, "read_datum: token=%d, \"%s\"\n",
	      token, buffer.get_output_string());
#endif

      if (current_context.state == READER_STATE_EXPECT_R_PAREN
	  && token != TOKEN_R_PAREN)
	{
	  error_message.clear()
	    ->append("Unexpected datum after . <datum>: `)' expected.");
	  goto error;
	}

      switch (token)
	{
	// <simple datum>
	case TOKEN_BOOLEAN_TRUE:
	  datum = Cell::t();
	  break;


	case TOKEN_BOOLEAN_FALSE:
	  datum = Cell::f();
	  break;


	case TOKEN_NUMBER:
	  datum = vm.mk_number(number);
	  break;


	case TOKEN_CHARACTER:
	  datum = parse_character_spec();
	  if (datum != nil)
	    break;

	  goto error;


	case TOKEN_STRING:
	  datum = vm.mk_counted_string(buffer.get_output_string(),
				       buffer.get_output_strlen())
	    ->set_immutable();
	  break;


	case TOKEN_IDENTIFIER:
	  datum = vm.core.mk_symbol(buffer.get_output_string());
	  break;


	case TOKEN_REGEXP:
	  datum = parse_regexp_spec();
	  if (datum != nil)
	    break;

	  goto error;


	// <compound datum>
	case TOKEN_L_PAREN:
	  reader_stack.push_back(current_context);
	  current_context.state = READER_STATE_LIST;
	  current_context.datum = nil;
	  continue;


	case TOKEN_L_VECTOR:
	  reader_stack.push_back(current_context);
	  current_context.state = READER_STATE_VECTOR;
	  current_context.datum = nil;
	  continue;


	case TOKEN_QUOTE:
	  reader_stack.push_back(current_context);
	  current_context.state = READER_STATE_ABBREV_QUOTE;
	  current_context.datum = nil;
	  continue;


	case TOKEN_BACKQUOTE:
	  reader_stack.push_back(current_context);
	  current_context.state = READER_STATE_ABBREV_QUASIQUOTE;
	  current_context.datum = nil;
	  continue;


	case TOKEN_COMMA:
	  reader_stack.push_back(current_context);
	  current_context.state = READER_STATE_ABBREV_UNQUOTE;
	  current_context.datum = nil;
	  continue;


	case TOKEN_COMMA_ATSIGN:
	  reader_stack.push_back(current_context);
	  current_context.state = READER_STATE_ABBREV_UNQUOTE_SPLICING;
	  current_context.datum = nil;
	  continue;


	case TOKEN_PERIOD:
	  if (current_context.datum->is_pair())
	    {
	      current_context.state = READER_STATE_IMPROPER_LIST;
	      continue;
	    }

	  error_message.clear()
	    ->append("Unexpected .: . must follow non-null list");
	  goto error;


	// complete a <compound datum> in the current nesting level
	case TOKEN_R_PAREN:
	  if (!reader_stack.empty())
	    {
	      switch (current_context.state)
		{
		case READER_STATE_LIST:
		  datum = R5RSListProcedure::list_reverse_in_place(
							vm.memory_system,
							nil,
							current_context.datum,
							true);
		  break;


		case READER_STATE_VECTOR:
		  {
		    long len = R5RSListProcedure::list_length(current_context.datum);
		    Cell *vect = vm.mk_vector(len, nil);
		    datum = current_context.datum;
		    while (datum != nil)
		      {
			vm.memory_system.set_vector_element(vect, --len,
							    CAR(datum));
			datum = CDR(datum);
		      }

		    datum = vect->set_immutable();
		  }
		  break;


		case READER_STATE_EXPECT_R_PAREN:
		  datum = current_context.datum;
		  break;


		case READER_STATE_INITIAL:
		case READER_STATE_IMPROPER_LIST:
		case READER_STATE_ABBREV_QUOTE:
		case READER_STATE_ABBREV_QUASIQUOTE:
		case READER_STATE_ABBREV_UNQUOTE:
		case READER_STATE_ABBREV_UNQUOTE_SPLICING:
		  goto unexpected_r_paren;
		}

	      current_context = reader_stack.back();
	      reader_stack.pop_back();
	      break;
	    }


	unexpected_r_paren:
	  error_message.clear()
	    ->append("Unexpected ): unmatched parentheses");
	  goto error;


	case TOKEN_COMMENT:
	  datum = nil;
	  break;	// Never reached.


	case TOKEN_UNKNOWN_SHARP:
	case TOKEN_UNTERMINATED_STRING:
	case TOKEN_UNTERMINATED_BLOCK_COMMENT:
	case TOKEN_UNTERMINATED_REGEXP:
	case TOKEN_INCOMPLETE_NUMBER:
	case TOKEN_INCOMPLETE_CHARACTER:
	case TOKEN_UNSUPPORTED_NUMBER:
	case TOKEN_BAD_FORMED_EXPRESSION:
	case TOKEN_INVALID_IDENTIFIER:
	  error_message.clear()
	    ->append(error_message_string[token - TOKEN_UNKNOWN_SHARP]);
	  error_message.append(buffer);
	  goto error;


	case TOKEN_UNEXPECTED_EOF:
	  error_message.clear()->append("Unexpected EOF");
	  goto error;


	case TOKEN_EOF:
	  if (reader_stack.empty())
	    {
	      datum = Cell::eof();
	      break;
	    }

	  error_message.clear()
	    ->append("Unexpected EOF: unmatched parentheses");
	  goto error;
	}


      while (true)
	{
	  switch (current_context.state)
	    {
	    case READER_STATE_LIST:
	    case READER_STATE_VECTOR:
	      current_context.datum = vm.cons(datum, current_context.datum);
	      break;


	    case READER_STATE_IMPROPER_LIST:
	      {
		Cell *tail = current_context.datum;
		current_context.datum
		  = R5RSListProcedure::list_reverse_in_place(vm.memory_system,
							     nil, tail, true);
		vm.set_cdr(tail, datum);
		current_context.state = READER_STATE_EXPECT_R_PAREN;
		break;
	      }


	    case READER_STATE_ABBREV_QUOTE:
	      datum = vm.cons(datum, nil);
	      datum = vm.cons(vm.core.quote_symbol, datum);
	      current_context = reader_stack.back();
	      reader_stack.pop_back();
	      continue;


	    case READER_STATE_ABBREV_QUASIQUOTE:
	      datum = vm.cons(datum, nil);
	      datum = vm.cons(vm.core.quasiquote_symbol, datum);
	      current_context = reader_stack.back();
	      reader_stack.pop_back();
	      continue;


	    case READER_STATE_ABBREV_UNQUOTE:
	      datum = vm.cons(datum, nil);
	      datum = vm.cons(vm.core.unquote_symbol, datum);
	      current_context = reader_stack.back();
	      reader_stack.pop_back();
	      continue;


	    case READER_STATE_ABBREV_UNQUOTE_SPLICING:
	      datum = vm.cons(datum, nil);
	      datum = vm.cons(vm.core.unquote_splicing_symbol, datum);
	      current_context = reader_stack.back();
	      reader_stack.pop_back();
	      continue;


	    default:
	      current_context.datum = datum;
	      break;
	    }

	  break;
	}
    }
  while (!reader_stack.empty());

  datum = current_context.datum;
  current_context.datum = nil;
  return datum;


 error:
  // Here we should have a chance to construct more friendly error messages.
  while (!reader_stack.empty())
    {
      current_context = reader_stack.back();
      reader_stack.pop_back();
    }

  current_context.datum = nil;
  return NULL;
}


Reader::Reader(VirtualMachine &vm) : vm(vm)
{
  current_context.state = READER_STATE_INITIAL;
  current_context.datum = Cell::nil();
  vm.memory_system.register_static_root(&current_context.datum);
  vm.memory_system.register_root_enumerator(enumerate_dynamic_roots, this);
}


Reader::~Reader()
{
  reader_stack.erase(reader_stack.begin(), reader_stack.end());
  current_context.datum = Cell::nil();
  vm.memory_system.unregister_static_root(&current_context.datum);
  vm.memory_system.unregister_root_enumerator(enumerate_dynamic_roots, this);
}


void
Reader::enumerate_dynamic_roots(vector<Cell **> &dynamic_roots, void *data)
{
  Reader *reader = static_cast<Reader *>(data);
  reader->enumerate_dynamic_roots(dynamic_roots);
}


void
Reader::enumerate_dynamic_roots(vector<Cell **> &dynamic_roots)
{
  // mark Cells reachable from reader_stack
  vector<ReaderContext>::iterator iter = reader_stack.begin();
  vector<ReaderContext>::iterator end_iter = reader_stack.end();
  while (iter != end_iter)
    {
      ReaderContext &context = *iter++;
      dynamic_roots.push_back(&context.datum);
    }
}

}
