/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: reader.h,v 1.16 2009/01/01 07:47:01 fuyu Exp $
 */

#ifndef _LIBTSCORE_READER_H_
#define _LIBTSCORE_READER_H_

#ifdef __cplusplus

#if defined(__GNUG__) && !defined(__ICC)
#pragma interface
#endif

#include "libtscore/cell.h"
#include "libtscore/number.h"
#include "libtscore/port.h"

#include <vector>


using namespace std;
namespace LibTSCore
{

  enum Radix {
    RADIX_2 = 0,
    RADIX_8,
    RADIX_10,
    RADIX_16,
    RADIX_UNKNOWN
  };


  enum Exactness {
    EXACTNESS_UNKNOWN,
    EXACTNESS_EXACT,
    EXACTNESS_INEXACT
  };


  enum Token {
    TOKEN_EOF,
    TOKEN_BOOLEAN_TRUE,
    TOKEN_BOOLEAN_FALSE,
    TOKEN_NUMBER,
    TOKEN_CHARACTER,
    TOKEN_STRING,
    TOKEN_IDENTIFIER,
    TOKEN_L_PAREN,
    TOKEN_R_PAREN,
    TOKEN_L_VECTOR,
    TOKEN_REGEXP,
    TOKEN_QUOTE,
    TOKEN_BACKQUOTE,
    TOKEN_COMMA,
    TOKEN_COMMA_ATSIGN,
    TOKEN_PERIOD,
    TOKEN_COMMENT,
    TOKEN_UNKNOWN_SHARP,
    TOKEN_UNTERMINATED_STRING,
    TOKEN_UNTERMINATED_BLOCK_COMMENT,
    TOKEN_UNTERMINATED_REGEXP,
    TOKEN_INCOMPLETE_NUMBER,
    TOKEN_INCOMPLETE_CHARACTER,
    TOKEN_UNSUPPORTED_NUMBER,
    TOKEN_BAD_FORMED_EXPRESSION,
    TOKEN_INVALID_IDENTIFIER,
    TOKEN_UNEXPECTED_EOF
  };


  enum ReaderState {
    READER_STATE_INITIAL,
    READER_STATE_LIST,
    READER_STATE_IMPROPER_LIST,
    READER_STATE_EXPECT_R_PAREN,
    READER_STATE_VECTOR,
    READER_STATE_ABBREV_QUOTE,
    READER_STATE_ABBREV_QUASIQUOTE,
    READER_STATE_ABBREV_UNQUOTE,
    READER_STATE_ABBREV_UNQUOTE_SPLICING
  };


  struct ReaderContext
  {
    ReaderState state;
    Cell *datum;
  };


  /*
   * Reader is a S expression reader that read character stream from
   * Ports and construct Scheme level representation of S expression.
   *
   * This is just a building block of TSCore not for users.
   */
  class Reader
  {
    friend class VirtualMachine;

  private:
    VirtualMachine &vm;
    OutputStringPort buffer;
    OutputStringPort error_message;
    Number number;

    ReaderContext current_context;
    vector<ReaderContext> reader_stack;


    static void enumerate_dynamic_roots(vector<Cell **> &dynamic_roots,
					void *data);
    void enumerate_dynamic_roots(vector<Cell **> &dynamic_roots);


    /*
     * Reads one <real R> from input_port and initialize the `real'
     * with it.  This returns the successor character of <real R> such as
     * delimiter, leading <sign> or @ of the complex imaginary or polar part
     * and `i' terminates the current imaginary part of a complex number.
     *
     * This may break the contents of `buffer' field!
     */
    long read_real(Port *input_port, Radix radix, Exactness exactness,
		   Number &real);


    /*
     * Initializes radix and exactness by reading from input_port with
     * assumption that input_port gives a number.
     */
    long check_prefix_r(Port *input_port, Radix &radix, Exactness &exactness);


    /*
     * Parses one external representation of number read from given
     * input port.  This parses one number and initialize the `number'
     * field if the parsing succeeded and return TOKEN_NUMBER, otherwise,
     * returns TOKEN_EOF (another case?).
     *
     * NOTE: We don't have numbers other than exact integer, long in C,
     * or inexact real, double in C.  Therefore, this reader will accept
     * other representation in R5RS as possible as we can but those value
     * may be sometime rounded or wrongly converted into Number we
     * currently have, e.g. exact rational will be converted to inexact
     * real, complex will be converted into inexact real by dropping
     * imaginary part!
     *
     * Although supporting more types of Number defined in R5RS is also
     * planned, to do that has almost lowest priority for us because
     * we don't need them at all at this time.
     */
    Token read_number(Port *input_port,
		      Radix radix = RADIX_UNKNOWN,
		      Exactness exactness = EXACTNESS_UNKNOWN);


    /*
     * Returns the Token read from given input_port.  For non-trivial token
     * that have non-constant external representation, the external
     * representation is stored in the `buffer'.
     */
    Token read_token(Port *input_port);


    /*
     * Returns the character (Cell *) that corresponds to the character
     * specified by string contained within the `buffer'.
     *
     * When buffer contains invalid specification, `error_message' will be
     * updated and then nil will be returned.
     */
    Cell *parse_character_spec();


    /*
     * Returns the regular expression (Cell *) that represents the
     * regular expression contained within the `buffer'.
     *
     * When buffer contains invalid specification, `error_message' will be
     * updated and then nil will be returned
     */
    Cell *parse_regexp_spec();


  public:
    Reader(VirtualMachine &vm);
    ~Reader();


    /*
     * Returns an S expression if it can be completely read from given
     * input_port.
     * Returns EOF object if input_port reaches EOF.
     * Returns (native) NULL pointer if reader met unexpected EOF or
     * not well formed expression.
     */
    Cell *read_datum(Port *input_port);


    /*
     * Returns a recent error message.
     */
    inline const char *get_error_message() const
    {
      return error_message.get_output_string();
    }

  };


}


#endif	/* __cplusplus */

#endif	/* _LIBTSCORE_READER_H_ */
