/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: srfi28_procedure.cc,v 1.4 2009/01/01 07:47:01 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libtscore/srfi28_procedure.h"

#include "libtscore/tscore.h"
#include "libtscore/port.h"
#include "libtscore/writer.h"


namespace LibTSCore
{

NativeProcedure::Info SRFI28Procedure::procedures[] = {
  { &SRFI28Procedure::format,
    "format", TST_STRING TST_ANY, 1, INF_ARG }
};


Cell *
SRFI28Procedure::assign_procedures(Cell *environment, Cell *frame)
{
  TSStackRoot env(core, environment);
  TSStackRoot frm(core, frame);
  size_t n_procs = sizeof(procedures) / sizeof(Info);
  for (size_t i = 0; i < n_procs; i++)
    assign_procedure(env, frm, &procedures[i], this);

  return env;
}


Cell *
SRFI28Procedure::format(VirtualMachine &vm, Register &context,
			unsigned long argument_pointer,
			unsigned long nargs, void *data)
{
  // ~a The corresponding value is inserted into the string as if printed
  //    with display.
  // ~s The corresponding value is inserted into the string as if printed
  //    with write.
  // ~% A newline is inserted.
  // ~~ A tilde '~' is inserted.
  //
  // Other escape sequences throw error.

  Cell *arg0 = context.frame->load_variable(argument_pointer);
  const char *cur_pos = arg0->get_string();
  const char *tail_pos = cur_pos + arg0->get_strlen();

  OutputStringPort buf;
  unsigned long ap = argument_pointer + 1;
  unsigned long max_ap = argument_pointer + nargs;
  while (cur_pos < tail_pos)
    {
      int c = *cur_pos++;
      if (c != '~')
	buf.append(c);
      else
	{
	  if (cur_pos < tail_pos)
	    {
	      c = *cur_pos++;
	      if (c == 'a')
		{
		  if (ap < max_ap)
		    Writer::write_datum(context.frame->load_variable(ap++),
					&buf, false, false);
		  else
		    return signal_error(vm,
					"format: No value for escape sequence ~a: ",
					arg0);
		}
	      else if (c == 's')
		{
		  if (ap < max_ap)
		    Writer::write_datum(context.frame->load_variable(ap++),
					&buf, true, false);
		  else
		    return signal_error(vm,
					"format: No value for escape sequence ~s: ",
					arg0);
		}
	      else if (c == '%')
		buf.append('\n');
	      else if (c == '~')
		buf.append('~');
	      else
		{
		  char err_msg[128];
		  snprintf(err_msg, 128,
			   "format: Unrecognized escape sequence: ~%c: ", c);
		  return signal_error(vm, err_msg, arg0);
		}
	    }
	  else
	    return signal_error(vm, "format: Incomplete escape sequence:",
				arg0);
	}
    }

  return vm.mk_counted_string(buf.get_output_string(), buf.get_output_strlen());
}

}
