/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: symbol_table.cc,v 1.15 2009/01/01 07:47:01 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libtscore/symbol_table.h"

#include "libtscore/cell.h"
#include "libtscore/tscore.h"

#include <cassert>
#include <cstdio>
#include <cstring>


namespace LibTSCore
{

SymbolTable::SymbolTable(TSCore &tscore)
  : core(tscore),
    table(Cell::nil())
{
  tscore.memory_system.register_static_root(&table);
  table = tscore.memory_system.get_vector(461);
}


SymbolTable::~SymbolTable()
{
  table = Cell::nil();
  core.memory_system.unregister_static_root(&table);
}


void
SymbolTable::purge_all()
{
  table = Cell::nil();
}


Cell *
SymbolTable::lookup(const char *name, size_t name_len) const
{
  long location
    = static_cast<unsigned long>(Cell::symbol_hash(name, name_len))
    % static_cast<unsigned long>(table->get_vector_length());
  for (Cell *x = table->get_vector_element(location);
       x != Cell::nil(); x = CDR(x))
    {
      Cell *sym_string = x->get_symbol_string();
      const char *sym_cstr = sym_string->get_string();
      size_t sym_len = sym_string->get_strlen();
      if (name_len == sym_len && strncasecmp(name, sym_cstr, name_len) == 0)
	return x;
    }

  return Cell::nil();
}


Cell *
SymbolTable::insert0(Cell *&name)
{
  assert(name->is_string());
  assert(name->is_immutable());

  Cell *symbol = core.memory_system.get_cell(name);
  if (core.memory_system.out_of_memory())
    return Cell::nil();

  long location
    = static_cast<unsigned long>(name->string_hash())
    % static_cast<unsigned long>(table->get_vector_length());
  symbol->mk_symbol(name, table->get_vector_element(location));
  core.memory_system.set_vector_element(table, location, symbol);
  return symbol;
}


Cell *
SymbolTable::insert(Cell *&name)
{
  assert(name->is_string());

  Cell *x = lookup(name);
  if (x != Cell::nil())
    return x;

  const char *p = name->get_string();
  size_t len = name->get_strlen();
  const char *p_tail = p + len;
  while (p < p_tail)
    if (*p != tolower(*p))
      break;
    else
      ++p;

  if (p == p_tail && name->is_immutable())
    return insert0(name);	// we can use `name' as symbol's name.

  char *canonical_name = static_cast<char *>(malloc(len + 1));
  if (canonical_name == NULL)
    {
      fprintf(stderr, "Out of Memory!\n");

      return Cell::nil();
    }

  p = name->get_string();
  char *q = canonical_name;
  while (p < p_tail)
    *q++ = tolower(*p++);

  *q = '\0';

  Cell *symname = core.memory_system.get_cell();
  if (core.memory_system.out_of_memory())
    {
      fprintf(stderr, "Out of Memory!\n");

      free(canonical_name);
      return Cell::nil();
    }

  symname->mk_string(canonical_name, len)->set_immutable();
  return insert0(symname);
}


Cell *
SymbolTable::insert(const char *name)
{
  Cell *x = lookup(name);
  if (x != Cell::nil())
    return x;

  size_t len = strlen(name);
  char *canonical_name = static_cast<char *>(malloc(len + 1));
  if (canonical_name == NULL)
    {
      fprintf(stderr, "Out of Memory!\n");

      return Cell::nil();
    }

  const char *p = name;
  const char *p_tail = p + len;
  char *q = canonical_name;
  while (p < p_tail)
    *q++ = tolower(*p++);

  *q = '\0';

  Cell *symname = core.memory_system.get_cell();
  if (core.memory_system.out_of_memory())
    {
      fprintf(stderr, "Out of Memory!\n");

      free(canonical_name);
      return Cell::nil();
    }

  symname->mk_string(canonical_name, strlen(canonical_name))->set_immutable();
  return insert0(symname);
}


Cell *
SymbolTable::list_all_symbols() const
{
  Cell *list = Cell::nil();
  for (size_t i = 0, len = table->get_vector_length(); i < len; i++)
    {
      for (Cell *x = table->get_vector_element(i);
	   x != Cell::nil(); x = CDR(x))
	{
	  Cell *symname = CAR(x);
	  list = core.memory_system.get_cons(symname, list);
	}
    }

  return list;
}

}
