/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: syntax.cc,v 1.18 2010/08/15 15:39:47 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libtscore/syntax.h"

#include "libtscore/bytecode_buffer.h"
#include "libtscore/virtual_machine.h"


#include <cassert>
#include <vector>


namespace LibTSCore
{


inline bool
Syntax::is_toplevel_environment(VirtualMachine &vm, const Cell *environment)
{
  assert(environment != Cell::nil());

  if (environment->get_parent_environment() != Cell::nil())
    return false;

  assert(environment == vm.core.interaction_environment
	 || environment == vm.core.r5rs_environment
	 || environment == vm.core.null_environment);

  return true;
}


bool
Syntax::is_quasiquote_syntax()
{
  return false;
}


size_t
Syntax::emit_list_qq_template(VirtualMachine &vm, Register &context,
			      BytecodeBuffer *buffer,
			      Cell *qq_template, long nesting,
			      Cell *environment, Cell *constant_pool,
			      Cell *hint, bool list_qq)
{
  if (vm.get_bytecompile_tracing(context))
    {
      Port *port = context.output_port->get_port();
      char buf[256];
      snprintf(buf, sizeof(buf),
	       "Syntax::emit_list_qq_template<%ld>: ", nesting);
      port->append(buf);
      Writer::write_datum(qq_template, port, false);
      port->append('\n');
    }

  VMStackRoot form(vm, qq_template);
  VMStackRoot env(vm, environment);
  VMStackRoot cpool(vm, constant_pool);
  VMStackRoot hnt(vm, hint);

  vector<VMStackRoot *> element_stack;
  Cell *list = form;
  while (list->is_pair())
    {
      element_stack.push_back(new VMStackRoot(vm, CAR(list)));
      list = CDR(list);
    }
  assert(!element_stack.empty());

  while (!element_stack.empty())
    {
      VMStackRoot *root = element_stack.back();
      Cell *prev = *root;
      if (!Quasiquote::find_unquote(vm, context, environment, prev, nesting))
	list = vm.mk_cons(prev, list, true);
      else
	break;

      element_stack.pop_back();
      delete root;
    }
  assert(!element_stack.empty());

  size_t len = emit_literal_expression(vm, context, buffer, list, cpool);

  long fp = env->allocate_stack_slot();	// for current continuation
  long nv = env->allocate_stack_slot();	// for new value
  long pv = env->allocate_stack_slot();	// for previous value

  while (!element_stack.empty())
    {
      len += emit_local_variable_assignment(buffer, pv);
      VMStackRoot *root = element_stack.back();
      Cell *next = *root;
      element_stack.pop_back();
      delete root;
      assert(Quasiquote::find_unquote(vm, context, env, next, nesting));

#if 0
      if (vm.get_bytecompile_tracing(context))
	{
	  Port *port = context.output_port->get_port();
	  char buf[256];
	  snprintf(buf, sizeof(buf), "* fp=%ld: subform: ", fp);
	  port->append(buf);
	  Writer::write_datum(next, port, false);
	  port->append('\n');
	}
#endif

      Syntax *syntax = lookup_syntax(next, env);
      size_t qq_len;
      if (syntax != NULL)
	qq_len = syntax->emit_qq_template(vm, context, buffer, next, nesting,
					  env, cpool, hnt, true);
      else
	qq_len = Syntax::emit_qq_template(vm, context, buffer, next, nesting,
					  env, cpool, hnt, true);

      if (qq_len == 0)
	return 0;	// error

      next = Cell::nil();
      while (!element_stack.empty())
	{
	  root = element_stack.back();
	  Cell *prev = *root;
	  if (Quasiquote::find_unquote(vm, context, env, prev, nesting))
	    break;

	  element_stack.pop_back();
	  delete root;
	  next = vm.mk_cons(prev, next, true);
	}

      if (next != Cell::nil())
	{
	  len += emit_local_variable_assignment(buffer, pv);
	  Cell *proc;
	  if (CDR(next) == Cell::nil())
	    {
	      len += emit_literal_expression(vm, context, buffer,
					     CAR(next), cpool);
	      proc = vm.core.internal_cons_procedure;
	    }
	  else
	    {
	      len += emit_literal_expression(vm, context, buffer, next, cpool);
	      proc = vm.core.internal_append_procedure;
	    }
#if 0
	  if (vm.get_bytecompile_tracing(context))
	    {
	      Port *port = context.output_port->get_port();
	      char buf[256];
	      snprintf(buf, sizeof(buf), "* fp=%ld: sub literal form: ", fp);
	      port->append(buf);
	      Writer::write_datum(next, port, false);
	      port->append('\n');
	    }
#endif

	  len += emit_local_variable_assignment(buffer, nv);
	  len += emit_known_procedure_call(vm, context, buffer,
					   proc, env, cpool,
					   fp, 3);
	}
    }

  env->dispose_stack_slots(fp);

  if (list_qq)
    {
      unsigned long fp = env->get_current_stack_pointer() - 3;
      assert(fp > 0);

      len += emit_local_variable_assignment(buffer, fp + 1);
      len += emit_known_procedure_call(vm, context, buffer,
				       vm.core.internal_cons_procedure,
				       env, cpool, fp, 3);
    }

  return len;
}


size_t
Syntax::emit_vector_qq_template(VirtualMachine &vm, Register &context,
				BytecodeBuffer *buffer,
				Cell *qq_template, long nesting,
				Cell *environment, Cell *constant_pool,
				Cell *hint, bool list_qq)
{
  assert(qq_template->is_vector());

  if (vm.get_bytecompile_tracing(context))
    {
      Port *port = context.output_port->get_port();
      char buf[256];
      snprintf(buf, sizeof(buf),
	       "Syntax::emit_vector_qq_template<%ld>: ", nesting);
      port->append(buf);
      Writer::write_datum(qq_template, port, false);
      port->append('\n');
    }

  VMStackRoot form(vm, qq_template);
  VMStackRoot env(vm, environment);
  VMStackRoot cpool(vm, constant_pool);
  VMStackRoot hnt(vm, hint);

  vector<VMStackRoot *> element_stack;
  long vector_len = form->get_vector_length();
  for (long i = 0; i < vector_len; i++)
    element_stack.push_back(new VMStackRoot(vm,
					    form->get_vector_element(i)));
  assert(!element_stack.empty());

  Cell *list = Cell::nil();
  while (!element_stack.empty())
    {
      VMStackRoot *root = element_stack.back();
      Cell *prev = *root;
      if (!Quasiquote::find_unquote(vm, context, environment, prev, nesting))
	list = vm.mk_cons(prev, list, true);
      else
	break;

      element_stack.pop_back();
      delete root;
    }
  assert(!element_stack.empty());

  size_t len = emit_literal_expression(vm, context, buffer, list, cpool);

  long fp = env->allocate_stack_slot();	// for current continuation
  long nv = env->allocate_stack_slot();	// for new value
  long pv = env->allocate_stack_slot();	// for previous value

  while (!element_stack.empty())
    {
      len += emit_local_variable_assignment(buffer, pv);
      VMStackRoot *root = element_stack.back();
      Cell *next = *root;
      element_stack.pop_back();
      delete root;

      assert(Quasiquote::find_unquote(vm, context, env, next, nesting));

      Syntax *syntax = lookup_syntax(next, env);
      size_t qq_len;
      if (syntax != NULL)
	qq_len = syntax->emit_qq_template(vm, context, buffer, next, nesting,
					  env, cpool, hnt, true);
      else
	qq_len = Syntax::emit_qq_template(vm, context, buffer, next, nesting,
					  env, cpool, hnt, true);
      if (qq_len == 0)
	return 0;	// error

      next = Cell::nil();
      while (!element_stack.empty())
	{
	  root = element_stack.back();
	  Cell *prev = *root;
	  if (Quasiquote::find_unquote(vm, context, env, prev, nesting))
	    break;

	  element_stack.pop_back();
	  delete root;
	  next = vm.mk_cons(prev, next, true);
	}

      if (next != Cell::nil())
	{
	  len += emit_local_variable_assignment(buffer, pv);
	  Cell *proc;
	  if (CDR(next) == Cell::nil())
	    {
	      len += emit_literal_expression(vm, context, buffer,
					     CAR(next), cpool);
	      proc = vm.core.internal_cons_procedure;
	    }
	  else
	    {
	      len += emit_literal_expression(vm, context, buffer, next, cpool);
	      proc = vm.core.internal_append_procedure;
	    }

	  len += emit_local_variable_assignment(buffer, nv);
	  len += emit_known_procedure_call(vm, context, buffer,
					   proc, env, cpool,
					   fp, 3);
	}
    }

  len += emit_local_variable_assignment(buffer, nv);
  len += emit_known_procedure_call(vm, context, buffer,
				   vm.core.internal_list2vector_procedure,
				   env, cpool, fp, 2);

  env->dispose_stack_slots(fp);

  if (list_qq)
    {
      unsigned long fp = env->get_current_stack_pointer() - 3;
      assert(fp > 0);

      len += emit_local_variable_assignment(buffer, fp + 1);
      len += emit_known_procedure_call(vm, context, buffer,
				       vm.core.internal_cons_procedure,
				       env, cpool, fp, 3);
    }

  return len;
}


size_t
Syntax::emit_known_procedure_call(VirtualMachine &vm, Register &context,
				  BytecodeBuffer *buffer,
				  Cell *procedure,
				  Cell *environment, Cell *constant_pool,
				  unsigned long frame_pointer,
				  unsigned long nargs)
{
  size_t len = emit_literal_expression(vm, context, buffer,
				       procedure, constant_pool);
  size_t imm_len;
  if (frame_pointer < nargs)
    {
      if (nargs <= 0xff)
	imm_len = 1;
      else if (nargs <= 0xffff)
	imm_len = 2;
      else
	{
#if SIZEOF_LONG > 4
	  assert(nargs <= 0xffffffff);	// not supported yet
#endif
	  imm_len = 4;
	}
    }
  else
    {
      if (frame_pointer <= 0xff)
	imm_len = 1;
      else if (frame_pointer <= 0xffff)
	imm_len = 2;
      else
	{
#if SIZEOF_LONG > 4
	  assert(frame_pointer <= 0xffffffff);	// not supported yet
#endif
	  imm_len = 4;
	}
    }

  switch (imm_len)
    {
    case 1:
      len += buffer->append(OP_CALL_CC1);
      len += buffer->append_unsigned_8(frame_pointer);
      len += buffer->append_unsigned_8(nargs);
      return len;


    case 2:
      len += buffer->append(OP_CALL_CC2);
      len += buffer->append_unsigned_16(frame_pointer);
      len += buffer->append_unsigned_16(nargs);
      return len;


    case 4:
      len += buffer->append(OP_CALL_CC4);
      len += buffer->append_unsigned_32(frame_pointer);
      len += buffer->append_unsigned_32(nargs);
      return len;
    }

#ifndef NDEBUG
  assert(0);	// should never reached
#endif
  return 0;
}

				 
size_t
Syntax::emit_qq_template(VirtualMachine &vm, Register &context,
			 BytecodeBuffer *buffer,
			 Cell *qq_template, long nesting,
			 Cell *environment, Cell *constant_pool,
			 Cell *hint, bool list_qq)
{
  if (vm.get_bytecompile_tracing(context))
    {
      Port *port = context.output_port->get_port();
      char buf[256];
      snprintf(buf, sizeof(buf), "Syntax::emit_qq_template<%ld>: ", nesting);
      port->append(buf);
      Writer::write_datum(qq_template, port, false);
      port->append('\n');
    }

  if (nesting == 0)
    {
      VMStackRoot env(vm, environment);
      return bytecompile(vm, context, buffer, qq_template, environment,
			 constant_pool, hint, false);
    }

  if (!Quasiquote::find_unquote(vm, context, environment,
				qq_template, nesting))
    return emit_literal_expression(vm, context, buffer,
				   qq_template, constant_pool);

  Syntax *syntax = lookup_syntax(qq_template, environment);
  if (syntax != NULL && syntax->is_quasiquote_syntax())
    return syntax->emit_qq_template(vm, context, buffer, qq_template, nesting,
				    environment, constant_pool, hint,
				    list_qq);

  if (qq_template->is_pair())
    return emit_list_qq_template(vm, context, buffer, qq_template, nesting,
				 environment, constant_pool, hint, list_qq);

  return emit_vector_qq_template(vm, context, buffer, qq_template, nesting,
				 environment, constant_pool, hint, list_qq);
}


size_t
Syntax::emit_variable_reference(VirtualMachine &vm, Register &context,
				BytecodeBuffer *buffer,
				const Cell *expression, Cell *environment,
				Cell *constant_pool)
{
  const Cell *binding;
  if (expression->is_symbol())
    binding = environment->lookup(expression, false);
  else
    {
      assert(expression->is_binding());

      if (expression->get_bound_environment() == environment)
	binding = expression;
      else
	binding = Cell::nil();
    }

  if (binding->is_binding())
    {
      size_t off = buffer->get_current_offset();

      buffer->append(OP_NOP);	// dummy;

      unsigned long slot = binding->get_slot_location();
      size_t imm_len = buffer->append_unsigned(slot);
      switch (imm_len)
	{
	case 1:
	  buffer->set_opcode(off, OP_LOAD_L1);
	  break;

	case 2:
	  buffer->set_opcode(off, OP_LOAD_L2);
	  break;

	case 4:
	  buffer->set_opcode(off, OP_LOAD_L4);
	  break;

	default:
	  assert(0);	// should never reached!
	}

      return imm_len + 1;
    }

  if (expression->is_symbol())
    binding = environment->lookup(expression, true);
  else
    {
      assert(expression->is_binding());

      if (expression->get_slot_location() >= 0)
	binding = expression;
      else
	expression = expression->get_bound_symbol();
    }

  if (binding->is_binding())
    {
      const Cell *bound_env = binding->get_bound_environment();
      if (is_toplevel_environment(vm, bound_env))
	{
	  size_t off = buffer->get_current_offset();
	  buffer->append(OP_NOP);	// dummy;
	  long slot = binding->get_slot_location();
	  size_t imm_len = buffer->append_unsigned(slot);
	  switch (imm_len)
	    {
	    case 1:
	      buffer->set_opcode(off, OP_LOAD_T1);
	      break;

	    case 2:
	      buffer->set_opcode(off, OP_LOAD_T2);
	      break;

	    case 4:
	      buffer->set_opcode(off, OP_LOAD_T4);
	      break;

	    default:
	      assert(0);	// should never reached!
	    }

	  return imm_len + 1;
	}

      long depth = environment->calculate_environment_distance(bound_env);
      // We should allow nested environment deeper than 255.
      assert(depth > 0 && depth <= 255);

      size_t off = buffer->get_current_offset();
      buffer->append(OP_NOP);	// dummy;
      buffer->append_unsigned_8(depth);
      long slot = binding->get_slot_location();
      size_t imm_len = buffer->append_unsigned(slot);
      switch (imm_len)
	{
	case 1:
	  buffer->set_opcode(off, OP_LOAD_G1);
	  break;

	case 2:
	  buffer->set_opcode(off, OP_LOAD_G2);
	  break;

	case 4:
	  buffer->set_opcode(off, OP_LOAD_G4);
	  break;

	default:
	  assert(0);	// should never reached!
	}

      return imm_len + 2;
    }

  // variable not bound at byte compilation time.
  long index = constant_pool->register_constant(expression);
  size_t off = buffer->get_current_offset();
  buffer->append(OP_NOP);	// dummy;
  size_t imm_len = buffer->append_unsigned(index);
  switch (imm_len)
    {
    case 1:
      buffer->set_opcode(off, OP_LOAD_F1);
      break;

    case 2:
      buffer->set_opcode(off, OP_LOAD_F2);
      break;

    case 4:
      buffer->set_opcode(off, OP_LOAD_F4);
      break;

    default:
      assert(0);	// should never reached!
    }

  return imm_len + 1;
}


size_t
Syntax::emit_variable_assignment(VirtualMachine &vm, Register &context,
				 BytecodeBuffer *buffer,
				 const Cell *symbol, Cell *environment,
				 Cell *constant_pool)
{
  assert(symbol->is_symbol());

  const Cell *binding = environment->lookup(symbol, true);
  if (!is_toplevel_environment(vm, environment))
    {
      if (binding->is_binding()
	  && binding->get_bound_environment() == environment)
	{
	  // variable is a local variable.
	  unsigned long slot = binding->get_slot_location();
	  return emit_local_variable_assignment(buffer, slot);
	}
    }

  if (binding->is_binding())
    {
      const Cell *bound_env = binding->get_bound_environment();
      if (is_toplevel_environment(vm, bound_env))
	{
	  size_t off = buffer->get_current_offset();
	  buffer->append(OP_NOP);	// dummy;
	  long slot = binding->get_slot_location();
	  size_t imm_len = buffer->append_unsigned(slot);
	  switch (imm_len)
	    {
	    case 1:
	      buffer->set_opcode(off, OP_STORE_T1);
	      break;

	    case 2:
	      buffer->set_opcode(off, OP_STORE_T2);
	      break;

	    case 4:
	      buffer->set_opcode(off, OP_STORE_T4);
	      break;

	    default:
	      assert(0);	// should never reached!
	    }

	  return imm_len + 1;
	}

      if (bound_env->get_parent_environment() == Cell::nil())
	return signal_error(vm, "cannot alter immutable variable: ", symbol);

      long depth = environment->calculate_environment_distance(bound_env);
      // Should we allow nested environment deeper than 255.
      if (depth < 0 || depth > 255)
	return 0;

      size_t off = buffer->get_current_offset();
      buffer->append(OP_NOP);	// dummy;
      buffer->append_unsigned_8(depth);
      long slot = binding->get_slot_location();
      size_t imm_len = buffer->append_unsigned(slot);
      switch (imm_len)
	{
	case 1:
	  buffer->set_opcode(off, OP_STORE_G1);
	  break;

	case 2:
	  buffer->set_opcode(off, OP_STORE_G2);
	  break;

	case 4:
	  buffer->set_opcode(off, OP_STORE_G4);
	  break;

	default:
	  assert(0);	// should never reached!
	}

      return imm_len + 2;
    }

  Cell *tmp_env = environment;
  while (true)
    {
      Cell *parent = tmp_env->get_parent_environment();
      if (parent == Cell::nil())
	{
	  if (tmp_env == vm.core.get_interaction_environment())
	    break;

	  return signal_error(vm, "cannot alter immutable variable: ", symbol);
	}

      tmp_env = parent;
    }

  // variable not bound at byte compilation time.
  long index = constant_pool->register_constant(symbol);
  size_t off = buffer->get_current_offset();
  buffer->append(OP_NOP);	// dummy;
  size_t imm_len = buffer->append_unsigned(index);
  switch (imm_len)
    {
    case 1:
      buffer->set_opcode(off, OP_STORE_F1);
      break;

    case 2:
      buffer->set_opcode(off, OP_STORE_F2);
      break;

    case 4:
      buffer->set_opcode(off, OP_STORE_F4);
      break;

    default:
#ifndef NDEBUG
      assert(0);	// should never reached
#else
      abort();		// should never reached
#endif
    }

  return imm_len + 1;
}


size_t
Syntax::emit_literal_expression(VirtualMachine &vm, Register &context,
				BytecodeBuffer *buffer,
				const Cell *expression, Cell *constant_pool)
{
  if (expression->is_integer())
    {
      static TSVMOpcode iconst_n_opcodes[] = {
	OP_ICONST_M1, OP_ICONST_0, OP_ICONST_1, OP_ICONST_2,
	OP_ICONST_3, OP_ICONST_4
      };

      long val = expression->get_integer0();
      if (val >= -1 && val <= 4)
	return buffer->append(iconst_n_opcodes[val + 1]);

      size_t off = buffer->get_current_offset();
      buffer->append(OP_NOP);	// dummy
      size_t imm_len = buffer->append_signed(val);
      switch (imm_len)
	{
	case 1:
	  buffer->set_opcode(off, OP_ICONST_I1);
	  break;


	case 2:
	  buffer->set_opcode(off, OP_ICONST_I2);
	  break;


	case 4:
	  buffer->set_opcode(off, OP_ICONST_I4);
	  break;


	case 8:
	  buffer->set_opcode(off, OP_ICONST_I8);
	  break;


	default:
	  assert(0);	// should never reached!
	}

      return imm_len + 1;
    }

  if (expression == Cell::nil())
    return buffer->append(OP_CONST_NULL);

  if (expression == Cell::t())
    return buffer->append(OP_CONST_T);

  if (expression == Cell::f())
    return buffer->append(OP_CONST_F);

  if (expression == Cell::unspecified())
    return buffer->append(OP_CONST_UNSPECIFIED);

  long index = constant_pool->register_constant(expression);
  assert(index >= 0);

  size_t off = buffer->get_current_offset();
  buffer->append(OP_NOP);	// dummy;
  size_t imm_len = buffer->append_unsigned(index);
  switch (imm_len)
    {
    case 1:
      buffer->set_opcode(off, OP_CONST_P1);
      break;


    case 2:
      buffer->set_opcode(off, OP_CONST_P2);
      break;


    case 4:
      buffer->set_opcode(off, OP_CONST_P4);
      break;


    default:
      assert(0);	// should never reached!
    }

  return imm_len + 1;
}


size_t
Syntax::emit_local_variable_assignment(BytecodeBuffer *buffer,
				       unsigned long slot)
{
  size_t off = buffer->get_current_offset();
  buffer->append(OP_NOP);	// dummy;
  size_t imm_len = buffer->append_unsigned(slot);
  switch (imm_len)
    {
    case 1:
      buffer->set_opcode(off, OP_STORE_L1);
      break;


    case 2:
      buffer->set_opcode(off, OP_STORE_L2);
      break;


    case 4:
      buffer->set_opcode(off, OP_STORE_L4);
      break;


    default:
      assert(0);	// should never reached!
    }

  return imm_len + 1;
}


size_t
Syntax::emit_procedure_call(VirtualMachine &vm, Register &context,
			    BytecodeBuffer *buffer,
			    Cell *expression, Cell *environment,
			    Cell *constant_pool, Cell *hint,
			    bool is_tail_context)
{
  assert(expression->is_pair());

  VMStackRoot form(vm, expression);
  VMStackRoot env(vm, environment);
  VMStackRoot cpool(vm, constant_pool);
  VMStackRoot hnt(vm, hint);

  // local variable slot #frame_bottom will be filled with the current
  // continuation at runtime by the virtual machine.
  unsigned long stack_bottom = env->allocate_stack_slot();
  // We evaluate the `operator' at last.

  size_t len = 0;
  VMStackRoot args(vm, CDR(form));
  while (args != Cell::nil())
    {
      size_t arg_len = bytecompile(vm, context, buffer, CAR(args),
				   env, cpool, hnt, false);
      if (arg_len > 0)
	{
	  unsigned long fp = env->allocate_stack_slot();
	  len += arg_len + emit_local_variable_assignment(buffer, fp);
	  args = CDR(args);
	  continue;
	}

      signal_error(vm, "invalid argument: ", CAR(args));
      return signal_error(vm, "syntax error procedure call: ", form);
    }

  size_t op_len = bytecompile(vm, context, buffer,
			      CAR(form), env, cpool, hnt, false);
  if (op_len == 0)
    {
      signal_error(vm, "invalid operator: ", CAR(form));
      return signal_error(vm, "syntax error procedure call: ", form);
    }

  len += op_len;
  unsigned long nargs = env->get_current_stack_pointer() - stack_bottom;
  size_t imm_len;
  if (stack_bottom < nargs)
    {
      if (nargs <= 0xff)
	imm_len = 1;
      else if (nargs <= 0xffff)
	imm_len = 2;
      else
	{
#if SIZEOF_LONG > 4
	  assert(nargs <= 0xffffffff);	// not supported yet
#endif
	  imm_len = 4;
	}
    }
  else
    {
      if (stack_bottom <= 0xff)
	imm_len = 1;
      else if (stack_bottom <= 0xffff)
	imm_len = 2;
      else
	{
#if SIZEOF_LONG > 4
	  assert(stack_bottom <= 0xffffffff);	// not supported yet
#endif
	  imm_len = 4;
	}
    }

  if (is_tail_context)
    {
      switch (imm_len)
	{
	case 1:
	  len += buffer->append(OP_APPLY1);
	  len += buffer->append_unsigned_8(stack_bottom);
	  len += buffer->append_unsigned_8(nargs);
	  break;


	case 2:
	  len += buffer->append(OP_APPLY2);
	  len += buffer->append_unsigned_16(stack_bottom);
	  len += buffer->append_unsigned_16(nargs);
	  break;


	case 4:
	  len += buffer->append(OP_APPLY4);
	  len += buffer->append_unsigned_32(stack_bottom);
	  len += buffer->append_unsigned_32(nargs);
	  break;


	default:
	  assert(0);	// should never reached!
	}
    }
  else
    {
      switch (imm_len)
	{
	case 1:
	  len += buffer->append(OP_CALL_CC1);
	  len += buffer->append_unsigned_8(stack_bottom);
	  len += buffer->append_unsigned_8(nargs);
	  break;


	case 2:
	  len += buffer->append(OP_CALL_CC2);
	  len += buffer->append_unsigned_16(stack_bottom);
	  len += buffer->append_unsigned_16(nargs);
	  break;


	case 4:
	  len += buffer->append(OP_CALL_CC4);
	  len += buffer->append_unsigned_32(stack_bottom);
	  len += buffer->append_unsigned_32(nargs);
	  break;


	default:
	  assert(0);	// should never reached!
	}
    }

  env->dispose_stack_slots(stack_bottom);
  return len;
}


const Cell *
Syntax::bind_variable(VirtualMachine &vm, Cell *environment,
		      const Cell *symbol, Cell *hint)
{
  const Cell *binding = environment->lookup(symbol, false);
  if (binding->is_binding())
    return binding;

  binding = environment->lookup(symbol, true);
  if (binding->is_syntax())
    {
      while (hint != Cell::nil())
	{
	  if (CAR(hint) == symbol)
	    {
	      signal_error(vm, "cannot shadow syntactic keyword in use: ",
			   symbol->get_symbol_string());
	      return Cell::nil();
	    }

	  hint = CDR(hint);
	}
    }

  return environment->bind_variable(symbol);
}


size_t
Syntax::bytecompile(VirtualMachine &vm, Register &context,
		    BytecodeBuffer *buffer,
		    Cell *expression, Cell *environment,
		    Cell *constant_pool, Cell *hint,
		    bool is_tail_context)
{
  if (expression == Cell::nil())
    return signal_error(vm, "empty combination: ", expression);

  if (vm.get_bytecompile_tracing(context))
    {
      Port *port = context.output_port->get_port();
      if (is_tail_context)
	port->append("Syntax::bytecompile <TAIL> (hint): ");
      else
	port->append("Syntax::bytecompile (hint): ");

      Writer::write_datum(hint, port, false);
      port->append("\n                    (form): ");
      Writer::write_datum(expression, port, false);
      port->append('\n');
    }

  Syntax *syntax = lookup_syntax(expression, environment);
  if (syntax != NULL)
    return syntax->emit_bytecode(vm, context, buffer, expression, environment,
				 constant_pool, hint,
				 is_tail_context);

  if (expression->is_pair())
    return emit_procedure_call(vm, context, buffer, expression, environment,
			       constant_pool, hint, is_tail_context);

  size_t len;
  if (expression->is_symbol() || expression->is_binding())
    len = emit_variable_reference(vm, context, buffer, expression, environment,
				  constant_pool);
  else
    len = emit_literal_expression(vm, context, buffer,
				  expression, constant_pool);

  if (!is_tail_context)
    return len;

  // emit application of default continuation.
  return len + buffer->append(OP_APPLY_CONTINUATION0);
}


size_t
Syntax::bytecompile_sequence(VirtualMachine &vm, Register &context,
			     BytecodeBuffer *buffer,
			     Cell *expression, Cell *environment,
			     Cell *constant_pool, Cell *hint,
			     bool is_tail_context)
{
#if 1
  if (!expression->is_pair())
    return signal_error(vm, "internal error: not a sequence: ", expression);
#else
  assert(expression->is_pair());
#endif

  if (vm.get_bytecompile_tracing(context))
    {
      Port *port = context.output_port->get_port();
      if (is_tail_context)
	port->append("Syntax::bytecompile_sequence <TAIL> (hint): ");
      else
	port->append("Syntax::bytecompile_sequence (hint): ");

      Writer::write_datum(hint, port, false);
      port->append("\n                             (form): ");
      Writer::write_datum(expression, port, false);
      port->append('\n');
    }

  VMStackRoot exp(vm, expression);
  VMStackRoot env(vm, environment);
  VMStackRoot cpool(vm, constant_pool);
  VMStackRoot hnt(vm, hint);

  if (!is_toplevel(vm, env, hnt))
    {
      VMStackRoot bindings(vm, Cell::nil());
      VMStackRoot seq(vm, exp);
      for (; seq != Cell::nil(); seq = CDR(seq))
	{
	  Cell *form = CAR(seq);
	  if (!form->is_pair())
	    break;

	  const Cell *binding;
	  if (CAR(form)->is_syntax() || CAR(form)->is_binding())
	    binding = CAR(form);
	  else
	    binding = env->lookup(CAR(form), true);

	  if (!binding->is_syntax()
	      || binding->get_syntax() != &vm.core.syntax_define)
	    break;

	  // internal definition found
	  if (!CDR(form)->is_pair()
	      || !CDDR(form)->is_pair())
	    return signal_error(vm, "syntax error define: ", form);

	  Cell *first = CADR(form);
	  if (first->is_pair())
	    {
	      // procedure definition
	      if (!CAR(first)->is_symbol() && !CAR(first)->is_binding())
		return signal_error(vm, "syntax error define: ", form);

	      Cell *args = CDR(first);
	      Cell *body = CDDR(form);
	      Cell *tmp_form = vm.mk_cons(args, body, true);
	      tmp_form = vm.mk_cons(vm.core.internal_lambda_symbol,
				    tmp_form, true);
	      tmp_form = vm.mk_cons(tmp_form, Cell::nil0(), true);
	      Cell *name = CAADR(form);
	      if (name->is_binding())
		name = const_cast<Cell *>(name->get_bound_symbol());

	      tmp_form = vm.mk_cons(name, tmp_form, true);
	      bindings = vm.mk_cons(tmp_form, bindings, true);
	    }
	  else
	    {
	      // variable definition
	      Cell *rest = CDDR(form);
	      if (!first->is_symbol() && !first->is_binding())
		return signal_error(vm, "syntax error define: ", form);

	      if (CDR(rest) != Cell::nil())
		return signal_error(vm, "syntax error define: ", form);

	      Cell *binding = CDR(form);
	      bindings = vm.mk_cons(binding, bindings, true);
	    }
	}

      Cell *body = seq;
      for (; seq != Cell::nil(); seq = CDR(seq))
	{
	  // Ensure no more internal definitions
	  Cell *form = CAR(seq);
	  if (form->is_pair())
	    {
	      const Cell *binding;
	      if (CAR(form)->is_symbol())
		binding = env->lookup(CAR(form), true);
	      else
		binding = CAR(form);

	      if (binding->is_syntax())
		{
		  Syntax *syntax = binding->get_syntax();
		  if (syntax == &vm.core.syntax_define
		      || syntax == &vm.core.syntax_define_syntax)
		    {
		      signal_error(vm, "unexpected sub form: ", form);
		      return 0;
		    }
		}
	    }
	}

      if (bindings != Cell::nil())
	{
	  bindings
	    = R5RSListProcedure::list_reverse_in_place(vm.get_memory_system(),
						       Cell::nil(), bindings,
						       true);
	  return Letrec::emit_letrec(vm, context, buffer, exp, env, cpool, hnt,
				     bindings, body, is_tail_context);
	}
    }

  // toplevel sequence or sequence without internal definition
  size_t len = 0;
  while (CDR(exp)->is_pair())
    {
      size_t sub_len = bytecompile(vm, context, buffer, CAR(exp),
				   env, cpool, hnt, false);
      if (sub_len == 0)
	return 0;	// error

      len += sub_len;
      exp = CDR(exp);
    }

  size_t sub_len = bytecompile(vm, context, buffer, CAR(exp),
			       env, cpool, hnt, is_tail_context);
  if (sub_len == 0)
    return 0;	// error

  if (CDR(exp) == Cell::nil())
    return len + sub_len;

  return 0;	// error
}


size_t
Syntax::signal_error(VirtualMachine &vm, const char *message,
		     const Cell *expression)
{
  OutputStringPort *error_buf = vm.get_error_buffer();
  error_buf->append(message);
  if (expression != NULL)
    Writer::write_datum(expression, error_buf, false);

  error_buf->append('\n');
  return 0;
}


Cell *
Syntax::rename_variable(VirtualMachine &vm, Cell *expression,
			const Cell *rename_from, const Cell *rename_to)
{
  if (expression == rename_from)
    return const_cast<Cell *>(rename_to);

  if (expression->is_pair())
    {
      VMStackRoot expr(vm, expression);
      VMStackRoot from(vm, const_cast<Cell *>(rename_from));
      VMStackRoot to(vm, const_cast<Cell *>(rename_to));
      Cell *tmp = rename_variable(vm, CAR(expr), from, to);
      vm.set_car(expr, tmp);
      tmp = rename_variable(vm, CDR(expr), from, to);
      vm.set_cdr(expr, tmp);
    }
  else if (expression->is_vector())
    {
      VMStackRoot expr(vm, expression);
      VMStackRoot from(vm, const_cast<Cell *>(rename_from));
      VMStackRoot to(vm, const_cast<Cell *>(rename_to));
      long size = expr->get_vector_length();
      for (long i = 0; i < size; i++)
	{
	  Cell *tmp = rename_variable(vm, expr->get_vector_element(i),
				      from, to);
	  vm.set_vector_element(expr, i, tmp);
	}
    }

  return expression;
}


const Cell *
Syntax::get_unique_symbol(VirtualMachine &vm, Cell *environment,
			  const Cell *symbol)
{
  if (symbol->is_binding())
    symbol = symbol->get_bound_symbol();
  assert(symbol->is_symbol());

  char symbol_template[256];
  if (symbol != NULL)
    snprintf(symbol_template, 256, "#<gensym: %s-%%ld>",
	     symbol->get_symname());
  else
    snprintf(symbol_template, 256, "#<gensym: %%ld>");

  VMStackRoot env(vm, environment);
  char gensym_buffer[256];
  long i = 0;
  while (true)
    {
      snprintf(gensym_buffer, 256, symbol_template, i++);
      symbol = vm.core.mk_symbol(gensym_buffer);
      if (env->lookup(symbol, false) == Cell::nil())
	return symbol;
    }
}

}
