/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: syntax.h,v 1.13 2009/01/01 07:47:01 fuyu Exp $
 */

#ifndef _LIBTSCORE_SYNTAX_H_
#define _LIBTSCORE_SYNTAX_H_

#ifdef __cplusplus

#if defined(__GNUG__) && !defined(__ICC)
#pragma interface
#endif

#include "libtscore/cell.h"


using namespace std;
namespace LibTSCore
{
  class BytecodeBuffer;
  class Label;
  class Register;
  class TSCore;
  class VirtualMachine;


  /*
   * Syntax abstracts the execution of Scheme syntax using our virtual
   * machine bytecodes.
   *
   * Essentially each concrete instance of Syntax behaves like a bytecode
   * compiler tailored to compile a Scheme program fragments that uses
   * a specific syntax.  Of course, Scheme program fragments usually have
   * another form that uses another syntax within a fragment.  In such
   * case, each compiler can query which Syntax is bound to an symbol
   * within a program fragment in the given environment and delegate
   * the compilation of the sub fragment to appropriate Syntax.
   */
  class Syntax
  {
    friend class SyntaxRules;
    friend class VirtualMachine;

  protected:
    /*
     * Appends bytecodes that represents the execution of `expression'
     * in the `environment' to `buffer'.
     *
     * The `constant_pool' is a constant pool Cell that can be used
     * to store pointers to Cell that can be treated as constant at
     * bytecompilation time.
     */
    virtual size_t emit_bytecode(VirtualMachine &vm, Register &context,
				 BytecodeBuffer *buffer,
				 Cell *expression, Cell *environment,
				 Cell *constant_pool, Cell *hint,
				 bool is_tail_context) = 0;
    virtual bool is_quasiquote_syntax();
    virtual size_t emit_qq_template(VirtualMachine &vm, Register &context,
				    BytecodeBuffer *buffer,
				    Cell *qq_template, long nesting,
				    Cell *environment, Cell *constant_pool,
				    Cell *hint, bool list_qq);


    size_t emit_list_qq_template(VirtualMachine &vm, Register &context,
				 BytecodeBuffer *buffer,
				 Cell *qq_template, long nesting,
				 Cell *environment, Cell *constant_pool,
				 Cell *hint, bool list_qq);
				 
    size_t emit_vector_qq_template(VirtualMachine &vm, Register &context,
				   BytecodeBuffer *buffer,
				   Cell *qq_template, long nesting,
				   Cell *environment, Cell *constant_pool,
				   Cell *hint, bool list_qq);


    /*
     * Emits bytecodes to load a value from a variable.
     */
    static size_t emit_variable_reference(VirtualMachine &vm,
					  Register &context,
					  BytecodeBuffer *buffer,
					  const Cell *symbol,
					  Cell *environment,
					  Cell *constant_pool);


    /*
     * Emits bytecodes to store a value into a variable.
     *
     * If hint is not the Cell::nil(), the assignment is treated as
     * toplevel definition, i.e. assignment to unbound variable will be
     * succeeded at runtime.
     */
    static size_t emit_variable_assignment(VirtualMachine &vm,
					   Register &context,
					   BytecodeBuffer *buffer,
					   const Cell *symbol,
					   Cell *environment,
					   Cell *constant_pool);


    /*
     * Emits bytecodes to setup a literal expression.
     */
    static size_t emit_literal_expression(VirtualMachine &vm,
					  Register &context,
					  BytecodeBuffer *buffer,
					  const Cell *expression,
					  Cell *constant_pool);


    /*
     * Emits bytecodes to call a procedure.
     */
    static size_t emit_procedure_call(VirtualMachine &vm,
				      Register &context,
				      BytecodeBuffer *buffer,
				      Cell *expression, Cell *environment,
				      Cell *constant_pool, Cell *hint,
				      bool is_tail_context);


    static size_t emit_known_procedure_call(VirtualMachine &vm,
					    Register &context,
					    BytecodeBuffer *buffer,
					    Cell *procedure,
					    Cell *environment,
					    Cell *constant_pool,
					    unsigned long frame_pointer,
					    unsigned long nargs);


    static inline Syntax *lookup_syntax(Cell *expression, Cell *environment)
    {
      if (expression->is_pair())
	{
	  Cell *op = CAR(expression);
	  if (op->is_syntax())
	    return op->get_syntax();

	  if (op->is_symbol())
	    {
	      const Cell *binding = environment->lookup(op, true);
	      if (binding->is_syntax())
		return binding->get_syntax();
	    }
	}
      return NULL;
    }


    /*
     * Bytecompiles given `expression' into given buffer.  This is
     * used by each concrete Syntax object to compile its sub forms.
     */
    static size_t bytecompile(VirtualMachine &vm, Register &context,
			      BytecodeBuffer *buffer,
			      Cell *expression, Cell *environment,
			      Cell *constant_pool, Cell *hint,
			      bool is_tail_context);


    /*
     * Bytecompiles given `expression' as a sequence of expression.
     * This is used by each concreate Syntax object to compile its sub
     * forms.
     */
    static size_t bytecompile_sequence(VirtualMachine &vm, Register &context,
				       BytecodeBuffer *buffer,
				       Cell *sequence, Cell *environment,
				       Cell *constant_pool, Cell *hint,
				       bool is_tail_context);

    static inline bool is_toplevel_environment(VirtualMachine &vm,
					       const Cell *environment);
    static inline bool is_toplevel(VirtualMachine &vm, const Cell *environment,
				   Cell *hint);


    static size_t signal_error(VirtualMachine &vm,
			       const char *message,
			       const Cell *expression = NULL);


    /*
     * Rename all appearances of symbol `rename_from' in the `expression'
     * with `rename_to'.
     *
     * This returns the top of renamed expression to allow renaming of
     * single symbol, i.e. `expression' is `rename_from'.
     */
    static Cell *rename_variable(VirtualMachine &vm, Cell *expression,
				 const Cell *rename_from,
				 const Cell *rename_to);


    /*
     * Finds a unique symbol within the `environment'.
     */
    static const Cell *get_unique_symbol(VirtualMachine &vm, Cell *environment,
					 const Cell *symbol = NULL);


    /*
     * Extends `hint' with a syntactic keyword to be used for compilation of
     * sub-forms.
     *
     * We use `hint' through bytecompilation to represent a list
     * of syntactic keywords in use for the compilation unit.
     */
    static inline Cell *extend_hint(VirtualMachine &vm, Cell *keyword,
				    Cell *hint);


    /*
     * Binds `symbol' as a variable within given `environment'.
     *
     * This also checks the binding of symbol not to shadow the syntactic
     * keywords in use within the same compilation unit that is an error.
     *
     * Result is the binding object or Cell::nil() iff in error case.
     */
    static const Cell *bind_variable(VirtualMachine &vm, Cell *environment,
				     const Cell *symbol, Cell *hint);


    static size_t emit_local_variable_assignment(BytecodeBuffer *buffer,
						 unsigned long slot);


  public:
    Syntax() {}
    virtual ~Syntax() {}
  };

}


#endif	/* __cplusplus */

#define _SYNTAX_DEFINED_	1
#include "libtscore/tscore.h"
#include "libtscore/virtual_machine.h"

#endif	/* _LIBTSCORE_SYNTAX_H_ */


#ifdef __cplusplus
#if defined(_TSCORE_DEFINED_) && defined(_VIRTUAL_MACHINE_DEFINED_) && !defined(_SYNTAX_INLINE_FUNC_DEFINED_)
namespace LibTSCore
{
  inline Cell *Syntax::extend_hint(VirtualMachine &vm,
				   Cell *keyword, Cell *hint)
  {
    if (keyword->is_syntax())
      keyword = keyword->get_syntax_symbol();

    assert(keyword->is_symbol());

    return vm.mk_cons(keyword, hint, false);
  }

  inline bool Syntax::is_toplevel(VirtualMachine &vm,
				  const Cell *environment, Cell *hint)
  {
    if (hint == Cell::nil())
      return true;

    if (CDR(hint) != Cell::nil())
      return false;

    const Cell *syntax = environment->lookup(CAR(hint), true);
    assert(syntax->is_syntax());

    return syntax->get_syntax() == &vm.core.syntax_begin;
  }
}

#define _SYNTAX_INLINE_FUNC_DEFINED_	1
#endif
#endif	/* __cplusplus */
