/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: tscore.cc,v 1.53 2009/01/05 12:16:19 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libtscore/tscore.h"

#include "libtscore/vmopcode.h"

#include <algorithm>
#include <string>

#include <cassert>
#include <cstdlib>


namespace LibTSCore
{

#define INTERNAL_LAMBDA		"#<syntax: lambda>"


static const char *r5rs_cond_syntax_rules =
  "(syntax-rules (else =>) ((cond (else result1 result2 ...))(begin result1 result2 ...))((cond (test => result))(let ((temp test))(if temp (result temp))))((cond (test => result) clause1 clause2 ...)(let ((temp test))(if temp (result temp)(cond clause1 clause2 ...))))((cond (test)) test)((cond (test) clause1 clause2 ...)(let ((temp test))(if temp temp (cond clause1 clause2 ...))))((cond (test result1 result2 ...))(if test (begin result1 result2 ...)))((cond (test result1 result2 ...) clause1 clause2 ...)(if test (begin result1 result2 ...)(cond clause1 clause2 ...))))";


static const char *r5rs_case_syntax_rules =
  "(syntax-rules (else) ((case (key ...) clauses ...)(let ((atom-key (key ...)))(case atom-key clauses ...)))((case key (else result1 result2 ...))(begin result1 result2 ...))((case key ((atoms ...) result1 result2 ...))(if (memv key '(atoms ...))(begin result1 result2 ...)))((case key ((atoms ...) result1 result2 ...) clause clauses ...)(if (memv key '(atoms ...))(begin result1 result2 ...)(case key clause clauses ...))))";


static const char *r5rs_let_syntax_rules =
  "(syntax-rules () ((let ((name val) ...) body1 body2 ...)((lambda (name ...) body1 body2 ...) val ...))((let tag ((name val) ...) body1 body2 ...)((letrec ((tag (lambda (name ...) body1 body2 ...))) tag) val ...)))";


static const char *r5rs_do_syntax_rules =
  "(syntax-rules () ((do ((var init step ...) ...)(test expr ...) command ...)(letrec ((loop (lambda (var ...)(if test (begin (if #f #f) expr ...) (begin command ... (loop (do \"step\" var step ...) ...))))))(loop init ...)))((do \"step\" x) x)((do \"step\" x y) y))";


TSCore::TSCore(size_t initial_n_cell_segs, size_t maximum_n_cell_segs,
	       size_t cell_segment_size)
  :
    memory_system(*this, initial_n_cell_segs, maximum_n_cell_segs,
		  cell_segment_size),
    symbol_table(*this),

    gettext_procedure(*this),
    r5rs_char_procedure(*this),
    r5rs_io_procedure(*this),
    r5rs_list_procedure(*this),
    r5rs_math_procedure(*this),
    r5rs_misc_procedure(*this),
    r5rs_string_procedure(*this),
    srfi6_procedure(*this),
    srfi19_procedure(*this),
    srfi28_procedure(*this),
    tscore_procedure(*this),
    interaction_environment(Cell::nil()),
    interaction_toplevel_frame(Cell::nil()),
    r5rs_environment(Cell::nil()),
    r5rs_toplevel_frame(Cell::nil()),
    null_environment(Cell::nil()),
    internal_cond_syntax(Cell::nil()),
    internal_case_syntax(Cell::nil()),
    internal_let_syntax(Cell::nil()),
    internal_do_syntax(Cell::nil()),
    interaction_input_port(Cell::nil()),
    interaction_output_port(Cell::nil())
{
  memory_system.register_static_root(&interaction_environment);
  memory_system.register_static_root(&interaction_toplevel_frame);
  memory_system.register_static_root(&r5rs_environment);
  memory_system.register_static_root(&r5rs_toplevel_frame);
  memory_system.register_static_root(&null_environment);


  // initialize interaction_environment
  interaction_environment
    = Cell::toplevel_environment_new(memory_system);
  interaction_toplevel_frame
    = Cell::toplevel_frame_new(memory_system, 256);
  init_tscore_environment(interaction_environment, interaction_toplevel_frame);



  lambda_symbol = mk_static_symbol("lambda");
  memory_system.register_static_root(&lambda_symbol);
  quote_symbol = mk_static_symbol("quote");
  memory_system.register_static_root(&quote_symbol);
  quasiquote_symbol = mk_static_symbol("quasiquote");
  memory_system.register_static_root(&quasiquote_symbol);
  unquote_symbol = mk_static_symbol("unquote");
  memory_system.register_static_root(&unquote_symbol);
  unquote_splicing_symbol = mk_static_symbol("unquote-splicing");
  memory_system.register_static_root(&unquote_splicing_symbol);



  else_symbol = mk_static_symbol("else");
  memory_system.register_static_root(&else_symbol);
  feed_to_symbol = mk_static_symbol("=>");
  memory_system.register_static_root(&feed_to_symbol);

  syntax_rules_symbol = mk_static_symbol("syntax-rules");
  memory_system.register_static_root(&syntax_rules_symbol);
  ellipsis_symbol = mk_static_symbol("...");
  memory_system.register_static_root(&ellipsis_symbol);

  internal_lambda_symbol = mk_static_symbol(INTERNAL_LAMBDA);
  memory_system.register_static_root(&internal_lambda_symbol);



  internal_unquote_syntax = mk_static_syntax("unquote", &syntax_unquote);
  memory_system.register_static_root(&internal_unquote_syntax);
  internal_unquote_splicing_syntax
    = mk_static_syntax("unquote-splicing", &syntax_unquote_splicing);
  memory_system.register_static_root(&internal_unquote_splicing_syntax);



  const Cell *binding
    = interaction_environment->lookup(mk_static_symbol("cons"), false);
  assert(binding->is_binding());

  internal_cons_procedure
    = interaction_toplevel_frame->load_toplevel_variable(binding->get_slot_location());
  assert(internal_cons_procedure->is_procedure());

  memory_system.register_static_root(&internal_cons_procedure);



  binding = interaction_environment->lookup(mk_static_symbol("append"), false);
  assert(binding->is_binding());

  internal_append_procedure
    = interaction_toplevel_frame->load_toplevel_variable(binding->get_slot_location());
  assert(internal_append_procedure->is_procedure());

  memory_system.register_static_root(&internal_append_procedure);



  binding = interaction_environment->lookup(mk_static_symbol("list->vector"),
					    false);
  assert(binding->is_binding());

  internal_list2vector_procedure
    = interaction_toplevel_frame->load_toplevel_variable(binding->get_slot_location());
  assert(internal_list2vector_procedure->is_procedure());

  memory_system.register_static_root(&internal_list2vector_procedure);


  static const char meta_return_bytecode[] = OPS_META_RETURN;
  meta_return = memory_system.get_cell();
  meta_return->mk_constant_continuation(meta_return_bytecode);
  memory_system.register_static_root(&meta_return);

  memory_system.register_static_root(&interaction_input_port);
  memory_system.register_static_root(&interaction_output_port);

  internal_cond_syntax = mk_static_macro_syntax("cond",
						r5rs_cond_syntax_rules);
  memory_system.register_static_root(&internal_cond_syntax);
  if (internal_cond_syntax != Cell::nil())
    interaction_environment->bind_syntax(internal_cond_syntax);

  internal_case_syntax = mk_static_macro_syntax("case",
						r5rs_case_syntax_rules);
  memory_system.register_static_root(&internal_case_syntax);
  if (internal_case_syntax != Cell::nil())
    interaction_environment->bind_syntax(internal_case_syntax);

  internal_let_syntax = mk_static_macro_syntax("let",
					      r5rs_let_syntax_rules);
  memory_system.register_static_root(&internal_let_syntax);
  if (internal_let_syntax != Cell::nil())
    interaction_environment->bind_syntax(internal_let_syntax);

  internal_do_syntax = mk_static_macro_syntax("do",
					      r5rs_do_syntax_rules);
  memory_system.register_static_root(&internal_do_syntax);
  if (internal_do_syntax != Cell::nil())
    interaction_environment->bind_syntax(internal_do_syntax);
}


TSCore::~TSCore()
{
  Cell *nil = Cell::nil();

  while (!idle_vm.empty())
    {
      delete idle_vm.back();
      idle_vm.pop_back();
    }

  while (!external_procedures.empty())
    {
      delete external_procedures.back();
      external_procedures.pop_back();
    }

  interaction_environment = nil;
  interaction_toplevel_frame = nil;
  r5rs_environment = nil;
  r5rs_toplevel_frame = nil;
  null_environment = nil;

  memory_system.unregister_static_root(&interaction_environment);
  memory_system.unregister_static_root(&interaction_toplevel_frame);
  memory_system.unregister_static_root(&r5rs_environment);
  memory_system.unregister_static_root(&r5rs_toplevel_frame);
  memory_system.unregister_static_root(&null_environment);

  lambda_symbol = nil;
  quote_symbol = nil;
  quasiquote_symbol = nil;
  unquote_symbol = nil;
  unquote_splicing_symbol = nil;
  else_symbol = nil;
  feed_to_symbol = nil;
  syntax_rules_symbol = nil;
  ellipsis_symbol = nil;

  memory_system.unregister_static_root(&lambda_symbol);
  memory_system.unregister_static_root(&quote_symbol);
  memory_system.unregister_static_root(&quasiquote_symbol);
  memory_system.unregister_static_root(&unquote_symbol);
  memory_system.unregister_static_root(&unquote_splicing_symbol);
  memory_system.unregister_static_root(&else_symbol);
  memory_system.unregister_static_root(&feed_to_symbol);
  memory_system.unregister_static_root(&syntax_rules_symbol);
  memory_system.unregister_static_root(&ellipsis_symbol);

  internal_lambda_symbol = nil;

  memory_system.unregister_static_root(&internal_lambda_symbol);

  internal_unquote_syntax = nil;
  internal_unquote_splicing_syntax = nil;

  memory_system.unregister_static_root(&internal_unquote_syntax);
  memory_system.unregister_static_root(&internal_unquote_splicing_syntax);

  internal_cons_procedure = nil;
  internal_append_procedure = nil;
  internal_list2vector_procedure = nil;

  memory_system.register_static_root(&internal_cons_procedure);
  memory_system.register_static_root(&internal_append_procedure);
  memory_system.register_static_root(&internal_list2vector_procedure);

  meta_return = nil;

  memory_system.unregister_static_root(&meta_return);

  interaction_input_port = nil;
  interaction_output_port = nil;

  memory_system.unregister_static_root(&interaction_input_port);
  memory_system.unregister_static_root(&interaction_output_port);

  internal_cond_syntax = nil;
  internal_case_syntax = nil;
  internal_let_syntax = nil;
  internal_do_syntax = nil;

  memory_system.unregister_static_root(&internal_cond_syntax);
  memory_system.unregister_static_root(&internal_case_syntax);
  memory_system.unregister_static_root(&internal_let_syntax);
  memory_system.unregister_static_root(&internal_do_syntax);

  symbol_table.purge_all();

  memory_system.set_verbose_gc(false);
  memory_system.gc(nil, nil);
}


void
TSCore::output_message(const char *message)
{
  if (interaction_output_port->is_output_port())
    interaction_output_port->get_port()->append(message)->flush();
  else
    printf("%s", message);
}


Cell *
TSCore::mk_static_macro_syntax(const char *keyword,
			       const char *syntax_rules)
{
  StackRoot name(memory_system, mk_static_symbol(keyword));
  VirtualMachine *vm = get_idle_virtual_machine();
  InputStringPort port(syntax_rules);
  Cell *rules = vm->get_reader()->read_datum(&port);
  if (rules == NULL)
    fprintf(stderr, "bad formed expression: %s\n", syntax_rules);

  const Cell *transformer = SyntaxRules::parse(*vm, name, rules,
					       interaction_environment);
  if (transformer == NULL)
    {
      fprintf(stderr, "bad syntax-rules: %s\n", syntax_rules);
      return Cell::nil();
    }

  Cell *syntax = mk_macro_syntax(name, transformer);
  return_idle_virtual_machine(vm);

  return syntax;
}


VirtualMachine *
TSCore::get_idle_virtual_machine()
{
  if (!idle_vm.empty())
    {
      VirtualMachine *vm = idle_vm.back();
      idle_vm.pop_back();
      return vm;
    }

  return new VirtualMachine(*this);
}


void
TSCore::return_idle_virtual_machine(VirtualMachine *vm)
{
  idle_vm.push_back(vm);
}


Cell *
TSCore::init_null_environment(Cell *environment)
{
  StackRoot env(memory_system, environment);

  env->bind_syntax(mk_static_syntax("define", &syntax_define));
  env->bind_syntax(mk_static_syntax("define-syntax", &syntax_define_syntax));
  env->bind_syntax(mk_static_syntax("let-syntax", &syntax_let_syntax));
  env->bind_syntax(mk_static_syntax("letrec-syntax", &syntax_letrec_syntax));

  env->bind_syntax(mk_static_syntax("quote", &syntax_quote));
  env->bind_syntax(mk_static_syntax("lambda", &syntax_lambda));
  env->bind_syntax(mk_static_syntax(INTERNAL_LAMBDA, &syntax_lambda));
  env->bind_syntax(mk_static_syntax("if", &syntax_if));
  env->bind_syntax(mk_static_syntax("set!", &syntax_set));

  env->bind_syntax(mk_static_syntax("and", &syntax_and));
  env->bind_syntax(mk_static_syntax("or", &syntax_or));
  env->bind_syntax(mk_static_syntax("let*", &syntax_let_asterisk));
  env->bind_syntax(mk_static_syntax("letrec", &syntax_letrec));

  env->bind_syntax(mk_static_syntax("begin", &syntax_begin));

  env->bind_syntax(mk_static_syntax("delay", &syntax_delay));

  env->bind_syntax(mk_static_syntax("quasiquote", &syntax_quasiquote));

  if (internal_cond_syntax != Cell::nil())
    env->bind_syntax(internal_cond_syntax);

  if (internal_case_syntax != Cell::nil())
    env->bind_syntax(internal_case_syntax);

  if (internal_let_syntax != Cell::nil())
    env->bind_syntax(internal_let_syntax);

  if (internal_do_syntax != Cell::nil())
    env->bind_syntax(internal_do_syntax);

  return env;
}


Cell *
TSCore::get_null_environment()
{
  if (null_environment == Cell::nil())
    {
      null_environment
	= Cell::toplevel_environment_new(memory_system);
      init_null_environment(null_environment);
    }

  return null_environment;
}


Cell *
TSCore::init_r5rs_environment(Cell *env, Cell *frame)
{
  StackRoot tmp_env(memory_system, env);
  StackRoot tmp_frame(memory_system, frame);

  init_null_environment(tmp_env);

  r5rs_char_procedure.assign_procedures(tmp_env, tmp_frame);
  r5rs_io_procedure.assign_procedures(tmp_env, tmp_frame);
  r5rs_list_procedure.assign_procedures(tmp_env, tmp_frame);
  r5rs_math_procedure.assign_procedures(tmp_env, tmp_frame);
  r5rs_misc_procedure.assign_procedures(tmp_env, tmp_frame);

  return r5rs_string_procedure.assign_procedures(tmp_env, tmp_frame);
}


Cell *
TSCore::get_r5rs_environment()
{
  if (r5rs_environment == Cell::nil())
    {
      r5rs_environment = Cell::toplevel_environment_new(memory_system);
      r5rs_toplevel_frame = Cell::toplevel_frame_new(memory_system, 128);
      init_r5rs_environment(r5rs_environment, r5rs_toplevel_frame);
    }

  return r5rs_environment;
}


Cell *
TSCore::get_r5rs_toplevel_frame()
{
  if (r5rs_toplevel_frame == Cell::nil())
    {
      get_r5rs_environment();
      assert(r5rs_toplevel_frame != Cell::nil());
    }

  return r5rs_toplevel_frame;
}


Cell *
TSCore::init_tscore_environment(Cell *env, Cell *frame)
{
  StackRoot tmp_env(memory_system, env);
  StackRoot tmp_frame(memory_system, frame);

  init_r5rs_environment(tmp_env, tmp_frame);
  tmp_env->bind_syntax(mk_static_syntax("syntax-object",
					&syntax_syntax_object));

  gettext_procedure.assign_procedures(tmp_env, tmp_frame);
  srfi6_procedure.assign_procedures(tmp_env, tmp_frame);
  srfi19_procedure.assign_procedures(tmp_env, tmp_frame);
  srfi28_procedure.assign_procedures(tmp_env, tmp_frame);

  return tscore_procedure.assign_procedures(tmp_env, tmp_frame);
}


Cell*
TSCore::get_toplevel_frame(Cell *env)
{
  if (env == Cell::nil())
    return NULL;

  if (env == interaction_environment)
    return interaction_toplevel_frame;

  if (env == r5rs_environment)
    return get_r5rs_toplevel_frame();

  if (env == null_environment)
    return Cell::nil();

  return NULL;
}


void
TSCore::define_procedure(NativeProcedure *procedure)
{
  external_procedures.push_back(procedure);
  procedure->assign_procedures(get_interaction_environment(),
			       get_interaction_toplevel_frame());
}


bool
TSCore::load_file(FILE *file, bool interactive)
{
  Port *port = new InputFilePort(file);
  Cell *input_port = mk_port(port);
  VirtualMachine *vm = get_idle_virtual_machine();

  const char *prompt = interactive ? "\n>" : NULL;
  bool result = vm->read_eval_print(input_port, get_interaction_output_port(),
				    prompt);
  return_idle_virtual_machine(vm);

  return result;
}


bool
TSCore::load_string(const char *cmd)
{
  Port *port = new InputStringPort(cmd, 0, false);
  Cell *input_port = mk_port(port);
  VirtualMachine *vm = get_idle_virtual_machine();
  bool result = vm->read_eval_print(input_port, get_interaction_output_port(),
				    NULL);
  return_idle_virtual_machine(vm);

  return result;
}


void
TSCore::define(Cell *environment, Cell *frame, Cell *symbol, Cell *value)
{
  StackRoot env(memory_system, environment);
  StackRoot frm(memory_system, frame);
  StackRoot val(memory_system, value);
  Cell *binding = const_cast<Cell *>(env->bind_variable(symbol));
  frm->store_toplevel_variable(env, binding->get_slot_location(), val);
}


Cell *
TSCore::apply(Cell *procedure, Cell *args)
{
  VirtualMachine *vm = get_idle_virtual_machine();
  Cell *result = vm->apply(procedure, args);
  return_idle_virtual_machine(vm);

  return result;
}


Cell *
TSCore::applyf(Cell *procedure, const char *fmt, ...)
{
  Cell *result;
  VirtualMachine *vm;
  va_list ap;

  va_start(ap, fmt);
  vm = get_idle_virtual_machine();
  result = vm->applyf(procedure, fmt, ap);
  va_end(ap);

  return_idle_virtual_machine(vm);

  return result;
}


Cell *
TSCore::evalf0(Cell *symbol, const char *fmt, ...)
{
  Cell *result;
  VirtualMachine *vm;
  va_list ap;

  va_start(ap, fmt);
  vm = get_idle_virtual_machine();
  result = vm->applyf(symbol, fmt, ap);
  va_end(ap);

  if (result == NULL)
    {
      string errstr("Error: ");
      errstr.append(vm->get_error_message()).append("\n");
      output_message(errstr.c_str());
    }

  return_idle_virtual_machine(vm);

  return result;
}


Cell *
TSCore::evalf(const char *name, const char *fmt, ...)
{
  Cell *result;
  VirtualMachine *vm;
  va_list ap;
  va_start(ap, fmt);
  vm = get_idle_virtual_machine();
  result = vm->applyf(mk_symbol(name), fmt, ap);
  va_end(ap);

  return_idle_virtual_machine(vm);

  return result;
}

}
