/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: tscore.h,v 1.50 2009/01/05 12:16:19 fuyu Exp $
 */

#ifndef _LIBTSCORE_TSCORE_H_
#define _LIBTSCORE_TSCORE_H_

#ifdef __cplusplus

#if defined(__GNUG__) && !defined(__ICC)
#pragma interface
#endif

#include "libtscore/cell.h"
#include "libtscore/gettext_procedure.h"
#include "libtscore/memory_system.h"
#include "libtscore/number.h"
#include "libtscore/port.h"
#include "libtscore/procedure.h"
#include "libtscore/r5rs_char_procedure.h"
#include "libtscore/r5rs_io_procedure.h"
#include "libtscore/r5rs_list_procedure.h"
#include "libtscore/r5rs_math_procedure.h"
#include "libtscore/r5rs_misc_procedure.h"
#include "libtscore/r5rs_string_procedure.h"
#include "libtscore/reader.h"
#include "libtscore/srfi6_procedure.h"
#include "libtscore/srfi19_procedure.h"
#include "libtscore/srfi28_procedure.h"
#include "libtscore/symbol_table.h"
#include "libtscore/tscore_procedure.h"
#include "libtscore/writer.h"

#include "libtscore/syntax.h"
#include "libtscore/r5rs_derived_syntax.h"
#include "libtscore/r5rs_macro_syntax.h"
#include "libtscore/r5rs_primitive_syntax.h"
#include "libtscore/tscore_syntax.h"
#include "libtscore/macro_syntax.h"


#include <vector>

#include <cctype>
#include <cstdarg>
#include <cstdio>
#include <cstdlib>
#include <cstring>


using namespace std;
namespace LibTSCore
{
  /*
   * TSCore represents the entire Scheme interpreter system.
   *
   * Usually, users only need to instantiate this and to use public methods
   * provided by this for executing Scheme programs.
   */
  class TSCore
  {
    friend class DefineSyntax;
    friend class Lambda;
    friend class LetSyntax;
    friend class LetrecSyntax;
    friend class MacroSyntax;
    friend class MemorySystem;
    friend class NativeProcedure;
    friend class Quasiquote;
    friend class R5RSListProcedure;
    friend class R5RSMiscProcedure;
    friend class Reader;
    friend class Register;
    friend class StackRoot;
    friend class SymbolTable;
    friend class Syntax;
    friend class SyntaxRules;
    friend class TSCoreProcedure;
    friend class TSStackRoot;
    friend class Unquote;
    friend class UnquoteSplicing;
    friend class VirtualMachine;

  private:
    MemorySystem memory_system;
    SymbolTable symbol_table;

    Define syntax_define;
    DefineSyntax syntax_define_syntax;
    LetSyntax syntax_let_syntax;
    LetrecSyntax syntax_letrec_syntax;
    MacroSyntax syntax_macro_syntax;

    Quote syntax_quote;
    Lambda syntax_lambda;
    If syntax_if;
    Assignment syntax_set;

    And syntax_and;
    Or syntax_or;
    LetAsterisk syntax_let_asterisk;
    Letrec syntax_letrec;

    Begin syntax_begin;

    Delay syntax_delay;

    Quasiquote syntax_quasiquote;

    Unquote syntax_unquote;
    UnquoteSplicing syntax_unquote_splicing;

    SyntaxObject syntax_syntax_object;


    GettextProcedure gettext_procedure;
    R5RSCharProcedure r5rs_char_procedure;
    R5RSIOProcedure r5rs_io_procedure;
    R5RSListProcedure r5rs_list_procedure;
    R5RSMathProcedure r5rs_math_procedure;
    R5RSMiscProcedure r5rs_misc_procedure;
    R5RSStringProcedure r5rs_string_procedure;
    SRFI6Procedure srfi6_procedure;
    SRFI19Procedure srfi19_procedure;
    SRFI28Procedure srfi28_procedure;
    TSCoreProcedure tscore_procedure;


    Cell *interaction_environment;
    Cell *interaction_toplevel_frame;

    Cell *r5rs_environment;
    Cell *r5rs_toplevel_frame;

    Cell *null_environment;


    // constants
    // global special syntax/symbols
    Cell *lambda_symbol;
    Cell *quote_symbol;
    Cell *quasiquote_symbol;
    Cell *unquote_symbol;
    Cell *unquote_splicing_symbol;
    Cell *else_symbol;
    Cell *feed_to_symbol;		// => symbol

    Cell *syntax_rules_symbol;
    Cell *ellipsis_symbol;

    Cell *internal_lambda_symbol;

    Cell *internal_cond_syntax;
    Cell *internal_case_syntax;
    Cell *internal_let_syntax;
    Cell *internal_do_syntax;

    Cell *internal_unquote_syntax;
    Cell *internal_unquote_splicing_syntax;

    Cell *internal_cons_procedure;
    Cell *internal_append_procedure;
    Cell *internal_list2vector_procedure;

    Cell *meta_return;

    Cell *interaction_input_port;
    Cell *interaction_output_port;

    vector<VirtualMachine *> idle_vm;
    vector<NativeProcedure *> external_procedures;

    VirtualMachine *get_idle_virtual_machine();
    void return_idle_virtual_machine(VirtualMachine *vm);

    void output_message(const char *message);

    inline Cell *mk_static_syntax(const char *name, Syntax *syntax)
    {
      Cell *symbol = mk_static_symbol(name);
      Cell *cell = memory_system.get_cell(symbol);
      return cell->mk_static_syntax(symbol, syntax);
    }

    inline Cell *mk_macro_syntax(const Cell *symbol, const Cell *transformer)
    {
      Cell *sym = const_cast<Cell *>(symbol);
      Cell *tr = const_cast<Cell *>(transformer);
      Cell *cell = memory_system.get_cell(sym, tr);
      return cell->mk_static_syntax(sym, &syntax_macro_syntax, tr);
    }

    inline Cell *mk_native_procedure(NativeProcedure *procedure,
				     const NativeProcedure::Info *info,
				     void *data)
    {
      return memory_system.get_cell()->mk_native_procedure(procedure,
							   info, data);
    }
    Cell *mk_static_macro_syntax(const char *keyword,
				 const char *syntax_rules);


    Cell *init_null_environment(Cell *env);
    Cell *get_null_environment();
    inline Cell *get_null_toplevel_frame() const { return Cell::nil(); }

    Cell *init_r5rs_environment(Cell *env, Cell *frame);
    Cell *get_r5rs_environment();
    Cell *get_r5rs_toplevel_frame();

    Cell *init_tscore_environment(Cell *env, Cell *frame);

    Cell *get_toplevel_frame(Cell *env);

    Cell *get_meta_return() const { return meta_return; }

    void define(Cell *environment, Cell *frame, Cell *symbol, Cell *value);

    inline Cell *mk_closure_template(Cell *&cpool, Cell *&code,
				     BytecodeProcedure *bytecode)
    {
      Cell *templ = memory_system.get_cell(cpool, code);
      return templ->mk_closure_template(cpool, code, bytecode);
    }

    inline Cell *mk_closure(Cell *&tmpl, Cell *&frame)
    {
      Cell *closure = memory_system.get_cell(tmpl, frame);
      return closure->mk_closure(tmpl, frame);
    }


  public:
    TSCore(size_t initial_n_cell_segs = MemorySystem::INITIAL_N_CELL_SEGMENTS,
	   size_t maximum_n_cell_segs = MemorySystem::MAXIMUM_N_CELL_SEGMENTS,
	   size_t cell_segment_size = MemorySystem::DEFAULT_CELL_SEGMENT_SIZE);
    ~TSCore();

    inline bool is_out_of_memory() const
    {
      return memory_system.out_of_memory();
    }

    inline bool get_verbose_gc() const
    {
      return memory_system.get_verbose_gc();
    }

    inline void set_verbose_gc(bool verbose)
    {
      memory_system.set_verbose_gc(verbose);
    }

    inline long get_retcode() { return 0; }	// to be removed.

    inline Cell *get_interaction_environment() const
    {
      return interaction_environment;
    }

    inline Cell *get_interaction_toplevel_frame() const
    {
      return interaction_toplevel_frame;
    }

    inline Cell *get_interaction_input_port() { return interaction_input_port; }
    inline Cell *get_interaction_output_port()
    {
      return interaction_output_port;
    }

    inline void set_interaction_output_port_file(FILE *file)
    {
      interaction_output_port = mk_port(new OutputFilePort(file, false));
    }


    inline Cell *mk_static_symbol(const char *name)
    {
      Cell *sym_name = mk_static_string(name);
      return symbol_table.insert(sym_name);
    }

    inline Cell *mk_symbol(const char *name)
    {
      return symbol_table.insert(name);
    }
    inline Cell *mk_symbol(Cell *&name) { return symbol_table.insert(name); }

    inline Cell *mk_integer(long val)
    {
      Cell *x = Cell::mk_small_integer(val);
      return x != NULL ? x : memory_system.get_cell()->mk_integer(val);
    }

    inline Cell *mk_real(double val)
    {
      return memory_system.get_cell()->mk_real(val);
    }

    inline Cell *mk_string(const char *str)
    {
      return memory_system.get_counted_string(str, strlen(str));
    }

    inline Cell *mk_static_string(const char *str)
    {
      return memory_system.get_static_string(str);
    }

    inline Cell *mk_cons(Cell *&a, Cell *&d, bool immutable)
    {
      return memory_system.get_cons(a, d, immutable);
    }

    inline Cell *mk_port(Port *port)
    {
      return memory_system.get_cons()->mk_port(port);
    }

    inline Cell *mk_foreign_object(void *pointer,
				   ForeignObjectDestructFunc destructor = NULL,
				   void *option_data = NULL)
    {
      if (destructor == NULL && option_data == NULL)
	{
	  Cell *cell = Cell::mk_foreign_object(pointer);
	  if (cell != NULL)
	    return cell;
	}

      return memory_system.get_cell()
	->mk_foreign_object(pointer, destructor, option_data);
    }


    inline Cell *list_reverse_in_place(Cell *term, Cell *list,
				bool set_immutable = false)
    {
      return R5RSListProcedure::list_reverse_in_place(memory_system,
						      term, list,
						      set_immutable);
    }

    inline void register_external_root(Cell **root)
    {
      memory_system.register_external_root(root);
    }

    inline void unregister_external_root(Cell **root)
    {
      memory_system.unregister_external_root(root);
    }

    bool load_file(FILE *fin, bool interactive = false);
    bool load_string(const char *cmd);

    Cell *apply(Cell *procedure, Cell *args);
    Cell *applyf(Cell *procedure, const char *fmt, ...);

    Cell *evalf0(Cell *symbol, const char *fmt, ...);
    Cell *evalf(const char *name, const char *fmt, ...);


    inline void define_global(Cell *symbol, Cell *value)
    {
      define(get_interaction_environment(), get_interaction_toplevel_frame(),
	     symbol, value);
    }

    inline void define_global(const char *name, Cell *value)
    {
      StackRoot val(memory_system, value);
      Cell *sym = mk_symbol(name);
      define_global(sym, val);
    }

    void define_procedure(NativeProcedure *procedure);
    inline void define_global_foreign_object(const char *name, void *object,
					     ForeignObjectDestructFunc destructor = NULL,
					     void *option_data = NULL)
    {
      define_global(name, mk_foreign_object(object, destructor, option_data));
    }

    inline void define_global_integer(const char *name, long val)
    {
      define_global(name, mk_integer(val));
    }

    inline void define_global_real(const char *name, double val)
    {
      define_global(name, mk_real(val));
    }

    inline void define_global_string(const char *name, const char *val)
    {
      define_global(name, mk_string(val));
    }

  };


  class TSStackRoot : public StackRoot
  {
  public:
    inline TSStackRoot(TSCore &core, Cell *cell = Cell::nil())
      : StackRoot(core.memory_system, cell) {}
    inline TSStackRoot(TSCore &core, const Cell *cell)
      : StackRoot(core.memory_system, const_cast<Cell *>(cell)) {}

    inline TSStackRoot &operator=(Cell *cell)
    {
      static_cast<StackRoot &>(*this) = cell;
      return *this;
    }

    inline TSStackRoot &operator=(const Cell *cell)
    {
      static_cast<StackRoot &>(*this) = cell;
      return *this;
    }

    inline TSStackRoot &operator=(const TSStackRoot &root)
    {
      static_cast<StackRoot &>(*this) = static_cast<const StackRoot &>(root);
      return *this;
    }

  };


  class CellHandle
  {
  private:
    TSCore &core;
    Cell *cell;


  public:
    inline CellHandle(TSCore &core)
      : core(core), cell(Cell::nil())
    {
      core.register_external_root(&cell);
    }

    inline CellHandle(TSCore &core, Cell *obj)
      : core(core), cell(obj)
    {
      core.register_external_root(&cell);
    }

    inline ~CellHandle() { core.unregister_external_root(&cell); }

    inline Cell *get_cell() const { return cell; }
    inline operator Cell *() { return cell; }
    inline Cell *operator->() { return cell; }
    inline CellHandle &operator=(Cell *obj) { cell = obj; return *this; }
    inline CellHandle &operator=(const CellHandle &ch)
    {
      cell = ch.cell;
      return *this;
    }

    inline bool operator==(CellHandle &ch) const { return cell == ch.cell; }
    inline bool operator==(const CellHandle &ch) const
    {
      return cell == ch.cell;
    }

    inline bool operator==(Cell *obj) const { return cell == obj; }
    inline bool operator==(const Cell *obj) const { return cell == obj; }
  };


}


#endif	/* __cplusplus */


#define _TSCORE_DEFINED_	1
#include "libtscore/virtual_machine.h"
#include "libtscore/syntax.h"


#endif	/* _LIBTSCORE_TSCORE_H_ */
