/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: tscore_procedure.cc,v 1.17 2009/01/01 07:47:01 fuyu Exp $
 */

#include "config.h"

#include <oniguruma.h>

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libtscore/tscore_procedure.h"

#include "libtscore/r5rs_list_procedure.h"
#include "libtscore/tscore.h"
#include "libtscore/virtual_machine.h"
#include "libtscore/writer.h"

#include <cstdio>
#include <climits>

#ifdef HAVE_TIME_H
#include <time.h>
#endif

#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif


namespace LibTSCore
{

NativeProcedure::Info TSCoreProcedure::procedures[] = {
  { &TSCoreProcedure::gc,
    "gc", TST_NONE, 0, 0 },
  { &TSCoreProcedure::gc_verbose,
    "gc-verbose", TST_ANY, 0, 1 },
  { &TSCoreProcedure::gensym,
    "gensym", TST_NONE, 0, 0 },
  { &TSCoreProcedure::get_closure_code,
    "get-closure-code", TST_CLOSURE, 1, 1 },
  { &TSCoreProcedure::is_closure,
    "closure?", TST_ANY, 1, 1 },
  { &TSCoreProcedure::is_defined,
    "defined?", TST_SYMBOL TST_ENVIRONMENT, 1, 2 },
  { &TSCoreProcedure::is_environment,
    "environment?", TST_ANY, 1, 1 },
  { &TSCoreProcedure::oblist,
    "oblist", TST_NONE, 0, 0 },

  { &TSCoreProcedure::set_input_port,
    "set-input-port", TST_INPORT, 1, 1 },
  { &TSCoreProcedure::set_output_port,
    "set-output-port", TST_OUTPORT, 1, 1 },

  { &TSCoreProcedure::tracing,
    "tracing", TST_ANY, 0, 1 },
  { &TSCoreProcedure::bytecompile_tracing,
    "bytecompile-tracing", TST_ANY, 0, 1 },
#ifndef NDEBUG
  { &TSCoreProcedure::bytecode_tracing,
    "bytecode-tracing", TST_ANY, 0, 1 },
#endif
  { &TSCoreProcedure::inspect,
    "inspect", TST_ANY TST_OUTPORT, 1, 2 },
  { &TSCoreProcedure::current_seconds,
    "current-seconds", TST_NONE, 0, 0 },
  { &TSCoreProcedure::current_milliseconds,
    "current-milliseconds", TST_NONE, 0, 0 },
#if 0
  { &TSCoreProcedure::cons_asterisk,
    "cons*", TST_NONE, 0, INF_ARG },
  { &TSCoreProcedure::append_in_place,
    "append!", TST_ANY, 1, INF_ARG },
#endif
  { &TSCoreProcedure::reverse_in_place,
    "reverse!", TST_ANY, 1, 2 },

  { &TSCoreProcedure::is_regexp,
    "regexp?", TST_ANY, 1, 1 },
  { &TSCoreProcedure::string2regexp,
    "string->regexp", TST_STRING TST_ANY, 1, 2 },
  { &TSCoreProcedure::regexp_match_positions,
    "regexp-match-positions", TST_REGEXP TST_STRING TST_INTEGER, 2, 4 },
  { &TSCoreProcedure::regexp_match,
    "regexp-match", TST_REGEXP TST_STRING TST_INTEGER, 2, 4 },
  { &TSCoreProcedure::regexp_split,
    "regexp-split", TST_REGEXP TST_STRING, 2, 2 },
  { &TSCoreProcedure::regexp_replace,
    "regexp-replace", TST_REGEXP TST_STRING, 3, 3 },
  { &TSCoreProcedure::regexp_replace_all,
    "regexp-replace*", TST_REGEXP TST_STRING, 3, 3 },
  { &TSCoreProcedure::regexp_quote,
    "regexp-quote", TST_STRING, 1, 1 }
};


#define CAR0(p)			((p)->car())
#define CDR0(p)			((p)->cdr())


Cell *
TSCoreProcedure::cons_asterisk(MemorySystem &memory_system, Cell *&d)
{
  Cell *p, *q;
  if (CDR(d) == Cell::nil())
    return CAR(d);

  p = memory_system.get_cons(CAR0(d), CDR0(d));
  q = p;
  while (CDDR(p) != Cell::nil())
    {
      d = memory_system.get_cons(CAR0(p), CDR0(p));
      if (CDDR(p) != Cell::nil())
        p = CDR(d);
    }

  memory_system.set_cdr(p, CADR(p));
  return q;
}


Cell *
TSCoreProcedure::assign_procedures(Cell *environment, Cell *frame)
{
  TSStackRoot env(core, environment);
  TSStackRoot frm(core, frame);
  size_t n_procs = sizeof(procedures) / sizeof(Info);
  for (size_t i = 0; i < n_procs; i++)
    assign_procedure(env, frm, &procedures[i], this);

  return env;
}


#define ARG(n)	context.frame->load_variable(argument_pointer + (n))


Cell *
TSCoreProcedure::gc(VirtualMachine &vm, Register &context,
		    unsigned long argument_pointer,
		    unsigned long nargs, void *data)
{
  vm.core.memory_system.gc();
  return Cell::t();
}


Cell *
TSCoreProcedure::gc_verbose(VirtualMachine &vm, Register &context,
			    unsigned long argument_pointer,
			    unsigned long nargs, void *data)
{
  bool was = vm.core.get_verbose_gc();
  if (nargs == 1)
    vm.core.set_verbose_gc(ARG(0) != Cell::f());

  return Cell::ts_bool(was);
}


Cell *
TSCoreProcedure::gensym(VirtualMachine &vm, Register &context,
			unsigned long argument_pointer,
			unsigned long nargs, void *data)
{
  char name[32];
  TSCoreProcedure *proc = static_cast<TSCoreProcedure *>(data);
  while (proc->gensym_count < LONG_MAX)
    {
      snprintf(name, 32, "gensym-%ld", proc->gensym_count++);
      if (vm.core.symbol_table.lookup(name) != Cell::nil())
	continue;

      return vm.core.mk_symbol(name);
    }

  return Cell::nil();
}


Cell *
TSCoreProcedure::get_closure_code(VirtualMachine &vm, Register &context,
				  unsigned long argument_pointer,
				  unsigned long nargs, void *data)
{
  return ARG(0)->get_closure_template();
}


Cell *
TSCoreProcedure::is_closure(VirtualMachine &vm, Register &context,
			    unsigned long argument_pointer,
			    unsigned long nargs, void *data)
{
  return Cell::ts_bool(ARG(0)->is_closure());
}


Cell *
TSCoreProcedure::is_defined(VirtualMachine &vm, Register &context,
			    unsigned long argument_pointer,
			    unsigned long nargs, void *data)
{
  Cell *env;
  if (nargs == 1)
    env = context.toplevel_environment;
  else
    env = ARG(1);

  return Cell::ts_bool(env->lookup(ARG(0), true)->is_binding());
}


Cell *
TSCoreProcedure::is_environment(VirtualMachine &vm, Register &context,
				unsigned long argument_pointer,
				unsigned long nargs, void *data)
{
  return Cell::ts_bool(ARG(0)->is_environment());
}


Cell *
TSCoreProcedure::oblist(VirtualMachine &vm, Register &context,
			unsigned long argument_pointer,
			unsigned long nargs, void *data)
{
  return vm.core.symbol_table.list_all_symbols();
}


Cell *
TSCoreProcedure::set_input_port(VirtualMachine &vm, Register &context,
				unsigned long argument_pointer,
				unsigned long nargs, void *data)
{
  context.input_port = ARG(0);
  return Cell::t();
}


Cell *
TSCoreProcedure::set_output_port(VirtualMachine &vm, Register &context,
				 unsigned long argument_pointer,
				 unsigned long nargs, void *data)
{
  context.output_port = ARG(0);
  return Cell::t();
}


Cell *
TSCoreProcedure::tracing(VirtualMachine &vm, Register &context,
			 unsigned long argument_pointer,
			 unsigned long nargs, void *data)
{
  if (nargs == 1)
    return Cell::ts_bool(vm.set_tracing(ARG(0) != Cell::f()));

  return Cell::ts_bool(vm.get_tracing(context));
}


Cell *
TSCoreProcedure::bytecompile_tracing(VirtualMachine &vm, Register &context,
				     unsigned long argument_pointer,
				     unsigned long nargs, void *data)
{
  if (nargs == 1)
    return Cell::ts_bool(vm.set_bytecompile_tracing(ARG(0) != Cell::f()));

  return Cell::ts_bool(vm.get_bytecompile_tracing(context));
}


#ifndef NDEBUG
Cell *
TSCoreProcedure::bytecode_tracing(VirtualMachine &vm, Register &context,
				  unsigned long argument_pointer,
				  unsigned long nargs, void *data)
{
  if (nargs == 1)
    return Cell::ts_bool(vm.set_bytecode_tracing(ARG(0) != Cell::f()));

  return Cell::ts_bool(vm.get_bytecode_tracing(context));
}
#endif


Cell *
TSCoreProcedure::current_seconds(VirtualMachine &vm, Register &context,
				 unsigned long argument_pointer,
				 unsigned long nargs, void *data)
{
#ifdef HAVE_TIME_H
  time_t t = time(NULL);
  if (t != static_cast<time_t>(-1))
    return vm.mk_integer(static_cast<long>(t));

  return signal_error(vm, "current-seconds: unexpected error: ", Cell::f());
#else
#if defined(HAVE_SYS_TIME_H) && defined(HAVE_GETTIMEOFDAY)
  struct timeval t;
  if (gettimeofday(&t, NULL) == 0)
    return vm.mk_integer(static_cast<long>(t.tv_sec));

  return signal_error(vm, "current-seconds: unexpected error: ", Cell::f());
#else
  return signal_error(vm, "current-seconds: not supported: ", Cell::f());
#endif
#endif
}


Cell *
TSCoreProcedure::current_milliseconds(VirtualMachine &vm, Register &context,
				      unsigned long argument_pointer,
				      unsigned long nargs, void *data)
{
#if defined(HAVE_SYS_TIME_H) && defined(HAVE_GETTIMEOFDAY)
  struct timeval t;
  if (gettimeofday(&t, NULL) == 0)
    return vm.mk_integer(static_cast<long>(t.tv_usec / 1000));

  return signal_error(vm, "current-milliseconds: unexpected error: ",
		      Cell::f());
#else
  return signal_error(vm, "current-milliseconds: not supported: ", Cell::f());
#endif
}


Cell *
TSCoreProcedure::inspect(VirtualMachine &vm, Register &context,
			 unsigned long argument_pointer,
			 unsigned long nargs, void *data)
{
  Port *output_port;
  if (nargs == 1)
    output_port = context.output_port->get_port();
  else
    output_port = ARG(1)->get_port();
#ifndef NDEBUG
  const Cell *result
    = Writer::write_datum(ARG(0), output_port, false, true);
  assert(result == Cell::nil());
#else
  Writer::write_datum(ARG(0), output_port, false, true);
#endif
  return Cell::t();
}


#if 0
Cell *
TSCoreProcedure::cons_asterisk(VirtualMachine &vm, Register &context,
			       unsigned long argument_pointer,
			       unsigned long nargs, void *data)
{
}


Cell *
TSCoreProcedure::append_in_place(VirtualMachine &vm, Register &context,
				 unsigned long argument_pointer,
				 unsigned long nargs, void *data)
{
}
#endif


Cell *
TSCoreProcedure::reverse_in_place(VirtualMachine &vm, Register &context,
				  unsigned long argument_pointer,
				  unsigned long nargs, void *data)
{
  Cell *list = ARG(0);
  if (list == Cell::nil())
    {
      if (nargs == 1)
	return Cell::nil();

      return ARG(1);
    }

  if (list->is_pair())
    {
      Cell *result;
      if (nargs == 1)
	result
	  = R5RSListProcedure::list_reverse_in_place(vm.get_memory_system(),
						     Cell::nil(), list, false);
      else
	result
	  = R5RSListProcedure::list_reverse_in_place(vm.get_memory_system(),
						     ARG(1), list, false);
      if (result != NULL)
	return result;
    }

  return signal_error(vm, "reverse!: argument #1 must be of list: ", list);
}


Cell *
TSCoreProcedure::is_regexp(VirtualMachine &vm, Register &context,
			   unsigned long argument_pointer,
			   unsigned long nargs, void *data)
{
  return Cell::ts_bool(ARG(0)->is_regexp());
}


Cell *
TSCoreProcedure::string2regexp(VirtualMachine &vm, Register &context,
			       unsigned long argument_pointer,
			       unsigned long nargs, void *data)
{
  bool ignore_case = (nargs == 2 && ARG(1) != Cell::f());

  Cell *rx_src = ARG(0);
  return vm.mk_regexp(rx_src, ignore_case);
}


/*
 * Returns an (OnigRegion *) if regexp matches, otherwise returns NULL
 */
OnigRegion *
TSCoreProcedure::regexp_match_region(Cell *regexp, Cell *string,
				     long start_off, long end_off)
{
  long str_len = static_cast<long>(string->get_strlen());
  if (end_off == -1)
    end_off = str_len;

  if (start_off < 0 || start_off >= end_off || end_off > str_len)
    return NULL;

  OnigRegion *region = onig_region_new();
  UChar *str
    = reinterpret_cast<UChar *>(const_cast<char *>(string->get_string()));
  if (onig_search(regexp->get_regexp(),
		  str, str + str_len, str + start_off, str + end_off,
		  region, 0) == ONIG_MISMATCH)
    {
      onig_region_free(region, 1);
      return NULL;
    }

  return region;
}


Cell *
TSCoreProcedure::regexp_match_positions(VirtualMachine &vm, Register &context,
					unsigned long argument_pointer,
					unsigned long nargs, void *data)
{
  Cell *regexp = ARG(0);
  Cell *string = ARG(1);
  long start_off;
  long end_off;
  if (nargs == 4)
    {
      start_off = ARG(2)->get_integer();
      end_off = ARG(3)->get_integer();
    }
  else if (nargs == 3)
    {
      start_off = ARG(2)->get_integer();
      end_off = -1;
    }
  else
    {
      start_off = 0;
      end_off = -1;
    }

  OnigRegion *region = regexp_match_region(regexp, string, start_off, end_off);
  if (region == NULL)
    return Cell::f();

  VMStackRoot result(vm, Cell::nil());
  for (int i = region->num_regs - 1; i >= 0; i--)
    {
      Cell *nil = Cell::nil();
      Cell *val = vm.cons(nil, nil);
      result = vm.cons(val, result);
      val = vm.mk_integer(static_cast<long>(region->beg[i]));
      vm.set_car(CAR(result), val);
      val = vm.mk_integer(static_cast<long>(region->end[i]));
      vm.set_cdr(CAR(result), val);
    }

  onig_region_free(region, 1);
  return result;
}


Cell *
TSCoreProcedure::regexp_match(VirtualMachine &vm, Register &context,
			      unsigned long argument_pointer,
			      unsigned long nargs, void *data)
{
  long start_off;
  long end_off;
  if (nargs == 4)
    {
      start_off = ARG(2)->get_integer();
      end_off = ARG(3)->get_integer();
    }
  else if (nargs == 3)
    {
      start_off = ARG(2)->get_integer();
      end_off = -1;
    }
  else
    {
      start_off = 0;
      end_off = -1;
    }

  Cell *string = ARG(1);
  OnigRegion *region = regexp_match_region(ARG(0), string, start_off, end_off);
  if (region == NULL)
    return Cell::f();

  if (!string->is_immutable())
    string = vm.mk_counted_string(string->get_string(),
				  string->get_strlen())->set_immutable();

  VMStackRoot str(vm, string);
  VMStackRoot result(vm, Cell::nil());
  for (int i = region->num_regs - 1; i >= 0; i--)
    {
      Cell *substr
	= vm.mk_substring_reference(str,
				    static_cast<long>(region->beg[i]),
				    static_cast<long>(region->end[i]));
      result = vm.cons(substr, result);
    }

  onig_region_free(region, 1);
  return result;
}


Cell *
TSCoreProcedure::regexp_split(VirtualMachine &vm, Register &context,
			      unsigned long argument_pointer,
			      unsigned long nargs, void *data)
{
  regex_t *rx = ARG(0)->get_regexp();
  Cell *src_string = ARG(1);
  if (!src_string->is_immutable())
    src_string = vm.mk_counted_string(src_string->get_string(),
				      src_string->get_strlen())
      ->set_immutable();

  UChar *str_start
    = reinterpret_cast<UChar *>(const_cast<char *>(src_string->get_string()));
  UChar *str = str_start;
  UChar *str_tail = str + src_string->get_strlen();
  OnigRegion *region = onig_region_new();

  VMStackRoot tmp_string(vm, src_string);
  VMStackRoot list(vm, Cell::nil());
  while (str < str_tail)
    {
      if (onig_search(rx, str_start, str_tail, str, str_tail, region, 0)
	  != ONIG_MISMATCH)
	{
	  if (region->beg[0] == region->end[0])
	    {
	      // we treat the NUL pattern case specially here to simulate
	      // pregexp's pregexp-split function.
	      onig_region_free(region, 1);
	      return string_split_characters(vm, tmp_string);
	    }

	  // append substring of [str...region->beg[0]) into
	  // the list (of string), and advance str to
	  // str + region->end[0]
	  if (str - str_start < region->beg[0])
	    {
	      Cell *substr = vm.mk_substring_reference(tmp_string,
						       str - str_start,
						       region->beg[0]);
	      list = vm.cons(substr, list);
	    }

	  str = str_start + region->end[0];
	}
      else
	{
	  // append substring(tmp_string, offset) left.
	  Cell *substr = vm.mk_substring_reference(tmp_string, str - str_start,
						   str_tail - str_start);
	  list = vm.cons(substr, list);
	  str = str_tail;
	}

      onig_region_clear(region);
    }

  onig_region_free(region, 1);
  return R5RSListProcedure::list_reverse_in_place(vm.get_memory_system(),
						  Cell::nil(), list, false);
}


const char *
TSCoreProcedure::reflect_match_result(OnigRegion *region,
				      OutputStringPort &string_buf,
				      const char *string,
				      Cell *replacement)
{
  const char *p = replacement->get_string();
  const char *rep_end = p + replacement->get_strlen();
  while (p < rep_end)
    {
      if (*p != '\\')
	string_buf.append(*p++);
      else
	{
	  int reg;
	  if (++p < rep_end && isdigit(*p)
	      && (reg = *p - '0') < region->num_regs)
	    {
	      string_buf.append(string + region->beg[reg],
				region->end[reg] - region->beg[reg]);
	      p++;
	    }
	  else
	    string_buf.append('\\');
	}
    }

  return string_buf.get_output_string();
}


Cell *
TSCoreProcedure::regexp_replace(VirtualMachine &vm, Register &context,
				unsigned long argument_pointer,
				unsigned long nargs, void *data)
{
  Cell *string = ARG(1);
  Cell *replacement = ARG(2);

  UChar *ustr_start
    = reinterpret_cast<UChar *>(const_cast<char *>(string->get_string()));
  UChar *ustr = ustr_start;
  long len = string->get_strlen();
  UChar *ustr_tail = ustr + len;
  OnigRegion *region = onig_region_new();

  if (onig_search(ARG(0)->get_regexp(), ustr_start, ustr_tail, ustr, ustr_tail,
		  region, 0) == ONIG_MISMATCH)
    return string;

  const char *str_start = reinterpret_cast<const char *>(ustr_start);
  OutputStringPort rep_buf;
  const char *rep_str = reflect_match_result(region, rep_buf,
					     str_start, replacement);
  OutputStringPort string_buf;

  if (region->beg[0] > 0)
    string_buf.append(str_start, region->beg[0]);

  if (rep_str != NULL)
    string_buf.append(rep_str);

  if (region->end[0] < len)
    string_buf.append(str_start + region->end[0], len - region->end[0]);

  return vm.mk_counted_string(string_buf.get_output_string(),
			      string_buf.get_output_strlen());
}


Cell *
TSCoreProcedure::regexp_replace_all(VirtualMachine &vm, Register &context,
				    unsigned long argument_pointer,
				    unsigned long nargs, void *data)
{
  Cell *string = ARG(1);

  UChar *ustr_start
    = reinterpret_cast<UChar *>(const_cast<char *>(string->get_string()));
  UChar *ustr = ustr_start;
  long len = string->get_strlen();
  UChar *ustr_tail = ustr + len;
  OnigRegion *region = onig_region_new();
  regex_t *rx = ARG(0)->get_regexp();

  if (onig_search(rx, ustr_start, ustr_tail, ustr, ustr_tail,
		  region, 0) == ONIG_MISMATCH)
    return string;

  const char *str_start = reinterpret_cast<const char *>(ustr_start);
  OutputStringPort rep_buf;
  OutputStringPort string_buf;

  Cell *replacement = ARG(2);
  do
    {
      if (region->beg[0] > 0)
	string_buf.append(reinterpret_cast<const char *>(ustr),
			  region->beg[0] - (ustr - ustr_start));

      rep_buf.clear();
      const char *rep_str = reflect_match_result(region, rep_buf,
						 str_start, replacement);
      if (rep_str != NULL)
	string_buf.append(rep_str);

      ustr = ustr_start + region->end[0];
    }
  while (onig_search(rx, ustr_start, ustr_tail, ustr, ustr_tail, region, 0)
	 != ONIG_MISMATCH);

  if (ustr < ustr_tail)
    string_buf.append(reinterpret_cast<const char *>(ustr), ustr_tail - ustr);

  return vm.mk_counted_string(string_buf.get_output_string(),
			      string_buf.get_output_strlen());
}


static inline long
calc_one_utf8_len(const char *string)
{
  // We don't validate the sequence headed by string is actually UTF-8.
  // It is possible to do here, but it also takes costs.
  unsigned char c = *reinterpret_cast<const unsigned char *>(string);
  if ((c & 0x80) == 0)
    return 1;

  if ((c & 0xe0) == 0xc0)
    return 2;

  if ((c & 0xf0) == 0xe0)
    return 3;

  if ((c & 0xf8) == 0xf0)
    return 4;

  if ((c & 0xfc) == 0xf8)
    return 5;

  if ((c & 0xfe) == 0xfc)
    return 6;

  return 1;	// seems not like a valid UTF-8 sequence.
}


Cell *
TSCoreProcedure::string_split_characters(VirtualMachine &vm, Cell *string)
{
  const char *str_start = string->get_string();
  const char *str = str_start;
  const char *str_tail = str + string->get_strlen();

  VMStackRoot tmp_string(vm, string);
  VMStackRoot list(vm, Cell::nil());
  while (str < str_tail)
    {
      long len = calc_one_utf8_len(str);
      long offset = str - str_start;
      Cell *substr = vm.mk_substring_reference(tmp_string, offset,
					       offset + len);
      list = vm.cons(substr, list);
      str += len;
    }

  return R5RSListProcedure::list_reverse_in_place(vm.get_memory_system(),
						  Cell::nil(), list, false);
}


Cell *
TSCoreProcedure::regexp_quote(VirtualMachine &vm, Register &context,
			      unsigned long argument_pointer,
			      unsigned long nargs, void *data)
{
  Cell *string = ARG(0);
  const char *p = string->get_string();
  const char *q = p + string->get_strlen();
  OutputStringPort string_buf;

  while (p < q)
    {
      if (*p == '\0')
	break;

      if (strchr("\\|[]?+*.(){}$^", *p) != NULL)
	string_buf.append("\\");

      string_buf.append(*p++);
    }

  return vm.mk_counted_string(string_buf.get_output_string(),
			      string_buf.get_output_strlen());
}

}
