/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: virtual_machine.cc,v 1.24 2010/08/15 15:39:47 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libtscore/virtual_machine.h"

#include "libtscore/bytecode_buffer.h"
#include "libtscore/cell.h"
#include "libtscore/tscore.h"
#include "libtscore/r5rs_list_procedure.h"
#include "libtscore/vmopcode.h"

#include <cassert>
#include <stdint.h>


namespace LibTSCore
{

Register::Register(VirtualMachine &vm, Cell *toplevel_env,
		   Cell *proc, Cell *caller_frame,
		   Cell *cpool, Cell *inport, Cell *outport)
  : memory_system(vm.memory_system),
    toplevel_environment(toplevel_env),
    toplevel_frame(Cell::nil()),
    procedure(proc),
    frame(caller_frame),
    constant_pool(cpool),
    input_port(inport),
    output_port(outport),
    value(Cell::unspecified()),
    pc(NULL),
    saved_pc(NULL),
    trampoline_frame(Cell::nil()),
    nargs(0)
{
  assert(toplevel_frame != NULL);

  vm.memory_system.stack_roots.push_back(&toplevel_environment);
  vm.memory_system.stack_roots.push_back(&toplevel_frame);
  vm.memory_system.stack_roots.push_back(&procedure);
  vm.memory_system.stack_roots.push_back(&frame);
  vm.memory_system.stack_roots.push_back(&constant_pool);
  vm.memory_system.stack_roots.push_back(&input_port);
  vm.memory_system.stack_roots.push_back(&output_port);
  vm.memory_system.stack_roots.push_back(&value);
  vm.memory_system.stack_roots.push_back(&trampoline_frame);

  Cell *frame = vm.core.get_toplevel_frame(toplevel_environment);
  toplevel_frame = frame != NULL ? frame : Cell::nil();
}


Register::~Register()
{
  assert(memory_system.stack_roots.back() == &trampoline_frame);

  memory_system.stack_roots.pop_back();
  memory_system.stack_roots.pop_back();
  memory_system.stack_roots.pop_back();
  memory_system.stack_roots.pop_back();
  memory_system.stack_roots.pop_back();
  memory_system.stack_roots.pop_back();
  memory_system.stack_roots.pop_back();
  memory_system.stack_roots.pop_back();
  memory_system.stack_roots.pop_back();
}


inline Cell *
Register::load_toplevel_variable(long slot) const
{
  return toplevel_frame->load_toplevel_variable(slot);
}


inline void
Register::store_toplevel_variable(long slot, Cell *value)
{
  toplevel_frame->store_toplevel_variable(toplevel_environment, slot, value);
}


inline const Cell *
Register::lookup_toplevel_binding(const Cell *symbol) const
{
  return toplevel_environment->lookup(symbol, false);
}


inline const Cell *
Register::bind_toplevel_variable(const Cell *symbol)
{
  return toplevel_environment->bind_variable(symbol);
}


VirtualMachine::VirtualMachine(TSCore &tscore)
  : core(tscore),
    memory_system(tscore.memory_system),
    reader(*this),
    tracing(false),
    bytecompile_tracing(false),
    bytecode_tracing(false)
{
  continuation_cache = Cell::nil();
  memory_system.register_static_root(&continuation_cache);
}


VirtualMachine::~VirtualMachine()
{
  continuation_cache = Cell::nil();
  memory_system.unregister_static_root(&continuation_cache);
}


Cell *
VirtualMachine::unbound_variable_error(const char *name)
{
  error_buffer.clear();
  error_buffer.append("undefined variable: ")->append(name);
  return NULL;
}


// This is only for car/cdr at this time.
Cell *
VirtualMachine::type_mismatch_error(const char *name, Cell *val)
{
  error_buffer.clear();
  error_buffer.append(name);
  error_buffer.append(": argument #1 must be of pair : ");
  Writer::write_datum(val, &error_buffer, false);
  return NULL;
}


Cell *
VirtualMachine::application_error(Cell *val)
{
  error_buffer.clear();
  error_buffer.append("bad procedure: ");
  Writer::write_datum(val, &error_buffer, false);
  return NULL;
}


Cell *
VirtualMachine::mk_input_file_port(const char *name)
{
  FILE *file = fopen(name, "r");
  if (file == NULL)
    return Cell::nil();

  Port *port = new InputFilePort(file, true);
  return mk_port(port);
}


Cell *
VirtualMachine::mk_output_file_port(const char *name)
{
  FILE *file = fopen(name, "w");
  if (file == NULL)
    return Cell::nil();

  Port *port = new OutputFilePort(file, true);
  return mk_port(port);
}


Cell *
VirtualMachine::mk_input_string_port(Cell *&str)
{
  Port *port = new InputStringPort(str->get_string(), str->get_strlen());
  if (port == NULL)
    return Cell::nil();

  return mk_port(port);
}


Cell *
VirtualMachine::mk_output_string_port()
{
  Port *port = new OutputStringPort();
  if (port == NULL)
    return Cell::nil();

  return mk_port(port);
}


Cell *
VirtualMachine::mk_foreign_object(void *pointer,
				  ForeignObjectDestructFunc destructor,
				  void *option_data)
{
  return core.mk_foreign_object(pointer, destructor, option_data);
}


#ifndef WORDS_BIGENDIAN
# if !defined(__i386) && !defined(__i386__) && !defined(_M_IX86) && \
  !(defined(__ppc__) && defined(__APPLE__)) &&			   \
  !defined(__x86_64) && !defined(__x86_64__) &&			   \
  !defined(__mc68020__)
#  define WORDS_BIGENDIAN 1
# endif
#endif


static inline long
read_signed_8(const char *pc)
{
  return static_cast<long>(*pc);
}


static inline long
read_signed_16(const char *pc)
{
#if defined(WORDS_BIGENDIAN) || SIZEOF_SHORT != 2
  long value = static_cast<long>(pc[1]) << 8;
  return value | reinterpret_cast<const unsigned char *>(pc)[0];
#else
  return static_cast<long>(reinterpret_cast<const short *>(pc)[0]);
#endif
}


static inline long
read_signed_32(const char *pc)
{
#if defined(WORDS_BIGENDIAN) || SIZEOF_INT != 4
  const unsigned char *upc = reinterpret_cast<const unsigned char *>(pc);
  return static_cast<long>(pc[3]) << 24
    | static_cast<long>(upc[2]) << 16
    | static_cast<long>(upc[1]) << 8
    | static_cast<long>(upc[0]);
#else
  return static_cast<long>(reinterpret_cast<const int *>(pc)[0]);
#endif
}


static inline long
read_signed_64(const char *pc)
{
#if SIZEOF_LONG >= 8
# if defined(WORDS_BIGENDIAN) || SIZEOF_LONG != 8
  const unsigned char *upc = reinterpret_cast<const unsigned char *>(pc);
  return static_cast<long>(pc[7]) << 56
    | static_cast<long>(upc[6]) << 48
    | static_cast<long>(upc[5]) << 40
    | static_cast<long>(upc[4]) << 32
    | static_cast<long>(upc[3]) << 24
    | static_cast<long>(upc[2]) << 16
    | static_cast<long>(upc[1]) << 8
    | static_cast<long>(upc[0]);
# else
  return static_cast<long>(reinterpret_cast<const long *>(pc)[0]);
# endif
#else
  // We just ignore higher 32 bits of 64 bits integer here.
  return read_signed_32(pc);
#endif
}


static inline unsigned long
read_unsigned_8(const char *pc)
{
  return static_cast<unsigned long>(reinterpret_cast<const unsigned char *>(pc)[0]);
}


static inline unsigned long
read_unsigned_16(const char *pc)
{
#if defined(WORDS_BIGENDIAN) || SIZEOF_SHORT != 2
  const unsigned char *upc = reinterpret_cast<const unsigned char *>(pc);
  return static_cast<unsigned long>(upc[1]) << 8 | static_cast<long>(upc[0]);
#else
  return static_cast<unsigned long>(reinterpret_cast<const short *>(pc)[0]);
#endif
}


static inline unsigned long
read_unsigned_32(const char *pc)
{
#if defined(WORDS_BIGENDIAN) || SIZEOF_INT != 4
  const unsigned char *upc = reinterpret_cast<const unsigned char *>(pc);
  return static_cast<unsigned long>(upc[3]) << 24
    | static_cast<unsigned long>(upc[2]) << 16
    | static_cast<unsigned long>(upc[1]) << 8
    | static_cast<unsigned long>(upc[0]);
#else
  return static_cast<unsigned long>(reinterpret_cast<const unsigned int *>(pc)[0]);
#endif
}


static inline unsigned long
read_unsigned_64(const char *pc)
{
#if SIZEOF_LONG >= 8
# if defined(WORDS_BIGENDIAN) || SIZEOF_LONG != 8
  const unsigned char *upc = reinterpret_cast<const unsigned char *>(pc);
  return static_cast<unsigned long>(upc[7]) << 56
    | static_cast<unsigned long>(upc[6]) << 48
    | static_cast<unsigned long>(upc[5]) << 40
    | static_cast<unsigned long>(upc[4]) << 32
    | static_cast<unsigned long>(upc[3]) << 24
    | static_cast<unsigned long>(upc[2]) << 16
    | static_cast<unsigned long>(upc[1]) << 8
    | static_cast<unsigned long>(upc[0]);
# else
  return static_cast<unsigned long>(reinterpret_cast<const unsigned long *>(pc)[0]);
# endif
#else
  // We just ignore higher 32 bits of 64 bits integer here.
  return read_unsigned_32(pc);
#endif
}


void
VirtualMachine::print_bytecode(Register &context)
{
  if (!context.output_port->is_output_port())
    return;

  Port *port = context.output_port->get_port();
  port->append("=> value: ");
  Writer::write_datum(context.value, port, false);
  port->append('\n');
  OpcodeInfo::print_bytecode(port, context.pc, context.constant_pool);
}


Cell *
VirtualMachine::apply(Cell *procedure, Register &caller_context,
		      unsigned long argument_pointer,
		      unsigned long number_of_arguments)
{
  Register context(*this, caller_context.toplevel_environment,
		   Cell::nil(), caller_context.frame, Cell::nil(),
		   caller_context.input_port, caller_context.output_port);


#ifdef __GNUG__
# define LABEL(label)		label:
# ifdef PIC
#  define LABEL_OFF(label)	(reinterpret_cast<intptr_t>(&&label) - reinterpret_cast<intptr_t>(&&op_iconst_m1))
#  ifdef NDEBUG
#   define GOTO_NEXT_OPCODE() \
  { goto *reinterpret_cast<void *>(reinterpret_cast<intptr_t>(&&op_iconst_m1) + label_offsets[static_cast<long>(*context.pc++)]); }
#  else
#   define GOTO_NEXT_OPCODE()						\
  { if (bytecode_tracing)						\
      print_bytecode(context);						\
    goto *reinterpret_cast<void *>(reinterpret_cast<intptr_t>(&&op_iconst_m1) + label_offsets[static_cast<long>(*context.pc++)]); }
#  endif	// !NDEBUG
  static const ptrdiff_t label_offsets[] = {

# else	// non-PIC

#  define LABEL_OFF(label)	(&&label)
#  ifdef NDEBUG
#   define GOTO_NEXT_OPCODE()	{ goto *const_cast<void *>(labels[static_cast<long>(*context.pc++)]); }
#  else
#   define GOTO_NEXT_OPCODE()						\
  { if (bytecode_tracing)						\
      print_bytecode(context);						\
    goto *const_cast<void *>(labels[static_cast<long>(*context.pc++)]); }
#  endif	// !NDEBUG
  static const void *labels[] = {

# endif

    0,
    LABEL_OFF(op_iconst_m1), LABEL_OFF(op_iconst_0), LABEL_OFF(op_iconst_1),
    LABEL_OFF(op_iconst_2), LABEL_OFF(op_iconst_3), LABEL_OFF(op_iconst_4),

    LABEL_OFF(op_iconst_i1), LABEL_OFF(op_iconst_i2),
    LABEL_OFF(op_iconst_i4), LABEL_OFF(op_iconst_i8),

    LABEL_OFF(op_const_null), LABEL_OFF(op_const_t), LABEL_OFF(op_const_f),
    LABEL_OFF(op_const_unspecified),

    LABEL_OFF(op_const_p1), LABEL_OFF(op_const_p2), LABEL_OFF(op_const_p4),

    LABEL_OFF(op_load_l1), LABEL_OFF(op_load_l2), LABEL_OFF(op_load_l4),

    LABEL_OFF(op_store_l1), LABEL_OFF(op_store_l2), LABEL_OFF(op_store_l4),

    LABEL_OFF(op_load_g1), LABEL_OFF(op_load_g2), LABEL_OFF(op_load_g4),

    LABEL_OFF(op_store_g1), LABEL_OFF(op_store_g2), LABEL_OFF(op_store_g4),

    LABEL_OFF(op_load_t1), LABEL_OFF(op_load_t2), LABEL_OFF(op_load_t4),

    LABEL_OFF(op_store_t1), LABEL_OFF(op_store_t2), LABEL_OFF(op_store_t4),

    LABEL_OFF(op_load_f1), LABEL_OFF(op_load_f2), LABEL_OFF(op_load_f4),

    LABEL_OFF(op_store_f1), LABEL_OFF(op_store_f2), LABEL_OFF(op_store_f4),

    LABEL_OFF(op_if_t1), LABEL_OFF(op_if_t2), LABEL_OFF(op_if_t4),
    LABEL_OFF(op_if_f1), LABEL_OFF(op_if_f2), LABEL_OFF(op_if_f4),
    LABEL_OFF(op_goto1), LABEL_OFF(op_goto2), LABEL_OFF(op_goto4),
    LABEL_OFF(op_cons0),
    LABEL_OFF(op_cons1), LABEL_OFF(op_cons2), LABEL_OFF(op_cons4),
    LABEL_OFF(op_car), LABEL_OFF(op_cdr),
    LABEL_OFF(op_set_car1), LABEL_OFF(op_set_car2), LABEL_OFF(op_set_car4),
    LABEL_OFF(op_set_cdr1), LABEL_OFF(op_set_cdr2), LABEL_OFF(op_set_cdr4),

    LABEL_OFF(op_apply1), LABEL_OFF(op_apply2), LABEL_OFF(op_apply4),
    LABEL_OFF(op_call_cc1), LABEL_OFF(op_call_cc2), LABEL_OFF(op_call_cc4),
    LABEL_OFF(op_apply_continuation0),
    LABEL_OFF(op_lambda), LABEL_OFF(op_promise),
    LABEL_OFF(op_if_forced1), LABEL_OFF(op_if_forced2), LABEL_OFF(op_if_forced4),
    LABEL_OFF(op_force_l1), LABEL_OFF(op_force_l2), LABEL_OFF(op_force_l4),
    LABEL_OFF(op_trampoline),
    LABEL_OFF(op_meta_return),
    LABEL_OFF(op_unbound_f1), LABEL_OFF(op_unbound_f2), LABEL_OFF(op_unbound_f4),
    LABEL_OFF(op_nop)
  };
#else
#define GOTO_NEXT_OPCODE()	opcode = static_cast<TSVMOpcode>(*context.pc++); continue
#define LABEL(label)
#endif


  if (procedure->is_procedure())
    {
      Procedure *invoker = procedure->get_procedure_invoker();
      Cell *result = invoker->call(*this, procedure,
				   context, context.frame,
				   argument_pointer, number_of_arguments);
      if (result == NULL)
	return NULL;	// error
    }
  else
    return application_error(procedure);

#ifndef NDEBUG
  if (bytecode_tracing)
    print_bytecode(context);
#endif

  TSVMOpcode opcode = static_cast<TSVMOpcode>(*context.pc++);

  while (true)
    {
      switch (opcode)
	{
	case OP_ICONST_M1: LABEL(op_iconst_m1);
	  context.value = mk_integer(-1);
	  GOTO_NEXT_OPCODE();


	case OP_ICONST_0: LABEL(op_iconst_0);
	  context.value = mk_integer(0);
	  GOTO_NEXT_OPCODE();


	case OP_ICONST_1: LABEL(op_iconst_1);
	  context.value = mk_integer(1);
	  GOTO_NEXT_OPCODE();


	case OP_ICONST_2: LABEL(op_iconst_2);
	  context.value = mk_integer(2);
	  GOTO_NEXT_OPCODE();


	case OP_ICONST_3: LABEL(op_iconst_3);
	  context.value = mk_integer(3);
	  GOTO_NEXT_OPCODE();


	case OP_ICONST_4: LABEL(op_iconst_4);
	  context.value = mk_integer(4);
	  GOTO_NEXT_OPCODE();


	case OP_ICONST_I1: LABEL(op_iconst_i1);
	  context.value = mk_integer(read_signed_8(context.pc++));
	  GOTO_NEXT_OPCODE();


	case OP_ICONST_I2: LABEL(op_iconst_i2);
	  context.value = mk_integer(read_signed_16(context.pc));
	  context.pc += 2;
	  GOTO_NEXT_OPCODE();


	case OP_ICONST_I4: LABEL(op_iconst_i4);
	  context.value = mk_integer(read_signed_32(context.pc));
	  context.pc += 4;
	  GOTO_NEXT_OPCODE();


	case OP_ICONST_I8: LABEL(op_iconst_i8);
	  context.value = mk_integer(read_signed_64(context.pc));
	  context.pc += 8;
	  GOTO_NEXT_OPCODE();


	case OP_CONST_NULL: LABEL(op_const_null);
	  context.value = Cell::nil();
	  GOTO_NEXT_OPCODE();


	case OP_CONST_T: LABEL(op_const_t);
	  context.value = Cell::t();
	  GOTO_NEXT_OPCODE();


	case OP_CONST_F: LABEL(op_const_f);
	  context.value = Cell::f();
	  GOTO_NEXT_OPCODE();


	case OP_CONST_UNSPECIFIED: LABEL(op_const_unspecified);
	  context.value = Cell::unspecified();
	  GOTO_NEXT_OPCODE();


#define LOAD_CONSTANT(index) \
	  context.value = context.constant_pool->load_constant(index)


	case OP_CONST_P1: LABEL(op_const_p1);
	  LOAD_CONSTANT(read_unsigned_8(context.pc++));
	  GOTO_NEXT_OPCODE();


	case OP_CONST_P2: LABEL(op_const_p2);
	  LOAD_CONSTANT(read_unsigned_16(context.pc));
	  context.pc += 2;
	  GOTO_NEXT_OPCODE();


	case OP_CONST_P4: LABEL(op_const_p4);
	  LOAD_CONSTANT(read_unsigned_32(context.pc));
	  context.pc += 4;
	  GOTO_NEXT_OPCODE();


#define LOAD_LOCAL_VARIABLE(slot) \
  context.value = context.frame->load_variable(slot)


	case OP_LOAD_L1: LABEL(op_load_l1);
	  LOAD_LOCAL_VARIABLE(read_unsigned_8(context.pc++));
	  GOTO_NEXT_OPCODE();


	case OP_LOAD_L2: LABEL(op_load_l2);
	  LOAD_LOCAL_VARIABLE(read_unsigned_16(context.pc));
	  context.pc += 2;
	  GOTO_NEXT_OPCODE();


	case OP_LOAD_L4: LABEL(op_load_l4);
	  LOAD_LOCAL_VARIABLE(read_unsigned_32(context.pc));
	  context.pc += 4;
	  GOTO_NEXT_OPCODE();


#define STORE_LOCAL_VARIABLE(slot) \
  context.frame->store_variable(memory_system, slot, context.value)


	case OP_STORE_L1: LABEL(op_store_l1);
	  STORE_LOCAL_VARIABLE(read_unsigned_8(context.pc++));
	  GOTO_NEXT_OPCODE();


	case OP_STORE_L2: LABEL(op_store_l2);
	  STORE_LOCAL_VARIABLE(read_unsigned_16(context.pc));
	  context.pc += 2;
	  GOTO_NEXT_OPCODE();


	case OP_STORE_L4: LABEL(op_store_l4);
	  STORE_LOCAL_VARIABLE(read_unsigned_32(context.pc));
	  context.pc += 4;
	  GOTO_NEXT_OPCODE();


#define LOAD_GLOBAL_VARIABLE(nesting, slot)		\
  do				    			\
    {							\
      long n = nesting;					\
      Cell *frame = context.frame;			\
      for (long i = 0; i < n; i++)			\
	frame = frame->get_parent_frame();		\
      context.value = frame->load_variable(slot);	\
    } while (0)

	case OP_LOAD_G1: LABEL(op_load_g1);
	  LOAD_GLOBAL_VARIABLE(read_unsigned_8(context.pc),
			       read_unsigned_8(context.pc + 1));
	  context.pc += 2;
	  GOTO_NEXT_OPCODE();


	case OP_LOAD_G2: LABEL(op_load_g2);
	  LOAD_GLOBAL_VARIABLE(read_unsigned_8(context.pc),
			       read_unsigned_16(context.pc + 1));
	  context.pc += 3;
	  GOTO_NEXT_OPCODE();


	case OP_LOAD_G4: LABEL(op_load_g4);
	  LOAD_GLOBAL_VARIABLE(read_unsigned_8(context.pc),
			       read_unsigned_32(context.pc + 1));
	  context.pc += 5;
	  GOTO_NEXT_OPCODE();


#define STORE_GLOBAL_VARIABLE(nesting, slot)				\
  do						    			\
    {									\
      long n = nesting;							\
      Cell *frame = context.frame;					\
      for (long i = 0; i < n; i++)					\
	frame = frame->get_parent_frame();				\
      frame->store_variable(memory_system, slot, context.value);	\
    } while (0)

	case OP_STORE_G1: LABEL(op_store_g1);
	  STORE_GLOBAL_VARIABLE(read_unsigned_8(context.pc),
				read_unsigned_8(context.pc + 1));
	  context.pc += 2;
	  GOTO_NEXT_OPCODE();


	case OP_STORE_G2: LABEL(op_store_g2);
	  STORE_GLOBAL_VARIABLE(read_unsigned_8(context.pc),
				read_unsigned_16(context.pc + 1));
	  context.pc += 3;
	  GOTO_NEXT_OPCODE();


	case OP_STORE_G4: LABEL(op_store_g4);
	  STORE_GLOBAL_VARIABLE(read_unsigned_8(context.pc),
				read_unsigned_32(context.pc + 1));
	  context.pc += 5;
	  GOTO_NEXT_OPCODE();


	case OP_LOAD_T1: LABEL(op_load_t1);
	  context.value
	    = context.load_toplevel_variable(read_unsigned_8(context.pc++));
	  GOTO_NEXT_OPCODE();


	case OP_LOAD_T2: LABEL(op_load_t2);
	  context.value
	    = context.load_toplevel_variable(read_unsigned_16(context.pc));
	  context.pc += 2;
	  GOTO_NEXT_OPCODE();


	case OP_LOAD_T4: LABEL(op_load_t4);
	  context.value
	    = context.load_toplevel_variable(read_unsigned_32(context.pc));
	  context.pc += 4;
	  GOTO_NEXT_OPCODE();


	case OP_STORE_T1: LABEL(op_store_t1);
	  context.store_toplevel_variable(read_unsigned_8(context.pc++),
					  context.value);
	  GOTO_NEXT_OPCODE();


	case OP_STORE_T2: LABEL(op_store_t2);
	  context.store_toplevel_variable(read_unsigned_16(context.pc),
					  context.value);
	  context.pc += 2;
	  GOTO_NEXT_OPCODE();


	case OP_STORE_T4: LABEL(op_store_t4);
	  context.store_toplevel_variable(read_unsigned_32(context.pc),
					  context.value);
	  context.pc += 4;
	  GOTO_NEXT_OPCODE();


	case OP_LOAD_F1: LABEL(op_load_f1);
	  {
	    const Cell *symbol = LOAD_CONSTANT(read_unsigned_8(context.pc++));
	    const Cell *binding = context.lookup_toplevel_binding(symbol);
	    if (binding->is_binding())
	      {
		long slot = binding->get_slot_location();
		if (slot >= 0)
		  {
		    context.value = context.load_toplevel_variable(slot);
		    GOTO_NEXT_OPCODE();
		  }
	      }

	    return unbound_variable_error(symbol->get_symname());
	  }


	case OP_LOAD_F2: LABEL(op_load_f2);
	  {
	    const Cell *symbol = LOAD_CONSTANT(read_unsigned_16(context.pc));
	    context.pc += 2;
	    const Cell *binding = context.lookup_toplevel_binding(symbol);
	    if (binding->is_binding())
	      {
		long slot = binding->get_slot_location();
		if (slot >= 0)
		  {
		    context.value = context.load_toplevel_variable(slot);
		    GOTO_NEXT_OPCODE();
		  }
	      }

	    return unbound_variable_error(symbol->get_symname());
	  }


	case OP_LOAD_F4: LABEL(op_load_f4);
	  {
	    const Cell *symbol = LOAD_CONSTANT(read_unsigned_32(context.pc));
	    context.pc += 4;
	    const Cell *binding = context.lookup_toplevel_binding(symbol);
	    if (binding->is_binding())
	      {
		long slot = binding->get_slot_location();
		if (slot >= 0)
		  {
		    context.value = context.load_toplevel_variable(slot);
		    GOTO_NEXT_OPCODE();
		  }
	      }

	    return unbound_variable_error(symbol->get_symname());
	  }


	case OP_STORE_F1: LABEL(op_store_f1);
	  {
	    const Cell *symbol = LOAD_CONSTANT(read_unsigned_8(context.pc++));
	    const Cell *binding = context.lookup_toplevel_binding(symbol);
	    if (binding->is_binding())
	      {
		long slot = binding->get_slot_location();
		if (slot >= 0)
		  {
		    context.store_toplevel_variable(slot, context.value);
		    GOTO_NEXT_OPCODE();
		  }
	      }

	    return unbound_variable_error(symbol->get_symname());
	  }


	case OP_STORE_F2: LABEL(op_store_f2);
	  {
	    const Cell *symbol = LOAD_CONSTANT(read_unsigned_16(context.pc));
	    context.pc += 2;
	    const Cell *binding = context.lookup_toplevel_binding(symbol);
	    if (binding->is_binding())
	      {
		long slot = binding->get_slot_location();
		if (slot >= 0)
		  {
		    context.store_toplevel_variable(slot, context.value);
		    GOTO_NEXT_OPCODE();
		  }
	      }

	    return unbound_variable_error(symbol->get_symname());
	  }


	case OP_STORE_F4: LABEL(op_store_f4);
	  {
	    const Cell *symbol = LOAD_CONSTANT(read_unsigned_32(context.pc));
	    context.pc += 4;
	    const Cell *binding = context.lookup_toplevel_binding(symbol);
	    if (binding->is_binding())
	      {
		long slot = binding->get_slot_location();
		if (slot >= 0)
		  {
		    context.store_toplevel_variable(slot, context.value);
		    GOTO_NEXT_OPCODE();
		  }
	      }

	    return unbound_variable_error(symbol->get_symname());
	  }


	case OP_IF_T1: LABEL(op_if_t1);
	  if (context.value != Cell::f())
	    context.pc += read_signed_8(context.pc);
	  else
	    ++context.pc;

	  GOTO_NEXT_OPCODE();


	case OP_IF_T2: LABEL(op_if_t2);
	  if (context.value != Cell::f())
	    context.pc += read_signed_16(context.pc);
	  else
	    context.pc += 2;

	  GOTO_NEXT_OPCODE();


	case OP_IF_T4: LABEL(op_if_t4);
	  if (context.value != Cell::f())
	    context.pc += read_signed_32(context.pc);
	  else
	    context.pc += 4;

	  GOTO_NEXT_OPCODE();


	case OP_IF_F1: LABEL(op_if_f1);
	  if (context.value == Cell::f())
	    context.pc += read_signed_8(context.pc);
	  else
	    ++context.pc;

	  GOTO_NEXT_OPCODE();


	case OP_IF_F2: LABEL(op_if_f2);
	  if (context.value == Cell::f())
	    context.pc += read_signed_16(context.pc);
	  else
	    context.pc += 2;

	  GOTO_NEXT_OPCODE();


	case OP_IF_F4: LABEL(op_if_f4);
	  if (context.value == Cell::f())
	    context.pc += read_signed_32(context.pc);
	  else
	    context.pc += 4;

	  GOTO_NEXT_OPCODE();


	case OP_GOTO1: LABEL(op_goto1);
	  context.pc += read_signed_8(context.pc);
	  GOTO_NEXT_OPCODE();


	case OP_GOTO2: LABEL(op_goto2);
	  context.pc += read_signed_16(context.pc);
	  GOTO_NEXT_OPCODE();


	case OP_GOTO4: LABEL(op_goto4);
	  context.pc += read_signed_32(context.pc);
	  GOTO_NEXT_OPCODE();


	case OP_CONS0: LABEL(op_cons0);
	  context.value = cons(Cell::nil0(), Cell::nil0());
	  GOTO_NEXT_OPCODE();


	case OP_CONS1: LABEL(op_cons1);
	  {
	    unsigned long car_idx = read_unsigned_8(context.pc);
	    unsigned long cdr_idx = read_unsigned_8(context.pc + 1);
	    context.pc += 2;
	    Cell *car_value = context.frame->load_variable(car_idx);
	    Cell *cdr_value = context.frame->load_variable(cdr_idx);
	    context.value = cons(car_value, cdr_value);
	  }
	  GOTO_NEXT_OPCODE();


	case OP_CONS2: LABEL(op_cons2);
	  {
	    unsigned long car_idx = read_unsigned_16(context.pc);
	    unsigned long cdr_idx = read_unsigned_16(context.pc + 2);
	    context.pc += 4;
	    Cell *car_value = context.frame->load_variable(car_idx);
	    Cell *cdr_value = context.frame->load_variable(cdr_idx);
	    context.value = cons(car_value, cdr_value);
	  }
	  GOTO_NEXT_OPCODE();


	case OP_CONS4: LABEL(op_cons4);
	  {
	    unsigned long car_idx = read_unsigned_32(context.pc);
	    unsigned long cdr_idx = read_unsigned_32(context.pc + 4);
	    context.pc += 8;
	    Cell *car_value = context.frame->load_variable(car_idx);
	    Cell *cdr_value = context.frame->load_variable(cdr_idx);
	    context.value = cons(car_value, cdr_value);
	  }
	  GOTO_NEXT_OPCODE();


	case OP_CAR: LABEL(op_car);
	  if (context.value->is_pair())
	    context.value = CAR(context.value);
	  else
	    return type_mismatch_error("car", context.value);

	  GOTO_NEXT_OPCODE();


	case OP_CDR: LABEL(op_cdr);
	  if (context.value->is_pair())
	    context.value = CDR(context.value);
	  else
	    return type_mismatch_error("cdr", context.value);

	  GOTO_NEXT_OPCODE();


	case OP_SET_CAR1: LABEL(op_set_car1);
	  {
	    unsigned long pair_idx = read_unsigned_8(context.pc++);
	    Cell *pair = context.frame->load_variable(pair_idx);
	    assert(pair->is_pair() && !pair->is_immutable());

	    set_car(pair, context.value);
	  }
	  GOTO_NEXT_OPCODE();


	case OP_SET_CAR2: LABEL(op_set_car2);
	  {
	    unsigned long pair_idx = read_unsigned_16(context.pc);
	    context.pc += 2;
	    Cell *pair = context.frame->load_variable(pair_idx);
	    assert(pair->is_pair() && !pair->is_immutable());

	    set_car(pair, context.value);
	  }
	  GOTO_NEXT_OPCODE();


	case OP_SET_CAR4: LABEL(op_set_car4);
	  {
	    unsigned long pair_idx = read_unsigned_32(context.pc);
	    context.pc += 4;
	    Cell *pair = context.frame->load_variable(pair_idx);
	    assert(pair->is_pair() && !pair->is_immutable());

	    set_car(pair, context.value);
	  }
	  GOTO_NEXT_OPCODE();


	case OP_SET_CDR1: LABEL(op_set_cdr1);
	  {
	    unsigned long pair_idx = read_unsigned_8(context.pc++);
	    Cell *pair = context.frame->load_variable(pair_idx);
	    assert(pair->is_pair() && !pair->is_immutable());

	    set_cdr(pair, context.value);
	  }
	  GOTO_NEXT_OPCODE();


	case OP_SET_CDR2: LABEL(op_set_cdr2);
	  {
	    unsigned long pair_idx = read_unsigned_16(context.pc);
	    context.pc += 2;
	    Cell *pair = context.frame->load_variable(pair_idx);
	    assert(pair->is_pair() && !pair->is_immutable());

	    set_cdr(pair, context.value);
	  }
	  GOTO_NEXT_OPCODE();


	case OP_SET_CDR4: LABEL(op_set_cdr4);
	  {
	    unsigned long pair_idx = read_unsigned_32(context.pc);
	    context.pc += 4;
	    Cell *pair = context.frame->load_variable(pair_idx);
	    assert(pair->is_pair() && !pair->is_immutable());

	    set_cdr(pair, context.value);
	  }
	  GOTO_NEXT_OPCODE();


#define COPY_CONTINUATION(slot)						\
  context.frame->store_variable(memory_system, slot,			\
				context.frame->load_variable(0))

	case OP_APPLY1: LABEL(op_apply1);
	  if (context.value->is_procedure())
	    {
	      unsigned long args = read_unsigned_8(context.pc);
	      unsigned long nargs = read_unsigned_8(context.pc + 1);
	      context.pc += 2;
	      COPY_CONTINUATION(args);
	      Cell *proc = context.value;
	      Procedure *invoker = proc->get_procedure_invoker();
	      Cell *result = invoker->call(*this, proc, context,
					   context.frame, args, nargs);
	      if (result != NULL)
		GOTO_NEXT_OPCODE();

	      return NULL;	// error
	    }

	  return application_error(context.value);


	case OP_APPLY2: LABEL(op_apply2);
	  if (context.value->is_procedure())
	    {
	      unsigned long args = read_unsigned_16(context.pc);
	      unsigned long nargs = read_unsigned_16(context.pc + 2);
	      context.pc += 4;
	      COPY_CONTINUATION(args);
	      Cell *proc = context.value;
	      Procedure *invoker = proc->get_procedure_invoker();
	      Cell *result = invoker->call(*this, proc, context,
					   context.frame, args, nargs);
	      if (result != NULL)
		GOTO_NEXT_OPCODE();

	      return NULL;	// error
	    }

	  return application_error(context.value);


	case OP_APPLY4: LABEL(op_apply4);
	  if (context.value->is_procedure())
	    {
	      unsigned long args = read_unsigned_32(context.pc);
	      unsigned long nargs = read_unsigned_32(context.pc + 4);
	      context.pc += 8;
	      COPY_CONTINUATION(args);
	      Cell *proc = context.value;
	      Procedure *invoker = proc->get_procedure_invoker();
	      Cell *result = invoker->call(*this, proc, context,
					   context.frame, args, nargs);
	      if (result != NULL)
		GOTO_NEXT_OPCODE();

	      return NULL;	// error
	    }

	  return application_error(context.value);


#define SET_CURRENT_CONTINUATION(slot)	\
  context.frame->store_variable(memory_system, slot, Cell::nil())

	case OP_CALL_CC1: LABEL(op_call_cc1);
	  if (context.value->is_procedure())
	    {
	      unsigned long args = read_unsigned_8(context.pc);
	      unsigned long nargs = read_unsigned_8(context.pc + 1);
	      context.pc += 2;
	      SET_CURRENT_CONTINUATION(args);
	      Cell *proc = context.value;
	      Procedure *invoker = proc->get_procedure_invoker();
	      Cell *result = invoker->call(*this, proc, context,
					   context.frame, args, nargs);
	      if (result != NULL)
		GOTO_NEXT_OPCODE();

	      return NULL;	// error
	    }

	  return application_error(context.value);


	case OP_CALL_CC2: LABEL(op_call_cc2);
	  if (context.value->is_procedure())
	    {
	      unsigned long args = read_unsigned_16(context.pc);
	      unsigned long nargs = read_unsigned_16(context.pc + 2);
	      context.pc += 4;
	      SET_CURRENT_CONTINUATION(args);
	      Cell *proc = context.value;
	      Procedure *invoker = proc->get_procedure_invoker();
	      Cell *result = invoker->call(*this, proc, context,
					   context.frame, args, nargs);
	      if (result != NULL)
		GOTO_NEXT_OPCODE();

	      return NULL;	// error
	    }

	  return application_error(context.value);


	case OP_CALL_CC4: LABEL(op_call_cc4);
	  if (context.value->is_procedure())
	    {
	      unsigned long args = read_unsigned_32(context.pc);
	      unsigned long nargs = read_unsigned_32(context.pc + 4);
	      context.pc += 8;
	      SET_CURRENT_CONTINUATION(args);
	      Cell *proc = context.value;
	      Procedure *invoker = proc->get_procedure_invoker();
	      Cell *result = invoker->call(*this, proc, context,
					   context.frame, args, nargs);
	      if (result != NULL)
		GOTO_NEXT_OPCODE();

	      return NULL;	// error
	    }

	  return application_error(context.value);


	case OP_APPLY_CONTINUATION0: LABEL(op_apply_continuation0);
	  {
	    Cell *continuation = context.frame->load_variable(0);
	    continuation->get_continuation()->apply(*this,
						    continuation, context,
						    context.value);
	    GOTO_NEXT_OPCODE();
	  }


	case OP_LAMBDA: LABEL(op_lambda);
	  context.value = mk_closure(context.value, context.frame);
	  GOTO_NEXT_OPCODE();


	case OP_PROMISE: LABEL(op_promise);
	  context.value = mk_promise(context.value);
	  GOTO_NEXT_OPCODE();


	case OP_IF_FORCED1: LABEL(op_if_forced1);
	  if (context.value->is_promise())
	    {
	      if (context.value->is_forced_promise())
		{
		  context.value = context.value->get_memoized_value();
		  context.pc += read_signed_8(context.pc);
		}
	      else
		{
		  context.value = context.value->get_promise_procedure();
		  ++context.pc;
		}
	    }
	  else
	    context.pc += read_signed_8(context.pc);

	  GOTO_NEXT_OPCODE();


	case OP_IF_FORCED2: LABEL(op_if_forced2);
	  if (context.value->is_promise())
	    {
	      if (context.value->is_forced_promise())
		{
		  context.value = context.value->get_memoized_value();
		  context.pc += read_signed_16(context.pc);
		}
	      else
		{
		  context.value = context.value->get_promise_procedure();
		  context.pc += 2;
		}
	    }
	  else
	    context.pc += read_signed_16(context.pc);

	  GOTO_NEXT_OPCODE();


	case OP_IF_FORCED4: LABEL(op_if_forced4);
	  if (context.value->is_promise())
	    {
	      if (context.value->is_forced_promise())
		{
		  context.value = context.value->get_memoized_value();
		  context.pc += read_signed_32(context.pc);
		}
	      else
		{
		  context.value = context.value->get_promise_procedure();
		  context.pc += 4;
		}
	    }
	  else
	    context.pc += read_signed_32(context.pc);

	  GOTO_NEXT_OPCODE();


#define MEMOIZE_LOCAL_VARIABLE(slot) \
  do									\
    {									\
      Cell *promise = context.frame->load_variable(slot);		\
      promise->memoize_promise_value(memory_system, context.value);	\
    }									\
  while (0)

	case OP_FORCE_L1: LABEL(op_force_l1);
	  MEMOIZE_LOCAL_VARIABLE(read_unsigned_8(context.pc++));
	  GOTO_NEXT_OPCODE();


	case OP_FORCE_L2: LABEL(op_force_l2);
	  MEMOIZE_LOCAL_VARIABLE(read_unsigned_16(context.pc));
	  context.pc += 2;
	  GOTO_NEXT_OPCODE();


	case OP_FORCE_L4: LABEL(op_force_l4);
	  MEMOIZE_LOCAL_VARIABLE(read_unsigned_32(context.pc));
	  context.pc += 4;
	  GOTO_NEXT_OPCODE();


	case OP_TRAMPOLINE: LABEL(op_trampoline);
	  {
	    // When you change the logic around here,
	    // you must also change `apply' and `call/cc'.
	    assert(context.trampoline_frame->is_scheme_frame());

	    context.pc = context.saved_pc;
	    context.saved_pc = NULL;
	    Cell *proc = context.value;
	    Procedure *invoker = proc->get_procedure_invoker();
	    Cell *result = invoker->call(*this, proc, context,
					 context.trampoline_frame, 0,
					 context.nargs);
	    if (result != NULL)
	      GOTO_NEXT_OPCODE();

	    return NULL;	// error
	  }


	case OP_META_RETURN: LABEL(op_meta_return);
	  return context.value;


	case OP_UNBOUND_F1: LABEL(op_unbound_f1);
	  {
	    const Cell *symbol = LOAD_CONSTANT(read_unsigned_8(context.pc));
	    return unbound_variable_error(symbol->get_symname());
	  }


	case OP_UNBOUND_F2: LABEL(op_unbound_f2);
	  {
	    const Cell *symbol = LOAD_CONSTANT(read_unsigned_16(context.pc));
	    return unbound_variable_error(symbol->get_symname());
	  }


	case OP_UNBOUND_F4: LABEL(op_unbound_f4);
	  {
	    const Cell *symbol = LOAD_CONSTANT(read_unsigned_16(context.pc));
	    return unbound_variable_error(symbol->get_symname());
	  }


	case OP_NOP: LABEL(op_nop);
	  // do nothing
	  GOTO_NEXT_OPCODE();


	default:
	  // never reached here!
	  abort();
	}
    }
}


Cell *
VirtualMachine::eval(Register &context, Cell *expression, Cell *environment)
{
  error_buffer.clear();

  StackRoot exp(memory_system, mk_cons(expression, Cell::nil0(), false));
  StackRoot cpool(memory_system, Cell::constant_pool_new(memory_system));
  Cell *env = environment_new(environment);
  env->allocate_anonymous_slot();	// dummy
  Cell *closure_template
    = Lambda::closure_template_new(*this, context, exp, env,
				   cpool,
				   Cell::nil(), Cell::nil(), exp);
  if (closure_template == Cell::nil())
    return NULL;	// bytecompilation error

  context.frame = frame_new(Cell::nil(), 28)->set_frame_closed();
  context.frame->store_variable(memory_system, 0, core.get_meta_return());
  Cell *procedure = mk_closure(closure_template, Cell::nil0());

  return apply(procedure, context, 0, 1);
}


bool
VirtualMachine::read_eval_print(Cell *input_port, Cell *output_port,
				const char *prompt)
{
  Register repl_context(*this, core.get_interaction_environment(),
			Cell::nil(), Cell::nil(), Cell::nil(),
			input_port, output_port);
  if (!input_port->is_input_port())
    {
      Port *outport = NULL;
      if (output_port != NULL && output_port->is_output_port())
	outport = output_port->get_port();

      if (outport != NULL)
	outport->append("Invalid input port\n");
      else
	fprintf(stderr, "Invalid input port\n");

      return false;
    }

  Port *inport = input_port->get_port();
  if (prompt != NULL && output_port->is_output_port())
    {
      // interactive repl
      Port *outport = output_port->get_port();
      while (true)
	{
	  outport->append(prompt);

	  Cell *expression = reader.read_datum(inport);
	  const char *error_message = NULL;
	  if (expression != NULL)
	    {
	      Cell *result = eval(repl_context, expression,
				  core.get_interaction_environment());
	      if (result != NULL)
		{
		  Writer::write_datum(result, outport, true, true);
		  if (result == Cell::eof())
		    return true;

		  continue;
		}

	      error_message = error_buffer.get_output_string();
	    }
	  else
	    error_message = reader.get_error_message();

	  outport->append("Error: ")->append(error_message);
	  inport->flush_remaining_input();
	}
    }

  // non interactive repl
  while (true)
    {
      Cell *expression = reader.read_datum(inport);
      if (expression != NULL)
	{
	  Cell *result = eval(repl_context, expression,
			      core.get_interaction_environment());
	  if (result != NULL)
	    {
	      if (result == Cell::eof())
		return true;

	      continue;
	    }
	}

      return false;	// error
    }
}


Cell *
VirtualMachine::apply(Cell *procedure, Cell *args)
{
  if (args == NULL || !R5RSListProcedure::is_list(args))
    {
      error_buffer.clear();
      error_buffer.append("args must be of list: ");
      if (args != NULL)
	Writer::write_datum(args, &error_buffer, false);
      else
	error_buffer.append("NULL");

      return NULL;
    }

  Register context(*this, core.get_interaction_environment(),
		   Cell::nil(), Cell::nil(), Cell::nil(),
		   core.get_interaction_input_port(),
		   core.get_interaction_output_port());

  if (procedure->is_symbol())
    {
      const Cell *binding = context.lookup_toplevel_binding(procedure);
      if (!binding->is_binding())
	{
	  error_buffer.clear();
	  error_buffer.append("unbound variable: ");
	  error_buffer.append(procedure->get_symname());
	  return NULL;
	}

      procedure = context.load_toplevel_variable(binding->get_slot_location());
    }

  if (!procedure->is_procedure())
    {
      error_buffer.clear();
      error_buffer.append("cannot apply: ");
      Writer::write_datum(procedure, &error_buffer, false);
      return NULL;
    }

  StackRoot proc(memory_system, procedure);

  long frame_size = R5RSListProcedure::list_length(args) + 1;
  if (frame_size < 28)
    frame_size = 28;

  StackRoot temporary_frame(memory_system,
			    frame_new(Cell::nil(), frame_size)
			    ->set_frame_closed());
  temporary_frame->store_variable(memory_system, 0, core.get_meta_return());

  long nargs = 1;
  while (args->is_pair())
    {
      temporary_frame->store_variable(memory_system, nargs++, CAR(args));
      args = CDR(args);
    }

  context.frame = temporary_frame;
  return apply(proc, context, 0, nargs);
}


Cell *
VirtualMachine::applyf(Cell *procedure, const char *fmt, va_list ap)
{
  error_buffer.clear();

  Register context(*this, core.get_interaction_environment(),
		   Cell::nil(), Cell::nil(), Cell::nil(),
		   core.get_interaction_input_port(),
		   core.get_interaction_output_port());

  if (procedure->is_symbol())
    {
      const Cell *binding = context.lookup_toplevel_binding(procedure);
      if (!binding->is_binding())
	{
	  error_buffer.append("unbound variable: ");
	  error_buffer.append(procedure->get_symname());
	  return NULL;
	}

      procedure = context.load_toplevel_variable(binding->get_slot_location());
    }

  if (!procedure->is_procedure())
    {
      error_buffer.append("cannot apply: ");
      Writer::write_datum(procedure, &error_buffer, false);
      return NULL;
    }

  StackRoot proc(memory_system, procedure);
  StackRoot temporary_frame(memory_system,
			    frame_new(Cell::nil(), 28)->set_frame_closed());
  temporary_frame->store_variable(memory_system, 0, core.get_meta_return());

  long nargs = 1;
  long frame_size = temporary_frame->get_frame_size();
  char error_buf[256];
  if (fmt != NULL)
    {
      bool post_percent_state = false;
      bool long_argument = false;

      while (*fmt != '\0')
	{
	  if (frame_size <= nargs)
	    {
	      snprintf(error_buf, 256,
		      "Too many arguments for this VirtualMachine: %ld\n",
		      nargs);
	      error_buffer.append(error_buf);
	      return NULL;
	    }
	  switch (*fmt)
	    {
	    case '%':
	      post_percent_state = !post_percent_state;
	      break;

	    case 'l':
	      if (!post_percent_state)
		break;

	      long_argument = true;
	      break;

	    case 'd':
	      if (!post_percent_state)
		break;

	      if (long_argument)
		temporary_frame->store_variable(memory_system, nargs++,
						mk_integer(va_arg(ap, long)));
	      else
		temporary_frame->store_variable(memory_system, nargs++,
				mk_integer(static_cast<long>(va_arg(ap, int))));

	      post_percent_state = false;
	      long_argument = false;
	      break;

	    case 'f':
	      if (!post_percent_state)
		break;

	      temporary_frame->store_variable(memory_system, nargs++,
					      mk_real(va_arg(ap, double)));
	      post_percent_state = false;
	      long_argument = false;
	      break;

	    case 'c':
	      if (!post_percent_state)
		break;

	      temporary_frame->store_variable(memory_system, nargs++,
					      mk_character(va_arg(ap, int)));
	      post_percent_state = false;
	      long_argument = false;
	      break;

	    case 'p':
	      if (!post_percent_state)
		break;

	      temporary_frame
		->store_variable(memory_system, nargs++,
				 mk_foreign_object(va_arg(ap, char *)));
	      post_percent_state = false;
	      long_argument = false;
	      break;

	    case 's':
	      if (!post_percent_state)
		break;

	      temporary_frame->store_variable(memory_system, nargs++,
					      mk_static_string(va_arg(ap,
								      char *)));
	      post_percent_state = false;
	      long_argument = false;
	      break;

	    default:
	      if (post_percent_state)
		snprintf(error_buf, 256, "unknown format '%%%c'\n", *fmt);
	      else
		snprintf(error_buf, 256,
			 "not a format specifier '%c' ignored\n", *fmt);

	      error_buffer.append(error_buf);
	      post_percent_state = false;
	      long_argument = false;
	      break;
	    }
	  fmt++;
	}
    }

  context.frame = temporary_frame;
  return apply(proc, context, 0, nargs);
}

}
