/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: virtual_machine.h,v 1.25 2009/01/02 03:07:32 fuyu Exp $
 */

#ifndef _LIBTSCORE_VIRTUAL_MACHINE_H_
#define _LIBTSCORE_VIRTUAL_MACHINE_H_

#ifdef __cplusplus

#if defined(__GNUG__) && !defined(__ICC)
#pragma interface
#endif

#include "libtscore/cell.h"
#include "libtscore/port.h"
#include "libtscore/reader.h"
#include "libtscore/tscore.h"
#include "libtscore/writer.h"

#include <vector>

#include <cassert>
#include <cstdarg>


using namespace std;
namespace LibTSCore
{

  class Cell;
  class MemorySystem;
  class TSCore;
  class VirtualMachine;

  class Register
  {
  private:
    MemorySystem &memory_system;


  public:
    Cell *toplevel_environment;
    Cell *toplevel_frame;
    Cell *procedure;
    Cell *frame;
    Cell *constant_pool;
    Cell *input_port;
    Cell *output_port;
    Cell *value;
    const char *pc;

    const char *saved_pc;
    Cell *trampoline_frame;
    long nargs;

    Register(VirtualMachine &vm, Cell *toplevel_environment,
	     Cell *procedure, Cell *frame, Cell *constant_pool,
	     Cell *input_port, Cell *output_port);
    ~Register();

    Cell *load_toplevel_variable(long slot) const;
    void store_toplevel_variable(long slot, Cell *value);
    const Cell *lookup_toplevel_binding(const Cell *symbol) const;
    const Cell *bind_toplevel_variable(const Cell *symbol);
  };


  class VirtualMachine
  {
    friend class BytecodeProcedure;
    friend class Continuation;
    friend class Define;
    friend class DefineSyntax;
    friend class Lambda;
    friend class LetSyntax;
    friend class LetrecSyntax;
    friend class MacroSyntax;
    friend class NativeProcedure;
    friend class Procedure;
    friend class Quasiquote;
    friend class R5RSIOProcedure;
    friend class R5RSMiscProcedure;
    friend class Reader;
    friend class Register;
    friend class Syntax;
    friend class SyntaxRules;
    friend class TSCore;
    friend class TSCoreProcedure;
    friend class Unquote;
    friend class UnquoteSplicing;
    friend class VMStackRoot;

  private:
    TSCore &core;
    MemorySystem &memory_system;
    Reader reader;
    OutputStringPort error_buffer;

    Cell *continuation_cache;

    bool tracing;
    bool bytecompile_tracing;
    bool bytecode_tracing;


    inline Cell *environment_new(Cell *parent)
    { return Cell::closed_environment_new(memory_system, parent); }
    inline Cell *frame_new(Cell *parent, long size)
    { return Cell::scheme_frame_new(memory_system, parent, size); }

    inline Cell *mk_current_continuation(Register &context)
    {
      Cell *continuation = continuation_cache;
      continuation_cache = Cell::nil();
      if (continuation == Cell::nil())
	continuation = memory_system.get_cell();
      else
	memory_system.write_barrier(continuation);

      continuation->mk_continuation(context.procedure, context.frame,
				    context.pc);
      return continuation;
    }
    inline void recycle_continuation(Cell *continuation)
    { continuation_cache = continuation; }

    inline Cell *mk_current_continuation_full(Register &context)
    {
      return Cell::continuation_new(memory_system,
				    context.procedure, context.frame,
				    context.pc,
				    0,	// reserved for `dynamic-wind'
				    context.input_port, context.output_port);
    }

    Cell *unbound_variable_error(const char *name);
    // This is only for car/cdr at this time.
    Cell *type_mismatch_error(const char *name, Cell *val);
    Cell *application_error(Cell *val);

    inline OutputStringPort *get_error_buffer() { return &error_buffer; }

    Cell *apply(Cell *procedure, Register &caller_context,
		unsigned long args_pointer, unsigned long nargs);

    void print_bytecode(Register &state);


  public:
    VirtualMachine(TSCore &tscore);
    ~VirtualMachine();

    Cell *eval(Register &context, Cell *expression, Cell *environment);

    bool read_eval_print(Cell *input_port, Cell *output_port,
			 const char *prompt);

    inline const char *get_error_message() const
    { return error_buffer.get_output_string(); }

    inline Reader *get_reader() { return &reader; }


    inline Cell *mk_cons(Cell *&a, Cell *&d, bool immutable)
    { return memory_system.get_cons(a, d, immutable); }
    inline Cell *set_car(Cell *pair, Cell *value)
    { return memory_system.set_car(pair, value); }
    inline Cell *set_cdr(Cell *pair, Cell *value)
    { return memory_system.set_cdr(pair, value); }
    inline Cell *set_vector_element(Cell *vec, long index, Cell *elm)
    { return memory_system.set_vector_element(vec, index, elm); }

    inline Cell *mk_integer(long val)
    {
      Cell *x = Cell::mk_small_integer(val);
      if (x != NULL)
	return x;

      return memory_system.get_cell()->mk_integer(val);
    }

    inline Cell *mk_real(double val)
    { return memory_system.get_cell()->mk_real(val); }

    inline Cell *mk_number(Number &number)
    {
      if (number.is_fixnum())
	return mk_integer(number.get_ivalue0());

      return mk_real(number.get_rvalue0());
    }

    inline Cell *mk_counted_string(const char *str, size_t len)
    { return memory_system.get_counted_string(str, len); }
    inline Cell *mk_empty_string(size_t len, char fill)
    { return memory_system.get_empty_string(len, fill); }
    inline Cell *mk_string(const char *str)
    { return mk_counted_string(str, strlen(str)); }
    inline Cell *mk_static_string(const char *str)
    { return memory_system.get_static_string(str); }
    inline Cell *mk_substring_copy(Cell *&src, long start, long end)
    { return memory_system.get_substring_copy(src, start, end); }
    inline Cell *mk_substring_reference(Cell *&src, long start, long end)
    { return memory_system.get_substring_reference(src, start, end); }


    // This may return string cell that explains the error instead of
    // regexp cell if compilation of regexp fails.
    inline Cell *mk_regexp(Cell *&rx_src, bool ignore_case = false)
    {
      Cell *regexp = memory_system.get_cell(rx_src);
      return regexp->mk_regexp(rx_src, ignore_case);
    }

    inline Cell *mk_time(long time_type, long nanosecond, long second)
    { return memory_system.get_cell()->mk_time(time_type, nanosecond, second); }

#ifdef HAVE_TIME_H
    inline Cell *mk_date(const struct tm *value, long nano, long zone_offset)
    { return memory_system.get_cell()->mk_date(value, nano, zone_offset); }
#endif

    inline Cell *mk_character(long c) { return Cell::mk_character(c); }
    inline Cell *mk_closure_template(Cell *&cpool, Cell *&code,
				     BytecodeProcedure *bytecode)
    {
      Cell *templ = memory_system.get_cell(cpool, code);
      return templ->mk_closure_template(cpool, code, bytecode);
    }

    inline Cell *mk_closure(Cell *&closure_template, Cell *&frame)
    {
      Cell *closure = memory_system.get_cell(closure_template, frame);
      return closure->mk_closure(closure_template, frame);
    }

    inline Cell *mk_promise(Cell *&proc)
    {
      Cell *promise = memory_system.get_cell(proc);
      return promise->mk_promise(proc);
    }

    inline Cell *mk_vector(size_t len) { return memory_system.get_vector(len); }
    inline Cell *mk_vector(size_t len, Cell *&fill)
    {
      return memory_system.get_vector(len, fill);
    }

    inline Cell *mk_port(Port *port)
    {
      return memory_system.get_cell()->mk_port(port);
    }

    Cell *mk_input_file_port(const char *name);
    Cell *mk_input_file_port(FILE *file, bool interactive = false);
    Cell *mk_output_file_port(const char *name);
    Cell *mk_output_file_port(FILE *file);

    Cell *mk_input_string_port(Cell *&str);
    Cell *mk_output_string_port();
    Cell *mk_foreign_object(void *pointer,
			    ForeignObjectDestructFunc destructor = NULL,
			    void *option_data = NULL);

    inline Cell *cons(Cell *&a, Cell *&d = Cell::nil0())
    { return mk_cons(a, d, false); }
    inline Cell *immutable_cons(Cell *&a, Cell *&d = Cell::nil0())
    { return mk_cons(a, d, true); }

    inline MemorySystem &get_memory_system() { return memory_system; }

    Cell *apply(Cell *procedure, Cell *args = Cell::nil());
    Cell *applyf(Cell *procedure, const char *fmt, va_list ap);

    inline bool set_tracing(bool enable)
    {
      bool prev = tracing;
      tracing = enable;
      return prev;
    }
    inline bool get_tracing(Register &context) const
    { return tracing && context.output_port->is_output_port(); }

    inline bool set_bytecompile_tracing(bool enable)
    {
      bool prev = bytecompile_tracing;
      bytecompile_tracing = enable;
      return prev;
    }
    inline bool get_bytecompile_tracing(Register &context) const
    { return bytecompile_tracing && context.output_port->is_output_port(); }

    inline bool set_bytecode_tracing(bool enable)
    {
      bool prev = bytecode_tracing;
      bytecode_tracing = enable;
      return prev;
    }
    inline bool get_bytecode_tracing(Register &context) const
    { return bytecode_tracing && context.output_port->is_output_port(); }
  };


  class VMStackRoot : public StackRoot
  {
  public:
    inline VMStackRoot(VirtualMachine &vm, Cell *cell = Cell::nil())
      : StackRoot(vm.memory_system, cell) {}
    inline VMStackRoot(VirtualMachine &vm, const Cell *cell)
      : StackRoot(vm.memory_system, const_cast<Cell *>(cell)) {}

    inline VMStackRoot &operator=(Cell *cell)
    {
      static_cast<StackRoot &>(*this) = cell;
      return *this;
    }

    inline VMStackRoot &operator=(const Cell *cell)
    {
      static_cast<StackRoot &>(*this) = cell;
      return *this;
    }

    inline VMStackRoot &operator=(const VMStackRoot &root)
    {
      static_cast<StackRoot &>(*this) = static_cast<const StackRoot &>(root);
      return *this;
    }
  };


}


#define _VIRTUAL_MACHINE_DEFINED_	1

#endif	/* __cplusplus */

#endif	/* _LIBTSCORE_VIRTUAL_MACHINE_H_ */
