/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: vmopcode.cc,v 1.2 2010/08/15 15:39:47 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libtscore/vmopcode.h"

#include "libtscore/cell.h"
#include "libtscore/port.h"
#include "libtscore/writer.h"


namespace LibTSCore
{

const OpcodeInfo OpcodeInfo::opcode_info[] = {
  { "*not used*", 1 },
  { "iconst_m1", 1 },
  { "iconst_0", 1 },
  { "iconst_1", 1 },
  { "iconst_2", 1 },
  { "iconst_3", 1 },
  { "iconst_4", 1 },
  { "iconst_i1", 2 },
  { "iconst_i2", 3 },
  { "iconst_i4", 5 },
  { "iconst_i8", 9 },
  { "const_null", 1 },
  { "const_t", 1 },
  { "const_f", 1 },
  { "const_unspecified", 1 },
  { "const_p_0", 1 },
  { "const_p_1", 1 },
  { "const_p_2", 1 },
  { "const_p_3", 1 },
  { "const_p1", 2 },
  { "const_p2", 3 },
  { "const_p4", 5 },
  { "load_l1", 2 },
  { "load_l2", 3 },
  { "load_l4", 5 },
  { "store_l1", 2 },
  { "store_l2", 3 },
  { "store_l4", 5 },
  { "load_g1", 3 },
  { "load_g2", 4 },
  { "load_g4", 6 },
  { "store_g1", 3 },
  { "store_g2", 4 },
  { "store_g4", 6 },
  { "load_t1", 2 },
  { "load_t2", 3 },
  { "load_t4", 5 },
  { "store_t1", 2 },
  { "store_t2", 3 },
  { "store_t4", 5 },
  { "load_f1", 2 },
  { "load_f2", 3 },
  { "load_f4", 5 },
  { "store_f1", 2 },
  { "store_f2", 3 },
  { "store_f4", 5 },
  { "if_t1", 2 },
  { "if_t2", 3 },
  { "if_t4", 5 },
  { "if_f1", 2 },
  { "if_f2", 3 },
  { "if_f4", 5 },
  { "goto1", 2 },
  { "goto2", 3 },
  { "goto4", 5 },
  { "cons0", 1 },
  { "cons1", 3 },
  { "cons2", 5 },
  { "cons4", 9 },
  { "car", 1 },
  { "cdr", 1 },
  { "set_car1", 2 },
  { "set_car2", 3 },
  { "set_car4", 5 },
  { "set_cdr1", 2 },
  { "set_cdr2", 3 },
  { "set_cdr4", 5 },
  { "apply1", 3 },
  { "apply2", 5 },
  { "apply4", 9 },
  { "call_cc1", 3 },
  { "call_cc2", 5 },
  { "call_cc4", 9 },
  { "apply_continuation0", 1 },
  { "lambda", 1 },
  { "promise", 1 },
  { "if_forced1", 2 },
  { "if_forced2", 3 },
  { "if_forced4", 5 },
  { "force_l1", 2 },
  { "force_l2", 3 },
  { "force_l4", 5 },
  { "trampoline", 1 },
  { "meta_return", 1 },
  { "unbound_f1", 2 },
  { "unbound_f2", 3 },
  { "unbound_f4", 5 },
  { "nop", 1 }
};


#ifndef WORDS_BIGENDIAN
# if !defined(__i386) && !defined(__i386__) && !defined(_M_IX86) && \
  !(defined(__ppc__) && defined(__APPLE__)) &&			   \
  !defined(__x86_64) && !defined(__x86_64__) &&			   \
  !defined(__mc68020__)
#  define WORDS_BIGENDIAN 1
# endif
#endif


static inline long
read_signed_8(const char *pc)
{
  return static_cast<long>(*pc);
}


static inline long
read_signed_16(const char *pc)
{
#if defined(WORDS_BIGENDIAN) || SIZEOF_SHORT != 2
  long value = static_cast<long>(pc[1]) << 8;
  return value | reinterpret_cast<const unsigned char *>(pc)[0];
#else
  return static_cast<long>(reinterpret_cast<const short *>(pc)[0]);
#endif
}


static inline long
read_signed_32(const char *pc)
{
#if defined(WORDS_BIGENDIAN) || SIZEOF_INT != 4
  const unsigned char *upc = reinterpret_cast<const unsigned char *>(pc);
  return static_cast<long>(pc[3]) << 24
    | static_cast<long>(upc[2]) << 16
    | static_cast<long>(upc[1]) << 8
    | static_cast<long>(upc[0]);
#else
  return static_cast<long>(reinterpret_cast<const int *>(pc)[0]);
#endif
}


static inline long
read_signed_64(const char *pc)
{
#if SIZEOF_LONG >= 8
# if defined(WORDS_BIGENDIAN) || SIZEOF_LONG != 8
  const unsigned char *upc = reinterpret_cast<const unsigned char *>(pc);
  return static_cast<long>(pc[7]) << 56
    | static_cast<long>(upc[6]) << 48
    | static_cast<long>(upc[5]) << 40
    | static_cast<long>(upc[4]) << 32
    | static_cast<long>(upc[3]) << 24
    | static_cast<long>(upc[2]) << 16
    | static_cast<long>(upc[1]) << 8
    | static_cast<long>(upc[0]);
# else
  return static_cast<long>(reinterpret_cast<const long *>(pc)[0]);
# endif
#else
  // We just ignore higher 32 bits of 64 bits integer here.
  return read_signed_32(pc);
#endif
}


static inline unsigned long
read_unsigned_8(const char *pc)
{
  return static_cast<unsigned long>(reinterpret_cast<const unsigned char *>(pc)[0]);
}


static inline unsigned long
read_unsigned_16(const char *pc)
{
#if defined(WORDS_BIGENDIAN) || SIZEOF_SHORT != 2
  const unsigned char *upc = reinterpret_cast<const unsigned char *>(pc);
  return static_cast<unsigned long>(upc[1]) << 8 | static_cast<long>(upc[0]);
#else
  return static_cast<unsigned long>(reinterpret_cast<const short *>(pc)[0]);
#endif
}


static inline unsigned long
read_unsigned_32(const char *pc)
{
#if defined(WORDS_BIGENDIAN) || SIZEOF_INT != 4
  const unsigned char *upc = reinterpret_cast<const unsigned char *>(pc);
  return static_cast<unsigned long>(upc[3]) << 24
    | static_cast<unsigned long>(upc[2]) << 16
    | static_cast<unsigned long>(upc[1]) << 8
    | static_cast<unsigned long>(upc[0]);
#else
  return static_cast<unsigned long>(reinterpret_cast<const unsigned int *>(pc)[0]);
#endif
}


static inline unsigned long
read_unsigned_64(const char *pc)
{
#if SIZEOF_LONG >= 8
# if defined(WORDS_BIGENDIAN) || SIZEOF_LONG != 8
  const unsigned char *upc = reinterpret_cast<const unsigned char *>(pc);
  return static_cast<unsigned long>(upc[7]) << 56
    | static_cast<unsigned long>(upc[6]) << 48
    | static_cast<unsigned long>(upc[5]) << 40
    | static_cast<unsigned long>(upc[4]) << 32
    | static_cast<unsigned long>(upc[3]) << 24
    | static_cast<unsigned long>(upc[2]) << 16
    | static_cast<unsigned long>(upc[1]) << 8
    | static_cast<unsigned long>(upc[0]);
# else
  return static_cast<unsigned long>(reinterpret_cast<const unsigned long *>(pc)[0]);
# endif
#else
  // We just ignore higher 32 bits of 64 bits integer here.
  return read_unsigned_32(pc);
#endif
}


void
OpcodeInfo::print_bytecode(Port *port, const char *from, Cell *cpool)
{
  char buf[256];
  switch (static_cast<TSVMOpcode>(*from))
    {
    case OP_NOT_USED:
    case OP_ICONST_M1:
    case OP_ICONST_0:
    case OP_ICONST_1:
    case OP_ICONST_2:
    case OP_ICONST_3:
    case OP_ICONST_4:
    case OP_CONST_NULL:
    case OP_CONST_T:
    case OP_CONST_F:
    case OP_CONST_UNSPECIFIED:
    case OP_APPLY_CONTINUATION0:
    case OP_LAMBDA:
    case OP_PROMISE:
    case OP_TRAMPOLINE:
    case OP_META_RETURN:
    case OP_NOP:
    case OP_CONS0:
    case OP_CAR:
    case OP_CDR:
      snprintf(buf, 256, "%p: %s\n", from, get_opcode_name(*from));
      port->append(buf);
      return;

    case OP_ICONST_I1:
      snprintf(buf, 256, "%p: %s %ld\n",
	       from, get_opcode_name(*from), read_signed_8(from + 1));
      port->append(buf);
      return;

    case OP_ICONST_I2:
      snprintf(buf, 256, "%p: %s %ld\n",
	       from, get_opcode_name(*from), read_signed_16(from + 1));
      port->append(buf);
      return;

    case OP_ICONST_I4:
      snprintf(buf, 256, "%p: %s %ld\n",
	       from, get_opcode_name(*from), read_signed_32(from + 1));
      port->append(buf);
      return;

    case OP_ICONST_I8:
      snprintf(buf, 256, "%p: %s %ld\n",
	       from, get_opcode_name(*from), read_signed_64(from + 1));
      port->append(buf);
      return;

    case OP_CONST_P1:
    case OP_LOAD_F1:
    case OP_STORE_F1:
    case OP_UNBOUND_F1:
      snprintf(buf, 256, "%p: %s %lu  // ",
	       from, get_opcode_name(*from), read_unsigned_8(from + 1));
      port->append(buf);
      Writer::write_datum(cpool->load_constant(read_unsigned_8(from + 1)), port, false);
      port->append('\n');
      return;

    case OP_CONST_P2:
    case OP_LOAD_F2:
    case OP_STORE_F2:
    case OP_UNBOUND_F2:
      snprintf(buf, 256, "%p: %s %lu  // ",
	       from, get_opcode_name(*from), read_unsigned_16(from + 1));
      port->append(buf);
      Writer::write_datum(cpool->load_constant(read_unsigned_16(from + 1)), port, false);
      port->append('\n');
      return;

    case OP_CONST_P4:
    case OP_LOAD_F4:
    case OP_STORE_F4:
    case OP_UNBOUND_F4:
      snprintf(buf, 256, "%p: %s %lu  // ",
	       from, get_opcode_name(*from), read_unsigned_32(from + 1));
      port->append(buf);
      Writer::write_datum(cpool->load_constant(read_unsigned_32(from + 1)), port, false);
      port->append('\n');
      return;

    case OP_LOAD_L1:
    case OP_STORE_L1:
    case OP_LOAD_T1:
    case OP_STORE_T1:
    case OP_SET_CAR1:
    case OP_SET_CDR1:
    case OP_FORCE_L1:
      snprintf(buf, 256, "%p: %s %lu\n",
	       from, get_opcode_name(*from), read_unsigned_8(from + 1));
      port->append(buf);
      return;

    case OP_LOAD_L2:
    case OP_STORE_L2:
    case OP_LOAD_T2:
    case OP_STORE_T2:
    case OP_SET_CAR2:
    case OP_SET_CDR2:
    case OP_FORCE_L2:
      snprintf(buf, 256, "%p: %s %lu\n",
	       from, get_opcode_name(*from), read_unsigned_16(from + 1));
      port->append(buf);
      return;


    case OP_LOAD_L4:
    case OP_STORE_L4:
    case OP_LOAD_T4:
    case OP_STORE_T4:
    case OP_SET_CAR4:
    case OP_SET_CDR4:
    case OP_FORCE_L4:
      snprintf(buf, 256, "%p: %s %lu\n",
	       from, get_opcode_name(*from), read_unsigned_32(from + 1));
      port->append(buf);
      return;


    case OP_LOAD_G1:
    case OP_STORE_G1:
      snprintf(buf, 256, "%p: %s %lu %lu\n",
	       from, get_opcode_name(*from),
	       read_unsigned_8(from + 1), read_unsigned_8(from + 2));
      port->append(buf);
      return;

    case OP_LOAD_G2:
    case OP_STORE_G2:
      snprintf(buf, 256, "%p: %s %lu %lu\n",
	       from, get_opcode_name(*from),
	       read_unsigned_8(from + 1), read_unsigned_16(from + 2));
      port->append(buf);
      return;

    case OP_LOAD_G4:
    case OP_STORE_G4:
      snprintf(buf, 256, "%p: %s %lu %lu\n",
	       from, get_opcode_name(*from),
	       read_unsigned_8(from + 1), read_unsigned_32(from + 2));
      port->append(buf);
      return;

    case OP_IF_T1:
    case OP_IF_F1:
    case OP_GOTO1:
    case OP_IF_FORCED1:
      snprintf(buf, 256, "%p: %s %p\n",
	       from, get_opcode_name(*from),
	       from + read_signed_8(from + 1) + 1);
      port->append(buf);
      return;

    case OP_IF_T2:
    case OP_IF_F2:
    case OP_GOTO2:
    case OP_IF_FORCED2:
      snprintf(buf, 256, "%p: %s %p\n",
	       from, get_opcode_name(*from),
	       from + read_signed_16(from + 1) + 1);
      port->append(buf);
      return;

    case OP_IF_T4:
    case OP_IF_F4:
    case OP_GOTO4:
    case OP_IF_FORCED4:
      snprintf(buf, 256, "%p: %s %p\n",
	       from, get_opcode_name(*from),
	       from + read_signed_32(from + 1) + 1);
      port->append(buf);
      return;

    case OP_CONS1:
    case OP_APPLY1:
    case OP_CALL_CC1:
      snprintf(buf, 256, "%p: %s %lu %lu\n",
	       from, get_opcode_name(*from),
	       read_unsigned_8(from + 1), read_unsigned_8(from + 2));
      port->append(buf);
      return;

    case OP_CONS2:
    case OP_APPLY2:
    case OP_CALL_CC2:
      snprintf(buf, 256, "%p: %s %lu %lu\n",
	       from, get_opcode_name(*from),
	       read_unsigned_16(from + 1), read_unsigned_16(from + 3));
      port->append(buf);
      return;

    case OP_CONS4:
    case OP_APPLY4:
    case OP_CALL_CC4:
      snprintf(buf, 256, "%p: %s %lu %lu\n",
	       from, get_opcode_name(*from),
	       read_unsigned_32(from + 1), read_unsigned_32(from + 5));
      port->append(buf);
      return;
    }
}


}
