/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: vmopcode.h,v 1.2 2010/08/15 15:39:47 fuyu Exp $
 */

#ifndef _LIBTSCORE_VMOPCODE_H_
#define _LIBTSCORE_VMOPCODE_H_

#ifdef __cplusplus

#pragma interface


using namespace std;
namespace LibTSCore
{

  class Cell;
  class Port;

  struct OpcodeInfo
  {
    const char *name;
    long length;

    static const OpcodeInfo opcode_info[];

    static inline const char *get_opcode_name(unsigned char code)
    {
      return opcode_info[static_cast<long>(code)].name;
    }

    static inline long get_opcode_length(unsigned char code)
    {
      return opcode_info[static_cast<long>(code)].length;
    }

    static inline const char *get_next_pc(const char *current_pc)
    {
      return current_pc + get_opcode_length(*current_pc);
    }

    static void print_bytecode(Port *port, const char *from, Cell *cpool);
  };


  enum TSVMOpcode {
    OP_NOT_USED = 0,

    // Load very small integer, -1, 0, 1, 2, 3 or 4, into `value'.
    OP_ICONST_M1,
#define OPS_ICONST_M1	"\001"
    OP_ICONST_0,
#define OPS_ICONST_0	"\002"
    OP_ICONST_1,
#define OPS_ICONST_1	"\003"
    OP_ICONST_2,
#define OPS_ICONST_2	"\004"
    OP_ICONST_3,
#define OPS_ICONST_3	"\005"
    OP_ICONST_4,
#define OPS_ICONST_4	"\006"


    // Load small integer value given as immediate into `value'.
    // Immediates are stored in little endian order within bytecode.
    // Postfixes, I1, I2, I4 and I8 shows the byte length of immediate
    // value.  Each immediate values are treated as `signed'.
    // OP_ICONST_I8 only used on ISA having 8 bytes `long'.
    OP_ICONST_I1,	// imm8
#define OPS_ICONST_I1	"\007"
    OP_ICONST_I2,	// imm16
#define OPS_ICONST_I2	"\010"
    OP_ICONST_I4,	// imm32
#define OPS_ICONST_I4	"\011"
    OP_ICONST_I8,	// imm64
#define OPS_ICONST_I8	"\012"


    // Load well known constant, '(), #t or #f, into `value'.
    OP_CONST_NULL,
#define OPS_CONST_NULL	"\013"
    OP_CONST_T,
#define OPS_CONST_T	"\014"
    OP_CONST_F,
#define OPS_CONST_F	"\015"

    // for internal implementation
    OP_CONST_UNSPECIFIED,
#define OPS_CONST_UNSPECIFIED	"\016"

    // Load constant value within constant pool specified by index given as
    // immediate into `value'.  Immediates are stored in little endian
    // order within bytecode.
    // Postfixes, P1, P2 and P4 shows the byte length of immediate
    // index.  Each immediate values are treated as `unsigned'.
    // P2 and P4 will be only rarely used we think, just reserved.
    OP_CONST_P1,	// index8
#define OPS_CONST_P1	"\017"
    OP_CONST_P2,	// index16
#define OPS_CONST_P2	"\020"
    OP_CONST_P4,	// index32
#define OPS_CONST_P4	"\021"


    // Load value within local variable slot specified by immediate
    // slot number into `value'.  Immediates are stored in little endian
    // order within bytecode.
    // Postfixes, L1, L2 and L4 shows the byte length of immediate
    // slot number.  Each immediate slot numbers are treated as `unsigned'.
    // L2 and L4 will be only rarely used we think, just reserved.
    OP_LOAD_L1,		// slot8
#define OPS_LOAD_L1	"\022"
    OP_LOAD_L2,		// slot16
#define OPS_LOAD_L2	"\023"
    OP_LOAD_L4,		// slot32
#define OPS_LOAD_L4	"\024"


    // Store value within `value' into local variable slot specified by
    // immediate slot number.  Immediates are stored in little endian
    // order within bytecode.
    // Postfixes, L1, L2 and L4 shows the byte length of immediate
    // slot number.  Each immediate slot numbers are treated as `unsigned'.
    // L2 and L4 will be only rarely used we think, just reserved.
    OP_STORE_L1,	// slot8
#define OPS_STORE_L1	"\025"
    OP_STORE_L2,	// slot16
#define OPS_STORE_L2	"\026"
    OP_STORE_L4,	// slot32
#define OPS_STORE_L4	"\027"


    // Load value within variable slot within outer environment specified
    // by immediate count and slot number into `value'.  Immediates are
    // stored in little endian order within bytecode.
    // `env8' specifies the environment `env8' times outer environment
    // from current environment in which the variable slot located.
    // Postfixes, G1, G2 and G4 shows the byte length of immediate
    // slot number.  Each immediate slot numbers are treated as `unsigned'.
    // G2 and G4 will be only rarely used we think, just reserved.
    OP_LOAD_G1,		// env8 slot8
#define OPS_LOAD_G1	"\030"
    OP_LOAD_G2,		// env8 slot16
#define OPS_LOAD_G2	"\031"
    OP_LOAD_G4,		// env8 slot32
#define OPS_LOAD_G4	"\032"


    // Store value within `value' into variable slot within outer environment
    // specified by immediate count and slot number.  Immediates are stored
    // in little endian order within bytecode.
    // `env8' specifies the environment `env8' times outer environment
    // from current environment in which the variable slot located.
    // Postfixes, G1, G2 and G4 shows the byte length of immediate
    // slot number.  Each immediate slot numbers are treated as `unsigned'.
    // G2 and G4 will be only rarely used we think, just reserved.
    OP_STORE_G1,	// env8 slot8
#define OPS_STORE_G1	"\033"
    OP_STORE_G2,	// env8 slot16
#define OPS_STORE_G2	"\034"
    OP_STORE_G4,	// env8 slot32
#define OPS_STORE_G4	"\035"


    // Load value within variable slot of top level environment specified
    // by immediate slot number into `value'.  Immediates are stored in
    // little endian order within bytecode.
    // Postfixes, T1, T2 and T4 shows the byte length of immediate
    // slot number.  Each immediate slot numbers are treated as `unsigned'.
    // T4 will be only rarely used we think, just reserved.
    OP_LOAD_T1,		// slot8
#define OPS_LOAD_T1	"\036"
    OP_LOAD_T2,		// slot16
#define OPS_LOAD_T2	"\037"
    OP_LOAD_T4,		// slot32
#define OPS_LOAD_T4	"\040"


    // Store value within `value' into variable slot of top level environment
    // specified by immediate slot number.  Immediates are stored in little
    // endian order within bytecode.
    // Postfixes, T1, T2 and T4 shows the byte length of immediate
    // slot number.  Each immediate slot numbers are treated as `unsigned'.
    // T4 will be only rarely used we think, just reserved.
    OP_STORE_T1,	// slot8
#define OPS_STORE_T1	"\041"
    OP_STORE_T2,	// slot16
#define OPS_STORE_T2	"\042"
    OP_STORE_T4,	// slot32
#define OPS_STORE_T4	"\043"


    // Load value within variable specified by symbol given by constant
    // pool whose indices are given as immediates into `value'.  Immediates
    // are stored in little endian order within bytecode.
    // Postfixes, F1, F2 and F4 shows the byte length of immediate index.
    // Each immediate values are treated as `unsigned'.
    // These bytecodes are for `free variable's that aren't known to be
    // bound at bytecompilation time.
    OP_LOAD_F1,		// index8
#define OPS_LOAD_F1	"\044"
    OP_LOAD_F2,		// index16
#define OPS_LOAD_F2	"\045"
    OP_LOAD_F4,		// index32
#define OPS_LOAD_F4	"\046"


    // Store value within `value' into variable specified by symbol given
    // by constant pool whose indices are given as immediates.  Immediates
    // are stored in little endian order within bytecode.
    // Postfixes, F1, F2 and F4 shows the byte length of immediate index.
    // Each immediate values are treated as `unsigned'.
    // These bytecodes are for `free variable's that aren't known to be
    // bound at bytecompilation time.
    OP_STORE_F1,	// index8
#define OPS_STORE_F1	"\047"
    OP_STORE_F2,	// index16
#define OPS_STORE_F2	"\050"
    OP_STORE_F4,	// index32
#define OPS_STORE_F3	"\051"


    // If the value within `value' is considered as `true' in Scheme semantics,
    // i.e. not #f, flow of control is jumped with immediate offsets.
    // Immediates are stored in little endian order within bytecode.
    // Postfixes, T1, T2 and T4 shows the byte length of immediate offset.
    // Each immediate values are treated as `signed'.
    // T2 and T4 will be only rarely used we think, just reserved.
    // `offset's are treated as the offset from the first byte of
    // immediate `offset', i.e. (offset == -1) means self jump.
    OP_IF_T1,		// off8
#define OPS_IF_T1	"\052"
    OP_IF_T2,		// off16
#define OPS_IF_T2	"\053"
    OP_IF_T4,		// off32
#define OPS_IF_T4	"\054"


    // If the value within `value' is considered as `false' in Scheme
    // semantics, i.e. #f, flow of control is jumped with immediate offsets.
    // Immediates are stored in little endian order within bytecode.
    // Postfixes, T1, T2 and T4 shows the byte length of immediate offset.
    // Each immediate values are treated as `signed'.
    // T2 and T4 will be only rarely used we think, just reserved.
    // `offset's are treated as the offset from the first byte of
    // immediate `offset', i.e. (offset == -1) means self jump.
    OP_IF_F1,		// off8
#define OPS_IF_F1	"\055"
    OP_IF_F2,		// off16
#define OPS_IF_F2	"\056"
    OP_IF_F4,		// off32
#define OPS_IF_F4	"\057"


    // Unconditionally jumps the flow of control with immediate offsets.
    // Immediates are stored in little endian order within bytecode.
    // Postfixes, T1, T2 and T4 shows the byte length of immediate offset.
    // Each immediate values are treated as `signed'.
    // T2 and T4 will be only rarely used we think, just reserved.
    // `offset's are treated as the offset from the first byte of
    // immediate `offset', i.e. (offset == -1) means self jump.
    OP_GOTO1,		// off8
#define OPS_GOTO1	"\060"
    OP_GOTO2,		// off16
#define OPS_GOTO2	"\061"
    OP_GOTO4,		// off32
#define OPS_GOTO4	"\062"


    // allocate a new empty cons cell into `value'.
    OP_CONS0,
#define OPS_CONS0	"\063"


    // allocate a new cons cell with car and cdr from local variable slot
    // into `value'.
    OP_CONS1,		// slot8, slot8
#define OPS_CONS1	"\064"
    OP_CONS2,		// slot16, slot16
#define OPS_CONS2	"\065"
    OP_CONS4,		// slot32, slot32
#define OPS_CONS4	"\066"


    // Get the value of (car `value') into `value'.
    OP_CAR,
#define OPS_CAR		"\067"


    // Get the value of (cdr `value') into `value'.
    OP_CDR,
#define OPS_CDR		"\070"


    // Set the value of `value' as car of cons cell within the local
    // variable slot.
    OP_SET_CAR1,	// slot8
#define OPS_SET_CAR1	"\071"
    OP_SET_CAR2,	// slot16
#define OPS_SET_CAR2	"\072"
    OP_SET_CAR4,	// slot32
#define OPS_SET_CAR4	"\073"


    // Set the value of `value' as cdr of cons cell within the local
    // variable slot.
    OP_SET_CDR1,	// slot8
#define OPS_SET_CDR1	"\074"
    OP_SET_CDR2,	// slot16
#define OPS_SET_CDR2	"\075"
    OP_SET_CDR4,	// slot32
#define OPS_SET_CDR4	"\076"


    // OP_APPLY[124] and OP_CALL_CC[124] are almost same but OP_APPLY[124]
    // refer the implicit continuation argument from caller's first
    // argument, i.e. caller's continuation.  On the other hand,
    // OP_CALL_CC[124] fill the implicit continuation argument with
    // '() to show the fact that it isn't a tail-call.  Essentially,
    // these instructions are macro that combine sequence of bytecode
    // like below:
    // { OP_LOAD_L1 0, OP_STORE_L1 args_pointer,
    //   OP_LOAD_L1 proc_pointer, OP_CALL1 } for OP_APPLY[124]
    // { OP_CONST_NULL, OP_STORE_L1 args_pointer,
    //   OP_LOAD_L1 proc_pointer, OP_CALL1 } for OP_CALL_CC[124]
    //
    // NOTE: Since we don't actually have OP_CALL[124], these aren't usable.

    // Applies procedure with computed arguments on frame.
    // These are used for `tail-call's.
    OP_APPLY1,		// off8 nargs8
#define OPS_APPLY1	"\077"
    OP_APPLY2,		// off16 nargs16
#define OPS_APPLY2	"\100"
    OP_APPLY4,		// off32 nargs32
#define OPS_APPLY4	"\101"


    // Applies procedure with computed arguments on frame with current
    // continuation.
    // These are used for non `tail-call's.
    OP_CALL_CC1,	// off8 nargs8
#define OPS_CALL_CC1	"\102"
    OP_CALL_CC2,	// off16 nargs16
#define OPS_CALL_CC2	"\103"
    OP_CALL_CC4,	// off32 nargs32
#define OPS_CALL_CC4	"\104"


    // Applies implicit continuation introduced by bytecompiler to the
    // current context.
    // In our convention, the `continuation' is always given as the
    // first argument of all procedures.
    OP_APPLY_CONTINUATION0,
#define OPS_APPLY_CONTINUATION0	"\105"


    // Creates a closure from closure template within `value' and the
    // current frame.
    OP_LAMBDA,
#define OPS_LAMBDA	"\106"


    // Creates a promise from closure within `value' and place it
    // into `value'.
    OP_PROMISE,
#define OPS_PROMISE	"\107"


    // Tests whether `value' is promise or not and adjust `value'
    // based on the condition of current `value' like below and
    // branch if `value' isn't a not-computed promise:
    // (1) If `value' isn't a promise, simply ignore and do branch.
    // (2) If `value' is a computed promise, place the computed
    //     value into `value' and do branch.
    // (3) If `value' is a not-computed promise, place the
    //     procedure, with no-argument other than our implicit
    //     continuation, to compute the promise into `value'.
    OP_IF_FORCED1,	// off8
#define OPS_IF_FORCED1	"\110"
    OP_IF_FORCED2,	// off16
#define OPS_IF_FORCED2	"\111"
    OP_IF_FORCED4,	// off32
#define OPS_IF_FORCED4	"\112"


    // Memoize `value' as the computed promise within promise
    // placed at local variable slot.
    OP_FORCE_L1,	// slot8
#define OPS_FORCE_L1	"\113"
    OP_FORCE_L2,	// slot16
#define OPS_FORCE_L2	"\114"
    OP_FORCE_L4,	// slot32
#define OPS_FORCE_L4	"\115"


    // Applies procedure with prepared arguments on trampoline frame
    // as they are in current frame.
    OP_TRAMPOLINE,
#define OPS_TRAMPOLINE	"\116"


    // Returns the `value' to the caller and terminates current
    // evaluation cycle.
    OP_META_RETURN,
#define OPS_META_RETURN	"\117"


    // Causes `unbound variable' error for symbol taken from constant pool.
    // These will only be used by compiled code with non `interactive
    // environment', e.g. procedures created by R5RS `eval' procedure.
    // Postfixes, F1, F2 and F4 shows the byte length of immediate index.
    // Each immediate values are treated as `unsigned' index into the
    // current constant pool.
    OP_UNBOUND_F1,
#define OPS_UNBOUND_F1	"\120"
    OP_UNBOUND_F2,
#define OPS_UNBOUND_F2	"\121"
    OP_UNBOUND_F4,
#define OPS_UNBOUND_F4	"\122"


    // no operation
    OP_NOP
#define OPS_NOP		"\123"
  };


}


#endif	/* __cplusplus */

#endif	/* _LIBTSCORE_VMOPCODE_H_ */
