/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: writer.cc,v 1.18 2009/01/02 03:07:32 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libtscore/writer.h"

#include "libtscore/bytecode_buffer.h"
#include "libtscore/virtual_machine.h"
#include "libtscore/vmopcode.h"

#include <vector>

#include <cctype>


namespace LibTSCore
{

void
Writer::write_atom(const Cell *atom, Port *port,
		   bool external, bool inspection)
{
  if (!external)
    {
      if (inspection)
	{
	  if (atom->is_symbol())
	    {
	      port->append("#<symbol: \"")->append(atom->get_symname())
		->append("\">");
	      return;
	    }

	  if (atom->is_binding())
	    {
	      port->append("#<binding: \"")->append(CAR(atom)->get_symname());
	      char buf[256];
	      snprintf(buf, sizeof(buf), "\" slot=%ld env=%p>",
		       atom->get_slot_location(),
		       atom->get_bound_environment());
	      port->append(buf);
	      return;
	    }

	  if (atom->is_macro_syntax())
	    {
	      port->append("#<syntax: ")->append(atom->get_syntax_name())
		->append(":\n  ");
	      write_datum(atom->get_macro_transformer(), port, false, true);
	      port->append('>');
	      return;
	    }

	  if (atom->is_macro_transformer())
	    {
	      port->append("#<syntax-rules ");
	      write_datum(atom->get_macro_transformer_literals(), port, true);
	      port->append("\n");
	      const Cell *rules = atom->get_macro_transformer_syntax_rules();
	      while (rules->is_pair())
		{
		  port->append("    <pattern>: ");
		  write_datum(CAAR(rules), port, true);
		  port->append("\n   <template>: ");
		  write_datum(CADAR(rules), port, true);

		  const Cell *alist = CDDAR(rules);
		  while (alist != Cell::nil())
		    {
		      port->append("\n        <sub template>: ");
		      write_datum(CAAR(alist), port, true);
		      port->append("\n         <sub pattern>: ");
		      write_datum(CDAR(alist), port, true);
		      alist = CDR(alist);
		    }

		  if (CDR(rules)->is_pair())
		    port->append("\n");

		  rules = CDR(rules);
		}

	      port->append(">");
	      return;
	    }

	  if (atom->is_closure_template())
	    {
	      char buf[256];
	      Cell *cpool = atom->get_closure_constant_pool();
	      snprintf(buf, sizeof(buf),
		       "#<closure_template: constant_pool=%p\n    code: ",
		       cpool);
	      port->append(buf);
	      write_datum(atom->get_closure_code(), port, true);
	      BytecodeProcedure *proc = atom->get_closure_bytecode();
	      snprintf(buf, sizeof(buf),
		       "\nbytecode: "
		       "frame_size=%ld, min_arity=%ld, max_arity=%ld {\n",
		       proc->get_frame_size(),
		       proc->get_minimum_arity(), proc->get_maximum_arity());
	      port->append(buf);
	      BytecodeBuffer *bytecode = proc->get_buffer();
	      const char *pc = bytecode->get_bytecode();
	      const char *tail = pc + bytecode->get_current_offset();
	      while (pc < tail)
		{
		  OpcodeInfo::print_bytecode(port, pc, cpool);
		  pc = OpcodeInfo::get_next_pc(pc);
		}

	      port->append("}>\n");
	      return;
	    }

	  if (atom->is_closure())
	    {
	      char buf[256];
	      Cell *cpool = atom->get_closure_constant_pool();
	      snprintf(buf, sizeof(buf),
		       "#<closure: parent_frame=%p\n     code: ",
		       atom->get_closure_frame());
	      port->append(buf);
	      Cell *templ = atom->get_closure_template();
	      write_datum(templ->get_closure_code(), port, true);
	      BytecodeProcedure *proc = atom->get_closure_bytecode();
	      snprintf(buf, sizeof(buf),
		       "\n bytecode: constant_pool=%p"
		       ", frame_size=%ld, min_arity=%ld, max_arity=%ld {\n",
		       cpool, proc->get_frame_size(),
		       proc->get_minimum_arity(), proc->get_maximum_arity());
	      port->append(buf);
	      BytecodeBuffer *bytecode = proc->get_buffer();
	      const char *pc = bytecode->get_bytecode();
	      const char *tail = pc + bytecode->get_current_offset();
	      while (pc < tail)
		{
		  OpcodeInfo::print_bytecode(port, pc, cpool);
		  pc = OpcodeInfo::get_next_pc(pc);
		}

	      port->append("}>\n");
	      return;
	    }

	  if (atom->is_promise())
	    {
	      if (atom->is_forced_promise())
		{
		  port->append("#<promise(forced): ");
		  port->append(atom->get_memoized_value());
		  port->append(">\n");
		  return;
		}

	      port->append("#<promise: expression: {\n");
	      write_datum(atom->get_promise_procedure(), port, false, true);
	      port->append("}>\n");
	      return;
	    }
	}
      else
	{
	  if (atom->is_string())
	    {
	      port->append(atom->get_string(), atom->get_strlen());
	      return;
	    }

	  if (atom->is_character())
	    {
	      port->append(atom->get_character());
	      return;
	    }
	}
    }

  port->append(atom);
}


enum WriterState
{
  WRITER_STATE_INITIAL,
  WRITER_STATE_PAIR,
  WRITER_STATE_VECTOR
};


struct WriterContext
{
  WriterState state;
  long vector_index;
  const Cell *datum;
};


const Cell *
Writer::write_datum(const Cell *datum, Port *port,
		    bool external, bool inspection)
{
  WriterContext current_context = { WRITER_STATE_INITIAL, 0, datum };
  vector<WriterContext> writer_stack;

  while (true)
    {
      datum = current_context.datum;

      if (datum->is_pair())
	{
	  if (current_context.state == WRITER_STATE_INITIAL)
	    {
	      port->append('(');
	      current_context.state = WRITER_STATE_PAIR;
	      writer_stack.push_back(current_context);
	      current_context.state = WRITER_STATE_INITIAL;
	      current_context.datum = CAR(datum);
	      continue;
	    }

	  assert(current_context.state == WRITER_STATE_PAIR);

	  datum = CDR(datum);
	  if (datum == Cell::nil())
	    {
	      port->append(')');
	      current_context.state = WRITER_STATE_INITIAL;
	      goto done;
	    }

	  if (datum->is_pair())
	    {
	      port->append(' ');
	      current_context.datum = datum;
	      writer_stack.push_back(current_context);
	      current_context.state = WRITER_STATE_INITIAL;
	      current_context.datum = CAR(datum);
	      continue;
	    }

	  port->append(" . ");
	}

      if (datum->is_vector())
	{
	  if (current_context.vector_index == 0)
	    {
	      current_context.state = WRITER_STATE_VECTOR;
	      port->append("#(");
	    }

	  assert(current_context.state == WRITER_STATE_VECTOR);

	  if (datum->get_vector_length() > current_context.vector_index)
	    {
	      if (current_context.vector_index > 0)
		port->append(' ');

	      assert(current_context.datum == datum);

	      writer_stack.push_back(current_context);
	      current_context.state = WRITER_STATE_INITIAL;
	      current_context.datum = datum->get_vector_element(current_context.vector_index++);
	      current_context.vector_index = 0;
	      continue;
	    }
	  else
	    {
	      port->append(')');
	      current_context.state = WRITER_STATE_INITIAL;
	      goto done;
	    }
	}

      if (external && !inspection
	  && (datum->is_environment()
	      || datum == Cell::eof() || datum == Cell::unspecified()
	      || datum->is_port()
	      || datum->is_native_procedure()
	      || datum->is_closure() || datum->is_promise()
	      || datum->is_continuation()
	      || datum->is_foreign_object()))
	return datum;	// error

      write_atom(datum, port, external, inspection);

      if (current_context.state == WRITER_STATE_PAIR)
	{ // This is only for the last element of improper list.
	  port->append(')');
	  current_context.state = WRITER_STATE_INITIAL;
	}

    done:
      assert(current_context.state == WRITER_STATE_INITIAL);

      if (current_context.state == WRITER_STATE_INITIAL)
	{
	  if (writer_stack.empty())
	    break;
	  current_context = writer_stack.back();
	  writer_stack.pop_back();
	  if (current_context.state == WRITER_STATE_VECTOR)
	    ++current_context.vector_index;
	}
    }
  return Cell::nil();
}


static const char * const short_binary[] = {
  "0",	// not used.
  "1",
  "10",
  "11",
  "100",
  "101",
  "110",
  "111",
  "1000",
  "1001",
  "1010",
  "1011",
  "1100",
  "1101",
  "1110",
  "1111"
};


static const char * const full_binary[] = {
  "0000",
  "0001",
  "0010",
  "0011",
  "0100",
  "0101",
  "0110",
  "0111",
  "1000",
  "1001",
  "1010",
  "1011",
  "1100",
  "1101",
  "1110",
  "1111"
};


const Cell *
Writer::write_number(const Cell *datum, Port *output_port, long radix)
{
  if (!datum->is_integer())
    {
      output_port->append(datum);
      return Cell::nil();
    }

  long value = datum->get_integer0();
  char buf[256];
  switch (radix)
    {
    case 2:
      break;

    case 8:
      snprintf(buf, sizeof(buf), "%lo", value);
      output_port->append(buf);
      return Cell::nil();

    case 10:
      snprintf(buf, sizeof(buf), "%ld", value);
      output_port->append(buf);
      return Cell::nil();

    case 16:
      snprintf(buf, sizeof(buf), "%lx", value);
      output_port->append(buf);
      return Cell::nil();

    default:
      return datum;
    }

  // radix == 2
  if (value == 0)
    {
      output_port->append('0');
      return Cell::nil();
    }

  unsigned long uvalue = value;
  long i = 0;
  while (uvalue != 0)
    {
      buf[i++] = uvalue & 0xf;
      uvalue >>= 4;
    }

  assert(i > 0);

  output_port->append(short_binary[static_cast<unsigned long>(buf[--i])]);
  while (i > 0)
    output_port->append(full_binary[static_cast<unsigned long>(buf[--i])]);

  return Cell::nil();
}

}
