/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: bbs_thread_renderer.cc,v 1.41 2009/01/01 07:47:06 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "bbs_thread_renderer.h"

#include "libochusha/bulletin_board.h"
#include "libochusha/constants.h"
#include "libochusha/misc_utils.h"
#include "libochusha/text_utils.h"
#include "libochusha/uri_utils.h"

#include "controller.h"
#include "image_manager.h"
#include "ui_constants.h"

#include "libtscore/tscore.h"

#include "libochushagtk/hypertext_buffer.h"
#include "libochushagtk/hypertext_tag_table.h"

#include <string>

#include <cstdlib>
#include <cstring>


namespace Ochusha
{

//#define DISABLE_SCHEME_REWRITER	1


#define SKIN_KEYWORD_PATTERN	"<([A-Z]+)/>"
#define UTF8_DIGITS		"[0-9\357\274\220-\357\274\231]"
#define DIGITS_PATTERN		"[[:space:]]*" UTF8_DIGITS "+[[:space:]]*"
#define BE_ID_PATTERN		"<?BE:([0-9]+)[:-]([0-9]+|#*)>?"
#define ID_ID_PATTERN		"(?:ID|HOST|" HASSHINMOTO_UTF8 "):[^? \t[:^ascii:]]+"
#define BE_ICON_PATTERN		"sssp://img.2ch.net/ico/([^? \t[:^ascii:]]+)"
#define LARGE_COLON		"\357\274\232"


static const char *default_skin_header =
  "<html><body>"
  "<h1 class=\"ThreadTitle\"><THREADNAME/></h1>"
  "<dl>";

static const char *default_skin_res =
  "<dt><br><NUMBER/> " LARGE_COLON
  "<span class=\"Name\"><MAILNAME/></span> <MAIL/> " LARGE_COLON
  "<DATE/> <ID/></dt><dd class=\"ResponseBody\"><MESSAGE/><br></dd>";

static const char *default_skin_new_res = default_skin_res;

static const char *default_skin_footer = 
  "</dl>-<font size=\"-1\"><b><font color=\"#ff99cc\"><SIZEKB/>KB</font></b>"
  " (<SIZE/>Byte)</font>-</font></body></html>";

static const char *default_skin_popup_res = default_skin_res;

static const char *default_skin_popup_title =
  "<html><body>"\
  "<h1 class=\"ThreadTitle\"><THREADNAME/></h1>";

static const char *default_skin_popup_id_header =
  "<html><body>"
  "<h1 class=\"ThreadTitle\"><THREADNAME/></h1>"
  "<dl><dd class=\"ResponseBody\">";

static const char *default_skin_popup_id_res =
  "<a href=\"<PLAINNUMBER/>\">&gt;&gt;<PLAINNUMBER/></a>";

static const char *default_skin_popup_id_res_nl =
  "<a href=\"<PLAINNUMBER/>\">&gt;&gt;<PLAINNUMBER/></a><br>";


BBSThreadRenderer::BBSThreadRenderer(Controller &controller,
				     BBSThread *thread,
				     OchushaHypertextBuffer *buffer,
				     OchushaHypertextView *view)
  : HTMLRenderer(controller, buffer, view),
    thread(thread),
    skin_keyword(SKIN_KEYWORD_PATTERN),
    digits(DIGITS_PATTERN),
    be_id(BE_ID_PATTERN),
    id_id(ID_ID_PATTERN),
    be_icon(BE_ICON_PATTERN),
    number_of_responses_got(thread->get_number_of_responses_got())
{
}


void
BBSThreadRenderer::start_element(const char *name, const char *const attrs[])
{
  HTMLRenderer::start_element(name, attrs);

  if (strcasecmp(name, "a") != 0)
    return;

  if (controller.get_prefs().get_enable_inline_image())
    {
      int i = 0;
      const char *link = NULL;
      while (attrs[i * 2] != NULL)
	{
	  if (strcasecmp(attrs[i * 2], "href") == 0)
	    {
	      link = attrs[i * 2 + 1];
	      canonicalizer.clear();
	      if (strncmp(link, "http", 4) == 0
		  || strncmp(link, "https", 5) == 0)
		link = canonicalizer.canon_url(link, -1);
	      break;
	    }
	  i++;
	}

      if (link == NULL)
	return;

      if (controller.get_image_manager().is_image_link(link))
	{
	  char width_buf[16];
	  char height_buf[16];
	  snprintf(width_buf, 16, "%d",
		   controller.get_prefs().get_inline_image_width());
	  snprintf(height_buf, 16, "%d",
		   controller.get_prefs().get_inline_image_height());

	  string image_link;
	  image_link.append(link);
	  const char * const img_attrs[] = {
	    "src", image_link.c_str(),
	    "width", width_buf,
	    "height", height_buf,
	    "hspace", "6",
	    "vspace", "0",
	    "class", "Secret",
	    NULL
	  };

	  HTMLRenderer::start_element("img", img_attrs);
	}
    }
}


bool
BBSThreadRenderer::is_visited_link(const char *link)
{
  return strcmp(link, "mailto:sage") == 0
    || strncmp(link, "mailto:sage<FAKE>", 17) == 0
    || (controller.get_image_manager().is_image_link(link)
	&& controller.get_image_manager().is_tenured(link));
}


void
BBSThreadRenderer::render_skin_fragment(const char *skin_fragment,
					const char *title,
					guint res_num,
					const char *name, const char *mailto,
					const char *date, const char *id,
					const char *message, size_t message_len)
{
  g_return_if_fail(skin_fragment != NULL);

  char tmp_buffer[64];
  string html_fragment;

  const char *cur_pos = skin_fragment;
  size_t len = strlen(cur_pos);
  while (skin_keyword.match(cur_pos, len))
    {
      const char *match_begin0 = skin_keyword.match_begin(0);
      if (cur_pos < match_begin0)
	html_fragment.append(cur_pos, match_begin0 - cur_pos);

      const char *keyword = skin_keyword.match_begin(1);
      size_t keyword_len = skin_keyword.match_len(1);
      switch (keyword_len)
	{
	case 2:
	  if (strncmp(keyword, "ID", 2) == 0)
	    {
	      if (id != NULL)
		html_fragment.append(id);
	    }
	case 4:
	  if (strncmp(keyword, "NAME", 4) == 0)
	    {
	      if (name != NULL)
		html_fragment.append(name);
	    }
	  else if (strncmp(keyword, "MAIL", 4) == 0)
	    {
	      if (mailto != NULL)
		{
		  html_fragment.append("[");
		  html_fragment.append(mailto);
		  html_fragment.append("]");
		}
	    }
	  else if (strncmp(keyword, "DATE", 4) == 0)
	    {
	      if (date != NULL)
		html_fragment.append(date);
	    }
	  else if (strncmp(keyword, "SIZE", 4) == 0)
	    {
	      snprintf(tmp_buffer, 64, "%lu",
		       static_cast<unsigned long>(thread->get_dat_filesize()));
	      html_fragment.append(tmp_buffer);
	    }
	  break;

	case 6:
	  if (strncmp(keyword, "NUMBER", 6) == 0)
	    {
	      html_fragment.append("<a href=&lt;RES&gt;");
	      snprintf(tmp_buffer, 64, "%d>%d</a>", res_num, res_num);
	      html_fragment.append(tmp_buffer);
	    }
	  else if (strncmp(keyword, "DATEID", 6) == 0)
	    {
	      if (date != NULL)
		{
		  html_fragment.append(date);
		  if (id != NULL)
		    html_fragment.append(" ");
		}
	      if (id != NULL)
		html_fragment.append(id);
	    }
	  else if (strncmp(keyword, "SIZEKB", 6) == 0)
	    {
	      snprintf(tmp_buffer, 64, "%lu",
		       static_cast<unsigned long>(thread->get_dat_filesize()) / 1024);
	      html_fragment.append(tmp_buffer);
	    }
	  break;

	case 7:
	  if (strncmp(keyword, "MESSAGE", 7) == 0)
	    {
	      if (message != NULL)
		html_fragment.append(message, message_len);
	    }
	  else if (strncmp(keyword, "BBSNAME", 7) == 0)
	    {
	      int bbs_type = thread->get_board()->get_bbs_type();
	      const char *bbs_name
		= BulletinBoard::get_bbs_type_description(bbs_type);
	      html_fragment.append(bbs_name);
	    }
	  break;

	case 8:
	  if (strncmp(keyword, "MAILNAME", 8) == 0)
	    {
	      if (name != NULL)
		html_fragment.append(name);
	    }
	  else if (strncmp(keyword, "SKINPATH", 8) == 0)
	    {
	      // 褺̵
	    }
	  else if (strncmp(keyword, "BOARDURL", 8) == 0)
	    {
	      html_fragment.append(thread->get_board()->get_base_url());
	    }
	  break;

	case 9:
	  if (strncmp(keyword, "THREADURL", 9) == 0)
	    {
	      char *url = thread->get_url();
	      html_fragment.append(url);
	      free(url);
	    }
	  else if (strncmp(keyword, "BOARDNAME", 9) == 0)
	    {
	      html_fragment.append(thread->get_board()->get_name());
	    }
	  break;

	case 10:
	  if (strncmp(keyword, "THREADNAME", 10) == 0)
	    {
	      if (title != NULL)
		html_fragment.append(title);
	      else if (thread->get_title() != NULL)
		html_fragment.append(thread->get_title());
	    }
	  break;

	case 11:
	  if (strncmp(keyword, "PLAINNUMBER", 11) == 0)
	    {
	      snprintf(tmp_buffer, 64, "%d", res_num);
	      html_fragment.append(tmp_buffer);
	    }
	  else if (strncmp(keyword, "GETRESCOUNT", 11) == 0)
	    {
	      snprintf(tmp_buffer, 64, "%d", number_of_responses_got);
	      html_fragment.append(tmp_buffer);
	    }
	  else if (strncmp(keyword, "NEWRESCOUNT", 11) == 0)
	    {
	      snprintf(tmp_buffer, 64, "%d",
		       thread->get_number_of_responses_available()
		       - number_of_responses_got);
	      html_fragment.append(tmp_buffer);
	    }
	  else if (strncmp(keyword, "ALLRESCOUNT", 11) == 0)
	    {
	      snprintf(tmp_buffer, 64, "%d",
		       thread->get_number_of_responses_available());
	      html_fragment.append(tmp_buffer);
	    }
	  break;

	default:
	  break;	// ̵
	}

      len -= skin_keyword.match_end(0) - cur_pos;

      if (len == 0)
	break;
      cur_pos = skin_keyword.match_end(0);
    }

  if (len > 0)
    html_fragment.append(cur_pos);

  parse_text(html_fragment.c_str(), html_fragment.length());
}


void
BBSThreadRenderer::render_header(const char *title)
{
  assert(title != NULL && *title != '\0');
  render_skin_fragment(default_skin_header, title,
		       0, NULL, NULL, NULL, NULL, NULL, 0);
}


void
BBSThreadRenderer::render_a_bone(guint res_num)
{
  string message;
  char *url = thread->get_url(res_num);

  const char *a_bone = _("A Bone");
  if (!controller.get_prefs().get_disable_popup_a_bone() && url != NULL)
    {
      message.append("<a href=\"");
      message.append(url);
      message.append("\">");
      message.append(a_bone);
      message.append("</a>");
      free(url);
    }
  else
    {
      message.append(a_bone);
    }

  const char *skin_fragment;
  if (res_num <= number_of_responses_got)
    skin_fragment = default_skin_res;
  else
    skin_fragment = default_skin_new_res;

  string name;
  name.append("<a href=\"mailto:").append(a_bone).append("\"><b>")
    .append(a_bone).append("</b></a>");
  render_skin_fragment(skin_fragment, NULL, res_num,
		       name.c_str(), a_bone, a_bone, NULL,
		       message.c_str(), message.length());
}


void
BBSThreadRenderer::render_broken(guint res_num)
{
  const char *skin_fragment;
  if (res_num <= number_of_responses_got)
    skin_fragment = default_skin_res;
  else
    skin_fragment = default_skin_new_res;

  render_skin_fragment(skin_fragment, NULL, res_num,
		       NULL, NULL, _("[broken here]"), NULL,
		       _("[broken here]"), strlen(_("[broken here]")));
}


void
BBSThreadRenderer::render_response(guint res_num, Response &response,
				   const char *skin_fragment)
{
  /*
   * ̾
   */
  // ad-hocʽ񤭴ˤschemeؿƤ٤褦ˤɤ
  const char *name = response.get_name();
  if (name == NULL)
    name = "";
  size_t len = strlen(name);
  const char *mailto = response.get_mailto();
  char internal_mailto[PATH_MAX];
  int number;
  char *link;
  if (digits.match(name, len) && digits.match_len(0) == len
      && (number = parse_int(name, strlen(name))) != 0
      && (link = thread->get_url(number)) != NULL)
    {
      // ̾󤬿ξ̰
      if (mailto == NULL || *mailto == '\0')
	snprintf(internal_mailto, PATH_MAX, "&lt;FAKE&gt;%s", link);
      else
	snprintf(internal_mailto, PATH_MAX, "%s&lt;FAKE&gt;%s", mailto, link);
      free(link);
      mailto = internal_mailto;
    }
  else if (mailto == NULL)
    mailto = "";

  string name_buf;
  if (*mailto != '\0')
    {
      name_buf.append("<a href=\"mailto:");
      name_buf.append(mailto);
      if (strncmp(mailto, "&lt;FAKE&gt;", 12) != 0)
	name_buf.append("\"><b>");
      else
	name_buf.append("\" class=\"Secret\"><b>");
      name_buf.append(name);
      name_buf.append("</b></a>");
      name = name_buf.c_str();
    }
  else
    {
      name_buf.append("<b>");
      name_buf.append(name);
      name_buf.append("</b>");
      name = name_buf.c_str();
    }

  if (controller.get_thread_ui().show_mailto(thread))
    mailto = response.get_mailto();
  else
    mailto = NULL;


  /*
   * 
   */
  const char *date = response.get_date();
  if (date != NULL && *date == '\0')
    date = NULL;


  /*
   * ID
   */
  const char *id = response.get_id();

  string entire_id;
  string id_buffer;

  if (id != NULL)
    {
      size_t id_len = strlen(id);
      if (id_id.match(id, id_len))
	{
	  const char *match_begin0 = id_id.match_begin(0);
	  size_t match_len0 = id_id.match_len(0);
	  id_buffer.append(id, match_begin0 - id);

	  id_buffer.append("<a href=\"&lt;ID&gt;");
	  id_buffer.append(match_begin0, match_len0);
	  id_buffer.append("\" class=\"Secret\">");
	  id_buffer.append(match_begin0, match_len0);
	  id_buffer.append("</a>");

	  const char *match_end0 = id_id.match_end(0);
	  id_buffer.append(match_end0, id_len - (match_end0 - id));
	  id = id_buffer.c_str();
	  id_len = strlen(id);
	}

      char *url;
      if (be_id.match(id, id_len) && (url = thread->get_url(res_num)) != NULL)
	{
	  const char *match_begin0 = be_id.match_begin(0);
	  entire_id.append(id, match_begin0 - id);
	  entire_id.append("<a href=\"http://be.2ch.net/test/p.php?i=");
	  entire_id.append(be_id.match_begin(1), be_id.match_len(1));
	  entire_id.append("&u=d:");
	  entire_id.append(url);
	  if (*match_begin0 == '<')
	    {
	      // ͤ˥塼ĤθԤBe ID
	      // s|<BE:([0-9])+:([0-9])+>|<a href="http://be.2ch.net/test/p.php?i=\1
	      // &u=d:URL">Lv.\2</a>|ߤѴURLϥ쥹URLִ
	      entire_id.append("\">Lv.");
	    }
	  else
	    {
	      // ʳĤBe ID
	      // s|BE:([0-9])-(#)*|<a href="http://be.2ch.net/test/p.php?i=\1
	      // &u=d:URL">?\2</a>ߤѴURLϥ쥹URLִ
	      entire_id.append("\">?");
	    }
	  entire_id.append(be_id.match_begin(2), be_id.match_len(2));
	  entire_id.append("</a>");

	  const char *match_end0 = be_id.match_end(0);
	  entire_id.append(match_end0, id_len - (match_end0 - id));
	  free(url);
	  id = entire_id.c_str();
	}
    }


  // ʲʸ
  // Τ褦ˤʤäHTMLѡ󥰤κ˥󥯤ʤΤ
  // ˽񤭴Τ϶ڤɤʤΤǡѡˤޤȤƽ롣
#ifndef DISABLE_SCHEME_REWRITER
  SchemeManager &scheme = controller.get_scheme_manager();
  Cell *message = scheme.get_core().evalf0(scheme.rewrite_message.get_cell(),
					   "%s", response.get_message());
  const char *body;
  size_t body_len;
  if (message != NULL && message->is_string())
    {
      body = message->get_string();
      body_len = message->get_strlen();
    }
  else
    {
      body = response.get_message();
      body_len = strlen(body);
    }
#else
  const char *body = response.get_message();
  size_t body_len = strlen(body);
#endif

#if 1	// Beicon
  string body_with_be_icon;
  if ((controller.get_prefs().get_enable_be_icon()
       || controller.get_prefs().get_enable_inline_image())
      && be_icon.match(body, body_len))
    {
      size_t pred_len = be_icon.match_begin(0) - body;
      if (pred_len > 0)
	body_with_be_icon.append(body, pred_len);

      body_with_be_icon.append("<img src=\"http://img.2ch.net/ico/");
      body_with_be_icon.append(be_icon.match_begin(1), be_icon.match_len(1));
      body_with_be_icon.append("\">");

      size_t be_icon_len = be_icon.match_len(0);
      body_with_be_icon.append(be_icon.match_begin(0) + be_icon_len,
			       body_len - (be_icon_len + pred_len));

      body = body_with_be_icon.c_str();
      body_len = strlen(body);
    }
#endif

  render_skin_fragment(skin_fragment, NULL, res_num,
		       name, mailto, date, id, body, body_len);
}


void
BBSThreadRenderer::render_response(guint res_num, Response &response)
{
  const char *skin_fragment;
  if (res_num <= number_of_responses_got)
    skin_fragment = default_skin_res;
  else
    skin_fragment = default_skin_new_res;

  render_response(res_num, response, skin_fragment);
}


void
BBSThreadRenderer::prepare_for_popup()
{
  parse_text("<html><body><dl>", 16);
}


void
BBSThreadRenderer::render_popup_response(guint res_num, Response &response)
{
  render_response(res_num, response, default_skin_popup_res);
}


void
BBSThreadRenderer::render_popup_broken(guint res_num)
{
  render_skin_fragment(default_skin_popup_res, NULL, res_num,
		       NULL, NULL, _("[broken here]"), NULL,
		       _("[broken here]"), strlen(_("[broken here]")));
}


void
BBSThreadRenderer::prepare_for_popup_id(const gchar *title, const gchar *id)
{
  gchar *popup_title = g_strdup_printf(_("Responses by %s in this thread"), id);
  render_skin_fragment(default_skin_popup_id_header, popup_title, 0,
		       NULL, NULL, NULL, id, NULL, 0);
}


void
BBSThreadRenderer::render_popup_id(guint res_num, Response &response, int count)
{
  const char *skin = (count % 10) == 0
    ? default_skin_popup_id_res_nl : default_skin_popup_id_res;
  const char *message = response.get_message();
  size_t len = message != NULL ? strlen(message) : 0;
  render_skin_fragment(skin, NULL, res_num, response.get_name(),
		       response.get_mailto(), response.get_date(),
		       response.get_id(), message, len);
}


void
BBSThreadRenderer::render_popup_title(const gchar *title)
{
  render_skin_fragment(default_skin_popup_title, title, 0,
		       NULL, NULL, NULL, NULL,
		       NULL, 0);
}


void
BBSThreadRenderer::render_image(GtkWidget *image, const gchar *link)
{
  ochusha_hypertext_buffer_push_link_tag(buffer,
					 tag_manager.get_link_tag(),
					 link);
  ochusha_hypertext_view_append_widget(view, image, 0, 0);
  ochusha_hypertext_buffer_pop_link_tag(buffer,
					tag_manager.get_link_tag());
}


#define UTF8_DASHES	"[-\343\203\274\343\200\234\342\224\200]"
static Regexp response_numbers("^(" UTF8_DIGITS "+)(?:" UTF8_DASHES ")?(" UTF8_DIGITS "+)?");


// ХѥURLХѥURLѤꡢԹŪ
// ղäƤ뱣ɤURLΥԡ֤
// linkñΤǤޤȤURLä⥳ԡ֤
// linkä[-]ʾ硢쥹ֻȸʤ
// 쥹֤бURL֤
char *
BBSThreadRenderer::canon_link(BBSThread *thread, const char *link)
{
  if (strncmp("ID:", link, 3) == 0)
    return NULL;

  const char *fake = strstr(link, "<FAKE>");
  if (fake != NULL)
    link = fake + 6;

  URI uri(link);
  if (uri.is_valid() && uri.get_scheme() != NULL)
    return strdup(link);

  if (strncmp("<RES>", link, 5) == 0)
    {
      // <NUMBER/>б쥹URLξ
      // "<RES>쥹"ˤʤäƤ롣
      int num;
      if (sscanf(link + 5, "%d", &num) == 1)
	return thread->get_url(num);
      return NULL;	// ͤ
    }

  if (strstr(link, thread->get_id()) != NULL)
    {
      // link˥IDƱʸ󤬴ޤޤƤ硢ۤХѥ
      // URLİ㤤βǽϤ뤬̩˲ᤷƤӤϾʤΤ
      // Ψ˽
      int from = 0;
      int to = 0;
      BBSThread *dummy
	= thread->get_board()->lookup_bbs_thread_by_url(link, &from, &to);
      if (dummy != NULL)
	return thread->get_url(from, to);

      return NULL;	// н˺櫡
    }

  assert(OCHUSHA_IN_GDK_THREADS());
  if (response_numbers.match(link, strlen(link)))
    {
      int from = parse_int(response_numbers.match_begin(1),
			   response_numbers.match_len(1));
      if (from == 0)
	return NULL;
      assert(from > 0);
      int to = 0;
      if (from > 0)
	{
	  if (response_numbers.match_len(2) > 0)
	    to = parse_int(response_numbers.match_begin(2),
			   response_numbers.match_len(2));
	}
      return thread->get_url(from, to);
    }

  return NULL;
}


gboolean
BBSThreadRenderer::link_mouse_over(OchushaHypertextView *widget,
				   GdkEventMotion *event,
				   const gchar *link)
{
  return FALSE;
}


gboolean
BBSThreadRenderer::link_mouse_out(OchushaHypertextView *widget,
				  GdkEventMotion *event)
{
  return FALSE;
}


gboolean
BBSThreadRenderer::link_mouse_press(OchushaHypertextView *widget,
				    GdkEventButton *event,
				    const gchar *link)
{
  return FALSE;
}


gboolean
BBSThreadRenderer::link_mouse_release(OchushaHypertextView *widget,
				      GdkEventButton *event,
				      const gchar *link)
{
  return FALSE;
}


}
