/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: bbs_thread_view.cc,v 1.79 2009/01/02 03:08:56 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "bbs_thread_view.h"

#include "libochusha/bbs_thread.h"
#include "libochusha/constants.h"
#include "libochusha/misc_utils.h"
#include "libochusha/text_utils.h"

#include "controller.h"
#include "gtk_utils.h"
#include "image_manager.h"
#include "scheme_manager.h"
#include "tab_label.h"
#include "ui_constants.h"

#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <string>

//#define ENABLE_BENCHMARKING	1

#ifdef ENABLE_BENCHMARKING
# ifdef HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  undef ENABLE_BENCHMARKING
# endif
#endif


namespace Ochusha
{


#define THREAD_VIEW_UPDATE_PRIORITY		GDK_PRIORITY_REDRAW - 5
  //#define MAX_NUM_OF_RES_TO_APPEND_AT_ONCE	256
#define MAX_NUM_OF_RES_TO_APPEND_AT_ONCE	1024

#define THREAD_VIEW_EXTRA_WIDTH		30
#define POPUP_OFFSET			8

//#define DEBUG_FOCUS_NOTIFIER	1
//#define VERBOSE_DESTRUCTOR	1


BBSThreadViewSelectionData::BBSThreadViewSelectionData(BBSThread *thread,
						       GtkTextView *view,
						       bool is_static)
  : TextViewSelectionData(view, is_static),
    thread(thread),
    link(NULL),
    canonical_link(NULL)
{
}


BBSThreadViewSelectionData::~BBSThreadViewSelectionData()
{
#ifdef VERBOSE_DESTRUCTOR
  fprintf(stderr,
	  "BBSThreadViewSelectionData::~BBSThreadViewSelectionData(): %p\n",
	  this);
#endif

  if (link != NULL)
    free(link);

  if (canonical_link != NULL)
    free(canonical_link);
}


bool
BBSThreadViewSelectionData::deletable()
{
  return false;
}


bool
BBSThreadViewSelectionData::cuttable()
{
  return false;
}


bool
BBSThreadViewSelectionData::copiable()
{
  return link != NULL ? true : TextViewSelectionData::copiable();
}


bool
BBSThreadViewSelectionData::pastable(GtkClipboard *clipboard)
{
  return false;
}


bool
BBSThreadViewSelectionData::do_delete()
{
  return false;
}


bool
BBSThreadViewSelectionData::do_paste(GtkClipboard *clipboard)
{
  return false;
}


SelectionData *
BBSThreadViewSelectionData::snapshot()
{
  BBSThreadViewSelectionData *data
    = new BBSThreadViewSelectionData(thread, get_view(), false);

  data->set_link(link);

  return data;
}


#if GTK_CHECK_VERSION(2,10,0)
static GdkAtom internal_decorated_link_atom =
  gdk_atom_intern_static_string("ochusha/internal-decorated-link");
#else
static GdkAtom internal_decorated_link_atom =
  gdk_atom_intern("ochusha/internal-decorated-link", FALSE);
#endif


static const GtkTargetEntry link_targets[] = {
  { const_cast<gchar *>("ochusha/internal-decorated-link"),
    GTK_TARGET_SAME_APP, 1 },
  { const_cast<gchar *>("text/uri-list"), 0, 0 }
};


static const GtkTargetEntry internal_targets[] = {
  { const_cast<gchar *>("ochusha/internal-decorated-link"),
    GTK_TARGET_SAME_APP, 1 }
};


static const GtkTargetEntry empty_targets[] = {
};


const GtkTargetEntry *
BBSThreadViewSelectionData::get_targets()
{
  if (link != NULL && canonical_link == NULL)
    canonical_link = BBSThreadRenderer::canon_link(thread, link);

  // link != NULL && canonical_link == NULLꤦ롣
  // ξ硢ochusha/internal-decorated-linkΥåȤȸ

  return canonical_link != NULL
    ? link_targets
    : (link != NULL ? internal_targets : TextViewSelectionData::get_targets());
}


guint
BBSThreadViewSelectionData::get_n_targets()
{
  if (link != NULL && canonical_link == NULL)
    canonical_link = BBSThreadRenderer::canon_link(thread, link);

  // link != NULL && canonical_link == NULLꤦ롣
  // ξ硢ochusha/internal-decorated-linkΥåȤȸ

  return canonical_link != NULL
    ? G_N_ELEMENTS(link_targets)
    : (link != NULL
       ? G_N_ELEMENTS(internal_targets)
       : TextViewSelectionData::get_n_targets());
}


void
BBSThreadViewSelectionData::get_data(GtkClipboard *clipboard,
				     GtkSelectionData *data, guint info)
{
  if (canonical_link != NULL)
    {
      g_return_if_fail(info == 0);

      char *buf[2] = { canonical_link, NULL };
      gtk_selection_data_set_uris(data, buf);

      return;
    }

  TextViewSelectionData::get_data(clipboard, data, info);
}


void
BBSThreadViewSelectionData::set_link(const gchar *value)
{
  char *prev_link = link;

  link = value != NULL ? strdup(value) : NULL;

  if (prev_link != NULL)
    free(prev_link);

  if (canonical_link != NULL)
    {
      free(canonical_link);
      canonical_link = NULL;
    }
}


const char *
BBSThreadViewSelectionData::get_canonical_link()
{
  if (canonical_link == NULL)
    canonical_link = BBSThreadRenderer::canon_link(thread, link);

  return canonical_link;
}


BBSThreadViewFocusNotifier::BBSThreadViewFocusNotifier(BBSThreadViewBase &view,
						       GtkWidget *widget)
  : FocusNotifier(view.controller, widget, false),
    view(view)
{
}


BBSThreadViewFocusNotifier::~BBSThreadViewFocusNotifier()
{
#ifdef VERBOSE_DESTRUCTOR
  fprintf(stderr,
	  "BBSThreadViewFocusNotifier::~BBSThreadViewFocusNotifier(): selection_data=%p\n",
	  &view.selection_data);
#endif
  view.controller.clear_clipboard_internal(&view.selection_data);
}


gboolean
BBSThreadViewFocusNotifier::notify_focus_in()
{
#ifdef DEBUG_FOCUS_NOTIFIER
  fprintf(stderr, "BBSThreadViewFocusNotifier::notify_focus_in(): %p, %d\n",
	  this, has_pointer());
#endif
  view.controller.update_clipboard_internal(&view.selection_data);

  view.get_primogenitor_view()->notify_focus_in();

  return FALSE;
}


gboolean
BBSThreadViewFocusNotifier::notify_enter()
{
  FocusNotifier::notify_enter();
  return view.notify_enter();
}


gboolean
BBSThreadViewFocusNotifier::notify_leave()
{
  FocusNotifier::notify_leave();
  return view.notify_leave();
}


void
BBSThreadViewFocusNotifier::notify_unmap()
{
  view.controller.clear_clipboard_internal(&view.selection_data);
  view.notify_unmap();
}


BBSThreadViewBase::BBSThreadViewBase(Controller &controller,
				     TabLabel *tab_label, BBSThread *thread,
				     GtkWidget *window,
				     BBSThreadViewBase *parent_view)
  : View(tab_label, window),
    controller(controller),
    thread(thread),
    buffer(ochusha_hypertext_buffer_new(NULL)),
    view(OCHUSHA_HYPERTEXT_VIEW(ochusha_hypertext_view_new(buffer))),
    renderer(controller, thread, buffer, view),
    parent_view(parent_view),
    window_title(NULL),
    combo_title(NULL),
    thread_url(NULL),
    message_id(0),
    selection_done_id(0),
    selection_data(thread, GTK_TEXT_VIEW(view), true),
    focus_notifier(*this, GTK_WIDGET(view)),
    popup_link(NULL),
    popup_view(NULL),
    popup_view_destroy_id(0)
{
  g_signal_connect(view, "popup_menu",
		   reinterpret_cast<GCallback>(popup_menu_cb), this);
  g_signal_connect(view, "populate_popup",
		   reinterpret_cast<GCallback>(populate_popup_cb), this);
  g_signal_connect(view, "link_mouse_over",
		   reinterpret_cast<GCallback>(link_mouse_over_cb), this);
  g_signal_connect(view, "link_mouse_out",
		   reinterpret_cast<GCallback>(link_mouse_out_cb), this);
  g_signal_connect(view, "link_mouse_press",
		   reinterpret_cast<GCallback>(link_mouse_press_cb), this);
  g_signal_connect(view, "link_mouse_release",
		   reinterpret_cast<GCallback>(link_mouse_release_cb), this);
  g_signal_connect(view, "copy_clipboard",
		   reinterpret_cast<GCallback>(copy_clipboard_cb), this);
  g_signal_connect(view, "start_interactive_search",
		   reinterpret_cast<GCallback>(start_interactive_search_cb),
		   this);

  new TextViewKeyPressHandler(GTK_TEXT_VIEW(view), controller);
}


BBSThreadViewBase::~BBSThreadViewBase()
{
  assert(OCHUSHA_IN_GDK_THREADS());

#ifdef VERBOSE_DESTRUCTOR
  fprintf(stderr, "BBSThreadViewBase destructed: %p\n", this);
#endif

  if (popup_view != NULL)
    {
      assert(popup_view_destroy_id != 0);

      BBSThreadViewBase *view = popup_view;

      g_signal_handler_disconnect(view->get_widget(), popup_view_destroy_id);
      popup_view_destroy_id = 0;

      free(popup_link);
      popup_link = NULL;
      popup_view = NULL;

      view->hide(true);
    }



  GDK_THREADS_LEAVE();

  disconnect_all();

  GDK_THREADS_ENTER();


  g_object_unref(buffer);

  if (message_id != 0)
    {
      GtkStatusbar *bar = controller.get_statusbar();
      gtk_statusbar_remove(bar, controller.get_link_context_id(), message_id);
    }

  if (thread_url != NULL)
    free(thread_url);

  if (combo_title != NULL)
    free(combo_title);

  if (window_title != NULL)
    g_free(window_title);

  g_return_if_fail(popup_link == NULL);
}


gboolean
BBSThreadViewBase::internal_decorated_link_available(GtkClipboard *clipboard)
{
  g_return_val_if_fail(GTK_IS_CLIPBOARD(clipboard), FALSE);

  return gtk_clipboard_wait_is_target_available(clipboard,
						internal_decorated_link_atom);
}


BBSThreadViewSelectionData *
BBSThreadViewBase::get_bbs_thread_view_selection_data_from_clipboard(GtkClipboard *clipboard)
{
  g_return_val_if_fail(internal_decorated_link_available(clipboard), NULL);

  SelectionData *data = SelectionData::get_from(clipboard);
  g_return_val_if_fail(data != NULL, NULL);

  BBSThreadViewSelectionData *selection_data
    = dynamic_cast<BBSThreadViewSelectionData *>(data);
  g_return_val_if_fail(selection_data != NULL, NULL);

  return selection_data;
}


BBSThreadViewBase *
BBSThreadViewBase::get_primogenitor_view()
{
  if (parent_view == NULL)
    return this;

  BBSThreadViewBase *ancestor_view = parent_view;
  while (ancestor_view->parent_view != NULL)
    ancestor_view = ancestor_view->parent_view;

  return ancestor_view;
}


const gchar *
BBSThreadViewBase::get_combo_title()
{
  if (combo_title != NULL)
    return combo_title;

  if (static_cast<BBSThread *>(thread) == NULL || thread->is_tmp_thread())
    return _("Anonymous Thread");

  const gchar *title = thread->get_title();
  g_return_val_if_fail(title != NULL, _("Anonymous Thread"));

  SchemeManager &scheme = controller.get_scheme_manager();
  Cell *func = scheme.abbreviate_thread_title.get_cell();
  if (func != NULL)
    {
      Cell *result = scheme.get_core().evalf0(func, "%s", title);
      if (result != NULL && result->is_string())
	{
	  combo_title = strndup(result->get_string(), result->get_strlen());
	  return combo_title;
	}
    }

  combo_title = strdup(title);
  return title;	// NULLå򥵥ܤ롣
}


const gchar *
BBSThreadViewBase::get_menu_title()
{
  return (thread != NULL && !thread->is_tmp_thread())
    ? thread->get_title() : _("Anonymous Thread");
}


const gchar *
BBSThreadViewBase::get_window_title()
{
  if (parent_view != NULL)
    return get_primogenitor_view()->get_window_title();

  if (window_title != NULL)
    return window_title;

  const gchar *thread_title = NULL;
  const gchar *board_name = NULL;
  if (thread != NULL)
    {
      thread_title = thread->get_title();
      board_name = thread->get_board()->get_name();
    }

  if (thread_title == NULL)
    thread_title = _("Anonymous Thread");

  if (board_name == NULL)
    board_name = _("Anonymous Board");

  window_title = g_strdup_printf(_("%s @ %s - Ochusha"),
				 thread_title, board_name);
  return window_title;
}


const gchar *
BBSThreadViewBase::get_url()
{
  if (parent_view != NULL)
    return get_primogenitor_view()->get_url();

  if (thread_url != NULL)
    return thread_url;

  thread_url = thread->get_url();
  return thread_url;
}


void
BBSThreadViewBase::grab_focus()
{
  gtk_widget_grab_focus(GTK_WIDGET(view));
}


bool
BBSThreadViewBase::contains(GtkWidget *child)
{
  return child == get_widget() || child == GTK_WIDGET(view)
    || (parent_view != NULL && parent_view->contains(child));
}


const gchar *
BBSThreadViewBase::get_search_title()
{
  return _("Find in Thread");
}


OchushaSearchable *
BBSThreadViewBase::get_search_target()
{
  return OCHUSHA_SEARCHABLE(get_primogenitor_view()->view);
}


void
BBSThreadViewBase::notify_closed()
{
  g_return_if_fail(parent_view == NULL);

  BBSThreadView *thread_view = dynamic_cast<BBSThreadView *>(this);
  g_return_if_fail(thread_view != NULL);

  thread_view->remember_current_view_position();
}


bool
BBSThreadViewBase::bookmarkable()
{
  return true;
}


bool
BBSThreadViewBase::bookmark()
{
  controller.bookmark(thread);
  return true;
}


bool
BBSThreadViewBase::refreshable()
{
  return true;
}


bool
BBSThreadViewBase::refresh()
{
  controller.refresh(thread);
  return true;
}


bool
BBSThreadViewBase::has_info()
{
  return true;
}


bool
BBSThreadViewBase::show_info()
{
  controller.get_thread_ui().open_properties(thread);
  return true;
}


char *
BBSThreadViewBase::get_current_view_url()
{
  g_return_val_if_fail(parent_view == NULL, NULL);

  BBSThreadView *thread_view = dynamic_cast<BBSThreadView *>(this);
  g_return_val_if_fail(thread_view != NULL, NULL);

  return thread->get_url(thread_view->estimate_current_view_position());
}


GtkWindow *
BBSThreadViewBase::get_window()
{
  if (GTK_IS_WINDOW(get_widget()))
    return GTK_WINDOW(get_widget());
  return controller.get_toplevel();
}


void
BBSThreadViewBase::notify_unmap()
{
  if (popup_view != NULL)
    {
      BBSThreadViewBase *view = popup_view;
      view->hide(true);

      assert(popup_view == NULL);
      assert(popup_link == NULL);
    }
}


gboolean
BBSThreadViewBase::notify_enter()
{
  return FALSE;
}


gboolean
BBSThreadViewBase::notify_leave()
{
  return FALSE;
}


void
BBSThreadViewBase::popup_destroy_cb(GtkWidget *widget, BBSThreadViewBase *view)
{
  // : destroyʥϥɥϿδطˤꡢ
  // View::get_view(widget)ϤViewϴdestructƤ롣
  // view->popup_viewλؤ֥ȤϡλǥߤˤʤäƤ롣
  assert(OCHUSHA_IN_GDK_THREADS());
  assert(view->popup_link != NULL);
  assert(view->popup_view_destroy_id != 0);

  free(view->popup_link);
  view->popup_link = NULL;

  view->popup_view = NULL;
  view->popup_view_destroy_id = 0;

  view->hide(false);
}


gboolean
BBSThreadViewBase::link_mouse_over_cb(OchushaHypertextView *widget,
				      GdkEventMotion *event, const gchar *link,
				      BBSThreadViewBase *view)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  view->selection_data.set_link(link);
  view->controller.update_clipboard_internal(&view->selection_data);

  if (view->message_id != 0)
    {
      GtkStatusbar *bar = view->controller.get_statusbar();
      gtk_statusbar_remove(bar, view->controller.get_link_context_id(),
			   view->message_id);
      view->message_id = 0;
    }

  if (strncmp(link, "<RES>", 5) == 0)
    return FALSE;

  if (view->popup_link != NULL && strcmp(view->popup_link, link) != 0)
    {
      if (view->popup_view != NULL)
	{
	  BBSThreadViewBase *popup_view = view->popup_view;
	  popup_view->hide(true);

	  assert(view->popup_view == NULL);
	  assert(view->popup_link == NULL);
	}
      else
	{
	  free(view->popup_link);
	  view->popup_link = NULL;
	}
    }

  if (view->get_response_cursor() != NULL)
    {
      // ݥåץåפ䥿ȥݥåץå׻ˤϲ⤹ɬפʤΤ
      // 
      if (view->popup_view == NULL)
	{ // ߥݥåץåɽǤʤ
	  assert(view->popup_view_destroy_id == 0);

	  char *popup_link = BBSThreadRenderer::canon_link(view->thread, link);
	  if (popup_link != NULL)
	    {
	      Controller &controller = view->controller;
	      int from;
	      int to;
	      if (view->thread->parse_url(popup_link, &from, &to))
		{
		  if (from == 0)
		    {
		      from = 1;
		      to = 1;
		    }
		  else if (from < 0)
		    {
		      to = view->thread->get_number_of_responses_got();
		      from += to;
		    }
		  view->popup_view
		    = new BBSThreadPopupView(controller, view, from, to);
		}
	      else
		{
		  BBSTable &table = controller.get_bbs_table();
		  BulletinBoard *board = table.guess_board(popup_link);
		  if (board != NULL)
		    {
		      GDK_THREADS_LEAVE();

		      board->read_threadlist_xml(controller.get_repository());

		      GDK_THREADS_ENTER();


		      char title_buf[256];
		      BBSThread *thread
			= board->lookup_bbs_thread_by_url(popup_link);
		      if (thread != NULL && thread->get_title() != NULL)
			snprintf(title_buf, 256, _("%s@%s"),
				 thread->get_title(), board->get_name());
		      else if (table.lookup_board_by_url(popup_link) == NULL)
			snprintf(title_buf, 256, _("Unknown thread@%s"),
				 board->get_name());
		      else
			snprintf(title_buf, 256, "%s", board->get_name());
		      view->popup_view
			= new BBSThreadPopupView(controller, view,
						 title_buf, popup_link);
		    }
		  else if (controller.get_prefs().get_enable_popup_image()
			   && controller.get_image_manager().is_image_link(popup_link))
		    {
		      int width
			= controller.get_prefs().get_popup_image_width();
		      int height
			= controller.get_prefs().get_popup_image_height();
		      GtkWidget *image
			= controller.get_image_manager().get_widget(popup_link,
								    width,
								    height,
								    true);
		      if (image != NULL)
			{
			  view->popup_view
			    = new BBSThreadPopupView(controller, view,
						     image, popup_link);
			}
		    }
		}
	    }
	  else
	    {
	      string tmp_link;
	      if (strncmp("ID:", link, 3) == 0)
		tmp_link.append("<ID>");
	      tmp_link.append(link);

	      if (strncmp(tmp_link.c_str(), "<ID>", 4) == 0
		  && view->controller.get_prefs().get_enable_popup_id())
		{
		  view->popup_view
		    = new BBSThreadPopupView(view->controller, view,
					     tmp_link.c_str() + 4);
		  popup_link = strdup(tmp_link.c_str());
		}
	    }

	  if (view->popup_view != NULL)
	    {
	      view->popup_link = popup_link;
	      BBSThreadViewBase *popup_view = view->popup_view;
	      popup_view->show();
	      view->popup_view_destroy_id
		= g_signal_connect(popup_view->get_widget(), "destroy",
				   reinterpret_cast<GCallback>(popup_destroy_cb),
				   view);
	    }
	  else if (popup_link != NULL)
	    free(popup_link);
	}
      else
	{ // ɽΥݥåץåפƱ̤٤󥯤ä
	  BBSThreadViewBase *popup_view = view->popup_view;
	  popup_view->show();
	}
    }


  // ʲlink򥹥ơСɽ
  char tmp_buffer[PATH_MAX];
  snprintf(tmp_buffer, PATH_MAX, "%s", link);
  char *tmp_link = strstr(tmp_buffer, "<FAKE>");

  if (tmp_link != NULL)
    *tmp_link = '\0';
  else if (strncmp(tmp_buffer, "<ID>", 4) == 0)
    snprintf(tmp_buffer, PATH_MAX, "%s", link + 4);

  tmp_link = strchr(tmp_buffer, '\n');
  if (tmp_link != NULL)
    *tmp_link = '\0';

  GtkStatusbar *bar = view->controller.get_statusbar();
  view->message_id
    = gtk_statusbar_push(bar, view->controller.get_link_context_id(),
			 tmp_buffer);

  return FALSE;
}


gboolean
BBSThreadViewBase::link_mouse_out_cb(OchushaHypertextView *widget,
				     GdkEventMotion *event,
				     BBSThreadViewBase *view)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  if (view->message_id != 0)
    {
      GtkStatusbar *bar = view->controller.get_statusbar();
      gtk_statusbar_remove(bar, view->controller.get_link_context_id(),
			   view->message_id);
      view->message_id = 0;
    }

  if (view->popup_view != NULL)
    {
      BBSThreadViewBase *popup_view = view->popup_view;
      popup_view->hide(false);
    }

  if (view->selection_done_id == 0)
    {
      view->selection_data.set_link(NULL);
      view->controller.update_clipboard_internal(&view->selection_data);
    }

  return FALSE;
}


gboolean
BBSThreadViewBase::popup_context_menu(const gchar *link,
				      guint button, guint32 event_time)
{
  const gchar *menu_path;
  if (link == NULL || strncmp(link, "mailto:", 7) == 0)
    {
#if 0
      // ͳΥݥåץåץ˥塼ƤȱåǽФƤ
      // GtkTextViewͳΥݥåץåץ˥塼β¤ǤƤ̯˰ۤʤ롣
      // åΥ٥Ȥϥɥ뤷ƤĤ⼫Υ˥塼ȤΤ
      // 񤷤ʤʤ顢link_mouse_press/link_mouse_releaseŪ
      // OchushaHypertextViewΥ٥Ȥ
      // button_press_event/button_release_eventΥϥɥ餫ȯԤ
      // ᡢƱ٥ȤꤷƤޤΤ꤬ġĤ褦ʵ
      // ֤ΤG_SIGNA_RUN_LASTƤ
      // button_press_event/button_release_eventʥ٥Ȥ
      // OchushaHypertextViewǥϥɥǤʤ褦ʡ
      // ǽ

      // 괺ƱƥȤǰ㤦˥塼ФΤϷʤΤǡ
      // ̵뤷Ƥ
      menu_path = "/ThreadViewPopupMenu";
#else
      return FALSE;
#endif
    }
  else if (strncmp(link, "<RES>", 5) == 0)
    menu_path = "/ThreadResponseNumberPopupMenu";
  else if (strncmp(link, "<ID>", 4) == 0)
    menu_path = "/ThreadIDPopupMenu";
  else if (controller.get_image_manager().is_image_link(link))
    menu_path = "/ThreadImagePopupMenu";
  else if (controller.get_thread_ui().is_link_in_thread(thread, link)
	   || controller.is_bbs_link(link))
    menu_path = "/ThreadBBSLinkPopupMenu";
  else
    menu_path = "/ThreadLinkPopupMenu";

#if 0
  selection_data.set_link(link);
  controller.update_clipboard_internal(&selection_data);
#endif

  g_return_val_if_fail(selection_done_id == 0, TRUE);

  GtkUIManager *manager = controller.get_ui_manager();
  GtkWidget *menu_widget = gtk_ui_manager_get_widget(manager, menu_path);
  g_return_val_if_fail(GTK_IS_MENU(menu_widget), TRUE);

  selection_done_id
    = g_signal_connect(menu_widget, "selection-done",
		       reinterpret_cast<GCallback>(selection_done_cb),
		       this);

  gtk_menu_popup(GTK_MENU(menu_widget), NULL, NULL, NULL, NULL,
		 button, event_time);

  return TRUE;
}


gboolean
BBSThreadViewBase::popup_menu_cb(GtkWidget *widget, BBSThreadViewBase *view)
{
  return view->popup_context_menu(view->selection_data.get_link(),
				  0, gtk_get_current_event_time());
}


void
BBSThreadViewBase::populate_popup_cb(GtkTextView *text_view,
				     GtkMenu *popup_menu,
				     BBSThreadViewBase *view)
{
  view->controller.get_scheme_manager()
    .attach_external_text_tools_menu(popup_menu, &view->selection_data);

  view->controller.get_view_manager().attach_thread_view_menu(popup_menu);

  GtkMenuShell *shell = GTK_MENU_SHELL(popup_menu);
  BBSThreadUI &thread_ui = view->controller.get_thread_ui();

  GtkWidget *separator = gtk_separator_menu_item_new();
  gtk_widget_show(separator);
  gtk_menu_shell_prepend(shell, separator);

  GtkWidget *menu = gtk_action_create_menu_item(thread_ui.goto_bottom);
  gtk_widget_show(menu);
  gtk_menu_shell_prepend(shell, menu);

  menu = gtk_action_create_menu_item(thread_ui.jump_to_bookmark);
  gtk_widget_show(menu);
  gtk_menu_shell_prepend(shell, menu);

  menu = gtk_action_create_menu_item(thread_ui.goto_top);
  gtk_widget_show(menu);
  gtk_menu_shell_prepend(shell, menu);

  menu = gtk_action_create_menu_item(thread_ui.goto_next);
  gtk_widget_show(menu);
  gtk_menu_shell_prepend(shell, menu);

  menu = gtk_action_create_menu_item(thread_ui.goto_prev);
  gtk_widget_show(menu);
  gtk_menu_shell_prepend(shell, menu);

  assert(view->selection_done_id == 0);

  view->selection_done_id
    = g_signal_connect(popup_menu, "selection-done",
		       reinterpret_cast<GCallback>(selection_done_cb), view);
}


void
BBSThreadViewBase::selection_done_cb(GtkWidget *widget,
				     BBSThreadViewBase *view)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  g_return_if_fail(view->selection_done_id != 0);
  g_signal_handler_disconnect(widget, view->selection_done_id);

  view->selection_done_id = 0;
  view->selection_data.set_link(NULL);
}


gboolean
BBSThreadViewBase::link_mouse_press_cb(OchushaHypertextView *widget,
				       GdkEventButton *event, const gchar *link,
				       BBSThreadViewBase *view)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  view->selection_data.set_link(NULL);

  g_return_val_if_fail(link != NULL, FALSE);

  if (event->button == 3)
    {
      view->selection_data.set_link(link);
      view->controller.update_clipboard_internal(&view->selection_data);

      return view->popup_context_menu(link, event->button, event->time);
    }

  if (strncmp(link, "<RES>", 5) == 0
      || strncmp(link, "<ID>", 4) == 0
      || strncmp(link, "mailto:", 7) == 0)
    return FALSE;

  view->selection_data.set_link(link);

  return FALSE;
}


gboolean
BBSThreadViewBase::link_mouse_release_cb(OchushaHypertextView *widget,
					 GdkEventButton *event,
					 const gchar *link,
					 BBSThreadViewBase *view)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  const char *last_event_link = view->get_last_event_link();
  if (last_event_link == NULL)
    return FALSE;

  if (strcmp(last_event_link, link) == 0 && event->button != 3)
    {
      Preferences &prefs = view->controller.get_prefs();
      bool in_tab = (event->button == 2) ^ prefs.get_default_open_in_tab();
      char *real_link = BBSThreadRenderer::canon_link(view->thread, link);

      // ϥ󥯤򥿥֤ǳˤΤ̣߰ġ
      view->selection_data.set_link(NULL);

      if (real_link != NULL)
	{
	  view->controller.open(real_link, true, in_tab, false,
				event->button == 1);
	  free(real_link);
	}

      return TRUE;
    }

  if (event->button != 3)
    view->selection_data.set_link(NULL);

  return FALSE;
}


void
BBSThreadViewBase::copy_clipboard_cb(OchushaHypertextView *widget,
				     BBSThreadViewBase *view)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  gchar *text = ochusha_hypertext_view_get_selected_text(widget);
  if (text != NULL)
    {
      view->controller.copy(text);
      g_free(text);
    }
}


gboolean
BBSThreadViewBase::start_interactive_search_cb(OchushaHypertextView *widget,
					       OchushaSearchActionType action,
					       BBSThreadViewBase *view)
{
  view->get_primogenitor_view()->request_search();
  return TRUE;
}


BBSThreadPopupView::BBSThreadPopupView(Controller &controller,
				       BBSThreadViewBase *parent_view,
				       guint from, guint to)
  : BBSThreadViewBase(controller, NULL, parent_view->get_thread(),
		      gtk_window_new(GTK_WINDOW_POPUP), parent_view),
    response_cursor(parent_view->get_response_cursor()),
    popup_id(0),
    popdown_id(0),
    pointed(false)
{
  // 쥹fromtoޤǤΥ쥹ɽݥåץåפ(Ǥڤꤢ)
  initialize_common();

  ResponseCursor *cursor = response_cursor;
  assert(cursor != NULL);

  int lock_id = cursor->trylock();
  if (lock_id != 0)
    {
      renderer.prepare_for_popup();
      if (from > to)
	to = from;

      int num_res
	= controller.get_prefs().get_maximum_number_of_popup_responses();

      while (from <= to && num_res > 0 && cursor->has_response(from, lock_id))
	{
	  Response &response = cursor->get_response(from, lock_id);
	  if (!response.is_deleted())
	    {
	      if (!response.is_broken())
		renderer.render_popup_response(from, response);
	      else
		renderer.render_popup_broken(from);
	      --num_res;
	    }
	  ++from;
	}

      cursor->unlock(lock_id);
    }
}


BBSThreadPopupView::BBSThreadPopupView(Controller &controller,
				       BBSThreadViewBase *parent_view,
				       const gchar *id)
  : BBSThreadViewBase(controller, NULL, parent_view->get_thread(),
		      gtk_window_new(GTK_WINDOW_POPUP), parent_view),
    response_cursor(parent_view->get_response_cursor()),
    popup_id(0),
    popdown_id(0),
    pointed(false)
{
  // IDidʥ쥹֤򽸤ɽݥåץåפ롣
  initialize_common();
  g_return_if_fail(id != NULL);

  ResponseCursor *cursor = response_cursor;
  assert(cursor != NULL);

  int lock_id = cursor->trylock();
  if (lock_id != 0)
    {
      renderer.prepare_for_popup_id(thread->get_title(), id);
      int count = 0;
      for (guint res_num = 1; cursor->has_response(res_num, lock_id); ++res_num)
	{
	  Response &response = cursor->get_response(res_num, lock_id);
	  if (!response.is_deleted())
	    {
	      const char *res_id = response.get_id();
	      if (res_id != NULL && strstr(res_id, id) != NULL)
		renderer.render_popup_id(res_num, response, ++count);
	    }
	}

      cursor->unlock(lock_id);
    }
}


BBSThreadPopupView::BBSThreadPopupView(Controller &controller,
				       BBSThreadViewBase *parent_view,
				       GtkWidget *image, const gchar *link)
  : BBSThreadViewBase(controller, NULL, parent_view->get_thread(),
		      gtk_window_new(GTK_WINDOW_POPUP), parent_view),
    response_cursor(NULL),
    popup_id(0),
    popdown_id(0),
    pointed(false)
{
  // linkimageɽݥåץåפ롣
  initialize_common();
  renderer.render_image(image, link);
}


BBSThreadPopupView::BBSThreadPopupView(Controller &controller,
				       BBSThreadViewBase *parent_view,
				       const gchar *title, const gchar *link)
  : BBSThreadViewBase(controller, NULL, parent_view->get_thread(),
		      gtk_window_new(GTK_WINDOW_POPUP), parent_view),
    response_cursor(NULL),
    popup_id(0),
    popdown_id(0),
    pointed(false)
{
  // 쥿̾ɽݥåץåפ롣
  initialize_common();
  renderer.render_popup_title(title);
}


BBSThreadPopupView::~BBSThreadPopupView()
{
  assert(OCHUSHA_IN_GDK_THREADS());

#ifdef VERBOSE_DESTRUCTOR
  fprintf(stderr, "BBSThreadPopupView(@%p) destructed\n", this);
#endif

  if (popup_id != 0)
    g_source_remove(popup_id);

  if (popdown_id != 0)
    g_source_remove(popdown_id);


  GDK_THREADS_LEAVE();

  response_cursor = NULL;

  GDK_THREADS_ENTER();
}


void
BBSThreadPopupView::initialize_common()
{
  assert(OCHUSHA_IN_GDK_THREADS());
  assert(parent_view != NULL);

  GtkWindow *parent = parent_view->get_window();
  GtkWindow *window = GTK_WINDOW(get_widget());
  gtk_window_set_resizable(window, FALSE);
  gtk_window_set_transient_for(window, parent);

  GtkWidget *frame = gtk_frame_new(NULL);
  gtk_widget_show(frame);

  gtk_container_add(GTK_CONTAINER(window), frame);

  GtkWidget *widget = GTK_WIDGET(view);
  gtk_widget_show(widget);
  gtk_container_add(GTK_CONTAINER(frame), widget);

  gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(view), GTK_WRAP_NONE);

  g_signal_connect(widget, "size_request",
		   reinterpret_cast<GCallback>(size_request_cb), this);
}


gboolean
BBSThreadPopupView::notify_enter()
{
  assert(OCHUSHA_IN_GDK_THREADS());
  assert(popup_id == 0);

  pointed = true;

  if (popdown_id != 0)
    {
      g_source_remove(popdown_id);
      popdown_id = 0;
    }

  return FALSE;
}


gboolean
BBSThreadPopupView::notify_leave()
{
  assert(OCHUSHA_IN_GDK_THREADS());
  assert(popup_id == 0);

  pointed = false;

  if (popdown_id == 0)
    hide(false);

  return FALSE;
}


void
BBSThreadPopupView::size_request_cb(GtkWidget *widget,
				    GtkRequisition *requisition,
				    BBSThreadPopupView *view)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  if (requisition->width <= 0 || requisition->height <= 0)
    return;

  gint width;
  if (view->response_cursor != NULL)
    width = requisition->width + THREAD_VIEW_EXTRA_WIDTH;
  else
    width = requisition->width;
  gint height = requisition->height;

  GtkWindow *window = GTK_WINDOW(view->get_widget());

  gtk_widget_set_size_request(widget, width, height);
  gtk_window_resize(window, width, height);

  GdkScreen *screen = gtk_widget_get_screen(view->get_widget());
  GdkDisplay *display = gdk_screen_get_display(screen);
  gint x_pos;
  gint y_pos;
  gdk_display_get_pointer(display, NULL, &x_pos, &y_pos, NULL);
  x_pos += POPUP_OFFSET;
  y_pos -= (height + POPUP_OFFSET);
  y_pos = y_pos > 0 ? y_pos : 0;

  gint screen_width = gdk_screen_get_width(screen);
  if (screen_width > 0 && (x_pos + width) > screen_width)
    {
      x_pos = screen_width - width;
      if (x_pos < 0)
	x_pos = 0;
    }

  gtk_window_move(window, x_pos, y_pos);
}


ResponseCursor *
BBSThreadPopupView::get_response_cursor()
{
  return response_cursor;
}


gboolean
BBSThreadPopupView::show_now(BBSThreadPopupView *view)
{
  GDK_THREADS_ENTER();

  view->popup_id = 0;
  gtk_widget_show(view->get_widget());

  GDK_THREADS_LEAVE();

  return FALSE;
}


gboolean
BBSThreadPopupView::hide_now(BBSThreadPopupView *view)
{
  GDK_THREADS_ENTER();

  if (view->popdown_id != 0)
    {
      g_source_remove(view->popdown_id);
      view->popdown_id = 0;
    }

  if (!view->has_popup() && !view->has_popup_menu())
    {
      // ҥݥåץåפʤ

      gtk_widget_destroy(view->get_widget());
      // λviewdeleteƤ롣
    }
  // ҥݥåץåפ̤ΤäԤġ

  GDK_THREADS_LEAVE();

  return FALSE;
}


void
BBSThreadPopupView::show()
{
  assert(OCHUSHA_IN_GDK_THREADS());

  if (popdown_id != 0)
    {
      g_source_remove(popdown_id);
      popdown_id = 0;

      assert(popup_id == 0);
    }
  else if (popup_id != 0)
    return;

  // gtk_widget_map()Ƥʤȡǽsize_requestʥ
  // OchushaHypertextView(ʤGtkTextView)Υ{ 0, 0 }
  // ˤʤäƤޤθäƤallocationΤľˤʤ롣
  // η̡ݥåץåפФƸ줷
  gtk_widget_map(GTK_WIDGET(view));

  guint popup_delay = controller.get_prefs().get_popup_delay();
  if (popup_delay > 0)
    popup_id = g_timeout_add(popup_delay,
			     reinterpret_cast<GSourceFunc>(show_now), this);
  else
    popup_id = g_idle_add(reinterpret_cast<GSourceFunc>(show_now), this);
}


void
BBSThreadPopupView::hide(bool now)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  if (popup_id != 0 || now)
    {
      gtk_widget_destroy(get_widget());
      return;
    }

  if (popdown_id != 0)
    return;

  if (pointed)
    return;

  guint popdown_delay = controller.get_prefs().get_popdown_delay();
  if (popdown_delay > 0)
    popdown_id = g_timeout_add(popdown_delay,
			       reinterpret_cast<GSourceFunc>(hide_now), this);
  else
    popdown_id = g_idle_add(reinterpret_cast<GSourceFunc>(hide_now), this);
}


BBSThreadView::BBSThreadView(Controller &controller, TabLabel *tab_label,
			     BBSThread *thread)
  : BBSThreadViewBase(controller, tab_label, thread,
		      gtk_scrolled_window_new(NULL, NULL), NULL),
    response_cursor(NULL),
    number_of_responses_shown(0),
    append_responses_id(0),
    initial_view_position(0),
    my_own_bookmark(NULL),
    my_own_bookmark_position(0),
    post_window(NULL),
    post_window_destroy_id(0),
    auto_refresher_id(0),
    previous_interval(0),
    previous_number_of_responses_got(0),
    enable_transparent_a_bone(controller.get_prefs().get_enable_transparent_a_bone()),
    disable_a_bone(controller.get_prefs().get_disable_a_bone()),
    enable_chain_a_bone(controller.get_prefs().get_enable_chain_a_bone())
{
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(get_widget()),
				 GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_widget_show(GTK_WIDGET(view));
  gtk_container_add(GTK_CONTAINER(get_widget()), GTK_WIDGET(view));
}


BBSThreadView::~BBSThreadView()
{
  assert(OCHUSHA_IN_GDK_THREADS());

#ifdef VERBOSE_DESTRUCTOR
  fprintf(stderr, "BBSThreadView destructed: %p\n", this);
#endif


  if (auto_refresher_id != 0)
    g_source_remove(auto_refresher_id);

  if (append_responses_id != 0)
    g_source_remove(append_responses_id);

  // 񤭹ߥɥƱĤȤ⤢뤬̯
  if (post_window_destroy_id != 0)
    g_signal_handler_disconnect(post_window, post_window_destroy_id);


  GDK_THREADS_LEAVE();

  response_cursor = NULL;

  GDK_THREADS_ENTER();


  thread->set_number_of_responses_read(thread->get_number_of_responses_got());
  GtkWidget *board_widget
    = controller.get_view_manager().get_current_board_widget();
  if (board_widget != NULL)
    gtk_widget_queue_draw(board_widget);
}


// Viewvirtualμ
bool
BBSThreadView::move_view(const char *target_url)
{
  int from = 0;
  if (!thread->parse_url(target_url, &from))
    return false;

  controller.get_thread_ui().notify_thread_view_state(this, thread);

  if (from <= 0 || records.size() < static_cast<size_t>(from))
    return true;

  // fromΥ쥹ɽ֤viewportư
  GtkTextMark *mark
    = ochusha_hypertext_buffer_create_mark_at_offset(buffer, "view-position",
						     records[from - 1].offset);
  ochusha_hypertext_view_scroll_to_mark(view, mark);

  return true;
}
// Viewvirtualμ


ResponseCursor *
BBSThreadView::get_response_cursor()
{
  return response_cursor;
}


// 鲶ѤϢ

// OchushaHypertextBuffer/Viewoffset֤widget
// records򹹿
void
BBSThreadView::insert_widget_at_offset(GtkWidget *widget, int offset)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  GSList *tags = buffer->tags;
  buffer->tags = NULL;
  GtkTextTag *bookmark_tag = controller.get_tag_manager().get_my_bookmark_tag();
  ochusha_hypertext_buffer_push_tag(buffer, bookmark_tag);
  gint b_off = ochusha_hypertext_view_insert_widget_at_offset(view, widget,
							      0, 20, offset);
  ochusha_hypertext_buffer_pop_tags(buffer, bookmark_tag);
  buffer->tags = tags;

  g_return_if_fail(b_off >= 0);

  assert(offset == b_off);

  // recordsϥ쥹ֹбƤΤǡoffsetˤĤƾ硣
  // binary searchƴطΤ֤ƬõƤ
  // offsetĴ򤹤ȤŬ;ϤϤ뤬
  // 쥹NˤĤO(N)ʤΤݤʤȤϤʤǤ
  vector<ResponseRecord>::iterator iter = records.begin();
  vector<ResponseRecord>::iterator end_iter = records.end();
  while (iter != end_iter)
    {
      if ((*iter).offset >= offset)
	(*iter).offset++;	// widgetΤä֤offset1餹
      ++iter;
    }
}


// OchushaHypertextBuffer/Viewwidgetrecords
// 
void
BBSThreadView::remove_widget(GtkWidget *widget)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  gint offset = ochusha_hypertext_view_remove_widget(view, widget);
  if (offset < 0)
    fprintf(stderr, "ochusha_hypertext_view_remove_widget() returns %d\n",
	    offset);
  g_return_if_fail(offset >= 0);

  // recordsϥ쥹ֹбƤΤǡoffsetˤĤƾ硣
  // binary searchƴطΤ֤ƬõƤ
  // offsetĴ򤹤ȤŬ;ϤϤ뤬
  // 쥹NˤĤO(N)ʤΤݤʤȤϤʤǤ
  vector<ResponseRecord>::iterator iter = records.begin();
  vector<ResponseRecord>::iterator end_iter = records.end();
  while (iter != end_iter)
    {
      if ((*iter).offset >= offset)
	(*iter).offset--;	// widgetΤä֤offset1餹
      iter++;
    }
}


// 쥹ֹres_numΥ쥹ľ˲ѤΤ
void
BBSThreadView::insert_my_own_bookmark(guint res_num, bool interactive)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  if (my_own_bookmark != NULL)
    {
      remove_widget(my_own_bookmark);
      my_own_bookmark = NULL;
    }

  if (records.size() < res_num - 1)
    return;

  TabLabel *tab_label = get_tab_label();
  if (interactive && tab_label != NULL && !tab_label->is_error_state())
    tab_label->notice(thread->get_number_of_responses_got()
		      > static_cast<int>(res_num));

  int bookmark_pos = thread->get_bookmark_position();
  if (bookmark_pos == 0 && controller.get_prefs().get_hide_bookmark())
    {
      controller.get_thread_ui().notify_thread_view_state(this, thread);
      return;	// ɽ
    }

  if (bookmark_pos != 0 && bookmark_pos == static_cast<int>(res_num)
      && bookmark_pos != static_cast<int>(my_own_bookmark_position))
    {
      gchar *name = g_strdup_printf(_("%d of %s"), res_num,
				    thread->get_title());
      controller.get_bookmarks_ui().insert_my_bookmark(thread, res_num, name);
      g_free(name);
    }

  GtkWidget *bookmark = create_bookmark();
  if (bookmark == NULL)
    return;	// ɽ̵֤ԲΥ顼ۤä롣

  ResponseRecord &res = records[res_num - 1];
  ochusha_hypertext_buffer_create_mark_at_offset(buffer, "my-own-bookmark",
						 res.offset);
  insert_widget_at_offset(bookmark, res.offset + res.length);
  my_own_bookmark = bookmark;
  my_own_bookmark_position = res_num;

  controller.get_thread_ui().notify_thread_view_state(this, thread);
}


GtkWidget *
BBSThreadView::create_bookmark()
{
  assert(OCHUSHA_IN_GDK_THREADS());

  GtkWidget *button
    = gtk_button_new_with_label(controller.get_prefs().get_bookmark_text());
  GTK_WIDGET_UNSET_FLAGS(button, GTK_CAN_FOCUS);
  gtk_button_set_focus_on_click(GTK_BUTTON(button), FALSE);
  gtk_widget_set_name(button, "mybookmark");
#if GTK_CHECK_VERSION(2,12,0)
  if (!controller.get_prefs().get_hide_bookmark())
    gtk_widget_set_tooltip_text(button, _("Click to move this to the tail of thread"));
  else
    gtk_widget_set_tooltip_text(button, _("Click to remove this bookmark"));
#endif

  g_signal_connect(button, "clicked",
		   reinterpret_cast<GCallback>(bookmark_clicked_cb), this);
  return button;
}


void
BBSThreadView::bookmark_clicked_cb(GtkWidget *widget, BBSThreadView *view)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  if (view->my_own_bookmark_position == static_cast<int>(view->records.size())
      && !view->controller.get_prefs().get_hide_bookmark())
    return;	// ˤȤϲ⤷ʤ

  view->thread->set_bookmark_position(0);
  view->thread->set_number_of_responses_read(view->records.size());

  view->controller.get_bookmarks_ui().remove_my_bookmark(view->thread);

  view->my_own_bookmark_position = view->records.size();
  view->insert_my_own_bookmark(view->records.size(), true);
  GtkWidget *board_widget
    = view->controller.get_view_manager().get_current_board_widget();
  if (board_widget != NULL)
    gtk_widget_queue_draw(board_widget);
}


// ޤǲѤϢ


void
BBSThreadView::append_response(guint res_num, Response &response)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  if (res_num == 1)
    {
      // 쥹1λϥȥɽ
      const char *title;
      if (!response.is_broken())
	{
	  title = response.get_title();
	  if (thread->get_title() == NULL
	      || thread->get_title()[0] == '\0'
	      || thread->is_tmp_thread())
	    {
	      thread->set_title(title);
	      GtkWidget *label = get_label();
	      if (GTK_IS_LABEL(label))
		gtk_label_set_text(GTK_LABEL(label), thread->get_title());
	    }
	}
      else
	title = thread->get_title();
      if (title != NULL && *title != '\0')
	renderer.render_header(title);
    }

  assert(records.size() == res_num - 1);

  int offset = ochusha_hypertext_buffer_get_current_offset(buffer);
  if (response.is_deleted())
    {
      records.push_back(ResponseRecord(offset, 0));
      return;
    }

  if (!response.is_broken())
    {
      bool a_bone = !disable_a_bone
	&& (thread->is_a_boned(res_num)
	    || thread->get_a_bone().is_a_boned(response)
	    || thread->get_board()->get_a_bone().is_a_boned(response)
	    || controller.get_config().get_a_bone().is_a_boned(response));

      if (!a_bone)
	renderer.render_response(res_num, response);
      else
	{
	  if (!enable_transparent_a_bone)
	    renderer.render_a_bone(res_num);
	}
    }
  else
    renderer.render_broken(res_num);

  int tail_offset = ochusha_hypertext_buffer_get_current_offset(buffer);
  records.push_back(ResponseRecord(offset, tail_offset - offset));
}


gboolean
BBSThreadView::append_responses_on_idle(BBSThreadView *view)
{
  // g_idle_addΥХåGLibɳʤΤGDK_THREADS_ENTER()Ǥʤ
  GDK_THREADS_ENTER();

  ResponseCursorHandle cursor = view->response_cursor;
  gboolean result;

  int lock_id = cursor->trylock();
  if (lock_id != 0)
    {
#ifdef ENABLE_BENCHMARKING
      struct timeval start_time;
      gettimeofday(&start_time, NULL);
#endif
      // MAX_NUM_OF_RES_TO_APPEND_AT_ONCE쥹⤷ϺǸޤɲ
      int count = 0;
      int res_num = view->number_of_responses_shown;
      while (cursor->has_response(++res_num, lock_id))
	{
	  Response &response = cursor->get_response(res_num, lock_id);
	  view->append_response(res_num, response);
	  view->number_of_responses_shown = res_num;

	  if (res_num == view->my_own_bookmark_position)
	    view->insert_my_own_bookmark(res_num, false);

	  if (res_num == view->initial_view_position)
	    {
	      GtkTextMark *mark
		= ochusha_hypertext_buffer_create_mark_at_offset(
					view->buffer, "view-position",
					view->records[res_num - 1].offset);
	      ochusha_hypertext_view_scroll_to_mark(view->view, mark);
	    }

	  if (++count >= MAX_NUM_OF_RES_TO_APPEND_AT_ONCE)
	    break;
	}

      if (view->auto_refresher_id != 0)
	{
	  GtkTextMark *mark
	    = ochusha_hypertext_buffer_create_mark(view->buffer,
						   "tail-view-position");
	  ochusha_hypertext_view_scroll_to_mark(view->view, mark);
	}

      if (!cursor->has_response(res_num, lock_id))
	{ // ΤȤʾɤ߼쥹ʤ
	  view->append_responses_id = 0;
	  result = FALSE;
	  if (cursor->finished(lock_id))
	    {
	      BBSThread *thread = view->thread;
	      TabLabel *tab_label = view->get_tab_label();
	      if (tab_label != NULL && !tab_label->is_error_state())
		tab_label->notice(thread->get_number_of_responses_to_read()
				  > 0);
	    }
	}
      else
	result = TRUE;

      cursor->unlock(lock_id);
#ifdef ENABLE_BENCHMARKING
      struct timeval end_time;
      gettimeofday(&end_time, NULL);
      time_t sec = end_time.tv_sec - start_time.tv_sec;
      suseconds_t usec = end_time.tv_usec - start_time.tv_usec;
      while (usec < 0)
	{
	  sec--;
	  usec += 1000000;
	}
      printf("Rendering of %d responses consumes: %ld sec, %lf msec\n",
	     count, static_cast<long>(sec), static_cast<double>(usec / 1000.0));
#endif
      GtkWidget *board_widget
	= view->controller.get_view_manager().get_current_board_widget();
      if (board_widget != NULL)
	gtk_widget_queue_draw(board_widget);
    }
  else
    result = TRUE;	// åˤ꺣Υ쥹ɲäϸ

  GDK_THREADS_LEAVE();


  return result;
}


void
BBSThreadView::state_changed_cb()
{
  // sigslotʥϥɥ
  GDK_THREADS_ENTER();

  if (append_responses_id == 0 && response_cursor != NULL)
    // idleؿϿ
    append_responses_id
      = g_idle_add_full(THREAD_VIEW_UPDATE_PRIORITY,
			reinterpret_cast<GSourceFunc>(append_responses_on_idle),
			this, NULL);

  GDK_THREADS_LEAVE();
}


BBSThreadView *
BBSThreadView::create(Controller &controller, BBSThread *thread,
		      TabLabel *label, int initial_pos, bool refresh)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  BBSThreadView *view = new BBSThreadView(controller, label, thread);

  if (initial_pos == 0)
    {
      ViewPosition pos;
      if (thread->is_alive() && !controller.get_config().get_offline())
	pos = controller.get_prefs().get_initial_view_position_of_active_thread();
      else
	pos = controller.get_prefs().get_initial_view_position_of_inactive_thread();

      switch (pos)
	{
	case MY_BOOKMARK:
	  break;	// initial_pos == 0Τޤޤɤ

	case LAST_VIEW_POSITION:
	  initial_pos = thread->get_last_view_response();
	  break;

	case THREAD_TOP:
	  initial_pos = 1;
	  break;
	}
    }

  view->initial_view_position = initial_pos;
  view->update(refresh, false);	// ɽϤ

  return view;
}


void
BBSThreadView::update(bool refresh, bool automatic)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  TabLabel *label = get_tab_label();

  thread->set_number_of_responses_read(thread->get_number_of_responses_got());
  ResponseCursor *cursor = response_cursor;
  if (cursor == NULL)
    {
      // BBSThread::get_responses()sigslotȤΤglobal lock롣
      // ʤΤǡǰöGDK_THREADS_LEAVE()롣
      GDK_THREADS_LEAVE();

      ResponseCursorHandle new_cursor
	= thread->get_responses(controller.get_config(),
				controller.get_repository(),
				controller.get_broker());
      new_cursor->access_started.connect(label, &TabLabel::access_started);
      new_cursor->access_progressed.connect(label,
					    &TabLabel::access_progressed);
      new_cursor->access_finished.connect(label, &TabLabel::access_finished);
      new_cursor->access_failed.connect(label, &TabLabel::access_failed);
      new_cursor->access_terminated.connect(label,
					    &TabLabel::access_terminated);
      new_cursor->state_changed.connect(this, &BBSThreadView::state_changed_cb);
      response_cursor = new_cursor;

      int bookmark_pos = thread->get_number_of_responses_read();
      if (new_cursor != NULL)
	{
	  if (refresh && new_cursor->update())
	    my_own_bookmark_position = bookmark_pos;
	  else
	    {
	      if (my_own_bookmark_position == 0)
		my_own_bookmark_position = bookmark_pos;
	      label->access_finished();
	    }
	  if (initial_view_position == 0)
	    initial_view_position = bookmark_pos;
	}

      GDK_THREADS_ENTER();
    }
  else
    {
      if (!automatic)
	{
	  GDK_THREADS_LEAVE();

	  if (refresh && cursor->update())
	    {
	      GDK_THREADS_ENTER();

	      int bookmark_pos = thread->get_number_of_responses_read();
	      if (my_own_bookmark_position != bookmark_pos)
		{
		  my_own_bookmark_position = bookmark_pos;
		  insert_my_own_bookmark(my_own_bookmark_position, false);
		  initial_view_position = bookmark_pos;
		}
	    }
	  else
	    GDK_THREADS_ENTER();
	}
    }

  if (append_responses_id == 0 && response_cursor != NULL)
    append_responses_id
      = g_idle_add_full(THREAD_VIEW_UPDATE_PRIORITY,
			reinterpret_cast<GSourceFunc>(append_responses_on_idle),
			this, NULL);

  if (automatic)
    {
      if (auto_refresher_id == 0)
	{
	  GdkColor *fg = controller.get_prefs().get_auto_refresh_go();
	  GtkWidget *label_widget = label->get_widget();
	  gtk_widget_modify_fg(label_widget, GTK_STATE_NORMAL, fg);
	  gtk_widget_modify_fg(label_widget, GTK_STATE_ACTIVE, fg);
	  previous_interval = 10000;
	  previous_number_of_responses_got
	    = thread->get_number_of_responses_read();
	  auto_refresher_id
	    = g_timeout_add(previous_interval,
			    reinterpret_cast<GSourceFunc>(auto_refresher_timeout_cb), this);
	}
      else
	{
	  g_source_remove(auto_refresher_id);
	  auto_refresher_id = 0;
	  GtkWidget *label_widget = label->get_widget();
	  gtk_widget_modify_fg(label_widget, GTK_STATE_NORMAL, NULL);
	  gtk_widget_modify_fg(label_widget, GTK_STATE_ACTIVE, NULL);
	}
    }
}


void
BBSThreadView::show()
{
  // ʤˤ⤻
}


void
BBSThreadView::hide(bool now)
{
  // ʤˤ⤻
}


bool
BBSThreadView::is_head_visible()
{
  GtkTextView *text_view = GTK_TEXT_VIEW(view);
  GtkTextBuffer *text_buffer = gtk_text_view_get_buffer(text_view);
  GdkRectangle rect;
  GtkTextIter iter;
  gint y;
  gint height;

  gtk_text_view_get_visible_rect(text_view, &rect);
  gtk_text_buffer_get_start_iter(text_buffer, &iter);
  gtk_text_view_get_line_yrange(text_view, &iter, &y, &height);

  return rect.y <= y;
}


bool
BBSThreadView::is_tail_visible()
{
  GtkTextView *text_view = GTK_TEXT_VIEW(view);
  GtkTextBuffer *text_buffer = gtk_text_view_get_buffer(text_view);
  GdkRectangle rect;
  GtkTextIter iter;
  gint y;
  gint height;

  gtk_text_view_get_visible_rect(text_view, &rect);
  gtk_text_buffer_get_end_iter(text_buffer, &iter);
  gtk_text_view_get_line_yrange(text_view, &iter, &y, &height);

  return (rect.y + rect.height) >= y;
}


bool
BBSThreadView::advance_viewport()
{
  GdkRectangle rect;
  gint pre_y;

  gtk_text_view_get_visible_rect(GTK_TEXT_VIEW(view), &rect);
  pre_y = rect.y;

  g_signal_emit_by_name(view, "move_viewport", GTK_SCROLL_PAGES, 1);

  gtk_text_view_get_visible_rect(GTK_TEXT_VIEW(view), &rect);
  if (pre_y != rect.y)
    return true;

  int lock_id = response_cursor->trylock();
  if (lock_id == 0)
    return true;	// ¿ʬʤΤǿʤ᤿դ

  bool finished = response_cursor->finished(lock_id);

  response_cursor->unlock(lock_id);

  if (finished)
    {
      if (number_of_responses_shown
	  == thread->get_number_of_responses_available())
	return false;

      // 顼ʤɤˤ깹˼ԤƤ顢ʤ
      refresh();
    }

  return true;	// 󥰤äƤʤΤǿʤ᤿դ
}


bool
BBSThreadView::retreat_viewport()
{
  bool head_visible = is_head_visible();

  g_signal_emit_by_name(view, "move_viewport", GTK_SCROLL_PAGES, -1);

  return !head_visible;
}


void
BBSThreadView::show_prefered_response_editor()
{
  controller.get_thread_ui().get_prefered_response_editor(this);
}


ResponseEditorWindow *
BBSThreadView::get_post_window()
{
  return ResponseEditorWindow::get_window(post_window);
}


void
BBSThreadView::show_post_window()
{
  assert(OCHUSHA_IN_GDK_THREADS());

  ResponseEditorWindow *window;
  if (post_window != NULL)
    {
      gtk_window_present(GTK_WINDOW(post_window));
      window = ResponseEditorWindow::get_window(post_window);
      g_return_if_fail(window != NULL);
    }
  else
    {
      window = new ResponseEditorWindow(controller, thread);
      post_window = window->get_widget();
      post_window_destroy_id
	= g_signal_connect(post_window, "destroy",
			   reinterpret_cast<GCallback>(post_window_destroy_cb),
			   this);
      gtk_widget_show(post_window);
    }
  window->grab_focus_into_editor();
}


void
BBSThreadView::jump_to_bookmark()
{
  assert(OCHUSHA_IN_GDK_THREADS());

  GtkTextMark *mark = gtk_text_buffer_get_mark(GTK_TEXT_BUFFER(buffer),
					       "my-own-bookmark");
  g_return_if_fail(mark != NULL);

  backward_history_push(get_current_view_url());
  forward_history_clear();

  controller.get_thread_ui().notify_thread_view_state(this, thread);

  ochusha_hypertext_view_scroll_to_mark(view, mark);
}


void
BBSThreadView::specify_view_position(int position)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  guint pos;
  if (position > 0)
    pos = position;
  else
    pos = records.size() + position;
 if (records.size() < pos)
    return;

  backward_history_push(get_current_view_url());
  forward_history_clear();

  controller.get_thread_ui().notify_thread_view_state(this, thread);

  GtkTextMark *mark
    = ochusha_hypertext_buffer_create_mark_at_offset(buffer, "view-position",
						     records[pos - 1].offset);
  ochusha_hypertext_view_scroll_to_mark(view, mark);
}


int
BBSThreadView::estimate_current_view_position()
{
  gint offset = ochusha_hypertext_view_get_visible_offset(view);

  vector<ResponseRecord>::iterator iter = records.begin();
  vector<ResponseRecord>::iterator end_iter = records.end();
  int res_num = 1;
  // binary searchŬѤ٤⡣
  while (iter != end_iter)
    {
      ResponseRecord &record = *iter++;
      if (record.offset >= offset
	  && offset < static_cast<gint>(record.offset + record.length))
	return res_num;

      ++res_num;
    }

  return 0;
}


void
BBSThreadView::remember_current_view_position()
{
  int pos = estimate_current_view_position();
  if (pos > 0)
    thread->set_last_view_response(pos);
}


void
BBSThreadView::goto_bottom()
{
  assert(OCHUSHA_IN_GDK_THREADS());

  backward_history_push(get_current_view_url());
  forward_history_clear();

  controller.get_thread_ui().notify_thread_view_state(this, thread);

  ochusha_hypertext_view_scroll_to_end(view);
}


void
BBSThreadView::goto_top()
{
  assert(OCHUSHA_IN_GDK_THREADS());

  backward_history_push(get_current_view_url());
  forward_history_clear();

  controller.get_thread_ui().notify_thread_view_state(this, thread);

  ochusha_hypertext_view_scroll_to_start(view);
}


void
BBSThreadView::post_window_destroy_cb(GtkWidget *widget, BBSThreadView *view)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  view->post_window = NULL;
  view->post_window_destroy_id = 0;
}


gboolean
BBSThreadView::auto_refresher_timeout_cb(BBSThreadView *view)
{
  GDK_THREADS_ENTER();

  bool skip_refresh;
  if (!view->is_tail_visible())
    {
      // ɽƤʤϼư򥹥å
      skip_refresh = TRUE;
      GdkColor *fg = view->controller.get_prefs().get_auto_refresh_wait();
      TabLabel *tab_label = view->get_tab_label();
      if (tab_label != NULL)
	{
	  GtkWidget *label_widget = tab_label->get_widget();
	  gtk_widget_modify_fg(label_widget, GTK_STATE_NORMAL, fg);
	  gtk_widget_modify_fg(label_widget, GTK_STATE_ACTIVE, fg);
	}
    }
  else
    {
      skip_refresh = FALSE;
      GdkColor *fg = view->controller.get_prefs().get_auto_refresh_go();
      TabLabel *tab_label = view->get_tab_label();
      if (tab_label != NULL)
	{
	  GtkWidget *label_widget = tab_label->get_widget();
	  gtk_widget_modify_fg(label_widget, GTK_STATE_NORMAL, fg);
	  gtk_widget_modify_fg(label_widget, GTK_STATE_ACTIVE, fg);
	}
    }

  int num_got = view->thread->get_number_of_responses_got();
  if (view->previous_number_of_responses_got != num_got)
    {
      if (!skip_refresh)
	{
	  view->previous_interval -= 1000;
	  if (view->previous_interval < 10000)
	    view->previous_interval = 10000;
	}
      else
	num_got = view->previous_number_of_responses_got;
    }
  else
    {
      view->previous_interval += 10000;
      if (view->thread->is_stopped() || view->previous_interval > 300000)
	{
	  // ߤ
	  TabLabel *tab_label = view->get_tab_label();
	  if (tab_label != NULL)
	    {
	      GtkWidget *label_widget = tab_label->get_widget();
	      gtk_widget_modify_fg(label_widget, GTK_STATE_NORMAL, NULL);
	      gtk_widget_modify_fg(label_widget, GTK_STATE_ACTIVE, NULL);
	      view->previous_interval = 0;
	    }
	}
    }
  view->previous_number_of_responses_got = num_got;

  if (view->previous_interval != 0)
    view->auto_refresher_id
      = g_timeout_add(view->previous_interval,
		      reinterpret_cast<GSourceFunc>(auto_refresher_timeout_cb),
		      view);
  else
    view->auto_refresher_id = 0;

  if (!skip_refresh)
    {
      GDK_THREADS_LEAVE();

      bool result = view->response_cursor->update();
      if (result)
	{
	  GDK_THREADS_ENTER();

	  if (view->append_responses_id == 0)
	    view->append_responses_id
	      = g_idle_add_full(THREAD_VIEW_UPDATE_PRIORITY,
				reinterpret_cast<GSourceFunc>(append_responses_on_idle), view, NULL);

	  GDK_THREADS_LEAVE();
	}

      return FALSE;
    }

  GDK_THREADS_LEAVE();

  return FALSE;
}

}
