/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: board_properties_ui.cc,v 1.23 2009/01/01 07:47:06 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "board_properties_ui.h"

#include "a_bone_ui.h"
#include "focus_utils.h"
#include "controller.h"

#include "libochusha/bulletin_board.h"
#include "libochusha/ref_count.h"

#include "libochushagtk/regex_editor.h"

#include <algorithm>

#include <cassert>
#include <cstdio>
#include <cstdlib>

#include <gtk/gtk.h>


namespace Ochusha
{


enum {
  BBS_TYPE_DESCRIPTION = 0,
  BBS_TYPE,
  BBS_TYPE_N_COLUMNS
};


static void
set_bbs_type_description(GtkCellLayout *layout, GtkCellRenderer *renderer,
			 GtkTreeModel *model, GtkTreeIter *iter,
			 gpointer unused)
{
  const gchar *desc = NULL;

  gtk_tree_model_get(model, iter, BBS_TYPE_DESCRIPTION, &desc, -1);
  if (desc != NULL)
    g_object_set(renderer, "text", desc, NULL);
}


class BoardProperties
{
private:
  BoardPropertiesUI &ui;
  BulletinBoardHandle board;
  bool add_board;

  ABoneSettings a_bone_settings;

  GtkNotebook *notebook;

  GtkWidget *general_info_page;
  GtkWidget *thread_view_page;


  // ľ󥿥
  GtkEntry *name_entry;
  GtkEntry *short_name_entry;

  GtkTreeModel *bbs_types;
  GtkComboBox *bbs_types_combo;

  GtkEntry *url_entry;

  GtkTextBuffer *description_buffer;

  GtkTextBuffer *movement_history_buffer;


  // ɽϢ
  GtkToggleButton *use_global_setting_button;
  GtkToggleButton *show_mailto_button;
  GtkToggleButton *hide_mailto_button;


  ~BoardProperties();

  static void dialog_destroyed_cb(GtkWidget *widget,
				  BoardProperties *properties);
  static void excavate_button_clicked_cb(GtkButton *button,
					 BoardProperties *properties);

  GtkWidget *create_general_info_page();
  GtkWidget *create_thread_view_page();

  void update_movement_history_buffer();
  void reflect_movement_history_buffer();


public:
  BoardProperties(BoardPropertiesUI &ui, GtkDialog *dialog,
		  BulletinBoard *board);

  inline BulletinBoard *get_board() { return board; }

  void process_response(GtkDialog *dialog, int response_id,
			BulletinBoard *board);

  void update_show_mailto(bool show_mailto);
  void append_name_to_be_a_boned(const char *name);
  void append_id_to_be_a_boned(const char *id);
  void append_content_to_be_a_boned(const char *content);

  inline void set_add_board(bool value) { add_board = value; }
};


#define TABLE_ATTACH_CAPTION(table, widget, l, r, t, b)			\
  gtk_table_attach(table, widget, l, r, t, b,				\
		   static_cast<GtkAttachOptions>(GTK_SHRINK | GTK_FILL),\
		   static_cast<GtkAttachOptions>(GTK_SHRINK | GTK_FILL), 3, 0)


GtkWidget *
BoardProperties::create_general_info_page()
{
  GtkWidget *widget = gtk_vbox_new(FALSE, 0);
  GtkBox *outer_vbox = GTK_BOX(widget);

  widget = gtk_table_new(6, 3, FALSE);
  gtk_widget_show(widget);
  gtk_box_pack_start(outer_vbox, widget, FALSE, FALSE, 3);

  GtkTable *table = GTK_TABLE(widget);


  GtkWidget *label = gtk_label_new_with_mnemonic(_("_Name:"));
  gtk_misc_set_alignment(GTK_MISC(label), 1.0, 0.5);
  gtk_widget_show(label);
  TABLE_ATTACH_CAPTION(table, label, 0, 1, 0, 1);

  GtkWidget *entry = EntryFocusNotifier::gtk_entry_new(ui.get_controller());
  gtk_widget_show(entry);
  gtk_table_attach_defaults(table, entry, 1, 2, 0, 1);
  gtk_label_set_mnemonic_widget(GTK_LABEL(label), entry);

  name_entry = GTK_ENTRY(entry);

  const char *name = board->get_name();
  gtk_entry_set_text(name_entry, name != NULL ? name : _("Anonymous Board"));
  gtk_entry_set_width_chars(name_entry, 24);
  gtk_entry_set_activates_default(name_entry, TRUE);


  GtkListStore *store = gtk_list_store_new(BBS_TYPE_N_COLUMNS,
					   G_TYPE_POINTER, G_TYPE_INT);
  bbs_types = GTK_TREE_MODEL(store);
  GtkWidget *combo = gtk_combo_box_new_with_model(bbs_types);
  gtk_widget_show(combo);
  gtk_table_attach(table, combo, 2, 3, 0, 1, GTK_SHRINK, GTK_SHRINK, 0, 0);

  bbs_types_combo = GTK_COMBO_BOX(combo);
  gtk_combo_box_set_focus_on_click(bbs_types_combo, FALSE);

  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(combo), renderer, TRUE);
  gtk_cell_layout_set_cell_data_func(GTK_CELL_LAYOUT(combo), renderer,
				     set_bbs_type_description, NULL, NULL);

  int board_type = board->get_bbs_type();
  {
    vector<int> bbs_type_list;
    BulletinBoard::get_all_bbs_type_list(bbs_type_list);
    vector<int>::iterator iter = bbs_type_list.begin();
    vector<int>::iterator end_iter = bbs_type_list.end();

    while (iter != end_iter)
      {
	int bbs_type = *iter++;
	if ((board_type >= OCHUSHA_BBS_TYPE_2CH
	     && bbs_type >= OCHUSHA_BBS_TYPE_2CH)
	    || board_type == bbs_type)
	  {
	    const char *desc
	      = BulletinBoard::get_bbs_type_description(bbs_type);
	    if (desc != NULL)
	      {
		GtkTreeIter tree_iter;
		gtk_list_store_insert_with_values(store, &tree_iter, G_MAXINT32,
						  BBS_TYPE_DESCRIPTION, desc,
						  BBS_TYPE, bbs_type,
						  -1);
		if (bbs_type == board_type)
		  gtk_combo_box_set_active_iter(bbs_types_combo, &tree_iter);
	      }
	  }
      }
  }



  label = gtk_label_new_with_mnemonic(_("_Short Name:"));
  gtk_misc_set_alignment(GTK_MISC(label), 1.0, 0.5);
  gtk_widget_show(label);
  TABLE_ATTACH_CAPTION(table, label, 0, 1, 1, 2);

  entry = EntryFocusNotifier::gtk_entry_new(ui.get_controller());
  gtk_widget_show(entry);
  gtk_table_attach_defaults(table, entry, 1, 3, 1, 2);
  gtk_label_set_mnemonic_widget(GTK_LABEL(label), entry);

  short_name_entry = GTK_ENTRY(entry);

  name = board->get_short_name();
  gtk_entry_set_text(short_name_entry,
		     name != NULL ? name : _("Anonymous Board"));
  gtk_entry_set_width_chars(short_name_entry, 24);
  gtk_entry_set_activates_default(short_name_entry, TRUE);


  label = gtk_label_new_with_mnemonic(_("_URL:"));
  gtk_misc_set_alignment(GTK_MISC(label), 1.0, 0.5);
  TABLE_ATTACH_CAPTION(table, label, 0, 1, 2, 3);

  entry = EntryFocusNotifier::gtk_entry_new(ui.get_controller());
  gtk_table_attach_defaults(table, entry, 1, 3, 2, 3);
  gtk_label_set_mnemonic_widget(GTK_LABEL(label), entry);

  url_entry = GTK_ENTRY(entry);

  const char *url = board->get_base_url();
  if (url != NULL)
    gtk_entry_set_text(url_entry, url);
  gtk_entry_set_width_chars(url_entry, 40);
  gtk_entry_set_activates_default(url_entry, TRUE);

  if (board_type >= OCHUSHA_BBS_TYPE_2CH
      || board_type == OCHUSHA_BBS_TYPE_ACTIVE_FOLDER)
    {
      gtk_widget_show(label);
      gtk_widget_show(entry);
    }

  if (board_type == OCHUSHA_BBS_TYPE_ACTIVE_FOLDER)
    {
      gtk_widget_set_sensitive(label, FALSE);
      gtk_widget_set_sensitive(entry, FALSE);
    }


  widget = gtk_vbox_new(FALSE, 0);
  gtk_widget_show(widget);
  gtk_table_attach(table, widget, 0, 1, 3, 4,
		   static_cast<GtkAttachOptions>(GTK_SHRINK | GTK_FILL),
		   static_cast<GtkAttachOptions>(GTK_EXPAND | GTK_FILL), 3, 3);
  GtkBox *vbox = GTK_BOX(widget);

  label = gtk_label_new_with_mnemonic(_("_Description:"));
  gtk_misc_set_alignment(GTK_MISC(label), 1.0, 0.0);
  gtk_widget_show(label);
  gtk_box_pack_start(vbox, label, TRUE, TRUE, 3);


  description_buffer = gtk_text_buffer_new(NULL);
  const char *desc = board->get_description();
  if (desc != NULL)
    gtk_text_buffer_insert_at_cursor(description_buffer, desc, -1);

  GtkWidget *win = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(win),
				 GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  GtkWidget *view
    = TextViewFocusNotifier::gtk_text_view_new_with_buffer(ui.get_controller(),
							   description_buffer);
  gtk_widget_show(view);
  gtk_container_add(GTK_CONTAINER(win), view);
  gtk_label_set_mnemonic_widget(GTK_LABEL(label), view);

  gtk_widget_show(win);
  gtk_table_attach(table, win, 1, 3, 3, 4,
		   static_cast<GtkAttachOptions>(GTK_EXPAND | GTK_FILL),
		   static_cast<GtkAttachOptions>(GTK_EXPAND | GTK_FILL), 3, 3);


  widget = gtk_vbox_new(FALSE, 0);
  gtk_widget_show(widget);
  gtk_table_attach(table, widget, 0, 1, 4, 5,
		   static_cast<GtkAttachOptions>(GTK_SHRINK | GTK_FILL),
		   static_cast<GtkAttachOptions>(GTK_EXPAND | GTK_FILL), 3, 3);
  vbox = GTK_BOX(widget);

  label = gtk_label_new_with_mnemonic(_("Movement _History"));
  gtk_misc_set_alignment(GTK_MISC(label), 1.0, 0.0);
  if (board_type >= OCHUSHA_BBS_TYPE_2CH)
    gtk_widget_show(label);
  gtk_box_pack_start(vbox, label, TRUE, TRUE, 3);


  movement_history_buffer = gtk_text_buffer_new(NULL);

  update_movement_history_buffer();

  win = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(win),
				 GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  view
    = TextViewFocusNotifier::gtk_text_view_new_with_buffer(ui.get_controller(),
							   movement_history_buffer);
  gtk_widget_show(view);
  gtk_container_add(GTK_CONTAINER(win), view);
  gtk_label_set_mnemonic_widget(GTK_LABEL(label), view);

  if (board_type >= OCHUSHA_BBS_TYPE_2CH)
    gtk_widget_show(win);
  gtk_table_attach(table, win, 1, 3, 4, 5,
		   static_cast<GtkAttachOptions>(GTK_EXPAND | GTK_FILL),
		   static_cast<GtkAttachOptions>(GTK_EXPAND | GTK_FILL), 3, 3);

  if (board_type < OCHUSHA_BBS_TYPE_2CH)
    {
      char message[512];
      GtkWidget *appendix = NULL;
      if (board_type == OCHUSHA_BBS_TYPE_ACTIVE_FOLDER)
	{
	  Folder *folder = board->get_correspond_folder();
	  if (folder != NULL)
	    {
	      snprintf(message, 512,
		       _("This is a pseudo board to show contents of the bookmark folder currently named \"%s\" as a threadlist."),
		       folder->get_name());
	      appendix = gtk_label_new(message);
	    }
	  else
	    {
	      appendix = gtk_label_new(_("This is a pseudo board used to be for showing contents of a already deleted bookmark folder."));
	    }
	}
      else if (board_type == OCHUSHA_BBS_TYPE_PSEUDO_FOLDER)
	{
	  desc = BulletinBoard::get_description_about_pseudo_folder_url(board->get_base_url());
	  if (desc != NULL)
	    {
	      snprintf(message, 512,
		       _("This is a pseudo board to show \"%s\" as a threadlist."),
		       desc);
	      appendix = gtk_label_new(message);
	    }
	  else
	    appendix = gtk_label_new(_("Unknown pseudo folder"));
	}

      if (appendix != NULL)
	{
	  gtk_widget_set_size_request(appendix, 300, -1);
	  gtk_label_set_line_wrap(GTK_LABEL(appendix), TRUE);
	  gtk_widget_show(appendix);
	  gtk_table_attach(table, appendix, 1, 3, 5, 6,
		   static_cast<GtkAttachOptions>(GTK_EXPAND | GTK_FILL),
		   static_cast<GtkAttachOptions>(GTK_EXPAND | GTK_FILL), 0, 0);
	}
    }


  widget = gtk_button_new_with_mnemonic(_("_Excavate DAT files"));
  g_signal_connect(widget, "clicked",
		   reinterpret_cast<GCallback>(excavate_button_clicked_cb),
		   this);
  gtk_button_set_focus_on_click(GTK_BUTTON(widget), FALSE);
  if (board_type >= OCHUSHA_BBS_TYPE_2CH)
    gtk_widget_show(widget);
  gtk_box_pack_end(outer_vbox, widget, FALSE, FALSE, 3);

  return GTK_WIDGET(outer_vbox);
}


GtkWidget *
BoardProperties::create_thread_view_page()
{
  GtkWidget *widget = gtk_vbox_new(FALSE, 0);
  GtkBox *outer_vbox = GTK_BOX(widget);


  widget = gtk_hbox_new(FALSE, 0);
  gtk_widget_show(widget);
  gtk_box_pack_start(outer_vbox, widget, FALSE, FALSE, 0);

  GtkBox *hbox = GTK_BOX(widget);

  GtkWidget *label = gtk_label_new(_("Contents of mailto:"));
  gtk_widget_show(label);
  gtk_box_pack_start(hbox, label, FALSE, FALSE, 5);


  GtkWidget *button
    = gtk_radio_button_new_with_mnemonic(NULL,
					 ui.get_global_show_mailto_mode()
					 ? _("_Global Setting [Show]")
					 : _("_Global Setting [Hide]"));
  gtk_widget_show(button);
  gtk_box_pack_start(hbox, button, FALSE, FALSE, 0);
  use_global_setting_button = GTK_TOGGLE_BUTTON(button);


  button
    = gtk_radio_button_new_with_mnemonic_from_widget(GTK_RADIO_BUTTON(button),
						     _("_Show"));
  gtk_widget_show(button);
  gtk_box_pack_start(hbox, button, FALSE, FALSE, 0);
  show_mailto_button = GTK_TOGGLE_BUTTON(button);


  button
    = gtk_radio_button_new_with_mnemonic_from_widget(GTK_RADIO_BUTTON(button),
						     _("_Hide"));
  gtk_widget_show(button);
  gtk_box_pack_start(hbox, button, FALSE, FALSE, 0);
  hide_mailto_button = GTK_TOGGLE_BUTTON(button);

  switch (board->get_mailto_mode())
    {
    case MAILTO_MODE_DEFAULT:
      gtk_toggle_button_set_active(use_global_setting_button, TRUE);
      break;

    case MAILTO_MODE_SHOW:
      gtk_toggle_button_set_active(show_mailto_button, TRUE);
      break;

    case MAILTO_MODE_HIDE:
      gtk_toggle_button_set_active(hide_mailto_button, TRUE);
      break;
    }


  // ĸͭܡ
  widget = a_bone_settings.get_widget();
  gtk_widget_show(widget);
  gtk_box_pack_start(outer_vbox, widget, TRUE, TRUE, 0);


  return GTK_WIDGET(outer_vbox);
}


BoardProperties::BoardProperties(BoardPropertiesUI &ui, GtkDialog *dialog,
				 BulletinBoard *board)
  : ui(ui),
    board(board),
    add_board(false),
    a_bone_settings(board->get_a_bone(), ui.controller),
    general_info_page(create_general_info_page()),
    thread_view_page(create_thread_view_page())
{
  GtkWidget *widget = gtk_notebook_new();
  gtk_widget_show(widget);
  gtk_box_pack_start(GTK_BOX(dialog->vbox), widget, TRUE, TRUE, 0);
  notebook = GTK_NOTEBOOK(widget);


  gtk_widget_show(general_info_page);
  gtk_notebook_append_page(notebook,
			   general_info_page, gtk_label_new(_("General")));


  if (board->get_bbs_type() >= OCHUSHA_BBS_TYPE_2CH)
    gtk_widget_show(thread_view_page);
  gtk_notebook_append_page(notebook,
			   thread_view_page,
			   gtk_label_new(_("Thread View Settings")));

  g_signal_connect(dialog, "destroy",
		   reinterpret_cast<GCallback>(dialog_destroyed_cb), this);
}


BoardProperties::~BoardProperties()
{
  ui.properties_destroyed(this);
  g_object_unref(movement_history_buffer);
  g_object_unref(description_buffer);
  g_object_unref(bbs_types);
}


void
BoardProperties::update_movement_history_buffer()
{
  GtkTextIter buffer_iter;
  GtkTextIter buffer_end_iter;
  gtk_text_buffer_get_start_iter(movement_history_buffer, &buffer_iter);
  gtk_text_buffer_get_end_iter(movement_history_buffer, &buffer_end_iter);
  gtk_text_buffer_delete(movement_history_buffer,
			 &buffer_iter, &buffer_end_iter);

  list<char *>::const_iterator iter = board->movement_history_begin();
  list<char *>::const_iterator end_iter = board->movement_history_end();
  while (iter != end_iter)
    {
      const char *server = *iter++;
      gtk_text_buffer_insert_at_cursor(movement_history_buffer, server, -1);
      if (iter != end_iter)
	gtk_text_buffer_insert_at_cursor(movement_history_buffer, "\n", 1);
    }
}


void
BoardProperties::reflect_movement_history_buffer()
{
  // İžˤȿ
  GtkTextIter iter;
  GtkTextIter end_iter;

  gtk_text_buffer_get_start_iter(movement_history_buffer, &iter);
  gtk_text_buffer_get_end_iter(movement_history_buffer, &end_iter);
  gchar *server_list = gtk_text_buffer_get_text(movement_history_buffer,
						&iter, &end_iter, FALSE);
  if (server_list != NULL)	// ¿ʬɬ
    {
      list<char *> &history = board->get_movement_history();
      while (!history.empty())
	{
	  char *server = history.front();
	  history.pop_front();
	  free(server);
	}

      URI base_uri(board->get_base_url());
      const char *current_server = base_uri.get_server();

      gchar *cur_pos = server_list;
      while (*cur_pos != '\0')
	{
	  gchar *nl_pos = strchr(cur_pos, '\n');
	  if (nl_pos != NULL)
	    *nl_pos = '\0';

	  // g_strstrip()in-placeʸ򤤤
	  const gchar *server = g_strstrip(cur_pos);
	  if (strlen(server) > 0 && strcmp(server, current_server) != 0)
	    {
	      list<char *>::iterator where;
	      list<char *>::iterator end_iter = history.end();
	      where = find_if(history.begin(), end_iter, CStringEqual(server));
	      if (where == end_iter)
		history.push_back(strdup(server));
	    }

	  if (nl_pos != NULL)
	    cur_pos = nl_pos + 1;
	  else
	    break;
	}
      g_free(server_list);
    }
}


void
BoardProperties::excavate_button_clicked_cb(GtkButton *button,
					    BoardProperties *properties)
{
  properties->reflect_movement_history_buffer();
  properties->ui.controller.get_board_ui()
    .excavate_dat_files(properties->board);
  properties->update_movement_history_buffer();
}


void
BoardProperties::process_response(GtkDialog *dialog, int response_id,
				  BulletinBoard *board)
{
  g_return_if_fail(board == get_board());

  if (response_id != GTK_RESPONSE_OK)
    {
      a_bone_settings.cancel_settings();
      return;
    }

  board->set_name(gtk_entry_get_text(name_entry));
  board->set_short_name(gtk_entry_get_text(short_name_entry));

  GtkTreeIter tree_iter;
  if (gtk_combo_box_get_active_iter(bbs_types_combo, &tree_iter))
    {
      gint bbs_type = OCHUSHA_BBS_TYPE_UNKNOWN;
      gtk_tree_model_get(bbs_types, &tree_iter, BBS_TYPE, &bbs_type, -1);
      board->set_bbs_type(bbs_type);
    }

  string old_base_url(board->get_base_url());
  board->set_base_url(gtk_entry_get_text(url_entry));

  GtkTextIter iter;
  GtkTextIter end_iter;

  gtk_text_buffer_get_start_iter(description_buffer, &iter);
  gtk_text_buffer_get_end_iter(description_buffer, &end_iter);
  gchar *desc = gtk_text_buffer_get_text(description_buffer,
					 &iter, &end_iter, FALSE);
  board->set_description(desc);
  g_free(desc);


  reflect_movement_history_buffer();


  if (gtk_toggle_button_get_active(use_global_setting_button))
    board->set_mailto_mode(MAILTO_MODE_DEFAULT);
  else if (gtk_toggle_button_get_active(show_mailto_button))
    board->set_mailto_mode(MAILTO_MODE_SHOW);
  else
    board->set_mailto_mode(MAILTO_MODE_HIDE);

  if (add_board)
    ui.confirm_addition_of_board(board);

  if (strcmp(board->get_base_url(), old_base_url.c_str()) != 0)
    board->notify_major_modification();
}


void
BoardProperties::dialog_destroyed_cb(GtkWidget *widget,
				     BoardProperties *properties)
{
  delete properties;
}


void
BoardProperties::update_show_mailto(bool show_mailto)
{
  gtk_button_set_label(GTK_BUTTON(use_global_setting_button),
		       show_mailto
		       ? _("_Global Setting (Show)")
		       : _("_Global Setting (Hide)"));
}


void
BoardProperties::append_name_to_be_a_boned(const char *name)
{
  a_bone_settings.append_name_to_be_a_boned(name);

  gint page = gtk_notebook_page_num(notebook, thread_view_page);
  g_return_if_fail(page != -1);
  gtk_notebook_set_current_page(notebook, page);
}


void
BoardProperties::append_id_to_be_a_boned(const char *id)
{
  a_bone_settings.append_id_to_be_a_boned(id);

  gint page = gtk_notebook_page_num(notebook, thread_view_page);
  g_return_if_fail(page != -1);
  gtk_notebook_set_current_page(notebook, page);
}


void
BoardProperties::append_content_to_be_a_boned(const char *content)
{
  a_bone_settings.append_content_to_be_a_boned(content);

  gint page = gtk_notebook_page_num(notebook, thread_view_page);
  g_return_if_fail(page != -1);
  gtk_notebook_set_current_page(notebook, page);
}


BoardPropertiesUI::BoardPropertiesUI(Controller &controller, Preferences &prefs)
  : controller(controller),
    global_show_mailto_mode(prefs.get_show_mailto())
{
  g_signal_connect(prefs.show_mailto, "toggled",
		   reinterpret_cast<GCallback>(show_mailto_toggled_cb), this);
}


BoardPropertiesUI::~BoardPropertiesUI()
{
  assert(properties_list.begin() == properties_list.end());
}


void
BoardPropertiesUI::show_mailto_toggled_cb(GtkToggleAction *action,
					  BoardPropertiesUI *ui)
{
  bool show_mailto = gtk_toggle_action_get_active(action);
  if (show_mailto == ui->global_show_mailto_mode)
    return;	// nothing has changed.

  ui->global_show_mailto_mode = show_mailto;

  list<BoardProperties *>::iterator iter = ui->properties_list.begin();
  list<BoardProperties *>::iterator end_iter = ui->properties_list.end();

  while (iter != end_iter)
    (*iter++)->update_show_mailto(show_mailto);
}




static GQuark properties_id
  = g_quark_from_static_string("BoardPropertiesUI::BoardProperties");


void
BoardPropertiesUI::process_response(GtkDialog *dialog, int response_id,
				    BulletinBoard *board)
{
  BoardProperties *properties
    = static_cast<BoardProperties *>(g_object_get_qdata(G_OBJECT(dialog),
							properties_id));
  g_return_if_fail(properties != NULL);
  properties->process_response(dialog, response_id, board);
}


void
BoardPropertiesUI::construct_properties_contents(GtkDialog *dialog,
						 BulletinBoard *board)
{
  BoardProperties *properties
    = new BoardProperties(*this, dialog, board);
  g_object_set_qdata(G_OBJECT(dialog), properties_id, properties);
  properties_list.push_front(properties);
}


BoardProperties *
BoardPropertiesUI::lookup_properties(BulletinBoard *board)
{
  {
    list<BoardProperties *>::iterator iter = properties_list.begin();
    list<BoardProperties *>::iterator end_iter = properties_list.end();

    while (iter != end_iter)
      {
	BoardProperties *properties = *iter++;
	if (properties->get_board() == board)
	  return properties;
      }
  }

  // threadlist.xmlɤ߹˼ưİžбURLѹ򤹤
  // ƼƤޤΤ
  GDK_THREADS_LEAVE();

  board->read_threadlist_xml(controller.get_repository());

  GDK_THREADS_ENTER();


  controller.get_boardlist_ui().open_properties(board);

  // ƤϤ
  list<BoardProperties *>::iterator iter = properties_list.begin();
  list<BoardProperties *>::iterator end_iter = properties_list.end();

  while (iter != end_iter)
    {
      BoardProperties *properties = *iter++;
      if (properties->get_board() == board)
	return properties;
    }

  return NULL;
}


void
BoardPropertiesUI::confirm_addition_of_board(BulletinBoard *board)
{
  if (board->get_base_url() == NULL)
    return;

  controller.get_bbs_table().register_board(board, board->get_base_url());
  controller.get_boardlist_ui().add_element(board, _("Add Board"));
}


void
BoardPropertiesUI::add_board(BulletinBoard *board)
{
  BoardProperties *properties = lookup_properties(board);
  if (properties != NULL)
    properties->set_add_board(true);
}


void
BoardPropertiesUI::append_id_to_be_a_boned(BulletinBoard *board, const char *id)
{
  BoardProperties *properties = lookup_properties(board);
  if (properties != NULL)
    properties->append_id_to_be_a_boned(id);
}


void
BoardPropertiesUI::properties_destroyed(BoardProperties *properties)
{
  properties_list.erase(find(properties_list.begin(),
			     properties_list.end(), properties));
}

}
