/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: board_ui.cc,v 1.28 2009/01/05 12:18:03 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "board_ui.h"

#include "libochusha/lock.h"
#include "libochusha/ochusha_procedure.h"

#include "controller.h"
#include "gtk_utils.h"
#include "threadlist_view.h"
#include "ui_constants.h"
#include "view.h"

#include "libochushagtk/combo_action.h"

#include <algorithm>
#include <list>

#include <cassert>
#include <cstdio>
#include <cstdlib>


namespace Ochusha
{


enum {
  THREADLIST_RULE_LABEL = 0,
  THREADLIST_RULE_NAME,
  THREADLIST_RULE_DESCRIPTION,
  THREADLIST_RULE_N_COLUMNS,
};


static GQuark rule_name_id;


BoardUI::BoardUI(Controller &controller)
  : controller(controller),
    recovery_manager(NULL),
    action_group(create_action_group())
{
  rule_name_id = g_quark_from_static_string("BoardUI::RuleName");

  g_signal_connect(controller.get_prefs().show_hidden_threads, "toggled",
		   reinterpret_cast<GCallback>(show_hidden_threads_toggled_cb),
		   this);
}


BoardUI::~BoardUI()
{
  g_object_unref(threadlist_rule_model);
  g_object_unref(default_threadlist_rule_model);
  g_object_unref(threadlist_rule);
  g_object_unref(default_threadlist_rule);
  g_object_unref(action_group);
}


static void
set_label_text(GtkMenuItem *item,
	       GtkLabel *widget, GtkTreeModel *model, GtkTreeIter *iter,
	       gpointer unused)
{
  const char *label = NULL;
  gtk_tree_model_get(model, iter, THREADLIST_RULE_LABEL, &label, -1);
  gtk_label_set_text(GTK_LABEL(widget), label);
}


static void
set_combo_text(GtkCellLayout *layout, GtkCellRenderer *renderer,
	       GtkTreeModel *model, GtkTreeIter *iter, gpointer unused)
{
  const char *label = NULL;
  gtk_tree_model_get(model, iter, THREADLIST_RULE_LABEL, &label, -1);
  if (label == NULL)
    label = _("Implementation Error?");

  g_object_set(renderer, "text", label, NULL);
}


static void
set_combo_tooltip(GtkTooltip *tooltip, GtkTreeModel *model, GtkTreeIter *iter,
		  gpointer unused)
{
  const char *desc = NULL;
  gtk_tree_model_get(model, iter, THREADLIST_RULE_DESCRIPTION, &desc, -1);
  if (desc == NULL)
    desc = _("Implementation Error?");

  gtk_tooltip_set_text(tooltip, desc);
}


GtkActionGroup *
BoardUI::create_action_group()
{
  static GtkActionEntry entries[] = {
    {
      "GoForwardThreadlist", GTK_STOCK_GO_FORWARD,
      N_("Go _Forward Threadlist"), "",
      N_("Go Forward Threadlist"),
      reinterpret_cast<GCallback>(go_forward_cb)
    },
    {
      "GoBackThreadlist", GTK_STOCK_GO_BACK,
      N_("Go _Back Threadlist"), "",
      N_("Go Back Threadlist"),
      reinterpret_cast<GCallback>(go_back_cb)
    },
    {
      "NewThread", GTK_STOCK_NEW,
      N_("_New Thread"), "",
      N_("New Thread"),
      reinterpret_cast<GCallback>(new_thread_cb)
    },
    {
      "BoardProperties", GTK_STOCK_PROPERTIES,
      N_("Board _Properties"), "",
      N_("Board Properties"),
      reinterpret_cast<GCallback>(board_properties_cb)
    },
    {
      "RestoreAllBoards", GTK_STOCK_EXECUTE,
      N_("Restore All Boards"), "",
      NULL,
      reinterpret_cast<GCallback>(restore_all_boards_cb)
    }
  };

  action_group = gtk_action_group_new("BoardActions");
  gtk_action_group_set_translation_domain(action_group, PACKAGE_NAME);
  gtk_action_group_add_actions(action_group, entries,
			       G_N_ELEMENTS(entries), this);

  go_forward = gtk_action_group_get_action(action_group, "GoForwardThreadlist");
  gtk_action_set_sensitive(go_forward, FALSE);

  go_back = gtk_action_group_get_action(action_group, "GoBackThreadlist");
  gtk_action_set_sensitive(go_back, FALSE);

  new_thread = gtk_action_group_get_action(action_group, "NewThread");
  gtk_action_set_sensitive(new_thread, FALSE);

  board_properties = gtk_action_group_get_action(action_group,
						 "BoardProperties");


  GtkListStore *store = gtk_list_store_new(THREADLIST_RULE_N_COLUMNS,
					   G_TYPE_POINTER,
					   G_TYPE_POINTER,
					   G_TYPE_POINTER);
  default_threadlist_rule_model = GTK_TREE_MODEL(store);

  OchushaComboAction *combo_action
    = ochusha_combo_action_new_with_model("DefaultThreadlistRule",
					  _("_Default Threadlist Rule"),
					  _("Default Threadlist Rule"),
					  NULL,
					  default_threadlist_rule_model);
  ochusha_combo_action_set_menu_data_func(combo_action,
					  set_label_text, this, NULL);
  ochusha_combo_action_set_cell_data_func(combo_action,
					  set_combo_text, this, NULL);
  ochusha_combo_action_set_tooltip_data_func(combo_action,
					     set_combo_tooltip, this, NULL);
  g_signal_connect(combo_action, "changed",
		   reinterpret_cast<GCallback>(default_threadlist_rule_changed_cb),
		   this);

  default_threadlist_rule = GTK_ACTION(combo_action);
  gtk_action_group_add_action_with_accel(action_group, default_threadlist_rule,
					 NULL);


  store = gtk_list_store_new(THREADLIST_RULE_N_COLUMNS,
			     G_TYPE_POINTER,
			     G_TYPE_POINTER,
			     G_TYPE_POINTER);
  threadlist_rule_model = GTK_TREE_MODEL(store);

  combo_action = ochusha_combo_action_new_with_model("ThreadlistRule",
						     _("Current Threadlist _Rule"),
						     _("Current Threadlist Rule"),
						     NULL,
						     threadlist_rule_model);
  ochusha_combo_action_set_menu_data_func(combo_action,
					  set_label_text, this, NULL);
  ochusha_combo_action_set_cell_data_func(combo_action,
					  set_combo_text, this, NULL);
  ochusha_combo_action_set_tooltip_data_func(combo_action,
					     set_combo_tooltip, this, NULL);
  g_signal_connect(combo_action, "changed",
		   reinterpret_cast<GCallback>(threadlist_rule_changed_cb),
		   this);

  threadlist_rule = GTK_ACTION(combo_action);
  gtk_action_group_add_action_with_accel(action_group, threadlist_rule, NULL);


  gtk_ui_manager_insert_action_group(controller.get_ui_manager(),
				     action_group, 0);
  return action_group;
}


void
BoardUI::go_forward_cb(GtkAction *action, BoardUI *ui)
{
  View *view = ui->controller.get_view_manager().get_current_board_view();
  g_return_if_fail(view != NULL);

  view->go_forward_view();
}


void
BoardUI::go_back_cb(GtkAction *action, BoardUI *ui)
{
  View *view = ui->controller.get_view_manager().get_current_board_view();
  g_return_if_fail(view != NULL);

  view->go_back_view();
}


void
BoardUI::new_thread_cb(GtkAction *action, BoardUI *ui)
{
  View *view = ui->controller.get_view_manager().get_current_board_view();
  g_return_if_fail(view != NULL);

  ThreadlistView *threadlist_view = dynamic_cast<ThreadlistView *>(view);
  g_return_if_fail(threadlist_view != NULL);

  threadlist_view->show_new_thread_window();
}


void
BoardUI::board_properties_cb(GtkAction *action, BoardUI *ui)
{
  BulletinBoard *board = ui->controller.get_view_manager().get_current_board();
  if (board != NULL)
    ui->controller.get_boardlist_ui().open_properties(board);
}


void
BoardUI::restore_all_boards_cb(GtkAction *action, BoardUI *ui)
{
  ui->restore_all_boards();
}


void
BoardUI::prepare_session()
{
}


View *
BoardUI::create_board_view(BulletinBoard *board, TabLabel *label, bool refresh)
{
  return new ThreadlistView(controller, board, label, refresh);
}


void
BoardUI::notify_current_board_view(View *view, BulletinBoard *board)
{
  if (board == NULL)
    {
      gtk_action_set_sensitive(go_forward, FALSE);
      gtk_action_set_sensitive(go_back, FALSE);
      gtk_action_set_sensitive(new_thread, FALSE);
      gtk_action_set_sensitive(board_properties, FALSE);
      gtk_action_set_sensitive(threadlist_rule, FALSE);
      set_current_threadlist_rule(NULL);
    }
  else
    {
      gtk_action_set_sensitive(go_forward, view->has_forward_view());
      gtk_action_set_sensitive(go_back, view->has_back_view());

      switch (board->get_bbs_type())
	{
	case OCHUSHA_BBS_TYPE_2CH:
	case OCHUSHA_BBS_TYPE_2CHLIKE_EUCJP:
	case OCHUSHA_BBS_TYPE_2CH_COMPATIBLE:
	case OCHUSHA_BBS_TYPE_2CH_BE:
	case OCHUSHA_BBS_TYPE_BBSPINK:
	case OCHUSHA_BBS_TYPE_MACHIBBS:
	case OCHUSHA_BBS_TYPE_JBBS_LIVEDOOR:
	  gtk_action_set_sensitive(new_thread, TRUE);
	  break;

	default:
	  gtk_action_set_sensitive(new_thread, FALSE);
	  break;
	}

      gtk_action_set_sensitive(board_properties, TRUE);
      gtk_action_set_sensitive(threadlist_rule, TRUE);
      set_current_threadlist_rule(board->get_rule());
    }
}


void
BoardUI::default_threadlist_rule_changed_cb(OchushaComboAction *action,
					    GtkTreeModel *model,
					    GtkTreeIter *iter,
					    GtkTreePath *path,
					    BoardUI *ui)
{
  const char *name = NULL;
  gtk_tree_model_get(model, iter, THREADLIST_RULE_NAME, &name, -1);

  g_return_if_fail(name != NULL);
  ui->set_default_threadlist_rule(name);
}


void
BoardUI::threadlist_rule_changed_cb(OchushaComboAction *action,
				    GtkTreeModel *model,
				    GtkTreeIter *iter,
				    GtkTreePath *path,
				    BoardUI *ui)
{
  const char *name = NULL;
  gtk_tree_model_get(model, iter, THREADLIST_RULE_NAME, &name, -1);

  BulletinBoard *board = ui->controller.get_view_manager().get_current_board();
  if (board != NULL)
    {
      const char *previous_rule = board->get_rule();
      if (previous_rule == NULL)
	previous_rule
	  = ui->controller.get_prefs().get_default_threadlist_rule();
      const char *effective_rule
	= name != NULL
	? name : ui->controller.get_prefs().get_default_threadlist_rule();

      bool redraw = strcmp(previous_rule, effective_rule) != 0;

      board->set_rule(name);
      if (redraw)
	ui->controller.refresh(board, false);
    }

  ui->set_current_threadlist_rule(name);
}


void
BoardUI::set_current_threadlist_rule(const char *name)
{
  GtkTreeIter iter;
  gboolean result = gtk_tree_model_get_iter_first(threadlist_rule_model, &iter);

  if (result)
    {
      if (name != NULL)
	{
	  result = FALSE;
	  while (gtk_tree_model_iter_next(threadlist_rule_model, &iter))
	    {
	      const char *rule_name = NULL;
	      gtk_tree_model_get(threadlist_rule_model, &iter,
				 THREADLIST_RULE_NAME, &rule_name, -1);
	      g_return_if_fail(rule_name != NULL);

	      if (strcmp(rule_name, name) == 0)
		{
		  result = TRUE;
		  break;
		}
	    }
	}

      if (result)
	ochusha_combo_action_set_active_iter(OCHUSHA_COMBO_ACTION(threadlist_rule),
					     &iter);
    }
}


static void
redraw_threadlist(GtkWidget *view, BulletinBoard *board, void *data)
{
  if (board->get_rule() == NULL)
    {
      Controller *controller = static_cast<Controller *>(data);
      controller->refresh(board, false);
    }
}


void
BoardUI::set_default_threadlist_rule(const char *name)
{
  const OchushaProcedure &procedure
    = controller.get_scheme_manager().get_ochusha_procedure();
  const ThreadlistRuleInfo *info = procedure.lookup_threadlist_rule_info(name);
  assert(info != NULL);

  const char *previous_rule
    = controller.get_prefs().get_default_threadlist_rule();
  bool rule_changed = previous_rule == NULL || strcmp(previous_rule, name) != 0;
  if (rule_changed)
    {
      GtkTreeIter iter;
      if (gtk_tree_model_get_iter_first(threadlist_rule_model, &iter))
	{
	  gchar *label = NULL;
	  gchar *desc = NULL;
	  gtk_tree_model_get(threadlist_rule_model, &iter,
			     THREADLIST_RULE_LABEL, &label,
			     THREADLIST_RULE_DESCRIPTION, &desc,
			     -1);
	  g_free(label);
	  g_free(desc);
	  label = g_strdup_printf("%s (*)", info->get_label());
	  desc = g_strdup_printf(_("%s [Global Default]"),
				 info->get_description());
	  gtk_list_store_set(GTK_LIST_STORE(threadlist_rule_model), &iter,
			     THREADLIST_RULE_LABEL, label,
			     THREADLIST_RULE_NAME, NULL,
			     THREADLIST_RULE_DESCRIPTION, desc,
			     -1);
	}

      controller.get_prefs().set_default_threadlist_rule(name);
      controller.get_view_manager().foreach_board(redraw_threadlist,
						  &controller);
      BulletinBoard *board = controller.get_view_manager().get_current_board();
      if (board == NULL || board->get_rule() == NULL)
	set_current_threadlist_rule(NULL);
    }
}


void
BoardUI::update_threadlist_rules()
{
  const OchushaProcedure &procedure
    = controller.get_scheme_manager().get_ochusha_procedure();

  vector<ThreadlistRuleInfo *>::const_iterator iter
    = procedure.threadlist_rule_list_begin();
  vector<ThreadlistRuleInfo *>::const_iterator end_iter
    = procedure.threadlist_rule_list_end();

  const char *default_rule_name
    = controller.get_prefs().get_default_threadlist_rule();
  assert(default_rule_name != NULL);


  GtkTreeIter dummy_iter;
  GtkListStore *current_store = GTK_LIST_STORE(threadlist_rule_model);
  gtk_list_store_insert_with_values(current_store, &dummy_iter, G_MAXINT32,
				    THREADLIST_RULE_LABEL, g_strdup("dummy"),
				    THREADLIST_RULE_NAME, NULL,
				    THREADLIST_RULE_DESCRIPTION, NULL,
				    -1);

  GtkListStore *default_store = GTK_LIST_STORE(default_threadlist_rule_model);
  ThreadlistRuleInfo *default_info = NULL;
  while (iter != end_iter)
    {
      ThreadlistRuleInfo *info = *iter++;

      gtk_list_store_insert_with_values(current_store, &dummy_iter, G_MAXINT32,
			THREADLIST_RULE_LABEL, info->get_label(),
			THREADLIST_RULE_NAME, info->get_name(),
			THREADLIST_RULE_DESCRIPTION, info->get_description(),
			-1);

      GtkTreeIter tree_iter;
      gtk_list_store_insert_with_values(default_store, &tree_iter, G_MAXINT32,
			THREADLIST_RULE_LABEL, info->get_label(),
			THREADLIST_RULE_NAME, info->get_name(),
			THREADLIST_RULE_DESCRIPTION, info->get_description(),
			-1);
      if (strcmp(info->get_name(), default_rule_name) == 0)
	{
	  default_info = info;
	  ochusha_combo_action_set_active_iter(OCHUSHA_COMBO_ACTION(default_threadlist_rule),
					       &tree_iter);
	}
    }

  if (default_info != NULL)
    {
      GtkTreeIter tree_iter;
      if (gtk_tree_model_get_iter_first(threadlist_rule_model, &tree_iter))
	{
	  gchar *label = NULL;
	  gchar *desc = NULL;
	  gtk_tree_model_get(threadlist_rule_model, &tree_iter,
			     THREADLIST_RULE_LABEL, &label,
			     THREADLIST_RULE_DESCRIPTION, &desc,
			     -1);
	  if (label != NULL)
	    g_free(label);

	  if (desc != NULL)
	    g_free(desc);

	  label = g_strdup_printf("%s (*)", default_info->get_label());
	  desc = g_strdup_printf(_("%s [Global Default]"),
				 default_info->get_description());
	  gtk_list_store_set(current_store, &tree_iter,
			     THREADLIST_RULE_LABEL, label,
			     THREADLIST_RULE_NAME, NULL,
			     THREADLIST_RULE_DESCRIPTION, desc,
			     -1);
	}
    }
}


static void
redraw_threadlist_all(GtkWidget *view, BulletinBoard *board, void *data)
{
  static_cast<Controller *>(data)->refresh(board, false);
}


void
BoardUI::show_hidden_threads_toggled_cb(GtkAction *action, BoardUI *ui)
{
  ui->controller.get_view_manager().foreach_board(redraw_threadlist_all,
						  &ui->controller);
}


class DATFileExplorerWatcher;

class BoardRecoveryManager
{
  friend class DATFileExplorerWatcher;

private:
  BoardUI &ui;

  GtkWidget *dialog;

  GtkListStore *list_store;
  GtkWidget *tree_view;

  GtkTextBuffer *text_buffer;
  GtkWidget *text_view;
  GtkTextIter tail_iter;
  GtkTextMark *tail_mark;

  GtkWidget *progress_bar;

  BulletinBoard *current_board;
  DATFileExplorerWatcher *current_watcher;

  int number_of_boards;

  BulletinBoardList pending_boards;
  BulletinBoardList busy_boards;


  static void response_cb(GtkWidget *widget, int response_id,
			  BoardRecoveryManager *manager);
  static void column_set_board_name(GtkTreeViewColumn *column,
				    GtkCellRenderer *renderer,
				    GtkTreeModel *model, GtkTreeIter *iter,
				    gpointer manager);

  void start_recovery();

  void update_board_list();
  void update_progress_bar();

  bool look_for_board_iter(GtkTreeIter *iter, BulletinBoard *board);
  void output_message(const gchar *message);


  // DATFileExplorerWatcherΥХå
  void notify_dat_file_found(BulletinBoard *board, BBSThread *thread);
  void notify_information_recovered(BulletinBoard *board, BBSThread *thread);
  void notify_finished(BulletinBoard *board, DATFileExplorerWatcher *watcher);


public:
  BoardRecoveryManager(BoardUI &ui);
  ~BoardRecoveryManager();

  void show();
};


class DATFileExplorerWatcher : public has_slots
{
  friend class BoardRecoveryManager;

private:
  Controller &controller;
  BoardRecoveryManager *manager;


public:
  DATFileExplorerWatcher(Controller &controller, DATFileExplorer &explorer,
			 BoardRecoveryManager *manager = NULL);

  // DATFileExplorersigslotʥΥϥɥ
  void dat_file_found_cb(BulletinBoard *board, BBSThread *thread);
  void information_recovered_cb(BulletinBoard *board, BBSThread *thread);
  void finished_cb(BulletinBoard *board);
};


DATFileExplorerWatcher::DATFileExplorerWatcher(Controller &controller,
					       DATFileExplorer &explorer,
					       BoardRecoveryManager *manager)
  : controller(controller),
    manager(manager)
{
  explorer.dat_file_found
    .connect(this, &DATFileExplorerWatcher::dat_file_found_cb);
  explorer.information_recovered
    .connect(this, &DATFileExplorerWatcher::information_recovered_cb);
  explorer.finished
    .connect(this, &DATFileExplorerWatcher::finished_cb);
}


void
DATFileExplorerWatcher::dat_file_found_cb(BulletinBoard *board,
					  BBSThread *thread)
{
  // sigslotʥϥɥ
  if (manager != NULL)
    {
      manager->notify_dat_file_found(board, thread);
      return;
    }

  char message[1024];

  const char *title = thread->get_title();
  if (title == NULL)
    title = _("Anonymous Thread");

  const char *name = board->get_name();
  if (name == NULL)
    name = _("Anonymous Board");

  snprintf(message, 1024,
	   _("DAT file of \"%s\" thread in \"%s\" board is excavated.\n"),
	   title, name);


  GDK_THREADS_ENTER();

  controller.notification.output(message);
  controller.notification.show();

  GDK_THREADS_LEAVE();
}


void
DATFileExplorerWatcher::information_recovered_cb(BulletinBoard *board,
						 BBSThread *thread)
{
  // sigslotʥϥɥ
  if (manager != NULL)
    {
      manager->notify_information_recovered(board, thread);
      return;
    }

  char message[1024];

  const char *title = thread->get_title();
  if (title == NULL)
    title = _("Anonymous Thread");

  const char *name = board->get_name();
  if (name == NULL)
    name = _("Anonymous Board");

  snprintf(message, 1024,
	   _("Information about \"%s\" thread in \"%s\" board is recovered.\n"),
	   title, name);


  GDK_THREADS_ENTER();

  controller.notification.output(message);
  controller.notification.show();

  GDK_THREADS_LEAVE();
}


void
DATFileExplorerWatcher::finished_cb(BulletinBoard *board)
{
  // sigslotʥϥɥ
  if (manager != NULL)
    {
      manager->notify_finished(board, this);
      delete this;
      return;
    }

  char message[1024];
  const char *name = board->get_name();
  if (name == NULL)
    name = _("Anonymous Board");

  snprintf(message, 1024, _("DAT file excavation for \"%s\" board finished.\n"),
	   name);


  GDK_THREADS_ENTER();

  controller.notification.output(message);
  controller.notification.show();

  GDK_THREADS_LEAVE();

  delete this;
}


BoardRecoveryManager::BoardRecoveryManager(BoardUI &ui)
  : ui(ui),
    dialog(NULL),
    list_store(NULL),
    tree_view(NULL),
    text_buffer(NULL),
    text_view(NULL),
    tail_mark(NULL),
    progress_bar(NULL),
    current_board(NULL),
    current_watcher(NULL),
    number_of_boards(0)
{
  dialog = gtk_message_dialog_new(ui.controller.get_toplevel(),
				  GTK_DIALOG_DESTROY_WITH_PARENT,
				  GTK_MESSAGE_INFO,
				  GTK_BUTTONS_OK_CANCEL,
				  _("Recover Information for All Boards"));
  gtk_message_dialog_format_secondary_text(GTK_MESSAGE_DIALOG(dialog),
					   _("The recovering process DO explore of DAT files you have and parse them if needed.  This may take a while.  How long it takes is depending on your situation, i.e. more lost information need more time.  You can continue using Ochusha during this process any way.  Of course, you can cancel this process any time you want."));

  g_signal_connect(dialog, "response",
		   reinterpret_cast<GCallback>(response_cb), this);
  gtk_dialog_set_response_sensitive(GTK_DIALOG(dialog), GTK_RESPONSE_OK, FALSE);

  //  gtk_window_set_default_size(GTK_WINDOW(dialog), 640, 480);
  gtk_window_set_default_size(GTK_WINDOW(dialog), -1, 400);
  gtk_window_set_resizable(GTK_WINDOW(dialog), TRUE);
  gtk_window_set_title(GTK_WINDOW(dialog),
		       _("Recover Information for All Boards"));



  GtkWidget *paned = gtk_hpaned_new();



  list_store = gtk_list_store_new(2,
				  G_TYPE_POINTER,	// BulletinBoard *
				  G_TYPE_INT);		// flag

  BBSTable &table = ui.controller.get_bbs_table();
  TreeElementVector::iterator iter = table.all_boards_begin();
  TreeElementVector::iterator end_iter = table.all_boards_end();

  while (iter != end_iter)
    {
      TreeElement *elm = *iter++;
      BulletinBoard *board = dynamic_cast<BulletinBoard *>(elm);
      if (board != NULL && board->get_bbs_type() >= OCHUSHA_BBS_TYPE_2CH)
	pending_boards.push_back(board);
    }

  number_of_boards = pending_boards.size();
  update_board_list();


  tree_view = gtk_tree_view_new_with_model(GTK_TREE_MODEL(list_store));

  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  GtkTreeViewColumn *column
    = gtk_tree_view_column_new_with_attributes(_("Bulletin Board"), renderer, NULL);
  gtk_tree_view_column_set_resizable(column, FALSE);
  gtk_tree_view_column_set_cell_data_func(column, renderer,
					  column_set_board_name, this, NULL);
  gtk_tree_view_append_column(GTK_TREE_VIEW(tree_view), column);
  if (column->button != NULL)
    {
      GTK_WIDGET_UNSET_FLAGS(column->button, GTK_CAN_FOCUS);
      gtk_button_set_focus_on_click(GTK_BUTTON(column->button), FALSE);
    }


  GtkWidget *scrolled = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled),
				 GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

  gtk_widget_show(tree_view);
  gtk_container_add(GTK_CONTAINER(scrolled), tree_view);

  gtk_widget_show(scrolled);
  gtk_paned_pack1(GTK_PANED(paned), scrolled, TRUE, TRUE);



  text_buffer = gtk_text_buffer_new(NULL);
  gtk_text_buffer_get_start_iter(text_buffer, &tail_iter);
  tail_mark = gtk_text_buffer_create_mark(text_buffer, "tail-mark",
					  &tail_iter, FALSE);
  text_view = gtk_text_view_new_with_buffer(text_buffer);
  new TextViewKeyPressHandler(GTK_TEXT_VIEW(text_view), ui.controller);

  scrolled = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled),
				 GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

  gtk_widget_show(text_view);
  gtk_container_add(GTK_CONTAINER(scrolled), text_view);

  gtk_widget_show(scrolled);
  gtk_paned_pack2(GTK_PANED(paned), scrolled, TRUE, TRUE);



#if GTK_CHECK_VERSION(2,14,0)
  GtkWidget *vbox = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
#else
  GtkWidget *vbox = GTK_DIALOG(dialog)->vbox;
#endif

  gtk_widget_show(paned);
  gtk_box_pack_start(GTK_BOX(vbox), paned, TRUE, TRUE, 0);


  progress_bar = gtk_progress_bar_new();
  update_progress_bar();

  gtk_widget_show(progress_bar);
  gtk_box_pack_start(GTK_BOX(vbox), progress_bar, FALSE, FALSE, 0);


  gtk_widget_show(dialog);

  start_recovery();
}


BoardRecoveryManager::~BoardRecoveryManager()
{
  g_object_unref(text_buffer);
  g_object_unref(list_store);
}


void
BoardRecoveryManager::show()
{
  g_return_if_fail(dialog != NULL);

  gtk_window_present(GTK_WINDOW(dialog));
}


void
BoardRecoveryManager::response_cb(GtkWidget *widget, int response_id,
				  BoardRecoveryManager *manager)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  if (response_id == GTK_RESPONSE_OK)
    {
      assert(manager->current_watcher == NULL);
      assert(manager->current_board == NULL);
      assert(manager->busy_boards.empty());

      if (!manager->pending_boards.empty())
	{
	  gtk_dialog_set_response_sensitive(GTK_DIALOG(widget),
					    GTK_RESPONSE_OK, FALSE);
	  manager->start_recovery();

	  return;
	}
    }
  else
    {
      manager->pending_boards.clear();
      manager->busy_boards.clear();
    }

  assert(manager->pending_boards.empty());
  assert(manager->busy_boards.empty());

  while (manager->current_watcher != NULL)
    ochusha_gdk_threads_lock.wait();

  gtk_widget_destroy(manager->dialog);

  g_return_if_fail(manager->ui.recovery_manager == manager);

  manager->ui.recovery_manager = NULL;
  delete manager;
}


void
BoardRecoveryManager::column_set_board_name(GtkTreeViewColumn *column,
					    GtkCellRenderer *renderer,
					    GtkTreeModel *model,
					    GtkTreeIter *iter,
					    gpointer data)
{
  BoardRecoveryManager *manager = static_cast<BoardRecoveryManager *>(data);
  BulletinBoard *board = NULL;
  gint flags = 0;
  gtk_tree_model_get(model, iter, 0, &board, 1, &flags, -1);
  g_return_if_fail(board != NULL);

  g_object_set(renderer,
	       "text", board->get_name(),
	       "strikethrough", flags == 1,
	       "strikethrough-set", flags == 1,
	       "weight", manager->current_board == board ? 600 : 400,
	       "weight-set", manager->current_board == board,
	       NULL);
}


bool
BoardRecoveryManager::look_for_board_iter(GtkTreeIter *iter,
					  BulletinBoard *board)
{
  g_return_val_if_fail(iter != NULL, false);
  g_return_val_if_fail(board != NULL, false);

  GtkTreeModel *model = GTK_TREE_MODEL(list_store);
  if (!gtk_tree_model_get_iter_first(model, iter))
    return false;

  do
    {
      BulletinBoard *iter_board = NULL;
      gtk_tree_model_get(model, iter, 0, &iter_board, -1);
      g_return_val_if_fail(iter_board != NULL, false);

      if (iter_board == board)
	return true;
    }
  while (gtk_tree_model_iter_next(model, iter));

  return false;
}


void
BoardRecoveryManager::start_recovery()
{
  assert(OCHUSHA_IN_GDK_THREADS());

  char message[1024];
  Repository &repository = ui.controller.get_repository();
  Configuration &config = ui.controller.get_config();
  while (!pending_boards.empty())
    {
      BulletinBoard *board = pending_boards.front();
      pending_boards.pop_front();


      GDK_THREADS_LEAVE();

      board->read_threadlist_xml(repository);

      DATFileExplorer *explorer
	= board->get_dat_file_explorer(config, repository);
      DATFileExplorerWatcher *watcher;
      if (explorer != NULL)
	watcher = new DATFileExplorerWatcher(ui.controller, *explorer, this);
      else
	watcher = NULL;

      GDK_THREADS_ENTER();


      if (explorer != NULL)
	{
	  assert(watcher != NULL);

	  current_watcher = watcher;
	  current_board = board;

	  const char *name = board->get_name();
	  if (name == NULL)
	    name = _("Anonymous Board");

	  snprintf(message, 1024,
		   _("Starting recovering process of \"%s\" board.\n"), name);
	  output_message(message);

	  GtkTreeIter iter;
	  if (look_for_board_iter(&iter, board))
	    {
	      TreePathHandle path
		= gtk_tree_model_get_path(GTK_TREE_MODEL(list_store), &iter);
	      if (path != NULL)
		{
		  gtk_tree_view_set_cursor(GTK_TREE_VIEW(tree_view), path,
					   NULL, FALSE);
		  gtk_tree_view_scroll_to_cell(GTK_TREE_VIEW(tree_view), path,
					       NULL, TRUE, 0.5, 0.0);
		}
	    }


	  GDK_THREADS_LEAVE();

	  explorer->explore(ui.controller.get_thread_pool());

	  GDK_THREADS_ENTER();


	  return;
	}

      const char *name = board->get_name();
      if (name == NULL)
	name = _("Anonymous Board");

      snprintf(message, 1024,
	       _("\"%s\" is busy: recovering process of this board is pended\n"),
	       name);
      output_message(message);

      busy_boards.push_back(board);
    }

  current_watcher = NULL;
  current_board = NULL;

  if (busy_boards.empty())
    {
      gtk_message_dialog_format_secondary_text(GTK_MESSAGE_DIALOG(dialog),
					       _("The recovering process has successfuly done.\n"));
      gtk_dialog_set_response_sensitive(GTK_DIALOG(dialog),
					GTK_RESPONSE_OK, TRUE);
      gtk_dialog_set_response_sensitive(GTK_DIALOG(dialog),
					GTK_RESPONSE_CANCEL, FALSE);
      return;
    }


  gtk_message_dialog_format_secondary_text(GTK_MESSAGE_DIALOG(dialog),
					   _("There's %d board(s) pended for busy situation.  Choose OK if you want to retry recovering of them.\n"),
					   busy_boards.size());

  while (!busy_boards.empty())
    {
      pending_boards.push_back(busy_boards.front());
      busy_boards.pop_front();
    }

  update_board_list();
  update_progress_bar();

  gtk_dialog_set_response_sensitive(GTK_DIALOG(dialog),
				    GTK_RESPONSE_OK, TRUE);
}


void
BoardRecoveryManager::update_board_list()
{
  if (tree_view != NULL)
    gtk_tree_view_set_model(GTK_TREE_VIEW(tree_view), NULL);

  gtk_list_store_clear(list_store);

  BulletinBoardList::const_iterator iter = pending_boards.begin();
  BulletinBoardList::const_iterator end_iter = pending_boards.end();
  while (iter != end_iter)
    gtk_list_store_insert_with_values(list_store, NULL, G_MAXINT32,
				      0, *iter++, 1, 0, -1);

  if (tree_view != NULL)
    gtk_tree_view_set_model(GTK_TREE_VIEW(tree_view),
			    GTK_TREE_MODEL(list_store));
}


void
BoardRecoveryManager::update_progress_bar()
{
  g_return_if_fail(progress_bar != NULL);

  int rest = pending_boards.size() + busy_boards.size();
  g_return_if_fail(rest <= number_of_boards);

  double fraction
    = (number_of_boards > 0)
    ? (static_cast<double>(number_of_boards - rest)
       / static_cast<double>(number_of_boards))
    : 1.0;

  gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(progress_bar), fraction);

  char text[128];
  snprintf(text, 128, _("%d %% done (%d/%d)"),
	   static_cast<int>(fraction * 100),
	   number_of_boards - rest, number_of_boards);

  gtk_progress_bar_set_text(GTK_PROGRESS_BAR(progress_bar), text);
}


void
BoardRecoveryManager::output_message(const gchar *message)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  if (text_view == NULL)
    {
      gtk_text_buffer_insert(text_buffer, &tail_iter, message, -1);
      return;
    }

  g_return_if_fail(GTK_IS_TEXT_VIEW(text_view));

  GdkRectangle rect;
  gtk_text_view_get_visible_rect(GTK_TEXT_VIEW(text_view), &rect);

  gint y;
  gint height;
  gtk_text_view_get_line_yrange(GTK_TEXT_VIEW(text_view), &tail_iter,
				&y, &height);

  gtk_text_buffer_insert(text_buffer, &tail_iter, message, -1);

  if (rect.y + rect.height >= y)
    gtk_text_view_scroll_to_mark(GTK_TEXT_VIEW(text_view), tail_mark,
				 0.3, FALSE, 0.0, 0.0);
}


void
BoardRecoveryManager::notify_dat_file_found(BulletinBoard *board,
					    BBSThread *thread)
{
  assert(!OCHUSHA_IN_GDK_THREADS());

  char message[1024];
  const char *title = thread->get_title();
  if (title == NULL)
    title = _("Anonymous Thread");

  const char *name = board->get_name();
  if (name == NULL)
    name = _("Anonymous Board");

  snprintf(message, 1024,
	   _("DAT file of \"%s\" thread in \"%s\" board is excavated.\n"),
	   title, name);

  GDK_THREADS_ENTER();

  output_message(message);

  GDK_THREADS_LEAVE();
}


void
BoardRecoveryManager::notify_information_recovered(BulletinBoard *board,
						   BBSThread *thread)
{
  assert(!OCHUSHA_IN_GDK_THREADS());

  assert(!OCHUSHA_IN_GDK_THREADS());

  char message[1024];
  const char *title = thread->get_title();
  if (title == NULL)
    title = _("Anonymous Thread");

  const char *name = board->get_name();
  if (name == NULL)
    name = _("Anonymous Board");

  snprintf(message, 1024,
	   _("Information about \"%s\" thread in \"%s\" board is recovered.\n"),
	   title, name);

  GDK_THREADS_ENTER();

  output_message(message);

  GDK_THREADS_LEAVE();
}


void
BoardRecoveryManager::notify_finished(BulletinBoard *board,
				      DATFileExplorerWatcher *watcher)
{
  assert(!OCHUSHA_IN_GDK_THREADS());
  assert(current_board == board);
  assert(current_watcher == watcher);


  char message[1024];
  const char *name = board->get_name();
  if (name == NULL)
    name = _("Anonymous Board");

  snprintf(message, 1024, _("\"%s\" board was successfully recovered.\n"),
	   name);


  GDK_THREADS_ENTER();

  output_message(message);
  update_progress_bar();

  GtkTreeIter iter;
  if (look_for_board_iter(&iter, board))
    gtk_list_store_set(list_store, &iter, 1, 1, -1);

  start_recovery();

  ochusha_gdk_threads_lock.notify_all();

  GDK_THREADS_LEAVE();
}


void
BoardUI::excavate_dat_files(BulletinBoard *board)
{
  char message[1024];
  const char *name = board->get_name();
  if (name == NULL)
    name = _("Anonymous Board");


  GDK_THREADS_LEAVE();

  board->read_threadlist_xml(controller.get_repository());

  DATFileExplorer *explorer
    = board->get_dat_file_explorer(controller.get_config(),
				   controller.get_repository());
  if (explorer != NULL)
    {
      new DATFileExplorerWatcher(controller, *explorer);

      snprintf(message, 1024,
	       _("Starting DAT file excavation for \"%s\" board.\n"), name);
    }
  else
    snprintf(message, 1024,
	     _("Cannot start DAT file excavation for \"%s\": board is busy.\n"),
	     name);

  GDK_THREADS_ENTER();


  controller.notification.output(message);
  controller.notification.show();

  if (explorer == NULL)
    return;


  GDK_THREADS_LEAVE();

  explorer->explore(controller.get_thread_pool());

  GDK_THREADS_ENTER();
}


void
BoardUI::restore_all_boards()
{
  assert(OCHUSHA_IN_GDK_THREADS());

  if (recovery_manager != NULL)
    {
      recovery_manager->show();
      return;
    }

  recovery_manager = new BoardRecoveryManager(*this);
}


class BoardMovedDialog : public has_slots
{
private:
  Controller &controller;
  BulletinBoardHandle board;
  smart_ptr<BoardURLSeeker> seeker;

  char *new_url;

  GtkWidget *dialog;


  BoardMovedDialog(Controller &controller, BulletinBoard *board);
  static void response_cb(GtkWidget *widget, gint rid,
			  BoardMovedDialog *dialog);

  void url_found_cb(BoardURLSeeker *seeker, const char *url);
  void seek_failed_cb(BoardURLSeeker *seeker);


public:
  ~BoardMovedDialog();

  inline BulletinBoard *get_board() { return board; }
  static void show_dialog(Controller &controller, BulletinBoard *board);
};


static list<BoardMovedDialog *> dialog_list;


BoardMovedDialog::BoardMovedDialog(Controller &controller, BulletinBoard *board)
  : controller(controller),
    board(board),
    new_url(NULL)
{
  GDK_THREADS_ENTER();

  dialog = gtk_message_dialog_new(controller.get_toplevel(),
				  GTK_DIALOG_DESTROY_WITH_PARENT,
				  GTK_MESSAGE_INFO,
				  GTK_BUTTONS_OK_CANCEL,
				  _("%s seems to be moved."),
				  board->get_name());
  gtk_message_dialog_format_secondary_text(GTK_MESSAGE_DIALOG(dialog),
					   _("URL: %s\n=> %s"),
					   board->get_base_url(),
					   _("seeking"));
  gtk_window_set_title(GTK_WINDOW(dialog), _("Board Moved Information"));

  gtk_dialog_set_response_sensitive(GTK_DIALOG(dialog),
				    GTK_RESPONSE_OK, FALSE);

  g_signal_connect(dialog, "response",
		   reinterpret_cast<GCallback>(response_cb), this);

  GDK_THREADS_LEAVE();


  seeker = new BoardURLSeeker(board, controller.get_config(),
			      controller.get_broker().employ_agent());
  seeker->url_found.connect(this, &BoardMovedDialog::url_found_cb);
  seeker->seek_failed.connect(this, &BoardMovedDialog::seek_failed_cb);

  seeker->seek();


  GDK_THREADS_ENTER();

  gtk_widget_show(dialog);

  GDK_THREADS_LEAVE();
}


BoardMovedDialog::~BoardMovedDialog()
{
  list<BoardMovedDialog *>::iterator where
    = find(dialog_list.begin(), dialog_list.end(), this);
  g_return_if_fail(where != dialog_list.end());

  dialog_list.erase(where);

  if (new_url != NULL)
    free(new_url);
}


void
BoardMovedDialog::response_cb(GtkWidget *widget, gint rid,
			      BoardMovedDialog *dialog)
{
  g_return_if_fail(dialog->dialog == widget);

  gtk_widget_destroy(dialog->dialog);

  if (rid == GTK_RESPONSE_OK && dialog->new_url != NULL)
    dialog->controller.treat_board_movement(dialog->board, dialog->new_url,
					    false);


  GDK_THREADS_LEAVE();

  delete dialog;

  GDK_THREADS_ENTER();
}


void
BoardMovedDialog::url_found_cb(BoardURLSeeker *seeker, const char *url)
{
  // sigslotʥϥɥ
  GDK_THREADS_ENTER();

  gtk_message_dialog_format_secondary_text(GTK_MESSAGE_DIALOG(dialog),
					   _("URL: %s\n=> %s"),
					   board->get_base_url(),
					   url);
  GtkWidget *icon = gtk_image_new_from_stock(GTK_STOCK_DIALOG_QUESTION,
					     GTK_ICON_SIZE_DIALOG);
  gtk_widget_show(icon);
  gtk_message_dialog_set_image(GTK_MESSAGE_DIALOG(dialog), icon);
  gtk_dialog_set_response_sensitive(GTK_DIALOG(dialog),
				    GTK_RESPONSE_OK, TRUE);
  gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_OK);

  if (new_url != NULL)
    free(new_url);

  new_url = strdup(url);

  GDK_THREADS_LEAVE();
}


void
BoardMovedDialog::seek_failed_cb(BoardURLSeeker *seeker)
{
  // sigslotʥϥɥ
  GDK_THREADS_ENTER();

  gtk_message_dialog_format_secondary_text(GTK_MESSAGE_DIALOG(dialog),
					   _("URL: %s\n=> %s"),
					   board->get_base_url(),
					   _("cannot determine URL"));
  GtkWidget *icon = gtk_image_new_from_stock(GTK_STOCK_DIALOG_WARNING,
					     GTK_ICON_SIZE_DIALOG);
  gtk_widget_show(icon);
  gtk_message_dialog_set_image(GTK_MESSAGE_DIALOG(dialog), icon);

  GDK_THREADS_LEAVE();
}


void
BoardMovedDialog::show_dialog(Controller &controller, BulletinBoard *board)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  list<BoardMovedDialog *>::iterator iter = dialog_list.begin();
  list<BoardMovedDialog *>::iterator end_iter = dialog_list.end();
  while (iter != end_iter)
    {
      BoardMovedDialog *dialog = *iter++;
      if (dialog->get_board() == board)
	return;
    }


  GDK_THREADS_LEAVE();

  dialog_list.push_front(new BoardMovedDialog(controller, board));

  GDK_THREADS_ENTER();
}


void
BoardUI::notify_board_moved(BulletinBoard *board)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  BoardMovedDialog::show_dialog(controller, board);
}


}

