/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: command_pipe.cc,v 1.19 2009/01/01 07:47:06 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"
#include "libochusha/repository.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "command_pipe.h"

#include "controller.h"

#include <glib.h>
#include <gtk/gtk.h>

#include <cassert>
#include <cerrno>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <climits>

#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif


#ifndef PIPE_BUF
#define PIPE_BUF 512	// POSIXǤκǾ׷餷
#endif

#define OCHUSHA_COMMAND_PIPE_FILE	"command-pipe"


namespace Ochusha
{

CommandPipe::CommandPipe(Repository &repository)
  : repository(repository),
    controller(NULL),
    pipe_filename(NULL),
    pipe_fd(-1),
    poll_id(0),
    buffer(NULL),
    pipe_input(NULL)
{
  assert(repository.get_home_path() != NULL);

  char pipe_file[PATH_MAX];
  if (snprintf(pipe_file, PATH_MAX, "%s" OCHUSHA_COMMAND_PIPE_FILE,
	       repository.get_home_path()) >= PATH_MAX)
    {
      fprintf(stderr, _("Cannot create named pipe \"%s%s\""),
	      repository.get_home_path(), OCHUSHA_COMMAND_PIPE_FILE);
      return;
    }

  ErrorMessage error_message;

 retry_mkfifo:
  int mkfifo_result = mkfifo(pipe_file, O_RDWR | S_IRUSR | S_IWUSR);
  if (mkfifo_result < 0)
    {
      if (errno != EEXIST)
	{
	  error_message.set_errno(errno);
	  fprintf(stderr, _("Cannot create named pipe \"%s\": %s\n"),
		  pipe_file, error_message.get_message());
	  return;
	}

      int fd = open(pipe_file, O_WRONLY | O_NONBLOCK);
      if (fd >= 0)
	{
	  /* ˼¹Τץ¸ߤ */
	  pipe_input = fdopen(fd, "a");
	  if (pipe_input == NULL)
	    {
	      error_message.set_errno(errno);
	      fprintf(stderr, _("Cannot fdopen() for remote control: %s\n"),
		      error_message.get_message());
	      return;
	    }

	  return;
	}
      else if (errno == ENXIO)
	{
	  /*
	   * 餫ͳư㤬ѥץե
	   * ä»ʤä
	   */
	  if (unlink(pipe_file) < 0)
	    {
	      error_message.set_errno(errno);
	      fprintf(stderr, _("Cannot delete file \"%s\": %s\n"),
		      pipe_file, error_message.get_message());
	      return;
	    }
	  goto retry_mkfifo;
	}
      else
	{
	  error_message.set_errno(errno);
	  fprintf(stderr, _("Cannot start ochusha: %s\n"),
		  error_message.get_message());
	  return;
	}
    }
  else
    {
      /* ư줿㤬ͣGUIץȤʤ */
      int fd = open(pipe_file, O_RDWR | O_NONBLOCK);
      if (fd < 0)
	{
	  error_message.set_errno(errno);
	  fprintf(stderr, _("Cannot open named pipe \"%s\": %s\n"),
		  pipe_file, error_message.get_message());
	  if (unlink(pipe_file) < 0)
	    {
	      error_message.set_errno(errno);
	      fprintf(stderr, _("Cannot delete file \"%s\": %s\n"),
		      pipe_file, error_message.get_message());
	    }
	  return;
	}

      pipe_fd = fd;
      pipe_filename = strdup(pipe_file);
      buffer = g_string_new(NULL);
    }
}


CommandPipe::~CommandPipe()
{
  if (poll_id != 0)
    g_source_remove(poll_id);

  if (pipe_fd >= 0)
    close(pipe_fd);

  if (pipe_filename != NULL)
    {
      unlink(pipe_filename);
      free(pipe_filename);
    }

  if (pipe_input != NULL)
    fclose(pipe_input);

  if (buffer != NULL)
    g_string_free(buffer, TRUE);
}


void
CommandPipe::initialize_command_handler(Controller *controller)
{
  assert(pipe_fd >= 0);
  assert(poll_id == 0);

  this->controller = controller;
  GIOChannel *channel = g_io_channel_unix_new(pipe_fd);
  // priorityˤĴ;ϤΤg_io_add_watch_full()
  // G_PRIORITY_DEFAULTΤޤޤɤʤg_io_add_watch()ǽʬ
  poll_id = g_io_add_watch_full(channel, G_PRIORITY_DEFAULT,
				static_cast<GIOCondition>(G_IO_IN | G_IO_PRI),
				reinterpret_cast<GIOFunc>(poll_cb), this, NULL);
  g_io_channel_unref(channel);
}


void
CommandPipe::finalize_command_handler()
{
  if (poll_id == 0)
    return;

  g_source_remove(poll_id);
  poll_id = 0;
}


gboolean
CommandPipe::poll_cb(GIOChannel *channel, GIOCondition condition,
		     CommandPipe *pipe)
{
#if 0
  fprintf(stderr, "CommandPipe::poll_cb: condition=%x\n", condition);
#endif
  if (condition & (G_IO_IN | G_IO_PRI))
    {
      GError *error = NULL;
      if (g_io_channel_read_line_string(channel, pipe->buffer,
					NULL, &error) == G_IO_STATUS_NORMAL)
	{
	  // g_io_addΥХåGLibɳʤΤGDK_THREADS_ENTER()Ǥʤ
	  GDK_THREADS_ENTER();

	  pipe->process(pipe->buffer->str);

	  GDK_THREADS_LEAVE();
	}

      if (error != NULL)
	{
	  fprintf(stderr,
		  _("g_io_channel_read_line_string(...) error: %s (%d)\n"),
		  error->message, error->code);
	  g_error_free(error);
	}
    }

  return TRUE;	/* ʤ*/
}


bool
CommandPipe::is_master()
{
  return pipe_filename != NULL;
}


/*
 * TODO: Ūˤϰ̤Schemeդͽꡣ
 *
 * ߼դ륳ޥ
 *   (open-url "<URL>")
 *   (open-url-in-tab "<URL>")
 *
 *   (login-2ch-be)
 *   (login-2ch-viewer)
 *   (refresh-boardlist)
 *   (restore-opened-tabs)
 *   (raise)
 *   (convert-virtual-boards)
 *   (read-all-threadlists)
 *
 * ѡ񤯤ΤݤʤΤʸˡϤʤ긷(;ʬʶȤʤ)
 * Scheme󥿥ץ꥿ˤǤͽʤΤǡλ̤
 * Schemeʸˡǽ񤱤褦ˤʤ뤬Ǥ1Ըͽꡣ
 */
void
CommandPipe::process(const char *command)
{
  g_return_if_fail(command != NULL);

  if (strcmp(command, "(login-2ch-be)\n") == 0)
    {
      controller->bbs2ch.login_2ch_be();
      return;
    }

  if (strcmp(command, "(login-2ch-viewer)\n") == 0)
    {
      controller->bbs2ch.login_2ch_viewer();
      return;
    }

  if (strcmp(command, "(refresh-boardlist)\n") == 0)
    {
      controller->boardlist_ui.refresh_boardlist();
      return;
    }

  if (strcmp(command, "(raise)\n") == 0)
    {
      gtk_window_present(controller->get_toplevel());
      return;
    }

  if (strcmp(command, "(convert-virtual-boards)\n") == 0)
    {
      controller->convert_virtual_boards();
      return;
    }

  if (strcmp(command, "(read-all-threadlists)\n") == 0)
    {
      controller->read_all_threadlists();
      return;
    }

  if (strcmp(command, "(restore-opened-tabs)\n") == 0)
    {
      controller->main_window.restore_tab_related_session();
      return;
    }

  gboolean in_tab;
  size_t len;
  const char *tmp_pos;
  char url[PATH_MAX];

  if (strncmp(command, "(open-url \"", 11) == 0)
    {
      in_tab = FALSE;
      tmp_pos = command + 11;
    }
  else if (strncmp(command, "(open-url-in-tab \"", 18) == 0)
    {
      in_tab = TRUE;
      tmp_pos = command + 18;
    }
  else
    {
      fprintf(stderr, _("Unknown remote command: %s\n"), command);
      return;
    }

  len = strlen(tmp_pos);
  if (len < 3 || strcmp(tmp_pos + len - 3, "\")\n") != 0)
    {
      fprintf(stderr, _("Unknown remote command: %s\n"), command);
      return;
    }

  if (len - 4 > PATH_MAX)
    {
      fprintf(stderr, _("Command too long: %s\n"), command);
      return;
    }

  memcpy(url, tmp_pos, len - 3);
  url[len - 3] = '\0';
  controller->open(url, true, in_tab, false);
}


bool
CommandPipe::send(const char *command)
{
  int len = strlen(command);
  if (len < 0 || len > PIPE_BUF)
    {
      fprintf(stderr, _("Command too long: %s\n"), command);

      return false;
    }

  if (pipe_filename == NULL)
    {
      // ⡼Ȥ˥ޥɤ
      assert(pipe_input != NULL);

      return fputs(command, pipe_input) == len;
    }

  // ʬȤ˥ޥɤ
  assert(pipe_fd >= 0);

  return write(pipe_fd, command, len) == len;
}

}
