/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: console.cc,v 1.24 2010/08/15 15:42:20 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "console.h"

#include "libochusha/misc_utils.h"

#include "gtk_utils.h"

#include <glib.h>
#include <gtk/gtk.h>

#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <cstring>


namespace Ochusha
{

Console::Console(Controller &controller,
		 const char *title, int fd, GtkToggleAction *action)
  : controller(controller),
    title(title != NULL ? strdup(title) : strdup(_("Anonymous"))),
    fd(fd),
    poll_id(0),
    string_buffer(g_string_new(NULL)),
    dialog(NULL),
    buffer(gtk_text_buffer_new(NULL)),
    view(NULL),
    tail_mark(NULL),
    action(action)
{
  gtk_text_buffer_get_start_iter(buffer, &tail_iter);
  tail_mark = gtk_text_buffer_create_mark(buffer, "tail-mark",
					  &tail_iter, FALSE);

  if (fd >= 0)
    {
      GIOChannel *channel = g_io_channel_unix_new(fd);
      GError *error = NULL;
      GIOStatus status = g_io_channel_set_flags(channel,
						G_IO_FLAG_NONBLOCK,
						&error);
      if (status != G_IO_STATUS_NORMAL)
	{
	  fprintf(stderr, "Console::Console() g_io_channel_set_flags failed\n");
	  if (error != NULL)
	    {
	      fprintf(stderr, "=> %s (%d)\n", error->message, error->code);
	      g_error_free(error);
	    }
	}
      poll_id = g_io_add_watch(channel,
			       static_cast<GIOCondition>(G_IO_IN | G_IO_PRI),
			       reinterpret_cast<GIOFunc>(poll_cb), this);
      g_io_channel_unref(channel);
    }
}


Console::~Console()
{
  if (title != NULL)
    free(title);

  if (poll_id != 0)
    g_source_remove(poll_id);

  g_string_free(string_buffer, TRUE);
  g_object_unref(buffer);
}


void
Console::prepare_shutdown()
{
  if (title != NULL)
    {
      free(title);
      title = NULL;
    }

  if (poll_id != 0)
    {
      g_source_remove(poll_id);
      poll_id = 0;
    }

  if (dialog != NULL)
    gtk_widget_destroy(dialog);

  g_return_if_fail(dialog == NULL);
}


gboolean
Console::poll_cb(GIOChannel *channel, GIOCondition condition,
		 Console *console)
{
  GIOStatus status = G_IO_STATUS_NORMAL;
  if (condition & (G_IO_IN | G_IO_PRI))
    {
      GError *error = NULL;
      bool read_done = false;
      if ((status = g_io_channel_read_line_string(channel,
						  console->string_buffer,
						  NULL, &error))
	  == G_IO_STATUS_NORMAL)
	{
	  // g_io_addΥХåGLibɳʤΤGDK_THREADS_ENTER()Ǥʤ
	  GDK_THREADS_ENTER();

	  console->output(console->string_buffer->str);

	  GDK_THREADS_LEAVE();

#if 0
	  fprintf(stderr, "Console::poll_cb(): %s\n",
		  console->string_buffer->str);
#endif

	  read_done = true;
	}

      if (error != NULL)
	{
	  gchar *message
	    = g_strdup_printf("g_io_channel_read_line_string error: %s(%d)\n",
			      error->message, error->code);


	  GDK_THREADS_ENTER();

	  console->output(message);

	  GDK_THREADS_LEAVE();

#if 0
	  fprintf(stderr, "Console::poll_cb(): *Error* %s\n", message);
#endif
	  g_free(message);
	  g_error_free(error);
	}
#if 0
      else
	{
	  if (status != G_IO_STATUS_NORMAL)
	    {
	      fprintf(stderr, "poll_cb: condition=%d\n", condition);
	      fprintf(stderr, "poll_cb(): status=%d\n", status);
	    }	      
	  assert(read_done);
	}
#endif
    }


#if 1
  if (status == G_IO_STATUS_EOF)
    {
      fprintf(stderr, "poll_cb(): EOF\n");
      return FALSE;
    }
  if (status == G_IO_STATUS_AGAIN)
    {
      // ʤȤʡġġ
      fprintf(stderr, "poll_cb(): AGAIN?\n");
      return FALSE;
    }
#endif

  return TRUE;	// ޤޤ衣
}


void
Console::output(const char *message)
{
  if (action != NULL && !gtk_toggle_action_get_active(action))
    return;

  g_return_if_fail(GTK_IS_TEXT_BUFFER(buffer));

  if (view == NULL)
    {
      gtk_text_buffer_insert(buffer, &tail_iter, message, -1);
      return;
    }

  GdkRectangle rect;
  gtk_text_view_get_visible_rect(view, &rect);

  gint y;
  gint height;
  gtk_text_view_get_line_yrange(view, &tail_iter, &y, &height);

  gtk_text_buffer_insert(buffer, &tail_iter, message, -1);

  if (rect.y + rect.height >= y)
    gtk_text_view_scroll_to_mark(view, tail_mark, 0.3, FALSE, 0.0, 0.0);
}


void
Console::clear()
{
  g_return_if_fail(GTK_IS_TEXT_BUFFER(buffer));
  GtkTextIter start_iter;
  GtkTextIter end_iter;

  gtk_text_buffer_get_start_iter(buffer, &start_iter);
  gtk_text_buffer_get_end_iter(buffer, &end_iter);
  gtk_text_buffer_delete(buffer, &start_iter, &end_iter);
  gtk_text_buffer_get_start_iter(buffer, &tail_iter);
}


void
Console::response_cb(GtkWidget *widget, int id, Console *console)
{
  switch (id)
    {
    case GTK_RESPONSE_REJECT:
      console->clear();
      break;

    case GTK_RESPONSE_OK:
    default:
      gtk_widget_hide(widget);
    }
}


void
Console::prepare_dialog()
{
  if (dialog != NULL)
    return;

  dialog = gtk_dialog_new_with_buttons(title, NULL,
				       GTK_DIALOG_DESTROY_WITH_PARENT,
				       GTK_STOCK_CLEAR, GTK_RESPONSE_REJECT,
				       GTK_STOCK_CLOSE, GTK_RESPONSE_OK,
				       NULL);
  g_signal_connect(dialog, "response",
		   reinterpret_cast<GCallback>(response_cb), this);
  g_signal_connect(dialog, "destroy",
		   reinterpret_cast<GCallback>(gtk_widget_destroyed), &dialog);
  gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_OK);

  GtkWidget *scroll = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll),
				 GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  GtkWidget *widget = gtk_text_view_new_with_buffer(buffer);
  new TextViewKeyPressHandler(GTK_TEXT_VIEW(widget), controller);
  gtk_widget_show(widget);
  gtk_container_add(GTK_CONTAINER(scroll), widget);

  view = GTK_TEXT_VIEW(widget);
  gtk_text_view_set_editable(view, FALSE);

  gtk_widget_show(scroll);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), scroll, TRUE, TRUE, 0);

  if (action != NULL)
    {
      GtkWidget *button = gtk_check_button_new();
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button),
				   gtk_toggle_action_get_active(action));
#if GTK_CHECK_VERSION(2,16,0)
      gtk_activatable_set_related_action(GTK_ACTIVATABLE(button),
					 GTK_ACTION(action));
#else
      gtk_action_connect_proxy(GTK_ACTION(action), button);
#endif
      gtk_widget_show(button);
      gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), button,
			 FALSE, FALSE, 0);
    }
}


void
Console::show()
{
  if (dialog == NULL)
    prepare_dialog();

  g_return_if_fail(dialog != NULL);
  gtk_widget_show_all(dialog);
  gtk_window_present(GTK_WINDOW(dialog));
}

}
