/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: controller.cc,v 1.101 2009/01/01 11:02:28 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "controller.h"

#include "ochusha_accel_map.h"

#include "libochusha/misc_utils.h"
#include "libochusha/ochusha_procedure.h"

#include "bbs_thread_view.h"
#include "command_pipe.h"
#include "focus_utils.h"
#include "image_manager.h"
#include "selection_utils.h"
#include "session_client.h"
#include "threadlist_view.h"
#include "ui_constants.h"

#include "libochushagtk/entry_action.h"
#include "libochushagtk/searchbar.h"
#include "libochushagtk/stock.h"


#include <iostream>

#include <cerrno>
#include <cstddef>
#include <cstdlib>
#include <cstdio>
#include <cstring>

#include <vector>

#include <gtk/gtk.h>

#include <fcntl.h>
#include <unistd.h>


#undef SHOW_STDERR_IN_DEBUG_CONSOLE


namespace Ochusha
{


static Lock pending_logs_lock("Controller::output_log");


#ifdef SHOW_STDERR_IN_DEBUG_CONSOLE
static int
piped_err()
{
  static int fds[] = { -1, -1 };
  static int fd = -1;
  ErrorMessage error_message;

  if (fds[0] == -1)
    {
      if (pipe(fds) == -1)
	{
	  error_message.set_errno(errno);
	  fprintf(stderr, "pipe(2) failed: %s\n", error_message.get_message());
	  return -1;
	}

      fd = dup2(fds[1], fileno(stderr));
      if (fd == -1)
	{
	  error_message.set_errno(errno);
	  fprintf(stderr, "dup2(2) failed: %s\n", error_message.get_message());
	  return -1;
	}
      fcntl(fd, F_SETFL, fcntl(fd, F_GETFL) | O_NONBLOCK);
    }

  return fds[0];
}
#endif


class BoardMovementWatcher : public has_slots
{
private:
  Controller &controller;

  // BBSTablesigslotʥϥɥ
  void board_moved_cb(BBSTable *table, BulletinBoard *board,
		      const char *new_url);


public:
  BoardMovementWatcher(Controller &controller, BBSTable &bbs_table);
};


BoardMovementWatcher::BoardMovementWatcher(Controller &controller,
					   BBSTable &bbs_table)
  : controller(controller)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  GDK_THREADS_LEAVE();

  bbs_table.board_moved.connect(this, &BoardMovementWatcher::board_moved_cb);

  GDK_THREADS_ENTER();
}


void
BoardMovementWatcher::board_moved_cb(BBSTable *table, BulletinBoard *board,
				     const char *new_url)
{
  controller.notify_board_moved(table, board, new_url);
}


class ImageManagerWatcher : public has_slots
{
private:
  Controller &controller;

  // ImageManagersigslotʥϥɥ
  void output_log_cb(LogCategory category, const char *message);


public:
  ImageManagerWatcher(Controller &controller, ImageManager &image_manager);
};


ImageManagerWatcher::ImageManagerWatcher(Controller &controller,
					 ImageManager &image_manager)
  : controller(controller)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  GDK_THREADS_LEAVE();

  image_manager.output_log.connect(this, &ImageManagerWatcher::output_log_cb);

  GDK_THREADS_ENTER();
}


void
ImageManagerWatcher::output_log_cb(LogCategory category, const char *message)
{
  controller.output_log(category, message);
}


class NetworkStatusWatcher : public has_slots
{
private:
  Controller &controller;

  // NetworkAgentBrokersigslotʥϥɥ
  void output_log_cb(LogCategory category, const char *message);
  void pre_prepare_request_cb(NetworkAgent *agent);
  void post_prepare_request_cb(NetworkAgent *agent);


public:
  NetworkStatusWatcher(Controller &controller);
};


NetworkStatusWatcher::NetworkStatusWatcher(Controller &controller)
  : controller(controller)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  GDK_THREADS_LEAVE();

  controller.broker->output_log
    .connect(this, &NetworkStatusWatcher::output_log_cb);
  controller.broker->pre_prepare_request
    .connect(this, &NetworkStatusWatcher::pre_prepare_request_cb);
  controller.broker->post_prepare_request
    .connect(this, &NetworkStatusWatcher::post_prepare_request_cb);

  GDK_THREADS_ENTER();
}


void
NetworkStatusWatcher::output_log_cb(LogCategory category, const char *message)
{
  controller.output_log(category, message);
}


void
NetworkStatusWatcher::pre_prepare_request_cb(NetworkAgent *agent)
{
  controller.pre_prepare_request(agent);
}


void
NetworkStatusWatcher::post_prepare_request_cb(NetworkAgent *agent)
{
  controller.post_prepare_request(agent);
}


#if 0	// XXX: ¸
static void
check_accel_path(const gchar *accel_path)
{
  GtkAccelKey accel_key = { 0, };
  if (gtk_accel_map_lookup_entry(accel_path, &accel_key))
    fprintf(stderr, "check_accel_path(\"%s\"): key=%u, mods=0x%x, flags=0x%x\n",
	    accel_path, accel_key.accel_key,
	    static_cast<unsigned int>(accel_key.accel_mods),
	    accel_key.accel_flags);
  else
    fprintf(stderr, "Cannot find accelerator entry for: \"%s\"\n", accel_path);
}


static void
print_accel_entry(gpointer unused, const gchar *accel_path, guint accel_key,
		  GdkModifierType accel_mods, gboolean changed)
{
  fprintf(stderr,
	  "accel-entry \"%s\" { accel_key=%u, accel_mods=0x%x, changed=%d }\n",
	  accel_path, accel_key, static_cast<unsigned int>(accel_mods),
	  changed);
}


static void
dump_accel_paths(const char *title)
{
  fprintf(stderr, "=== BEGIN %s ===\n", title);
  gtk_accel_map_foreach_unfiltered(NULL, print_accel_entry);
  fprintf(stderr, "=== END %s ===\n", title);
}
#endif


Controller::Controller(Repository &repository, CommandPipe &command_pipe,
		       SessionClient &session, CookieManager &cookie_manager,
		       HistoryManager &history_manager,
		       HistoryManager &post_history_manager)
  : repository(repository),
    command_pipe(command_pipe),
    session(session),
    bbs_table(new BBSTable()),
    cookie_manager(cookie_manager),
    thread_pool(new ThreadPool(INITIAL_NUM_WORKER_THREADS,
			       MAXIMUM_NUM_WORKER_THREADS)),
    broker(new NetworkAgentBroker(cookie_manager, poll_manager, thread_pool)),
    history_manager(history_manager),
    post_history_manager(post_history_manager),
    clipboard(gtk_clipboard_get(GDK_SELECTION_CLIPBOARD)),
    clipboard_primary(gtk_clipboard_get(GDK_SELECTION_PRIMARY)),
    clipboard_internal(gtk_clipboard_get(gdk_atom_intern("_OCHUSHA_INTERNAL_CLIPBOARD", FALSE))),
    ui_manager(gtk_ui_manager_new()),
    action_group(create_action_group()),
    prefs(*this),
    scheme_manager(*this),
    tag_manager(*this),
    bbs2ch(*this),
    bookmarks_ui(*this),
    thread_ui(*this),
    board_properties_ui(*this, prefs),
    board_ui(*this),
    boardlist_ui(*this),
    link_ui(*this),
    threadlist_ui(*this),
    tree_ui(*this),
    main_window(*this),
    view_manager(main_window.view_manager),
    prefs_ui(*this),
    fonts_ui(*this),
    about(*this),
    notification(*this, _("Notification"), -1, NULL),
#ifdef SHOW_STDERR_IN_DEBUG_CONSOLE
    debug_console(*this, _("Debug Console"), piped_err(), prefs.debug_console),
#else
    debug_console(*this, _("Debug Console"), -1, prefs.debug_console),
#endif
    image_manager(NULL),
    board_movement_watcher(new BoardMovementWatcher(*this, *bbs_table)),
    image_manager_watcher(NULL),
    network_status_watcher(new NetworkStatusWatcher(*this)),
    open_url_dialog(NULL),
    process_pending_logs_on_idle_id(0)

{
  // 󥹥ȥ饯ƤФλsession.xmlƤsession¸
  config.read_config_xml(repository);
  gtk_toggle_action_set_active(work_offline_action, config.get_offline());

  // config.xmlߴΤɤ߼äȿǤ롣
  restore_configuration();
  // ޤǤǡեͳξȿǺ

  // schemeĶѡschemeĶνˤꡢġ
  // ɽ§ϿƤΤǡȿǤ롣

  board_ui.update_threadlist_rules();


  // ̯ġġ
  // ImageManagerΥ󥹥ȥ饯prefsɤΤǡ
  // config.xmlͳξȿǤΤԤġ
  // Ȥʤȡդޤ٤餻뤳Ȥˤʤ롣
  image_manager = new ImageManager(*this);
  image_manager_watcher = new ImageManagerWatcher(*this, *image_manager);


  if (bbs_table->all_boards_begin() != bbs_table->all_boards_end())
    command_pipe.send("(convert-virtual-boards)\n");

  if (prefs.get_login_2ch_viewer_at_startup())
    command_pipe.send("(login-2ch-viewer)\n");
  if (prefs.get_login_2ch_be_at_startup())
    command_pipe.send("(login-2ch-be)\n");

  g_signal_connect(clipboard, "owner-change",
		   reinterpret_cast<GCallback>(owner_change_cb), this);
  g_signal_connect(clipboard_primary, "owner-change",
		   reinterpret_cast<GCallback>(owner_change_cb), this);
  g_signal_connect(clipboard_internal, "owner-change",
		   reinterpret_cast<GCallback>(owner_change_cb), this);

  main_window.show();

  boardlist_ui.restore_session();
  bookmarks_ui.restore_session();

  main_window.restore_session();

  command_pipe.send("(restore-opened-tabs)\n");

#if 0	// XXX: ¸
  check_accel_path("<Actions>/OchushaActions/OpenURL");

  dump_accel_paths("PRE LOADING OF ACCEL-MAP");
#endif

  char filename[PATH_MAX];
  if (repository.find_file(OCHUSHA_ACCEL_MAP_RC, filename, PATH_MAX) != NULL)
    {
      gtk_accel_map_load(filename);
    }
  else
    {
      GScanner *scanner = g_scanner_new(NULL);
      if (scanner == NULL)
	{
	  g_message(_("*FATAL* Cannot prepare accel-map: Out of Memory?\n"));
	  exit(EXIT_FAILURE);
	}

      g_scanner_input_text(scanner,
			   default_accel_map, sizeof(default_accel_map));
      gtk_accel_map_load_scanner(scanner);

      g_scanner_destroy(scanner);
    }

#if 0	// XXX: ¸
  check_accel_path("<Actions>/OchushaActions/OpenURL");

  dump_accel_paths("POST LOADING OF ACCEL-MAP");
#endif
}


Controller::~Controller()
{
  assert(OCHUSHA_IN_GDK_THREADS());

  if (open_url_dialog != NULL)
    gtk_widget_destroy(open_url_dialog);

  if (action_group != NULL)
    g_object_unref(action_group);

  if (ui_manager != NULL)
    g_object_unref(ui_manager);


  GDK_THREADS_LEAVE();

  delete network_status_watcher;
  delete image_manager_watcher;
  delete board_movement_watcher;
  delete image_manager;

  delete broker;

  bbs_table = NULL;

  GDK_THREADS_ENTER();
}


void
Controller::prepare_shutdown()
{
  assert(OCHUSHA_IN_GDK_THREADS());

  image_manager->prepare_shutdown();

  // ǰΤġġ
  GDK_THREADS_LEAVE();

  thread_pool->terminate_workers();
  poll_manager.terminate_all();

  GDK_THREADS_ENTER();


  if (open_url_dialog != NULL)
    gtk_widget_destroy(open_url_dialog);

  prefs_ui.prepare_shutdown();
  notification.prepare_shutdown();
  debug_console.prepare_shutdown();
  fonts_ui.prepare_shutdown();
  main_window.prepare_shutdown();

  if (action_group != NULL)
    {
      g_object_unref(action_group);
      action_group = NULL;
    }

  if (ui_manager != NULL)
    {
      g_object_unref(ui_manager);
      ui_manager = NULL;
    }

  g_return_if_fail(open_url_dialog == NULL);
}


void
Controller::restore_configuration()
{
  if (config.get_config_version() < CONFIG_VERSION_OCHUSHA_0_6)
    {
      // ʲ¸꤬session.xmlˤʤä
      bbs2ch.set_be_id(config.get_config_string("be_id"));
      bbs2ch.set_be_passwd(config.get_config_string("be_passwd"));
      bbs2ch.set_viewer_id(config.get_config_string("dolib_id"));
      bbs2ch.set_viewer_passwd(config.get_config_string("dolib_passwd"));

      const char *str = config.get_config_string("last_name");
      if (str != NULL)
	session.set_state("PostUI::name", str);

      str = config.get_config_string("last_mail");
      if (str != NULL)
	session.set_state("PostUI::mail", str);

      str = config.get_config_string("last_download_directory");
      if (str != NULL)
	session.set_state("DownloadUI::directory", str);


      // ʲ¸꤬prefs.xmlˤʤä
      prefs.set_save_2ch_be_passwd(config.get_config_bool("be_save_passwd"));
      prefs.set_login_2ch_be_at_startup(config.get_config_bool("be_login_at_startup"));
      prefs.set_save_2ch_viewer_passwd(config.get_config_bool("dolib_save_passwd"));
      prefs.set_login_2ch_viewer_at_startup(config.get_config_bool("dolib_login_at_startup"));

      prefs.set_board_tab_shrinkable(config.get_config_bool("board_tab_shrinkable"));
      prefs.set_board_tab_minimum_size(config.get_config_int("board_tab_minimum_size"));
      prefs.set_board_tab_enable_tooltips(config.get_config_bool("board_tab_enable_tooltips"));
      prefs.set_board_tab_always_show(config.get_config_bool("board_tab_always_show"));
      prefs.set_vertical_placement(config.get_config_int("threadlist_pane_style") != 0);
      prefs.set_thread_tab_shrinkable(config.get_config_bool("thread_tab_shrinkable"));
      prefs.set_thread_tab_minimum_size(config.get_config_int("thread_tab_minimum_size"));
      prefs.set_thread_tab_enable_tooltips(config.get_config_bool("thread_tab_enable_tooltips"));
      prefs.set_thread_tab_always_show(config.get_config_bool("thread_tab_always_show"));
      prefs.set_threadlist_view_contents(config.get_config_string("threadlist_view_contents"));
      prefs.set_thread_view_font_name(config.get_config_string("thread_view_font_name"));
      prefs.set_enable_transparent_a_bone(config.get_config_bool("enable_transparent_a_bone"));
      prefs.set_disable_popup_a_bone(config.get_config_bool("disable_popup_a_bone"));
      prefs.set_hide_bookmark(config.get_config_bool("hide_bookmark"));

      const char *bt = config.get_config_string("bookmark_text");
      if (bt != NULL)
	{
	  char *text = URI::decode_string(bt);
	  prefs.set_bookmark_text(text);
	  free(text);
	}

      prefs.set_default_open_in_tab(config.get_config_bool("default_open_in_tab"));
      prefs.set_select_tab_when_opened(config.get_config_bool("select_tab_when_opened"));
      prefs.set_show_mailto_mode(static_cast<MailtoMode>(config.get_config_int("show_mailto_mode")));
      prefs.set_enable_inline_image(config.get_config_bool("enable_inline_image"));
      prefs.set_inline_image_width(config.get_config_int("inline_image_width"));
      prefs.set_inline_image_height(config.get_config_int("inline_image_height"));
      prefs.set_enable_popup_image(config.get_config_bool("enable_popup_image"));
      prefs.set_popup_image_width(config.get_config_int("popup_image_width"));
      prefs.set_popup_image_height(config.get_config_int("popup_image_height"));
      prefs.set_capacity_of_image_cache(config.get_config_int("capacity_of_image_cache"));
      prefs.set_maximum_number_of_image_cache_files(config.get_config_int("maximum_number_of_cache_files"));
      prefs.set_enable_popup_title(config.get_config_bool("enable_popup_title"));
      prefs.set_enable_popup_response(config.get_config_bool("enable_popup_response"));
      prefs.set_popup_delay(config.get_config_int("popup_response_delay"));
      prefs.set_popdown_delay(config.get_config_int("popup_close_delay"));
      prefs.set_maximum_number_of_popup_responses(config.get_config_int("maximum_number_of_popup_responses"));
      prefs.set_enable_popup_id(config.get_config_bool("enable_popup_id"));
      prefs.set_web_browser_template(config.get_config_string("web_browser_template"));
      prefs.set_show_hidden_boardlist_items(!config.get_config_bool("hide_hidden_boardlist_items"));
      prefs.set_show_hidden_threads(!config.get_config_bool("hide_hidden_threads"));
#if 0
      prefs.set_aalist_filename(config.get_config_string("aalist_filename"));
      prefs.set_aalist_char_encoding(config.get_config_string("aalist_char_encoding"));
#endif
      prefs.set_focus_follow_mouse(config.get_config_bool("focus_follow_mouse"));
    }
}


void
Controller::convert_virtual_board(BulletinBoard *board)
{
  FolderHandle folder = bookmarks_ui.folder_new(board->get_name());


  GDK_THREADS_LEAVE();

  board->read_threadlist_xml(repository);

  ThreadlistIteratorHandle threadlist
    = board->get_threadlist(config, repository, *thread_pool, *broker, false);

  GDK_THREADS_ENTER();


  int lock_id = threadlist->lock();
  assert(threadlist->finished(lock_id));

  while (threadlist->has_more_threads(lock_id))
    {
      BBSThread *thread = threadlist->next_thread(lock_id);
      assert(thread != NULL);

      bookmarks_ui.bookmark(folder, thread->get_id(), thread->get_title());
    }
  threadlist->unlock(lock_id);


  GDK_THREADS_LEAVE();

  // threadlistdestructorƤФglobal lockݤ뤿ᡢ
  // GDKΥåƤɬפ롣
  threadlist = NULL;

  GDK_THREADS_ENTER();


  board->set_bbs_type(OCHUSHA_BBS_TYPE_ACTIVE_FOLDER);
  char *url = BulletinBoard::synthesize_active_folder_url(folder->get_serial_number());
  board->set_base_url(url);
  free(url);
  boardlist_ui.notify_element_updated(board);
}


void
Controller::convert_virtual_boards()
{
  TreeElementVector::iterator iter = bbs_table->all_boards_begin();
  TreeElementVector::iterator end_iter = bbs_table->all_boards_end();

  while (iter != end_iter)
    {
      TreeElement *element = *iter++;
      if (!element->is_folder())
	{
	  BulletinBoard *board = dynamic_cast<BulletinBoard *>(element);
	  if (board != NULL)
	    {
	      if (board->get_bbs_type() == OCHUSHA_BBS_TYPE_VIRTUAL_BOARD)
		{
		  // urlhttp:///all-threads/ǤʤСѴ롣
		  const char *base_url = board->get_base_url();
		  if (base_url != NULL)
		    {
		      if (strcmp(base_url, "http:///all-threads/") != 0)
			convert_virtual_board(board);
		      else
			{
			  board->set_bbs_type(OCHUSHA_BBS_TYPE_PSEUDO_FOLDER);
			  char *new_url = BulletinBoard::synthesize_pseudo_folder_url("all-threads");
			  if (new_url != NULL)
			    {
			      board->set_base_url(new_url);
			      free(new_url);
			    }
			}
		    }
		}
	    }
	}
    }
}


class ReadThreadlistJob : public WorkerJob
{
private:
  Controller &controller;


public:
  ReadThreadlistJob(Controller &controller) : controller(controller) {}
  virtual void work(const WorkerThread *employee);
  virtual const char *description() const
  { return "reading threadlist.xml for all boards"; }
};


void
ReadThreadlistJob::work(const WorkerThread *employee)
{
  BBSTable &table = controller.get_bbs_table();
  TreeElementVector::iterator iter = table.all_boards_begin();
  TreeElementVector::iterator end_iter = table.all_boards_end();
  while (employee->get_state() != WorkerThread::DIE_NOW
	 && iter != end_iter)
    {
      TreeElement *element = *iter++;
      BulletinBoard *board = dynamic_cast<BulletinBoard *>(element);
      assert(board != NULL);

      if (board->get_bbs_type() >= OCHUSHA_BBS_TYPE_2CH)
	{
	  board->read_threadlist_xml(controller.get_repository());

	  char message[512];
	  snprintf(message, 512,
		   _("Reading of threadlist.xml for %s done.\n"),
		   board->get_name());


	  controller.output_log(DEBUG_LOG, message);
	}
      else if (board->get_bbs_type() == OCHUSHA_BBS_TYPE_PSEUDO_FOLDER)
	{
	  // 0.6ϳȯpseudo folderλͷΤΤĴ
	  const char *url = board->get_base_url();
	  if (strcmp(url, "http:///all-threads/") == 0)
	    {
	      char *new_url
		= BulletinBoard::synthesize_pseudo_folder_url("all-threads");
	      if (new_url != NULL)
		board->set_base_url(new_url);
	      free(new_url);
	    }
	}
    }
}


void
Controller::read_all_threadlists()
{
  thread_pool->commit(new ReadThreadlistJob(*this));
}


void
Controller::spawn_command(const char *command_line)
{
  char *escaped_command_line = shell_escape_text(command_line);
  g_return_if_fail(escaped_command_line != NULL);
  g_return_if_fail(*escaped_command_line != '\0');

  GError *error = NULL;
  if (!g_spawn_command_line_async(escaped_command_line, &error))
    {
      // 顼
      GtkWidget *dialog
	= gtk_message_dialog_new(get_toplevel(),
				 GTK_DIALOG_DESTROY_WITH_PARENT,
				 GTK_MESSAGE_ERROR,
				 GTK_BUTTONS_CLOSE,
				 _("Spawn Error"));
      gtk_message_dialog_format_secondary_markup(
			GTK_MESSAGE_DIALOG(dialog),
			"<span foreground=\"red\">%s</span>\n%s (%d)",
			escaped_command_line, error->message, error->code);
      g_error_free(error);
      g_signal_connect_swapped(dialog, "response",
			       reinterpret_cast<GCallback>(gtk_widget_destroy),
			       dialog);
      gtk_widget_show_all(dialog);
    }

  free(escaped_command_line);
}


struct RefreshMatchedBoardArg
{
  BulletinBoard *board;
  Controller &controller;
};


static void
refresh_matched_board(GtkWidget *view, BulletinBoard *board, void *data)
{
  RefreshMatchedBoardArg *arg = static_cast<RefreshMatchedBoardArg *>(data);
  if (arg->board == board)
    arg->controller.refresh(board, true);
}


void
Controller::treat_board_movement(BulletinBoard *board, const char *new_url,
				 bool show_notification)
{
  const char *name = board->get_name();
  if (name == NULL)
    name = _("Anonymous Board");

  const char *old_url = board->get_base_url();
  if (old_url == NULL)
    old_url = "unknown";

  if (strcmp(old_url, new_url) != 0)
    {
      char message[1024];
      snprintf(message, 1024, _("%s moved: \"%s\" => \"%s\"\n"),
	       name, old_url, new_url);

      notification.output(message);

      if (show_notification)
	notification.show();
    }

  GDK_THREADS_LEAVE();

  board->read_threadlist_xml(repository);

  GDK_THREADS_ENTER();


  board->set_base_url(new_url);
  board->notify_major_modification();

  RefreshMatchedBoardArg arg = { board, *this };
  get_view_manager().foreach_board(refresh_matched_board, &arg);
}


void
Controller::notify_board_moved(BBSTable *table, BulletinBoard *board,
			       const char *new_url)
{
  GDK_THREADS_ENTER();

  treat_board_movement(board, new_url, true);

  GDK_THREADS_LEAVE();
}


void
Controller::process_pending_logs()
{
  assert(pending_logs_lock.is_locked_by_self());
  assert(OCHUSHA_IN_GDK_THREADS());
  assert(process_pending_logs_on_idle_id == 0);

  while (!pending_logs.empty())
    {
      PendingLog &log = pending_logs.front();
      debug_console.output(log.message.c_str());

      pending_logs.pop_front();
    }
}


gboolean
Controller::process_pending_logs_on_idle_cb(Controller *controller)
{
  LockBlock lock_obj(pending_logs_lock);

  if (OCHUSHA_GDK_THREADS_TRYENTER())
    {
      controller->process_pending_logs_on_idle_id = 0;
      controller->process_pending_logs();

      GDK_THREADS_LEAVE();

      return FALSE;
    }

  return TRUE;
}


void
Controller::output_log(LogCategory category, const char *message)
{
  // sigslotʥϥɥ餫ƤӽФ롣
  LockBlock lock_obj(pending_logs_lock);
  pending_logs.push_back(PendingLog(category, message));

  if (OCHUSHA_GDK_THREADS_TRYENTER())
    {
      if (process_pending_logs_on_idle_id != 0)
	{
	  g_source_remove(process_pending_logs_on_idle_id);
	  process_pending_logs_on_idle_id = 0;
	}

      process_pending_logs();

      GDK_THREADS_LEAVE();
    }
  else if (process_pending_logs_on_idle_id == 0)
    process_pending_logs_on_idle_id
      = g_idle_add(reinterpret_cast<GSourceFunc>(process_pending_logs_on_idle_cb),
		   this);
}


void
Controller::pre_prepare_request(NetworkAgent *agent)
{
  // sigslotʥϥɥ餫ƤӽФ
#if 0
  GDK_THREADS_ENTER();
  GDK_THREADS_LEAVE();
#endif
}


void
Controller::post_prepare_request(NetworkAgent *agent)
{
  // sigslotʥϥɥ餫ƤӽФ롣
#if 0
  GDK_THREADS_ENTER();
  GDK_THREADS_LEAVE();
#endif
}


GtkActionGroup *
Controller::create_action_group()
{
  static GtkActionEntry entries[] = {
    {
      "ClearCookies", GTK_STOCK_DELETE,
      N_("Clear Cookies"), "",
      NULL,
      reinterpret_cast<GCallback>(clear_cookies_cb)
    },
    {
      "OpenFile", GTK_STOCK_OPEN,
      N_("Open _File"), "",
      N_("Open File"),
      reinterpret_cast<GCallback>(open_file_cb)
    },
    {
      "Quit", GTK_STOCK_QUIT,
      N_("_Quit"), "",
      N_("Quit"),
      reinterpret_cast<GCallback>(quit_cb)
    },
    {
      "CutSelection", GTK_STOCK_CUT,
      N_("Cu_t"), "",
      N_("Cut"),
      reinterpret_cast<GCallback>(cut_cb)
    },
    {
      "CopySelection", GTK_STOCK_COPY,
      N_("_Copy"), "",
      N_("Copy"),
      reinterpret_cast<GCallback>(copy_cb)
    },
    {
      "PasteSelection", GTK_STOCK_PASTE,
      N_("_Paste"), "",
      N_("Paste"),
      reinterpret_cast<GCallback>(paste_cb)
    },
    {
      "DeleteSelection", GTK_STOCK_DELETE,
      N_("_Delete"), "",
      N_("Delete"),
      reinterpret_cast<GCallback>(delete_cb)
    },
    {
      "Preferences", GTK_STOCK_PREFERENCES,
      N_("Prefere_nces"), "",
      N_("Preferences"),
      reinterpret_cast<GCallback>(preferences_cb)
    },
    {
      "SelectListViewFonts", GTK_STOCK_SELECT_FONT,
      N_("Select Board_list and Threadlist View Fonts"), "",
      N_("Select Boardlist and Threadlist View Fonts"),
      reinterpret_cast<GCallback>(select_list_fonts_cb)
    },
    {
      "SelectThreadViewFonts", GTK_STOCK_SELECT_FONT,
      N_("Select _Thread View Fonts"), "",
      N_("Select Thread View Fonts"),
      reinterpret_cast<GCallback>(select_thread_fonts_cb)
    },
    {
      "Stop", GTK_STOCK_STOP,
      N_("_Stop"), "",
      N_("Stop"),
      reinterpret_cast<GCallback>(stop_cb)
    },
    {
      "Notification", NULL,
      N_("_Notification"), NULL,
      N_("Notification"),
      reinterpret_cast<GCallback>(notification_cb)
    },
    {
      "DebugConsole", NULL,
      N_("_Debug Console"), NULL,
      N_("Debug Console"),
      reinterpret_cast<GCallback>(debug_console_cb)
    },
    {
      "SchemeConsole", NULL,
      N_("_Scheme Console"), NULL,
      N_("Scheme Console"),
      reinterpret_cast<GCallback>(scheme_console_cb)
    },
    {
      "About", GTK_STOCK_ABOUT,
      N_("_About Ochusha"), "",
      N_("About Ochusha"),
      reinterpret_cast<GCallback>(about_cb)
    },
    {
      "Help", GTK_STOCK_HELP,
      N_("_Help Contents"), "",
      N_("Help Contents"),
      reinterpret_cast<GCallback>(help_cb)
    }
  };

  static GtkToggleActionEntry toggle_entries[] = {
    {
      "WorkOffline", NULL,
      N_("_Work Offline"), NULL,
      N_("Work Offline"),
      reinterpret_cast<GCallback>(work_offline_toggled_cb),
      FALSE
    }
  };

  action_group = gtk_action_group_new("OchushaActions");
  gtk_action_group_set_translation_domain(action_group, PACKAGE_NAME);
  gtk_action_group_add_actions(action_group, entries,
			       G_N_ELEMENTS(entries), this);
  gtk_action_group_add_toggle_actions(action_group, toggle_entries,
				      G_N_ELEMENTS(toggle_entries), this);

  cut_action = gtk_action_group_get_action(action_group, "CutSelection");
  gtk_action_set_sensitive(cut_action, FALSE);

  copy_action = gtk_action_group_get_action(action_group, "CopySelection");
  gtk_action_set_sensitive(copy_action, FALSE);

  paste_action = gtk_action_group_get_action(action_group, "PasteSelection");
  gtk_action_set_sensitive(paste_action, FALSE);

  delete_action = gtk_action_group_get_action(action_group, "DeleteSelection");
  gtk_action_set_sensitive(delete_action, FALSE);

  work_offline_action
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "WorkOffline"));
  stop_action = gtk_action_group_get_action(action_group, "Stop");
  gtk_action_set_sensitive(stop_action, FALSE);

  open_url_action
    = ochusha_entry_action_new("OpenURL",
			       _("Open _Location"),
			       _("Open Location"),
			       GTK_STOCK_OPEN);
  ochusha_entry_action_set_default_width_chars(open_url_action, 48);
  ochusha_entry_action_set_enable_completion(open_url_action, TRUE);
  g_signal_connect(open_url_action, "entry_activate",
		   reinterpret_cast<GCallback>(open_url_action_activate_cb),
		   this);
  gtk_action_group_add_action_with_accel(action_group,
					 GTK_ACTION(open_url_action), "");
  g_signal_connect(open_url_action, "activate",
		   reinterpret_cast<GCallback>(open_url_cb), this);


  gtk_ui_manager_insert_action_group(ui_manager, action_group, 0);

  return action_group;
}


void
Controller::save_yourself()
{
  main_window.prepare_session();

  thread_ui.prepare_session();
  board_ui.prepare_session();
  boardlist_ui.prepare_session();
  bookmarks_ui.prepare_session();
  threadlist_ui.prepare_session();

  boardlist_ui.save_boardlist();
  bookmarks_ui.save_bookmarks();
  config.write_config_xml(repository);
  prefs.write_prefs_xml(repository);
  tag_manager.write_prefs_css(repository);
}


void
Controller::quit()
{
  assert(OCHUSHA_IN_GDK_THREADS());

  image_manager->prepare_shutdown();

  GDK_THREADS_LEAVE();

  thread_pool->terminate_workers();
  poll_manager.terminate_all();

  GDK_THREADS_ENTER();


  save_yourself();
  session.write_session_xml();

  prepare_shutdown();

  gtk_main_quit();
}


GtkWidget *
Controller::create_open_url_dialog()
{
  GtkWidget *dialog = gtk_message_dialog_new(get_toplevel(),
					     GTK_DIALOG_DESTROY_WITH_PARENT,
					     GTK_MESSAGE_QUESTION,
					     GTK_BUTTONS_OK_CANCEL,
					     _("Open Location"));
  gtk_message_dialog_format_secondary_text(GTK_MESSAGE_DIALOG(dialog),
			_("Input the URL of document you want to open."));
  gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_OK);
  g_signal_connect(dialog, "response",
		   reinterpret_cast<GCallback>(open_url_dialog_response_cb),
		   this);
  g_signal_connect(dialog, "destroy",
		   reinterpret_cast<GCallback>(gtk_widget_destroyed),
		   &open_url_dialog);

  GtkWidget *hbox = gtk_hbox_new(FALSE, 0);
  GtkWidget *label = gtk_label_new_with_mnemonic(_("_URL:"));
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 3);

  GtkWidget *widget = ochusha_entry_action_create_entry_widget(open_url_action);
  GtkWidget *entry = gtk_bin_get_child(GTK_BIN(widget));
  gtk_label_set_mnemonic_widget(GTK_LABEL(label), entry);
  gtk_entry_set_activates_default(GTK_ENTRY(entry), TRUE);
  new EntryFocusNotifier(*this, GTK_ENTRY(entry), true);

  gtk_box_pack_start(GTK_BOX(hbox), widget, TRUE, TRUE, 0);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), hbox, TRUE, TRUE, 0);
  g_object_set_data(G_OBJECT(dialog), "entry", entry);

  GtkWidget *button = gtk_check_button_new_with_mnemonic(_("Open in _Tab"));
  const char *value = session.get_state("Controller::open_url_in_tab");
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button),
			       value != NULL && *value != '0');

  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox),
		     button, FALSE, FALSE, 0);
  g_object_set_data(G_OBJECT(dialog), "button", button);

  return dialog;
}


void
Controller::open_url_dialog_response_cb(GtkWidget *dialog, int response_id,
					Controller *controller)
{
  bool in_tab;
  GtkToggleButton *button
    = (GtkToggleButton *)g_object_get_data(G_OBJECT(dialog), "button");
  if (button != NULL)
    {
      in_tab = gtk_toggle_button_get_active(button);

      SessionClient &session = controller->get_session();
      char value[128];
      snprintf(value, 128, "%u", in_tab);
      session.set_state("Controller::open_url_in_tab", value);
    }
  else
    in_tab = false;

  if (response_id == GTK_RESPONSE_OK)
    {
      GtkEntry *entry
	= static_cast<GtkEntry *>(g_object_get_data(G_OBJECT(dialog), "entry"));

      if (entry != NULL)
	{
	  const char *url = gtk_entry_get_text(entry);
	  controller->open(url, true, in_tab, false, true);
	}
    }

  gtk_widget_hide(dialog);
}


void
Controller::show_open_url_dialog()
{
  if (open_url_dialog == NULL)
    open_url_dialog = create_open_url_dialog();

  GtkEntry *entry
    = static_cast<GtkEntry *>(g_object_get_data(G_OBJECT(open_url_dialog),
						"entry"));
  gtk_editable_select_region(GTK_EDITABLE(entry), 0, -1);

  gtk_widget_show_all(open_url_dialog);
}


void
Controller::open_url_cb(GtkAction *action, Controller *controller)
{
  if (!controller->main_window.url_entry_grab_focus())
    controller->show_open_url_dialog();
}


void
Controller::open_url_action_activate_cb(OchushaEntryAction *action,
					const gchar *text,
					Controller *controller)
{
  ochusha_entry_action_insert_completion_text(action, 0, text);
}


void
Controller::clear_cookies_cb(GtkAction *action, Controller *controller)
{
  controller->get_cookie_manager().delete_all_cookies();
}


void
Controller::open_file_cb(GtkAction *action, Controller *controller)
{
  // controller->open_open_file_dialog();
}


void
Controller::cut_cb(GtkAction *action, Controller *controller)
{
  SelectionData *data = SelectionData::get_from(controller->clipboard_internal);
  g_return_if_fail(data != NULL);

  data->do_cut(controller->clipboard);
}


void
Controller::copy_cb(GtkAction *action, Controller *controller)
{
  SelectionData *data = SelectionData::get_from(controller->clipboard_internal);
  g_return_if_fail(data != NULL);

  data->do_copy(controller->clipboard);
}


void
Controller::paste_cb(GtkAction *action, Controller *controller)
{
  SelectionData *data = SelectionData::get_from(controller->clipboard_internal);
  g_return_if_fail(data != NULL);

  data->do_paste(controller->clipboard);
}


void
Controller::delete_cb(GtkAction *action, Controller *controller)
{
  SelectionData *data = SelectionData::get_from(controller->clipboard_internal);
  g_return_if_fail(data != NULL);

  data->do_delete();
}


void
Controller::quit_cb(GtkAction *action, Controller *controller)
{
  controller->session.discard_session();
  controller->quit();
}


void
Controller::stop_cb(GtkAction *action, Controller *controller)
{
  controller->poll_manager.terminate_all();
}


void
Controller::notification_cb(GtkAction *action, Controller *controller)
{
  controller->notification.show();
}


void
Controller::debug_console_cb(GtkAction *action, Controller *controller)
{
  controller->debug_console.show();
}


void
Controller::scheme_console_cb(GtkAction *action, Controller *controller)
{
  controller->scheme_manager.console.show();
}


void
Controller::preferences_cb(GtkAction *action, Controller *controller)
{
  controller->prefs_ui.show_dialog();
}


void
Controller::select_list_fonts_cb(GtkAction *action, Controller *controller)
{
  controller->fonts_ui.show_tree_view_font_selection_dialog();
}


void
Controller::select_thread_fonts_cb(GtkAction *action, Controller *controller)
{
  controller->fonts_ui.show_thread_view_font_selection_dialog();
}


void
Controller::about_cb(GtkAction *action, Controller *controller)
{
  controller->about.show();
}


void
Controller::help_cb(GtkAction *action, Controller *controller)
{
  controller->open(OCHUSHA_HELP_URL, true, true, true);
}


void
Controller::work_offline_toggled_cb(GtkToggleAction *action,
				    Controller *controller)
{
  controller->config.set_offline(gtk_toggle_action_get_active(action));
}


bool
Controller::is_bbs_link(const char *link)
{
  BulletinBoard *board = bbs_table->guess_board(link);
  return board != NULL;
}


bool
Controller::is_board_url(const char *url)
{
  g_return_val_if_fail(url != NULL && *url != '\0', false);

  BulletinBoard *board = bbs_table->guess_board(url);
  if (board != NULL)
    {
      GDK_THREADS_LEAVE();

      board->read_threadlist_xml(repository);

      GDK_THREADS_ENTER();


      return board->lookup_bbs_thread_by_url(url) == NULL;
    }

  return false;
}


bool
Controller::is_thread_url(const char *url)
{
  g_return_val_if_fail(url != NULL && *url != '\0', false);

  BulletinBoard *board = bbs_table->guess_board(url);
  if (board != NULL)
    {
      GDK_THREADS_LEAVE();

      board->read_threadlist_xml(repository);

      GDK_THREADS_ENTER();


      return board->lookup_bbs_thread_by_url(url) != NULL;
    }

  return false;
}


void
Controller::open(const char *url, bool refresh, bool in_tab, bool with_browser,
		 bool select_tab)
{
  if (url == NULL || *url == '\0')
    return;

  if (BulletinBoard::is_active_folder_url(url))
    {
      if (with_browser)
	return;		// ƥ֥եϥ֥֥饦Ǥϳʤ

      BulletinBoard *board = boardlist_ui.lookup_active_folder(url);
      if (board != NULL)
	open(board, refresh, in_tab, false, select_tab);

      // оݤȤʤ֥åޡեäƤۤäƼԡ
      return;
    }

  if (!with_browser)
    {
      BulletinBoard *board = bbs_table->guess_board(url);
      if (board != NULL)
	{
	  GDK_THREADS_LEAVE();

	  board->read_threadlist_xml(repository);

	  GDK_THREADS_ENTER();


	  int from = 0;
	  BBSThread *thread = board->lookup_bbs_thread_by_url(url, &from);
	  if (thread != NULL)
	    {
	      open(thread, refresh, from, in_tab, false, select_tab);
	      return;
	    }

	  open(board, refresh, in_tab, false, select_tab);
	  return;
	}
    }

  assert(strncmp("oabf:///sn-", url, 11) != 0);

  history_manager.record(url);

  // web browserǳ
  char command_line[4096];
  if (snprintf(command_line, 4096,
	       prefs.get_web_browser_template(), url) < 4096)
    spawn_command(command_line);
  else
    {
      gchar *long_command_line
	= g_strdup_printf(prefs.get_web_browser_template(), url);
      spawn_command(long_command_line);
      g_free(long_command_line);
    }
}


void
Controller::open(BulletinBoard *board, bool refresh, bool in_tab,
		 bool with_browser, bool select_tab)
{
  if (with_browser)
    {
      open(board->get_base_url(), refresh, in_tab, with_browser, select_tab);
      return;
    }

  if (view_manager.find_board_view(board, (!in_tab || select_tab)) != NULL)
    return;	// ɽ

  history_manager.record(board->get_base_url());

  const gchar *stock_id;
  switch (board->get_bbs_type())
    {
    case OCHUSHA_BBS_TYPE_PSEUDO_FOLDER:
      stock_id = OCHUSHA_STOCK_PSEUDO_FOLDER;
      break;

    case OCHUSHA_BBS_TYPE_PASSIVE_FOLDER:
      stock_id = OCHUSHA_STOCK_BOOKMARK_FOLDER;
      break;

    case OCHUSHA_BBS_TYPE_ACTIVE_FOLDER:
      stock_id = OCHUSHA_STOCK_ACTIVE_FOLDER;
      break;

    default:
      stock_id = OCHUSHA_STOCK_BOARD;
    }

  TabLabel *label = new TabLabel(board->get_short_name(), stock_id);
  View *view = board_ui.create_board_view(board, label, refresh);
  view->notify_attached(&main_window);

  select_tab |= prefs.get_select_tab_when_opened();
  view_manager.append_board_view(board, view, in_tab, select_tab);
}


void
Controller::open(BBSThread *thread, bool refresh, int initial_pos, bool in_tab,
		 bool with_browser, bool select_tab)
{
  if (with_browser)
    {
      char *url = thread->get_url();
      open(url, refresh, in_tab, with_browser, select_tab);
      free(url);
      return;
    }

  if (view_manager.find_thread_view(thread, (!in_tab || select_tab)) != NULL)
    {
      if (initial_pos != 0)
	thread_ui.specify_thread_view_position(thread, initial_pos);
      else if (refresh && thread->get_number_of_responses_to_get() > 0)
	this->refresh(thread, false);
      return;	// ɽ
    }

  char *url = thread->get_url();
  history_manager.record(url);
  free(url);

  char title_buf[256];
  const char *title = thread->get_title();
  if (title == NULL)
    {
      snprintf(title_buf, 256, _("(tmp)%s"), thread->get_id());
      title = title_buf;
    }
  TabLabel *label = new TabLabel(title, OCHUSHA_STOCK_THREAD);

  BBSThreadView *view
    = thread_ui.create_thread_view(thread, label, initial_pos, refresh);
  view->notify_attached(&main_window);

  select_tab |= prefs.get_select_tab_when_opened();
  view_manager.append_thread_view(thread, view, in_tab, select_tab);
}


void
Controller::refresh(BulletinBoard *board, bool reload)
{
  View *view = view_manager.find_board_view(board, false);
  g_return_if_fail(view != NULL);

  ThreadlistView *threadlist_view = dynamic_cast<ThreadlistView *>(view);
  g_return_if_fail(threadlist_view != NULL);

  threadlist_view->update(reload);

  history_manager.record(board->get_base_url());
}


void
Controller::refresh(BBSThread *thread, bool automatic)
{
  View *view = view_manager.find_thread_view(thread, false);
  g_return_if_fail(view != NULL);

  BBSThreadView *thread_view = dynamic_cast<BBSThreadView *>(view);
  g_return_if_fail(thread_view != NULL);

  thread_view->update(true, automatic);

  char *url = thread->get_url();
  history_manager.record(url);
  free(url);
}


gchar *
Controller::get_bookmark_title(const char *url)
{
  g_return_val_if_fail(url != NULL, NULL);

  BulletinBoard *board = bbs_table->guess_board(url);
  if (board != NULL)
    {
      int from = 0;
      int to = 0;
      BBSThread *thread = board->lookup_bbs_thread_by_url(url, &from, &to);
      if (thread != NULL && !thread->is_tmp_thread())
	{
	  if (from > 0)
	    return g_strdup_printf(_("%d of %s"), from, thread->get_title());

	  return g_strdup(thread->get_title());
	}

      return g_strdup(board->get_name());
    }

  return NULL;
}


void
Controller::bookmark(const char *url, const char *name)
{
  g_return_if_fail(url != NULL);
  g_return_if_fail(*url != '\0');

  if (name != NULL)
    {
      bookmarks_ui.bookmark(url, name);
      return;
    }

  gchar *bookmark_title = get_bookmark_title(url);
  bookmarks_ui.bookmark(url, bookmark_title);
  if (bookmark_title != NULL)
    g_free(bookmark_title);
}


void
Controller::bookmark(BulletinBoard *board)
{
  bookmark(board->get_base_url(), board->get_name());
}


void
Controller::bookmark(BBSThread *thread)
{
  bookmark(thread->get_url(), thread->get_title());
}


// GtkClipboardϢ
void
Controller::owner_change_cb(GtkClipboard *clip, GdkEvent *event,
			    Controller *controller)
{
  g_return_if_fail(GTK_IS_CLIPBOARD(clip));

  if (controller->get_toplevel() == NULL)
    return;	// λ

  if (clip == controller->clipboard_internal)
    {
      controller->notify_clipboard_internal_changed(clip);
      return;
    }

  if (clip == controller->clipboard)
    {
      controller->notify_clipboard_changed(clip);
      return;
    }

  g_return_if_fail(clip == controller->clipboard_primary);
  controller->notify_clipboard_primary_changed(clip);
}


void
Controller::copy(const char *text)
{
  gtk_clipboard_set_text(clipboard_primary, text, -1);
  gtk_clipboard_set_text(clipboard, text, -1);
}


void
Controller::notify_clipboard_changed(GtkClipboard *clip)
{
  g_return_if_fail(clip == clipboard);

  SelectionData *data = SelectionData::get_from(clipboard_internal);
  if (data != NULL)
    gtk_action_set_sensitive(paste_action, data->pastable(clip));
  else
    gtk_action_set_sensitive(paste_action, FALSE);
}


bool
Controller::update_clipboard(SelectionData *data)
{
  return data->set_to(clipboard);
}


void
Controller::clear_clipboard(SelectionData *data)
{
  SelectionData *current_data = SelectionData::get_from(clipboard);
  if (current_data == data)
    gtk_clipboard_clear(clipboard);
}


void
Controller::notify_clipboard_primary_changed(GtkClipboard *clip)
{
  g_return_if_fail(clip == clipboard_primary);

  // MEMO: ѲΤGDK_SELECTION_PRIMARYбGtkClipboard
  //       Ƕ̣Τclipboard_internal
  //       GtkEntryʤGTK+͡ʥãϡselectionѲ˱
  //       primary򹹿ƤΤǡΥߥ󥰤
  //       clipboard_internalåprimary򹹿GtkWidget
  //       selectionϤŪˤƱclipboard_internal˼
  //       ޤäƤȦ
  SelectionData *data = SelectionData::get_from(clipboard_internal);
  if (data != NULL)
    {
      gtk_action_set_sensitive(cut_action, data->cuttable());
      gtk_action_set_sensitive(copy_action, data->copiable());
      gtk_action_set_sensitive(paste_action, data->pastable(clipboard));
      gtk_action_set_sensitive(delete_action, data->deletable());
    }
  else
    {
      gtk_action_set_sensitive(cut_action, FALSE);
      gtk_action_set_sensitive(copy_action, FALSE);
      gtk_action_set_sensitive(paste_action, FALSE);
      gtk_action_set_sensitive(delete_action, FALSE);
    }
}


bool
Controller::update_clipboard_primary(SelectionData *data)
{
  return data->set_to(clipboard_primary);
}


void
Controller::clear_clipboard_primary(SelectionData *data)
{
  SelectionData *current_data = SelectionData::get_from(clipboard_primary);
  if (current_data == data)
    gtk_clipboard_clear(clipboard_primary);
}


void
Controller::notify_clipboard_internal_changed(GtkClipboard *clip)
{
  g_return_if_fail(clip == clipboard_internal);

  SelectionData *data = SelectionData::get_from(clip);
  if (data != NULL)
    {
      gtk_action_set_sensitive(cut_action, data->cuttable());
      gtk_action_set_sensitive(copy_action, data->copiable());
      gtk_action_set_sensitive(paste_action, data->pastable(clipboard));
      gtk_action_set_sensitive(delete_action, data->deletable());
    }
  else
    {
      gtk_action_set_sensitive(cut_action, FALSE);
      gtk_action_set_sensitive(copy_action, FALSE);
      gtk_action_set_sensitive(paste_action, FALSE);
      gtk_action_set_sensitive(delete_action, FALSE);
    }
}


bool
Controller::update_clipboard_internal(SelectionData *data)
{
  return data->set_to(clipboard_internal);
}


void
Controller::clear_clipboard_internal(SelectionData *data)
{
  SelectionData *current_data = SelectionData::get_from(clipboard_internal);
  if (current_data == data)
    gtk_clipboard_clear(clipboard_internal);
}


}
