/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: focus_utils.cc,v 1.4 2010/08/15 15:42:20 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "focus_utils.h"

#include "controller.h"
#include "gtk_utils.h"
#include "preferences.h"

#include <cstdio>

#include <gtk/gtk.h>


namespace Ochusha
{

//#define DEBUG_FOCUS_NOTIFIER		1
//#define DEBUG_FOCUS_NOTIFIER_MOST	1
//#define VERBOSE_DESTRUCTOR		1


FocusNotifier::FocusNotifier(Controller &controller, GtkWidget *widget,
			     bool delete_when_widget_destroyed)
  : controller(controller),
    widget(widget),
    focus_in_event_id(0),
    focus_out_event_id(0),
    enter_notify_event_id(0),
    leave_notify_event_id(0),
    realize_id(0),
    unmap_id(0),
    destroy_id(0),
    hide_id(0),
    state_changed_id(0),
    in_focus(false),
    with_pointer(false),
    delete_when_widget_destroyed(delete_when_widget_destroyed)
{
  set_widget(widget);
}


FocusNotifier::~FocusNotifier()
{
  disconnect_all();
}


void
FocusNotifier::disconnect_all()
{
  if (widget == NULL)
    return;

  if (focus_in_event_id > 0)
    g_signal_handler_disconnect(widget, focus_in_event_id);

  if (focus_out_event_id > 0)
    g_signal_handler_disconnect(widget, focus_out_event_id);

  if (enter_notify_event_id > 0)
    g_signal_handler_disconnect(widget, enter_notify_event_id);

  if (leave_notify_event_id > 0)
    g_signal_handler_disconnect(widget, leave_notify_event_id);

  if (realize_id > 0)
    g_signal_handler_disconnect(widget, realize_id);

  if (unmap_id > 0)
    g_signal_handler_disconnect(widget, unmap_id);

  if (destroy_id > 0)
    g_signal_handler_disconnect(widget, destroy_id);

  if (hide_id > 0)
    g_signal_handler_disconnect(widget, hide_id);

  if (state_changed_id > 0)
    g_signal_handler_disconnect(widget, state_changed_id);

  widget = NULL;
}


void
FocusNotifier::set_events_mask(GdkWindow *window)
{
  g_return_if_fail(window != NULL);

  gdk_window_set_events(window,
			static_cast<GdkEventMask>(gdk_window_get_events(window)
						  | GDK_ENTER_NOTIFY_MASK
						  | GDK_LEAVE_NOTIFY_MASK
						  | GDK_FOCUS_CHANGE_MASK));
}


gboolean
FocusNotifier::focus_in_event_cb(GtkWidget *widget, GdkEventFocus *event,
				 FocusNotifier *notifier)
{
  g_return_val_if_fail(notifier->widget == widget, FALSE);

  notifier->in_focus = true;

  return notifier->notify_focus_in();
}


gboolean
FocusNotifier::focus_out_event_cb(GtkWidget *widget, GdkEventFocus *event,
				  FocusNotifier *notifier)
{
  g_return_val_if_fail(notifier->widget == widget, FALSE);

  notifier->in_focus = false;

  return notifier->notify_focus_out();
}


gboolean
FocusNotifier::enter_notify_event_cb(GtkWidget *widget, GdkEventCrossing *event,
				     FocusNotifier *notifier)
{
  g_return_val_if_fail(notifier->widget == widget, FALSE);

  notifier->with_pointer = true;

  return notifier->notify_enter();
}


gboolean
FocusNotifier::leave_notify_event_cb(GtkWidget *widget, GdkEventCrossing *event,
				     FocusNotifier *notifier)
{
  g_return_val_if_fail(notifier->widget == widget, FALSE);

  notifier->with_pointer = false;

  return notifier->notify_leave();
}


void
FocusNotifier::realize_cb(GtkWidget *widget, FocusNotifier *notifier)
{
  notifier->set_events_mask(notifier->get_window());
  notifier->notify_realize();
}


void
FocusNotifier::unmap_cb(GtkWidget *widget, FocusNotifier *notifier)
{
  notifier->notify_unmap();
}


void
FocusNotifier::destroy_cb(GtkWidget *widget, FocusNotifier *notifier)
{
#ifdef DEBUG_FOCUS_NOTIFIER_MOST
  fprintf(stderr, "FocusNotifier::destroy_cb(%p, %p)\n", widget, notifier);
#endif
  notifier->widget = NULL;
  notifier->focus_in_event_id = 0;
  notifier->focus_out_event_id = 0;
  notifier->enter_notify_event_id = 0;
  notifier->leave_notify_event_id = 0;
  notifier->realize_id = 0;
  notifier->unmap_id = 0;
  notifier->destroy_id = 0;
  notifier->hide_id = 0;
  notifier->state_changed_id = 0;

  if (notifier->delete_when_widget_destroyed)
    delete notifier;
}


void
FocusNotifier::hide_cb(GtkWidget *widget, FocusNotifier *notifier)
{
  notifier->notify_hide();
}


void
FocusNotifier::state_changed_cb(GtkWidget *widget, GtkStateType state,
				FocusNotifier *notifier)
{
  notifier->notify_state_changed(state);
}


GdkWindow *
FocusNotifier::get_window()
{
#if GTK_CHECK_VERSION(2,14,0)
  return gtk_widget_get_window(widget);
#else
  g_return_val_if_fail(widget != NULL, NULL);
  return widget->window;
#endif
}


gboolean
FocusNotifier::notify_focus_in()
{
#ifdef DEBUG_FOCUS_NOTIFIER_MOST
  fprintf(stderr, "FocusNotifier::notify_focus_in(): %p\n", this);
#endif

  return FALSE;
}


gboolean
FocusNotifier::notify_focus_out()
{
#ifdef DEBUG_FOCUS_NOTIFIER_MOST
  fprintf(stderr, "FocusNotifier::notify_focus_out(): %p\n", this);
#endif

  return FALSE;
}


gboolean
FocusNotifier::notify_enter()
{
#ifdef DEBUG_FOCUS_NOTIFIER_MOST
  fprintf(stderr, "FocusNotifier::notify_enter(): %p\n", this);
#endif

  if (controller.get_prefs().get_focus_follow_mouse())
    gtk_widget_grab_focus(widget);

  return FALSE;
}


gboolean
FocusNotifier::notify_leave()
{
#ifdef DEBUG_FOCUS_NOTIFIER_MOST
  fprintf(stderr, "FocusNotifier::notify_leave(): %p\n", this);
#endif

  return FALSE;
}


void
FocusNotifier::notify_hide()
{
}


void
FocusNotifier::notify_realize()
{
}


void
FocusNotifier::notify_unmap()
{
}


void
FocusNotifier::notify_state_changed(GtkStateType state)
{
}


void
FocusNotifier::set_widget(GtkWidget *widget)
{
  if (this->widget != NULL)
    disconnect_all();

  if (widget != NULL)
    {
      focus_in_event_id
	= g_signal_connect(widget, "focus-in-event",
			   reinterpret_cast<GCallback>(focus_in_event_cb),
			   this);
      focus_out_event_id
	= g_signal_connect(widget, "focus-out-event",
			   reinterpret_cast<GCallback>(focus_out_event_cb),
			   this);
      enter_notify_event_id
	= g_signal_connect(widget, "enter-notify-event",
			   reinterpret_cast<GCallback>(enter_notify_event_cb),
			   this);
      leave_notify_event_id
	= g_signal_connect(widget, "leave-notify-event",
			   reinterpret_cast<GCallback>(leave_notify_event_cb),
			   this);

      realize_id = g_signal_connect(widget, "realize",
				    reinterpret_cast<GCallback>(realize_cb),
				    this);
      unmap_id = g_signal_connect(widget, "unmap",
				  reinterpret_cast<GCallback>(unmap_cb), this);
      destroy_id
	= g_signal_connect_after(widget, "destroy",
				 reinterpret_cast<GCallback>(destroy_cb), this);
      hide_id = g_signal_connect(widget, "hide",
				 reinterpret_cast<GCallback>(hide_cb), this);
      state_changed_id
	= g_signal_connect(widget, "state-changed",
			   reinterpret_cast<GCallback>(state_changed_cb), this);

      this->widget = widget;

      GdkWindow *window = get_window();
      if (window != NULL)
	set_events_mask(window);
    }
}


TextViewFocusNotifier::TextViewFocusNotifier(Controller &controller,
					     GtkTextView *view,
					     bool delete_when_widget_destroyed)
  : FocusNotifier(controller, GTK_WIDGET(view), delete_when_widget_destroyed),
    selection_data(view, true),
    has_selection(FALSE)
{
  GtkTextBuffer *buffer = gtk_text_view_get_buffer(view);
  if (buffer != NULL)
    {
      g_signal_connect(view, "notify::has-selection",
		       reinterpret_cast<GCallback>(notify_has_selection_cb),
		       this);
#if GTK_CHECK_VERSION(2,10,0)
      has_selection = gtk_text_buffer_get_has_selection(buffer);
#else
      has_selection = gtk_text_buffer_get_selection_bounds(buffer, NULL, NULL);
#endif
    }

  g_signal_connect(view, "populate_popup",
		   reinterpret_cast<GCallback>(populate_popup_cb), this);
}


TextViewFocusNotifier::~TextViewFocusNotifier()
{
#ifdef VERBOSE_DESTRUCTOR
  fprintf(stderr, "TextViewFocusNotifier::~TextViewFocusNotifier(): %p\n",
	  this);
#endif
  get_controller().clear_clipboard_internal(&selection_data);
}


void
TextViewFocusNotifier::notify_has_selection_cb(GtkTextBuffer *buffer,
					       GParamSpec *pspec,
					       TextViewFocusNotifier *notifier)
{
#ifdef DEBUG_FOCUS_NOTIFIER
  fprintf(stderr, "TextViewFocusNotifier::notify_has_selection_cb(): %p\n",
	  notifier);
#endif

#if GTK_CHECK_VERSION(2,10,0)
  gboolean has_selection = gtk_text_buffer_get_has_selection(buffer);
#else
  gboolean has_selection
    = gtk_text_buffer_get_selection_bounds(buffer, NULL, NULL);
#endif

  if (has_selection == notifier->has_selection)
    return;

  notifier->has_selection = has_selection;
  notifier->get_controller()
    .update_clipboard_internal(&notifier->selection_data);
}


void
TextViewFocusNotifier::populate_popup_cb(GtkTextView *view, GtkMenu *popup_menu,
					 TextViewFocusNotifier *notifier)
{
  notifier->get_controller().get_scheme_manager()
    .attach_external_text_tools_menu(popup_menu, &notifier->selection_data);
}


gboolean
TextViewFocusNotifier::notify_focus_in()
{
#ifdef DEBUG_FOCUS_NOTIFIER_MOST
  fprintf(stderr, "TextViewFocusNotifier::notify_focus_in(): %p\n", this);
#endif
  get_controller().update_clipboard_internal(&selection_data);

  return FALSE;
}


void
TextViewFocusNotifier::notify_hide()
{
#ifdef DEBUG_FOCUS_NOTIFIER
  fprintf(stderr, "TextViewFocusNotifier::notify_hide(): %p\n", this);
#endif

  get_controller().clear_clipboard_internal(&selection_data);
}


void
TextViewFocusNotifier::notify_unmap()
{
#ifdef DEBUG_FOCUS_NOTIFIER
  fprintf(stderr, "TextViewFocusNotifier::notify_unmap(): %p\n", this);
#endif

  get_controller().clear_clipboard_internal(&selection_data);
}


void
TextViewFocusNotifier::notify_state_changed(GtkStateType state)
{
#ifdef DEBUG_FOCUS_NOTIFIER
  fprintf(stderr, "TextViewFocusNotifier::notify_state_changed(%d): %p\n",
	  state, this);
#endif

  if (state == GTK_STATE_INSENSITIVE)
    get_controller().clear_clipboard_internal(&selection_data);
}


GtkWidget *
TextViewFocusNotifier::gtk_text_view_new_with_buffer(Controller &controller,
						     GtkTextBuffer *buffer)
{
  GtkWidget *widget = ::gtk_text_view_new_with_buffer(buffer);
  g_return_val_if_fail(widget != NULL, NULL);

  new TextViewFocusNotifier(controller, GTK_TEXT_VIEW(widget), true);
  new TextViewKeyPressHandler(GTK_TEXT_VIEW(widget), controller);

  return widget;
}


EntryFocusNotifier::EntryFocusNotifier(Controller &controller, GtkEntry *entry,
				       bool delete_when_widget_destroyed)
  : FocusNotifier(controller, GTK_WIDGET(entry), delete_when_widget_destroyed),
    selection_data(GTK_WIDGET(entry), true),
    has_selection(gtk_editable_get_selection_bounds(GTK_EDITABLE(entry), NULL,
						    NULL))
{
  g_signal_connect(entry, "notify::selection-bound",
		   reinterpret_cast<GCallback>(notify_selection_bound_cb),
		   this);
  g_signal_connect(entry, "populate_popup",
		   reinterpret_cast<GCallback>(populate_popup_cb), this);
}


EntryFocusNotifier::~EntryFocusNotifier()
{
#ifdef VERBOSE_DESTRUCTOR
  fprintf(stderr, "EntryFocusNotifier::~EntryFocusNotifier(): %p\n", this);
#endif
  get_controller().clear_clipboard_internal(&selection_data);
}


void
EntryFocusNotifier::notify_selection_bound_cb(GtkEntry *entry,
					      GParamSpec *pspec,
					      EntryFocusNotifier *notifier)
{
  // MEMO: ݤäݤ
  // GtkEntryselectionξ֤ѤäƤnotify::selection-bound
  // ȯԤƤʤˤؤ餺ʸϤȯԤƤġġ
  // GTK+ХäƤ͡
#ifdef DEBUG_FOCUS_NOTIFIER_MOST
  fprintf(stderr, "EntryFocusNotifier::notify_selection_bound_cb(): %p\n",
	  notifier);
#endif
  gboolean has_selection
    = gtk_editable_get_selection_bounds(GTK_EDITABLE(entry), NULL, NULL);

  if (has_selection == notifier->has_selection)
    return;

  notifier->has_selection = has_selection;
  notifier->get_controller()
    .update_clipboard_internal(&notifier->selection_data);
}


void
EntryFocusNotifier::populate_popup_cb(GtkEntry *entry, GtkMenu *popup_menu,
				      EntryFocusNotifier *notifier)
{
  notifier->get_controller().get_scheme_manager()
    .attach_external_text_tools_menu(popup_menu, &notifier->selection_data);
}


gboolean
EntryFocusNotifier::notify_focus_in()
{
#ifdef DEBUG_FOCUS_NOTIFIER_MOST
  fprintf(stderr, "EntryFocusNotifier::notify_focus_in(): %p\n", this);
#endif
  get_controller().update_clipboard_internal(&selection_data);

  return FALSE;
}


gboolean
EntryFocusNotifier::notify_enter()
{
#ifdef DEBUG_FOCUS_NOTIFIER_MOST
  fprintf(stderr, "EntryFocusNotifier::notify_enter(): %p\n", this);
#endif

  if (!get_controller().get_prefs().get_focus_follow_mouse_modestly())
    return FocusNotifier::notify_enter();

  return FALSE;
}


void
EntryFocusNotifier::notify_hide()
{
#ifdef DEBUG_FOCUS_NOTIFIER
  fprintf(stderr, "EntryFocusNotifier::notify_hide(): %p\n", this);
#endif

  get_controller().clear_clipboard_internal(&selection_data);
}


void
EntryFocusNotifier::notify_unmap()
{
#ifdef DEBUG_FOCUS_NOTIFIER
  fprintf(stderr, "EntryFocusNotifier::notify_unmap(): %p\n", this);
#endif

  get_controller().clear_clipboard_internal(&selection_data);
}


void
EntryFocusNotifier::notify_state_changed(GtkStateType state)
{
#ifdef DEBUG_FOCUS_NOTIFIER
  fprintf(stderr, "EntryFocusNotifier::notify_state_changed(%d): %p\n",
	  state, this);
#endif

  if (state == GTK_STATE_INSENSITIVE)
    get_controller().clear_clipboard_internal(&selection_data);
}


GtkWidget *
EntryFocusNotifier::gtk_entry_new(Controller &controller)
{
  GtkWidget *widget = ::gtk_entry_new();
  g_return_val_if_fail(widget != NULL, NULL);

  new EntryFocusNotifier(controller, GTK_ENTRY(widget), true);

  return widget;
}


}
