/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: gdk_procedure.cc,v 1.9 2009/01/01 07:47:06 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "gdk_procedure.h"


#include "libtscore/tscore.h"

#include <gdk/gdk.h>


namespace Ochusha
{

NativeProcedure::Info GdkProcedure::procedures[] = {
  { &GdkProcedure::color_parse,
    "gdk-color-parse", TST_STRING, 1, 1 },
  { &GdkProcedure::event_key_get_keyval,
    "gdk-event-key-get-keyval", TST_NONE, 1, 1 },
  { &GdkProcedure::event_key_get_modifier_state,
    "gdk-event-key-get-modifier-state", TST_NONE, 1, 1 },
  { &GdkProcedure::keyval_name,
    "gdk-keyval-name", TST_INTEGER, 1, 1 },
  { &GdkProcedure::keyval_from_name,
    "gdk-keyval-from-name", TST_STRING, 1, 1 }
};


GdkProcedure::GdkProcedure(TSCore &tscore) : NativeProcedure(tscore)
{
  tscore.define_global_integer("GDK_SHIFT_MASK", GDK_SHIFT_MASK);
  tscore.define_global_integer("GDK_LOCK_MASK", GDK_LOCK_MASK);
  tscore.define_global_integer("GDK_CONTROL_MASK", GDK_CONTROL_MASK);
  tscore.define_global_integer("GDK_MOD1_MASK", GDK_MOD1_MASK);
  tscore.define_global_integer("GDK_MOD2_MASK", GDK_MOD2_MASK);
  tscore.define_global_integer("GDK_MOD3_MASK", GDK_MOD3_MASK);
  tscore.define_global_integer("GDK_MOD4_MASK", GDK_MOD4_MASK);
  tscore.define_global_integer("GDK_MOD5_MASK", GDK_MOD5_MASK);
}


Cell *
GdkProcedure::assign_procedures(Cell *environment, Cell *frame)
{
  TSStackRoot env(core, environment);
  TSStackRoot frm(core, frame);
  size_t n_procs = sizeof(procedures) / sizeof(Info);
  for (size_t i = 0; i < n_procs; i++)
    assign_procedure(env, frm, &procedures[i], this);

  return env;
}


static const char *gdk_color_cell_id = "GdkProcedure::GdkColor";


static void
gdk_color_cell_destroy(void *data, void *option_data)
{
  g_return_if_fail(option_data == gdk_color_cell_id);

  g_free(data);
}


Cell *
GdkProcedure::color_parse(VirtualMachine &vm, Register &context,
			  unsigned long argument_pointer,
			  unsigned long nargs, void *data)
{
  Cell *arg = context.frame->load_variable(argument_pointer);
  assert(arg->is_string());

  GdkColor *color = g_new0(GdkColor, 1);
  if (gdk_color_parse(arg->get_string(), color))
    return vm.mk_foreign_object(color,
				gdk_color_cell_destroy,
				const_cast<char *>(gdk_color_cell_id));

  g_free(color);
  return Cell::nil();
}


Cell *
GdkProcedure::event_key_get_keyval(VirtualMachine &vm, Register &context,
				   unsigned long argument_pointer,
				   unsigned long nargs, void *data)
{
  Cell *arg = context.frame->load_variable(argument_pointer);
  if (!arg->is_foreign_object())
    return signal_error(vm, "gdk-event-key-get-keyval: argument must be of foreign object: ", arg);

  GdkEventKey *event = static_cast<GdkEventKey *>(arg->get_foreign_pointer());
  return vm.mk_integer(event->keyval);
}


Cell *
GdkProcedure::event_key_get_modifier_state(VirtualMachine &vm,
					   Register &context,
					   unsigned long argument_pointer,
					   unsigned long nargs, void *data)
{
  Cell *arg = context.frame->load_variable(argument_pointer);
  if (!arg->is_foreign_object())
    return signal_error(vm, "gdk-event-key-get-modifier-state: argument must be of foreign object: ", arg);

  GdkEventKey *event = static_cast<GdkEventKey *>(arg->get_foreign_pointer());
  return vm.mk_integer(event->state & GDK_MODIFIER_MASK);
}


Cell *
GdkProcedure::keyval_name(VirtualMachine &vm, Register &context,
			  unsigned long argument_pointer,
			  unsigned long nargs, void *data)
{
  Cell *arg = context.frame->load_variable(argument_pointer);
  assert(arg->is_integer());

  gchar *name = gdk_keyval_name(static_cast<guint>(arg->get_integer0()));
  return name != NULL ? vm.mk_static_string(name) : Cell::nil();
}


Cell *
GdkProcedure::keyval_from_name(VirtualMachine &vm, Register &context,
			       unsigned long argument_pointer,
			       unsigned long nargs, void *data)
{
  Cell *arg = context.frame->load_variable(argument_pointer);
  assert(arg->is_string());

  return vm.mk_integer(gdk_keyval_from_name(arg->get_string()));
}

}
