/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: html_document.cc,v 1.3 2009/01/01 07:47:06 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "html_document.h"

#include "libochusha/misc_utils.h"

#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <cstring>


namespace Ochusha
{


HTMLElement::HTMLElement(const char *name, bool inline_element)
  : name(strdup(name)),
    inline_element(inline_element)
{
}


HTMLElement::~HTMLElement()
{
  free(name);
}


class HTMLElementHtml : public HTMLElement
{
public:
  HTMLElementHtml();

  virtual bool is_containable(const HTMLElement *content) const;
  virtual bool is_empty_element() const;
};


HTMLElementHtml::HTMLElementHtml()
  : HTMLElement("html", false)
{
}


bool
HTMLElementHtml::is_containable(const HTMLElement *content) const
{
  const char *content_name = content->get_name();
  return strcasecmp(content_name, "head") == 0
    || strcasecmp(content_name, "body") == 0;
}


bool
HTMLElementHtml::is_empty_element() const
{
  return false;
}


class HTMLElementHead : public HTMLElement
{
public:
  HTMLElementHead();

  virtual bool is_containable(const HTMLElement *content) const;
  virtual bool is_empty_element() const;
};


HTMLElementHead::HTMLElementHead()
  : HTMLElement("head", false)
{
}


bool
HTMLElementHead::is_containable(const HTMLElement *content) const
{
  const char *content_name = content->get_name();
  return strcasecmp(content_name, "title") == 0
    || strcasecmp(content_name, "base") == 0
    || strcasecmp(content_name, "meta") == 0
    || strcasecmp(content_name, "link") == 0
    || strcasecmp(content_name, "script") == 0
    || strcasecmp(content_name, "style") == 0
    || strcasecmp(content_name, "object") == 0;
}


bool
HTMLElementHead::is_empty_element() const
{
  return false;
}


class HTMLElementInlinePCData : public HTMLElement
{
public:
  HTMLElementInlinePCData(const char *name);

  virtual bool is_containable(const HTMLElement *content) const;
  virtual bool is_empty_element() const;
};


HTMLElementInlinePCData::HTMLElementInlinePCData(const char *name)
  : HTMLElement(name, true)
{
}


bool
HTMLElementInlinePCData::is_containable(const HTMLElement *content) const
{
  return false;
}


bool
HTMLElementInlinePCData::is_empty_element() const
{
  return false;
}


class HTMLElementBody : public HTMLElement
{
public:
  HTMLElementBody();

  virtual bool is_containable(const HTMLElement *content) const;
  virtual bool is_empty_element() const;
};


HTMLElementBody::HTMLElementBody()
  : HTMLElement("body", false)
{
}


bool
HTMLElementBody::is_containable(const HTMLElement *content) const
{
  // ̩ˤhtmlheadϵʤ
  // Strict DTDǤϥ֥åǰʳϵʤ
  // ˤƤäȾ꤯Ԥ
  return true;
}


bool
HTMLElementBody::is_empty_element() const
{
  return false;
}


class HTMLElementInline : public HTMLElement
{
public:
  HTMLElementInline(const char *name);

  virtual bool is_containable(const HTMLElement *content) const;
  virtual bool is_empty_element() const;
};


HTMLElementInline::HTMLElementInline(const char *name)
  : HTMLElement(name, true)
{
}


bool
HTMLElementInline::is_containable(const HTMLElement *content) const
{
  return content->is_inline_element();
}


bool
HTMLElementInline::is_empty_element() const
{
  return false;
}


class HTMLElementAnchor : public HTMLElement
{
public:
  HTMLElementAnchor(const char *name);

  virtual bool is_containable(const HTMLElement *content) const;
  virtual bool is_empty_element() const;
};


HTMLElementAnchor::HTMLElementAnchor(const char *name)
  : HTMLElement(name, true)
{
}


bool
HTMLElementAnchor::is_containable(const HTMLElement *content) const
{
  return content->is_inline_element()
    && strcasecmp(content->get_name(), "a") != 0;
}


bool
HTMLElementAnchor::is_empty_element() const
{
  return false;
}


class HTMLElementInlineEmpty : public HTMLElement
{
public:
  HTMLElementInlineEmpty(const char *name);

  virtual bool is_containable(const HTMLElement *content) const;
  virtual bool is_empty_element() const;
};


HTMLElementInlineEmpty::HTMLElementInlineEmpty(const char *name)
  : HTMLElement(name, true)
{
}


bool
HTMLElementInlineEmpty::is_containable(const HTMLElement *content) const
{
  return false;
}


bool
HTMLElementInlineEmpty::is_empty_element() const
{
  return true;
}


class HTMLElementInlineAny : public HTMLElement
{
public:
  HTMLElementInlineAny(const char *name);

  virtual bool is_containable(const HTMLElement *content) const;
  virtual bool is_empty_element() const;
};


HTMLElementInlineAny::HTMLElementInlineAny(const char *name)
  : HTMLElement(name, true)
{
}


bool
HTMLElementInlineAny::is_containable(const HTMLElement *content) const
{
  return true;
}


bool
HTMLElementInlineAny::is_empty_element() const
{
  return false;
}


class HTMLElementInlineForm : public HTMLElement
{
public:
  HTMLElementInlineForm(const char *name);

  virtual bool is_containable(const HTMLElement *content) const;
  virtual bool is_empty_element() const;
};


HTMLElementInlineForm::HTMLElementInlineForm(const char *name)
  : HTMLElement(name, true)
{
}


bool
HTMLElementInlineForm::is_containable(const HTMLElement *content) const
{
  if (strcasecmp(name, "input") == 0 || strcasecmp(name, "textarea") == 0)
    return false;

  if (strcasecmp(name, "select") == 0)
    {
      const char *content_name = content->get_name();
      return strcasecmp(content_name, "optgroup") == 0
	|| strcasecmp(content_name, "option") == 0;
    }

  if (strcasecmp(name, "label") == 0)
    return content->is_inline_element()
      && strcasecmp(content->get_name(), "label") != 0;

  const char *content_name = content->get_name();
  return strcasecmp(content_name, "input") != 0
    && strcasecmp(content_name, "select") != 0
    && strcasecmp(content_name, "textarea") != 0
    && strcasecmp(content_name, "label") != 0
    && strcasecmp(content_name, "button") != 0
    && strcasecmp(content_name, "a") != 0
    && strcasecmp(content_name, "form") != 0
    && strcasecmp(content_name, "fieldset") != 0;
}


bool
HTMLElementInlineForm::is_empty_element() const
{
  return strcasecmp(name, "input") == 0;
}


class HTMLElementBlockAny : public HTMLElement
{
public:
  HTMLElementBlockAny(const char *name);

  virtual bool is_containable(const HTMLElement *content) const;
  virtual bool is_empty_element() const;
};


HTMLElementBlockAny::HTMLElementBlockAny(const char *name)
  : HTMLElement(name, false)
{
}


bool
HTMLElementBlockAny::is_containable(const HTMLElement *content) const
{
  return true;
}


bool
HTMLElementBlockAny::is_empty_element() const
{
  return false;
}


class HTMLElementBlockAnyInline : public HTMLElement
{
public:
  HTMLElementBlockAnyInline(const char *name);

  virtual bool is_containable(const HTMLElement *content) const;
  virtual bool is_empty_element() const;
};


HTMLElementBlockAnyInline::HTMLElementBlockAnyInline(const char *name)
  : HTMLElement(name, false)
{
}


bool
HTMLElementBlockAnyInline::is_containable(const HTMLElement *content) const
{
  return content->is_inline_element();
}


bool
HTMLElementBlockAnyInline::is_empty_element() const
{
  return false;
}


class HTMLElementBlockAnyBlock : public HTMLElement
{
public:
  HTMLElementBlockAnyBlock(const char *name);

  virtual bool is_containable(const HTMLElement *content) const;
  virtual bool is_empty_element() const;
};


HTMLElementBlockAnyBlock::HTMLElementBlockAnyBlock(const char *name)
  : HTMLElement(name, false)
{
}


bool
HTMLElementBlockAnyBlock::is_containable(const HTMLElement *content) const
{
  return content->is_block_element();
}


bool
HTMLElementBlockAnyBlock::is_empty_element() const
{
  return false;
}


class HTMLElementBlockTable : public HTMLElement
{
public:
  HTMLElementBlockTable(const char *name);

  virtual bool is_containable(const HTMLElement *content) const;
  virtual bool is_empty_element() const;
};


HTMLElementBlockTable::HTMLElementBlockTable(const char *name)
  : HTMLElement(name, false)
{
}


bool
HTMLElementBlockTable::is_containable(const HTMLElement *content) const
{
  const char *content_name = content->get_name();
  return strcasecmp(content_name, "caption") == 0
    || strcasecmp(content_name, "tbody") == 0
    || strcasecmp(content_name, "thead") == 0
    || strcasecmp(content_name, "tfoot") == 0
    || strcasecmp(content_name, "colgroup") == 0;
}


bool
HTMLElementBlockTable::is_empty_element() const
{
  return false;
}


class HTMLElementTableGroup : public HTMLElement
{
public:
  HTMLElementTableGroup(const char *name);

  virtual bool is_containable(const HTMLElement *content) const;
  virtual bool is_empty_element() const;
};


HTMLElementTableGroup::HTMLElementTableGroup(const char *name)
  : HTMLElement(name, false)
{
}


bool
HTMLElementTableGroup::is_containable(const HTMLElement *content) const
{
  if (strcasecmp(name, "colgroup") == 0)
    return strcasecmp(content->get_name(), "col") == 0;
  if (strcasecmp(name, "tr") == 0)
    {
      const char *content_name = content->get_name();
      return strcasecmp(content_name, "th") == 0
	|| strcasecmp(content_name, "td") == 0;
    }
  return strcasecmp(content->get_name(), "tr") == 0;
}


bool
HTMLElementTableGroup::is_empty_element() const
{
  return false;
}


class HTMLElementList : public HTMLElement
{
public:
  HTMLElementList(const char *name);

  virtual bool is_containable(const HTMLElement *content) const;
  virtual bool is_empty_element() const;
};


HTMLElementList::HTMLElementList(const char *name)
  : HTMLElement(name, false)
{
}


bool
HTMLElementList::is_containable(const HTMLElement *content) const
{
  return strcasecmp(content->get_name(), "li") == 0;
}


bool
HTMLElementList::is_empty_element() const
{
  return false;
}


class HTMLElementDefinitionList : public HTMLElement
{
public:
  HTMLElementDefinitionList(const char *name);

  virtual bool is_containable(const HTMLElement *content) const;
  virtual bool is_empty_element() const;
};


HTMLElementDefinitionList::HTMLElementDefinitionList(const char *name)
  : HTMLElement(name, false)
{
}


bool
HTMLElementDefinitionList::is_containable(const HTMLElement *content) const
{
  const char *content_name = content->get_name();
  return strcasecmp(content_name, "dt") == 0
    || strcasecmp(content_name, "dd") == 0;
}


bool
HTMLElementDefinitionList::is_empty_element() const
{
  return false;
}


class HTMLElementBlockEmpty : public HTMLElement
{
public:
  HTMLElementBlockEmpty(const char *name);

  virtual bool is_containable(const HTMLElement *content) const;
  virtual bool is_empty_element() const;
};


HTMLElementBlockEmpty::HTMLElementBlockEmpty(const char *name)
  : HTMLElement(name, false)
{
}


bool
HTMLElementBlockEmpty::is_containable(const HTMLElement *content) const
{
  return false;
}


bool
HTMLElementBlockEmpty::is_empty_element() const
{
  return true;
}


class HTMLElementBlockSpecial : public HTMLElement
{
public:
  HTMLElementBlockSpecial(const char *name);

  virtual bool is_containable(const HTMLElement *content) const;
  virtual bool is_empty_element() const;
};


HTMLElementBlockSpecial::HTMLElementBlockSpecial(const char *name)
  : HTMLElement(name, false)
{
}


bool
HTMLElementBlockSpecial::is_containable(const HTMLElement *content) const
{
  if (strcasecmp(name, "pre") == 0)
    {
      const char *content_name = content->get_name();
      return content->is_inline_element()
	&& strcasecmp(content_name, "img") != 0
	&& strcasecmp(content_name, "object") != 0
	&& strcasecmp(content_name, "big") != 0
	&& strcasecmp(content_name, "small") != 0
	&& strcasecmp(content_name, "sub") != 0
	&& strcasecmp(content_name, "sup") != 0;
    }

  if (strcasecmp(name, "form") == 0)
    {
      const char *content_name = content->get_name();
      return strcasecmp(content_name, "form") != 0
	&& (content->is_block_element()
	    || strcasecmp(content_name, "script") == 0);
    }

  return true;
}


bool
HTMLElementBlockSpecial::is_empty_element() const
{
  return false;
}


HTMLElementManager::Entry::Entry(unsigned int hash, const char *key,
				 HTMLElement *value,
				 HTMLElementManager::HashtableEntry *next)
  : HTMLElementManager::HashtableEntry(hash, strdup(key), value, next)
{
}


HTMLElementManager::Entry::~Entry()
{
  delete static_cast<HTMLElement *>(value);
}


unsigned int
HTMLElementManager::calculate_hash(const char *key) const
{
  return cstring_case_insensitive_hash(key);
}


bool
HTMLElementManager::equals(const char *key1, const char *key2) const
{
  return strcasecmp(key1, key2) == 0;
}


HTMLElementManager::HashtableEntry *
HTMLElementManager::entry_new(unsigned int hash, const char *key,
			      HTMLElement *value,
			      HTMLElementManager::HashtableEntry *next)
{
  return new Entry(hash, key, value, next);
}


HTMLElementManager::HTMLElementManager(int initial_capacity)
  : hashtable<const char *, HTMLElement *>(initial_capacity)
{
  insert(new HTMLElementHtml());

  insert(new HTMLElementHead());
  insert(new HTMLElementInlinePCData("title"));
  insert(new HTMLElementInlineEmpty("base"));
  insert(new HTMLElementInlineEmpty("meta"));
  insert(new HTMLElementInlineEmpty("link"));
  insert(new HTMLElementInlinePCData("script"));
  insert(new HTMLElementInlinePCData("style"));
  insert(new HTMLElementInlineAny("object"));

  insert(new HTMLElementBody());

  insert(new HTMLElementBlockAnyInline("p"));
  insert(new HTMLElementBlockAnyInline("h1"));
  insert(new HTMLElementBlockAnyInline("h2"));
  insert(new HTMLElementBlockAnyInline("h3"));
  insert(new HTMLElementBlockAnyInline("h4"));
  insert(new HTMLElementBlockAnyInline("h5"));
  insert(new HTMLElementBlockAnyInline("h6"));
  insert(new HTMLElementBlockAnyInline("address"));

  insert(new HTMLElementBlockAnyBlock("div"));
  insert(new HTMLElementBlockAnyBlock("blockquote"));
  insert(new HTMLElementBlockAnyBlock("noscript"));

  insert(new HTMLElementBlockTable("table"));
  insert(new HTMLElementInline("caption"));
  insert(new HTMLElementTableGroup("thead"));
  insert(new HTMLElementTableGroup("tbody"));
  insert(new HTMLElementTableGroup("tfoot"));
  insert(new HTMLElementTableGroup("colgroup"));
  insert(new HTMLElementInlineEmpty("col"));
  insert(new HTMLElementTableGroup("tr"));
  insert(new HTMLElementBlockAny("th"));
  insert(new HTMLElementBlockAny("td"));

  insert(new HTMLElementList("ul"));
  insert(new HTMLElementList("ol"));
  insert(new HTMLElementBlockAny("li"));

  insert(new HTMLElementDefinitionList("dl"));
  insert(new HTMLElementBlockAnyInline("dt"));
  insert(new HTMLElementBlockAny("dd"));

  insert(new HTMLElementBlockEmpty("hr"));


  insert(new HTMLElementBlockSpecial("form"));
  insert(new HTMLElementBlockSpecial("fieldset"));
  insert(new HTMLElementBlockSpecial("pre"));

  insert(new HTMLElementInline("em"));
  insert(new HTMLElementInline("strong"));
  insert(new HTMLElementInline("dfn"));
  insert(new HTMLElementInline("code"));
  insert(new HTMLElementInline("samp"));
  insert(new HTMLElementInline("kbd"));
  insert(new HTMLElementInline("var"));
  insert(new HTMLElementInline("cite"));
  insert(new HTMLElementInline("abbr"));
  insert(new HTMLElementInline("acronym"));

  insert(new HTMLElementAnchor("a"));
  insert(new HTMLElementInlineEmpty("img"));
  insert(new HTMLElementInlineEmpty("br"));
  insert(new HTMLElementInline("q"));
  insert(new HTMLElementInline("sub"));
  insert(new HTMLElementInline("sup"));
  insert(new HTMLElementInline("span"));
  insert(new HTMLElementInline("bdo"));

  insert(new HTMLElementInline("tt"));
  insert(new HTMLElementInline("i"));
  insert(new HTMLElementInline("b"));
  insert(new HTMLElementInline("big"));
  insert(new HTMLElementInline("small"));
  insert(new HTMLElementInline("font"));

  insert(new HTMLElementInlineForm("input"));
  insert(new HTMLElementInlineForm("select"));
  insert(new HTMLElementInlineForm("textarea"));
  insert(new HTMLElementInlineForm("label"));
  insert(new HTMLElementInlineForm("button"));
}


void
HTMLElementManager::insert(HTMLElement *element)
{
  hashtable<const char *, HTMLElement *>::insert(element->get_name(), element);
}


}
