/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: html_renderer.cc,v 1.31 2010/08/15 15:42:20 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "html_renderer.h"

#include "libochusha/buffer.h"
#include "libochusha/html_utils.h"
#include "libochusha/misc_utils.h"
#include "libochusha/ref_count.h"
#include "libochusha/repository.h"
#include "libochusha/text_utils.h"

#include "controller.h"
#include "html_document.h"
#include "image_manager.h"
#include "preferences.h"
#include "style_sheet.h"
#include "ui_constants.h"

#include "libochushagtk/hypertext_buffer.h"
#include "libochushagtk/hypertext_tag_table.h"

#include <vector>

#include <cassert>
#include <string>


namespace Ochusha
{


// 줬ˤΤϥʥʤ
static const char ochusha_default_style_sheet[] =
  ":link { color: -ochusha-link-color }\n"
  ":visited { color: -ochusha-visited-color }\n"
  ".ThreadTitle { color: -ochusha-thread-title-color; font-size: 1.2em }\n"
  ".Name { color: -ochusha-name-color }\n"
  ".ResponseBody { margin-left: 40px }\n"
  ".Secret { text-decoration: none; color: inherit }\n";


HTMLRendererTagManager::HTMLRendererTagManager(Controller &controller)
  : controller(controller)
{
  OchushaHypertextTagTable *table
    = ochusha_hypertext_buffer_get_default_tag_table();

  // ǤˤĤ(ܤˤϱƶͿʤ)
  GtkTextTag *tag = gtk_text_tag_new("link");
  ochusha_hypertext_tag_table_add_link_tag(table, tag);
  link_tag = tag;


  // Ѥѡ
  tag = gtk_text_tag_new("my-bookmark");
  g_object_set(tag,
	       "left-margin", 40,
	       "left-margin-set", TRUE,
	       NULL);
  ochusha_hypertext_tag_table_add_tag(table, tag);
  tag_table.insert("div::class=my-bookmark", tag);
  my_bookmark_tag = tag;


  html_styles.define_styles(ochusha_default_style_sheet,
			    sizeof(ochusha_default_style_sheet));

  read_prefs_css(controller.get_repository());

  Preferences &prefs = controller.get_prefs();
  set_colors(prefs);


  prefs.prefs_changed.connect(this, &HTMLRendererTagManager::prefs_changed_cb);
}


HTMLRendererTagManager::~HTMLRendererTagManager()
{
}


GtkTextTag *
HTMLRendererTagManager::create_tag(const char *name)
{
  OchushaHypertextTagTable *table
    = ochusha_hypertext_buffer_get_default_tag_table();

  GtkTextTag *tag = gtk_text_tag_new(name);
  ochusha_hypertext_tag_table_add_tag(table, tag);
  return tag;
}


void
HTMLRendererTagManager::insert_tag(const char *name, GtkTextTag *tag)
{
  OchushaHypertextTagTable *table
    = ochusha_hypertext_buffer_get_default_tag_table();
  ochusha_hypertext_tag_table_add_tag(table, tag);
  gint priority = gtk_text_tag_get_priority(link_tag);
  gtk_text_tag_set_priority(tag, priority);
  tag_table.insert(name, tag);
}


static void
output_color_setting(FILE *file, const char *name, GdkColor *fg, GdkColor *bg)
{
  if (fg == NULL && bg == NULL)
    return;

  if (fg != NULL && bg != NULL)
    fprintf(file,
	    "%s { color: #%02x%02x%02x; background-color: #%02x%02x%02x }\n",
	    name,
	    (fg->red >> 8) & 0xff, (fg->green >> 8) & 0xff,
	    (fg->blue >> 8) & 0xff,
	    (bg->red >> 8) & 0xff, (bg->green >> 8) & 0xff,
	    (bg->blue >> 8) & 0xff);
  else if (bg == NULL)
    fprintf(file, "%s { color: #%02x%02x%02x }\n",
	    name,
	    (fg->red >> 8) & 0xff, (fg->green >> 8) & 0xff,
	    (fg->blue >> 8) & 0xff);
  else
    fprintf(file, "%s { background-color: #%02x%02x%02x }\n",
	    name,
	    (bg->red >> 8) & 0xff, (bg->green >> 8) & 0xff,
	    (bg->blue >> 8) & 0xff);
}


void
HTMLRendererTagManager::write_prefs_css(Repository &repository)
{
  int fd = repository.open_file(OCHUSHA_PREFS_CSS,
				O_WRONLY | O_TRUNC | O_CREAT);
  Preferences &prefs = controller.get_prefs();
  if (fd >= 0)
    {
      FILE *file = fdopen(fd, "w");
      if (file != NULL)
	{
	  fprintf(file, "/* " OCHUSHA_PREFS_CSS " generated by "
		  PACKAGE_STRING ". */\n");
	  fprintf(file, "/* Do NOT Edit! */\n\n");
	  output_color_setting(file, "-ochusha-thread-title",
			       prefs.get_thread_title(), NULL);
	  output_color_setting(file, "-ochusha-name",
			       prefs.get_name(), NULL);
	  output_color_setting(file, "-ochusha-link",
			       prefs.get_link(), NULL);
	  output_color_setting(file, "-ochusha-visited",
			       prefs.get_visited(), NULL);
	  output_color_setting(file, "-ochusha-list-normal",
			       prefs.get_fg_normal(), prefs.get_bg_normal());
	  output_color_setting(file, "-ochusha-list-emph",
			       prefs.get_fg_emph(), prefs.get_bg_emph());
	  output_color_setting(file, "-ochusha-list-strong",
			       prefs.get_fg_strong(), prefs.get_bg_strong());
	  output_color_setting(file, "-ochusha-list-hidden",
			       prefs.get_fg_hidden(), prefs.get_bg_hidden());
	  output_color_setting(file, "-ochusha-auto-refresh-go",
			       prefs.get_auto_refresh_go(), NULL);
	  output_color_setting(file, "-ochusha-auto-refresh-wait",
			       prefs.get_auto_refresh_wait(), NULL);
	}
      fclose(file);
    }
  else
    close(fd);
}


void
HTMLRendererTagManager::read_prefs_css(Repository &repository)
{
  int fd = repository.open_file(OCHUSHA_PREFS_CSS, O_RDONLY);
  if (fd < 0)
    return;


  {
    BufferHandle buffer = new Buffer();
    buffer->map_file(fd);


    int lock_id = buffer->get_lock()->rdlock();

    const char *buf_top = buffer->get_buffer(lock_id);
    size_t buf_len = buffer->get_length(lock_id);
    html_styles.define_styles(buf_top, buf_len);

    buffer->get_lock()->unlock(lock_id);
  }


  Preferences &prefs = controller.get_prefs();
  StyleSheet &ss = html_styles.get_style_sheet();
  const StyleMatcher *matcher = ss.lookup_style(NULL, "-ochusha-thread-title",
						NULL, CSS_PSEUDO_CLASS_NO,
						CSS_PSEUDO_ELEMENT_NO);
  if (matcher != NULL)
    {
      const CSSStyle &style = matcher->get_style();
      parse_color(style.get_color(), &prefs.thread_title_color);
    }

  matcher = ss.lookup_style(NULL, "-ochusha-name",
			    NULL, CSS_PSEUDO_CLASS_NO, CSS_PSEUDO_ELEMENT_NO);
  if (matcher != NULL)
    {
      const CSSStyle &style = matcher->get_style();
      parse_color(style.get_color(), &prefs.name_color);
    }

  matcher = ss.lookup_style(NULL, "-ochusha-link",
			    NULL, CSS_PSEUDO_CLASS_NO, CSS_PSEUDO_ELEMENT_NO);
  if (matcher != NULL)
    {
      const CSSStyle &style = matcher->get_style();
      parse_color(style.get_color(), &prefs.link_color);
    }

  matcher = ss.lookup_style(NULL, "-ochusha-visited",
			    NULL, CSS_PSEUDO_CLASS_NO, CSS_PSEUDO_ELEMENT_NO);
  if (matcher != NULL)
    {
      const CSSStyle &style = matcher->get_style();
      parse_color(style.get_color(), &prefs.visited_color);
    }

  matcher = ss.lookup_style(NULL, "-ochusha-list-normal",
			    NULL, CSS_PSEUDO_CLASS_NO, CSS_PSEUDO_ELEMENT_NO);
  if (matcher != NULL)
    {
      const CSSStyle &style = matcher->get_style();
      if (parse_color(style.get_color(), &prefs.fg_normal_color))
	prefs.set_fg_normal_color_set(true);

      if (parse_color(style.get_background_color(), &prefs.bg_normal_color))
	prefs.set_bg_normal_color_set(true);
    }

  matcher = ss.lookup_style(NULL, "-ochusha-list-emph",
			    NULL, CSS_PSEUDO_CLASS_NO, CSS_PSEUDO_ELEMENT_NO);
  if (matcher != NULL)
    {
      const CSSStyle &style = matcher->get_style();
      if (parse_color(style.get_color(), &prefs.fg_emph_color))
	prefs.set_fg_emph_color_set(true);

      if (parse_color(style.get_background_color(), &prefs.bg_emph_color))
	prefs.set_bg_emph_color_set(true);
    }

  matcher = ss.lookup_style(NULL, "-ochusha-list-strong",
			    NULL, CSS_PSEUDO_CLASS_NO, CSS_PSEUDO_ELEMENT_NO);
  if (matcher != NULL)
    {
      const CSSStyle &style = matcher->get_style();
      if (parse_color(style.get_color(), &prefs.fg_strong_color))
	prefs.set_fg_strong_color_set(true);

      if (parse_color(style.get_background_color(), &prefs.bg_strong_color))
	prefs.set_bg_strong_color_set(true);
    }

  matcher = ss.lookup_style(NULL, "-ochusha-list-hidden",
			    NULL, CSS_PSEUDO_CLASS_NO, CSS_PSEUDO_ELEMENT_NO);
  if (matcher != NULL)
    {
      const CSSStyle &style = matcher->get_style();
      if (parse_color(style.get_color(), &prefs.fg_hidden_color))
	prefs.set_fg_hidden_color_set(true);

      if (parse_color(style.get_background_color(), &prefs.bg_hidden_color))
	prefs.set_bg_hidden_color_set(true);
    }

  matcher = ss.lookup_style(NULL, "-ochusha-auto-refresh-go",
			    NULL, CSS_PSEUDO_CLASS_NO, CSS_PSEUDO_ELEMENT_NO);
  if (matcher != NULL)
    {
      const CSSStyle &style = matcher->get_style();
      parse_color(style.get_color(), &prefs.auto_refresh_go_color);
    }

  matcher = ss.lookup_style(NULL, "-ochusha-auto-refresh-wait",
			    NULL, CSS_PSEUDO_CLASS_NO, CSS_PSEUDO_ELEMENT_NO);
  if (matcher != NULL)
    {
      const CSSStyle &style = matcher->get_style();
      parse_color(style.get_color(), &prefs.auto_refresh_wait_color);
    }
}


static bool
set_color_rgb8(guint16 red, guint16 green, guint16 blue, GdkColor *color)
{
  assert((red & ~(0xff)) == 0);
  assert((green & ~(0xff)) == 0);
  assert((blue & ~(0xff)) == 0);
  color->pixel = 0;
  color->red = (red << 8) | red;
  color->green = (green << 8) | green;
  color->blue = (blue << 8) | blue;
  return true;
}


bool
HTMLRendererTagManager::parse_color(const CSSValue &value, GdkColor *color)
{
  if (!value.is_value_set())
    return false;

  if (value.is_keyword())
    {
      CSSKeyword keyword = value.get_keyword();
      if (keyword == css_black)
	return set_color_rgb8(0, 0, 0, color);
      if (keyword == css_silver)
	return set_color_rgb8(0xc0, 0xc0, 0xc0, color);
      if (keyword == css_gray)
	return set_color_rgb8(0x80, 0x80, 0x80, color);
      if (keyword == css_white)
	return set_color_rgb8(0xff, 0xff, 0xff, color);
      if (keyword == css_maroon)
	return set_color_rgb8(0x80, 0, 0, color);
      if (keyword == css_red)
	return set_color_rgb8(0xff, 0, 0, color);
      if (keyword == css_purple)
	return set_color_rgb8(0x80, 0, 0x80, color);
      if (keyword == css_fuchsia)
	return set_color_rgb8(0xff, 0, 0xff, color);
      if (keyword == css_green)
	return set_color_rgb8(0, 0x80, 0, color);
      if (keyword == css_lime)
	return set_color_rgb8(0, 0xff, 0, color);
      if (keyword == css_olive)
	return set_color_rgb8(0x80, 0x80, 0, color);
      if (keyword == css_yellow)
	return set_color_rgb8(0xff, 0xff, 0, color);
      if (keyword == css_navy)
	return set_color_rgb8(0, 0, 0x80, color);
      if (keyword == css_blue)
	return set_color_rgb8(0, 0, 0xff, color);
      if (keyword == css_teal)
	return set_color_rgb8(0, 0x80, 0x80, color);
      if (keyword == css_aqua)
	return set_color_rgb8(0, 0xff, 0xff, color);
      if (keyword == css_orange)
	return set_color_rgb8(0xff, 0xa5, 0, color);
      return gdk_color_parse(reinterpret_cast<const gchar *>(keyword), color);
    }

  if (value.is_color_rgb())
    return set_color_rgb8(value.get_color_rgb_red(),
			  value.get_color_rgb_green(),
			  value.get_color_rgb_blue(),
			  color);

  if (value.is_color_string())
    {
      const char *val = value.get_color_string();

      if (strcmp(val, "-ochusha-link-color") == 0)
	{
	  Preferences &prefs = controller.get_prefs();
	  GdkColor *c = prefs.get_link();
	  if (c == NULL || c == color)
	    return false;
	  color->pixel = c->pixel;
	  color->red = c->red;
	  color->green = c->green;
	  color->blue = c->blue;
	  return true;
	}
      if (strcmp(val, "-ochusha-visited-color") == 0)
	{
	  Preferences &prefs = controller.get_prefs();
	  GdkColor *c = prefs.get_visited();
	  if (c == NULL || c == color)
	    return false;
	  color->pixel = c->pixel;
	  color->red = c->red;
	  color->green = c->green;
	  color->blue = c->blue;
	  return true;
	}
      if (strcmp(val, "-ochusha-thread-title-color") == 0)
	{
	  Preferences &prefs = controller.get_prefs();
	  GdkColor *c = prefs.get_thread_title();
	  if (c == NULL || c == color)
	    return false;
	  color->pixel = c->pixel;
	  color->red = c->red;
	  color->green = c->green;
	  color->blue = c->blue;
	  return true;
	}
      if (strcmp(val, "-ochusha-name-color") == 0)
	{
	  Preferences &prefs = controller.get_prefs();
	  GdkColor *c = prefs.get_name();
	  if (c == NULL || c == color)
	    return false;
	  color->pixel = c->pixel;
	  color->red = c->red;
	  color->green = c->green;
	  color->blue = c->blue;
	  return true;
	}
      if (strcmp(val, "-ochusha-text-color") == 0)
	return false;

      return gdk_color_parse(reinterpret_cast<const gchar *>(val), color);
    }

  return false;
}


GtkTextTag *
HTMLRendererTagManager::set_foreground(GtkTextTag *tag, const CSSValue &value)
{
  if (!value.is_value_set())
    {
      g_object_set(tag, "foreground-set", FALSE, NULL);
      return tag;
    }

  if (value.is_color_string())
    {
      const char *val = value.get_color_string();

      if (strcmp(val, "-ochusha-link-color") == 0)
	{
	  Preferences &prefs = controller.get_prefs();
	  GdkColor *c = prefs.get_link();

	  g_object_set(tag,
		       "foreground-gdk", c,
		       "foreground-set", c != NULL,
		       NULL);

	  vector<GtkTextTag *>::iterator iter = link_color_tags.begin();
	  vector<GtkTextTag *>::iterator end_iter = link_color_tags.end();
	  while (iter != end_iter)
	    {
	      if (*iter == tag)
		break;
	      ++iter;
	    }

	  if (iter == end_iter)
	    link_color_tags.push_back(tag);

	  return tag;
	}
      if (strcmp(val, "-ochusha-visited-color") == 0)
	{
	  Preferences &prefs = controller.get_prefs();
	  GdkColor *c = prefs.get_visited();

	  g_object_set(tag,
		       "foreground-gdk", c,
		       "foreground-set", c != NULL,
		       NULL);

	  vector<GtkTextTag *>::iterator iter = visited_color_tags.begin();
	  vector<GtkTextTag *>::iterator end_iter = visited_color_tags.end();
	  while (iter != end_iter)
	    {
	      if (*iter == tag)
		break;
	      ++iter;
	    }

	  if (iter == end_iter)
	    visited_color_tags.push_back(tag);

	  return tag;
	}
      if (strcmp(val, "-ochusha-thread-title-color") == 0)
	{
	  Preferences &prefs = controller.get_prefs();
	  GdkColor *c = prefs.get_thread_title();

	  g_object_set(tag,
		       "foreground-gdk", c,
		       "foreground-set", c != NULL,
		       NULL);

	  vector<GtkTextTag *>::iterator iter = title_color_tags.begin();
	  vector<GtkTextTag *>::iterator end_iter = title_color_tags.end();
	  while (iter != end_iter)
	    {
	      if (*iter == tag)
		break;
	      ++iter;
	    }

	  if (iter == end_iter)
	    title_color_tags.push_back(tag);

	  return tag;
	}
      if (strcmp(val, "-ochusha-name-color") == 0)
	{
	  Preferences &prefs = controller.get_prefs();
	  GdkColor *c = prefs.get_name();

	  g_object_set(tag,
		       "foreground-gdk", c,
		       "foreground-set", c != NULL,
		       NULL);

	  vector<GtkTextTag *>::iterator iter = name_color_tags.begin();
	  vector<GtkTextTag *>::iterator end_iter = name_color_tags.end();
	  while (iter != end_iter)
	    {
	      if (*iter == tag)
		break;
	      ++iter;
	    }

	  if (iter == end_iter)
	    name_color_tags.push_back(tag);

	  return tag;
	}
      if (strcmp(val, "-ochusha-text-color") == 0)
	{
	  g_object_set(tag, "foreground-set", FALSE, NULL);
	  return tag;
	}
    }

  GdkColor color = { 0, 0, 0, 0 };
  if (parse_color(value, &color))
    g_object_set(tag,
		 "foreground-gdk", &color,
		 "foreground-set", TRUE,
		 NULL);
  else
    g_object_set(tag, "foreground-set", FALSE, NULL);

  return tag;
}


GtkTextTag *
HTMLRendererTagManager::set_background(GtkTextTag *tag, const CSSValue &value)
{
  GdkColor color = { 0, 0, 0, 0 };
  if (parse_color(value, &color))
    g_object_set(tag,
		 "background-gdk", &color,
		 "background-set", TRUE,
		 NULL);
  else
    g_object_set(tag, "background-set", FALSE, NULL);

  return tag;
}


GtkTextTag *
HTMLRendererTagManager::set_paragraph_background(GtkTextTag *tag,
						 const CSSValue &value)
{
  if (gtk_major_version <= 2 && gtk_minor_version < 8)
    return tag;

  GdkColor color = { 0, 0, 0, 0 };
  if (parse_color(value, &color))
    g_object_set(tag,
		 "paragraph-background-gdk", &color,
		 "paragraph-background-set", TRUE,
		 NULL);
  else
    g_object_set(tag, "paragraph-background-set", FALSE, NULL);

  return tag;
}


void
HTMLRendererTagManager::set_colors(Preferences &prefs)
{
  // ι
  if (!title_color_tags.empty())
    {
      GdkColor *color = prefs.get_thread_title();
      gboolean is_set = color != NULL;
      vector<GtkTextTag *>::iterator iter = title_color_tags.begin();
      vector<GtkTextTag *>::iterator end_iter = title_color_tags.end();
      while (iter != end_iter)
	g_object_set(*iter++,
		     "foreground-gdk", color,
		     "foreground-set", is_set,
		     NULL);
    }

  if (!name_color_tags.empty())
    {
      GdkColor *color = prefs.get_name();
      gboolean is_set = color != NULL;
      vector<GtkTextTag *>::iterator iter = name_color_tags.begin();
      vector<GtkTextTag *>::iterator end_iter = name_color_tags.end();
      while (iter != end_iter)
	g_object_set(*iter++,
		     "foreground-gdk", color,
		     "foreground-set", is_set,
		     NULL);
    }

  if (!link_color_tags.empty())
    {
      GdkColor *color = prefs.get_link();
      gboolean is_set = color != NULL;
      vector<GtkTextTag *>::iterator iter = link_color_tags.begin();
      vector<GtkTextTag *>::iterator end_iter = link_color_tags.end();
      while (iter != end_iter)
	g_object_set(*iter++,
		     "foreground-gdk", color,
		     "foreground-set", is_set,
		     NULL);
    }

  if (!visited_color_tags.empty())
    {
      GdkColor *color = prefs.get_visited();
      gboolean is_set = color != NULL;
      vector<GtkTextTag *>::iterator iter = visited_color_tags.begin();
      vector<GtkTextTag *>::iterator end_iter = visited_color_tags.end();
      while (iter != end_iter)
	g_object_set(*iter++,
		     "foreground-gdk", color,
		     "foreground-set", is_set,
		     NULL);
    }
}


void
HTMLRendererTagManager::prefs_changed_cb(Preferences &prefs)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  set_colors(prefs);
}


HTMLElementManager HTMLRenderer::html_elements;


HTMLRenderer::HTMLRenderer(Controller &controller,
			   OchushaHypertextBuffer *buffer,
			   OchushaHypertextView *view)
  : controller(controller),
    tag_manager(controller.get_tag_manager()),
    html_styles(tag_manager.get_html_styles()),
    buffer(buffer),
    view(view),
    canonicalizer(canon_buf),
    link_tag(NULL),
    message_id(0),
    last_event_link(NULL),
    at_beginning_of_line(true),
    has_leading_space(false)
{
#if 0
  // ɽΥ󥯤֤äƽƤޤΤǡöα
  // ʥϥɥvirtualؿФ褤
  g_signal_connect(view, "link_mouse_over",
		   reinterpret_cast<GCallback>(link_mouse_over_cb), this);
  g_signal_connect(view, "link_mouse_out",
		   reinterpret_cast<GCallback>(link_mouse_out_cb), this);
  g_signal_connect(view, "link_mouse_press",
		   reinterpret_cast<GCallback>(link_mouse_press_cb), this);
  g_signal_connect(view, "link_mouse_release",
		   reinterpret_cast<GCallback>(link_mouse_release_cb), this);
#endif
}


HTMLRenderer::~HTMLRenderer()
{
  if (message_id != 0)
    {
      GtkStatusbar *bar = controller.get_statusbar();
      gtk_statusbar_remove(bar, controller.get_link_context_id(),
			   message_id);
      message_id = 0;
    }

  if (last_event_link != NULL)
    {
      free(last_event_link);
      last_event_link = NULL;
    }
}


gboolean
HTMLRenderer::link_mouse_over_cb(OchushaHypertextView *widget,
				 GdkEventMotion *event,
				 const gchar *link,
				 HTMLRenderer *renderer)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  return renderer->link_mouse_over(widget, event, link);
}


gboolean
HTMLRenderer::link_mouse_out_cb(OchushaHypertextView *widget,
				GdkEventMotion *event,
				HTMLRenderer *renderer)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  return renderer->link_mouse_out(widget, event);
}


gboolean
HTMLRenderer::link_mouse_press_cb(OchushaHypertextView *widget,
				  GdkEventButton *event,
				  const gchar *link,
				  HTMLRenderer *renderer)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  return renderer->link_mouse_press(widget, event, link);
}


gboolean
HTMLRenderer::link_mouse_release_cb(OchushaHypertextView *widget,
				    GdkEventButton *event,
				    const gchar *link,
				    HTMLRenderer *renderer)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  return renderer->link_mouse_release(widget, event, link);
}


gboolean
HTMLRenderer::link_mouse_over(OchushaHypertextView *widget,
			      GdkEventMotion *event,
			      const gchar *link)
{
  GtkStatusbar *bar = controller.get_statusbar();
  if (message_id != 0)
    {
      gtk_statusbar_remove(bar, controller.get_link_context_id(), message_id);
      message_id = 0;
    }

  message_id = gtk_statusbar_push(bar, controller.get_link_context_id(), link);

  return FALSE;
}


gboolean
HTMLRenderer::link_mouse_out(OchushaHypertextView *widget,
			     GdkEventMotion *event)
{
  if (message_id != 0)
    {
      GtkStatusbar *bar = controller.get_statusbar();
      gtk_statusbar_remove(bar, controller.get_link_context_id(), message_id);
      message_id = 0;
    }

  return FALSE;
}


gboolean
HTMLRenderer::link_mouse_press(OchushaHypertextView *widget,
			       GdkEventButton *event,
			       const gchar *link)
{
  g_return_val_if_fail(link != NULL, FALSE);

  if (last_event_link != NULL)
    {
      if (strcmp(link, last_event_link) == 0)
	return FALSE;
      free(last_event_link);
      last_event_link = NULL;
    }

  last_event_link = strdup(link);

  return TRUE;
}


gboolean
HTMLRenderer::link_mouse_release(OchushaHypertextView *widget,
				 GdkEventButton *event,
				 const gchar *link)
{
  char *last_link = last_event_link;
  if (last_link == NULL)
    return FALSE;

  last_event_link = NULL;

  if (strcmp(last_link, link) == 0 && event->button != 3)
    {
      Preferences &prefs = controller.get_prefs();
      bool in_tab = (event->button == 2) ^ prefs.get_default_open_in_tab();
      controller.open(link, true, in_tab, false, event->button == 1);
    }

  free(last_link);

  return TRUE;
}


void
HTMLRenderer::start_element(const char *name, const char *const attrs[])
{
  const HTMLElement *element = html_elements.lookup(name);
  if ((!at_beginning_of_line && element != NULL && element->is_block_element())
      || strcasecmp(name, "br") == 0)
    flush_pcdata(true);
  else
    flush_pcdata(false);

  if (element != NULL)
    {
      while (!context_stack.empty())
	{
	  if (context_stack.back()->get_element()->is_containable(element))
	    break;
	  context_stack.pop_back();
	}

      const HTMLStyle *parent;
      CSSPseudoClass pseudo_class = CSS_PSEUDO_CLASS_NO;
      if (!context_stack.empty())
	parent = context_stack.back();
      else
	parent = NULL;

      if (strcasecmp(name, "a") == 0)
	{
	  int i = 0;
	  const char *link = NULL;
	  const char *klass = NULL;
	  while (attrs[i * 2] != NULL)
	    {
	      if (strcasecmp(attrs[i * 2], "href") == 0)
		{
		  link = attrs[i * 2 + 1];
		  canonicalizer.clear();
		  if (strncmp(link, "http", 4) == 0
		      || strncmp(link, "https", 5) == 0
		      || strncmp(link, "ftp", 3) == 0)
		    link = canonicalizer.canon_url(link, -1);
		  else
		    link = canonicalizer.canon(link, -1);
		}
	      else if (strcasecmp(attrs[i * 2], "class") == 0)
		klass = attrs[i * 2 + 1];	// 괺ad-hocн
	      i++;
	    }

	  if (link != NULL)
	    {
	      if (!is_visited_link(link))
		pseudo_class = CSS_PSEUDO_CLASS_LINK;
	      else
		pseudo_class = CSS_PSEUDO_CLASS_VISITED;
	    }
	  else
	    pseudo_class = CSS_PSEUDO_CLASS_NO;
	}

      const HTMLStyle *style
	= html_styles.lookup_style(parent, name, element, attrs, pseudo_class,
				   CSS_PSEUDO_ELEMENT_NO);
      context_stack.push_back(style);
    }

  if (strcasecmp(name, "a") == 0)
    {
      int i = 0;
      const char *link = NULL;
      const char *klass = NULL;
      while (attrs[i * 2] != NULL)
	{
	  if (strcasecmp(attrs[i * 2], "href") == 0)
	    {
	      link = attrs[i * 2 + 1];
	      canonicalizer.clear();
	      if (strncmp(link, "http", 4) == 0
		  || strncmp(link, "https", 5) == 0
		  || strncmp(link, "ftp", 3) == 0)
		link = canonicalizer.canon_url(link, -1);
	      else
		link = canonicalizer.canon(link, -1);
	    }
	  else if (strcasecmp(attrs[i * 2], "class") == 0)
	    klass = attrs[i * 2 + 1];	// 괺ad-hocн
	  i++;
	}

      link_tag = tag_manager.get_link_tag();
      ochusha_hypertext_buffer_push_link_tag(buffer, link_tag, link);
      return;
    }

  if (strcasecmp(name, "img") == 0)
    {
      int i = 0;
      int width = 160;
      int height = 120;
      int hborder = 0;
      int vborder = 0;
      const char *url = NULL;
      while (attrs[i * 2] != NULL)
	{
	  if (strcasecmp(attrs[i * 2], "src") == 0)
	    {
	      canonicalizer.clear();
	      url = canonicalizer.canon_url(attrs[i * 2 + 1], -1);
	    }
	  else if (strcasecmp(attrs[i * 2], "width") == 0)
	    {
	      if (sscanf(attrs[i * 2 + 1], "%d", &width) != 1)
		width = 160;
	    }
	  else if (strcasecmp(attrs[i * 2], "height") == 0)
	    {
	      if (sscanf(attrs[i * 2 + 1], "%d", &height) != 1)
		height = 120;
	    }
	  else if (strcasecmp(attrs[i * 2], "border") == 0)
	    {
	      int border;
	      if (sscanf(attrs[i * 2 + 1], "%d", &border) == 1)
		hborder = vborder = border;
	    }
	  else if (strcasecmp(attrs[i * 2], "hspace") == 0)
	    {
	      if (sscanf(attrs[i * 2 + 1], "%d", &hborder) != 1)
		hborder = 0;
	    }
	  else if (strcasecmp(attrs[i * 2], "vspace") == 0)
	    {
	      if (sscanf(attrs[i * 2 + 1], "%d", &vborder) != 1)
		vborder = 0;
	    }
	  ++i;
	}

      if (url != NULL)
	{
	  GtkWidget *image
	    = controller.get_image_manager().get_widget(url, width, height);
	  if (image != NULL)
	    {
	      GtkTextTag *tag;
	      if (!context_stack.empty())
		{
		  HTMLStyle *style
		    = const_cast<HTMLStyle *>(context_stack.back());
		  tag = style->get_gtk_text_tag(tag_manager,
					reinterpret_cast<GtkWidget *>(view));
		}
	      else
		tag = NULL;
	      if (tag != NULL)
		ochusha_hypertext_buffer_push_tag(buffer, tag);

	      ochusha_hypertext_view_append_widget(view, image,
						   hborder, vborder);

	      if (tag != NULL)
		ochusha_hypertext_buffer_pop_tags(buffer, tag);
	    }
	}

      return;
    }
}


void
HTMLRenderer::end_element(const char *name)
{
  const HTMLElement *element = html_elements.lookup(name);
  if (element != NULL && element->is_block_element())
    flush_pcdata(true);
  else
    flush_pcdata(false);

  if (element != NULL)
    {
      if (element == context_stack.back()->get_element())
	context_stack.pop_back();
      else
	{
	  vector<const HTMLStyle *>::const_iterator iter = context_stack.begin();
	  vector<const HTMLStyle *>::const_iterator end_iter = context_stack.end();
	  while (iter != end_iter)
	    {
	      const HTMLStyle *ctx = *iter;
	      if (ctx->get_element() == element)
		break;
	      ++iter;
	    }

	  if (iter != end_iter)
	    {
	      while (!context_stack.empty())
		{
		  const HTMLStyle *ctx = context_stack.back();
		  context_stack.pop_back();
		  if (ctx->get_element() == element)
		    break;
		}
	    }
	}
    }

  if (strcasecmp(name, "title") == 0)
    {
      GtkWidget *toplevel = gtk_widget_get_toplevel(GTK_WIDGET(view));
#if GTK_CHECK_VERSION(2,18,0)
      if (toplevel != NULL
	  && gtk_widget_is_toplevel(toplevel) && GTK_IS_WINDOW(toplevel))
	gtk_window_set_title(GTK_WINDOW(toplevel), title_text.c_str());
#else
      if (toplevel != NULL
	  && GTK_WIDGET_TOPLEVEL(toplevel) && GTK_IS_WINDOW(toplevel))
	gtk_window_set_title(GTK_WINDOW(toplevel), title_text.c_str());
#endif

      return;
    }

  if (strcasecmp(name, "a") == 0 && link_tag != NULL)
    {
      ochusha_hypertext_buffer_pop_link_tag(buffer, link_tag);
      link_tag = NULL;
      return;
    }
}


void
HTMLRenderer::flush_pcdata(bool with_newline)
{
  if (with_newline)
    {
      if (has_leading_space)
	pcdata[pcdata.length() - 1] = '\n';
      else
	pcdata.push_back('\n');
      at_beginning_of_line = true;
      has_leading_space = false;
    }

  if (pcdata.empty())
    return;

  // ǸߤΥƥȤΥ˹礦GtkTextTagѰդ
  // Ȥäbuffer˽񤭹ࡣ
  if (context_stack.empty()
      || (context_stack.size() >= 2
	  && (strcasecmp(context_stack[1]->get_element()->get_name(), "head")
	      != 0)))
    {
      // Ѥ餺ġ
      HTMLStyle *style = const_cast<HTMLStyle *>(context_stack.back());
      GtkTextTag *tag
	= style->get_gtk_text_tag(tag_manager,
				  reinterpret_cast<GtkWidget *>(view));

      ochusha_hypertext_buffer_append_text_with_tag(buffer, pcdata.c_str(),
						    pcdata.length(), tag);
    }
  else if (strcasecmp(context_stack.back()->get_element()->get_name(), "title")
	   == 0)
    {
      title_text.append(pcdata);
    }

  pcdata.clear();
}


void
HTMLRenderer::characters(const char *ch, size_t len)
{
  while (!context_stack.empty()
	 && context_stack.back()->get_element()->is_empty_element())
    context_stack.pop_back();

  const char *cur_pos = ch;
  const char *tail_pos = cur_pos + len;
  bool at_bol = at_beginning_of_line;
  bool has_ls = has_leading_space;
  while (cur_pos < tail_pos)
    {
      char c = *cur_pos++;
      if (c != ' ' && c != '\t' && c != '\n' && c != '\r')
	{
	  pcdata.push_back(c);
	  at_bol = false;
	  has_ls = false;
	}
      else if (!at_bol && !has_ls)
	{
	  pcdata.push_back(' ');
	  has_ls = true;
	}
    }
  at_beginning_of_line = at_bol;
  has_leading_space = has_ls;
}


bool
HTMLRenderer::is_visited_link(const char *link)
{
  return false;	// ޤɤ
}


void
HTMLRenderer::render_html(const char *html)
{
  parse_text(html, strlen(html));
}

}
