/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: image_access_db.cc,v 1.2 2009/01/01 07:47:06 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "image_access_db.h"

#include "libochusha/lock.h"
#include "libochusha/misc_utils.h"
#include "libochusha/repository.h"
#include "libochusha/sqlite_utils.h"

#include "ui_constants.h"

#include <cassert>
#include <climits>
#include <cstddef>
#include <cstdlib>
#include <cstdio>
#include <cstring>

#include <vector>


namespace Ochusha
{


AccessResult::AccessResult(const char *url, ResultType result, int status_code,
			   time_t access_time)
  : url(strdup(url)),
    result(result),
    status_code(status_code),
    last_accessed(access_time)
{
}


AccessResult::~AccessResult()
{
  free(url);
}


void
AccessResult::set_url(const char *value)
{
  if (url != NULL)
    free(url);

  url = value != NULL ? strdup(value) : NULL;
}


#define OCHUSHA_IMAGE_ACCESSES_TABLE_DEFINITION "CREATE TABLE IF NOT EXISTS ochusha_image_accesses (id INTEGER PRIMARY KEY, url TEXT, result INTEGER, statusCode INTEGER, lastAccessed INTEGER);"

#define COLUMN_ID		0
#define COLUMN_URL		1
#define COLUMN_RESULT		2
#define COLUMN_STATUS_CODE	3
#define COLUMN_LAST_ACCESSED	4

#define OCHUSHA_IMAGE_ACCESSES_DELETE_ALL		"DELETE FROM ochusha_image_accesses;"

#define INDEX_ID		1
#define INDEX_URL		2
#define INDEX_RESULT		3
#define INDEX_STATUS_CODE	4
#define INDEX_LAST_ACCESSED	5


#define OCHUSHA_IMAGE_ACCESSES_SELECT_ALL		"SELECT * FROM ochusha_image_accesses;"
#define OCHUSHA_IMAGE_ACCESSES_SELECT_MATCHED_URL	"SELECT * FROM ochusha_image_accesses WHERE url = ?2;"
#define OCHUSHA_IMAGE_ACCESSES_INSERT_NEW_URL		"INSERT INTO ochusha_image_accesses (url, result, statusCode, lastAccessed) VALUES (?2, ?3, ?4, ?5);"
#define OCHUSHA_IMAGE_ACCESSES_UPDATE_MATCHED_ID	"UPDATE ochusha_image_accesses SET result = ?3, statusCode = ?4, lastAccessed = ?5 WHERE id = ?1;"
#define OCHUSHA_IMAGE_ACCESSES_DELETE_MATCHED_ID	"DELETE FROM ochusha_image_accesses WHERE id = ?1;"




ImageAccessDB::ImageAccessDB(Repository &repository)
  : lock("ImageAccessDB")
{
  char pathname[PATH_MAX];

  if (repository.expand_path(OCHUSHA_IMAGE_ACCESSES_SQLITE, pathname, PATH_MAX))
    db_handle = new SQLite3(pathname);
  else
    db_handle = new SQLite3("");

  // table
  db_handle->get_table(OCHUSHA_IMAGE_ACCESSES_TABLE_DEFINITION);

  // prepared statementν
  select_all_urls
    = db_handle->statement_new(OCHUSHA_IMAGE_ACCESSES_SELECT_ALL);

  select_matched_url
    = db_handle->statement_new(OCHUSHA_IMAGE_ACCESSES_SELECT_MATCHED_URL);

  insert_new_url
    = db_handle->statement_new(OCHUSHA_IMAGE_ACCESSES_INSERT_NEW_URL);

  update_matched_id
    = db_handle->statement_new(OCHUSHA_IMAGE_ACCESSES_UPDATE_MATCHED_ID);

  delete_matched_id
    = db_handle->statement_new(OCHUSHA_IMAGE_ACCESSES_DELETE_MATCHED_ID);
}


ImageAccessDB::~ImageAccessDB()
{
  delete delete_matched_id;
  delete update_matched_id;
  delete insert_new_url;
  delete select_matched_url;
  delete select_all_urls;

  delete db_handle;
}


void
ImageAccessDB::register_access_result(const char *url,
				      AccessResult::ResultType result,
				      int status_code, time_t access_time)
{
  assert(url != NULL);

  vector<sqlite3_int64> duplicated_entries;

  LockBlock lock_obj(lock);

  select_matched_url->clear_bindings();
  select_matched_url->reset();

  sqlite3_int64 row_id = 0;
  int db_result = SQLITE_DONE;
  if (select_matched_url->bind_static(INDEX_URL, url))
    {
      while ((db_result = select_matched_url->step()) == SQLITE_ROW)
	{
	  if (row_id != 0)
	    duplicated_entries.push_back(row_id);
	  row_id = select_matched_url->get_column_int64(COLUMN_ID);
	}
    }

  select_matched_url->clear_bindings();
  select_matched_url->reset();

  if (!duplicated_entries.empty())
    {
      duplicated_entries.push_back(row_id);
      row_id = 0;

      delete_matched_id->clear_bindings();
      delete_matched_id->reset();

      while (!duplicated_entries.empty())
	{
	  if (delete_matched_id->bind(INDEX_ID, duplicated_entries.back()))
	    {
	      int del_result;
	      do
		{
		  del_result = delete_matched_id->step();
		}
	      while (del_result == SQLITE_ROW);
	    }

	  duplicated_entries.pop_back();
	}

      delete_matched_id->clear_bindings();
      delete_matched_id->reset();
    }

  if (db_result == SQLITE_DONE)
    {
      SQLite3Statement *stmt;
      bool bindings_ok = true;

      if (row_id == 0)
	{
	  // INSERT
	  stmt = insert_new_url;
	  stmt->clear_bindings();
	  stmt->reset();

	  bindings_ok = stmt->bind_static(INDEX_URL, url);
	}
      else
	{
	  // UPDATE
	  stmt = update_matched_id;
	  stmt->clear_bindings();
	  stmt->reset();

	  bindings_ok = stmt->bind(INDEX_ID, row_id);
	}

      if (bindings_ok
	  && stmt->bind(INDEX_RESULT, static_cast<int>(result))
	  && stmt->bind(INDEX_STATUS_CODE, status_code)
	  && stmt->bind(INDEX_LAST_ACCESSED,
			static_cast<sqlite3_int64>(access_time)))
	stmt->step();

      stmt->clear_bindings();
      stmt->reset();
    }
}


AccessResult *
ImageAccessDB::query_access_result(const char *url, AccessResult &result)
{
  assert(url != NULL);

  result.set_url(url);

  LockBlock lock_obj(lock);

  select_matched_url->clear_bindings();
  select_matched_url->reset();

  if (select_matched_url->bind_static(INDEX_URL, url)
      && select_matched_url->step() == SQLITE_ROW)
    {
      result.set_result(static_cast<AccessResult::ResultType>(select_matched_url->get_column_int(COLUMN_RESULT)));
      result.set_status_code(select_matched_url->get_column_int(COLUMN_STATUS_CODE));
      result.set_last_accessed(select_matched_url->get_column_int64(COLUMN_LAST_ACCESSED));

      return &result;
    }

  result.set_result(AccessResult::UNKNOWN);

  return NULL;
}


}
