/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: image_loader.cc,v 1.14 2009/01/01 07:47:06 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "image_loader.h"

#include "libochusha/lock.h"
#include "libochusha/misc_utils.h"
#include "libochusha/network_agent.h"
#include "libochusha/text_utils.h"

#include "ui_constants.h"

#include "composite_icon.h"
#include "image_manager.h"

#include "libochushagtk/stock.h"

#include <algorithm>
#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <cstring>


#ifdef HAVE_TIME_H
#include <time.h>
#endif


// #define DEBUG_IMAGE_LOADER_MOST	1
// #define VERBOSE_DESTRUCTOR	1


#ifdef DEBUG_IMAGE_LOADER_MOST
# ifdef HAVE_UNISTD_H
#  include <unistd.h>
# endif
#endif


#ifdef HAVE_MMAP

# ifdef HAVE_SYS_MMAN_H
#  include <sys/mman.h>
# endif

# ifndef MAP_NOCORE
#  define MAP_NOCORE 0
# endif

# ifndef MAP_NOSYNC
#  define MAP_NOSYNC 0
# endif

#endif	// HAVE_MMAP

#include <gtk/gtk.h>
#include <gdk-pixbuf/gdk-pixbuf.h>


namespace Ochusha
{


class ImageLoader
{
private:
  ImageLoaderInfo &info;
  GdkPixbufLoader *loader;
  GdkPixbufLoader *loader_rem;

  enum {
    LOADER_STATE_INITIAL = 0,
    LOADER_STATE_WAITING,
    LOADER_STATE_STARTED,
    LOADER_STATE_CONNECTED,
    LOADER_STATE_SIZE_PREPARED,
    LOADER_STATE_AREA_PREPARED,
    LOADER_STATE_LOADING,
    LOADER_STATE_DONE,
    LOADER_STATE_SCALING_ERROR,
    LOADER_STATE_FAILURE,
    LOADER_STATE_A_BONED
  } state;

  int width;
  int height;

  GdkPixbufAnimation *animation;
  GdkPixbuf *pixbuf;

  list<GtkImage *> image_list;

  guint timeout_id;
  gulong size_prepared_id;
  gulong area_prepared_id;
  gulong area_updated_id;
  gulong closed_id;

  char *description;

  bool loader_closed;

  void unref_loader();
  void update_pixbuf(GtkImage *image);
  void update_pixbuf();


  static gboolean timeout_cb(ImageLoader *loader);
  static void destroy_cb(GtkWidget *widget, ImageLoader *loader);

  static void size_prepared_cb(GdkPixbufLoader *pixbuf_loader, int width,
			       int height, ImageLoader *loader);
  static void area_prepared_cb(GdkPixbufLoader *pixbuf_loader,
			       ImageLoader *loader);
  static void area_updated_cb(GdkPixbufLoader *pixbuf_loader, int x, int y,
			      int width, int height, ImageLoader *loader);
  static void closed_cb(GdkPixbufLoader *pixbuf_loader, ImageLoader *loader);


public:
  ImageLoader(ImageLoaderInfo &info, int width, int height);
  ~ImageLoader();

  inline ImageLoaderInfo *get_info() { return &info; }
  inline bool match_size(int w, int h) const
  { return width == w && height == h; }
  inline int get_state() const { return state; }
  inline bool empty() const { return image_list.empty(); }
  inline bool fixed() const { return state >= LOADER_STATE_DONE; }
  inline GdkPixbufLoader *get_loader_rem() const { return loader_rem; }

  void notify_waiting();
  void notify_started();
  void notify_connected();
  void notify_error();
  void notify_a_boned();

  GtkWidget *get_widget(bool interactive);

  const char *to_string();	// ǥХå
};


ImageLoader::ImageLoader(ImageLoaderInfo &info, int width, int height)
  : info(info),
    loader(NULL),
    loader_rem(NULL),
    state(LOADER_STATE_INITIAL),
    width(width),
    height(height),
    animation(NULL),
    pixbuf(NULL),
    timeout_id(0),
    size_prepared_id(0),
    area_prepared_id(0),
    area_updated_id(0),
    closed_id(0),
    description(NULL),
    loader_closed(true)
{
}


ImageLoader::~ImageLoader()
{
  assert(image_list.empty());
  assert(OCHUSHA_IN_GDK_THREADS());

#ifdef VERBOSE_DESTRUCTOR
  fprintf(stderr, "ImageLoader::~ImageLoader(): %s, state=%d\n",
	  to_string(), state);
#endif

  if (timeout_id != 0)
    g_source_remove(timeout_id);

  unref_loader();

  if (animation != NULL)
    g_object_unref(animation);

  if (pixbuf != NULL)
    g_object_unref(pixbuf);

  if (description != NULL)
    free(description);
}


void
ImageLoader::unref_loader()
{
  assert(OCHUSHA_IN_GDK_THREADS());

  if (loader != NULL)
    {
      g_signal_handler_disconnect(loader, size_prepared_id);
      size_prepared_id = 0;

      g_signal_handler_disconnect(loader, area_prepared_id);
      area_prepared_id = 0;

      g_signal_handler_disconnect(loader, area_updated_id);
      area_updated_id = 0;

      g_signal_handler_disconnect(loader, closed_id);
      closed_id = 0;

      GdkPixbufLoader *pixbuf_loader = loader;
      loader = NULL;

      g_object_unref(pixbuf_loader);
    }
}


const char *
ImageLoader::to_string()
{
  if (description != NULL)
    return description;

  char buf[1024];
  snprintf(buf, 1024, "ImageLoader(@%p), %dx%d, url=%s",
	   this, width, height, info.get_url());
  description = strdup(buf);

  return description;
}


void
ImageLoader::update_pixbuf(GtkImage *image)
{
  g_return_if_fail(GTK_IS_IMAGE(image));
  assert(OCHUSHA_IN_GDK_THREADS());

  switch (state)
    {
    case LOADER_STATE_INITIAL:
    case LOADER_STATE_WAITING:
      {
	CompositeIcon *icon
	  = CompositeIcon::get_composite_icon(GTK_WIDGET(image));
	g_return_if_fail(icon != NULL);

	gtk_image_set_from_animation(image, icon->get_wait_animation());
	break;
      }

    case LOADER_STATE_STARTED:
      {
	CompositeIcon *icon
	  = CompositeIcon::get_composite_icon(GTK_WIDGET(image));
	g_return_if_fail(icon != NULL);

	gtk_image_set_from_animation(image, icon->get_preparation_animation());
	break;
      }

    case LOADER_STATE_CONNECTED:
    case LOADER_STATE_SIZE_PREPARED:
      {
	CompositeIcon *icon
	  = CompositeIcon::get_composite_icon(GTK_WIDGET(image));
	g_return_if_fail(icon != NULL);

	gtk_image_set_from_animation(image, icon->get_progress_animation());
	break;
      }

    case LOADER_STATE_AREA_PREPARED:
    case LOADER_STATE_LOADING:
    case LOADER_STATE_DONE:
      {
	if (animation != NULL)
	  gtk_image_set_from_animation(image, animation);
	else if (pixbuf != NULL)
	  gtk_image_set_from_pixbuf(image, pixbuf);
	else
	  {
	    CompositeIcon *icon
	      = CompositeIcon::get_composite_icon(GTK_WIDGET(image));
	    g_return_if_fail(icon != NULL);

	    gtk_image_set_from_animation(image, icon->get_progress_animation());
	  }
	break;
      }

    case LOADER_STATE_SCALING_ERROR:
      {
	CompositeIcon *icon
	  = CompositeIcon::get_composite_icon(GTK_WIDGET(image));
	g_return_if_fail(icon != NULL);

	gtk_image_set_from_pixbuf(image, icon->get_error_icon());
	break;
      }

    case LOADER_STATE_FAILURE:
      gtk_image_set_from_stock(image, GTK_STOCK_MISSING_IMAGE,
			       GTK_ICON_SIZE_DIALOG);
      break;

    case LOADER_STATE_A_BONED:
      gtk_image_set_from_stock(image, GTK_STOCK_DIALOG_WARNING,
			       GTK_ICON_SIZE_DIALOG);
      break;
    }

  gtk_widget_queue_resize(GTK_WIDGET(image));
}


void
ImageLoader::update_pixbuf()
{
  assert(OCHUSHA_IN_GDK_THREADS());

  // GdkPixbufLoaderΤloaderñ̤ĤƤΤȸơ
  // thread safeäݤäƤɤդ󤭤ʤФ
  // GdkPixbufGdkPixbufAnimationGTK+ϤȡҤ롣
  // loaderλŻοĽˤꡢGdkPixbufGdkPixbufAnimation
  // ǡ¤thread safeǤʤѹ礬褦ǡ
  // Υߥ󥰤̥åɤGTK+Υʥ˥ΤȤƤ
  // Segmentation Faultꤹ褦
  //
  // äƤȤǡֵɤ᤿ʬ龯ɽפʤɤ
  // ưäƤޤȤϳи礷Ĥġloader"closed"ʥȯ
  // ޤloaderФGdkPixbufGdkPixbufAnimationˤϿʤ
  // 褦ѹƤ
#if 1
  if (state > LOADER_STATE_SIZE_PREPARED && state <= LOADER_STATE_DONE
      && loader_closed && pixbuf == NULL && animation == NULL)
#else
  if (state > LOADER_STATE_SIZE_PREPARED && state <= LOADER_STATE_DONE
      && pixbuf == NULL && animation == NULL)
#endif
    {
      assert(loader != NULL);

      GdkPixbufFormat *pf = gdk_pixbuf_loader_get_format(loader);
      if (pf != NULL)
	{
	  gchar *name = gdk_pixbuf_format_get_name(pf);
	  GdkPixbufAnimation *anim = NULL;
	  if (name != NULL)
	    {
	      if (strcmp(name, "gif") == 0)
		{
		  GDK_THREADS_LEAVE();

		  anim = gdk_pixbuf_loader_get_animation(loader);

		  GDK_THREADS_ENTER();
		}
	      g_free(name);
	    }
	  else
	    anim = NULL;

	  if (anim != NULL)
	    {
	      assert(pixbuf == NULL);
	      assert(animation == NULL);

	      g_object_ref(anim);
	      animation = anim;

	      state = LOADER_STATE_LOADING;
	    }
	  else
	    {
	      GDK_THREADS_LEAVE();

	      GdkPixbuf *pb = gdk_pixbuf_loader_get_pixbuf(loader);

	      GDK_THREADS_ENTER();

	      if (pb != NULL)
		{
		  assert(animation == NULL);
		  assert(pixbuf == NULL);

		  g_object_ref(pb);
		  pixbuf = pb;

		  state = LOADER_STATE_LOADING;
		}
	    }
	}
    }

  if (state == LOADER_STATE_FAILURE && loader_closed && loader != NULL)
    unref_loader();

  list<GtkImage *>::iterator iter = image_list.begin();
  list<GtkImage *>::iterator end_iter = image_list.end();

#ifdef DEBUG_IMAGE_LOADER_MOST
  fprintf(stderr, "ImageLoader::update_pixbuf(): %s, state=%d\n",
	  to_string(), state);
#endif

  while (iter != end_iter)
    update_pixbuf(*iter++);
}


void
ImageLoader::notify_waiting()
{
  assert(OCHUSHA_IN_GDK_THREADS());
  assert(state == LOADER_STATE_INITIAL);

  assert(loader == NULL);
  assert(animation == NULL);
  assert(pixbuf == NULL);
  assert(timeout_id == 0);

  state = LOADER_STATE_WAITING;

  update_pixbuf();
}


void
ImageLoader::notify_started()
{
  assert(OCHUSHA_IN_GDK_THREADS());
  assert(state < LOADER_STATE_STARTED);
  assert(loader == NULL);
  assert(animation == NULL);
  assert(pixbuf == NULL);

  if (timeout_id != 0)
    {
      assert(info.is_alive());
      g_source_remove(timeout_id);
      timeout_id = 0;
    }

  loader = gdk_pixbuf_loader_new();

  if (loader != NULL)
    {
      size_prepared_id
	= g_signal_connect(loader, "size_prepared",
			   reinterpret_cast<GCallback>(size_prepared_cb), this);
      area_prepared_id
	= g_signal_connect(loader, "area_prepared",
			   reinterpret_cast<GCallback>(area_prepared_cb), this);
      area_updated_id
	= g_signal_connect(loader, "area_updated",
			   reinterpret_cast<GCallback>(area_updated_cb), this);
      closed_id
	= g_signal_connect(loader, "closed",
			   reinterpret_cast<GCallback>(closed_cb), this);

      state = LOADER_STATE_STARTED;
      loader_closed = false;
      loader_rem = loader;
    }
  else
    state = LOADER_STATE_FAILURE;

  update_pixbuf();
}


void
ImageLoader::notify_connected()
{
  assert(OCHUSHA_IN_GDK_THREADS());
  assert(state == LOADER_STATE_STARTED);

  state = LOADER_STATE_CONNECTED;

  update_pixbuf();
}


void
ImageLoader::notify_error()
{
  assert(OCHUSHA_IN_GDK_THREADS());
  assert(state <= LOADER_STATE_DONE);

  state = LOADER_STATE_FAILURE;

  update_pixbuf();

#ifdef DEBUG_IMAGE_LOADER_MOST
  fprintf(stderr, "ImageLoader::notify_error(): %s, state=%d\n",
	  to_string(), state);
#endif
}


void
ImageLoader::notify_a_boned()
{
  assert(OCHUSHA_IN_GDK_THREADS());

  state = LOADER_STATE_A_BONED;

  update_pixbuf();
}


GtkWidget *
ImageLoader::get_widget(bool interactive)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  GtkWidget *widget = gtk_image_new();

  if (state < LOADER_STATE_DONE && animation == NULL && pixbuf == NULL)
    {
      // CompositeIcon֥ȤwidgetȰϡʤΤǺäѤ
      new CompositeIcon(widget, OCHUSHA_STOCK_MISSING_IMAGE,
			GTK_ICON_SIZE_DIALOG);
    }

  g_signal_connect(widget, "destroy",
		   reinterpret_cast<GCallback>(destroy_cb), this);
  image_list.push_back(reinterpret_cast<GtkImage *>(widget));

  if (timeout_id != 0)
    {
      g_source_remove(timeout_id);
      timeout_id = 0;
    }

  update_pixbuf(GTK_IMAGE(widget));

  info.enqueue(this, interactive);

  return widget;
}


gboolean
ImageLoader::timeout_cb(ImageLoader *loader)
{
  assert(!OCHUSHA_IN_GDK_THREADS());

#ifdef DEBUG_IMAGE_LOADER_MOST
  fprintf(stderr, "ImageLoader::timeout_cb(): %s, state=%d\n",
	  loader->to_string(), loader->state);
#endif


  GDK_THREADS_ENTER();

  loader->timeout_id = 0;

  if (loader->image_list.empty())
    {
      if (loader->loader != NULL && loader->loader_closed)
	loader->unref_loader();
      loader->info.dequeue(loader);
    }

  GDK_THREADS_LEAVE();


  return FALSE;
}


void
ImageLoader::destroy_cb(GtkWidget *widget, ImageLoader *loader)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  loader->image_list.erase(find(loader->image_list.begin(),
				loader->image_list.end(),
				reinterpret_cast<GtkImage *>(widget)));

  g_return_if_fail(loader->timeout_id == 0);

  if (loader->empty())
    loader->timeout_id
      = g_timeout_add(IMAGE_CACHE_TIMEOUT,
		      reinterpret_cast<GSourceFunc>(timeout_cb), loader);

  // XXX: ɬ̵ϤǰΤ
  gtk_image_set_from_pixbuf(GTK_IMAGE(widget), NULL);

#ifdef DEBUG_IMAGE_LOADER_MOST
  fprintf(stderr, "ImageLoader::destroy_cb(): %s, state=%d\n",
	  loader->to_string(), loader->state);
#endif
}


void
ImageLoader::size_prepared_cb(GdkPixbufLoader *pixbuf_loader, int width,
				  int height, ImageLoader *loader)
{
  assert(!OCHUSHA_IN_GDK_THREADS());
  assert(loader->loader == pixbuf_loader);

#ifdef DEBUG_IMAGE_LOADER_MOST
  fprintf(stderr, "ImageLoader::size_prepared_cb(): %s, state=%d\n",
	  loader->to_string(), loader->state);
#endif

  g_return_if_fail(loader->state < LOADER_STATE_SIZE_PREPARED);

  double width_request = loader->width;
  double height_request = loader->height;

  if (width > width_request || height > height_request)
    {
      double width_scale = width_request / width;
      double height_scale = height_request / height;
      double scale;
      if (width_scale > height_scale)
	scale = height_scale;
      else
	scale = width_scale;

      int scaled_width = static_cast<int>(width * scale);
      int scaled_height = static_cast<int>(height * scale);
      
#warning "Revisit here!"
#define SCALED_IMAGE_SIZE_MINIMUM	16
      // դad-hoc᤮롣äɤ꤬ꤽʴ
      if (scale < 0.01
	  && (scaled_width <= SCALED_IMAGE_SIZE_MINIMUM
	      || scaled_height <= SCALED_IMAGE_SIZE_MINIMUM))
	{
	  /*
	   * ɽ꤬󸽼Ūäꤷintoverflow
	   * ü˽Ĺ䲣Ĺβ̾褦Ȥ̡⤵
	   * 0ˤʤGTK+ΤǤк
	   * ˲ä餫ξ郎4x120դؤscalingǤ
	   * GdkPixbufLoader̵¥롼פäݤ˴٤Τǡ
	   * 괺ad-hocˤ򤱤Ƥ
	   */
	  char message[512];
	  snprintf(message, 512,
		   _("Image has extremely tall or fat in size so as to make GdkPixbufLoader confused when scaled: %s, width=%d, height=%d\n"),
		   loader->info.get_url(), width, height);
	  loader->info.manager.output_log(NON_FATAL_ERROR_LOG, message);

	  loader->state = LOADER_STATE_SCALING_ERROR;

	  return;
	}

      gdk_pixbuf_loader_set_size(pixbuf_loader, scaled_width, scaled_height);
    }

  loader->state = LOADER_STATE_SIZE_PREPARED;
}


void
ImageLoader::area_prepared_cb(GdkPixbufLoader *pixbuf_loader,
			      ImageLoader *loader)
{
  assert(!OCHUSHA_IN_GDK_THREADS());
  assert(loader->loader == pixbuf_loader);

#ifdef DEBUG_IMAGE_LOADER_MOST
  fprintf(stderr, "ImageLoader::area_prepared_cb()#1: %s, state=%d\n",
	  loader->to_string(), loader->state);
#endif

  if (loader->state >= LOADER_STATE_AREA_PREPARED)
    {
      g_return_if_fail(loader->state == LOADER_STATE_SCALING_ERROR
		       || loader->state == LOADER_STATE_FAILURE);
      return;
    }

  loader->state = LOADER_STATE_AREA_PREPARED;

  GDK_THREADS_ENTER();

  loader->update_pixbuf();

  if (loader->loader_closed)
    {
      if (loader->state == LOADER_STATE_LOADING
	  && (loader->animation != NULL || loader->pixbuf != NULL))
	loader->state = LOADER_STATE_DONE;
      else
	{
	  loader->state = LOADER_STATE_FAILURE;
	  loader->update_pixbuf();
	}

      if (loader != NULL)
	loader->unref_loader();
    }

  GDK_THREADS_LEAVE();

#ifdef DEBUG_IMAGE_LOADER_MOST
  fprintf(stderr, "ImageLoader::area_prepared_cb()#2: %s, state=%d\n",
	  loader->to_string(), loader->state);
#endif
}


void
ImageLoader::area_updated_cb(GdkPixbufLoader *pixbuf_loader, int x, int y,
				 int width, int height, ImageLoader *loader)
{
  assert(!OCHUSHA_IN_GDK_THREADS());
  assert(loader->loader == pixbuf_loader);

#ifdef DEBUG_IMAGE_LOADER_MOST
  fprintf(stderr, "ImageLoader::area_updated_cb()#1: %s, state=%d\n",
	  loader->to_string(), loader->state);
#endif

  if (loader->state >= LOADER_STATE_FAILURE)
    return;


  GDK_THREADS_ENTER();

  loader->update_pixbuf();

  if (loader->loader_closed)
    {
      if (loader->state == LOADER_STATE_LOADING
	  && (loader->animation != NULL || loader->pixbuf != NULL))
	loader->state = LOADER_STATE_DONE;
      else
	{
	  loader->state = LOADER_STATE_FAILURE;
	  loader->update_pixbuf();
	}
      if (loader != NULL)
	loader->unref_loader();
    }

  GDK_THREADS_LEAVE();

#ifdef DEBUG_IMAGE_LOADER_MOST
  fprintf(stderr, "ImageLoader::area_updated_cb()#2: %s, state=%d\n",
	  loader->to_string(), loader->state);
#endif
}


void
ImageLoader::closed_cb(GdkPixbufLoader *pixbuf_loader, ImageLoader *loader)
{
  assert(!OCHUSHA_IN_GDK_THREADS());
  assert(loader->loader == pixbuf_loader);

#ifdef DEBUG_IMAGE_LOADER_MOST
  fprintf(stderr, "ImageLoader::closed_cb()#1: %s, state=%d\n",
	  loader->to_string(), loader->state);
#endif


  GDK_THREADS_ENTER();

  loader->loader_closed = true;
  loader->update_pixbuf();

  if (loader->state >= LOADER_STATE_DONE)
    loader->unref_loader();

#ifdef DEBUG_IMAGE_LOADER_MOST
  fprintf(stderr, "ImageLoader::closed_cb()#2: %s, state=%d\n",
	  loader->to_string(), loader->state);
#endif

  if (loader->timeout_id == 0 && loader->empty())
    {
      if (loader->loader != NULL)
	loader->unref_loader();
      loader->info.dequeue(loader);
    }

  GDK_THREADS_LEAVE();
}


class ImageLoaderContext : public RefCount, public has_slots
{
  friend class ImageBuffer;

private:
  ImageManager &manager;
  ImageLoaderInfoHandle info;
  NetworkAgentHandle agent;
  TmpfileHandle tmpfile;
  string content_type;
  string redirect_page;

  enum {
    LOADER_CONTEXT_STATE_INITIAL = 0,
    LOADER_CONTEXT_STATE_REDIRECT_PAGE,
    LOADER_CONTEXT_STATE_IMAGE_DATA,
    LOADER_CONTEXT_STATE_REDIRECT_PAGE_FINISHED,
    LOADER_CONTEXT_STATE_IMAGE_DATA_FINISHED,
    LOADER_CONTEXT_STATE_FAILURE
  } state;

  void prepare_request_cb(NetworkAgent *agent, RefCount *data);
  void connection_established_cb(NetworkAgent *agent, RefCount *data);
  void response_header_received_cb(NetworkAgent *agent, RefCount *data);
  void access_finished_cb(NetworkAgent *agent, RefCount *data);
  void access_failed_cb(NetworkAgent *agent, const char *reason,
			RefCount *data);
  void access_terminated_cb(NetworkAgent *agent, RefCount *data);

  bool append_data(const char *data, size_t length);

  inline const char *get_effective_url() const
  { return info->get_effective_url(); }
  inline const char *get_effective_referer() const
  { return info->get_effective_referer(); }

  void notify_access_finished(int status_code);

  bool check_content_type(const char *content_type);
  bool check_redirection();


public:
  ImageLoaderContext(ImageLoaderInfo &info);
  ~ImageLoaderContext();

  bool start_access();
};


class ImageBuffer : public Buffer
{
private:
  smart_ptr<ImageLoaderContext> context;


public:
  ImageBuffer(ImageLoaderContext *context, int fd)
    : Buffer(fd, false), context(context)
  {}
  virtual ~ImageBuffer();
  virtual bool append_data(const char *data, size_t length, int lock_id);
};


ImageLoaderContext::ImageLoaderContext(ImageLoaderInfo &info)
  : manager(info.manager),
    info(&info),
    state(LOADER_CONTEXT_STATE_INITIAL)
{
}


ImageLoaderContext::~ImageLoaderContext()
{
  assert(static_cast<NetworkAgent *>(agent) == NULL);
  assert(state == LOADER_CONTEXT_STATE_INITIAL);

#ifdef VERBOSE_DESTRUCTOR
  fprintf(stderr, "ImageLoaderContext::~ImageLoaderContext(): %p\n", this);
#endif
}


bool
ImageLoaderContext::start_access()
{
  assert(!OCHUSHA_IN_GDK_THREADS());
  assert(state == LOADER_CONTEXT_STATE_INITIAL);
  assert(static_cast<NetworkAgent *>(agent) == NULL);

  content_type.clear();
  redirect_page.clear();

  Repository &repository = manager.controller.get_repository();
  tmpfile = repository.create_tmpfile(O_WRONLY | O_CREAT | O_TRUNC);
  if (static_cast<Tmpfile *>(tmpfile) == NULL)
    {
      state = LOADER_CONTEXT_STATE_FAILURE;
      return false;
    }
  ImageBuffer *buffer = new ImageBuffer(this, dup(tmpfile->get_fd()));

  agent = manager.controller.get_broker().employ_agent();

  agent->prepare_request.connect(this, &ImageLoaderContext::prepare_request_cb);
  agent->connection_established
    .connect(this, &ImageLoaderContext::connection_established_cb);
  agent->response_header_received
    .connect(this, &ImageLoaderContext::response_header_received_cb);
  agent->access_finished.connect(this, &ImageLoaderContext::access_finished_cb);
  agent->access_failed.connect(this, &ImageLoaderContext::access_failed_cb);
  agent->access_terminated
    .connect(this, &ImageLoaderContext::access_terminated_cb);

  agent->get(get_effective_url(), buffer, this);

  return true;
}


void
ImageLoaderContext::notify_access_finished(int status_code)
{
  // NetworkAgentȯ줿ʥϥɥ餫Τ߸ƤӽФΤǡ
  // λǤthisΥե󥹤ɬͭ

  assert(!OCHUSHA_IN_GDK_THREADS());
  assert(state >= LOADER_CONTEXT_STATE_REDIRECT_PAGE_FINISHED);

  NetworkAgentHandle tmp_agent = agent;
  agent = NULL;

  if (check_redirection())
    {
      // 쥯ؤΥϡ
      state = LOADER_CONTEXT_STATE_INITIAL;
      if (start_access())
	return;
    }

  if (!info->close_loaders(state != LOADER_CONTEXT_STATE_IMAGE_DATA_FINISHED))
    state = LOADER_CONTEXT_STATE_FAILURE;

  if (state == LOADER_CONTEXT_STATE_IMAGE_DATA_FINISHED)
    {
      char message[512];
      // å¸
      if (strcmp(info->get_url(), get_effective_url()) != 0)
	{
	  snprintf(message, 512,
		   _("Caching image file for %s as it is for %s\n"),
		   get_effective_url(), info->get_url());
	  manager.output_log(DEBUG_LOG, message);
	}


      GDK_THREADS_ENTER();

      if (manager.cache(info->get_url(), tmpfile))
	{
	  GDK_THREADS_LEAVE();


	  tmpfile->close();
	}
      else
	{
	  GDK_THREADS_LEAVE();


	  snprintf(message, 512,
		   _("Couldn't cache image file for %s\n"),
		   info->get_url());
	  manager.output_log(NON_FATAL_ERROR_LOG, message);
	}
    }

  if (state != LOADER_CONTEXT_STATE_IMAGE_DATA_FINISHED && status_code == 200)
    status_code = 0;

  state = LOADER_CONTEXT_STATE_INITIAL;

  // info˽λ롣
  info->notify_access_finished(status_code);
}


void
ImageLoaderContext::prepare_request_cb(NetworkAgent *agent, RefCount *data)
{
  // sigslotʥϥɥ
  ImageLoaderContext *context = dynamic_cast<ImageLoaderContext *>(data);
  assert(context != NULL);

  agent->set_header(HTTPHeader::Accept_Encoding, "deflate, gzip");


  const char *referer = context->get_effective_referer();
  if (referer != NULL)
    agent->set_header(HTTPHeader::Referer, referer);


  ImageManager &manager = context->manager;
  agent->set_header(HTTPHeader::User_Agent, manager.get_user_agent());

  const Preferences &prefs = manager.controller.get_prefs();
  if (prefs.get_enable_proxy_for_images())
    {
      agent->set_proxy_authinfo(NULL, NULL);
      agent->set_proxy_url(prefs.get_proxy_uri_for_images().get_uri());
    }

  agent->set_non_interactive_mode(true);
}


bool
ImageLoaderContext::check_content_type(const char *content_type)
{
  if (content_type == NULL)
    {
      state = LOADER_CONTEXT_STATE_FAILURE;
      return false;
    }


  char message[512];
  snprintf(message, 512,
	   _("\"Content-Type: %s\" specified for %s\n"),
	   content_type, get_effective_url());
  manager.output_log(DEBUG_LOG, message);

  this->content_type = content_type;

  if (strstr(content_type, "text/html") != NULL)
    {
      state = LOADER_CONTEXT_STATE_REDIRECT_PAGE;
      return true;
    }

  if (strstr(content_type, "application/octet-stream") != NULL)
    {
      state = LOADER_CONTEXT_STATE_IMAGE_DATA;
      return true;
    }

  if (!manager.is_content_type_supported(content_type))
    {
      snprintf(message, 512,
	       _("\"Content-Type: %s\" is not supported as an image format: %s\n"),
	       content_type, get_effective_url());
      manager.output_log(NON_FATAL_ERROR_LOG, message);

      state = LOADER_CONTEXT_STATE_FAILURE;
      return false;
    }

  state = LOADER_CONTEXT_STATE_IMAGE_DATA;
  return true;
}


bool
ImageLoaderContext::check_redirection()
{
  assert(state >= LOADER_CONTEXT_STATE_REDIRECT_PAGE_FINISHED);

  if (state != LOADER_CONTEXT_STATE_REDIRECT_PAGE_FINISHED
      || redirect_page.empty())
    return false;

  const char *ct = content_type.c_str();
  assert(strstr(ct, "text/html") != NULL);

  const char *cs = strcasestr(ct, "charset=");
  string cs_buf;
  if (cs != NULL)
    {
      cs += 8;	// "charset="ʬ
      const char *cur_pos = cs;
      while (*cur_pos != '\0' && *cur_pos != ';')
	++cur_pos;

      cs_buf.append(cs, cur_pos - cs);
    }
  else
    cs_buf = IconvBuffer::UTF8_ENCODING;

  IconvBuffer body(IconvBuffer::UTF8_ENCODING, cs_buf.c_str());
  body.append(redirect_page.c_str(), redirect_page.length());

  return info->check_redirection(body.get_string());
}


void
ImageLoaderContext::connection_established_cb(NetworkAgent *agent,
					      RefCount *data)
{
  // sigslotʥϥɥ

  info->notify_access_started();
}


void
ImageLoaderContext::response_header_received_cb(NetworkAgent *agent,
						RefCount *data)
{
  // sigslotʥϥɥ

  char message[512];
  ImageLoaderContext *context = dynamic_cast<ImageLoaderContext *>(data);
  assert(context != NULL);

  if (agent->get_status_code() != 200)
    {
      const char *location = agent->get_header(HTTPHeader::Location);
      if (location != NULL)
	{
	  snprintf(message, 512,
		   _("ImageLoaderContext::response_header_received_cb(): \"Location: %s\" specified for %s: status_code = %d\n"),
		   location, context->get_effective_url(),
		   agent->get_status_code());
	  context->manager.output_log(DEBUG_LOG, message);
	}
      else
	{
	  snprintf(message, 512,
		   "ImageLoaderContext::response_header_received_cb(): %s: status_code = %d\n",
		   context->get_effective_url(), agent->get_status_code());
	  context->manager.output_log(DEBUG_LOG, message);
	}

      // 200ʤϤɤɤʤΤǤȤäȤ롣
      // 301: Moved Permanently
      // 302: Moved Temporarily
      // 303: See Other
      // 307: Temporary Redirect
      // դϡLocationإå򸫤ƥ쥯դ礦٤
      // ȻפʤǤʤ㤨itmedia.co.jpβ
      // ǤϡRefererƤʤ302֤äƤ롣
      // äѡΤɤ
      context->state = LOADER_CONTEXT_STATE_FAILURE;

      return;
    }

  if (context->state != LOADER_CONTEXT_STATE_INITIAL)
    return;

  if (!context->check_content_type(agent->get_header(HTTPHeader::Content_Type)))
    return;	// Գ

  // եΥ¤Τ褦ʲĤʤ顢ǥå
}


void
ImageLoaderContext::access_finished_cb(NetworkAgent *agent, RefCount *data)
{
  // sigslotʥϥɥ

  ImageLoaderContext *context = dynamic_cast<ImageLoaderContext *>(data);
  assert(context != NULL);

  char message[512];
  snprintf(message, 512,
	   "ImageLoaderContext::access_finished_cb(): url=\"%s\"\n",
	   get_effective_url());
  context->manager.output_log(DEBUG_LOG, message);

  if (context->state == LOADER_CONTEXT_STATE_IMAGE_DATA)
    context->state = LOADER_CONTEXT_STATE_IMAGE_DATA_FINISHED;
  else if (context->state == LOADER_CONTEXT_STATE_REDIRECT_PAGE)
    context->state = LOADER_CONTEXT_STATE_REDIRECT_PAGE_FINISHED;

  context->notify_access_finished(agent->get_status_code());
}


void
ImageLoaderContext::access_failed_cb(NetworkAgent *agent, const char *reason,
				     RefCount *data)
{
  // sigslotʥϥɥ

  ImageLoaderContext *context = dynamic_cast<ImageLoaderContext *>(data);
  assert(context != NULL);

  char message[512];
  snprintf(message, 512,
	   "ImageLoaderContext::access_failed_cb(): url=\"%s\": %s\n",
	   get_effective_url(), reason != NULL ? reason : _("Unknown reason"));
  context->manager.output_log(DEBUG_LOG, message);

  redirect_page.clear();
  context->state = LOADER_CONTEXT_STATE_FAILURE;
  context->notify_access_finished(agent->get_status_code());
}


void
ImageLoaderContext::access_terminated_cb(NetworkAgent *agent, RefCount *data)
{
  // sigslotʥϥɥ

  ImageLoaderContext *context = dynamic_cast<ImageLoaderContext *>(data);
  assert(context != NULL);

  char message[512];
  snprintf(message, 512,
	   "ImageLoaderContext::access_terminated_cb(): url=\"%s\"\n",
	   get_effective_url());
  context->manager.output_log(DEBUG_LOG, message);

  redirect_page.clear();
  context->state = LOADER_CONTEXT_STATE_FAILURE;
  context->notify_access_finished(agent->get_status_code());
}


bool
ImageLoaderContext::append_data(const char *data, size_t length)
{
  if (state == LOADER_CONTEXT_STATE_IMAGE_DATA)
    {
      if (info->append_data(data, length))
	return true;

      state = LOADER_CONTEXT_STATE_FAILURE;
      return false;
    }
  else if (state == LOADER_CONTEXT_STATE_REDIRECT_PAGE)
    {
      redirect_page.append(data, length);
      return true;
    }

  return false;
}


ImageBuffer::~ImageBuffer()
{
#ifdef VERBOSE_DESTRUCTOR
  fprintf(stderr, "ImageBuffer::~ImageBuffer(): %p\n", this);
#endif
}


bool
ImageBuffer::append_data(const char *data, size_t length, int lock_id)
{
  if (context->append_data(data, length))
    return Buffer::append_data(data, length, lock_id);

  return false;
}


ImageLoaderInfo::ImageLoaderInfo(ImageManager &manager, const char *url)
  : manager(manager),
    url(strdup(url)),
    referer_url(NULL),
    redirected_url(NULL),
    status_code(0),
    last_accessed(0),
    description(NULL),
    state(FREE_ELEMENT)
{
}


ImageLoaderInfo::~ImageLoaderInfo()
{
  assert(active_loaders.empty());

  free(url);

  if (referer_url != NULL)
    free(referer_url);

  if (redirected_url != NULL)
    free(redirected_url);

  if (description != NULL)
    free(description);
}


void
ImageLoaderInfo::set_referer_url(const char *value)
{
  if (value != NULL && referer_url != NULL && strcmp(value, referer_url) == 0)
    return;	// ѹ̵

  if (value != NULL && *value == '\0')
    value = NULL;

  if (referer_url != NULL)
    free(referer_url);

  referer_url = value != NULL ? strdup(value) : NULL;
}


void
ImageLoaderInfo::set_redirected_url(const char *value)
{
  if (value != NULL && redirected_url != NULL
      && strcmp(value, redirected_url) == 0)
    return;	// ѹ̵

  if (value != NULL && *value == '\0')
    value = NULL;

  if (redirected_url != NULL)
    free(redirected_url);

  redirected_url = value != NULL ? strdup(value) : NULL;
}


const char *
ImageLoaderInfo::to_string()
{
  if (description != NULL)
    return description;

  char buf[1024];
  snprintf(buf, 1024, "ImageLoaderInfo(@%p), url=%s", this, url);
  description = strdup(buf);

  return description;
}


const char *
ImageLoaderInfo::get_effective_url() const
{
  if (redirected_url == NULL)
    return url;

  return redirected_url;
}


const char *
ImageLoaderInfo::get_effective_referer() const
{
  if (redirected_url != NULL)
    return url;

  return referer_url;
}


void
ImageLoaderInfo::enqueue(ImageLoader *loader, bool interactive)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  if ((find(done_loader_list.begin(), done_loader_list.end(), loader)
       != done_loader_list.end())
      || (find(active_loader_list.begin(), active_loader_list.end(), loader)
	  != active_loader_list.end())
      || (find(pending_loader_list.begin(), pending_loader_list.end(), loader)
	  != pending_loader_list.end()))
    return;	// ˥塼äƤ

  if (loader->fixed())
    {
      done_loader_list.push_front(loader);
      return;
    }

  pending_loader_list.push_front(loader);
  loader->notify_waiting();
  if (active_loader_list.empty())
    manager.enqueue(this, interactive);
}


void
ImageLoaderInfo::dequeue(ImageLoader *loader)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  bool found = false;

#ifdef DEBUG_IMAGE_LOADER_MOST
  fprintf(stderr, "ImageLoaderInfo::dequeue(%p): %s, state=%d\n",
	  loader, loader->to_string(), loader->get_state());
#endif

  ImageLoaderList::iterator iter
    = find(done_loader_list.begin(), done_loader_list.end(), loader);
  if (iter != done_loader_list.end())
    {
      found = true;
      done_loader_list.erase(iter);
#ifdef NDEBUG
      goto done;
#endif
    }

  iter = find(pending_loader_list.begin(), pending_loader_list.end(), loader);
  if (iter != pending_loader_list.end())
    {
      assert(!found);
      found = true;
      pending_loader_list.erase(iter);
#ifdef NDEBUG
      goto done;
#endif
    }

  iter = find(active_loader_list.begin(), active_loader_list.end(), loader);
  if (iter != active_loader_list.end())
    {
      // XXX: ΤϥХ
      assert(!found);
      found = true;
      active_loader_list.erase(iter);
    }

#ifdef NDEBUG
 done:
#endif
  assert(found);
  delete loader;

  if (active_loader_list.empty()
      && pending_loader_list.empty()
      && done_loader_list.empty())
    manager.dequeue(this);
}


bool
ImageLoaderInfo::is_alive()
{
  assert(OCHUSHA_IN_GDK_THREADS());

  return manager.is_alive();
}


bool
ImageLoaderInfo::hungry()
{
  assert(OCHUSHA_IN_GDK_THREADS());

  if (!active_loader_list.empty())
    return false;

  ImageLoaderList::iterator iter = pending_loader_list.begin();
  ImageLoaderList::iterator end_iter = pending_loader_list.end();
  while (iter != end_iter)
    {
      ImageLoader *loader = *iter++;
      if (!loader->empty())
	return true;
    }

  return false;
}


void
ImageLoaderInfo::notify_a_boned()
{
  assert(OCHUSHA_IN_GDK_THREADS());

  ImageLoaderList::iterator iter = done_loader_list.begin();
  ImageLoaderList::iterator end_iter = done_loader_list.end();
  while (iter != end_iter)
    {
      ImageLoader *loader = *iter++;
      loader->notify_a_boned();
    }

  iter = active_loader_list.begin();
  end_iter = active_loader_list.end();
  while (iter != end_iter)
    {
      ImageLoader *loader = *iter++;
      loader->notify_a_boned();
    }

  iter = pending_loader_list.begin();
  end_iter = pending_loader_list.end();
  while (iter != end_iter)
    {
      ImageLoader *loader = *iter++;
      loader->notify_a_boned();
    }
}


GtkWidget *
ImageLoaderInfo::get_widget(int width, int height, bool interactive)
{
  assert(OCHUSHA_IN_GDK_THREADS());

#ifndef NDEBUG
  GtkWidget *widget = NULL;
#endif
  ImageLoaderList::iterator iter = done_loader_list.begin();
  ImageLoaderList::iterator end_iter = done_loader_list.end();
  while (iter != end_iter)
    {
      ImageLoader *loader = *iter++;
      if (loader->match_size(width, height))
	{
#ifndef NDEBUG
	  assert(widget == NULL);
	  widget = loader->get_widget(interactive);
#else
	  return loader->get_widget(interactive);
#endif
	}
    }

  iter = active_loader_list.begin();
  end_iter = active_loader_list.end();
  while (iter != end_iter)
    {
      ImageLoader *loader = *iter++;
      if (loader->match_size(width, height))
	{
#ifndef NDEBUG
	  assert(widget == NULL);
	  widget = loader->get_widget(interactive);
#else
	  return loader->get_widget(interactive);
#endif
	}
    }

  iter = pending_loader_list.begin();
  end_iter = pending_loader_list.end();
  while (iter != end_iter)
    {
      ImageLoader *loader = *iter++;
      if (loader->match_size(width, height))
	{
#ifndef NDEBUG
	  assert(widget == NULL);
	  widget = loader->get_widget(interactive);
#else
	  return loader->get_widget(interactive);
#endif
	}
    }

#ifndef NDEBUG
  if (widget != NULL)
    return widget;
#endif

  ImageLoader *loader = new ImageLoader(*this, width, height);

  if (status_code != 200 && last_accessed != 0)
    {
      if ((status_code != 0 && status_code != 503)
	  || time(NULL) < last_accessed + IMAGE_LOADER_VOLUNTARY_INTERVAL)
	loader->notify_error();
    }

#ifndef NDEBUG
  widget = loader->get_widget(interactive);

  assert((find(active_loader_list.begin(), active_loader_list.end(), loader)
	  != active_loader_list.end())
	 || (find(pending_loader_list.begin(), pending_loader_list.end(), loader) != pending_loader_list.end())
	 || (find(done_loader_list.begin(), done_loader_list.end(), loader)
	     != done_loader_list.end()));

  return widget;
#else
  return loader->get_widget(interactive);
#endif
}


bool
ImageLoaderInfo::prepare_loading()
{
  assert(OCHUSHA_IN_GDK_THREADS());

  assert(active_loaders.empty());
  assert(active_loader_list.empty());

#ifdef DEBUG_IMAGE_LOADER_MOST
  fprintf(stderr, "ImageLoaderInfo::prepare_loading(): %s\n", to_string());
#endif

  while (!pending_loader_list.empty())
    {
      ImageLoader *loader = pending_loader_list.front();
      active_loader_list.push_front(loader);
      pending_loader_list.pop_front();

      loader->notify_started();

      GdkPixbufLoader *pl = loader->get_loader_rem();
      assert(pl != NULL);
      g_object_ref(pl);
      active_loaders.push_back(pl);
    }

  return true;
}


bool
ImageLoaderInfo::append_data(const char *data, size_t length)
{
  assert(!OCHUSHA_IN_GDK_THREADS());

#ifdef DEBUG_IMAGE_LOADER_MOST
  fprintf(stderr, "ImageLoaderInfo::append_data(): active_loaders.empty()=%d: %s\n", active_loaders.empty(), to_string());
#endif

  if (active_loaders.empty())
    return true;	// ۤäȸ

  bool result = false;

  vector<GdkPixbufLoader *> succeeded_loaders;

  while (!active_loaders.empty())
    {
      GdkPixbufLoader *loader = active_loaders.back();
      active_loaders.pop_back();

      GError *error = NULL;
      if (gdk_pixbuf_loader_write(loader,
				  reinterpret_cast<const guchar *>(data),
				  length, &error))
	{
	  // GdkPixbufLoaderΣĤǤƤ֤
	  // dataϥߤǤʤȸ
	  result = true;
	  succeeded_loaders.push_back(loader);
	}
      else
	{
	  char message[512];
	  snprintf(message, 512,
		   _("%s: gdk_pixbuf_loader_write(...) failed: %s\n"),
		   url, error != NULL ? error->message : _("Unknown reason"));
	  manager.output_log(NON_FATAL_ERROR_LOG, message);

	  if (error != NULL)
	    {
	      g_error_free(error);
	      error = NULL;
	    }


	  GDK_THREADS_ENTER();

	  ImageLoaderList::iterator iter = active_loader_list.begin();
	  ImageLoaderList::iterator end_iter = active_loader_list.end();
	  while (iter != end_iter)
	    {
	      if ((*iter)->get_loader_rem() == loader)
		break;
	      ++iter;
	    }

	  // XXX: assert(iter != end_iter);
	  if (iter != end_iter)
	    {
	      ImageLoader *il = *iter;
	      il->notify_error();

	      done_loader_list.push_front(il);
	      active_loader_list.erase(iter);
	    }

	  GDK_THREADS_LEAVE();


	  if (!gdk_pixbuf_loader_close(loader, &error))
	    {
	      char message[512];
	      snprintf(message, 512,
		       _("%s: gdk_pixbuf_loader_close(%p) failed: %s\n"),
		       url, loader,
		       error != NULL ? error->message : _("Unknown reason"));
	      manager.output_log(NON_FATAL_ERROR_LOG, message);

	      if (error != NULL)
		{
		  g_error_free(error);
		  error = NULL;
		}
	    }

	  g_object_unref(loader);
	}
    }

  assert(active_loaders.empty());

  while (!succeeded_loaders.empty())
    {
      active_loaders.push_back(succeeded_loaders.back());
      succeeded_loaders.pop_back();
    }


  return result;
}


bool
ImageLoaderInfo::close_loaders(bool error_exit)
{
  assert(!OCHUSHA_IN_GDK_THREADS());

#ifdef DEBUG_IMAGE_LOADER_MOST
  fprintf(stderr, "ImageLoaderInfo::close_loaders(): %s\n", to_string());
#endif

  bool result = false;
  if (active_loaders.empty())
    result = true;	// ǧΤ褦ʤΤǤkǡ
  else while (!active_loaders.empty())
    {
      GdkPixbufLoader *loader = active_loaders.back();
      active_loaders.pop_back();


      ImageLoader *il = NULL;

      GDK_THREADS_ENTER();

      // MEMO1: gdk_pixbuf_loader_close()η̡ImageLoader
      // GdkPixbufLoaderΤƤƤޤΤǡGdkPixbufLoaderΥݥ
      // 򸵤бImageLoaderõΤϺ
      ImageLoaderList::iterator iter = active_loader_list.begin();
      ImageLoaderList::iterator end_iter = active_loader_list.end();
      while (iter != end_iter)
	{
	  if ((*iter)->get_loader_rem() == loader)
	    {
	      il = *iter;
	      break;
	    }
	  ++iter;
	}

      GDK_THREADS_LEAVE();


      GError *error = NULL;
      bool close_error = false;
      if (!gdk_pixbuf_loader_close(loader, &error))
	{
	  if (!error_exit)
	    {
	      char message[512];
	      snprintf(message, 512,
		       _("%s: gdk_pixbuf_loader_close(%p) failed: %s\n"),
		       url, loader,
		       error != NULL ? error->message : _("Unknown reason"));
	      manager.output_log(NON_FATAL_ERROR_LOG, message);
	    }

	  if (error != NULL)
	    g_error_free(error);

	  close_error = true;
	}
      else
	result = true;	// ĤǤФk


      if (il != NULL)
	{
	  // ilNULLȤʤΤϡǡɹκФ
	  // ImageLoader˴줿ΤߤǤꡢdone_loader_list
	  // ɬפʤdeleteѤʤΤǤդ̵뤷ɤ

	  GDK_THREADS_ENTER();

	  // MEMO2: gdk_pixbuf_loader_close()η̡ImageLoader
	  // ImageLoaderInfo::dequeue()ƤӽФƾǤƤǽΤǡ
	  // ¸ߤǧ롣
	  iter = find(active_loader_list.begin(), active_loader_list.end(), il);
	  if (iter != active_loader_list.end())
	    {
	      // ilƱݥ󥿤active_loader_list¸ߤ硢
	      // MEMO1ľõilؤƤ֥ȤΤΤ
	      // ؤ
	      // active_loader_listؤǤɲäprepare_loading()Τߤ
	      // Ƥꡢǡɤ߹˺ƤӼ¹Ԥ뤳Ȥ
	      // ʤ
	      // äơilؤƤImageLoaderǤˡthisλؤ
	      // ImageLoaderInfo갷ƱURLβɤ߹ि
	      // ̤ImageLoaderޤilƱɥ쥹֤ˤݤ
	      // ƤġĤʤɤȤԹʾˤϤʤʤ
	      if (close_error || error_exit)
		il->notify_error();

	      done_loader_list.push_front(il);
	      active_loader_list.erase(iter);
	    }

	  GDK_THREADS_LEAVE();
	}

      g_object_unref(loader);
    }

  return result;
}


void
ImageLoaderInfo::notify_access_started()
{
  assert(!OCHUSHA_IN_GDK_THREADS());

  GDK_THREADS_ENTER();

  ImageLoaderList::iterator iter = active_loader_list.begin();
  ImageLoaderList::iterator end_iter = active_loader_list.end();
  while (iter != end_iter)
    {
      ImageLoader *loader = *iter++;
      loader->notify_connected();
    }

  GDK_THREADS_LEAVE();
}


void
ImageLoaderInfo::notify_access_finished(int status_code)
{
  assert(!OCHUSHA_IN_GDK_THREADS());

  this->status_code = status_code;
  last_accessed = time(NULL);

  manager.notify_access_finished(this);
}


bool
ImageLoaderInfo::check_redirection(const char *source)
{
  assert(!OCHUSHA_IN_GDK_THREADS());

  bool result = false;

  GDK_THREADS_ENTER();

  if (redirected_url == NULL)
    {
      // 괺ǡġ
      redirected_url = manager.check_redirection(url, source);
      result = redirected_url != NULL;
    }

  GDK_THREADS_LEAVE();

  return result;
}


bool
ImageLoaderInfo::load_cache(int fd)
{
  assert(!OCHUSHA_IN_GDK_THREADS());

#ifdef DEBUG_IMAGE_LOADER_MOST
  fprintf(stderr, "ImageLoaderInfo::load_cache(%d): %s\n", fd, to_string());
#endif


  GDK_THREADS_ENTER();

  prepare_loading();

  GDK_THREADS_LEAVE();


  if (active_loaders.empty())
    {
      char message[512];
      snprintf(message, 512,
	       _("%s: ImageLoaderInfo::load_cache(%d): No loader available.  Skip loading.\n"),
	       url, fd);
      manager.output_log(DEBUG_LOG, message);

      return true;
    }

  notify_access_started();

  bool result = false;

#ifdef HAVE_MMAP
  size_t buf_len = static_cast<size_t>(lseek(fd, 0, SEEK_END));
  char *buf = static_cast<char *>(mmap(NULL, buf_len, PROT_READ,
				       MAP_PRIVATE | MAP_NOCORE | MAP_NOSYNC,
				       fd, 0));
  if (buf != MAP_FAILED)
    {
      result = append_data(buf, buf_len);
      munmap(buf, buf_len);

      goto done;
    }
  buf = static_cast<char *>(malloc(LOAD_IMAGE_CHUNKSIZE));
#else
  char *buf = static_cast<char *>(malloc(LOAD_IMAGE_CHUNKSIZE));
#endif

  if (buf != NULL)
    {
      do
	{
	  ssize_t len = read(fd, buf, LOAD_IMAGE_CHUNKSIZE);
	  if (len > 0)
	    result = append_data(buf, static_cast<size_t>(len));
	  else if (len == 0)
	    break;
	  else
	    result = false;
	}
      while (result);

      free(buf);
    }
  else
    {
      char message[512];
      snprintf(message, 512,
	       _("%s: ImageLoaderInfo::load_cache(%d) failed: Out of Memory\n"),
	       url, fd);
      manager.output_log(NON_FATAL_ERROR_LOG, message);

      result = false;
    }


#ifdef HAVE_MMAP
 done:
#endif
  close_loaders(!result);


  return result;
}


bool
ImageLoaderInfo::download_image()
{
  assert(OCHUSHA_IN_GDK_THREADS());

  prepare_loading();

  if (active_loaders.empty())
    {
      const Preferences &prefs = manager.controller.get_prefs();
      if (prefs.get_download_shown_images_only())
	{
	  char message[512];
	  snprintf(message, 512,
		   _("%s: ImageLoaderInfo::download_image(): No loader available.  Skip loading.\n"),
		   url);


	  GDK_THREADS_LEAVE();

	  manager.output_log(DEBUG_LOG, message);

	  GDK_THREADS_ENTER();

	  
	  return false;	// ̵̤ʥɤϤʤ
	}
    }


  smart_ptr<ImageLoaderContext> context = new ImageLoaderContext(*this);


  GDK_THREADS_LEAVE();

  bool result =  context->start_access();

  GDK_THREADS_ENTER();


  return result;
}


}
