/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: image_manager.cc,v 1.48 2009/01/01 07:47:06 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "image_manager.h"

#include "ui_constants.h"

#include "libochusha/constants.h"
#include "libochusha/misc_utils.h"
#include "libochusha/uri_utils.h"

#include "libochushagtk/stock.h"


#include <algorithm>
#include <list>

#include <cassert>

#include <gtk/gtk.h>


namespace Ochusha
{

// #define DISABLE_SCHEME_REWRITER	1
// #define CHASE_REF_COUNTS		1
// #define VERBOSE_DESTRUCTOR		1


class ImageLoaderJob : public WorkerJob
{
private:
  ImageManager &manager;


public:
  ImageLoaderJob(ImageManager &manager) : manager(manager) {}
  virtual void work(const WorkerThread *employee);
  virtual const char *description() const
  { return "Processing ImageManager's Job queue"; }
};


void
ImageLoaderJob::work(const WorkerThread *employee)
{
  manager.queue_monitor.enter();

  if (manager.number_of_workers_available++ == 0)
    manager.queue_monitor.notify();

  manager.queue_monitor.exit();


  while (employee->get_state() != WorkerThread::DIE_NOW)
    {
      if (!manager.process(employee))
	break;
    }


  manager.queue_monitor.enter();

  --manager.number_of_workers_available;

  manager.queue_monitor.notify_all();
  manager.queue_monitor.exit();
}


ImageManager::ImageManager(Controller &controller)
  : CacheManager(controller.get_repository(),
		 "image_cache",	// 켡å
		 "image",	// Ҹ
		 controller.get_prefs().get_capacity_of_image_cache(),
		 controller.get_prefs().get_maximum_number_of_image_cache_files()),
    controller(controller),
    a_bone_image_db(controller.get_repository()),
    supported_format_list(gdk_pixbuf_get_formats()),
    user_agent(strdup(OCHUSHA_USER_AGENT)),
    queue_monitor("ImageManager"),
    image_access_db(controller.get_repository()),
    number_of_downloaders_active(0),
    number_of_workers_available(0),
    terminate_workers(false)
{
  GSList *list = supported_format_list;
  while (list != NULL)
    {
      GdkPixbufFormat *format = static_cast<GdkPixbufFormat *>(list->data);
      gchar **mime_types = gdk_pixbuf_format_get_mime_types(format);
      gchar **a_mime_type = mime_types;
      while (*a_mime_type != NULL)
	supported_format_mime_types.push_back(strdup(*a_mime_type++));

      g_strfreev(mime_types);
      list = g_slist_next(list);
    }

  assert(OCHUSHA_IN_GDK_THREADS());


  queue_monitor.enter();

  controller.get_thread_pool().commit(new ImageLoaderJob(*this));

  queue_monitor.wait();
  queue_monitor.exit();


  controller.get_prefs().prefs_changed
    .connect(this, &ImageManager::prefs_changed_cb);
}


ImageManager::~ImageManager()
{
  assert(!OCHUSHA_IN_GDK_THREADS());

  g_slist_free(supported_format_list);
  vector<char *>::iterator iter = supported_format_mime_types.begin();
  vector<char *>::iterator end_iter = supported_format_mime_types.end();
  while (iter != end_iter)
    free(*iter++);
}


void
ImageManager::prefs_changed_cb(Preferences &prefs)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  set_cache_capacity(prefs.get_capacity_of_image_cache());
  set_maximum_number_of_files(prefs.get_maximum_number_of_image_cache_files());
}


GtkWidget *
ImageManager::get_widget(const char *url, int width, int height,
			 bool interactive)
{
  assert(OCHUSHA_IN_GDK_THREADS());
  g_return_val_if_fail(url != NULL, NULL);

  if (*url == '\0')
    return NULL;

  if (is_a_boned(url))
    return NULL;

  const char *referer = NULL;
#ifndef DISABLE_SCHEME_REWRITER
  SchemeManager &sm = controller.get_scheme_manager();
  Cell *result = sm.get_core().evalf0(sm.rewrite_image_link.get_cell(),
				      "%s", url);
  string url_buf;
  string referer_buf;
  if (result != NULL)
    {
      const char *orig_url = url;
      if (result->is_string())
	{
	  url_buf.append(result->get_string(), result->get_strlen());

	  if (strcmp(url_buf.c_str(), url) != 0)
	    url = url_buf.c_str();
	}
      else if (result->is_pair())
	{
	  Cell *url_obj = result->get_car();
	  if (url_obj->is_string())
	    {
	      url_buf.append(url_obj->get_string(), url_obj->get_strlen());
	      if (strcmp(url_buf.c_str(), url) != 0)
		url = url_buf.c_str();
	    }

	  url_obj = result->get_cdr();
	  if (url_obj->is_string())
	    {
	      referer_buf.append(url_obj->get_string(), url_obj->get_strlen());
	      if (!referer_buf.empty())
		referer = referer_buf.c_str();
	    }
	}

      if (url != orig_url || referer != NULL)
	{
	  string message;
	  message.append("(rewrite-image-link \"").append(orig_url)
	    .append("\") => ");
	  if (referer == NULL)
	    message.append("\"").append(url).append("\"\n");
	  else
	    message.append("(\"").append(url).append("\" . \"").append(referer)
	      .append("\")\n");
	  sm.console.output(message.c_str());
	}
    }
#endif

  URI uri(url);
  if (!uri.is_valid()
      || uri.get_scheme() == NULL
      || (strcmp(uri.get_scheme(), "http") != 0
	  && strcmp(uri.get_scheme(), "https") != 0)
      || uri.get_server() == NULL)
    return NULL;


  ImageLoaderInfo *info
    = dynamic_cast<ImageLoaderInfo *>(loader_table.lookup(url));

  if (info == NULL)
    {
      info = new ImageLoaderInfo(*this, url);
      info->set_referer_url(referer);

      AccessResult previous_result(url, AccessResult::UNKNOWN, 0, 0);
      if (image_access_db.query_access_result(url, previous_result) != NULL)
	{
	  info->set_status_code(previous_result.get_status_code());
	  info->set_last_accessed(previous_result.get_last_accessed());
	}

      loader_table.insert(url, info);
    }

  return info->get_widget(width, height, interactive);
}


char *
ImageManager::check_redirection(const char *url, const char *source)
{
  assert(OCHUSHA_IN_GDK_THREADS());

#ifndef DISABLE_SCHEME_REWRITER
  // SchemeؿƤӽФƥå
  SchemeManager &sm = controller.get_scheme_manager();
  Cell *result = sm.get_core().evalf0(sm.redirect_image_link.get_cell(),
				      "%s%s", url, source);
  string url_buf;
  if (result != NULL)
    {
      if (result->is_string())
	{
	  url_buf.append(result->get_string(), result->get_strlen());
	  if (strcmp(url_buf.c_str(), url) == 0)
	    url_buf.clear();
	}

      if (!url_buf.empty())
	{
	  string message;
	  message.append("(redirect-image-link \"").append(url).append("\" \"")
	    .append(source).append("\") => ").append(url_buf);
	  sm.console.output(message.c_str());
	}
    }

  return url_buf.empty() ? NULL : strdup(url_buf.c_str());
#else
  return NULL;
#endif
}


void
ImageManager::prepare_shutdown()
{
  assert(OCHUSHA_IN_GDK_THREADS());

  queue_monitor.enter();

  GDK_THREADS_LEAVE();

  loader_queue.clear();
  terminate_workers = true;

  while (number_of_workers_available > 0)
    {
      queue_monitor.notify_all();
      queue_monitor.wait();
    }
  queue_monitor.exit();

  GDK_THREADS_ENTER();
}


void
ImageManager::set_user_agent(const char *value)
{
  if (value != NULL && user_agent != NULL && strcmp(value, user_agent) == 0)
    return;

  if (user_agent != NULL)
    free(user_agent);

  user_agent = value != NULL ? strdup(value) : NULL;
}


void
ImageManager::enqueue(ImageLoaderInfo *info, bool interactive)
{
  assert(OCHUSHA_IN_GDK_THREADS());

#ifdef DEBUG_IMAGE_LOADER
  fprintf(stderr, "ImageManager::enqueue(%p, %d): url=%s\n",
	  info, interactive, info->get_url());
#endif

  queue_monitor.enter();

  if (info->is_in_busy_queue())
    {
      queue_monitor.exit();
      return;			// ߽
    }

  ImageLoaderInfoHandle tmp_ref = info;

  if (info->is_in_loader_queue())
    {
      if (interactive)
	{
	  ImageLoaderInfoList::iterator iter = loader_queue.begin();
	  ImageLoaderInfoList::iterator end_iter = loader_queue.end();
	  while (iter != end_iter)
	    {
	      if (info == *iter)
		{
		  if (interactive)
		    {
		      // 塼Ƭ˰ư
		      loader_queue.erase(iter);
		      loader_queue.push_front(info);
		    }
		  goto wakeup_workers;
		}
	      ++iter;
	    }
	  // NEVER REACHED
	  assert(iter != end_iter);
	}
    }
  else
    info->set_in_loader_queue();

  if (!interactive)
    loader_queue.push_back(info);
  else
    loader_queue.push_front(info);

 wakeup_workers:
  queue_monitor.notify();
  queue_monitor.exit();

  activate_downloaders();
}


void
ImageManager::dequeue(ImageLoaderInfo *info)
{
  assert(OCHUSHA_IN_GDK_THREADS());

#ifdef DEBUG_IMAGE_LOADER
  fprintf(stderr, "ImageManager::dequeue(%p): url=%s\n",
	  info, info->get_url());
#endif

  queue_monitor.enter();

  if (info->is_in_busy_queue() || info->is_free_element())
    {
      loader_table.remove(info->get_url());

      queue_monitor.exit();
      return;
    }

  assert(info->is_in_loader_queue());

  ImageLoaderInfoList::iterator iter = loader_queue.begin();
  ImageLoaderInfoList::iterator end_iter = loader_queue.end();
  while (iter != end_iter)
    {
      if (info == *iter)
	{
	  loader_queue.erase(iter);
	  info->set_free_element();
	  break;
	}
      ++iter;
    }

  assert(iter != end_iter);

  queue_monitor.exit();

  loader_table.remove(info->get_url());
}


/*
 * managerΥ塼Ż򣱤ļФƽ롣
 * ŻʤСqueue_monitor̲롣
 * åɤλ٤ˤϡfalse֤
 */
bool
ImageManager::process(const WorkerThread *employee)
{
  assert(!OCHUSHA_IN_GDK_THREADS());

#ifdef DEBUG_IMAGE_LOADER
  fprintf(stderr, "ImageManager::process(%p)\n", employee);
#endif


  GDK_THREADS_ENTER();

  queue_monitor.enter();

  int fd = -1;
  ImageLoaderInfoHandle job = NULL;

  ImageLoaderInfoList::iterator iter = loader_queue.begin();
  ImageLoaderInfoList::iterator end_iter = loader_queue.end();

  while (iter != end_iter)
    {
      ImageLoaderInfo *info = *iter;
      fd = open_cache(info->get_url(), O_RDONLY);
      if (fd >= 0)
	{
	  touch(info->get_url());
	  job = info;
	  break;
	}
      ++iter;
    }

  GDK_THREADS_LEAVE();


  if (fd >= 0)
    {
      assert(job != NULL);
      assert(iter != end_iter);

      loader_queue.erase(iter);
      busy_queue.push_front(job);
      job->set_in_busy_queue();

      queue_monitor.exit();


      if (!job->load_cache(fd))
	{
	  char message[512];
	  snprintf(message, 512,
		   _("cached file for \"%s\" seems not a good image file\n"),
		   job->get_url());
	  output_log(NON_FATAL_ERROR_LOG, message);
	  // invalidate(job->get_url());
	}
      close(fd);


      queue_monitor.enter();

      ImageLoaderInfoList::iterator iter
	= find(busy_queue.begin(), busy_queue.end(), job);
      if (iter != busy_queue.end())
	busy_queue.erase(iter);

      job->set_free_element();

      queue_monitor.exit();
    }
  else
    {
      // åɤŻʤˤĤơġ
      queue_monitor.wait();
      queue_monitor.exit();
    }

  return !terminate_workers;
}


void
ImageManager::activate_downloaders()
{
  assert(OCHUSHA_IN_GDK_THREADS());

  queue_monitor.enter();

  ImageLoaderInfoList downloader_queue;

  ImageLoaderInfoList::iterator iter = loader_queue.begin();
  ImageLoaderInfoList::iterator end_iter = loader_queue.end();

  if (!terminate_workers)
    {
      while (iter != end_iter)
	{
	  ImageLoaderInfo *info = *iter++;
	  if (!cache_exist(info->get_url()))
	    downloader_queue.push_back(info);
	}
    }

  while (number_of_downloaders_active < MAXIMUM_NUM_IMAGE_DOWNLOADERS
	 && !downloader_queue.empty())
    {
      ImageLoaderInfoHandle info = downloader_queue.front();
      downloader_queue.pop_front();

      ImageLoaderInfoList::iterator iter
	= find(loader_queue.begin(), loader_queue.end(), info);
      if (iter != loader_queue.end())
	{
	  loader_queue.erase(iter);

	  busy_queue.push_front(info);
	  info->set_in_busy_queue();

	  ++number_of_downloaders_active;

	  // info->download_image()
	  // GDK_THREADS_LEAVE()/GDK_THREADS_ENTER()롣
	  // ΰ١queue_monitor.enter()ʾ֤ǸƤӽФȥå
	  // queue_monitorGDKåȤʤꡢǥåɥå
	  queue_monitor.exit();


	  bool result = info->download_image();


	  queue_monitor.enter();

	  if (!result)
	    {
	      info->set_free_element();

	      iter = find(busy_queue.begin(), busy_queue.end(), info);
	      assert(iter != busy_queue.end());
	      busy_queue.erase(iter);

	      --number_of_downloaders_active;
	    }
	}
    }

  queue_monitor.exit();
}


void
ImageManager::notify_access_finished(ImageLoaderInfo *info)
{
  assert(!OCHUSHA_IN_GDK_THREADS());

  AccessResult::ResultType result;
  int status_code = info->get_status_code();
  switch (info->get_status_code())
    {
    case 0:
      result = AccessResult::UNKNOWN;
      break;

    case 200:
      // Ԥ褦˲ǡɤ߹᤿ɤ
      // GOODBROKEN_IMAGEȤʬϸ塹Ȥǡġġ
      result = AccessResult::GOOD;
      break;

    default:
      result = AccessResult::ACCESS_ERROR;
    }

  image_access_db.register_access_result(info->get_url(), result, status_code,
					 info->get_last_accessed());


  queue_monitor.enter();

  --number_of_downloaders_active;

  if (info->is_in_busy_queue())
    {
      info->set_free_element();

      ImageLoaderInfoList::iterator iter
	= find(busy_queue.begin(), busy_queue.end(), info);
      assert(iter != busy_queue.end());
      busy_queue.erase(iter);
    }

  assert(info->is_free_element());

  queue_monitor.exit();


  GDK_THREADS_ENTER();

  activate_downloaders();

  GDK_THREADS_LEAVE();
}


void
ImageManager::a_bone(const char *url)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  if (url == NULL || *url == '\0')
    return;

  if (a_bone_image_db.is_a_boned_url(url))
    return;	// ˤܡ󤵤Ƥ

  URI uri(url);
  if (!uri.is_valid())
    return;

  a_bone_image_db.register_a_bone_image_url(url, time(NULL));

  ImageLoaderInfo *info
    = dynamic_cast<ImageLoaderInfo *>(loader_table.lookup(url));
  if (info != NULL)
    info->notify_a_boned();

  invalidate(url);
}


void
ImageManager::a_bone_server(const char *url)
{
  if (url == NULL || *url == '\0')
    return;

  URI uri(url);
  const char *server = uri.get_server();
  if (!uri.is_valid() || server == NULL
      || a_bone_image_db.is_a_boned_domain(server))
    return;

  a_bone_image_db.register_a_bone_image_domain(server, time(NULL));

  vector<char *> cached_urls;
  get_cached_urls(cached_urls);

  vector<char *>::iterator iter = cached_urls.begin();
  vector<char *>::iterator end_iter = cached_urls.end();
  while (iter != end_iter)
    {
      char *cached_url = *iter++;
      URI cached_uri(cached_url);
      if (cached_uri.is_valid() && cached_uri.get_server() != NULL
	  && strcasecmp(cached_uri.get_server(), server) == 0)
	{
	  ImageLoaderInfo *info
	    = dynamic_cast<ImageLoaderInfo *>(loader_table.lookup(cached_url));
	  if (info != NULL)
	    info->notify_a_boned();

	  invalidate(cached_url);
	}

      free(cached_url);
    }
}


bool
ImageManager::is_a_boned(const char *url)
{
  assert(url != NULL);

  if (a_bone_image_db.is_a_boned_url(url))
    return true;

  URI uri(url);
  if (!uri.is_valid() || uri.get_server() == NULL)
    return false;

  return a_bone_image_db.is_a_boned_domain(uri.get_server());
}


bool
ImageManager::is_image_link(const char *url)
{
  URI uri(url);
  if (!uri.is_valid())
    return false;

  const char *scheme = uri.get_scheme();
  if (scheme == NULL
      || (strcmp(scheme, "http") != 0 && strcmp(scheme, "https") != 0))
    return false;

  const char *path = uri.get_path();
  if (path != NULL
      && (g_str_has_suffix(path, ".jpg") || g_str_has_suffix(path, ".png")
	  || g_str_has_suffix(path, ".gif") || g_str_has_suffix(path, ".jpeg")
	  || g_str_has_suffix(path, ".JPG") || g_str_has_suffix(path, ".PNG")
	  || g_str_has_suffix(path, ".GIF") || g_str_has_suffix(path, ".JPEG")))
    return !is_a_boned(url);

  const char *query = uri.get_query();
  if (query != NULL
      && (strcasestr(query, ".jpg") != NULL
	  || strcasestr(query, ".png") != NULL
	  || strcasestr(query, ".gif") != NULL
	  || strcasestr(query, ".jpeg") != NULL))
    return !is_a_boned(url);

#ifndef DISABLE_SCHEME_REWRITER
  assert(OCHUSHA_IN_GDK_THREADS());

  SchemeManager &sm = controller.get_scheme_manager();
  Cell *result = sm.get_core().evalf0(sm.is_image_link.get_cell(), "%s", url);

  return result != NULL && result != Cell::f();
#else
  return false;
#endif
}


bool
ImageManager::is_content_type_supported(const char *content_type)
{
  if (content_type == NULL || *content_type == '\0')
    return false;

  vector<char *>::iterator iter = supported_format_mime_types.begin();
  vector<char *>::iterator end_iter = supported_format_mime_types.end();
  while (iter != end_iter)
    {
      if (strcmp(*iter, content_type) == 0)
	return true;

      ++iter;
    }

  return false;
}


void
ImageManager::invalidate(const char *url)
{
  // MEMO: åϢξĤ褦ˤʤä顢פʤʤä
  // Ϥ˴Ф褤
  CacheManager::invalidate(url);
}

}
