/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: main.cc,v 1.27 2009/01/05 12:18:03 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#include "libochusha/constants.h"
#include "libochusha/cookie_manager.h"
#include "libochusha/lock.h"
#include "libochusha/repository.h"

#include "command_pipe.h"
#include "controller.h"
#include "session_client.h"
#include "ui_constants.h"

#include "libochushagtk/stock.h"

#include <glib.h>
#include <gtk/gtk.h>
#include <libxml/parser.h>

#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <cstring>

#include <iostream>


#define ENABLE_STRICT_CHECK		1


#include <pthread.h>

using namespace Ochusha;

Monitor ochusha_gdk_threads_lock("GDK");


static void
ochusha_gdk_threads_enter()
{
  ochusha_gdk_threads_lock.lock();

#ifdef ENABLE_STRICT_CHECK
  assert(ochusha_gdk_threads_lock.get_count() == 1);
#else
  if (ochusha_gdk_threads_lock.get_count() > 1)
    cerr << "Warning: gdk_threads_enter() called recursively.\n";
#endif
}


bool
ochusha_gdk_threads_tryenter()
{
  if (ochusha_gdk_threads_lock.trylock())
    {
#ifdef ENABLE_STRICT_CHECK
      assert(ochusha_gdk_threads_lock.get_count() == 1);
#else
      if (ochusha_gdk_threads_lock.get_count() > 1)
	cerr << "Warning: gdk_threads_enter() called recursively.\n";
#endif
      return true;
    }

  return false;
}


bool
ochusha_in_gdk_threads()
{
  return ochusha_gdk_threads_lock.is_locked_by_self();
}


static void
ochusha_gdk_threads_leave()
{
  ochusha_gdk_threads_lock.unlock();
}


int
main(int argc, char *argv[])
{
  bindtextdomain(PACKAGE_NAME, LOCALEDIR);
  textdomain(PACKAGE_NAME);

  char buf[PATH_MAX];
  const char *ochusha_home = getenv("OCHUSHA_HOME");
  if (ochusha_home == NULL)
    {
      const char *home = g_get_home_dir();
      if (home == NULL)
	{
	  fprintf(stderr, _("Cannot locate your home directory.\n"));
	  return EXIT_FAILURE;
	}

      if (snprintf(buf, PATH_MAX, "%s/.ochusha/", home) >= PATH_MAX)
	{
	  fprintf(stderr,
		  _("Your home directory is located too deep to use: %s\n"),
		  home);
	  return EXIT_FAILURE;
	}

      ochusha_home = buf;
    }

  Repository repository(ochusha_home);
  if (repository.get_home_path() == NULL)
    {
      fprintf(stderr, _("Cannot prepare ochusha home: %s\n"), ochusha_home);
      return EXIT_FAILURE;
    }

  SessionClient session_client(repository);


  char filepath[PATH_MAX];
  if (repository.expand_path(OCHUSHA_COOKIES_SQLITE, filepath, PATH_MAX)
      == NULL)
    filepath[0] = '\0';

  CookieManager cookie_manager(filepath);


  if (repository.expand_path(OCHUSHA_HISTORIES_SQLITE, filepath, PATH_MAX)
      == NULL)
    filepath[0] = '\0';

  HistoryManager history_manager(filepath, false);


  if (repository.expand_path(OCHUSHA_POST_HISTORIES_SQLITE, filepath, PATH_MAX)
      == NULL)
    filepath[0] = '\0';

  HistoryManager post_history_manager(filepath, true);


  g_set_application_name(_("Ochusha"));


  // ޥɥ饤ɾ
  GOptionContext *option_context
    = g_option_context_new(N_("[URL...]"));

  gboolean version = FALSE;
  GOptionEntry option_entries[] = {
    { "version", 0, 0, G_OPTION_ARG_NONE, &version,
      N_("Show version"), NULL },
    { NULL }
  };
  GOptionGroup *option_group = g_option_group_new("ochusha",
						  N_("Ochusha options"),
						  N_("Show ochusha options"),
						  NULL, NULL);
  g_option_group_add_entries(option_group, option_entries);
  g_option_group_set_translation_domain(option_group, PACKAGE_NAME);
  g_option_context_set_main_group(option_context, option_group);
  g_option_context_add_group(option_context, session_client.get_option_group());
  g_option_context_add_group(option_context, gtk_get_option_group(TRUE));
  g_option_context_set_help_enabled(option_context, TRUE);

  GError *error = NULL;
  g_option_context_parse(option_context, &argc, &argv, &error);

  if (error != NULL)
    {
      fprintf(stderr, "Error: %s\n", error->message);
      g_error_free(error);
      return EXIT_FAILURE;
    }

  if (version)
    {
      fprintf(stdout, "%s version " PACKAGE_VERSION "\n", _("Ochusha"));
      return 0;
    }

  if (!session_client.process_command_line())
    return 0;	// --discard=IDξϽλ

  g_option_context_free(option_context);

  CommandPipe command_pipe(repository);

  if (argc > 1)
    {
      char command_buf[PATH_MAX];
      for (int i = 1; i < argc; i++)
	{
	  snprintf(command_buf, PATH_MAX,
		   "(open-url-in-tab \"%s\")\n", argv[i]);
	  command_pipe.send(command_buf);
	}
    }

  if (!command_pipe.is_master())
    {
      command_pipe.send("(raise)\n");
      exit(0);	/* ޥǽλ */
    }

  // GTK+ʤΤUTF-8ò
  bind_textdomain_codeset(PACKAGE_NAME, "UTF-8");

  xmlInitParser();

  if (!g_thread_supported())
    g_thread_init(NULL);

  gdk_threads_set_lock_functions(ochusha_gdk_threads_enter,
				 ochusha_gdk_threads_leave);

  gdk_threads_init();

#ifdef CHECK_LOCK_ORDER
  new ThreadInfo("<<Ochusha Main>>");
#endif

  GDK_THREADS_ENTER();

  gtk_set_locale();
  gtk_init(&argc, &argv);

  // ̵̣äݤ͡Ȥפɤġġ
  gtk_rc_parse_string("gtk-toolbar-style = icons\n"
		      "gtk-toolbar-icon-size = small-toolbar\n");
  // ̵̣äݤ͡Ȥפɤġġ

  gtk_rc_add_default_file(PKGDATADIR "/" OCHUSHA_GTKRC);

  if (repository.find_file(OCHUSHA_GTKRC, filepath, PATH_MAX) != NULL)
    {
      gtk_rc_parse(filepath);
      gtk_rc_add_default_file(filepath);
    }

  if (repository.find_file(OCHUSHA_PREFS_GTKRC, filepath, PATH_MAX) != NULL)
    {
      gtk_rc_parse(filepath);
      gtk_rc_add_default_file(filepath);
    }

  if (repository.find_file(OCHUSHA_ANARCHY_GTKRC, filepath, PATH_MAX) != NULL)
    {
      gtk_rc_parse(filepath);
      gtk_rc_add_default_file(filepath);
    }

  ochusha_initialize_stock_items();

  {
    Controller controller(repository, command_pipe, session_client,
			  cookie_manager, history_manager,
			  post_history_manager);
    session_client.open_session(&controller, argc, argv);
    command_pipe.initialize_command_handler(&controller);


    gtk_main();


    controller.prepare_shutdown();


    xmlCleanupParser();

    command_pipe.finalize_command_handler();
    session_client.close_session();
  }

  GDK_THREADS_LEAVE();


  return 0;
}
