/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: main_window.cc,v 1.66 2010/08/15 15:42:20 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "main_window.h"

#include "controller.h"
#include "focus_utils.h"
#include "session_client.h"
#include "ui_constants.h"
#include "view.h"
#include "view_manager.h"

#include "libochushagtk/searchbar.h"
#include "libochushagtk/stock.h"
#include "libochushagtk_lgpl/mod_notebook.h"

#include "ochusha_ui.h"

#include "ochusha48_pixdata.h"

#include <cassert>
#include <cstdlib>

#include <vector>


#include "stable_icon_pixdata.h"
#include "progress_icon_data.h"


#define DEFAULT_WINDOW_WIDTH	600
#define DEFAULT_WINDOW_HEIGHT	400
#define DEFAULT_SIDEBAR_WIDTH	150


namespace Ochusha
{


static Lock pending_status_lock("MainWindow::pending_status");


static void
find_again_notify_sensitive_cb(GtkWidget *button, GParamSpec *arg1,
			       GtkAction *action)
{
#if GTK_CHECK_VERSION(2,18,0)
  gtk_action_set_sensitive(action, gtk_widget_is_sensitive(button));
#else
  gtk_action_set_sensitive(action, GTK_WIDGET_IS_SENSITIVE(button));
#endif
}


MainWindow::MainWindow(Controller &controller)
  : controller(controller),
    menubar_action_group(create_menubar_action_group()),
    current_view_action_group(create_current_view_action_group()),
    url_entry(NULL),
    history_view(NULL),
    post_history_view(NULL),
    shown_sidebar_items(0),
    view_manager(controller, *this),
    search_target(NULL),
    search_message_id(0),
    search_response_id(0),
    search_request_id(0),
    current_focus_view(NULL),
    current_main_view(NULL),
    process_pending_status_on_idle_id(0),
    stable_icon(NULL),
    progress_icon(NULL),
    progress_icon_iter(NULL),
    progress_icon_frame_timeout_id(0),
    active_board_url(NULL),
    active_thread_url(NULL),
    maximized(false)
{
  static GdkGeometry geometry = { 0, 0, 0, 0, 0, 0, 0, 0, 0.0, 0.0,
				  GDK_GRAVITY_NORTH_WEST };

  const char *id = controller.get_session().get_sm_client_id();
  if (id != NULL)
    gdk_set_sm_client_id(id);

  setup_main_ui();

  toplevel = GTK_WINDOW(gtk_window_new(GTK_WINDOW_TOPLEVEL));
  g_signal_connect(toplevel, "destroy",
		   reinterpret_cast<GCallback>(gtk_widget_destroyed),
		   &toplevel);
  g_signal_connect(toplevel, "delete_event",
		   reinterpret_cast<GCallback>(delete_event_cb), this);
  g_signal_connect(toplevel, "window_state_event",
		   reinterpret_cast<GCallback>(window_state_event_cb), this);

  GtkAccelGroup *accel_group
    = gtk_ui_manager_get_accel_group(controller.get_ui_manager());
  gtk_window_add_accel_group(toplevel, accel_group);

  gtk_window_set_title(toplevel, _("Ochusha"));
  gtk_window_set_icon(toplevel,
		      gdk_pixbuf_new_from_inline(-1, ochusha48, FALSE, NULL));

  // 礭
  int width;
  const char *value = controller.get_session().get_state("MainWindow::width");
  if (value == NULL || sscanf(value, "%d", &width) == 0 || width == 0)
    width = DEFAULT_WINDOW_WIDTH;

  int height;
  value = controller.get_session().get_state("MainWindow::height");
  if (value == NULL || sscanf(value, "%d", &height) == 0 || height == 0)
    height = DEFAULT_WINDOW_HEIGHT;

  gtk_widget_set_size_request(GTK_WIDGET(toplevel), width, height);

  int x;
  value = controller.get_session().get_state("MainWindow::x");
  if (value == NULL || sscanf(value, "%d", &x) != 1)
    x = -1;

  int y;
  value = controller.get_session().get_state("MainWindow::y");
  if (value == NULL || sscanf(value, "%d", &y) != 1)
    y = -1;

  if (x >= 0 && y >= 0)
    {
      gtk_widget_realize(GTK_WIDGET(toplevel));
      gtk_window_move(toplevel, x, y);
      gtk_widget_queue_resize(GTK_WIDGET(toplevel));
    }

  int m = 0;
  value = controller.get_session().get_state("MainWindow::maximized");
  if (value != NULL && sscanf(value, "%d", &m) == 1 && m != 0)
    {
      maximized = true;
      gtk_window_maximize(toplevel);
    }

  main_box = GTK_BOX(gtk_vbox_new(FALSE, 0));
  gtk_container_add(GTK_CONTAINER(toplevel), GTK_WIDGET(main_box));
  gtk_window_set_geometry_hints(toplevel, GTK_WIDGET(main_box),
				&geometry, GDK_HINT_MIN_SIZE);

  menubar = gtk_ui_manager_get_widget(controller.get_ui_manager(), "/MainMenu");


  // XXX: ¸
  gtk_widget_set_app_paintable(menubar, TRUE);
  g_signal_connect_after(menubar, "expose-event",
			 reinterpret_cast<GCallback>(menubar_expose_event_cb),
			 this);

  // GtkImageɽƤ뤻˥塼СطʤȥƩʬ
  // ˺Τˤʤġġ
  // ˥塼СGTK_APP_PAINTABLEˤľ񤭤뤷ʤΤ
  stable_icon = gdk_pixbuf_new_from_inline(-1, stable_icon_16, FALSE, NULL);

  GdkPixbufLoader *loader = gdk_pixbuf_loader_new();

  g_signal_connect(loader, "area-prepared",
		   reinterpret_cast<GCallback>(area_prepared_cb), this);
  g_signal_connect(loader, "area-updated",
		   reinterpret_cast<GCallback>(area_updated_cb), this);
  g_signal_connect(loader, "closed",
		   reinterpret_cast<GCallback>(closed_cb), this);

  GError *error = NULL;
  if (!gdk_pixbuf_loader_write(loader,
			       progress_icon_16, sizeof(progress_icon_16),
			       &error))
    {
      fprintf(stderr, _("Cannot prepare progress icon: %s\n"),
	      error != NULL ? error->message : _("Unknown reason"));
      if (error != NULL)
	{
	  g_error_free(error);
	  error = NULL;
	}
    }

  if (!gdk_pixbuf_loader_close(loader, &error))
    {
      fprintf(stderr, _("Cannot prepare progress icon: %s\n"),
	      error != NULL ? error->message : _("Unknown reason"));
      if (error != NULL)
	{
	  g_error_free(error);
	  error = NULL;
	}
    }

  g_object_unref(loader);

  gtk_widget_show(menubar);
  gtk_box_pack_start(main_box, menubar, FALSE, TRUE, 0);


  navigation_toolbar = gtk_ui_manager_get_widget(controller.get_ui_manager(),
						 "/NavigationToolbar");
  GtkToggleAction *action = controller.get_prefs().show_navigation_toolbar;
  if (navigation_toolbar != NULL)
    {
      gtk_toolbar_set_style(GTK_TOOLBAR(navigation_toolbar),
			    GTK_TOOLBAR_ICONS);
      gtk_box_pack_start(main_box, navigation_toolbar, FALSE, TRUE, 0);

      if (gtk_toggle_action_get_active(action))
	gtk_widget_show(navigation_toolbar);
      else
	gtk_widget_hide(navigation_toolbar);
      g_signal_connect(action, "toggled",
		       reinterpret_cast<GCallback>(show_navigation_toolbar_toggled_cb), this);

      GtkWidget *tool_item
	= gtk_ui_manager_get_widget(controller.get_ui_manager(),
				    "/NavigationToolbar/OpenURLEntry");
      // ʳξ֤줿ˤбǤʤդʤä
      //   ȤˤƤ
      if (tool_item != NULL)
	{
	  GtkWidget *combo_box = gtk_bin_get_child(GTK_BIN(tool_item));
	  if (GTK_IS_COMBO_BOX_ENTRY(combo_box))
	    {
	      GtkWidget *entry = gtk_bin_get_child(GTK_BIN(combo_box));
	      if (GTK_IS_ENTRY(entry))
		{
		  url_entry = GTK_ENTRY(entry);
		  gtk_entry_set_width_chars(url_entry, -1);
		  new EntryFocusNotifier(controller, url_entry, true);
		  g_signal_connect(entry, "activate",
				   reinterpret_cast<GCallback>(url_entry_activate_cb),
				   this);
		}
	    }
	}
    }
  else
    gtk_action_set_sensitive(GTK_ACTION(action), FALSE);

  main_container = gtk_hpaned_new();
  g_signal_connect(main_container, "notify::position-set",
		   reinterpret_cast<GCallback>(notify_position_set_cb), this);
  gtk_widget_show(main_container);
  gtk_box_pack_start(main_box, main_container, TRUE, TRUE, 1);

  value = controller.get_session().get_state("MainWindow::sidebar_width");
  if (value != NULL && sscanf(value, "%d", &width) == 1)
    gtk_paned_set_position(GTK_PANED(main_container), width);


  // ɥСν
  GtkWidget *widget = mod_notebook_new();
  g_signal_connect(widget, "button_press_event",
		   reinterpret_cast<GCallback>(sidebar_button_press_event_cb),
		   this);
  sidebar_book = MOD_NOTEBOOK(widget);
  mod_notebook_set_scrollable(sidebar_book, TRUE);
  mod_notebook_set_show_tabs(sidebar_book, TRUE);
  mod_notebook_set_minimum_tab_label_size(sidebar_book, 0);
  mod_notebook_set_tab_shrinkable(sidebar_book, TRUE);
  mod_notebook_tooltips_enable(sidebar_book);


  // İ
  boardlist_view = controller.boardlist_ui.get_master_view();
  boardlist_view->notify_attached(this);
  GtkWidget *tmp_widget = boardlist_view->get_widget();
  action = controller.get_prefs().show_boardlist_sidebar;
  gtk_widget_show(tmp_widget);
  if (gtk_toggle_action_get_active(action))
    shown_sidebar_items |= BOARDLIST_SIDEBAR;
  g_signal_connect(action, "toggled",
		   reinterpret_cast<GCallback>(show_boardlist_sidebar_toggled_cb), this);
  mod_notebook_append_page(sidebar_book, tmp_widget,
			   boardlist_view->get_label());


  // ֥åޡ
  bookmarks_view = controller.bookmarks_ui.get_master_view();
  bookmarks_view->notify_attached(this);
  tmp_widget = bookmarks_view->get_widget();
  action = controller.get_prefs().show_bookmarks_sidebar;
  gtk_widget_show(tmp_widget);
  if (gtk_toggle_action_get_active(action))
    shown_sidebar_items |= BOOKMARKS_SIDEBAR;
  g_signal_connect(action, "toggled",
		   reinterpret_cast<GCallback>(show_bookmarks_sidebar_toggled_cb), this);
  mod_notebook_append_page(sidebar_book, tmp_widget,
			   bookmarks_view->get_label());

#if 0
  // 
  history_view = controller.history_ui.get_master_view();
  history_view->notify_attached(this);
  tmp_widget = history_view->get_widget();
  action = controller.get_prefs().show_history_sidebar;
  gtk_widget_show(tmp_widget);
  if (gtk_toggle_action_get_active(action))
    shown_sidebar_items |= HISTORY_SIDEBAR;
  g_signal_connect(action, "toggled",
		   reinterpret_cast<GCallback>(show_history_sidebar_toggled_cb), this);
  mod_notebook_append_page(sidebar_book, tmp_widget, history_view->get_label());


  // 
  post_history_view = controller.post_history_ui.get_master_view();
  post_history_view->notify_attached(this);
  tmp_widget = post_history_view->get_widget();
  action = controller.get_prefs().show_post_history_sidebar;
  gtk_widget_show(tmp_widget);
  if (gtk_toggle_action_get_active(action))
    shown_sidebar_items |= POST_HISTORY_SIDEBAR;
  g_signal_connect(action, "toggled",
		   reinterpret_cast<GCallback>(show_post_history_sidebar_toggled_cb), this);
  mod_notebook_append_page(sidebar_book, tmp_widget,
			   post_history_view->get_label());
#endif

  if ((shown_sidebar_items & SHOWN_SIDEBAR_MASK) != 0)
    {
      int item = 0;
      value = controller.get_session().get_state("MainWindow::current_sidebar_item");
      if (value != NULL && sscanf(value, "%d", &item) == 1)
	{
	  View *view;
	  if (item == BOARDLIST_SIDEBAR)
	    view = boardlist_view;
	  else if (item == BOOKMARKS_SIDEBAR)
	    view = bookmarks_view;
	  else if (item == HISTORY_SIDEBAR)
	    view = history_view;
	  else if (item == POST_HISTORY_SIDEBAR)
	    view = post_history_view;
	  else
	    view = NULL;

	  if (view != NULL)
	    {
	      gint page = mod_notebook_page_num(sidebar_book,
						view->get_widget());
	      if (page != -1)
		mod_notebook_set_current_page(sidebar_book, page);

	      view->grab_focus();
	      notify_current_focus_view(view);
	    }
	}
    }

  gtk_widget_show(widget);
  gtk_paned_add1(GTK_PANED(main_container), widget);
  update_sidebar_view();



  widget = view_manager.get_widget();
  gtk_widget_show(widget);
  gtk_paned_add2(GTK_PANED(main_container), widget);


  statusbar = GTK_STATUSBAR(gtk_statusbar_new());
  //gtk_statusbar_set_has_resize_grip(statusbar, FALSE);


  // СϢ
  searchbar = ochusha_searchbar_new();
  gtk_widget_grab_focus(searchbar);
  gtk_box_pack_start(main_box, searchbar, FALSE, FALSE, 0);

  OchushaSearchbar *sb = OCHUSHA_SEARCHBAR(searchbar);
  new EntryFocusNotifier(controller, sb->entry, true);

  ochusha_searchbar_set_enable_incremental_search(sb, TRUE);
  GtkAction *find_again
    = gtk_action_group_get_action(menubar_action_group, "FindAgain");
#if GTK_CHECK_VERSION(2,18,0)
  gtk_action_set_sensitive(find_again,
			   gtk_widget_is_sensitive(sb->find_next_button));
#else
  gtk_action_set_sensitive(find_again,
			   GTK_WIDGET_IS_SENSITIVE(sb->find_next_button));
#endif
  g_signal_connect_swapped(find_again, "activate",
			   reinterpret_cast<GCallback>(gtk_button_clicked),
			   sb->find_next_button);
  g_signal_connect(sb->find_next_button, "notify::sensitive",
		   reinterpret_cast<GCallback>(find_again_notify_sensitive_cb),
		   find_again);

  int b;
  value = controller.get_session().get_state("Searchbar::match_case");
  if (value != NULL && sscanf(value, "%d", &b) == 1)
    gtk_toggle_button_set_active(sb->match_case_button, b);

  value = controller.get_session().get_state("Searchbar::use_regexp");
  if (value != NULL && sscanf(value, "%d", &b) == 1)
    gtk_toggle_button_set_active(sb->use_regexp_button, b);

  g_signal_connect(sb->match_case_button, "toggled",
		   reinterpret_cast<GCallback>(match_case_toggled_cb), this);
  g_signal_connect(sb->use_regexp_button, "toggled",
		   reinterpret_cast<GCallback>(use_regexp_toggled_cb), this);

  searchbar_context_id
    = gtk_statusbar_get_context_id(statusbar, "searchbar-target");
  boardlist_context_id
    = gtk_statusbar_get_context_id(statusbar, "boardlist-info");
  link_context_id
    = gtk_statusbar_get_context_id(statusbar, "link-target");
  network_context_id
    = gtk_statusbar_get_context_id(statusbar, "network-access-status");


  GDK_THREADS_LEAVE();

  controller.get_broker()
    .notify_access_status.connect(this, &MainWindow::notify_access_status_cb);

  GDK_THREADS_ENTER();


  // ơС
  status_box = GTK_BOX(gtk_hbox_new(FALSE, 0));


  widget = gtk_event_box_new();
#if GTK_CHECK_VERSION(2,12,0)
  gtk_widget_set_tooltip_text(widget, _("Work Offline"));
#endif
  g_signal_connect(widget, "button_press_event",
		   reinterpret_cast<GCallback>(connection_image_pressed_cb),
		   &controller);

  connection_image
    = gtk_image_new_from_stock(controller.get_config().get_offline()
			       ? GTK_STOCK_DISCONNECT : GTK_STOCK_CONNECT,
			       GTK_ICON_SIZE_MENU);
  gtk_widget_show(connection_image);
  gtk_container_add(GTK_CONTAINER(widget), connection_image);
  gtk_event_box_set_visible_window(GTK_EVENT_BOX(widget), FALSE);
  gtk_widget_show(widget);
  gtk_box_pack_start(status_box, widget, FALSE, FALSE, 2);


  widget = gtk_event_box_new();
#if GTK_CHECK_VERSION(2,12,0)
  gtk_widget_set_tooltip_text(widget, _("Embeded Response Editor"));
#endif
  g_signal_connect(widget, "button_press_event",
		   reinterpret_cast<GCallback>(edit_image_pressed_cb),
		   controller.prefs.enable_embeded_response_editor);

  g_signal_connect_swapped(view_manager.thread_pane, "hide",
			   reinterpret_cast<GCallback>(gtk_widget_hide),
			   widget);
  g_signal_connect_swapped(view_manager.thread_pane, "show",
			   reinterpret_cast<GCallback>(gtk_widget_show),
			   widget);

  edit_image
    = gtk_image_new_from_stock(controller.prefs.get_enable_embeded_response_editor()
			       ? OCHUSHA_STOCK_EDIT_ENABLED
			       : OCHUSHA_STOCK_EDIT_DISABLED,
			       GTK_ICON_SIZE_MENU);


  gtk_widget_show(edit_image);
  gtk_container_add(GTK_CONTAINER(widget), edit_image);
  gtk_event_box_set_visible_window(GTK_EVENT_BOX(widget), FALSE);
  //gtk_widget_show(widget);
  gtk_box_pack_end(GTK_BOX(statusbar), widget, FALSE, FALSE, 0);


  gtk_widget_show(GTK_WIDGET(statusbar));
  gtk_box_pack_start(status_box, GTK_WIDGET(statusbar), TRUE, TRUE, 0);
  gtk_widget_show(GTK_WIDGET(status_box));
  gtk_box_pack_end(main_box, GTK_WIDGET(status_box), FALSE, FALSE, 0);

  g_signal_connect(controller.work_offline_action, "toggled",
		   reinterpret_cast<GCallback>(work_offline_toggled_cb), this);
  g_signal_connect(controller.prefs.enable_embeded_response_editor, "toggled",
		   reinterpret_cast<GCallback>(embeded_response_editor_toggled_cb),
		   this);

  gtk_widget_show(GTK_WIDGET(main_box));
}


MainWindow::~MainWindow()
{
  if (process_pending_status_on_idle_id != 0)
    g_source_remove(process_pending_status_on_idle_id);


  if (stable_icon != NULL)
    g_object_unref(stable_icon);

  progress_animation_disable();

  if (progress_icon != NULL)
    g_object_unref(progress_icon);

  g_object_unref(current_view_action_group);
  g_object_unref(menubar_action_group);
}


void
MainWindow::progress_animation_disable()
{
  if (progress_icon_frame_timeout_id != 0)
    {
      g_source_remove(progress_icon_frame_timeout_id);
      progress_icon_frame_timeout_id = 0;
    }

  if (progress_icon_iter != NULL)
    {
      g_object_unref(progress_icon_iter);
      progress_icon_iter = NULL;
    }

  if (toplevel != NULL)
    {
      gtk_widget_queue_draw(menubar);

#if GTK_CHECK_VERSION(2,18,0)
      if (gtk_widget_is_drawable(menubar))
	gdk_window_process_updates(menubar->window, TRUE);
#else
      if (GTK_WIDGET_DRAWABLE(menubar))
	gdk_window_process_updates(menubar->window, TRUE);
#endif
    }
}


void
MainWindow::progress_animation_enable()
{
  g_return_if_fail(progress_icon != NULL);

  if (progress_icon_iter == NULL)
    {
      progress_icon_iter = gdk_pixbuf_animation_get_iter(progress_icon, NULL);
      g_return_if_fail(progress_icon_frame_timeout_id == 0);

      int delay = gdk_pixbuf_animation_iter_get_delay_time(progress_icon_iter);
      if (delay >= 0)
	progress_icon_frame_timeout_id
	  = g_timeout_add(delay,
			  reinterpret_cast<GSourceFunc>(progress_animation_timeout_cb),
			  this);
    }

  if (toplevel != NULL)
    {
      gtk_widget_queue_draw(menubar);

#if GTK_CHECK_VERSION(2,18,0)
      if (gtk_widget_is_drawable(menubar))
	gdk_window_process_updates(menubar->window, TRUE);
#else
      if (GTK_WIDGET_DRAWABLE(menubar))
	gdk_window_process_updates(menubar->window, TRUE);
#endif
    }
}


gboolean
MainWindow::menubar_expose_event_cb(GtkWidget *widget, GdkEventExpose *event,
				    MainWindow *window)
{
#if GTK_CHECK_VERSION(2,18,0)
  if (!gtk_widget_get_visible(widget) || !gtk_widget_get_mapped(widget))
    return FALSE;
#else
  if (!GTK_WIDGET_VISIBLE(widget) || !GTK_WIDGET_MAPPED(widget))
    return FALSE;
#endif

  GdkWindow *gdk_window = widget->window;
  GtkAllocation &allocation = widget->allocation;
  GdkPixbuf *pixbuf;
  if (window->progress_icon_iter != NULL)
    pixbuf = gdk_pixbuf_animation_iter_get_pixbuf(window->progress_icon_iter);
  else
    pixbuf = window->stable_icon;

  int width = gdk_pixbuf_get_width(pixbuf);
  int height = gdk_pixbuf_get_width(pixbuf);

  if (allocation.width < width)
    width = allocation.width;

  if (allocation.height < height)
    height = allocation.height;

  g_return_val_if_fail(width > 0 && height > 0, FALSE);

  int xadj;
  int yadj;
  if (height <= width)
    {
      yadj = (allocation.height - height) / 2 + 1;
      xadj = 1;
    }
  else
    {
      yadj = 1;
      xadj = (height - width) / 2 + 1;
    }

  gint dest_x;
  if (gtk_widget_get_direction(window->menubar) != GTK_TEXT_DIR_RTL)
    dest_x = allocation.x + allocation.width - (width + xadj);
  else
    dest_x = allocation.x + xadj;

  gint dest_y = allocation.y + yadj;

  if (pixbuf != NULL)
    gdk_draw_pixbuf(gdk_window,
		    widget->style->black_gc,		/* Τ */
		    pixbuf,
		    0, 0,				/* src_x, src_y */
		    dest_x, dest_y,
		    width, height,
		    GDK_RGB_DITHER_NONE, 0, 0);

  return FALSE;
}


gboolean
MainWindow::progress_animation_timeout_cb(MainWindow *window)
{
  GDK_THREADS_ENTER();

  GdkPixbufAnimationIter *iter = window->progress_icon_iter;
  if (iter != NULL)
    {
      gdk_pixbuf_animation_iter_advance(iter, NULL);
      int delay = gdk_pixbuf_animation_iter_get_delay_time(iter);

      if (delay >= 0)
	window->progress_icon_frame_timeout_id
	  = g_timeout_add(delay,
			  reinterpret_cast<GSourceFunc>(progress_animation_timeout_cb),
			  window);

      gtk_widget_queue_draw(window->menubar);

#if GTK_CHECK_VERSION(2,18,0)
      if (gtk_widget_is_drawable(window->menubar))
	gdk_window_process_updates(window->menubar->window, TRUE);
#else
      if (GTK_WIDGET_DRAWABLE(window->menubar))
	gdk_window_process_updates(window->menubar->window, TRUE);
#endif
    }

  GDK_THREADS_LEAVE();


  return FALSE;
}


bool
MainWindow::url_entry_grab_focus()
{
#if GTK_CHECK_VERSION(2,18,0)
  if (url_entry == NULL || !gtk_widget_get_mapped(GTK_WIDGET(url_entry)))
    return false;
#else
  if (url_entry == NULL || !GTK_WIDGET_MAPPED(GTK_WIDGET(url_entry)))
    return false;
#endif

  gtk_widget_grab_focus(GTK_WIDGET(url_entry));
  return true;
}


void
MainWindow::search_response_cb(GtkWidget *widget, int response_id,
			       MainWindow *window)
{
  assert(window->search_target != NULL);
  assert(window->search_message_id != 0);
  assert(window->search_response_id != 0);
  assert(window->search_request_id != 0);

  if (response_id == GTK_RESPONSE_CANCEL)
    {
      gtk_widget_hide(widget);
      gtk_statusbar_remove(window->statusbar, window->searchbar_context_id,
			   window->search_message_id);
      g_signal_handler_disconnect(widget, window->search_response_id);
      g_signal_handler_disconnect(widget, window->search_request_id);

      View *target = window->search_target;
      window->search_target = NULL;
      window->search_message_id = 0;
      window->search_response_id = 0;
      window->search_request_id = 0;

      if (target == window->boardlist_view)
	{
	  bool shown = (window->shown_sidebar_items & BOARDLIST_SIDEBAR) != 0;
	  window->controller.get_prefs().set_show_boardlist_sidebar(shown);
	}
      else if (target == window->bookmarks_view)
	{
	  bool shown = (window->shown_sidebar_items & BOOKMARKS_SIDEBAR) != 0;
	  window->controller.get_prefs().set_show_bookmarks_sidebar(shown);
	}
      else if (target == window->history_view)
	{
	  bool shown = (window->shown_sidebar_items & HISTORY_SIDEBAR) != 0;
	  window->controller.get_prefs().set_show_history_sidebar(shown);
	}
      else if (target == window->post_history_view)
	{
	  bool shown = (window->shown_sidebar_items & POST_HISTORY_SIDEBAR) != 0;
	  window->controller.get_prefs().set_show_post_history_sidebar(shown);
	}

      if (window->current_focus_view != NULL)
	window->current_focus_view->grab_focus();
    }
}


static OchushaSearchable *
search_request_cb(gpointer unused, View *view)
{
  return view->get_search_target();
}


void
MainWindow::setup_search(View *target)
{
  if (target == search_target)
    {
      ochusha_searchbar_grab_focus(OCHUSHA_SEARCHBAR(searchbar));
      return;
    }

  if (search_target != NULL)
    ochusha_searchbar_response(OCHUSHA_SEARCHBAR(searchbar),
			       GTK_RESPONSE_CANCEL);

  if (target == boardlist_view)
    {
      gtk_widget_show(boardlist_view->get_widget());
      shown_sidebar_items |= BOARDLIST_SIDEBAR;
      gint page = mod_notebook_page_num(sidebar_book,
					boardlist_view->get_widget());
      assert(page != -1);

      mod_notebook_set_current_page(sidebar_book, page);
    }
  else if (target == bookmarks_view)
    {
      gtk_widget_show(bookmarks_view->get_widget());
      shown_sidebar_items |= BOOKMARKS_SIDEBAR;
      gint page = mod_notebook_page_num(sidebar_book,
					bookmarks_view->get_widget());
      assert(page != -1);

      mod_notebook_set_current_page(sidebar_book, page);
    }
  else if (target == history_view)
    {
      gtk_widget_show(history_view->get_widget());
      shown_sidebar_items |= HISTORY_SIDEBAR;
      gint page = mod_notebook_page_num(sidebar_book,
					history_view->get_widget());
      assert(page != -1);

      mod_notebook_set_current_page(sidebar_book, page);
    }
  else if (target == post_history_view)
    {
      gtk_widget_show(post_history_view->get_widget());
      shown_sidebar_items |= POST_HISTORY_SIDEBAR;
      gint page = mod_notebook_page_num(sidebar_book,
					post_history_view->get_widget());
      assert(page != -1);

      mod_notebook_set_current_page(sidebar_book, page);
    }

  assert(search_target == NULL);
  assert(search_message_id == 0);
  assert(search_response_id == 0);
  assert(search_request_id == 0);

  search_target = target;
  search_response_id
    = g_signal_connect(searchbar, "response",
		       reinterpret_cast<GCallback>(search_response_cb), this);
  search_request_id
    = g_signal_connect(searchbar, "search_request",
		       reinterpret_cast<GCallback>(search_request_cb), target);
  search_message_id
    = gtk_statusbar_push(statusbar, searchbar_context_id,
			 target->get_search_title());

  gtk_widget_show(searchbar);
  ochusha_searchbar_set_key_selected(OCHUSHA_SEARCHBAR(searchbar));
}


void
MainWindow::prepare_shutdown()
{
  if (toplevel != NULL)
    gtk_widget_destroy(GTK_WIDGET(toplevel));

  g_return_if_fail(toplevel == NULL);
}


GtkActionGroup *
MainWindow::create_menubar_action_group()
{
  static GtkActionEntry action_entries[] = {
    {
      "FileMenu", NULL,
      N_("_File"), NULL,
      NULL,
      NULL,
    },
    {
      "EditMenu", NULL,
      N_("_Edit"), NULL,
      NULL,
      NULL,
    },
    {
      "ViewMenu", NULL,
      N_("_View"), NULL,
      NULL,
      NULL,
    },
    {
      "GoMenu", NULL,
      N_("_Go"), NULL,
      NULL,
      NULL,
    },
    {
      "BoardlistMenu", NULL,
      N_("Boar_dlist"), NULL,
      N_("Boardlist"),
      NULL,
    },
    {
      "BookmarksMenu", NULL,
      N_("_Bookmarks"), NULL,
      N_("Bookmarks"),
      NULL,
    },
    {
      "ToolsMenu", NULL,
      N_("_Tools"), NULL,
      N_("Tools"),
      NULL,
    },
    {
      "HelpMenu", NULL,
      N_("_Help"), NULL,
      NULL,
      NULL,
    },
    {
      "ViewToolbarMenu", NULL,
      N_("_Toolbar"), NULL,
      NULL,
      NULL,
    },
    {
      "BoardlistViewMenu", NULL,
      N_("Boar_dlist"), NULL,
      N_("Boardlist"),
      NULL,
    },
    {
      "ThreadlistViewMenu", NULL,
      N_("Thread_list"), NULL,
      N_("Threadlist"),
      NULL
    },
    {
      "ThreadViewMenu", NULL,
      N_("_Thread"), NULL,
      N_("Thread"),
      NULL
    },
    {
      "ViewSidebarMenu", NULL,
      N_("Sid_ebar"), NULL,
      NULL,
      NULL,
    },
    {
      "FontsMenu", NULL,
      N_("_Fonts"), NULL,
      NULL,
      NULL,
    },
    {
      "FindAgain", NULL,
      N_("Find A_gain"), NULL,
      N_("Find Again"),
      NULL,
    },
  };

  GtkActionGroup *action_group = gtk_action_group_new("MenubarActions");
  gtk_action_group_set_translation_domain(action_group, PACKAGE_NAME);
  gtk_action_group_add_actions(action_group,
			       action_entries, G_N_ELEMENTS(action_entries),
			       this);


  /* GtkUIManager˻Ź */
  gtk_ui_manager_insert_action_group(controller.get_ui_manager(),
				     action_group, 0);

  return action_group;
}


GtkActionGroup *
MainWindow::create_current_view_action_group()
{
  static GtkActionEntry action_entries[] = {
    {
      "BookmarkThisPage", OCHUSHA_STOCK_ADD_BOOKMARK,
      N_("_Bookmark This Page"), NULL,
      N_("Bookmark This Page"),
      reinterpret_cast<GCallback>(bookmark_this_page_cb)
    },
    {
      "CloseThisPage", GTK_STOCK_CLOSE,
      N_("_Close This Page"), "",
      N_("Close This Page"),
      reinterpret_cast<GCallback>(close_this_page_cb)
    },
    {
      "FindInThisPage", GTK_STOCK_FIND,
      N_("_Find in This Page"), "",
      N_("Find in This Page"),
      reinterpret_cast<GCallback>(find_in_this_page_cb)
    },
    {
      "GoBack", GTK_STOCK_GO_BACK,
      N_("_Back"), "",
      N_("Back"),
      reinterpret_cast<GCallback>(go_back_cb)
    },
    {
      "GoForward", GTK_STOCK_GO_FORWARD,
      N_("_Forward"), "",
      N_("Forward"),
      reinterpret_cast<GCallback>(go_forward_cb)
    },
    {
#ifdef GTK_STOCK_INFO
      "PageInfo", GTK_STOCK_INFO,
#else
      "PageInfo", GTK_STOCK_PROPERTIES,
#endif
      N_("Page _Info"), "",
      N_("Page Info"),
      reinterpret_cast<GCallback>(page_info_cb)
    },
    {
      "RefreshThisPage", GTK_STOCK_REFRESH,
      N_("_Refresh This Page"), "",
      N_("Refresh This Page"),
      reinterpret_cast<GCallback>(refresh_this_page_cb)
    }
  };

  GtkActionGroup *action_group = gtk_action_group_new("CurrentViewActions");
  gtk_action_group_set_translation_domain(action_group, PACKAGE_NAME);
  gtk_action_group_add_actions(action_group,
			       action_entries,
			       G_N_ELEMENTS(action_entries),
			       this);

  bookmark_this_page
    = gtk_action_group_get_action(action_group, "BookmarkThisPage");
  gtk_action_set_sensitive(bookmark_this_page, FALSE);

  close_this_page
    = gtk_action_group_get_action(action_group, "CloseThisPage");
  gtk_action_set_sensitive(close_this_page, FALSE);

  find_in_this_page
    = gtk_action_group_get_action(action_group, "FindInThisPage");
  gtk_action_set_sensitive(find_in_this_page, FALSE);

  go_back = gtk_action_group_get_action(action_group, "GoBack");
  gtk_action_set_sensitive(go_back, FALSE);

  go_forward = gtk_action_group_get_action(action_group, "GoForward");
  gtk_action_set_sensitive(go_forward, FALSE);

  page_info = gtk_action_group_get_action(action_group, "PageInfo");
  gtk_action_set_sensitive(page_info, FALSE);

  refresh_this_page
    = gtk_action_group_get_action(action_group, "RefreshThisPage");
  gtk_action_set_sensitive(refresh_this_page, FALSE);


  /* GtkUIManager˻Ź */
  gtk_ui_manager_insert_action_group(controller.get_ui_manager(),
				     action_group, 0);

  return action_group;
}


void
MainWindow::setup_main_ui()
{
  char filename[PATH_MAX];
  GError *error = NULL;
  if (controller.get_repository().find_file(OCHUSHA_UI_XML,
					    filename, PATH_MAX) == NULL
      || !gtk_ui_manager_add_ui_from_file(controller.get_ui_manager(),
					  filename, &error))
    {
      if (error != NULL)
	{
	  g_message(_("Building UI from file \"%s\" failed: %s\nFallback to built-in UI.\n"),
		    filename, error->message);
	  g_error_free(error);
	  error = NULL;
	}

      if (!gtk_ui_manager_add_ui_from_string(controller.get_ui_manager(),
					     default_ui, -1, &error))
	{
	  g_message(_("*FATAL* Building built-in UI failed: %s\n"),
		    error->message);
	  g_error_free(error);
	  exit(EXIT_FAILURE);
	}
    }

  controller.boardlist_ui.setup_menubar();
  controller.bookmarks_ui.setup_menubar();
}


void
MainWindow::bookmark_this_page_cb(GtkAction *action, MainWindow *window)
{
  g_return_if_fail(window->current_focus_view != NULL);
  g_return_if_fail(window->current_focus_view->bookmarkable());

  window->current_focus_view->bookmark();
}


void
MainWindow::close_this_page_cb(GtkAction *action, MainWindow *window)
{
  g_return_if_fail(window->current_focus_view != NULL);

  View *view = window->current_focus_view;
  if (view == window->boardlist_view)
    {
      window->controller.get_prefs().set_show_boardlist_sidebar(false);
      return;
    }
  else if (view == window->bookmarks_view)
    {
      window->controller.get_prefs().set_show_bookmarks_sidebar(false);
      return;
    }
  else if (view == window->history_view)
    {
      window->controller.get_prefs().set_show_history_sidebar(false);
      return;
    }
  else if (view == window->post_history_view)
    {
      window->controller.get_prefs().set_show_post_history_sidebar(false);
      return;
    }
    
  window->view_manager.close_tab(view);
}


void
MainWindow::find_in_this_page_cb(GtkAction *action, MainWindow *window)
{
  g_return_if_fail(window->current_focus_view != NULL);

  window->setup_search(window->current_focus_view);
}


void
MainWindow::go_back_cb(GtkAction *action, MainWindow *window)
{
  g_return_if_fail(window->current_focus_view != NULL);
  g_return_if_fail(window->current_focus_view->has_back_view());

  window->current_focus_view->go_back_view();
}


void
MainWindow::go_forward_cb(GtkAction *action, MainWindow *window)
{
  g_return_if_fail(window->current_focus_view != NULL);
  g_return_if_fail(window->current_focus_view->has_forward_view());

  window->current_focus_view->go_forward_view();
}


void
MainWindow::page_info_cb(GtkAction *action, MainWindow *window)
{
  g_return_if_fail(window->current_focus_view != NULL);
  g_return_if_fail(window->current_focus_view->has_info());

  window->current_focus_view->show_info();
}


void
MainWindow::refresh_this_page_cb(GtkAction *action, MainWindow *window)
{
  g_return_if_fail(window->current_focus_view != NULL);
  g_return_if_fail(window->current_focus_view->refreshable());

  window->current_focus_view->refresh();
}


void
MainWindow::notify_position_set_cb(GtkPaned *paned, GParamSpec *arg1,
				   MainWindow *window)
{
  char value[64];
  snprintf(value, 64, "%d", gtk_paned_get_position(paned));
  window->controller.get_session().set_state("MainWindow::sidebar_width",
					     value);
}


void
MainWindow::match_case_toggled_cb(GtkToggleButton *button, MainWindow *window)
{
  char value[64];
  snprintf(value, 64, "%d", gtk_toggle_button_get_active(button));
  window->controller.get_session().set_state("Searchbar::match_case", value);
}


void
MainWindow::use_regexp_toggled_cb(GtkToggleButton *button, MainWindow *window)
{
  char value[64];
  snprintf(value, 64, "%d", gtk_toggle_button_get_active(button));
  window->controller.get_session().set_state("Searchbar::use_regexp", value);
}


void
MainWindow::delete_event_cb(GtkWidget *widget, GdkEventAny *event,
			    MainWindow *window)
{
  window->controller.quit();
}


gboolean
MainWindow::window_state_event_cb(GtkWidget *widget, GdkEventWindowState *event,
				  MainWindow *window)
{
  int changed_mask = event->changed_mask;
  if ((changed_mask & static_cast<int>(GDK_WINDOW_STATE_MAXIMIZED)) == 0)
    return FALSE;

  int new_state = event->new_window_state;
  window->maximized
    = (new_state & static_cast<int>(GDK_WINDOW_STATE_MAXIMIZED)) != 0;

  char value[64];
  snprintf(value, 64, "%d", window->maximized);

  window->controller.get_session().set_state("MainWindow::maximized", value);

  return FALSE;
}


void
MainWindow::show_navigation_toolbar_toggled_cb(GtkToggleAction *action,
					       MainWindow *window)
{
  g_return_if_fail(window->navigation_toolbar != NULL);

  if (gtk_toggle_action_get_active(action))
    gtk_widget_show(window->navigation_toolbar);
  else
    gtk_widget_hide(window->navigation_toolbar);
}


gboolean
MainWindow::connection_image_pressed_cb(GtkWidget *event_box,
					GdkEventButton *event,
					Controller *controller)
{
  gboolean val = !gtk_toggle_action_get_active(controller->work_offline_action);
  gtk_toggle_action_set_active(controller->work_offline_action, val);

  return FALSE;
}


void
MainWindow::work_offline_toggled_cb(GtkToggleAction *action,
				    MainWindow *window)
{
  if (window->connection_image != NULL)
    gtk_image_set_from_stock(GTK_IMAGE(window->connection_image),
			     gtk_toggle_action_get_active(action)
			     ? GTK_STOCK_DISCONNECT : GTK_STOCK_CONNECT,
			     GTK_ICON_SIZE_MENU);
}


gboolean
MainWindow::edit_image_pressed_cb(GtkWidget *event_box,
				  GdkEventButton *event,
				  GtkToggleAction *action)
{
  gboolean value = !gtk_toggle_action_get_active(action);
  gtk_toggle_action_set_active(action, value);

  return FALSE;
}


void
MainWindow::embeded_response_editor_toggled_cb(GtkToggleAction *action,
					       MainWindow *window)
{
  if (window->edit_image != NULL)
    gtk_image_set_from_stock(GTK_IMAGE(window->edit_image),
			     gtk_toggle_action_get_active(action)
			     ? OCHUSHA_STOCK_EDIT_ENABLED
			     : OCHUSHA_STOCK_EDIT_DISABLED,
			     GTK_ICON_SIZE_MENU);

  window->controller.get_prefs().set_prefer_embeded_response_editor(true);
}


gboolean
MainWindow::sidebar_button_press_event_cb(ModNotebook *notebook,
					  GdkEventButton *event,
					  MainWindow *window)
{
  if (event->button != 2)
    return FALSE;

  gint page = mod_notebook_get_mouse_event_page_num(window->sidebar_book,
					reinterpret_cast<GdkEventAny *>(event));
  if (page != -1)
    {
      GtkWidget *widget = mod_notebook_get_nth_page(window->sidebar_book, page);
      if (widget == window->boardlist_view->get_widget())
	window->controller.get_prefs().set_show_boardlist_sidebar(false);
      else if (widget == window->bookmarks_view->get_widget())
	window->controller.get_prefs().set_show_bookmarks_sidebar(false);
      else if (widget == window->history_view->get_widget())
	window->controller.get_prefs().set_show_history_sidebar(false);
      else if (widget == window->post_history_view->get_widget())
	window->controller.get_prefs().set_show_post_history_sidebar(false);
      return TRUE;
    }
  return FALSE;
}


void
MainWindow::url_entry_activate_cb(GtkEntry *entry, MainWindow *window)
{
  const gchar *url = gtk_entry_get_text(entry);
  bool in_tab = window->controller.get_prefs().get_default_open_in_tab();
  window->controller.open(url, true, in_tab, false);
}


void
MainWindow::update_sidebar_view()
{
  if ((shown_sidebar_items & SHOWN_SIDEBAR_MASK) != 0)
    {
      gtk_widget_show(GTK_WIDGET(sidebar_book));

      if ((shown_sidebar_items & BOARDLIST_SIDEBAR) != 0)
	gtk_widget_show(boardlist_view->get_widget());
      else
	gtk_widget_hide(boardlist_view->get_widget());

      if ((shown_sidebar_items & BOOKMARKS_SIDEBAR) != 0)
	gtk_widget_show(bookmarks_view->get_widget());
      else
	gtk_widget_hide(bookmarks_view->get_widget());

#if 0	// not exist yet.
      if ((shown_sidebar_items & HISTORY_SIDEBAR) != 0)
	gtk_widget_show(history_view->get_widget());
      else
	gtk_widget_hide(history_view->get_widget());

      if ((shown_sidebar_items & POST_HISTORY_SIDEBAR) != 0)
	gtk_widget_show(post_history_view->get_widget());
      else
	gtk_widget_hide(post_history_view->get_widget());
#endif
    }
  else
    gtk_widget_hide(GTK_WIDGET(sidebar_book));
}


void
MainWindow::show_boardlist_sidebar_toggled_cb(GtkToggleAction *action,
					      MainWindow *window)
{
  if (gtk_toggle_action_get_active(action))
    {
      gtk_widget_show(window->boardlist_view->get_widget());
      window->shown_sidebar_items |= BOARDLIST_SIDEBAR;
      gint page = mod_notebook_page_num(window->sidebar_book,
					window->boardlist_view->get_widget());
      assert(page != -1);

      mod_notebook_set_current_page(window->sidebar_book, page);

      window->boardlist_view->grab_focus();
      window->notify_current_focus_view(window->boardlist_view);
    }
  else
    {
      gtk_widget_hide(window->boardlist_view->get_widget());
      window->shown_sidebar_items &= ~BOARDLIST_SIDEBAR;

      window->notify_view_vanished(window->boardlist_view);
    }

  window->update_sidebar_view();
}


void
MainWindow::show_bookmarks_sidebar_toggled_cb(GtkToggleAction *action,
					      MainWindow *window)
{
  if (gtk_toggle_action_get_active(action))
    {
      gtk_widget_show(window->bookmarks_view->get_widget());
      window->shown_sidebar_items |= BOOKMARKS_SIDEBAR;
      gint page = mod_notebook_page_num(window->sidebar_book,
					window->bookmarks_view->get_widget());
      assert(page != -1);

      mod_notebook_set_current_page(window->sidebar_book, page);

      window->bookmarks_view->grab_focus();
      window->notify_current_focus_view(window->bookmarks_view);
    }
  else
    {
      gtk_widget_hide(window->bookmarks_view->get_widget());
      window->shown_sidebar_items &= ~BOOKMARKS_SIDEBAR;

      window->notify_view_vanished(window->bookmarks_view);
    }

  window->update_sidebar_view();
}


void
MainWindow::show_history_sidebar_toggled_cb(GtkToggleAction *action,
					    MainWindow *window)
{
  if (gtk_toggle_action_get_active(action))
    {
      gtk_widget_show(window->history_view->get_widget());
      window->shown_sidebar_items |= HISTORY_SIDEBAR;
      gint page = mod_notebook_page_num(window->sidebar_book,
					window->history_view->get_widget());
      assert(page != -1);

      mod_notebook_set_current_page(window->sidebar_book, page);

      window->history_view->grab_focus();
      window->notify_current_focus_view(window->history_view);
    }
  else
    {
      gtk_widget_hide(window->history_view->get_widget());
      window->shown_sidebar_items &= ~HISTORY_SIDEBAR;

      window->notify_view_vanished(window->history_view);
    }

  window->update_sidebar_view();
}


void
MainWindow::show_post_history_sidebar_toggled_cb(GtkToggleAction *action,
						 MainWindow *window)
{
  if (gtk_toggle_action_get_active(action))
    {
      gtk_widget_show(window->post_history_view->get_widget());
      window->shown_sidebar_items |= POST_HISTORY_SIDEBAR;
      gint page = mod_notebook_page_num(window->sidebar_book,
					window->post_history_view->get_widget());
      assert(page != -1);

      mod_notebook_set_current_page(window->sidebar_book, page);

      window->post_history_view->grab_focus();
      window->notify_current_focus_view(window->post_history_view);
    }
  else
    {
      gtk_widget_hide(window->post_history_view->get_widget());
      window->shown_sidebar_items &= POST_HISTORY_SIDEBAR;

      window->notify_view_vanished(window->post_history_view);
    }

  window->update_sidebar_view();
}


void
MainWindow::notify_view_vanished(View *view)
{
  if (toplevel == NULL)
    return;	// ɥ˴(ḽǤϤνλ)

  if (view == search_target)
    ochusha_searchbar_response(OCHUSHA_SEARCHBAR(searchbar),
			       GTK_RESPONSE_CANCEL);

  if (view == current_focus_view)
    notify_current_focus_view(NULL);

  if (view == current_main_view)
    notify_current_main_view(NULL);
}


void
MainWindow::notify_current_main_view(View *view)
{
  if (view == current_main_view)
    return;	// Ѳ̵

  current_main_view = view;

  if (view != NULL)
    {
      const gchar *title = view->get_window_title();
      if (title != NULL)
	gtk_window_set_title(toplevel, title);

      const gchar *url = view->get_url();
      if (url_entry != NULL && url != NULL)
	gtk_entry_set_text(url_entry, url);
    }
  else
    {
      gtk_window_set_title(toplevel, _("Ochusha"));

      if (url_entry != NULL)
	gtk_entry_set_text(url_entry, "");
    }
}


void
MainWindow::notify_current_focus_view(View *view)
{
  if (view == current_focus_view)
    return;	// Ѳ̵

  current_focus_view = view;

  if (view != NULL && view->get_window_title() != NULL)
    notify_current_main_view(view);

  update_actions();
}


void
MainWindow::update_actions()
{
  View *view = current_focus_view;

  if (view != NULL)
    {
      gtk_action_set_sensitive(bookmark_this_page, view->bookmarkable());
      gtk_action_set_sensitive(close_this_page, TRUE);
      gtk_action_set_sensitive(find_in_this_page, TRUE);
      gtk_action_set_sensitive(go_back, view->has_back_view());
      gtk_action_set_sensitive(go_forward, view->has_forward_view());
      gtk_action_set_sensitive(page_info, view->has_info());
      gtk_action_set_sensitive(refresh_this_page, view->refreshable());
    }
  else
    {
      gtk_action_set_sensitive(bookmark_this_page, FALSE);
      gtk_action_set_sensitive(close_this_page, FALSE);
      gtk_action_set_sensitive(find_in_this_page, FALSE);
      gtk_action_set_sensitive(go_back, FALSE);
      gtk_action_set_sensitive(go_forward, FALSE);
      gtk_action_set_sensitive(page_info, FALSE);
      gtk_action_set_sensitive(refresh_this_page, FALSE);
    }
}


void
MainWindow::notify_state_changed(View *view)
{
  if (view == current_focus_view)
    update_actions();
}


void
MainWindow::area_prepared_cb(GdkPixbufLoader *loader, MainWindow *window)
{
  if (window->progress_icon == NULL)
    {
      window->progress_icon = gdk_pixbuf_loader_get_animation(loader);
      if (window->progress_icon != NULL)
	g_object_ref(window->progress_icon);
    }
}


void
MainWindow::area_updated_cb(GdkPixbufLoader *loader, gint x, gint y,
			    gint width, gint height, MainWindow *window)
{
  if (window->progress_icon == NULL)
    {
      window->progress_icon = gdk_pixbuf_loader_get_animation(loader);
      if (window->progress_icon != NULL)
	g_object_ref(window->progress_icon);
    }
}


void
MainWindow::closed_cb(GdkPixbufLoader *loader, MainWindow *window)
{
  if (window->progress_icon == NULL)
    {
      window->progress_icon = gdk_pixbuf_loader_get_animation(loader);
      if (window->progress_icon != NULL)
	g_object_ref(window->progress_icon);
    }
}


void
MainWindow::process_pending_status()
{
  assert(pending_status_lock.is_locked_by_self());
  assert(OCHUSHA_IN_GDK_THREADS());
  assert(process_pending_status_on_idle_id == 0);

      while (!pending_status.empty())
	{
	  PendingStatus &info = pending_status.front();
	  notify_access_status(info.url.c_str(), info.status);

	  pending_status.pop_front();
	}


}


gboolean
MainWindow::process_pending_status_on_idle_cb(MainWindow *window)
{
  LockBlock lock_obj(pending_status_lock);

  if (OCHUSHA_GDK_THREADS_TRYENTER())
    {
      window->process_pending_status_on_idle_id = 0;
      window->process_pending_status();

      GDK_THREADS_LEAVE();

      return FALSE;
    }

  return TRUE;
}


void
MainWindow::notify_access_status_cb(const char *url, NetworkAccessStatus status)
{
  LockBlock lock_obj(pending_status_lock);

  pending_status.push_back(PendingStatus(url, status));

  if (OCHUSHA_GDK_THREADS_TRYENTER())
    {
      if (process_pending_status_on_idle_id != 0)
	{
	  g_source_remove(process_pending_status_on_idle_id);
	  process_pending_status_on_idle_id = 0;
	}
      process_pending_status();

      GDK_THREADS_LEAVE();
    }
  else if (process_pending_status_on_idle_id == 0)
    process_pending_status_on_idle_id
      = g_idle_add(reinterpret_cast<GSourceFunc>(process_pending_status_on_idle_cb),
		   this);
}


void
MainWindow::notify_access_status(const char *url, NetworkAccessStatus status)
{
  if (toplevel == NULL)
    return;

  list<AccessMessage>::iterator iter = access_message_list.begin();
  list<AccessMessage>::iterator end_iter = access_message_list.end();
  while (iter != end_iter)
    {
      if (strcmp((*iter).url.c_str(), url) == 0)
	break;
      iter++;
    }

  if (iter != end_iter)
    gtk_statusbar_remove(statusbar, network_context_id, (*iter).message_id);

  if (status == ACCESS_FINISHED)
    {
      g_return_if_fail(iter != end_iter);
      access_message_list.erase(iter);

      if (access_message_list.empty())
	{
	  gtk_action_set_sensitive(controller.stop_action, FALSE);

	  progress_animation_disable();
	}

      return;
    }

  URI uri(url);
  gchar message[128];

  const char *server = uri.get_server();
  if (server == NULL)
    server = "unknown";	// ꤨ͡

  switch (status)
    {
    case LOOK_UP_SERVER:
      snprintf(message, 128, _("Looking up %s..."), server);
      break;

    case CONNECT_TO_SERVER:
      snprintf(message, 128, _("Connecting to %s..."), server);
      break;

    case SEND_REQUEST:
      snprintf(message, 128, _("Sending request to %s..."), server);
      break;

    case WAIT_FOR_SERVER:
      snprintf(message, 128, _("Waiting for %s..."), server);
      break;

    case TRANSFER_DATA:
      snprintf(message, 128, _("Transferring data from %s..."), server);
      break;

    default:
      abort();	// ꤨ͡
    }

  guint message_id = gtk_statusbar_push(statusbar, network_context_id, message);

  if (iter != end_iter)
    {
      (*iter).message_id = message_id;

      return;
    }

  access_message_list.push_front(AccessMessage(url, message_id));

  if (!gtk_action_get_sensitive(controller.stop_action))
    {
      gtk_action_set_sensitive(controller.stop_action, TRUE);
      if (progress_icon != NULL)
	progress_animation_enable();
    }
}


void
MainWindow::show()
{
  gtk_widget_show(GTK_WIDGET(toplevel));
}


void
MainWindow::save_states()
{
}


void
MainWindow::prepare_session()
{
  g_return_if_fail(toplevel != NULL);

  GtkWidget *widget = GTK_WIDGET(toplevel);
  g_return_if_fail(GDK_IS_WINDOW(widget->window));

  SessionClient &session = controller.get_session();
  char buf[64];
  if (!maximized || session.get_state("MainWindow::x") == NULL)
    {
      int sx = gdk_screen_width();
      int sy = gdk_screen_height();
      int x, y;
      gdk_window_get_root_origin(widget->window, &x, &y);
      x %= sx;
      y %= sy;

      snprintf(buf, 64, "%d", x);
      session.set_state("MainWindow::x", buf);
      snprintf(buf, 64, "%d", y);
      session.set_state("MainWindow::y", buf);
      snprintf(buf, 64, "%d", widget->allocation.width);
      session.set_state("MainWindow::width", buf);
      snprintf(buf, 64, "%d", widget->allocation.height);
      session.set_state("MainWindow::height", buf);
    }

  if (shown_sidebar_items != 0)
    {
      gint page = mod_notebook_get_current_page(sidebar_book);
      g_return_if_fail(page != -1);
      GtkWidget *widget = mod_notebook_get_nth_page(sidebar_book, page);
      g_return_if_fail(widget != NULL);

      View *view = View::get_view(widget);
      g_return_if_fail(view != NULL);

      int current_sidebar_item;
      if (view == boardlist_view)
	current_sidebar_item = BOARDLIST_SIDEBAR;
      else if (view == bookmarks_view)
	current_sidebar_item = BOOKMARKS_SIDEBAR;
      else if (view == history_view)
	current_sidebar_item = HISTORY_SIDEBAR;
      else if (view == post_history_view)
	current_sidebar_item = POST_HISTORY_SIDEBAR;
      else
	current_sidebar_item = 0;

      if (current_sidebar_item != 0)
	{
	  snprintf(buf, 64, "%d", current_sidebar_item);
	  session.set_state("MainWindow::current_sidebar_item", buf);
	}
    }

  clear_tab_related_session_data();
  prepare_tab_related_session();
}


void
MainWindow::clear_tab_related_session_data()
{
  char name_buf[512];
  SessionClient &session = controller.get_session();

  int i = 0;
  while (true)
    {
      if (snprintf(name_buf, 512, "MainWindow::view_tab::%d", ++i) >= 512
	  || session.get_state(name_buf) == NULL)
	break;
      session.remove_state(name_buf);
    }

  session.remove_state("MainWindow::active_board");
  session.remove_state("MainWindow::active_thread");
}


static void
gather_board_urls(GtkWidget *widget, BulletinBoard *unused, void *data)
{
  vector<char *> &urls = *static_cast<vector<char *> *>(data);
  View *view = View::get_view(widget);
  g_return_if_fail(view != NULL);

  urls.push_back(view->get_current_view_url());
}


static void
gather_thread_urls(GtkWidget *widget, BBSThread *unused, void *data)
{
  vector<char *> &urls = *static_cast<vector<char *> *>(data);
  View *view = View::get_view(widget);
  g_return_if_fail(view != NULL);

  urls.push_back(view->get_current_view_url());
}


void
MainWindow::prepare_tab_related_session()
{
  if (!controller.get_prefs().get_restore_opened_tabs())
    return;

  vector<char *> urls;

  view_manager.foreach_board(gather_board_urls, &urls);
  view_manager.foreach_thread(gather_thread_urls, &urls);

  char name_buf[512];
  SessionClient &session = controller.get_session();

  int i = 0;
  vector<char *>::const_iterator iter = urls.begin();
  vector<char *>::const_iterator end_iter = urls.end();
  while (iter != end_iter)
    {
      if (snprintf(name_buf, 512, "MainWindow::view_tab::%d", ++i) >= 512)
	break;

      char *url = *iter++;
      session.set_state(name_buf, url);
      free(url);
    }

  View *view = view_manager.get_current_board_view();
  if (view != NULL)
    {
      char *url = view->get_current_view_url();
      if (url != NULL)
	{
	  session.set_state("MainWindow::active_board", url);
	  free(url);
	}
    }

  view = view_manager.get_current_thread_view();
  if (view != NULL)
    {
      char *url = view->get_current_view_url();
      if (url != NULL)
	{
	  session.set_state("MainWindow::active_thread", url);
	  free(url);
	}
    }
}


void
MainWindow::restore_tab_related_session()
{
  // 󳫤Ƥ֤򳫤
  // commandͳGLibΥᥤ롼⤫ƤФ롣

  // session managerˤ륻åΤߤ˸ꤹ٤Ȥ
  // ʤ겼ʤǤϤ뤬prefsˤͭ줿ˤ
  // Ȥ褦ˤΤȤݤʤΤprefs򸫤Ƥ롣
  bool really_restore = controller.get_prefs().get_restore_opened_tabs();

  // session managerˤ륻åˤϡǽʸ¤λƱ
  // ֤٤ʤΤǡĤ䥹򳫤˹٤Ǥʤ
  // ΤȤsession managerˤɤǧƤʤΤǡ
  // Ƥޤ
  bool refresh = true;

  vector<char *>::const_iterator iter = opened_tab_urls.begin();
  vector<char *>::const_iterator end_iter = opened_tab_urls.end();
  while (iter != end_iter)
    {
      char *url = *iter++;
      if (really_restore)
	controller.open(url, refresh, true, false, false);
      free(url);
    }

  opened_tab_urls.clear();

  if (active_board_url != NULL)
    {
      if (really_restore)
	controller.open(active_board_url, false, true, false, true);
      free(active_board_url);
      active_board_url = NULL;
    }

  if (active_thread_url != NULL)
    {
      if (really_restore)
	controller.open(active_thread_url, false, true, false, true);
      free(active_thread_url);
      active_thread_url = NULL;
    }
}


void
MainWindow::restore_session()
{
  // åϿͳνؤƥ󥹥ȥ饯ѡ

  // İ֥åޡΥɥСrestore_session()֤κƸ
  // Թgtk_widget_show()ƤޤΤǡ֤᤹
  update_sidebar_view();

  while (!opened_tab_urls.empty())
    {
      char *url = opened_tab_urls.back();
      free(url);
      opened_tab_urls.pop_back();
    }

  char name_buf[512];
  SessionClient &session = controller.get_session();

  int i = 0;
  while (true)
    {
      if (snprintf(name_buf, 512, "MainWindow::view_tab::%d", ++i) < 512)
	{	
	  const char * url = session.get_state(name_buf);
	  if (url != NULL)
	    {
	      opened_tab_urls.push_back(strdup(url));
	      continue;
	    }
	}
      break;
    }

  if (active_board_url != NULL)
    {
      free(active_board_url);
      active_board_url = NULL;
    }

  if (active_thread_url != NULL)
    {
      free(active_thread_url);
      active_thread_url = NULL;
    }

  const char *url = session.get_state("MainWindow::active_board");
  if (url != NULL)
    active_board_url = strdup(url);

  url = session.get_state("MainWindow::active_thread");
  if (url != NULL)
    active_thread_url = strdup(url);
}


}
