/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: preferences.cc,v 1.46 2010/08/15 15:42:20 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "preferences.h"

#include "libochusha/ochusha_intl.h"
#include "libochusha/base_utils.h"
#include "libochusha/hashtable.h"
#include "libochusha/misc_utils.h"
#include "libochusha/repository.h"
#include "libochusha/sax_context.h"
#include "libochusha/uri_utils.h"

#include "controller.h"
#include "ui_constants.h"

#include "libochushagtk/stock.h"

#include <cassert>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <string>

#include <zlib.h>


namespace Ochusha
{

Preferences::Preferences(Controller &controller)
  : controller(controller),
    changed(false),
    fg_normal(NULL),
    bg_normal(NULL),
    fg_emph(NULL),
    bg_emph(NULL),
    fg_strong(NULL),
    bg_strong(NULL),
    fg_hidden(NULL),
    bg_hidden(NULL),
    board_tab_minimum_size(DEFAULT_BOARD_TAB_MINIMUM_SIZE),
    threadlist_view_contents(strdup(DEFAULT_THREADLIST_VIEW_CONTENTS)),
    default_threadlist_rule(strdup(DEFAULT_DEFAULT_THREADLIST_RULE)),
    thread_tab_minimum_size(DEFAULT_THREAD_TAB_MINIMUM_SIZE),
    thread_view_font_name(NULL),
    initial_view_position_of_active_thread(DEFAULT_INITIAL_VIEW_POSITION_ACTIVE),
    initial_view_position_of_inactive_thread(DEFAULT_INITIAL_VIEW_POSITION_INACTIVE),
    auto_refresh_go(&auto_refresh_go_color),
    auto_refresh_wait(&auto_refresh_wait_color),
    thread_title(&thread_title_color),
    name(&name_color),
    link(&link_color),
    visited(&visited_color),
    bookmark_text(strdup(_("==== My bookmark: I have read above. ===="))),
    popup_delay(DEFAULT_POPUP_DELAY),
    popdown_delay(DEFAULT_POPDOWN_DELAY),
    maximum_number_of_popup_responses(DEFAULT_MAX_NUM_OF_POPUP_RESPONSES),
    inline_image_width(DEFAULT_INLINE_IMAGE_WIDTH),
    inline_image_height(DEFAULT_INLINE_IMAGE_HEIGHT),
    popup_image_width(DEFAULT_POPUP_IMAGE_WIDTH),
    popup_image_height(DEFAULT_POPUP_IMAGE_HEIGHT),
    capacity_of_image_cache(DEFAULT_CAPACITY_OF_IMAGE_CACHE),
    maximum_number_of_image_cache_files(DEFAULT_MAX_NUM_OF_IMAGE_CACHE_FILES),
    web_browser_template(strdup(DEFAULT_WEB_BROWSER_TEMPLATE)),
    prefs_changed(OCHUSHA_GDK_THREADS_LOCK())
{
  static GtkToggleActionEntry entries[] = {
    {
      "NavigationToolbar", NULL,
      N_("_Navigation Toolbar"), NULL,
      N_("Navigation Toolbar"),
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_SHOW_NAVIGATION_TOOLBAR
    },
    {
      "ThreadlistToolbar", NULL,
      N_("Thread_list Toolbar"), NULL,
      N_("Threadlist Toolbar"),
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_SHOW_THREADLIST_TOOLBAR
    },
    {
      "ThreadToolbar", NULL,
      N_("_Thread Toolbar"), NULL,
      N_("Thread Toolbar"),
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_SHOW_THREAD_TOOLBAR
    },
    {
      "BoardlistToolbar", NULL,
      N_("Boar_dlist Toolbar"), NULL,
      N_("Boardlist Toolbar"),
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_SHOW_BOARDLIST_TOOLBAR
    },
    {
      "BookmarksToolbar", NULL,
      N_("_Bookmarks Toolbar"), NULL,
      N_("Bookmarks Toolbar"),
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_SHOW_BOOKMARKS_TOOLBAR
    },
    {
      "BoardlistSidebar", NULL,
      N_("Boar_dlist"), NULL,
      N_("Boardlist"),
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_SHOW_BOARDLIST_SIDEBAR
    },
    {
      "BookmarksSidebar", NULL,
      N_("_Bookmarks"), NULL,
      N_("Bookmarks"),
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_SHOW_BOOKMARKS_SIDEBAR
    },
    {
      "HistorySidebar", NULL,
      N_("Hi_story"), NULL,
      N_("History"),
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_SHOW_HISTORY_SIDEBAR
    },
    {
      "PostHistorySidebar", NULL,
      N_("_Post History"), NULL,
      N_("Post History"),
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_SHOW_POST_HISTORY_SIDEBAR
    },
    {
      "ShowHiddenBoardlistItems", NULL,
      N_("Show Hidden Items"), NULL,
      NULL,
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_SHOW_HIDDEN_BOARDLIST_ITEMS
    },
    {
      "ShowHiddenThreads", NULL,
      N_("Show Hidden Threads"), NULL,
      NULL,
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_SHOW_HIDDEN_THREADS
    },
    {
      "VerticalPlacement", NULL,
      N_("Place Threadlist View Pane and Thread View Pane Vertically"), NULL,
      NULL,
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_VERTICAL_PLACEMENT
    },
    {
      "DefaultOpenInTab", NULL,
      N_("Open Pages in New Tab by Default"), NULL,
      NULL,
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_DEFAULT_OPEN_IN_TAB
    },
    {
      "SelectTabWhenOpened", NULL,
      N_("Select Tab When New Tab Opened"), NULL,
      NULL,
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_SELECT_TAB_WHEN_OPENED
    },
    {
      "FocusFollowMouse", NULL,
      N_("Focus Follow Mouse"), NULL,
      NULL,
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_FOCUS_FOLLOW_MOUSE
    },
    {
      "FocusFollowMouseModestly", NULL,
      N_("Make Modest Focus Follow Mouse"), NULL,
      NULL,
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_FOCUS_FOLLOW_MOUSE_MODESTLY
    },
    {
      "EmbededResponseEditor", OCHUSHA_STOCK_EMBEDED_EDITOR,
      N_("_Embeded Response Editor"), NULL,
      N_("Embeded Response Editor"),
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_EMBEDED_RESPONSE_EDITOR
    },
    {
      "PreferEmbededResponseEditor", OCHUSHA_STOCK_EMBEDED_EDITOR,
      N_("Prefer Embeded Response Editor"), NULL,
      NULL,
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_PREFER_EMBEDED_RESPONSE_EDITOR
    },
    {
      "ShrinkableBoardTab", NULL,
      N_("Shrinkable Board Tab"), NULL,
      NULL,
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_BOARD_TAB_SHRINKABLE
    },
    {
      "BoardTabTooltips", NULL,
      N_("Board Tab Tooltips"), NULL,
      NULL,
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_BOARD_TAB_ENABLE_TOOLTIPS
    },
    {
      "ShowBoardTabAlways", NULL,
      N_("Always Show Board Tab"), NULL,
      NULL,
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_BOARD_TAB_ALWAYS_SHOW
    },
    {
      "ShowBoardPaneAlways", NULL,
      N_("Always Show Threadlist View Pane"), NULL,
      NULL,
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_BOARD_PANE_ALWAYS_SHOW
    },
    {
      "ShrinkableThreadTab", NULL,
      N_("Shrinkable Thread Tab"), NULL,
      NULL,
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_THREAD_TAB_SHRINKABLE
    },
    {
      "ThreadTabTooltips", NULL,
      N_("Thread Tab Tooltips"), NULL,
      NULL,
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_THREAD_TAB_ENABLE_TOOLTIPS
    },
    {
      "ShowThreadTabAlways", NULL,
      N_("Always Show Thread Tab"), NULL,
      NULL,
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_THREAD_TAB_ALWAYS_SHOW
    },
    {
      "ShowThreadPaneAlways", NULL,
      N_("Always Show Thread View Pane"), NULL,
      NULL,
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_THREAD_PANE_ALWAYS_SHOW
    },
    {
      "TransparentABone", NULL,
      N_("Enable Transparent a Bone"), NULL,
      NULL,
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_ENABLE_TRANSPARENT_A_BONE
    },
    {
      "DisablePopupABone", NULL,
      N_("Disable Popup of a Boned Response"), NULL,
      NULL,
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_DISABLE_POPUP_A_BONE
    },
    {
      "DisableABone", NULL,
      N_("Disable Local a Bone"), NULL,
      NULL,
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_DISABLE_A_BONE
    },
    {
      "EnableChainABone", NULL,
      N_("Enable Chain a Bone"), NULL,
      NULL,
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_ENABLE_CHAIN_A_BONE
    },
    {
      "ShowMailto", NULL,
      N_("Show mailto in Thread View"), NULL,
      NULL,
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_SHOW_MAILTO
    },
    {
      "BeIcon", NULL,
      N_("_Be Icon"), NULL,
      N_("Be Icon"),
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_ENABLE_BE_ICON
    },
    {
      "InlineImage", NULL,
      N_("_Inline Image"), NULL,
      N_("Inline Image"),
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_ENABLE_INLINE_IMAGE
    },
    {
      "DownloadShownImagesOnly", NULL,
      N_("_Download Shown Images Only"), NULL,
      N_("Download Shown Images Only"),
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_DOWNLOAD_SHOWN_IMAGES_ONLY
    },
    {
      "HideMyBookmark", NULL,
      N_("_Suppress Automatic Insertion of My Own Bookmark"), NULL,
      NULL,
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_HIDE_MY_BOOKMARK
    },
    {
      "PopupResponse", NULL,
      N_("Enable Popup of Responses"), NULL,
      NULL,
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_ENABLE_POPUP_RESPONSE
    },
    {
      "PopupImage", NULL,
      N_("_Popup Image"), NULL,
      N_("Popup Image"),
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_ENABLE_POPUP_IMAGE
    },
    {
      "PopupTitle", NULL,
      N_("Enable Popup of Titles"), NULL,
      NULL,
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_ENABLE_POPUP_TITLE
    },
    {
      "PopupID", NULL,
      N_("Enable Popup Responses by ID"), NULL,
      NULL,
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_ENABLE_POPUP_ID
    },
    {
      "EnableProxyToRead", NULL,
      N_("Use Proxy Server to Read"), NULL,
      NULL,
      reinterpret_cast<GCallback>(enable_proxy_to_read_toggled_cb),
      FALSE
    },
    {
      "EnableProxyToPost", NULL,
      N_("Use Proxy Server to Post"), NULL,
      NULL,
      reinterpret_cast<GCallback>(enable_proxy_to_post_toggled_cb),
      FALSE
    },
    {
      "EnableProxyForImages", NULL,
      N_("Use Proxy Server to Get Images"), NULL,
      NULL,
      reinterpret_cast<GCallback>(prefs_changed_cb),
      FALSE
    },
    {
      "EnableDebugConsole", NULL,
      N_("Enable Outputs in Debug Console"), NULL,
      NULL,
      reinterpret_cast<GCallback>(prefs_changed_cb),
      FALSE,
    },
    {
      "EnableSchemeConsole", NULL,
      N_("Show Outputs in Scheme Console"), NULL,
      NULL,
      reinterpret_cast<GCallback>(prefs_changed_cb),
      FALSE,
    },
    {
      "Login2chBeAtStartup", NULL,
      N_("Login 2ch Be at Startup"), NULL,
      NULL,
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_LOGIN_2CH_BE_AT_STARTUP
    },
    {
      "Save2chBePasswd", NULL,
      N_("Save 2ch Be Password"), NULL,
      NULL,
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_SAVE_2CH_BE_PASSWD
    },
    {
      "Login2chViewerAtStartup", NULL,
      N_("Login 2ch Viewer at Startup"), NULL,
      NULL,
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_LOGIN_2CH_VIEWER_AT_STARTUP
    },
    {
      "Save2chViewerPasswd", NULL,
      N_("Save 2ch Viewer Password"), NULL,
      NULL,
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_SAVE_2CH_VIEWER_PASSWD
    },
    {
      "RestoreOpenedTabs", NULL,
      N_("Restore Opened Tab Pages as Previous Session"), NULL,
      NULL,
      reinterpret_cast<GCallback>(prefs_changed_cb),
      DEFAULT_RESTORE_OPENED_TABS
    }
  };

  action_group = gtk_action_group_new("PreferenceActions");
  gtk_action_group_set_translation_domain(action_group, PACKAGE_NAME);
  gtk_action_group_add_toggle_actions(action_group,
				      entries, G_N_ELEMENTS(entries), this);
  gtk_ui_manager_insert_action_group(controller.get_ui_manager(),
				     action_group, 0);

  show_navigation_toolbar
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "NavigationToolbar"));

  show_threadlist_toolbar
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "ThreadlistToolbar"));

  show_thread_toolbar
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "ThreadToolbar"));

  show_boardlist_toolbar
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "BoardlistToolbar"));
  gtk_action_set_sensitive(GTK_ACTION(show_boardlist_toolbar), FALSE);

  show_bookmarks_toolbar
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "BookmarksToolbar"));
  gtk_action_set_sensitive(GTK_ACTION(show_bookmarks_toolbar), FALSE);

  show_boardlist_sidebar
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "BoardlistSidebar"));

  show_bookmarks_sidebar
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "BookmarksSidebar"));

  show_history_sidebar
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "HistorySidebar"));
  gtk_action_set_sensitive(GTK_ACTION(show_history_sidebar), FALSE);

  show_post_history_sidebar
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "PostHistorySidebar"));
  gtk_action_set_sensitive(GTK_ACTION(show_post_history_sidebar), FALSE);

  show_hidden_boardlist_items
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "ShowHiddenBoardlistItems"));

  show_hidden_threads
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "ShowHiddenThreads"));

  vertical_placement
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "VerticalPlacement"));

  default_open_in_tab
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "DefaultOpenInTab"));

  select_tab_when_opened
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "SelectTabWhenOpened"));

  focus_follow_mouse
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "FocusFollowMouse"));

  focus_follow_mouse_modestly
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "FocusFollowMouseModestly"));

  enable_embeded_response_editor
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "EmbededResponseEditor"));

  prefer_embeded_response_editor
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "PreferEmbededResponseEditor"));

  board_tab_shrinkable
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "ShrinkableBoardTab"));

  board_tab_enable_tooltips
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "BoardTabTooltips"));

  board_tab_always_show
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "ShowBoardTabAlways"));

  board_pane_always_show
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "ShowBoardPaneAlways"));

  thread_tab_shrinkable
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "ShrinkableThreadTab"));

  thread_tab_enable_tooltips
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "ThreadTabTooltips"));

  thread_tab_always_show
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "ShowThreadTabAlways"));

  thread_pane_always_show
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "ShowThreadPaneAlways"));

  enable_transparent_a_bone
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "TransparentABone"));

  disable_popup_a_bone
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "DisablePopupABone"));

  disable_a_bone
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "DisableABone"));

  enable_chain_a_bone
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "EnableChainABone"));
#if 1	// until implemented.
  gtk_action_set_sensitive(GTK_ACTION(enable_chain_a_bone), FALSE);
#endif

  show_mailto
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "ShowMailto"));

  enable_be_icon
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "BeIcon"));

  enable_inline_image
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "InlineImage"));

  download_shown_images_only
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "DownloadShownImagesOnly"));

  hide_bookmark
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "HideMyBookmark"));


  enable_popup_response
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "PopupResponse"));

  enable_popup_image
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "PopupImage"));

  enable_popup_title
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "PopupTitle"));

  enable_popup_id
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "PopupID"));

  enable_proxy_to_read
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "EnableProxyToRead"));

  enable_proxy_to_post
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "EnableProxyToPost"));

  enable_proxy_for_images
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "EnableProxyForImages"));

  debug_console
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "EnableDebugConsole"));

  scheme_console
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "EnableSchemeConsole"));

  login_2ch_be_at_startup
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "Login2chBeAtStartup"));

  save_2ch_be_passwd
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "Save2chBePasswd"));

  login_2ch_viewer_at_startup
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "Login2chViewerAtStartup"));

  save_2ch_viewer_passwd
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "Save2chViewerPasswd"));

  restore_opened_tabs
    = GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						    "RestoreOpenedTabs"));


  GdkColor fg = { 0, 0, 0, 0 };
  GdkColor bg = { 0, 0xffff, 0xffff, 0xffff };

  if (strlen(DEFAULT_FG_NORMAL_COLOR) > 0
      && gdk_color_parse(DEFAULT_FG_NORMAL_COLOR, &fg_normal_color))
    set_fg_normal_color_set(true);
  else
    {
      fg_normal_color = fg;
      set_fg_normal_color_set(false);
    }

  if (strlen(DEFAULT_BG_NORMAL_COLOR) > 0
      && gdk_color_parse(DEFAULT_BG_NORMAL_COLOR, &bg_normal_color))
    set_bg_normal_color_set(true);
  else
    {
      bg_normal_color = bg;
      set_bg_normal_color_set(false);
    }

  if (strlen(DEFAULT_FG_EMPH_COLOR) > 0
      && gdk_color_parse(DEFAULT_FG_EMPH_COLOR, &fg_emph_color))
    set_fg_emph_color_set(true);
  else
    {
      fg_emph_color = fg;
      set_fg_emph_color_set(false);
    }

  if (strlen(DEFAULT_BG_EMPH_COLOR) > 0
      && gdk_color_parse(DEFAULT_BG_EMPH_COLOR, &bg_emph_color))
    set_bg_emph_color_set(true);
  else
    {
      bg_normal_color = bg;
      set_bg_emph_color_set(false);
    }

  if (strlen(DEFAULT_FG_STRONG_COLOR) > 0
      && gdk_color_parse(DEFAULT_FG_STRONG_COLOR, &fg_strong_color))
    set_fg_strong_color_set(true);
  else
    {
      fg_strong_color = fg;
      set_fg_strong_color_set(false);
    }

  if (strlen(DEFAULT_BG_STRONG_COLOR) > 0
      && gdk_color_parse(DEFAULT_BG_STRONG_COLOR, &bg_strong_color))
    set_bg_strong_color_set(true);
  else
    {
      bg_strong_color = bg;
      set_bg_strong_color_set(false);
    }

  if (strlen(DEFAULT_FG_HIDDEN_COLOR) > 0
      && gdk_color_parse(DEFAULT_FG_HIDDEN_COLOR, &fg_hidden_color))
    set_fg_hidden_color_set(true);
  else
    {
      fg_hidden_color = fg;
      set_fg_hidden_color_set(false);
    }

  if (strlen(DEFAULT_BG_HIDDEN_COLOR) > 0
      && gdk_color_parse(DEFAULT_BG_HIDDEN_COLOR, &bg_hidden_color))
    set_bg_hidden_color_set(true);
  else
    {
      bg_hidden_color = bg;
      set_bg_hidden_color_set(false);
    }

  if (gdk_color_parse(DEFAULT_AUTO_REFRESH_GO_COLOR, &auto_refresh_go_color))
    auto_refresh_go = &auto_refresh_go_color;
  else
    {
      auto_refresh_go_color = fg;
      auto_refresh_go = NULL;
    }

  if (gdk_color_parse(DEFAULT_AUTO_REFRESH_WAIT_COLOR,
		      &auto_refresh_wait_color))
    auto_refresh_wait = &auto_refresh_wait_color;
  else
    {
      auto_refresh_wait_color = fg;
      auto_refresh_wait = NULL;
    }

  if (gdk_color_parse(DEFAULT_THREAD_TITLE_COLOR, &thread_title_color))
    thread_title = &thread_title_color;
  else
    {
      thread_title_color = fg;
      thread_title = NULL;
    }

  if (gdk_color_parse(DEFAULT_NAME_COLOR, &name_color))
    name = &name_color;
  else
    {
      name_color = fg;
      name = NULL;
    }

  if (gdk_color_parse(DEFAULT_LINK_COLOR, &link_color))
    link = &link_color;
  else
    {
      link_color = fg;
      link = NULL;
    }

  if (gdk_color_parse(DEFAULT_VISITED_COLOR, &visited_color))
    visited = &visited_color;
  else
    {
      visited_color = fg;
      visited = NULL;
    }

  read_prefs_xml(controller.get_repository());
}


Preferences::~Preferences()
{
  if (threadlist_view_contents != NULL)
    free(threadlist_view_contents);
  if (default_threadlist_rule != NULL)
    free(default_threadlist_rule);
  if (thread_view_font_name != NULL)
    free(thread_view_font_name);
  if (bookmark_text != NULL)
    free(bookmark_text);
  if (web_browser_template != NULL)
    free(web_browser_template);

  g_object_unref(action_group);
}


void
Preferences::prefs_changed_cb(gpointer unused, Preferences *prefs)
{
  prefs->changed = true;
}


void
Preferences::set_fg_normal_color_set(bool val)
{
  if (val)
    {
      if (fg_normal != NULL)
	return;
      fg_normal = &fg_normal_color;
    }
  else
    {
      if (fg_normal == NULL)
	return;
      fg_normal = NULL;
    }


  prefs_changed(*this);
}


void
Preferences::set_bg_normal_color_set(bool val)
{
  if (val)
    {
      if (bg_normal != NULL)
	return;
      bg_normal = &bg_normal_color;
    }
  else
    {
      if (bg_normal == NULL)
	return;
      bg_normal = NULL;
    }


  prefs_changed(*this);
}


void
Preferences::set_fg_emph_color_set(bool val)
{
  if (val)
    {
      if (fg_emph != NULL)
	return;
      fg_emph = &fg_emph_color;
    }
  else
    {
      if (fg_emph == NULL)
	return;
      fg_emph = NULL;
    }


  prefs_changed(*this);
}


void
Preferences::set_bg_emph_color_set(bool val)
{
  if (val)
    {
      if (bg_emph != NULL)
	return;
      bg_emph = &bg_emph_color;
    }
  else
    {
      if (bg_emph == NULL)
	return;
      bg_emph = NULL;
    }


  prefs_changed(*this);
}


void
Preferences::set_fg_strong_color_set(bool val)
{
  if (val)
    {
      if (fg_strong != NULL)
	return;
      fg_strong = &fg_strong_color;
    }
  else
    {
      if (fg_strong == NULL)
	return;
      fg_strong = NULL;
    }


  prefs_changed(*this);
}


void
Preferences::set_bg_strong_color_set(bool val)
{
  if (val)
    {
      if (bg_strong != NULL)
	return;
      bg_strong = &bg_strong_color;
    }
  else
    {
      if (bg_strong == NULL)
	return;
      bg_strong = NULL;
    }


  prefs_changed(*this);
}


void
Preferences::set_fg_hidden_color_set(bool val)
{
  if (val)
    {
      if (fg_hidden != NULL)
	return;
      fg_hidden = &fg_hidden_color;
    }
  else
    {
      if (fg_hidden == NULL)
	return;
      fg_hidden = NULL;
    }


  prefs_changed(*this);
}


void
Preferences::set_bg_hidden_color_set(bool val)
{
  if (val)
    {
      if (bg_hidden != NULL)
	return;
      bg_hidden = &bg_hidden_color;
    }
  else
    {
      if (bg_hidden == NULL)
	return;
      bg_hidden = NULL;
    }


  prefs_changed(*this);
}


void
Preferences::set_board_tab_minimum_size(unsigned int val)
{
  if (board_tab_minimum_size == val)
    return;
  board_tab_minimum_size = val;


  prefs_changed(*this);
}


void
Preferences::set_thread_tab_minimum_size(unsigned int val)
{
  if (thread_tab_minimum_size == val)
    return;
  thread_tab_minimum_size = val;


  prefs_changed(*this);
}


void
Preferences::set_popup_delay(unsigned int val)
{
  if (popup_delay == val)
    return;
  popup_delay = val;


  prefs_changed(*this);
}


void
Preferences::set_popdown_delay(unsigned int val)
{
  if (popdown_delay == val)
    return;
  popdown_delay = val;


  prefs_changed(*this);
}


void
Preferences::set_maximum_number_of_popup_responses(int val)
{
  if (maximum_number_of_popup_responses == val)
    return;
  maximum_number_of_popup_responses = val;


  prefs_changed(*this);
}


void
Preferences::set_inline_image_width(int val)
{
  if (inline_image_width == val)
    return;
  inline_image_width = val;


  prefs_changed(*this);
}


void
Preferences::set_inline_image_height(int val)
{
  if (inline_image_height == val)
    return;
  inline_image_height = val;


  prefs_changed(*this);
}


void
Preferences::set_popup_image_width(int val)
{
  if (popup_image_width == val)
    return;
  popup_image_width = val;


  prefs_changed(*this);
}


void
Preferences::set_popup_image_height(int val)
{
  if (popup_image_height == val)
    return;
  popup_image_height = val;


  prefs_changed(*this);
}


void
Preferences::set_capacity_of_image_cache(uintmax_t val)
{
  if (capacity_of_image_cache == val)
    return;
  capacity_of_image_cache = val;


  prefs_changed(*this);
}


void
Preferences::set_maximum_number_of_image_cache_files(int val)
{
  if (maximum_number_of_image_cache_files == val)
    return;
  maximum_number_of_image_cache_files = val;


  prefs_changed(*this);
}


void
Preferences::set_threadlist_view_contents(const char *val)
{
  assert(threadlist_view_contents != NULL);

  if (val == NULL || *val == '\0')
    val = DEFAULT_THREADLIST_VIEW_CONTENTS;

  if (strcmp(threadlist_view_contents, val) == 0)
    return;

  free(threadlist_view_contents);
  threadlist_view_contents = strdup(val);


  prefs_changed(*this);
}


void
Preferences::set_default_threadlist_rule(const char *val)
{
  assert(default_threadlist_rule != NULL);
  if (val == NULL)
    return;

  if (strcmp(default_threadlist_rule, val) == 0)
    return;
  free(default_threadlist_rule);
  default_threadlist_rule = strdup(val);
  

  prefs_changed(*this);
}


void
Preferences::set_thread_view_font_name(const char *val)
{
  if (thread_view_font_name == val)	// NULL
    return;

  if (thread_view_font_name != NULL && val != NULL
      && strcmp(thread_view_font_name, val) == 0)
    return;

  if (thread_view_font_name != NULL)
    free(thread_view_font_name);

  thread_view_font_name = val != NULL ? strdup(val) : NULL;


  prefs_changed(*this);
}


void
Preferences::set_bookmark_text(const char *val)
{
  if (val != NULL && strcmp(bookmark_text, val) == 0)
    return;

  if (bookmark_text != NULL)
    free(bookmark_text);

  if (val == NULL || *val == '\0')
    bookmark_text = strdup(_("==== My bookmark: I have read above. ===="));
  else
    bookmark_text = strdup(val);


  prefs_changed(*this);
}


#define COMMAND_LINE_BUFFER_SIZE	4096
  /*
   * templʸ%s⡹ĤʤȤݾڤʸ
   * ԡ֤
   */
static char *
validate_command_line_template(const char *templ)
{
  char *result = strdup(templ);
  char *tmp_pos = result;
  gboolean percent_s_exist = FALSE;

  while (*tmp_pos != '\0')
    {
      if (*tmp_pos++ == '%')
	{
	  if (*tmp_pos == 's')
	    {
	      if (percent_s_exist)
		{
		  tmp_pos[-1] = ' ';
		  tmp_pos[0] = ' ';
		}
	      else
		percent_s_exist = TRUE;
	      tmp_pos++;
	    }
	}
    }
#if 0
  if (strcmp(templ, result) != 0)
    {
      fprintf(stderr, "multiple %%s isn't allowed\n");
      fprintf(stderr, "PRE validate: \"%s\"\n", templ);
      fprintf(stderr, "POST validate: \"%s\"\n", result);
    }
#endif
  return result;
}


void
Preferences::set_web_browser_template(const char *val)
{
  char *templ;
  if (val != NULL && *val != '\0')
    {
      templ = validate_command_line_template(val);
      if (strcmp(web_browser_template, templ) == 0)	// Ѳʤ
	{
	  free(templ);
	  return;
	}
    }
  else if (strcmp(web_browser_template, DEFAULT_WEB_BROWSER_TEMPLATE) == 0)
    return;
  else
    templ = strdup(DEFAULT_WEB_BROWSER_TEMPLATE);

  if (web_browser_template != NULL)
    free(web_browser_template);

  web_browser_template = templ;


  prefs_changed(*this);
}


const URI &
Preferences::get_proxy_uri_to_read() const
{
  return controller.get_config().get_proxy_uri_to_read();
}


const URI &
Preferences::get_proxy_uri_to_post() const
{
  return controller.get_config().get_proxy_uri_to_post();
}


const URI &
Preferences::get_proxy_uri_for_images() const
{
  return proxy_uri_for_images;
}


void
Preferences::enable_proxy_to_read_toggled_cb(GtkToggleAction *action,
					     Preferences *prefs)
{
  prefs->controller.get_config().set_enable_proxy_to_read(gtk_toggle_action_get_active(action));

  prefs_changed_cb(action, prefs);
}


static void
compose_proxy_uri(string &uri_buf, const char *server, int port,
		  const char *user, const char *passwd)
{
  uri_buf.clear();
  if (server != NULL && *server != '\0' && port != 0)
    {
      uri_buf.append("http://");

      if (user != NULL && *user != '\0')
	{
	  uri_buf.append(user).append(":");
	  if (passwd != NULL)
	    uri_buf.append(passwd);
	  uri_buf.append("@");
	}

      uri_buf.append(server);

      uri_buf.append(":");
      char port_buf[16];
      snprintf(port_buf, 16, "%u", port & 0xffff);
      uri_buf.append(port_buf);
    }
}


void
Preferences::set_proxy_to_read(const char *server, int port,
			       const char *user, const char *passwd)
{
  string uri_buf;
  compose_proxy_uri(uri_buf, server, port, user, passwd);

  const char *prev_uri
    = controller.get_config().get_proxy_uri_to_read().get_uri();
  if (prev_uri != NULL && strcmp(prev_uri, uri_buf.c_str()) == 0)
    return;	// Ѳ̵

  controller.get_config().set_proxy_uri_to_read(uri_buf.c_str());


  prefs_changed(*this);
}


void
Preferences::enable_proxy_to_post_toggled_cb(GtkToggleAction *action,
					     Preferences *prefs)
{
  prefs->controller.get_config().set_enable_proxy_to_post(gtk_toggle_action_get_active(action));

  prefs_changed_cb(action, prefs);
}


void
Preferences::set_proxy_to_post(const char *server, int port,
			       const char *user, const char *passwd)
{
  string uri_buf;
  compose_proxy_uri(uri_buf, server, port, user, passwd);

  const char *prev_uri
    = controller.get_config().get_proxy_uri_to_post().get_uri();
  if (prev_uri != NULL && strcmp(prev_uri, uri_buf.c_str()) == 0)
    return;	// Ѳ̵

  controller.get_config().set_proxy_uri_to_post(uri_buf.c_str());


  prefs_changed(*this);
}


void
Preferences::set_proxy_for_images(const char *server, int port,
				  const char *user, const char *passwd)
{
  string uri_buf;
  compose_proxy_uri(uri_buf, server, port, user, passwd);

  const char *prev_uri = proxy_uri_for_images.get_uri();
  if (prev_uri != NULL && strcmp(prev_uri, uri_buf.c_str()) == 0)
    return;	// Ѳ̵

  proxy_uri_for_images.parse_uri(uri_buf.c_str());


  prefs_changed(*this);
}


void
Preferences::register_recent_name(const char *val)
{
  if (val == NULL)
    return;

  list<char *>::iterator iter = recent_name_list.begin();
  list<char *>::iterator end_iter = recent_name_list.end();

  int count = 0;
  while (iter != end_iter)
    {
      ++count;
      char *name = *iter;
      if (strcmp(name, val) == 0)
	{
	  recent_name_list.erase(iter);
	  recent_name_list.push_front(name);
	  return;
	}
      ++iter;
    }

  if (count == MAXIMUM_NUMBER_OF_RECENT_NAME_LIST_ENTRIES)
    {
      char *tmp = recent_name_list.back();
      recent_name_list.pop_back();
      free(tmp);
    }
  recent_name_list.push_front(strdup(val));
}


void
Preferences::register_recent_mail(const char *val)
{
  if (val == NULL)
    return;

  list<char *>::iterator iter = recent_mail_list.begin();
  list<char *>::iterator end_iter = recent_mail_list.end();

  int count = 0;
  while (iter != end_iter)
    {
      ++count;
      char *mail = *iter;
      if (strcmp(mail, val) == 0)
	{
	  recent_mail_list.erase(iter);
	  recent_mail_list.push_front(mail);
	  return;
	}
      ++iter;
    }

  if (count == MAXIMUM_NUMBER_OF_RECENT_MAIL_LIST_ENTRIES)
    {
      char *tmp = recent_mail_list.back();
      recent_mail_list.pop_back();
      free(tmp);
    }
  recent_mail_list.push_front(strdup(val));
}


#define OUTPUT_ATTRIBUTE_INT(gzfile, attribute)				\
  gzprintf(gzfile,							\
	   "    <attribute name=\"" #attribute	"\">\n"			\
	   "      <int val=\"%d\"/>\n"					\
	   "    </attribute>\n", attribute)

#define OUTPUT_ATTRIBUTE_LONG(gzfile, attribute)			\
  gzprintf(gzfile,							\
	   "    <attribute name=\"" #attribute	"\">\n"			\
	   "      <long val=\"%d\"/>\n"					\
	   "    </attribute>\n", attribute)

#define OUTPUT_ATTRIBUTE_BOOLEAN(gzfile, attribute)			\
  gzprintf(gzfile,							\
	   "    <attribute name=\"" #attribute	"\">\n"			\
	   "      <boolean val=\"%s\"/>\n"				\
	   "    </attribute>\n",					\
	   gtk_toggle_action_get_active(attribute) ? "true" : "false" )

#define OUTPUT_ATTRIBUTE_STRING(gzfile, attribute)			\
  do									\
    {									\
      if (attribute != NULL)						\
	{								\
	  string text;							\
	  gzprintf(gzfile,						\
		   "    <attribute name=\"" #attribute	"\">\n"		\
		   "      <string>%s</string>\n"			\
		   "    </attribute>\n",				\
		   append_markup_escaped_text(text, attribute).c_str());\
	}								\
    }									\
  while (0)

#define OUTPUT_ATTRIBUTE_COLOR(gzfile, attribute)			\
  do									\
    {									\
      if (attribute != NULL)						\
	gzprintf(gzfile,						\
		 "    <attribute name=\"" #attribute "_color\">\n"	\
		 "      <string>#%02x%02x%02x</string>\n"		\
		 "    </attribute>\n",					\
		 (attribute##_color.red >> 8) & 0xff,			\
		 (attribute##_color.green >> 8) & 0xff,			\
		 (attribute##_color.blue >> 8) & 0xff);			\
      else								\
	gzprintf(gzfile,						\
		 "    <attribute name=\"" #attribute "_color\">\n"	\
		 "      <string></string>\n"				\
		 "    </attribute>\n");					\
    }									\
  while (0)

#define OUTPUT_ATTRIBUTE_URI(gzfile, attribute)				\
  do									\
    {									\
      const char *uri = (attribute).get_uri();				\
      if (uri != NULL)							\
	{								\
	  char *enc_uri = Base64Encoder::encode_text_to_base64url(uri); \
	  gzprintf(gzfile,						\
		   "    <attribute name=\"" #attribute "\">\n"		\
		   "      <string>%s</string>\n"			\
		   "    </attribute>\n", enc_uri);			\
	  free(enc_uri);						\
	}								\
    }									\
  while (0)


bool
Preferences::write_prefs_xml(Repository &repository)
{
  int fd = repository.open_file(OCHUSHA_PREFS_XML, O_WRONLY|O_TRUNC|O_CREAT);
  if (fd < 0)
    {
      cerr << "Couldn't open " OCHUSHA_PREFS_XML " to write.\n";
      return false;
    }

  gzFile prefs_xml = gzdopen(fd, "w");
  if (prefs_xml == NULL)
    {
      cerr << "Couldn't gzdopen for prefs.xml.\n";
      return false;
    }

  gzprintf(prefs_xml, "<?xml version=\"1.0\"?>\n");
  gzprintf(prefs_xml, "<ochusha>\n");
  gzprintf(prefs_xml, "  <preferences>\n");

  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, show_navigation_toolbar);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, show_threadlist_toolbar);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, show_thread_toolbar);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, show_boardlist_toolbar);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, show_bookmarks_toolbar);

  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, show_boardlist_sidebar);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, show_bookmarks_sidebar);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, show_history_sidebar);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, show_post_history_sidebar);

  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, show_hidden_boardlist_items);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, show_hidden_threads);

#if 0
  OUTPUT_ATTRIBUTE_COLOR(prefs_xml, fg_normal);
  OUTPUT_ATTRIBUTE_COLOR(prefs_xml, bg_normal);
  OUTPUT_ATTRIBUTE_COLOR(prefs_xml, fg_emph);
  OUTPUT_ATTRIBUTE_COLOR(prefs_xml, bg_emph);
  OUTPUT_ATTRIBUTE_COLOR(prefs_xml, fg_strong);
  OUTPUT_ATTRIBUTE_COLOR(prefs_xml, bg_strong);
  OUTPUT_ATTRIBUTE_COLOR(prefs_xml, fg_hidden);
  OUTPUT_ATTRIBUTE_COLOR(prefs_xml, bg_hidden);
#endif

  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, vertical_placement);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, default_open_in_tab);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, select_tab_when_opened);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, focus_follow_mouse);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, focus_follow_mouse_modestly);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, enable_embeded_response_editor);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, prefer_embeded_response_editor);

  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, board_tab_shrinkable);
  OUTPUT_ATTRIBUTE_INT(prefs_xml, board_tab_minimum_size);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, board_tab_enable_tooltips);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, board_tab_always_show);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, board_pane_always_show);

  OUTPUT_ATTRIBUTE_STRING(prefs_xml, threadlist_view_contents);
  OUTPUT_ATTRIBUTE_STRING(prefs_xml, default_threadlist_rule);

  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, thread_tab_shrinkable);
  OUTPUT_ATTRIBUTE_INT(prefs_xml, thread_tab_minimum_size);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, thread_tab_enable_tooltips);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, thread_tab_always_show);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, thread_pane_always_show);

  OUTPUT_ATTRIBUTE_STRING(prefs_xml, thread_view_font_name);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, enable_transparent_a_bone);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, disable_popup_a_bone);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, disable_a_bone);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, enable_chain_a_bone);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, show_mailto);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, enable_be_icon);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, enable_inline_image);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, download_shown_images_only);

  OUTPUT_ATTRIBUTE_INT(prefs_xml, initial_view_position_of_active_thread);
  OUTPUT_ATTRIBUTE_INT(prefs_xml, initial_view_position_of_inactive_thread);

#if 0	// ʲεϿ¾ذܤä
  OUTPUT_ATTRIBUTE_COLOR(prefs_xml, auto_refresh_go);
  OUTPUT_ATTRIBUTE_COLOR(prefs_xml, auto_refresh_wait);

  OUTPUT_ATTRIBUTE_COLOR(prefs_xml, thread_title);
  OUTPUT_ATTRIBUTE_COLOR(prefs_xml, name);
  OUTPUT_ATTRIBUTE_COLOR(prefs_xml, link);
  OUTPUT_ATTRIBUTE_COLOR(prefs_xml, visited);
#endif

  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, hide_bookmark);

  char *bookmark_text = URI::encode_string(this->bookmark_text);
  OUTPUT_ATTRIBUTE_STRING(prefs_xml, bookmark_text);
  free(bookmark_text);

  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, enable_popup_response);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, enable_popup_image);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, enable_popup_title);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, enable_popup_id);

  OUTPUT_ATTRIBUTE_INT(prefs_xml, popup_delay);
  OUTPUT_ATTRIBUTE_INT(prefs_xml, popdown_delay);

  OUTPUT_ATTRIBUTE_INT(prefs_xml, maximum_number_of_popup_responses);

  OUTPUT_ATTRIBUTE_INT(prefs_xml, inline_image_width);
  OUTPUT_ATTRIBUTE_INT(prefs_xml, inline_image_height);
  OUTPUT_ATTRIBUTE_INT(prefs_xml, popup_image_width);
  OUTPUT_ATTRIBUTE_INT(prefs_xml, popup_image_height);

  OUTPUT_ATTRIBUTE_LONG(prefs_xml, capacity_of_image_cache);
  OUTPUT_ATTRIBUTE_INT(prefs_xml, maximum_number_of_image_cache_files);

  OUTPUT_ATTRIBUTE_STRING(prefs_xml, web_browser_template);

  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, enable_proxy_for_images);
  OUTPUT_ATTRIBUTE_URI(prefs_xml, proxy_uri_for_images);

  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, debug_console);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, scheme_console);

  {
    list<char *>::iterator iter = recent_name_list.begin();
    list<char *>::iterator end_iter = recent_name_list.end();

    unsigned int count = 0;
    while (iter != end_iter
	   && count < MAXIMUM_NUMBER_OF_RECENT_NAME_LIST_ENTRIES)
      {
	string text;
	gzprintf(prefs_xml,
		 "    <attribute name=\"recent_name%u\">\n"
		 "      <string>%s</string>\n"
		 "    </attribute>\n", count,
		 append_markup_escaped_text(text, *iter).c_str());
	++iter;
	++count;
      }
  }

  {
    list<char *>::iterator iter = recent_mail_list.begin();
    list<char *>::iterator end_iter = recent_mail_list.end();

    unsigned int count = 0;
    while (iter != end_iter
	   && count < MAXIMUM_NUMBER_OF_RECENT_MAIL_LIST_ENTRIES)
      {
	string text;
	gzprintf(prefs_xml,
		 "    <attribute name=\"recent_mail%u\">\n"
		 "      <string>%s</string>\n"
		 "    </attribute>\n", count,
		 append_markup_escaped_text(text, *iter).c_str());
	++iter;
	++count;
      }
  }

  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, login_2ch_be_at_startup);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, save_2ch_be_passwd);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, login_2ch_viewer_at_startup);
  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, save_2ch_viewer_passwd);

  OUTPUT_ATTRIBUTE_BOOLEAN(prefs_xml, restore_opened_tabs);

  gzprintf(prefs_xml, "  </preferences>\n");
  gzprintf(prefs_xml, "</ochusha>\n");

  return gzclose(prefs_xml) == 0;
}


class PrefsXMLContext : public SAXContext
{
private:
  friend class Preferences;

  enum SAXState
  {
    SAX_INITIAL,
    SAX_OCHUSHA,
    SAX_PREFERENCES,
    SAX_PREFERENCES_ATTRIBUTE,
    SAX_PREFERENCES_ATTRIBUTE_BOOLEAN,
    SAX_PREFERENCES_ATTRIBUTE_INT,
    SAX_PREFERENCES_ATTRIBUTE_LONG,
    SAX_PREFERENCES_ATTRIBUTE_STRING,
    SAX_ACCEPTED,
    SAX_ERROR
  } state;

  AttributeTable attributes;

protected:
  void start_element_handler(const char *name, const char **attrs);
  void end_element_handler(const char *name);
  void characters_handler(const char *ch, size_t len);

public:
  PrefsXMLContext(int size = 101) : attributes(size)
  {
    state = SAX_INITIAL;
  }

  inline bool is_accepted()
  {
    return state == SAX_ACCEPTED;
  }
};


void
PrefsXMLContext::start_element_handler(const char *name, const char **attrs)
{
  switch (state)
    {
    case SAX_INITIAL:
      if (strcmp(name, "ochusha") == 0)
	{ state = SAX_OCHUSHA; return; }
      break;	/* 顼 */

    case SAX_OCHUSHA:
      if (strcmp(name, "preferences") == 0)
	{ state = SAX_PREFERENCES; return; }
      break;	/* 顼 */

    case SAX_PREFERENCES:
      if (strcmp(name, "attribute") == 0
	  && attrs != NULL && strcmp(attrs[0], "name") == 0)
	{
	  state = SAX_PREFERENCES_ATTRIBUTE;
	  attributes.set_attr_name(attrs[1]);
	  return;
	}
      break;	/* 顼 */

    case SAX_PREFERENCES_ATTRIBUTE:
      if (attrs != NULL && strcmp(attrs[0], "val") == 0)
	{
	  /* int/long/booleanβǽ */
	  if (strcmp(name, "int") == 0)
	    {
	      state = SAX_PREFERENCES_ATTRIBUTE_INT;
	    }
	  else if (strcmp(name, "long") == 0)
	    {
	      state = SAX_PREFERENCES_ATTRIBUTE_LONG;
	    }
	  else if (strcmp(name, "boolean") == 0)
	    {
	      state = SAX_PREFERENCES_ATTRIBUTE_BOOLEAN;
	    }
	  else
	    {
	      break;	/* 顼 */
	    }

	  attributes.set_attr_value(attrs[1]);
	  return;
	}
      else if (strcmp(name, "string") == 0)
	{
	  state = SAX_PREFERENCES_ATTRIBUTE_STRING;
	  return;
	}
      break;	/* 顼 */

    case SAX_PREFERENCES_ATTRIBUTE_STRING:
    case SAX_PREFERENCES_ATTRIBUTE_BOOLEAN:
    case SAX_PREFERENCES_ATTRIBUTE_INT:
    case SAX_ACCEPTED:
    case SAX_ERROR:
      break;	/* 顼 */

    default:
      cerr << "start_handler is called in unknown state: " << state << "\n";
    }
  state = SAX_ERROR;
}


void
PrefsXMLContext::end_element_handler(const char *name)
{
  switch (state)
    {
    case SAX_OCHUSHA:
      if (strcmp(name, "ochusha") == 0)
	{ state = SAX_ACCEPTED; return; }
      break;	/* 顼 */

    case SAX_PREFERENCES:
      if (strcmp(name, "preferences") == 0)
	{ state = SAX_OCHUSHA; return; }
      break;	/* 顼 */

    case SAX_PREFERENCES_ATTRIBUTE:
      if (strcmp(name, "attribute") == 0)
	{
	  state = SAX_PREFERENCES;
	  attributes.register_current_attribute();
	  return;
	}
      break;	/* 顼 */

    case SAX_PREFERENCES_ATTRIBUTE_STRING:
      if (strcmp(name, "string") == 0)
	{
	  state = SAX_PREFERENCES_ATTRIBUTE;
	  return;
	}
      break;	/* 顼 */

    case SAX_PREFERENCES_ATTRIBUTE_BOOLEAN:
      if (strcmp(name, "boolean") == 0)
	{ state = SAX_PREFERENCES_ATTRIBUTE; return; }
      break;	/* 顼 */

    case SAX_PREFERENCES_ATTRIBUTE_INT:
      if (strcmp(name, "int") == 0)
	{ state = SAX_PREFERENCES_ATTRIBUTE; return; }
      break;	/* 顼 */

    case SAX_PREFERENCES_ATTRIBUTE_LONG:
      if (strcmp(name, "long") == 0)
	{ state = SAX_PREFERENCES_ATTRIBUTE; return; }
      break;	/* 顼 */

    case SAX_INITIAL:
    case SAX_ACCEPTED:
    case SAX_ERROR:
      break;	/* 顼 */

    default:
#if DEBUG_SAX_HANDLER
      cerr << "endHandler called in unknown state: " << state << "\n";
#endif
      break;	/* 顼 */
    }
#if DEBUG_SAX_HANDLER
  cerr << "Invalid document: </" << name << "> appeared in state="
       << state << "\n";
#endif
  state = SAX_ERROR;
}


void
PrefsXMLContext::characters_handler(const char *ch, size_t len)
{
  if (state == SAX_PREFERENCES_ATTRIBUTE_STRING)
    attributes.append_attr_value(ch, len);
}


bool
Preferences::read_prefs_xml(Repository &repository)
{
  char *pathname = repository.find_file(OCHUSHA_PREFS_XML);
  if (pathname == NULL)
    return true;

  PrefsXMLContext context;
  context.parse_xml_file(pathname);
  free(pathname);

  if (!context.is_accepted())
    return false;

  AttributeTable &attributes = context.attributes;

  set_show_navigation_toolbar(attributes.get_bool("show_navigation_toolbar"));
  set_show_threadlist_toolbar(attributes.get_bool("show_threadlist_toolbar"));
  set_show_thread_toolbar(attributes.get_bool("show_thread_toolbar"));
  set_show_boardlist_toolbar(attributes.get_bool("show_boardlist_toolbar"));
  set_show_bookmarks_toolbar(attributes.get_bool("show_bookmarks_toolbar"));

  set_show_boardlist_sidebar(attributes.get_bool("show_boardlist_sidebar"));
  set_show_bookmarks_sidebar(attributes.get_bool("show_bookmarks_sidebar"));
  set_show_history_sidebar(attributes.get_bool("show_history_sidebar"));
  set_show_post_history_sidebar(attributes.get_bool("show_post_history_sidebar"));

  set_show_hidden_boardlist_items(attributes.get_bool("show_hidden_boardlist_items"));
  set_show_hidden_threads(attributes.get_bool("show_hidden_threads"));

  const char *color = attributes.get_string("fg_normal_color");
  if (color != NULL)
    {
      if (*color != '\0' && gdk_color_parse(color, &fg_normal_color))
	set_fg_normal_color_set(true);
      else
	set_fg_normal_color_set(false);
    }

  color = attributes.get_string("bg_normal_color");
  if (color != NULL)
    {
      if (*color != '\0' && gdk_color_parse(color, &bg_normal_color))
	set_bg_normal_color_set(true);
      else
	set_bg_normal_color_set(false);
    }

  color = attributes.get_string("fg_emph_color");
  if (color != NULL)
    {
      if (*color != '\0' && gdk_color_parse(color, &fg_emph_color))
	set_fg_emph_color_set(true);
      else
	set_fg_emph_color_set(false);
    }

  color = attributes.get_string("bg_emph_color");
  if (color != NULL)
    {
      if (*color != '\0' && gdk_color_parse(color, &bg_emph_color))
	set_bg_emph_color_set(true);
      else
	set_bg_emph_color_set(false);
    }

  color = attributes.get_string("fg_strong_color");
  if (color != NULL)
    {
      if (*color != '\0' && gdk_color_parse(color, &fg_strong_color))
	set_fg_strong_color_set(true);
      else
	set_fg_strong_color_set(false);
    }

  color = attributes.get_string("bg_strong_color");
  if (color != NULL)
    {
      if (*color != '\0' && gdk_color_parse(color, &bg_strong_color))
	set_bg_strong_color_set(true);
      else
	set_bg_strong_color_set(false);
    }

  color = attributes.get_string("fg_hidden_color");
  if (color != NULL)
    {
      if (*color != '\0' && gdk_color_parse(color, &fg_hidden_color))
	set_fg_hidden_color_set(true);
      else
	set_fg_hidden_color_set(false);
    }

  color = attributes.get_string("bg_hidden_color");
  if (color != NULL)
    {
      if (*color != '\0' && gdk_color_parse(color, &bg_hidden_color))
	set_bg_hidden_color_set(true);
      else
	set_bg_hidden_color_set(false);
    }

  set_vertical_placement(attributes.get_bool("vertical_placement"));
  set_default_open_in_tab(attributes.get_bool("default_open_in_tab"));
  set_select_tab_when_opened(attributes.get_bool("select_tab_when_opened"));
  set_focus_follow_mouse(attributes.get_bool("focus_follow_mouse"));
  set_focus_follow_mouse_modestly(attributes.get_bool("focus_follow_mouse_modestly"));
  set_enable_embeded_response_editor(attributes.get_bool("enable_embeded_response_editor"));
  set_prefer_embeded_response_editor(attributes.get_bool("prefer_embeded_response_editor"));

  set_board_tab_shrinkable(attributes.get_bool("board_tab_shrinkable"));
  set_board_tab_minimum_size(attributes.get_int("board_tab_minimum_size"));
  set_board_tab_enable_tooltips(attributes.get_bool("board_tab_enable_tooltips"));
  set_board_tab_always_show(attributes.get_bool("board_tab_always_show"));
  set_board_pane_always_show(attributes.get_bool("board_pane_always_show"));

  set_threadlist_view_contents(attributes.get_string("threadlist_view_contents"));
  set_default_threadlist_rule(attributes.get_string("default_threadlist_rule"));

  set_thread_tab_shrinkable(attributes.get_bool("thread_tab_shrinkable"));
  set_thread_tab_minimum_size(attributes.get_int("thread_tab_minimum_size"));
  set_thread_tab_enable_tooltips(attributes.get_bool("thread_tab_enable_tooltips"));
  set_thread_tab_always_show(attributes.get_bool("thread_tab_always_show"));
  set_thread_pane_always_show(attributes.get_bool("thread_pane_always_show"));

  set_thread_view_font_name(attributes.get_string("thread_view_font_name"));
  set_enable_transparent_a_bone(attributes.get_bool("enable_transparent_a_bone"));
  set_disable_popup_a_bone(attributes.get_bool("disable_popup_a_bone"));
  set_disable_a_bone(attributes.get_bool("disable_a_bone"));
  set_enable_chain_a_bone(attributes.get_bool("enable_chain_a_bone"));
  set_show_mailto(attributes.get_bool("show_mailto"));
  set_enable_be_icon(attributes.get_bool("enable_be_icon"));
  set_enable_inline_image(attributes.get_bool("enable_inline_image"));
  set_download_shown_images_only(attributes.get_bool("download_shown_images_only"));

  set_initial_view_position_of_active_thread(static_cast<ViewPosition>(attributes.get_int("initial_view_position_of_active_thread")));
  set_initial_view_position_of_inactive_thread(static_cast<ViewPosition>(attributes.get_int("initial_view_position_of_inactive_thread")));

  color = attributes.get_string("auto_refresh_go_color");
  if (color != NULL && *color != '\0')
    {
      GdkColor tmp_color;
      if (gdk_color_parse(color, &tmp_color))
	{
	  auto_refresh_go_color = tmp_color;
	  auto_refresh_go = &auto_refresh_go_color;
	}
    }

  color = attributes.get_string("auto_refresh_wait_color");
  if (color != NULL && *color != '\0')
    {
      GdkColor tmp_color;
      if (gdk_color_parse(color, &tmp_color))
	{
	  auto_refresh_wait_color = tmp_color;
	  auto_refresh_wait = &auto_refresh_wait_color;
	}
    }

  color = attributes.get_string("thread_title_color");
  if (color != NULL && *color != '\0')
    {
      GdkColor tmp_color;
      if (gdk_color_parse(color, &tmp_color))
	{
	  thread_title_color = tmp_color;
	  thread_title = &thread_title_color;
	}
    }

  color = attributes.get_string("name_color");
  if (color != NULL && *color != '\0')
    {
      GdkColor tmp_color;
      if (gdk_color_parse(color, &tmp_color))
	{
	  name_color = tmp_color;
	  name = &name_color;
	}
    }

  color = attributes.get_string("link_color");
  if (color != NULL && *color != '\0')
    {
      GdkColor tmp_color;
      if (gdk_color_parse(color, &tmp_color))
	{
	  link_color = tmp_color;
	  link = &link_color;
	}
    }

  color = attributes.get_string("visited_color");
  if (color != NULL && *color != '\0')
    {
      GdkColor tmp_color;
      if (gdk_color_parse(color, &tmp_color))
	{
	  visited_color = tmp_color;
	  visited = &visited_color;
	}
    }

  set_hide_bookmark(attributes.get_bool("hide_bookmark"));

  const char *bt = attributes.get_string("bookmark_text");
  if (bt != NULL)
    {
      char *text = URI::decode_string(bt);
      set_bookmark_text(text);
      free(text);
    }

  set_enable_popup_response(attributes.get_bool("enable_popup_response"));
  set_enable_popup_image(attributes.get_bool("enable_popup_image"));
  set_enable_popup_title(attributes.get_bool("enable_popup_title"));
  set_enable_popup_id(attributes.get_bool("enable_popup_id"));

  set_popup_delay(attributes.get_int("popup_delay"));
  set_popdown_delay(attributes.get_int("popdown_delay"));

  set_maximum_number_of_popup_responses(attributes.get_int("maximum_number_of_popup_responses"));

  set_inline_image_width(attributes.get_int("inline_image_width"));
  set_inline_image_height(attributes.get_int("inline_image_height"));
  set_popup_image_width(attributes.get_int("popup_image_width"));
  set_popup_image_height(attributes.get_int("popup_image_height"));

  set_capacity_of_image_cache(static_cast<uintmax_t>(attributes.get_long("capacity_of_image_cache")));
  set_maximum_number_of_image_cache_files(attributes.get_int("maximum_number_of_image_cache_files"));

  set_web_browser_template(attributes.get_string("web_browser_template"));

#if 1
  /* ۤΰ֤ˤä */
  // 0.6꡼ޤǤ˾õͽꡣ
  if (attributes.lookup("enable_proxy") != NULL)
    set_enable_proxy_to_read(attributes.get_bool("enable_proxy"));

  const char *server = attributes.get_string("proxy_server");
  int port = attributes.get_int("proxy_port");
  const char *user = attributes.get_string("proxy_user");
  const char *passwd = attributes.get_string("proxy_passwd");

  if (server != NULL && port != 0)
    set_proxy_to_read(server, port, user, passwd);


  if (attributes.lookup("enable_proxy_to_post") != NULL)
    set_enable_proxy_to_post(attributes.get_bool("enable_proxy_to_post"));

  server = attributes.get_string("proxy_server_to_post");
  port = attributes.get_int("proxy_port_to_post");
  user = attributes.get_string("proxy_user_to_post");
  passwd = attributes.get_string("proxy_passwd_to_post");

  if (server != NULL && port != 0)
    set_proxy_to_post(server, port, user, passwd);


  // ۤΰ֤ȤäƤǡ
  server = attributes.get_string("proxy_server_for_images");
  port = attributes.get_int("proxy_port_for_images");
  user = attributes.get_string("proxy_user_for_images");
  passwd = attributes.get_string("proxy_passwd_for_images");

  if (server != NULL && port != 0)
    set_proxy_for_images(server, port, user, passwd);
#endif

  set_enable_proxy_for_images(attributes.get_bool("enable_proxy_for_images"));

  const char *value = attributes.get_string("proxy_uri_for_images");
  if (value != NULL && *value != '\0')
    {
      char *uri = Base64Decoder::decode_text_from_base64url(value);
      proxy_uri_for_images.parse_uri(uri);
      free(uri);
    }


  set_debug_console(attributes.get_bool("debug_console"));
  set_scheme_console(attributes.get_bool("scheme_console"));


  unsigned int count = MAXIMUM_NUMBER_OF_RECENT_NAME_LIST_ENTRIES;
  char attr_name[64];
  while (count > 0)
    {
      snprintf(attr_name, 64, "recent_name%u", --count);
      register_recent_name(attributes.get_string(attr_name));
    }

  count = MAXIMUM_NUMBER_OF_RECENT_MAIL_LIST_ENTRIES;
  while (count > 0)
    {
      snprintf(attr_name, 64, "recent_mail%u", --count);
      register_recent_mail(attributes.get_string(attr_name));
    }


  set_login_2ch_be_at_startup(attributes.get_bool("login_2ch_be_at_startup"));
  set_save_2ch_be_passwd(attributes.get_bool("save_2ch_be_passwd"));
  set_login_2ch_viewer_at_startup(attributes.get_bool("login_2ch_viewer_at_startup"));
  set_save_2ch_viewer_passwd(attributes.get_bool("save_2ch_viewer_passwd"));

  set_restore_opened_tabs(attributes.get_bool("restore_opened_tabs"));

  changed = false;	// prefs.xmlɤߤƤʤΤѲʤȻפäɤ


  prefs_changed(*this);

  return true;
}

}
