/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: selection_utils.cc,v 1.6 2009/01/01 07:47:07 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "selection_utils.h"

#include "libochushagtk/hypertext_buffer.h"

#include <cstdlib>
#include <cstring>


namespace Ochusha
{


static GQuark data_id
  = g_quark_from_static_string("SelectionUtils::SelectionData");


SelectionData::~SelectionData()
{
}


void
SelectionData::get_data(GtkClipboard *clipboard,
			GtkSelectionData *data,
			guint info, gpointer user_data)
{
  GObject *owner = G_OBJECT(user_data);
  g_return_if_fail(owner != NULL);

  SelectionData *selection_data
    = static_cast<SelectionData *>(g_object_get_qdata(owner, data_id));
  g_return_if_fail(selection_data != NULL);

  selection_data->get_data(clipboard, data, info);
}


void
SelectionData::clear_data(GtkClipboard *clipboard, gpointer user_data)
{
  g_return_if_fail(G_IS_OBJECT(user_data));

  GObject *owner = G_OBJECT(user_data);
  g_return_if_fail(owner != NULL);

  SelectionData *selection_data
    = static_cast<SelectionData *>(g_object_get_qdata(owner, data_id));
  g_return_if_fail(selection_data != NULL);

  if (selection_data->clear_data(clipboard))
    delete selection_data;

  g_object_unref(owner);
}


bool
SelectionData::cuttable()
{
  return deletable() && copiable();
}


bool
SelectionData::do_copy(GtkClipboard *clipboard)
{
  SelectionData *data = snapshot();
  if (data != NULL)
    return data->set_to(clipboard);

  return false;
}


bool
SelectionData::do_cut(GtkClipboard *clipboard)
{
  if (do_copy(clipboard))
    return do_delete();

  return false;
}


gboolean
SelectionData::set_to(GtkClipboard *clipboard)
{
  g_return_val_if_fail(GTK_IS_CLIPBOARD(clipboard), FALSE);

  GObject *owner = G_OBJECT(g_object_new(G_TYPE_OBJECT, NULL));
  g_object_set_qdata(owner, data_id, this);

  if (gtk_clipboard_set_with_owner(clipboard, get_targets(), get_n_targets(),
				   get_data, clear_data, owner))
    return true;

  clear_data(clipboard, owner);
  return false;
}


SelectionData *
SelectionData::get_from(GtkClipboard *clipboard)
{
  g_return_val_if_fail(GTK_IS_CLIPBOARD(clipboard), NULL);

  GObject *owner = gtk_clipboard_get_owner(clipboard);
  if (owner == NULL)
    return NULL;

  return static_cast<SelectionData *>(g_object_get_qdata(owner, data_id));
}


gboolean
SelectionData::set_uris(GtkSelectionData *data,
			const vector<const char *> &uris)
{
  if (uris.empty())
    return FALSE;

  gchar **uri_list
    = static_cast<gchar **>(malloc(sizeof(gchar *) * (uris.size() + 1)));

  int i = 0;
  vector<const char *>::const_iterator iter = uris.begin();
  vector<const char *>::const_iterator end_iter = uris.end();
  while (iter != end_iter)
    uri_list[i++] = const_cast<char *>(*iter++);
  uri_list[i] = NULL;

  gboolean result = gtk_selection_data_set_uris(data, uri_list);
  free(uri_list);

  return result;
}


gboolean
SelectionData::set_uris(GtkSelectionData *data,
			const vector<char *> &uris)
{
  if (uris.empty())
    return FALSE;

  gchar **uri_list
    = static_cast<gchar **>(malloc(sizeof(gchar *) * (uris.size() + 1)));

  int i = 0;
  vector<char *>::const_iterator iter = uris.begin();
  vector<char *>::const_iterator end_iter = uris.end();
  while (iter != end_iter)
    uri_list[i++] = *iter++;
  uri_list[i] = NULL;

  gboolean result = gtk_selection_data_set_uris(data, uri_list);
  free(uri_list);

  return result;
}


TextViewSelectionData::TextViewSelectionData(GtkTextView *view, bool is_static)
  : view(view),
    buffer(gtk_text_view_get_buffer(view)),
    text(NULL),
    destroy_id(0),
    frozen(false),
    is_static(is_static)
{
  g_return_if_fail(buffer != NULL);

  g_object_ref(buffer);

  destroy_id = g_signal_connect(view, "destroy",
				reinterpret_cast<GCallback>(destroy_cb), this);
}


TextViewSelectionData::~TextViewSelectionData()
{
  if (destroy_id != 0)
    {
      g_signal_handler_disconnect(view, destroy_id);
      destroy_id = 0;
    }

  if (text != NULL)
    g_free(text);

  g_return_if_fail(buffer != NULL);
  g_object_unref(buffer);
}


void
TextViewSelectionData::destroy_cb(GtkWidget *widget,
				  TextViewSelectionData *data)
{
  g_return_if_fail(data->view == GTK_TEXT_VIEW(widget));

  data->view = NULL;
  data->destroy_id = 0;
}


bool
TextViewSelectionData::deletable()
{
  return !frozen
    && view != NULL && gtk_text_view_get_editable(view) && copiable();
}


bool
TextViewSelectionData::cuttable()
{
  return !frozen
    && view != NULL && gtk_text_view_get_editable(view) && copiable();
}


bool
TextViewSelectionData::copiable()
{
  if (frozen)
    return text != NULL;

#if GTK_CHECK_VERSION(2,10,0)
  return gtk_text_buffer_get_has_selection(buffer) != FALSE;
#else
  return gtk_text_buffer_get_selection_bounds(buffer, NULL, NULL) != FALSE;
#endif
}


bool
TextViewSelectionData::pastable(GtkClipboard *clipboard)
{
  if (frozen)
    return false;

  return gtk_text_view_get_editable(view)
    && gtk_clipboard_wait_is_text_available(clipboard);
}


bool
TextViewSelectionData::do_delete()
{
  g_return_val_if_fail(deletable(), false);

  return gtk_text_buffer_delete_selection(buffer, TRUE, TRUE) == TRUE;
}


bool
TextViewSelectionData::do_paste(GtkClipboard *clipboard)
{
  g_return_val_if_fail(pastable(clipboard), false);

  gtk_text_buffer_paste_clipboard(buffer, clipboard, NULL, TRUE);

  return false;
}


SelectionData *
TextViewSelectionData::snapshot()
{
  TextViewSelectionData *data = new TextViewSelectionData(view, false);

  data->frozen = true;

  GtkTextIter start_iter;
  GtkTextIter end_iter;
  if (gtk_text_buffer_get_selection_bounds(buffer, &start_iter, &end_iter))
    data->text = gtk_text_buffer_get_text(buffer, &start_iter, &end_iter,
					  FALSE);

  return data;
}


static const GtkTargetEntry text_targets[] = {
  { const_cast<char *>("UTF8_STRING"), 0, 0 },
  { const_cast<char *>("STRING"), 0, 0 },
  { const_cast<char *>("TEXT"), 0, 0 },
  { const_cast<char *>("COMPOUND_TEXT"), 0, 0 },
  { const_cast<char *>("text/plain;charset=utf-8"), 0, 0 },
  { const_cast<char *>("text/plain"), 0, 0 }
};


static const GtkTargetEntry empty_targets[] = {
};


const GtkTargetEntry *
TextViewSelectionData::get_targets()
{
#if GTK_CHECK_VERSION(2,10,0)
  return ((frozen && text != NULL) || gtk_text_buffer_get_has_selection(buffer))
    ? text_targets : empty_targets;
#else
  return ((frozen && text != NULL)
	  || gtk_text_buffer_get_selection_bounds(buffer, NULL, NULL))
    ? text_targets : empty_targets;
#endif
}


guint
TextViewSelectionData::get_n_targets()
{
#if GTK_CHECK_VERSION(2,10,0)
  return ((frozen && text != NULL) || gtk_text_buffer_get_has_selection(buffer))
    ? G_N_ELEMENTS(text_targets) : 0;
#else
  return ((frozen && text != NULL)
	  || gtk_text_buffer_get_selection_bounds(buffer, NULL, NULL))
    ? G_N_ELEMENTS(text_targets) : 0;
#endif
}


void
TextViewSelectionData::get_data(GtkClipboard *clipboard,
				GtkSelectionData *data, guint info)
{
  g_return_if_fail(info == 0);

  if (frozen)
    {
      if (text != NULL)
	gtk_selection_data_set_text(data, text, -1);
      return;
    }

  GtkTextIter start_iter;
  GtkTextIter end_iter;
  if (gtk_text_buffer_get_selection_bounds(buffer, &start_iter, &end_iter))
    {
      gchar *current_text
	= gtk_text_buffer_get_text(buffer, &start_iter, &end_iter, FALSE);
      gtk_selection_data_set_text(data, current_text, -1);
      g_free(current_text);
    }
}


bool
TextViewSelectionData::clear_data(GtkClipboard *clipboard)
{
  return !is_static;
}


EditableSelectionData::EditableSelectionData(GtkWidget *widget, bool is_static)
  : widget(widget),
    editable(GTK_EDITABLE(widget)),
    text(NULL),
    destroy_id(0),
    frozen(false),
    is_static(is_static)
{
  g_return_if_fail(editable != NULL);

  destroy_id = g_signal_connect(widget, "destroy",
				reinterpret_cast<GCallback>(destroy_cb), this);
}


EditableSelectionData::~EditableSelectionData()
{
  if (destroy_id != 0)
    {
      g_signal_handler_disconnect(widget, destroy_id);
      destroy_id = 0;
    }

  if (text != NULL)
    g_free(text);
}


void
EditableSelectionData::destroy_cb(GtkWidget *widget,
				  EditableSelectionData *data)
{
  g_return_if_fail(data->widget == widget);

  data->widget = NULL;
  data->editable = NULL;
  data->destroy_id = 0;
}


bool
EditableSelectionData::deletable()
{
  return !frozen
    && widget != NULL && gtk_editable_get_editable(editable) && copiable();
}


bool
EditableSelectionData::cuttable()
{
  return !frozen
    && widget != NULL && gtk_editable_get_editable(editable) && copiable();
}


bool
EditableSelectionData::copiable()
{
  if (frozen)
    return text != NULL;

  return gtk_editable_get_selection_bounds(editable, NULL, NULL) != FALSE;
}


bool
EditableSelectionData::pastable(GtkClipboard *clipboard)
{
  if (frozen)
    return false;

  return gtk_editable_get_editable(editable)
    && gtk_clipboard_wait_is_text_available(clipboard);
}


bool
EditableSelectionData::do_delete()
{
  g_return_val_if_fail(deletable(), false);

  gtk_editable_delete_selection(editable);
  return true;
}


bool
EditableSelectionData::do_paste(GtkClipboard *clipboard)
{
  g_return_val_if_fail(pastable(clipboard), false);

  // XXX: Υ᥽åɤcallerο񤤤GTK+μ˶¸Ƥ롣

  // ͿclipboardGDK_SELECTION_CLIPBOARDclipboardǤꡢ
  // gtk_editable_paste_clipboard(editable)ǻѤclipboard
  // ƱGDK_SELECTION_CLIPBOARDclipboardǤ롣
  gtk_editable_paste_clipboard(editable);

  return false;
}


SelectionData *
EditableSelectionData::snapshot()
{
  EditableSelectionData *data = new EditableSelectionData(widget, false);

  data->frozen = true;

  gint start;
  gint end;
  if (gtk_editable_get_selection_bounds(editable, &start, &end))
    data->text = gtk_editable_get_chars(editable, start, end);

  return data;
}


const GtkTargetEntry *
EditableSelectionData::get_targets()
{
  return ((frozen && text != NULL)
	  || gtk_editable_get_selection_bounds(editable, NULL, NULL))
    ? text_targets : empty_targets;
}


guint
EditableSelectionData::get_n_targets()
{
  return ((frozen && text != NULL)
	  || gtk_editable_get_selection_bounds(editable, NULL, NULL))
    ? G_N_ELEMENTS(text_targets) : 0;
}


void
EditableSelectionData::get_data(GtkClipboard *clipboard,
				GtkSelectionData *data, guint info)
{
  g_return_if_fail(info == 0);

  if (frozen)
    {
      if (text != NULL)
	gtk_selection_data_set_text(data, text, -1);
      return;
    }

  gint start;
  gint end;
  if (gtk_editable_get_selection_bounds(editable, &start, &end))
    {
      gchar *current_text = gtk_editable_get_chars(editable, start, end);
      gtk_selection_data_set_text(data, current_text, -1);
      g_free(current_text);
    }
}


bool
EditableSelectionData::clear_data(GtkClipboard *clipboard)
{
  return !is_static;
}


SelectionDataProxy::SelectionDataProxy(SelectionData *data, GtkWidget *widget)
  : data(data),
    widget(widget)
{
  destroy_id = g_signal_connect(widget, "destroy",
				reinterpret_cast<GCallback>(destroy_cb), this);
}


SelectionDataProxy::~SelectionDataProxy()
{
  if (destroy_id != 0)
    {
      g_return_if_fail(widget != NULL);
      g_signal_handler_disconnect(widget, destroy_id);
    }
}


bool
SelectionDataProxy::deletable()
{
  return data != NULL && data->deletable();
}


bool
SelectionDataProxy::cuttable()
{
  return data != NULL && data->cuttable();
}


bool
SelectionDataProxy::copiable()
{
  return data != NULL && data->copiable();
}


bool
SelectionDataProxy::pastable(GtkClipboard *clipboard)
{
  return data != NULL && data->pastable(clipboard);
}


bool
SelectionDataProxy::do_delete()
{
  if (data == NULL)
    return false;

  return data->do_delete();
}


bool
SelectionDataProxy::do_paste(GtkClipboard *clipboard)
{
  if (data == NULL)
    return false;

  return data->do_paste(clipboard);
}


SelectionData *
SelectionDataProxy::snapshot()
{
  g_return_val_if_fail(data != NULL, NULL);
  return data->snapshot();
}


void
SelectionDataProxy::destroy_cb(GtkWidget *widget, SelectionDataProxy *proxy)
{
  proxy->data = NULL;
  proxy->widget = NULL;
  proxy->destroy_id = 0;
}


const GtkTargetEntry *
SelectionDataProxy::get_targets()
{
  return data != NULL ? data->get_targets() : empty_targets;
}


guint
SelectionDataProxy::get_n_targets()
{
  return data != NULL ? data->get_n_targets() : 0;
}


void
SelectionDataProxy::get_data(GtkClipboard *clipboard,
			     GtkSelectionData *selection_data, guint info)
{
  if (data != NULL)
    data->get_data(clipboard, selection_data, info);
}


bool
SelectionDataProxy::clear_data(GtkClipboard *clipboard)
{
  return true;
}


}
