/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: session_client.cc,v 1.14 2009/01/01 07:47:07 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "session_client.h"

#include "libochusha/hashtable.h"
#include "libochusha/misc_utils.h"
#include "libochusha/repository.h"
#include "libochusha/sax_context.h"

#include "controller.h"
#include "ui_constants.h"

#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <cstring>

#include <string>

#include <zlib.h>


#define ENABLE_LEGACY_SUPPORT	1


namespace Ochusha
{

SessionClient::SessionClient(Repository &repository)
  : repository(repository),
    controller(NULL),
    states(101)
{
}


SessionClient::~SessionClient()
{
}


void
SessionClient::open_session(Controller *controller, int argc, char **argv)
{
  this->controller = controller;
  XSMPHandler::open_session(argc, argv);
}


void
SessionClient::close_session()
{
  XSMPHandler::close_session();
  controller = NULL;
}


const char *
SessionClient::get_state(const char *name)
{
  assert(name != NULL);
  return static_cast<const char *>(states.lookup(name));
}


void
SessionClient::set_state(const char *name, const char *value)
{
  assert(name != NULL);
  if (value == NULL)
    {
      remove_state(name);
      return;
    }

  states.insert(name, value);
}


void
SessionClient::remove_state(const char *name)
{
  assert(name != NULL);
  states.remove(name);
}


void
SessionClient::discard_session()
{
  const char *smc_id = get_sm_client_id();
  if (smc_id != NULL)
    discard(smc_id);
}


void
SessionClient::save_yourself(const char *smc_id)
{
  assert(smc_id != NULL);

  GDK_THREADS_ENTER();

  controller->save_yourself();
  write_session_xml(smc_id);

  GDK_THREADS_LEAVE();
}


void
SessionClient::die(const char *smc_id)
{
  GDK_THREADS_ENTER();

  XSMPHandler::close_session();
  controller->quit();
  // äƤʤϤġ

  GDK_THREADS_LEAVE();
}


void
SessionClient::save_complete(const char *smc_id)
{
  // ⤷ʤɤϤ
}


void
SessionClient::shutdown_cancelled(const char *smc_id)
{
  // ⤷ʤɤϤ
}


void
SessionClient::restore(const char *smc_id)
{
  read_session_xml(smc_id);
}


void
SessionClient::discard(const char *smc_id)
{
  char pathname[PATH_MAX];

  if (snprintf(pathname, PATH_MAX, ".sm/%s", smc_id) >= PATH_MAX)
    return;	// ʤȤˤϡ⤽⥻å¸ǤʤϤ

  repository.unlink_directory(pathname);
}


static void
write_state(const char *value, gzFile session_xml, const char *name)
{
  assert(value != NULL);

  string text;
  gzprintf(session_xml,
	   "    <state name=\"%s\">\n"
	   "      <string>%s</string>\n"
	   "    </state>\n",
	   name, append_markup_escaped_text(text, value).c_str());
}


void
SessionClient::write_session_xml(const char *smc_id)
{
  string relpath;
  if (smc_id != NULL)
    relpath.append(".sm/").append(smc_id).append("/");

  relpath.append(OCHUSHA_SESSION_XML);

  int fd = repository.open_file(relpath.c_str(), O_WRONLY | O_TRUNC | O_CREAT);
  if (fd < 0)
    {
      fprintf(stderr, "Couldn't open \"%s\"\n", relpath.c_str());
      return;
    }

  gzFile session_xml = gzdopen(fd, "w");
  if (session_xml == NULL)
    {
      close(fd);
      fprintf(stderr, "Couldn't gzdopen \"%s\"\n", relpath.c_str());
      return;
    }

  gzprintf(session_xml, "<?xml version=\"1.0\"?>\n");
  gzprintf(session_xml, "<ochusha>\n");
  gzprintf(session_xml, "  <session>\n");

  states.for_each(reinterpret_cast<AttributeTable::Scanner>(write_state),
		  session_xml);

  gzprintf(session_xml, "  </session>\n");
  gzprintf(session_xml, "</ochusha>\n");

  gzclose(session_xml);
}


class SessionXMLContext : public SAXContext
{
private:
  enum SAXState {
    SAX_INITIAL,
    SAX_OCHUSHA,
    SAX_SESSION,
    SAX_SESSION_STATE,
#ifdef ENABLE_LEGACY_SUPPORT
    SAX_SESSION_STATE_BOOLEAN,
    SAX_SESSION_STATE_INT,
#endif
    SAX_SESSION_STATE_STRING,
    SAX_ACCEPTED,
    SAX_ERROR
  } state;

  AttributeTable &table;

protected:
  void start_element_handler(const char *name, const char **attrs);
  void end_element_handler(const char *name);
  void characters_handler(const char *ch, size_t len);

public:
  SessionXMLContext(AttributeTable &table) : state(SAX_INITIAL), table(table)
  {
    table.clear();
  }

  inline bool is_accepted() { return state == SAX_ACCEPTED; }
};


void
SessionXMLContext::start_element_handler(const char *name, const char **attrs)
{
  switch (state)
    {
    case SAX_INITIAL:
      if (strcmp(name, "ochusha") == 0)
	{ state = SAX_OCHUSHA; return; }
      break;	/* 顼 */

    case SAX_OCHUSHA:
      if (strcmp(name, "session") == 0)
	{ state = SAX_SESSION; return; }
      break;	/* 顼 */

    case SAX_SESSION:
#ifdef ENABLE_LEGACY_SUPPORT
      if ((strcmp(name, "state") == 0 || strcmp(name, "attribute") == 0)
	  && attrs != NULL && strcmp(attrs[0], "name") == 0)
#else
      if (strcmp(name, "state") == 0
	  && attrs != NULL && strcmp(attrs[0], "name") == 0)
#endif
	{
	  state = SAX_SESSION_STATE;
#ifdef ENABLE_LEGACY_SUPPORT
	  const char *attr_name;
	  if (strcmp(attrs[1], "x_pos") == 0)
	    attr_name = "MainWindow::x";
	  else if (strcmp(attrs[1], "y_pos") == 0)
	    attr_name = "MainWindow::y";
	  else if (strcmp(attrs[1], "width") == 0)
	    attr_name = "MainWindow::width";
	  else if (strcmp(attrs[1], "height") == 0)
	    attr_name = "MainWindow::height";
	  else if (strcmp(attrs[1], "boardlist_width") == 0)
	    attr_name = "MainWindow::sidebar_width";
	  else if (strcmp(attrs[1], "threadlist_height") == 0)
	    attr_name = "ViewManager::board_pane_size";
	  else
	    attr_name = attrs[1];
	  table.set_attr_name(attr_name);
#else
	  table.set_attr_name(attrs[1]);
#endif
	  return;
	}
      break;	/* 顼 */

    case SAX_SESSION_STATE:
#ifdef ENABLE_LEGACY_SUPPORT
      if (attrs != NULL && strcmp(attrs[0], "val") == 0)
	{
	  /* int/booleanβǽ */
	  if (strcmp(name, "int") == 0)
	    {
	      state = SAX_SESSION_STATE_INT;
	    }
	  else if (strcmp(name, "boolean") == 0)
	    {
	      state = SAX_SESSION_STATE_BOOLEAN;
	    }
	  else
	    {
	      break;	/* 顼 */
	    }

	  table.set_attr_value(attrs[1]);
	  return;
	}
      else if (strcmp(name, "string") == 0)
#else
      if (strcmp(name, "string") == 0)
#endif
	{
	  state = SAX_SESSION_STATE_STRING;
	  return;
	}
      break;	/* 顼 */

    case SAX_SESSION_STATE_STRING:
#ifdef ENABLE_LEGACY_SUPPORT
    case SAX_SESSION_STATE_BOOLEAN:
    case SAX_SESSION_STATE_INT:
#endif
    case SAX_ACCEPTED:
    case SAX_ERROR:
      break;	/* 顼 */

    default:
      cerr << "start_handler is called in unknown state: " << state << "\n";
    }
  state = SAX_ERROR;
}


void
SessionXMLContext::end_element_handler(const char *name)
{
  switch (state)
    {
    case SAX_OCHUSHA:
      if (strcmp(name, "ochusha") == 0)
	{ state = SAX_ACCEPTED; return; }
      break;	/* 顼 */

    case SAX_SESSION:
      if (strcmp(name, "session") == 0)
	{ state = SAX_OCHUSHA; return; }
      break;	/* 顼 */

    case SAX_SESSION_STATE:
#ifdef ENABLE_LEGACY_SUPPORT
      if (strcmp(name, "state") == 0 || strcmp(name, "attribute") == 0)
#else
      if (strcmp(name, "state") == 0)
#endif
	{
	  state = SAX_SESSION;
	  table.register_current_attribute();
	  return;
	}
      break;	/* 顼 */

    case SAX_SESSION_STATE_STRING:
      if (strcmp(name, "string") == 0)
	{
	  state = SAX_SESSION_STATE;
	  return;
	}
      break;	/* 顼 */

#ifdef ENABLE_LEGACY_SUPPORT
    case SAX_SESSION_STATE_BOOLEAN:
      if (strcmp(name, "boolean") == 0)
	{ state = SAX_SESSION_STATE; return; }
      break;	/* 顼 */

    case SAX_SESSION_STATE_INT:
      if (strcmp(name, "int") == 0)
	{ state = SAX_SESSION_STATE; return; }
      break;	/* 顼 */
#endif

    case SAX_INITIAL:
    case SAX_ACCEPTED:
    case SAX_ERROR:
      break;	/* 顼 */

    default:
#if DEBUG_SAX_HANDLER
      cerr << "endHandler called in unknown state: " << state << "\n";
#endif
      break;	/* 顼 */
    }
#if DEBUG_SAX_HANDLER
  cerr << "Invalid document: </" << name << "> appeared in state="
       << state << "\n";
#endif
  state = SAX_ERROR;
}


void
SessionXMLContext::characters_handler(const char *ch, size_t len)
{
  if (state == SAX_SESSION_STATE_STRING)
    table.append_attr_value(ch, len);
}


void
SessionClient::read_session_xml(const char *smc_id)
{
  string relpath;
  if (smc_id != NULL)
    relpath.append(".sm/").append(smc_id).append("/");

  relpath.append(OCHUSHA_SESSION_XML);

  char *pathname = repository.find_file(relpath.c_str());
  if (pathname == NULL)
    return;

  SessionXMLContext context(states);
  context.parse_xml_file(pathname);

  free(pathname);
}




}
