/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: tab_label.cc,v 1.28 2009/01/01 07:47:07 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "tab_label.h"

#include "libochusha/lock.h"

#include "ui_constants.h"

#include "libochushagtk/icon_label.h"
#include "libochushagtk/stock.h"

#include "libochushagtk_lgpl/mod_pixbuf_simple_anim.h"

#include <gtk/gtk.h>

#include <cstring>
#include <set>


//#define VERBOSE_DESTRUCTOR	1


namespace Ochusha
{


// GtkWidgetTabLabel֥ȤФȤ˻Ȥ
static GQuark label_id = g_quark_from_static_string("TabLabel::label");

static set<TabLabel *> tab_labels;

static Lock idle_func_lock("TabLabel::idle-id");
static guint update_tab_label_icon_on_idle_id = 0;


gboolean
TabLabel::update_tab_label_icon_on_idle(gpointer unused)
{
  {
    LockBlock lock_obj(idle_func_lock);
    g_return_val_if_fail(update_tab_label_icon_on_idle_id != 0, FALSE);

    update_tab_label_icon_on_idle_id = 0;

    if (tab_labels.empty())
      return FALSE;
  }


  GDK_THREADS_ENTER();

  set<TabLabel *>::iterator iter = tab_labels.begin();
  set<TabLabel *>::iterator end_iter = tab_labels.end();
  while (iter != end_iter)
    (*iter++)->update_tab_label_icon();

  GDK_THREADS_LEAVE();


  return FALSE;
}


void
TabLabel::ensure_idle_func_alive()
{
  LockBlock lock_obj(idle_func_lock);
  if (update_tab_label_icon_on_idle_id != 0)
    return;

  update_tab_label_icon_on_idle_id
    = g_idle_add(reinterpret_cast<GSourceFunc>(update_tab_label_icon_on_idle),
		 NULL);
}


TabLabel::TabLabel(const gchar *text, const gchar *stock_id)
  : label(OCHUSHA_ICON_LABEL(ochusha_icon_label_new(text))),
    icon_set(*new CompositeIcon(GTK_WIDGET(label),
				stock_id, GTK_ICON_SIZE_MENU)),
    state(INITIAL)
{
  ochusha_icon_label_set_icon_from_animation(label,
					     icon_set.get_wait_animation());
  gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);

  g_object_set_qdata_full(G_OBJECT(label), label_id, this,
			  reinterpret_cast<GDestroyNotify>(destroy));

  tab_labels.insert(this);
}


TabLabel::~TabLabel()
{
#ifdef VERBOSE_DESTRUCTOR
  cerr << "TabLabel(@" << this << ") destructed.\n";
#endif
  GDK_THREADS_LEAVE();

  disconnect_all();

  GDK_THREADS_ENTER();


  tab_labels.erase(this);
}


void
TabLabel::destroy(TabLabel *label)
{
  delete label;
}


TabLabel *
TabLabel::get_label(GtkWidget *widget)
{
  return static_cast<TabLabel *>(g_object_get_qdata(G_OBJECT(widget),
						    label_id));
}


void
TabLabel::update_tab_label_icon()
{
  g_return_if_fail(OCHUSHA_IN_GDK_THREADS());

  switch (state)
    {
    case INITIAL:
      ochusha_icon_label_set_icon_from_animation(label,
						 icon_set.get_wait_animation());
      return;


    case STARTED:
      ochusha_icon_label_set_icon_from_animation(label,
						 icon_set.get_preparation_animation());
      return;


    case PROGRESSED:
      ochusha_icon_label_set_icon_from_animation(label,
						 icon_set.get_progress_animation());
      return;


    case FINISHED:
      ochusha_icon_label_set_icon_from_pixbuf(label, icon_set.get_done_icon());
      return;


    case NOTICE:
      ochusha_icon_label_set_icon_from_pixbuf(label,
					      icon_set.get_notice_icon());
      return;


    case FAILED:
    case TERMINATED:
      ochusha_icon_label_set_icon_from_pixbuf(label, icon_set.get_error_icon());
      return;
    }
}


void
TabLabel::access_started()
{
  // sigslotʥϥɥ
  state = STARTED;

  ensure_idle_func_alive();
}


void
TabLabel::access_progressed()
{
  // sigslotʥϥɥ
  state = PROGRESSED;

  ensure_idle_func_alive();
}


void
TabLabel::access_finished()
{
  // sigslotʥϥɥ
  if (state != NOTICE)
    state = FINISHED;

  ensure_idle_func_alive();
}


void
TabLabel::access_failed()
{
  // sigslotʥϥɥ
  state = FAILED;

  ensure_idle_func_alive();
}


void
TabLabel::access_terminated()
{
  // sigslotʥϥɥ
  state = TERMINATED;

  ensure_idle_func_alive();
}


void
TabLabel::notice(bool value)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  state = value ? NOTICE : FINISHED;
  update_tab_label_icon();
}


}
