/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: threadlist_view.cc,v 1.62 2009/01/01 07:47:07 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "threadlist_view.h"

#include "libochusha/ochusha_procedure.h"
#include "libochusha/regexp.h"
#include "libochusha/thread_pool.h"

#include "bbs_thread_view.h"
#include "controller.h"
#include "session_client.h"
#include "tab_label.h"
#include "ui_constants.h"

#include "libochushagtk/searchable.h"
#include "libochushagtk/tree_selector_view.h"

#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>

#include <climits>


namespace Ochusha
{

#define THREADLIST_UPDATE_PRIORITY		GDK_PRIORITY_REDRAW + 5
#define MAX_NUM_OF_THREADS_TO_APPEND_AT_ONCE	500

// #define VERBOSE_DESTRUCTOR	1
// #define SHOW_INTERNAL_STATUS	1


enum {
  THREAD_DATA = 0,
  THREAD_COUNT,
  THREAD_WEIGHT,
  THREAD_SEARCH_WEIGHT,
  N_COLUMNS
};


// ThreadlistViewSelectionDataμ
enum {
  TARGET_INTERNAL_LIST_OF_THREADS = 0,
  TARGET_TEXT_URI_LIST,
  TARGET_UTF8_STRING
};


#if GTK_CHECK_VERSION(2,10,0)
static GdkAtom internal_list_of_threads_atom =
  gdk_atom_intern_static_string("ochusha/internal-list-of-threads");
#else
static GdkAtom internal_list_of_threads_atom =
  gdk_atom_intern("ochusha/internal-list-of-threads", FALSE);
#endif


static const GtkTargetEntry all_targets[] = {
  { const_cast<gchar *>("ochusha/internal-list-of-threads"),
    GTK_TARGET_SAME_APP, TARGET_INTERNAL_LIST_OF_THREADS },
  { const_cast<gchar *>("text/uri-list"), 0, TARGET_TEXT_URI_LIST },
  { const_cast<gchar *>("UTF8_STRING"), 0, TARGET_UTF8_STRING },
  { const_cast<gchar *>("STRING"), 0, TARGET_UTF8_STRING },
  { const_cast<gchar *>("TEXT"), 0, TARGET_UTF8_STRING },
  { const_cast<gchar *>("COMPOUND_TEXT"), 0, TARGET_UTF8_STRING },
  { const_cast<gchar *>("text/plain;charset=utf-8"), 0, TARGET_UTF8_STRING },
  { const_cast<gchar *>("text/plain"), 0, TARGET_UTF8_STRING }
};


static const GtkTargetEntry empty_targets[] = {
};


ThreadlistViewSelectionData::ThreadlistViewSelectionData(GtkTreeSelection *selection,
							 bool is_static)
  : selection(selection),
    is_static(is_static)
{
  if (is_static)
    g_signal_connect(selection, "changed",
		     reinterpret_cast<GCallback>(changed_cb), this);
  else
    changed_cb(selection, this);
}


ThreadlistViewSelectionData::~ThreadlistViewSelectionData()
{
}


bool
ThreadlistViewSelectionData::deletable()
{
  return false;
}


bool
ThreadlistViewSelectionData::cuttable()
{
  return false;
}


bool
ThreadlistViewSelectionData::copiable()
{
  return !thread_list.empty();
}


bool
ThreadlistViewSelectionData::pastable(GtkClipboard *clipboard)
{
  return false;
}


bool
ThreadlistViewSelectionData::do_delete()
{
  return false;
}


bool
ThreadlistViewSelectionData::do_paste(GtkClipboard *clipboard)
{
  return false;
}


SelectionData *
ThreadlistViewSelectionData::snapshot()
{
  return new ThreadlistViewSelectionData(selection, false);
}


void
ThreadlistViewSelectionData::accumulate_threads(GtkTreeModel *model,
						GtkTreePath *path,
						GtkTreeIter *iter,
						gpointer data)
{
  BBSThread *thread = NULL;
  gtk_tree_model_get(model, iter, THREAD_DATA, &thread, -1);
  g_return_if_fail(thread != NULL);

  ThreadlistViewSelectionData *selection_data
    = static_cast<ThreadlistViewSelectionData *>(data);
  selection_data->thread_list.push_back(thread);
}


void
ThreadlistViewSelectionData::changed_cb(GtkTreeSelection *selection,
					ThreadlistViewSelectionData *data)
{
  data->thread_list.clear();

  gtk_tree_selection_selected_foreach(selection, accumulate_threads, data);
}


const GtkTargetEntry *
ThreadlistViewSelectionData::get_targets()
{
  return thread_list.size() > 0 ? all_targets : empty_targets;
}


guint
ThreadlistViewSelectionData::get_n_targets()
{
  return thread_list.size() > 0 ? 6 : 0;
}


void
ThreadlistViewSelectionData::get_data(GtkClipboard *clipboard,
				      GtkSelectionData *data, guint info)
{
  set_data(data, info);
}


bool
ThreadlistViewSelectionData::clear_data(GtkClipboard *clipboard)
{
  return !is_static;
}


bool
ThreadlistViewSelectionData::set_data(GtkSelectionData *data, guint info)
{
  switch (info)
    {
    case TARGET_INTERNAL_LIST_OF_THREADS:
      break;

    case TARGET_TEXT_URI_LIST:
      {
	char **uri_list
	  = static_cast<char **>(malloc(sizeof(char *)
					* (thread_list.size() + 1)));

	vector<BBSThread *>::iterator iter = thread_list.begin();
	vector<BBSThread *>::iterator end_iter = thread_list.end();
	int i = 0;
	while (iter != end_iter)
	  uri_list[i++] = (*iter++)->get_url();
	uri_list[i] = NULL;
	gboolean result = gtk_selection_data_set_uris(data, uri_list);
	while (--i >= 0)
	  free(uri_list[i]);
	free(uri_list);
	return result != FALSE;
      }

    case TARGET_UTF8_STRING:
      {
	string buf;
	vector<BBSThread *>::iterator iter = thread_list.begin();
	vector<BBSThread *>::iterator end_iter = thread_list.end();
	while (iter != end_iter)
	  {
	    BBSThread *thread = *iter++;
	    const char *title = thread->get_title();
	    if (title != NULL)
	      buf.append(title);
	    else
	      buf.append(_("Anonymous Thread"));
	    buf.append("\n");
	    char *url = thread->get_url();
	    buf.append(url).append("\n");
	    free(url);
	  }
	return gtk_selection_data_set_text(data, buf.c_str(), -1) != FALSE;
      }
    }

  return false;
}


bool
ThreadlistViewSelectionData::is_selected(GtkTreePath *path)
{
  return gtk_tree_selection_path_is_selected(selection, path);
}


SelectionData *
ThreadlistViewSelectionData::create_proxy()
{
  GtkWidget *widget = GTK_WIDGET(gtk_tree_selection_get_tree_view(selection));
  return new SelectionDataProxy(this, widget);
}
// ThreadlistViewSelectionDataμ



// ThreadlistViewFocusNotifierμ
ThreadlistViewFocusNotifier::ThreadlistViewFocusNotifier(ThreadlistView &view,
							 GtkWidget *widget)
  : FocusNotifier(view.controller, widget, false),
    view(view)
{
}


ThreadlistViewFocusNotifier::~ThreadlistViewFocusNotifier()
{
#if VERBOSE_DESTRUCTOR
  fprintf(stderr, "ThreadlistViewFocusNotifier::~ThreadlistViewFocusNotifier(): selection_data=%p\n",
	  &view.selection_data);
#endif
  view.controller.clear_clipboard_internal(&view.selection_data);
}


gboolean
ThreadlistViewFocusNotifier::notify_focus_in()
{
#ifdef DEBUG_FOCUS_NOTIFIER
  fprintf(stderr, "ThreadlistViewFocusNotifier::notify_focus_in(): %p, %d\n",
	  this, has_pointer());
#endif
  view.controller.update_clipboard_internal(&view.selection_data);

  view.notify_focus_in();

  return FALSE;
}


void
ThreadlistViewFocusNotifier::notify_unmap()
{
  view.controller.clear_clipboard_internal(&view.selection_data);
}
// ThreadlistViewFocusNotifierμ



// ThreadlistViewμ
enum {
  SORT_BY_COUNT,
  SORT_BY_BOARD,
  SORT_BY_RANK,
  SORT_BY_MARK,
  SORT_BY_TITLE,
  SORT_BY_N_RES_VERBOSE,
  SORT_BY_N_RES_VERBOSE2,
  SORT_BY_N_RES_AVAILABLE,
  SORT_BY_N_RES_READ,
  SORT_BY_N_RES_UNGOT,
  SORT_BY_N_RES_TO_READ,
  SORT_BY_N_NEW_RESPONSES,
  SORT_BY_BIRTHTIME,
  SORT_BY_LAST_MODIFIED,
  SORT_BY_WEIGHT,
  SORT_BY_SEARCH_WEIGHT
};


#define BOARD_WIDTH		"ThreadlistView::board::width"
#define TITLE_WIDTH		"ThreadlistView::title::width"
#define RANK_WIDTH		"ThreadlistView::rank::width"
#define MARK_WIDTH		"ThreadlistView::mark::width"
#define COUNT_WIDTH		"ThreadlistView::count::width"
#define N_RES_VERBOSE_WIDTH	"ThreadlistView::n_res_verbose::width"
#define N_RES_VERBOSE2_WIDTH	"ThreadlistView::n_res_verbose2::width"
#define N_RES_AVAILABLE_WIDTH	"ThreadlistView::n_res_available::width"
#define N_RES_READ_WIDTH	"ThreadlistView::n_res_read::width"
#define N_RES_UNGOT_WIDTH	"ThreadlistView::n_res_ungot::width"
#define N_RES_TO_READ_WIDTH	"ThreadlistView::n_res_to_read::width"
#define N_NEW_RES_WIDTH		"ThreadlistView::n_new_res::width"
#define RANK_DIFF_WIDTH		"ThreadlistView::rank_diff::width"
#define BIRTHTIME_WIDTH		"ThreadlistView::birthtime::width"
#define LAST_MODIFIED_WIDTH	"ThreadlistView::last_modified::width"
#define WEIGHT_WIDTH		"ThreadlistView::weight::width"
#define SEARCH_WEIGHT_WIDTH	"ThreadlistView::search_weight::width"


static GQuark sort_column_id
  = g_quark_from_static_string("ThreadlistView::sort_column");
static GQuark sort_order_id
  = g_quark_from_static_string("ThreadlistView::sort_order");


static int compare_count(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
			 gpointer unused);
static int compare_board(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
			 gpointer unused);
static int compare_rank(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
			gpointer unused);
static int compare_mark(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
			gpointer unused);
static int compare_title(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
			 gpointer unused);
static int compare_n_res_verbose(GtkTreeModel *model, GtkTreeIter *a,
				 GtkTreeIter *b, gpointer unused);
static int compare_n_res_verbose2(GtkTreeModel *model, GtkTreeIter *a,
				  GtkTreeIter *b, gpointer unused);
static int compare_n_res_avail(GtkTreeModel *model, GtkTreeIter *a,
			       GtkTreeIter *b, gpointer unused);
static int compare_n_res_read(GtkTreeModel *model, GtkTreeIter *a,
			      GtkTreeIter *b, gpointer unused);
static int compare_n_res_ungot(GtkTreeModel *model, GtkTreeIter *a,
			       GtkTreeIter *b, gpointer unused);
static int compare_n_res_to_read(GtkTreeModel *model, GtkTreeIter *a,
				 GtkTreeIter *b, gpointer unused);
static int compare_n_new_res(GtkTreeModel *model, GtkTreeIter *a,
			     GtkTreeIter *b, gpointer unused);
static int compare_birthtime(GtkTreeModel *model, GtkTreeIter *a,
			     GtkTreeIter *b, gpointer unused);
static int compare_last_modified(GtkTreeModel *model, GtkTreeIter *a,
				 GtkTreeIter *b, gpointer unused);
static int compare_weight(GtkTreeModel *model, GtkTreeIter *a,
			  GtkTreeIter *b, gpointer unused);
static int compare_search_weight(GtkTreeModel *model, GtkTreeIter *a,
				 GtkTreeIter *b, gpointer unused);


static gboolean
title_match_func(OchushaTreeSelectorView *view,
		 GtkTreeModel *model, GtkTreeIter *iter,
		 OchushaSearchKey *key, gpointer unused)
{
  BBSThread *thread = NULL;
  gtk_tree_model_get(model, iter, THREAD_DATA, &thread, -1);
  g_return_val_if_fail(thread != NULL, FALSE);
  const char *title = thread->get_title();
  return ochusha_search_key_match_text(key, title != NULL ? title : "", -1);
}


static Lock list_iter_lock("ThreadlistView::list_iter_lock");


ThreadlistView::ThreadlistView(Controller &controller, BulletinBoard *board,
			       TabLabel *label, bool refresh)
  : View(label, gtk_scrolled_window_new(NULL, NULL)),
    controller(controller),
    model(GTK_TREE_MODEL(gtk_list_store_new(N_COLUMNS,
					    G_TYPE_POINTER,	// BBSThread *
					    G_TYPE_INT,		// ɲý
					    G_TYPE_LONG,	// 
					    G_TYPE_LONG))),	// 
    selector_widget(ochusha_tree_selector_view_new_with_model(model)),
    window_title(g_strdup_printf(_("%s - Ochusha"), board->get_name())),
    selection_data(gtk_tree_view_get_selection(GTK_TREE_VIEW(selector_widget)),
		   true),
    focus_notifier(*this, selector_widget),
    new_thread_window(NULL),
    new_thread_window_destroy_id(0),
    weight_column(NULL),
    last_sort_column(NULL),
    press_start_x(0),
    press_start_y(0),
    press_start_x_root(0),
    press_start_y_root(0),
    press_start_button(0),
    press_start_state(0),
    popup_delay_timeout_id(0),
    board(board),
    rule_info(NULL),
    append_threads_id(0),
    number_of_threads(0),
    number_of_interest_threads(0),
    updating(false)
{
  GtkWidget *widget = get_widget();
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(widget),
				 GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_widget_show(selector_widget);
  gtk_container_add(GTK_CONTAINER(widget), selector_widget);

  g_signal_connect(selector_widget, "start_interactive_search",
		   reinterpret_cast<GCallback>(start_interactive_search_cb),
		   this);

  gtk_tree_view_set_enable_search(GTK_TREE_VIEW(selector_widget), FALSE);

  g_signal_connect(selector_widget, "key_press_event",
		   reinterpret_cast<GCallback>(key_press_event_cb), this);

  GtkTreeSortable *sortable = GTK_TREE_SORTABLE(model);
  gtk_tree_sortable_set_sort_func(sortable,
				  SORT_BY_COUNT, compare_count,
				  NULL, NULL);
  gtk_tree_sortable_set_sort_func(sortable,
				  SORT_BY_BOARD, compare_board,
				  NULL, NULL);
  gtk_tree_sortable_set_sort_func(sortable,
				  SORT_BY_RANK, compare_rank,
				  NULL, NULL);
  gtk_tree_sortable_set_sort_func(sortable,
				  SORT_BY_MARK, compare_mark,
				  NULL, NULL);
  gtk_tree_sortable_set_sort_func(sortable,
				  SORT_BY_TITLE, compare_title,
				  NULL, NULL);
  gtk_tree_sortable_set_sort_func(sortable,
				  SORT_BY_N_RES_VERBOSE, compare_n_res_verbose,
				  NULL, NULL);
  gtk_tree_sortable_set_sort_func(sortable,
				  SORT_BY_N_RES_VERBOSE2,
				  compare_n_res_verbose2,
				  NULL, NULL);
  gtk_tree_sortable_set_sort_func(sortable,
				  SORT_BY_N_RES_AVAILABLE,
				  compare_n_res_avail,
				  NULL, NULL);
  gtk_tree_sortable_set_sort_func(sortable,
				  SORT_BY_N_RES_READ, compare_n_res_read,
				  NULL, NULL);
  gtk_tree_sortable_set_sort_func(sortable,
				  SORT_BY_N_RES_UNGOT, compare_n_res_ungot,
				  NULL, NULL);
  gtk_tree_sortable_set_sort_func(sortable,
				  SORT_BY_N_RES_TO_READ, compare_n_res_to_read,
				  NULL, NULL);
  gtk_tree_sortable_set_sort_func(sortable,
				  SORT_BY_N_NEW_RESPONSES, compare_n_new_res,
				  NULL, NULL);
  gtk_tree_sortable_set_sort_func(sortable,
				  SORT_BY_BIRTHTIME, compare_birthtime,
				  NULL, NULL);
  gtk_tree_sortable_set_sort_func(sortable,
				  SORT_BY_LAST_MODIFIED, compare_last_modified,
				  NULL, NULL);
  gtk_tree_sortable_set_sort_func(sortable,
				  SORT_BY_WEIGHT, compare_weight,
				  NULL, NULL);
  gtk_tree_sortable_set_sort_func(sortable,
				  SORT_BY_SEARCH_WEIGHT, compare_search_weight,
				  NULL, NULL);

  gtk_tree_sortable_set_sort_column_id(sortable,
				       SORT_BY_WEIGHT,
				       GTK_SORT_ASCENDING);
  GtkTreeSelection *selection
    = gtk_tree_view_get_selection(GTK_TREE_VIEW(selector_widget));
  gtk_tree_selection_set_mode(selection, GTK_SELECTION_MULTIPLE);

  g_signal_connect(selection, "changed",
		   reinterpret_cast<GCallback>(selection_changed_cb), this);

  setup_columns();
  ochusha_tree_selector_view_set_search_equal_func(OCHUSHA_TREE_SELECTOR_VIEW(selector_widget),
						   title_match_func,
						   NULL, NULL);
  g_signal_connect(selector_widget, "row_button_press",
		   reinterpret_cast<GCallback>(row_button_press_cb), this);
  g_signal_connect(selector_widget, "row_button_release",
		   reinterpret_cast<GCallback>(row_button_release_cb), this);
  g_signal_connect(selector_widget, "popup_menu",
		   reinterpret_cast<GCallback>(popup_menu_cb), this);
  g_signal_connect(selector_widget, "motion_notify_event",
		   reinterpret_cast<GCallback>(motion_notify_cb), this);
  g_signal_connect(selector_widget, "column_resized",
		   reinterpret_cast<GCallback>(column_resized_cb), this);

  // DnDݡȤμ¸Ƥ
  initialize_for_drag_source();
  // ޤDnDϢ

  update(refresh);			// ȤϤ
}


ThreadlistView::~ThreadlistView()
{
  // ThreadlistViewsigslotΥϥɥĤΤǥϥɥ餬ʥ
  // ³줿ޤޤǤdestructorglobal lock롣
  // å뤿ᡢĴ롣
  assert(OCHUSHA_IN_GDK_THREADS());

  if (popup_delay_timeout_id != 0)
    g_source_remove(popup_delay_timeout_id);

  if (append_threads_id != 0)
    g_source_remove(append_threads_id);


  GDK_THREADS_LEAVE();

  list_iter = NULL;
  disconnect_all();

  GDK_THREADS_ENTER();


  // ΩƥɥƱĤȤ⤢뤬̯
  if (new_thread_window_destroy_id != 0)
    g_signal_handler_disconnect(new_thread_window,
				new_thread_window_destroy_id);

#if VERBOSE_DESTRUCTOR
  fprintf(stderr, "ThreadlistView destructed: %p\n", this);
#endif

  if (window_title != NULL)
    g_free(window_title);

  g_object_unref(model);
}


void
ThreadlistView::update(bool reload)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  {
    LockBlock lock_obj(list_iter_lock);
    if (list_iter != NULL)
      {
#ifdef SHOW_INTERNAL_STATUS
	cerr << "Threadlist is busy: " << board->get_name() << endl;
#endif
	return;	// 
      }
  }

  const char *rule_name = board->get_rule();
  const OchushaProcedure *procedure
    = controller.get_scheme_manager().ochusha_procedure;
  if (rule_name != NULL)
    rule_info = procedure->lookup_threadlist_rule_info(rule_name);
  else
    rule_info = NULL;

  if (rule_info == NULL)
    rule_info
      = procedure->lookup_threadlist_rule_info(controller.get_prefs()
					       .get_default_threadlist_rule());

  gtk_list_store_clear(GTK_LIST_STORE(model));
  number_of_threads = 0;
  number_of_interest_threads = 0;

  while (!shown_threads.empty())
    shown_threads.pop_back();


  GDK_THREADS_LEAVE();

  // ϡ1ܤ̣ʤ
  board->read_threadlist_xml(controller.get_repository());

  ThreadlistIteratorHandle iter;
  {
    iter = board->get_threadlist(controller.get_config(),
				 controller.get_repository(),
				 controller.get_thread_pool(),
				 controller.get_broker(),
				 reload);

    LockBlock lock_obj(list_iter_lock);
    list_iter = iter;
  }
  updating = reload;

  int lock_id = iter->lock();

  TabLabel *label = get_tab_label();
  if (label != NULL)
    {
      switch (iter->get_agent_state(lock_id))
	{
	case NetworkAgent::STATE_INITIAL:
	  if (iter->finished(lock_id))
	    label->access_finished();
	  break;

	case NetworkAgent::STATE_STARTED:
	  label->access_started();
	  break;

	case NetworkAgent::STATE_TERMINATED:
	  label->access_terminated();
	  break;

	default:
	  label->access_progressed();
	  break;
	}

      iter->access_started.connect(label, &TabLabel::access_started);
      iter->access_progressed.connect(label, &TabLabel::access_progressed);
      iter->access_finished.connect(label, &TabLabel::access_finished);
      iter->access_failed.connect(label, &TabLabel::access_failed);
      iter->access_terminated.connect(label, &TabLabel::access_terminated);
    }

  iter->state_changed.connect(this, &ThreadlistView::state_changed_cb);
  iter->board_moved.connect(this, &ThreadlistView::board_moved_cb);

  bool has_threads = iter->has_more_threads(lock_id);

  iter->unlock(lock_id);


  GDK_THREADS_ENTER();


  if (append_threads_id == 0 && has_threads)
    append_threads_id
      = g_idle_add_full(THREADLIST_UPDATE_PRIORITY,
			reinterpret_cast<GSourceFunc>(append_threads_on_idle),
			this, NULL);
}


void
ThreadlistView::new_thread_window_destroy_cb(GtkWidget *widget,
					     ThreadlistView *view)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  view->new_thread_window = NULL;
  view->new_thread_window_destroy_id = 0;
}


void
ThreadlistView::show_new_thread_window()
{
  assert(OCHUSHA_IN_GDK_THREADS());

  NewThreadWindow *window;
  if (new_thread_window != NULL)
    {
      gtk_window_present(GTK_WINDOW(new_thread_window));
      window = NewThreadWindow::get_window(new_thread_window);
    }
  else
    {
      window = new NewThreadWindow(controller, board);
      new_thread_window = window->get_widget();
      new_thread_window_destroy_id
	= g_signal_connect(new_thread_window, "destroy",
			   reinterpret_cast<GCallback>(new_thread_window_destroy_cb),
			   this);
      gtk_widget_show(new_thread_window);
    }
  window->grab_focus_into_editor();
}


GtkTreeSelection *
ThreadlistView::get_selection()
{
  return gtk_tree_view_get_selection(GTK_TREE_VIEW(selector_widget));
}



// Viewvirtualμ
const gchar *
ThreadlistView::get_combo_title()
{
  const gchar *title = board->get_short_name();
  return title != NULL ? title : _("Anonymous");
}


const gchar *
ThreadlistView::get_menu_title()
{
  const gchar *title = board->get_short_name();
  return title != NULL ? title : _("Anonymous");
}


const gchar *
ThreadlistView::get_window_title()
{
  return window_title;
}


const gchar *
ThreadlistView::get_url()
{
  return board->get_base_url();
}


void
ThreadlistView::grab_focus()
{
  gtk_widget_grab_focus(selector_widget);
}


bool
ThreadlistView::contains(GtkWidget *child)
{
  return child == selector_widget || child == get_widget();
}


const gchar *
ThreadlistView::get_search_title()
{
  return _("Find in Threadlist");
}


OchushaSearchable *
ThreadlistView::get_search_target()
{
  return OCHUSHA_SEARCHABLE(selector_widget);
}


void
ThreadlistView::notify_closed()
{
  // ⤷ʤ
}


bool
ThreadlistView::bookmarkable()
{
  return true;
}


bool
ThreadlistView::bookmark()
{
  controller.bookmark(board);
  return true;
}


bool
ThreadlistView::refreshable()
{
  return true;
}


bool
ThreadlistView::refresh()
{
  controller.refresh(board, true);
  return true;
}


bool
ThreadlistView::has_info()
{
  return true;
}


bool
ThreadlistView::show_info()
{
  controller.get_boardlist_ui().open_properties(board);
  return true;
}
// Viewvirtualμ



// GtkTreeSelectionϢ
void
ThreadlistView::selection_changed_cb(GtkTreeSelection *selection,
				     ThreadlistView *view)
{
  view->controller.update_clipboard_internal(&view->selection_data);
}
// GtkTreeSelectionϢ



// OchushaTreeSelectorViewΥʥϥɥ
gboolean
ThreadlistView::start_interactive_search_cb(OchushaTreeSelectorView *widget,
					    ThreadlistView *view)
{
  view->request_search();
  return TRUE;
}


gboolean
ThreadlistView::key_press_event_cb(OchushaTreeSelectorView *widget,
				   GdkEventKey *event, ThreadlistView *view)
{
  g_return_val_if_fail(OCHUSHA_IS_TREE_SELECTOR_VIEW(widget), FALSE);

  // MEMO: ϡɥǥ󥰤Τ򤱤Τǡ
  //       schemeؿꤲ褦ˤʤͽꡣ

  /*
   * ʲ괺ΤΥХǥ󥰤Ƹ
   *
   * GDK_o: open-selection
   * GDK_t: open-selection-in-tab
   * GDK_r: reset-thread
   *
   * GDK_asterisk: mark-thread
   * GDK_u: unmark-thread
   * GDK_d: toggle-thread-hidden-state
   * GDK_space: advance-thread-view
   * GDK_BackSpace: retreat-thread-view
   * GDK_a: write-response
   *
   * GDK_p: find_interest_thread_previous
   * GDK_n: find_interest_thread_next
   */

  Controller &controller = view->controller;
  LinkUI &link_ui = controller.get_link_ui();
  ThreadlistUI &threadlist_ui = controller.get_threadlist_ui();

  GtkAction *action;
  switch (event->keyval)
    {
    case GDK_o:
    case GDK_Return:
    case GDK_KP_Enter:
    case GDK_ISO_Enter:
      action = link_ui.open_link_location;
      break;

    case GDK_t:
      action = link_ui.open_link_location_in_tab;
      break;

    case GDK_r:
      action = threadlist_ui.reset_thread;
      break;

    case GDK_asterisk:
      action = threadlist_ui.mark_thread;
      break;

    case GDK_u:
      action = threadlist_ui.unmark_thread;
      break;

    case GDK_d:
      if (controller.get_prefs().get_show_hidden_threads())
	action = threadlist_ui.toggle_hide_thread;
      else
	action = threadlist_ui.hide_thread;
      break;

    case GDK_space:
    case GDK_BackSpace:
    case GDK_a:
    case GDK_n:
    case GDK_p:
      action = NULL;
      break;

    default:
      return FALSE;
    }

  if (event->keyval != GDK_asterisk && (event->state & GDK_MODIFIER_MASK) != 0)
    return FALSE;	// ǥեդΥХɤϤƤʤ

  GtkTreePath *path = NULL;
  gtk_tree_view_get_cursor(GTK_TREE_VIEW(widget), &path, NULL);
  TreePathHandle cursor_path = path;

  if (path == NULL)
    return FALSE;

  gtk_tree_view_scroll_to_cell(GTK_TREE_VIEW(widget),
			       path, NULL, TRUE, 0.5, 0.0);

  if (!view->selection_data.is_selected(path))
    {
      gboolean result = FALSE;
      g_signal_emit_by_name(widget, "select-cursor-row", FALSE, &result);
      controller.update_clipboard_internal(&view->selection_data);
    }

  if (action != NULL)
    {
      gtk_action_activate(action);

      if (!controller.get_prefs().get_select_tab_when_opened())
	view->grab_focus();	// եȥ५ĤΤ

      return TRUE;
    }

  if (event->keyval == GDK_n)
    {
      view->find_interest_thread_next();
      return TRUE;
    }

  if (event->keyval == GDK_p)
    {
      view->find_interest_thread_previous();
      return TRUE;
    }

  BBSThread *thread = view->activate_view_at_cursor();
  if (thread == NULL)
    {
      if (!controller.get_prefs().get_select_tab_when_opened())
	view->grab_focus();	// եȥ५ĤΤ
      return TRUE;
    }

  if (event->keyval == GDK_space)
    view->advance_view_at_cursor(thread);
  else if (event->keyval == GDK_BackSpace)
    view->retreat_view_at_cursor(thread);
  else
    {
      g_return_val_if_fail(event->keyval == GDK_a, TRUE);

      ViewManager &view_manager = view->controller.get_view_manager();
      g_return_val_if_fail(view_manager.get_current_thread() == thread, TRUE);

      View *tmp_view = view_manager.get_current_thread_view();
      g_return_val_if_fail(tmp_view != NULL, TRUE);

      BBSThreadView *thread_view = dynamic_cast<BBSThreadView *>(tmp_view);
      g_return_val_if_fail(thread_view != NULL, TRUE);

      thread_view->show_prefered_response_editor();
    }

  return TRUE;
}


bool
ThreadlistView::is_interest_thread(BBSThread *thread)
{
  SchemeManager &scheme = controller.get_scheme_manager();
  Cell *func = scheme.is_interest_thread.get_cell();
  if (func != NULL)
    {
      Cell *result = scheme.get_core().evalf0(func, "%p", thread);
      if (result != NULL)
	return result != Cell::f();

      // schemeؿǤΥ顼ȯǥեȵ§fallback
    }

  return thread->get_number_of_responses_got() > 0
    && thread->get_number_of_responses_to_read() > 0;
}


BBSThread *
ThreadlistView::find_interest_thread_previous()
{
  GtkTreePath *path = NULL;
  gtk_tree_view_get_cursor(GTK_TREE_VIEW(selector_widget), &path, NULL);
  TreePathHandle cursor_path = path;
  g_return_val_if_fail(path != NULL, NULL);

  while (gtk_tree_path_prev(path))
    {
      GtkTreeIter iter;
      gboolean result = gtk_tree_model_get_iter(model, &iter, path);
      g_return_val_if_fail(result, NULL);

      BBSThread *thread = NULL;
      gtk_tree_model_get(model, &iter, THREAD_DATA, &thread, -1);
      g_return_val_if_fail(thread != NULL, NULL);

      if (is_interest_thread(thread))
	{
	  path = gtk_tree_model_get_path(model, &iter);
	  cursor_path = path;
	  g_return_val_if_fail(path != NULL, NULL);

	  gtk_tree_view_set_cursor(GTK_TREE_VIEW(selector_widget), path,
				   NULL, FALSE);
	  gtk_tree_view_scroll_to_cell(GTK_TREE_VIEW(selector_widget),
				       path, NULL, TRUE, 0.5, 0.0);
	  return thread;
	}
    }

  return NULL;
}


BBSThread *
ThreadlistView::find_interest_thread_next()
{
  GtkTreePath *path = NULL;
  gtk_tree_view_get_cursor(GTK_TREE_VIEW(selector_widget), &path, NULL);
  TreePathHandle cursor_path = path;
  g_return_val_if_fail(path != NULL, NULL);

  GtkTreeIter iter;
  gboolean result = gtk_tree_model_get_iter(model, &iter, path);
  g_return_val_if_fail(result, NULL);

  while (gtk_tree_model_iter_next(model, &iter))
    {
      BBSThread *thread = NULL;
      gtk_tree_model_get(model, &iter, THREAD_DATA, &thread, -1);
      g_return_val_if_fail(thread != NULL, NULL);

      if (is_interest_thread(thread))
	{
	  path = gtk_tree_model_get_path(model, &iter);
	  cursor_path = path;
	  g_return_val_if_fail(path != NULL, NULL);

	  gtk_tree_view_set_cursor(GTK_TREE_VIEW(selector_widget), path,
				   NULL, FALSE);
	  gtk_tree_view_scroll_to_cell(GTK_TREE_VIEW(selector_widget),
				       path, NULL, TRUE, 0.5, 0.0);
	  return thread;
	}
    }

  return NULL;
}


// ֤Υ줬ɽɽƤ֤ˤ롣
// ɽƤˤϥBBSThread *֤ʳξˤ
// NULL֤
BBSThread *
ThreadlistView::activate_view_at_cursor()
{
  GtkTreePath *path = NULL;
  gtk_tree_view_get_cursor(GTK_TREE_VIEW(selector_widget), &path, NULL);
  TreePathHandle cursor_path = path;

  GtkTreeIter iter;
  if (!gtk_tree_model_get_iter(model, &iter, path))
    return NULL;

  BBSThread *thread = NULL;
  gtk_tree_model_get(model, &iter, THREAD_DATA, &thread, -1);
  g_return_val_if_fail(thread != NULL, NULL);

  ViewManager &view_manager = controller.get_view_manager();
  if (view_manager.get_current_thread() == thread)
    return thread;	// ɽɽƤ硣

  if (view_manager.find_thread_view(thread, true) != NULL)
    return NULL;	// 줿֤ɽƤ硣

  controller.open(thread, true, 0,
		  controller.get_prefs().get_default_open_in_tab(),
		  false, true);

  return NULL;
}


void
ThreadlistView::advance_view_at_cursor(BBSThread *thread)
{
  ViewManager &view_manager = controller.get_view_manager();
  g_return_if_fail(view_manager.get_current_thread() == thread);

  View *tmp_view = view_manager.get_current_thread_view();
  g_return_if_fail(tmp_view != NULL);

  BBSThreadView *thread_view = dynamic_cast<BBSThreadView *>(tmp_view);
  g_return_if_fail(thread_view != NULL);

  if (thread_view->advance_viewport())
    return;

  // ޤãƤ硢õ
  thread = find_interest_thread_next();
  if (thread != NULL)
    {
      controller.open(thread, true, 0,
		      controller.get_prefs().get_default_open_in_tab(),
		      false, true);
      grab_focus();	// ǥեΤϵʬ
    }
}


void
ThreadlistView::retreat_view_at_cursor(BBSThread *thread)
{
  ViewManager &view_manager = controller.get_view_manager();
  g_return_if_fail(view_manager.get_current_thread() == thread);

  View *tmp_view = view_manager.get_current_thread_view();
  g_return_if_fail(tmp_view != NULL);

  BBSThreadView *thread_view = dynamic_cast<BBSThreadView *>(tmp_view);
  g_return_if_fail(thread_view != NULL);

  thread_view->retreat_viewport();
}


gboolean
ThreadlistView::row_button_press_cb(OchushaTreeSelectorView *widget,
				    GtkTreePath *path,
				    GtkTreeViewColumn *column,
				    GdkEventButton *event,
				    ThreadlistView *view)
{
  view->last_event_path = NULL;

  if (event->button == 3)
    {
      GtkTreeIter iter;
      GtkTreeView *tree_view = GTK_TREE_VIEW(widget);
      GtkTreeModel *model = gtk_tree_view_get_model(tree_view);

      if (!gtk_tree_model_get_iter(model, &iter, path))
	return FALSE;

      view->press_start_x = static_cast<gint>(event->x);
      view->press_start_y = static_cast<gint>(event->y);

      view->press_start_x_root = static_cast<gint>(event->x_root);
      view->press_start_y_root = static_cast<gint>(event->y_root);
      view->press_start_button = event->button;

      if (view->popup_delay_timeout_id != 0)
	g_source_remove(view->popup_delay_timeout_id);

      view->popup_delay_timeout_id
	= g_timeout_add(POPUP_MENU_DELAY_TIMEOUT,
			reinterpret_cast<GSourceFunc>(popup_delay_timeout_cb),
			view);

      return view->selection_data.is_selected(path);
    }
  else if ((event->state & (GDK_CONTROL_MASK | GDK_SHIFT_MASK)) == 0)
    view->last_event_path = gtk_tree_path_copy(path);

  return FALSE;
}


gboolean
ThreadlistView::popup_menu_cb(GtkWidget *widget, ThreadlistView *view)
{
  if (view->popup_delay_timeout_id != 0)
    {
      g_source_remove(view->popup_delay_timeout_id);
      view->popup_delay_timeout_id = 0;
    }

  view->popup_context_menu(0, gtk_get_current_event_time());

  return TRUE;
}


gboolean
ThreadlistView::motion_notify_cb(GtkWidget *widget, GdkEventMotion *event,
				 ThreadlistView *view)
{
  if (view->popup_delay_timeout_id != 0
      && gtk_drag_check_threshold(widget,
				  view->press_start_x, view->press_start_y,
				  static_cast<gint>(event->x),
				  static_cast<gint>(event->y)))
    {
      g_source_remove(view->popup_delay_timeout_id);
      view->popup_delay_timeout_id = 0;
    }

  return FALSE;
}


void
ThreadlistView::popup_context_menu(guint button, guint32 event_time)
{
  GtkWidget *menu_widget
    = gtk_ui_manager_get_widget(controller.get_ui_manager(),
				controller.get_prefs().get_show_hidden_threads()
				? "/UglyThreadlistPopupMenu"
				: "/ThreadlistPopupMenu");
  g_return_if_fail(GTK_IS_MENU(menu_widget));

  gtk_menu_popup(GTK_MENU(menu_widget), NULL, NULL, NULL, NULL,
		 button, event_time);
}


gboolean
ThreadlistView::popup_delay_timeout_cb(ThreadlistView *view)
{
  // g_timeout_addΥХåGLibɳʤΤGDK_THREADS_ENTER()Ǥʤ
  GDK_THREADS_ENTER();

  view->popup_context_menu(view->press_start_button,
			   gtk_get_current_event_time());
  view->popup_delay_timeout_id = 0;

  GDK_THREADS_LEAVE();


  return FALSE;
}


gboolean
ThreadlistView::row_button_release_cb(OchushaTreeSelectorView *widget,
				      GtkTreePath *path,
				      GtkTreeViewColumn *column,
				      GdkEventButton *event,
				      ThreadlistView *view)
{
  if (view->popup_delay_timeout_id != 0)
    {
      g_source_remove(view->popup_delay_timeout_id);
      view->popup_delay_timeout_id = 0;
      view->popup_context_menu(0, event->time);
    }

  GtkTreeIter iter;
  GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(widget));

  gboolean result = FALSE;
  BBSThread *thread = NULL;
  if (static_cast<GtkTreePath *>(view->last_event_path) == NULL
      || gtk_tree_path_compare(view->last_event_path, path) != 0
      || !gtk_tree_model_get_iter(model, &iter, path))
    goto done;

  gtk_tree_model_get(model, &iter, THREAD_DATA, &thread, -1);
  if (thread != NULL)
    {
      gint *id_p = static_cast<gint *>(g_object_get_qdata(G_OBJECT(column),
							  sort_column_id));
      if (id_p != NULL && *id_p == SORT_BY_MARK)
	{
	  thread->set_marked(!thread->is_marked());
	  gtk_tree_model_row_changed(model, path, &iter);
	}
      else
	{
	  Controller &controller = view->controller;
	  bool in_tab = ((event->button == 2)
			 ^ (controller.get_prefs().get_default_open_in_tab()));
	  controller.open(thread, true, 0, in_tab, false);
	}
    }

 done:
  view->last_event_path = NULL;

  return result;
}


void
ThreadlistView::column_resized_cb(OchushaTreeSelectorView *widget,
				  GtkTreeViewColumn *column,
				  gboolean use_resized_width,
				  gint resized_width,
				  ThreadlistView *view)
{
  gint *id_p = static_cast<gint *>(g_object_get_qdata(G_OBJECT(column),
						      sort_column_id));
  const char *name;

  if (id_p == NULL)
    name = RANK_DIFF_WIDTH;	// rank_diffΤϤġ
  else
    switch (*id_p)
      {
      case SORT_BY_COUNT:
	name = COUNT_WIDTH;
	break;

      case SORT_BY_BOARD:
	name = BOARD_WIDTH;
	break;

      case SORT_BY_RANK:
	name = RANK_WIDTH;
	break;

      case SORT_BY_MARK:
	name = MARK_WIDTH;
	break;

      case SORT_BY_TITLE:
	name = TITLE_WIDTH;
	break;

      case SORT_BY_N_RES_VERBOSE:
	name = N_RES_VERBOSE_WIDTH;
	break;

      case SORT_BY_N_RES_VERBOSE2:
	name = N_RES_VERBOSE2_WIDTH;
	break;

      case SORT_BY_N_RES_AVAILABLE:
	name = N_RES_AVAILABLE_WIDTH;
	break;

      case SORT_BY_N_RES_READ:
	name = N_RES_READ_WIDTH;
	break;

      case SORT_BY_N_RES_UNGOT:
	name = N_RES_UNGOT_WIDTH;
	break;

      case SORT_BY_N_RES_TO_READ:
	name = N_RES_TO_READ_WIDTH;
	break;

      case SORT_BY_N_NEW_RESPONSES:
	name = N_NEW_RES_WIDTH;
	break;

      case SORT_BY_BIRTHTIME:
	name = BIRTHTIME_WIDTH;
	break;

      case SORT_BY_LAST_MODIFIED:
	name = LAST_MODIFIED_WIDTH;
	break;

      case SORT_BY_WEIGHT:
	name = WEIGHT_WIDTH;
	break;

      case SORT_BY_SEARCH_WEIGHT:
	name = SEARCH_WEIGHT_WIDTH;
	break;

      default:
	name = NULL;
      }

  if (name != NULL)
    {
      SessionClient &session = view->controller.get_session();
      if (use_resized_width)
	{
	  char buf[64];
	  snprintf(buf, 64, "%d", resized_width);
	  session.set_state(name, buf);
	}
      else
	session.remove_state(name);
    }

  gtk_tree_view_column_set_max_width(column, -1);

  if (name == NULL)
    fprintf(stderr, "Cannot correctly handle id=%d\n", *id_p);
}
// OchushaTreeSelectorViewΥʥϥɥ



// DnDϢ
void
ThreadlistView::initialize_for_drag_source()
{
  int buttons = GDK_BUTTON1_MASK | GDK_BUTTON2_MASK | GDK_BUTTON3_MASK;
  int actions = GDK_ACTION_COPY;

  gtk_tree_view_enable_model_drag_source(GTK_TREE_VIEW(selector_widget),
					 static_cast<GdkModifierType>(buttons),
					 all_targets, G_N_ELEMENTS(all_targets),
					 static_cast<GdkDragAction>(actions));
  g_signal_connect(selector_widget, "drag_data_get",
		   reinterpret_cast<GCallback>(drag_data_get_cb), this);
}


void
ThreadlistView::drag_data_get_cb(GtkWidget *widget,
				 GdkDragContext *drag_context,
				 GtkSelectionData *data, guint info, guint time,
				 ThreadlistView *view)
{
  g_return_if_fail(GTK_IS_TREE_VIEW(widget));
  view->selection_data.set_data(data, info);
}
// ޤDnDϢ


gboolean
ThreadlistView::internal_list_of_threads_available(GtkClipboard *clipboard)
{
  g_return_val_if_fail(GTK_IS_CLIPBOARD(clipboard), FALSE);

  return gtk_clipboard_wait_is_target_available(clipboard,
						internal_list_of_threads_atom);
}


GtkTreeSelection *
ThreadlistView::get_gtk_tree_selection_from_clipboard(GtkClipboard *clipboard)
{
  g_return_val_if_fail(internal_list_of_threads_available(clipboard), NULL);

  SelectionData *data = SelectionData::get_from(clipboard);
  g_return_val_if_fail(data != NULL, NULL);

  ThreadlistViewSelectionData *selection_data
    = dynamic_cast<ThreadlistViewSelectionData *>(data);
  g_return_val_if_fail(selection_data != NULL, NULL);

  GtkTreeSelection *selection = selection_data->get_selection();
  g_return_val_if_fail(GTK_IS_TREE_SELECTION(selection), NULL);

  return selection;
}


BBSThread *
ThreadlistView::get_thread(GtkTreeModel *model, GtkTreeIter *iter)
{
  BBSThread *thread = NULL;
  gtk_tree_model_get(model, iter, THREAD_DATA, &thread, -1);
  return thread;
}


void
ThreadlistView::setup_columns()
{
  const char *column = controller.get_prefs().get_threadlist_view_contents();
  g_return_if_fail(strlen(column) > 0);

  while (*column != '\0')
    {
      switch (*column)
	{
	case 'A':
	  append_n_res_avail_column();
	  break;

	case 'B':
	  if (board->get_bbs_type() < OCHUSHA_BBS_TYPE_2CH)
	    append_board_name_column();
	  break;

	case 'b':
	  append_birthtime_column();
	  break;

	case 'D':
	  append_rank_diff_column();
	  break;

	case 'L':
	  append_last_modified_column();
	  break;

	case 'M':
	  append_mark_column();
	  break;

	case 'N':
	  append_n_new_res_column();
	  break;

	case 'n':
	  append_n_res_read_column();
	  break;

	case 'R':
	  append_rank_column();
	  break;

	case 'T':
	  append_title_column();
	  break;

	case 'U':
	  append_n_res_ungot_column();
	  break;

	case 'u':
	  append_n_res_to_read_column();
	  break;

	case 'V':
	  append_n_res_verbose_column();
	  break;

	case 'v':
	  append_n_res_verbose2_column();
	  break;

	case 'W':
	  append_weight_column();
	  break;

	case 'w':
	  append_search_weight_column();
	  break;

	default:
	  fprintf(stderr, "'%c' doesn't match any column type\n", *column);
	  break;
	}
      column++;
    }
}


void
ThreadlistView::column_set_mark(GtkTreeViewColumn *column,
				GtkCellRenderer *renderer,
				GtkTreeModel *model, GtkTreeIter *iter,
				gpointer data)
{
  BBSThread *thread = NULL;
  gtk_tree_model_get(model, iter, THREAD_DATA, &thread, -1);
  g_return_if_fail(thread != NULL);

  GdkColor *bg = NULL;
  ThreadlistView *view = static_cast<ThreadlistView *>(data);
  SchemeManager &scheme = view->controller.get_scheme_manager();
  Cell *func = scheme.decorate_list_entry.get_cell();
  if (func != NULL)
    {
      /* SchemeؿƤǿ򤹤 */
      Cell *color_pair = scheme.get_core().evalf0(func, "%p", thread);
      if (color_pair != NULL && color_pair->is_pair())
	{
	  Cell *bg_obj = color_pair->get_cdr();
	  if (bg_obj->is_foreign_object())
	    bg = static_cast<GdkColor *>(bg_obj->get_foreign_pointer());
	}
    }

  g_object_set(renderer,
	       "active", thread->is_marked(),
	       "cell_background_gdk", bg,
	       NULL);
}


void
ThreadlistView::column_set_text(GtkCellRenderer *renderer, BBSThread *thread,
				const gchar *text, gpointer data)
{
  bool hidden = thread->is_hidden();

  GdkColor *fg = NULL;
  GdkColor *bg = NULL;
  ThreadlistView *view = static_cast<ThreadlistView *>(data);
  SchemeManager &scheme = view->controller.get_scheme_manager();
  Cell *func = scheme.decorate_list_entry.get_cell();

  if (func != NULL)
    {
      /* SchemeؿƤǿ򤹤 */
      Cell *color_pair = scheme.get_core().evalf0(func, "%p", thread);
      if (color_pair != NULL && color_pair->is_pair())
	{
	  Cell *fg_obj = color_pair->get_car();
	  if (fg_obj->is_foreign_object())
	    fg = static_cast<GdkColor *>(fg_obj->get_foreign_pointer());

	  Cell *bg_obj = color_pair->get_cdr();
	  if (bg_obj->is_foreign_object())
	    bg = static_cast<GdkColor *>(bg_obj->get_foreign_pointer());
	}
    }

  g_object_set(renderer,
	       "text", text,
	       "strikethrough", hidden,
	       "strikethrough_set", hidden,
	       "foreground_gdk", fg,
	       "cell_background_gdk", bg,
	       NULL);
}


void
ThreadlistView::column_set_board_name(GtkTreeViewColumn *column,
				      GtkCellRenderer *renderer,
				      GtkTreeModel *model, GtkTreeIter *iter,
				      gpointer view)
{
  BBSThread *thread = NULL;
  gtk_tree_model_get(model, iter, THREAD_DATA, &thread, -1);
  g_return_if_fail(thread != NULL);

  column_set_text(renderer, thread, thread->get_board()->get_short_name(),
		  view);
}


void
ThreadlistView::column_set_title(GtkTreeViewColumn *column,
				 GtkCellRenderer *renderer,
				 GtkTreeModel *model, GtkTreeIter *iter,
				 gpointer view)
{
  BBSThread *thread = NULL;
  gtk_tree_model_get(model, iter, THREAD_DATA, &thread, -1);
  g_return_if_fail(thread != NULL);

  const char *title = thread->get_title();
  if (title != NULL)
    {
      column_set_text(renderer, thread, title, view);
      return;
    }

  char title_buf[256];
  snprintf(title_buf, 256, _("(tmp)%s"), thread->get_id());
  column_set_text(renderer, thread, title_buf, view);
}


void
ThreadlistView::column_set_rank(GtkTreeViewColumn *column,
				GtkCellRenderer *renderer,
				GtkTreeModel *model, GtkTreeIter *iter,
				gpointer view)
{
  BBSThread *thread = NULL;
  gtk_tree_model_get(model, iter, THREAD_DATA, &thread, -1);
  g_return_if_fail(thread != NULL);

  gchar rank_buf[16];
  const gchar *text;
  if (thread->is_alive())
    {
      snprintf(rank_buf, 16, "%d", thread->get_rank());
      text = rank_buf;
    }
  else if (thread->is_kako_thread())
    text = _("Past");
  else if (thread->is_viewer_used())
    text = _("$");
  else
    text = _("Dead");

  column_set_text(renderer, thread, text, view);
}


void
ThreadlistView::column_set_n_res_verbose(GtkTreeViewColumn *column,
					 GtkCellRenderer *renderer,
					 GtkTreeModel *model, GtkTreeIter *iter,
					 gpointer view)
{
  BBSThread *thread = NULL;
  gtk_tree_model_get(model, iter, THREAD_DATA, &thread, -1);
  g_return_if_fail(thread != NULL);

  gchar text[64];

  int num_read = thread->get_number_of_responses_got();
  int num_to_read = thread->get_number_of_responses_to_read();
  int num_avail = thread->get_number_of_responses_available();
  if (num_read == 0)
    snprintf(text, 64, "%d", num_avail);
  else if (num_to_read == 0)
    snprintf(text, 64, "%d (0)", num_avail);
  else
    snprintf(text, 64, "%d (+%d)", num_avail, num_to_read);

  column_set_text(renderer, thread, text, view);
}


void
ThreadlistView::column_set_n_res_verbose2(GtkTreeViewColumn *column,
					  GtkCellRenderer *renderer,
					  GtkTreeModel *model,
					  GtkTreeIter *iter,
					  gpointer view)
{
  BBSThread *thread = NULL;
  gtk_tree_model_get(model, iter, THREAD_DATA, &thread, -1);
  g_return_if_fail(thread != NULL);

  gchar text[64];

  int num_read = thread->get_number_of_responses_got();
  int num_to_get = thread->get_number_of_responses_to_get();
  int num_avail = thread->get_number_of_responses_available();
  if (num_read == 0)
    snprintf(text, 64, "%d", num_avail);
  else if (num_to_get == 0)
    snprintf(text, 64, "%d (0)", num_avail);
  else
    snprintf(text, 64, "%d (+%d)", num_avail, num_to_get);

  column_set_text(renderer, thread, text, view);
}


void
ThreadlistView::column_set_n_res_avail(GtkTreeViewColumn *column,
				       GtkCellRenderer *renderer,
				       GtkTreeModel *model, GtkTreeIter *iter,
				       gpointer view)
{
  BBSThread *thread = NULL;
  gtk_tree_model_get(model, iter, THREAD_DATA, &thread, -1);
  g_return_if_fail(thread != NULL);

  gchar text[64];
  snprintf(text, 64, "%d", thread->get_number_of_responses_available());
  column_set_text(renderer, thread, text, view);
}


void
ThreadlistView::column_set_n_res_read(GtkTreeViewColumn *column,
				      GtkCellRenderer *renderer,
				      GtkTreeModel *model, GtkTreeIter *iter,
				      gpointer view)
{
  BBSThread *thread = NULL;
  gtk_tree_model_get(model, iter, THREAD_DATA, &thread, -1);
  g_return_if_fail(thread != NULL);

  gchar text[64];
  snprintf(text, 64, "%d", thread->get_number_of_responses_got());
  column_set_text(renderer, thread, text, view);
}

   
void
ThreadlistView::column_set_n_res_ungot(GtkTreeViewColumn *column,
				       GtkCellRenderer *renderer,
				       GtkTreeModel *model, GtkTreeIter *iter,
				       gpointer view)
{
  BBSThread *thread = NULL;
  gtk_tree_model_get(model, iter, THREAD_DATA, &thread, -1);
  g_return_if_fail(thread != NULL);

  gchar text[64];
  snprintf(text, 64, "%d", thread->get_number_of_responses_to_get());
  column_set_text(renderer, thread, text, view);
}


void
ThreadlistView::column_set_n_res_to_read(GtkTreeViewColumn *column,
					 GtkCellRenderer *renderer,
					 GtkTreeModel *model, GtkTreeIter *iter,
					 gpointer view)
{
  BBSThread *thread = NULL;
  gtk_tree_model_get(model, iter, THREAD_DATA, &thread, -1);
  g_return_if_fail(thread != NULL);

  gchar text[64];
  snprintf(text, 64, "%d", thread->get_number_of_responses_to_read());
  column_set_text(renderer, thread, text, view);
}


void
ThreadlistView::column_set_n_new_res(GtkTreeViewColumn *column,
				     GtkCellRenderer *renderer,
				     GtkTreeModel *model, GtkTreeIter *iter,
				     gpointer view)
{
  BBSThread *thread = NULL;
  gtk_tree_model_get(model, iter, THREAD_DATA, &thread, -1);
  g_return_if_fail(thread != NULL);

  gchar text_buf[64];
  const gchar *text;
  int num = thread->get_number_of_new_responses();
  if (num == 0)
    text = "";
  else
    {
      snprintf(text_buf, 64, "+%d", num);
      text = text_buf;
    }

  column_set_text(renderer, thread, text, view);
}


void
ThreadlistView::column_set_rank_diff(GtkTreeViewColumn *column,
				     GtkCellRenderer *renderer,
				     GtkTreeModel *model, GtkTreeIter *iter,
				     gpointer view)
{
  BBSThread *thread = NULL;
  gtk_tree_model_get(model, iter, THREAD_DATA, &thread, -1);
  g_return_if_fail(thread != NULL);

  gchar text_buf[64];
  const gchar *text;

  int rank = thread->get_rank();
  int prev_rank = thread->get_previous_rank();
  if (rank == prev_rank || rank == 0)
    text = "";	// ѲʤǤ륹졣
  else if (prev_rank == 0)
    text = _("New Face!");
  else if (rank > prev_rank)
    text = _("down");
  else
    {
      snprintf(text_buf, 64, _("%d Rank Up"), prev_rank - rank);
      text = text_buf;
    }

  column_set_text(renderer, thread, text, view);
}


void
ThreadlistView::column_set_birthtime(GtkTreeViewColumn *column,
				     GtkCellRenderer *renderer,
				     GtkTreeModel *model, GtkTreeIter *iter,
				     gpointer view)
{
  BBSThread *thread = NULL;
  gtk_tree_model_get(model, iter, THREAD_DATA, &thread, -1);
  g_return_if_fail(thread != NULL);

  gchar text_buf[128];
  const gchar *text;

  time_t utc = thread->get_birthtime();
  if (utc == -1)
    text = "";	// Τ͡
  else
    {
#ifdef HAVE_LOCALTIME_R
      struct tm tm;
      size_t result;
      localtime_r(&utc, &tm);
      result = strftime(text_buf, 128, "%Y-%m-%d %T", &tm);
#else	/* GDK_THREADS_ENTER()ǼƤĤ */
      struct tm *tm = localtime(&utc);
      size_t result = strftime(text_buf, 128, "%Y-%m-%d %T", tm);
#endif
      if (result == 0)
	text_buf[0] = '\0';
      text = text_buf;
    }

  column_set_text(renderer, thread, text, view);
}

   
void
ThreadlistView::column_set_last_modified(GtkTreeViewColumn *column,
					 GtkCellRenderer *renderer,
					 GtkTreeModel *model, GtkTreeIter *iter,
					 gpointer view)
{
  BBSThread *thread = NULL;
  gtk_tree_model_get(model, iter, THREAD_DATA, &thread, -1);
  g_return_if_fail(thread != NULL);

  gchar text_buf[128];
  const gchar *text;

  time_t utc = thread->get_last_modified_utc();
  if (utc == -1)
    text = "";	/* Τ͡ */
  else
    {
#ifdef HAVE_LOCALTIME_R
      struct tm tm;
      size_t result;
      localtime_r(&utc, &tm);
      result = strftime(text_buf, 128, "%Y-%m-%d %T", &tm);
#else	/* GDK_THREADS_ENTER()ǼƤĤ */
      struct tm *tm = localtime(&utc);
      size_t result = strftime(text_buf, 128, "%Y-%m-%d %T", tm);
#endif
      if (result == 0)
	text_buf[0] = '\0';
      text = text_buf;
    }

  column_set_text(renderer, thread, text, view);
}


void
ThreadlistView::column_set_weight(GtkTreeViewColumn *column,
				  GtkCellRenderer *renderer,
				  GtkTreeModel *model, GtkTreeIter *iter,
				  gpointer view)
{
  BBSThread *thread = NULL;
  long weight = 0;
  gtk_tree_model_get(model, iter,
		     THREAD_DATA, &thread,
		     THREAD_WEIGHT, &weight,
		     -1);
  g_return_if_fail(thread != NULL);

  gchar text[64];
  snprintf(text, 64, "%ld", weight + LONG_MAX);

  column_set_text(renderer, thread, text, view);
}

   
void
ThreadlistView::column_set_search_weight(GtkTreeViewColumn *column,
					 GtkCellRenderer *renderer,
					 GtkTreeModel *model, GtkTreeIter *iter,
					 gpointer view)
{
  BBSThread *thread = NULL;
  long weight = 0;
  gtk_tree_model_get(model, iter,
		     THREAD_DATA, &thread,
		     THREAD_SEARCH_WEIGHT, &weight,
		     -1);
  g_return_if_fail(thread != NULL);

  gchar text[64];
  snprintf(text, 64, "%ld", weight + LONG_MAX);

  column_set_text(renderer, thread, text, view);
}


void
ThreadlistView::column_clicked_cb(GtkTreeViewColumn *column,
				  ThreadlistView *view)
{
  gint *id_p = static_cast<gint *>(g_object_get_qdata(G_OBJECT(column),
						      sort_column_id));
  g_return_if_fail(id_p != NULL);

  GtkSortType *order_p = static_cast<GtkSortType *>(g_object_get_qdata(G_OBJECT(column),
								       sort_order_id));
  g_return_if_fail(order_p != NULL);

  GtkTreeSortable *sortable = GTK_TREE_SORTABLE(view->model);

  gint current_id;
  GtkSortType order;
  if (gtk_tree_sortable_get_sort_column_id(sortable, &current_id, &order))
    {
      if (current_id == *id_p)
	{
	  gtk_tree_view_column_set_sort_indicator(column, FALSE);
	  gtk_tree_sortable_set_sort_column_id(sortable,
					       SORT_BY_WEIGHT,
					       GTK_SORT_ASCENDING);

	  view->last_sort_column = view->weight_column;
	  if (view->weight_column != NULL)
	    gtk_tree_view_column_set_sort_indicator(view->weight_column, TRUE);
	}
      else
	{
	  if (view->last_sort_column != NULL)
	    gtk_tree_view_column_set_sort_indicator(view->last_sort_column,
						    FALSE);

	  gtk_tree_view_column_set_sort_indicator(column, TRUE);
	  gtk_tree_sortable_set_sort_column_id(sortable,
					       *id_p, *order_p);
	  view->last_sort_column = column;
	}
      return;
    }

  fprintf(stderr, "Why non sortable column clickable?\n");
}


void
ThreadlistView::set_sort_column_id(GtkTreeViewColumn *column, gint column_id,
				   GtkSortType order)
{
  gint *id_p = g_new0(gint, 1);
  *id_p = column_id;
  g_object_set_qdata_full(G_OBJECT(column), sort_column_id,
			  id_p, reinterpret_cast<GDestroyNotify>(g_free));

  GtkSortType *order_p = g_new0(GtkSortType, 1);
  *order_p = order;
  g_object_set_qdata_full(G_OBJECT(column), sort_order_id,
			  order_p, reinterpret_cast<GDestroyNotify>(g_free));

  gtk_tree_view_column_set_clickable(column, TRUE);
  g_signal_connect(column, "clicked",
		   reinterpret_cast<GCallback>(column_clicked_cb), this);
}


void
ThreadlistView::restore_session(GtkTreeViewColumn *column, const char *name)
{
  SessionClient &session = controller.get_session();
  const char *value = session.get_state(name);
  if (value != NULL)
    {
      gint width;
      if (sscanf(value, "%d", &width) == 1)
	{
#if 0
	  // ݤ졽ɽʤʤ롣
	  column->use_resized_width = 1;
	  column->width = width;
#endif
	  // max_widthȤꤷƤ뤬column_resizedʥ
	  // Υϥɥ-1ꤷľƤΤmax_widthresize
	  // ¤뤳ȤϤʤ
	  gtk_tree_view_column_set_max_width(column, width);
	}
    }
}


void
ThreadlistView::append_board_name_column()
{
  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  GtkTreeViewColumn *column
    = gtk_tree_view_column_new_with_attributes(_("Board"), renderer, NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_cell_data_func(column, renderer,
					  column_set_board_name,
					  this, NULL);
  set_sort_column_id(column, SORT_BY_BOARD, GTK_SORT_ASCENDING);
  gtk_tree_view_append_column(GTK_TREE_VIEW(selector_widget), column);
  if (column->button != NULL)
    {
      GTK_WIDGET_UNSET_FLAGS(column->button, GTK_CAN_FOCUS);
      gtk_button_set_focus_on_click(GTK_BUTTON(column->button), FALSE);
    }

  restore_session(column, BOARD_WIDTH);
}


void
ThreadlistView::append_title_column()
{
  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  GtkTreeViewColumn *column
    = gtk_tree_view_column_new_with_attributes(_("Thread Title"), renderer,
					       NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_cell_data_func(column, renderer,
					  column_set_title,
					  this, NULL);
  set_sort_column_id(column, SORT_BY_TITLE, GTK_SORT_ASCENDING);
  gtk_tree_view_append_column(GTK_TREE_VIEW(selector_widget), column);
  if (column->button != NULL)
    {
      GTK_WIDGET_UNSET_FLAGS(column->button, GTK_CAN_FOCUS);
      gtk_button_set_focus_on_click(GTK_BUTTON(column->button), FALSE);
    }

  restore_session(column, TITLE_WIDTH);
}


void
ThreadlistView::append_rank_column()
{
  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  GtkTreeViewColumn *column
    = gtk_tree_view_column_new_with_attributes(_("Rank"), renderer, NULL);
  g_object_set(renderer, "xalign", 0.9, NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_cell_data_func(column, renderer,
					  column_set_rank,
					  this, NULL);
  set_sort_column_id(column, SORT_BY_RANK, GTK_SORT_ASCENDING);
  gtk_tree_view_append_column(GTK_TREE_VIEW(selector_widget), column);
  if (column->button != NULL)
    {
      GTK_WIDGET_UNSET_FLAGS(column->button, GTK_CAN_FOCUS);
      gtk_button_set_focus_on_click(GTK_BUTTON(column->button), FALSE);
    }

  restore_session(column, RANK_WIDTH);
}


void
ThreadlistView::append_mark_column()
{
  GtkCellRenderer *renderer = gtk_cell_renderer_toggle_new();
  GtkTreeViewColumn *column
    = gtk_tree_view_column_new_with_attributes(_("Mark"), renderer, NULL);
  gtk_cell_renderer_toggle_set_radio(GTK_CELL_RENDERER_TOGGLE(renderer),
				     FALSE);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_cell_data_func(column, renderer,
					  column_set_mark,
					  this, NULL);
  set_sort_column_id(column, SORT_BY_MARK, GTK_SORT_ASCENDING);
  gtk_tree_view_append_column(GTK_TREE_VIEW(selector_widget), column);
  if (column->button != NULL)
    {
      GTK_WIDGET_UNSET_FLAGS(column->button, GTK_CAN_FOCUS);
      gtk_button_set_focus_on_click(GTK_BUTTON(column->button), FALSE);
    }

  restore_session(column, MARK_WIDTH);
}


void
ThreadlistView::append_n_res_verbose_column()
{
  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  GtkTreeViewColumn *column
    = gtk_tree_view_column_new_with_attributes(_("#Res"), renderer, NULL);
  g_object_set(renderer, "xalign", 0.9, NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_cell_data_func(column, renderer,
					  column_set_n_res_verbose,
					  this, NULL);
  set_sort_column_id(column, SORT_BY_N_RES_VERBOSE, GTK_SORT_DESCENDING);
  gtk_tree_view_append_column(GTK_TREE_VIEW(selector_widget), column);
  if (column->button != NULL)
    {
      GTK_WIDGET_UNSET_FLAGS(column->button, GTK_CAN_FOCUS);
      gtk_button_set_focus_on_click(GTK_BUTTON(column->button), FALSE);
    }

  restore_session(column, N_RES_VERBOSE_WIDTH);
}


void
ThreadlistView::append_n_res_verbose2_column()
{
  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  GtkTreeViewColumn *column
    = gtk_tree_view_column_new_with_attributes(_("#Res"), renderer, NULL);
  g_object_set(renderer, "xalign", 0.9, NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_cell_data_func(column, renderer,
					  column_set_n_res_verbose2,
					  this, NULL);
  set_sort_column_id(column, SORT_BY_N_RES_VERBOSE2, GTK_SORT_DESCENDING);
  gtk_tree_view_append_column(GTK_TREE_VIEW(selector_widget), column);
  if (column->button != NULL)
    {
      GTK_WIDGET_UNSET_FLAGS(column->button, GTK_CAN_FOCUS);
      gtk_button_set_focus_on_click(GTK_BUTTON(column->button), FALSE);
    }

  restore_session(column, N_RES_VERBOSE2_WIDTH);
}


void
ThreadlistView::append_n_res_avail_column()
{
  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  GtkTreeViewColumn *column
    = gtk_tree_view_column_new_with_attributes(_("#Res"), renderer, NULL);
  g_object_set(renderer, "xalign", 0.9, NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_cell_data_func(column, renderer,
					  column_set_n_res_avail,
					  this, NULL);
  set_sort_column_id(column, SORT_BY_N_RES_AVAILABLE, GTK_SORT_DESCENDING);
  gtk_tree_view_append_column(GTK_TREE_VIEW(selector_widget), column);
  if (column->button != NULL)
    {
      GTK_WIDGET_UNSET_FLAGS(column->button, GTK_CAN_FOCUS);
      gtk_button_set_focus_on_click(GTK_BUTTON(column->button), FALSE);
    }

  restore_session(column, N_RES_AVAILABLE_WIDTH);
}


void
ThreadlistView::append_n_res_read_column()
{
  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  GtkTreeViewColumn *column
    = gtk_tree_view_column_new_with_attributes(_("#Got"), renderer, NULL);
  g_object_set(renderer, "xalign", 0.9, NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_cell_data_func(column, renderer,
					  column_set_n_res_read,
					  this, NULL);
  set_sort_column_id(column, SORT_BY_N_RES_READ, GTK_SORT_DESCENDING);
  gtk_tree_view_append_column(GTK_TREE_VIEW(selector_widget), column);
  if (column->button != NULL)
    {
      GTK_WIDGET_UNSET_FLAGS(column->button, GTK_CAN_FOCUS);
      gtk_button_set_focus_on_click(GTK_BUTTON(column->button), FALSE);
    }

  restore_session(column, N_RES_READ_WIDTH);
}


void
ThreadlistView::append_n_res_ungot_column()
{
  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  GtkTreeViewColumn *column
    = gtk_tree_view_column_new_with_attributes(_("#Ungot"), renderer, NULL);
  g_object_set(renderer, "xalign", 0.9, NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_cell_data_func(column, renderer,
					  column_set_n_res_ungot,
					  this, NULL);
  set_sort_column_id(column, SORT_BY_N_RES_UNGOT, GTK_SORT_DESCENDING);
  gtk_tree_view_append_column(GTK_TREE_VIEW(selector_widget), column);
  if (column->button != NULL)
    {
      GTK_WIDGET_UNSET_FLAGS(column->button, GTK_CAN_FOCUS);
      gtk_button_set_focus_on_click(GTK_BUTTON(column->button), FALSE);
    }

  restore_session(column, N_RES_UNGOT_WIDTH);
}


void
ThreadlistView::append_n_res_to_read_column()
{
  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  GtkTreeViewColumn *column
    = gtk_tree_view_column_new_with_attributes(_("#Unread"), renderer, NULL);
  g_object_set(renderer, "xalign", 0.9, NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_cell_data_func(column, renderer,
					  column_set_n_res_to_read,
					  this, NULL);
  set_sort_column_id(column, SORT_BY_N_RES_TO_READ, GTK_SORT_DESCENDING);
  gtk_tree_view_append_column(GTK_TREE_VIEW(selector_widget), column);
  if (column->button != NULL)
    {
      GTK_WIDGET_UNSET_FLAGS(column->button, GTK_CAN_FOCUS);
      gtk_button_set_focus_on_click(GTK_BUTTON(column->button), FALSE);
    }

  restore_session(column, N_RES_TO_READ_WIDTH);
}


void
ThreadlistView::append_n_new_res_column()
{
  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  GtkTreeViewColumn *column
    = gtk_tree_view_column_new_with_attributes(_("#NewRes"), renderer, NULL);
  g_object_set(renderer, "xalign", 0.9, NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_cell_data_func(column, renderer,
					  column_set_n_new_res,
					  this, NULL);
  set_sort_column_id(column, SORT_BY_N_NEW_RESPONSES, GTK_SORT_DESCENDING);
  gtk_tree_view_append_column(GTK_TREE_VIEW(selector_widget), column);
  if (column->button != NULL)
    {
      GTK_WIDGET_UNSET_FLAGS(column->button, GTK_CAN_FOCUS);
      gtk_button_set_focus_on_click(GTK_BUTTON(column->button), FALSE);
    }

  restore_session(column, N_NEW_RES_WIDTH);
}


void
ThreadlistView::append_rank_diff_column()
{
  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  GtkTreeViewColumn *column
    = gtk_tree_view_column_new_with_attributes(_("Rank Diffs"),
					       renderer, NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_cell_data_func(column, renderer,
					  column_set_rank_diff,
					  this, NULL);
  gtk_tree_view_append_column(GTK_TREE_VIEW(selector_widget), column);
  if (column->button != NULL)
    {
      GTK_WIDGET_UNSET_FLAGS(column->button, GTK_CAN_FOCUS);
      gtk_button_set_focus_on_click(GTK_BUTTON(column->button), FALSE);
    }

  restore_session(column, RANK_DIFF_WIDTH);
}


void
ThreadlistView::append_birthtime_column()
{
  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  GtkTreeViewColumn *column
    = gtk_tree_view_column_new_with_attributes(_("Birthtime"), renderer, NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_cell_data_func(column, renderer,
					  column_set_birthtime,
					  this, NULL);
  set_sort_column_id(column, SORT_BY_BIRTHTIME, GTK_SORT_DESCENDING);
  gtk_tree_view_append_column(GTK_TREE_VIEW(selector_widget), column);
  if (column->button != NULL)
    {
      GTK_WIDGET_UNSET_FLAGS(column->button, GTK_CAN_FOCUS);
      gtk_button_set_focus_on_click(GTK_BUTTON(column->button), FALSE);
    }

  restore_session(column, BIRTHTIME_WIDTH);
}


void
ThreadlistView::append_last_modified_column()
{
  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  GtkTreeViewColumn *column
    = gtk_tree_view_column_new_with_attributes(_("Last Modified"),
					       renderer, NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_cell_data_func(column, renderer,
					  column_set_last_modified,
					  this, NULL);
  set_sort_column_id(column, SORT_BY_LAST_MODIFIED, GTK_SORT_DESCENDING);
  gtk_tree_view_append_column(GTK_TREE_VIEW(selector_widget), column);
  if (column->button != NULL)
    {
      GTK_WIDGET_UNSET_FLAGS(column->button, GTK_CAN_FOCUS);
      gtk_button_set_focus_on_click(GTK_BUTTON(column->button), FALSE);
    }

  restore_session(column, LAST_MODIFIED_WIDTH);
}


void
ThreadlistView::append_weight_column()
{
  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  GtkTreeViewColumn *column
    = gtk_tree_view_column_new_with_attributes(_("Weight"), renderer, NULL);
  g_object_set(renderer, "xalign", 0.9, NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_cell_data_func(column, renderer,
					  column_set_weight,
					  this, NULL);
  set_sort_column_id(column, SORT_BY_WEIGHT, GTK_SORT_ASCENDING);
  gtk_tree_view_append_column(GTK_TREE_VIEW(selector_widget), column);
  if (column->button != NULL)
    {
      GTK_WIDGET_UNSET_FLAGS(column->button, GTK_CAN_FOCUS);
      gtk_button_set_focus_on_click(GTK_BUTTON(column->button), FALSE);
    }

  restore_session(column, WEIGHT_WIDTH);
  weight_column = column;
}


void
ThreadlistView::append_search_weight_column()
{
  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  GtkTreeViewColumn *column
    = gtk_tree_view_column_new_with_attributes(_("S Weight"), renderer, NULL);
  g_object_set(renderer, "xalign", 0.9, NULL);
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_cell_data_func(column, renderer,
					  column_set_search_weight,
					  this, NULL);
  set_sort_column_id(column, SORT_BY_SEARCH_WEIGHT, GTK_SORT_ASCENDING);
  gtk_tree_view_append_column(GTK_TREE_VIEW(selector_widget), column);
  if (column->button != NULL)
    {
      GTK_WIDGET_UNSET_FLAGS(column->button, GTK_CAN_FOCUS);
      gtk_button_set_focus_on_click(GTK_BUTTON(column->button), FALSE);
    }

  restore_session(column, SEARCH_WEIGHT_WIDTH);
}


bool
ThreadlistView::append_thread(BBSThread *thread)
{
  assert(OCHUSHA_IN_GDK_THREADS());

  if (thread->is_tmp_thread())
    return false;

  SchemeManager &scheme = controller.get_scheme_manager();
  long weight = 0;

  // MEMO: ºݤ˥ɽ뤫ɤϤλǤ餫Ǥʤ
  //       ɽȷޤäǥȤ䤹褦ˤȡweight׻
  //       ɽthread->get_shown_count()ͤۤʤ̤Ȥʤ롣
  //       ξ硢weight׻ˤϡֿװƤΤˡºݤɽ
  //       ǤϡֿǤʤפȽǤ礬ꡢܤŪ
  //       ̷֤ˤꡢ桼𤷤롣
  //       ΰ١ǤϡֲɽȤ鲿ܤפȤ
  //       ꤷƤweight׻줬ɽʤˤ
  //       shown_countͤ򸵤᤹ȤƧࡣ
  if (updating)
    thread->set_shown_count(thread->get_shown_count() + 1);

  if (rule_info != NULL)
    {
      Cell *cell = scheme.get_core().applyf(rule_info->get_procedure(),
					    "%p", thread);
      if (cell != NULL)
	{
	  if (cell->is_number())
	    weight = cell->get_integer() - LONG_MAX;
	  else if (cell == Cell::f())
	    {
	      if (updating)
		thread->set_shown_count(thread->get_shown_count() - 1);

	      return false;
	    }
	}
    }

  if (is_interest_thread(thread))
    ++number_of_interest_threads;

  shown_threads.push_back(thread);
  gtk_list_store_insert_with_values(GTK_LIST_STORE(model), NULL,
				    G_MAXINT32,
				    THREAD_DATA, thread,
				    THREAD_COUNT, ++number_of_threads,
				    THREAD_WEIGHT, weight,
				    THREAD_SEARCH_WEIGHT, 0L,
				    -1);

  return true;
}


gboolean
ThreadlistView::append_threads_on_idle(ThreadlistView *view)
{
  ThreadlistIteratorHandle iter;

  {
    LockBlock lock_obj(list_iter_lock);
    iter = view->list_iter;
  }

  g_return_val_if_fail(iter != NULL, FALSE);


  GDK_THREADS_ENTER();

  bool finished = false;
  gboolean result;

  int lock_id = iter->lock();

  // MAX_NUM_THREADS_TO_APPEND_AT_ONCE⤷list_iter鸽
  // ɤ߹ʬΥ줬̵ʤޤǥǥɲ
  int count = 0;
  while (iter->has_more_threads(lock_id))
    {
      if (view->append_thread(iter->next_thread(lock_id))
	  && ++count >= MAX_NUM_OF_THREADS_TO_APPEND_AT_ONCE)
	break;
    }

  bool error_occurred = false;
  if (!iter->has_more_threads(lock_id))
    {
      view->append_threads_id = 0;
      result = FALSE;

      finished = iter->finished(lock_id);
      if (finished)
	{
	  error_occurred = iter->has_error_occurred();
	  if (!error_occurred && view->number_of_interest_threads > 0)
	    {
	      TabLabel *tab_label = view->get_tab_label();
	      if (tab_label != NULL)
		tab_label->notice(true);
	    }

	  list_iter_lock.lock();
	}
    }
  else
    result = TRUE;

  iter->unlock(lock_id);

  GDK_THREADS_LEAVE();

  if (finished)
    {
      assert(!result);

      view->list_iter = NULL;

      list_iter_lock.unlock();

      TabLabel *tab_label = view->get_tab_label();
      if (tab_label != NULL)
	{
	  if (!error_occurred)
	    tab_label->access_finished();
	  else
	    tab_label->access_failed();
	}
    }


  return result;
}


void
ThreadlistView::state_changed_cb()
{
  // sigslotʥϥɥ
  ThreadlistIteratorHandle iter;
  {
    LockBlock lock_obj(list_iter_lock);
    iter = list_iter;
  }


  GDK_THREADS_ENTER();

  if (append_threads_id == 0 && iter != NULL)
    {
      int lock_id = iter->lock();

      if (iter->has_more_threads(lock_id))
	{
	  // idleؿϿ
	  append_threads_id
	    = g_idle_add_full(THREADLIST_UPDATE_PRIORITY,
			      reinterpret_cast<GSourceFunc>(append_threads_on_idle),
			      this, NULL);
	}

      iter->unlock(lock_id);
    }

  GDK_THREADS_LEAVE();
}


void
ThreadlistView::board_moved_cb()
{
  // sigslotʥϥɥ
  GDK_THREADS_ENTER();

  controller.get_board_ui().notify_board_moved(board);

  GDK_THREADS_LEAVE();
}


static int
compare_count(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
	      gpointer unused)
{
  int a_count;
  int b_count;
  gtk_tree_model_get(model, a, THREAD_COUNT, &a_count, -1);
  gtk_tree_model_get(model, b, THREAD_COUNT, &b_count, -1);

  return a_count - b_count;
}


static int
compare_board(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
	      gpointer unused)
{
  BBSThread *a_thread;
  BBSThread *b_thread;
  int result = 0;
  gtk_tree_model_get(model, a, THREAD_DATA, &a_thread, -1);
  gtk_tree_model_get(model, b, THREAD_DATA, &b_thread, -1);
  if (a_thread == NULL || b_thread == NULL)
    return 0;

  result = strcmp(a_thread->get_board()->get_name(),
		  b_thread->get_board()->get_name());
  if (result != 0)
    return result;

  return compare_weight(model, a, b, unused);
}


static int
compare_rank(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
	     gpointer unused)
{
  BBSThread *a_thread;
  BBSThread *b_thread;
  int result = 0;
  gtk_tree_model_get(model, a, THREAD_DATA, &a_thread, -1);
  gtk_tree_model_get(model, b, THREAD_DATA, &b_thread, -1);
  if (a_thread == NULL || b_thread == NULL)
    return 0;

  if (a_thread->is_alive() && b_thread->is_alive())
    result = a_thread->get_rank() - b_thread->get_rank();
  else if (a_thread->is_alive())
    return -1;
  else if (b_thread->is_alive())
    return 1;
  else
    return compare_title(model, a, b, unused);

  if (result != 0)
    return result;

  return compare_title(model, a, b, unused);
}


static int
compare_mark(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
	     gpointer unused)
{
  BBSThread *a_thread;
  BBSThread *b_thread;
  int result = 0;
  gtk_tree_model_get(model, a, THREAD_DATA, &a_thread, -1);
  gtk_tree_model_get(model, b, THREAD_DATA, &b_thread, -1);
  if (a_thread == NULL || b_thread == NULL)
    return 0;

  result = b_thread->is_marked() - a_thread->is_marked();
  if (result != 0)
    return result;

  return compare_weight(model, a, b, unused);
}


static Regexp title_pattern("(?:[\343\200\220\\[][^\343\200\220\\]]*[\343\200\221\\]])*\\s*([^\343\200\220\\[]*)\\s*(?:[\343\200\220\\[][^\343\200\220\\]]*[\343\200\221\\]])*");


static int
compare_title(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
	      gpointer unused)
{
  BBSThread *a_thread;
  BBSThread *b_thread;
  int result = 0;
  gtk_tree_model_get(model, a, THREAD_DATA, &a_thread, -1);
  gtk_tree_model_get(model, b, THREAD_DATA, &b_thread, -1);
  if (a_thread == NULL || b_thread == NULL)
    return 0;

  const char *a_title0 = a_thread->get_title();
  const char *b_title0 = b_thread->get_title();

  if (a_title0 == NULL && b_title0 == NULL)
    return 0;
  if (a_title0 == NULL)
    return 1;
  if (b_title0 == NULL)
    return -1;

  string a_title;
  if (title_pattern.match(a_title0, strlen(a_title0)))
    a_title.append(title_pattern.match_begin(1), title_pattern.match_len(1));

  string b_title;
  if (title_pattern.match(b_title0, strlen(b_title0)))
    b_title.append(title_pattern.match_begin(1), title_pattern.match_len(1));

  result = strcmp(a_title.c_str(), b_title.c_str());
  if (result != 0)
    return result;

  return compare_birthtime(model, a, b, unused);
}


static int
compare_n_res_verbose(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
		      gpointer unused)
{
  BBSThread *a_thread;
  BBSThread *b_thread;
  int result = 0;
  gtk_tree_model_get(model, a, THREAD_DATA, &a_thread, -1);
  gtk_tree_model_get(model, b, THREAD_DATA, &b_thread, -1);
  if (a_thread == NULL || b_thread == NULL)
    return 0;

  if ((!a_thread->is_alive()) || (!b_thread->is_alive()))
    {
      if (a_thread->is_alive())
	return 1;
      else
	return -1;
    }

  int a_read = a_thread->get_number_of_responses_got();
  int b_read = b_thread->get_number_of_responses_got();
  if ((a_read > 0 || b_read > 0) && (a_read == 0 || b_read == 0))
    return a_read - b_read;

  result = a_thread->get_number_of_responses_to_read()
    - b_thread->get_number_of_responses_to_read();
  if (result != 0)
    return result;

  return compare_weight(model, a, b, unused);
}


static int
compare_n_res_verbose2(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
		       gpointer unused)
{
  BBSThread *a_thread;
  BBSThread *b_thread;
  int result = 0;
  gtk_tree_model_get(model, a, THREAD_DATA, &a_thread, -1);
  gtk_tree_model_get(model, b, THREAD_DATA, &b_thread, -1);
  if (a_thread == NULL || b_thread == NULL)
    return 0;

  int a_read = a_thread->get_number_of_responses_got();
  int b_read = b_thread->get_number_of_responses_got();
  if ((a_read > 0 || b_read > 0) && (a_read == 0 || b_read == 0))
    return a_read - b_read;

  result = a_thread->get_number_of_responses_to_get()
    - b_thread->get_number_of_responses_to_get();
  if (result != 0)
    return result;

  return compare_weight(model, a, b, unused);
}


static int
compare_n_res_avail(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
		    gpointer unused)
{
  BBSThread *a_thread;
  BBSThread *b_thread;
  int result = 0;
  gtk_tree_model_get(model, a, THREAD_DATA, &a_thread, -1);
  gtk_tree_model_get(model, b, THREAD_DATA, &b_thread, -1);
  if (a_thread == NULL || b_thread == NULL)
    return 0;

  result = a_thread->get_number_of_responses_available()
    - b_thread->get_number_of_responses_available();
  if (result != 0)
    return result;

  return compare_weight(model, a, b, unused);
}


static int
compare_n_res_read(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
		   gpointer unused)
{
  BBSThread *a_thread;
  BBSThread *b_thread;
  int result = 0;
  gtk_tree_model_get(model, a, THREAD_DATA, &a_thread, -1);
  gtk_tree_model_get(model, b, THREAD_DATA, &b_thread, -1);
  if (a_thread == NULL || b_thread == NULL)
    return 0;

  result = a_thread->get_number_of_responses_got()
    - b_thread->get_number_of_responses_got();
  if (result != 0)
    return result;

  return compare_weight(model, a, b, unused);
}


static int
compare_n_res_ungot(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
		    gpointer unused)
{
  BBSThread *a_thread;
  BBSThread *b_thread;
  int result = 0;
  gtk_tree_model_get(model, a, THREAD_DATA, &a_thread, -1);
  gtk_tree_model_get(model, b, THREAD_DATA, &b_thread, -1);
  if (a_thread == NULL || b_thread == NULL)
    return 0;

  result = a_thread->get_number_of_responses_to_get()
    - b_thread->get_number_of_responses_to_get();
  if (result != 0)
    return result;

  return compare_weight(model, a, b, unused);
}


static int
compare_n_res_to_read(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
		      gpointer unused)
{
  BBSThread *a_thread;
  BBSThread *b_thread;
  int result = 0;
  gtk_tree_model_get(model, a, THREAD_DATA, &a_thread, -1);
  gtk_tree_model_get(model, b, THREAD_DATA, &b_thread, -1);
  if (a_thread == NULL || b_thread == NULL)
    return 0;

  result = a_thread->get_number_of_responses_to_read()
    - b_thread->get_number_of_responses_to_read();
  if (result != 0)
    return result;

  return compare_weight(model, a, b, unused);
}


static int
compare_n_new_res(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
		  gpointer unused)
{
  BBSThread *a_thread;
  BBSThread *b_thread;
  int result = 0;
  gtk_tree_model_get(model, a, THREAD_DATA, &a_thread, -1);
  gtk_tree_model_get(model, b, THREAD_DATA, &b_thread, -1);
  if (a_thread == NULL || b_thread == NULL)
    return 0;

  result = a_thread->get_number_of_new_responses()
    - b_thread->get_number_of_new_responses();
  if (result != 0)
    return result;

  return compare_weight(model, a, b, unused);
}


static int
compare_birthtime(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
		  gpointer unused)
{
  BBSThread *a_thread;
  BBSThread *b_thread;
  time_t result = 0;
  gtk_tree_model_get(model, a, THREAD_DATA, &a_thread, -1);
  gtk_tree_model_get(model, b, THREAD_DATA, &b_thread, -1);
  if (a_thread == NULL || b_thread == NULL)
    return 0;

  result = a_thread->get_birthtime() - b_thread->get_birthtime();
  if (result != 0)
    return static_cast<int>(result);

  return compare_count(model, a, b, unused);
}


static int
compare_last_modified(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
		      gpointer unused)
{
  BBSThread *a_thread;
  BBSThread *b_thread;
  time_t result = 0;
  gtk_tree_model_get(model, a, THREAD_DATA, &a_thread, -1);
  gtk_tree_model_get(model, b, THREAD_DATA, &b_thread, -1);
  if (a_thread == NULL || b_thread == NULL)
    return 0;

  result = a_thread->get_last_modified_utc()
    - b_thread->get_last_modified_utc();
  if (result != 0)
    return static_cast<int>(result);

  return compare_weight(model, a, b, unused);
}


static int
compare_weight(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
	       gpointer unused)
{
  BBSThread *a_thread;
  BBSThread *b_thread;
  long a_weight;
  long b_weight;
  int result = 0;
  gtk_tree_model_get(model, a,
		     THREAD_DATA, &a_thread,
		     THREAD_WEIGHT, &a_weight,
		     -1);
  gtk_tree_model_get(model, b,
		     THREAD_DATA, &b_thread,
		     THREAD_WEIGHT, &b_weight,
		     -1);
  result = static_cast<int>(a_weight - b_weight);
  if (result != 0)
    return result;

  if (a_thread->is_alive() || b_thread->is_alive())
    return compare_count(model, a, b, unused);

  return compare_title(model, a, b, unused);
}


static int
compare_search_weight(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
		      gpointer unused)
{
  BBSThread *a_thread;
  BBSThread *b_thread;
  long a_weight;
  long b_weight;
  int result = 0;
  gtk_tree_model_get(model, a,
		     THREAD_DATA, &a_thread,
		     THREAD_SEARCH_WEIGHT, &a_weight,
		     -1);
  gtk_tree_model_get(model, b,
		     THREAD_DATA, &b_thread,
		     THREAD_SEARCH_WEIGHT, &b_weight,
		     -1);
  result = static_cast<int>(a_weight - b_weight);
  if (result != 0)
    return result;

  if (a_thread->is_alive() || b_thread->is_alive())
    return compare_count(model, a, b, unused);

  return compare_title(model, a, b, unused);
}


}
