/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: tree_ui.cc,v 1.46 2009/01/01 07:47:07 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "tree_ui.h"

#include "controller.h"
#include "tree_ui_impl.h"
#include "tree_view.h"
#include "ui_constants.h"

#include <algorithm>
#include <cassert>
#include <list>
#include <vector>


namespace Ochusha
{


TreeUI::TreeUI(Controller &controller)
  : controller(controller),
    action_group(create_action_group())
{
}


TreeUI::~TreeUI()
{
  g_object_unref(action_group);
}


GtkActionGroup *
TreeUI::create_action_group()
{
  static GtkActionEntry entries[] = {
    {
      "NewFolder", GTK_STOCK_DIRECTORY,
      N_("New _Folder"), "",
      N_("New Folder"),
      reinterpret_cast<GCallback>(new_folder_cb)
    },
    {
      "HideTreeItem", NULL,
      N_("_Hide"), "",
      N_("Hide"),
      reinterpret_cast<GCallback>(hide_cb)
    },
    {
      "ToggleTreeItemHiddenState", NULL,
      N_("Toggle _Hide"), "",
      N_("Toggle Hide"),
      reinterpret_cast<GCallback>(toggle_hide_cb)
    },
    {
      "CopyTreeItemName", GTK_STOCK_COPY,
      N_("Copy _Name"), "",
      N_("Copy Name"),
      reinterpret_cast<GCallback>(copy_name_cb)
    },
    {
      "CopyTreeItemNameURL", GTK_STOCK_COPY,
      N_("Copy Name _and URL"), "",
      N_("Copy Name and URL"),
      reinterpret_cast<GCallback>(copy_name_url_cb)
    },
    {
      "TreeItemProperties", GTK_STOCK_PROPERTIES,
      N_("_Properties"), NULL,
      N_("Properties"),
      reinterpret_cast<GCallback>(properties_cb)
    }
  };

  GtkActionGroup *action_group = gtk_action_group_new("TreeUIActions");

  gtk_action_group_set_translation_domain(action_group, PACKAGE_NAME);
  gtk_action_group_add_actions(action_group, entries,
			       G_N_ELEMENTS(entries),
			       this);


  new_folder = gtk_action_group_get_action(action_group, "NewFolder");
  hide = gtk_action_group_get_action(action_group, "HideTreeItem");
  toggle_hide = gtk_action_group_get_action(action_group,
					    "ToggleTreeItemHiddenState");
  copy_name = gtk_action_group_get_action(action_group, "CopyTreeItemName");
  copy_name_url = gtk_action_group_get_action(action_group,
					      "CopyTreeItemNameURL");
  properties = gtk_action_group_get_action(action_group, "TreeItemProperties");


  gtk_action_group_set_sensitive(action_group, FALSE);

  g_signal_connect(controller.get_clipboard_internal(), "owner-change",
		   reinterpret_cast<GCallback>(owner_change_cb), this);

  gtk_ui_manager_insert_action_group(controller.get_ui_manager(),
				     action_group, 0);

  return action_group;
}


void
TreeUI::owner_change_cb(GtkClipboard *clipboard, GdkEvent *event, TreeUI *ui)
{
  g_return_if_fail(GTK_IS_CLIPBOARD(clipboard));

  gboolean selection_available
    = TreeView::internal_list_of_elements_available(clipboard);

  gtk_action_group_set_sensitive(ui->action_group, selection_available);

  bool show_hidden
    = ui->controller.get_prefs().get_show_hidden_boardlist_items();
  gtk_action_set_sensitive(ui->hide,
			   !show_hidden && selection_available);
  gtk_action_set_sensitive(ui->toggle_hide,
			   show_hidden && selection_available);
}


void
TreeUI::notify_item_selected(bool item_selected)
{
  gtk_action_set_sensitive(copy_name, item_selected);
  gtk_action_set_sensitive(copy_name_url, item_selected);
}


TreeUIImpl *
TreeUI::get_target_tree_ui_impl()
{
  GtkClipboard *clipboard = controller.get_clipboard_internal();
  g_return_val_if_fail(TreeView::internal_list_of_elements_available(clipboard),
		       NULL);

  return TreeView::get_tree_ui_impl_from_clipboard(clipboard);
}


void
TreeUI::new_folder_cb(GtkAction *action, TreeUI *ui)
{
  TreeUIImpl *impl = ui->get_target_tree_ui_impl();
  g_return_if_fail(impl != NULL);

  impl->new_folder_selection();
}


void
TreeUI::hide_cb(GtkAction *action, TreeUI *ui)
{
  TreeUIImpl *impl = ui->get_target_tree_ui_impl();
  g_return_if_fail(impl != NULL);

  impl->hide_selection();
}


void
TreeUI::toggle_hide_cb(GtkAction *action, TreeUI *ui)
{
  TreeUIImpl *impl = ui->get_target_tree_ui_impl();
  g_return_if_fail(impl != NULL);

  impl->toggle_hide_selection();
}


void
TreeUI::copy_name_cb(GtkAction *action, TreeUI *ui)
{
  TreeUIImpl *impl = ui->get_target_tree_ui_impl();
  g_return_if_fail(impl != NULL);

  impl->copy_name_selection();
}


void
TreeUI::copy_name_url_cb(GtkAction *action, TreeUI *ui)
{
  TreeUIImpl *impl = ui->get_target_tree_ui_impl();
  g_return_if_fail(impl != NULL);

  impl->copy_name_url_selection();
}


void
TreeUI::properties_cb(GtkAction *action, TreeUI *ui)
{
  TreeUIImpl *impl = ui->get_target_tree_ui_impl();
  g_return_if_fail(impl != NULL);

  impl->open_properties_selection();
}


}
