/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: tree_ui_impl.cc,v 1.30 2009/01/01 07:47:07 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "tree_ui_impl.h"

#include "controller.h"
#include "focus_utils.h"
#include "gtk_utils.h"
#include "tree_view.h"
#include "ui_constants.h"

#include "libochushagtk/stock.h"

#include <algorithm>
#include <cassert>
#include <list>
#include <vector>


namespace Ochusha
{


TreeUIImpl::TreeUIImpl(Controller &controller, RootFolder *root)
  : controller(controller),
    root_folder(root),
    master_store(gtk_tree_store_new(TREE_ROW_N_COLUMNS,
				    G_TYPE_POINTER, G_TYPE_INT)),
    view_model(gtk_tree_model_filter_new(GTK_TREE_MODEL(master_store), NULL)),
    menu_model(gtk_tree_model_filter_new(GTK_TREE_MODEL(master_store), NULL)),
    folder_model(gtk_tree_model_filter_new(GTK_TREE_MODEL(master_store), NULL)),
    menu_helper(ochusha_tree_menu_helper_new_with_model(menu_model)),
    menu_shell(NULL),
    last_view (NULL),
    last_selection(NULL),
    periodic_save_id(0),
    save_on_idle_id(0),
    master_view(NULL),
    ignore_notify_folder_modified(false)
{
  gtk_tree_model_filter_set_visible_func(GTK_TREE_MODEL_FILTER(view_model),
					 is_view_iter, this, NULL);


  gtk_tree_model_filter_set_visible_func(GTK_TREE_MODEL_FILTER(menu_model),
					 is_visible_menu_iter, this, NULL);
  g_signal_connect(menu_helper, "activate",
		   reinterpret_cast<GCallback>(menu_activate_cb), this);
  ochusha_tree_menu_helper_set_menu_data_func(menu_helper, update_menu_entry,
					      this, NULL);
  ochusha_tree_menu_helper_set_row_separator_func(menu_helper,
						  is_separator_iter,
						  this, NULL);
  ochusha_tree_menu_helper_set_row_sensitive_func(menu_helper,
						  is_sensitive_iter,
						  this, NULL);


  gtk_tree_model_filter_set_visible_func(GTK_TREE_MODEL_FILTER(folder_model),
					 is_folder_iter, this, NULL);


  g_signal_connect(master_store, "row_deleted",
		   reinterpret_cast<GCallback>(row_deleted_cb), this);


  register_important_folder(root);


  GDK_THREADS_LEAVE();

  root_folder->tree_updated.connect(this, &TreeUIImpl::tree_updated_cb);

  GDK_THREADS_ENTER();


  periodic_save_id
    = g_timeout_add(PERIODIC_SAVE_TIMEOUT,
		    reinterpret_cast<GSourceFunc>(periodic_save_timeout_cb),
		    this);

  const gchar *title = canonicalize_element_name(root_folder);
  master_view = new TreeView(view_model, title, *this, true);
  GtkWidget *widget = master_view->get_widget();
  g_signal_connect(widget, "destroy",
		   reinterpret_cast<GCallback>(gtk_widget_destroyed),
		   &master_view);
}


TreeUIImpl::~TreeUIImpl()
{
  GDK_THREADS_LEAVE();

  disconnect_all();
  important_folders.clear();
  all_folders.clear();

  root_folder = NULL;

  GDK_THREADS_ENTER();


  if (periodic_save_id != 0)
    g_source_remove(periodic_save_id);

  if (save_on_idle_id != 0)
    g_source_remove(save_on_idle_id);

  g_object_unref(menu_helper);
  g_object_unref(folder_model);
  g_object_unref(menu_model);
  g_object_unref(view_model);
  g_object_unref(master_store);
}


gboolean
TreeUIImpl::periodic_save_timeout_cb(TreeUIImpl *impl)
{
  // g_timeout_addΥХåGLibɳʤΤGDK_THREADS_ENTER()Ǥʤ
  GDK_THREADS_ENTER();

  if (impl->save_on_idle_id == 0 && impl->root_folder->is_dirty())
    impl->save_on_idle_id
      = g_idle_add_full(G_PRIORITY_LOW,
			reinterpret_cast<GSourceFunc>(save_on_idle_cb),
			impl, NULL);

  GDK_THREADS_LEAVE();

  return TRUE;	// Ǹޤǽʤ
}


gboolean
TreeUIImpl::save_on_idle_cb(TreeUIImpl *impl)
{
  if (libochusha_giant.trylock())
    {
      GDK_THREADS_ENTER();

      impl->save_tree_xml();
      impl->save_on_idle_id = 0;

      GDK_THREADS_LEAVE();


      libochusha_giant.unlock();

      return FALSE;	// saveλ
    }

  return TRUE;		// saveäΤǺĩ
}


void
TreeUIImpl::save_tree_xml()
{
}


TreeView *
TreeUIImpl::create_tree_view(const gchar *title, bool is_master_view)
{
  TreeView *view = new TreeView(view_model, title, *this, is_master_view);

  return view;
}


TreeView *
TreeUIImpl::create_folder_tree_view(const gchar *title)
{
  TreeView *view
    = new FolderTreeView(folder_model,
			 (title != NULL
			  ? title : canonicalize_element_name(root_folder)),
			 *this);

  GtkTreeView *tree_view = GTK_TREE_VIEW(view->get_tree_view());
  gtk_tree_view_set_headers_clickable(tree_view, TRUE);

  return view;
}


View *
TreeUIImpl::get_master_view()
{
  return master_view;
}


GtkWidget *
TreeUIImpl::get_master_widget()
{
  return master_view->get_widget();
}


void
TreeUIImpl::restore_session()
{
  master_view->restore_session(get_ui_name(), controller.get_session());
}


void
TreeUIImpl::prepare_session()
{
  master_view->prepare_session(get_ui_name(), controller.get_session());
}


const char *
TreeUIImpl::canonicalize_element_name(TreeElement *element)
{
  const char *name = element->get_name();
  if (name == NULL)
    return _("Anonymous Item");

  if (!element->is_folder())
    return name;

  Folder *folder = dynamic_cast<Folder *>(element);
  g_return_val_if_fail(folder != NULL, name);

  switch (folder->get_folder_type())
    {
    case FOLDER_TYPE_ROOT_FOLDER:
      return dgettext(PACKAGE, name);

    default:
      return name;
    }
}


// եȤοѲ˱ơ˥塼ɽƤѲ뤿
// row_deletedʥȿ롣
void
TreeUIImpl::row_deleted_cb(GtkTreeModel *model, GtkTreePath *path,
			   TreeUIImpl *impl)
{
  // modelmaster_store
  if (gtk_tree_path_get_depth(path) == 1)
    return;

  TreePathHandle copy_path(gtk_tree_path_copy(path));
  if (gtk_tree_path_up(copy_path))
    {
      GtkTreeIter iter;
      if (gtk_tree_model_get_iter(model, &iter, copy_path))
	{
	  TreeElement *element = NULL;
	  TreeRowType row_type = TREE_ROW_TYPE_UNKNOWN;
	  gtk_tree_model_get(model, &iter,
			     TREE_ELEMENT_DATA, &element,
			     TREE_ROW_TYPE, &row_type,
			     -1);
	  if (row_type == TREE_ROW_TYPE_ELEMENT
	      && element != NULL && element->is_folder())
	    impl->notify_folder_modified(dynamic_cast<Folder *>(element));
	}
    }
}


gboolean
TreeUIImpl::is_view_iter(GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
  TreeRowType row_type = TREE_ROW_TYPE_UNKNOWN;
  gtk_tree_model_get(model, iter, TREE_ROW_TYPE, &row_type, -1);

  return row_type == TREE_ROW_TYPE_ELEMENT;
}


gboolean
TreeUIImpl::is_visible_menu_iter(GtkTreeModel *model, GtkTreeIter *iter,
				 gpointer data)
{
  TreeElement *element = NULL;
  TreeRowType row_type = TREE_ROW_TYPE_UNKNOWN;
  gtk_tree_model_get(model, iter,
		     TREE_ELEMENT_DATA, &element,
		     TREE_ROW_TYPE, &row_type,
		     -1);

  switch (row_type)
    {
    case TREE_ROW_TYPE_ELEMENT:
      return TRUE;

    case TREE_ROW_TYPE_EMPTY_FOLDER_INDICATOR:
      {
	g_return_val_if_fail(element != NULL, FALSE);
	g_return_val_if_fail(element->is_folder(), FALSE);
	Folder *folder = dynamic_cast<Folder *>(element);

	TreeElementList::const_iterator list_iter = folder->elements_begin();
	TreeElementList::const_iterator end_iter = folder->elements_end();
	bool show_all = static_cast<TreeUIImpl *>(data)->show_all();
	int n = 0;
	while (list_iter != end_iter)
	  {
	    TreeElement *elm = *list_iter;
	    if (!elm->is_folder() && !elm->is_separator()
		&& (show_all || !elm->is_hidden()) && ++n > 0)
	      return FALSE;
	    ++list_iter;
	  }
	return TRUE;
      }

    case TREE_ROW_TYPE_INTERNAL_SEPARATOR:
    case TREE_ROW_TYPE_OPEN_IN_TAB_MENU:
      {
	g_return_val_if_fail(element != NULL, FALSE);
	g_return_val_if_fail(element->is_folder(), FALSE);
	Folder *folder = dynamic_cast<Folder *>(element);
	TreeElementList::const_iterator list_iter = folder->elements_begin();
	TreeElementList::const_iterator end_iter = folder->elements_end();
	bool show_all = static_cast<TreeUIImpl *>(data)->show_all();
	int n = 0;
	while (list_iter != end_iter)
	  {
	    TreeElement *elm = *list_iter;
	    if (!elm->is_folder() && !elm->is_separator()
		&& (show_all || !elm->is_hidden()) && ++n > 1)
	      return TRUE;
	    ++list_iter;
	  }
	return FALSE;
      }

    case TREE_ROW_TYPE_UNKNOWN:
      return FALSE;
    }

  g_return_val_if_fail(FALSE, FALSE);	// should never reached here!
  return FALSE;
}


gboolean
TreeUIImpl::is_separator_iter(GtkTreeModel *model, GtkTreeIter *iter,
			      gpointer data)
{
  TreeElement *element = NULL;
  TreeRowType row_type = TREE_ROW_TYPE_UNKNOWN;
  gtk_tree_model_get(model, iter,
		     TREE_ELEMENT_DATA, &element,
		     TREE_ROW_TYPE, &row_type,
		     -1);

  return row_type == TREE_ROW_TYPE_INTERNAL_SEPARATOR
    || (row_type == TREE_ROW_TYPE_ELEMENT && element->is_separator());
}


gboolean
TreeUIImpl::is_sensitive_iter(GtkTreeModel *model, GtkTreeIter *iter,
			      gpointer data)
{
  TreeRowType row_type = TREE_ROW_TYPE_UNKNOWN;
  gtk_tree_model_get(model, iter, TREE_ROW_TYPE, &row_type, -1);

  return row_type != TREE_ROW_TYPE_EMPTY_FOLDER_INDICATOR;
}


gboolean
TreeUIImpl::is_folder_iter(GtkTreeModel *model, GtkTreeIter *iter,
			   gpointer data)
{
  TreeElement *element = NULL;
  TreeRowType row_type = TREE_ROW_TYPE_UNKNOWN;
  gtk_tree_model_get(model, iter,
		     TREE_ELEMENT_DATA, &element,
		     TREE_ROW_TYPE, &row_type,
		     -1);

  if (row_type != TREE_ROW_TYPE_ELEMENT)
    return FALSE;

#if GTK_CHECK_VERSION(2,10,0)
  g_return_val_if_fail(element != NULL, FALSE);
#else
  if (element == NULL)
    return FALSE;
#endif

  if (!element->is_folder())
    return FALSE;

  Folder *folder = dynamic_cast<Folder *>(element);
  g_return_val_if_fail(folder != NULL, FALSE);

  switch (folder->get_folder_type())
    {
    case FOLDER_TYPE_OCHUSHA_MY_BOOKMARK:
    case FOLDER_TYPE_PSEUDO_FOLDER:
    case FOLDER_TYPE_OABF_FOLDER:
      return FALSE;

    default:
      return TRUE;
    }
}


void
TreeUIImpl::menu_activate_cb(OchushaTreeMenuHelper *helper,
			     GtkTreeModel *model, GtkTreeIter *iter,
			     GtkTreePath *path, guint button, TreeUIImpl *impl)
{
  // modelmenu_model
  TreeElement *element = NULL;
  TreeRowType row_type = TREE_ROW_TYPE_UNKNOWN;
  gtk_tree_model_get(model, iter,
		     TREE_ELEMENT_DATA, &element,
		     TREE_ROW_TYPE, &row_type,
		     -1);

  g_return_if_fail(row_type != TREE_ROW_TYPE_UNKNOWN);
  g_return_if_fail(row_type != TREE_ROW_TYPE_EMPTY_FOLDER_INDICATOR);
  g_return_if_fail(row_type != TREE_ROW_TYPE_INTERNAL_SEPARATOR);

  if (row_type == TREE_ROW_TYPE_OPEN_IN_TAB_MENU)
    {
      g_return_if_fail(element != NULL);
      g_return_if_fail(element->is_folder());
      impl->open(element, true, false, true);
      return;
    }

  g_return_if_fail(row_type == TREE_ROW_TYPE_ELEMENT);
  g_return_if_fail(element != NULL);
  g_return_if_fail(!element->is_separator());

  if (element->is_folder())
    {
      if (button == 2)
	{
	  impl->open(element, true, false, true);
	  // XXX
	}
      return;
    }

  bool in_tab
    = (button == 2) ^ (impl->controller.get_prefs().get_default_open_in_tab());
  impl->open(element, in_tab, false, true);
}


struct FolderCompare
{
  inline bool operator()(FolderHandle &x, FolderHandle &y)
  {
    return static_cast<Folder *>(x)->get_time_of_last_modification()
      >= static_cast<Folder *>(y)->get_time_of_last_modification();
  }
};


// all_foldersĹnФO(n)ʤΤǡˡա
void
TreeUIImpl::register_folder(Folder *folder)
{
  list<FolderHandle> new_folder_list;
  new_folder_list.push_front(folder);

  list<FolderHandle>::iterator where
    = find(all_folders.begin(), all_folders.end(), folder);
  if (where != all_folders.end())
    all_folders.erase(where);

  all_folders.merge(new_folder_list, FolderCompare());
}


void
TreeUIImpl::register_important_folder(Folder *folder)
{
  if (find(important_folders.begin(), important_folders.end(), folder)
      != important_folders.end())
    return;	// Ͽ

  important_folders.push_back(folder);
}


const gchar *
TreeUIImpl::get_element_icon_stock_id(TreeElement *element, gboolean expanded)
{
  return element->is_folder()
    ? (expanded ? OCHUSHA_STOCK_FOLDER_OPEN : OCHUSHA_STOCK_FOLDER_CLOSED)
    : NULL;
}


void
TreeUIImpl::update_menu_entry(GtkMenuItem *menu_item, GtkLabel *menu_label,
			      GtkTreeModel *model, GtkTreeIter *iter,
			      gpointer data)
{
  TreeUIImpl *impl = static_cast<TreeUIImpl *>(data);
  TreeElement *element = NULL;
  TreeRowType row_type = TREE_ROW_TYPE_UNKNOWN;
  gtk_tree_model_get(model, iter,
		     TREE_ELEMENT_DATA, &element,
		     TREE_ROW_TYPE, &row_type,
		     -1);

#if GTK_CHECK_VERSION(2,10,0)
  g_return_if_fail(row_type != TREE_ROW_TYPE_UNKNOWN);
#else
  if (row_type == TREE_ROW_TYPE_UNKNOWN)
    return;
#endif

  g_return_if_fail(row_type != TREE_ROW_TYPE_INTERNAL_SEPARATOR);

  if (row_type == TREE_ROW_TYPE_ELEMENT)
    {
      g_return_if_fail(element != NULL);
      g_return_if_fail(!element->is_separator());

      gtk_label_set_text(menu_label, canonicalize_element_name(element));
      if (!GTK_IS_IMAGE_MENU_ITEM(menu_item))
	return;

      const gchar *id = impl->get_element_icon_stock_id(element, FALSE);
      GtkWidget *icon;
      if (id != NULL)
	{
	  icon = gtk_image_new_from_stock(id, GTK_ICON_SIZE_MENU);
	  gtk_widget_show(icon);
	}
      else
	icon = NULL;
      gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(menu_item), icon);
      return;
    }

  const char *name;
  if (row_type == TREE_ROW_TYPE_EMPTY_FOLDER_INDICATOR)
    name = _("(Empty)");
  else
    {
      g_return_if_fail(row_type == TREE_ROW_TYPE_OPEN_IN_TAB_MENU);
      name = _("_Open in Tab");
    }
  gtk_label_set_text_with_mnemonic(menu_label, name);
  gtk_label_set_mnemonic_widget(menu_label, GTK_WIDGET(menu_item));
  if (GTK_IS_IMAGE_MENU_ITEM(menu_item))
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(menu_item), NULL);
}


bool
TreeUIImpl::locate_element(GtkTreeModel *model, GtkTreeIter *iter,
			   Folder **parent_p, TreeElement **element_p)
{
  TreeElement *element = NULL;
  TreeRowType row_type = TREE_ROW_TYPE_UNKNOWN;
  gtk_tree_model_get(model, iter,
		     TREE_ELEMENT_DATA, &element,
		     TREE_ROW_TYPE, &row_type,
		     -1);
  g_return_val_if_fail(row_type == TREE_ROW_TYPE_ELEMENT, false);
  g_return_val_if_fail(element != NULL, false);

  Folder *folder;
  GtkTreeIter parent_iter;
  if (gtk_tree_model_iter_parent(model, &parent_iter, iter))
    {
      TreeElement *parent = NULL;
      gtk_tree_model_get(model, &parent_iter, TREE_ELEMENT_DATA, &parent, -1);
      g_return_val_if_fail(parent != NULL, false);
      g_return_val_if_fail(parent->is_folder(), false);

      folder = dynamic_cast<Folder *>(parent);
      g_return_val_if_fail(folder != NULL, false);
    }
  else
    folder = root_folder;

  *parent_p = folder;
  *element_p = element;
  return true;
}


bool
TreeUIImpl::convert_iter_to_master_iter(GtkTreeModel *model,
					GtkTreeIter *master_iter,
					GtkTreeIter *iter)
{
  g_return_val_if_fail(master_store != NULL, false);

  GtkTreeModel *master_model = GTK_TREE_MODEL(master_store);
  if (model == master_model)
    {
      TreePathHandle path(gtk_tree_model_get_path(model, iter));
      g_return_val_if_fail(path != NULL, false);

      gboolean result = gtk_tree_model_get_iter(model, master_iter, path);
      return result == TRUE;
    }

  if (GTK_IS_TREE_MODEL_FILTER(model))
    {
      GtkTreeModelFilter *filter_model = GTK_TREE_MODEL_FILTER(model);
      g_return_val_if_fail(gtk_tree_model_filter_get_model(filter_model)
			   == master_model, false);
      gtk_tree_model_filter_convert_iter_to_child_iter(filter_model,
						       master_iter,
						       iter);
      return true;
    }

  return false;
}


bool
TreeUIImpl::convert_master_iter_to_iter(GtkTreeModel *model,
					GtkTreeIter *iter,
					GtkTreeIter *master_iter)
{
  g_return_val_if_fail(master_store != NULL, false);

  GtkTreeModel *master_model = GTK_TREE_MODEL(master_store);
  if (model == master_model)
    {
      TreePathHandle path(gtk_tree_model_get_path(model, master_iter));
      g_return_val_if_fail(path != NULL, false);

      gboolean result = gtk_tree_model_get_iter(model, iter, path);
      return result == TRUE;
    }

  if (GTK_IS_TREE_MODEL_FILTER(model))
    {
      GtkTreeModelFilter *filter_model = GTK_TREE_MODEL_FILTER(model);
#if GTK_CHECK_VERSION(2,10,0)
      return (gtk_tree_model_filter_convert_child_iter_to_iter(filter_model,
							       iter,
							       master_iter)
	      == TRUE);
#else
      gtk_tree_model_filter_convert_child_iter_to_iter(filter_model,
						       iter, master_iter);
      return gtk_tree_store_iter_is_valid(master_store, master_iter);
#endif
    }

  return false;
}


bool
TreeUIImpl::get_master_iter(GtkTreeModel *model, GtkTreeIter *master_iter,
			    GtkTreeRowReference *reference)
{
  g_return_val_if_fail(master_store != NULL, false);
#if GTK_CHECK_VERSION(2,8,0)
  g_return_val_if_fail(model == gtk_tree_row_reference_get_model(reference),
		       false);
#endif
  TreePathHandle path(gtk_tree_row_reference_get_path(reference));
  g_return_val_if_fail(path != NULL, false);

  GtkTreeModel *master_model = GTK_TREE_MODEL(master_store);
  if (master_model == model)
    return gtk_tree_model_get_iter(model, master_iter, path) != FALSE;

  if (GTK_IS_TREE_MODEL_FILTER(model))
    {
      GtkTreeIter filter_iter;
      if (gtk_tree_model_get_iter(model, &filter_iter, path))
	{
	  GtkTreeModelFilter *filter_model = GTK_TREE_MODEL_FILTER(model);
	  gtk_tree_model_filter_convert_iter_to_child_iter(filter_model,
							   master_iter,
							   &filter_iter);
	  return true;
	}
    }

  return false;
}


struct ParentIterFinder
{
  FolderHandle parent;
  TreeElementHandle sibling;

  GtkTreeIter iter;

  GtkTreeIter *parent_iter;
  TreePathHandle sibling_path;

  static gboolean foreach_func(GtkTreeModel *model, GtkTreePath *path,
			       GtkTreeIter *iter, gpointer data);

  ParentIterFinder(GtkTreeModel *model, Folder *parent, TreeElement *sibling);
};


ParentIterFinder::ParentIterFinder(GtkTreeModel *model,
				   Folder *parent, TreeElement *sibling)
  : parent(parent), sibling(sibling), parent_iter(NULL), sibling_path(NULL)
{
  gtk_tree_model_foreach(model, foreach_func, this);
}


gboolean
ParentIterFinder::foreach_func(GtkTreeModel *model, GtkTreePath *path,
			       GtkTreeIter *iter, gpointer data)
{
  ParentIterFinder *finder = static_cast<ParentIterFinder *>(data);

  TreeElement *element = NULL;
  gtk_tree_model_get(model, iter, TREE_ELEMENT_DATA, &element, -1);
  g_return_val_if_fail(element != NULL, FALSE);

  if (static_cast<TreeElement *>(finder->sibling) != NULL)
    {
      if (static_cast<TreeElement *>(finder->sibling) != element)
	return FALSE;

      // ȯ
      finder->sibling_path = gtk_tree_path_copy(path);

      if (gtk_tree_model_iter_parent(model, &finder->iter, iter))
	finder->parent_iter = &finder->iter;
      return TRUE;
    }

  assert(static_cast<Folder *>(finder->parent) != NULL);

  if (!element->is_folder() || static_cast<Folder *>(finder->parent) != element)
    return FALSE;

  if (gtk_tree_model_get_iter(model, &finder->iter, path))
    finder->parent_iter = &finder->iter;
  return TRUE;
}


struct FolderIterFinder
{
  FolderHandle folder;

  GtkTreeIter iter;

  GtkTreeIter *folder_iter;

  static gboolean foreach_func(GtkTreeModel *model, GtkTreePath *path,
			       GtkTreeIter *iter, gpointer data);

  FolderIterFinder(GtkTreeModel *model, Folder *folder);
};


FolderIterFinder::FolderIterFinder(GtkTreeModel *model, Folder *folder)
  : folder(folder), folder_iter(NULL)
{
  assert(folder != NULL);

  gtk_tree_model_foreach(model, foreach_func, this);
}


gboolean
FolderIterFinder::foreach_func(GtkTreeModel *model, GtkTreePath *path,
			       GtkTreeIter *iter, gpointer data)
{
  FolderIterFinder *finder = static_cast<FolderIterFinder *>(data);

  TreeElement *element = NULL;
  gtk_tree_model_get(model, iter, TREE_ELEMENT_DATA, &element, -1);
  g_return_val_if_fail(element != NULL, FALSE);

  if (!element->is_folder() || static_cast<Folder *>(finder->folder) != element)
    return FALSE;

  if (gtk_tree_model_get_iter(model, &finder->iter, path))
    finder->folder_iter = &finder->iter;

  return TRUE;
}


struct ElementIterFinder
{
  TreeElementHandle element;

  GtkTreeIter iter;

  GtkTreeIter *element_iter;
  TreePathHandle element_path;

  static gboolean foreach_func(GtkTreeModel *model, GtkTreePath *path,
			       GtkTreeIter *iter, gpointer data);

  ElementIterFinder(GtkTreeModel *model, TreeElement *element);
};


ElementIterFinder::ElementIterFinder(GtkTreeModel *model, TreeElement *element)
  : element(element),
    element_iter(NULL),
    element_path(NULL)
{
  gtk_tree_model_foreach(model, foreach_func, this);
}


gboolean
ElementIterFinder::foreach_func(GtkTreeModel *model, GtkTreePath *path,
				GtkTreeIter *iter, gpointer data)
{
  ElementIterFinder *finder = static_cast<ElementIterFinder *>(data);

  TreeElement *element = NULL;
  gtk_tree_model_get(model, iter, TREE_ELEMENT_DATA, &element, -1);
  g_return_val_if_fail(element != NULL, FALSE);

  if (element == finder->element)
    {
      // ȯ
      finder->element_path = gtk_tree_path_copy(path);
      if (gtk_tree_model_get_iter(model, &finder->iter, path))
	finder->element_iter = &finder->iter;

      return TRUE;
    }

  return FALSE;
}


enum {
  FOLDER_LIST_FOLDER_DATA = 0,
  FOLDER_LIST_ROW_REFERENCE,
  FOLDER_LIST_N_COLUMNS
};


class AddElementDialog
{
private:
  TreeUIImpl &impl;
  TreeElementHandle element;
  GtkWidget *dialog;

  Folder *current_folder;

  GtkEntry *name_entry;

  GtkTreeModel *folder_list;
  GtkComboBox *folder_combo;

  vector<FolderHandle> folder_combo_folders;

  GtkWidget *folder_tree_window;
  GtkTreeView *folder_tree_view;
  GtkTreeSelection *selection;

  guint row_deleted_id;

  static gboolean collect_invalid_references(GtkTreeModel *model,
					     GtkTreePath *path,
					     GtkTreeIter *iter, gpointer data);

  static void row_deleted_cb(GtkTreeModel *model, GtkTreePath *path,
			     AddElementDialog *dialog);

  static gboolean set_current_folder(GtkTreeModel *model, GtkTreePath *path,
				     GtkTreeIter *iter, gpointer data);
  static gboolean select_current_folder(GtkTreeModel *model, GtkTreePath *path,
					GtkTreeIter *iter, gpointer data);

  static void header_clicked_cb(GtkTreeViewColumn *column,
				AddElementDialog *dialog);
  static void combo_changed_cb(GtkComboBox *combo, AddElementDialog *dialog);
  static void selection_changed_cb(GtkTreeSelection *selection,
				   AddElementDialog *dialog);
  static void new_folder_cb(GtkButton *button, AddElementDialog *dialog);
  static void response_cb(GtkWidget *widget, int response_id,
			  AddElementDialog *dialog);
  static void dialog_destroyed_cb(GtkWidget *widget, AddElementDialog *dialog);

  bool ensure_combo_entry(Folder *folder, bool select_entry);

  ~AddElementDialog();


public:
  AddElementDialog(TreeUIImpl &impl, TreeElement *element, const gchar *title);
};


static gboolean
is_separator(GtkTreeModel *model, GtkTreeIter *iter, gpointer unused)
{
  Folder *folder = NULL;
  gtk_tree_model_get(model, iter, 0, &folder, -1);
  return folder == NULL;
}


static void
set_folder_name(GtkCellLayout *layout, GtkCellRenderer *renderer,
		GtkTreeModel *model, GtkTreeIter *iter, gpointer unused)
{
  Folder *folder = NULL;
  gtk_tree_model_get(model, iter, 0, &folder, -1);
  if (folder != NULL)
    g_object_set(renderer, "text",
		 TreeUIImpl::canonicalize_element_name(folder), NULL);
}


#define RESPONSE_NEW_FOLDER	1

#define TABLE_ATTACH_CAPTION(table, widget, l, r, t, b)			\
  gtk_table_attach(GTK_TABLE(table), widget, l, r, t, b,		\
		   static_cast<GtkAttachOptions>(GTK_SHRINK | GTK_FILL),\
		   static_cast<GtkAttachOptions>(GTK_SHRINK | GTK_FILL), 3, 3)


AddElementDialog::AddElementDialog(TreeUIImpl &impl, TreeElement *element,
				   const gchar *title)
  : impl(impl),
    element(element),
    current_folder(NULL)
{
  dialog = gtk_dialog_new_with_buttons(title,
				       impl.get_controller().get_toplevel(),
				       GTK_DIALOG_DESTROY_WITH_PARENT,
				       _("New F_older"), RESPONSE_NEW_FOLDER,
				       GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
				       GTK_STOCK_ADD, GTK_RESPONSE_OK,
				       NULL);
  gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_OK);
  gtk_window_set_default_size(GTK_WINDOW(dialog), -1, 300);
  g_signal_connect(dialog, "response",
		   reinterpret_cast<GCallback>(response_cb), this);
  g_signal_connect(dialog, "destroy",
		   reinterpret_cast<GCallback>(dialog_destroyed_cb), this);

  GtkBox *outer_vbox = GTK_BOX(GTK_DIALOG(dialog)->vbox);

  GtkWidget *table = gtk_table_new(3, 2, FALSE);
  gtk_widget_show(table);
  gtk_box_pack_start(outer_vbox, table, TRUE, TRUE, 10);


  GtkWidget *label = gtk_label_new_with_mnemonic(_("_Name:"));
  gtk_misc_set_alignment(GTK_MISC(label), 1.0, 0.5);
  gtk_widget_show(label);
  TABLE_ATTACH_CAPTION(table, label, 0, 1, 0, 1);

  GtkWidget *entry = EntryFocusNotifier::gtk_entry_new(impl.get_controller());
  name_entry = GTK_ENTRY(entry);
  gtk_entry_set_width_chars(name_entry, 36);
  gtk_entry_set_activates_default(name_entry, TRUE);
  const char *name = element->get_name();
  if (name == NULL)
    name = _("New Element");
  gtk_entry_set_text(name_entry, name);
  gtk_widget_show(entry);
  gtk_table_attach(GTK_TABLE(table), entry, 1, 2, 0, 1,
		   static_cast<GtkAttachOptions>(GTK_EXPAND|GTK_FILL),
		   static_cast<GtkAttachOptions>(GTK_SHRINK|GTK_FILL), 3, 3);
  gtk_label_set_mnemonic_widget(GTK_LABEL(label), entry);


  label = gtk_label_new_with_mnemonic(_("_Create in:"));
  gtk_misc_set_alignment(GTK_MISC(label), 1.0, 0.5);
  gtk_widget_show(label);
  TABLE_ATTACH_CAPTION(table, label, 0, 1, 1, 2);

  GtkListStore *store = gtk_list_store_new(FOLDER_LIST_N_COLUMNS,
					   G_TYPE_POINTER,
					   G_TYPE_POINTER);
  folder_list = GTK_TREE_MODEL(store);

  GtkWidget *combo = gtk_combo_box_new_with_model(folder_list);
  gtk_widget_show(combo);
  gtk_table_attach(GTK_TABLE(table), combo, 1, 2, 1, 2,
		   static_cast<GtkAttachOptions>(GTK_EXPAND|GTK_FILL),
		   static_cast<GtkAttachOptions>(GTK_SHRINK|GTK_FILL), 3, 3);
  g_signal_connect(combo, "changed",
		   reinterpret_cast<GCallback>(combo_changed_cb), this);

  gtk_label_set_mnemonic_widget(GTK_LABEL(label), combo);
  folder_combo = GTK_COMBO_BOX(combo);
  gtk_combo_box_set_row_separator_func(folder_combo, is_separator,
				       NULL, NULL);
  gtk_combo_box_set_focus_on_click(folder_combo, FALSE);

  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(combo), renderer, TRUE);
  gtk_cell_layout_set_cell_data_func(GTK_CELL_LAYOUT(combo), renderer,
				     set_folder_name, NULL, NULL);
  

  TreeView *folder_view = impl.create_folder_tree_view();
  folder_tree_window = folder_view->get_widget();
  GtkWidget *view = folder_view->get_tree_view();
  GtkTreeModel *folder_model = gtk_tree_view_get_model(GTK_TREE_VIEW(view));
  row_deleted_id
    = g_signal_connect(folder_model, "row_deleted",
		       reinterpret_cast<GCallback>(row_deleted_cb), this);

  gtk_widget_show(folder_tree_window);
  gtk_table_attach(GTK_TABLE(table), folder_tree_window, 0, 2, 2, 3,
		   static_cast<GtkAttachOptions>(GTK_EXPAND | GTK_FILL),
		   static_cast<GtkAttachOptions>(GTK_EXPAND | GTK_FILL), 3, 5);

  folder_tree_view = GTK_TREE_VIEW(view);
  selection = gtk_tree_view_get_selection(folder_tree_view);
  g_signal_connect(selection, "changed",
		   reinterpret_cast<GCallback>(selection_changed_cb), this);

  GtkTreeViewColumn *column = gtk_tree_view_get_column(folder_tree_view, 0);
  gtk_tree_view_column_set_clickable(column, TRUE);
  GTK_WIDGET_UNSET_FLAGS(column->button, GTK_CAN_FOCUS);
  gtk_button_set_focus_on_click(GTK_BUTTON(column->button), FALSE);
#if GTK_CHECK_VERSION(2,12,0)
  gtk_widget_set_tooltip_text(column->button,
			      _("Click to choose the root folder"));
#endif
  g_signal_connect(column, "clicked",
		   reinterpret_cast<GCallback>(header_clicked_cb), this);


  // ץեϿ
  list<FolderHandle>::iterator iter = impl.important_folders.begin();
  list<FolderHandle>::iterator end_iter = impl.important_folders.end();
  while (iter != end_iter)
    ensure_combo_entry(*iter++, false);

  // ѥ졼
  GtkTreeIter dummy_iter;
  gtk_list_store_insert_with_values(store, &dummy_iter, G_MAXINT32,
				    FOLDER_LIST_FOLDER_DATA, NULL,
				    FOLDER_LIST_ROW_REFERENCE, NULL,
				    -1);

  // ǺǶῨäե5ĤϿ
  iter = impl.all_folders.begin();
  end_iter = impl.all_folders.end();
  for (int i = 0; i < NUMBER_OF_RECENT_FOLDERS && iter != end_iter; i++)
    if (!ensure_combo_entry(*iter++, i == 0))
      --i;

  gtk_widget_show(dialog);
}


gboolean
AddElementDialog::collect_invalid_references(GtkTreeModel *model,
					     GtkTreePath *path,
					     GtkTreeIter *iter, gpointer data)
{
  Folder *folder = NULL;
  gtk_tree_model_get(model, iter,
		     FOLDER_LIST_FOLDER_DATA, &folder,
		     -1);

  if (folder == NULL)	// ѥ졼
    return FALSE;

  if (folder->get_folder_type() == FOLDER_TYPE_ROOT_FOLDER)
    return FALSE;

  GtkTreeRowReference *reference = NULL;
  gtk_tree_model_get(model, iter,
		     FOLDER_LIST_ROW_REFERENCE, &reference,
		     -1);

  g_return_val_if_fail(reference != NULL, FALSE);

  if (folder->get_folder_type() != FOLDER_TYPE_ROOT_FOLDER
      && !gtk_tree_row_reference_valid(reference))
    {
      GSList **list_p = static_cast<GSList **>(data);
      *list_p = g_slist_prepend(*list_p,
				gtk_tree_row_reference_new(model, path));
      gtk_tree_row_reference_free(reference);
    }

  return FALSE;
}


void
AddElementDialog::row_deleted_cb(GtkTreeModel *model, GtkTreePath *path,
				 AddElementDialog *dialog)
{
  GSList *invalidated_rows = NULL;
  gtk_tree_model_foreach(dialog->folder_list,
			 collect_invalid_references, &invalidated_rows);

  if (invalidated_rows != NULL)
    {
      GSList *list = invalidated_rows;
      GtkListStore *store = GTK_LIST_STORE(dialog->folder_list);
      while (list != NULL)
	{
	  GtkTreeRowReference *ref
	    = static_cast<GtkTreeRowReference *>(list->data);
	  TreePathHandle invalid_path(gtk_tree_row_reference_get_path(ref));
	  GtkTreeIter iter;
	  if (gtk_tree_model_get_iter(dialog->folder_list, &iter, invalid_path))
	    gtk_list_store_remove(store, &iter);

	  gtk_tree_row_reference_free(ref);
	  list = g_slist_next(list);
	}

      g_slist_free(invalidated_rows);
    }

  if (gtk_combo_box_get_active(dialog->folder_combo) == -1)
    gtk_combo_box_set_active(dialog->folder_combo, 0);

  return;
}


gboolean
AddElementDialog::set_current_folder(GtkTreeModel *model, GtkTreePath *path,
				     GtkTreeIter *iter, gpointer data)
{
  AddElementDialog *dialog = static_cast<AddElementDialog *>(data);
  Folder *folder = NULL;

  gtk_tree_model_get(model, iter,
		     FOLDER_LIST_FOLDER_DATA, &folder,
		     -1);

  if (folder != dialog->current_folder)
    return FALSE;

  gtk_combo_box_set_active_iter(dialog->folder_combo, iter);

  return TRUE;
}


struct ExpandToFolderArgs
{
  Folder *folder;
  GtkTreeView *view;
  bool scroll;
  GtkTreeRowReference *reference;

  ExpandToFolderArgs(Folder *folder, GtkTreeView *view, bool scroll);
  ~ExpandToFolderArgs();
};


ExpandToFolderArgs::ExpandToFolderArgs(Folder *folder, GtkTreeView *view,
				       bool scroll)
  : folder(folder), view(view), scroll(scroll), reference(NULL)
{
}


ExpandToFolderArgs::~ExpandToFolderArgs()
{
  if (reference != NULL)
    gtk_tree_row_reference_free(reference);
}


static gboolean
expand_to_folder(GtkTreeModel *model, GtkTreePath *path,
		 GtkTreeIter *iter, gpointer data)
{
  ExpandToFolderArgs *args = static_cast<ExpandToFolderArgs *>(data);

  TreeElement *element = NULL;
  gtk_tree_model_get(model, iter, TREE_ELEMENT_DATA, &element, -1);
  g_return_val_if_fail(element != NULL, FALSE);
  g_return_val_if_fail(element->is_folder(), FALSE);

  if (element != args->folder)
    return FALSE;

  assert(args->reference == NULL);
  args->reference = gtk_tree_row_reference_new(model, path);

  gtk_tree_view_expand_to_path(args->view, path);

  if (args->scroll)
    gtk_tree_view_scroll_to_cell(args->view, path, NULL, TRUE, 0.5, 0.0);

  return TRUE;
}


bool
AddElementDialog::ensure_combo_entry(Folder *folder, bool select_entry)
{
  g_return_val_if_fail(folder != NULL, false);

  vector<FolderHandle>::iterator iter
    = find(folder_combo_folders.begin(), folder_combo_folders.end(), folder);

  if (iter == folder_combo_folders.end())
    {
      GtkTreeModel *model = gtk_tree_view_get_model(folder_tree_view);
      ExpandToFolderArgs args(folder, folder_tree_view, select_entry);
      gtk_tree_model_foreach(model, expand_to_folder, &args);
      if (folder->get_folder_type() != FOLDER_TYPE_ROOT_FOLDER
	  && !gtk_tree_row_reference_valid(args.reference))
	return false;

      GtkTreeIter tree_iter;
      folder_combo_folders.push_back(folder);
      gtk_list_store_insert_with_values(GTK_LIST_STORE(folder_list), &tree_iter,
					G_MAXINT32,
					FOLDER_LIST_FOLDER_DATA, folder,
					FOLDER_LIST_ROW_REFERENCE, args.reference,
					-1);
      args.reference = NULL;

      if (select_entry)
	gtk_combo_box_set_active_iter(folder_combo, &tree_iter);

      return true;
    }

  if (!select_entry || current_folder == folder)
    return true;	// nothing to be done

  GtkTreeModel *model = gtk_tree_view_get_model(folder_tree_view);
  ExpandToFolderArgs args(folder, folder_tree_view, true);
  gtk_tree_model_foreach(model, expand_to_folder, &args);
  if (folder->get_folder_type() != FOLDER_TYPE_ROOT_FOLDER
      && !gtk_tree_row_reference_valid(args.reference))
    return false;

  current_folder = folder;
  gtk_tree_model_foreach(folder_list, set_current_folder, this);
  return true;
}


static gboolean
free_row_reference(GtkTreeModel *model, GtkTreePath *path,
		   GtkTreeIter *iter, gpointer data)
{
  GtkTreeRowReference *reference = NULL;
  gtk_tree_model_get(model, iter,
		     FOLDER_LIST_ROW_REFERENCE, &reference,
		     -1);
  if (reference != NULL)
    gtk_tree_row_reference_free(reference);

  return FALSE;
}


AddElementDialog::~AddElementDialog()
{
  GtkTreeModel *folder_model = gtk_tree_view_get_model(folder_tree_view);
  g_signal_handler_disconnect(folder_model, row_deleted_id);
  gtk_tree_model_foreach(folder_list, free_row_reference, NULL);
  g_object_unref(folder_list);
}


gboolean
AddElementDialog::select_current_folder(GtkTreeModel *model, GtkTreePath *path,
					GtkTreeIter *iter, gpointer data)
{
  AddElementDialog *dialog = static_cast<AddElementDialog *>(data);

  TreeElement *element = NULL;
  gtk_tree_model_get(model, iter, TREE_ELEMENT_DATA, &element, -1);

  if (element == NULL)	// ѥ졼
    return FALSE;

  g_return_val_if_fail(element->is_folder(), FALSE);

  Folder *folder = dynamic_cast<Folder *>(element);
  assert(folder != NULL);

  if (folder != dialog->current_folder)
    return FALSE;

  gtk_tree_view_expand_to_path(dialog->folder_tree_view, path);
  gtk_tree_view_set_cursor(dialog->folder_tree_view, path, NULL, FALSE);
  gtk_tree_view_scroll_to_cell(dialog->folder_tree_view, path, NULL,
			       TRUE, 0.5, 0.0);
  gtk_tree_selection_select_iter(dialog->selection, iter);

  return TRUE;
}


void
AddElementDialog::header_clicked_cb(GtkTreeViewColumn *column,
				    AddElementDialog *dialog)
{
  if (dialog->current_folder == dialog->impl.get_root_folder())
    return;

  dialog->ensure_combo_entry(dialog->impl.get_root_folder(), true);
  gtk_tree_selection_unselect_all(dialog->selection);
}


void
AddElementDialog::combo_changed_cb(GtkComboBox *combo, AddElementDialog *dialog)
{
  GtkTreeIter iter;
  if (gtk_combo_box_get_active_iter(dialog->folder_combo, &iter))
    {
      Folder *folder = NULL;
      gtk_tree_model_get(dialog->folder_list, &iter, 0, &folder, -1);
      if (folder != NULL && dialog->current_folder != folder)
	{
	  dialog->current_folder = folder;

	  gtk_tree_selection_unselect_all(dialog->selection);
	  if (folder != dialog->impl.get_root_folder())
	    {
	      GtkTreeView *view
		= gtk_tree_selection_get_tree_view(dialog->selection);
	      GtkTreeModel *model = gtk_tree_view_get_model(view);

	      gtk_tree_model_foreach(model, select_current_folder, dialog);
	      return;
	    }
	}
    }
}


void
AddElementDialog::selection_changed_cb(GtkTreeSelection *selection,
				       AddElementDialog *dialog)
{
  GtkTreeModel *model = NULL;
  GtkTreeIter iter;
  if (gtk_tree_selection_get_selected(selection, &model, &iter))
    {
      TreeElement *element = NULL;
      gtk_tree_model_get(model, &iter, TREE_ELEMENT_DATA, &element, -1);
      g_return_if_fail(element != NULL);
      g_return_if_fail(element->is_folder());

      Folder *folder = dynamic_cast<Folder *>(element);
      assert(folder != NULL);

      if (dialog->current_folder == folder)
	return;

      dialog->ensure_combo_entry(folder, true);
    }
}


void
AddElementDialog::response_cb(GtkWidget *widget, int response_id,
			      AddElementDialog *dialog)
{
  if (response_id == RESPONSE_NEW_FOLDER)
    {
      GtkTreeIter iter;
      if (gtk_combo_box_get_active_iter(dialog->folder_combo, &iter))
	{
	  Folder *parent = NULL;
	  gtk_tree_model_get(dialog->folder_list, &iter, 0, &parent, -1);

	  if (parent != NULL)
	    {
	      Folder *folder = dialog->impl.new_folder(parent, NULL, NULL);
	      if (folder != NULL)
		dialog->ensure_combo_entry(folder, true);
	    }
	}
      return;
    }

  if (response_id == GTK_RESPONSE_OK)
    {
      dialog->element->set_name(gtk_entry_get_text(dialog->name_entry));
      dialog->impl.new_element(dialog->current_folder, dialog->element);
    }

  gtk_widget_hide(widget);
  gtk_widget_unrealize(widget);
  gtk_widget_destroy(widget);
}


void
AddElementDialog::dialog_destroyed_cb(GtkWidget *widget,
				      AddElementDialog *dialog)
{
  delete dialog;
}


void
TreeUIImpl::add_element(TreeElement *element, const gchar *dialog_title)
{
  new AddElementDialog(*this, element, dialog_title);
}


void
TreeUIImpl::remove_element(Folder *folder, TreeElement *element)
{
  ElementIterFinder finder(GTK_TREE_MODEL(master_store), element);
  if (finder.element_iter == NULL)
    return;	// elementmaster_store¸ߤʤ

  Folder *fldr = NULL;
  TreeElement *elm = NULL;
  if (locate_element(GTK_TREE_MODEL(master_store), finder.element_iter,
		     &fldr, &elm))
    {
      g_return_if_fail(folder == fldr);
      g_return_if_fail(elm == element);

      gtk_tree_store_remove(master_store, finder.element_iter);
      folder->remove(element);
      notify_folder_modified(folder);
    }
}


static void
emit_row_changed(GtkTreeModel *model, GtkTreeIter *iter)
{
  TreePathHandle path(gtk_tree_model_get_path(model, iter));
  gtk_tree_model_row_changed(model, path, iter);
}


void
TreeUIImpl::notify_folder_modified(Folder *folder)
{
  if (folder == root_folder)
    return;

  if (ignore_notify_folder_modified)
    return;

  ElementIterFinder finder(GTK_TREE_MODEL(master_store), folder);

  if (finder.element_iter != NULL && finder.element_path != NULL)
    {
      GtkTreeModel *model = GTK_TREE_MODEL(master_store);
      gint n_children = gtk_tree_model_iter_n_children(model,
						       finder.element_iter);
      g_return_if_fail(n_children >= 3);

      GtkTreeIter child_iter;
      if (gtk_tree_model_iter_nth_child(model, &child_iter,
					finder.element_iter, 0))
	emit_row_changed(model, &child_iter);

      if (gtk_tree_model_iter_nth_child(model, &child_iter,
					finder.element_iter, --n_children))
	emit_row_changed(model, &child_iter);

      if (gtk_tree_model_iter_nth_child(model, &child_iter,
					finder.element_iter, --n_children))
	emit_row_changed(model, &child_iter);
    }
}


void
TreeUIImpl::notify_element_updated(TreeElement *element)
{
  ElementIterFinder finder(GTK_TREE_MODEL(master_store), element);

  if (finder.element_iter != NULL && finder.element_path != NULL)
    gtk_tree_model_row_changed(GTK_TREE_MODEL(master_store),
			       finder.element_path, finder.element_iter);
}


// siblingNULLʾ硢siblingƱ٥siblingľ˿ե
// 
// siblingNULLʾ硢parentλҥ٥˿ե
Folder *
TreeUIImpl::new_folder(Folder *parent, TreeElement *sibling,
		       GtkTreeSelection *selection)
{
  g_return_val_if_fail(parent != NULL, NULL);

  if (sibling != NULL)
    {
      // siblingNULLʾ硢siblingparentλҥ٥
      // ǤʤФʤʤ
      TreeElementList::const_iterator iter = parent->elements_begin();
      TreeElementList::const_iterator end_iter = parent->elements_end();

      while (iter != end_iter)
	if (*iter != sibling)
	  ++iter;
	else
	  break;

      g_return_val_if_fail(iter != end_iter, NULL);
    }

  ParentIterFinder finder(GTK_TREE_MODEL(master_store), parent, sibling);

  g_return_val_if_fail(finder.parent_iter != NULL || parent == root_folder,
		       NULL);

  gint position;
  if (finder.sibling_path != NULL)
    {
      gint *indices = gtk_tree_path_get_indices(finder.sibling_path);
      g_return_val_if_fail(indices != NULL, NULL);
      position = indices[gtk_tree_path_get_depth(finder.sibling_path) - 1];
    }
  else
    {
      if (finder.parent_iter != NULL)
	{
	  position
	    = gtk_tree_model_iter_n_children(GTK_TREE_MODEL(master_store),
					     finder.parent_iter);
	  position -= 2;
	}
      else
	position = G_MAXINT32;
    }

  Folder *folder = get_root_folder()->folder_new();
  folder->set_name(_("New Folder"));
  if (sibling != NULL)
    parent->prepend(folder, sibling);
  else
    parent->append(folder);

  GtkTreeIter new_iter;
#if GTK_CHECK_VERSION(2,10,0)
  gtk_tree_store_insert_with_values(master_store, &new_iter,
				    finder.parent_iter, position,
				    TREE_ELEMENT_DATA, folder,
				    TREE_ROW_TYPE, TREE_ROW_TYPE_ELEMENT,
				    -1);
#else
  gtk_tree_store_insert(master_store,
			&new_iter, finder.parent_iter, position);
  gtk_tree_store_set(master_store, &new_iter,
		     TREE_ELEMENT_DATA, folder,
		     TREE_ROW_TYPE, TREE_ROW_TYPE_ELEMENT,
		     -1);
#endif

  if (selection != NULL)
    {
      gtk_tree_selection_unselect_all(selection);
      GtkTreeView *sel_view = gtk_tree_selection_get_tree_view(selection);
      GtkTreeModel *sel_model = gtk_tree_view_get_model(sel_view);
      GtkTreeModelFilter *filter = GTK_TREE_MODEL_FILTER(sel_model);
      GtkTreeIter filter_iter;
      gtk_tree_model_filter_convert_child_iter_to_iter(filter,
						       &filter_iter,
						       &new_iter);
      gtk_tree_selection_select_iter(selection, &filter_iter);
    }

  TreePathHandle path(gtk_tree_model_get_path(GTK_TREE_MODEL(master_store),
					      &new_iter));
  GtkTreeRowReference *reference
    = gtk_tree_row_reference_new(GTK_TREE_MODEL(master_store), path);


#if GTK_CHECK_VERSION(2,10,0)
  gtk_tree_store_insert_with_values(
			master_store, NULL,
			&new_iter, G_MAXINT32,
			TREE_ELEMENT_DATA, folder,
			TREE_ROW_TYPE, TREE_ROW_TYPE_EMPTY_FOLDER_INDICATOR,
			-1);
  gtk_tree_store_insert_with_values(
			master_store, NULL,
			&new_iter, G_MAXINT32,
			TREE_ELEMENT_DATA, folder,
			TREE_ROW_TYPE, TREE_ROW_TYPE_INTERNAL_SEPARATOR,
			-1);
  gtk_tree_store_insert_with_values(
			master_store, NULL,
			&new_iter, G_MAXINT32,
			TREE_ELEMENT_DATA, folder,
			TREE_ROW_TYPE, TREE_ROW_TYPE_OPEN_IN_TAB_MENU,
			-1);
#else
  GtkTreeIter child_iter;
  gtk_tree_store_append(master_store, &child_iter, &new_iter);
  gtk_tree_store_set(master_store, &child_iter,
		     TREE_ELEMENT_DATA, folder,
		     TREE_ROW_TYPE, TREE_ROW_TYPE_EMPTY_FOLDER_INDICATOR,
		     -1);
  gtk_tree_store_append(master_store, &child_iter, &new_iter);
  gtk_tree_store_set(master_store, &child_iter,
		     TREE_ELEMENT_DATA, folder,
		     TREE_ROW_TYPE, TREE_ROW_TYPE_INTERNAL_SEPARATOR,
		     -1);
  gtk_tree_store_append(master_store, &child_iter, &new_iter);
  gtk_tree_store_set(master_store, &child_iter,
		     TREE_ELEMENT_DATA, folder,
		     TREE_ROW_TYPE, TREE_ROW_TYPE_OPEN_IN_TAB_MENU,
		     -1);
#endif

  // Ūrow_changedäƤ
  gtk_tree_model_row_changed(GTK_TREE_MODEL(master_store), path, &new_iter);
  gtk_tree_model_filter_refilter(GTK_TREE_MODEL_FILTER(menu_model));

  open_properties(NULL, folder, reference);

  return folder;
}


void
TreeUIImpl::remove(GtkTreeRowReference *row_reference)
{
  if (!gtk_tree_row_reference_valid(row_reference))
    return;

  GtkTreeModel *model = GTK_TREE_MODEL(master_store);
#if GTK_CHECK_VERSION(2,8,0)
  g_return_if_fail(model == gtk_tree_row_reference_get_model(row_reference));
#endif
  TreePathHandle path(gtk_tree_row_reference_get_path(row_reference));
  GtkTreeIter iter;
  if (gtk_tree_model_get_iter(model, &iter, path))
    {
      Folder *folder = NULL;
      TreeElement *element = NULL;
      if (locate_element(model, &iter, &folder, &element))
	{
	  gtk_tree_store_remove(master_store, &iter);
	  folder->remove(element);
	}
    }
}


void
TreeUIImpl::emit_row_changed_from_filter(GtkTreeModel *model, GtkTreeIter *iter)
{
  g_return_if_fail(GTK_IS_TREE_MODEL_FILTER(model));
  GtkTreeModelFilter *filter = GTK_TREE_MODEL_FILTER(model);
  GtkTreeModel *master_model = gtk_tree_model_filter_get_model(filter);
  g_return_if_fail(master_model == GTK_TREE_MODEL(master_store));

  GtkTreeIter master_iter;
  gtk_tree_model_filter_convert_iter_to_child_iter(filter, &master_iter, iter);
  TreePathHandle master_path(gtk_tree_model_get_path(master_model,
						     &master_iter));
  gtk_tree_model_row_changed(master_model, master_path, &master_iter);
  gtk_tree_model_filter_refilter(GTK_TREE_MODEL_FILTER(menu_model));
}


class TreeItemPropertiesDialog
{
private:
  TreeUIImpl &impl;
  TreeElementHandle element;
  GtkWidget *dialog;

  GtkEntry *name_entry;
  GtkTextBuffer *description_buffer;

  GtkTreeRowReference *row_reference;


  static void response_cb(GtkWidget *widget, int response_id,
			  TreeItemPropertiesDialog *dialog);
  static void destroy_cb(GtkWidget *widget, TreeItemPropertiesDialog *dialog);

public:
  TreeItemPropertiesDialog(TreeUIImpl &impl, const gchar *title,
			   TreeElement *element,
			   GtkTreeRowReference *row_reference);
  ~TreeItemPropertiesDialog();

  inline TreeElement *get_element() { return element; }

  void show();
};


TreeItemPropertiesDialog::TreeItemPropertiesDialog(TreeUIImpl &impl,
						   const char *title,
						   TreeElement *element,
						   GtkTreeRowReference *row_ref)
  : impl(impl),
    element(element),
    name_entry(NULL),
    description_buffer(NULL),
    row_reference(row_ref)
{
  char *tmp_title;
  if (title != NULL)
    tmp_title = NULL;
  else
    {
      const char *name = element->get_name();
      if (name == NULL)
	name = _("Anonymous Item");
      tmp_title = g_strdup_printf(_("Properties for \"%s\""), name);
      title = tmp_title;
    }

  dialog = gtk_dialog_new_with_buttons(title,
				       impl.get_controller().get_toplevel(),
				       GTK_DIALOG_DESTROY_WITH_PARENT,
				       GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
				       GTK_STOCK_OK, GTK_RESPONSE_OK,
				       NULL);
  if (tmp_title != NULL)
    g_free(tmp_title);

  gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_OK);
  g_signal_connect(dialog, "response",
		   reinterpret_cast<GCallback>(response_cb), this);
  g_signal_connect(dialog, "destroy",
		   reinterpret_cast<GCallback>(destroy_cb), this);

  if (!impl.construct_properties_contents(GTK_DIALOG(dialog), element))
    {
      GtkBox *outer_vbox = GTK_BOX(GTK_DIALOG(dialog)->vbox);
      GtkWidget *table = gtk_table_new(2, 2, FALSE);
      gtk_widget_show(table);
      gtk_box_pack_start(outer_vbox, table, FALSE, FALSE, 10);


      GtkWidget *label = gtk_label_new_with_mnemonic(_("_Name:"));
      gtk_misc_set_alignment(GTK_MISC(label), 1.0, 0.5);
      gtk_widget_show(label);
      TABLE_ATTACH_CAPTION(table, label, 0, 1, 0, 1);

      GtkWidget *entry
	= EntryFocusNotifier::gtk_entry_new(impl.get_controller());
      name_entry = GTK_ENTRY(entry);
      gtk_entry_set_width_chars(name_entry, 36);

      const char *name = element->get_name();
      gtk_entry_set_text(name_entry, name != NULL ? name : _("Anonymous Item"));
      gtk_entry_set_activates_default(name_entry, TRUE);
      gtk_widget_show(entry);
      gtk_table_attach(GTK_TABLE(table), entry, 1, 2, 0, 1,
		       static_cast<GtkAttachOptions>(GTK_EXPAND | GTK_FILL),
		       static_cast<GtkAttachOptions>(GTK_EXPAND | GTK_FILL),
		       3, 3);
      gtk_label_set_mnemonic_widget(GTK_LABEL(label), entry);


      label = gtk_label_new_with_mnemonic(_("_Description:"));
      gtk_misc_set_alignment(GTK_MISC(label), 1.0, 0.0);
      gtk_widget_show(label);
      GtkWidget *vbox = gtk_vbox_new(FALSE, 0);
      gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 3);

      gtk_widget_show(vbox);
      TABLE_ATTACH_CAPTION(table, vbox, 0, 1, 1, 2);

      description_buffer = gtk_text_buffer_new(NULL);
      const char *desc = element->get_description();
      if (desc != NULL)
	gtk_text_buffer_insert_at_cursor(description_buffer, desc, -1);

      GtkWidget *win = gtk_scrolled_window_new(NULL, NULL);
      gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(win),
				     GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
      GtkWidget *view
	= TextViewFocusNotifier::gtk_text_view_new_with_buffer(impl.get_controller(),
							       description_buffer);
      gtk_widget_show(view);
      gtk_container_add(GTK_CONTAINER(win), view);
      gtk_label_set_mnemonic_widget(GTK_LABEL(label), view);

      gtk_widget_show(win);
      gtk_table_attach(GTK_TABLE(table), win, 1, 2, 1, 2,
		       static_cast<GtkAttachOptions>(GTK_EXPAND | GTK_FILL),
		       static_cast<GtkAttachOptions>(GTK_EXPAND | GTK_FILL),
		       3, 3);
    }
}


TreeItemPropertiesDialog::~TreeItemPropertiesDialog()
{
  impl.notify_dialog_destroyed(this);

  if (description_buffer != NULL)
    g_object_unref(description_buffer);

  if (row_reference != NULL)
    gtk_tree_row_reference_free(row_reference);
}


void
TreeItemPropertiesDialog::response_cb(GtkWidget *widget, int response_id,
				      TreeItemPropertiesDialog *dialog)
{
  dialog->impl.process_response(GTK_DIALOG(widget), response_id,
				dialog->get_element());

  if (dialog->name_entry != NULL)
    {
      if (response_id == GTK_RESPONSE_OK)
	{
	  TreeElement *element = dialog->get_element();
	  const char *name = gtk_entry_get_text(dialog->name_entry);
	  element->set_name(name);

	  GtkTextIter iter;
	  GtkTextIter end_iter;

	  gtk_text_buffer_get_start_iter(dialog->description_buffer, &iter);
	  gtk_text_buffer_get_end_iter(dialog->description_buffer, &end_iter);
	  gchar *desc = gtk_text_buffer_get_text(dialog->description_buffer,
						 &iter, &end_iter, FALSE);
	  element->set_description(desc);

	  g_free(desc);
	}
    }

  if (response_id == GTK_RESPONSE_OK)
    dialog->impl.notify_element_updated(dialog->get_element());

  if (dialog->row_reference != NULL)
    {
      TreeElement *element = dialog->get_element();
      if (response_id == GTK_RESPONSE_OK)
	{
	  if (element->is_folder())
	    dialog->impl.register_folder(dynamic_cast<Folder *>(element));
	}
      else
	dialog->impl.remove(dialog->row_reference);
    }

  gtk_widget_hide(widget);
  gtk_widget_unrealize(widget);
  gtk_widget_destroy(widget);
}


void
TreeItemPropertiesDialog::destroy_cb(GtkWidget *widget,
				     TreeItemPropertiesDialog *dialog)
{
  delete dialog;
}


void
TreeItemPropertiesDialog::show()
{
  gtk_widget_show(dialog);
  gtk_window_present(GTK_WINDOW(dialog));
}


void
TreeUIImpl::notify_dialog_destroyed(TreeItemPropertiesDialog *dialog)
{
  dialogs.erase(find(dialogs.begin(), dialogs.end(), dialog));
}


TreeItemPropertiesDialog *
TreeUIImpl::lookup_dialog(const gchar *title, TreeElement *element,
			  GtkTreeRowReference *row_reference)
{
  list<TreeItemPropertiesDialog *>::iterator iter = dialogs.begin();
  list<TreeItemPropertiesDialog *>::iterator end_iter = dialogs.end();

  while (iter != end_iter)
    {
      TreeItemPropertiesDialog *dialog = *iter++;
      if (dialog->get_element() == element)
	{
	  assert(row_reference == NULL);
	  return dialog;
	}
    }

  TreeItemPropertiesDialog *dialog
    = new TreeItemPropertiesDialog(*this, title, element, row_reference);
  if (dialog != NULL)
    dialogs.push_front(dialog);

  return dialog;
}


void
TreeUIImpl::open_properties(const gchar *title, TreeElement *element,
			    GtkTreeRowReference *row_reference)
{
  TreeItemPropertiesDialog *dialog
    = lookup_dialog(title, element, row_reference);
  if (dialog != NULL)
    dialog->show();
}


void
TreeUIImpl::open_properties(const gchar *title, TreeElement *element)
{
  open_properties(title, element, NULL);
}


void
TreeUIImpl::open_properties(TreeElement *element)
{
  open_properties(NULL, element, NULL);
}


void
TreeUIImpl::notify_selection_changed(GtkTreeSelection *selection,
				     bool has_folder, bool has_item)
{
}


void
TreeUIImpl::popup_context_menu(TreeView *view, GtkTreeSelection *selection,
			       guint button, guint32 event_time)
{
  last_view = view;
  last_selection = selection;

  GtkWidget *menu_widget
    = gtk_ui_manager_get_widget(controller.get_ui_manager(),
				get_popup_menu_root());
  g_return_if_fail(GTK_IS_MENU(menu_widget));

  gtk_menu_popup(GTK_MENU(menu_widget), NULL, NULL, NULL, NULL,
		 button, event_time);
}


void
TreeUIImpl::popup_folder_tree_context_menu(TreeView *view,
					   GtkTreeSelection *selection,
					   guint button, guint32 event_time)
{
  last_view = view;
  last_selection = selection;

  GtkWidget *menu_widget
    = gtk_ui_manager_get_widget(controller.get_ui_manager(),
				get_folder_tree_popup_menu_root());
  g_return_if_fail(GTK_IS_MENU(menu_widget));

  gtk_menu_popup(GTK_MENU(menu_widget), NULL, NULL, NULL, NULL,
		 button, event_time);
}


void
TreeUIImpl::tree_updated_cb(RootFolder *root)
{
  // sigslotʥϥɥ
#ifndef NDEBUG
  RootFolder *prev_root = root_folder;
  assert(prev_root == root);
#endif


  GDK_THREADS_ENTER();

  reconstruct();

  GDK_THREADS_LEAVE();
}


struct AppendElementFunc
{
  TreeUIImpl &impl;
  Folder *parent_folder;
  GtkTreeIter *parent_folder_iter;
  GtkTreeStore *store;
  bool show_all;

  AppendElementFunc(TreeUIImpl &impl, bool show_all)
    : impl(impl),
      parent_folder(NULL), parent_folder_iter(NULL),
      store(impl.master_store), show_all(show_all)
  {
  }
  AppendElementFunc(TreeUIImpl &impl, Folder *parent_folder,
		    GtkTreeIter *parent_folder_iter, bool show_all)
    : impl(impl),
      parent_folder(parent_folder), parent_folder_iter(parent_folder_iter),
      store(impl.master_store), show_all(show_all)
  {
  }

  inline void append_leaf(TreeElement *leaf);
  inline void append_folder(Folder *folder);
  inline void operator()(TreeElement *element);
};


void
AppendElementFunc::append_leaf(TreeElement *leaf)
{
  g_return_if_fail(leaf != NULL);

  if (!show_all && leaf->is_hidden())
    return;

#if GTK_CHECK_VERSION(2,10,0)
  gtk_tree_store_insert_with_values(store, NULL,
				    parent_folder_iter, G_MAXINT32,
				    TREE_ELEMENT_DATA, leaf,
				    TREE_ROW_TYPE, TREE_ROW_TYPE_ELEMENT,
				    -1);
#else
  GtkTreeIter new_iter;
  gtk_tree_store_append(store, &new_iter, parent_folder_iter);
  gtk_tree_store_set(store, &new_iter,
		     TREE_ELEMENT_DATA, leaf,
		     TREE_ROW_TYPE, TREE_ROW_TYPE_ELEMENT,
		     -1);
#endif
}


void
AppendElementFunc::append_folder(Folder *folder)
{
  g_return_if_fail(folder != NULL);

  if (!show_all && folder->is_hidden())
    return;

  impl.all_folders.push_front(folder);

  GtkTreeIter iter;
#if GTK_CHECK_VERSION(2,10,0)
  gtk_tree_store_insert_with_values(store, &iter,
				    parent_folder_iter, G_MAXINT32,
				    TREE_ELEMENT_DATA, folder,
				    TREE_ROW_TYPE, TREE_ROW_TYPE_ELEMENT,
				    -1);
#else
  gtk_tree_store_append(store, &iter, parent_folder_iter);
  gtk_tree_store_set(store, &iter,
		     TREE_ELEMENT_DATA, folder,
		     TREE_ROW_TYPE, TREE_ROW_TYPE_ELEMENT,
		     -1);
#endif

  GtkTreeModel *model = GTK_TREE_MODEL(store);
  TreePathHandle path(gtk_tree_model_get_path(model, &iter));

  parent_folder = folder;
  parent_folder_iter = &iter;

#if GTK_CHECK_VERSION(2,10,0)
  gtk_tree_store_insert_with_values(
			store, NULL, &iter, G_MAXINT32,
			TREE_ELEMENT_DATA, folder,
			TREE_ROW_TYPE, TREE_ROW_TYPE_EMPTY_FOLDER_INDICATOR,
			-1);
#else
  GtkTreeIter empty_iter;
  gtk_tree_store_append(store, &empty_iter, &iter);
  gtk_tree_store_set(store, &empty_iter,
		     TREE_ELEMENT_DATA, folder,
		     TREE_ROW_TYPE, TREE_ROW_TYPE_EMPTY_FOLDER_INDICATOR,
		     -1);
#endif

  for_each(folder->elements_begin(), folder->elements_end(), *this);

#if GTK_CHECK_VERSION(2,10,0)
  gtk_tree_store_insert_with_values(
				store, NULL, &iter, G_MAXINT32,
				TREE_ELEMENT_DATA, folder,
				TREE_ROW_TYPE, TREE_ROW_TYPE_INTERNAL_SEPARATOR,
				-1);
  gtk_tree_store_insert_with_values(
				store, NULL, &iter, G_MAXINT32,
				TREE_ELEMENT_DATA, folder,
				TREE_ROW_TYPE, TREE_ROW_TYPE_OPEN_IN_TAB_MENU,
				-1);
#else
  GtkTreeIter trailer_iter;
  gtk_tree_store_append(store, &trailer_iter, &iter);
  gtk_tree_store_set(store, &trailer_iter,
		     TREE_ELEMENT_DATA, folder,
		     TREE_ROW_TYPE, TREE_ROW_TYPE_INTERNAL_SEPARATOR,
		     -1);
  gtk_tree_store_append(store, &trailer_iter, &iter);
  gtk_tree_store_set(store, &trailer_iter,
		     TREE_ELEMENT_DATA, folder,
		     TREE_ROW_TYPE, TREE_ROW_TYPE_OPEN_IN_TAB_MENU,
		     -1);
#endif

  // եξ϶Ūrow_changedäƤ
  gtk_tree_model_row_changed(model, path, &iter);
}


void
AppendElementFunc::operator()(TreeElement *element)
{
  if (element->is_folder())
    {
      Folder *prev_parent_folder = parent_folder;
      GtkTreeIter *prev_parent_folder_iter = parent_folder_iter;

      append_folder(dynamic_cast<Folder *>(element));

      parent_folder = prev_parent_folder;
      parent_folder_iter = prev_parent_folder_iter;
    }
  else
    append_leaf(element);
}


void
TreeUIImpl::new_element(Folder *folder, TreeElement *element)
{
  g_return_if_fail(folder != NULL);
  g_return_if_fail(element != NULL);

  FolderIterFinder finder(GTK_TREE_MODEL(master_store), folder);

  if (finder.folder_iter == NULL && folder != root_folder)
    {
      // folderɽ֡
      folder->append(element);
      return;
    }

  gint position;
  if (finder.folder_iter != NULL)
    {
      position = gtk_tree_model_iter_n_children(GTK_TREE_MODEL(master_store),
						finder.folder_iter);
      position -= 2;
    }
  else
    position = G_MAXINT32;

  folder->append(element);

  GtkTreeIter new_iter;
#if GTK_CHECK_VERSION(2,10,0)
  gtk_tree_store_insert_with_values(master_store, &new_iter,
				    finder.folder_iter, position,
				    TREE_ELEMENT_DATA, element,
				    TREE_ROW_TYPE, TREE_ROW_TYPE_ELEMENT,
				    -1);
#else
  gtk_tree_store_insert(master_store, &new_iter, finder.folder_iter, position);
  gtk_tree_store_set(master_store, &new_iter,
		     TREE_ELEMENT_DATA, element,
		     TREE_ROW_TYPE, TREE_ROW_TYPE_ELEMENT,
		     -1);
#endif

  notify_folder_modified(folder);
  register_folder(folder);

  if (!element->is_folder())
    return;

  Folder *src_folder = dynamic_cast<Folder *>(element);
  g_return_if_fail(src_folder != NULL);


  if (menu_shell != NULL)
    ochusha_tree_menu_helper_detach_tree_menu(menu_helper, menu_shell);

#if GTK_CHECK_VERSION(2,10,0)
  gtk_tree_store_insert_with_values(
			master_store, NULL, &new_iter, G_MAXINT32,
			TREE_ELEMENT_DATA, src_folder,
			TREE_ROW_TYPE, TREE_ROW_TYPE_EMPTY_FOLDER_INDICATOR,
			-1);
#else
  GtkTreeIter empty_iter;
  gtk_tree_store_append(master_store, &empty_iter, &new_iter);
  gtk_tree_store_set(master_store, &empty_iter,
		     TREE_ELEMENT_DATA, src_folder,
		     TREE_ROW_TYPE, TREE_ROW_TYPE_EMPTY_FOLDER_INDICATOR,
		     -1);
#endif

  for_each(src_folder->elements_begin(), src_folder->elements_end(),
	   AppendElementFunc(*this, src_folder, &new_iter, show_all()));

#if GTK_CHECK_VERSION(2,10,0)
  gtk_tree_store_insert_with_values(
			master_store, NULL, &new_iter, G_MAXINT32,
			TREE_ELEMENT_DATA, src_folder,
			TREE_ROW_TYPE, TREE_ROW_TYPE_INTERNAL_SEPARATOR,
			-1);
  gtk_tree_store_insert_with_values(
			master_store, NULL, &new_iter, G_MAXINT32,
			TREE_ELEMENT_DATA, src_folder,
			TREE_ROW_TYPE, TREE_ROW_TYPE_OPEN_IN_TAB_MENU,
			-1);
#else
  GtkTreeIter trailer_iter;
  gtk_tree_store_append(master_store, &trailer_iter, &new_iter);
  gtk_tree_store_set(master_store, &trailer_iter,
		     TREE_ELEMENT_DATA, src_folder,
		     TREE_ROW_TYPE, TREE_ROW_TYPE_INTERNAL_SEPARATOR,
		     -1);
  gtk_tree_store_append(master_store, &trailer_iter, &new_iter);
  gtk_tree_store_set(master_store, &trailer_iter,
		     TREE_ELEMENT_DATA, src_folder,
		     TREE_ROW_TYPE, TREE_ROW_TYPE_OPEN_IN_TAB_MENU,
		     -1);
#endif

  gtk_tree_model_filter_refilter(GTK_TREE_MODEL_FILTER(menu_model));

  if (menu_shell != NULL)
    ochusha_tree_menu_helper_attach_tree_menu(menu_helper, menu_shell);
}


/*
 * src_folder
 * dest_folder/dest_element`position'ǻꤵ줿֤롣
 */
bool
TreeUIImpl::insert_folder(Folder *src_folder,
			  Folder *dest_folder, TreeElement *dest_element,
			  GtkTreeViewDropPosition position)
{
  g_return_val_if_fail(src_folder->is_alone(), false);
  g_return_val_if_fail(src_folder != root_folder, false);
  g_return_val_if_fail(src_folder->is_folder(), false);


  GtkTreeModel *model = GTK_TREE_MODEL(master_store);


  ParentIterFinder dest_finder(model, dest_folder, dest_element);
  g_return_val_if_fail(dest_finder.parent_iter != NULL
		       || dest_folder == root_folder,
		       false);
  g_return_val_if_fail(dest_finder.sibling_path != NULL, false);

  gint *indices = gtk_tree_path_get_indices(dest_finder.sibling_path);
  g_return_val_if_fail(indices != NULL, false);


  if (dest_element->is_folder()
      && (position == GTK_TREE_VIEW_DROP_INTO_OR_BEFORE
	  || position == GTK_TREE_VIEW_DROP_INTO_OR_AFTER))
    dest_folder = dynamic_cast<Folder *>(dest_element);

  if (dest_folder == dest_element)
    {
      new_element(dest_folder, src_folder);
      return true;
    }

  gint insert_pos = indices[gtk_tree_path_get_depth(dest_finder.sibling_path) - 1];
  // ͤdest_elementľؤƤ롣

  if (position == GTK_TREE_VIEW_DROP_BEFORE
      || position == GTK_TREE_VIEW_DROP_INTO_OR_BEFORE)
    {
      dest_folder->prepend(src_folder, dest_element);
    }
  else
    {
      ++insert_pos;
      dest_folder->append(src_folder, dest_element);
    }

  if (menu_shell != NULL)
    ochusha_tree_menu_helper_detach_tree_menu(menu_helper, menu_shell);

  GtkTreeIter folder_iter;
#if GTK_CHECK_VERSION(2,10,0)
  gtk_tree_store_insert_with_values(master_store, &folder_iter,
				    dest_finder.parent_iter, insert_pos,
				    TREE_ELEMENT_DATA, src_folder,
				    TREE_ROW_TYPE, TREE_ROW_TYPE_ELEMENT,
				    -1);
#else
  gtk_tree_store_insert(master_store, &folder_iter,
			dest_finder.parent_iter, insert_pos);
  gtk_tree_store_set(master_store, &folder_iter,
		     TREE_ELEMENT_DATA, src_folder,
		     TREE_ROW_TYPE, TREE_ROW_TYPE_ELEMENT,
		     -1);
#endif

#if GTK_CHECK_VERSION(2,10,0)
  gtk_tree_store_insert_with_values(
			master_store, NULL, &folder_iter, G_MAXINT32,
			TREE_ELEMENT_DATA, src_folder,
			TREE_ROW_TYPE, TREE_ROW_TYPE_EMPTY_FOLDER_INDICATOR,
			-1);
#else
  GtkTreeIter empty_iter;
  gtk_tree_store_append(master_store, &empty_iter, &folder_iter);
  gtk_tree_store_set(master_store, &empty_iter,
		     TREE_ELEMENT_DATA, src_folder,
		     TREE_ROW_TYPE, TREE_ROW_TYPE_EMPTY_FOLDER_INDICATOR,
		     -1);
#endif

  for_each(src_folder->elements_begin(), src_folder->elements_end(),
	   AppendElementFunc(*this, src_folder, &folder_iter, show_all()));

#if GTK_CHECK_VERSION(2,10,0)
  gtk_tree_store_insert_with_values(
			master_store, NULL, &folder_iter, G_MAXINT32,
			TREE_ELEMENT_DATA, src_folder,
			TREE_ROW_TYPE, TREE_ROW_TYPE_INTERNAL_SEPARATOR,
			-1);
  gtk_tree_store_insert_with_values(
			master_store, NULL, &folder_iter, G_MAXINT32,
			TREE_ELEMENT_DATA, src_folder,
			TREE_ROW_TYPE, TREE_ROW_TYPE_OPEN_IN_TAB_MENU,
			-1);
#else
  GtkTreeIter trailer_iter;
  gtk_tree_store_append(master_store, &trailer_iter, &folder_iter);
  gtk_tree_store_set(master_store, &trailer_iter,
		     TREE_ELEMENT_DATA, src_folder,
		     TREE_ROW_TYPE, TREE_ROW_TYPE_INTERNAL_SEPARATOR,
		     -1);
  gtk_tree_store_append(master_store, &trailer_iter, &folder_iter);
  gtk_tree_store_set(master_store, &trailer_iter,
		     TREE_ELEMENT_DATA, src_folder,
		     TREE_ROW_TYPE, TREE_ROW_TYPE_OPEN_IN_TAB_MENU,
		     -1);
#endif

  gtk_tree_model_filter_refilter(GTK_TREE_MODEL_FILTER(menu_model));

  if (menu_shell != NULL)
    ochusha_tree_menu_helper_attach_tree_menu(menu_helper, menu_shell);


  return true;
}


/*
 * parent_folder/src_folder
 * dest_folder/dest_element`position'ǻꤵ줿֤˰ư롣
 */
bool
TreeUIImpl::move_folder(Folder *parent_folder, Folder *src_folder,
			Folder *dest_folder, TreeElement *dest_element,
			GtkTreeViewDropPosition position)
{
  g_return_val_if_fail(parent_folder->is_folder(), false);
  g_return_val_if_fail(src_folder != root_folder, false);
  g_return_val_if_fail(src_folder->is_folder(), false);


  GtkTreeModel *model = GTK_TREE_MODEL(master_store);


  ParentIterFinder dest_finder(model, dest_folder, dest_element);
  g_return_val_if_fail(dest_finder.parent_iter != NULL
		       || dest_folder == root_folder,
		       false);
  g_return_val_if_fail(dest_finder.sibling_path != NULL, false);

  gint *indices = gtk_tree_path_get_indices(dest_finder.sibling_path);
  g_return_val_if_fail(indices != NULL, false);


  Folder *effective_dest_folder;
  if (dest_element->is_folder()
      && (position == GTK_TREE_VIEW_DROP_INTO_OR_BEFORE
	  || position == GTK_TREE_VIEW_DROP_INTO_OR_AFTER))
    effective_dest_folder = dynamic_cast<Folder *>(dest_element);
  else
    effective_dest_folder = dest_folder;

  if (!src_folder->is_alone())
    {
      FolderHandle folder = src_folder;
      FolderIterFinder src_finder(model, src_folder);
      g_return_val_if_fail(src_finder.folder_iter != NULL, false);

      // ư褬ʬλ¹ǤʤȤǧɬפ롣
      if (effective_dest_folder != root_folder)
	{
	  FolderIterFinder folder_finder(model, effective_dest_folder);
	  g_return_val_if_fail(folder_finder.folder_iter != NULL, false);

	  TreePathHandle src_path
	    = gtk_tree_model_get_path(model, src_finder.folder_iter);
	  TreePathHandle folder_path
	    = gtk_tree_model_get_path(model, folder_finder.folder_iter);
	  if (gtk_tree_path_compare(folder_path, src_path) == 0
	      || gtk_tree_path_is_descendant(folder_path, src_path))
	    {
#ifndef NDEBUG
	      fprintf(stderr, "Cannot move folder into its descendant!\n");
#endif
	      return false;
	    }
	}

      // ޤmaster_storesrc_folderrow롣
      gtk_tree_store_remove(master_store, src_finder.folder_iter);
      parent_folder->remove(src_folder);
      g_return_val_if_fail(src_folder->is_alone(), false);
    }

  return insert_folder(src_folder, dest_folder, dest_element, position);
}


/*
 * src_element
 * dest_folder/dest_element`position'ǻꤵ줿֤롣
 */
bool
TreeUIImpl::insert_element(TreeElement *src_element,
			   Folder *dest_folder, TreeElement *dest_element,
			   GtkTreeViewDropPosition position)
{
  g_return_val_if_fail(!src_element->is_folder(), false);

  ParentIterFinder dest_finder(GTK_TREE_MODEL(master_store),
			       dest_folder, dest_element);
  g_return_val_if_fail(dest_finder.parent_iter != NULL
		       || dest_folder == root_folder,
		       false);
  g_return_val_if_fail(dest_finder.sibling_path != NULL, false);

  gint *indices = gtk_tree_path_get_indices(dest_finder.sibling_path);
  g_return_val_if_fail(indices != NULL, false);


  TreeElementHandle element = src_element;
  if (dest_element->is_folder()
      && (position == GTK_TREE_VIEW_DROP_INTO_OR_BEFORE
	  || position == GTK_TREE_VIEW_DROP_INTO_OR_AFTER))
    {
      new_element(dynamic_cast<Folder *>(dest_element), element);
      return true;
    }


  gint insert_pos
    = indices[gtk_tree_path_get_depth(dest_finder.sibling_path) - 1];
  // ͤdest_elementľؤƤ롣

  if (position == GTK_TREE_VIEW_DROP_BEFORE
      || position == GTK_TREE_VIEW_DROP_INTO_OR_BEFORE)
    {
      dest_folder->prepend(element, dest_element);
    }
  else
    {
      ++insert_pos;
      dest_folder->append(element, dest_element);
    }

#if GTK_CHECK_VERSION(2,10,0)
  gtk_tree_store_insert_with_values(master_store, NULL,
				    dest_finder.parent_iter, insert_pos,
				    TREE_ELEMENT_DATA, src_element,
				    TREE_ROW_TYPE, TREE_ROW_TYPE_ELEMENT,
				    -1);
#else
  GtkTreeIter new_iter;
  gtk_tree_store_insert(master_store, &new_iter,
			dest_finder.parent_iter, insert_pos);
  gtk_tree_store_set(master_store, &new_iter,
		     TREE_ELEMENT_DATA, src_element,
		     TREE_ROW_TYPE, TREE_ROW_TYPE_ELEMENT,
		     -1);
#endif

  return true;
}


/*
 * src_folder/src_element
 * dest_folder/dest_element`position'ǻꤵ줿֤˰ư롣
 */
bool
TreeUIImpl::move_element(Folder *src_folder, TreeElement *src_element,
			 Folder *dest_folder, TreeElement *dest_element,
			 GtkTreeViewDropPosition position)
{
  if (src_element->is_folder())
    return move_folder(src_folder, dynamic_cast<Folder *>(src_element),
		       dest_folder, dest_element, position);


  ParentIterFinder dest_finder(GTK_TREE_MODEL(master_store),
			       dest_folder, dest_element);
  g_return_val_if_fail(dest_finder.parent_iter != NULL
		       || dest_folder == root_folder,
		       false);
  g_return_val_if_fail(dest_finder.sibling_path != NULL, false);

  gint *indices = gtk_tree_path_get_indices(dest_finder.sibling_path);
  g_return_val_if_fail(indices != NULL, false);


  TreeElementHandle element = src_element;

  if (!src_element->is_alone())
    {
      // src_elementϤޤĥ꡼ˤ롣
      ParentIterFinder src_finder(GTK_TREE_MODEL(master_store),
				  src_folder, src_element);
      g_return_val_if_fail(src_finder.parent_iter != NULL
			   || src_folder == root_folder,
			   false);
      g_return_val_if_fail(src_finder.sibling_path != NULL, false);

      GtkTreeIter src_iter;
      if (!gtk_tree_model_get_iter(GTK_TREE_MODEL(master_store), &src_iter,
				   src_finder.sibling_path))
	return false;


      // ޤmaster_storesrc_folder/src_elementrow롣
      gtk_tree_store_remove(master_store, &src_iter);
      src_folder->remove(src_element);
      g_return_val_if_fail(src_element->is_alone(), false);
    }

  return insert_element(element, dest_folder, dest_element, position);
}


void
TreeUIImpl::reconstruct()
{
  if (menu_shell != NULL)
    ochusha_tree_menu_helper_detach_tree_menu(menu_helper, menu_shell);

  ignore_notify_folder_modified = true;
  gtk_tree_store_clear(master_store);
  ignore_notify_folder_modified = false;

  all_folders.clear();
  all_folders.push_front(static_cast<Folder *>(root_folder));

  // İGtkTreeModelȤ
  for_each(root_folder->elements_begin(), root_folder->elements_end(),
	   AppendElementFunc(*this, show_all()));

  all_folders.sort(FolderCompare());

  if (menu_shell != NULL)
    ochusha_tree_menu_helper_attach_tree_menu(menu_helper, menu_shell);
}


void
TreeUIImpl::setup_menubar()
{
  if (menu_shell != NULL)
    return;

  GtkWidget *menu_widget = get_menu_widget();
  if (menu_widget == NULL)
    {
      fprintf(stderr, "Cannot setup menu\n");
      return;
    }

  GtkMenuShell *shell
    = GTK_MENU_SHELL(gtk_menu_item_get_submenu(GTK_MENU_ITEM(menu_widget)));
  GtkWidget *separator = gtk_separator_menu_item_new();
  gtk_widget_show(separator);
  gtk_menu_shell_append(shell, separator);
  ochusha_tree_menu_helper_attach_tree_menu(menu_helper, shell);
  menu_shell = shell;

  
}


void
TreeUIImpl::new_folder_selection()
{
  if (last_selection == NULL)
    return;

  GtkTreeModel *model = NULL;
  GList *rows = gtk_tree_selection_get_selected_rows(last_selection, &model);
  if (rows == NULL)
    return;

  GtkTreeIter sibling;
  if (gtk_tree_model_get_iter(model, &sibling,
			      static_cast<GtkTreePath *>(rows->data)))
    {
      Folder *parent = NULL;
      TreeElement *next = NULL;
      if (locate_element(model, &sibling, &parent, &next))
	new_folder(parent, next, last_selection);
    }

  g_list_foreach(rows, reinterpret_cast<GFunc>(gtk_tree_path_free), NULL);
  g_list_free(rows);

  last_view = NULL;
  last_selection = NULL;
}


enum WhatToDo {
  COPY_NAME_ELEMENTS,
  COPY_NAME_URL_ELEMENTS
};


class DoOnSelectionHelper
{
private:
  TreeUIImpl &impl;
  WhatToDo what_to_do;
  GtkTreeSelection *selection;

  string text;

  static void do_on_selection(GtkTreeModel *model, GtkTreePath *path,
			      GtkTreeIter *iter, gpointer data);


public:
  DoOnSelectionHelper(TreeUIImpl &impl, WhatToDo what_to_do,
		      GtkTreeSelection *selection);
  ~DoOnSelectionHelper();
};


void
DoOnSelectionHelper::do_on_selection(GtkTreeModel *model, GtkTreePath *path,
				     GtkTreeIter *iter, gpointer data)
{
  TreeElement *element = NULL;
  TreeRowType row_type = TREE_ROW_TYPE_UNKNOWN;
  gtk_tree_model_get(model, iter,
		     TREE_ELEMENT_DATA, &element,
		     TREE_ROW_TYPE, &row_type,
		     -1);
  g_return_if_fail(row_type == TREE_ROW_TYPE_ELEMENT);
  g_return_if_fail(element != NULL);
  g_return_if_fail(!element->is_separator());

  DoOnSelectionHelper *helper = static_cast<DoOnSelectionHelper *>(data);
  switch (helper->what_to_do)
    {
    case COPY_NAME_ELEMENTS:
      if (helper->text.size() > 0)
	helper->text.append("\n");
      helper->text.append(TreeUIImpl::canonicalize_element_name(element));
      return;


    case COPY_NAME_URL_ELEMENTS:
      if (helper->text.size() > 0)
	helper->text.append("\n");
      helper->text.append(TreeUIImpl::canonicalize_element_name(element));

      if (element->is_folder())
	return;

      helper->text.append("\n");
      helper->text.append(helper->impl.get_url(element));
      return;
    }
}


DoOnSelectionHelper::DoOnSelectionHelper(TreeUIImpl &impl, WhatToDo what_to_do,
					 GtkTreeSelection *selection)
  : impl(impl), what_to_do(what_to_do), selection(selection)
{
  if (selection == NULL)
    return;

  gtk_tree_selection_selected_foreach(selection, do_on_selection, this);
  gtk_tree_selection_unselect_all(selection);

  switch (what_to_do)
    {
    case COPY_NAME_ELEMENTS:
    case COPY_NAME_URL_ELEMENTS:
      impl.controller.copy(text.c_str());
      return;
    }
}


DoOnSelectionHelper::~DoOnSelectionHelper()
{
}


void
TreeUIImpl::toggle_hide_selection()
{
  if (last_view == NULL)
    return;

  last_view->selection_data.toggle_hide();

  last_view = NULL;
  last_selection = NULL;
}


void
TreeUIImpl::hide_selection()
{
  if (last_view == NULL)
    return;

  last_view->selection_data.hide();

  last_view = NULL;
  last_selection = NULL;
}


void
TreeUIImpl::copy_name_selection()
{
  if (last_selection == NULL)
    return;

  DoOnSelectionHelper(*this, COPY_NAME_ELEMENTS, last_selection);
  last_view = NULL;
  last_selection = NULL;
}


void
TreeUIImpl::copy_name_url_selection()
{
  if (last_selection == NULL)
    return;

  DoOnSelectionHelper(*this, COPY_NAME_URL_ELEMENTS, last_selection);

  last_view = NULL;
  last_selection = NULL;
}


void
TreeUIImpl::open_properties_selection()
{
  if (last_view == NULL)
    return;

  last_view->selection_data.open_properties();

  last_view = NULL;
  last_selection = NULL;
}


void
TreeUIImpl::cut_tree_item_selection()
{
  if (last_view == NULL)
    return;

  g_return_if_fail(last_selection != NULL);

  last_view = NULL;
  last_selection = NULL;
}


void
TreeUIImpl::copy_tree_item_selection()
{
  if (last_view == NULL)
    return;

  g_return_if_fail(last_selection != NULL);

  controller.update_clipboard(new TreeViewSelectionData(*this, last_selection,
							false));
  last_view = NULL;
  last_selection = NULL;
}


void
TreeUIImpl::paste_tree_item_selection()
{
  if (last_view == NULL)
    return;

  g_return_if_fail(last_selection != NULL);

  last_view = NULL;
  last_selection = NULL;
}


}
