/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: tree_view.cc,v 1.24 2009/01/01 07:47:07 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "tree_view.h"

#include "controller.h"
#include "preferences.h"
#include "session_client.h"
#include "ui_constants.h"

#if !GTK_CHECK_VERSION(2,14,0)
#include "libochushagtk/misc_utils.h"
#endif
#include "libochushagtk/searchbar.h"
#include "libochushagtk/stock.h"

#include <cassert>

#include <gdk/gdkkeysyms.h>


namespace Ochusha
{


//#define VERBOSE_DESTRUCTOR	1
//#define DEBUG_FOCUS_NOTIFIER	1
//#define DEBUG_DND		1
//#define DEBUG_DND_MORE	1


// TreeViewSelectionDataμ
#if GTK_CHECK_VERSION(2,10,0)
static GdkAtom list_of_elements_atom =
  gdk_atom_intern_static_string("ochusha/internal-list-of-elements");
#else
static GdkAtom list_of_elements_atom =
  gdk_atom_intern("ochusha/internal-list-of-elements", FALSE);
#endif


enum {
  TARGET_INTERNAL_LIST_OF_ELEMENTS = 0,
  TARGET_TEXT_URI_LIST,
  TARGET_UTF8_STRING
};


static const GtkTargetEntry target_internal_list_of_elements = {
  const_cast<gchar *>("ochusha/internal-list-of-elements"),
  GTK_TARGET_SAME_APP, TARGET_INTERNAL_LIST_OF_ELEMENTS
};


static const GtkTargetEntry target_text_uri_list = {
  const_cast<gchar *>("text/uri-list"), 0, TARGET_TEXT_URI_LIST
};


static const GtkTargetEntry target_utf8_string = {
  const_cast<gchar *>("UTF8_STRING"), 0, TARGET_UTF8_STRING
};


static const GtkTargetEntry target_string = {
  const_cast<gchar *>("STRING"), 0, TARGET_UTF8_STRING
};


static const GtkTargetEntry target_text = {
  const_cast<gchar *>("TEXT"), 0, TARGET_UTF8_STRING
};


static const GtkTargetEntry target_ctext = {
  const_cast<gchar *>("COMPOUND_TEXT"), 0, TARGET_UTF8_STRING
};


static const GtkTargetEntry text_plain_utf8 = {
  const_cast<gchar *>("text/plain;charset=utf-8"), 0, TARGET_UTF8_STRING
};


static const GtkTargetEntry text_plain = {
  const_cast<gchar *>("text/plain"), 0, TARGET_UTF8_STRING
};


static const GtkTargetEntry all_targets[] = {
  target_internal_list_of_elements,
  target_text_uri_list,
  target_utf8_string,
  target_string,
  target_text,
  target_ctext,
  text_plain_utf8,
  text_plain
};


static const GtkTargetEntry linkless_targets[] = {
  target_internal_list_of_elements,
  target_utf8_string,
  target_string,
  target_text,
  target_ctext,
  text_plain_utf8,
  text_plain
};


static const GtkTargetEntry empty_targets[] = {
};


TreeViewSelectionData::TreeViewSelectionData(TreeUIImpl &impl,
					     GtkTreeSelection *selection,
					     bool is_static)
  : impl(impl),
    selection(selection),
    is_static(is_static)
{
  if (is_static)
    g_signal_connect(selection, "changed",
		     reinterpret_cast<GCallback>(changed_cb), this);
  else
    changed_cb(selection, this);
}


TreeViewSelectionData::~TreeViewSelectionData()
{
}


bool
TreeViewSelectionData::deletable()
{
  return copiable();
}


bool
TreeViewSelectionData::cuttable()
{
  return copiable();
}


bool
TreeViewSelectionData::copiable()
{
  return !rows.empty();
}


bool
TreeViewSelectionData::pastable(GtkClipboard *clipboard)
{
#if 1
#warning "Revisit here!"
  // XXX: do_paste(...)μӾʤΤ
  //Ʊĥ꡼ͳΥǡΤߥڡȲǽäƤȤǤƤ
  if (TreeView::internal_list_of_elements_available(clipboard) == FALSE)
    return false;

  SelectionData *selection_data = SelectionData::get_from(clipboard);
  g_return_val_if_fail(selection_data != NULL, false);

  TreeViewSelectionData *data
    = dynamic_cast<TreeViewSelectionData *>(selection_data);
  g_return_val_if_fail(data != NULL, false);

  return &data->impl == &impl;
#else
  // ̤URLĥդεưİ֥åޡ˼餳
#if GTK_CHECK_VERSION(2,14,0)
  return gtk_clipboard_wait_is_uris_available(clipboard);
#else
  return ochusha_clipboard_wait_is_uris_available(clipboard);
#endif
#endif
}


bool
TreeViewSelectionData::do_delete()
{
  GtkTreeView *view = gtk_tree_selection_get_tree_view(selection);
  g_return_val_if_fail(view != NULL, false);

  GtkTreeModel *model = gtk_tree_view_get_model(view);
  g_return_val_if_fail(model != NULL, false);

  g_return_val_if_fail(!rows.empty(), false);

  list<GtkTreeRowReference *> rows_to_be_removed;

  vector<SelectedRow>::const_iterator iter = rows_begin();
  vector<SelectedRow>::const_iterator end_iter = rows_end();
  while (iter != end_iter)
    {
      const SelectedRow &row = *iter++;
      GtkTreeIter tree_iter;
      if (gtk_tree_model_get_iter(model, &tree_iter, row.get_path()))
	{
	  row.get_folder()->remove(row.get_element());
	  GtkTreeRowReference *reference
	    = gtk_tree_row_reference_new(model, row.get_path());
	  rows_to_be_removed.push_front(reference);
	}
      row.get_folder()->remove(row.get_element());
    }

  while (!rows_to_be_removed.empty())
    {
      GtkTreeRowReference *reference = rows_to_be_removed.front();
      rows_to_be_removed.pop_front();

      GtkTreeIter master_iter;
      if (impl.get_master_iter(model, &master_iter, reference))
	gtk_tree_store_remove(impl.master_store, &master_iter);
      gtk_tree_row_reference_free(reference);
    }

  return true;
}


bool
TreeViewSelectionData::do_paste(GtkClipboard *clipboard)
{
#warning "Revisit here!"
  // XXX: ҤȤޤƱĥ꡼ͳΥǡΤߥڡȲǽäƤȤǤ
  g_return_val_if_fail(pastable(clipboard), false);

  SelectionData *selection_data = SelectionData::get_from(clipboard);
  g_return_val_if_fail(selection_data != NULL, false);

  TreeViewSelectionData *data
    = dynamic_cast<TreeViewSelectionData *>(selection_data);
  g_return_val_if_fail(data != NULL, false);

  assert(&data->impl == &impl);

  // tree_viewpasteGtkTreeView
  GtkTreeView *tree_view = gtk_tree_selection_get_tree_view(selection);
  g_return_val_if_fail(tree_view != NULL, false);

  GtkTreePath *path = NULL;
  gtk_tree_view_get_cursor(tree_view, &path, NULL);
  TreePathHandle cursor_path = path;
  Folder *dest_folder = NULL;
  TreeElement *dest_element = NULL;
  if (path != NULL
      && gtk_tree_selection_path_is_selected(selection, path))
    {
      // pathΰ֤paste
      GtkTreeModel *model = gtk_tree_view_get_model(tree_view);
      g_return_val_if_fail(model != NULL, false);

      GtkTreeIter cursor_iter;
      if (!gtk_tree_model_get_iter(model, &cursor_iter, path))
	return false;

      TreeRowType row_type = TREE_ROW_TYPE_UNKNOWN;
      gtk_tree_model_get(model, &cursor_iter,
			 TREE_ELEMENT_DATA, &dest_element,
			 TREE_ROW_TYPE, &row_type,
			 -1);
      g_return_val_if_fail(row_type == TREE_ROW_TYPE_ELEMENT, false);

      GtkTreeIter parent_iter;
      if (gtk_tree_model_iter_parent(model, &parent_iter, &cursor_iter))
	{
	  gtk_tree_model_get(model, &parent_iter,
			     TREE_ELEMENT_DATA, &dest_folder,
			     TREE_ROW_TYPE, &row_type,
			     -1);
	  g_return_val_if_fail(row_type == TREE_ROW_TYPE_ELEMENT, false);
	  g_return_val_if_fail(dest_folder != NULL, false);
	  g_return_val_if_fail(dest_folder->is_folder(), false);
	}
    }

  if (dest_folder == NULL)
    dest_folder = impl.get_root_folder();	// root_folderpaste

  vector<SelectedRow>::const_iterator iter = data->rows_begin();
  vector<SelectedRow>::const_iterator end_iter = data->rows_end();
  if (dest_element != NULL)
    {
      while (iter != end_iter)
	{
	  const SelectedRow &row = *iter++;
	  TreeElement *elm = row.get_element();
	  if (!elm->is_alone())
	    {
	      if (elm->is_folder())
		elm = impl.get_root_folder()->copy(dynamic_cast<Folder *>(elm));
	      else
		elm = impl.get_root_folder()->copy(elm);
	    }

	  if (elm->is_folder())
	    impl.insert_folder(dynamic_cast<Folder *>(elm),
			       dest_folder, dest_element,
			       GTK_TREE_VIEW_DROP_INTO_OR_AFTER);
	  else
	    impl.insert_element(elm, dest_folder, dest_element,
				GTK_TREE_VIEW_DROP_INTO_OR_AFTER);
	}
    }
  else
    {
      while (iter != end_iter)
	{
	  const SelectedRow &row = *iter++;
	  TreeElement *elm = row.get_element();
	  if (!elm->is_alone())
	    {
	      if (elm->is_folder())
		elm = impl.get_root_folder()->copy(dynamic_cast<Folder *>(elm));
	      else
		elm = impl.get_root_folder()->copy(elm);
	    }

	  impl.new_element(dest_folder, elm);
	}
    }

  return true;
}


SelectionData *
TreeViewSelectionData::snapshot()
{
  return new TreeViewSelectionData(impl, selection, false);
}


void
TreeViewSelectionData::accumulate_elements(GtkTreeModel *model,
					   GtkTreePath *path,
					   GtkTreeIter *iter, gpointer data)
{
  TreeElement *element = NULL;
  TreeRowType row_type = TREE_ROW_TYPE_UNKNOWN;

  gtk_tree_model_get(model, iter,
		     TREE_ELEMENT_DATA, &element,
		     TREE_ROW_TYPE, &row_type,
		     -1);

  g_return_if_fail(row_type == TREE_ROW_TYPE_ELEMENT);
  g_return_if_fail(element != NULL);

  TreeViewSelectionData *selection_data
    = static_cast<TreeViewSelectionData *>(data);

  Folder *folder;
  GtkTreeIter parent_iter;
  if (gtk_tree_model_iter_parent(model, &parent_iter, iter))
    {
      TreeElement *parent;
      gtk_tree_model_get(model, &parent_iter,
			 TREE_ELEMENT_DATA, &parent,
			 TREE_ROW_TYPE, &row_type,
			 -1);
      g_return_if_fail(row_type == TREE_ROW_TYPE_ELEMENT);
      g_return_if_fail(parent != NULL);
      g_return_if_fail(parent->is_folder());

      folder = dynamic_cast<Folder *>(parent);
      g_return_if_fail(folder != NULL);
    }
  else
    folder = selection_data->impl.get_root_folder();

  selection_data->rows.push_back(SelectedRow(folder, element, path));

  if (!element->is_folder())
    selection_data->non_folders.push_back(element);
}


void
TreeViewSelectionData::changed_cb(GtkTreeSelection *selection,
				  TreeViewSelectionData *data)
{
  data->rows.clear();
  data->non_folders.clear();

  gtk_tree_selection_selected_foreach(selection, accumulate_elements, data);
}


const GtkTargetEntry *
TreeViewSelectionData::get_targets()
{
  if (!non_folders.empty())
    return all_targets;

  if (!rows.empty())
    return linkless_targets;

  return empty_targets;
}


guint
TreeViewSelectionData::get_n_targets()
{
  if (!non_folders.empty())
    return G_N_ELEMENTS(all_targets);

  if (!rows.empty())
    return G_N_ELEMENTS(linkless_targets);

  return 0;
}


void
TreeViewSelectionData::get_data(GtkClipboard *clipboard,
				GtkSelectionData *data, guint info)
{
  set_data(data, info, GDK_NONE);
}


bool
TreeViewSelectionData::clear_data(GtkClipboard *clipboard)
{
  return !is_static;
}


bool
TreeViewSelectionData::set_data(GtkSelectionData *data, guint info,
				GdkAtom type)
{
  switch (info)
    {
    case TARGET_INTERNAL_LIST_OF_ELEMENTS:
      {
	g_return_val_if_fail(type != GDK_NONE, false);

	char pointer_buf[256];
	if (snprintf(pointer_buf, 256,
		     "Ochusha::TreeViewSelectionData *data=%p;", this) < 256)
	  {
	    gtk_selection_data_set(data, type, 8,
				   reinterpret_cast<guchar *>(pointer_buf),
				   strlen(pointer_buf));
	    return true;
	  }
	return false;
      }

    case TARGET_TEXT_URI_LIST:
      {
	gchar **uri_list
	  = static_cast<gchar **>(malloc(sizeof(gchar *)
					 * (non_folders.size() + 1)));
	vector<TreeElement *>::iterator iter = non_folders.begin();
	vector<TreeElement *>::iterator end_iter = non_folders.end();
	int i = 0;
	while (iter != end_iter)
	  uri_list[i++] = const_cast<gchar *>(impl.get_url(*iter++));
	uri_list[i] = NULL;
	gboolean result = gtk_selection_data_set_uris(data, uri_list);
	free(uri_list);
	return result != FALSE;
      }

    case TARGET_UTF8_STRING:
      {
	string buf;
	vector<SelectedRow>::const_iterator iter = rows_begin();
	vector<SelectedRow>::const_iterator end_iter = rows_end();
	while (iter != end_iter)
	  {
	    const SelectedRow &row = *iter++;
	    TreeElement *element = row.get_element();
	    buf.append(TreeUIImpl::canonicalize_element_name(element))
	      .append("\n");
	    if (!element->is_folder())
	      buf.append(impl.get_url(element)).append("\n");
	  }
	return gtk_selection_data_set_text(data, buf.c_str(), -1) != FALSE;
      }
    }

  return false;
}


bool
TreeViewSelectionData::is_selected(GtkTreePath *path)
{
  return gtk_tree_selection_path_is_selected(selection, path);
}


SelectionData *
TreeViewSelectionData::create_proxy()
{
  GtkWidget *widget = GTK_WIDGET(gtk_tree_selection_get_tree_view(selection));
  return new SelectionDataProxy(this, widget);
}


void
TreeViewSelectionData::toggle_hide()
{
  GtkTreeView *view = gtk_tree_selection_get_tree_view(selection);
  g_return_if_fail(view != NULL);

  GtkTreeModel *model = gtk_tree_view_get_model(view);
  g_return_if_fail(model != NULL);

  vector<SelectedRow>::const_iterator iter = rows_begin();
  vector<SelectedRow>::const_iterator end_iter = rows_end();
  while (iter != end_iter)
    {
      const SelectedRow &row = *iter++;
      GtkTreeIter tree_iter;
      if (gtk_tree_model_get_iter(model, &tree_iter, row.get_path()))
	{
	  TreeElement *element = row.get_element();
	  element->set_hidden(element->is_shown());
	  impl.emit_row_changed_from_filter(model, &tree_iter);
	}
    }
}


void
TreeViewSelectionData::hide()
{
  GtkTreeView *view = gtk_tree_selection_get_tree_view(selection);
  g_return_if_fail(view != NULL);

  GtkTreeModel *model = gtk_tree_view_get_model(view);
  g_return_if_fail(model != NULL);

  list<GtkTreeRowReference *> rows_to_be_removed;

  vector<SelectedRow>::const_iterator iter = rows_begin();
  vector<SelectedRow>::const_iterator end_iter = rows_end();
  while (iter != end_iter)
    {
      const SelectedRow &row = *iter++;
      GtkTreeIter tree_iter;
      if (gtk_tree_model_get_iter(model, &tree_iter, row.get_path()))
	{
	  TreeElement *element = row.get_element();
	  element->hide();
	  GtkTreeRowReference *reference
	    = gtk_tree_row_reference_new(model, row.get_path());
	  rows_to_be_removed.push_front(reference);
	}
    }

  while (!rows_to_be_removed.empty())
    {
      GtkTreeRowReference *reference = rows_to_be_removed.front();
      rows_to_be_removed.pop_front();

      GtkTreeIter master_iter;
      if (impl.get_master_iter(model, &master_iter, reference))
	gtk_tree_store_remove(impl.master_store, &master_iter);
      gtk_tree_row_reference_free(reference);
    }
}


void
TreeViewSelectionData::open_properties()
{
  vector<SelectedRow>::const_iterator iter = rows_begin();
  vector<SelectedRow>::const_iterator end_iter = rows_end();
  while (iter != end_iter)
    {
      const SelectedRow &row = *iter++;
      impl.open_properties(row.get_element());
    }
}
// TreeViewSelectionDataμ



// TreeViewFocusNotifierμ
TreeViewFocusNotifier::TreeViewFocusNotifier(TreeView &view, GtkWidget *widget)
  : FocusNotifier(view.impl.get_controller(), widget, false),
    view(view)
{
}


TreeViewFocusNotifier::~TreeViewFocusNotifier()
{
#ifdef VERBOSE_DESTRUCTOR
  fprintf(stderr, "TreeViewFocusNotifier::~TreeViewFocusNotifier(): selection_data=%p\n",
	  &view.selection_data);
#endif
  view.impl.get_controller().clear_clipboard_internal(&view.selection_data);
}


gboolean
TreeViewFocusNotifier::notify_focus_in()
{
#ifdef DEBUG_FOCUS_NOTIFIER
  fprintf(stderr, "TreeViewFocusNotifier::notify_focus_in(): %p, %d\n",
	  this, has_pointer());
#endif
  view.impl.get_controller().update_clipboard_internal(&view.selection_data);
  view.notify_focus_in();

  return FALSE;
}


void
TreeViewFocusNotifier::notify_unmap()
{
  view.impl.get_controller().clear_clipboard_internal(&view.selection_data);
}
// TreeViewFocusNotifierμ



// TreeViewμ
static gboolean name_match_func(OchushaTreeSelectorView *view,
				GtkTreeModel *model, GtkTreeIter *iter,
				OchushaSearchKey *key, gpointer unused);


TreeView::TreeView(GtkTreeModel *model, const gchar *title, TreeUIImpl &impl,
		   bool is_master_view)
  : View(gtk_label_new(title), gtk_scrolled_window_new(NULL, NULL)),
    impl(impl),
    title(title != NULL ? g_strdup(title) : NULL),
    tree_view(ochusha_tree_selector_view_new_with_model(model)),
    selection_data(impl,
		   gtk_tree_view_get_selection(GTK_TREE_VIEW(tree_view)), true),
    focus_notifier(*this, tree_view),
    last_event_path(NULL),
    press_start_x(0),
    press_start_y(0),
    press_start_state(0),
    press_start_button(0),
    press_start_type(GDK_NOTHING),
    popup_delay_timeout_id(0),
    is_master_view(is_master_view)
{
  GtkWidget *widget = get_widget();
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(widget),
				 GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

#if GTK_CHECK_VERSION(2,10,0)
  gtk_tree_view_set_enable_tree_lines(GTK_TREE_VIEW(tree_view), TRUE);
#endif
  gtk_tree_view_set_enable_search(GTK_TREE_VIEW(tree_view), FALSE);

  gtk_tree_view_set_row_separator_func(GTK_TREE_VIEW(tree_view),
				       TreeUIImpl::is_separator_iter,
				       &impl, NULL);

  if (is_master_view)
    {
      g_signal_connect(impl.master_store, "row_changed",
		       reinterpret_cast<GCallback>(row_changed_cb), this);
      g_signal_connect(tree_view, "row_expanded",
		       reinterpret_cast<GCallback>(row_expanded_cb), this);
      g_signal_connect(tree_view, "row_collapsed",
		       reinterpret_cast<GCallback>(row_collapsed_cb), this);
      g_signal_connect(tree_view, "start_interactive_search",
		       reinterpret_cast<GCallback>(start_interactive_search_cb),
		       this);
      g_signal_connect(tree_view, "key_press_event",
		       reinterpret_cast<GCallback>(key_press_event_cb), this);
    }

  g_signal_connect(tree_view, "popup_menu",
		   reinterpret_cast<GCallback>(popup_menu_cb), this);

  g_signal_connect(tree_view, "row_button_press",
		   reinterpret_cast<GCallback>(row_button_press_cb), this);
  g_signal_connect(tree_view, "row_button_release",
		   reinterpret_cast<GCallback>(row_button_release_cb), this);
  g_signal_connect(tree_view, "motion_notify_event",
		   reinterpret_cast<GCallback>(motion_notify_event_cb), this);

  GtkCellRenderer *renderer = gtk_cell_renderer_pixbuf_new();
  GtkTreeViewColumn *column = gtk_tree_view_column_new();
  if (title != NULL && !is_master_view)
    gtk_tree_view_column_set_title(column, title);
  else
    gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(tree_view), FALSE);

  gtk_tree_view_column_set_alignment(column, 0.0);
  gtk_tree_view_column_pack_start(column, renderer, FALSE);
  gtk_tree_view_column_set_cell_data_func(column, renderer, column_set_icon,
					  this, NULL);

  renderer = gtk_cell_renderer_text_new();
  gtk_tree_view_column_pack_start(column, renderer, TRUE);
  gtk_tree_view_column_set_cell_data_func(column, renderer, column_set_name,
					  this, NULL);

  gtk_tree_view_append_column(GTK_TREE_VIEW(tree_view), column);

  gtk_widget_show(tree_view);
  gtk_container_add(GTK_CONTAINER(widget), tree_view);

  ochusha_tree_selector_view_set_search_equal_func(
					OCHUSHA_TREE_SELECTOR_VIEW(tree_view),
					name_match_func,
					NULL, NULL);

  initialize_for_drag_source();
  initialize_for_drag_dest();

  GtkTreeSelection *selection
    = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree_view));
  gtk_tree_selection_set_mode(selection, GTK_SELECTION_MULTIPLE);

  g_signal_connect(selection, "changed",
		   reinterpret_cast<GCallback>(selection_changed_cb), this);
}


TreeView::~TreeView()
{
  if (popup_delay_timeout_id != 0)
    g_source_remove(popup_delay_timeout_id);

  if (title != NULL)
    g_free(title);
}


gboolean
TreeView::internal_list_of_elements_available(GtkClipboard *clipboard)
{
  g_return_val_if_fail(GTK_IS_CLIPBOARD(clipboard), FALSE);

  return gtk_clipboard_wait_is_target_available(clipboard,
						list_of_elements_atom);
}


TreeUIImpl *
TreeView::get_tree_ui_impl_from_clipboard(GtkClipboard *clipboard)
{
  g_return_val_if_fail(internal_list_of_elements_available(clipboard), NULL);

  SelectionData *data = SelectionData::get_from(clipboard);
  g_return_val_if_fail(data != NULL, NULL);

  TreeViewSelectionData *selection_data
    = dynamic_cast<TreeViewSelectionData *>(data);
  g_return_val_if_fail(selection_data != NULL, NULL);

  return &selection_data->get_impl();
}


void
TreeView::popup_context_menu(GtkTreeSelection *selection, guint button,
			     guint32 event_time)
{
  impl.popup_context_menu(this, selection, button, event_time);
}


bool
TreeView::activate_element(TreeElement *element, GdkEventType event_type,
			   guint button)
{
  if (!element->is_folder() || button == 2)
    {
      Controller &controller = impl.controller;
      Preferences &prefs = controller.get_prefs();
      bool in_tab;
      if (!element->is_folder())
	in_tab = (button == 2) ^ (prefs.get_default_open_in_tab());
      else
	in_tab = true;
      bool select_page = prefs.get_select_tab_when_opened();
      impl.open(element, in_tab, false, select_page);
      return true;
    }

  return false;
}


void
TreeView::column_set_icon(GtkCellRendererPixbuf *renderer,
			  TreeElement *element, gboolean expanded)
{
  g_object_set(renderer,
	       "stock-id", impl.get_element_icon_stock_id(element, expanded),
	       "stock-size", GTK_ICON_SIZE_MENU,
	       NULL);
}


void
TreeView::column_set_name(GtkCellRendererText *renderer, TreeElement *element)
{
  gboolean hidden = element->is_hidden();
  Preferences &prefs = impl.controller.get_prefs();
  GdkColor *fg;
  GdkColor *bg;
  if (!hidden)
    {
      fg = prefs.get_fg_normal();
      bg = prefs.get_bg_normal();
    }
  else
    {
      fg = prefs.get_fg_hidden();
      bg = prefs.get_bg_hidden();
    }

  g_object_set(renderer,
	       "text", TreeUIImpl::canonicalize_element_name(element),
	       "strikethrough", hidden,
	       "strikethrough_set", hidden,
	       "foreground_gdk", fg,
	       "cell_background_gdk", bg,
	       NULL);
}


void
TreeView::column_set_icon(GtkTreeViewColumn *column, GtkCellRenderer *renderer,
			  GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
  g_return_if_fail(GTK_IS_CELL_RENDERER_PIXBUF(renderer));

  TreeElement *element = NULL;
  TreeRowType row_type = TREE_ROW_TYPE_UNKNOWN;
  gtk_tree_model_get(model, iter,
		     TREE_ELEMENT_DATA, &element,
		     TREE_ROW_TYPE, &row_type,
		     -1);

#if GTK_CHECK_VERSION(2,10,0)
  g_return_if_fail(row_type != TREE_ROW_TYPE_UNKNOWN);
#else
  if (row_type == TREE_ROW_TYPE_UNKNOWN)
    return;
#endif

  if (row_type != TREE_ROW_TYPE_ELEMENT)
    return;

  g_return_if_fail(element != NULL);
  g_return_if_fail(!element->is_separator());

  TreePathHandle path(gtk_tree_model_get_path(model, iter));
  TreeView *view = static_cast<TreeView *>(data);
  gboolean expanded
    = gtk_tree_view_row_expanded(GTK_TREE_VIEW(view->tree_view), path);
  view->column_set_icon(reinterpret_cast<GtkCellRendererPixbuf *>(renderer),
			element, expanded);
}


void
TreeView::column_set_name(GtkTreeViewColumn *column, GtkCellRenderer *renderer,
			  GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
  g_return_if_fail(GTK_IS_CELL_RENDERER_TEXT(renderer));

  TreeElement *element = NULL;
  TreeRowType row_type = TREE_ROW_TYPE_UNKNOWN;
  gtk_tree_model_get(model, iter,
		     TREE_ELEMENT_DATA, &element,
		     TREE_ROW_TYPE, &row_type,
		     -1);

#if GTK_CHECK_VERSION(2,10,0)
  g_return_if_fail(row_type != TREE_ROW_TYPE_UNKNOWN);
#else
  if (row_type == TREE_ROW_TYPE_UNKNOWN)
    return;
#endif
  g_return_if_fail(row_type == TREE_ROW_TYPE_ELEMENT);
  g_return_if_fail(element != NULL);
  g_return_if_fail(!element->is_separator());

  TreeView *view = static_cast<TreeView *>(data);
  view->column_set_name(reinterpret_cast<GtkCellRendererText *>(renderer),
			element);
}



// Viewvirtualμ
const gchar *
TreeView::get_combo_title()
{
  return title != NULL ? title : _("Anonymous");
}


const gchar *
TreeView::get_menu_title()
{
  return title != NULL ? title : _("Anonymous");
}


const gchar *
TreeView::get_window_title()
{
  return NULL;
}


const gchar *
TreeView::get_url()
{
  return NULL;
}


void
TreeView::grab_focus()
{
  gtk_widget_grab_focus(tree_view);
}


bool
TreeView::contains(GtkWidget *child)
{
  return child == tree_view || child == get_widget();
}


const gchar *
TreeView::get_search_title()
{
  return impl.get_search_title();
}


OchushaSearchable *
TreeView::get_search_target()
{
  return OCHUSHA_SEARCHABLE(tree_view);
}


void
TreeView::notify_closed()
{
  // ⤻
}
// Viewvirtualμ



// GtkTreeSelectionϢ
struct SelectedItemState
{
  bool has_folder;
  bool has_item;
};


static void
check_selected_items(GtkTreeModel *model, GtkTreePath *path,
		     GtkTreeIter *iter, gpointer data)
{
  SelectedItemState *state = static_cast<SelectedItemState *>(data);

  TreeElement *element = NULL;
  TreeRowType row_type = TREE_ROW_TYPE_UNKNOWN;
  gtk_tree_model_get(model, iter,
		     TREE_ELEMENT_DATA, &element,
		     TREE_ROW_TYPE, &row_type,
		     -1);
  g_return_if_fail(row_type == TREE_ROW_TYPE_ELEMENT);
  g_return_if_fail(element != NULL);
  g_return_if_fail(!element->is_separator());

  if (element->is_folder())
    state->has_folder = true;
  else
    state->has_item = true;
}


void
TreeView::selection_changed_cb(GtkTreeSelection *selection,
			       TreeView *view)
{
  SelectedItemState state = { false, false };
  gtk_tree_selection_selected_foreach(selection, check_selected_items, &state);

  view->impl.get_controller().get_tree_ui()
    .notify_item_selected(state.has_item);
  view->impl.notify_selection_changed(selection,
				      state.has_folder, state.has_item);
  view->impl.get_controller().update_clipboard_internal(&view->selection_data);
}
// GtkTreeSelectionϢ



// GtkTreeModelΥʥϥɥ
void
TreeView::row_changed_cb(GtkTreeModel *model, GtkTreePath *path,
			 GtkTreeIter *iter, TreeView *view)
{
  g_return_if_fail(view->is_master_view);

  // modelmaster_store
  GtkTreeView *tree_view = GTK_TREE_VIEW(view->tree_view);

  TreeElement *element = NULL;
  TreeRowType row_type = TREE_ROW_TYPE_UNKNOWN;
  gtk_tree_model_get(model, iter,
		     TREE_ELEMENT_DATA, &element,
		     TREE_ROW_TYPE, &row_type,
		     -1);
  if (row_type != TREE_ROW_TYPE_ELEMENT)
    return;
  g_return_if_fail(element != NULL);
  if (!element->is_folder())
    return;

  Folder *folder = dynamic_cast<Folder *>(element);
  assert(folder != NULL);

  GtkTreeModel *view_model = gtk_tree_view_get_model(tree_view);
  TreeUIImpl &impl = view->impl;
  GtkTreeIter view_iter;
  if (!impl.convert_master_iter_to_iter(view_model, &view_iter, iter))
    return;

  TreePathHandle view_path(gtk_tree_model_get_path(view_model, &view_iter));
  if (folder->is_expanded() != gtk_tree_view_row_expanded(tree_view, view_path))
    {
      if (folder->is_expanded())
	gtk_tree_view_expand_row(tree_view, view_path, FALSE);
      else
	gtk_tree_view_collapse_row(tree_view, view_path);
    }
}
// GtkTreeModelΥʥϥɥ



// OchushaTreeSelectorViewΥʥϥɥ
gboolean
TreeView::start_interactive_search_cb(OchushaTreeSelectorView *view,
				      TreeView *tree_view)
{
  tree_view->request_search();
  return TRUE;
}


gboolean
TreeView::key_press_event_cb(OchushaTreeSelectorView *widget,
			     GdkEventKey *event, TreeView *view)
{
  g_return_val_if_fail(OCHUSHA_IS_TREE_SELECTOR_VIEW(widget), FALSE);

  // MEMO: ϡɥǥ󥰤Τ򤱤Τǡ
  //       schemeؿꤲ褦ˤʤͽꡣ

  /*
   * ʲ괺ΤΥХǥ󥰤Ƹ
   *
   * GDK_o: open-selection
   * GDK_t: open-selection-in-tab
   *
   * GDK_d: toggle-thread-hidden-state
   * GDK_space: open-selection
   */

  if ((event->state & GDK_MODIFIER_MASK) != 0)
    return FALSE;	// ǥեդΥХɤϤƤʤ


  Controller &controller = view->impl.controller;
  LinkUI &link_ui = controller.get_link_ui();
  TreeUI &tree_ui = controller.get_tree_ui();

  GtkAction *action;
  switch (event->keyval)
    {
    case GDK_space:
    case GDK_Return:
    case GDK_KP_Enter:
    case GDK_ISO_Enter:
    case GDK_o:
      action = link_ui.open_link_location;
      break;

    case GDK_p:
      {
	gboolean result = FALSE;
	g_signal_emit_by_name(widget, "move-cursor",
			      GTK_MOVEMENT_DISPLAY_LINES, -1, &result);
	action = NULL;
	break;
      }

    case GDK_n:
      {
	gboolean result = FALSE;
	g_signal_emit_by_name(widget, "move-cursor",
			      GTK_MOVEMENT_DISPLAY_LINES, 1, &result);
	action = NULL;
	break;
      }

    case GDK_t:
      action = link_ui.open_link_location_in_tab;
      break;

    case GDK_d:
      if (controller.get_prefs().get_show_hidden_boardlist_items())
	action = tree_ui.toggle_hide;
      else
	action = tree_ui.hide;
      break;
      
    default:
      return FALSE;
    }

  GtkTreePath *path = NULL;
  gtk_tree_view_get_cursor(GTK_TREE_VIEW(widget), &path, NULL);
  TreePathHandle cursor_path = path;

  if (path == NULL)
    return FALSE;

  gtk_tree_view_scroll_to_cell(GTK_TREE_VIEW(widget),
			       path, NULL, TRUE, 0.5, 0.0);

  if (!view->selection_data.is_selected(path))
    {
      gboolean result = FALSE;
      g_signal_emit_by_name(widget, "select-cursor-row", FALSE, &result);
      controller.update_clipboard_internal(&view->selection_data);
    }

  switch (event->keyval)
    {
    case GDK_space:
    case GDK_Return:
    case GDK_KP_Enter:
    case GDK_ISO_Enter:
      {
	GtkTreeIter iter;
	GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(widget));
	g_return_val_if_fail(model != NULL, TRUE);

	if (!gtk_tree_model_get_iter(model, &iter, path))
	  return TRUE;

	TreeElement *element = NULL;
	TreeRowType row_type = TREE_ROW_TYPE_UNKNOWN;
	gtk_tree_model_get(model, &iter,
			   TREE_ELEMENT_DATA, &element,
			   TREE_ROW_TYPE, &row_type,
			   -1);
	if (row_type != TREE_ROW_TYPE_ELEMENT)
	  return TRUE;

	if (!element->is_folder())
	  break;

	Folder *folder = dynamic_cast<Folder *>(element);
	g_return_val_if_fail(folder != NULL, TRUE);

	folder->set_expanded(!folder->is_expanded());
	view->impl.emit_row_changed_from_filter(model, &iter);

	return TRUE;
      }

    default:
      break;
    }

  if (action == NULL)
    return TRUE;

  gtk_action_activate(action);
  //activate_view_at_cursor();
  if (!controller.get_prefs().get_select_tab_when_opened())
    view->grab_focus();		// եȥ५ĤΤ

  return TRUE;
}


// ֤Ĥ줬ɽɽƤ֤ˤ롣
// ɽƤˤоݤView *֤ʳξˤ
// NULL֤
View *
TreeView::activate_view_at_cursor()
{
  return NULL;	// ̤
}


gboolean
TreeView::popup_menu_cb(GtkWidget *widget, TreeView *view)
{
  GtkTreeSelection *selection
    = gtk_tree_view_get_selection(GTK_TREE_VIEW(view->tree_view));
  view->popup_context_menu(selection, 0, gtk_get_current_event_time());

  return TRUE;
}


static gboolean
name_match_func(OchushaTreeSelectorView *view, GtkTreeModel *model,
		GtkTreeIter *iter, OchushaSearchKey *key, gpointer unused)
{
  TreeElement *element = NULL;
  TreeRowType row_type = TREE_ROW_TYPE_UNKNOWN;
  gtk_tree_model_get(model, iter,
		     TREE_ELEMENT_DATA, &element,
		     TREE_ROW_TYPE, &row_type,
		     -1);

  g_return_val_if_fail(row_type == TREE_ROW_TYPE_ELEMENT, FALSE);
  g_return_val_if_fail(element != NULL, FALSE);

  return ochusha_search_key_match_text(
				key,
				TreeUIImpl::canonicalize_element_name(element),
				-1);
}


void
TreeView::row_expanded_cb(GtkTreeView *tree_view, GtkTreeIter *iter,
			  GtkTreePath *path, TreeView *view)
{
  g_return_if_fail(view->is_master_view);

  GtkTreeModel *model = gtk_tree_view_get_model(tree_view);

  TreeElement *element = NULL;
  TreeRowType row_type = TREE_ROW_TYPE_UNKNOWN;
  gtk_tree_model_get(model, iter,
		     TREE_ELEMENT_DATA, &element,
		     TREE_ROW_TYPE, &row_type,
		     -1);
  if (row_type != TREE_ROW_TYPE_ELEMENT)
    return;
  g_return_if_fail(element != NULL);
  g_return_if_fail(element->is_folder());

  GtkTreeIter children;
  if (gtk_tree_model_iter_children(model, &children, iter))
    {
      do
	{
	  TreeElement *child_element = NULL;
	  TreeRowType child_row_type = TREE_ROW_TYPE_UNKNOWN;
	  gtk_tree_model_get(model, &children,
			     TREE_ELEMENT_DATA, &child_element,
			     TREE_ROW_TYPE, &child_row_type,
			     -1);
	  if (child_row_type == TREE_ROW_TYPE_ELEMENT)
	    {
	      g_return_if_fail(child_element != NULL);
	      if (!child_element->is_folder())
		continue;

	      Folder *child_folder = dynamic_cast<Folder *>(child_element);
	      assert(child_folder != NULL);

	      TreePathHandle child_path(gtk_tree_model_get_path(model,
								&children));
	      if (child_folder->is_expanded())
		gtk_tree_view_expand_row(tree_view, child_path, FALSE);
	      else
		gtk_tree_view_collapse_row(tree_view, child_path);
	    }
	}
      while (gtk_tree_model_iter_next(model, &children));
    }

  Folder *folder = dynamic_cast<Folder *>(element);
  assert(folder != NULL);

  if (folder->is_expanded())
    return;

  folder->set_expanded(true);
  view->impl.emit_row_changed_from_filter(model, iter);
}


void
TreeView::row_collapsed_cb(GtkTreeView *tree_view, GtkTreeIter *iter,
			   GtkTreePath *path, TreeView *view)
{
  g_return_if_fail(view->is_master_view);

  GtkTreeModel *model = gtk_tree_view_get_model(tree_view);
  TreeElement *element = NULL;
  TreeRowType row_type = TREE_ROW_TYPE_UNKNOWN;
  gtk_tree_model_get(model, iter,
		     TREE_ELEMENT_DATA, &element,
		     TREE_ROW_TYPE, &row_type,
		     -1);
  if (row_type != TREE_ROW_TYPE_ELEMENT)
    return;
  g_return_if_fail(element != NULL);
  g_return_if_fail(element->is_folder());

  Folder *folder = dynamic_cast<Folder *>(element);
  assert(folder != NULL);

  if (!folder->is_expanded())
    return;

  folder->set_expanded(false);
  view->impl.emit_row_changed_from_filter(model, iter);
}


#define EXPANDER_EXTRA_PADDING	4	// gtk+-2.6.10
// ɬ׺Ǿ¤åʤ٥Ȥexpander columnǵä
// ꤷơ줬expander arrow()ξ夫ɤȽꤹ롣
static bool
coords_are_over_arrow(GtkTreeView *view, GtkTreePath *path,
		      GdkEventButton *event)
{
  gint depth = gtk_tree_path_get_depth(path) - 1;
  if (depth < 0)
    return false;

  gint expander_size;
  gtk_widget_style_get(GTK_WIDGET(view), "expander_size", &expander_size, NULL);
  expander_size += EXPANDER_EXTRA_PADDING;
  gint offset = expander_size * depth;
  gint arrow_width = expander_size + 1;
  gint x = static_cast<gint>(event->x);

  return x >= offset && x < (offset + arrow_width);
}


gboolean
TreeView::row_button_press_cb(OchushaTreeSelectorView *tree_view,
			      GtkTreePath *path, GtkTreeViewColumn *column,
			      GdkEventButton *event, TreeView *view)
{
  view->last_event_path = NULL;

  if (coords_are_over_arrow(GTK_TREE_VIEW(tree_view), path, event))
    {
      view->press_start_type = GDK_NOTHING;
      return FALSE;
    }

  view->press_start_type = event->type;
  view->press_start_x = static_cast<gint>(event->x);
  view->press_start_y = static_cast<gint>(event->y);
  view->press_start_state = event->state;
  view->press_start_button = event->button;

  if (event->button == 3)
    {
      GtkTreeIter iter;
      GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(tree_view));

      if (!gtk_tree_model_get_iter(model, &iter, path))
	return FALSE;

      if (view->popup_delay_timeout_id != 0)
	g_source_remove(view->popup_delay_timeout_id);

      view->popup_delay_timeout_id
	= g_timeout_add(POPUP_MENU_DELAY_TIMEOUT,
			reinterpret_cast<GSourceFunc>(popup_delay_timeout_cb),
			view);

      return view->selection_data.is_selected(path);
    }

  if ((event->state & (GDK_CONTROL_MASK | GDK_SHIFT_MASK)) == 0)
    view->last_event_path = gtk_tree_path_copy(path);

  return FALSE;
}


gboolean
TreeView::motion_notify_event_cb(GtkWidget *widget, GdkEventMotion *event,
				 TreeView *view)
{
  if (view->popup_delay_timeout_id != 0
      && gtk_drag_check_threshold(widget,
				  view->press_start_x, view->press_start_y,
				  static_cast<gint>(event->x),
				  static_cast<gint>(event->y)))
    {
      g_source_remove(view->popup_delay_timeout_id);
      view->popup_delay_timeout_id = 0;
    }

  return FALSE;
}


gboolean
TreeView::popup_delay_timeout_cb(TreeView *view)
{
  // g_timeout_addΥХåGLibɳʤΤGDK_THREADS_ENTER()Ǥʤ
  GDK_THREADS_ENTER();

  view->popup_delay_timeout_id = 0;
  GtkTreeSelection *selection
    = gtk_tree_view_get_selection(GTK_TREE_VIEW(view->tree_view));
  view->popup_context_menu(selection, view->press_start_button,
			   gtk_get_current_event_time());

  GDK_THREADS_LEAVE();


  return FALSE;
}


gboolean
TreeView::row_button_release_cb(OchushaTreeSelectorView *tree_view,
				GtkTreePath *path, GtkTreeViewColumn *column,
				GdkEventButton *event, TreeView *view)
{
  GdkEventType press_type = view->press_start_type;
  view->press_start_type = GDK_NOTHING;

  if (view->popup_delay_timeout_id != 0)
    {
      g_source_remove(view->popup_delay_timeout_id);
      view->popup_delay_timeout_id = 0;
      GtkTreeSelection *selection
	= gtk_tree_view_get_selection(GTK_TREE_VIEW(tree_view));
      view->popup_context_menu(selection, 0, event->time);
      return FALSE;
    }

  if (coords_are_over_arrow(GTK_TREE_VIEW(tree_view), path, event))
    {
      view->last_event_path = NULL;
      return FALSE;
    }

  if (static_cast<GtkTreePath *>(view->last_event_path) == NULL)
    return FALSE;

  GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(tree_view));
  GtkTreeIter iter;
  if (gtk_tree_path_compare(view->last_event_path, path) != 0
      || !gtk_tree_model_get_iter(model, &iter, path))
    {
      view->last_event_path = NULL;
      return FALSE;
    }

  view->last_event_path = NULL;

  TreeElement *element = NULL;
  TreeRowType row_type = TREE_ROW_TYPE_UNKNOWN;
  gtk_tree_model_get(model, &iter,
		     TREE_ELEMENT_DATA, &element,
		     TREE_ROW_TYPE, &row_type,
		     -1);
  g_return_val_if_fail(row_type == TREE_ROW_TYPE_ELEMENT, FALSE);
  g_return_val_if_fail(element != NULL, FALSE);
  g_return_val_if_fail(!element->is_separator(), FALSE);

  if (view->activate_element(element, press_type, event->button))
    return FALSE;

  if (element->is_folder())
    {
      Folder *folder = dynamic_cast<Folder *>(element);
      assert(folder != NULL);

      if (view->is_master_view)
	{
	  folder->set_expanded(!folder->is_expanded());
	  view->impl.emit_row_changed_from_filter(model, &iter);
	}
    }

  return FALSE;
}



// DnDϢ
// drag sourceϢ
void
TreeView::initialize_for_drag_source()
{
  // (GTK+-2.2κ)GDK_BUTTON3_MASKդƤʤƤⱦå
  // ɥå֤ˤʤäƤޤȤХä(GTK+2.6.10)ϡ
  // GDK_BUTTON3_MASKդƤƤⱦåǥɥå֤ˤʤʤ
  // äɤХ
  // GHIǤϥǥեȤCtrl+ɥåCOPY
  // Shift+ɥåMOVE
  // Ƥ롣
  // åɥåġġ¿ʬ˥塼ФƤޤä
  // ǥեʤʤ뤫ɥå٥Ȥȯʤ
  // button_press٥Ȥλǥݥåץåץ˥塼ФƤ뤬
  // ߥ٤timeoutꤷƤޤǤδ֤
  // gtk_drag_check_threshold()TRUEˤʤ٤motion_notifyȯ
  // ݥåץåפɽ֤⤷ʤץ⡼ɤȴ褦ˤ
  // ߤ褦timeoutbutton_release褿餽λǥݥåץåפ
  // ɽ롣timeoutޤǤgtk_drag_check_threshold()TRUEˤʤ餺
  // button_releaseȯʤä餽λϥݥåץåץ˥塼Ф
  int buttons = GDK_BUTTON1_MASK | GDK_BUTTON2_MASK | GDK_BUTTON3_MASK;
  int actions = GDK_ACTION_COPY | GDK_ACTION_MOVE | GDK_ACTION_ASK;
  gtk_tree_view_enable_model_drag_source(GTK_TREE_VIEW(tree_view),
					 static_cast<GdkModifierType>(buttons),
					 all_targets, G_N_ELEMENTS(all_targets),
					 static_cast<GdkDragAction>(actions));
  g_signal_connect(tree_view, "drag_data_get",
		   reinterpret_cast<GCallback>(drag_data_get_cb), this);
  g_signal_connect(tree_view, "drag_data_delete",
		   reinterpret_cast<GCallback>(drag_data_delete_cb), this);
  g_signal_connect(tree_view, "drag_end",
		   reinterpret_cast<GCallback>(drag_end_cb), this);
}


void
TreeView::drag_data_get_cb(GtkWidget *widget, GdkDragContext *context,
			   GtkSelectionData *data, guint info, guint time,
			   TreeView *view)
{
  g_return_if_fail(GTK_IS_TREE_VIEW(widget));

  GdkAtom type;
  if (info == TARGET_INTERNAL_LIST_OF_ELEMENTS)
    type = list_of_elements_atom;
  else
    type = GDK_NONE;

  view->selection_data.set_data(data, info, type);
}


void
TreeView::drag_data_delete_cb(GtkWidget *widget, GdkDragContext *context,
			      TreeView *view)
{
#ifdef DEBUG_DND_MORE
  fprintf(stderr, "TreeView::drag_data_delete_cb(%p, %p, %p): actions=%x suggested_action=%x action=%x\n",
	  widget, context, view,
	  context->actions, context->suggested_action, context->action);
#endif
  g_signal_stop_emission_by_name(widget, "drag_data_delete");
}


void
TreeView::drag_end_cb(GtkWidget *widget, GdkDragContext *context,
		      TreeView *view)
{
  view->press_start_state = 0;
}


static const GtkTargetEntry drop_targets[] = {
  target_internal_list_of_elements,
  target_text_uri_list
};



// drag destϢ
void
TreeView::initialize_for_drag_dest()
{
  int actions = GDK_ACTION_COPY | GDK_ACTION_MOVE | GDK_ACTION_ASK;
  gtk_tree_view_enable_model_drag_dest(GTK_TREE_VIEW(tree_view),
				       drop_targets, G_N_ELEMENTS(drop_targets),
				       static_cast<GdkDragAction>(actions));
  g_signal_connect(tree_view, "drag_drop",
		   reinterpret_cast<GCallback>(drag_drop_cb), this);
  g_signal_connect(tree_view, "drag_data_received",
		   reinterpret_cast<GCallback>(drag_data_received_cb), this);
}


gboolean
TreeView::drag_drop_cb(GtkWidget *widget, GdkDragContext *context,
		       gint x, gint y, guint time, TreeView *view)
{
  GdkAtom target = gtk_drag_dest_find_target(widget, context, NULL);

#ifdef DEBUG_DND_MORE
  fprintf(stderr, "TreeView::drag_drop_cb()\n");
#endif

  if (target != GDK_NONE)
    {
      gtk_drag_get_data(widget, context, target, time);
      g_signal_stop_emission_by_name(widget, "drag_drop");
      return TRUE;
    }

  return FALSE;
}


void
TreeView::drag_data_received_cb(GtkWidget *widget, GdkDragContext *context,
				gint x, gint y, GtkSelectionData *data,
				guint info, guint time, TreeView *view)
{
#ifdef DEBUG_DND_MORE
  fprintf(stderr,
	  "TreeView::drag_data_received_cb(%p, %p, %d, %d, %p, %d, %d, %p): actions=%x suggested_action=%x\n",
	  widget, context, x, y, data, info, time, view,
	  context->actions, context->suggested_action);
#endif

  bool is_drop_succeeded = false;

  GtkTreePath *path;
  GtkTreeViewDropPosition position;
  if (gtk_tree_view_get_dest_row_at_pos(GTK_TREE_VIEW(widget), x, y,
					&path, &position))
    {
      TreePathHandle handle(path);

#ifdef DEBUG_DND_MORE
      fprintf(stderr, "TreeView::drag_data_received_cb: position=%d\n",
	      position);
#endif
      GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(widget));
      GtkTreeIter drop_iter;
      TreeElement *drop_element = NULL;
      Folder *drop_folder = NULL;
      TreeRowType row_type = TREE_ROW_TYPE_UNKNOWN;
      if (gtk_tree_model_get_iter(model, &drop_iter, path))
	{
	  gtk_tree_model_get(model, &drop_iter,
			     TREE_ELEMENT_DATA, &drop_element,
			     TREE_ROW_TYPE, &row_type,
			     -1);
	  if (row_type == TREE_ROW_TYPE_ELEMENT)
	    {
	      GtkTreeIter parent_iter;
	      if (gtk_tree_model_iter_parent(model, &parent_iter, &drop_iter))
		{
		  gtk_tree_model_get(model, &parent_iter,
				     TREE_ELEMENT_DATA, &drop_folder,
				     TREE_ROW_TYPE, &row_type,
				     -1);
		  if (row_type != TREE_ROW_TYPE_ELEMENT
		      || !drop_folder->is_folder())
		    drop_element = NULL;
		}
	      else
		drop_folder = view->impl.get_root_folder();
	    }
	  else
	    drop_element = NULL;
	}

      if (drop_element != NULL)
	{
	  assert(drop_folder != NULL);

	  if (info == TARGET_INTERNAL_LIST_OF_ELEMENTS)
	    {
	      TreeViewSelectionData *selection_data;
	      if (data->type == list_of_elements_atom
		  && data->format == 8
		  && data->data != NULL
		  && sscanf(reinterpret_cast<char *>(data->data),
			    "Ochusha::TreeViewSelectionData *data=%p;",
			    &selection_data) == 1)
		{
		  bool is_move_action;
		  if (selection_data->get_impl().get_root_folder()
		      == view->impl.get_root_folder()
		      && (context->suggested_action == GDK_ACTION_MOVE
			  || view->press_start_state == 0)
		      && (context->actions & GDK_ACTION_MOVE) != 0)
		    is_move_action = true;
		  else
		    is_move_action = false;

		  vector<SelectedRow> drop_data;

		  vector<SelectedRow>::const_iterator s_iter
		    = selection_data->rows_begin();
		  vector<SelectedRow>::const_iterator s_end_iter
		    = selection_data->rows_end();

		  while (s_iter != s_end_iter)
		    drop_data.push_back(SelectedRow(*s_iter++));

		  TreeUIImpl &dest_impl = view->impl;

		  if (position == GTK_TREE_VIEW_DROP_BEFORE
		      || position == GTK_TREE_VIEW_DROP_INTO_OR_BEFORE)
		    {
		      // ˽
		      for (size_t i = 0; i < drop_data.size(); i++)
			{
			  Folder *src_folder = drop_data[i].get_folder();
			  TreeElement *src_element = drop_data[i].get_element();
			  is_drop_succeeded
			    = (dest_impl.process_drop_data(src_folder,
							   src_element,
							   drop_folder,
							   drop_element,
							   position,
							   is_move_action)
			       || is_drop_succeeded);
			}
		    }
		  else
		    {
		      // ս˽
		      for (size_t i = drop_data.size(); i-- > 0;)
			{
			  Folder *src_folder = drop_data[i].get_folder();
			  TreeElement *src_element = drop_data[i].get_element();
			  is_drop_succeeded
			    = (dest_impl.process_drop_data(src_folder,
							   src_element,
							   drop_folder,
							   drop_element,
							   position,
							   is_move_action)
			       || is_drop_succeeded);
			}
		    }
		}
	    }
	  else
	    {
	      // text/uri-list
	      gchar **uri_list = gtk_selection_data_get_uris(data);
	      if (uri_list != NULL)
		{
		  TreeUIImpl &dest_impl = view->impl;
		  gchar **uri_p = uri_list;
		  while (*uri_p != NULL)
		    {
		      is_drop_succeeded
			= (dest_impl.process_drop_data(*uri_p, drop_folder,
						       drop_element, position)
			   || is_drop_succeeded);
		      ++uri_p;
		    }
		  g_strfreev(uri_list);
		}
	    }
	}
      else
	{
#ifdef DEBUG_DND
	  fprintf(stderr, "Couldn't interpret drop position.\n");
#endif
	  is_drop_succeeded = false;
	}
    }
  else
    {
      // TODO: 롼ȥեؤΥɥåפȽǤ٤ʤġ
#ifdef DEBUG_DND
      fprintf(stderr, "Couldn't know drop position.\n");
#endif
      is_drop_succeeded = false;
    }

  // MOVEξǤ⡢source/destδطˤºݤεưѤΤǡ
  // DnDμŪˤϾCOPYäȻפäƤ

  // ºݤMOVE줿sourceκ⤳ãǴλ
  // ΤGTK+ݤ򸫤㤦ɬפϤʤ

  // ʤߤMOVEΤϡsourcedestmaster_modelƱǤ
  // Τߡ
  // COPYξǤ⡢ԹCOPY¸ǤʤˤMOVE롣
  // İǤƱեƱĤͣĤ¸ߤǤʤ
  gtk_drag_finish(context, is_drop_succeeded, FALSE, time);

  g_signal_stop_emission_by_name(widget, "drag_data_received");
}
// ޤDnDϢ



void
TreeView::restore_session(const char *name, SessionClient &session)
{
  GtkWidget *widget = get_widget();
  gtk_widget_show(widget);

  char name_buf[256];

  snprintf(name_buf, 256, "%s::hadj", name);
  const char *state = session.get_state(name_buf);
  if (state != NULL)
    {
      GtkAdjustment *adj
	= gtk_scrolled_window_get_hadjustment(GTK_SCROLLED_WINDOW(widget));
      gdouble value;
      if (adj != NULL && sscanf(state, "%lf", &value) == 1)
	gtk_adjustment_set_value(adj, value);
    }

  snprintf(name_buf, 256, "%s::vadj", name);
  state = session.get_state(name_buf);
  if (state != NULL)
    {
      GtkAdjustment *adj
	= gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(widget));
      gdouble value;
      if (adj != NULL && sscanf(state, "%lf", &value) == 1)
	gtk_adjustment_set_value(adj, value);
    }
}


void
TreeView::prepare_session(const char *name, SessionClient &session)
{
  GtkWidget *widget = get_widget();
  char name_buf[256];

  GtkAdjustment *adj
    = gtk_scrolled_window_get_hadjustment(GTK_SCROLLED_WINDOW(widget));
  if (adj != NULL)
    {
      gdouble value = gtk_adjustment_get_value(adj);
      char buf[64];
      snprintf(buf, 64, "%lf", value);
      snprintf(name_buf, 256, "%s::hadj", name);
      session.set_state(name_buf, buf);
    }

  adj = gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(widget));
  if (adj != NULL)
    {
      gdouble value = gtk_adjustment_get_value(adj);
      char buf[64];
      snprintf(buf, 64, "%lf", value);
      snprintf(name_buf, 256, "%s::vadj", name);
      session.set_state(name_buf, buf);
    }
}


FolderTreeView::FolderTreeView(GtkTreeModel *model, const gchar *title,
			       TreeUIImpl &impl)
  : TreeView(model, title, impl, false)
{
  GtkTreeSelection *selection
    = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree_view));
  gtk_tree_selection_set_mode(selection, GTK_SELECTION_SINGLE);
}


void
FolderTreeView::popup_context_menu(GtkTreeSelection *selection, guint button,
				   guint32 event_time)
{
  impl.popup_folder_tree_context_menu(this, selection, button, event_time);
}




}
