/* -*- c++ -*-
 * Copyright (c) 2003-2009 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: xsmp_handler.cc,v 1.15 2009/01/01 07:47:07 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "xsmp_handler.h"

#include "libochusha/ochusha_intl.h"

#include <glib.h>

#include <X11/SM/SMlib.h>
#include <X11/ICE/ICElib.h>

#include <cstdio>
#include <cstdlib>
#include <cstring>


namespace Ochusha
{

class XSMPHandlerPrivate
{
public:
  XSMPHandlerPrivate(XSMPHandler *handler)
  {
    this->handler = handler;
    smc_connection = NULL;
    ice_connection = NULL;
    ice_connection_poll_id = 0;
  }
  ~XSMPHandlerPrivate()
  {
    close_xsmp_connection();
  }

  bool open_xsmp_connection();
  void close_xsmp_connection();


private:
  XSMPHandler *handler;

  SmcConn smc_connection;

  IceConn ice_connection;
  guint ice_connection_poll_id;


  static gboolean ice_connection_poll_cb(GIOChannel *channel,
					 GIOCondition condition,
					 XSMPHandler *handler);
  static void ice_connection_watch_proc(IceConn ice_conn,
					IcePointer client_data, Bool opening,
					IcePointer *watch_data);
  static void save_yourself_cb(SmcConn smc_conn, SmPointer client_data,
			       int save_type, Bool shutdown,
			       int interact_style, Bool fast);
  static void die_cb(SmcConn smc_conn, SmPointer client_data);
  static void save_complete_cb(SmcConn smc_conn, SmPointer client_data);
  static void shutdown_cancelled_cb(SmcConn smc_conn, SmPointer client_data);
};


bool
XSMPHandlerPrivate::open_xsmp_connection()
{
  char error_string[4096];
  SmcCallbacks smc_callbacks =
    {
      {
	save_yourself_cb,
	handler
      },
      {
	die_cb,
	handler
      },
      {
	save_complete_cb,
	handler
      },
      {
	shutdown_cancelled_cb,
	handler
      }
    };

  IceInitThreads();	/* ᥤ󥹥åɤΤߤSMlibȤΤ¿ʬ */

  if (IceAddConnectionWatch(ice_connection_watch_proc, handler) == 0)
    {
      return false;
    }

#if DEBUG_XSMP
  if (handler->sm_client_id != NULL)
    fprintf(stderr, "previous sm_client_id=\"%s\"\n", handler->sm_client_id);
  else
    fprintf(stderr, "starting new session\n");
#endif

  smc_connection
    = SmcOpenConnection(NULL, NULL, 1, 0,
			(SmcSaveYourselfProcMask
			 | SmcDieProcMask
			 | SmcSaveCompleteProcMask
			 | SmcShutdownCancelledProcMask),
			&smc_callbacks,
			handler->previous_sm_client_id,
			&handler->sm_client_id,
			4096, error_string);

  if (smc_connection == NULL)
    {
#if DEBUG_XSMP
      fprintf(stderr, "Couldn't open SMC connection.\n");
#endif
      handler->sm_client_id = NULL;
      return false;
    }

  return true;
}


void
XSMPHandlerPrivate::close_xsmp_connection()
{
  if (ice_connection_poll_id != 0)
    {
      g_source_remove(ice_connection_poll_id);
      ice_connection_poll_id = 0;
    }

  if (smc_connection != NULL)
    {
      SmcCloseConnection(smc_connection, 0, NULL);
      smc_connection = NULL;
    }
}


gboolean
XSMPHandlerPrivate::ice_connection_poll_cb(GIOChannel *channel,
					   GIOCondition condition,
					   XSMPHandler *handler)
{
  XSMPHandlerPrivate *priv = handler->priv;
  if (priv->ice_connection == NULL)
    return FALSE;

  if (condition & (G_IO_IN | G_IO_PRI))
    {
      IceProcessMessagesStatus status
	= IceProcessMessages(static_cast<IceConn>(priv->ice_connection),
			     NULL, NULL);

      if (status == IceProcessMessagesIOError)
	{
	  IceCloseConnection(priv->ice_connection);
	  priv->ice_connection = NULL;
	  return FALSE;
	}
    }

  return TRUE;
}


void
XSMPHandlerPrivate::ice_connection_watch_proc(IceConn ice_conn,
					      IcePointer client_data,
					      Bool opening,
					      IcePointer *watch_data)
{
  XSMPHandler *handler = static_cast<XSMPHandler *>(client_data);
  XSMPHandlerPrivate *priv = handler->priv;

  priv->ice_connection = ice_conn;

  if (priv->ice_connection_poll_id != 0)
    {
      g_source_remove(priv->ice_connection_poll_id);
      priv->ice_connection_poll_id = 0;
    }

  if (opening)
    {
      int fd = IceConnectionNumber(ice_conn);
      if (fd >= 0)
	{
	  GIOChannel *channel = g_io_channel_unix_new(fd);
	  *watch_data = handler;
	  priv->ice_connection_poll_id
	    = g_io_add_watch_full(channel,
				  G_PRIORITY_HIGH,
				  static_cast<GIOCondition>(G_IO_IN|G_IO_PRI|G_IO_ERR|G_IO_HUP|G_IO_NVAL),
				  reinterpret_cast<GIOFunc>(ice_connection_poll_cb),
				  handler, NULL);
	  g_io_channel_unref(channel);
	}
    }
}


void
XSMPHandlerPrivate::save_yourself_cb(SmcConn smc_conn, SmPointer client_data,
				     int save_type, Bool shutdown,
				     int interact_style, Bool fast)
{
  XSMPHandler *handler = static_cast<XSMPHandler *>(client_data);
  SmProp prop = { NULL, NULL, 0, NULL };
  SmProp *props[] = { &prop, NULL };
  SmPropValue *prop_values = g_new0(SmPropValue, handler->argc + 1);
  int i;

  /* Initialize common parts of prop_values */
  for (i = 0; i < handler->argc; i++)
    {
      prop_values[i].length = strlen(handler->argv[i]);
      prop_values[i].value = handler->argv[i];
    }
  prop.type = const_cast<char *>(SmLISTofARRAY8);
  prop.num_vals = i;
  prop.vals = prop_values;


  /* set SmCloneCommand */
  prop.name = const_cast<char *>(SmCloneCommand);
  SmcSetProperties(smc_conn, 1, props);


  prop.num_vals = i + 1;

  /* set SmRestartCommand */
  prop_values[i].value = g_strconcat("--sm-client-id=", handler->sm_client_id,
				     NULL);
  prop_values[i].length = strlen(static_cast<char *>(prop_values[i].value));
  prop.name = const_cast<char *>(SmRestartCommand);
  SmcSetProperties(smc_conn, 1, props);
  g_free(prop_values[i].value);


  /* set SmDiscardCommand */
  prop_values[i].value = g_strconcat("--discard=", handler->sm_client_id,
				     NULL);
  prop_values[i].length = strlen(static_cast<char *>(prop_values[i].value));
  prop.name = const_cast<char *>(SmDiscardCommand);
  SmcSetProperties(smc_conn, 1, props);
  g_free(prop_values[i].value);


  /* set SmProgram */
  prop.name = const_cast<char *>(SmProgram);
  prop.type = const_cast<char *>(SmARRAY8);
  prop.num_vals = 1;
  prop.vals = prop_values;
  SmcSetProperties(smc_conn, 1, props);


  /* set SmUserID */
  prop.name = const_cast<char *>(SmUserID);
  prop.type = const_cast<char *>(SmARRAY8);
  prop.num_vals = 1;
  prop.vals = prop_values;
  prop_values[0].value = const_cast<char *>(g_get_user_name());
  prop_values[0].length = strlen(static_cast<char *>(prop_values[0].value));
  SmcSetProperties(smc_conn, 1, props);


  /* set SmCurrentDirectory */
  prop.name = const_cast<char *>(SmCurrentDirectory);
  prop.type = const_cast<char *>(SmARRAY8);
  prop.num_vals = 1;
  prop.vals = prop_values;
  prop_values[0].value = g_get_current_dir();
  prop_values[0].length = strlen(static_cast<char *>(prop_values[0].value));
  SmcSetProperties(smc_conn, 1, props);
  g_free(prop_values[0].value);

  g_free(prop_values);

  handler->save_yourself(handler->sm_client_id);

  SmcSaveYourselfDone(smc_conn, True);

}


void
XSMPHandlerPrivate::die_cb(SmcConn smc_conn, SmPointer client_data)
{
  XSMPHandler *handler = static_cast<XSMPHandler *>(client_data);
  XSMPHandlerPrivate *priv = handler->priv;

  SmcCloseConnection(priv->smc_connection, 0, NULL);
  priv->smc_connection = NULL;

  handler->die(handler->sm_client_id);
}


void
XSMPHandlerPrivate::save_complete_cb(SmcConn smc_conn, SmPointer client_data)
{
  XSMPHandler *handler = static_cast<XSMPHandler *>(client_data);
  handler->save_complete(handler->sm_client_id);
}


void
XSMPHandlerPrivate::shutdown_cancelled_cb(SmcConn smc_conn,
					  SmPointer client_data)
{
  XSMPHandler *handler = static_cast<XSMPHandler *>(client_data);
  handler->shutdown_cancelled(handler->sm_client_id);
}


XSMPHandler::XSMPHandler()
  : priv(NULL),
    sm_client_id(NULL),
    previous_sm_client_id(NULL),
    sm_disable(FALSE),
    discard_session(NULL),
    argc(0),
    argv(NULL)
{
}


XSMPHandler::~XSMPHandler()
{
  if (priv != NULL)
    delete priv;
  if (sm_client_id != NULL)
    free(sm_client_id);
}


GOptionGroup *
XSMPHandler::get_option_group()
{
  GOptionEntry option_entries[] = {
    { "sm-client-id", 0, 0, G_OPTION_ARG_STRING, &previous_sm_client_id,
      N_("Specify session management ID"), "ID" },
    { "sm-disable", 0, 0, G_OPTION_ARG_NONE, &sm_disable,
      N_("Disable connection to session manager"), NULL },
    { "discard", 0, 0, G_OPTION_ARG_STRING, &discard_session,
      N_("Discard saved configuration for ID"), "ID" },
    { NULL }
  };

  GOptionGroup *group
    = g_option_group_new("sm",
			 N_("Session Management Options:"),
			 N_("Show session management options"),
			 NULL, NULL);
  g_option_group_add_entries(group, option_entries);
  g_option_group_set_translation_domain(group, PACKAGE_NAME);
  return group;
}


bool
XSMPHandler::process_command_line()
{
  if (discard_session != NULL)
    {
      discard(discard_session);
      return false;
    }

  restore(previous_sm_client_id);
  return true;
}


bool
XSMPHandler::open_session(int argc, char **argv)
{
  this->argc = argc;
  this->argv = argv;

  if (sm_disable)
    return true;

  if (priv != NULL)
    delete priv;

  priv = new XSMPHandlerPrivate(this);
  return priv->open_xsmp_connection();
}


void
XSMPHandler::close_session()
{
  if (priv == NULL)
    return;
  delete priv;
  priv = NULL;
}

}
