/* -*- Mode: C; c-file-style: "gnu"; tab-width: 8 -*- */
/* GTK - The GIMP Toolkit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * Modified by the GTK+ Team and others 1997-2000.  See the AUTHORS
 * file for a list of people on the GTK+ Team.  See the ChangeLog
 * files for a list of changes.  These files are distributed with
 * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
 */

/*
 * This is a modified version of GtkNotebook from GTK+-2.10.6 by the
 * Ochusha Project.  See the AUTHORS file for a list of people on the
 * Ochusha Project.  See the ChangeLog files for a list of changes.
 * These files are distributed with ochusha at
 * http://ochusha.sourceforge.jp/.
 *
 * Since libochushagtk_lgpl only has limited set of modified widgets and
 * is intended to be used with original GTK+, I don't want to cause name
 * space conflictions with original widget sets from GTK+.  Therefore,
 * ALL PUBLICLY AVAILABLE NAMES ARE RENAMED.
 *
 * $Id: mod_notebook.c,v 1.14 2010/08/15 15:38:57 fuyu Exp $
 */

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <stdio.h>

#include "gtkintl.h"
#include "mod_notebook.h"
#include "marshal.h"

#if !GTK_CHECK_VERSION(2,10,0)
#define gdk_atom_intern_static_string(String) gdk_atom_intern((String), TRUE)
#endif

#define DEFAULT_TIMEOUT_INITIAL 200
#define DEFAULT_TIMEOUT_REPEAT   20
#define DEFAULT_TIMEOUT_EXPAND  500

#define SCROLL_DELAY_FACTOR   5
#define SCROLL_THRESHOLD      12
#define DND_THRESHOLD_MULTIPLIER 4
#define FRAMES_PER_SECOND     45
#define MSECS_BETWEEN_UPDATES (1000 / FRAMES_PER_SECOND)

#define ARROW_SIZE                     16
#define CLOSE_BUTTON_SIZE              20
#define TOOLTIPS_STICKY_DELAY          (0)
#define TOOLTIPS_STICKY_REVERT_DELAY   (1000)
#define TOOLTIPS_HIDE_DELAY            (5000)

#if !GLIB_CHECK_VERSION(2,8,0)
#define G_PARAM_STATIC_NAME	G_PARAM_PRIVATE
#define G_PARAM_STATIC_NICK	0
#define G_PARAM_STATIC_BLURB	0
#endif

#define MOD_PARAM_READABLE G_PARAM_READABLE|G_PARAM_STATIC_NAME|G_PARAM_STATIC_NICK|G_PARAM_STATIC_BLURB
#define MOD_PARAM_WRITABLE G_PARAM_WRITABLE|G_PARAM_STATIC_NAME|G_PARAM_STATIC_NICK|G_PARAM_STATIC_BLURB
#define MOD_PARAM_READWRITE G_PARAM_READWRITE|G_PARAM_STATIC_NAME|G_PARAM_STATIC_NICK|G_PARAM_STATIC_BLURB


enum {
  SWITCH_PAGE,
  FOCUS_TAB,
  SELECT_PAGE,
  CHANGE_CURRENT_PAGE,
  MOVE_FOCUS_OUT,
  REORDER_TAB,
  PAGE_REORDERED,
  PAGE_REMOVED,
  PAGE_ADDED,
  CLOSE_BUTTON_CLICKED,
  LAST_SIGNAL
};

enum {
  STEP_PREV,
  STEP_NEXT
};

typedef enum
{
  ARROW_NONE,
  ARROW_LEFT_BEFORE,
  ARROW_RIGHT_BEFORE,
  ARROW_LEFT_AFTER,
  ARROW_RIGHT_AFTER
} ModNotebookArrow;

typedef enum
{
  POINTER_BEFORE,
  POINTER_AFTER,
  POINTER_BETWEEN
} ModNotebookPointerPosition;

typedef enum
{
  DRAG_OPERATION_NONE,
  DRAG_OPERATION_REORDER,
  DRAG_OPERATION_DETACH
} ModNotebookDragOperation;

#define ARROW_IS_LEFT(arrow)  ((arrow) == ARROW_LEFT_BEFORE || (arrow) == ARROW_LEFT_AFTER)
#define ARROW_IS_BEFORE(arrow) ((arrow) == ARROW_LEFT_BEFORE || (arrow) == ARROW_RIGHT_BEFORE)

enum {
  PROP_0,
  PROP_TAB_POS,
  PROP_SHOW_TABS,
  PROP_SHOW_BORDER,
  PROP_SCROLLABLE,
  PROP_TAB_BORDER,
  PROP_TAB_HBORDER,
  PROP_TAB_VBORDER,
  PROP_PAGE,
  PROP_ENABLE_POPUP,
  PROP_GROUP_ID,
  PROP_HOMOGENEOUS,
  PROP_TAB_SHRINKABLE,
  PROP_MINIMUM_TAB_LABEL_SIZE,
  PROP_ENABLE_TOOLTIPS,
  PROP_ENABLE_CLOSE_BUTTON
};

enum {
  CHILD_PROP_0,
  CHILD_PROP_TAB_LABEL,
  CHILD_PROP_MENU_LABEL,
  CHILD_PROP_POSITION,
  CHILD_PROP_TAB_EXPAND,
  CHILD_PROP_TAB_FILL,
  CHILD_PROP_TAB_PACK,
  CHILD_PROP_REORDERABLE,
  CHILD_PROP_DETACHABLE
};

#define MOD_NOTEBOOK_PAGE(_glist_)         ((ModNotebookPage *)((GList *)(_glist_))->data)

/* some useful defines for calculating coords */
#define PAGE_LEFT_X(_page_)   (((ModNotebookPage *) (_page_))->allocation.x)
#define PAGE_RIGHT_X(_page_)  (((ModNotebookPage *) (_page_))->allocation.x + ((ModNotebookPage *) (_page_))->allocation.width)
#define PAGE_MIDDLE_X(_page_) (((ModNotebookPage *) (_page_))->allocation.x + ((ModNotebookPage *) (_page_))->allocation.width / 2)
#define PAGE_TOP_Y(_page_)    (((ModNotebookPage *) (_page_))->allocation.y)
#define PAGE_BOTTOM_Y(_page_) (((ModNotebookPage *) (_page_))->allocation.y + ((ModNotebookPage *) (_page_))->allocation.height)
#define PAGE_MIDDLE_Y(_page_) (((ModNotebookPage *) (_page_))->allocation.y + ((ModNotebookPage *) (_page_))->allocation.height / 2)
#define NOTEBOOK_IS_TAB_LABEL_PARENT(_notebook_,_page_) (((ModNotebookPage *) (_page_))->tab_label->parent == ((GtkWidget *) (_notebook_)))

struct _ModNotebookPage
{
  GtkWidget *child;
  GtkWidget *tab_label;
  GtkWidget *menu_label;
  GtkWidget *last_focus_child;	/* Last descendant of the page that had focus */

  guint default_menu : 1;	/* If true, we create the menu label ourself */
  guint default_tab  : 1;	/* If true, we create the tab label ourself */
  guint expand       : 1;
  guint fill         : 1;
  guint pack         : 1;
  guint reorderable  : 1;
  guint detachable   : 1;

  GtkRequisition requisition;
  GtkAllocation allocation;

  gulong mnemonic_activate_signal;
  gulong notify_visible_handler;
  guint ref_count;
};

#define MOD_NOTEBOOK_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), MOD_TYPE_NOTEBOOK, ModNotebookPrivate))

typedef struct _ModNotebookPrivate ModNotebookPrivate;

struct _ModNotebookPrivate
{
  gint  group_id;
  gint  mouse_x;
  gint  mouse_y;
  gint  pressed_button;
  guint dnd_timer;
  guint switch_tab_timer;

  gint  drag_begin_x;
  gint  drag_begin_y;

  gint  drag_offset_x;
  gint  drag_offset_y;

  GtkWidget *dnd_window;
  GtkTargetList *source_targets;
  ModNotebookDragOperation operation;
  GdkWindow *drag_window;
  gint drag_window_x;
  gint drag_window_y;
  ModNotebookPage *detached_tab;

  guint32 timestamp;

  gboolean during_reorder : 1;
  gboolean during_detach  : 1;
  gboolean has_scrolled   : 1;
};

static const GtkTargetEntry notebook_source_targets [] = {
  { "MOD_NOTEBOOK_TAB", GTK_TARGET_SAME_APP, 0 },
  { "application/x-rootwindow-drop", 0, 1 }
};

static const GtkTargetEntry notebook_dest_targets[] = {
  { "MOD_NOTEBOOK_TAB", GTK_TARGET_SAME_APP, 0 }
};

#ifdef G_DISABLE_CHECKS
#define CHECK_FIND_CHILD(notebook, child)                           \
 mod_notebook_find_child (notebook, child, G_STRLOC)
#else
#define CHECK_FIND_CHILD(notebook, child)                           \
 mod_notebook_find_child (notebook, child, NULL)
#endif
 
/*** ModNotebook Methods ***/
static gboolean mod_notebook_select_page         (ModNotebook      *notebook,
						  gboolean          move_focus);
static gboolean mod_notebook_focus_tab           (ModNotebook      *notebook,
						  GtkNotebookTab    type);
static gboolean mod_notebook_change_current_page (ModNotebook      *notebook,
						  gint              offset);
static void     mod_notebook_move_focus_out      (ModNotebook      *notebook,
						  GtkDirectionType  direction_type);
static gboolean mod_notebook_reorder_tab         (ModNotebook      *notebook,
						  GtkDirectionType  direction_type,
						  gboolean          move_to_last);

/*** GtkObject Methods ***/
static void mod_notebook_destroy             (GtkObject        *object);
static void mod_notebook_set_property	     (GObject         *object,
					      guint            prop_id,
					      const GValue    *value,
					      GParamSpec      *pspec);
static void mod_notebook_get_property	     (GObject         *object,
					      guint            prop_id,
					      GValue          *value,
					      GParamSpec      *pspec);

/*** GtkWidget Methods ***/
static void mod_notebook_map                 (GtkWidget        *widget);
static void mod_notebook_unmap               (GtkWidget        *widget);
static void mod_notebook_realize             (GtkWidget        *widget);
static void mod_notebook_unrealize           (GtkWidget        *widget);
static void mod_notebook_size_request        (GtkWidget        *widget,
					      GtkRequisition   *requisition);
static void mod_notebook_size_allocate       (GtkWidget        *widget,
					      GtkAllocation    *allocation);
static gint mod_notebook_expose              (GtkWidget        *widget,
					      GdkEventExpose   *event);
static gboolean mod_notebook_scroll          (GtkWidget        *widget,
                                              GdkEventScroll   *event);
static gint mod_notebook_button_press        (GtkWidget        *widget,
					      GdkEventButton   *event);
static gint mod_notebook_button_release      (GtkWidget        *widget,
					      GdkEventButton   *event);
static gboolean mod_notebook_popup_menu      (GtkWidget        *widget);
static gint mod_notebook_leave_notify        (GtkWidget        *widget,
					      GdkEventCrossing *event);
static gint mod_notebook_motion_notify       (GtkWidget        *widget,
					      GdkEventMotion   *event);
static gint mod_notebook_focus_in            (GtkWidget        *widget,
					      GdkEventFocus    *event);
static gint mod_notebook_focus_out           (GtkWidget        *widget,
					      GdkEventFocus    *event);
static void mod_notebook_grab_notify         (GtkWidget          *widget,
					      gboolean            was_grabbed);
static void mod_notebook_state_changed       (GtkWidget          *widget,
					      GtkStateType        previous_state);
static void mod_notebook_draw_focus          (GtkWidget        *widget);
static gint mod_notebook_focus               (GtkWidget        *widget,
					      GtkDirectionType  direction);
static void mod_notebook_style_set           (GtkWidget        *widget,
					      GtkStyle         *previous);

/*** Drag and drop Methods ***/
#if GTK_CHECK_VERSION(2,10,0)
static void mod_notebook_drag_begin          (GtkWidget        *widget,
					      GdkDragContext   *context);
static void mod_notebook_drag_end            (GtkWidget        *widget,
					      GdkDragContext   *context);
static gboolean mod_notebook_drag_motion     (GtkWidget        *widget,
					      GdkDragContext   *context,
					      gint              x,
					      gint              y,
					      guint             time);
static void mod_notebook_drag_leave          (GtkWidget        *widget,
					      GdkDragContext   *context,
					      guint             time);
static gboolean mod_notebook_drag_drop       (GtkWidget        *widget,
					      GdkDragContext   *context,
					      gint              x,
					      gint              y,
					      guint             time);
static void mod_notebook_drag_data_get       (GtkWidget        *widget,
					      GdkDragContext   *context,
					      GtkSelectionData *data,
					      guint             info,
					      guint             time);
static void mod_notebook_drag_data_received  (GtkWidget        *widget,
					      GdkDragContext   *context,
					      gint              x,
					      gint              y,
					      GtkSelectionData *data,
					      guint             info,
					      guint             time);
#endif

/*** GtkContainer Methods ***/
static void mod_notebook_set_child_property  (GtkContainer     *container,
					      GtkWidget        *child,
					      guint             property_id,
					      const GValue     *value,
					      GParamSpec       *pspec);
static void mod_notebook_get_child_property  (GtkContainer     *container,
					      GtkWidget        *child,
					      guint             property_id,
					      GValue           *value,
					      GParamSpec       *pspec);
static void mod_notebook_add                 (GtkContainer     *container,
					      GtkWidget        *widget);
static void mod_notebook_remove              (GtkContainer     *container,
					      GtkWidget        *widget);
static void mod_notebook_set_focus_child     (GtkContainer     *container,
					      GtkWidget        *child);
static GType mod_notebook_child_type       (GtkContainer     *container);
static void mod_notebook_forall              (GtkContainer     *container,
					      gboolean		include_internals,
					      GtkCallback       callback,
					      gpointer          callback_data);

/*** ModNotebook Methods ***/
static gint mod_notebook_real_insert_page    (ModNotebook      *notebook,
					      GtkWidget        *child,
					      GtkWidget        *tab_label,
					      GtkWidget        *menu_label,
					      gint              position);

/*** ModNotebook Private Functions ***/
static void mod_notebook_redraw_tabs         (ModNotebook      *notebook);
static void mod_notebook_redraw_arrows       (ModNotebook      *notebook);
static void mod_notebook_redraw_close_button (ModNotebook      *notebook);
static void mod_notebook_real_remove         (ModNotebook      *notebook,
					      GList            *list);
static void mod_notebook_update_labels       (ModNotebook      *notebook);
static gint mod_notebook_timer               (ModNotebook      *notebook);
static void mod_notebook_set_scroll_timer    (ModNotebook *notebook);
static gint mod_notebook_page_compare        (gconstpointer     a,
					      gconstpointer     b);
static GList* mod_notebook_find_child        (ModNotebook      *notebook,
					      GtkWidget        *child,
					      const gchar      *function);
static gint  mod_notebook_real_page_position (ModNotebook      *notebook,
					      GList            *list);
static GList * mod_notebook_search_page      (ModNotebook      *notebook,
					      GList            *list,
					      gint              direction,
					      gboolean          find_visible);
static gboolean mod_notebook_in_close_button (ModNotebook      *notebook,
					      gint              x,
					      gint              y);
static void mod_notebook_page_ref            (ModNotebookPage  *page);
static void mod_notebook_page_unref          (ModNotebookPage  *page);

/*** ModNotebook Drawing Functions ***/
static void mod_notebook_paint               (GtkWidget        *widget,
					      GdkRectangle     *area);
static void mod_notebook_draw_tab            (ModNotebook      *notebook,
					      ModNotebookPage  *page,
					      GdkRectangle     *area);
static void mod_notebook_draw_arrow          (ModNotebook      *notebook,
					      ModNotebookArrow  arrow);
static void mod_notebook_draw_close_button   (ModNotebook      *notebook);

/*** ModNotebook Size Allocate Functions ***/
static void mod_notebook_pages_allocate      (ModNotebook      *notebook);
static void mod_notebook_page_allocate       (ModNotebook      *notebook,
					      ModNotebookPage  *page);
static void mod_notebook_calc_tabs           (ModNotebook      *notebook,
			                      GList            *start,
					      GList           **end,
					      gint             *tab_space,
					      gint              typical_tab_size,
					      guint             direction);

/*** ModNotebook Page Switch Methods ***/
static void mod_notebook_real_switch_page    (ModNotebook      *notebook,
					      ModNotebookPage  *page,
					      guint             page_num);

/*** ModNotebook Page Switch Functions ***/
static void mod_notebook_switch_page         (ModNotebook      *notebook,
					      ModNotebookPage  *page,
					      gint              page_num);
static gint mod_notebook_page_select         (ModNotebook      *notebook,
					      gboolean          move_focus);
static void mod_notebook_switch_focus_tab    (ModNotebook      *notebook,
                                              GList            *new_child);
static void mod_notebook_menu_switch_page    (GtkWidget        *widget,
					      ModNotebookPage  *page);

/*** ModNotebook Menu Functions ***/
static void mod_notebook_menu_item_create    (ModNotebook      *notebook,
					      GList            *list);
static void mod_notebook_menu_label_unparent (GtkWidget        *widget,
					      gpointer          data);
static void mod_notebook_menu_detacher       (GtkWidget        *widget,
					      GtkMenu          *menu);

/*** ModNotebook Private Setters ***/
static void mod_notebook_set_homogeneous_tabs_internal (ModNotebook *notebook,
							gboolean     homogeneous);
static void mod_notebook_set_tab_border_internal       (ModNotebook *notebook,
							guint        border_width);
static void mod_notebook_set_tab_hborder_internal      (ModNotebook *notebook,
							guint        tab_hborder);
static void mod_notebook_set_tab_vborder_internal      (ModNotebook *notebook,
							guint        tab_vborder);

static void mod_notebook_update_tab_states             (ModNotebook *notebook);
static gboolean mod_notebook_mnemonic_activate_switch_page (GtkWidget *child,
							    gboolean overload,
							    gpointer data);

static gboolean focus_tabs_in  (ModNotebook      *notebook);
static gboolean focus_child_in (ModNotebook      *notebook,
				GtkDirectionType  direction);

static void stop_scrolling (ModNotebook *notebook);


static void mod_notebook_tooltips_set_active_page    (ModNotebook *notebook,
						      ModNotebookPage *page);
static void mod_notebook_tooltips_start_delay        (ModNotebook *notebook,
					              ModNotebookPage *page);
static gboolean mod_notebook_tooltips_recently_shown (ModNotebook *notebook);
static gboolean mod_notebook_tooltips_timeout        (gpointer data);
static gboolean mod_notebook_tooltips_hide_timeout   (gpointer data);
static void mod_notebook_tooltips_draw_tips          (ModNotebook *notebook);
static void mod_notebook_tooltips_force_window       (ModNotebook *notebook);
static gboolean mod_notebook_tooltips_paint_window   (ModNotebook *notebook);
static void mod_notebook_tooltips_update_screen      (ModNotebook *notebook,
						      gboolean     new_window);
static gboolean get_keyboard_mode                    (GtkWidget   *widget);
static void disconnect_tip_window_display_closed     (ModNotebook *notebook);
static void tip_window_display_closed                (GdkDisplay  *display,
						      gboolean     was_error,
						      ModNotebook *notebook);
static void mod_notebook_tooltips_unset_tip_window   (ModNotebook *notebook);

static ModNotebookPage *mod_notebook_get_tab_page    (ModNotebook *notebook,
						      gint x, gint y);


static ModNotebookWindowCreationFunc window_creation_hook = NULL;
static gpointer window_creation_hook_data;
static GDestroyNotify window_creation_hook_destroy = NULL;

static guint notebook_signals[LAST_SIGNAL] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

G_DEFINE_TYPE (ModNotebook, mod_notebook, GTK_TYPE_CONTAINER)

static void
add_tab_bindings (GtkBindingSet    *binding_set,
		  GdkModifierType   modifiers,
		  GtkDirectionType  direction)
{
  gtk_binding_entry_add_signal (binding_set, GDK_Tab, modifiers,
                                "move_focus_out", 1,
                                GTK_TYPE_DIRECTION_TYPE, direction);
  gtk_binding_entry_add_signal (binding_set, GDK_KP_Tab, modifiers,
                                "move_focus_out", 1,
                                GTK_TYPE_DIRECTION_TYPE, direction);
}

static void
add_arrow_bindings (GtkBindingSet    *binding_set,
		    guint             keysym,
		    GtkDirectionType  direction)
{
  guint keypad_keysym = keysym - GDK_Left + GDK_KP_Left;
  
  gtk_binding_entry_add_signal (binding_set, keysym, GDK_CONTROL_MASK,
                                "move_focus_out", 1,
                                GTK_TYPE_DIRECTION_TYPE, direction);
  gtk_binding_entry_add_signal (binding_set, keypad_keysym, GDK_CONTROL_MASK,
                                "move_focus_out", 1,
                                GTK_TYPE_DIRECTION_TYPE, direction);
}

static void
add_reorder_bindings (GtkBindingSet    *binding_set,
		      guint             keysym,
		      GtkDirectionType  direction,
		      gboolean          move_to_last)
{
  guint keypad_keysym = keysym - GDK_Left + GDK_KP_Left;

  gtk_binding_entry_add_signal (binding_set, keysym, GDK_MOD1_MASK,
				"reorder_tab", 2,
				GTK_TYPE_DIRECTION_TYPE, direction,
				G_TYPE_BOOLEAN, move_to_last);
  gtk_binding_entry_add_signal (binding_set, keypad_keysym, GDK_MOD1_MASK,
				"reorder_tab", 2,
				GTK_TYPE_DIRECTION_TYPE, direction,
				G_TYPE_BOOLEAN, move_to_last);
}

static void
mod_notebook_class_init (ModNotebookClass *class)
{
  GObjectClass   *gobject_class = G_OBJECT_CLASS (class);
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (class);
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (class);
  GtkContainerClass *container_class = GTK_CONTAINER_CLASS (class);
  GtkBindingSet *binding_set;
  
  gobject_class->set_property = mod_notebook_set_property;
  gobject_class->get_property = mod_notebook_get_property;
  object_class->destroy = mod_notebook_destroy;

  widget_class->map = mod_notebook_map;
  widget_class->unmap = mod_notebook_unmap;
  widget_class->realize = mod_notebook_realize;
  widget_class->unrealize = mod_notebook_unrealize;
  widget_class->size_request = mod_notebook_size_request;
  widget_class->size_allocate = mod_notebook_size_allocate;
  widget_class->expose_event = mod_notebook_expose;
  widget_class->scroll_event = mod_notebook_scroll;
  widget_class->button_press_event = mod_notebook_button_press;
  widget_class->button_release_event = mod_notebook_button_release;
  widget_class->popup_menu = mod_notebook_popup_menu;
  widget_class->leave_notify_event = mod_notebook_leave_notify;
  widget_class->motion_notify_event = mod_notebook_motion_notify;
  widget_class->grab_notify = mod_notebook_grab_notify;
  widget_class->state_changed = mod_notebook_state_changed;
  widget_class->focus_in_event = mod_notebook_focus_in;
  widget_class->focus_out_event = mod_notebook_focus_out;
  widget_class->focus = mod_notebook_focus;
  widget_class->style_set = mod_notebook_style_set;

#if GTK_CHECK_VERSION(2,10,0)
  widget_class->drag_begin = mod_notebook_drag_begin;
  widget_class->drag_end = mod_notebook_drag_end;
  widget_class->drag_motion = mod_notebook_drag_motion;
  widget_class->drag_leave = mod_notebook_drag_leave;
  widget_class->drag_drop = mod_notebook_drag_drop;
  widget_class->drag_data_get = mod_notebook_drag_data_get;
  widget_class->drag_data_received = mod_notebook_drag_data_received;
#endif

  container_class->add = mod_notebook_add;
  container_class->remove = mod_notebook_remove;
  container_class->forall = mod_notebook_forall;
  container_class->set_focus_child = mod_notebook_set_focus_child;
  container_class->get_child_property = mod_notebook_get_child_property;
  container_class->set_child_property = mod_notebook_set_child_property;
  container_class->child_type = mod_notebook_child_type;

  class->switch_page = mod_notebook_real_switch_page;
  class->insert_page = mod_notebook_real_insert_page;

  class->focus_tab = mod_notebook_focus_tab;
  class->select_page = mod_notebook_select_page;
  class->change_current_page = mod_notebook_change_current_page;
  class->move_focus_out = mod_notebook_move_focus_out;
  class->reorder_tab = mod_notebook_reorder_tab;
  class->close_button_clicked = NULL;
  
  g_object_class_install_property (gobject_class,
				   PROP_PAGE,
				   g_param_spec_int ("page",
 						     P_("Page"),
 						     P_("The index of the current page"),
 						     0,
 						     G_MAXINT,
 						     0,
 						     MOD_PARAM_READWRITE));
  g_object_class_install_property (gobject_class,
				   PROP_TAB_POS,
				   g_param_spec_enum ("tab-pos",
 						      P_("Tab Position"),
 						      P_("Which side of the notebook holds the tabs"),
 						      GTK_TYPE_POSITION_TYPE,
 						      GTK_POS_TOP,
 						      MOD_PARAM_READWRITE));
  g_object_class_install_property (gobject_class,
				   PROP_TAB_BORDER,
				   g_param_spec_uint ("tab-border",
 						      P_("Tab Border"),
 						      P_("Width of the border around the tab labels"),
 						      0,
 						      G_MAXUINT,
 						      2,
 						      MOD_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class,
				   PROP_TAB_HBORDER,
				   g_param_spec_uint ("tab-hborder",
 						      P_("Horizontal Tab Border"),
 						      P_("Width of the horizontal border of tab labels"),
 						      0,
 						      G_MAXUINT,
 						      2,
 						      MOD_PARAM_READWRITE));
  g_object_class_install_property (gobject_class,
				   PROP_TAB_VBORDER,
				   g_param_spec_uint ("tab-vborder",
 						      P_("Vertical Tab Border"),
 						      P_("Width of the vertical border of tab labels"),
 						      0,
 						      G_MAXUINT,
 						      2,
 						      MOD_PARAM_READWRITE));
  g_object_class_install_property (gobject_class,
				   PROP_SHOW_TABS,
				   g_param_spec_boolean ("show-tabs",
 							 P_("Show Tabs"),
 							 P_("Whether tabs should be shown or not"),
 							 TRUE,
 							 MOD_PARAM_READWRITE));
  g_object_class_install_property (gobject_class,
				   PROP_SHOW_BORDER,
				   g_param_spec_boolean ("show-border",
 							 P_("Show Border"),
 							 P_("Whether the border should be shown or not"),
 							 TRUE,
 							 MOD_PARAM_READWRITE));
  g_object_class_install_property (gobject_class,
				   PROP_SCROLLABLE,
				   g_param_spec_boolean ("scrollable",
 							 P_("Scrollable"),
 							 P_("If TRUE, scroll arrows are added if there are too many tabs to fit"),
 							 FALSE,
 							 MOD_PARAM_READWRITE));
  g_object_class_install_property (gobject_class,
				   PROP_ENABLE_POPUP,
				   g_param_spec_boolean ("enable-popup",
 							 P_("Enable Popup"),
 							 P_("If TRUE, pressing the right mouse button on the notebook pops up a menu that you can use to go to a page"),
 							 FALSE,
 							 MOD_PARAM_READWRITE));
  g_object_class_install_property (gobject_class,
				   PROP_HOMOGENEOUS,
				   g_param_spec_boolean ("homogeneous",
 							 P_("Homogeneous"),
 							 P_("Whether tabs should have homogeneous sizes"),
 							 FALSE,
							 MOD_PARAM_READWRITE));
  g_object_class_install_property (gobject_class,
				   PROP_GROUP_ID,
				   g_param_spec_int ("group-id",
						     P_("Group ID"),
						     P_("Group ID for tabs drag and drop"),
						     -1,
						     G_MAXINT,
						     -1,
						     MOD_PARAM_READWRITE));
  g_object_class_install_property (gobject_class,
				   PROP_MINIMUM_TAB_LABEL_SIZE,
				   g_param_spec_uint ("minimum-tab-label-size",
 						      P_("Minimum Size of Tab Labels"),
 						      P_("Minimum size of tab labels that specifies the minimum size of each tab labels if they are shrinked"),
 						      1,
 						      G_MAXUINT,
 						      ARROW_SIZE,
 						      MOD_PARAM_READWRITE));
  g_object_class_install_property (gobject_class,
				   PROP_TAB_SHRINKABLE,
				   g_param_spec_boolean ("tab-shrinkable",
							 P_("Tab Shrinkable"),
							 P_("Whether tabs are allowed to be shrinked if there are too many tabs to fit"),
							 FALSE,
							 MOD_PARAM_READWRITE));
  g_object_class_install_property (gobject_class,
				   PROP_ENABLE_TOOLTIPS,
				   g_param_spec_boolean ("enable-tooltips",
							 P_("Enable Tooltips"),
							 P_("If TRUE, tooltip will be shown when pointer is placed on a tab label"),
							 FALSE,
							 MOD_PARAM_READWRITE));
  g_object_class_install_property (gobject_class,
				   PROP_ENABLE_CLOSE_BUTTON,
				   g_param_spec_boolean ("enable-close-button",
							 P_("Enable Close Button"),
							 P_("If TRUE, close button will be shown when tab bookmarks are shown"),
							 FALSE,
							 MOD_PARAM_READWRITE));

  gtk_container_class_install_child_property (container_class,
					      CHILD_PROP_TAB_LABEL,
					      g_param_spec_string ("tab-label", 
								   P_("Tab label"),
								   P_("The string displayed on the child's tab label"),
								   NULL,
								   MOD_PARAM_READWRITE));
  gtk_container_class_install_child_property (container_class,
					      CHILD_PROP_MENU_LABEL,
					      g_param_spec_string ("menu-label", 
								   P_("Menu label"), 
								   P_("The string displayed in the child's menu entry"),
								   NULL,
								   MOD_PARAM_READWRITE));
  gtk_container_class_install_child_property (container_class,
					      CHILD_PROP_POSITION,
					      g_param_spec_int ("position", 
								P_("Position"), 
								P_("The index of the child in the parent"),
								-1, G_MAXINT, 0,
								MOD_PARAM_READWRITE));
  gtk_container_class_install_child_property (container_class,
					      CHILD_PROP_TAB_EXPAND,
					      g_param_spec_boolean ("tab-expand", 
								    P_("Tab expand"), 
								    P_("Whether to expand the child's tab or not"),
								    TRUE,
								    MOD_PARAM_READWRITE));
  gtk_container_class_install_child_property (container_class,
					      CHILD_PROP_TAB_FILL,
					      g_param_spec_boolean ("tab-fill", 
								    P_("Tab fill"), 
								    P_("Whether the child's tab should fill the allocated area or not"),
								    TRUE,
								    MOD_PARAM_READWRITE));
  gtk_container_class_install_child_property (container_class,
					      CHILD_PROP_TAB_PACK,
					      g_param_spec_enum ("tab-pack", 
								 P_("Tab pack type"),
								 P_("A GtkPackType indicating whether the child is packed with reference to the start or end of the parent"),
								 GTK_TYPE_PACK_TYPE, GTK_PACK_START,
								 MOD_PARAM_READWRITE));
  gtk_container_class_install_child_property (container_class,
					      CHILD_PROP_REORDERABLE,
					      g_param_spec_boolean ("reorderable",
								    P_("Tab reorderable"),
								    P_("Whether the tab is reorderable by user action or not"),
								    FALSE,
								    MOD_PARAM_READWRITE));
  gtk_container_class_install_child_property (container_class,
					      CHILD_PROP_DETACHABLE,
					      g_param_spec_boolean ("detachable",
								    P_("Tab detachable"),
								    P_("Whether the tab is detachable"),
								    FALSE,
								    MOD_PARAM_READWRITE));

/**
 * ModNotebook:has-secondary-backward-stepper:
 *
 * The "has-secondary-backward-stepper" property determines whether 
 * a second backward arrow button is displayed on the opposite end 
 * of the tab area.
 *
 * Since: 2.4
 */  
  gtk_widget_class_install_style_property (widget_class,
					   g_param_spec_boolean ("has-secondary-backward-stepper",
								 P_("Secondary backward stepper"),
								 P_("Display a second backward arrow button on the opposite end of the tab area"),
								 FALSE,
								 MOD_PARAM_READABLE));

/**
 * ModNotebook:has-secondary-forward-stepper:
 *
 * The "has-secondary-forward-stepper" property determines whether 
 * a second forward arrow button is displayed on the opposite end 
 * of the tab area.
 *
 * Since: 2.4
 */  
  gtk_widget_class_install_style_property (widget_class,
					   g_param_spec_boolean ("has-secondary-forward-stepper",
								 P_("Secondary forward stepper"),
								 P_("Display a second forward arrow button on the opposite end of the tab area"),
								 FALSE,
								 MOD_PARAM_READABLE));

/**
 * ModNotebook:has-backward-stepper:
 *
 * The "has-backward-stepper" property determines whether 
 * the standard backward arrow button is displayed.
 *
 * Since: 2.4
 */  
  gtk_widget_class_install_style_property (widget_class,
					   g_param_spec_boolean ("has-backward-stepper",
								 P_("Backward stepper"),
								 P_("Display the standard backward arrow button"),
								 TRUE,
								 MOD_PARAM_READABLE));

/**
 * ModNotebook:has-forward-stepper:
 *
 * The "has-forward-stepper" property determines whether 
 * the standard forward arrow button is displayed.
 *
 * Since: 2.4
 */  
  gtk_widget_class_install_style_property (widget_class,
					   g_param_spec_boolean ("has-forward-stepper",
								 P_("Forward stepper"),
								 P_("Display the standard forward arrow button"),
								 TRUE,
								 MOD_PARAM_READABLE));
  
/**
 * ModNotebook:tab-overlap:
 *
 * The "tab-overlap" property defines size of tab overlap
 * area.
 *
 * Since: 2.10
 */  
  gtk_widget_class_install_style_property (widget_class,
					   g_param_spec_int ("tab-overlap",
							     P_("Tab overlap"),
							     P_("Size of tab overlap area"),
							     G_MININT,
							     G_MAXINT,
							     2,
							     MOD_PARAM_READABLE));

/**
 * ModNotebook:tab-curvature:
 *
 * The "tab-curvature" property defines size of tab curvature.
 *
 * Since: 2.10
 */  
  gtk_widget_class_install_style_property (widget_class,
					   g_param_spec_int ("tab-curvature",
							     P_("Tab curvature"),
							     P_("Size of tab curvature"),
							     0,
							     G_MAXINT,
							     1,
							     MOD_PARAM_READABLE));

  /**
   * ModNotebook:arrow-spacing:
   *
   * The "arrow-size" property defines the spacing between the scroll
   * arrows and the tabs.
   *
   * Since: 2.10
   */
  gtk_widget_class_install_style_property (widget_class,
                                           g_param_spec_int ("arrow-spacing",
                                                             _("Arrow spacing"),
                                                             _("Scroll arrow spacing"),
                                                             0,
                                                             G_MAXINT,
                                                             0,
                                                             MOD_PARAM_READABLE));

  /**
   * ModNotebook:close-button-size:
   *
   * The "close-button-size" property defines the size of close button.
   */
  gtk_widget_class_install_style_property (widget_class,
                                           g_param_spec_int ("close-button-size",
                                                             _("Close button size"),
                                                             _("Size of close button"),
                                                             1,
                                                             G_MAXINT,
                                                             CLOSE_BUTTON_SIZE,
                                                             MOD_PARAM_READABLE));

  /**
   * ModNotebook:close-button-spacing:
   *
   * The "close-button-spacing" property defines the spacing between the
   * close button and the scroll arrows/the tabs.
   */
  gtk_widget_class_install_style_property (widget_class,
                                           g_param_spec_int ("close-button-spacing",
                                                             _("Close button spacing"),
                                                             _("Spacing around the close button"),
                                                             0,
                                                             G_MAXINT,
                                                             2,
                                                             MOD_PARAM_READABLE));

  if (gtk_major_version == 2 && gtk_minor_version <= 8)
    {
      /*
       * Some style properties of GtkWidget used by this widget implementation
       * only exist in gtk+-2.10 or later.  We define them here just for
       * simplification on such older gtk+.
       */

      /**
       * GtkWidget:scroll-arrow-hlength:
       *
       * The "scroll-arrow-hlength" property defines the length of 
       * horizontal scroll arrows.
       *
       * Since: 2.10
       */
      gtk_widget_class_install_style_property (widget_class,
					       g_param_spec_int ("scroll-arrow-hlength",
								 P_("Horizontal Scroll Arrow Length"),
								 P_("The length of horizontal scroll arrows"),
								 1, G_MAXINT, ARROW_SIZE,
								 MOD_PARAM_READABLE));

      /**
       * GtkWidget:scroll-arrow-vlength:
       *
       * The "scroll-arrow-vlength" property defines the length of 
       * vertical scroll arrows.
       *
       * Since: 2.10
       */
      gtk_widget_class_install_style_property (widget_class,
					       g_param_spec_int ("scroll-arrow-vlength",
								 P_("Vertical Scroll Arrow Length"),
								 P_("The length of vertical scroll arrows"),
								 1, G_MAXINT, ARROW_SIZE,
								 MOD_PARAM_READABLE));
    }

  notebook_signals[SWITCH_PAGE] =
    g_signal_new (I_("switch_page"),
		  G_TYPE_FROM_CLASS (gobject_class),
		  G_SIGNAL_RUN_LAST,
		  G_STRUCT_OFFSET (ModNotebookClass, switch_page),
		  NULL, NULL,
		  modgtk2_marshal_VOID__POINTER_UINT,
		  G_TYPE_NONE, 2,
		  G_TYPE_POINTER,
		  G_TYPE_UINT);
  notebook_signals[FOCUS_TAB] = 
    g_signal_new (I_("focus_tab"),
                  G_TYPE_FROM_CLASS (gobject_class),
                  G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
                  G_STRUCT_OFFSET (ModNotebookClass, focus_tab),
                  NULL, NULL,
                  modgtk2_marshal_BOOLEAN__ENUM,
                  G_TYPE_BOOLEAN, 1,
                  GTK_TYPE_NOTEBOOK_TAB);
  notebook_signals[SELECT_PAGE] = 
    g_signal_new (I_("select_page"),
                  G_TYPE_FROM_CLASS (gobject_class),
                  G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
                  G_STRUCT_OFFSET (ModNotebookClass, select_page),
                  NULL, NULL,
                  modgtk2_marshal_BOOLEAN__BOOLEAN,
                  G_TYPE_BOOLEAN, 1,
                  G_TYPE_BOOLEAN);
  notebook_signals[CHANGE_CURRENT_PAGE] = 
    g_signal_new (I_("change_current_page"),
                  G_TYPE_FROM_CLASS (gobject_class),
                  G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
                  G_STRUCT_OFFSET (ModNotebookClass, change_current_page),
                  NULL, NULL,
                  modgtk2_marshal_BOOLEAN__INT,
                  G_TYPE_NONE, 1,
                  G_TYPE_INT);
  notebook_signals[MOVE_FOCUS_OUT] =
    g_signal_new (I_("move_focus_out"),
                  G_TYPE_FROM_CLASS (gobject_class),
                  G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
                  G_STRUCT_OFFSET (ModNotebookClass, move_focus_out),
                  NULL, NULL,
                  modgtk2_marshal_VOID__ENUM,
                  G_TYPE_NONE, 1,
                  GTK_TYPE_DIRECTION_TYPE);
  notebook_signals[REORDER_TAB] =
    g_signal_new (I_("reorder_tab"),
                  G_TYPE_FROM_CLASS (gobject_class),
                  G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
                  G_STRUCT_OFFSET (ModNotebookClass, reorder_tab),
                  NULL, NULL,
                  modgtk2_marshal_BOOLEAN__ENUM_BOOLEAN,
                  G_TYPE_NONE, 2,
                  GTK_TYPE_DIRECTION_TYPE,
		  G_TYPE_BOOLEAN);
  /**
   * ModNotebook::page-reordered:
   * @notebook: the #ModNotebook
   * @child: the child #GtkWidget affected
   * @page_num: the new page number for @child
   *
   * the ::page-reordered signal is emitted in the notebook
   * right after a page has been reordered.
   *
   * Since: 2.10
   **/
  notebook_signals[PAGE_REORDERED] =
    g_signal_new (I_("page_reordered"),
                  G_TYPE_FROM_CLASS (gobject_class),
                  G_SIGNAL_RUN_LAST,
                  0, NULL, NULL,
		  modgtk2_marshal_VOID__OBJECT_UINT,
                  G_TYPE_NONE, 2,
		  GTK_TYPE_WIDGET,
		  G_TYPE_UINT);
  /**
   * ModNotebook::page-removed:
   * @notebook: the #ModNotebook
   * @child: the child #GtkWidget affected
   * @page_num: the @child page number
   *
   * the ::page-removed signal is emitted in the notebook
   * right after a page is removed from the notebook.
   *
   * Since: 2.10
   **/
  notebook_signals[PAGE_REMOVED] =
    g_signal_new (I_("page_removed"),
                  G_TYPE_FROM_CLASS (gobject_class),
                  G_SIGNAL_RUN_LAST,
                  0, NULL, NULL,
		  modgtk2_marshal_VOID__OBJECT_UINT,
                  G_TYPE_NONE, 2,
		  GTK_TYPE_WIDGET,
		  G_TYPE_UINT);
  /**
   * ModNotebook::page-added:
   * @notebook: the #ModNotebook
   * @child: the child #GtkWidget affected
   * @page_num: the new page number for @child
   *
   * the ::page-added signal is emitted in the notebook
   * right after a page is added to the notebook.
   *
   * Since: 2.10
   **/
  notebook_signals[PAGE_ADDED] =
    g_signal_new (I_("page_added"),
                  G_TYPE_FROM_CLASS (gobject_class),
                  G_SIGNAL_RUN_LAST,
                  0, NULL, NULL,
		  modgtk2_marshal_VOID__OBJECT_UINT,
                  G_TYPE_NONE, 2,
		  GTK_TYPE_WIDGET,
		  G_TYPE_UINT);

  notebook_signals[CLOSE_BUTTON_CLICKED] =
    g_signal_new (I_("close_button_clicked"),
		  G_TYPE_FROM_CLASS (gobject_class),
		  G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
		  G_STRUCT_OFFSET (ModNotebookClass, close_button_clicked),
		  NULL, NULL,
		  modgtk2_marshal_VOID__VOID,
		  G_TYPE_NONE, 0);

  binding_set = gtk_binding_set_by_class (class);
  gtk_binding_entry_add_signal (binding_set,
                                GDK_space, 0,
                                "select_page", 1, 
                                G_TYPE_BOOLEAN, FALSE);
  gtk_binding_entry_add_signal (binding_set,
                                GDK_KP_Space, 0,
                                "select_page", 1, 
                                G_TYPE_BOOLEAN, FALSE);
  
  gtk_binding_entry_add_signal (binding_set,
                                GDK_Home, 0,
                                "focus_tab", 1, 
                                GTK_TYPE_NOTEBOOK_TAB, GTK_NOTEBOOK_TAB_FIRST);
  gtk_binding_entry_add_signal (binding_set,
                                GDK_KP_Home, 0,
                                "focus_tab", 1, 
                                GTK_TYPE_NOTEBOOK_TAB, GTK_NOTEBOOK_TAB_FIRST);
  gtk_binding_entry_add_signal (binding_set,
                                GDK_End, 0,
                                "focus_tab", 1, 
                                GTK_TYPE_NOTEBOOK_TAB, GTK_NOTEBOOK_TAB_LAST);
  gtk_binding_entry_add_signal (binding_set,
                                GDK_KP_End, 0,
                                "focus_tab", 1, 
                                GTK_TYPE_NOTEBOOK_TAB, GTK_NOTEBOOK_TAB_LAST);

  gtk_binding_entry_add_signal (binding_set,
                                GDK_Page_Up, GDK_CONTROL_MASK,
                                "change_current_page", 1,
                                G_TYPE_INT, -1);
  gtk_binding_entry_add_signal (binding_set,
                                GDK_Page_Down, GDK_CONTROL_MASK,
                                "change_current_page", 1,
                                G_TYPE_INT, 1);

  gtk_binding_entry_add_signal (binding_set,
                                GDK_Page_Up, GDK_CONTROL_MASK | GDK_MOD1_MASK,
                                "change_current_page", 1,
                                G_TYPE_INT, -1);
  gtk_binding_entry_add_signal (binding_set,
                                GDK_Page_Down, GDK_CONTROL_MASK | GDK_MOD1_MASK,
                                "change_current_page", 1,
                                G_TYPE_INT, 1);

  add_arrow_bindings (binding_set, GDK_Up, GTK_DIR_UP);
  add_arrow_bindings (binding_set, GDK_Down, GTK_DIR_DOWN);
  add_arrow_bindings (binding_set, GDK_Left, GTK_DIR_LEFT);
  add_arrow_bindings (binding_set, GDK_Right, GTK_DIR_RIGHT);

  add_reorder_bindings (binding_set, GDK_Up, GTK_DIR_UP, FALSE);
  add_reorder_bindings (binding_set, GDK_Down, GTK_DIR_DOWN, FALSE);
  add_reorder_bindings (binding_set, GDK_Left, GTK_DIR_LEFT, FALSE);
  add_reorder_bindings (binding_set, GDK_Right, GTK_DIR_RIGHT, FALSE);
  add_reorder_bindings (binding_set, GDK_Home, GTK_DIR_LEFT, TRUE);
  add_reorder_bindings (binding_set, GDK_Home, GTK_DIR_UP, TRUE);
  add_reorder_bindings (binding_set, GDK_End, GTK_DIR_RIGHT, TRUE);
  add_reorder_bindings (binding_set, GDK_End, GTK_DIR_DOWN, TRUE);

  add_tab_bindings (binding_set, GDK_CONTROL_MASK, GTK_DIR_TAB_FORWARD);
  add_tab_bindings (binding_set, GDK_CONTROL_MASK | GDK_SHIFT_MASK, GTK_DIR_TAB_BACKWARD);

  g_type_class_add_private (class, sizeof (ModNotebookPrivate));
}

static void
mod_notebook_init (ModNotebook *notebook)
{
  ModNotebookPrivate *priv;

#if GTK_CHECK_VERSION(2,18,0)
  gtk_widget_set_can_focus (GTK_WIDGET (notebook), TRUE);
  gtk_widget_set_has_window (GTK_WIDGET (notebook), FALSE);
#else
  GTK_WIDGET_SET_FLAGS (notebook, GTK_CAN_FOCUS);
  GTK_WIDGET_SET_FLAGS (notebook, GTK_NO_WINDOW);
#endif

  priv = MOD_NOTEBOOK_GET_PRIVATE (notebook);

  notebook->cur_page = NULL;
  notebook->children = NULL;
  notebook->first_tab = NULL;
  notebook->focus_tab = NULL;
  notebook->event_window = NULL;
  notebook->menu = NULL;
  notebook->active_tips_page = NULL;
  notebook->tip_window = NULL;
  notebook->tip_label = NULL;
  notebook->tip_timer = 0;

  notebook->tab_hborder = 2;
  notebook->tab_vborder = 2;

  notebook->minimum_tab_label_size = ARROW_SIZE;

  notebook->show_tabs = TRUE;
  notebook->show_border = TRUE;
  notebook->tab_pos = GTK_POS_TOP;
  notebook->scrollable = FALSE;

  notebook->tab_shrinkable = FALSE;
  notebook->enable_tooltips = FALSE;
  notebook->enable_close_button = FALSE;
  notebook->in_close_button = FALSE;
  notebook->click_close_button = FALSE;

  notebook->in_child = 0;
  notebook->click_child = 0;
  notebook->button = 0;
  notebook->need_timer = 0;
  notebook->child_has_focus = FALSE;
  notebook->have_visible_child = FALSE;
  notebook->focus_out = FALSE;

  notebook->has_before_previous = 1;
  notebook->has_before_next     = 0;
  notebook->has_after_previous  = 0;
  notebook->has_after_next      = 1;

  notebook->delay = 100;

  priv->group_id = -1;
  priv->pressed_button = -1;
  priv->dnd_timer = 0;
  priv->switch_tab_timer = 0;
  priv->source_targets = gtk_target_list_new (notebook_source_targets,
					      G_N_ELEMENTS (notebook_source_targets));
  priv->operation = DRAG_OPERATION_NONE;
  priv->detached_tab = NULL;
  priv->during_detach = FALSE;
  priv->has_scrolled = FALSE;

#if GTK_CHECK_VERSION(2,10,0)
  gtk_drag_dest_set (GTK_WIDGET (notebook),
		     GTK_DEST_DEFAULT_MOTION | GTK_DEST_DEFAULT_DROP,
		     notebook_dest_targets, G_N_ELEMENTS (notebook_dest_targets),
                     GDK_ACTION_MOVE);

  gtk_drag_dest_set_track_motion (GTK_WIDGET (notebook), TRUE);
#endif
}

static gboolean
mod_notebook_select_page (ModNotebook *notebook,
                          gboolean     move_focus)
{
  if (gtk_widget_is_focus (GTK_WIDGET (notebook)))
    {
      mod_notebook_page_select (notebook, move_focus);
      return TRUE;
    }
  else
    return FALSE;
}

static gboolean
mod_notebook_focus_tab (ModNotebook       *notebook,
                        GtkNotebookTab     type)
{
  GList *list;

  if (gtk_widget_is_focus (GTK_WIDGET (notebook)))
    {
      switch (type)
	{
	case GTK_NOTEBOOK_TAB_FIRST:
	  list = mod_notebook_search_page (notebook, NULL, STEP_NEXT, TRUE);
	  if (list)
	    mod_notebook_switch_focus_tab (notebook, list);
	  break;
	case GTK_NOTEBOOK_TAB_LAST:
	  list = mod_notebook_search_page (notebook, NULL, STEP_PREV, TRUE);
	  if (list)
	    mod_notebook_switch_focus_tab (notebook, list);
	  break;
	}

      return TRUE;
    }
  else
    return FALSE;
}

static gboolean
mod_notebook_change_current_page (ModNotebook *notebook,
				  gint         offset)
{
  GList *current = NULL;

  if (!notebook->show_tabs)
    return FALSE;

  if (notebook->cur_page)
    current = g_list_find (notebook->children, notebook->cur_page);

  while (offset != 0)
    {
      current = mod_notebook_search_page (notebook, current, offset < 0 ? STEP_PREV : STEP_NEXT, TRUE);
      offset += offset < 0 ? 1 : -1;
    }

  if (current)
    mod_notebook_switch_page (notebook, current->data, -1);
  else
    gdk_display_beep (gtk_widget_get_display (GTK_WIDGET (notebook)));

  return TRUE;
}

static GtkDirectionType
get_effective_direction (ModNotebook      *notebook,
			 GtkDirectionType  direction)
{
  /* Remap the directions into the effective direction it would be for a
   * GTK_POS_TOP notebook
   */

#define D(rest) GTK_DIR_##rest

  static const GtkDirectionType translate_direction[2][4][6] = {
    /* LEFT */   {{ D(TAB_FORWARD),  D(TAB_BACKWARD), D(LEFT), D(RIGHT), D(UP),   D(DOWN) },
    /* RIGHT */  { D(TAB_BACKWARD), D(TAB_FORWARD),  D(LEFT), D(RIGHT), D(DOWN), D(UP)   },
    /* TOP */    { D(TAB_FORWARD),  D(TAB_BACKWARD), D(UP),   D(DOWN),  D(LEFT), D(RIGHT) },
    /* BOTTOM */ { D(TAB_BACKWARD), D(TAB_FORWARD),  D(DOWN), D(UP),    D(LEFT), D(RIGHT) }},
    /* LEFT */  {{ D(TAB_BACKWARD), D(TAB_FORWARD),  D(LEFT), D(RIGHT), D(DOWN), D(UP)   },
    /* RIGHT */  { D(TAB_FORWARD),  D(TAB_BACKWARD), D(LEFT), D(RIGHT), D(UP),   D(DOWN) },
    /* TOP */    { D(TAB_FORWARD),  D(TAB_BACKWARD), D(UP),   D(DOWN),  D(RIGHT), D(LEFT) },
    /* BOTTOM */ { D(TAB_BACKWARD), D(TAB_FORWARD),  D(DOWN), D(UP),    D(RIGHT), D(LEFT) }},
  };

#undef D

  int text_dir = gtk_widget_get_direction (GTK_WIDGET (notebook)) == GTK_TEXT_DIR_RTL ? 1 : 0;

  return translate_direction[text_dir][notebook->tab_pos][direction];
}

static gint
get_effective_tab_pos (ModNotebook *notebook)
{
  if (gtk_widget_get_direction (GTK_WIDGET (notebook)) == GTK_TEXT_DIR_RTL)
    {
      switch (notebook->tab_pos) 
	{
	case GTK_POS_LEFT:
	  return GTK_POS_RIGHT;
	case GTK_POS_RIGHT:
	  return GTK_POS_LEFT;
	default: ;
	}
    }

  return notebook->tab_pos;
}

static gint
get_tab_gap_pos (ModNotebook *notebook)
{
  gint tab_pos = get_effective_tab_pos (notebook);
  gint gap_side = GTK_POS_BOTTOM;
  
  switch (tab_pos)
    {
    case GTK_POS_TOP:
      gap_side = GTK_POS_BOTTOM;
      break;
    case GTK_POS_BOTTOM:
      gap_side = GTK_POS_TOP;
      break;
    case GTK_POS_LEFT:
      gap_side = GTK_POS_RIGHT;
      break;
    case GTK_POS_RIGHT:
      gap_side = GTK_POS_LEFT;
      break;
    }

  return gap_side;
}

static void
mod_notebook_move_focus_out (ModNotebook      *notebook,
			     GtkDirectionType  direction_type)
{
  GtkDirectionType effective_direction = get_effective_direction (notebook, direction_type);
  GtkWidget *toplevel;
  
  if (GTK_CONTAINER (notebook)->focus_child && effective_direction == GTK_DIR_UP)
    if (focus_tabs_in (notebook))
      return;
  if (gtk_widget_is_focus (GTK_WIDGET (notebook)) && effective_direction == GTK_DIR_DOWN)
    if (focus_child_in (notebook, GTK_DIR_TAB_FORWARD))
      return;

  /* At this point, we know we should be focusing out of the notebook entirely. We
   * do this by setting a flag, then propagating the focus motion to the notebook.
   */
  toplevel = gtk_widget_get_toplevel (GTK_WIDGET (notebook));
#if GTK_CHECK_VERSION(2,18,0)
  if (!gtk_widget_is_toplevel (toplevel))
    return;
#else
  if (!GTK_WIDGET_TOPLEVEL (toplevel))
    return;
#endif

  g_object_ref (notebook);
  
  notebook->focus_out = TRUE;
  g_signal_emit_by_name (toplevel, "move_focus", direction_type);
  notebook->focus_out = FALSE;
  
  g_object_unref (notebook);
}

static gint
reorder_tab (ModNotebook *notebook, GList *position, GList *tab)
{
  GList *elem;

  if (position == tab)
    return g_list_position (notebook->children, tab);

  /* check that we aren't inserting the tab in the
   * same relative position, taking packing into account */
  elem = (position) ? position->prev : g_list_last (notebook->children);

  while (elem && elem != tab && MOD_NOTEBOOK_PAGE (elem)->pack != MOD_NOTEBOOK_PAGE (tab)->pack)
    elem = elem->prev;

  if (elem == tab)
    return g_list_position (notebook->children, tab);

  /* now actually reorder the tab */
  if (notebook->first_tab == tab)
    notebook->first_tab = mod_notebook_search_page (notebook, notebook->first_tab,
						    STEP_NEXT, TRUE);

  notebook->children = g_list_remove_link (notebook->children, tab);

  if (!position)
    elem = g_list_last (notebook->children);
  else
    {
      elem = position->prev;
      position->prev = tab;
    }

  if (elem)
    elem->next = tab;
  else
    notebook->children = tab;

  tab->prev = elem;
  tab->next = position;

  return g_list_position (notebook->children, tab);
}

static gboolean
mod_notebook_reorder_tab (ModNotebook      *notebook,
			  GtkDirectionType  direction_type,
			  gboolean          move_to_last)
{
  GtkDirectionType effective_direction = get_effective_direction (notebook, direction_type);
  ModNotebookPage *page;
  GList *last, *child;
  gint page_num;

  if (!gtk_widget_is_focus (GTK_WIDGET (notebook)) || !notebook->show_tabs)
    return FALSE;

  if (!notebook->cur_page ||
      !notebook->cur_page->reorderable)
    return FALSE;

  if (effective_direction != GTK_DIR_LEFT &&
      effective_direction != GTK_DIR_RIGHT)
    return FALSE;

  if (move_to_last)
    {
      child = notebook->focus_tab;

      do
	{
	  last = child;
	  child = mod_notebook_search_page (notebook, last, 
					    (effective_direction == GTK_DIR_RIGHT) ? STEP_NEXT : STEP_PREV,
					    TRUE);
	}
      while (child && MOD_NOTEBOOK_PAGE (last)->pack == MOD_NOTEBOOK_PAGE (child)->pack);

      child = last;
    }
  else
    child = mod_notebook_search_page (notebook, notebook->focus_tab,
				      (effective_direction == GTK_DIR_RIGHT) ? STEP_NEXT : STEP_PREV,
				      TRUE);

  if (!child || child->data == notebook->cur_page)
    return FALSE;

  page = child->data;

  if (page->pack == notebook->cur_page->pack)
    {
      if (effective_direction == GTK_DIR_RIGHT)
	page_num = reorder_tab (notebook, (page->pack == GTK_PACK_START) ? child->next : child, notebook->focus_tab);
      else
	page_num = reorder_tab (notebook, (page->pack == GTK_PACK_START) ? child : child->next, notebook->focus_tab);

      mod_notebook_pages_allocate (notebook);

      g_signal_emit (notebook,
		     notebook_signals[PAGE_REORDERED],
		     0,
		     ((ModNotebookPage *) notebook->focus_tab->data)->child,
		     page_num);
      return TRUE;
    }

  return FALSE;
}

/**
 * mod_notebook_new:
 * 
 * Creates a new #ModNotebook widget with no pages.

 * Return value: the newly created #ModNotebook
 **/
GtkWidget*
mod_notebook_new (void)
{
  return g_object_new (MOD_TYPE_NOTEBOOK, NULL);
}

/* Private GtkObject Methods :
 * 
 * mod_notebook_destroy
 * mod_notebook_set_arg
 * mod_notebook_get_arg
 */
static void
mod_notebook_destroy (GtkObject *object)
{
  ModNotebook *notebook = MOD_NOTEBOOK (object);
  ModNotebookPrivate *priv = MOD_NOTEBOOK_GET_PRIVATE (notebook);
  
  if (notebook->menu)
    mod_notebook_popup_disable (notebook);

  if (priv->source_targets)
    {
      gtk_target_list_unref (priv->source_targets);
      priv->source_targets = NULL;
    }

  if (priv->switch_tab_timer)
    {
      g_source_remove (priv->switch_tab_timer);
      priv->switch_tab_timer = 0;
    }

  GTK_OBJECT_CLASS (mod_notebook_parent_class)->destroy (object);
}

static void
mod_notebook_set_property (GObject         *object,
			   guint            prop_id,
			   const GValue    *value,
			   GParamSpec      *pspec)
{
  ModNotebook *notebook;

  notebook = MOD_NOTEBOOK (object);

  switch (prop_id)
    {
    case PROP_SHOW_TABS:
      mod_notebook_set_show_tabs (notebook, g_value_get_boolean (value));
      break;
    case PROP_SHOW_BORDER:
      mod_notebook_set_show_border (notebook, g_value_get_boolean (value));
      break;
    case PROP_SCROLLABLE:
      mod_notebook_set_scrollable (notebook, g_value_get_boolean (value));
      break;
    case PROP_ENABLE_POPUP:
      if (g_value_get_boolean (value))
	mod_notebook_popup_enable (notebook);
      else
	mod_notebook_popup_disable (notebook);
      break;
    case PROP_HOMOGENEOUS:
      mod_notebook_set_homogeneous_tabs_internal (notebook, g_value_get_boolean (value));
      break;  
    case PROP_PAGE:
      mod_notebook_set_current_page (notebook, g_value_get_int (value));
      break;
    case PROP_TAB_POS:
      mod_notebook_set_tab_pos (notebook, g_value_get_enum (value));
      break;
    case PROP_TAB_BORDER:
      mod_notebook_set_tab_border_internal (notebook, g_value_get_uint (value));
      break;
    case PROP_TAB_HBORDER:
      mod_notebook_set_tab_hborder_internal (notebook, g_value_get_uint (value));
      break;
    case PROP_TAB_VBORDER:
      mod_notebook_set_tab_vborder_internal (notebook, g_value_get_uint (value));
      break;
    case PROP_GROUP_ID:
      mod_notebook_set_group_id (notebook, g_value_get_int (value));
      break;
    case PROP_TAB_SHRINKABLE:
      mod_notebook_set_tab_shrinkable (notebook, g_value_get_boolean(value));
      break;
    case PROP_ENABLE_TOOLTIPS:
      if (g_value_get_boolean (value))
	mod_notebook_tooltips_enable (notebook);
      else
	mod_notebook_tooltips_disable (notebook);
      break;
    case PROP_ENABLE_CLOSE_BUTTON:
      if (g_value_get_boolean (value))
	mod_notebook_close_button_enable (notebook);
      else
	mod_notebook_close_button_disable (notebook);
      break;
    case PROP_MINIMUM_TAB_LABEL_SIZE:
      mod_notebook_set_minimum_tab_label_size (notebook, g_value_get_uint (value));
      break;
    default:
      break;
    }
}

static void
mod_notebook_get_property (GObject         *object,
			   guint            prop_id,
			   GValue          *value,
			   GParamSpec      *pspec)
{
  ModNotebook *notebook;
  ModNotebookPrivate *priv;

  notebook = MOD_NOTEBOOK (object);
  priv = MOD_NOTEBOOK_GET_PRIVATE (notebook);

  switch (prop_id)
    {
    case PROP_SHOW_TABS:
      g_value_set_boolean (value, notebook->show_tabs);
      break;
    case PROP_SHOW_BORDER:
      g_value_set_boolean (value, notebook->show_border);
      break;
    case PROP_SCROLLABLE:
      g_value_set_boolean (value, notebook->scrollable);
      break;
    case PROP_ENABLE_POPUP:
      g_value_set_boolean (value, notebook->menu != NULL);
      break;
    case PROP_HOMOGENEOUS:
      g_value_set_boolean (value, notebook->homogeneous);
      break;
    case PROP_PAGE:
      g_value_set_int (value, mod_notebook_get_current_page (notebook));
      break;
    case PROP_TAB_POS:
      g_value_set_enum (value, notebook->tab_pos);
      break;
    case PROP_TAB_HBORDER:
      g_value_set_uint (value, notebook->tab_hborder);
      break;
    case PROP_TAB_VBORDER:
      g_value_set_uint (value, notebook->tab_vborder);
      break;
    case PROP_GROUP_ID:
      g_value_set_int (value, priv->group_id);
      break;
    case PROP_TAB_SHRINKABLE:
      g_value_set_boolean (value, notebook->tab_shrinkable);
      break;
    case PROP_ENABLE_TOOLTIPS:
      g_value_set_boolean (value, notebook->enable_tooltips);
      break;
    case PROP_ENABLE_CLOSE_BUTTON:
      g_value_set_boolean (value, notebook->enable_close_button);
      break;
    case PROP_MINIMUM_TAB_LABEL_SIZE:
      g_value_set_uint (value, notebook->minimum_tab_label_size);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

/* Private GtkWidget Methods :
 * 
 * mod_notebook_map
 * mod_notebook_unmap
 * mod_notebook_realize
 * mod_notebook_size_request
 * mod_notebook_size_allocate
 * mod_notebook_expose
 * mod_notebook_scroll
 * mod_notebook_button_press
 * mod_notebook_button_release
 * mod_notebook_popup_menu
 * mod_notebook_leave_notify
 * mod_notebook_motion_notify
 * mod_notebook_focus_in
 * mod_notebook_focus_out
 * mod_notebook_draw_focus
 * mod_notebook_style_set
 * mod_notebook_drag_begin
 * mod_notebook_drag_end
 * mod_notebook_drag_motion
 * mod_notebook_drag_drop
 * mod_notebook_drag_data_get
 * mod_notebook_drag_data_received
 */
static gboolean
mod_notebook_get_event_window_position (ModNotebook  *notebook,
					GdkRectangle *rectangle)
{
  GtkWidget *widget = GTK_WIDGET (notebook);
  gint border_width = GTK_CONTAINER (notebook)->border_width;
  ModNotebookPage *visible_page = NULL;
  GList *tmp_list;
  gint tab_pos = get_effective_tab_pos (notebook);

  for (tmp_list = notebook->children; tmp_list; tmp_list = tmp_list->next)
    {
      ModNotebookPage *page = tmp_list->data;
#if GTK_CHECK_VERSION(2,18,0)
      if (gtk_widget_get_visible (page->child))
#else
      if (GTK_WIDGET_VISIBLE (page->child))
#endif
	{
	  visible_page = page;
	  break;
	}
    }

  if (notebook->show_tabs && visible_page)
    {
      if (rectangle)
	{
	  rectangle->x = widget->allocation.x + border_width;
	  rectangle->y = widget->allocation.y + border_width;
	  
	  switch (tab_pos)
	    {
	    case GTK_POS_TOP:
	    case GTK_POS_BOTTOM:
	      rectangle->width = widget->allocation.width - 2 * border_width;
	      rectangle->height = visible_page->requisition.height;
	      if (tab_pos == GTK_POS_BOTTOM)
		rectangle->y += widget->allocation.height - 2 * border_width - rectangle->height;
	      break;
	    case GTK_POS_LEFT:
	    case GTK_POS_RIGHT:
	      rectangle->width = visible_page->requisition.width;
	      rectangle->height = widget->allocation.height - 2 * border_width;
	      if (tab_pos == GTK_POS_RIGHT)
		rectangle->x += widget->allocation.width - 2 * border_width - rectangle->width;
	      break;
	    }
	}

      return TRUE;
    }
  else
    {
      if (rectangle)
	{
	  rectangle->x = rectangle->y = 0;
	  rectangle->width = rectangle->height = 10;
	}
    }

  return FALSE;
}

static void
mod_notebook_map (GtkWidget *widget)
{
  ModNotebook *notebook;
  ModNotebookPage *page;
  GList *children;

  g_return_if_fail (MOD_IS_NOTEBOOK (widget));

  GTK_WIDGET_SET_FLAGS (widget, GTK_MAPPED);

  notebook = MOD_NOTEBOOK (widget);

#if GTK_CHECK_VERSION(2,18,0)
  if (notebook->cur_page && 
      gtk_widget_get_visible (notebook->cur_page->child) &&
      !gtk_widget_get_mapped (notebook->cur_page->child))
    gtk_widget_map (notebook->cur_page->child);
#else
  if (notebook->cur_page && 
      GTK_WIDGET_VISIBLE (notebook->cur_page->child) &&
      !GTK_WIDGET_MAPPED (notebook->cur_page->child))
    gtk_widget_map (notebook->cur_page->child);
#endif

  if (notebook->scrollable)
    mod_notebook_pages_allocate (notebook);
  else
    {
      children = notebook->children;

      while (children)
	{
	  page = children->data;
	  children = children->next;

#if GTK_CHECK_VERSION(2,18,0)
	  if (page->tab_label &&
	      gtk_widget_get_visible (page->tab_label) &&
	      !gtk_widget_get_mapped (page->tab_label))
	    gtk_widget_map (page->tab_label);
#else
	  if (page->tab_label &&
	      GTK_WIDGET_VISIBLE (page->tab_label) &&
	      !GTK_WIDGET_MAPPED (page->tab_label))
	    gtk_widget_map (page->tab_label);
#endif
	}
    }

  if (mod_notebook_get_event_window_position (notebook, NULL))
    gdk_window_show_unraised (notebook->event_window);
}

static void
mod_notebook_unmap (GtkWidget *widget)
{
  g_return_if_fail (MOD_IS_NOTEBOOK (widget));

  stop_scrolling (MOD_NOTEBOOK (widget));
  
  GTK_WIDGET_UNSET_FLAGS (widget, GTK_MAPPED);

  gdk_window_hide (MOD_NOTEBOOK (widget)->event_window);

  GTK_WIDGET_CLASS (mod_notebook_parent_class)->unmap (widget);
}

static void
mod_notebook_realize (GtkWidget *widget)
{
  ModNotebook *notebook;
  GdkWindowAttr attributes;
  gint attributes_mask;
  GdkRectangle event_window_pos;

  g_return_if_fail (MOD_IS_NOTEBOOK (widget));

  notebook = MOD_NOTEBOOK (widget);
  GTK_WIDGET_SET_FLAGS (notebook, GTK_REALIZED);

  mod_notebook_get_event_window_position (notebook, &event_window_pos);
  
  widget->window = gtk_widget_get_parent_window (widget);
  g_object_ref (widget->window);
  
  attributes.window_type = GDK_WINDOW_CHILD;
  attributes.x = event_window_pos.x;
  attributes.y = event_window_pos.y;
  attributes.width = event_window_pos.width;
  attributes.height = event_window_pos.height;
  attributes.wclass = GDK_INPUT_ONLY;
  attributes.event_mask = gtk_widget_get_events (widget);
  attributes.event_mask |= (GDK_BUTTON_PRESS_MASK |
			    GDK_BUTTON_RELEASE_MASK | GDK_KEY_PRESS_MASK |
			    GDK_POINTER_MOTION_MASK | GDK_LEAVE_NOTIFY_MASK |
			    GDK_SCROLL_MASK);
  attributes_mask = GDK_WA_X | GDK_WA_Y;

  notebook->event_window = gdk_window_new (gtk_widget_get_parent_window (widget), 
					   &attributes, attributes_mask);
  gdk_window_set_user_data (notebook->event_window, notebook);

  widget->style = gtk_style_attach (widget->style, widget->window);
}

static void
mod_notebook_unrealize (GtkWidget *widget)
{
  ModNotebook *notebook;
  ModNotebookPrivate *priv;

  g_return_if_fail (MOD_IS_NOTEBOOK (widget));

  notebook = MOD_NOTEBOOK (widget);
  priv = MOD_NOTEBOOK_GET_PRIVATE (widget);

  gdk_window_set_user_data (notebook->event_window, NULL);
  gdk_window_destroy (notebook->event_window);
  notebook->event_window = NULL;

  if (priv->drag_window)
    {
      gdk_window_set_user_data (priv->drag_window, NULL);
      gdk_window_destroy (priv->drag_window);
      priv->drag_window = NULL;
    }

  if (GTK_WIDGET_CLASS (mod_notebook_parent_class)->unrealize)
    (* GTK_WIDGET_CLASS (mod_notebook_parent_class)->unrealize) (widget);
}

static void
mod_notebook_size_request (GtkWidget      *widget,
			   GtkRequisition *requisition)
{
  ModNotebook *notebook = MOD_NOTEBOOK (widget);
  ModNotebookPage *page;
  GList *children;
  GtkRequisition child_requisition;
  gboolean switch_page = FALSE;
  gint vis_pages;
  gint focus_width;
  gint tab_overlap;
  gint tab_curvature;
  gint arrow_spacing;
  gint scroll_arrow_hlength;
  gint scroll_arrow_vlength;

  gtk_widget_style_get (widget,
			"focus-line-width", &focus_width,
			"tab-overlap", &tab_overlap,
			"tab-curvature", &tab_curvature,
                        "arrow-spacing", &arrow_spacing,
                        "scroll-arrow-hlength", &scroll_arrow_hlength,
                        "scroll-arrow-vlength", &scroll_arrow_vlength,
			NULL);

  widget->requisition.width = 0;
  widget->requisition.height = 0;

  for (children = notebook->children, vis_pages = 0; children;
       children = children->next)
    {
      page = children->data;

#if GTK_CHECK_VERSION(2,18,0)
      if (gtk_widget_get_visible (page->child))
	{
	  vis_pages++;
	  gtk_widget_size_request (page->child, &child_requisition);
	  
	  widget->requisition.width = MAX (widget->requisition.width,
					   child_requisition.width);
	  widget->requisition.height = MAX (widget->requisition.height,
					    child_requisition.height);

	  if (notebook->menu && page->menu_label->parent &&
	      !gtk_widget_get_visible (page->menu_label->parent))
	    gtk_widget_show (page->menu_label->parent);
	}
      else
	{
	  if (page == notebook->cur_page)
	    switch_page = TRUE;
	  if (notebook->menu && page->menu_label->parent &&
	      gtk_widget_get_visible (page->menu_label->parent))
	    gtk_widget_hide (page->menu_label->parent);
	}
#else
      if (GTK_WIDGET_VISIBLE (page->child))
	{
	  vis_pages++;
	  gtk_widget_size_request (page->child, &child_requisition);
	  
	  widget->requisition.width = MAX (widget->requisition.width,
					   child_requisition.width);
	  widget->requisition.height = MAX (widget->requisition.height,
					    child_requisition.height);

	  if (notebook->menu && page->menu_label->parent &&
	      !GTK_WIDGET_VISIBLE (page->menu_label->parent))
	    gtk_widget_show (page->menu_label->parent);
	}
      else
	{
	  if (page == notebook->cur_page)
	    switch_page = TRUE;
	  if (notebook->menu && page->menu_label->parent &&
	      GTK_WIDGET_VISIBLE (page->menu_label->parent))
	    gtk_widget_hide (page->menu_label->parent);
	}
#endif
    }

  if (notebook->show_border || notebook->show_tabs)
    {
      widget->requisition.width += widget->style->xthickness * 2;
      widget->requisition.height += widget->style->ythickness * 2;

      if (notebook->show_tabs)
	{
	  gint tab_width = 0;
	  gint tab_height = 0;
	  gint tab_max = 0;
	  gint padding;
	  
	  for (children = notebook->children; children;
	       children = children->next)
	    {
	      page = children->data;

#if GTK_CHECK_VERSION(2,18,0)
	      if (gtk_widget_get_visible (page->child))
		{
		  if (!gtk_widget_get_visible (page->tab_label))
		    gtk_widget_show (page->tab_label);

		  gtk_widget_size_request (page->tab_label,
					   &child_requisition);

		  page->requisition.width = 
		    child_requisition.width +
		    2 * widget->style->xthickness;
		  page->requisition.height = 
		    child_requisition.height +
		    2 * widget->style->ythickness;
		  
		  switch (notebook->tab_pos)
		    {
		    case GTK_POS_TOP:
		    case GTK_POS_BOTTOM:
		      page->requisition.height += 2 * (notebook->tab_vborder +
						       focus_width);
		      tab_height = MAX (tab_height, page->requisition.height);
		      tab_max = MAX (tab_max, page->requisition.width);
		      break;
		    case GTK_POS_LEFT:
		    case GTK_POS_RIGHT:
		      page->requisition.width += 2 * (notebook->tab_hborder +
						      focus_width);
		      tab_width = MAX (tab_width, page->requisition.width);
		      tab_max = MAX (tab_max, page->requisition.height);
		      break;
		    }
		}
	      else if (gtk_widget_get_visible (page->tab_label))
		gtk_widget_hide (page->tab_label);
#else	      
	      if (GTK_WIDGET_VISIBLE (page->child))
		{
		  if (!GTK_WIDGET_VISIBLE (page->tab_label))
		    gtk_widget_show (page->tab_label);

		  gtk_widget_size_request (page->tab_label,
					   &child_requisition);

		  page->requisition.width = 
		    child_requisition.width +
		    2 * widget->style->xthickness;
		  page->requisition.height = 
		    child_requisition.height +
		    2 * widget->style->ythickness;
		  
		  switch (notebook->tab_pos)
		    {
		    case GTK_POS_TOP:
		    case GTK_POS_BOTTOM:
		      page->requisition.height += 2 * (notebook->tab_vborder +
						       focus_width);
		      tab_height = MAX (tab_height, page->requisition.height);
		      tab_max = MAX (tab_max, page->requisition.width);
		      break;
		    case GTK_POS_LEFT:
		    case GTK_POS_RIGHT:
		      page->requisition.width += 2 * (notebook->tab_hborder +
						      focus_width);
		      tab_width = MAX (tab_width, page->requisition.width);
		      tab_max = MAX (tab_max, page->requisition.height);
		      break;
		    }
		}
	      else if (GTK_WIDGET_VISIBLE (page->tab_label))
		gtk_widget_hide (page->tab_label);
#endif
	    }

	  children = notebook->children;

	  if (vis_pages)
	    {
	      switch (notebook->tab_pos)
		{
		case GTK_POS_TOP:
		case GTK_POS_BOTTOM:
		  if (tab_height == 0)
		    break;

		  if (notebook->scrollable && vis_pages > 1 && 
		      widget->requisition.width < tab_width)
		    tab_height = MAX (tab_height, scroll_arrow_hlength);

		  padding = 2 * (tab_curvature + focus_width +
				 notebook->tab_hborder) - tab_overlap;
		  tab_max += padding;
		  while (children)
		    {
		      page = children->data;
		      children = children->next;

#if GTK_CHECK_VERSION(2,18,0)
		      if (!gtk_widget_get_visible (page->child))
			continue;
#else		  
		      if (!GTK_WIDGET_VISIBLE (page->child))
			continue;
#endif

		      if (notebook->homogeneous)
			page->requisition.width = tab_max;
		      else
			page->requisition.width += padding;

		      tab_width += page->requisition.width;
		      page->requisition.height = tab_height;
		    }

		  if (notebook->scrollable && vis_pages > 1 &&
		      widget->requisition.width < tab_width)
		    tab_width = tab_max + 2 * (scroll_arrow_hlength + arrow_spacing);

                  if (notebook->homogeneous && !notebook->scrollable)
                    widget->requisition.width = MAX (widget->requisition.width,
                                                     vis_pages * tab_max +
                                                     tab_overlap);
                  else
                    widget->requisition.width = MAX (widget->requisition.width,
                                                     tab_width + tab_overlap);

		  widget->requisition.height += tab_height;
		  break;
		case GTK_POS_LEFT:
		case GTK_POS_RIGHT:
		  if (tab_width == 0)
		    break;

		  if (notebook->scrollable && vis_pages > 1 && 
		      widget->requisition.height < tab_height)
		    tab_width = MAX (tab_width,
                                     arrow_spacing + 2 * scroll_arrow_vlength);

		  padding = 2 * (tab_curvature + focus_width +
				 notebook->tab_vborder) - tab_overlap;
		  tab_max += padding;

		  while (children)
		    {
		      page = children->data;
		      children = children->next;

#if GTK_CHECK_VERSION(2,18,0)
		      if (!gtk_widget_get_visible (page->child))
			continue;
#else
		      if (!GTK_WIDGET_VISIBLE (page->child))
			continue;
#endif

		      page->requisition.width   = tab_width;

		      if (notebook->homogeneous)
			page->requisition.height = tab_max;
		      else
			page->requisition.height += padding;

		      tab_height += page->requisition.height;
		    }

		  if (notebook->scrollable && vis_pages > 1 && 
		      widget->requisition.height < tab_height)
		    tab_height = tab_max + scroll_arrow_vlength + arrow_spacing;

		  widget->requisition.width += tab_width;

                  if (notebook->homogeneous && !notebook->scrollable)
                    widget->requisition.height =
		      MAX (widget->requisition.height,
			   vis_pages * tab_max + tab_overlap);
                  else
                    widget->requisition.height =
		      MAX (widget->requisition.height,
			   tab_height + tab_overlap);

		  if (!notebook->homogeneous || notebook->scrollable)
		    vis_pages = 1;
		  widget->requisition.height = MAX (widget->requisition.height,
						    vis_pages * tab_max +
						    tab_overlap);
		  break;
		}
	    }
	}
      else
	{
	  for (children = notebook->children; children;
	       children = children->next)
	    {
	      page = children->data;

#if GTK_CHECK_VERSION(2,18,0)	      
	      if (page->tab_label && gtk_widget_get_visible (page->tab_label))
		gtk_widget_hide (page->tab_label);
#else
	      if (page->tab_label && GTK_WIDGET_VISIBLE (page->tab_label))
		gtk_widget_hide (page->tab_label);
#endif
	    }
	}
    }

  widget->requisition.width += GTK_CONTAINER (widget)->border_width * 2;
  widget->requisition.height += GTK_CONTAINER (widget)->border_width * 2;

  if (switch_page)
    {
      if (vis_pages)
	{
	  for (children = notebook->children; children;
	       children = children->next)
	    {
	      page = children->data;
#if GTK_CHECK_VERSION(2,18,0)
	      if (gtk_widget_get_visible (page->child))
		{
		  mod_notebook_switch_page (notebook, page, -1);
		  break;
		}
#else
	      if (GTK_WIDGET_VISIBLE (page->child))
		{
		  mod_notebook_switch_page (notebook, page, -1);
		  break;
		}
#endif
	    }
	}
#if GTK_CHECK_VERSION(2,18,0)
      else if (gtk_widget_get_visible (widget))
	{
	  widget->requisition.width = GTK_CONTAINER (widget)->border_width * 2;
	  widget->requisition.height= GTK_CONTAINER (widget)->border_width * 2;
	}
#else
      else if (GTK_WIDGET_VISIBLE (widget))
	{
	  widget->requisition.width = GTK_CONTAINER (widget)->border_width * 2;
	  widget->requisition.height= GTK_CONTAINER (widget)->border_width * 2;
	}
#endif
    }
  if (vis_pages && !notebook->cur_page)
    {
      children = mod_notebook_search_page (notebook, NULL, STEP_NEXT, TRUE);
      if (children)
	{
	  notebook->first_tab = children;
	  mod_notebook_switch_page (notebook, MOD_NOTEBOOK_PAGE (children),-1);
	}
    }
}

static void
mod_notebook_size_allocate (GtkWidget     *widget,
			    GtkAllocation *allocation)
{
  ModNotebook *notebook = MOD_NOTEBOOK (widget);
  gint tab_pos = get_effective_tab_pos (notebook);

  widget->allocation = *allocation;
#if GTK_CHECK_VERSION(2,18,0)
  if (gtk_widget_get_realized (widget))
#else
  if (GTK_WIDGET_REALIZED (widget))
#endif
    {
      GdkRectangle position;

      if (mod_notebook_get_event_window_position (notebook, &position))
	{
	  gdk_window_move_resize (notebook->event_window,
				  position.x, position.y,
				  position.width, position.height);
#if GTK_CHECK_VERSION(2,18,0)
	  if (gtk_widget_get_mapped (widget))
	    gdk_window_show_unraised (notebook->event_window);
#else
	  if (GTK_WIDGET_MAPPED (notebook))
	    gdk_window_show_unraised (notebook->event_window);
#endif
	}
      else
	gdk_window_hide (notebook->event_window);
    }

  if (notebook->children)
    {
      gint border_width = GTK_CONTAINER (widget)->border_width;
      ModNotebookPage *page;
      GtkAllocation child_allocation;
      GList *children;
      
      child_allocation.x = widget->allocation.x + border_width;
      child_allocation.y = widget->allocation.y + border_width;
      child_allocation.width = MAX (1, allocation->width - border_width * 2);
      child_allocation.height = MAX (1, allocation->height - border_width * 2);

      if (notebook->show_tabs || notebook->show_border)
	{
	  child_allocation.x += widget->style->xthickness;
	  child_allocation.y += widget->style->ythickness;
	  child_allocation.width = MAX (1, child_allocation.width -
					widget->style->xthickness * 2);
	  child_allocation.height = MAX (1, child_allocation.height -
					 widget->style->ythickness * 2);

	  if (notebook->show_tabs && notebook->children && notebook->cur_page)
	    {
	      switch (tab_pos)
		{
		case GTK_POS_TOP:
		  child_allocation.y += notebook->cur_page->requisition.height;
		case GTK_POS_BOTTOM:
		  child_allocation.height =
		    MAX (1, child_allocation.height -
			 notebook->cur_page->requisition.height);
		  break;
		case GTK_POS_LEFT:
		  child_allocation.x += notebook->cur_page->requisition.width;
		case GTK_POS_RIGHT:
		  child_allocation.width =
		    MAX (1, child_allocation.width -
			 notebook->cur_page->requisition.width);
		  break;
		}
	    }
	}

      children = notebook->children;
      while (children)
	{
	  page = children->data;
	  children = children->next;

#if GTK_CHECK_VERSION(2,18,0)	  
	  if (gtk_widget_get_visible (page->child))
	    gtk_widget_size_allocate (page->child, &child_allocation);
#else
	  if (GTK_WIDGET_VISIBLE (page->child))
	    gtk_widget_size_allocate (page->child, &child_allocation);
#endif
	}

      mod_notebook_pages_allocate (notebook);
    }
}

static gint
mod_notebook_expose (GtkWidget      *widget,
		     GdkEventExpose *event)
{
  ModNotebook *notebook;
  ModNotebookPrivate *priv;
  GdkRectangle child_area;

  g_return_val_if_fail (MOD_IS_NOTEBOOK (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  notebook = MOD_NOTEBOOK (widget);
  priv = MOD_NOTEBOOK_GET_PRIVATE (widget);

  if (event->window == priv->drag_window)
    {
      GdkRectangle area = { 0, };

      gdk_drawable_get_size (priv->drag_window,
			     &area.width, &area.height);
      mod_notebook_draw_tab (notebook,
			     notebook->cur_page,
			     &area);
    }
#if GTK_CHECK_VERSION(2,18,0)
  else if (gtk_widget_is_drawable (widget))
#else
  else if (GTK_WIDGET_DRAWABLE (widget))
#endif
    {
      mod_notebook_paint (widget, &event->area);
      if (notebook->show_tabs)
	{
	  if (notebook->cur_page && 
	      gtk_widget_intersect (notebook->cur_page->tab_label, 
				    &event->area, &child_area))
	    mod_notebook_draw_focus (widget);
	}

      
      if (notebook->cur_page)
	gtk_container_propagate_expose (GTK_CONTAINER (notebook),
					notebook->cur_page->child,
					event);
    }

  return FALSE;
}

static gboolean
mod_notebook_show_arrows (ModNotebook *notebook)
{
  gboolean show_arrow = FALSE;
  GList *children;
  
  if (!notebook->scrollable)
    return FALSE;

  children = notebook->children;
  while (children)
    {
      ModNotebookPage *page = children->data;

      if (page->tab_label && !gtk_widget_get_child_visible (page->tab_label))
	show_arrow = TRUE;

      children = children->next;
    }

  return show_arrow;
}

static void
mod_notebook_get_arrow_rect (ModNotebook     *notebook,
			     GdkRectangle    *rectangle,
			     ModNotebookArrow arrow)
{
  GdkRectangle event_window_pos;
  gboolean before = ARROW_IS_BEFORE (arrow);
  gboolean left = ARROW_IS_LEFT (arrow);

  if (mod_notebook_get_event_window_position (notebook, &event_window_pos))
    {
      gint scroll_arrow_hlength;
      gint scroll_arrow_vlength;
      gint close_button_size;
      gint close_button_spacing;

      gtk_widget_style_get (GTK_WIDGET (notebook),
                            "scroll-arrow-hlength", &scroll_arrow_hlength,
                            "scroll-arrow-vlength", &scroll_arrow_vlength,
			    "close-button-size", &close_button_size,
			    "close-button-spacing", &close_button_spacing,
                            NULL);

      if (notebook->enable_close_button)
	close_button_size += close_button_spacing * 2;
      else
	close_button_size = 0;

      switch (notebook->tab_pos)
	{
	case GTK_POS_LEFT:
	case GTK_POS_RIGHT:
          rectangle->width = scroll_arrow_vlength;
          rectangle->height = scroll_arrow_vlength;

	  if ((before && (notebook->has_before_previous != notebook->has_before_next)) ||
	      (!before && (notebook->has_after_previous != notebook->has_after_next))) 
	    rectangle->x = event_window_pos.x + (event_window_pos.width - rectangle->width) / 2;
	  else if (left)
	    rectangle->x = event_window_pos.x + event_window_pos.width / 2 - rectangle->width;
	  else 
	    rectangle->x = event_window_pos.x + event_window_pos.width / 2;
	  rectangle->y = event_window_pos.y;
	  if (!before)
	    rectangle->y += event_window_pos.height - rectangle->height -
	      close_button_size;

	  break;

	case GTK_POS_TOP:
	case GTK_POS_BOTTOM:
          rectangle->width = scroll_arrow_hlength;
          rectangle->height = scroll_arrow_hlength;

	  if (before)
	    {
	      if (left || !notebook->has_before_previous)
		rectangle->x = event_window_pos.x;
	      else
		rectangle->x = event_window_pos.x + rectangle->width;
	    }
	  else
	    {
	      if (!left || !notebook->has_after_next)
		rectangle->x = event_window_pos.x + event_window_pos.width - rectangle->width;
	      else
		rectangle->x = event_window_pos.x + event_window_pos.width - 2 * rectangle->width;
	      rectangle->x -= close_button_size;
	    }
	  rectangle->y = event_window_pos.y + (event_window_pos.height - rectangle->height) / 2;
	  break;
	}
    }
}

static ModNotebookArrow
mod_notebook_get_arrow (ModNotebook *notebook,
			gint         x,
			gint         y)
{
  GdkRectangle arrow_rect;
  GdkRectangle event_window_pos;
  gint i;
  gint x0, y0;
  ModNotebookArrow arrow[4];

  arrow[0] = notebook->has_before_previous ? ARROW_LEFT_BEFORE : ARROW_NONE;
  arrow[1] = notebook->has_before_next ? ARROW_RIGHT_BEFORE : ARROW_NONE;
  arrow[2] = notebook->has_after_previous ? ARROW_LEFT_AFTER : ARROW_NONE;
  arrow[3] = notebook->has_after_next ? ARROW_RIGHT_AFTER : ARROW_NONE;

  if (mod_notebook_show_arrows (notebook))
    {
      mod_notebook_get_event_window_position (notebook, &event_window_pos);
      for (i = 0; i < 4; i++)
	{
	  if (arrow[i] == ARROW_NONE)
	    continue;

	  mod_notebook_get_arrow_rect (notebook, &arrow_rect, arrow[i]);

	  x0 = x - arrow_rect.x;
	  y0 = y - arrow_rect.y;

	  if (y0 >= 0 && y0 < arrow_rect.height &&
	      x0 >= 0 && x0 < arrow_rect.width)
	    return arrow[i];
	}
    }

  return ARROW_NONE;
}

static void
mod_notebook_do_arrow (ModNotebook     *notebook,
		       GtkArrowType     arrow)
{
  GtkWidget *widget = GTK_WIDGET (notebook);
  GtkDirectionType dir;
  gboolean is_rtl, left;

  is_rtl = gtk_widget_get_direction (widget) == GTK_TEXT_DIR_RTL;
  left = (ARROW_IS_LEFT (arrow) && !is_rtl) || 
         (!ARROW_IS_LEFT (arrow) && is_rtl);

  if (!notebook->focus_tab ||
      mod_notebook_search_page (notebook, notebook->focus_tab,
				left ? STEP_PREV : STEP_NEXT,
				TRUE))
    {
      if (notebook->tab_pos == GTK_POS_LEFT ||
	  notebook->tab_pos == GTK_POS_RIGHT)
	dir = ARROW_IS_LEFT (arrow) ? GTK_DIR_UP : GTK_DIR_DOWN;
      else
	dir = ARROW_IS_LEFT (arrow) ? GTK_DIR_LEFT : GTK_DIR_RIGHT;
      gtk_widget_child_focus (widget, dir);
    }
}

static gboolean
mod_notebook_arrow_button_press (ModNotebook      *notebook,
				 ModNotebookArrow  arrow,
				 gint              button)
{
  GtkWidget *widget = GTK_WIDGET (notebook);

  gboolean is_rtl = gtk_widget_get_direction (widget) == GTK_TEXT_DIR_RTL;
  gboolean left = (ARROW_IS_LEFT (arrow) && !is_rtl) || 
                  (!ARROW_IS_LEFT (arrow) && is_rtl);

#if GTK_CHECK_VERSION(2,18,0)
  if (!gtk_widget_has_focus (widget))
    gtk_widget_grab_focus (widget);
#else
  if (!GTK_WIDGET_HAS_FOCUS (widget))
    gtk_widget_grab_focus (widget);
#endif
  
  notebook->button = button;
  notebook->click_child = arrow;

  if (button == 1)
    {
      mod_notebook_do_arrow (notebook, arrow);
      mod_notebook_set_scroll_timer (notebook);
    }
  else if (button == 2)
    mod_notebook_page_select (notebook, TRUE);
  else if (button == 3)
    mod_notebook_switch_focus_tab (notebook,
				   mod_notebook_search_page (notebook,
							     NULL,
							     left ? STEP_NEXT : STEP_PREV,
							     TRUE));
  mod_notebook_redraw_arrows (notebook);

  return TRUE;
}

static gboolean
get_widget_coordinates (GtkWidget *widget,
			GdkEvent  *event,
			gint      *x,
			gint      *y)
{
  GdkWindow *window = ((GdkEventAny *)event)->window;
  gdouble tx, ty;

  if (!gdk_event_get_coords (event, &tx, &ty))
    return FALSE;

  while (window && window != widget->window)
    {
      gint window_x, window_y;

      gdk_window_get_position (window, &window_x, &window_y);
      tx += window_x;
      ty += window_y;

      window = gdk_window_get_parent (window);
    }

  if (window)
    {
      *x = tx;
      *y = ty;

      return TRUE;
    }

  return FALSE;
}

static gboolean
mod_notebook_scroll (GtkWidget      *widget,
                     GdkEventScroll *event)
{
  ModNotebook *notebook = MOD_NOTEBOOK (widget);

  GtkWidget* child;
  GtkWidget* originator;

  if (!notebook->cur_page)
    return FALSE;

  child = notebook->cur_page->child;
  originator = gtk_get_event_widget ((GdkEvent *)event);

  /* ignore scroll events from the content of the page */
  if (!originator || gtk_widget_is_ancestor (originator, child))
    return FALSE;

  switch (event->direction)
    {
    case GDK_SCROLL_RIGHT:
    case GDK_SCROLL_DOWN:
      mod_notebook_next_page (notebook);
      break;
    case GDK_SCROLL_LEFT:
    case GDK_SCROLL_UP:
      mod_notebook_prev_page (notebook);
      break;
    }

  return TRUE;
}

static GList*
get_tab_at_pos (ModNotebook *notebook, gint x, gint y)
{
  ModNotebookPage *page;
  GList *children = notebook->children;

  while (children)
    {
      page = children->data;

#if GTK_CHECK_VERSION(2,18,0)
      if (gtk_widget_get_visible (page->child) &&
	  page->tab_label && gtk_widget_get_mapped (page->tab_label) &&
	  (x >= page->allocation.x) &&
	  (y >= page->allocation.y) &&
	  (x <= (page->allocation.x + page->allocation.width)) &&
	  (y <= (page->allocation.y + page->allocation.height)))
	return children;
#else      
      if (GTK_WIDGET_VISIBLE (page->child) &&
	  page->tab_label && GTK_WIDGET_MAPPED (page->tab_label) &&
	  (x >= page->allocation.x) &&
	  (y >= page->allocation.y) &&
	  (x <= (page->allocation.x + page->allocation.width)) &&
	  (y <= (page->allocation.y + page->allocation.height)))
	return children;
#endif

      children = children->next;
    }

  return NULL;
}

static gboolean
mod_notebook_button_press (GtkWidget      *widget,
			   GdkEventButton *event)
{
  ModNotebook *notebook = MOD_NOTEBOOK (widget);
  ModNotebookPrivate *priv = MOD_NOTEBOOK_GET_PRIVATE (notebook);
  ModNotebookPage *page;
  GList *tab;
  ModNotebookArrow arrow;
  gint x, y;

  if (event->type != GDK_BUTTON_PRESS || !notebook->children ||
      notebook->button)
    return FALSE;

  if (!get_widget_coordinates (widget, (GdkEvent *)event, &x, &y))
    return FALSE;

  arrow = mod_notebook_get_arrow (notebook, x, y);
  if (arrow != ARROW_NONE)
    return mod_notebook_arrow_button_press (notebook, arrow, event->button);

  if (event->button == 3 && notebook->menu)
    {
      gtk_menu_popup (GTK_MENU (notebook->menu), NULL, NULL, 
		      NULL, NULL, 3, event->time);
      return TRUE;
    }

  if (event->button != 1)
    return FALSE;

  if (mod_notebook_in_close_button (notebook, x, y))
    {
      notebook->click_close_button = TRUE;
      mod_notebook_redraw_close_button (notebook);
      return TRUE;
    }

  notebook->button = event->button;

  if ((tab = get_tab_at_pos (notebook, x, y)) != NULL)
    {
      gboolean page_changed, was_focus;

      page = tab->data;
      page_changed = page != notebook->cur_page;
      was_focus = gtk_widget_is_focus (widget);

      mod_notebook_switch_focus_tab (notebook, tab);
      gtk_widget_grab_focus (widget);

      if (page_changed && !was_focus)
	gtk_widget_child_focus (page->child, GTK_DIR_TAB_FORWARD);

      /* save press to possibly begin a drag */
      if (page->reorderable || page->detachable)
	{
	  priv->during_detach = FALSE;
	  priv->during_reorder = FALSE;
	  priv->pressed_button = event->button;

	  gdk_window_get_pointer (widget->window,
				  &priv->mouse_x,
				  &priv->mouse_y,
				  NULL);

	  priv->drag_begin_x = priv->mouse_x;
	  priv->drag_begin_y = priv->mouse_y;
	  priv->drag_offset_x = priv->drag_begin_x - page->allocation.x;
	  priv->drag_offset_y = priv->drag_begin_y - page->allocation.y;
	}
    }

  return TRUE;
}

static void
popup_position_func (GtkMenu  *menu,
                     gint     *x,
                     gint     *y,
                     gboolean *push_in,
                     gpointer  data)
{
  ModNotebook *notebook = data;
  GtkWidget *w;
  GtkRequisition requisition;

  if (notebook->focus_tab)
    {
      ModNotebookPage *page;

      page = notebook->focus_tab->data;
      w = page->tab_label;
    }
  else
   {
     w = GTK_WIDGET (notebook);
   }

  gdk_window_get_origin (w->window, x, y);
  gtk_widget_size_request (GTK_WIDGET (menu), &requisition);

  if (gtk_widget_get_direction (w) == GTK_TEXT_DIR_RTL)
    *x += w->allocation.x + w->allocation.width - requisition.width;
  else
    *x += w->allocation.x;

  *y += w->allocation.y + w->allocation.height;

  *push_in = FALSE;
}

static gboolean
mod_notebook_popup_menu (GtkWidget *widget)
{
  ModNotebook *notebook = MOD_NOTEBOOK (widget);

  if (notebook->menu)
    {
      gtk_menu_popup (GTK_MENU (notebook->menu), NULL, NULL, 
		      popup_position_func, notebook,
		      0, gtk_get_current_event_time ());
      gtk_menu_shell_select_first (GTK_MENU_SHELL (notebook->menu), FALSE);
      return TRUE;
    }

  return FALSE;
}

static void 
stop_scrolling (ModNotebook *notebook)
{
  if (notebook->timer)
    {
      g_source_remove (notebook->timer);
      notebook->timer = 0;
      notebook->need_timer = FALSE;
    }
  notebook->click_child = 0;
  notebook->button = 0;
  mod_notebook_redraw_arrows (notebook);
}

#if GTK_CHECK_VERSION(2,10,0)
static GList*
get_drop_position (ModNotebook *notebook,
		   guint        pack)
{
  ModNotebookPrivate *priv;
  GList *children, *last_child;
  ModNotebookPage *page;
  gboolean is_rtl;
  gint x, y;

  priv = MOD_NOTEBOOK_GET_PRIVATE (notebook);
  x = priv->mouse_x;
  y = priv->mouse_y;

  is_rtl = gtk_widget_get_direction ((GtkWidget *) notebook) == GTK_TEXT_DIR_RTL;
  children = notebook->children;
  last_child = NULL;

  while (children)
    {
      page = children->data;

      if ((priv->operation != DRAG_OPERATION_REORDER || page != notebook->cur_page) &&
#if GTK_CHECK_VERSION(2,18,0)
	  gtk_widget_get_visible (page->child) &&
	  page->tab_label &&
	  gtk_widget_get_mapped (page->tab_label) &&
#else
	  GTK_WIDGET_VISIBLE (page->child) &&
	  page->tab_label &&
	  GTK_WIDGET_MAPPED (page->tab_label) &&
#endif
	  page->pack == pack)
	{
	  switch (notebook->tab_pos)
	    {
	    case GTK_POS_TOP:
	    case GTK_POS_BOTTOM:
	      if (!is_rtl)
		{
		  if ((page->pack == GTK_PACK_START && PAGE_MIDDLE_X (page) > x) ||
		      (page->pack == GTK_PACK_END && PAGE_MIDDLE_X (page) < x))
		    return children;
		}
	      else
		{
		  if ((page->pack == GTK_PACK_START && PAGE_MIDDLE_X (page) < x) ||
		      (page->pack == GTK_PACK_END && PAGE_MIDDLE_X (page) > x))
		    return children;
		}

	      break;
	    case GTK_POS_LEFT:
	    case GTK_POS_RIGHT:
	      if ((page->pack == GTK_PACK_START && PAGE_MIDDLE_Y (page) > y) ||
		  (page->pack == GTK_PACK_END && PAGE_MIDDLE_Y (page) < y))
		return children;

	      break;
	    }

	  last_child = children->next;
	}

      children = children->next;
    }

  return last_child;
}

static void
show_drag_window (ModNotebook        *notebook,
		  ModNotebookPrivate *priv,
		  ModNotebookPage    *page)
{
  GtkWidget *widget = GTK_WIDGET (notebook);

  if (!priv->drag_window)
    {
      GdkWindowAttr attributes;
      guint attributes_mask;

      attributes.x = page->allocation.x;
      attributes.y = page->allocation.y;
      attributes.width = page->allocation.width;
      attributes.height = page->allocation.height;
      attributes.window_type = GDK_WINDOW_CHILD;
      attributes.wclass = GDK_INPUT_OUTPUT;
      attributes.visual = gtk_widget_get_visual (widget);
      attributes.colormap = gtk_widget_get_colormap (widget);
      attributes.event_mask = GDK_VISIBILITY_NOTIFY_MASK | GDK_EXPOSURE_MASK | GDK_POINTER_MOTION_MASK;
      attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;

      priv->drag_window = gdk_window_new (gtk_widget_get_parent_window (widget),
					  &attributes,
					  attributes_mask);
      gdk_window_set_user_data (priv->drag_window, widget);
    }

  g_object_ref (page->tab_label);
  gtk_widget_unparent (page->tab_label);
  gtk_widget_set_parent_window (page->tab_label, priv->drag_window);
  gtk_widget_set_parent (page->tab_label, widget);
  g_object_unref (page->tab_label);

  gdk_window_show (priv->drag_window);

  /* the grab will dissapear when the window is hidden */
  gdk_pointer_grab (priv->drag_window,
		    FALSE,
		    GDK_POINTER_MOTION_MASK | GDK_BUTTON_RELEASE_MASK,
		    NULL, NULL, GDK_CURRENT_TIME);
}

/* This function undoes the reparenting that happens both when drag_window
 * is shown for reordering and when the DnD icon is shown for detaching
 */
static void
hide_drag_window (ModNotebook        *notebook,
		  ModNotebookPrivate *priv,
		  ModNotebookPage    *page)
{
  GtkWidget *widget = GTK_WIDGET (notebook);
  GtkWidget *parent = page->tab_label->parent;

  if (page->tab_label->window != widget->window ||
      !NOTEBOOK_IS_TAB_LABEL_PARENT (notebook, page))
    {
      g_object_ref (page->tab_label);

      if (GTK_IS_WINDOW (parent))
	{
	  /* parent widget is the drag window */
	  gtk_container_remove (GTK_CONTAINER (parent), page->tab_label);
	}
      else
	gtk_widget_unparent (page->tab_label);

      gtk_widget_set_parent_window (page->tab_label, widget->window);
      gtk_widget_set_parent (page->tab_label, widget);
      g_object_unref (page->tab_label);
    }

  if (priv->drag_window &&
      gdk_window_is_visible (priv->drag_window))
    gdk_window_hide (priv->drag_window);
}

static void
mod_notebook_stop_reorder (ModNotebook *notebook)
{
  ModNotebookPrivate *priv;
  ModNotebookPage *page;

  priv = MOD_NOTEBOOK_GET_PRIVATE (notebook);
  page = notebook->cur_page;

  if (!page)
    return;

  priv->pressed_button = -1;

  if (page->reorderable || page->detachable)
    {
      if (priv->during_reorder)
	{
	  gint old_page_num, page_num;
	  GList *element;

	  element = get_drop_position (notebook, page->pack);
	  old_page_num = g_list_position (notebook->children, notebook->focus_tab);
	  page_num = reorder_tab (notebook, element, notebook->focus_tab);

	  if (priv->has_scrolled || old_page_num != page_num)
	    g_signal_emit (notebook,
			   notebook_signals[PAGE_REORDERED], 0,
			   page->child, page_num);

	  priv->has_scrolled = FALSE;
          priv->during_reorder = FALSE; 
	}

      hide_drag_window (notebook, priv, page);

      priv->operation = DRAG_OPERATION_NONE;
      mod_notebook_pages_allocate (notebook);

      if (priv->dnd_timer)
	{
	  g_source_remove (priv->dnd_timer);
	  priv->dnd_timer = 0;
	}
    }
}
#endif

static gint
mod_notebook_button_release (GtkWidget      *widget,
			     GdkEventButton *event)
{
  ModNotebook *notebook;
  ModNotebookPrivate *priv;
  ModNotebookPage *page;

  g_return_val_if_fail (MOD_IS_NOTEBOOK (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  if (event->type != GDK_BUTTON_RELEASE)
    return FALSE;

  notebook = MOD_NOTEBOOK (widget);

  if (notebook->click_close_button)
    {
      gint x, y;

      notebook->click_close_button = FALSE;
      mod_notebook_redraw_close_button (notebook);

      if (get_widget_coordinates (widget, (GdkEvent *)event, &x, &y))
	{
	  if (mod_notebook_in_close_button (notebook, x, y))
	    {
	      g_signal_emit (notebook,
			     notebook_signals[CLOSE_BUTTON_CLICKED],
			     0);
	      return TRUE;
	    }
	}
      return FALSE;
    }

  priv = MOD_NOTEBOOK_GET_PRIVATE (notebook);
  page = notebook->cur_page;

#if GTK_CHECK_VERSION(2,10,0)
  if (!priv->during_detach &&
      page->reorderable &&
      event->button == priv->pressed_button)
    mod_notebook_stop_reorder (notebook);
#endif

  if (event->button == notebook->button)
    {
      stop_scrolling (notebook);
      return TRUE;
    }
  else
    return FALSE;
}

static gint
mod_notebook_leave_notify (GtkWidget        *widget,
			   GdkEventCrossing *event)
{
  ModNotebook *notebook = MOD_NOTEBOOK (widget);
  ModNotebookArrow arrow;
  gint x, y;

  if (!get_widget_coordinates (widget, (GdkEvent *)event, &x, &y))
    return FALSE;

  if (notebook->in_child)
    {
      notebook->in_child = 0;
      mod_notebook_redraw_arrows (notebook);
    }

  if (notebook->in_close_button)
    {
      notebook->in_close_button = FALSE;
      mod_notebook_redraw_close_button (notebook);
    }

  arrow = mod_notebook_get_arrow (notebook, x, y);
  if (arrow == ARROW_NONE && notebook->active_tips_page)
    {
#if GTK_CHECK_VERSION(2,18,0)
      gboolean use_sticky_delay =
	notebook->tip_window && gtk_widget_get_visible (notebook->tip_window);
#else
      gboolean use_sticky_delay =
	notebook->tip_window && GTK_WIDGET_VISIBLE (notebook->tip_window);
#endif
      mod_notebook_tooltips_set_active_page (notebook, NULL);
      notebook->use_sticky_delay = use_sticky_delay;
    }

  return TRUE;
}

#if GTK_CHECK_VERSION(2,10,0)
static ModNotebookPointerPosition
get_pointer_position (ModNotebook *notebook)
{
  GtkWidget *widget = (GtkWidget *) notebook;
  GtkContainer *container = (GtkContainer *) notebook;
  ModNotebookPrivate *priv = MOD_NOTEBOOK_GET_PRIVATE (notebook);
  gboolean is_rtl;

  if (!notebook->scrollable)
    return POINTER_BETWEEN;

  if (notebook->tab_pos == GTK_POS_TOP ||
      notebook->tab_pos == GTK_POS_BOTTOM)
    {
      gint x;

      is_rtl = gtk_widget_get_direction (widget) == GTK_TEXT_DIR_RTL;
      x = priv->mouse_x - widget->allocation.x;

      if (x > widget->allocation.width - 2 * container->border_width - SCROLL_THRESHOLD)
	return (is_rtl) ? POINTER_BEFORE : POINTER_AFTER;
      else if (x < SCROLL_THRESHOLD + container->border_width)
	return (is_rtl) ? POINTER_AFTER : POINTER_BEFORE;
      else
	return POINTER_BETWEEN;
    }
  else
    {
      gint y;

      y = priv->mouse_y - widget->allocation.y;
      if (y > widget->allocation.height - 2 * container->border_width - SCROLL_THRESHOLD)
	return POINTER_AFTER;
      else if (y < SCROLL_THRESHOLD + container->border_width)
	return POINTER_BEFORE;
      else
	return POINTER_BETWEEN;
    }
}

static gboolean
scroll_notebook_timer (gpointer data)
{
  ModNotebook *notebook = (ModNotebook *) data;
  ModNotebookPrivate *priv;
  ModNotebookPointerPosition pointer_position;
  GList *element, *first_tab;

  GDK_THREADS_ENTER ();

  priv = MOD_NOTEBOOK_GET_PRIVATE (notebook);
  pointer_position = get_pointer_position (notebook);

  element = get_drop_position (notebook, notebook->cur_page->pack);
  reorder_tab (notebook, element, notebook->focus_tab);
  first_tab = mod_notebook_search_page (notebook, notebook->first_tab,
					(pointer_position == POINTER_BEFORE) ? STEP_PREV : STEP_NEXT,
					TRUE);
  if (first_tab)
    {
      notebook->first_tab = first_tab;
      mod_notebook_pages_allocate (notebook);

      gdk_window_move_resize (priv->drag_window,
			      priv->drag_window_x,
			      priv->drag_window_y,
			      notebook->cur_page->allocation.width,
			      notebook->cur_page->allocation.height);
      gdk_window_raise (priv->drag_window);
    }

  GDK_THREADS_LEAVE ();

  return TRUE;
}

static gboolean
check_threshold (ModNotebook *notebook,
		 gint         current_x,
		 gint         current_y)
{
  GtkWidget *widget;
  gint dnd_threshold;
  GdkRectangle rectangle = { 0, }; /* shut up gcc */
  GtkSettings *settings;
  
  widget = GTK_WIDGET (notebook);
  settings = gtk_widget_get_settings (GTK_WIDGET (notebook));
  g_object_get (G_OBJECT (settings), "gtk-dnd-drag-threshold", &dnd_threshold, NULL);

  /* we want a large threshold */
  dnd_threshold *= DND_THRESHOLD_MULTIPLIER;

  gdk_window_get_position (notebook->event_window, &rectangle.x, &rectangle.y);
  gdk_drawable_get_size (GDK_DRAWABLE (notebook->event_window), &rectangle.width, &rectangle.height);

  rectangle.x -= dnd_threshold;
  rectangle.width += 2 * dnd_threshold;
  rectangle.y -= dnd_threshold;
  rectangle.height += 2 * dnd_threshold;

  return (current_x < rectangle.x ||
	  current_x > rectangle.x + rectangle.width ||
	  current_y < rectangle.y ||
	  current_y > rectangle.y + rectangle.height);
}
#endif

static gint
mod_notebook_motion_notify (GtkWidget      *widget,
			    GdkEventMotion *event)
{
  ModNotebook *notebook = MOD_NOTEBOOK (widget);
  ModNotebookPrivate *priv = MOD_NOTEBOOK_GET_PRIVATE (notebook);
  ModNotebookPage *page;
  ModNotebookArrow arrow;
#if GTK_CHECK_VERSION(2,10,0)
  ModNotebookPointerPosition pointer_position;
  GtkSettings *settings;
  guint timeout;
#endif

  page = notebook->cur_page;

  if (!page)
    return FALSE;

  if (!(event->state & GDK_BUTTON1_MASK) &&
      priv->pressed_button != -1)
    {
#if GTK_CHECK_VERSION(2,10,0)
      mod_notebook_stop_reorder (notebook);
#endif
      stop_scrolling (notebook);
    }

  if (event->time < priv->timestamp + MSECS_BETWEEN_UPDATES)
    return FALSE;

  priv->timestamp = event->time;
  gdk_window_get_pointer (widget->window,
			  &priv->mouse_x,
			  &priv->mouse_y,
			  NULL);

  if (mod_notebook_in_close_button (notebook, priv->mouse_x, priv->mouse_y))
    {
      if (!notebook->in_close_button)
	{
	  notebook->in_close_button = TRUE;
	  mod_notebook_redraw_close_button (notebook);
	}
    }
  else if (notebook->in_close_button)
    {
      notebook->in_close_button = FALSE;
      mod_notebook_redraw_close_button (notebook);
    }

  arrow = mod_notebook_get_arrow (notebook, priv->mouse_x, priv->mouse_y);
  if (arrow != notebook->in_child)
    {
      notebook->in_child = arrow;
      mod_notebook_redraw_arrows (notebook);
    }

  if (priv->pressed_button == -1)
    {
      if (arrow == ARROW_NONE && !notebook->in_close_button)
	{
	  ModNotebookPage *page =
	    mod_notebook_get_tab_page (notebook, priv->mouse_x, priv->mouse_y);
	  if (page)
	    mod_notebook_tooltips_start_delay (notebook, page);
	  else
	    mod_notebook_tooltips_set_active_page (notebook, NULL);
	}

      return FALSE;
    }

#if GTK_CHECK_VERSION(2,10,0)
  if (page->detachable &&
      check_threshold (notebook, priv->mouse_x, priv->mouse_y))
    {
      priv->detached_tab = notebook->cur_page;
      priv->during_detach = TRUE;

      gtk_drag_begin (widget, priv->source_targets, GDK_ACTION_MOVE,
		      priv->pressed_button, (GdkEvent*) event);
      return TRUE;
    }

  if (page->reorderable &&
      (priv->during_reorder ||
       gtk_drag_check_threshold (widget, priv->drag_begin_x, priv->drag_begin_y, priv->mouse_x, priv->mouse_y)))
    {
      priv->during_reorder = TRUE;
      pointer_position = get_pointer_position (notebook);

      if (event->window == priv->drag_window &&
	  pointer_position != POINTER_BETWEEN &&
	  mod_notebook_show_arrows (notebook))
	{
	  /* scroll tabs */
	  if (!priv->dnd_timer)
	    {
	      priv->has_scrolled = TRUE;
	      settings = gtk_widget_get_settings (GTK_WIDGET (notebook));
	      g_object_get (settings, "gtk-timeout-repeat", &timeout, NULL);

	      priv->dnd_timer = g_timeout_add (timeout * SCROLL_DELAY_FACTOR,
					       (GSourceFunc) scroll_notebook_timer, 
					       (gpointer) notebook);
	    }
	}
      else
	{
	  if (priv->dnd_timer)
	    {
	      g_source_remove (priv->dnd_timer);
	      priv->dnd_timer = 0;
	    }
	}

      if (event->window == priv->drag_window ||
	  priv->operation != DRAG_OPERATION_REORDER)
	{
	  /* the drag operation is beginning, create the window */
	  if (priv->operation != DRAG_OPERATION_REORDER)
	    {
	      priv->operation = DRAG_OPERATION_REORDER;
	      show_drag_window (notebook, priv, page);
	    }

	  mod_notebook_pages_allocate (notebook);
	  gdk_window_move_resize (priv->drag_window,
				  priv->drag_window_x,
				  priv->drag_window_y,
				  page->allocation.width,
				  page->allocation.height);
	}
    }
#endif

  return TRUE;
}

static void
mod_notebook_grab_notify (GtkWidget *widget,
			  gboolean   was_grabbed)
{
  ModNotebook *notebook = MOD_NOTEBOOK (widget);

  if (!was_grabbed)
    {
#if GTK_CHECK_VERSION(2,10,0)
      mod_notebook_stop_reorder (notebook);
#endif
      stop_scrolling (notebook);
    }
}

static void
mod_notebook_state_changed (GtkWidget    *widget,
			    GtkStateType  previous_state)
{
#if GTK_CHECK_VERSION(2,18,0)
  if (!gtk_widget_is_sensitive (widget)) 
    stop_scrolling (MOD_NOTEBOOK (widget));
#else
  if (!GTK_WIDGET_IS_SENSITIVE (widget)) 
    stop_scrolling (MOD_NOTEBOOK (widget));
#endif
}

static gint
mod_notebook_focus_in (GtkWidget     *widget,
		       GdkEventFocus *event)
{
  MOD_NOTEBOOK (widget)->child_has_focus = FALSE;

  mod_notebook_redraw_tabs (MOD_NOTEBOOK (widget));

  return FALSE;
}

static gint
mod_notebook_focus_out (GtkWidget     *widget,
			GdkEventFocus *event)
{
  mod_notebook_redraw_tabs (MOD_NOTEBOOK (widget));

  return FALSE;
}

static void
mod_notebook_draw_focus (GtkWidget *widget)
{
  ModNotebook *notebook = MOD_NOTEBOOK (widget);

#if GTK_CHECK_VERSION(2,18,0)
  if (gtk_widget_is_drawable (widget) && notebook->show_tabs &&
      notebook->focus_tab)
#else
  if (GTK_WIDGET_DRAWABLE (widget) && notebook->show_tabs &&
      notebook->focus_tab)
#endif
    {
      ModNotebookPage *page;
      GdkRectangle area;
      gint focus_width;
      
      gtk_widget_style_get (widget, "focus-line-width", &focus_width, NULL);

      page = notebook->focus_tab->data;

      area.x = page->tab_label->allocation.x - focus_width;
      area.y = page->tab_label->allocation.y - focus_width;
      area.width = page->tab_label->allocation.width + 2 * focus_width;
      area.height = page->tab_label->allocation.height + 2 * focus_width;

      mod_notebook_draw_tab (MOD_NOTEBOOK (widget), page, &area);
    }
}

static void
mod_notebook_style_set  (GtkWidget *widget,
			 GtkStyle  *previous)
{
  ModNotebook *notebook;

  gboolean has_before_previous;
  gboolean has_before_next;
  gboolean has_after_previous;
  gboolean has_after_next;

  notebook = MOD_NOTEBOOK (widget);

  gtk_widget_style_get (widget,
                        "has-backward-stepper", &has_before_previous,
                        "has-secondary-forward-stepper", &has_before_next,
                        "has-secondary-backward-stepper", &has_after_previous,
                        "has-forward-stepper", &has_after_next,
                        NULL);

  notebook->has_before_previous = has_before_previous;
  notebook->has_before_next = has_before_next;
  notebook->has_after_previous = has_after_previous;
  notebook->has_after_next = has_after_next;

  (* GTK_WIDGET_CLASS (mod_notebook_parent_class)->style_set) (widget, previous);
}

#if GTK_CHECK_VERSION(2,10,0)
static gboolean
on_drag_icon_expose (GtkWidget      *widget,
		     GdkEventExpose *event,
		     gpointer        data)
{
  GtkWidget *notebook, *child = GTK_WIDGET (data);
  GtkRequisition requisition;
  gint gap_pos;

  notebook = GTK_WIDGET (data);
  child = GTK_BIN (widget)->child;
  gtk_widget_size_request (widget, &requisition);
  gap_pos = get_tab_gap_pos (MOD_NOTEBOOK (notebook));

  gtk_paint_extension (notebook->style, widget->window,
		       GTK_STATE_NORMAL, GTK_SHADOW_OUT,
		       NULL, widget, "tab",
		       0, 0,
		       requisition.width, requisition.height,
		       gap_pos);
  if (child)
    gtk_container_propagate_expose (GTK_CONTAINER (widget), child, event);

  return TRUE;
}

static void
mod_notebook_drag_begin (GtkWidget        *widget,
			 GdkDragContext   *context)
{
  ModNotebookPrivate *priv = MOD_NOTEBOOK_GET_PRIVATE (widget);
  ModNotebook *notebook = (ModNotebook*) widget;
  GtkWidget *tab_label;

  if (priv->dnd_timer)
    {
      g_source_remove (priv->dnd_timer);
      priv->dnd_timer = 0;
    }

  priv->operation = DRAG_OPERATION_DETACH;
  mod_notebook_pages_allocate (notebook);

  tab_label = priv->detached_tab->tab_label;

  hide_drag_window (notebook, priv, notebook->cur_page);
  g_object_ref (tab_label);
  gtk_widget_unparent (tab_label);

  priv->dnd_window = gtk_window_new (GTK_WINDOW_POPUP);
  gtk_container_add (GTK_CONTAINER (priv->dnd_window), tab_label);
  gtk_widget_set_size_request (priv->dnd_window,
			       priv->detached_tab->allocation.width,
			       priv->detached_tab->allocation.height);
  g_object_unref (tab_label);

  g_signal_connect (G_OBJECT (priv->dnd_window), "expose-event",
		    G_CALLBACK (on_drag_icon_expose), notebook);

  gtk_drag_set_icon_widget (context, priv->dnd_window, -2, -2);
}

static void
mod_notebook_drag_end (GtkWidget      *widget,
		       GdkDragContext *context)
{
  ModNotebookPrivate *priv = MOD_NOTEBOOK_GET_PRIVATE (widget);

  mod_notebook_stop_reorder (MOD_NOTEBOOK (widget));

  GTK_BIN (priv->dnd_window)->child = NULL;
  gtk_widget_destroy (priv->dnd_window);
  priv->dnd_window = NULL;

  priv->operation = DRAG_OPERATION_NONE;
}

static gboolean
mod_notebook_switch_tab_timeout (gpointer data)
{
  ModNotebook *notebook;
  ModNotebookPrivate *priv;
  GList *tab;
  gint x, y;

  GDK_THREADS_ENTER ();

  notebook = MOD_NOTEBOOK (data);
  priv = MOD_NOTEBOOK_GET_PRIVATE (notebook);

  priv->switch_tab_timer = 0;
  x = priv->mouse_x;
  y = priv->mouse_y;

  if ((tab = get_tab_at_pos (notebook, x, y)) != NULL)
    {
      /* FIXME: hack, we don't want the
       * focus to move fom the source widget
       */
      notebook->child_has_focus = FALSE;
      mod_notebook_switch_focus_tab (notebook, tab);
    }

  GDK_THREADS_LEAVE ();

  return FALSE;
}

static gboolean
mod_notebook_drag_motion (GtkWidget      *widget,
			  GdkDragContext *context,
			  gint            x,
			  gint            y,
			  guint           time)
{
  ModNotebook *notebook;
  ModNotebookPrivate *priv;
  GdkRectangle position;
  GtkSettings *settings;
  ModNotebookArrow arrow;
  guint timeout;
  GdkAtom target, tab_target;

  notebook = MOD_NOTEBOOK (widget);
  arrow = mod_notebook_get_arrow (notebook,
				  x + widget->allocation.x,
				  y + widget->allocation.y);
  if (arrow)
    {
      notebook->click_child = arrow;
      mod_notebook_set_scroll_timer (notebook);
      gdk_drag_status (context, 0, time);
      return TRUE;
    }

  stop_scrolling (notebook);
  target = gtk_drag_dest_find_target (widget, context, NULL);
  tab_target = gdk_atom_intern_static_string ("MOD_NOTEBOOK_TAB");

  if (target == tab_target)
    {
      gint widget_group, source_widget_group;
      GtkWidget *source_widget;

      source_widget = gtk_drag_get_source_widget (context);
      g_assert (source_widget);

      widget_group = mod_notebook_get_group_id (notebook);
      source_widget_group = mod_notebook_get_group_id (MOD_NOTEBOOK (source_widget));

      if (widget_group != -1 &&
	  source_widget_group != -1 &&
	  widget_group == source_widget_group &&
	  !(widget == MOD_NOTEBOOK (source_widget)->cur_page->child || 
	    gtk_widget_is_ancestor (widget, MOD_NOTEBOOK (source_widget)->cur_page->child)))
	{
	  gdk_drag_status (context, GDK_ACTION_MOVE, time);
	  return TRUE;
	}
    }

  priv = MOD_NOTEBOOK_GET_PRIVATE (widget);
  x += widget->allocation.x;
  y += widget->allocation.y;

  if (target == tab_target)
    {
      /* it's a tab, but doesn't share
       * ID with this notebook */
      gdk_drag_status (context, 0, time);
    }

  if (mod_notebook_get_event_window_position (notebook, &position) &&
      x >= position.x && x <= position.x + position.width &&
      y >= position.y && y <= position.y + position.height)
    {
      priv->mouse_x = x;
      priv->mouse_y = y;

      if (!priv->switch_tab_timer)
	{
          settings = gtk_widget_get_settings (widget);

          g_object_get (settings, "gtk-timeout-expand", &timeout, NULL);
	  priv->switch_tab_timer = g_timeout_add (timeout,
						  (GSourceFunc) mod_notebook_switch_tab_timeout,
						  widget);
	}
    }
  else
    {
      if (priv->switch_tab_timer)
	{
	  g_source_remove (priv->switch_tab_timer);
	  priv->switch_tab_timer = 0;
	}
    }

  return TRUE;
}

static void
mod_notebook_drag_leave (GtkWidget      *widget,
			 GdkDragContext *context,
			 guint           time)
{
  ModNotebookPrivate *priv;

  priv = MOD_NOTEBOOK_GET_PRIVATE (widget);

  if (priv->switch_tab_timer)
    {
      g_source_remove (priv->switch_tab_timer);
      priv->switch_tab_timer = 0;
    }

  stop_scrolling (MOD_NOTEBOOK (widget));
}

static gboolean
mod_notebook_drag_drop (GtkWidget        *widget,
			GdkDragContext   *context,
			gint              x,
			gint              y,
			guint             time)
{
  GdkAtom target;

  target = gtk_drag_dest_find_target (widget, context, NULL);

  if (target == GDK_NONE)
    gtk_drag_finish (context, FALSE, FALSE, time);

  return TRUE;
}

static void
do_detach_tab (ModNotebook     *from,
	       ModNotebook     *to,
	       GtkWidget       *child,
	       gint             x,
	       gint             y)
{
  ModNotebookPrivate *priv;
  GtkWidget *tab_label, *menu_label;
  gboolean tab_expand, tab_fill, reorderable, detachable;
  GList *element;
  guint tab_pack;
  gint page_num;

  menu_label = mod_notebook_get_menu_label (from, child);

  if (menu_label)
    g_object_ref (menu_label);

  tab_label = mod_notebook_get_tab_label (from, child);
  
  if (tab_label)
    g_object_ref (tab_label);

  g_object_ref (child);

  gtk_container_child_get (GTK_CONTAINER (from),
			   child,
			   "tab-expand", &tab_expand,
			   "tab-fill", &tab_fill,
			   "tab-pack", &tab_pack,
			   "reorderable", &reorderable,
			   "detachable", &detachable,
			   NULL);

  gtk_container_remove (GTK_CONTAINER (from), child);

  priv = MOD_NOTEBOOK_GET_PRIVATE (to);
  priv->mouse_x = x + GTK_WIDGET (to)->allocation.x;
  priv->mouse_y = y + GTK_WIDGET (to)->allocation.y;

  element = get_drop_position (to, tab_pack);
  page_num = g_list_position (to->children, element);
  mod_notebook_insert_page_menu (to, child, tab_label, menu_label, page_num);

  gtk_container_child_set (GTK_CONTAINER (to), child,
			   "tab-pack", tab_pack,
			   "tab-expand", tab_expand,
			   "tab-fill", tab_fill,
			   "reorderable", reorderable,
			   "detachable", detachable,
			   NULL);
  if (child)
    g_object_unref (child);

  if (tab_label)
    g_object_unref (tab_label);

  if (menu_label)
    g_object_unref (menu_label);

  mod_notebook_set_current_page (to, page_num);
}

static void
mod_notebook_drag_data_get (GtkWidget        *widget,
			    GdkDragContext   *context,
			    GtkSelectionData *data,
			    guint             info,
			    guint             time)
{
  ModNotebook *dest_notebook, *notebook;
  ModNotebookPrivate *priv;

  if (data->target != gdk_atom_intern_static_string ("MOD_NOTEBOOK_TAB") &&
      (data->target != gdk_atom_intern_static_string ("application/x-rootwindow-drop") ||
       !window_creation_hook))
    return;

  notebook = MOD_NOTEBOOK (widget);
  priv = MOD_NOTEBOOK_GET_PRIVATE (notebook);

  if (data->target == gdk_atom_intern_static_string ("MOD_NOTEBOOK_TAB"))
    {
      gtk_selection_data_set (data,
			      data->target,
			      8,
			      (void*) &priv->detached_tab->child,
			      sizeof (gpointer));
    }
  else
    {
      GdkDisplay *display;
      gint x, y;

      display = gtk_widget_get_display (widget);
      gdk_display_get_pointer (display, NULL, &x, &y, NULL);

      dest_notebook = (* window_creation_hook) (notebook,
						priv->detached_tab->child,
						x, y,
						window_creation_hook_data);
      if (dest_notebook)
	do_detach_tab (notebook, dest_notebook, priv->detached_tab->child, 0, 0);
    }
}

static void
mod_notebook_drag_data_received (GtkWidget        *widget,
				 GdkDragContext   *context,
				 gint              x,
				 gint              y,
				 GtkSelectionData *data,
				 guint             info,
				 guint             time)
{
  ModNotebook *notebook;
  GtkWidget *source_widget;
  GtkWidget **child;

  notebook = MOD_NOTEBOOK (widget);
  source_widget = gtk_drag_get_source_widget (context);

  if (source_widget &&
      data->target == gdk_atom_intern_static_string ("MOD_NOTEBOOK_TAB"))
    {
      child = (void*) data->data;

      do_detach_tab (MOD_NOTEBOOK (source_widget), notebook, *child, x, y);
      gtk_drag_finish (context, TRUE, FALSE, time);
    }
  else
    gtk_drag_finish (context, FALSE, FALSE, time);
}
#endif

/* Private GtkContainer Methods :
 * 
 * mod_notebook_set_child_arg
 * mod_notebook_get_child_arg
 * mod_notebook_add
 * mod_notebook_remove
 * mod_notebook_focus
 * mod_notebook_set_focus_child
 * mod_notebook_child_type
 * mod_notebook_forall
 */
static void
mod_notebook_set_child_property (GtkContainer    *container,
				 GtkWidget       *child,
				 guint            property_id,
				 const GValue    *value,
				 GParamSpec      *pspec)
{
  gboolean expand;
  gboolean fill;
  GtkPackType pack_type;

  /* not finding child's page is valid for menus or labels */
  if (!mod_notebook_find_child (MOD_NOTEBOOK (container), child, NULL))
    return;

  switch (property_id)
    {
    case CHILD_PROP_TAB_LABEL:
      /* a NULL pointer indicates a default_tab setting, otherwise
       * we need to set the associated label
       */
      mod_notebook_set_tab_label_text (MOD_NOTEBOOK (container), child,
				       g_value_get_string (value));
      break;
    case CHILD_PROP_MENU_LABEL:
      mod_notebook_set_menu_label_text (MOD_NOTEBOOK (container), child,
					g_value_get_string (value));
      break;
    case CHILD_PROP_POSITION:
      mod_notebook_reorder_child (MOD_NOTEBOOK (container), child,
				  g_value_get_int (value));
      break;
    case CHILD_PROP_TAB_EXPAND:
      mod_notebook_query_tab_label_packing (MOD_NOTEBOOK (container), child,
					    &expand, &fill, &pack_type);
      mod_notebook_set_tab_label_packing (MOD_NOTEBOOK (container), child,
					  g_value_get_boolean (value),
					  fill, pack_type);
      break;
    case CHILD_PROP_TAB_FILL:
      mod_notebook_query_tab_label_packing (MOD_NOTEBOOK (container), child,
					    &expand, &fill, &pack_type);
      mod_notebook_set_tab_label_packing (MOD_NOTEBOOK (container), child,
					  expand,
					  g_value_get_boolean (value),
					  pack_type);
      break;
    case CHILD_PROP_TAB_PACK:
      mod_notebook_query_tab_label_packing (MOD_NOTEBOOK (container), child,
					    &expand, &fill, &pack_type);
      mod_notebook_set_tab_label_packing (MOD_NOTEBOOK (container), child,
					  expand, fill,
					  g_value_get_enum (value));
      break;
    case CHILD_PROP_REORDERABLE:
      mod_notebook_set_tab_reorderable (MOD_NOTEBOOK (container), child,
					g_value_get_boolean (value));
      break;
    case CHILD_PROP_DETACHABLE:
      mod_notebook_set_tab_detachable (MOD_NOTEBOOK (container), child,
				       g_value_get_boolean (value));
      break;
    default:
      GTK_CONTAINER_WARN_INVALID_CHILD_PROPERTY_ID (container, property_id, pspec);
      break;
    }
}

static void
mod_notebook_get_child_property (GtkContainer    *container,
				 GtkWidget       *child,
				 guint            property_id,
				 GValue          *value,
				 GParamSpec      *pspec)
{
  GList *list;
  ModNotebook *notebook;
  GtkWidget *label;
  gboolean expand;
  gboolean fill;
  GtkPackType pack_type;

  notebook = MOD_NOTEBOOK (container);

  /* not finding child's page is valid for menus or labels */
  list = mod_notebook_find_child (notebook, child, NULL);
  if (!list)
    {
      /* nothing to set on labels or menus */
      g_param_value_set_default (pspec, value);
      return;
    }

  switch (property_id)
    {
    case CHILD_PROP_TAB_LABEL:
      label = mod_notebook_get_tab_label (notebook, child);

      if (label && GTK_IS_LABEL (label))
	g_value_set_string (value, GTK_LABEL (label)->label);
      else
	g_value_set_string (value, NULL);
      break;
    case CHILD_PROP_MENU_LABEL:
      label = mod_notebook_get_menu_label (notebook, child);

      if (label && GTK_IS_LABEL (label))
	g_value_set_string (value, GTK_LABEL (label)->label);
      else
	g_value_set_string (value, NULL);
      break;
    case CHILD_PROP_POSITION:
      g_value_set_int (value, g_list_position (notebook->children, list));
      break;
    case CHILD_PROP_TAB_EXPAND:
	mod_notebook_query_tab_label_packing (MOD_NOTEBOOK (container), child,
					      &expand, NULL, NULL);
	g_value_set_boolean (value, expand);
      break;
    case CHILD_PROP_TAB_FILL:
	mod_notebook_query_tab_label_packing (MOD_NOTEBOOK (container), child,
					      NULL, &fill, NULL);
	g_value_set_boolean (value, fill);
      break;
    case CHILD_PROP_TAB_PACK:
	mod_notebook_query_tab_label_packing (MOD_NOTEBOOK (container), child,
					      NULL, NULL, &pack_type);
	g_value_set_enum (value, pack_type);
      break;
    case CHILD_PROP_REORDERABLE:
      g_value_set_boolean (value,
			   mod_notebook_get_tab_reorderable (MOD_NOTEBOOK (container), child));
      break;
    case CHILD_PROP_DETACHABLE:
      g_value_set_boolean (value,
			   mod_notebook_get_tab_detachable (MOD_NOTEBOOK (container), child));
      break;
    default:
      GTK_CONTAINER_WARN_INVALID_CHILD_PROPERTY_ID (container, property_id, pspec);
      break;
    }
}

static void
mod_notebook_add (GtkContainer *container,
		  GtkWidget    *widget)
{
  g_return_if_fail (MOD_IS_NOTEBOOK (container));

  mod_notebook_insert_page_menu (MOD_NOTEBOOK (container), widget, 
				 NULL, NULL, -1);
}

static void
mod_notebook_remove (GtkContainer *container,
		     GtkWidget    *widget)
{
  ModNotebook *notebook;
  ModNotebookPage *page;
  GList *children;
  gint page_num = 0;

  g_return_if_fail (MOD_IS_NOTEBOOK (container));
  g_return_if_fail (widget != NULL);

  notebook = MOD_NOTEBOOK (container);

  children = notebook->children;
  while (children)
    {
      page = children->data;

      if (page->child == widget)
	break;

      page_num++;
      children = children->next;
    }
 
  if (children == NULL)
    return;
  
  g_object_ref (widget);

  mod_notebook_real_remove (notebook, children);

  g_signal_emit (notebook,
		 notebook_signals[PAGE_REMOVED],
		 0,
		 widget,
		 page_num);
  
  g_object_unref (widget);
}

static gboolean
focus_tabs_in (ModNotebook *notebook)
{
  if (notebook->show_tabs && notebook->cur_page)
    {
      gtk_widget_grab_focus (GTK_WIDGET (notebook));

      mod_notebook_switch_focus_tab (notebook,
				     g_list_find (notebook->children,
						  notebook->cur_page));

      return TRUE;
    }
  else
    return FALSE;
}

static gboolean
focus_tabs_move (ModNotebook     *notebook,
		 GtkDirectionType direction,
		 gint             search_direction)
{
  GList *new_page;

  new_page = mod_notebook_search_page (notebook, notebook->focus_tab,
				       search_direction, TRUE);
  if (new_page)
    mod_notebook_switch_focus_tab (notebook, new_page);
  else
    gdk_display_beep (gtk_widget_get_display (GTK_WIDGET (notebook)));
  
  return TRUE;
}

static gboolean
focus_child_in (ModNotebook     *notebook,
		GtkDirectionType direction)
{
  if (notebook->cur_page)
    return gtk_widget_child_focus (notebook->cur_page->child, direction);
  else
    return FALSE;
}

/* Focus in the notebook can either be on the pages, or on
 * the tabs.
 */
static gint
mod_notebook_focus (GtkWidget        *widget,
		    GtkDirectionType  direction)
{
  GtkWidget *old_focus_child;
  ModNotebook *notebook;
  GtkDirectionType effective_direction;

  gboolean widget_is_focus;
  GtkContainer *container;

  g_return_val_if_fail (MOD_IS_NOTEBOOK (widget), FALSE);

  container = GTK_CONTAINER (widget);
  notebook = MOD_NOTEBOOK (container);

  if (notebook->focus_out)
    {
      notebook->focus_out = FALSE; /* Clear this to catch the wrap-around case */
      return FALSE;
    }

  widget_is_focus = gtk_widget_is_focus (widget);
  old_focus_child = container->focus_child; 

  effective_direction = get_effective_direction (notebook, direction);

  if (old_focus_child)		/* Focus on page child */
    {
      if (gtk_widget_child_focus (old_focus_child, direction))
        return TRUE;
      
      switch (effective_direction)
	{
	case GTK_DIR_TAB_BACKWARD:
	case GTK_DIR_UP:
	  /* Focus onto the tabs */
	  return focus_tabs_in (notebook);
	case GTK_DIR_DOWN:
	case GTK_DIR_TAB_FORWARD:
	case GTK_DIR_LEFT:
	case GTK_DIR_RIGHT:
	  return FALSE;
	}
    }
  else if (widget_is_focus)	/* Focus was on tabs */
    {
      switch (effective_direction)
	{
	case GTK_DIR_TAB_BACKWARD:
	case GTK_DIR_UP:
	  return FALSE;
	case GTK_DIR_TAB_FORWARD:
	case GTK_DIR_DOWN:
	  /* We use TAB_FORWARD rather than direction so that we focus a more
	   * predictable widget for the user; users may be using arrow focusing
	   * in this situation even if they don't usually use arrow focusing.
	   */
	  return focus_child_in (notebook, GTK_DIR_TAB_FORWARD);
	case GTK_DIR_LEFT:
	  return focus_tabs_move (notebook, direction, STEP_PREV);
	case GTK_DIR_RIGHT:
	  return focus_tabs_move (notebook, direction, STEP_NEXT);
	}
    }
  else /* Focus was not on widget */
    {
      switch (effective_direction)
	{
	case GTK_DIR_TAB_FORWARD:
	case GTK_DIR_DOWN:
	  if (focus_tabs_in (notebook))
	    return TRUE;
	  if (focus_child_in (notebook, direction))
	    return TRUE;
	  return FALSE;
	case GTK_DIR_TAB_BACKWARD:
	case GTK_DIR_UP:
	  if (focus_child_in (notebook, direction))
	    return TRUE;
	  if (focus_tabs_in (notebook))
	    return TRUE;
	  return FALSE;
	case GTK_DIR_LEFT:
	case GTK_DIR_RIGHT:
	  return focus_child_in (notebook, direction);
	}
    }

  g_assert_not_reached ();
  return FALSE;
}  

static void
mod_notebook_set_focus_child (GtkContainer *container,
			      GtkWidget    *child)
{
  ModNotebook *notebook = MOD_NOTEBOOK (container);
  GtkWidget *page_child;
  GtkWidget *toplevel;

  /* If the old focus widget was within a page of the notebook,
   * (child may either be NULL or not in this case), record it
   * for future use if we switch to the page with a mnemonic.
   */

  toplevel = gtk_widget_get_toplevel (GTK_WIDGET (container));
#if GTK_CHECK_VERSION(2,18,0)
  if (toplevel && gtk_widget_is_toplevel (toplevel))
#else
  if (toplevel && GTK_WIDGET_TOPLEVEL (toplevel))
#endif
    {
      page_child = GTK_WINDOW (toplevel)->focus_widget; 
      while (page_child)
	{
	  if (page_child->parent == GTK_WIDGET (container))
	    {
	      GList *list = mod_notebook_find_child (notebook, page_child, NULL);
	      if (list != NULL) 
		{
		  ModNotebookPage *page = list->data;
	      
		  if (page->last_focus_child)
		    g_object_remove_weak_pointer (G_OBJECT (page->last_focus_child), (gpointer *)&page->last_focus_child);
		  
		  page->last_focus_child = GTK_WINDOW (toplevel)->focus_widget;
		  g_object_add_weak_pointer (G_OBJECT (page->last_focus_child), (gpointer *)&page->last_focus_child);
	      
		  break;
		}
	    }

	  page_child = page_child->parent;
	}
    }
  
  if (child)
    {
      g_return_if_fail (GTK_IS_WIDGET (child));

      notebook->child_has_focus = TRUE;
      if (!notebook->focus_tab)
	{
	  GList *children;
	  ModNotebookPage *page;

	  children = notebook->children;
	  while (children)
	    {
	      page = children->data;
	      if (page->child == child || page->tab_label == child)
		mod_notebook_switch_focus_tab (notebook, children);
	      children = children->next;
	    }
	}
    }

  GTK_CONTAINER_CLASS (mod_notebook_parent_class)->set_focus_child (container, child);
}

static void
mod_notebook_forall (GtkContainer *container,
		     gboolean      include_internals,
		     GtkCallback   callback,
		     gpointer      callback_data)
{
  ModNotebook *notebook;
  GList *children;

  g_return_if_fail (MOD_IS_NOTEBOOK (container));
  g_return_if_fail (callback != NULL);

  notebook = MOD_NOTEBOOK (container);

  children = notebook->children;
  while (children)
    {
      ModNotebookPage *page;
      
      page = children->data;
      children = children->next;
      (* callback) (page->child, callback_data);

      if (include_internals)
	{
	  if (page->tab_label)
	    (* callback) (page->tab_label, callback_data);
	}
    }
}

static GType
mod_notebook_child_type (GtkContainer     *container)
{
  return GTK_TYPE_WIDGET;
}

/* Private ModNotebook Methods:
 *
 * mod_notebook_real_insert_page
 */
static void
page_visible_cb (GtkWidget  *page,
                 GParamSpec *arg,
                 gpointer    data)
{
  ModNotebook *notebook = (ModNotebook *) data;
  GList *list;
  GList *next = NULL;

  if (notebook->cur_page &&
      notebook->cur_page->child == page &&
#if GTK_CHECK_VERSION(2,18,0)
      !gtk_widget_get_visible (page))
#else
      !GTK_WIDGET_VISIBLE (page))
#endif
    {
      list = g_list_find (notebook->children, notebook->cur_page);
      if (list)
        {
          next = mod_notebook_search_page (notebook, list, STEP_NEXT, TRUE);
          if (!next)
            next = mod_notebook_search_page (notebook, list, STEP_PREV, TRUE);
        }

      if (next)
        mod_notebook_switch_page (notebook, MOD_NOTEBOOK_PAGE (next), -1);
    }
}

static gint
mod_notebook_real_insert_page (ModNotebook *notebook,
			       GtkWidget   *child,
			       GtkWidget   *tab_label,
			       GtkWidget   *menu_label,
			       gint         position)
{
  ModNotebookPage *page;
  gint nchildren;

  g_return_val_if_fail (MOD_IS_NOTEBOOK (notebook), -1);
  g_return_val_if_fail (GTK_IS_WIDGET (child), -1);
  g_return_val_if_fail (tab_label == NULL || GTK_IS_WIDGET (tab_label), -1);
  g_return_val_if_fail (menu_label == NULL || GTK_IS_WIDGET (menu_label), -1);

  gtk_widget_freeze_child_notify (child);
  
  page = g_new (ModNotebookPage, 1);
  page->child = child;
  page->last_focus_child = NULL;
  page->requisition.width = 0;
  page->requisition.height = 0;
  page->allocation.x = 0;
  page->allocation.y = 0;
  page->allocation.width = 0;
  page->allocation.height = 0;
  page->default_menu = FALSE;
  page->default_tab = FALSE;
  page->mnemonic_activate_signal = 0;
  page->reorderable = FALSE;
  page->detachable = FALSE;
  page->ref_count = 0;
  mod_notebook_page_ref(page);
   
  nchildren = g_list_length (notebook->children);
  if ((position < 0) || (position > nchildren))
    position = nchildren;

  notebook->children = g_list_insert (notebook->children, page, position);

  if (!tab_label)
    {
      page->default_tab = TRUE;
      if (notebook->show_tabs)
	tab_label = gtk_label_new ("");
    }

  page->tab_label = tab_label;
  page->menu_label = menu_label;
  page->expand = FALSE;
  page->fill = TRUE;
  page->pack = GTK_PACK_START; 

  if (!menu_label)
    page->default_menu = TRUE;
  else  
    {
#if GTK_CHECK_VERSION(2,10,0)
      g_object_ref_sink (page->menu_label);
#else
      g_object_ref (page->menu_label);
      gtk_object_sink (GTK_OBJECT (page->menu_label));
#endif
    }

  if (notebook->menu)
    mod_notebook_menu_item_create (notebook,
				   g_list_find (notebook->children, page));

  gtk_widget_set_parent (child, GTK_WIDGET (notebook));
  if (tab_label)
    gtk_widget_set_parent (tab_label, GTK_WIDGET (notebook));

  mod_notebook_update_labels (notebook);

  if (!notebook->first_tab)
    notebook->first_tab = notebook->children;

  /* child visible will be turned on by switch_page below */
  gtk_widget_set_child_visible (child, FALSE);
  
  if (tab_label)
    {
#if GTK_CHECK_VERSION(2,18,0)
      if (notebook->show_tabs && gtk_widget_get_visible (child))
#else
      if (notebook->show_tabs && GTK_WIDGET_VISIBLE (child))
#endif
	gtk_widget_show (tab_label);
      else
	gtk_widget_hide (tab_label);

    page->mnemonic_activate_signal =
      g_signal_connect (tab_label,
			"mnemonic_activate",
			G_CALLBACK (mod_notebook_mnemonic_activate_switch_page),
			notebook);
    }

  page->notify_visible_handler = g_signal_connect (child, "notify::visible",
						   G_CALLBACK (page_visible_cb), notebook);

  g_signal_emit (notebook,
		 notebook_signals[PAGE_ADDED],
		 0,
		 child,
		 position);

  if (!notebook->cur_page)
    {
      mod_notebook_switch_page (notebook, page, 0);
      mod_notebook_switch_focus_tab (notebook, NULL);
    }

  mod_notebook_update_tab_states (notebook);

  gtk_widget_child_notify (child, "tab-expand");
  gtk_widget_child_notify (child, "tab-fill");
  gtk_widget_child_notify (child, "tab-pack");
  gtk_widget_child_notify (child, "tab-label");
  gtk_widget_child_notify (child, "menu-label");
  gtk_widget_child_notify (child, "position");
  gtk_widget_thaw_child_notify (child);

  return position;
}

/* Private ModNotebook Functions:
 *
 * mod_notebook_redraw_tabs
 * mod_notebook_real_remove
 * mod_notebook_update_labels
 * mod_notebook_timer
 * mod_notebook_set_scroll_timer
 * mod_notebook_page_compare
 * mod_notebook_real_page_position
 * mod_notebook_search_page
 */
static void
mod_notebook_redraw_tabs (ModNotebook *notebook)
{
  GtkWidget *widget;
  ModNotebookPage *page;
  GdkRectangle redraw_rect;
  gint border;
  gint tab_pos = get_effective_tab_pos (notebook);

  widget = GTK_WIDGET (notebook);
  border = GTK_CONTAINER (notebook)->border_width;

#if GTK_CHECK_VERSION(2,18,0)
  if (!gtk_widget_get_mapped (widget) || !notebook->first_tab)
    return;
#else
  if (!GTK_WIDGET_MAPPED (notebook) || !notebook->first_tab)
    return;
#endif

  page = notebook->first_tab->data;

  redraw_rect.x = border;
  redraw_rect.y = border;

  switch (tab_pos)
    {
    case GTK_POS_BOTTOM:
      redraw_rect.y = widget->allocation.height - border -
	page->allocation.height - widget->style->ythickness;

      if (page != notebook->cur_page)
	redraw_rect.y -= widget->style->ythickness;
      /* fall through */
    case GTK_POS_TOP:
      redraw_rect.width = widget->allocation.width - 2 * border;
      redraw_rect.height = page->allocation.height + widget->style->ythickness;

      if (page != notebook->cur_page)
	redraw_rect.height += widget->style->ythickness;
      break;
    case GTK_POS_RIGHT:
      redraw_rect.x = widget->allocation.width - border -
	page->allocation.width - widget->style->xthickness;

      if (page != notebook->cur_page)
	redraw_rect.x -= widget->style->xthickness;
      /* fall through */
    case GTK_POS_LEFT:
      redraw_rect.width = page->allocation.width + widget->style->xthickness;
      redraw_rect.height = widget->allocation.height - 2 * border;

      if (page != notebook->cur_page)
	redraw_rect.width += widget->style->xthickness;
      break;
    }

  redraw_rect.x += widget->allocation.x;
  redraw_rect.y += widget->allocation.y;

  gdk_window_invalidate_rect (widget->window, &redraw_rect, TRUE);
}

static void
mod_notebook_redraw_arrows (ModNotebook *notebook)
{
#if GTK_CHECK_VERSION(2,18,0)
  if (gtk_widget_get_mapped (GTK_WIDGET (notebook)) &&
      mod_notebook_show_arrows (notebook))
#else
  if (GTK_WIDGET_MAPPED (notebook) && mod_notebook_show_arrows (notebook))
#endif
    {
      GdkRectangle rect;
      gint i;
      ModNotebookArrow arrow[4];

      arrow[0] = notebook->has_before_previous ? ARROW_LEFT_BEFORE : ARROW_NONE;
      arrow[1] = notebook->has_before_next ? ARROW_RIGHT_BEFORE : ARROW_NONE;
      arrow[2] = notebook->has_after_previous ? ARROW_LEFT_AFTER : ARROW_NONE;
      arrow[3] = notebook->has_after_next ? ARROW_RIGHT_AFTER : ARROW_NONE;

      for (i = 0; i < 4; i++)
	{
	  if (arrow[i] == ARROW_NONE)
	    continue;

	  mod_notebook_get_arrow_rect (notebook, &rect, arrow[i]);
	  gdk_window_invalidate_rect (GTK_WIDGET (notebook)->window,
				      &rect, FALSE);
	}
    }
}

static gint
mod_notebook_timer (ModNotebook *notebook)
{
  gboolean retval = FALSE;

  GDK_THREADS_ENTER ();

  if (notebook->timer)
    {
      mod_notebook_do_arrow (notebook, notebook->click_child);

      if (notebook->need_timer)
	{
          guint        timeout;

	  if (gtk_major_version > 3 ||
	      (gtk_major_version == 2 && gtk_minor_version > 8))
	    {
	      GtkSettings *settings;

	      settings = gtk_widget_get_settings (GTK_WIDGET (notebook));
	      g_object_get (settings, "gtk-timeout-repeat", &timeout, NULL);
	    }
	  else
	    timeout = DEFAULT_TIMEOUT_REPEAT;

	  notebook->need_timer = FALSE;
	  notebook->timer = g_timeout_add (timeout * SCROLL_DELAY_FACTOR,
					   (GSourceFunc) mod_notebook_timer,
					   (gpointer) notebook);
	}
      else
	retval = TRUE;
    }

  GDK_THREADS_LEAVE ();

  return retval;
}

static void
mod_notebook_set_scroll_timer (ModNotebook *notebook)
{
  GtkWidget *widget = GTK_WIDGET (notebook);

  if (!notebook->timer)
    {
      guint timeout;

      if (gtk_major_version > 3 ||
	  (gtk_major_version == 2 && gtk_minor_version > 8))
	{
	  GtkSettings *settings = gtk_widget_get_settings (widget);

	  g_object_get (settings, "gtk-timeout-initial", &timeout, NULL);
	}
      else
	timeout = DEFAULT_TIMEOUT_INITIAL;

      notebook->timer = g_timeout_add (timeout,
				       (GSourceFunc) mod_notebook_timer,
				       (gpointer) notebook);
      notebook->need_timer = TRUE;
    }
}

static gint
mod_notebook_page_compare (gconstpointer a,
			   gconstpointer b)
{
  return (((ModNotebookPage *) a)->child != b);
}

static GList*
mod_notebook_find_child (ModNotebook *notebook,
			 GtkWidget   *child,
			 const gchar *function)
{
  GList *list = g_list_find_custom (notebook->children, child,
				    mod_notebook_page_compare);

#ifndef G_DISABLE_CHECKS
  if (!list && function)
    g_warning ("%s: unable to find child %p in notebook %p",
	       function, child, notebook);
#endif

  return list;
}

static void
mod_notebook_remove_tab_label (ModNotebook     *notebook,
			       ModNotebookPage *page)
{
  if (page->tab_label)
    {
      if (page->mnemonic_activate_signal)
	g_signal_handler_disconnect (page->tab_label,
				     page->mnemonic_activate_signal);
      page->mnemonic_activate_signal = 0;

      gtk_widget_set_state (page->tab_label, GTK_STATE_NORMAL);
      gtk_widget_unparent (page->tab_label);
      page->tab_label = NULL;
    }
}

static void
mod_notebook_real_remove (ModNotebook *notebook,
			  GList       *list)
{
  ModNotebookPrivate *priv;
  ModNotebookPage *page;
  GList * next_list;
  gint need_resize = FALSE;

  gboolean destroying;

  priv = MOD_NOTEBOOK_GET_PRIVATE (notebook);
  destroying = GTK_OBJECT_FLAGS (notebook) & GTK_IN_DESTRUCTION;
  
  next_list = mod_notebook_search_page (notebook, list, STEP_NEXT, TRUE);
  if (!next_list)
    next_list = mod_notebook_search_page (notebook, list, STEP_PREV, TRUE);

  if (notebook->cur_page == list->data)
    { 
      notebook->cur_page = NULL;
      if (next_list && !destroying)
	mod_notebook_switch_page (notebook, MOD_NOTEBOOK_PAGE (next_list), -1);
    }

  if (priv->detached_tab == list->data)
    priv->detached_tab = NULL;

  if (list == notebook->first_tab)
    notebook->first_tab = next_list;
  if (list == notebook->focus_tab && !destroying)
    mod_notebook_switch_focus_tab (notebook, next_list);

  page = list->data;
  
  g_signal_handler_disconnect (page->child, page->notify_visible_handler); 

#if GTK_CHECK_VERSION(2,18,0)
  if (gtk_widget_get_visible (page->child) &&
      gtk_widget_get_visible (GTK_WIDGET (notebook)))
#else
  if (GTK_WIDGET_VISIBLE (page->child) && GTK_WIDGET_VISIBLE (notebook))
#endif
    need_resize = TRUE;

  gtk_widget_unparent (page->child);

  if (notebook->active_tips_page == page)
    {
      notebook->active_tips_page = NULL;
      mod_notebook_page_unref(page);
      if (GTK_IS_WIDGET(notebook->tip_window))
	gtk_widget_hide(notebook->tip_window);
    }

  mod_notebook_remove_tab_label (notebook, page);
  
  if (notebook->menu)
    {
      gtk_container_remove (GTK_CONTAINER (notebook->menu), 
			    page->menu_label->parent);
      gtk_widget_queue_resize (notebook->menu);
    }
  if (!page->default_menu)
    g_object_unref (page->menu_label);
  
  notebook->children = g_list_remove_link (notebook->children, list);
  g_list_free (list);

  if (page->last_focus_child)
    {
      g_object_remove_weak_pointer (G_OBJECT (page->last_focus_child), (gpointer *)&page->last_focus_child);
      page->last_focus_child = NULL;
    }
  
  mod_notebook_page_unref (page);

  mod_notebook_update_labels (notebook);
  if (need_resize)
    gtk_widget_queue_resize (GTK_WIDGET (notebook));
}

static void
mod_notebook_update_labels (ModNotebook *notebook)
{
  ModNotebookPage *page;
  GList *list;
  gchar string[32];
  gint page_num = 1;

  for (list = mod_notebook_search_page (notebook, NULL, STEP_NEXT, FALSE);
       list;
       list = mod_notebook_search_page (notebook, list, STEP_NEXT, FALSE))
    {
      page = list->data;
      g_snprintf (string, sizeof(string), _("Page %u"), page_num++);
      if (notebook->show_tabs)
	{
	  if (page->default_tab)
	    {
	      if (!page->tab_label)
		{
		  page->tab_label = gtk_label_new (string);
		  gtk_widget_set_parent (page->tab_label,
					 GTK_WIDGET (notebook));
		}
	      else
		gtk_label_set_text (GTK_LABEL (page->tab_label), string);
	    }

#if GTK_CHECK_VERSION(2,18,0)
	  if (gtk_widget_get_visible (page->child) &&
	      !gtk_widget_get_visible (page->tab_label))
	    gtk_widget_show (page->tab_label);
	  else if (!gtk_widget_get_visible (page->child) &&
		   gtk_widget_get_visible (page->tab_label))
	    gtk_widget_hide (page->tab_label);
#else
	  if (GTK_WIDGET_VISIBLE (page->child) &&
	      !GTK_WIDGET_VISIBLE (page->tab_label))
	    gtk_widget_show (page->tab_label);
	  else if (!GTK_WIDGET_VISIBLE (page->child) &&
		   GTK_WIDGET_VISIBLE (page->tab_label))
	    gtk_widget_hide (page->tab_label);
#endif
	}
      if (notebook->menu && page->default_menu)
	{
	  if (page->tab_label && GTK_IS_LABEL (page->tab_label))
	    gtk_label_set_text (GTK_LABEL (page->menu_label),
			   GTK_LABEL (page->tab_label)->label);
	  else
	    gtk_label_set_text (GTK_LABEL (page->menu_label), string);
	}
    }  
}

static gint
mod_notebook_real_page_position (ModNotebook *notebook,
				 GList       *list)
{
  GList *work;
  gint count_start;

  g_return_val_if_fail (MOD_IS_NOTEBOOK (notebook), -1);
  g_return_val_if_fail (list != NULL, -1);

  for (work = notebook->children, count_start = 0;
       work && work != list; work = work->next)
    if (MOD_NOTEBOOK_PAGE (work)->pack == GTK_PACK_START)
      count_start++;

  if (!work)
    return -1;

  if (MOD_NOTEBOOK_PAGE (list)->pack == GTK_PACK_START)
    return count_start;

  return (count_start + g_list_length (list) - 1);
}

static GList *
mod_notebook_search_page (ModNotebook *notebook,
			  GList       *list,
			  gint         direction,
			  gboolean     find_visible)
{
  ModNotebookPage *page = NULL;
  GList *old_list = NULL;
  gint flag = 0;

  g_return_val_if_fail (MOD_IS_NOTEBOOK (notebook), NULL);

  switch (direction)
    {
    case STEP_PREV:
      flag = GTK_PACK_END;
      break;

    case STEP_NEXT:
      flag = GTK_PACK_START;
      break;
    }

  if (list)
    page = list->data;

  if (!page || page->pack == flag)
    {
      if (list)
	{
	  old_list = list;
	  list = list->next;
	}
      else
	list = notebook->children;

      while (list)
	{
	  page = list->data;
#if GTK_CHECK_VERSION(2,18,0)
	  if (page->pack == flag &&
	      (!find_visible ||
	       (gtk_widget_get_visible (page->child) &&
		(!page->tab_label || NOTEBOOK_IS_TAB_LABEL_PARENT (notebook, page)))))
	    return list;
#else
	  if (page->pack == flag &&
	      (!find_visible ||
	       (GTK_WIDGET_VISIBLE (page->child) &&
		(!page->tab_label || NOTEBOOK_IS_TAB_LABEL_PARENT (notebook, page)))))
	    return list;
#endif
	  old_list = list;
	  list = list->next;
	}
      list = old_list;
    }
  else
    {
      old_list = list;
      list = list->prev;
    }
  while (list)
    {
      page = list->data;
#if GTK_CHECK_VERSION(2,18,0)
      if (page->pack != flag &&
	  (!find_visible ||
	   (gtk_widget_get_visible (page->child) &&
	    (!page->tab_label || NOTEBOOK_IS_TAB_LABEL_PARENT (notebook, page)))))
	return list;
#else
      if (page->pack != flag &&
	  (!find_visible ||
	   (GTK_WIDGET_VISIBLE (page->child) &&
	    (!page->tab_label || NOTEBOOK_IS_TAB_LABEL_PARENT (notebook, page)))))
	return list;
#endif
      old_list = list;
      list = list->prev;
    }
  return NULL;
}

/* Private ModNotebook Drawing Functions:
 *
 * mod_notebook_paint
 * mod_notebook_draw_tab
 * mod_notebook_draw_arrow
 */
static void
mod_notebook_paint (GtkWidget    *widget,
		    GdkRectangle *area)
{
  ModNotebook *notebook;
  ModNotebookPrivate *priv;
  ModNotebookPage *page;
  GList *children;
  gboolean showarrow;
  gint width, height;
  gint x, y;
  gint border_width = GTK_CONTAINER (widget)->border_width;
  gint gap_x = 0, gap_width = 0, step = STEP_PREV;
  gboolean is_rtl;
  gint tab_pos;
   
  g_return_if_fail (MOD_IS_NOTEBOOK (widget));
  g_return_if_fail (area != NULL);

#if GTK_CHECK_VERSION(2,18,0)
  if (!gtk_widget_is_drawable (widget))
    return;
#else
  if (!GTK_WIDGET_DRAWABLE (widget))
    return;
#endif

  notebook = MOD_NOTEBOOK (widget);
  priv = MOD_NOTEBOOK_GET_PRIVATE (notebook);
  is_rtl = gtk_widget_get_direction (widget) == GTK_TEXT_DIR_RTL;
  tab_pos = get_effective_tab_pos (notebook);

#if GTK_CHECK_VERSION(2,18,0)
  if ((!notebook->show_tabs && !notebook->show_border) ||
      !notebook->cur_page || !gtk_widget_get_visible (notebook->cur_page->child))
    return;
#else
  if ((!notebook->show_tabs && !notebook->show_border) ||
      !notebook->cur_page || !GTK_WIDGET_VISIBLE (notebook->cur_page->child))
    return;
#endif

  x = widget->allocation.x + border_width;
  y = widget->allocation.y + border_width;
  width = widget->allocation.width - border_width * 2;
  height = widget->allocation.height - border_width * 2;

  if (notebook->show_border && (!notebook->show_tabs || !notebook->children))
    {
      gtk_paint_box (widget->style, widget->window,
		     GTK_STATE_NORMAL, GTK_SHADOW_OUT,
		     area, widget, "notebook",
		     x, y, width, height);
      return;
    }

  if (!notebook->first_tab)
    notebook->first_tab = notebook->children;

#if GTK_CHECK_VERSION(2,18,0)
  if (!NOTEBOOK_IS_TAB_LABEL_PARENT (notebook, notebook->cur_page) ||
      !gtk_widget_get_mapped (notebook->cur_page->tab_label))
    page = MOD_NOTEBOOK_PAGE (notebook->first_tab);
  else
    page = notebook->cur_page;
#else
  if (!NOTEBOOK_IS_TAB_LABEL_PARENT (notebook, notebook->cur_page) ||
      !GTK_WIDGET_MAPPED (notebook->cur_page->tab_label))
    page = MOD_NOTEBOOK_PAGE (notebook->first_tab);
  else
    page = notebook->cur_page;
#endif

  switch (tab_pos)
    {
    case GTK_POS_TOP:
      y += page->allocation.height;
      /* fall thru */
    case GTK_POS_BOTTOM:
      height -= page->allocation.height;
      break;
    case GTK_POS_LEFT:
      x += page->allocation.width;
      /* fall thru */
    case GTK_POS_RIGHT:
      width -= page->allocation.width;
      break;
    }

#if GTK_CHECK_VERSION(2,18,0)
  if (!NOTEBOOK_IS_TAB_LABEL_PARENT (notebook, notebook->cur_page) ||
      !gtk_widget_get_mapped (notebook->cur_page->tab_label))
#else
  if (!NOTEBOOK_IS_TAB_LABEL_PARENT (notebook, notebook->cur_page) ||
      !GTK_WIDGET_MAPPED (notebook->cur_page->tab_label))
#endif
    {
      gap_x = 0;
      gap_width = 0;
    }
  else
    {
      switch (tab_pos)
	{
	case GTK_POS_TOP:
	case GTK_POS_BOTTOM:
	  if (priv->operation == DRAG_OPERATION_REORDER)
	    gap_x = priv->drag_window_x - widget->allocation.x - border_width;
	  else
	    gap_x = notebook->cur_page->allocation.x - widget->allocation.x - border_width;

	  gap_width = notebook->cur_page->allocation.width;
	  step = is_rtl ? STEP_NEXT : STEP_PREV;
	  break;
	case GTK_POS_LEFT:
	case GTK_POS_RIGHT:
	  if (priv->operation == DRAG_OPERATION_REORDER)
	    gap_x = priv->drag_window_y - border_width - widget->allocation.y;
	  else
	    gap_x = notebook->cur_page->allocation.y - widget->allocation.y - border_width;

	  gap_width = notebook->cur_page->allocation.height;
	  step = STEP_PREV;
	  break;
	}
    }
  gtk_paint_box_gap (widget->style, widget->window,
		     GTK_STATE_NORMAL, GTK_SHADOW_OUT,
		     area, widget, "notebook",
		     x, y, width, height,
		     tab_pos, gap_x, gap_width);

  if (notebook->enable_close_button)
    mod_notebook_draw_close_button (notebook);

  showarrow = FALSE;
  children = mod_notebook_search_page (notebook, NULL, step, TRUE);
  while (children)
    {
      page = children->data;
      children = mod_notebook_search_page (notebook, children,
					   step, TRUE);
#if GTK_CHECK_VERSION(2,18,0)
      if (!gtk_widget_get_visible (page->child))
	continue;
      if (!gtk_widget_get_mapped (page->tab_label))
	showarrow = TRUE;
#else
      if (!GTK_WIDGET_VISIBLE (page->child))
	continue;
      if (!GTK_WIDGET_MAPPED (page->tab_label))
	showarrow = TRUE;
#endif
      else if (page != notebook->cur_page)
	mod_notebook_draw_tab (notebook, page, area);
    }

  if (showarrow && notebook->scrollable) 
    {
      if (notebook->has_before_previous)
	mod_notebook_draw_arrow (notebook, ARROW_LEFT_BEFORE);
      if (notebook->has_before_next)
	mod_notebook_draw_arrow (notebook, ARROW_RIGHT_BEFORE);
      if (notebook->has_after_previous)
	mod_notebook_draw_arrow (notebook, ARROW_LEFT_AFTER);
      if (notebook->has_after_next)
	mod_notebook_draw_arrow (notebook, ARROW_RIGHT_AFTER);
    }
  mod_notebook_draw_tab (notebook, notebook->cur_page, area);
}

static void
mod_notebook_draw_tab (ModNotebook     *notebook,
		       ModNotebookPage *page,
		       GdkRectangle    *area)
{
  ModNotebookPrivate *priv;
  GdkRectangle child_area;
  GdkRectangle page_area;
  GtkStateType state_type;
  GtkPositionType gap_side;
  GdkWindow *window;
  GtkWidget *widget;
  
  g_return_if_fail (notebook != NULL);
  g_return_if_fail (page != NULL);
  g_return_if_fail (area != NULL);

#if GTK_CHECK_VERSION(2,18,0)
  if (!NOTEBOOK_IS_TAB_LABEL_PARENT (notebook, page) ||
      !gtk_widget_get_mapped (page->tab_label) ||
      (page->allocation.width == 0) || (page->allocation.height == 0))
    return;
#else
  if (!NOTEBOOK_IS_TAB_LABEL_PARENT (notebook, page) ||
      !GTK_WIDGET_MAPPED (page->tab_label) ||
      (page->allocation.width == 0) || (page->allocation.height == 0))
    return;
#endif

  widget = GTK_WIDGET (notebook);
  priv = MOD_NOTEBOOK_GET_PRIVATE (notebook);

  if (priv->operation == DRAG_OPERATION_REORDER && page == notebook->cur_page)
    window = priv->drag_window;
  else
    window = widget->window;

  page_area.x = page->allocation.x;
  page_area.y = page->allocation.y;
  page_area.width = page->allocation.width;
  page_area.height = page->allocation.height;

  if (gdk_rectangle_intersect (&page_area, area, &child_area))
    {
      gap_side = get_tab_gap_pos (notebook);

      if (notebook->cur_page == page)
	state_type = GTK_STATE_NORMAL;
      else 
	state_type = GTK_STATE_ACTIVE;

      gtk_paint_extension (widget->style, window,
			   state_type, GTK_SHADOW_OUT,
			   area, widget, "tab",
			   page_area.x, page_area.y,
			   page_area.width, page_area.height,
			   gap_side);

#if GTK_CHECK_VERSION(2,18,0)
      if ((gtk_widget_has_focus (widget)) &&
	  notebook->focus_tab && (notebook->focus_tab->data == page))
#else
      if ((GTK_WIDGET_HAS_FOCUS (widget)) &&
	  notebook->focus_tab && (notebook->focus_tab->data == page))
#endif
	{
          gint focus_width;

	  gtk_widget_style_get (widget, "focus-line-width", &focus_width, NULL);

#if GTK_CHECK_VERSION(2,18,0)
	  gtk_paint_focus (widget->style, window, gtk_widget_get_state (widget),
			   area, widget, "tab",
			   page->tab_label->allocation.x - focus_width,
			   page->tab_label->allocation.y - focus_width,
			   page->tab_label->allocation.width + 2 * focus_width,
			   page->tab_label->allocation.height + 2 * focus_width);
#else
	  gtk_paint_focus (widget->style, window, GTK_WIDGET_STATE (widget),
			   area, widget, "tab",
			   page->tab_label->allocation.x - focus_width,
			   page->tab_label->allocation.y - focus_width,
			   page->tab_label->allocation.width + 2 * focus_width,
			   page->tab_label->allocation.height + 2 * focus_width);
#endif
	}

#if GTK_CHECK_VERSION(2,18,0)
      if (gtk_widget_intersect (page->tab_label, area, &child_area) &&
          gtk_widget_is_drawable (page->tab_label))
#else
      if (gtk_widget_intersect (page->tab_label, area, &child_area) &&
          GTK_WIDGET_DRAWABLE (page->tab_label))
#endif
        {
          GdkEvent *expose_event = gdk_event_new (GDK_EXPOSE);

          /* This is a lame hack since all this code needs rewriting anyhow */
          expose_event->expose.window = g_object_ref (page->tab_label->window);
          expose_event->expose.area = child_area;
          expose_event->expose.region = gdk_region_rectangle (&child_area);
          expose_event->expose.send_event = TRUE;
          expose_event->expose.count = 0;

	  gtk_widget_send_expose (page->tab_label, expose_event);
	  gdk_event_free (expose_event);
	}
    }
}

static void
mod_notebook_draw_arrow (ModNotebook      *notebook,
			 ModNotebookArrow  nbarrow)
{
  GtkStateType state_type;
  GtkShadowType shadow_type;
  GtkWidget *widget;
  GdkRectangle arrow_rect;
  GtkArrowType arrow;
  gboolean is_rtl, left;

  mod_notebook_get_arrow_rect (notebook, &arrow_rect, nbarrow);

  widget = GTK_WIDGET (notebook);

  is_rtl = gtk_widget_get_direction (widget) == GTK_TEXT_DIR_RTL;
  left = (ARROW_IS_LEFT (nbarrow) && !is_rtl) ||
         (!ARROW_IS_LEFT (nbarrow) && is_rtl); 

#if GTK_CHECK_VERSION(2,18,0)
  if (gtk_widget_is_drawable (widget))
#else
  if (GTK_WIDGET_DRAWABLE (notebook))
#endif
    {
      gint scroll_arrow_hlength;
      gint scroll_arrow_vlength;
      gint arrow_size;

      gtk_widget_style_get (widget,
                            "scroll-arrow-hlength", &scroll_arrow_hlength,
                            "scroll-arrow-vlength", &scroll_arrow_vlength,
                            NULL);

      if (notebook->in_child == nbarrow)
        {
          if (notebook->click_child == nbarrow)
            state_type = GTK_STATE_ACTIVE;
          else
            state_type = GTK_STATE_PRELIGHT;
        }
      else
#if GTK_CHECK_VERSION(2,18,0)
        state_type = gtk_widget_get_state (widget);
#else
        state_type = GTK_WIDGET_STATE (widget);
#endif

      if (notebook->click_child == nbarrow)
        shadow_type = GTK_SHADOW_IN;
      else
        shadow_type = GTK_SHADOW_OUT;

      if (notebook->focus_tab &&
	  !mod_notebook_search_page (notebook, notebook->focus_tab,
				      left? STEP_PREV : STEP_NEXT, TRUE))
	{
	  shadow_type = GTK_SHADOW_ETCHED_IN;
	  state_type = GTK_STATE_INSENSITIVE;
	}
      
      if (notebook->tab_pos == GTK_POS_LEFT ||
	  notebook->tab_pos == GTK_POS_RIGHT)
        {
          arrow = (ARROW_IS_LEFT (nbarrow) ? GTK_ARROW_UP : GTK_ARROW_DOWN);
          arrow_size = scroll_arrow_vlength;
        }
      else
        {
          arrow = (ARROW_IS_LEFT (nbarrow) ? GTK_ARROW_LEFT : GTK_ARROW_RIGHT);
          arrow_size = scroll_arrow_hlength;
        }
     
      gtk_paint_arrow (widget->style, widget->window, state_type, 
		       shadow_type, NULL, widget, "notebook",
		       arrow, TRUE, arrow_rect.x, arrow_rect.y, 
		       arrow_size, arrow_size);
    }
}

/* Private ModNotebook Size Allocate Functions:
 *
 * mod_notebook_tab_space
 * mod_notebook_calculate_shown_tabs
 * mod_notebook_calculate_tabs_allocation
 * mod_notebook_pages_allocate
 * mod_notebook_page_allocate
 * mod_notebook_calc_tabs
 */
static void
mod_notebook_tab_space (ModNotebook *notebook,
			gboolean    *show_arrows,
			gint        *min,
			gint        *max,
			gint        *typical_tab_size,
			gint        *tab_space)
{
  ModNotebookPrivate *priv;
  GtkWidget *widget;
  GList *children;
  gint tab_pos = get_effective_tab_pos (notebook);
  gint tab_overlap;
  gint arrow_spacing;
  gint scroll_arrow_hlength;
  gint scroll_arrow_vlength;
  gint close_button_size;
  gint close_button_spacing;
  gint number_of_pages = 0;
  gint min_size = notebook->minimum_tab_label_size;
  gint available_space;

  widget = GTK_WIDGET (notebook);
  priv = MOD_NOTEBOOK_GET_PRIVATE (notebook);
  children = notebook->children;

  gtk_widget_style_get (GTK_WIDGET (notebook),
                        "arrow-spacing", &arrow_spacing,
                        "scroll-arrow-hlength", &scroll_arrow_hlength,
                        "scroll-arrow-vlength", &scroll_arrow_vlength,
			"close-button-size", &close_button_size,
			"close-button-spacing", &close_button_spacing,
                        NULL);

  if (notebook->enable_close_button)
    close_button_size += close_button_spacing * 2;
  else
    close_button_size = 0;

  switch (tab_pos)
    {
    case GTK_POS_TOP:
    case GTK_POS_BOTTOM:
      *min = widget->allocation.x + GTK_CONTAINER (notebook)->border_width;
      *max = widget->allocation.x + widget->allocation.width - GTK_CONTAINER (notebook)->border_width;

      if (gtk_widget_get_direction(widget) == GTK_TEXT_DIR_RTL)
	*min += close_button_size;
      else
	*max -= close_button_size;

      while (children)
	{
          ModNotebookPage *page;

	  page = children->data;
	  children = children->next;

	  if (NOTEBOOK_IS_TAB_LABEL_PARENT (notebook, page) &&
#if GTK_CHECK_VERSION(2,18,0)
	      gtk_widget_get_visible (page->child))
#else
	      GTK_WIDGET_VISIBLE (page->child))
#endif
	    {
	      *tab_space += page->requisition.width;
	      min_size = MAX (min_size, page->requisition.height);
	      ++number_of_pages;
	    }
	}
      break;
    case GTK_POS_RIGHT:
    case GTK_POS_LEFT:
      *min = widget->allocation.y + GTK_CONTAINER (notebook)->border_width;
      *max = widget->allocation.y + widget->allocation.height - GTK_CONTAINER (notebook)->border_width - close_button_size;

      while (children)
	{
          ModNotebookPage *page;

	  page = children->data;
	  children = children->next;

	  if (NOTEBOOK_IS_TAB_LABEL_PARENT (notebook, page) &&
#if GTK_CHECK_VERSION(2,18,0)
	      gtk_widget_get_visible (page->child))
#else
	      GTK_WIDGET_VISIBLE (page->child))
#endif
	    {
	      *tab_space += page->requisition.height;
	      min_size = MAX (min_size, page->requisition.width);
	      ++number_of_pages;
	    }
	}
      break;
    }

  *show_arrows = FALSE;
  if (number_of_pages == 0 ||
      (!notebook->tab_shrinkable && !notebook->scrollable))
    {
      *typical_tab_size = 0;
      return;
    }

  gtk_widget_style_get (widget, "tab-overlap", &tab_overlap, NULL);
  available_space = *max - *min - tab_overlap;

  if (notebook->tab_shrinkable && *tab_space > available_space)
    {
      *typical_tab_size = MAX (available_space / number_of_pages, min_size);
      *tab_space = *typical_tab_size * number_of_pages;
    }
  else
    *typical_tab_size = 0;

  if (notebook->scrollable)
    {
      switch (tab_pos)
	{
	case GTK_POS_TOP:
	case GTK_POS_BOTTOM:
	  if (*tab_space > available_space)
	    {
	      *show_arrows = TRUE;

	      /* take arrows into account */
	      *tab_space = widget->allocation.width - tab_overlap -
		2 * GTK_CONTAINER (notebook)->border_width - close_button_size;

	      if (notebook->has_after_previous)
		{
		  *tab_space -= arrow_spacing + scroll_arrow_hlength;
		  *max -= arrow_spacing + scroll_arrow_hlength;
		}

	      if (notebook->has_after_next)
		{
		  *tab_space -= arrow_spacing + scroll_arrow_hlength;
		  *max -= arrow_spacing + scroll_arrow_hlength;
		}

	      if (notebook->has_before_previous)
		{
		  *tab_space -= arrow_spacing + scroll_arrow_hlength;
		  *min += arrow_spacing + scroll_arrow_hlength;
		}

	      if (notebook->has_before_next)
		{
		  *tab_space -= arrow_spacing + scroll_arrow_hlength;
		  *min += arrow_spacing + scroll_arrow_hlength;
		}
	    }
	  break;
	case GTK_POS_LEFT:
	case GTK_POS_RIGHT:
	  if (*tab_space > available_space)
	    {
	      *show_arrows = TRUE;

	      /* take arrows into account */
	      *tab_space = widget->allocation.height - tab_overlap -
		2 * GTK_CONTAINER (notebook)->border_width - close_button_size;

	      if (notebook->has_after_previous || notebook->has_after_next)
		{
		  *tab_space -= arrow_spacing + scroll_arrow_vlength;
		  *max -= arrow_spacing + scroll_arrow_vlength;
		}

	      if (notebook->has_before_previous || notebook->has_before_next)
		{
		  *tab_space -= arrow_spacing + scroll_arrow_vlength;
		  *min += arrow_spacing + scroll_arrow_vlength;
		}
	    }
	  break;
	}
    }
}

static void
mod_notebook_calculate_shown_tabs (ModNotebook *notebook,
				   gboolean     show_arrows,
				   gint         min,
				   gint         max,
				   gint         typical_tab_size,
				   gint         tab_space,
				   GList      **last_child,
				   gint        *n,
				   gint        *remaining_space)
{
  GtkWidget *widget;
  GtkContainer *container;
  GList *children;
  ModNotebookPage *page;
  gint tab_pos, tab_overlap;
  
  widget = GTK_WIDGET (notebook);
  container = GTK_CONTAINER (notebook);
  gtk_widget_style_get (widget, "tab-overlap", &tab_overlap, NULL);
  tab_pos = get_effective_tab_pos (notebook);

  if (show_arrows) /* first_tab <- focus_tab */
    {
      *remaining_space = tab_space;

      if (NOTEBOOK_IS_TAB_LABEL_PARENT (notebook, notebook->cur_page) &&
#if GTK_CHECK_VERSION(2,18,0)
	  gtk_widget_get_visible (notebook->cur_page->child))
#else
	  GTK_WIDGET_VISIBLE (notebook->cur_page->child))
#endif
	{
	  mod_notebook_calc_tabs (notebook,
				  notebook->focus_tab,
				  &(notebook->focus_tab),
				  remaining_space, typical_tab_size, STEP_NEXT);
	}

      if (*remaining_space <= 0)
	{
	  /* show 1 tab */
	  notebook->first_tab = notebook->focus_tab;
	  *last_child = mod_notebook_search_page (notebook, notebook->focus_tab,
						  STEP_NEXT, TRUE);
	}
      else
	{
	  children = NULL;

	  if (notebook->first_tab && notebook->first_tab != notebook->focus_tab)
	    {
	      /* Is first_tab really predecessor of focus_tab? */
	      page = notebook->first_tab->data;
	      if (NOTEBOOK_IS_TAB_LABEL_PARENT (notebook, page) &&
#if GTK_CHECK_VERSION(2,18,0)
		  gtk_widget_get_visible (page->child))
#else
		  GTK_WIDGET_VISIBLE (page->child))
#endif
		for (children = notebook->focus_tab;
		     children && children != notebook->first_tab;
		     children = mod_notebook_search_page (notebook,
							  children,
							  STEP_PREV,
							  TRUE));
	    }

	  if (!children)
	    {
	      if (NOTEBOOK_IS_TAB_LABEL_PARENT (notebook, notebook->cur_page))
		notebook->first_tab = notebook->focus_tab;
	      else
		notebook->first_tab = mod_notebook_search_page (notebook, notebook->focus_tab,
								STEP_NEXT, TRUE);
	    }
	  else
	    /* calculate shown tabs counting backwards from the focus tab */
	    mod_notebook_calc_tabs (notebook,
				    mod_notebook_search_page (notebook,
							      notebook->focus_tab,
							      STEP_PREV,
							      TRUE),
				    &(notebook->first_tab), remaining_space,
				    typical_tab_size,
				    STEP_PREV);

	  if (*remaining_space < 0)
	    {
	      notebook->first_tab =
		mod_notebook_search_page (notebook, notebook->first_tab,
					  STEP_NEXT, TRUE);
	      if (!notebook->first_tab)
		notebook->first_tab = notebook->focus_tab;

	      *last_child = mod_notebook_search_page (notebook, notebook->focus_tab,
						      STEP_NEXT, TRUE); 
	    }
	  else /* focus_tab -> end */   
	    {
	      if (!notebook->first_tab)
		notebook->first_tab = mod_notebook_search_page (notebook,
								NULL,
								STEP_NEXT,
								TRUE);
	      children = NULL;
	      mod_notebook_calc_tabs (notebook,
				      mod_notebook_search_page (notebook,
								notebook->focus_tab,
								STEP_NEXT,
								TRUE),
				      &children, remaining_space,
				      typical_tab_size, STEP_NEXT);

	      if (*remaining_space <= 0) 
		*last_child = children;
	      else /* start <- first_tab */
		{
		  *last_child = NULL;
		  children = NULL;

		  mod_notebook_calc_tabs (notebook,
					  mod_notebook_search_page (notebook,
								    notebook->first_tab,
								    STEP_PREV,
								    TRUE),
					  &children, remaining_space,
					  typical_tab_size, STEP_PREV);

		  if (*remaining_space == 0)
		    notebook->first_tab = children;
		  else
		    notebook->first_tab = mod_notebook_search_page(notebook,
								   children,
								   STEP_NEXT,
								   TRUE);
		}
	    }
	}

      if (*remaining_space < 0) 
	{
	  /* calculate number of tabs */
	  *remaining_space = - (*remaining_space);
	  *n = 0;

	  for (children = notebook->first_tab;
	       children && children != *last_child;
	       children = mod_notebook_search_page (notebook, children,
						    STEP_NEXT, TRUE))
	    (*n)++;
	}
      else 
	*remaining_space = 0;

      /* unmap all non-visible tabs */
      for (children = mod_notebook_search_page (notebook, NULL,
						STEP_NEXT, TRUE);
	   children && children != notebook->first_tab;
	   children = mod_notebook_search_page (notebook, children,
						STEP_NEXT, TRUE))
	{
	  page = children->data;

	  if (page->tab_label &&
	      NOTEBOOK_IS_TAB_LABEL_PARENT (notebook, page))
	    gtk_widget_set_child_visible (page->tab_label, FALSE);
	}

      for (children = *last_child; children;
	   children = mod_notebook_search_page (notebook, children,
						STEP_NEXT, TRUE))
	{
	  page = children->data;

	  if (page->tab_label &&
	      NOTEBOOK_IS_TAB_LABEL_PARENT (notebook, page))
	    gtk_widget_set_child_visible (page->tab_label, FALSE);
	}
    }
  else /* !show_arrows */
    {
      gint c = 0;
      *n = 0;

      *remaining_space = max - min - tab_overlap - tab_space;
      children = notebook->children;
      notebook->first_tab = mod_notebook_search_page (notebook, NULL,
						      STEP_NEXT, TRUE);
      while (children)
	{
	  page = children->data;
	  children = children->next;

#if GTK_CHECK_VERSION(2,18,0)
	  if (!NOTEBOOK_IS_TAB_LABEL_PARENT (notebook, page) ||
	      !gtk_widget_get_visible (page->child))
	    continue;
#else
	  if (!NOTEBOOK_IS_TAB_LABEL_PARENT (notebook, page) ||
	      !GTK_WIDGET_VISIBLE (page->child))
	    continue;
#endif

	  c++;

	  if (page->expand)
	    (*n)++;
	}

      /* if notebook is homogeneous, all tabs are expanded */
      if (notebook->homogeneous && *n)
	*n = c;
    }
}

static gboolean
get_allocate_at_bottom (GtkWidget *widget,
			gint       search_direction)
{
  gboolean is_rtl = (gtk_widget_get_direction (widget) == GTK_TEXT_DIR_RTL);
  gboolean tab_pos = get_effective_tab_pos (MOD_NOTEBOOK (widget));

  switch (tab_pos)
    {
    case GTK_POS_TOP:
    case GTK_POS_BOTTOM:
      if (!is_rtl)
	return (search_direction == STEP_PREV);
      else
	return (search_direction == STEP_NEXT);

      break;
    case GTK_POS_RIGHT:
    case GTK_POS_LEFT:
      return (search_direction == STEP_PREV);
      break;
    }

  return FALSE;
}

static gboolean
mod_notebook_calculate_tabs_allocation (ModNotebook  *notebook,
					GList       **children,
					GList        *last_child,
					gboolean      showarrow,
					gint          direction,
					gint         *remaining_space,
					gint         *expanded_tabs,
					gint          typical_tab_size,
					gint          min,
					gint          max)
{
  GtkWidget *widget;
  GtkContainer *container;
  ModNotebookPrivate *priv;
  ModNotebookPage *page;
  gboolean allocate_at_bottom;
  gint tab_overlap, tab_pos, tab_extra_space;
  gint left_x, right_x, top_y, bottom_y, anchor;
  gboolean gap_left, packing_changed;
  GtkAllocation child_allocation = { 0, };
  gboolean allocation_changed = FALSE;

  widget = GTK_WIDGET (notebook);
  container = GTK_CONTAINER (notebook);
  priv = MOD_NOTEBOOK_GET_PRIVATE (notebook);
  gtk_widget_style_get (widget, "tab-overlap", &tab_overlap, NULL);
  tab_pos = get_effective_tab_pos (notebook);
  allocate_at_bottom = get_allocate_at_bottom (widget, direction);
  anchor = 0;

  child_allocation.x = widget->allocation.x + container->border_width;
  child_allocation.y = widget->allocation.y + container->border_width;

  switch (tab_pos)
    {
    case GTK_POS_BOTTOM:
      child_allocation.y = widget->allocation.y + widget->allocation.height -
	notebook->cur_page->requisition.height - container->border_width;
      /* fall through */
    case GTK_POS_TOP:
      child_allocation.x = (allocate_at_bottom) ? max : min;
      child_allocation.height = notebook->cur_page->requisition.height;
      anchor = child_allocation.x;
      break;
      
    case GTK_POS_RIGHT:
      child_allocation.x = widget->allocation.x + widget->allocation.width -
	notebook->cur_page->requisition.width - container->border_width;
      /* fall through */
    case GTK_POS_LEFT:
      child_allocation.y = (allocate_at_bottom) ? max : min;
      child_allocation.width = notebook->cur_page->requisition.width;
      anchor = child_allocation.y;
      break;
    }

  left_x   = CLAMP (priv->mouse_x - priv->drag_offset_x,
		    min, max - notebook->cur_page->allocation.width);
  top_y    = CLAMP (priv->mouse_y - priv->drag_offset_y,
		    min, max - notebook->cur_page->allocation.height);
  right_x  = left_x + notebook->cur_page->allocation.width;
  bottom_y = top_y + notebook->cur_page->allocation.height;
  gap_left = packing_changed = FALSE;

  while (*children && *children != last_child)
    {
      page = (*children)->data;

      if (direction == STEP_NEXT && page->pack != GTK_PACK_START)
	{
	  if (!showarrow)
	    break;
	  else if (priv->operation == DRAG_OPERATION_REORDER)
	    packing_changed = TRUE;
	}

      if (direction == STEP_NEXT)
	*children = mod_notebook_search_page (notebook, *children, direction, TRUE);
      else
	{
	  *children = (*children)->next;

#if GTK_CHECK_VERSION(2,18,0)
          if (page->pack != GTK_PACK_END || !gtk_widget_get_visible (page->child))
	    continue;
#else
          if (page->pack != GTK_PACK_END || !GTK_WIDGET_VISIBLE (page->child))
	    continue;
#endif
	}

      if (!NOTEBOOK_IS_TAB_LABEL_PARENT (notebook, page))
	continue;

      tab_extra_space = 0;
      if (*expanded_tabs && (showarrow || page->expand || notebook->homogeneous))
	{
	  tab_extra_space = *remaining_space / *expanded_tabs;
	  *remaining_space -= tab_extra_space;
	  (*expanded_tabs)--;
	}

      switch (tab_pos)
	{
	case GTK_POS_TOP:
	case GTK_POS_BOTTOM:
	  if (typical_tab_size > 0)
	    child_allocation.width = typical_tab_size +
	      tab_overlap + tab_extra_space;
	  else
	    child_allocation.width = page->requisition.width +
	      tab_overlap + tab_extra_space;

	  /* make sure that the reordered tab doesn't go past the last position */
	  if (priv->operation == DRAG_OPERATION_REORDER &&
	      !gap_left && packing_changed)
	    {
	      if (!allocate_at_bottom)
		{
		  if ((notebook->cur_page->pack == GTK_PACK_START && left_x >= anchor) ||
		      (notebook->cur_page->pack == GTK_PACK_END && left_x < anchor))
		    {
		      left_x = priv->drag_window_x = anchor;
		      anchor += notebook->cur_page->allocation.width - tab_overlap;
		    }
		}
	      else
		{
		  if ((notebook->cur_page->pack == GTK_PACK_START && right_x <= anchor) ||
		      (notebook->cur_page->pack == GTK_PACK_END && right_x > anchor))
		    {
		      anchor -= notebook->cur_page->allocation.width;
		      left_x = priv->drag_window_x = anchor;
		      anchor += tab_overlap;
		    }
		}

	      gap_left = TRUE;
	    }

	  if (priv->operation == DRAG_OPERATION_REORDER && page == notebook->cur_page)
	    {
	      priv->drag_window_x = left_x;
	      priv->drag_window_y = child_allocation.y;
	    }
 	  else
 	    {
 	      if (allocate_at_bottom)
		anchor -= child_allocation.width;
 
 	      if (priv->operation == DRAG_OPERATION_REORDER && page->pack == notebook->cur_page->pack)
 		{
 		  if (!allocate_at_bottom &&
 		      left_x >= anchor &&
 		      left_x <= anchor + child_allocation.width / 2)
 		    anchor += notebook->cur_page->allocation.width - tab_overlap;
 		  else if (allocate_at_bottom &&
 			   right_x >= anchor + child_allocation.width / 2 &&
 			   right_x <= anchor + child_allocation.width)
 		    anchor -= notebook->cur_page->allocation.width - tab_overlap;
 		}

	      child_allocation.x = anchor;
 	    }

	  break;
	case GTK_POS_LEFT:
	case GTK_POS_RIGHT:
	  if (typical_tab_size > 0)
	    child_allocation.height = typical_tab_size +
	      tab_overlap + tab_extra_space;
	  else
	    child_allocation.height = page->requisition.height +
	      tab_overlap + tab_extra_space;

	  /* make sure that the reordered tab doesn't go past the last position */
	  if (priv->operation == DRAG_OPERATION_REORDER &&
	      !gap_left && packing_changed)
	    {
	      if (!allocate_at_bottom &&
		  ((notebook->cur_page->pack == GTK_PACK_START && top_y >= anchor) ||
		   (notebook->cur_page->pack == GTK_PACK_END && top_y < anchor)))
		{
		  top_y = priv->drag_window_y = anchor;
		  anchor += notebook->cur_page->allocation.height - tab_overlap;
		}
 
	      gap_left = TRUE;
	    }

	  if (priv->operation == DRAG_OPERATION_REORDER && page == notebook->cur_page)
	    {
	      priv->drag_window_x = child_allocation.x;
	      priv->drag_window_y = top_y;
	    }
 	  else
 	    {
	      if (allocate_at_bottom)
		anchor -= child_allocation.height;

 	      if (priv->operation == DRAG_OPERATION_REORDER && page->pack == notebook->cur_page->pack)
		{
		  if (!allocate_at_bottom &&
		      top_y >= anchor &&
		      top_y <= anchor + child_allocation.height / 2)
		    anchor += notebook->cur_page->allocation.height - tab_overlap;
		  else if (allocate_at_bottom &&
			   bottom_y >= anchor + child_allocation.height / 2 &&
			   bottom_y <= anchor + child_allocation.height)
		    anchor -= notebook->cur_page->allocation.height - tab_overlap;
		}

	      child_allocation.y = anchor;
 	    }

	  break;
	}

      if ((priv->operation != DRAG_OPERATION_REORDER || page != notebook->cur_page) &&
	  (page->allocation.x != child_allocation.x ||
	   page->allocation.y != child_allocation.y ||
	   page->allocation.width != child_allocation.width ||
	   page->allocation.height != child_allocation.height))
	allocation_changed = TRUE;

      page->allocation = child_allocation;

      if (page == notebook->cur_page &&
	  (priv->operation == DRAG_OPERATION_REORDER ||
	   priv->operation == DRAG_OPERATION_DETACH))
	{
	  /* needs to be allocated at 0,0
	   * to be shown in the drag window */
	  page->allocation.x = 0;
	  page->allocation.y = 0;
	}

      /* calculate whether to leave a gap based on reorder operation or not */
      switch (tab_pos)
	{
	case GTK_POS_TOP:
	case GTK_POS_BOTTOM:
 	  if (priv->operation != DRAG_OPERATION_REORDER ||
	      (priv->operation == DRAG_OPERATION_REORDER && page != notebook->cur_page))
 	    {
 	      if (priv->operation == DRAG_OPERATION_REORDER)
 		{
 		  if (page->pack == notebook->cur_page->pack &&
 		      !allocate_at_bottom &&
 		      left_x >  anchor + child_allocation.width / 2 &&
 		      left_x <= anchor + child_allocation.width)
 		    anchor += notebook->cur_page->allocation.width - tab_overlap;
 		  else if (page->pack == notebook->cur_page->pack &&
 			   allocate_at_bottom &&
 			   right_x >= anchor &&
 			   right_x <= anchor + child_allocation.width / 2)
 		    anchor -= notebook->cur_page->allocation.width - tab_overlap;
 		}
 
 	      if (!allocate_at_bottom)
 		anchor += child_allocation.width - tab_overlap;
 	      else
 		anchor += tab_overlap;
 	    }

	  break;
	case GTK_POS_LEFT:
	case GTK_POS_RIGHT:
 	  if (priv->operation != DRAG_OPERATION_REORDER  ||
	      (priv->operation == DRAG_OPERATION_REORDER && page != notebook->cur_page))
 	    {
 	      if (priv->operation == DRAG_OPERATION_REORDER)
		{
		  if (page->pack == notebook->cur_page->pack &&
		      !allocate_at_bottom &&
		      top_y >= anchor + child_allocation.height / 2 &&
		      top_y <= anchor + child_allocation.height)
		    anchor += notebook->cur_page->allocation.height - tab_overlap;
		  else if (page->pack == notebook->cur_page->pack &&
			   allocate_at_bottom &&
			   bottom_y >= anchor &&
			   bottom_y <= anchor + child_allocation.height / 2)
		    anchor -= notebook->cur_page->allocation.height - tab_overlap;
		}

	      if (!allocate_at_bottom)
		anchor += child_allocation.height - tab_overlap;
	      else
		anchor += tab_overlap;
 	    }

	  break;
	}

      /* set child visible */
      if (page->tab_label)
	gtk_widget_set_child_visible (page->tab_label, TRUE);
    }

  /* Don't move the current tab past the last position during tabs reordering */
  if (children &&
      priv->operation == DRAG_OPERATION_REORDER &&
      ((direction == STEP_NEXT && notebook->cur_page->pack == GTK_PACK_START) ||
       ((direction == STEP_PREV || packing_changed) && notebook->cur_page->pack == GTK_PACK_END)))
    {
      switch (tab_pos)
	{
	case GTK_POS_TOP:
	case GTK_POS_BOTTOM:
	  if (allocate_at_bottom)
	    anchor -= notebook->cur_page->allocation.width;

	  if ((!allocate_at_bottom && priv->drag_window_x > anchor) ||
	      (allocate_at_bottom && priv->drag_window_x < anchor))
	    priv->drag_window_x = anchor;
	  break;
	case GTK_POS_LEFT:
	case GTK_POS_RIGHT:
	  if (allocate_at_bottom)
	    anchor -= notebook->cur_page->allocation.height;

	  if ((!allocate_at_bottom && priv->drag_window_y > anchor) ||
	      (allocate_at_bottom && priv->drag_window_y < anchor))
	    priv->drag_window_y = anchor;
	  break;
	}
    }

  return allocation_changed;
}

static void
mod_notebook_pages_allocate (ModNotebook *notebook)
{
  GList *children = NULL;
  GList *last_child = NULL;
  gboolean showarrow = FALSE;
  gint tab_space, min, max, typical_tab_size, remaining_space;
  gint expanded_tabs;

  if (!notebook->show_tabs || !notebook->children || !notebook->cur_page)
    return;

  min = max = tab_space = remaining_space = 0;
  expanded_tabs = 1;

  mod_notebook_tab_space (notebook, &showarrow,
			  &min, &max, &typical_tab_size, &tab_space);

  mod_notebook_calculate_shown_tabs (notebook, showarrow,
				     min, max, typical_tab_size, tab_space,
				     &last_child,
				     &expanded_tabs, &remaining_space);

  children = notebook->first_tab;
  mod_notebook_calculate_tabs_allocation (notebook, &children, last_child,
					  showarrow, STEP_NEXT,
					  &remaining_space, &expanded_tabs,
					  typical_tab_size, min, max);
  if (children && children != last_child)
    {
      children = notebook->children;
      mod_notebook_calculate_tabs_allocation (notebook, &children, last_child,
					      showarrow, STEP_PREV,
					      &remaining_space, &expanded_tabs,
					      typical_tab_size, min, max);
    }

  children = notebook->children;

  while (children)
    {
      mod_notebook_page_allocate (notebook, MOD_NOTEBOOK_PAGE (children));
      children = children->next;
    }

  if (!notebook->first_tab)
    notebook->first_tab = notebook->children;

  mod_notebook_redraw_tabs (notebook);
}

static void
mod_notebook_page_allocate (ModNotebook     *notebook,
			    ModNotebookPage *page)
{
  GtkWidget *widget = GTK_WIDGET (notebook);
  GtkAllocation child_allocation;
  GtkRequisition tab_requisition;
  gint xthickness;
  gint ythickness;
  gint padding;
  gint focus_width;
  gint tab_curvature;
  gint tab_pos = get_effective_tab_pos (notebook);

  if (!page->tab_label)
    return;

  xthickness = widget->style->xthickness;
  ythickness = widget->style->ythickness;

  gtk_widget_get_child_requisition (page->tab_label, &tab_requisition);
  gtk_widget_style_get (widget,
			"focus-line-width", &focus_width,
			"tab-curvature", &tab_curvature,
			NULL);
  switch (tab_pos)
    {
    case GTK_POS_TOP:
    case GTK_POS_BOTTOM:
      padding = tab_curvature + focus_width + notebook->tab_hborder;
      if (page->fill)
	{
	  child_allocation.x = xthickness + focus_width + notebook->tab_hborder;
	  child_allocation.width = MAX (1, page->allocation.width - 2 * child_allocation.x);
	  child_allocation.x += page->allocation.x;
	}
      else
	{
	  child_allocation.x = page->allocation.x;
	  if (page->allocation.width > tab_requisition.width)
	    {
	      child_allocation.x += (page->allocation.width - tab_requisition.width) / 2;
	      child_allocation.width = tab_requisition.width;
	    }
	  else
	    {
	      gint extra = xthickness + focus_width + notebook->tab_hborder;
	      child_allocation.x += extra;
	      child_allocation.width = MAX (1, page->allocation.width - extra * 2);
	    }
	}

      child_allocation.y = notebook->tab_vborder + focus_width + page->allocation.y;

      if (tab_pos == GTK_POS_TOP)
	child_allocation.y += ythickness;

      child_allocation.height = MAX (1, (page->allocation.height - ythickness -
					 2 * (notebook->tab_vborder + focus_width)));
      break;
    case GTK_POS_LEFT:
    case GTK_POS_RIGHT:
      padding = tab_curvature + focus_width + notebook->tab_vborder;
      if (page->fill)
	{
	  child_allocation.y = ythickness + padding;
	  child_allocation.height = MAX (1, (page->allocation.height -
					     2 * child_allocation.y));
	  child_allocation.y += page->allocation.y;
	}
      else
	{
	  child_allocation.y = page->allocation.y +
	    (page->allocation.height - tab_requisition.height) / 2;

	  child_allocation.height = tab_requisition.height;
	}

      child_allocation.x = notebook->tab_hborder + focus_width + page->allocation.x;

      if (tab_pos == GTK_POS_LEFT)
	child_allocation.x += xthickness;

      child_allocation.width = MAX (1, (page->allocation.width - xthickness -
					2 * (notebook->tab_hborder + focus_width)));
      break;
    }

  gtk_widget_size_allocate (page->tab_label, &child_allocation);
}

static void 
mod_notebook_calc_tabs (ModNotebook  *notebook,
			GList        *start,
                        GList       **end,
			gint         *tab_space,
			gint          typical_tab_size,
                        guint         direction)
{
  ModNotebookPage *page = NULL;
  GList *children;
  GList *last_list = NULL;
  GList *last_calculated_child = NULL;
  gboolean pack;
  gint tab_pos = get_effective_tab_pos (notebook);
  guint real_direction;

  if (!start)
    return;

  children = start;
  pack = MOD_NOTEBOOK_PAGE (start)->pack;
  if (pack == GTK_PACK_END)
    real_direction = (direction == STEP_PREV) ? STEP_NEXT : STEP_PREV;
  else
    real_direction = direction;

  while (1)
    {
      switch (tab_pos)
	{
	case GTK_POS_TOP:
	case GTK_POS_BOTTOM:
	  while (children)
	    {
	      page = children->data;
	      if (NOTEBOOK_IS_TAB_LABEL_PARENT (notebook, page) &&
#if GTK_CHECK_VERSION(2,18,0)
		  gtk_widget_get_visible (page->child))
#else
		  GTK_WIDGET_VISIBLE (page->child))
#endif
		{
		  if (page->pack == pack)
		    {
		      gint width;
		      if (typical_tab_size > 0)
			width = typical_tab_size;
		      else
			width = page->requisition.width;

		      *tab_space -= width;
		      if (*tab_space < 0 || children == *end)
			{
			  if (*tab_space < 0) 
			    {
			      *tab_space = - (*tab_space + width);

			      if (*tab_space == 0 && direction == STEP_PREV)
				children = last_calculated_child;

			      *end = children;
			    }
			  return;
			}

		      last_calculated_child = children;
		    }
		  last_list = children;
		}
	      if (real_direction == STEP_NEXT)
		children = children->next;
	      else
		children = children->prev;
	    }
	  break;
	case GTK_POS_LEFT:
	case GTK_POS_RIGHT:
	  while (children)
	    {
	      page = children->data;
	      if (NOTEBOOK_IS_TAB_LABEL_PARENT (notebook, page) &&
#if GTK_CHECK_VERSION(2,18,0)
		  gtk_widget_get_visible (page->child))
#else
		  GTK_WIDGET_VISIBLE (page->child))
#endif
		{
		  if (page->pack == pack)
		    {
		      gint height;
		      if (typical_tab_size > 0)
			height = typical_tab_size;
		      else
			height = page->requisition.height;

		      *tab_space -= height;
		      if (*tab_space < 0 || children == *end)
			{
			  if (*tab_space < 0)
			    {
			      *tab_space = - (*tab_space + height);

			      if (*tab_space == 0 && direction == STEP_PREV)
				children = last_calculated_child;

			      *end = children;
			    }
			  return;
			}

		      last_calculated_child = children;
		    }
		  last_list = children;
		}
	      if (real_direction == STEP_NEXT)
		children = children->next;
	      else
		children = children->prev;
	    }
	  break;
	}
      if (real_direction == STEP_PREV)
	return;
      pack = (pack == GTK_PACK_END) ? GTK_PACK_START : GTK_PACK_END;
      real_direction = STEP_PREV;
      children = last_list;
    }
}

static void
mod_notebook_update_tab_states (ModNotebook *notebook)
{
  GList *list;

  for (list = notebook->children; list != NULL; list = list->next)
    {
      ModNotebookPage *page = list->data;
      
      if (page->tab_label)
	{
	  if (page == notebook->cur_page)
	    gtk_widget_set_state (page->tab_label, GTK_STATE_NORMAL);
	  else
	    gtk_widget_set_state (page->tab_label, GTK_STATE_ACTIVE);
	}
    }
}

/* Private ModNotebook Page Switch Methods:
 *
 * mod_notebook_real_switch_page
 */
static void
mod_notebook_real_switch_page (ModNotebook     *notebook,
			       ModNotebookPage *page,
			       guint            page_num)
{
  g_return_if_fail (MOD_IS_NOTEBOOK (notebook));
  g_return_if_fail (page != NULL);

#if GTK_CHECK_VERSION(2,18,0)
  if (notebook->cur_page == page || !gtk_widget_get_visible (page->child))
    return;
#else
  if (notebook->cur_page == page || !GTK_WIDGET_VISIBLE (page->child))
    return;
#endif

  if (notebook->cur_page)
    gtk_widget_set_child_visible (notebook->cur_page->child, FALSE);
  
  notebook->cur_page = page;

  if (!notebook->focus_tab ||
      notebook->focus_tab->data != (gpointer) notebook->cur_page)
    notebook->focus_tab = 
      g_list_find (notebook->children, notebook->cur_page);

  gtk_widget_set_child_visible (notebook->cur_page->child, TRUE);

  /* If the focus was on the previous page, move it to the first
   * element on the new page, if possible, or if not, to the
   * notebook itself.
   */
  if (notebook->child_has_focus)
    {
      if (notebook->cur_page->last_focus_child &&
	  gtk_widget_is_ancestor (notebook->cur_page->last_focus_child, notebook->cur_page->child))
	gtk_widget_grab_focus (notebook->cur_page->last_focus_child);
      else
	if (!gtk_widget_child_focus (notebook->cur_page->child, GTK_DIR_TAB_FORWARD))
	  gtk_widget_grab_focus (GTK_WIDGET (notebook));
    }
  
  mod_notebook_update_tab_states (notebook);
  gtk_widget_queue_resize (GTK_WIDGET (notebook));
  g_object_notify (G_OBJECT (notebook), "page");
}

/* Private ModNotebook Page Switch Functions:
 *
 * mod_notebook_switch_page
 * mod_notebook_page_select
 * mod_notebook_switch_focus_tab
 * mod_notebook_menu_switch_page
 */
static void
mod_notebook_switch_page (ModNotebook     *notebook,
			  ModNotebookPage *page,
			  gint             page_num)
{ 
  g_return_if_fail (MOD_IS_NOTEBOOK (notebook));
  g_return_if_fail (page != NULL);
 
  if (notebook->cur_page == page)
    return;

  if (page_num < 0)
    page_num = g_list_index (notebook->children, page);

  g_signal_emit (notebook,
		 notebook_signals[SWITCH_PAGE],
		 0,
		 page,
		 page_num);
}

static gint
mod_notebook_page_select (ModNotebook *notebook,
			  gboolean     move_focus)
{
  ModNotebookPage *page;
  GtkDirectionType dir = GTK_DIR_DOWN; /* Quiet GCC */
  gint tab_pos = get_effective_tab_pos (notebook);

  g_return_val_if_fail (MOD_IS_NOTEBOOK (notebook), FALSE);

  if (!notebook->focus_tab)
    return FALSE;

  page = notebook->focus_tab->data;
  mod_notebook_switch_page (notebook, page, -1);

  if (move_focus)
    {
      switch (tab_pos)
	{
	case GTK_POS_TOP:
	  dir = GTK_DIR_DOWN;
	  break;
	case GTK_POS_BOTTOM:
	  dir = GTK_DIR_UP;
	  break;
	case GTK_POS_LEFT:
	  dir = GTK_DIR_RIGHT;
	  break;
	case GTK_POS_RIGHT:
	  dir = GTK_DIR_LEFT;
	  break;
	}

      if (gtk_widget_child_focus (page->child, dir))
        return TRUE;
    }
  return FALSE;
}

static void
mod_notebook_switch_focus_tab (ModNotebook *notebook, 
			       GList       *new_child)
{
  GList *old_child;
  ModNotebookPage *page;

  g_return_if_fail (MOD_IS_NOTEBOOK (notebook));

  if (notebook->focus_tab == new_child)
    return;

  old_child = notebook->focus_tab;
  notebook->focus_tab = new_child;

  if (notebook->scrollable)
    mod_notebook_redraw_arrows (notebook);

  if (!notebook->show_tabs || !notebook->focus_tab)
    return;

  page = notebook->focus_tab->data;
#if GTK_CHECK_VERSION(2,18,0)
  if (gtk_widget_get_mapped (page->tab_label))
    mod_notebook_redraw_tabs (notebook);
  else
    mod_notebook_pages_allocate (notebook);
#else
  if (GTK_WIDGET_MAPPED (page->tab_label))
    mod_notebook_redraw_tabs (notebook);
  else
    mod_notebook_pages_allocate (notebook);
#endif
  
  mod_notebook_switch_page (notebook, page,
			    g_list_index (notebook->children, page));
}

static void
mod_notebook_menu_switch_page (GtkWidget       *widget,
			       ModNotebookPage *page)
{
  ModNotebook *notebook;
  GList *children;
  guint page_num;

  g_return_if_fail (widget != NULL);
  g_return_if_fail (page != NULL);

  notebook = MOD_NOTEBOOK (gtk_menu_get_attach_widget
			   (GTK_MENU (widget->parent)));

  if (notebook->cur_page == page)
    return;

  page_num = 0;
  children = notebook->children;
  while (children && children->data != page)
    {
      children = children->next;
      page_num++;
    }

  g_signal_emit (notebook,
		 notebook_signals[SWITCH_PAGE],
		 0,
		 page,
		 page_num);
}

/* Private ModNotebook Menu Functions:
 *
 * mod_notebook_menu_item_create
 * mod_notebook_menu_label_unparent
 * mod_notebook_menu_detacher
 */
static void
mod_notebook_menu_item_create (ModNotebook *notebook, 
			       GList       *list)
{	
  ModNotebookPage *page;
  GtkWidget *menu_item;

  page = list->data;
  if (page->default_menu)
    {
      if (page->tab_label && GTK_IS_LABEL (page->tab_label))
	page->menu_label = gtk_label_new (GTK_LABEL (page->tab_label)->label);
      else
	page->menu_label = gtk_label_new ("");
      gtk_misc_set_alignment (GTK_MISC (page->menu_label), 0.0, 0.5);
    }

  gtk_widget_show (page->menu_label);
  menu_item = gtk_menu_item_new ();
  gtk_container_add (GTK_CONTAINER (menu_item), page->menu_label);
  gtk_menu_shell_insert (GTK_MENU_SHELL (notebook->menu), menu_item,
			 mod_notebook_real_page_position (notebook, list));
  g_signal_connect (menu_item, "activate",
		    G_CALLBACK (mod_notebook_menu_switch_page), page);
#if GTK_CHECK_VERSION(2,18,0)
  if (gtk_widget_get_visible (page->child))
    gtk_widget_show (menu_item);
#else
  if (GTK_WIDGET_VISIBLE (page->child))
    gtk_widget_show (menu_item);
#endif
}

static void
mod_notebook_menu_label_unparent (GtkWidget *widget, 
				  gpointer  data)
{
  gtk_widget_unparent (GTK_BIN (widget)->child);
  GTK_BIN (widget)->child = NULL;
}

static void
mod_notebook_menu_detacher (GtkWidget *widget,
			    GtkMenu   *menu)
{
  ModNotebook *notebook;

  g_return_if_fail (MOD_IS_NOTEBOOK (widget));

  notebook = MOD_NOTEBOOK (widget);
  g_return_if_fail (notebook->menu == (GtkWidget*) menu);

  notebook->menu = NULL;
}

/* Private ModNotebook Setter Functions:
 *
 * mod_notebook_set_homogeneous_tabs_internal
 * mod_notebook_set_tab_border_internal
 * mod_notebook_set_tab_hborder_internal
 * mod_notebook_set_tab_vborder_internal
 */
static void
mod_notebook_set_homogeneous_tabs_internal (ModNotebook *notebook,
				            gboolean     homogeneous)
{
  g_return_if_fail (MOD_IS_NOTEBOOK (notebook));

  if (homogeneous == notebook->homogeneous)
    return;

  notebook->homogeneous = homogeneous;
  gtk_widget_queue_resize (GTK_WIDGET (notebook));

  g_object_notify (G_OBJECT (notebook), "homogeneous");
}

static void
mod_notebook_set_tab_border_internal (ModNotebook *notebook,
				      guint        border_width)
{
  g_return_if_fail (MOD_IS_NOTEBOOK (notebook));

  notebook->tab_hborder = border_width;
  notebook->tab_vborder = border_width;

#if GTK_CHECK_VERSION(2,18,0)
  if (gtk_widget_get_visible (GTK_WIDGET (notebook)) && notebook->show_tabs)
    gtk_widget_queue_resize (GTK_WIDGET (notebook));
#else
  if (GTK_WIDGET_VISIBLE (notebook) && notebook->show_tabs)
    gtk_widget_queue_resize (GTK_WIDGET (notebook));
#endif

  g_object_freeze_notify (G_OBJECT (notebook));
  g_object_notify (G_OBJECT (notebook), "tab-hborder");
  g_object_notify (G_OBJECT (notebook), "tab-vborder");
  g_object_thaw_notify (G_OBJECT (notebook));
}

static void
mod_notebook_set_tab_hborder_internal (ModNotebook *notebook,
				       guint        tab_hborder)
{
  g_return_if_fail (MOD_IS_NOTEBOOK (notebook));

  if (notebook->tab_hborder == tab_hborder)
    return;

  notebook->tab_hborder = tab_hborder;

#if GTK_CHECK_VERSION(2,18,0)
  if (gtk_widget_get_visible (GTK_WIDGET (notebook)) && notebook->show_tabs)
    gtk_widget_queue_resize (GTK_WIDGET (notebook));
#else
  if (GTK_WIDGET_VISIBLE (notebook) && notebook->show_tabs)
    gtk_widget_queue_resize (GTK_WIDGET (notebook));
#endif

  g_object_notify (G_OBJECT (notebook), "tab-hborder");
}

static void
mod_notebook_set_tab_vborder_internal (ModNotebook *notebook,
				       guint        tab_vborder)
{
  g_return_if_fail (MOD_IS_NOTEBOOK (notebook));

  if (notebook->tab_vborder == tab_vborder)
    return;

  notebook->tab_vborder = tab_vborder;

#if GTK_CHECK_VERSION(2,18,0)
  if (gtk_widget_get_visible (GTK_WIDGET (notebook)) && notebook->show_tabs)
    gtk_widget_queue_resize (GTK_WIDGET (notebook));
#else
  if (GTK_WIDGET_VISIBLE (notebook) && notebook->show_tabs)
    gtk_widget_queue_resize (GTK_WIDGET (notebook));
#endif

  g_object_notify (G_OBJECT (notebook), "tab-vborder");
}

/* Public ModNotebook Page Insert/Remove Methods :
 *
 * mod_notebook_append_page
 * mod_notebook_append_page_menu
 * mod_notebook_prepend_page
 * mod_notebook_prepend_page_menu
 * mod_notebook_insert_page
 * mod_notebook_insert_page_menu
 * mod_notebook_remove_page
 */
/**
 * mod_notebook_append_page:
 * @notebook: a #ModNotebook
 * @child: the #GtkWidget to use as the contents of the page.
 * @tab_label: the #GtkWidget to be used as the label for the page,
 *             or %NULL to use the default label, 'page N'.
 * 
 * Appends a page to @notebook.
 *
 * Return value: the index (starting from 0) of the appended
 * page in the notebook, or -1 if function fails
 **/
gint
mod_notebook_append_page (ModNotebook *notebook,
			  GtkWidget   *child,
			  GtkWidget   *tab_label)
{
  g_return_val_if_fail (MOD_IS_NOTEBOOK (notebook), -1);
  g_return_val_if_fail (GTK_IS_WIDGET (child), -1);
  g_return_val_if_fail (tab_label == NULL || GTK_IS_WIDGET (tab_label), -1);
  
  return mod_notebook_insert_page_menu (notebook, child, tab_label, NULL, -1);
}

/**
 * mod_notebook_append_page_menu:
 * @notebook: a #ModNotebook
 * @child: the #GtkWidget to use as the contents of the page.
 * @tab_label: the #GtkWidget to be used as the label for the page,
 *             or %NULL to use the default label, 'page N'.
 * @menu_label: the widget to use as a label for the page-switch
 *              menu, if that is enabled. If %NULL, and @tab_label
 *              is a #GtkLabel or %NULL, then the menu label will be
 *              a newly created label with the same text as @tab_label;
 *              If @tab_label is not a #GtkLabel, @menu_label must be
 *              specified if the page-switch menu is to be used.
 * 
 * Appends a page to @notebook, specifying the widget to use as the
 * label in the popup menu.
 *
 * Return value: the index (starting from 0) of the appended
 * page in the notebook, or -1 if function fails
 **/
gint
mod_notebook_append_page_menu (ModNotebook *notebook,
			       GtkWidget   *child,
			       GtkWidget   *tab_label,
			       GtkWidget   *menu_label)
{
  g_return_val_if_fail (MOD_IS_NOTEBOOK (notebook), -1);
  g_return_val_if_fail (GTK_IS_WIDGET (child), -1);
  g_return_val_if_fail (tab_label == NULL || GTK_IS_WIDGET (tab_label), -1);
  g_return_val_if_fail (menu_label == NULL || GTK_IS_WIDGET (menu_label), -1);
  
  return mod_notebook_insert_page_menu (notebook, child, tab_label, menu_label, -1);
}

/**
 * mod_notebook_prepend_page:
 * @notebook: a #ModNotebook
 * @child: the #GtkWidget to use as the contents of the page.
 * @tab_label: the #GtkWidget to be used as the label for the page,
 *             or %NULL to use the default label, 'page N'.
 *
 * Prepends a page to @notebook.
 *
 * Return value: the index (starting from 0) of the prepended
 * page in the notebook, or -1 if function fails
 **/
gint
mod_notebook_prepend_page (ModNotebook *notebook,
			   GtkWidget   *child,
			   GtkWidget   *tab_label)
{
  g_return_val_if_fail (MOD_IS_NOTEBOOK (notebook), -1);
  g_return_val_if_fail (GTK_IS_WIDGET (child), -1);
  g_return_val_if_fail (tab_label == NULL || GTK_IS_WIDGET (tab_label), -1);
  
  return mod_notebook_insert_page_menu (notebook, child, tab_label, NULL, 0);
}

/**
 * mod_notebook_prepend_page_menu:
 * @notebook: a #ModNotebook
 * @child: the #GtkWidget to use as the contents of the page.
 * @tab_label: the #GtkWidget to be used as the label for the page,
 *             or %NULL to use the default label, 'page N'.
 * @menu_label: the widget to use as a label for the page-switch
 *              menu, if that is enabled. If %NULL, and @tab_label
 *              is a #GtkLabel or %NULL, then the menu label will be
 *              a newly created label with the same text as @tab_label;
 *              If @tab_label is not a #GtkLabel, @menu_label must be
 *              specified if the page-switch menu is to be used.
 * 
 * Prepends a page to @notebook, specifying the widget to use as the
 * label in the popup menu.
 *
 * Return value: the index (starting from 0) of the prepended
 * page in the notebook, or -1 if function fails
 **/
gint
mod_notebook_prepend_page_menu (ModNotebook *notebook,
				GtkWidget   *child,
				GtkWidget   *tab_label,
				GtkWidget   *menu_label)
{
  g_return_val_if_fail (MOD_IS_NOTEBOOK (notebook), -1);
  g_return_val_if_fail (GTK_IS_WIDGET (child), -1);
  g_return_val_if_fail (tab_label == NULL || GTK_IS_WIDGET (tab_label), -1);
  g_return_val_if_fail (menu_label == NULL || GTK_IS_WIDGET (menu_label), -1);
  
  return mod_notebook_insert_page_menu (notebook, child, tab_label, menu_label, 0);
}

/**
 * mod_notebook_insert_page:
 * @notebook: a #ModNotebook
 * @child: the #GtkWidget to use as the contents of the page.
 * @tab_label: the #GtkWidget to be used as the label for the page,
 *             or %NULL to use the default label, 'page N'.
 * @position: the index (starting at 0) at which to insert the page,
 *            or -1 to append the page after all other pages.
 * 
 * Insert a page into @notebook at the given position.
 *
 * Return value: the index (starting from 0) of the inserted
 * page in the notebook, or -1 if function fails
 **/
gint
mod_notebook_insert_page (ModNotebook *notebook,
			  GtkWidget   *child,
			  GtkWidget   *tab_label,
			  gint         position)
{
  g_return_val_if_fail (MOD_IS_NOTEBOOK (notebook), -1);
  g_return_val_if_fail (GTK_IS_WIDGET (child), -1);
  g_return_val_if_fail (tab_label == NULL || GTK_IS_WIDGET (tab_label), -1);
  
  return mod_notebook_insert_page_menu (notebook, child, tab_label, NULL, position);
}


static gint
mod_notebook_page_compare_tab (gconstpointer a,
			       gconstpointer b)
{
  return (((ModNotebookPage *) a)->tab_label != b);
}

static gboolean
mod_notebook_mnemonic_activate_switch_page (GtkWidget *child,
					    gboolean overload,
					    gpointer data)
{
  ModNotebook *notebook = MOD_NOTEBOOK (data);
  GList *list;
  
  list = g_list_find_custom (notebook->children, child,
			     mod_notebook_page_compare_tab);
  if (list)
    {
      ModNotebookPage *page = list->data;

      gtk_widget_grab_focus (GTK_WIDGET (notebook));	/* Do this first to avoid focusing new page */
      mod_notebook_switch_page (notebook, page,  -1);
      focus_tabs_in (notebook);
    }

  return TRUE;
}

/**
 * mod_notebook_insert_page_menu:
 * @notebook: a #ModNotebook
 * @child: the #GtkWidget to use as the contents of the page.
 * @tab_label: the #GtkWidget to be used as the label for the page,
 *             or %NULL to use the default label, 'page N'.
 * @menu_label: the widget to use as a label for the page-switch
 *              menu, if that is enabled. If %NULL, and @tab_label
 *              is a #GtkLabel or %NULL, then the menu label will be
 *              a newly created label with the same text as @tab_label;
 *              If @tab_label is not a #GtkLabel, @menu_label must be
 *              specified if the page-switch menu is to be used.
 * @position: the index (starting at 0) at which to insert the page,
 *            or -1 to append the page after all other pages.
 * 
 * Insert a page into @notebook at the given position, specifying
 * the widget to use as the label in the popup menu.
 *
 * Return value: the index (starting from 0) of the inserted
 * page in the notebook
 **/
gint
mod_notebook_insert_page_menu (ModNotebook *notebook,
			       GtkWidget   *child,
			       GtkWidget   *tab_label,
			       GtkWidget   *menu_label,
			       gint         position)
{
  ModNotebookClass *class;

  g_return_val_if_fail (MOD_IS_NOTEBOOK (notebook), -1);
  g_return_val_if_fail (GTK_IS_WIDGET (child), -1);
  g_return_val_if_fail (tab_label == NULL || GTK_IS_WIDGET (tab_label), -1);
  g_return_val_if_fail (menu_label == NULL || GTK_IS_WIDGET (menu_label), -1);

  class = MOD_NOTEBOOK_GET_CLASS (notebook);

  return (class->insert_page) (notebook, child, tab_label, menu_label, position);
}

/**
 * mod_notebook_remove_page:
 * @notebook: a #ModNotebook.
 * @page_num: the index of a notebook page, starting
 *            from 0. If -1, the last page will
 *            be removed.
 * 
 * Removes a page from the notebook given its index
 * in the notebook.
 **/
void
mod_notebook_remove_page (ModNotebook *notebook,
			  gint         page_num)
{
  GList *list = NULL;

  g_return_if_fail (MOD_IS_NOTEBOOK (notebook));

  if (page_num >= 0)
    list = g_list_nth (notebook->children, page_num);
  else
    list = g_list_last (notebook->children);

  if (list)
    gtk_container_remove (GTK_CONTAINER (notebook),
			  ((ModNotebookPage *) list->data)->child);
}

/* Public ModNotebook Page Switch Methods :
 * mod_notebook_get_current_page
 * mod_notebook_page_num
 * mod_notebook_set_current_page
 * mod_notebook_next_page
 * mod_notebook_prev_page
 */
/**
 * mod_notebook_get_current_page:
 * @notebook: a #ModNotebook
 * 
 * Returns the page number of the current page.
 * 
 * Return value: the index (starting from 0) of the current
 * page in the notebook. If the notebook has no pages, then
 * -1 will be returned.
 **/
gint
mod_notebook_get_current_page (ModNotebook *notebook)
{
  g_return_val_if_fail (MOD_IS_NOTEBOOK (notebook), -1);

  if (!notebook->cur_page)
    return -1;

  return g_list_index (notebook->children, notebook->cur_page);
}

/**
 * mod_notebook_get_nth_page:
 * @notebook: a #ModNotebook
 * @page_num: the index of a page in the noteobok, or -1
 *            to get the last page.
 * 
 * Returns the child widget contained in page number @page_num.
 * 
 * Return value: the child widget, or %NULL if @page_num is
 * out of bounds.
 **/
GtkWidget*
mod_notebook_get_nth_page (ModNotebook *notebook,
			   gint         page_num)
{
  ModNotebookPage *page;
  GList *list;

  g_return_val_if_fail (MOD_IS_NOTEBOOK (notebook), NULL);

  if (page_num >= 0)
    list = g_list_nth (notebook->children, page_num);
  else
    list = g_list_last (notebook->children);

  if (list)
    {
      page = list->data;
      return page->child;
    }

  return NULL;
}

/**
 * mod_notebook_get_n_pages:
 * @notebook: a #ModNotebook
 * 
 * Gets the number of pages in a notebook.
 * 
 * Return value: the number of pages in the notebook.
 *
 * Since: 2.2
 **/
gint
mod_notebook_get_n_pages (ModNotebook *notebook)
{
  g_return_val_if_fail (MOD_IS_NOTEBOOK (notebook), 0);

  return g_list_length (notebook->children);
}

/**
 * mod_notebook_page_num:
 * @notebook: a #ModNotebook
 * @child: a #GtkWidget
 * 
 * Finds the index of the page which contains the given child
 * widget.
 * 
 * Return value: the index of the page containing @child, or
 *   -1 if @child is not in the notebook.
 **/
gint
mod_notebook_page_num (ModNotebook      *notebook,
		       GtkWidget        *child)
{
  GList *children;
  gint num;

  g_return_val_if_fail (MOD_IS_NOTEBOOK (notebook), -1);

  num = 0;
  children = notebook->children;
  while (children)
    {
      ModNotebookPage *page =  children->data;
      
      if (page->child == child)
	return num;

      children = children->next;
      num++;
    }

  return -1;
}

/**
 * mod_notebook_set_current_page:
 * @notebook: a #ModNotebook
 * @page_num: index of the page to switch to, starting from 0.
 *            If negative, the last page will be used. If greater
 *            than the number of pages in the notebook, nothing
 *            will be done.
 *                
 * Switches to the page number @page_num. 
 *
 * Note that due to historical reasons, ModNotebook refuses
 * to switch to a page unless the child widget is visible. 
 * Therefore, it is recommended to show child widgets before
 * adding them to a notebook. 
 */
void
mod_notebook_set_current_page (ModNotebook *notebook,
			       gint         page_num)
{
  GList *list;

  g_return_if_fail (MOD_IS_NOTEBOOK (notebook));

  if (page_num >= 0)
    list = g_list_nth (notebook->children, page_num);
  else
    list = g_list_last (notebook->children);

  page_num = g_list_index (notebook->children, list);

  if (list)
    mod_notebook_switch_page (notebook, MOD_NOTEBOOK_PAGE (list), page_num);
}

/**
 * mod_notebook_next_page:
 * @notebook: a #ModNotebook
 * 
 * Switches to the next page. Nothing happens if the current page is
 * the last page.
 **/
void
mod_notebook_next_page (ModNotebook *notebook)
{
  GList *list;

  g_return_if_fail (MOD_IS_NOTEBOOK (notebook));

  list = g_list_find (notebook->children, notebook->cur_page);
  if (!list)
    return;

  list = mod_notebook_search_page (notebook, list, STEP_NEXT, TRUE);
  if (!list)
    return;

  mod_notebook_switch_page (notebook, MOD_NOTEBOOK_PAGE (list), -1);
}

/**
 * mod_notebook_prev_page:
 * @notebook: a #ModNotebook
 * 
 * Switches to the previous page. Nothing happens if the current page
 * is the first page.
 **/
void
mod_notebook_prev_page (ModNotebook *notebook)
{
  GList *list;

  g_return_if_fail (MOD_IS_NOTEBOOK (notebook));

  list = g_list_find (notebook->children, notebook->cur_page);
  if (!list)
    return;

  list = mod_notebook_search_page (notebook, list, STEP_PREV, TRUE);
  if (!list)
    return;

  mod_notebook_switch_page (notebook, MOD_NOTEBOOK_PAGE (list), -1);
}

/* Public ModNotebook/Tab Style Functions
 *
 * mod_notebook_set_show_border
 * mod_notebook_set_show_tabs
 * mod_notebook_set_tab_pos
 * mod_notebook_set_homogeneous_tabs
 * mod_notebook_set_tab_border
 * mod_notebook_set_tab_hborder
 * mod_notebook_set_tab_vborder
 * mod_notebook_set_scrollable
 */
/**
 * mod_notebook_set_show_border:
 * @notebook: a #ModNotebook
 * @show_border: %TRUE if a bevel should be drawn around the notebook.
 * 
 * Sets whether a bevel will be drawn around the notebook pages.
 * This only has a visual effect when the tabs are not shown.
 * See mod_notebook_set_show_tabs().
 **/
void
mod_notebook_set_show_border (ModNotebook *notebook,
			      gboolean     show_border)
{
  g_return_if_fail (MOD_IS_NOTEBOOK (notebook));

  if (notebook->show_border != show_border)
    {
      notebook->show_border = show_border;

#if GTK_CHECK_VERSION(2,18,0)
      if (gtk_widget_get_visible (GTK_WIDGET (notebook)))
	gtk_widget_queue_resize (GTK_WIDGET (notebook));
#else
      if (GTK_WIDGET_VISIBLE (notebook))
	gtk_widget_queue_resize (GTK_WIDGET (notebook));
#endif
      
      g_object_notify (G_OBJECT (notebook), "show-border");
    }
}

/**
 * mod_notebook_get_show_border:
 * @notebook: a #ModNotebook
 *
 * Returns whether a bevel will be drawn around the notebook pages. See
 * mod_notebook_set_show_border().
 *
 * Return value: %TRUE if the bevel is drawn
 **/
gboolean
mod_notebook_get_show_border (ModNotebook *notebook)
{
  g_return_val_if_fail (MOD_IS_NOTEBOOK (notebook), FALSE);

  return notebook->show_border;
}

/**
 * mod_notebook_set_show_tabs:
 * @notebook: a #ModNotebook
 * @show_tabs: %TRUE if the tabs should be shown.
 * 
 * Sets whether to show the tabs for the notebook or not.
 **/
void
mod_notebook_set_show_tabs (ModNotebook *notebook,
			    gboolean     show_tabs)
{
  ModNotebookPage *page;
  GList *children;

  g_return_if_fail (MOD_IS_NOTEBOOK (notebook));

  show_tabs = show_tabs != FALSE;

  if (notebook->show_tabs == show_tabs)
    return;

  notebook->show_tabs = show_tabs;
  children = notebook->children;

  if (!show_tabs)
    {
      GTK_WIDGET_UNSET_FLAGS (notebook, GTK_CAN_FOCUS);
      
      while (children)
	{
	  page = children->data;
	  children = children->next;
	  if (page->default_tab)
	    {
	      gtk_widget_destroy (page->tab_label);
	      page->tab_label = NULL;
	    }
	  else
	    gtk_widget_hide (page->tab_label);
	}
    }
  else
    {
      GTK_WIDGET_SET_FLAGS (notebook, GTK_CAN_FOCUS);
      mod_notebook_update_labels (notebook);
    }
  gtk_widget_queue_resize (GTK_WIDGET (notebook));

  g_object_notify (G_OBJECT (notebook), "show-tabs");
}

/**
 * mod_notebook_get_show_tabs:
 * @notebook: a #ModNotebook
 *
 * Returns whether the tabs of the notebook are shown. See
 * mod_notebook_set_show_tabs().
 *
 * Return value: %TRUE if the tabs are shown
 **/
gboolean
mod_notebook_get_show_tabs (ModNotebook *notebook)
{
  g_return_val_if_fail (MOD_IS_NOTEBOOK (notebook), FALSE);

  return notebook->show_tabs;
}

/**
 * mod_notebook_set_tab_pos:
 * @notebook: a #ModNotebook.
 * @pos: the edge to draw the tabs at.
 * 
 * Sets the edge at which the tabs for switching pages in the
 * notebook are drawn.
 **/
void
mod_notebook_set_tab_pos (ModNotebook     *notebook,
			  GtkPositionType  pos)
{
  g_return_if_fail (MOD_IS_NOTEBOOK (notebook));

  if (notebook->tab_pos != pos)
    {
      notebook->tab_pos = pos;
#if GTK_CHECK_VERSION(2,18,0)
      if (gtk_widget_get_visible (GTK_WIDGET (notebook)))
	gtk_widget_queue_resize (GTK_WIDGET (notebook));
#else
      if (GTK_WIDGET_VISIBLE (notebook))
	gtk_widget_queue_resize (GTK_WIDGET (notebook));
#endif
    }

  g_object_notify (G_OBJECT (notebook), "tab-pos");
}

/**
 * mod_notebook_get_tab_pos:
 * @notebook: a #ModNotebook
 *
 * Gets the edge at which the tabs for switching pages in the
 * notebook are drawn.
 *
 * Return value: the edge at which the tabs are drawn
 **/
GtkPositionType
mod_notebook_get_tab_pos (ModNotebook *notebook)
{
  g_return_val_if_fail (MOD_IS_NOTEBOOK (notebook), GTK_POS_TOP);

  return notebook->tab_pos;
}

/**
 * mod_notebook_set_homogeneous_tabs:
 * @notebook: a #ModNotebook
 * @homogeneous: %TRUE if all tabs should be the same size.
 * 
 * Sets whether the tabs must have all the same size or not.
 **/
void
mod_notebook_set_homogeneous_tabs (ModNotebook *notebook,
				   gboolean     homogeneous)
{
  g_return_if_fail (MOD_IS_NOTEBOOK (notebook));

  mod_notebook_set_homogeneous_tabs_internal (notebook, homogeneous);
}

/**
 * mod_notebook_set_tab_border:
 * @notebook: a #ModNotebook
 * @border_width: width of the border around the tab labels.
 * 
 * Sets the width the border around the tab labels
 * in a notebook. This is equivalent to calling
 * mod_notebook_set_tab_hborder (@notebook, @border_width) followed
 * by mod_notebook_set_tab_vborder (@notebook, @border_width).
 **/
void
mod_notebook_set_tab_border (ModNotebook *notebook,
			     guint        border_width)
{
  g_return_if_fail (MOD_IS_NOTEBOOK (notebook));

  mod_notebook_set_tab_border_internal (notebook, border_width);
}

/**
 * mod_notebook_set_tab_hborder:
 * @notebook: a #ModNotebook
 * @tab_hborder: width of the horizontal border of tab labels.
 * 
 * Sets the width of the horizontal border of tab labels.
 **/
void
mod_notebook_set_tab_hborder (ModNotebook *notebook,
			      guint        tab_hborder)
{
  g_return_if_fail (MOD_IS_NOTEBOOK (notebook));

  mod_notebook_set_tab_hborder_internal (notebook, tab_hborder);
}

/**
 * mod_notebook_set_tab_vborder:
 * @notebook: a #ModNotebook
 * @tab_vborder: width of the vertical border of tab labels.
 * 
 * Sets the width of the vertical border of tab labels.
 **/
void
mod_notebook_set_tab_vborder (ModNotebook *notebook,
			      guint        tab_vborder)
{
  g_return_if_fail (MOD_IS_NOTEBOOK (notebook));

  mod_notebook_set_tab_vborder_internal (notebook, tab_vborder);
}

/**
 * mod_notebook_set_scrollable:
 * @notebook: a #ModNotebook
 * @scrollable: %TRUE if scroll arrows should be added
 * 
 * Sets whether the tab label area will have arrows for scrolling if
 * there are too many tabs to fit in the area.
 **/
void
mod_notebook_set_scrollable (ModNotebook *notebook,
			     gboolean     scrollable)
{
  g_return_if_fail (MOD_IS_NOTEBOOK (notebook));

  scrollable = (scrollable != FALSE);

  if (scrollable != notebook->scrollable)
    {
      notebook->scrollable = scrollable;

#if GTK_CHECK_VERSION(2,18,0)
      if (gtk_widget_get_visible (GTK_WIDGET (notebook)))
	gtk_widget_queue_resize (GTK_WIDGET (notebook));
#else
      if (GTK_WIDGET_VISIBLE (notebook))
	gtk_widget_queue_resize (GTK_WIDGET (notebook));
#endif

      g_object_notify (G_OBJECT (notebook), "scrollable");
    }
}

/**
 * mod_notebook_get_scrollable:
 * @notebook: a #ModNotebook
 *
 * Returns whether the tab label area has arrows for scrolling. See
 * mod_notebook_set_scrollable().
 *
 * Return value: %TRUE if arrows for scrolling are present
 **/
gboolean
mod_notebook_get_scrollable (ModNotebook *notebook)
{
  g_return_val_if_fail (MOD_IS_NOTEBOOK (notebook), FALSE);

  return notebook->scrollable;
}

/* Public ModNotebook Popup Menu Methods:
 *
 * mod_notebook_popup_enable
 * mod_notebook_popup_disable
 */


/**
 * mod_notebook_popup_enable:
 * @notebook: a #ModNotebook
 * 
 * Enables the popup menu: if the user clicks with the right mouse button on
 * the bookmarks, a menu with all the pages will be popped up.
 **/
void
mod_notebook_popup_enable (ModNotebook *notebook)
{
  GList *list;

  g_return_if_fail (MOD_IS_NOTEBOOK (notebook));

  if (notebook->menu)
    return;

  notebook->menu = gtk_menu_new ();
  for (list = mod_notebook_search_page (notebook, NULL, STEP_NEXT, FALSE);
       list;
       list = mod_notebook_search_page (notebook, list, STEP_NEXT, FALSE))
    mod_notebook_menu_item_create (notebook, list);

  mod_notebook_update_labels (notebook);
  gtk_menu_attach_to_widget (GTK_MENU (notebook->menu),
			     GTK_WIDGET (notebook),
			     mod_notebook_menu_detacher);

  g_object_notify (G_OBJECT (notebook), "enable-popup");
}

/**
 * mod_notebook_popup_disable:
 * @notebook: a #ModNotebook
 * 
 * Disables the popup menu.
 **/
void       
mod_notebook_popup_disable  (ModNotebook *notebook)
{
  g_return_if_fail (MOD_IS_NOTEBOOK (notebook));

  if (!notebook->menu)
    return;

  gtk_container_foreach (GTK_CONTAINER (notebook->menu),
			 (GtkCallback) mod_notebook_menu_label_unparent, NULL);
  gtk_widget_destroy (notebook->menu);

  g_object_notify (G_OBJECT (notebook), "enable-popup");
}

/* Public ModNotebook Page Properties Functions:
 *
 * mod_notebook_get_tab_label
 * mod_notebook_set_tab_label
 * mod_notebook_set_tab_label_text
 * mod_notebook_get_menu_label
 * mod_notebook_set_menu_label
 * mod_notebook_set_menu_label_text
 * mod_notebook_set_tab_label_packing
 * mod_notebook_query_tab_label_packing
 * mod_notebook_get_tab_reorderable
 * mod_notebook_set_tab_reorderable
 * mod_notebook_get_tab_detachable
 * mod_notebook_set_tab_detachable
 */

/**
 * mod_notebook_get_tab_label:
 * @notebook: a #ModNotebook
 * @child: the page
 * 
 * Returns the tab label widget for the page @child. %NULL is returned
 * if @child is not in @notebook or if no tab label has specifically
 * been set for @child.
 * 
 * Return value: the tab label
 **/
GtkWidget *
mod_notebook_get_tab_label (ModNotebook *notebook,
			    GtkWidget   *child)
{
  GList *list;

  g_return_val_if_fail (MOD_IS_NOTEBOOK (notebook), NULL);
  g_return_val_if_fail (GTK_IS_WIDGET (child), NULL);

  list = CHECK_FIND_CHILD (notebook, child);
  if (!list)  
    return NULL;

  if (MOD_NOTEBOOK_PAGE (list)->default_tab)
    return NULL;

  return MOD_NOTEBOOK_PAGE (list)->tab_label;
}  

/**
 * mod_notebook_set_tab_label:
 * @notebook: a #ModNotebook
 * @child: the page
 * @tab_label: the tab label widget to use, or %NULL for default tab
 *             label.
 * 
 * Changes the tab label for @child. If %NULL is specified
 * for @tab_label, then the page will have the label 'page N'.
 **/
void
mod_notebook_set_tab_label (ModNotebook *notebook,
			    GtkWidget   *child,
			    GtkWidget   *tab_label)
{
  ModNotebookPage *page;
  GList *list;

  g_return_if_fail (MOD_IS_NOTEBOOK (notebook));
  g_return_if_fail (GTK_IS_WIDGET (child));

  list = CHECK_FIND_CHILD (notebook, child);
  if (!list)  
    return;

  /* a NULL pointer indicates a default_tab setting, otherwise
   * we need to set the associated label
   */
  page = list->data;
  
  if (page->tab_label == tab_label)
    return;
  

  mod_notebook_remove_tab_label (notebook, page);
  
  if (tab_label)
    {
      page->default_tab = FALSE;
      page->tab_label = tab_label;
      gtk_widget_set_parent (page->tab_label, GTK_WIDGET (notebook));
    }
  else
    {
      page->default_tab = TRUE;
      page->tab_label = NULL;

      if (notebook->show_tabs)
	{
	  gchar string[32];

	  g_snprintf (string, sizeof(string), _("Page %u"), 
		      mod_notebook_real_page_position (notebook, list));
	  page->tab_label = gtk_label_new (string);
	  gtk_widget_set_parent (page->tab_label, GTK_WIDGET (notebook));
	}
    }

  if (page->tab_label)
    page->mnemonic_activate_signal =
      g_signal_connect (page->tab_label,
			"mnemonic_activate",
			G_CALLBACK (mod_notebook_mnemonic_activate_switch_page),
			notebook);

#if GTK_CHECK_VERSION(2,18,0)
  if (notebook->show_tabs && gtk_widget_get_visible (child))
#else
  if (notebook->show_tabs && GTK_WIDGET_VISIBLE (child))
#endif
    {
      gtk_widget_show (page->tab_label);
      gtk_widget_queue_resize (GTK_WIDGET (notebook));
    }

  mod_notebook_update_tab_states (notebook);
  gtk_widget_child_notify (child, "tab-label");
}

/**
 * mod_notebook_set_tab_label_text:
 * @notebook: a #ModNotebook
 * @child: the page
 * @tab_text: the label text
 * 
 * Creates a new label and sets it as the tab label for the page
 * containing @child.
 **/
void
mod_notebook_set_tab_label_text (ModNotebook *notebook,
				 GtkWidget   *child,
				 const gchar *tab_text)
{
  GtkWidget *tab_label = NULL;

  g_return_if_fail (MOD_IS_NOTEBOOK (notebook));

  if (tab_text)
    tab_label = gtk_label_new (tab_text);
  mod_notebook_set_tab_label (notebook, child, tab_label);
  gtk_widget_child_notify (child, "tab-label");
}

/**
 * mod_notebook_get_tab_label_text:
 * @notebook: a #ModNotebook
 * @child: a widget contained in a page of @notebook
 *
 * Retrieves the text of the tab label for the page containing
 *    @child.
 *
 * Returns value: the text of the tab label, or %NULL if the
 *                tab label widget is not a #GtkLabel. The
 *                string is owned by the widget and must not
 *                be freed.
 **/
G_CONST_RETURN gchar *
mod_notebook_get_tab_label_text (ModNotebook *notebook,
				 GtkWidget   *child)
{
  GtkWidget *tab_label;

  g_return_val_if_fail (MOD_IS_NOTEBOOK (notebook), NULL);
  g_return_val_if_fail (GTK_IS_WIDGET (child), NULL);

  tab_label = mod_notebook_get_tab_label (notebook, child);

  if (tab_label && GTK_IS_LABEL (tab_label))
    return gtk_label_get_text (GTK_LABEL (tab_label));
  else
    return NULL;
}

/**
 * mod_notebook_get_menu_label:
 * @notebook: a #ModNotebook
 * @child: a widget contained in a page of @notebook
 * 
 * Retrieves the menu label widget of the page containing @child.
 * 
 * Return value: the menu label, or %NULL if the
 *               notebook page does not have a menu label other
 *               than the default (the tab label).
 **/
GtkWidget*
mod_notebook_get_menu_label (ModNotebook *notebook,
			     GtkWidget   *child)
{
  GList *list;

  g_return_val_if_fail (MOD_IS_NOTEBOOK (notebook), NULL);
  g_return_val_if_fail (GTK_IS_WIDGET (child), NULL);

  list = CHECK_FIND_CHILD (notebook, child);
  if (!list)  
    return NULL;

  if (MOD_NOTEBOOK_PAGE (list)->default_menu)
    return NULL;

  return MOD_NOTEBOOK_PAGE (list)->menu_label;
}  

/**
 * mod_notebook_set_menu_label:
 * @notebook: a #ModNotebook
 * @child: the child widget
 * @menu_label: the menu label, or NULL for default
 * 
 * Changes the menu label for the page containing @child. 
 **/
void
mod_notebook_set_menu_label (ModNotebook *notebook,
			     GtkWidget   *child,
			     GtkWidget   *menu_label)
{
  ModNotebookPage *page;
  GList *list;

  g_return_if_fail (MOD_IS_NOTEBOOK (notebook));
  g_return_if_fail (GTK_IS_WIDGET (child));

  list = CHECK_FIND_CHILD (notebook, child);
  if (!list)  
    return;

  page = list->data;
  if (page->menu_label)
    {
      if (notebook->menu)
	gtk_container_remove (GTK_CONTAINER (notebook->menu), 
			      page->menu_label->parent);

      if (!page->default_menu)
	g_object_unref (page->menu_label);
    }

  if (menu_label)
    {
      page->menu_label = menu_label;
#if GTK_CHECK_VERSION(2,10,0)
      g_object_ref_sink (page->menu_label);
#else
      g_object_ref (page->menu_label);
      gtk_object_sink (GTK_OBJECT (page->menu_label));
#endif
      page->default_menu = FALSE;
    }
  else
    page->default_menu = TRUE;

  if (notebook->menu)
    mod_notebook_menu_item_create (notebook, list);
  gtk_widget_child_notify (child, "menu-label");
}

/**
 * mod_notebook_set_menu_label_text:
 * @notebook: a #ModNotebook
 * @child: the child widget
 * @menu_text: the label text
 * 
 * Creates a new label and sets it as the menu label of @child.
 **/
void
mod_notebook_set_menu_label_text (ModNotebook *notebook,
				  GtkWidget   *child,
				  const gchar *menu_text)
{
  GtkWidget *menu_label = NULL;

  g_return_if_fail (MOD_IS_NOTEBOOK (notebook));

  if (menu_text)
    menu_label = gtk_label_new (menu_text);
  mod_notebook_set_menu_label (notebook, child, menu_label);
  gtk_widget_child_notify (child, "menu-label");
}

/**
 * mod_notebook_get_menu_label_text:
 * @notebook: a #ModNotebook
 * @child: the child widget of a page of the notebook.
 *
 * Retrieves the text of the menu label for the page containing
 *    @child.
 *
 * Returns value: the text of the tab label, or %NULL if the
 *                widget does not have a menu label other than
 *                the default menu label, or the menu label widget
 *                is not a #GtkLabel. The string is owned by
 *                the widget and must not be freed.
 **/
G_CONST_RETURN gchar *
mod_notebook_get_menu_label_text (ModNotebook *notebook,
				  GtkWidget *child)
{
  GtkWidget *menu_label;

  g_return_val_if_fail (MOD_IS_NOTEBOOK (notebook), NULL);
  g_return_val_if_fail (GTK_IS_WIDGET (child), NULL);
 
  menu_label = mod_notebook_get_menu_label (notebook, child);

  if (menu_label && GTK_IS_LABEL (menu_label))
    return gtk_label_get_text (GTK_LABEL (menu_label));
  else
    return NULL;
}
  
/* Helper function called when pages are reordered
 */
static void
mod_notebook_child_reordered (ModNotebook     *notebook,
			      ModNotebookPage *page)
{
  if (notebook->menu)
    {
      GtkWidget *menu_item;
      
      menu_item = page->menu_label->parent;
      gtk_container_remove (GTK_CONTAINER (menu_item), page->menu_label);
      gtk_container_remove (GTK_CONTAINER (notebook->menu), menu_item);
      mod_notebook_menu_item_create (notebook, g_list_find (notebook->children, page));
    }

  mod_notebook_update_tab_states (notebook);
  mod_notebook_update_labels (notebook);
}

/**
 * mod_notebook_set_tab_label_packing:
 * @notebook: a #ModNotebook
 * @child: the child widget
 * @expand: whether to expand the bookmark or not
 * @fill: whether the bookmark should fill the allocated area or not
 * @pack_type: the position of the bookmark
 * 
 * Sets the packing parameters for the tab label of the page
 * containing @child. See gtk_box_pack_start() for the exact meaning
 * of the parameters.
 **/
void
mod_notebook_set_tab_label_packing (ModNotebook *notebook,
				    GtkWidget   *child,
				    gboolean     expand,
				    gboolean     fill,
				    GtkPackType  pack_type)
{
  ModNotebookPage *page;
  GList *list;

  g_return_if_fail (MOD_IS_NOTEBOOK (notebook));
  g_return_if_fail (GTK_IS_WIDGET (child));

  list = CHECK_FIND_CHILD (notebook, child);
  if (!list)  
    return;

  page = list->data;
  expand = expand != FALSE;
  fill = fill != FALSE;
  if (page->pack == pack_type && page->expand == expand && page->fill == fill)
    return;

  gtk_widget_freeze_child_notify (child);
  page->expand = expand;
  gtk_widget_child_notify (child, "tab-expand");
  page->fill = fill;
  gtk_widget_child_notify (child, "tab-fill");
  if (page->pack != pack_type)
    {
      page->pack = pack_type;
      mod_notebook_child_reordered (notebook, page);
    }
  gtk_widget_child_notify (child, "tab-pack");
  gtk_widget_child_notify (child, "position");
  if (notebook->show_tabs)
    mod_notebook_pages_allocate (notebook);
  gtk_widget_thaw_child_notify (child);
}  

/**
 * mod_notebook_query_tab_label_packing:
 * @notebook: a #ModNotebook
 * @child: the page
 * @expand: location to store the expand value (or NULL)
 * @fill: location to store the fill value (or NULL)
 * @pack_type: location to store the pack_type (or NULL)
 * 
 * Query the packing attributes for the tab label of the page
 * containing @child.
 **/
void
mod_notebook_query_tab_label_packing (ModNotebook *notebook,
				      GtkWidget   *child,
				      gboolean    *expand,
				      gboolean    *fill,
				      GtkPackType *pack_type)
{
  GList *list;

  g_return_if_fail (MOD_IS_NOTEBOOK (notebook));
  g_return_if_fail (GTK_IS_WIDGET (child));

  list = CHECK_FIND_CHILD (notebook, child);
  if (!list)
    return;

  if (expand)
    *expand = MOD_NOTEBOOK_PAGE (list)->expand;
  if (fill)
    *fill = MOD_NOTEBOOK_PAGE (list)->fill;
  if (pack_type)
    *pack_type = MOD_NOTEBOOK_PAGE (list)->pack;
}

/**
 * mod_notebook_reorder_child:
 * @notebook: a #ModNotebook
 * @child: the child to move
 * @position: the new position, or -1 to move to the end
 * 
 * Reorders the page containing @child, so that it appears in position
 * @position. If @position is greater than or equal to the number of
 * children in the list or negative, @child will be moved to the end
 * of the list.
 **/
void
mod_notebook_reorder_child (ModNotebook *notebook,
			    GtkWidget   *child,
			    gint         position)
{
  GList *list, *new_list;
  ModNotebookPage *page;
  gint old_pos;
  gint max_pos;

  g_return_if_fail (MOD_IS_NOTEBOOK (notebook));
  g_return_if_fail (GTK_IS_WIDGET (child));

  list = CHECK_FIND_CHILD (notebook, child);
  if (!list)
    return;

  max_pos = g_list_length (notebook->children) - 1;
  if (position < 0 || position > max_pos)
    position = max_pos;

  old_pos = g_list_position (notebook->children, list);

  if (old_pos == position)
    return;

  page = list->data;
  notebook->children = g_list_delete_link (notebook->children, list);

  notebook->children = g_list_insert (notebook->children, page, position);
  new_list = g_list_nth (notebook->children, position);

  /* Fix up GList references in ModNotebook structure */
  if (notebook->first_tab == list)
    notebook->first_tab = new_list;
  if (notebook->focus_tab == list)
    notebook->focus_tab = new_list;

  gtk_widget_freeze_child_notify (child);

  /* Move around the menu items if necessary */
  mod_notebook_child_reordered (notebook, page);
  gtk_widget_child_notify (child, "tab-pack");
  gtk_widget_child_notify (child, "position");

  if (notebook->show_tabs)
    mod_notebook_pages_allocate (notebook);

  gtk_widget_thaw_child_notify (child);

  g_signal_emit (notebook,
		 notebook_signals[PAGE_REORDERED],
		 0,
		 child,
		 position);
}

/**
 * mod_notebook_set_window_creation_hook:
 * @func: the #ModNotebookWindowCreationFunc, or %NULL
 * @data: user data for @func
 * @destroy: Destroy notifier for @data, or %NULL
 *
 * Installs a global function used to create a window
 * when a detached tab is dropped in an empty area.
 * 
 * Since: 2.10
 **/
void
mod_notebook_set_window_creation_hook (ModNotebookWindowCreationFunc  func,
				       gpointer                       data,
                                       GDestroyNotify                 destroy)
{
  if (window_creation_hook_destroy)
    window_creation_hook_destroy (window_creation_hook_data);

  window_creation_hook = func;
  window_creation_hook_data = data;
  window_creation_hook_destroy = destroy;
}

/**
 * mod_notebook_set_group_id:
 * @notebook: a #ModNotebook
 * @group_id: a group identificator, or -1 to unset it
 *
 * Sets an group identificator for @notebook, notebooks sharing
 * the same group identificator will be able to exchange tabs
 * via drag and drop. A notebook with group identificator -1 will
 * not be able to exchange tabs with any other notebook.
 * 
 * Since: 2.10
 **/
void
mod_notebook_set_group_id (ModNotebook *notebook,
			   gint         group_id)
{
  ModNotebookPrivate *priv;

  g_return_if_fail (MOD_IS_NOTEBOOK (notebook));

  priv = MOD_NOTEBOOK_GET_PRIVATE (notebook);

  if (priv->group_id != group_id)
    {
      priv->group_id = group_id;
      g_object_notify (G_OBJECT (notebook), "group-id");
    }
}

/**
 * mod_notebook_get_group_id:
 * @notebook: a #ModNotebook
 * 
 * Gets the current group identificator for @notebook.
 * 
 * Return Value: the group identificator, or -1 if none is set.
 *
 * Since: 2.10
 **/
gint
mod_notebook_get_group_id (ModNotebook *notebook)
{
  ModNotebookPrivate *priv;

  g_return_val_if_fail (MOD_IS_NOTEBOOK (notebook), -1);

  priv = MOD_NOTEBOOK_GET_PRIVATE (notebook);
  return priv->group_id;
}

/**
 * mod_notebook_get_tab_reorderable:
 * @notebook: a #ModNotebook
 * @child: a child #GtkWidget
 * 
 * Gets whether the tab can be reordered via drag and drop or not.
 * 
 * Return Value: %TRUE if the tab is reorderable.
 * 
 * Since: 2.10
 **/
gboolean
mod_notebook_get_tab_reorderable (ModNotebook *notebook,
				  GtkWidget   *child)
{
  GList *list;

  g_return_val_if_fail (MOD_IS_NOTEBOOK (notebook), FALSE);
  g_return_val_if_fail (GTK_IS_WIDGET (child), FALSE);

  list = CHECK_FIND_CHILD (notebook, child);
  if (!list)  
    return FALSE;

  return MOD_NOTEBOOK_PAGE (list)->reorderable;
}

/**
 * mod_notebook_set_tab_reorderable:
 * @notebook: a #ModNotebook
 * @child: a child #GtkWidget
 * @reorderable: whether the tab is reorderable or not.
 *
 * Sets whether the notebook tab can be reordered
 * via drag and drop or not.
 * 
 * Since: 2.10
 **/
void
mod_notebook_set_tab_reorderable (ModNotebook *notebook,
				  GtkWidget   *child,
				  gboolean     reorderable)
{
  GList *list;

  g_return_if_fail (MOD_IS_NOTEBOOK (notebook));
  g_return_if_fail (GTK_IS_WIDGET (child));

  list = CHECK_FIND_CHILD (notebook, child);
  if (!list)  
    return;

  if (MOD_NOTEBOOK_PAGE (list)->reorderable != reorderable)
    {
      MOD_NOTEBOOK_PAGE (list)->reorderable = (reorderable == TRUE);
      gtk_widget_child_notify (child, "reorderable");
    }
}

/**
 * mod_notebook_get_tab_detachable:
 * @notebook: a #ModNotebook
 * @child: a child #GtkWidget
 * 
 * Returns whether the tab contents can be detached from @notebook.
 * 
 * Return Value: TRUE if the tab is detachable.
 *
 * Since: 2.10
 **/
gboolean
mod_notebook_get_tab_detachable (ModNotebook *notebook,
				 GtkWidget   *child)
{
  GList *list;

  g_return_val_if_fail (MOD_IS_NOTEBOOK (notebook), FALSE);
  g_return_val_if_fail (GTK_IS_WIDGET (child), FALSE);

  list = CHECK_FIND_CHILD (notebook, child);
  if (!list)  
    return FALSE;

  return MOD_NOTEBOOK_PAGE (list)->detachable;
}

/**
 * mod_notebook_set_tab_detachable:
 * @notebook: a #ModNotebook
 * @child: a child #GtkWidget
 * @detachable: whether the tab is detachable or not
 *
 * Sets whether the tab can be detached from @notebook to another
 * notebook or widget.
 *
 * Note that 2 notebooks must share a common group identificator
 * (see mod_notebook_set_group_id ()) to allow automatic tabs
 * interchange between them.
 *
 * If you want a widget to interact with a notebook through DnD
 * (i.e.: accept dragged tabs from it) it must be set as a drop
 * destination and accept the target "MOD_NOTEBOOK_TAB". The notebook
 * will fill the selection with a GtkWidget** pointing to the child
 * widget that corresponds to the dropped tab.
 *
 * <informalexample><programlisting>
 *  static void
 *  on_drop_zone_drag_data_received (GtkWidget        *widget,
 *                                   GdkDragContext   *context,
 *                                   gint              x,
 *                                   gint              y,
 *                                   GtkSelectionData *selection_data,
 *                                   guint             info,
 *                                   guint             time,
 *                                   gpointer          user_data)
 *  {
 *    GtkWidget *notebook;
 *    GtkWidget **child;
 *    
 *    notebook = gtk_drag_get_source_widget (context);
 *    child = (void*) selection_data->data;
 *    
 *    process_widget (*child);
 *    gtk_container_remove (GTK_CONTAINER (notebook), *child);
 *  }
 * </programlisting></informalexample>
 *
 * If you want a notebook to accept drags from other widgets,
 * you will have to set your own DnD code to do it.
 *
 * Since: 2.10
 **/
void
mod_notebook_set_tab_detachable (ModNotebook *notebook,
				 GtkWidget  *child,
				 gboolean    detachable)
{
  GList *list;

  g_return_if_fail (MOD_IS_NOTEBOOK (notebook));
  g_return_if_fail (GTK_IS_WIDGET (child));

  list = CHECK_FIND_CHILD (notebook, child);
  if (!list)  
    return;

  if (MOD_NOTEBOOK_PAGE (list)->detachable != detachable)
    {
      MOD_NOTEBOOK_PAGE (list)->detachable = (detachable == TRUE);
      gtk_widget_child_notify (child, "detachable");
    }
}


/*
 * Ochusha's modifications.
 */

/* Public ModNotebook/Tab Style Functions
 *
 * mod_notebook_set_tab_shrinkable
 * mod_notebook_get_tab_shrinkable
 * mod_notebook_set_minimum_tab_label_size
 * mod_notebook_get_minimum_tab_label_size
 */
/**
 * mod_notebook_set_tab_shrinkable:
 * @notebook: a #ModNotebook
 * @shrinkable: %TRUE if bookmark tabs are allowed to be shrinked if the
 *              drawing area for bookmark tabs don't have enough space to
 *              show all tabs.
 * 
 * Sets whether larger tab labels are allowed to be shrinked if there are
 * too many tabs to fit in the area.
 **/
void
mod_notebook_set_tab_shrinkable (ModNotebook     *notebook,
				 gboolean         shrinkable)
{
  guint shrink;
  g_return_if_fail (MOD_IS_NOTEBOOK (notebook));
  
  shrink = (shrinkable != FALSE);

  if (shrink != notebook->tab_shrinkable)
    {
      notebook->tab_shrinkable = shrink;

#if GTK_CHECK_VERSION(2,18,0)
      if (gtk_widget_get_visible (GTK_WIDGET (notebook)))
	gtk_widget_queue_resize (GTK_WIDGET (notebook));
#else
      if (GTK_WIDGET_VISIBLE (notebook))
	gtk_widget_queue_resize (GTK_WIDGET (notebook));
#endif

      g_object_notify (G_OBJECT (notebook), "tab-shrinkable");
    }
}

/**
 * mod_notebook_get_tab_shrinkable:
 * @notebook: a #ModNotebook
 *
 * Returns whether tab labels are allowed to be shrinked if there are too
 * many tabs to fit tab areas. See mod_notebook_set_tab_shrinkable().
 *
 * Return value: %TRUE if bookmark tab labels are allowed to be shrinked.
 **/
gboolean
mod_notebook_get_tab_shrinkable (ModNotebook     *notebook)
{
  g_return_val_if_fail (MOD_IS_NOTEBOOK (notebook), FALSE);

  return notebook->tab_shrinkable;
}

/**
 * mod_notebook_set_minimum_tab_label_size:
 * @notebook: a #ModNotebook
 * @size: the minimum size of tab label for shrinking.
 *
 * Sets the minimum size of tab labels for shrinked tab labels.  ModNotebook
 * guarantees to shrink tab labels no smaller than the specified size.
 **/
void
mod_notebook_set_minimum_tab_label_size (ModNotebook *notebook,
					 guint        size)
{
  g_return_if_fail (MOD_IS_NOTEBOOK (notebook));

  if (notebook->minimum_tab_label_size != size)
    {
      notebook->minimum_tab_label_size = size;

#if GTK_CHECK_VERSION(2,18,0)
      if (gtk_widget_get_visible (GTK_WIDGET (notebook)))
	gtk_widget_queue_resize (GTK_WIDGET (notebook));
#else
      if (GTK_WIDGET_VISIBLE (notebook))
	gtk_widget_queue_resize (GTK_WIDGET (notebook));
#endif

      g_object_notify (G_OBJECT (notebook), "minimum_tab_label_size");
    }
}

/**
 * mod_notebook_get_minimum_tab_label_size:
 * @notebook: a #ModNotebook
 *
 * Returns the minimum size of tab labels that are guaranteed that each
 * tab labels displayed to have size no smaller than it.
 *
 * Return value: the minimum size of tab labels.
 **/
guint
mod_notebook_get_minimum_tab_label_size (ModNotebook *notebook)
{
  g_return_val_if_fail (MOD_IS_NOTEBOOK (notebook), ARROW_SIZE);

  return notebook->minimum_tab_label_size;
}

static void
mod_notebook_get_close_button_rect (ModNotebook  *notebook,
				    GdkRectangle *rectangle)
{
  GdkRectangle event_window_pos;

  if (!notebook->enable_close_button || !notebook->show_tabs)
    return;

  if (mod_notebook_get_event_window_position (notebook, &event_window_pos))
    {
      gint close_button_size;
      gint close_button_spacing;
      gint gap;

      gtk_widget_style_get (GTK_WIDGET (notebook),
                            "close-button-size", &close_button_size,
                            "close-button-spacing", &close_button_spacing,
                            NULL);

      switch (get_effective_tab_pos (notebook))
	{
	case GTK_POS_LEFT:
	case GTK_POS_RIGHT:
	  rectangle->width = close_button_size;
	  rectangle->height = close_button_size;
	  rectangle->x = event_window_pos.x + (event_window_pos.width - close_button_size) / 2;
	  rectangle->y = event_window_pos.y + event_window_pos.height -
			 close_button_size - close_button_spacing;
	  break;
	case GTK_POS_TOP:
	case GTK_POS_BOTTOM:
	  rectangle->width = close_button_size;
	  rectangle->height = close_button_size;
	  gap = (event_window_pos.height - close_button_size) / 2;
	  rectangle->y = event_window_pos.y + gap;
	  if (gtk_widget_get_direction (GTK_WIDGET (notebook)) != GTK_TEXT_DIR_RTL)
	    rectangle->x = event_window_pos.x + event_window_pos.width -
	      close_button_size - close_button_spacing;
	  else
	    rectangle->x = event_window_pos.x + close_button_spacing;
	  break;
	}
    }
}

static gboolean
mod_notebook_in_close_button (ModNotebook *notebook,
			      gint         x,
			      gint         y)
{
  GdkRectangle button_rect;
  gint x0, y0;

  if (!notebook->enable_close_button || !notebook->show_tabs)
    return FALSE;

  mod_notebook_get_close_button_rect (notebook, &button_rect);

  x0 = x - button_rect.x;
  y0 = y - button_rect.y;

  return (y0 >= 0 && y0 < button_rect.height &&
	  x0 >= 0 && x0 < button_rect.width);
}

static void
mod_notebook_draw_close_button (ModNotebook *notebook)
{
  GtkWidget *widget;
  GdkRectangle button_rect;

  g_return_if_fail (MOD_IS_NOTEBOOK (notebook));

  mod_notebook_get_close_button_rect (notebook, &button_rect);

  widget = GTK_WIDGET (notebook);

#if GTK_CHECK_VERSION(2,18,0)
  if (gtk_widget_is_drawable (widget))
#else
  if (GTK_WIDGET_DRAWABLE (widget))
#endif
    {
      GtkStateType state_type;
      GtkShadowType shadow_type;
      GtkStyle *style;
      GtkIconSet *icon_set;
      GdkPixbuf *button;
      int button_width;
      int button_height;
      int off_x;
      int off_y;
      if (notebook->click_close_button)
	{
	  state_type = GTK_STATE_ACTIVE;
	  shadow_type = GTK_SHADOW_IN;
	}
      else if (notebook->in_close_button)
	{
	  state_type = GTK_STATE_PRELIGHT;
	  shadow_type = GTK_SHADOW_OUT;
	}
      else
	{
	  state_type = GTK_STATE_NORMAL;
	  shadow_type = GTK_SHADOW_NONE;
	}

      style = gtk_widget_get_style (widget);

      gtk_paint_box (style, widget->window, state_type, shadow_type,
		     &button_rect, widget, "button",
		     button_rect.x, button_rect.y,
		     button_rect.width, button_rect.height);

      icon_set = gtk_style_lookup_icon_set (style, GTK_STOCK_CLOSE);
      if (icon_set == NULL)
	return;

      button = gtk_icon_set_render_icon (icon_set, style,
					 gtk_widget_get_direction (widget),
					 state_type,
					 GTK_ICON_SIZE_MENU,
					 widget,
					 NULL);

      button_width = gdk_pixbuf_get_width (button);
      button_height = gdk_pixbuf_get_height (button);

      if (button_width > button_rect.width)
	button_width = button_rect.width;
      if (button_height > button_rect.height)
	button_height = button_rect.height;

      off_x = (button_rect.width - button_width) / 2;
      off_y = (button_rect.height - button_height) / 2;

      gdk_draw_pixbuf(widget->window, NULL, button,
		      0, 0,
		      button_rect.x + off_x, button_rect.y + off_y,
		      button_width, button_height,
		      GDK_RGB_DITHER_NONE,
		      0, 0);

      g_object_unref(G_OBJECT(button));
    }
}

static void
mod_notebook_redraw_close_button (ModNotebook *notebook)
{
#if GTK_CHECK_VERSION(2,18,0)
  if (gtk_widget_get_mapped (GTK_WIDGET (notebook)) &&
      notebook->enable_close_button && notebook->first_tab)
#else
  if (GTK_WIDGET_MAPPED (notebook) && notebook->enable_close_button &&
      notebook->first_tab)
#endif
    {
      GdkRectangle button_rect;
      mod_notebook_get_close_button_rect (notebook, &button_rect);
      gdk_window_invalidate_rect (GTK_WIDGET (notebook)->window,
				  &button_rect, FALSE);
    }
}

static void
mod_notebook_page_ref (ModNotebookPage *page)
{
  page->ref_count++;
}

static void
mod_notebook_page_unref (ModNotebookPage *page)
{
  if (--page->ref_count == 0)
    g_free(page);
}

/* Public ModNotebook Tooltips Methods:
 *
 * mod_notebook_tooltips_enable
 * mod_notebook_tooltips_disable
 */

/**
 * mod_notebook_tooltips_enable:
 * @notebook: a #ModNotebook
 * 
 * Enables the popup menu.
 **/
void
mod_notebook_tooltips_enable (ModNotebook *notebook)
{
  g_return_if_fail (MOD_IS_NOTEBOOK(notebook));

  if (notebook->enable_tooltips)
    return;

  notebook->enable_tooltips = TRUE;
  g_object_notify (G_OBJECT (notebook), "enable-tooltips");
}

/**
 * mod_notebook_tooltips_disable:
 * @notebook: a #ModNotebook
 * 
 * Disables the tooltips.
 **/
void
mod_notebook_tooltips_disable (ModNotebook *notebook)
{
  g_return_if_fail (MOD_IS_NOTEBOOK(notebook));

  if (!notebook->enable_tooltips)
    return;

  notebook->enable_tooltips = FALSE;
  g_object_notify (G_OBJECT (notebook), "enable-tooltips");
}


/* Public ModNotebook Close Button Methods:
 *
 * mod_notebook_close_button_enable
 * mod_notebook_close_button_disable
 */

/**
 * mod_notebook_close_button_enable:
 * @notebook: a #ModNotebook
 * 
 * Enables the close button.
 **/
void
mod_notebook_close_button_enable (ModNotebook *notebook)
{
  g_return_if_fail (MOD_IS_NOTEBOOK(notebook));

  if (notebook->enable_close_button)
    return;

  notebook->enable_close_button = TRUE;
  gtk_widget_queue_resize (GTK_WIDGET (notebook));

  g_object_notify (G_OBJECT (notebook), "enable-close-button");
}

/**
 * mod_notebook_close_button_disable:
 * @notebook: a #ModNotebook
 * 
 * Disables the close button.
 **/
void
mod_notebook_close_button_disable (ModNotebook *notebook)
{
  g_return_if_fail (MOD_IS_NOTEBOOK(notebook));

  if (!notebook->enable_close_button)
    return;

  notebook->enable_close_button = FALSE;
  gtk_widget_queue_resize (GTK_WIDGET (notebook));

  g_object_notify (G_OBJECT (notebook), "enable-close-button");
}


static void
mod_notebook_tooltips_set_active_page (ModNotebook *notebook,
				       ModNotebookPage *new_page)
{
  if (notebook->tip_window)
    {
#if GTK_CHECK_VERSION(2,18,0)
      if (gtk_widget_get_visible (notebook->tip_window))
	g_get_current_time (&notebook->last_popdown);
#else
      if (GTK_WIDGET_VISIBLE (notebook->tip_window))
	g_get_current_time (&notebook->last_popdown);
#endif
      gtk_widget_hide (notebook->tip_window);
    }
  if (notebook->tip_timer)
    {
      g_source_remove (notebook->tip_timer);
      notebook->tip_timer = 0;
    }
  if (notebook->tip_hide_timer)
    {
      g_source_remove (notebook->tip_hide_timer);
      notebook->tip_hide_timer = 0;
    }
  if (notebook->active_tips_page != NULL)
    {
      mod_notebook_page_unref(notebook->active_tips_page);
      notebook->active_tips_page = NULL;
    }

  if (new_page)
    {
      GList *list;

      for (list = notebook->children; list; list = list->next)
	{
	  ModNotebookPage *page;

	  page = list->data;

#if GTK_CHECK_VERSION(2,18,0)
	  if (page == new_page && gtk_widget_is_drawable (page->tab_label))
#else
	  if (page == new_page && GTK_WIDGET_DRAWABLE (page->tab_label))
#endif
	    {
	      mod_notebook_page_ref(page);
	      notebook->active_tips_page = page;
	      break;
	    }
	}
    }
  else
    {
      notebook->use_sticky_delay = FALSE;
    }
}

typedef struct _TooltipsTimeoutArgs
{
  ModNotebook *notebook;
  ModNotebookPage *page;
} TooltipsTimeoutArgs;

static void
mod_notebook_tooltips_timeout_done(gpointer data)
{
  TooltipsTimeoutArgs *args = (TooltipsTimeoutArgs *)data;
  mod_notebook_page_unref(args->page);
  g_free(args);
}

static void
mod_notebook_tooltips_start_delay (ModNotebook *notebook,
				   ModNotebookPage *page)
{
  ModNotebookPage *old_tips_page;

  old_tips_page = notebook->active_tips_page;

  if (notebook->enable_tooltips &&
      (!old_tips_page || old_tips_page != page))
    {
      guint delay;
      TooltipsTimeoutArgs *args = g_new(TooltipsTimeoutArgs, 1);
      args->notebook = notebook;
      args->page = page;
      mod_notebook_page_ref(page);

      mod_notebook_tooltips_set_active_page (notebook, page);

      if (notebook->use_sticky_delay &&
	  mod_notebook_tooltips_recently_shown (notebook))
	delay = TOOLTIPS_STICKY_DELAY;
      else
	delay = notebook->delay;
      notebook->tip_timer
	= g_timeout_add_full (G_PRIORITY_DEFAULT,
			      delay,
			      mod_notebook_tooltips_timeout,
			      (gpointer) args,
			      mod_notebook_tooltips_timeout_done);
    }
}

static gboolean
mod_notebook_tooltips_recently_shown (ModNotebook *notebook)
{
  GTimeVal now;
  glong msec;

  g_get_current_time (&now);
  msec = (now.tv_sec - notebook->last_popdown.tv_sec) * 1000 +
          (now.tv_usec - notebook->last_popdown.tv_usec) / 1000;
  return (msec < TOOLTIPS_STICKY_REVERT_DELAY);
}

static gboolean
mod_notebook_tooltips_timeout (gpointer data)
{
  TooltipsTimeoutArgs *args = (TooltipsTimeoutArgs *)data;
  ModNotebook *notebook = (ModNotebook *)args->notebook;

  GDK_THREADS_ENTER ();

#if GTK_CHECK_VERSION(2,18,0)
  if (notebook->active_tips_page == args->page &&
      gtk_widget_is_drawable (notebook->active_tips_page->tab_label))
    mod_notebook_tooltips_draw_tips (notebook);
#else
  if (notebook->active_tips_page == args->page &&
      GTK_WIDGET_DRAWABLE (notebook->active_tips_page->tab_label))
    mod_notebook_tooltips_draw_tips (notebook);
#endif

  if (notebook->tip_hide_timer)
    g_source_remove (notebook->tip_hide_timer);
  notebook->tip_hide_timer = g_timeout_add (TOOLTIPS_HIDE_DELAY,
					    mod_notebook_tooltips_hide_timeout,
					    (gpointer) notebook);

  GDK_THREADS_LEAVE ();

  return FALSE;
}

static gboolean
mod_notebook_tooltips_hide_timeout (gpointer data)
{
  ModNotebook *notebook = MOD_NOTEBOOK(data);
  GDK_THREADS_ENTER ();

  if (notebook->tip_window)
    {
#if GTK_CHECK_VERSION(2,18,0)
      if (gtk_widget_get_visible (notebook->tip_window))
	g_get_current_time (&notebook->last_popdown);
#else
      if (GTK_WIDGET_VISIBLE (notebook->tip_window))
	g_get_current_time (&notebook->last_popdown);
#endif
      gtk_widget_hide (notebook->tip_window);
    }

  GDK_THREADS_LEAVE ();
  return FALSE;
}

static void
mod_notebook_tooltips_draw_tips (ModNotebook *notebook)
{
  GtkRequisition requisition;
  GtkWidget *widget;
  GtkStyle *style;
  gint x, y, w, h, scr_w, scr_h;
  ModNotebookPage *page;
  gboolean keyboard_mode;
  GdkScreen *screen;

  if (!notebook->tip_window)
    mod_notebook_tooltips_force_window (notebook);
#if GTK_CHECK_VERSION(2,18,0)
  else if (gtk_widget_get_visible (notebook->tip_window))
    g_get_current_time (&notebook->last_popdown);
#else
  else if (GTK_WIDGET_VISIBLE (notebook->tip_window))
    g_get_current_time (&notebook->last_popdown);
#endif

  gtk_widget_ensure_style (notebook->tip_window);
  style = notebook->tip_window->style;

  widget = notebook->active_tips_page->tab_label;

  keyboard_mode = get_keyboard_mode (widget);

  mod_notebook_tooltips_update_screen (notebook, FALSE);

  screen = gtk_widget_get_screen (widget);
  scr_w = gdk_screen_get_width (screen);
  scr_h = gdk_screen_get_height (screen);

  page = notebook->active_tips_page;

  if (page->menu_label && GTK_IS_LABEL (page->menu_label))
    gtk_label_set_text (GTK_LABEL (notebook->tip_label),
			gtk_label_get_text (GTK_LABEL (page->menu_label)));
  else if (page->tab_label && GTK_IS_LABEL (page->tab_label))
    gtk_label_set_text (GTK_LABEL (notebook->tip_label),
			gtk_label_get_text (GTK_LABEL (page->tab_label)));
  else
    gtk_label_set_text (GTK_LABEL (notebook->tip_label), "");

  gtk_widget_size_request (notebook->tip_window, &requisition);
  w = requisition.width;
  h = requisition.height;

  gdk_window_get_origin (widget->window, &x, &y);
#if GTK_CHECK_VERSION(2,18,0)
  if (!gtk_widget_get_has_window (widget))
#else
  if (GTK_WIDGET_NO_WINDOW (widget))
#endif
    {
      x += widget->allocation.x;
      y += widget->allocation.y;
    }

  x += widget->allocation.width / 2;

  if (!keyboard_mode)
    gdk_window_get_pointer (gdk_screen_get_root_window (screen),
			    &x, NULL, NULL);

  x -= (w / 2 + 4);

  if ((x + w) > scr_w)
    x -= (x + w) - scr_w;
  else if (x < 0)
    x = 0;

  if ((y + h + widget->allocation.height + 4) > scr_h)
    y = y - h - 4;
  else
    y = y + widget->allocation.height + 4;

  gtk_window_move (GTK_WINDOW (notebook->tip_window), x, y);
  gtk_widget_show (notebook->tip_window);
}

static void
mod_notebook_tooltips_force_window (ModNotebook *notebook)
{
  if (notebook->tip_window)
    return;

  notebook->tip_window = gtk_window_new (GTK_WINDOW_POPUP);
  mod_notebook_tooltips_update_screen (notebook, TRUE);
  gtk_widget_set_app_paintable (notebook->tip_window, TRUE);
  gtk_window_set_resizable (GTK_WINDOW (notebook->tip_window), FALSE);
  gtk_widget_set_name (notebook->tip_window, "gtk-tooltips");
  gtk_container_set_border_width (GTK_CONTAINER(notebook->tip_window), 4);

  g_signal_connect_swapped (notebook->tip_window,
			    "expose_event",
			    G_CALLBACK (mod_notebook_tooltips_paint_window),
			    notebook);
  notebook->tip_label = gtk_label_new (NULL);
  gtk_label_set_line_wrap (GTK_LABEL (notebook->tip_label), TRUE);
  gtk_misc_set_alignment (GTK_MISC (notebook->tip_label), 0.5, 0.5);
  gtk_widget_show (notebook->tip_label);

  gtk_container_add (GTK_CONTAINER (notebook->tip_window),
		     notebook->tip_label);
  g_signal_connect (notebook->tip_window,
		    "destroy",
		    G_CALLBACK (gtk_widget_destroyed),
		    &notebook->tip_window);
}

static gboolean
mod_notebook_tooltips_paint_window (ModNotebook *notebook)
{
  gtk_paint_flat_box (notebook->tip_window->style,
		      notebook->tip_window->window,
		      GTK_STATE_NORMAL, GTK_SHADOW_OUT,
		      NULL, notebook->tip_window, "tooltip",
		      0, 0, -1, -1);

  return FALSE;
}

static void
mod_notebook_tooltips_update_screen (ModNotebook *notebook,
				     gboolean     new_window)
{
  gboolean screen_changed = FALSE;

#if GTK_CHECK_VERSION(2,18,0)
  if (notebook->active_tips_page &&
      gtk_widget_is_drawable (notebook->active_tips_page->tab_label))
#else
  if (notebook->active_tips_page &&
      GTK_WIDGET_DRAWABLE (notebook->active_tips_page->tab_label))
#endif
    {
      GdkScreen *screen = gtk_widget_get_screen (notebook->active_tips_page->tab_label);

      screen_changed = (screen != gtk_widget_get_screen (notebook->tip_window));

      if (screen_changed)
	{
	  if (!new_window)
	    disconnect_tip_window_display_closed (notebook);

	  gtk_window_set_screen (GTK_WINDOW (notebook->tip_window), screen);
	}
    }

  if (screen_changed || new_window)
    g_signal_connect (gtk_widget_get_display (notebook->tip_window), "closed",
		      G_CALLBACK (tip_window_display_closed), notebook);
}

static gboolean
get_keyboard_mode (GtkWidget   *widget)
{
  GtkWidget *toplevel = gtk_widget_get_toplevel (widget);
  if (GTK_IS_WINDOW (toplevel))
    return GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (toplevel), "gtk-tooltips-keyboard-mode"));
  else
    return FALSE;
}

static void
disconnect_tip_window_display_closed (ModNotebook *notebook)
{
  g_signal_handlers_disconnect_by_func (gtk_widget_get_display (notebook->tip_window),
					(gpointer) tip_window_display_closed,
					notebook);
}

static void
tip_window_display_closed (GdkDisplay  *display,
			   gboolean     was_error,
			   ModNotebook *notebook)
{
  mod_notebook_tooltips_unset_tip_window(notebook);
}

static void
mod_notebook_tooltips_unset_tip_window (ModNotebook *notebook)
{
  if (notebook->tip_window)
    {
      disconnect_tip_window_display_closed (notebook);

      gtk_widget_destroy (notebook->tip_window);
      notebook->tip_window = NULL;
    }
}

static ModNotebookPage *
mod_notebook_get_tab_page (ModNotebook *notebook, gint x, gint y)
{
  GList *children;
  ModNotebookPage *page;

  children = notebook->children;
  while (children != NULL)
    {
      page = (ModNotebookPage *)children->data;

#if GTK_CHECK_VERSION(2,18,0)
      if (gtk_widget_get_visible (page->child) &&
	  page->tab_label && gtk_widget_get_mapped (page->tab_label) &&
	  (x >= page->allocation.x) && (y >= page->allocation.y) &&
	  (x <= (page->allocation.x + page->allocation.width)) &&
	  (y <= (page->allocation.y + page->allocation.height)))
	return page;
#else
      if (GTK_WIDGET_VISIBLE (page->child) &&
	  page->tab_label && GTK_WIDGET_MAPPED (page->tab_label) &&
	  (x >= page->allocation.x) && (y >= page->allocation.y) &&
	  (x <= (page->allocation.x + page->allocation.width)) &&
	  (y <= (page->allocation.y + page->allocation.height)))
	return page;
#endif

      children = children->next;
    }

  return NULL;
}

/**
 * mod_notebook_get_mouse_event_page_num:
 * @notebook: a #ModNotebook
 * @event: a #GdkEventButton
 *
 * Calculates the page number associated to bookmark tab on which the
 * specified event is occur.
 * 
 * Returns value: the page number or -1 for events not occur on bookmarks.
 **/
gint
mod_notebook_get_mouse_event_page_num (ModNotebook *notebook,
				       GdkEventAny *event)
{
  GList *children;
  ModNotebookPage *page;
  gint num;
  gint x, y;

  g_return_val_if_fail(MOD_IS_NOTEBOOK(notebook), -1);

  if (notebook->children == NULL)
    return -1;

  if (!get_widget_coordinates(GTK_WIDGET(notebook), (GdkEvent *)event, &x, &y))
    return -1;

  if (mod_notebook_get_arrow(notebook, x, y) != ARROW_NONE)
    return -1;

  num = 0;
  children = notebook->children;
  while (children != NULL)
    {
      page = (ModNotebookPage *)children->data;

#if GTK_CHECK_VERSION(2,18,0)
      if (gtk_widget_get_visible (page->child) &&
	  page->tab_label && gtk_widget_get_mapped (page->tab_label) &&
	  (x >= page->allocation.x) && (y >= page->allocation.y) &&
	  (x <= (page->allocation.x + page->allocation.width)) &&
	  (y <= (page->allocation.y + page->allocation.height)))
	return num;
#else
      if (GTK_WIDGET_VISIBLE (page->child) &&
	  page->tab_label && GTK_WIDGET_MAPPED (page->tab_label) &&
	  (x >= page->allocation.x) && (y >= page->allocation.y) &&
	  (x <= (page->allocation.x + page->allocation.width)) &&
	  (y <= (page->allocation.y + page->allocation.height)))
	return num;
#endif

      children = children->next;
      num++;
    }

  return -1;
}
